import codecs
import re

from MyCapytain.common.constants import Mimetypes
from MyCapytain.resolvers.cts.api import HttpCtsResolver
from MyCapytain.retrievers.cts5 import HttpCtsRetriever as CTS
from requests.exceptions import HTTPError
from cylleneus import settings
from cylleneus.utils import nrange, alnum

# Description
description = "Laboratoire d’Analyse Statistique des Langues Anciennes (LASLA)"

# Language
language = "lat"

# Glob pattern for indexing
glob = "*.BPN"

# Repo
repo = {
    "origin": "https://github.com/cylleneus/lasla.git",
    "raw": "http://raw.github.com/cylleneus/lasla/master/",
    "location": "remote",
}


# Fetch text
def fetch(work, meta, fragment):
    _, urn = work.urn[0]

    # Reference and hlite values
    divs = meta["meta"].split("-")

    ref_start = ", ".join(
        [
            f"{item}: {meta['start'][item]}"
            for item in meta["start"]
            if item in divs
        ]
    )
    ref_end = ", ".join(
        [
            f"{item}: {meta['end'][item]}"
            for item in meta["end"]
            if item in divs
        ]
    )
    reference = (
        "-".join([ref_start, ref_end]) if ref_end != ref_start else ref_start
    )

    _, file = work.filename[0]
    path = work.corpus.text_dir / file

    # Collect text and context
    if path.exists():
        with codecs.open(path, "r", "utf8") as fp:
            lines = fp.readlines()

        start = meta["start"]["pos"]
        end = meta["end"]["pos"]

        match = []
        match.append(
            f"<match>{' '.join([re.sub(r' ?<.+?> ?', '', parse_bpn(line)['form']) for line in lines[start:end+1]])}</match>"
        )

        if "poem" in divs or (
            len(divs) == 2 and divs[-1] in ["line", "verse"]
        ):
            joiner = "\n\n"
        else:
            joiner = " "
        parts = match
        text = f"{joiner}".join(parts)
    else:
        text = ""
    return urn, reference, text


subord = {
    "main clause": ["00"],
    "paratactic subjunctive": ["AA"],
    "paratactic subjunctive in interrogative clause": ["AB"],
    "ablative absolute": ["AD", "AE"],
    "infinitive clause": ["AG"],
    "AC": ["AK"],
    "ATQVE": ["AK"],
    "AN": ["AW"],
    "ANNE": ["BD"],
    "ANTEQVAM": ["BK"],
    "ANTEAQVAM": ["BK"],
    "adverbial or subordinating conjunction,": ["BL"],
    "CEV": ["BL"],
    "CVIVS": ["BM"],
    "CVIVSMODI": ["BM"],
    "CVM": ["BN"],
    "CVICVIMODI": ["BO"],
    "CVMCVMQVE": ["BS"],
    "interrogative": [
        "BX",
        "DG",
        "FN",
        "GA",
        "GF",
        "GK",
        "GS",
        "HG",
        "JG",
        "JJ",
        "JN",
        "JX",
        "KG",
        "KS",
        "LD",
        "LK",
        "PA",
        "PS",
        "RG",
        "RS",
        "WX",
        "YA",
    ],
    "CVR": ["BX"],
    "DONEC": ["CA"],
    "DVM": ["CD"],
    "DVMMODO": ["CG"],
    "ECQVIS": ["CL"],
    "ECQVID": ["CL"],
    "ECQVISNAM": ["CM"],
    "ECQVIDNAM": ["CM"],
    "ETIAMSI": ["CN"],
    "ECQVANDO": ["CR"],
    "ETSI": ["CS"],
    "LICET": ["CX"],
    "MODO": ["DA"],
    "subordinating conjunction": [
        "DD",
        "HX",
        "KX",
        "LE",
        "MK",
        "NN",
        "SX",
        "TB",
        "WK",
        "XK",
        "ZE",
    ],
    "NE": ["DD", "DG"],
    "NECNE": ["DP", "DS"],
    "NEC": ["DR"],
    "NECVBI": ["DT"],
    "NEDVM": ["DX"],
    "NEVE": ["EA"],
    "NEV": ["EA"],
    "NI": ["ED"],
    "NISI": ["EG"],
    "NONNE": ["EK"],
    "NVM": ["EN"],
    "NVMQVIS": ["EP"],
    "NVMQVID": ["EP"],
    "NVMQVANDO": ["EQ"],
    "POSTQVAM": ["ES"],
    "POSTEAQVAM": ["ES"],
    "PRAETERQVAM": ["EY"],
    "PRAEVT": ["EZ"],
    "PRIVSQVAM": ["FA"],
    "PROQVAM": ["FD"],
    "PROVT": ["FG"],
    "relative": [
        "FK",
        "FX",
        "GE",
        "GG",
        "GN",
        "HD",
        "JF",
        "JH",
        "JK",
        "JS",
        "KD",
        "KN",
        "LA",
        "LG",
        "NX",
        "PN",
        "RD",
        "RN",
        "SG",
        "WS",
        "XX",
    ],
    "QVA": ["FK", "FN"],
    "QVACVMQVE": ["FS"],
    "QVALIS": ["FX", "GA"],
    "QVALISCVMQVE": ["GD"],
    "QVALITER": ["GE", "GF"],
    "QVAM": ["GG", "GK"],
    "QVAMDIV": ["GN", "GS"],
    "QVAMDVDVM": ["GX"],
    "QVAMLIBET": ["HA"],
    "QVAMOBREM": ["HD", "HG"],
    "QVAMQVAM": ["HK"],
    "QVAMVIS": ["HN"],
    "QVANAM": ["HP"],
    "interrogative adverb": ["HS", "MA", "SW", "WG", "XG"],
    "QVANDO": ["HS", "HX"],
    "QVANDOCVMQVE": ["JA"],
    "QVANDOQVE": ["JD"],
    "QVANDOQVIDEM": ["JE"],
    "QVANTILLVS": ["JF", "JG"],
    "QVANTO": ["JH", "JJ"],
    "QVANTOPERE": ["JK", "JN"],
    "QVANTVLVS": ["JS", "JX"],
    "QVANTVLVSLIBET": ["JZ"],
    "QVANTVLVSCVMQVE": ["KA"],
    "QVANTVS": ["KD", "KG"],
    "QVANTVM": ["KD", "KG"],
    "QVANTVSCVMQVE": ["KK"],
    "QVAPROPTER": ["KL"],
    "QVARE": ["KN", "KS"],
    "relative adverb": ["KW", "LX", "TA", "WD", "XD"],
    "QVASI": ["KW", "KX"],
    "QVATENVS": ["LA", "LD", "LE"],
    "QVEMADMODVM": ["LG", "LK"],
    "pronom relative": ["LN"],
    "QVI": ["LN", "LS", "LX", "MA"],
    "pronom interrogative": ["LS"],
    "QVIA": ["MD"],
    "QVICVMQVE": ["MG"],
    "QVIN": ["MK"],
    "quin": ["ML"],
    "NON": ["ML"],
    "QVIPPE": ["MN"],
    "QVIS": ["MS"],
    "QVID": ["MS"],
    "QVISNAM": ["MX"],
    "QVIDNAM": ["MX"],
    "QVISQVE": ["MY"],
    "QVISQVIS": ["NA"],
    "adv. lieu relative": ["ND"],
    "QVO": ["ND", "NG", "NK", "NN"],
    "adv. lieu interrogative": ["NG"],
    "abl. diff. mes.": ["NK"],
    "QVOAD": ["NX", "PA"],
    "QVOADVSQVE": ["PC"],
    "QVOCVMQVE": ["PD"],
    "QVOD": ["PG"],
    "QVOMINVS": ["PK"],
    "QVOMODO": ["PN", "PS"],
    "QVOMODOCVMQVE": ["PT"],
    "QVONAM": ["PU"],
    "QVONIAM": ["PX"],
    "QVOQVO": ["RA"],
    "QVORSVS": ["RB"],
    "QVOT": ["RD", "RG"],
    "QVOTCVMQVE": ["RK"],
    "QVOTIENS": ["RN", "RS"],
    "QVOTIENSCVMQVE": ["RX"],
    "QVOTQVOT": ["SA"],
    "QVOTVS": ["SD"],
    "QVOTVSCVMQVE": ["SG"],
    "QVOTVSQVISQVE": ["SK"],
    "QVOVSQVE": ["SN"],
    "SEV": ["SS"],
    "SI": ["SW", "SX"],
    "SICVBI": ["SY"],
    "SICVNDE": ["SZ"],
    "SICVT": ["TA", "TB"],
    "SICVTI": ["TA", "TB"],
    "SIMVL": ["TC"],
    "SIMVLAC": ["TD"],
    "SIMVLATQVE": ["TD"],
    "SIN": ["TG"],
    "SIQVIDEM": ["TK"],
    "SIVE": ["TN"],
    "TAMENETSI": ["TR"],
    "TAMETSI": ["TS"],
    "TAMQVAM": ["TX"],
    "VBI": ["WD", "WG", "WK"],
    "VBICVMQVE": ["WN"],
    "VBINAM": ["WP"],
    "VBIVBI": ["WQ"],
    "VNDE": ["WS", "WX"],
    "VNDECVMQVE": ["XA"],
    "VSQVEQVO": ["XC"],
    "VT": ["XD", "XG", "XK"],
    "VTI": ["XD", "XG", "XK"],
    "VTCVMQVE": ["XS"],
    "VTER": ["XX", "YA"],
    "VTERCVMQVE": ["YD"],
    "VTQVI": ["YS"],
    "VTRVM": ["YW"],
    "VTRVMNE": ["YZ"],
    "adverbial": ["ZD"],
    "VELVT": ["ZD", "ZE"],
    "VELVTI": ["ZD", "ZE"],
}

subord_codes = {
    "00": (None, "main clause"),
    "AA": (None, "paratactic subjunctive"),
    "AB": (None, "paratactic subjunctive in interrogative clause"),
    "AD": (None, "ablative absolute"),
    "AE": (None, "ablative absolute"),
    "AG": (None, "infinitive clause"),
    "AK": (("AC", "ATQVE"), None),
    "AW": (("AN",), None),
    "BD": (("ANNE",), None),
    "BK": (("ANTEQVAM", "ANTEAQVAM"), None),
    "BL": (("CEV",), "adverbial or subordinating conjunction,"),
    "BM": (("CVIVS", "CVIVSMODI"), None),
    "BN": (("CVM",), None),
    "BO": (("CVICVIMODI",), None),
    "BS": (("CVMCVMQVE",), None),
    "BX": (("CVR",), "interrogative"),
    "CA": (("DONEC",), None),
    "CD": (("DVM",), None),
    "CG": (("DVMMODO",), None),
    "CL": (("ECQVIS", "ECQVID"), None),
    "CM": (("ECQVISNAM", "ECQVIDNAM"), None),
    "CN": (("ETIAMSI",), None),
    "CR": (("ECQVANDO",), None),
    "CS": (("ETSI",), None),
    "CX": (("LICET",), None),
    "DA": (("MODO",), None),
    "DD": (("NE",), "subordinating conjunction"),
    "DG": (("NE",), "interrogative"),
    "DP": (("NECNE",), None),
    "DR": (("NEC",), None),
    "DS": (("NECNE",), None),
    "DT": (("NECVBI",), None),
    "DX": (("NEDVM",), None),
    "EA": (("NEVE", "NEV"), None),
    "ED": (("NI",), None),
    "EG": (("NISI",), None),
    "EK": (("NONNE",), None),
    "EN": (("NVM",), None),
    "EP": (("NVMQVIS", "NVMQVID"), None),
    "EQ": (("NVMQVANDO",), None),
    "ES": (("POSTQVAM", "POSTEAQVAM"), None),
    "EY": (("PRAETERQVAM",), None),
    "EZ": (("PRAEVT",), None),
    "FA": (("PRIVSQVAM",), None),
    "FD": (("PROQVAM",), None),
    "FG": (("PROVT",), None),
    "FK": (("QVA",), "relative"),
    "FN": (("QVA",), "interrogative"),
    "FS": (("QVACVMQVE",), None),
    "FX": (("QVALIS",), "relative"),
    "GA": (("QVALIS",), "interrogative"),
    "GD": (("QVALISCVMQVE",), None),
    "GE": (("QVALITER",), "relative"),
    "GF": (("QVALITER",), "interrogative"),
    "GG": (("QVAM",), "relative"),
    "GK": (("QVAM",), "interrogative"),
    "GN": (("QVAMDIV",), "relative"),
    "GS": (("QVAMDIV",), "interrogative"),
    "GX": (("QVAMDVDVM",), None),
    "HA": (("QVAMLIBET",), None),
    "HD": (("QVAMOBREM",), "relative"),
    "HG": (("QVAMOBREM",), "interrogative"),
    "HK": (("QVAMQVAM",), None),
    "HN": (("QVAMVIS",), None),
    "HP": (("QVANAM",), None),
    "HS": (("QVANDO",), "interrogative adverb"),
    "HX": (("QVANDO",), "subordinating conjunction"),
    "JA": (("QVANDOCVMQVE",), None),
    "JD": (("QVANDOQVE",), None),
    "JE": (("QVANDOQVIDEM",), None),
    "JF": (("QVANTILLVS",), "relative"),
    "JG": (("QVANTILLVS",), "interrogative"),
    "JH": (("QVANTO",), "relative"),
    "JJ": (("QVANTO",), "interrogative"),
    "JK": (("QVANTOPERE",), "relative"),
    "JN": (("QVANTOPERE",), "interrogative"),
    "JS": (("QVANTVLVS",), "relative"),
    "JX": (("QVANTVLVS",), "interrogative"),
    "JZ": (("QVANTVLVSLIBET",), None),
    "KA": (("QVANTVLVSCVMQVE",), None),
    "KD": (("QVANTVS", "QVANTVM"), "relative"),
    "KG": (("QVANTVS", "QVANTVM"), "interrogative"),
    "KK": (("QVANTVSCVMQVE",), None),
    "KL": (("QVAPROPTER",), None),
    "KN": (("QVARE",), "relative"),
    "KS": (("QVARE",), "interrogative"),
    "KW": (("QVASI",), "relative adverb"),
    "KX": (("QVASI",), "subordinating conjunction"),
    "LA": (("QVATENVS",), "relative"),
    "LD": (("QVATENVS",), "interrogative"),
    "LE": (("QVATENVS",), "subordinating conjunction"),
    "LG": (("QVEMADMODVM",), "relative"),
    "LK": (("QVEMADMODVM",), "interrogative"),
    "LN": (("QVI",), "pronom relative"),
    "LS": (("QVI",), "pronom interrogative"),
    "LX": (("QVI",), "relative adverb"),
    "MA": (("QVI",), "interrogative adverb"),
    "MD": (("QVIA",), None),
    "MG": (("QVICVMQVE",), None),
    "MK": (("QVIN",), "subordinating conjunction"),
    "ML": (("NON",), "quin"),
    "MN": (("QVIPPE",), None),
    "MS": (("QVIS", "QVID"), None),
    "MX": (("QVISNAM", "QVIDNAM"), None),
    "MY": (("QVISQVE",), None),
    "NA": (("QVISQVIS",), None),
    "ND": (("QVO",), "adv. lieu relative"),
    "NG": (("QVO",), "adv. lieu interrogative"),
    "NK": (("QVO",), "abl. diff. mes."),
    "NN": (("QVO",), "subordinating conjunction"),
    "NX": (("QVOAD",), "relative"),
    "PA": (("QVOAD",), "interrogative"),
    "PC": (("QVOADVSQVE",), None),
    "PD": (("QVOCVMQVE",), None),
    "PG": (("QVOD",), None),
    "PK": (("QVOMINVS",), None),
    "PN": (("QVOMODO",), "relative"),
    "PS": (("QVOMODO",), "interrogative"),
    "PT": (("QVOMODOCVMQVE",), None),
    "PU": (("QVONAM",), None),
    "PX": (("QVONIAM",), None),
    "RA": (("QVOQVO",), None),
    "RB": (("QVORSVS",), None),
    "RD": (("QVOT",), "relative"),
    "RG": (("QVOT",), "interrogative"),
    "RK": (("QVOTCVMQVE",), None),
    "RN": (("QVOTIENS",), "relative"),
    "RS": (("QVOTIENS",), "interrogative"),
    "RX": (("QVOTIENSCVMQVE",), None),
    "SA": (("QVOTQVOT",), None),
    "SD": (("QVOTVS",), None),
    "SG": (("QVOTVSCVMQVE",), "relative"),
    "SK": (("QVOTVSQVISQVE",), None),
    "SN": (("QVOVSQVE",), None),
    "SS": (("SEV",), None),
    "SW": (("SI",), "interrogative adverb"),
    "SX": (("SI",), "subordinating conjunction"),
    "SY": (("SICVBI",), None),
    "SZ": (("SICVNDE",), None),
    "TA": (("SICVT", "SICVTI"), "relative adverb"),
    "TB": (("SICVT", "SICVTI"), "subordinating conjunction"),
    "TC": (("SIMVL",), None),
    "TD": (("SIMVLAC", "SIMVLATQVE"), None),
    "TG": (("SIN",), None),
    "TK": (("SIQVIDEM",), None),
    "TN": (("SIVE",), None),
    "TR": (("TAMENETSI",), None),
    "TS": (("TAMETSI",), None),
    "TX": (("TAMQVAM",), None),
    "WD": (("VBI",), "relative adverb"),
    "WG": (("VBI",), "interrogative adverb"),
    "WK": (("VBI",), "subordinating conjunction"),
    "WN": (("VBICVMQVE",), None),
    "WP": (("VBINAM",), None),
    "WQ": (("VBIVBI",), None),
    "WS": (("VNDE",), "relative"),
    "WX": (("VNDE",), "interrogative"),
    "XA": (("VNDECVMQVE",), None),
    "XC": (("VSQVEQVO",), None),
    "XD": (("VT", "VTI"), "relative adverb"),
    "XG": (("VT", "VTI"), "interrogative adverb"),
    "XK": (("VT", "VTI"), "subordinating conjunction"),
    "XS": (("VTCVMQVE",), None),
    "XX": (("VTER",), "relative"),
    "YA": (("VTER",), "interrogative"),
    "YD": (("VTERCVMQVE",), None),
    "YS": (("VTQVI",), None),
    "YW": (("VTRVM",), None),
    "YZ": (("VTRVMNE",), None),
    "ZD": (("VELVT", "VELVTI"), "adverbial"),
    "ZE": (("VELVT", "VELVTI"), "subordinating conjunction"),
}

MORPHS = [
    {  # Case
        "1": "n",
        "2": "v",
        "3": "a",
        "4": "g",
        "5": "d",
        "6": "b",
        "7": "l",
        "8": "-",
    },
    {"1": "s", "2": "p"},  # Number
    {"1": "p", "2": "c", "3": "s"},  # Degree
    {  # Mood
        "0": "-",  # Ignore (we had one for an agendis annotation)
        "1": "i",
        "3": "s",
        "2": "m",
        "4": "p",
        "7": "n",
        "5": "d",
        "6": "g",
        "8": "s",
        "9": "s",
    },
    {  # Tense
        "0": "-",  # Ignore (We had one for an agendis annotation)
        "1": "p",
        "2": "i",
        "3": "f",
        "4": "r",
        "5": "l",
        "6": "u",
        "7": "r",  # periphrasis?
        "8": "l",  # periphrasis?
        "9": "f",  # periphrasis?
    },
    {"1": "a", "2": "p", "3": "d", "4": "d"},  # Voice  # semi-deponent
    {"1": "1", "2": "2", "3": "3"},  # Person
    {
        # Gender
        "0": "-",  # Ignore (We had one for an agendis annotation)
        "1": "a",
        "2": "f",
        "3": "m/f",  # c
        "4": "m",
        "5": "m/n",  # c
        "6": "n",
    },
]

GROUP = {
    "A": {
        "1": {"group": "1", "stem": "-"},
        "2": {"group": "2", "stem": "-"},
        "3": {"group": "3", "stem": "-"},
        "4": {"group": "4", "stem": "-"},
        "5": {"group": "5", "stem": "-"},
        "6": {"group": "-", "stem": "-"},
        "7": {
            "group": "3",
            "stem": "g",
        },  # could be group 2 -os, -on or even group 1 -es/-as, -ae
    },
    "B": {
        "1": {"group": "1", "stem": "-"},
        "2": {"group": "2", "stem": "-"},
        "3": {"group": "3", "stem": "-"},
        "4": {"group": "4", "stem": "-"},
        "5": {"group": "3", "stem": "i"},
        "6": {
            "group": "3",
            "stem": "-",
        },  # irregulars are classed with group 3
    },
    "C": {
        "1": {"group": "1", "stem": "-"},
        "2": {"group": "3", "stem": "-"},
        "3": {"group": "3", "stem": "r"},
        "4": {"group": "3", "stem": "i"},
        "5": {"group": "3", "stem": "-"},
        "6": {"group": "-", "stem": "-"},  # irregulars
        "7": {"group": "1", "stem": "g"},  # greek decl in -os, -on
        " ": {"group": "-", "stem": "-"},  # unspecified
    },
    "D": {
        "1": {"group": "-", "stem": "-"},  # indecl except for unus, duo, tres
        "2": {"group": "1", "stem": "-"},  # -us
        "3": {"group": "1", "stem": "-"},  # -eni
        "4": {"group": "3", "stem": "-"},  # -ex
        "5": {"group": "-", "stem": "-"},  # adverbs
        "6": {"group": "-", "stem": "-"},
    },
}

POS_MAP = {
    "A": "n",
    "B": "v",
    "C": "a",
    "D": "a",
    "E": "n",
    "F": "n",
    "G": "n",
    "H": "n",
    "I": "n",
    "J": "n",
    "K": "n",
    "L": "n",
    "M": "r",
    "N": "r",
    "O": "r",
    "P": "r",
    "Q": "r",
    "R": "p",
    "S": "-",
    "T": "-",
    "U": "-",
    "#": "v",
    "0": "-",
    " ": "-",
}


def bpn2wn(morph_code: str) -> str:
    """
    Convert a BPN code string to a Perseus-like morphological string

    :return: str
    """
    if len(morph_code) < 11:
        morph_code += " " * (11 - len(morph_code))
    pos = POS_MAP[morph_code[0]]
    morph = morph_code[2:]
    if morph_code[0] in "ABCD":
        group = GROUP[morph_code[0]][morph_code[1]]
    else:
        group = {"group": "-", "stem": "-"}
    morpho = []
    for i, morph_char in enumerate(morph):
        if morph_char.strip():
            morpho.append(MORPHS[i].get(morph_char, "-"))
        else:
            morpho.append("-")
    if pos in "ar":  # aps---mn1-
        desc = f"{pos}{morpho[2]}{morpho[1]}{morpho[4]}{morpho[3]}{morpho[5]}{morpho[7]}{morpho[0]}{group['group']}{group['stem']}"
    else:
        desc = f"{pos}{morpho[6]}{morpho[1]}{morpho[4]}{morpho[3]}{morpho[5]}{morpho[7]}{morpho[0]}{group['group']}{group['stem']}"
    return desc


def parse_bpn(line: str) -> dict:
    """ Parse a single line of the LASLA corpus

    :param line: Input line
    :return: Dictionary
    """
    line = line.replace("\n", "")
    if not line.strip():
        return None

    author_code = line[0]
    work_code = line[1:3]
    form_code = line[3]
    sent_id = line[4:8]
    lemma, lemma_n = line[8:29].strip(), line[29]
    form = line[30:55].strip()
    refs = line[55:66].strip()
    ref_n = line[66]
    morpho = line[67:76]
    subord = line[76:78]
    if len(line) == 80:
        morpho += line[79]
    else:
        morpho += " "
    return {
        "author_code": author_code,
        "work_code": work_code,
        "form_code": form_code,
        "sent_id": sent_id,
        "refs": refs,
        "ref_n": ref_n,
        "lemma": lemma,
        "lemma_n": lemma_n,
        "form": form,
        "morpho": morpho,
        "subord": subord,
    }


DICT_MORPH = {
    "-": {"-": {"-": "----------"}},
    "N": {
        "2": {
            "1": "aps---mn1-",
            "2": "aps---an3-",
            "3": "aps---mn1r",
            "4": "aps---cn3i",
            "5": "aps---an3-",
            "6": "aps---an--",
            "7": "aps---mn1g",
            "&": "acs---cn3-",
        },
        "1": {
            "1": "n-s----n1-",
            "2": "n-s----n2-",
            "3": "n-s----n3-",
            "4": "n-s----n4-",
            "5": "n-s----n5-",
            "6": "n-s----n--",
            "7": "n-s----n-g",
        },
    },
    "A": {
        "1": {
            "1": "n-s----n1-",
            "2": "n-s----n2-",
            "3": "n-s----n3-",
            "4": "n-s----n4-",
            "5": "n-s----n5-",
            "6": "n-s----n--",
            "7": "n-s----n-g",
        },
        "2": {
            "1": "aps---mn1-",
            "2": "aps---an3-",
            "3": "aps---mn1r",
            "4": "aps---cn3i",
            "5": "aps---an3-",
            "6": "aps---an--",
            "7": "aps---mn1g",
            "&": "acs---cn3-",
        },
    },
    # NOVNS
    "1": {
        "1": "n-s----n1-",
        "2": "n-s----n2-",
        "3": "n-s----n3-",
        "4": "n-s----n4-",
        "5": "n-s----n5-",
        "6": "n-s----n--",
        "7": "n-s----n-g",
    },
    # ADJECTIVES
    "2": {
        "1": "aps---mn1-",
        "2": "aps---an3-",
        "3": "aps---mn1r",
        "4": "aps---cn3i",
        "5": "aps---an3-",
        "6": "aps---an--",
        "7": "aps---mn1g",
        "&": "acs---cn3-",
        "G": "aps---mn1g",
    },
    # NVMERALS
    "3": {
        "1": "----------",
        "2": "aps---mn1-",
        "3": "rp--------",
        "4": "aps---an3-",
        "5": "rp--------",
        "6": "rp--------",
        "N": "rp--------",
        "K": "aps---mn1-",
        "E": "rc--------",
        "B": "acs---cn3-",
    },
    # PRONOVNS
    "4": {
        "1": "o-----an--",
        "2": "aps---mn1-",
        "3": "o-----an--",  # reflexive
        "4": "aps---mn1-",
        "5": "aps---mn1-",  # some pronominal declensions
        "6": "aps---mn1-",  # some pronominal, some indeclinable
        "7": "----------",  # interrogatives
        "8": "----------",  # indefinites
    },
    # VERBS
    "5": {
        "1": "v1spia--1-",
        "2": "v1spia--2-",
        "3": "v1spia--3-",
        "4": "v1spia--4-",
        "5": "v1spia--3i",
        "6": "v1spia--3-",  # irregulars
        "7": "----------",  # greek participles?
        "J": "v1spid--1-",  # deponent
        "K": "v1spid--2-",
        "L": "v1spid--3-",
        "M": "v1spid--4-",
        "N": "v1spid--3i",
        "S": "v1spia--2-",  # semi-deponents
        "T": "v1spia--3-",
    },
    # ADVERBS
    "6": {
        "-": "rp--------",
        "&": "rp--------",
        "0": "rp--------",
        "6": "r---------",
        "7": "r---------",
        "8": "r---------",
        "9": "----------",  # conjunctions
    },
    # PREPOSITIONS
    "7": {"-": "p---------", "0": "p---------"},
    # CONJVNCTION?
    "8": {"1": "----------", "2": "----------"},
    # INTERJECTION
    "9": {"0": "----------"},
    "E": {"6": "v1spia--3-"},  # auxiliary verbs
}

lexicon = {
    "AB": {"-": ["70"]},
    "ABACTVS": {"-": ["14"]},
    "ABACVS": {"-": ["12"]},
    "ABALIENO": {"-": ["51"]},
    "ABATOS": {"-": ["27"]},
    "ABAVVS": {"-": ["12"]},
    "ABBAS": {"-": ["13"]},
    "ABDICO": {"1": ["51"]},
    "ABDITVM": {"-": ["12"]},
    "ABDITVS": {"C": ["21"], "-": ["21"]},
    "ABDO": {"-": ["53"]},
    "ABDOMEN": {"-": ["13"]},
    "ABDVCO": {"-": ["53"]},
    "ABEO": {"1": ["56"]},
    "ABERRO": {"2": ["51"]},
    "ABHINC": {"-": ["60"]},
    "ABHORREO": {"-": ["52"]},
    "ABHORRIDVS": {"-": ["21"]},
    "ABICIO": {"-": ["55"]},
    "ABIECTE": {"-": ["6&", "66"]},
    "ABIECTVS": {"-": ["21"]},
    "ABIEGNEVS": {"-": ["21"]},
    "ABIEGNVS": {"-": ["21"]},
    "ABIES": {"-": ["13"]},
    "ABIGO": {"-": ["53"]},
    "ABITVS": {"-": ["14"]},
    "ABIVDICO": {"-": ["51"]},
    "ABIVNGO": {"-": ["53"]},
    "ABIVRO": {"-": ["51"]},
    "ABLAQVEO": {"1": ["51"]},
    "ABLEGO": {"1": ["51"]},
    "ABLEVO": {"1": ["51"]},
    "ABLVDO": {"-": ["53"]},
    "ABLVO": {"-": ["53"]},
    "ABNEGO": {"-": ["51"]},
    "ABNEPOS": {"-": ["13"]},
    "ABNOCTO": {"-": ["51"]},
    "ABNORMIS": {"-": ["24"]},
    "ABNVO": {"-": ["53"]},
    "ABOLEO": {"-": ["52"]},
    "ABOLESCO": {"-": ["53"]},
    "ABOLITIO": {"-": ["13"]},
    "ABOLLA": {"-": ["11"]},
    "ABOMINOR": {"-": ["5J"]},
    "ABORIOR": {"-": ["5M"]},
    "ABORISCOR": {"-": ["5L"]},
    "ABORTIVVS": {"-": ["21"]},
    "ABORTVS": {"-": ["14"]},
    "ABRADO": {"-": ["53"]},
    "ABRIPIO": {"-": ["55"]},
    "ABROGO": {"-": ["51"]},
    "ABROTONVM": {"-": ["12"]},
    "ABROTONVS": {"-": ["12"]},
    "ABRVMPO": {"-": ["53"]},
    "ABRVPTVM": {"-": ["12"]},
    "ABRVPTVS": {"-": ["21"]},
    "ABSCEDO": {"1": ["53"]},
    "ABSCESSVS": {"-": ["14"]},
    "ABSCIDO": {"2": ["53"]},
    "ABSCINDO": {"-": ["53"]},
    "ABSCISIO": {"-": ["13"]},
    "ABSCISVS": {"-": ["21"]},
    "ABSCONDITVS": {"C": ["21"], "-": ["21"]},
    "ABSCONDO": {"-": ["53"]},
    "ABSCONSVS": {"C": ["21"], "-": ["21"]},
    "ABSENS": {"-": ["25"]},
    "ABSENTIA": {"-": ["11"]},
    "ABSILIO": {"-": ["54"]},
    "ABSIMILIS": {"-": ["24"]},
    "ABSINTHIVM": {"-": ["12"]},
    "ABSISTO": {"-": ["53"]},
    "ABSOLVO": {"-": ["53"]},
    "ABSOLVTE": {"-": ["60"]},
    "ABSOLVTIO": {"-": ["13"]},
    "ABSOLVTVS": {"-": ["21"]},
    "ABSONVS": {"-": ["21"]},
    "ABSORBEO": {"-": ["52"]},
    "ABSQVE": {"1": ["60"], "2": ["70"]},
    "ABSTANTIA": {"-": ["11"]},
    "ABSTEMIVS": {"-": ["21"]},
    "ABSTERGEO": {"-": ["52"]},
    "ABSTERREO": {"-": ["52"]},
    "ABSTINENS": {"-": ["25"]},
    "ABSTINENTER": {"-": ["60"]},
    "ABSTINENTIA": {"-": ["11"]},
    "ABSTINEO": {"-": ["52"]},
    "ABSTRAHO": {"-": ["53"]},
    "ABSTRVDO": {"-": ["53"]},
    "ABSTRVSVS": {"-": ["21"]},
    "ABSVM": {"1": ["56"]},
    "ABSVMO": {"-": ["53"]},
    "ABSVRDE": {"-": ["60"]},
    "ABSVRDVS": {"-": ["21"]},
    "ABVNDANS": {"1": ["13"], "2": ["25"]},
    "ABVNDANTIA": {"-": ["11"]},
    "ABVNDE": {"-": ["60"]},
    "ABVNDO": {"-": ["51"]},
    "ABVNDVS": {"-": ["21"]},
    "ABVSQVE": {"-": ["70"]},
    "ABVTOR": {"-": ["5L"]},
    "AC": {"1": ["81"], "2": ["82"]},
    "ACADEMICI": {"N": ["12"]},
    "ACADEMICVS": {"N": ["21"], "A": ["21"]},
    "ACALANTHIS": {"-": ["17"]},
    "ACANTHVS": {"-": ["12"]},
    "ACCEDO": {"1": ["53"]},
    "ACCELERO": {"-": ["51"]},
    "ACCENDO": {"-": ["53"]},
    "ACCENSEO": {"-": ["52"]},
    "ACCENSI": {"-": ["12"]},
    "ACCENSVS": {"1": ["12"], "2": ["14"]},
    "ACCEPTATVS": {"-": ["21"]},
    "ACCEPTIO": {"-": ["13"]},
    "ACCEPTO": {"-": ["51"]},
    "ACCEPTVM": {"-": ["12"]},
    "ACCEPTVS": {"-": ["21"]},
    "ACCESSIO": {"-": ["13"]},
    "ACCESSITO": {"-": ["51"]},
    "ACCESSO": {"-": ["51"]},
    "ACCESSVS": {"-": ["14"]},
    "ACCIDENS": {"-": ["13"]},
    "ACCIDO": {"1": ["53"], "2": ["53"]},
    "ACCIEO": {"-": ["52"]},
    "ACCINGO": {"-": ["53"]},
    "ACCIO": {"-": ["54"]},
    "ACCIPIO": {"-": ["55"]},
    "ACCIPITER": {"-": ["13"]},
    "ACCITO": {"1": ["51"]},
    "ACCITVS": {"-": ["14"]},
    "ACCLAMATIO": {"-": ["13"]},
    "ACCLAMO": {"-": ["51"]},
    "ACCLINIS": {"-": ["24"]},
    "ACCLIVIS": {"-": ["24"]},
    "ACCLIVITAS": {"-": ["13"]},
    "ACCLIVVS": {"-": ["21"]},
    "ACCOGNOSCO": {"-": ["53"]},
    "ACCOLA": {"-": ["11"]},
    "ACCOLO": {"2": ["53"]},
    "ACCOMMODATIO": {"-": ["13"]},
    "ACCOMMODATVS": {"-": ["21"]},
    "ACCOMMODO": {"-": ["51"]},
    "ACCOMMODVS": {"-": ["21"]},
    "ACCREDO": {"-": ["53"]},
    "ACCREMENTVM": {"-": ["12"]},
    "ACCRESCO": {"-": ["53"]},
    "ACCVBITIO": {"-": ["13"]},
    "ACCVBO": {"-": ["51"]},
    "ACCVMBO": {"-": ["53"]},
    "ACCVMVLATOR": {"-": ["13"]},
    "ACCVMVLO": {"-": ["51"]},
    "ACCVRATE": {"-": ["60"]},
    "ACCVRATVS": {"-": ["21"]},
    "ACCVRO": {"-": ["51"]},
    "ACCVRRO": {"-": ["53"]},
    "ACCVRSVS": {"-": ["14"]},
    "ACCVSATIO": {"-": ["13"]},
    "ACCVSATOR": {"-": ["13"]},
    "ACCVSATORIE": {"-": ["60"]},
    "ACCVSATORIVS": {"-": ["21"]},
    "ACCVSO": {"-": ["51"]},
    "ACEO": {"-": ["52"]},
    "ACER": {"1": ["13"], "2": ["23"]},
    "ACERBE": {"-": ["60"]},
    "ACERBITAS": {"-": ["13"]},
    "ACERBVM": {"-": ["12"]},
    "ACERBVS": {"-": ["21"]},
    "ACERNVS": {"-": ["21"]},
    "ACERRA": {"-": ["11"]},
    "ACERSECOMES": {"-": ["17"]},
    "ACERVATIM": {"-": ["60"]},
    "ACERVO": {"-": ["51"]},
    "ACERVVS": {"-": ["12"]},
    "ACESCO": {"-": ["53"]},
    "ACESTENSES": {"N": ["13"]},
    "ACETABVLVM": {"-": ["12"]},
    "ACETVM": {"-": ["12"]},
    "ACHAEVS": {"N": ["21"], "A": ["21"]},
    "ACHAIA": {"N": ["11"]},
    "ACHAICVS": {"N": ["21"], "A": ["21"]},
    "ACHIVI": {"N": ["12"]},
    "ACHIVVS": {"N": ["21"], "A": ["21"]},
    "ACHRADINA": {"N": ["11"]},
    "ACIDVS": {"-": ["21"]},
    "ACIES": {"-": ["15"]},
    "ACILIVS": {"A": ["21"], "N": ["12"]},
    "ACINA": {"1": ["11"], "-": ["11"]},
    "ACINACES": {"-": ["13"]},
    "ACINVM": {"-": ["12"]},
    "ACINVS": {"-": ["12"]},
    "ACIPENSER": {"-": ["13"]},
    "ACOENONOETVS": {"-": ["12"]},
    "ACONITVM": {"-": ["12"]},
    "ACOSMOS": {"-": ["17"]},
    "ACQVIESCO": {"-": ["53"]},
    "ACQVIRO": {"-": ["53"]},
    "ACQVISITIO": {"-": ["13"]},
    "ACRIMONIA": {"-": ["11"]},
    "ACRITER": {"-": ["60"]},
    "ACRITVDO": {"-": ["13"]},
    "ACROAMA": {"-": ["17"]},
    "ACROASIS": {"-": ["17"]},
    "ACROBATICOS": {"-": ["27"]},
    "ACROLITHVS": {"-": ["27"]},
    "ACROTERIVM": {"-": ["12"]},
    "ACTA": {"-": ["11"]},
    "ACTIACVS": {"N": ["21"], "A": ["21"]},
    "ACTIO": {"-": ["13"]},
    "ACTITO": {"-": ["51"]},
    "ACTIVITAS": {"-": ["13"]},
    "ACTIVS": {"-": ["21"], "A": ["21"]},
    "ACTIVVS": {"-": ["21"]},
    "ACTOR": {"-": ["13"], "N": ["13"]},
    "ACTVARIA": {"-": ["11"]},
    "ACTVARIVS": {"1": ["12"], "2": ["21"]},
    "ACTVM": {"-": ["12"]},
    "ACTVOSVS": {"-": ["21"]},
    "ACTVS": {"1": ["14"], "2": ["21"]},
    "ACTVTVM": {"-": ["60"]},
    "ACVLEATVS": {"-": ["21"]},
    "ACVLEVS": {"-": ["12"]},
    "ACVMEN": {"-": ["13"]},
    "ACVO": {"-": ["53"]},
    "ACVS": {"1": ["12"], "2": ["13"], "3": ["14"]},
    "ACVTE": {"-": ["60"]},
    "ACVTVLVS": {"-": ["21"]},
    "ACVTVM": {"-": ["12"]},
    "ACVTVS": {"-": ["21"]},
    "AD": {"1": ["60"], "2": ["70"]},
    "ADACTVS": {"-": ["14"]},
    "ADAEQVO": {"-": ["51"]},
    "ADAGGERO": {"1": ["51"]},
    "ADAMANTINVS": {"-": ["21"]},
    "ADAMAS": {"-": ["17"]},
    "ADAMO": {"-": ["51"]},
    "ADAPERIO": {"-": ["54"]},
    "ADAREO": {"-": ["52"]},
    "ADAVCTVS": {"-": ["14"]},
    "ADAVGEO": {"-": ["52"]},
    "ADAVGESCO": {"-": ["53"]},
    "ADAVGMEN": {"-": ["13"]},
    "ADBIBO": {"2": ["53"]},
    "ADDECET": {"-": ["52"]},
    "ADDICO": {"2": ["53"]},
    "ADDICTIO": {"-": ["13"]},
    "ADDISCO": {"-": ["53"]},
    "ADDITAMENTVM": {"-": ["12"]},
    "ADDO": {"-": ["53"]},
    "ADDOCEO": {"-": ["52"]},
    "ADDVBITO": {"-": ["51"]},
    "ADDVCO": {"-": ["53"]},
    "ADDVCTE": {"-": ["6&"]},
    "ADDVCTVS": {"-": ["21"]},
    "ADEDO": {"2": ["56"]},
    "ADEMO": {"-": ["53"]},
    "ADEMPTIO": {"-": ["13"]},
    "ADEO": {"1": ["56"], "2": ["60"]},
    "ADEPS": {"-": ["13"]},
    "ADEQVITO": {"-": ["51"]},
    "ADHAEREO": {"-": ["52"]},
    "ADHAERESCO": {"-": ["53"]},
    "ADHAESVS": {"-": ["14"]},
    "ADHIBEO": {"-": ["52"]},
    "ADHINNIO": {"-": ["54"]},
    "ADHORTATIO": {"-": ["13"]},
    "ADHORTATOR": {"-": ["13"]},
    "ADHORTOR": {"-": ["5J"]},
    "ADHVC": {"-": ["60"]},
    "ADIACEO": {"-": ["52"]},
    "ADICIO": {"-": ["55"]},
    "ADICO": {"2": ["53"]},
    "ADIECTIO": {"-": ["13"]},
    "ADIECTVS": {"-": ["14"]},
    "ADIGO": {"-": ["53"]},
    "ADIMO": {"-": ["53"]},
    "ADINDO": {"-": ["53"]},
    "ADIPATVM": {"-": ["12"]},
    "ADIPES": {"-": ["13"]},
    "ADIPISCOR": {"-": ["5L"]},
    "ADITIALIS": {"-": ["24"]},
    "ADITVS": {"-": ["14"]},
    "ADIVDICO": {"-": ["51"]},
    "ADIVMENTVM": {"-": ["12"]},
    "ADIVNCTIO": {"-": ["13"]},
    "ADIVNCTVM": {"-": ["12"]},
    "ADIVNCTVS": {"-": ["21"]},
    "ADIVNGO": {"-": ["53"]},
    "ADIVRO": {"-": ["51"]},
    "ADIVTO": {"-": ["51"]},
    "ADIVTOR": {"1": ["13"], "2": ["5J"]},
    "ADIVTORIVM": {"-": ["12"]},
    "ADIVTRIX": {"-": ["13"], "N": ["13"]},
    "ADIVVO": {"-": ["51"]},
    "ADMATVRO": {"-": ["51"]},
    "ADMETIOR": {"-": ["5M"]},
    "ADMINICVLVM": {"-": ["12"]},
    "ADMINISTER": {"-": ["12"]},
    "ADMINISTRA": {"-": ["11"]},
    "ADMINISTRATIO": {"-": ["13"]},
    "ADMINISTRO": {"-": ["51"]},
    "ADMIRABILIS": {"-": ["24"]},
    "ADMIRANDVS": {"-": ["21"]},
    "ADMIRATIO": {"-": ["13"]},
    "ADMIRATOR": {"-": ["13"]},
    "ADMIROR": {"-": ["5J"]},
    "ADMISCEO": {"-": ["52"]},
    "ADMISSIO": {"-": ["13"]},
    "ADMISSVM": {"-": ["12"]},
    "ADMITTO": {"-": ["53"]},
    "ADMODERATE": {"-": ["60"]},
    "ADMODVM": {"-": ["60"]},
    "ADMONEO": {"-": ["52"]},
    "ADMONITIO": {"-": ["13"]},
    "ADMONITOR": {"-": ["13"]},
    "ADMONITVS": {"-": ["14"]},
    "ADMORDEO": {"-": ["52"]},
    "ADMOVEO": {"-": ["52"]},
    "ADMVGIO": {"-": ["54"]},
    "ADMVRMVRATIO": {"-": ["13"]},
    "ADMVRMVRO": {"-": ["51"]},
    "ADOLEO": {"-": ["52"]},
    "ADOLESCENS": {"1": ["13"], "2": ["25"]},
    "ADOLESCENTIA": {"-": ["11"]},
    "ADOLESCENTVLVS": {"-": ["12"]},
    "ADOLESCO": {"1": ["53"], "2": ["53"]},
    "ADOPERIO": {"-": ["54"]},
    "ADOPINOR": {"-": ["5J"]},
    "ADOPTATIO": {"-": ["13"]},
    "ADOPTIO": {"-": ["13"]},
    "ADOPTIVVS": {"-": ["21"]},
    "ADOPTO": {"-": ["51"]},
    "ADOR": {"-": ["13"]},
    "ADOREA": {"-": ["11"]},
    "ADOREVS": {"-": ["21"]},
    "ADORIOR": {"-": ["5M"]},
    "ADORNO": {"-": ["51"]},
    "ADORO": {"-": ["51"]},
    "ADVECTICIVS": {"-": ["21"]},
    "ADVECTO": {"-": ["51"]},
    "ADVECTOR": {"-": ["13"]},
    "ADVECTVS": {"-": ["14"]},
    "ADVEHO": {"-": ["53"]},
    "ADVEHOR": {"-": ["5L"]},
    "ADVELO": {"-": ["51"]},
    "ADVENA": {"-": ["11"]},
    "ADVENIO": {"-": ["54"]},
    "ADVENTICIVM": {"-": ["12"]},
    "ADVENTICIVS": {"-": ["21"]},
    "ADVENTO": {"-": ["51"]},
    "ADVENTVS": {"-": ["14"]},
    "ADVERRO": {"-": ["53"]},
    "ADVERSARIVS": {"1": ["12"], "2": ["21"]},
    "ADVERSO": {"-": ["51"]},
    "ADVERSOR": {"-": ["5J"]},
    "ADVERSVM": {"-": ["12"]},
    "ADVERSVS": {"1": ["21"], "2": ["60"], "-": ["70"], "3": ["70"]},
    "ADVERTO": {"-": ["33"]},
    "ADVESPERASCO": {"-": ["53"]},
    "ADVIGILO": {"-": ["51"]},
    "ADVIVO": {"-": ["53"]},
    "ADVLATIO": {"-": ["13"]},
    "ADVLATOR": {"-": ["13"]},
    "ADVLATORIVS": {"-": ["21"]},
    "ADVLESCENTIA": {"-": ["11"]},
    "ADVLO": {"-": ["51"]},
    "ADVLOR": {"-": ["5J"]},
    "ADVLTER": {"1": ["12"], "2": ["21"]},
    "ADVLTERA": {"-": ["11"]},
    "ADVLTERINVS": {"-": ["21"]},
    "ADVLTERIVM": {"-": ["12"]},
    "ADVLTERO": {"-": ["51"]},
    "ADVLTVS": {"-": ["21"]},
    "ADVMBRATIM": {"-": ["60"]},
    "ADVMBRATIO": {"-": ["13"]},
    "ADVMBRATVS": {"-": ["21"]},
    "ADVMBRO": {"-": ["51"]},
    "ADVNCVS": {"-": ["21"]},
    "ADVOCATIO": {"-": ["13"]},
    "ADVOCATVS": {"-": ["12"]},
    "ADVOCO": {"-": ["51"]},
    "ADVOLO": {"2": ["51"]},
    "ADVOLVO": {"-": ["53"]},
    "ADVRGEO": {"-": ["52"]},
    "ADVRO": {"-": ["53"]},
    "ADVSTVS": {"-": ["21"]},
    "ADYTVM": {"-": ["12"]},
    "AEBVTIVS": {"N": ["12"], "A": ["21"]},
    "AEDES": {"-": ["13"]},
    "AEDICVLA": {"-": ["11"]},
    "AEDIFICATIO": {"-": ["13"]},
    "AEDIFICATOR": {"-": ["13"]},
    "AEDIFICIVM": {"-": ["12"]},
    "AEDIFICO": {"-": ["51"]},
    "AEDILICIVS": {"1": ["12"], "2": ["21"]},
    "AEDILIS": {"-": ["13"]},
    "AEDILITAS": {"-": ["13"]},
    "AEDITVENS": {"-": ["13"]},
    "AEDITVMVS": {"-": ["12"]},
    "AEDITVVS": {"-": ["12"]},
    "AEDON": {"-": ["17"]},
    "AEGAEVS": {"-": ["21"], "A": ["21"]},
    "AEGER": {"1": ["12"], "2": ["21"]},
    "AEGINIVM": {"N": ["12"]},
    "AEGIS": {"-": ["17"], "N": ["13"]},
    "AEGRE": {"-": ["60"]},
    "AEGREO": {"-": ["52"]},
    "AEGRESCO": {"-": ["53"]},
    "AEGRIMONIA": {"-": ["11"]},
    "AEGRITOMARVS": {"N": ["12"]},
    "AEGRITVDO": {"-": ["13"]},
    "AEGROR": {"-": ["13"]},
    "AEGROTATIO": {"-": ["13"]},
    "AEGROTO": {"-": ["51"]},
    "AEGROTVS": {"-": ["21"]},
    "AEGRVM": {"C": ["12"], "-": ["12"]},
    "AEGYPTIVS": {"N": ["21"], "A": ["21"]},
    "AEGYPTVS": {"N": ["12"]},
    "AELINOS": {"-": ["17"]},
    "AELVRVS": {"-": ["12"]},
    "AEMILIANVS": {"N": ["21"], "A": ["21"]},
    "AEMILIVS": {"N": ["12"], "A": ["21"]},
    "AEMVLA": {"-": ["11"]},
    "AEMVLATIO": {"-": ["13"]},
    "AEMVLATOR": {"-": ["13"]},
    "AEMVLATVS": {"-": ["14"]},
    "AEMVLOR": {"-": ["5J"]},
    "AEMVLVS": {"1": ["12"], "2": ["21"]},
    "AENEAS": {"N": ["17"]},
    "AENEATOR": {"-": ["13"]},
    "AENEVM": {"-": ["12"]},
    "AENEVS": {"-": ["21"]},
    "AENIGMA": {"-": ["17"]},
    "AENIPES": {"-": ["25"]},
    "AENVM": {"-": ["12"]},
    "AENVS": {"-": ["21"], "N": ["12"]},
    "AEQVABILIS": {"-": ["24"]},
    "AEQVABILITER": {"-": ["60"]},
    "AEQVAEVVS": {"-": ["21"]},
    "AEQVALIS": {"1": ["13"], "2": ["24"]},
    "AEQVALITAS": {"-": ["13"]},
    "AEQVALITER": {"-": ["60"]},
    "AEQVANIMITAS": {"-": ["13"]},
    "AEQVE": {"-": ["60"]},
    "AEQVI": {"N": ["12"]},
    "AEQVICOLI": {"N": ["12"]},
    "AEQVICVS": {"N": ["21"]},
    "AEQVILATATIO": {"-": ["13"]},
    "AEQVINOCTIALIS": {"-": ["24"]},
    "AEQVINOCTIVM": {"-": ["12"]},
    "AEQVIPARO": {"-": ["51"]},
    "AEQVIPOLLENS": {"-": ["25"]},
    "AEQVIPONDIVM": {"-": ["12"]},
    "AEQVITAS": {"-": ["13"]},
    "AEQVO": {"-": ["51"]},
    "AEQVOR": {"-": ["13"], "N": ["13"]},
    "AEQVOREVS": {"-": ["21"]},
    "AEQVVM": {"-": ["12"]},
    "AEQVVS": {"-": ["21"], "N": ["21"]},
    "AER": {"-": ["17"]},
    "AERACEVS": {"-": ["21"]},
    "AERARIA": {"-": ["11"]},
    "AERARIVM": {"-": ["12"]},
    "AERARIVS": {"1": ["12"], "2": ["21"]},
    "AERATVS": {"-": ["21"]},
    "AEREVS": {"1": ["12"], "2": ["21"]},
    "AERIPES": {"-": ["25"]},
    "AERIVS": {"-": ["21"]},
    "AERO": {"-": ["13"]},
    "AERVCA": {"-": ["11"]},
    "AERVGINOSVS": {"-": ["21"]},
    "AERVGO": {"-": ["13"]},
    "AERVMNA": {"-": ["11"]},
    "AERVMNABILIS": {"-": ["24"]},
    "AERVMNOSVS": {"-": ["21"]},
    "AES": {"-": ["13"]},
    "AESCHRIO": {"N": ["17"]},
    "AESCHYLVS": {"N": ["12"]},
    "AESCVLAPIVS": {"N": ["12"]},
    "AESCVLETVM": {"-": ["12"]},
    "AESCVLINVS": {"-": ["21"]},
    "AESCVLVS": {"-": ["12"]},
    "AESERNINVS": {"C": ["12"], "N": ["12"]},
    "AESOPEVS": {"N": ["21"], "A": ["21"]},
    "AESTAS": {"-": ["13"]},
    "AESTIFER": {"-": ["21"]},
    "AESTIMATIO": {"-": ["13"]},
    "AESTIMATOR": {"-": ["13"]},
    "AESTIMO": {"-": ["51"]},
    "AESTIVA": {"-": ["12"]},
    "AESTIVVS": {"-": ["21"]},
    "AESTVARIVM": {"-": ["12"]},
    "AESTVO": {"-": ["51"]},
    "AESTVOSE": {"-": ["6-"]},
    "AESTVOSVS": {"-": ["21"]},
    "AESTVS": {"-": ["14"]},
    "AETAS": {"-": ["13"]},
    "AETERNITAS": {"-": ["13"]},
    "AETERNO": {"1": ["51"], "2": ["60"]},
    "AETERNVM": {"1": ["12"], "2": ["60"]},
    "AETERNVS": {"-": ["21"]},
    "AETHER": {"-": ["17"]},
    "AETHERIVS": {"-": ["21"]},
    "AETHRA": {"-": ["11"], "N": ["11"]},
    "AETIOLOGIA": {"-": ["17"]},
    "AETNA": {"1N": ["11"], "2N": ["11"], "N": ["11"]},
    "AETNAEVS": {"N": ["21"], "A": ["21"]},
    "AETNENSES": {"N": ["13"]},
    "AETNENSIS": {"N": ["24"], "A": ["24"]},
    "AETOLIA": {"N": ["11"]},
    "AETOLICVS": {"N": ["21"]},
    "AETOLVS": {"N": ["21"], "A": ["21"]},
    "AEVVM": {"-": ["12"]},
    "AEVVS": {"-": ["12"]},
    "AFFABILIS": {"-": ["24"]},
    "AFFABRE": {"-": ["60"]},
    "AFFATIM": {"-": ["60"]},
    "AFFATVS": {"-": ["14"]},
    "AFFECTATIO": {"-": ["13"]},
    "AFFECTATOR": {"-": ["13"]},
    "AFFECTIO": {"-": ["13"]},
    "AFFECTO": {"-": ["51"]},
    "AFFECTVS": {"1": ["14"], "2": ["21"]},
    "AFFERO": {"-": ["56"]},
    "AFFICIO": {"-": ["55"]},
    "AFFIGO": {"-": ["53"]},
    "AFFINGO": {"-": ["53"]},
    "AFFINIS": {"1": ["13"], "2": ["24"]},
    "AFFINITAS": {"-": ["13"]},
    "AFFIRMATIO": {"-": ["13"]},
    "AFFIRMO": {"-": ["51"]},
    "AFFLATVS": {"-": ["14"]},
    "AFFLICTATIO": {"-": ["13"]},
    "AFFLICTO": {"-": ["51"]},
    "AFFLICTVS": {"1": ["14"], "2": ["21"]},
    "AFFLIGO": {"-": ["53"]},
    "AFFLO": {"-": ["51"]},
    "AFFLVENS": {"-": ["25"]},
    "AFFLVENTER": {"-": ["60"]},
    "AFFLVENTIA": {"-": ["11"]},
    "AFFLVO": {"-": ["53"]},
    "AFFOR": {"-": ["5J"]},
    "AFFRICO": {"-": ["51"]},
    "AFFVLGEO": {"-": ["52"]},
    "AFFVNDO": {"2": ["53"]},
    "AFRICA": {"N": ["11"]},
    "AFRICANVS": {"C": ["12"], "N": ["21"], "A": ["21"]},
    "AFRICVS": {"A": ["21"], "N": ["12"]},
    "AGASO": {"-": ["13"]},
    "AGATHINVS": {"N": ["12"]},
    "AGATHOCLES": {"N": ["17"]},
    "AGE": {"-": ["90"]},
    "AGEDVM": {"-": ["90"]},
    "AGELLVS": {"-": ["12"]},
    "AGENS": {"1": ["13"], "2": ["25"]},
    "AGER": {"-": ["12"], "N": ["12"]},
    "AGGER": {"-": ["13"]},
    "AGGERATIO": {"-": ["13"]},
    "AGGERO": {"1": ["51"], "2": ["53"]},
    "AGGESTVS": {"-": ["14"]},
    "AGGLOMERO": {"-": ["51"]},
    "AGGLVTINO": {"-": ["51"]},
    "AGGRAVO": {"-": ["51"]},
    "AGGREDIOR": {"-": ["5N"]},
    "AGGREGO": {"-": ["51"]},
    "AGILIS": {"-": ["24"]},
    "AGILITAS": {"-": ["13"]},
    "AGITABILIS": {"-": ["24"]},
    "AGITATIO": {"-": ["13"]},
    "AGITATOR": {"-": ["13"]},
    "AGITATVS": {"1": ["14"], "2": ["21"]},
    "AGITE": {"-": ["90"]},
    "AGITO": {"-": ["51"]},
    "AGMEN": {"-": ["13"]},
    "AGNA": {"-": ["11"]},
    "AGNASCOR": {"-": ["5L"]},
    "AGNATVS": {"-": ["12"]},
    "AGNINA": {"-": ["11"]},
    "AGNINVS": {"-": ["21"]},
    "AGNOSCO": {"-": ["53"]},
    "AGNVS": {"-": ["12"]},
    "AGO": {"-": ["53"]},
    "AGON": {"-": ["17"]},
    "AGRAMMATVS": {"-": ["21"]},
    "AGRARII": {"-": ["12"]},
    "AGRARIVS": {"-": ["21"]},
    "AGRESTIS": {"1": ["13"], "2": ["24"], "N": ["13"]},
    "AGRICOLA": {"-": ["11"], "N": ["11"]},
    "AGRIGENTINI": {"N": ["12"]},
    "AGRIGENTINVS": {"N": ["21"], "A": ["21"]},
    "AGRIGENTVM": {"N": ["12"]},
    "AGRIPPA": {"N": ["11"]},
    "AGRO": {"-": ["51"]},
    "AGYRINENSES": {"N": ["13"]},
    "AGYRINENSIS": {"N": ["24"], "A": ["24"]},
    "AGYRIVM": {"N": ["12"]},
    "AH": {"-": ["90"]},
    "AIAX": {"N": ["13"]},
    "AIO": {"-": ["56"]},
    "ALA": {"-": ["11"]},
    "ALACER": {"-": ["23"]},
    "ALACRITAS": {"-": ["13"]},
    "ALACRITER": {"-": ["60"]},
    "ALAPA": {"-": ["11"]},
    "ALARII": {"-": ["12"]},
    "ALARIVS": {"-": ["21"]},
    "ALATVS": {"-": ["21"]},
    "ALBA": {"1N": ["11"], "2N": ["11"], "N": ["11"]},
    "ALBANI": {"N": ["12"]},
    "ALBANVS": {"N": ["21"], "A": ["21"]},
    "ALBARIVS": {"-": ["21"]},
    "ALBATVS": {"-": ["21"]},
    "ALBEO": {"-": ["52"]},
    "ALBESCO": {"-": ["53"]},
    "ALBICERIS": {"-": ["24"]},
    "ALBICO": {"-": ["51"]},
    "ALBIDVS": {"-": ["21"]},
    "ALBINVS": {"N": ["12"]},
    "ALBVLA": {"N": ["11"]},
    "ALBVLVS": {"-": ["21"]},
    "ALBVM": {"-": ["12"]},
    "ALBVS": {"-": ["21"], "A": ["21"]},
    "ALCES": {"-": ["13"]},
    "ALCYON": {"-": ["17"]},
    "ALCYONEVM": {"-": ["12"]},
    "ALEA": {"-": ["11"]},
    "ALEATOR": {"-": ["13"]},
    "ALEBRIA": {"-": ["13"]},
    "ALEO": {"-": ["13"]},
    "ALES": {"1": ["13"], "2": ["25"], "N": ["13"]},
    "ALESCO": {"-": ["53"]},
    "ALEXANDER": {"N": ["17"]},
    "ALEXANDRIA": {"N": ["11"]},
    "ALGA": {"-": ["11"]},
    "ALGEO": {"-": ["52"]},
    "ALGIDVS": {"-": ["21"], "N": ["12"], "A": ["21"]},
    "ALGOR": {"-": ["13"]},
    "ALGVS": {"-": ["14"]},
    "ALIA": {"-": ["60"]},
    "ALIAS": {"-": ["60"]},
    "ALIBI": {"-": ["60"]},
    "ALICA": {"-": ["11"]},
    "ALICVBI": {"-": ["60"]},
    "ALICVNDE": {"-": ["60"]},
    "ALIENATIO": {"-": ["13"]},
    "ALIENIGENA": {"-": ["11"]},
    "ALIENIGENVS": {"-": ["21"]},
    "ALIENO": {"-": ["51"]},
    "ALIENVM": {"C": ["12"], "-": ["12"]},
    "ALIENVS": {"1": ["12"], "2": ["21"], "N": ["12"]},
    "ALIGER": {"-": ["21"]},
    "ALIMENTVM": {"-": ["12"]},
    "ALIMONIVM": {"-": ["12"]},
    "ALIO": {"-": ["60"]},
    "ALIOQVI": {"-": ["60"]},
    "ALIPES": {"-": ["25"]},
    "ALIPILVS": {"-": ["12"]},
    "ALIPTES": {"-": ["17"]},
    "ALIQVA": {"-": ["60"]},
    "ALIQVAM": {"-": ["60"]},
    "ALIQVAMDIV": {"-": ["60"]},
    "ALIQVAMMVLTVS": {"-": ["21"]},
    "ALIQVANDO": {"-": ["60"]},
    "ALIQVANTO": {"-": ["60"]},
    "ALIQVANTVLVM": {"1": ["48"], "2": ["60"]},
    "ALIQVANTVM": {"1": ["48"], "2": ["60"]},
    "ALIQVANTVS": {"-": ["48"]},
    "ALIQVATENVS": {"-": ["60"]},
    "ALIQVIS": {"-": ["48"]},
    "ALIQVO": {"-": ["60"]},
    "ALIQVOT": {"-": ["48"]},
    "ALIQVOTFARIAM": {"-": ["60"]},
    "ALIQVOTIENS": {"-": ["60"]},
    "ALIS": {"-": ["48"]},
    "ALITER": {"-": ["60"]},
    "ALIVBI": {"-": ["60"]},
    "ALIVNDE": {"-": ["60"]},
    "ALIVS": {"-": ["48"]},
    "ALIVSMODI": {"-": ["60"]},
    "ALLABOR": {"2": ["5L"]},
    "ALLABORO": {"-": ["51"]},
    "ALLAPSVS": {"-": ["14"]},
    "ALLAVDABILIS": {"-": ["24"]},
    "ALLEC": {"-": ["13"]},
    "ALLECTO": {"-": ["51"], "N": ["11"]},
    "ALLEGATIO": {"-": ["13"]},
    "ALLEGO": {"1": ["51"], "2": ["53"]},
    "ALLEVO": {"1": ["51"], "2": ["51"]},
    "ALLEX": {"-": ["13"]},
    "ALLICEFACIO": {"-": ["55"]},
    "ALLICIO": {"-": ["55"]},
    "ALLIDO": {"-": ["53"]},
    "ALLIGATIO": {"-": ["13"]},
    "ALLIGO": {"2": ["51"]},
    "ALLINO": {"-": ["53"]},
    "ALLIVM": {"-": ["12"]},
    "ALLOBROGES": {"N": ["17"]},
    "ALLOBROX": {"N": ["27"], "A": ["25"]},
    "ALLOCVTIO": {"-": ["13"]},
    "ALLOQVIVM": {"-": ["12"]},
    "ALLOQVOR": {"-": ["5L"]},
    "ALLVCEO": {"-": ["52"]},
    "ALLVDO": {"-": ["53"]},
    "ALLVO": {"-": ["53"]},
    "ALMVS": {"-": ["21"]},
    "ALNEVS": {"-": ["21"]},
    "ALNVS": {"-": ["12"]},
    "ALO": {"-": ["53"]},
    "ALOE": {"-": ["17"]},
    "ALPES": {"N": ["13"]},
    "ALPHA": {"-": ["16"]},
    "ALSIVS": {"-": ["21"]},
    "ALTAR": {"-": ["13"]},
    "ALTE": {"-": ["60"]},
    "ALTER": {"-": ["48"]},
    "ALTERAS": {"-": ["60"]},
    "ALTERCATIO": {"-": ["13"]},
    "ALTERCOR": {"-": ["5J"]},
    "ALTERNIS": {"-": ["60"]},
    "ALTERNO": {"-": ["51"]},
    "ALTERNVS": {"-": ["21"]},
    "ALTERO": {"-": ["51"]},
    "ALTERVTER": {"-": ["48"]},
    "ALTILIS": {"1": ["13"], "2": ["24"]},
    "ALTISONVS": {"-": ["21"]},
    "ALTITONANS": {"-": ["25"]},
    "ALTITVDO": {"-": ["13"]},
    "ALTIVOLANS": {"-": ["25"]},
    "A": {"-": ["13"]},
    "ALTRINSECVS": {"-": ["60"]},
    "ALTRIX": {"-": ["13"]},
    "ALTVM": {"-": ["12"]},
    "ALTVS": {"-": ["21"]},
    "ALVARIVM": {"-": ["12"]},
    "ALVCINATIO": {"-": ["13"]},
    "ALVEATVS": {"-": ["21"]},
    "ALVEOLATVS": {"-": ["21"]},
    "ALVEOLVS": {"-": ["12"]},
    "ALVEVS": {"-": ["12"]},
    "ALVMEN": {"-": ["13"]},
    "ALVMINOSVS": {"-": ["21"]},
    "ALVMNA": {"-": ["11"]},
    "ALVMNVS": {"1": ["12"], "2": ["21"]},
    "ALVNTINI": {"N": ["12"]},
    "ALVNTINVS": {"N": ["21"], "A": ["21"]},
    "ALVNTIVM": {"N": ["12"]},
    "ALVTA": {"1": ["11"], "2": ["11"]},
    "ALVVS": {"-": ["12"]},
    "AM": {"-": ["70"]},
    "AMABILIS": {"-": ["24"]},
    "AMABILITER": {"-": ["60"]},
    "AMANDO": {"2": ["51"], "-": ["51"]},
    "AMANS": {"1": ["13"], "2": ["25"]},
    "AMANTER": {"-": ["60"]},
    "AMARA": {"-": ["12"]},
    "AMARACINVM": {"-": ["12"]},
    "AMARACVS": {"-": ["12"]},
    "AMARANTVS": {"-": ["12"]},
    "AMARE": {"-": ["60"]},
    "AMARITAS": {"-": ["13"]},
    "AMARITIES": {"-": ["15"]},
    "AMARITVDO": {"-": ["13"]},
    "AMAROR": {"-": ["13"]},
    "AMARVS": {"-": ["21"]},
    "AMATOR": {"-": ["13"]},
    "AMATORIVM": {"-": ["12"]},
    "AMATORIVS": {"-": ["21"]},
    "AMBACTVS": {"-": ["12"]},
    "AMBAGES": {"-": ["13"]},
    "AMBARVALIS": {"-": ["24"]},
    "AMBEDO": {"2": ["56"]},
    "AMBEO": {"1": ["56"]},
    "AMBIGO": {"-": ["53"]},
    "AMBIGVE": {"-": ["60"]},
    "AMBIGVITAS": {"-": ["13"]},
    "AMBIGVVM": {"-": ["12"]},
    "AMBIGVVS": {"-": ["21"]},
    "AMBIO": {"-": ["56"]},
    "AMBITIO": {"-": ["13"], "N": ["13"]},
    "AMBITIOSE": {"-": ["60"]},
    "AMBITIOSVS": {"-": ["21"]},
    "AMBITVS": {"-": ["14"]},
    "AMBO": {"2": ["31"]},
    "AMBROSIA": {"-": ["11"]},
    "AMBROSIVS": {"-": ["21"], "N": ["12"]},
    "AMBVBAIA": {"-": ["11"]},
    "AMBVLATIO": {"-": ["13"]},
    "AMBVLATOR": {"-": ["13"]},
    "AMBVLATORIVS": {"-": ["21"]},
    "AMBVLATRIX": {"-": ["13"]},
    "AMBVLO": {"-": ["51"]},
    "AMBVRO": {"-": ["53"]},
    "AMELLVS": {"-": ["12"]},
    "AMENS": {"-": ["25"]},
    "AMENTIA": {"-": ["11"]},
    "AMENTVM": {"-": ["12"]},
    "AMERIOLA": {"N": ["11"]},
    "AMES": {"-": ["13"]},
    "AMESTRATINI": {"N": ["12"]},
    "AMESTRATINVS": {"N": ["21"], "A": ["21"]},
    "AMESTRATVS": {"N": ["12"]},
    "AMETHYSTINA": {"-": ["12"]},
    "AMETHYSTVS": {"-": ["12"]},
    "AMICA": {"-": ["11"]},
    "AMICE": {"-": ["60"]},
    "AMICIO": {"-": ["54"]},
    "AMICITIA": {"-": ["11"]},
    "AMICITIES": {"-": ["15"]},
    "AMICTVS": {"-": ["14"]},
    "AMICVLVM": {"-": ["12"]},
    "AMICVLVS": {"-": ["12"]},
    "AMICVS": {"1": ["12"], "2": ["21"]},
    "AMISSIO": {"-": ["13"]},
    "AMITA": {"-": ["11"]},
    "AMITTO": {"-": ["53"]},
    "AMMONIACVS": {"-": ["21"]},
    "AMNIS": {"-": ["13"]},
    "AMO": {"-": ["51"]},
    "AMOENITAS": {"-": ["13"]},
    "AMOENVS": {"-": ["21"]},
    "AMOLIOR": {"-": ["5M"]},
    "AMOMON": {"-": ["17"]},
    "AMOMVM": {"-": ["12"]},
    "AMOR": {"-": ["13"], "N": ["13"]},
    "AMOVEO": {"-": ["52"]},
    "AMPHERES": {"-": ["27"]},
    "AMPHILOCHI": {"N": ["12"]},
    "AMPHIO": {"-": ["17"]},
    "AMPHIPOLIS": {"N": ["17"]},
    "AMPHIPROSTYLOS": {"-": ["27"]},
    "AMPHITHALAMVS": {"-": ["17"]},
    "AMPHITHEATRVM": {"-": ["12"]},
    "AMPHORA": {"-": ["11"]},
    "AMPLA": {"-": ["11"]},
    "AMPLE": {"-": ["60"]},
    "AMPLECTOR": {"-": ["5L"]},
    "AMPLEXOR": {"-": ["5J"]},
    "AMPLEXVS": {"-": ["14"]},
    "AMPLIFICATIO": {"-": ["13"]},
    "AMPLIFICATOR": {"-": ["13"]},
    "AMPLIFICE": {"-": ["60"]},
    "AMPLIFICO": {"-": ["51"]},
    "AMPLIO": {"-": ["51"]},
    "AMPLITVDO": {"-": ["13"]},
    "AMPLIVS": {"-": ["13"]},
    "AMPLVS": {"-": ["21"]},
    "AMPVLLOR": {"-": ["5J"]},
    "AMPVTO": {"-": ["51"]},
    "AMVLIVS": {"N": ["12"]},
    "AMVRCA": {"-": ["11"]},
    "AMVRCARIVS": {"-": ["21"]},
    "AMVSOS": {"-": ["27"]},
    "AMVSSIVM": {"-": ["12"]},
    "AMYGDALVM": {"-": ["12"]},
    "AMYLVM": {"-": ["12"]},
    "AMYSTIS": {"-": ["17"]},
    "AN": {"-": ["67"]},
    "ANABATHRVM": {"-": ["12"]},
    "ANADEMA": {"-": ["17"]},
    "ANALECTA": {"-": ["11"]},
    "ANALEMMA": {"-": ["17"]},
    "ANALEMPTRIS": {"-": ["13"]},
    "ANALOGIA": {"-": ["17"]},
    "ANAPHORICVS": {"-": ["21"]},
    "ANATONVS": {"-": ["21"]},
    "ANCEPS": {"1": ["13"], "2": ["25"]},
    "ANCHISES": {"N": ["17"]},
    "ANCILE": {"-": ["16"]},
    "ANCILLA": {"-": ["11"]},
    "ANCILLARIOLVS": {"-": ["12"]},
    "ANCILLARIS": {"-": ["24"]},
    "ANCILLVLA": {"-": ["11"]},
    "ANCISVS": {"-": ["21"]},
    "ANCON": {"-": ["17"], "N": ["17"]},
    "ANCORA": {"-": ["11"]},
    "ANCVS": {"N": ["12"]},
    "ANDRO": {"N": ["17"]},
    "ANDRON": {"-": ["17"], "N": ["17"]},
    "ANDRONITIS": {"-": ["17"]},
    "ANDROSTHENES": {"N": ["13"]},
    "ANELLVS": {"-": ["12"]},
    "ANETHVM": {"-": ["12"]},
    "ANFRACTVS": {"1": ["14"], "2": ["21"]},
    "ANGELLVS": {"-": ["12"]},
    "ANGELVS": {"-": ["12"]},
    "ANGINA": {"-": ["11"]},
    "ANGIPORTVM": {"-": ["12"]},
    "ANGIPORTVS": {"-": ["14"]},
    "ANGO": {"-": ["53"]},
    "ANGOR": {"-": ["13"]},
    "ANGVBATAE": {"-": ["11"]},
    "ANGVICOMVS": {"-": ["21"]},
    "ANGVIFER": {"-": ["21"]},
    "ANGVILLA": {"-": ["11"]},
    "ANGVIMANVS": {"-": ["26"]},
    "ANGVINEVS": {"-": ["21"]},
    "ANGVINVS": {"-": ["21"]},
    "ANGVIS": {"-": ["13"], "N": ["13"]},
    "ANGVLARIS": {"-": ["24"]},
    "ANGVLVS": {"-": ["12"]},
    "ANGVSTE": {"-": ["60"]},
    "ANGVSTIA": {"-": ["11"]},
    "ANGVSTO": {"-": ["51"]},
    "ANGVSTVM": {"-": ["12"]},
    "ANGVSTVS": {"-": ["21"]},
    "ANHELITVS": {"-": ["14"]},
    "ANHELO": {"-": ["51"]},
    "ANHELVS": {"-": ["21"]},
    "ANIATROLOGETVS": {"-": ["21"]},
    "ANICVLA": {"-": ["11"]},
    "ANILIS": {"-": ["24"]},
    "ANILITAS": {"-": ["13"]},
    "ANILITER": {"-": ["60"]},
    "ANIMA": {"-": ["11"]},
    "ANIMADVERSIO": {"-": ["13"]},
    "ANIMADVERTO": {"-": ["53"]},
    "ANIMAL": {"-": ["13"]},
    "ANIMALIS": {"-": ["24"]},
    "ANIMANS": {"1": ["13"], "2": ["25"]},
    "ANIMATIO": {"-": ["13"]},
    "ANIMATVS": {"-": ["21"]},
    "ANIMO": {"-": ["51"]},
    "ANIMOSE": {"-": ["6&"]},
    "ANIMOSVS": {"-": ["21"]},
    "ANIMVS": {"-": ["12"]},
    "ANIO": {"N": ["16"]},
    "ANISOCYCLI": {"-": ["17"]},
    "ANISVM": {"-": ["12"]},
    "ANNAEA": {"N": ["11"]},
    "ANNAEVS": {"N": ["12"]},
    "ANNALIS": {"1": ["13"], "2": ["24"], "N": ["13"]},
    "ANNATO": {"-": ["51"]},
    "ANNE": {"-": ["67"]},
    "ANNECTO": {"-": ["53"]},
    "ANNIA": {"N": ["11"]},
    "ANNIANVS": {"A": ["21"]},
    "ANNICVLVS": {"-": ["21"]},
    "ANNITOR": {"2": ["5L"]},
    "ANNIVERSARIVS": {"-": ["21"]},
    "ANNIVS": {"N": ["12"]},
    "ANNO": {"1": ["51"], "2": ["51"]},
    "ANNONA": {"-": ["11"]},
    "ANNOSVS": {"-": ["21"]},
    "ANNOTINVS": {"-": ["21"]},
    "ANNOTO": {"-": ["51"]},
    "ANNVMERO": {"1": ["51"]},
    "ANNVNTIO": {"-": ["51"]},
    "ANNVO": {"-": ["53"]},
    "ANNVS": {"-": ["12"], "N": ["12"]},
    "ANNVVM": {"-": ["12"]},
    "ANNVVS": {"-": ["21"]},
    "ANQVIRO": {"-": ["53"]},
    "ANSA": {"-": ["11"]},
    "ANSER": {"-": ["13"], "N": ["12"]},
    "ANTAE": {"-": ["11"]},
    "ANTARIVS": {"-": ["21"]},
    "ANTE": {"1": ["60"], "2": ["70"]},
    "ANTEA": {"-": ["60"]},
    "ANTEACTVS": {"-": ["21"]},
    "ANTEAGO": {"-": ["53"]},
    "ANTEAQVAM": {"-": ["82"]},
    "ANTECAPIO": {"2": ["55"]},
    "ANTECEDENS": {"-": ["25"]},
    "ANTECEDENTIA": {"-": ["13"]},
    "ANTECEDO": {"1": ["53"]},
    "ANTECELLO": {"-": ["53"]},
    "ANTECESSOR": {"-": ["13"]},
    "ANTECHON": {"-": ["57"]},
    "ANTECVRRO": {"-": ["53"]},
    "ANTECVRSOR": {"-": ["13"]},
    "ANTEEO": {"1": ["56"]},
    "ANTEFERO": {"-": ["56"]},
    "ANTEFIXVM": {"-": ["12"]},
    "ANTEFIXVS": {"-": ["21"]},
    "ANTEHABEO": {"-": ["52"]},
    "ANTEHAC": {"-": ["60"]},
    "ANTELVCANVS": {"-": ["21"]},
    "ANTEMERIDIANVS": {"-": ["21"]},
    "ANTEMNATES": {"N": ["13"]},
    "ANTENNA": {"-": ["11"]},
    "ANTENOR": {"N": ["17"]},
    "ANTEPAGMENTVM": {"-": ["12"]},
    "ANTEPONO": {"-": ["53"]},
    "ANTEQVAM": {"-": ["82"], "1": ["6-"]},
    "ANTERIS": {"-": ["17"]},
    "ANTES": {"-": ["13"]},
    "ANTESIGNANVS": {"1": ["12"], "2": ["21"]},
    "ANTESTO": {"-": ["51"]},
    "ANTESTOR": {"-": ["5J"]},
    "ANTEVENIO": {"-": ["54"]},
    "ANTEVERTO": {"-": ["53"]},
    "ANTHROPINVS": {"N": ["12"]},
    "ANTIATES": {"N": ["13"]},
    "ANTIBASIS": {"-": ["17"]},
    "ANTIBOREVM": {"-": ["12"]},
    "ANTICIPO": {"-": ["51"]},
    "ANTIDOTVM": {"-": ["12"]},
    "ANTIGONVS": {"N": ["12"]},
    "ANTIOCHENSES": {"N": ["13"]},
    "ANTIOCHIA": {"N": ["11"]},
    "ANTIOCHVS": {"N": ["12"]},
    "ANTIPODES": {"-": ["17"]},
    "ANTIQVARIA": {"-": ["11"]},
    "ANTIQVARIVS": {"1": ["12"], "2": ["21"]},
    "ANTIQVE": {"-": ["60"]},
    "ANTIQVI": {"-": ["12"]},
    "ANTIQVITAS": {"-": ["13"]},
    "ANTIQVITVS": {"-": ["60"]},
    "ANTIQVVM": {"-": ["12"]},
    "ANTIQVVS": {"-": ["21"]},
    "ANTISTES": {"-": ["13"]},
    "ANTISTITA": {"-": ["11"]},
    "ANTISTIVS": {"N": ["12"]},
    "ANTITHETON": {"-": ["17"]},
    "ANTIVM": {"N": ["12"]},
    "ANTONIANI": {"N": ["12"]},
    "ANTONIANVS": {"N": ["21"], "A": ["21"]},
    "ANTONIVS": {"N": ["12"], "A": ["21"]},
    "ANTRVM": {"-": ["12"]},
    "ANVLARIVS": {"-": ["21"]},
    "ANVLVS": {"-": ["12"]},
    "ANVS": {"1": ["12"], "2": ["14"]},
    "ANXIE": {"-": ["60"]},
    "ANXIETAS": {"-": ["13"]},
    "ANXIVS": {"-": ["21"]},
    "AONIA": {"N": ["11"]},
    "APELLES": {"N": ["13"]},
    "APER": {"-": ["12"], "N": ["12"]},
    "APERIO": {"-": ["54"]},
    "APERTE": {"-": ["60"]},
    "APERTVM": {"-": ["12"]},
    "APERTVRA": {"-": ["11"]},
    "APERTVS": {"-": ["21"]},
    "APEX": {"-": ["13"]},
    "APIACVS": {"-": ["21"]},
    "APICIVS": {"-": ["21"], "N": ["12"]},
    "APIOLAE": {"N": ["11"]},
    "APIS": {"-": ["13"], "N": ["17"]},
    "APISCOR": {"-": ["5L"]},
    "APIVM": {"-": ["12"]},
    "APLVSTRE": {"-": ["16"]},
    "APOLLO": {"N": ["13"]},
    "APOLLODORVS": {"N": ["12"]},
    "APOLLONIA": {"N": ["11"]},
    "APOLLONIATAE": {"N": ["17"]},
    "APOLLONIENSIS": {"N": ["24"], "A": ["24"]},
    "APOLLONIVS": {"N": ["12"]},
    "APOLOGO": {"-": ["51"]},
    "APOPHYSIS": {"-": ["17"]},
    "APOTHECA": {"-": ["11"]},
    "APOTHESIS": {"-": ["17"]},
    "APPAGINECVLI": {"-": ["12"]},
    "APPARATIO": {"-": ["13"]},
    "APPARATVS": {"1": ["14"], "2": ["21"]},
    "APPAREO": {"-": ["52"]},
    "APPARIO": {"-": ["55"]},
    "APPARITIO": {"-": ["13"]},
    "APPARITOR": {"-": ["13"]},
    "APPARO": {"2": ["51"]},
    "APPELLATIO": {"-": ["13"]},
    "APPELLATOR": {"-": ["13"]},
    "APPELLITO": {"-": ["51"]},
    "APPELLO": {"1": ["51"], "2": ["53"]},
    "APPENDEO": {"-": ["52"]},
    "APPENDIX": {"-": ["13"]},
    "APPENDO": {"-": ["53"]},
    "APPERIOR": {"-": ["5M"]},
    "APPETENS": {"-": ["25"]},
    "APPETITIO": {"-": ["13"]},
    "APPETITVS": {"-": ["14"]},
    "APPETO": {"-": ["53"]},
    "APPIVS": {"N": ["12"], "A": ["21"]},
    "APPLAVDO": {"-": ["53"]},
    "APPLICATIO": {"-": ["13"]},
    "APPLICO": {"-": ["51"]},
    "APPLORO": {"-": ["51"]},
    "APPONO": {"-": ["53"]},
    "APPORTATIO": {"-": ["13"]},
    "APPORTO": {"-": ["51"]},
    "APPOSITVS": {"-": ["21"]},
    "APPRECOR": {"-": ["5J"]},
    "APPREHENDO": {"-": ["53"]},
    "APPREHENSIO": {"-": ["13"]},
    "APPRIME": {"-": ["60"]},
    "APPRIMO": {"-": ["53"]},
    "APPRIMVS": {"-": ["21"]},
    "APPROBATIO": {"-": ["13"]},
    "APPROBO": {"-": ["51"]},
    "APPROPERO": {"-": ["51"]},
    "APPROPINQVO": {"-": ["51"]},
    "APPVGNO": {"-": ["51"]},
    "APPVLSVS": {"-": ["14"]},
    "APRICVM": {"-": ["12"]},
    "APRICVS": {"-": ["21"]},
    "APRONIANVS": {"N": ["21"], "A": ["21"]},
    "APRONIVS": {"N": ["12"]},
    "APTE": {"-": ["60"]},
    "APTO": {"-": ["51"]},
    "APTVS": {"-": ["21"]},
    "APVD": {"-": ["70"]},
    "APYRENVM": {"-": ["12"]},
    "AQVA": {"-": ["11"]},
    "AQVALICVLVS": {"-": ["12"]},
    "AQVALIS": {"2": ["24"], "1": ["24"]},
    "AQVARIVM": {"-": ["12"]},
    "AQVARIVS": {"1": ["12"], "2": ["21"], "N": ["12"]},
    "AQVATICVS": {"-": ["21"]},
    "AQVATIO": {"-": ["13"]},
    "AQVATOR": {"-": ["13"]},
    "AQVATVS": {"-": ["21"]},
    "AQVIFOLIVS": {"-": ["21"]},
    "AQVILA": {"-": ["11"], "N": ["11"]},
    "AQVILIFER": {"-": ["12"]},
    "AQVILIVS": {"N": ["12"], "A": ["21"]},
    "AQVILO": {"-": ["13"], "N": ["13"]},
    "AQVOR": {"-": ["5J"]},
    "AQVOSVS": {"-": ["21"]},
    "AQVVLA": {"-": ["11"]},
    "ARA": {"-": ["11"], "N": ["11"]},
    "ARACHNE": {"-": ["17"], "N": ["11"]},
    "ARAEOSTYLOS": {"-": ["27"]},
    "ARANEA": {"-": ["11"]},
    "ARANEOSVS": {"-": ["21"]},
    "ARANEVS": {"1": ["12"], "2": ["21"]},
    "ARATIO": {"-": ["13"]},
    "ARATOR": {"-": ["13"]},
    "ARATRVM": {"-": ["12"]},
    "ARBITER": {"-": ["12"]},
    "ARBITRA": {"-": ["11"]},
    "ARBITRATVS": {"-": ["14"]},
    "ARBITRIVM": {"-": ["12"]},
    "ARBITROR": {"-": ["5J"]},
    "ARBOR": {"-": ["13"]},
    "ARBORARIVS": {"-": ["21"]},
    "ARBOREVS": {"-": ["21"]},
    "ARBVSCVLA": {"-": ["11"], "N": ["11"]},
    "ARBVSTIVVS": {"-": ["21"]},
    "ARBVSTVM": {"-": ["12"]},
    "ARBVTEVS": {"-": ["21"]},
    "ARBVTVM": {"-": ["12"]},
    "ARBVTVS": {"-": ["12"]},
    "ARCA": {"-": ["11"]},
    "ARCADES": {"N": ["13"]},
    "ARCADIA": {"N": ["11"]},
    "ARCADICVS": {"N": ["21"], "A": ["21"]},
    "ARCANO": {"-": ["60"]},
    "ARCANVM": {"-": ["12"]},
    "ARCANVS": {"-": ["21"]},
    "ARCATVS": {"1": ["12"], "2": ["21"]},
    "ARCEO": {"-": ["52"]},
    "ARCESSITVS": {"1": ["14"], "2": ["21"]},
    "ARCESSO": {"-": ["53"]},
    "ARCHAGATHVS": {"N": ["12"]},
    "ARCHETYPVS": {"-": ["21"]},
    "ARCHIMAGIRVS": {"-": ["12"]},
    "ARCHIMEDES": {"N": ["17"]},
    "ARCHIPIRATA": {"-": ["11"]},
    "ARCHITECTO": {"-": ["51"]},
    "ARCHITECTONICVS": {"-": ["21"]},
    "ARCHITECTOR": {"1": ["13"], "2": ["5J"], "-": ["5J"]},
    "ARCHITECTVRA": {"-": ["11"]},
    "ARCHITECTVS": {"-": ["12"]},
    "ARCHONIDAS": {"N": ["17"]},
    "ARCITENENS": {"1": ["13"], "2": ["25"]},
    "ARCTOS": {"-": ["17"]},
    "ARCTOVS": {"-": ["21"], "A": ["21"]},
    "ARCVLA": {"-": ["11"]},
    "ARCVS": {"-": ["14"], "N": ["12"]},
    "ARDEA": {"-": ["11"], "N": ["11"]},
    "ARDENS": {"-": ["25"], "-": ["22"]},
    "ARDENTER": {"-": ["60"]},
    "ARDEO": {"-": ["52"]},
    "ARDESCO": {"-": ["53"]},
    "ARDOR": {"-": ["13"]},
    "ARDVVM": {"-": ["12"]},
    "ARDVVS": {"-": ["21"]},
    "AREA": {"-": ["11"]},
    "AREFACIO": {"-": ["55"]},
    "ARENA": {"-": ["11"]},
    "ARENARIVM": {"-": ["12"]},
    "ARENATIO": {"-": ["13"]},
    "ARENATVM": {"-": ["12"]},
    "ARENATVS": {"-": ["21"]},
    "ARENOSVS": {"-": ["21"]},
    "AREO": {"-": ["52"]},
    "ARESCO": {"-": ["53"]},
    "ARETALOGVS": {"-": ["12"]},
    "ARETHVSA": {"N": ["11"]},
    "AREVS": {"N": ["12"], "A": ["21"]},
    "ARGEI": {"N": ["12"]},
    "ARGENTARIA": {"-": ["11"]},
    "ARGENTARIVS": {"1": ["12"], "2": ["21"]},
    "ARGENTEVS": {"1": ["12"], "2": ["21"]},
    "ARGENTIFODINA": {"-": ["11"]},
    "ARGENTVM": {"-": ["12"]},
    "ARGILETVM": {"N": ["12"]},
    "ARGILLA": {"-": ["11"]},
    "ARGVMENTATIO": {"-": ["13"]},
    "ARGVMENTOR": {"-": ["5J"]},
    "ARGVMENTVM": {"-": ["12"]},
    "ARGVO": {"-": ["53"]},
    "ARGVTATIO": {"-": ["13"]},
    "ARGVTE": {"-": ["60"]},
    "ARGVTIA": {"-": ["11"]},
    "ARGVTO": {"-": ["51"]},
    "ARGVTVS": {"-": ["21"]},
    "ARICIA": {"N": ["11"]},
    "ARICINI": {"N": ["12"]},
    "ARICINVS": {"N": ["21"], "A": ["21"]},
    "ARIDA": {"-": ["11"]},
    "ARIDVLVS": {"-": ["21"]},
    "ARIDVM": {"-": ["12"]},
    "ARIDVS": {"-": ["21"]},
    "ARIES": {"-": ["13"], "N": ["13"]},
    "ARIETARIVS": {"-": ["21"]},
    "ARIETO": {"-": ["51"]},
    "ARIMINENSES": {"N": ["13"]},
    "ARIMINVM": {"N": ["12"]},
    "ARISTA": {"-": ["11"]},
    "ARISTAEVS": {"N": ["12"]},
    "ARISTEVS": {"N": ["12"]},
    "ARISTIDES": {"N": ["17"]},
    "ARISTODAMVS": {"N": ["12"]},
    "ARISTODEMVS": {"N": ["12"]},
    "ARISTOTELES": {"N": ["17"]},
    "ARISTVS": {"N": ["12"]},
    "ARITHMETICA": {"-": ["17"]},
    "ARMA": {"-": ["12"]},
    "ARMAMAXA": {"-": ["11"]},
    "ARMAMENTARIVM": {"-": ["12"]},
    "ARMAMENTVM": {"-": ["12"]},
    "ARMARIVM": {"-": ["12"]},
    "ARMATI": {"-": ["12"]},
    "ARMATVRA": {"-": ["11"]},
    "ARMATVS": {"1": ["14"], "2": ["21"]},
    "ARMENTARIVS": {"1": ["12"], "2": ["21"]},
    "ARMENTVM": {"-": ["12"]},
    "ARMIFER": {"-": ["21"]},
    "ARMIGER": {"1": ["12"], "2": ["21"]},
    "ARMILLA": {"-": ["11"]},
    "ARMIPOTENS": {"-": ["25"], "N": ["13"]},
    "ARMISONVS": {"-": ["21"]},
    "ARMO": {"-": ["51"]},
    "ARMVS": {"-": ["12"]},
    "ARO": {"-": ["51"]},
    "ARRECTARIVM": {"-": ["12"]},
    "ARRECTVS": {"-": ["21"]},
    "ARREPO": {"-": ["53"]},
    "ARRIDEO": {"-": ["52"]},
    "ARRIGO": {"-": ["53"]},
    "ARRIPIO": {"-": ["55"]},
    "ARRISOR": {"-": ["13"]},
    "ARRIVS": {"N": ["12"]},
    "ARRODO": {"-": ["53"]},
    "ARROGANS": {"A": ["25"], "-": ["25"]},
    "ARROGANTER": {"-": ["60"]},
    "ARROGANTIA": {"-": ["11"]},
    "ARROGO": {"-": ["51"]},
    "ARROSOR": {"-": ["13"]},
    "ARRVNS": {"N": ["13"]},
    "ARS": {"-": ["13"], "N": ["13"]},
    "ARSENICON": {"-": ["17"]},
    "ARSIA": {"N": ["11"]},
    "ARSINEVM": {"-": ["12"]},
    "ARSINOE": {"N": ["17"]},
    "ARTE": {"-": ["60"]},
    "ARTEMIDORVS": {"N": ["12"]},
    "ARTEMO": {"-": ["13"], "N": ["17"]},
    "ARTERIA": {"-": ["16"]},
    "ARTERIACE": {"-": ["17"]},
    "ARTICVLARIVS": {"-": ["21"]},
    "ARTICVLATIM": {"-": ["60"]},
    "ARTICVLO": {"-": ["51"]},
    "ARTICVLVS": {"-": ["12"]},
    "ARTIFEX": {"-": ["13"]},
    "ARTIFICIOSE": {"-": ["60"]},
    "ARTIFICIOSVS": {"-": ["21"]},
    "ARTIFICIVM": {"-": ["12"]},
    "ARTIO": {"-": ["54"]},
    "ARTO": {"-": ["51"]},
    "ARTOCREAS": {"-": ["17"]},
    "ARTOPTA": {"-": ["11"]},
    "ARTVM": {"-": ["12"]},
    "ARTVS": {"1": ["14"], "2": ["21"]},
    "ARVLA": {"-": ["11"]},
    "ARVNDINETVM": {"-": ["12"]},
    "ARVNDINOSVS": {"-": ["21"]},
    "ARVNDO": {"-": ["13"]},
    "ARVVM": {"-": ["12"]},
    "ARX": {"-": ["13"], "N": ["13"]},
    "AS": {"-": ["13"]},
    "ASCANIVS": {"N": ["12"]},
    "ASCENDO": {"-": ["53"]},
    "ASCENSIO": {"-": ["13"]},
    "ASCENSVS": {"-": ["14"]},
    "ASCIA": {"-": ["11"]},
    "ASCIO": {"1": ["51"], "2": ["54"]},
    "ASCISCO": {"-": ["53"]},
    "ASCITVS": {"1": ["14"], "2": ["21"]},
    "ASCRIBO": {"-": ["53"]},
    "ASCRIPTICIVS": {"-": ["21"]},
    "ASCRIPTIO": {"-": ["13"]},
    "ASELLA": {"-": ["11"]},
    "ASELLVS": {"-": ["12"], "N": ["12"]},
    "ASIA": {"N": ["11"]},
    "ASIATICVS": {"N": ["21"], "A": ["21"]},
    "ASILVS": {"-": ["12"]},
    "ASINARIVS": {"1": ["12"], "2": ["21"]},
    "ASININVS": {"-": ["21"]},
    "ASINIVS": {"N": ["12"]},
    "ASINVS": {"-": ["12"]},
    "ASOTVS": {"-": ["21"]},
    "ASPARAGVS": {"-": ["12"]},
    "ASPECTO": {"-": ["51"]},
    "ASPECTVS": {"-": ["14"]},
    "ASPENDIVS": {"N": ["21"], "A": ["21"]},
    "ASPENDVS": {"N": ["17"]},
    "ASPER": {"-": ["21"], "N": ["12"]},
    "ASPERA": {"-": ["12"]},
    "ASPERE": {"-": ["60"]},
    "ASPERGO": {"1": ["13"], "2": ["53"]},
    "ASPERITAS": {"-": ["13"]},
    "ASPERNATIO": {"-": ["13"]},
    "ASPERNOR": {"-": ["5J"]},
    "ASPERO": {"-": ["51"]},
    "ASPERVM": {"-": ["12"]},
    "ASPHALTITIS": {"-": ["27"]},
    "ASPICIO": {"-": ["55"]},
    "ASPIRO": {"-": ["51"]},
    "ASPIS": {"-": ["17"]},
    "ASPLENOS": {"-": ["17"]},
    "ASPORTATIO": {"-": ["13"]},
    "ASPORTO": {"-": ["51"]},
    "ASSARIVS": {"-": ["21"]},
    "ASSECLA": {"-": ["11"]},
    "ASSECTATOR": {"-": ["13"]},
    "ASSECTOR": {"2": ["5J"]},
    "ASSENSIO": {"-": ["13"]},
    "ASSENSOR": {"-": ["13"]},
    "ASSENSVS": {"-": ["14"]},
    "ASSENTATIO": {"-": ["13"]},
    "ASSENTATOR": {"-": ["13"]},
    "ASSENTIO": {"-": ["54"]},
    "ASSENTIOR": {"-": ["5M"]},
    "ASSENTOR": {"-": ["5J"]},
    "ASSEQVOR": {"-": ["5L"]},
    "ASSER": {"-": ["13"]},
    "ASSERCVLVM": {"-": ["12"]},
    "ASSERO": {"2": ["53"], "3": ["53"]},
    "ASSERTOR": {"-": ["13"]},
    "ASSERVO": {"-": ["51"]},
    "ASSESSOR": {"-": ["13"]},
    "ASSEVERATIO": {"-": ["13"]},
    "ASSEVERO": {"-": ["51"]},
    "ASSICCO": {"-": ["51"]},
    "ASSIDEO": {"-": ["52"]},
    "ASSIDO": {"-": ["53"]},
    "ASSIDVE": {"-": ["60"]},
    "ASSIDVITAS": {"-": ["13"]},
    "ASSIDVVS": {"-": ["21"]},
    "ASSIGNO": {"-": ["51"]},
    "ASSILIO": {"-": ["54"]},
    "ASSIMILIS": {"-": ["24"]},
    "ASSIMVLATIO": {"-": ["13"]},
    "ASSIMVLO": {"-": ["51"]},
    "ASSISTO": {"-": ["53"]},
    "ASSOLEO": {"-": ["5S"]},
    "ASSONO": {"1": ["51"]},
    "ASSORINI": {"N": ["12"]},
    "ASSORINVS": {"N": ["21"], "A": ["21"]},
    "ASSORVM": {"N": ["12"]},
    "ASSVEFACIO": {"-": ["55"]},
    "ASSVESCO": {"-": ["53"]},
    "ASSVETVDO": {"-": ["13"]},
    "ASSVETVS": {"-": ["21"]},
    "ASSVGO": {"-": ["53"]},
    "ASSVLA": {"-": ["11"]},
    "ASSVLTO": {"-": ["51"]},
    "ASSVLTVS": {"-": ["14"]},
    "ASSVM": {"1": ["56"], "2": ["12"]},
    "ASSVMO": {"-": ["53"]},
    "ASSVRGO": {"-": ["53"]},
    "ASSVS": {"-": ["21"]},
    "ASTO": {"-": ["51"]},
    "ASTRAGALVS": {"-": ["12"]},
    "ASTREPO": {"-": ["53"]},
    "ASTRICTE": {"-": ["6&"]},
    "ASTRICTVS": {"A": ["21"], "-": ["21"]},
    "ASTRINGO": {"-": ["53"]},
    "ASTROLOGIA": {"-": ["11"]},
    "ASTROLOGVS": {"-": ["12"]},
    "ASTRONOMIA": {"-": ["11"]},
    "ASTRVM": {"-": ["12"]},
    "ASTRVO": {"-": ["53"]},
    "ASTV": {"-": ["17"]},
    "ASTVPEO": {"-": ["52"]},
    "ASTVRCO": {"-": ["13"]},
    "ASTVS": {"-": ["14"]},
    "ASTVTE": {"-": ["60"]},
    "ASTVTIA": {"-": ["11"]},
    "ASTVTVS": {"-": ["21"]},
    "ASYLVM": {"-": ["12"]},
    "AT": {"1": ["60"], "2": ["81"]},
    "ATAT": {"-": ["90"]},
    "ATAVI": {"-": ["12"]},
    "ATAVVS": {"-": ["12"]},
    "ATER": {"-": ["21"]},
    "ATHENAE": {"N": ["11"]},
    "ATHENIENSES": {"N": ["13"]},
    "ATHENIENSIS": {"N": ["24"], "A": ["24"]},
    "ATHENIO": {"N": ["17"]},
    "ATHLETA": {"-": ["11"]},
    "ATIDIVS": {"N": ["12"]},
    "ATILIVS": {"N": ["12"]},
    "ATINIVS": {"A": ["21"], "N": ["12"]},
    "ATLANTES": {"-": ["17"]},
    "ATOMVS": {"-": ["12"]},
    "ATQVE": {"1": ["81"], "2": ["82"]},
    "ATQVI": {"-": ["81"]},
    "ATRAMENTVM": {"-": ["12"]},
    "ATRATVS": {"-": ["21"]},
    "ATRIENSIS": {"1": ["13"], "2": ["24"]},
    "ATRIVM": {"-": ["12"], "N": ["12"]},
    "ATROCITAS": {"-": ["13"]},
    "ATROCITER": {"-": ["60"]},
    "ATROX": {"-": ["25"]},
    "ATTAGEN": {"-": ["13"]},
    "ATTALICVS": {"N": ["21"], "A": ["21"]},
    "ATTALVS": {"N": ["12"]},
    "ATTAMEN": {"-": ["60"]},
    "ATTEGIA": {"-": ["11"]},
    "ATTEMPERO": {"-": ["51"]},
    "ATTENDO": {"-": ["53"]},
    "ATTENTE": {"-": ["6&"]},
    "ATTENTO": {"-": ["51"]},
    "ATTENTVS": {"1": ["14"], "2": ["21"]},
    "ATTENVATVS": {"-": ["21"]},
    "ATTENVO": {"-": ["51"]},
    "ATTERO": {"-": ["53"]},
    "ATTEXO": {"-": ["53"]},
    "ATTICA": {"N": ["11"]},
    "ATTICVS": {"C": ["12"], "N": ["12"], "A": ["21"]},
    "ATTINEO": {"-": ["52"]},
    "ATTINGO": {"-": ["53"]},
    "ATTIS": {"N": ["17"]},
    "ATTOLLO": {"-": ["53"]},
    "ATTONDEO": {"-": ["52"]},
    "ATTONITVS": {"1": ["14"], "2": ["21"]},
    "ATTONO": {"-": ["51"]},
    "ATTRACTIO": {"-": ["13"]},
    "ATTRACTVS": {"-": ["21"]},
    "ATTRAHO": {"-": ["53"]},
    "ATTRECTO": {"-": ["51"]},
    "ATTRIBVO": {"-": ["53"]},
    "ATTRITVS": {"1": ["14"], "2": ["21"]},
    "ATTVS": {"N": ["12"]},
    "AVARA": {"-": ["11"]},
    "AVARE": {"-": ["60"]},
    "AVARITER": {"-": ["60"]},
    "AVARITIA": {"-": ["11"], "N": ["11"]},
    "AVARITIES": {"-": ["15"]},
    "AVARVS": {"1": ["12"], "2": ["21"]},
    "AVCEPS": {"-": ["13"]},
    "AVCTE": {"-": ["6&"]},
    "AVCTIFICVS": {"-": ["21"]},
    "AVCTIO": {"-": ["13"]},
    "AVCTIONARIVS": {"-": ["21"]},
    "AVCTIONOR": {"-": ["5J"]},
    "AVCTITO": {"-": ["51"]},
    "AVCTO": {"-": ["51"]},
    "AVCTOR": {"-": ["13"]},
    "AVCTORAMENTVM": {"-": ["12"]},
    "AVCTORITAS": {"-": ["13"]},
    "AVCTORO": {"-": ["51"]},
    "AVCTVMNVS": {"1": ["12"]},
    "AVCTVS": {"1": ["14"], "2": ["21"]},
    "AVCVPIVM": {"-": ["12"]},
    "AVCVPO": {"-": ["51"]},
    "AVCVPOR": {"-": ["5J"]},
    "AVDACIA": {"-": ["11"]},
    "AVDACITER": {"-": ["60"]},
    "AVDAX": {"-": ["25"]},
    "AVDENS": {"-": ["25"]},
    "AVDENTER": {"-": ["6&"]},
    "AVDENTIA": {"-": ["11"]},
    "AVDEO": {"-": ["5S"]},
    "AVDIENS": {"-": ["25"]},
    "AVDIO": {"-": ["54"]},
    "AVDITIO": {"-": ["13"]},
    "AVDITOR": {"-": ["13"]},
    "AVDITORIVM": {"-": ["12"]},
    "AVDITVM": {"-": ["12"]},
    "AVDITVS": {"-": ["14"]},
    "AVE": {"-": ["90"]},
    "AVEHO": {"-": ["53"]},
    "AVELLO": {"-": ["53"]},
    "AVENA": {"-": ["11"]},
    "AVENTINVS": {"1N": ["12"], "2N": ["16"], "A": ["21"], "N": ["12"]},
    "AVEO": {"-": ["52"]},
    "AVERNVS": {"N": ["21"], "A": ["21"]},
    "AVERRO": {"-": ["53"]},
    "AVERRVNCO": {"-": ["51"]},
    "AVERSABILIS": {"-": ["24"]},
    "AVERSATIO": {"-": ["13"]},
    "AVERSIO": {"-": ["13"]},
    "AVERSOR": {"1": ["13"], "2": ["5J"]},
    "AVERSVM": {"-": ["12", "12 "]},
    "AVERSVS": {"-": ["21"]},
    "AVERTO": {"-": ["53"]},
    "AVFERO": {"-": ["56"]},
    "AVFVGIO": {"-": ["55"]},
    "AVGEO": {"-": ["52"]},
    "AVGESCO": {"-": ["53"]},
    "AVGMEN": {"-": ["13"]},
    "AVGVR": {"-": ["13"]},
    "AVGVRALE": {"-": ["13"]},
    "AVGVRALIS": {"-": ["24"]},
    "AVGVRATVS": {"-": ["14"]},
    "AVGVRIVM": {"-": ["12"]},
    "AVGVROR": {"-": ["5J"]},
    "AVGVSTA": {"1N": ["11"], "2N": ["11"], "N": ["11"]},
    "AVGVSTE": {"-": ["60"]},
    "AVGVSTVS": {"-": ["21"], "C": ["12"], "N": ["21"], "A": ["21"]},
    "AVIA": {"-": ["11"]},
    "AVIARIVM": {"-": ["12"]},
    "AVIDE": {"-": ["60"]},
    "AVIDITAS": {"-": ["13"]},
    "AVIDVS": {"-": ["21"]},
    "AVIS": {"-": ["13"]},
    "AVITVS": {"-": ["21"], "N": ["12"]},
    "AVIVM": {"-": ["12"]},
    "AVIVS": {"-": ["21"]},
    "AVLA": {"1": ["11"], "2": ["11"]},
    "AVLAEVM": {"-": ["12"]},
    "AVLVS": {"N": ["12"]},
    "AVOCAMENTVM": {"-": ["12"]},
    "AVOCATIO": {"-": ["13"]},
    "AVOCO": {"-": ["51"]},
    "AVOLO": {"2": ["51"]},
    "AVRA": {"-": ["11"], "N": ["11"]},
    "AVRARIA": {"-": ["11"]},
    "AVRATVS": {"-": ["21"]},
    "AVRELIVS": {"N": ["12"], "A": ["21"]},
    "AVREOLVS": {"-": ["21"]},
    "AVREVS": {"1": ["12"], "2": ["21"]},
    "AVRICILLA": {"-": ["11"]},
    "AVRICOMVS": {"-": ["21"]},
    "AVRICVLA": {"-": ["11"]},
    "AVRIFER": {"-": ["21"]},
    "AVRIFERVS": {"-": ["21"]},
    "AVRIFEX": {"-": ["13"]},
    "AVRIGA": {"-": ["11"]},
    "AVRION": {"-": ["60"]},
    "AVRIPIGMENTVM": {"-": ["12"]},
    "AVRIS": {"-": ["13"]},
    "AVRITVS": {"-": ["21"]},
    "AVRO": {"-": ["51"]},
    "AVRORA": {"-": ["11"], "N": ["11"]},
    "AVRVM": {"-": ["12"]},
    "AVRVNCI": {"N": ["12"]},
    "AVRVNCVS": {"N": ["21"], "A": ["21"]},
    "AVSCVLTATIO": {"-": ["13"]},
    "AVSCVLTO": {"-": ["51"]},
    "AVSPEX": {"-": ["13"], "N": ["13"]},
    "AVSPICATVS": {"1": ["14"], "2": ["21"], "-": ["14"]},
    "AVSPICIVM": {"-": ["12"]},
    "AVSPICO": {"-": ["51"]},
    "AVSPICOR": {"-": ["5J"]},
    "AVSTER": {"-": ["12"], "N": ["12"]},
    "AVSTERITAS": {"-": ["13"]},
    "AVSTERVS": {"-": ["21"]},
    "AVSTRALIS": {"-": ["24"]},
    "AVSTRINVS": {"-": ["21"]},
    "AVSVM": {"A": ["12"], "-": ["12"]},
    "AVT": {"-": ["81"]},
    "AVTEM": {"-": ["81"]},
    "AVTIONOR": {"-": ["5J"]},
    "AVTOMATOPOEETVS": {"-": ["21"]},
    "AVTVMNALIS": {"-": ["24"]},
    "AVTVMNITAS": {"-": ["13"]},
    "AVTVMNVS": {"1": ["12"], "2": ["21"], "N": ["12"]},
    "AVTVMO": {"-": ["51"]},
    "AVVNCVLVS": {"-": ["12"]},
    "AVVS": {"-": ["12"]},
    "AVXILIARES": {"-": ["13"]},
    "AVXILIARII": {"-": ["12"]},
    "AVXILIARIS": {"-": ["24"]},
    "AVXILIARIVS": {"-": ["21"]},
    "AVXILIATOR": {"-": ["13"]},
    "AVXILIATVS": {"-": ["14"]},
    "AVXILIOR": {"-": ["5J"]},
    "AVXILIVM": {"-": ["12"]},
    "AXICVLVS": {"-": ["12"]},
    "AXIS": {"-": ["13"]},
    "AXON": {"-": ["17"]},
    "BABYLONICVM": {"-": ["12"], "N": ["12"]},
    "BACCA": {"-": ["11"]},
    "BACCAR": {"-": ["13"]},
    "BACCATVS": {"-": ["21"]},
    "BACCHA": {"-": ["17"], "N": ["11"]},
    "BACCHABVNDVS": {"-": ["21"]},
    "BACCHANAL": {"-": ["13"]},
    "BACCHANALIA": {"-": ["16"], "N": ["11"]},
    "BACCHATIO": {"-": ["13"]},
    "BACCHOR": {"-": ["5J"]},
    "BACCHVS": {"N": ["12"]},
    "BACCIFER": {"-": ["21"]},
    "BACILLVM": {"-": ["12"]},
    "BACVLVM": {"-": ["12"]},
    "BACVLVS": {"-": ["12"], "N": ["12"]},
    "BAIAE": {"N": ["11"]},
    "BALAENA": {"-": ["11"]},
    "BALANATVS": {"-": ["21"]},
    "BALANTES": {"-": ["13"]},
    "BALANVS": {"-": ["12"]},
    "BALATRO": {"-": ["13"], "N": ["13"]},
    "BALATVS": {"-": ["14"]},
    "BALBE": {"-": ["60"]},
    "BALBVS": {"-": ["21"], "N": ["12"]},
    "BALBVTIO": {"-": ["54"]},
    "BALEARES": {"N": ["13"]},
    "BALINEAE": {"-": ["11"]},
    "BALINEVM": {"-": ["12"]},
    "BALLISTA": {"-": ["11"]},
    "BALNEARIA": {"-": ["12"]},
    "BALNEARIVS": {"-": ["21"]},
    "BALNEATOR": {"-": ["13"]},
    "BALNEOLVM": {"-": ["12"]},
    "BALNEVM": {"-": ["16"]},
    "BALO": {"-": ["51"]},
    "BALSAMVM": {"-": ["12"]},
    "BALTEVM": {"-": ["12"]},
    "BALTEVS": {"-": ["12"]},
    "BAPTIZO": {"-": ["51"]},
    "BARATHRVM": {"-": ["12"]},
    "BARBA": {"-": ["11"], "N": ["11"]},
    "BARBARE": {"-": ["60"]},
    "BARBARIA": {"-": ["11"]},
    "BARBARICVS": {"-": ["21"]},
    "BARBARISMVS": {"-": ["12"]},
    "BARBARVS": {"1": ["12"], "2": ["21"]},
    "BARBATVS": {"2": ["21"]},
    "BARBIGER": {"-": ["21"]},
    "BARBITON": {"-": ["17"]},
    "BARBITOS": {"-": ["17"]},
    "BARDITVS": {"-": ["14"]},
    "BARIOBAL": {"N": ["17"]},
    "BARIS": {"-": ["17"]},
    "BARO": {"-": ["13"]},
    "BARRVS": {"-": ["12"], "N": ["12"]},
    "BARVLCOS": {"-": ["27"]},
    "BARYCEPHALVS": {"-": ["21"]},
    "BASCAVDA": {"-": ["11"]},
    "BASIATIO": {"-": ["13"]},
    "BASILICA": {"-": ["11"]},
    "BASILISCVS": {"N": ["12"]},
    "BASIO": {"-": ["51"]},
    "BASIS": {"-": ["17"]},
    "BASIVM": {"-": ["12"]},
    "BAVBOR": {"-": ["5J"]},
    "BEATE": {"-": ["60"]},
    "BEATVLVS": {"-": ["21"]},
    "BEATVM": {"-": ["12"]},
    "BEATVS": {"-": ["21"]},
    "BELLATOR": {"-": ["13"]},
    "BELLATRIX": {"-": ["13"]},
    "BELLE": {"-": ["60"]},
    "BELLICOSVS": {"-": ["21"]},
    "BELLICVM": {"-": ["12"]},
    "BELLICVS": {"-": ["21"]},
    "BELLIGER": {"-": ["21"]},
    "BELLIGERO": {"-": ["51"]},
    "BELLO": {"-": ["51"]},
    "BELLONA": {"N": ["11"]},
    "BELLVA": {"-": ["11"]},
    "BELLVM": {"-": ["12"], "N": ["12"]},
    "BELLVOSVS": {"-": ["21"]},
    "BELLVS": {"-": ["21"], "N": ["12"]},
    "BELVA": {"-": ["11"]},
    "BENE": {"-": ["60"]},
    "BENEDICO": {"2": ["53"]},
    "BENEDICTVM": {"-": ["12"]},
    "BENEFACIO": {"-": ["55"]},
    "BENEFACTVM": {"-": ["12"]},
    "BENEFICENTIA": {"-": ["11"]},
    "BENEFICIARIVS": {"1": ["12"], "2": ["21"]},
    "BENEFICIVM": {"-": ["12"]},
    "BENEFICVS": {"-": ["21"]},
    "BENEVENTANVS": {"N": ["21"], "A": ["21"]},
    "BENEVENTVM": {"N": ["12"]},
    "BENEVOLE": {"-": ["60"]},
    "BENEVOLENS": {"-": ["25"]},
    "BENEVOLENTIA": {"-": ["11"]},
    "BENEVOLVS": {"-": ["21"]},
    "BENIGNE": {"-": ["60"]},
    "BENIGNITAS": {"-": ["13"]},
    "BENIGNVS": {"-": ["21"], "N": ["12"]},
    "BEO": {"-": ["51"]},
    "BERYLLVS": {"-": ["12"]},
    "BES": {"-": ["16"]},
    "BESTIA": {"-": ["11"], "N": ["11"]},
    "BESTIARIVS": {"1": ["12"], "2": ["21"]},
    "BESTIOLA": {"-": ["11"]},
    "BETA": {"1": ["11"], "2": ["16"]},
    "BIBERARIVS": {"-": ["12"]},
    "BIBLIOTHECA": {"-": ["11"]},
    "BIBO": {"1": ["13"], "2": ["53"]},
    "BIBVLVS": {"-": ["21"], "N": ["12"]},
    "BICEPS": {"-": ["25"]},
    "BICOLOR": {"-": ["25"]},
    "BICORNES": {"-": ["13"]},
    "BICORNIS": {"-": ["24"]},
    "BIDENS": {"1": ["13"], "2": ["25"]},
    "BIDENTAL": {"-": ["13"]},
    "BIDINI": {"N": ["12"]},
    "BIDINVS": {"N": ["21"], "A": ["21"]},
    "BIDIS": {"N": ["13"]},
    "BIDVVM": {"-": ["12"]},
    "BIENNIS": {"-": ["24"]},
    "BIENNIVM": {"-": ["12"]},
    "BIFARIVS": {"-": ["21"]},
    "BIFER": {"-": ["21"]},
    "BIFORIS": {"-": ["24"]},
    "BIFORMIS": {"-": ["24"]},
    "BIGA": {"-": ["11"]},
    "BIGATVS": {"-": ["21"]},
    "BIIVGI": {"-": ["12"]},
    "BIIVGIS": {"-": ["24"]},
    "BIIVGVS": {"-": ["21"]},
    "BILIBRIS": {"-": ["24"]},
    "BILINGVIS": {"-": ["24"]},
    "BILIS": {"-": ["13"]},
    "BILVSTRIS": {"-": ["24"]},
    "BIMARIS": {"-": ["24"]},
    "BIMEMBRES": {"-": ["13"], "N": ["13"]},
    "BIMEMBRIS": {"-": ["24"]},
    "BIMESTRIS": {"-": ["24"]},
    "BIMVLVS": {"-": ["21"]},
    "BIMVS": {"-": ["21"]},
    "BINOCTIVM": {"-": ["12"]},
    "BINVS": {"-": ["33"]},
    "BIPALIVM": {"-": ["12"]},
    "BIPARTIOR": {"-": ["5M"]},
    "BIPARTITIO": {"-": ["13"]},
    "BIPARTITO": {"-": ["60"]},
    "BIPATENS": {"-": ["25"]},
    "BIPEDALIS": {"-": ["24"]},
    "BIPENNIS": {"1": ["13"], "2": ["24"]},
    "BIPES": {"1": ["13"], "2": ["25"]},
    "BIREMIS": {"1": ["13"], "2": ["24"]},
    "BIS": {"-": ["36"]},
    "BISON": {"-": ["17"]},
    "BISVLCVS": {"-": ["21"]},
    "BITHYNIA": {"N": ["11"]},
    "BITVMEN": {"-": ["13"]},
    "BITVMINOSVS": {"-": ["21"]},
    "BIVGVS": {"-": ["21"]},
    "BIVIVM": {"-": ["12"]},
    "BLAESVS": {"-": ["21"], "N": ["12"]},
    "BLANDE": {"-": ["60"]},
    "BLANDILOQVENTIA": {"-": ["11"]},
    "BLANDILOQVVS": {"-": ["21"]},
    "BLANDIMENTVM": {"-": ["12"]},
    "BLANDIOR": {"-": ["5M"]},
    "BLANDITIA": {"-": ["11"], "N": ["11"]},
    "BLANDVS": {"-": ["21"], "N": ["12"]},
    "BLATERO": {"-": ["51"]},
    "BLATTA": {"-": ["11"]},
    "BLATTARIVS": {"-": ["21"]},
    "BOARIVS": {"-": ["21"], "A": ["21"]},
    "BOEOTIA": {"N": ["11"]},
    "BOEOTII": {"N": ["12"]},
    "BOETHVS": {"N": ["12"]},
    "BOLETATIO": {"-": ["13"]},
    "BOLETVS": {"-": ["12"]},
    "BOMBVS": {"-": ["12"]},
    "BOMBYCINVS": {"-": ["21"]},
    "BOMBYX": {"-": ["13"]},
    "BONI": {"-": ["12"]},
    "BONITAS": {"-": ["13"]},
    "BONVM": {"-": ["12"]},
    "BONVS": {"-": ["21"], "N": ["21"], "A": ["21"]},
    "BOO": {"-": ["51"]},
    "BOREAS": {"-": ["17"], "N": ["17"]},
    "BOS": {"-": ["13"], "N": ["13"]},
    "BOTVLARIVS": {"-": ["12"]},
    "BRACA": {"-": ["11"]},
    "BRACATVS": {"-": ["21"]},
    "BRACHIOLVM": {"-": ["12"]},
    "BRACHIVM": {"-": ["12"]},
    "BRACTEA": {"-": ["11"]},
    "BRACTEATVS": {"-": ["21"]},
    "BRACTEOLA": {"-": ["11"]},
    "BRASSICA": {"-": ["11"]},
    "BREVE": {"1": ["13"], "2": ["60"]},
    "BREVI": {"-": ["60"]},
    "BREVIARIVM": {"-": ["12"]},
    "BREVIS": {"-": ["24"]},
    "BREVITAS": {"-": ["13"]},
    "BREVITER": {"-": ["60"]},
    "BRITANNIA": {"N": ["11"]},
    "BRITANNVS": {"N": ["21"], "A": ["21"]},
    "BROCCHVS": {"N": ["12"]},
    "BRVMA": {"-": ["11"]},
    "BRVMALIS": {"-": ["24"]},
    "BRVNDISIVM": {"N": ["12"]},
    "BRVTESCO": {"-": ["53"]},
    "BRVTVS": {"-": ["21"], "N": ["12"]},
    "BVBILE": {"-": ["13"]},
    "BVBO": {"-": ["13"]},
    "BVBVLA": {"-": ["11"]},
    "BVBVLCVS": {"-": ["12"]},
    "BVBVLVS": {"-": ["21"]},
    "BVCCA": {"-": ["11"]},
    "BVCCVLA": {"-": ["11"]},
    "BVCERIVS": {"-": ["21"]},
    "BVCERVS": {"-": ["21"]},
    "BVCINA": {"-": ["11"]},
    "BVCINATOR": {"-": ["13"]},
    "BVCVLA": {"-": ["11"]},
    "BVFO": {"-": ["13"]},
    "BVLBVS": {"-": ["12"], "N": ["12"]},
    "BVLLA": {"-": ["11"]},
    "BVLLATVS": {"-": ["21"]},
    "BVLLIO": {"-": ["54"]},
    "BVLLITVS": {"-": ["14"]},
    "BVLLO": {"-": ["51"]},
    "BVMASTVS": {"-": ["12"]},
    "BVRIS": {"-": ["13"]},
    "BVRRVS": {"N": ["12"]},
    "BVSIRIS": {"N": ["13"]},
    "BVSTVARIVS": {"1": ["12"], "2": ["21"]},
    "BVSTVM": {"-": ["12"]},
    "BVTHROTVM": {"N": ["17"]},
    "BVXIFER": {"-": ["21"]},
    "BVXVM": {"-": ["12"]},
    "BVXVS": {"-": ["12"]},
    "BYZANTII": {"N": ["12"]},
    "BYZANTIVM": {"N": ["12"]},
    "CABALLVS": {"-": ["12"]},
    "CACHINNO": {"1": ["13"], "2": ["51"]},
    "CACHINNVS": {"-": ["12"]},
    "CACO": {"-": ["51"]},
    "CACOETHES": {"-": ["13"]},
    "CACVMEN": {"-": ["13"]},
    "CACVRIVS": {"N": ["12"]},
    "CACVS": {"N": ["12"]},
    "CADAVER": {"-": ["13"]},
    "CADMEA": {"N": ["11"]},
    "CADMVS": {"-": ["12"], "N": ["12"]},
    "CADO": {"-": ["53"]},
    "CADVCEATOR": {"-": ["13"]},
    "CADVCEVM": {"-": ["12"]},
    "CADVCEVS": {"-": ["12"]},
    "CADVCIFER": {"-": ["21"], "N": ["12"]},
    "CADVCVM": {"-": ["12"]},
    "CADVCVS": {"-": ["21"]},
    "CADVRCVM": {"-": ["12"]},
    "CADVS": {"-": ["12"]},
    "CAECIGENVS": {"-": ["21"]},
    "CAECILIVS": {"N": ["12"], "A": ["21"]},
    "CAECITAS": {"-": ["13"]},
    "CAECO": {"-": ["51"]},
    "CAECVS": {"-": ["21"], "N": ["12"]},
    "CAEDES": {"-": ["13"]},
    "CAEDICIVS": {"N": ["12"]},
    "CAEDO": {"-": ["53"]},
    "CAEDVVS": {"-": ["21"]},
    "CAELATOR": {"-": ["13"]},
    "CAELATVRA": {"-": ["11"]},
    "CAELEBS": {"-": ["25"]},
    "CAELES": {"1": ["13"], "2": ["25"], "N": ["13"]},
    "CAELESTES": {"-": ["13"]},
    "CAELESTIA": {"-": ["13"]},
    "CAELESTIS": {"-": ["24"]},
    "CAELI": {"-": ["12"]},
    "CAELIBATVS": {"-": ["14"]},
    "CAELICOLA": {"-": ["11"]},
    "CAELIFER": {"-": ["21"]},
    "CAELITES": {"-": ["13"]},
    "CAELIVS": {"N": ["12"]},
    "CAELO": {"-": ["51"]},
    "CAELVM": {"1": ["12"], "2": ["12"]},
    "CAEMENTICIVS": {"-": ["21"]},
    "CAEMENTVM": {"-": ["12"]},
    "CAENINENSES": {"N": ["13"]},
    "CAENINVS": {"N": ["21"], "A": ["21"]},
    "CAENO": {"N": ["13"]},
    "CAENOSVS": {"-": ["21"]},
    "CAENVM": {"-": ["12"]},
    "CAEPA": {"-": ["17"]},
    "CAEPIO": {"N": ["13"]},
    "CAERE": {"N": ["16"]},
    "CAERIMONIA": {"-": ["11"]},
    "CAERVLA": {"-": ["12"]},
    "CAERVLEVM": {"-": ["12"]},
    "CAERVLEVS": {"-": ["21"]},
    "CAESAR": {"N": ["13"]},
    "CAESARIES": {"-": ["15"]},
    "CAESETIVS": {"N": ["12"]},
    "CAESIVS": {"-": ["21"], "N": ["12"], "A": ["21"]},
    "CAESO": {"N": ["13"]},
    "CAESONIVS": {"N": ["12"]},
    "CAESPES": {"-": ["13"]},
    "CAESTVS": {"-": ["14"]},
    "CALACTE": {"N": ["17"]},
    "CALACTINI": {"N": ["12"]},
    "CALACTINVS": {"N": ["21"], "A": ["21"]},
    "CALAMISTER": {"-": ["12"]},
    "CALAMITAS": {"-": ["13"]},
    "CALAMITOSVS": {"-": ["21"]},
    "CALAMVS": {"-": ["12"]},
    "CALATHISCVS": {"-": ["12"]},
    "CALATHVS": {"-": ["12"]},
    "CALCAR": {"-": ["13"]},
    "CALCARIVS": {"1": ["12"], "2": ["21"]},
    "CALCATVRA": {"-": ["11"]},
    "CALCEAMENTVM": {"-": ["12"]},
    "CALCEO": {"-": ["51"]},
    "CALCEVS": {"-": ["12"]},
    "CALCO": {"-": ["51"]},
    "CALCVLOSVS": {"-": ["21"]},
    "CALCVLVS": {"-": ["12"]},
    "CALDA": {"-": ["11"]},
    "CALDARIVM": {"-": ["12"]},
    "CALDARIVS": {"-": ["21"]},
    "CALEFACIO": {"-": ["55"]},
    "CALEFACTO": {"-": ["51"]},
    "CALEFIO": {"-": ["56"]},
    "CALENVS": {"N": ["12"], "A": ["21"]},
    "CALEO": {"-": ["52"]},
    "CALESCO": {"-": ["53"]},
    "CALICVLVS": {"-": ["12"]},
    "CALIDA": {"-": ["11"]},
    "CALIDIANVS": {"N": ["21"], "A": ["21"]},
    "CALIDIVS": {"N": ["12"]},
    "CALIDVS": {"-": ["21"]},
    "CALIENDRVM": {"-": ["12"]},
    "CALIGA": {"-": ["11"]},
    "CALIGATVS": {"-": ["21"]},
    "CALIGINOSVS": {"-": ["21"]},
    "CALIGO": {"1": ["13"], "2": ["51"]},
    "CALIX": {"-": ["13"]},
    "CALLEO": {"-": ["52"]},
    "CALLIDAMA": {"N": ["17"]},
    "CALLIDE": {"-": ["60"]},
    "CALLIDITAS": {"-": ["13"]},
    "CALLIDVS": {"-": ["21"]},
    "CALLIS": {"-": ["13"]},
    "CALLISCO": {"-": ["53"]},
    "CALLOSVS": {"-": ["21"]},
    "CALLVM": {"-": ["12"]},
    "CALO": {"1": ["13"], "2": ["51"]},
    "CALOR": {"-": ["13"]},
    "CALPVRNIVS": {"N": ["12"], "A": ["21"]},
    "CALTHA": {"-": ["11"]},
    "CALVARIA": {"1": ["11"], "2": ["11"]},
    "CALVITIVM": {"-": ["12"]},
    "CALVMNIA": {"-": ["11"]},
    "CALVMNIATOR": {"-": ["13"]},
    "CALVMNIOR": {"-": ["5J"]},
    "CALVVS": {"-": ["21"], "N": ["12"]},
    "CALX": {"1": ["13"], "2": ["13"]},
    "CAMELVS": {"-": ["12"]},
    "CAMENA": {"-": ["11"], "N": ["11"]},
    "CAMENAE": {"N": ["11"]},
    "CAMERA": {"-": ["11"]},
    "CAMERIA": {"N": ["11"]},
    "CAMINVS": {"-": ["12"]},
    "CAMMARVS": {"-": ["12"]},
    "CAMPANIA": {"N": ["11"]},
    "CAMPESTER": {"-": ["23"]},
    "CAMPESTRE": {"-": ["13"]},
    "CAMPESTRIA": {"-": ["13"]},
    "CAMPVS": {"1": ["12"], "2": ["12"], "N": ["12"]},
    "CAMVRVS": {"-": ["21"]},
    "CANALICVLVS": {"-": ["12"]},
    "CANALIS": {"1": ["13"], "2": ["24"]},
    "CANCELLVS": {"-": ["12"]},
    "CANCER": {"-": ["16"], "N": ["12"]},
    "CANDELA": {"-": ["11"]},
    "CANDELABRVM": {"-": ["12"]},
    "CANDENS": {"-": ["25", "22"]},
    "CANDENTIA": {"-": ["11"]},
    "CANDEO": {"-": ["52"]},
    "CANDESCO": {"-": ["53"]},
    "CANDIDATVS": {"1": ["12"], "2": ["21"]},
    "CANDIDVLVS": {"-": ["21"]},
    "CANDIDVS": {"-": ["21"], "N": ["12"]},
    "CANDOR": {"-": ["13"]},
    "CANEO": {"-": ["52"]},
    "CANEPHORA": {"N": ["17"]},
    "CANESCO": {"-": ["53"]},
    "CANI": {"-": ["12"]},
    "CANICVLA": {"-": ["11"], "N": ["11"]},
    "CANIDIA": {"-": ["11"], "N": ["11"]},
    "CANINVS": {"-": ["21"], "N": ["12"]},
    "CANIS": {"-": ["13"], "N": ["13"]},
    "CANISTRA": {"-": ["12"]},
    "CANISTRVM": {"-": ["12"]},
    "CANITIES": {"-": ["15"]},
    "CANNA": {"-": ["11"]},
    "CANNABIS": {"-": ["16"]},
    "CANNENSIS": {"N": ["24"], "A": ["24"]},
    "CANNVLA": {"-": ["11"]},
    "CANO": {"-": ["53"]},
    "CANON": {"-": ["17"]},
    "CANONICVS": {"1": ["12"], "2": ["21"]},
    "CANOR": {"-": ["13"]},
    "CANORVS": {"-": ["21"]},
    "CANTABRI": {"-": ["12"], "N": ["12"]},
    "CANTERIVS": {"-": ["12"]},
    "CANTHARVS": {"-": ["12"]},
    "CANTHVS": {"-": ["12"]},
    "CANTICVM": {"-": ["12"]},
    "CANTILENA": {"-": ["11"]},
    "CANTIO": {"-": ["13"]},
    "CANTITO": {"-": ["51"]},
    "CANTO": {"-": ["51"]},
    "CANTOR": {"-": ["13"]},
    "CANTVS": {"-": ["14"]},
    "CANVLEIANVS": {"N": ["21"], "A": ["21"]},
    "CANVLEIVS": {"N": ["12"]},
    "CANVS": {"-": ["21"], "N": ["12"]},
    "CAPACITAS": {"-": ["13"]},
    "CAPAX": {"-": ["25"], "N": ["13"]},
    "CAPEDVNCVLA": {"-": ["11"]},
    "CAPELLA": {"-": ["11"], "N": ["11"]},
    "CAPENA": {"N": ["11"]},
    "CAPENAS": {"A": ["25"]},
    "CAPER": {"-": ["12"]},
    "CAPESSO": {"-": ["53"]},
    "CAPETVS": {"N": ["12"]},
    "CAPILLAMENTVM": {"-": ["12"]},
    "CAPILLATVS": {"-": ["21"]},
    "CAPILLVS": {"-": ["12"]},
    "CAPIO": {"1": ["13"], "2": ["55"]},
    "CAPISTRO": {"-": ["51"]},
    "CAPISTRVM": {"-": ["12"]},
    "CAPITAL": {"-": ["13"]},
    "CAPITALIS": {"-": ["24"]},
    "CAPITATIO": {"-": ["13"]},
    "CAPITATVS": {"-": ["21"]},
    "CAPITINVS": {"N": ["21"], "A": ["21"]},
    "CAPITO": {"-": ["13"], "N": ["13"]},
    "CAPITOLINVS": {"A": ["21"], "N": ["21", "12"]},
    "CAPITOLIVM": {"N": ["12"]},
    "CAPITVLVM": {"-": ["12"]},
    "CAPPADOCIA": {"N": ["11"]},
    "CAPRA": {"-": ["11"], "N": ["11"]},
    "CAPREA": {"-": ["11"]},
    "CAPREOLVS": {"-": ["12"]},
    "CAPRIFICVS": {"-": ["12"]},
    "CAPRIGENVS": {"-": ["21"]},
    "CAPRILE": {"-": ["13"]},
    "CAPRIMVLGVS": {"-": ["12"]},
    "CAPRINVS": {"-": ["21"]},
    "CAPRIPES": {"-": ["25"]},
    "CAPSA": {"-": ["11"], "N": ["11"]},
    "CAPSVLA": {"-": ["11"]},
    "CAPSVS": {"-": ["12"]},
    "CAPTATIO": {"-": ["13"]},
    "CAPTATOR": {"-": ["13"]},
    "CAPTIO": {"-": ["13"]},
    "CAPTIOSVS": {"-": ["21"]},
    "CAPTIVA": {"-": ["11"]},
    "CAPTIVITAS": {"-": ["13"]},
    "CAPTIVVS": {"1": ["12"], "2": ["21"]},
    "CAPTO": {"-": ["51"]},
    "CAPTVRA": {"-": ["11"]},
    "CAPTVS": {"-": ["14"], "A": ["21"]},
    "CAPVLATOR": {"-": ["13"]},
    "CAPVLVS": {"-": ["12"]},
    "CAPVT": {"-": ["13"]},
    "CAPYS": {"N": ["17"]},
    "CARBASA": {"-": ["12"]},
    "CARBASEVS": {"-": ["21"]},
    "CARBASVS": {"1": ["12"], "2": ["21"]},
    "CARBO": {"-": ["13"], "N": ["13"]},
    "CARBVNCVLVS": {"-": ["12"]},
    "CARCER": {"-": ["13"]},
    "CARCHESIVM": {"-": ["12"]},
    "CARCINOMA": {"-": ["17"]},
    "CARDIACVS": {"1": ["12"], "2": ["21"]},
    "CARDINALIS": {"-": ["24"]},
    "CARDINATVS": {"-": ["21"]},
    "CARDO": {"-": ["13"]},
    "CARDVVS": {"-": ["12"]},
    "CARE": {"-": ["60"]},
    "CARECTVM": {"-": ["12"]},
    "CAREO": {"-": ["52"]},
    "CAREX": {"-": ["13"]},
    "CARICA": {"-": ["11"]},
    "CARIES": {"-": ["15"]},
    "CARINA": {"-": ["11"], "N": ["11"]},
    "CARIOSVS": {"-": ["21"]},
    "CARITAS": {"-": ["13"]},
    "CARMEN": {"1": ["13"], "2": ["13"]},
    "CARMENTALIS": {"N": ["24"], "A": ["24"]},
    "CARMENTIS": {"N": ["16"]},
    "CARMO": {"-": ["13"], "N": ["13"]},
    "CARNARIVM": {"-": ["12"]},
    "CARNIFEX": {"-": ["13"]},
    "CARNIFICINA": {"-": ["11"]},
    "CARNIFICO": {"-": ["51"]},
    "CARO": {"1": ["13"], "2": ["53"]},
    "CARPATINA": {"-": ["11"]},
    "CARPENTVM": {"-": ["12"]},
    "CARPINATIVS": {"N": ["12"]},
    "CARPINVS": {"-": ["12"]},
    "CARPO": {"-": ["53"]},
    "CARPTIM": {"-": ["60"]},
    "CARPTOR": {"-": ["13"]},
    "CARRVM": {"-": ["12"]},
    "CARRVS": {"-": ["12"]},
    "CARTHAGINIENSES": {"N": ["13"]},
    "CARTHAGINIENSIS": {"N": ["24"], "A": ["24"]},
    "CARTHAGO": {"N": ["13"]},
    "CARVS": {"-": ["21"], "N": ["12"]},
    "CARYATIS": {"-": ["17"]},
    "CASA": {"-": ["11"]},
    "CASEVM": {"-": ["12"]},
    "CASEVS": {"-": ["12"]},
    "CASIA": {"-": ["11"]},
    "CASSIANVS": {"N": ["21"], "A": ["21"]},
    "CASSIDA": {"-": ["11"]},
    "CASSIS": {"1": ["13"], "2": ["13"]},
    "CASSIVS": {"A": ["21"], "N": ["12"]},
    "CASSVM": {"-": ["12"]},
    "CASSVS": {"-": ["21"]},
    "CASTANEA": {"-": ["11"]},
    "CASTANEVS": {"-": ["21"]},
    "CASTE": {"-": ["60"]},
    "CASTELLANI": {"-": ["12"]},
    "CASTELLANVS": {"1": ["12"], "2": ["21"]},
    "CASTELLARIVS": {"-": ["12"]},
    "CASTELLVM": {"-": ["12"]},
    "CASTIFICVS": {"-": ["21"]},
    "CASTIGATIO": {"-": ["13"]},
    "CASTIGATOR": {"-": ["13"]},
    "CASTIGATVS": {"-": ["21"]},
    "CASTIGO": {"-": ["51"]},
    "CASTIMONIA": {"-": ["11"]},
    "CASTITAS": {"-": ["13"]},
    "CASTOR": {"-": ["17"], "N": ["17"]},
    "CASTOREVM": {"-": ["12"]},
    "CASTOREVS": {"N": ["21"], "A": ["21"]},
    "CASTRA": {"1": ["11"], "2": ["12"], "N": ["11"]},
    "CASTRENSIS": {"-": ["24"]},
    "CASTRICIVS": {"N": ["12"]},
    "CASTRO": {"-": ["51"]},
    "CASTRVM": {"-": ["12"], "N": ["12"]},
    "CASTVS": {"1": ["14"], "2": ["21"]},
    "CASV": {"-": ["60"]},
    "CASVLA": {"-": ["11"]},
    "CASVS": {"-": ["14"]},
    "CATAGRAPHVS": {"-": ["21"]},
    "CATAPHRACTVS": {"-": ["12"]},
    "CATAPLASMA": {"-": ["17"]},
    "CATAPLEXIS": {"-": ["13"]},
    "CATAPVLTA": {"-": ["11"]},
    "CATARACTA": {"-": ["11"]},
    "CATASCOPVS": {"-": ["12"]},
    "CATASTA": {"-": ["11"]},
    "CATATEXITECHNOS": {"-": ["27"]},
    "CATATONVS": {"-": ["21"]},
    "CATECHON": {"-": ["57"]},
    "CATELLA": {"-": ["11"], "2": ["11"], "1": ["12"]},
    "CATELLVS": {"-": ["12"]},
    "CATENA": {"-": ["11"]},
    "CATENARIVS": {"-": ["21"]},
    "CATENATIO": {"-": ["13"]},
    "CATENO": {"-": ["51"]},
    "CATERVA": {"-": ["11"]},
    "CATERVATIM": {"-": ["60"]},
    "CATHEDRA": {"-": ["11"]},
    "CATHEDRARIVS": {"-": ["21"]},
    "CATHETVS": {"-": ["17"]},
    "CATILINA": {"N": ["11"]},
    "CATILLVS": {"-": ["12"], "N": ["12"]},
    "CATINA": {"N": ["11"]},
    "CATINENSES": {"N": ["13"]},
    "CATINENSIS": {"N": ["24"], "A": ["24"]},
    "CATINVS": {"-": ["12"]},
    "CATLASTER": {"-": ["12"]},
    "CATO": {"N": ["13"]},
    "CATVLVS": {"-": ["12"], "N": ["12"]},
    "CATVS": {"-": ["21"], "N": ["12"]},
    "CAVATVS": {"-": ["21"]},
    "CAVDA": {"-": ["11"]},
    "CAVDEX": {"-": ["13"], "N": ["13"]},
    "CAVDICARIVS": {"-": ["21"]},
    "CAVEA": {"-": ["11"]},
    "CAVEO": {"-": ["52"]},
    "CAVERNA": {"-": ["11"]},
    "CAVILLATIO": {"-": ["13"]},
    "CAVILLATOR": {"-": ["13"]},
    "CAVILLOR": {"-": ["5J"]},
    "CAVLA": {"-": ["11"]},
    "CAVLICVLVS": {"-": ["12"]},
    "CAVLIS": {"-": ["13"]},
    "CAVO": {"1": ["51"], "2": ["53"]},
    "CAVPO": {"-": ["13"]},
    "CAVPONA": {"-": ["11"]},
    "CAVPONOR": {"-": ["5J"]},
    "CAVRVS": {"-": ["12"], "N": ["12"]},
    "CAVSA": {"-": ["11"]},
    "CAVSARIVS": {"-": ["21"]},
    "CAVSIDICVS": {"-": ["12"]},
    "CAVSOR": {"-": ["5J"]},
    "CAVSVLA": {"-": ["11"]},
    "CAVTE": {"-": ["60"]},
    "CAVTES": {"-": ["13"]},
    "CAVTIO": {"-": ["13"]},
    "CAVTVS": {"-": ["21"]},
    "CAVVM": {"-": ["12"]},
    "CAVVS": {"1": ["12"], "2": ["21"]},
    "CEDO": {"1": ["53"], "2": ["90"]},
    "CEDRIVM": {"-": ["12"]},
    "CEDRVS": {"-": ["12"]},
    "CELATVM": {"-": ["12"]},
    "CELEBER": {"-": ["23"]},
    "CELEBRATIO": {"-": ["13"]},
    "CELEBRATVS": {"-": ["21"]},
    "CELEBRITAS": {"-": ["13"]},
    "CELEBRITER": {"-": ["60"]},
    "CELEBRO": {"-": ["51"]},
    "CELER": {"-": ["23"], "N": ["12"]},
    "CELERIPES": {"-": ["25"]},
    "CELERITAS": {"-": ["13"]},
    "CELERITER": {"-": ["60"]},
    "CELERO": {"-": ["51"]},
    "CELLA": {"-": ["11"]},
    "CELLARIVS": {"1": ["12"], "2": ["21"]},
    "CELLO": {"-": ["53"]},
    "CELLVLA": {"-": ["11"]},
    "CELO": {"-": ["51"]},
    "CELSVS": {"-": ["21"], "N": ["12"]},
    "CENA": {"-": ["11"]},
    "CENACVLVM": {"-": ["12"]},
    "CENATIO": {"-": ["13"]},
    "CENDO": {"-": ["53"]},
    "CENO": {"-": ["51"]},
    "CENSEO": {"-": ["52"]},
    "CENSOR": {"-": ["13"]},
    "CENSORIVS": {"1": ["12"], "2": ["21"], "N": ["12"]},
    "CENSVRA": {"-": ["11"]},
    "CENSVS": {"-": ["14"]},
    "CENTAVREVM": {"-": ["12"]},
    "CENTAVRVS": {"-": ["12"], "N": ["12"]},
    "CENTENARIVS": {"-": ["21"]},
    "CENTENVS": {"-": ["33"]},
    "CENTESIMA": {"-": ["11"]},
    "CENTESIMVS": {"-": ["32"]},
    "CENTICEPS": {"-": ["25"]},
    "CENTIES": {"-": ["36"]},
    "CENTIMANVS": {"-": ["21"]},
    "CENTO": {"-": ["13"], "N": ["13"]},
    "CENTRVM": {"-": ["12"]},
    "CENTVM": {"-": ["31"]},
    "CENTVMGEMINVS": {"-": ["21"]},
    "CENTVMPONDIVM": {"-": ["12"]},
    "CENTVMVIR": {"-": ["12"]},
    "CENTVMVIRALIS": {"-": ["24"]},
    "CENTVNCVLVS": {"-": ["12"]},
    "CENTVRIA": {"-": ["11"]},
    "CENTVRIATIM": {"-": ["60"]},
    "CENTVRIATVS": {"1": ["14"], "2": ["21"]},
    "CENTVRIO": {"1": ["13"], "2": ["51"]},
    "CENTVRIONATVS": {"-": ["14"]},
    "CENTVRIPINI": {"N": ["12"]},
    "CENTVRIPINVS": {"N": ["21"], "A": ["21"]},
    "CENTVSSIS": {"-": ["13"]},
    "CENVLA": {"-": ["11"]},
    "CEPHALOEDITANI": {"N": ["12"]},
    "CEPHALOEDITANVS": {"N": ["21"], "A": ["21"]},
    "CEPHALOEDIVM": {"N": ["12"]},
    "CEPTO": {"-": ["51"]},
    "CERA": {"-": ["11"]},
    "CERARIVM": {"-": ["12"]},
    "CERASTES": {"-": ["17"]},
    "CERASVM": {"-": ["12"]},
    "CERASVS": {"-": ["12"]},
    "CERBERVS": {"N": ["12"]},
    "CERCEI": {"N": ["12"]},
    "CERCOPITHECVS": {"-": ["12"]},
    "CERDO": {"-": ["13"], "N": ["13"]},
    "CEREALIS": {"-": ["24"], "A": ["24"]},
    "CEREBROSVS": {"-": ["21"]},
    "CEREBRVM": {"-": ["12"]},
    "CERES": {"N": ["13"], "-": ["23"]},
    "CEREVS": {"1": ["12"], "2": ["21"]},
    "CERINTHA": {"-": ["17"]},
    "CERNO": {"-": ["53"]},
    "CERNVLO": {"-": ["51"]},
    "CERO": {"-": ["51"]},
    "CEROMA": {"-": ["13"]},
    "CEROMATICVS": {"-": ["21"]},
    "CERPO": {"-": ["53"]},
    "CERRITVS": {"-": ["21"]},
    "CERRVS": {"-": ["12"]},
    "CERTAMEN": {"-": ["13"]},
    "CERTATIM": {"-": ["60"]},
    "CERTATIO": {"-": ["13"]},
    "CERTE": {"-": ["60"]},
    "CERTO": {"1": ["51"], "2": ["60"]},
    "CERTVM": {"1": ["12"], "2": ["60"]},
    "CERTVS": {"-": ["21"], "N": ["12"]},
    "CERVA": {"-": ["11"]},
    "CERVICAL": {"-": ["13"]},
    "CERVICVLA": {"-": ["11"]},
    "CERVINVS": {"-": ["21"]},
    "CERVIVS": {"N": ["12"]},
    "CERVIX": {"-": ["13"]},
    "CERVSSA": {"-": ["11"]},
    "CERVVS": {"-": ["12"]},
    "CERYX": {"-": ["13"]},
    "CESSATOR": {"-": ["13"]},
    "CESSIM": {"-": ["60"]},
    "CESSO": {"-": ["51"]},
    "CETARINI": {"N": ["12"]},
    "CETARIVM": {"-": ["12"]},
    "CETARIVS": {"1": ["12"], "2": ["21"], "-": ["21"]},
    "CETERA": {"1": ["12"], "2": ["60"]},
    "CETERI": {"-": ["12"]},
    "CETERVM": {"1": ["12"], "2": ["60"]},
    "CETERVS": {"-": ["21"]},
    "CETOS": {"-": ["17"]},
    "CETRA": {"-": ["11"]},
    "CETRATI": {"-": ["12"]},
    "CETRATVS": {"-": ["21"]},
    "CETVS": {"-": ["12"]},
    "CEV": {"1": ["66"], "2": ["82"]},
    "CEVEO": {"-": ["52"]},
    "CHAERE": {"-": ["90"]},
    "CHALCIDICVM": {"-": ["12"]},
    "CHALYBS": {"-": ["13"]},
    "CHAOS": {"-": ["17"], "N": ["12"]},
    "CHARA": {"-": ["11"]},
    "CHARACTERISMVS": {"-": ["17"]},
    "CHARIDEMVS": {"N": ["12"]},
    "CHARINVS": {"N": ["12"]},
    "CHARIS": {"-": ["17"], "N": ["16"]},
    "CHARTA": {"-": ["11"]},
    "CHARYBDIS": {"N": ["17"]},
    "CHEIROTOMETOS": {"-": ["27"]},
    "CHELE": {"-": ["17"], "N": ["17"]},
    "CHELIDON": {"-": ["17"], "N": ["17"]},
    "CHELONIVM": {"-": ["12"]},
    "CHELYDRVS": {"-": ["12"]},
    "CHELYS": {"-": ["17"]},
    "CHERAGRA": {"-": ["11"]},
    "CHERSONESVS": {"N": ["12"]},
    "CHILIARCHES": {"-": ["11"]},
    "CHILO": {"N": ["13"]},
    "CHIRAGRA": {"-": ["11"]},
    "CHIROGRAPHVM": {"-": ["12"]},
    "CHIRONOMON": {"-": ["17"]},
    "CHIVS": {"A": ["21"], "N": ["17"]},
    "CHLAMYDATVS": {"-": ["21"]},
    "CHLAMYS": {"-": ["17"]},
    "CHLORVS": {"N": ["12"]},
    "CHOO": {"-": ["51"]},
    "CHORAGIVM": {"-": ["12"]},
    "CHORAVLES": {"-": ["17"]},
    "CHORDA": {"-": ["11"]},
    "CHORDVS": {"-": ["21"]},
    "CHOREA": {"-": ["11"]},
    "CHOROBATES": {"-": ["17"]},
    "CHOROGRAPHIA": {"-": ["11"]},
    "CHORVS": {"-": ["12"]},
    "CHRIA": {"-": ["17"]},
    "CHRISTIANVS": {"1": ["12"], "2": ["21"], "N": ["12"]},
    "CHROMA": {"-": ["17"]},
    "CHROMATICE": {"-": ["17"]},
    "CHROMATICOS": {"-": ["27"]},
    "CHRYSAS": {"N": ["17"]},
    "CHRYSOCOLLA": {"-": ["11"]},
    "CHRYSOGONVS": {"N": ["12"]},
    "CHRYSOLITHVS": {"-": ["12"]},
    "CIBARIVM": {"-": ["12"]},
    "CIBARIVS": {"-": ["21"]},
    "CIBATVS": {"-": ["14"]},
    "CIBORIVM": {"-": ["12"]},
    "CIBVS": {"-": ["12"]},
    "CIBYRA": {"N": ["11"]},
    "CIBYRATA": {"N": ["11"]},
    "CIBYRATICVS": {"N": ["21"], "A": ["21"]},
    "CICADA": {"-": ["11"]},
    "CICATRICOSVS": {"-": ["21"]},
    "CICATRIX": {"-": ["13"]},
    "CICER": {"-": ["13"]},
    "CICERO": {"N": ["13"]},
    "CICHORIVM": {"-": ["12"]},
    "CICONIA": {"-": ["11"]},
    "CICVR": {"-": ["25"]},
    "CICVTA": {"-": ["11"], "N": ["11"]},
    "CIDARIS": {"-": ["13"]},
    "CIDO": {"1": ["53"], "2": ["53"]},
    "CIEO": {"-": ["52"]},
    "CILICES": {"N": ["13"]},
    "CILICIA": {"N": ["11"]},
    "CILICIVM": {"-": ["12"]},
    "CILIO": {"-": ["51"]},
    "CIMEX": {"-": ["13"]},
    "CINAEDVS": {"1": ["12"], "2": ["21"]},
    "CINCINNVS": {"-": ["12"]},
    "CINEFACTVS": {"-": ["21"]},
    "CINERARIVS": {"1": ["12"], "2": ["21"]},
    "CINESIS": {"-": ["17"]},
    "CINGO": {"-": ["53"]},
    "CINGVLA": {"-": ["11"]},
    "CINGVLVM": {"-": ["12"], "N": ["12"]},
    "CINIFLO": {"-": ["13"]},
    "CINIS": {"-": ["13"]},
    "CINNA": {"N": ["11"]},
    "CINNAMVM": {"-": ["17"]},
    "CINNO": {"-": ["51"]},
    "CINNVS": {"-": ["12"]},
    "CINO": {"-": ["53"]},
    "CIO": {"-": ["54"]},
    "CIPIO": {"-": ["55"]},
    "CIPPVS": {"-": ["12"]},
    "CIRCA": {"1": ["60"], "2": ["70"]},
    "CIRCE": {"N": ["17"]},
    "CIRCENSES": {"-": ["13"]},
    "CIRCENSIS": {"-": ["24"], "A": ["24"]},
    "CIRCINATIO": {"-": ["13"]},
    "CIRCINO": {"-": ["51"]},
    "CIRCINVS": {"-": ["12"]},
    "CIRCITER": {"1": ["60"], "2": ["70"]},
    "CIRCITO": {"1": ["5J", "51"]},
    "CIRCVITIO": {"-": ["13"]},
    "CIRCVITVS": {"-": ["14"]},
    "CIRCVLARIS": {"-": ["24"]},
    "CIRCVLATOR": {"-": ["13"]},
    "CIRCVLOR": {"-": ["5J"]},
    "CIRCVLVS": {"-": ["12"]},
    "CIRCVM": {"1": ["60"], "2": ["70"]},
    "CIRCVMACTIO": {"-": ["13"]},
    "CIRCVMACTVS": {"-": ["14"]},
    "CIRCVMAGO": {"-": ["53"]},
    "CIRCVMAMPLECTOR": {"-": ["5L"]},
    "CIRCVMCAESVRA": {"-": ["11"]},
    "CIRCVMCIDANEVS": {"-": ["21"]},
    "CIRCVMCIDO": {"2": ["53"], "1": ["53"]},
    "CIRCVMCINGO": {"-": ["53"]},
    "CIRCVMCIRCA": {"-": ["60"]},
    "CIRCVMCISVS": {"C": ["21"], "-": ["21"]},
    "CIRCVMCLVDO": {"-": ["53"]},
    "CIRCVMCOLO": {"2": ["53"]},
    "CIRCVMCVRRO": {"-": ["53"]},
    "CIRCVMCVRSO": {"-": ["51"]},
    "CIRCVMDO": {"-": ["51"]},
    "CIRCVMDVCO": {"-": ["53"]},
    "CIRCVMDVCTIO": {"-": ["13"]},
    "CIRCVMEO": {"1": ["56"]},
    "CIRCVMFERO": {"-": ["56"]},
    "CIRCVMFIGO": {"-": ["53"]},
    "CIRCVMFLECTO": {"-": ["53"]},
    "CIRCVMFLO": {"-": ["51"]},
    "CIRCVMFLVO": {"-": ["53"]},
    "CIRCVMFLVVS": {"-": ["21"]},
    "CIRCVMFODIO": {"-": ["55"]},
    "CIRCVMFREMO": {"-": ["53"]},
    "CIRCVMFRICO": {"-": ["51"]},
    "CIRCVMFVNDO": {"2": ["53"]},
    "CIRCVMGEMO": {"-": ["53"]},
    "CIRCVMGREDIOR": {"-": ["5N"]},
    "CIRCVMIACEO": {"-": ["52"]},
    "CIRCVMICIO": {"-": ["55"]},
    "CIRCVMLATIO": {"-": ["13"]},
    "CIRCVMLATRO": {"-": ["51"]},
    "CIRCVMLEGO": {"2": ["53"]},
    "CIRCVMLEVIGO": {"-": ["51"]},
    "CIRCVMLEVO": {"1": ["51"], "-": ["51"]},
    "CIRCVMLIGO": {"2": ["51"]},
    "CIRCVMLINO": {"-": ["53"]},
    "CIRCVMLITIO": {"-": ["13"]},
    "CIRCVMLOQVOR": {"-": ["5L"]},
    "CIRCVMLVCENS": {"-": ["25"]},
    "CIRCVMMETIO": {"-": ["54"]},
    "CIRCVMMITTO": {"-": ["53"]},
    "CIRCVMMVGIO": {"-": ["54"]},
    "CIRCVMMVNIO": {"2": ["54"]},
    "CIRCVMMVNITIO": {"-": ["13"]},
    "CIRCVMPENDEO": {"-": ["52"]},
    "CIRCVMPLECTO": {"-": ["53"]},
    "CIRCVMPLECTOR": {"-": ["5L"]},
    "CIRCVMPLVMBO": {"-": ["51"]},
    "CIRCVMPONO": {"-": ["53"]},
    "CIRCVMQVAQVE": {"-": ["60"]},
    "CIRCVMRETIO": {"-": ["54"]},
    "CIRCVMRODO": {"-": ["53"]},
    "CIRCVMSAEPIO": {"-": ["54"]},
    "CIRCVMSCINDO": {"-": ["53"]},
    "CIRCVMSCRIBO": {"-": ["53"]},
    "CIRCVMSCRIPTIO": {"-": ["13"]},
    "CIRCVMSCRIPTOR": {"-": ["13"]},
    "CIRCVMSECO": {"-": ["51"]},
    "CIRCVMSEDEO": {"-": ["52"]},
    "CIRCVMSERO": {"3": ["53"]},
    "CIRCVMSESSIO": {"-": ["13"]},
    "CIRCVMSIDO": {"-": ["53"]},
    "CIRCVMSILIO": {"-": ["54"]},
    "CIRCVMSISTO": {"-": ["53"]},
    "CIRCVMSONO": {"1": ["51"]},
    "CIRCVMSONVS": {"-": ["21"]},
    "CIRCVMSPECTE": {"-": ["6&"]},
    "CIRCVMSPECTO": {"-": ["51"]},
    "CIRCVMSPECTVS": {"1": ["14"], "2": ["21"]},
    "CIRCVMSPICIO": {"-": ["55"]},
    "CIRCVMSTO": {"-": ["51"]},
    "CIRCVMSTREPO": {"-": ["53"]},
    "CIRCVMTENDO": {"-": ["53"]},
    "CIRCVMTENEO": {"-": ["52"]},
    "CIRCVMTERGEO": {"-": ["52"]},
    "CIRCVMTERO": {"-": ["53"]},
    "CIRCVMTEXO": {"-": ["53"]},
    "CIRCVMTONDEO": {"-": ["52"]},
    "CIRCVMTONO": {"-": ["51"]},
    "CIRCVMVADO": {"2": ["53"]},
    "CIRCVMVAGOR": {"2": ["5J"], "-": ["5J"]},
    "CIRCVMVAGVS": {"-": ["21"]},
    "CIRCVMVALLO": {"-": ["51"]},
    "CIRCVMVECTO": {"-": ["51"]},
    "CIRCVMVECTOR": {"-": ["5J"]},
    "CIRCVMVEHOR": {"-": ["5L"]},
    "CIRCVMVENIO": {"-": ["54"]},
    "CIRCVMVERRO": {"-": ["53"]},
    "CIRCVMVERSO": {"-": ["51"]},
    "CIRCVMVERTO": {"-": ["53"]},
    "CIRCVMVOLITO": {"-": ["51"]},
    "CIRCVMVOLO": {"2": ["51"]},
    "CIRCVMVOLVO": {"-": ["53"]},
    "CIRCVS": {"-": ["12"], "N": ["12"]},
    "CIRRATI": {"-": ["12"]},
    "CIRRVS": {"-": ["12"]},
    "CIS": {"-": ["70"]},
    "CISIVM": {"-": ["12"]},
    "CISTA": {"-": ["11"]},
    "CISTERNA": {"-": ["11"]},
    "CISTERNINVS": {"-": ["21"]},
    "CITATIM": {"-": ["60"]},
    "CITATVS": {"-": ["21"]},
    "CITE": {"-": ["6&"]},
    "CITER": {"-": ["21"]},
    "CITERIA": {"-": ["11"]},
    "CITERIOR": {"-": ["2&"], "A": ["2&"]},
    "CITHARA": {"-": ["11"]},
    "CITHARISTA": {"-": ["11"]},
    "CITHAROEDVS": {"-": ["12"]},
    "CITO": {"1": ["51"], "2": ["60"]},
    "CITRA": {"1": ["60"], "2": ["70"]},
    "CITREVS": {"-": ["21"]},
    "CITRO": {"-": ["60"]},
    "CITRVM": {"-": ["12"]},
    "CITRVS": {"-": ["12"]},
    "CITVS": {"-": ["21"]},
    "CIVICA": {"-": ["11"], "N": ["11"]},
    "CIVICVS": {"-": ["21"]},
    "CIVILIS": {"-": ["24"], "N": ["13"]},
    "CIVILITER": {"-": ["60"]},
    "CIVIS": {"-": ["13"]},
    "CIVITAS": {"-": ["13"]},
    "CLADES": {"-": ["13"]},
    "CLAM": {"1": ["60"], "2": ["70"]},
    "CLAMITO": {"-": ["51"]},
    "CLAMO": {"-": ["51"]},
    "CLAMOR": {"-": ["13"]},
    "CLAMOSVS": {"-": ["21"]},
    "CLANCVLVM": {"-": ["60"]},
    "CLANDESTINVS": {"-": ["21"]},
    "CLANGOR": {"-": ["13"]},
    "CLARE": {"-": ["60"]},
    "CLAREO": {"-": ["52"]},
    "CLARESCO": {"-": ["53"]},
    "CLARICITO": {"-": ["51"]},
    "CLARIGO": {"-": ["51"]},
    "CLARISONVS": {"-": ["21"]},
    "CLARITAS": {"-": ["13"]},
    "CLARITVDO": {"-": ["13"]},
    "CLARO": {"-": ["51"]},
    "CLARVS": {"-": ["21"], "N": ["12"]},
    "CLASSIARII": {"-": ["12"]},
    "CLASSIARIVS": {"-": ["21"]},
    "CLASSICI": {"-": ["12"]},
    "CLASSICVM": {"-": ["12"]},
    "CLASSICVS": {"1": ["12"], "2": ["21"], "N": ["12"]},
    "CLASSIS": {"-": ["13"]},
    "CLATHRATVS": {"-": ["21"]},
    "CLATHRI": {"-": ["12"]},
    "CLAVA": {"-": ["11"]},
    "CLAVDICO": {"-": ["51"]},
    "CLAVDIVS": {"A": ["21"], "N": ["12"]},
    "CLAVDO": {"1": ["53"], "2": ["53"]},
    "CLAVDVS": {"-": ["21"]},
    "CLAVICVLA": {"-": ["11"]},
    "CLAVIS": {"-": ["13"]},
    "CLAVSTRVM": {"-": ["12"]},
    "CLAVSVLA": {"-": ["11"]},
    "CLAVSVM": {"-": ["12"]},
    "CLAVSVS": {"N": ["12"]},
    "CLAVVLVS": {"-": ["12"]},
    "CLAVVS": {"-": ["12"]},
    "CLEMENS": {"-": ["25"], "N": ["13"]},
    "CLEMENTER": {"-": ["60"]},
    "CLEMENTIA": {"-": ["11"]},
    "CLEOMENES": {"N": ["17"]},
    "CLEPO": {"-": ["53"]},
    "CLEPSYDRA": {"-": ["11"]},
    "CLIENS": {"-": ["13"]},
    "CLIENTA": {"-": ["11"]},
    "CLIENTELA": {"-": ["11"]},
    "CLIENTVLVS": {"-": ["12"]},
    "CLIMA": {"-": ["17"]},
    "CLIMACHIAS": {"N": ["17"]},
    "CLIMACIS": {"-": ["17"]},
    "CLINAMEN": {"-": ["13"]},
    "CLINO": {"-": ["51"]},
    "CLIPEATVS": {"-": ["12"], "2": ["12"], "1": ["12"]},
    "CLIPEO": {"-": ["51"]},
    "CLIPEVS": {"-": ["12"]},
    "CLITELLAE": {"-": ["11"]},
    "CLITELLARIVS": {"-": ["21"]},
    "CLIVA": {"-": ["12"]},
    "CLIVOSVS": {"-": ["21"]},
    "CLIVVS": {"-": ["12"]},
    "CLOACA": {"-": ["11"]},
    "CLODIVS": {"N": ["12"], "A": ["21"]},
    "CLOELIA": {"N": ["11"]},
    "CLOELIVS": {"A": ["21"], "N": ["12"]},
    "CLVDO": {"-": ["53"]},
    "CLVEO": {"-": ["52"]},
    "CLVNIS": {"-": ["13"]},
    "CLVPEA": {"N": ["16"]},
    "CLVSINVS": {"N": ["21"], "A": ["21"]},
    "CNEMIS": {"-": ["17"]},
    "CNIDII": {"N": ["12"]},
    "CNODAX": {"-": ["17"]},
    "COACERVATIO": {"-": ["13"]},
    "COACERVO": {"-": ["51"]},
    "COACTA": {"-": ["12"]},
    "COACTO": {"-": ["51"]},
    "COACTOR": {"-": ["13"]},
    "COACTVS": {"-": ["14"]},
    "COADDO": {"-": ["53"]},
    "COAEDIFICO": {"-": ["51"]},
    "COAEQVO": {"-": ["51"]},
    "COAGMENTATIO": {"-": ["13"]},
    "COAGMENTO": {"-": ["51"]},
    "COAGMENTVM": {"-": ["12"]},
    "COAGVLO": {"-": ["51"]},
    "COAGVLVM": {"-": ["12"]},
    "COALESCO": {"-": ["53"]},
    "COANGVSTO": {"-": ["51"]},
    "COARCTATIO": {"-": ["13"]},
    "COARCTO": {"-": ["51"]},
    "COAREO": {"-": ["52"]},
    "COARGVO": {"-": ["53"]},
    "COARTATIO": {"-": ["13"]},
    "COARTO": {"-": ["51"]},
    "COAXATIO": {"-": ["13"]},
    "COAXO": {"-": ["51"]},
    "COCCEIVS": {"N": ["12"]},
    "COCCINVS": {"-": ["21"]},
    "COCCVM": {"-": ["12"]},
    "COCHLEA": {"-": ["11"]},
    "COCLES": {"N": ["13"]},
    "COCTILIS": {"-": ["24"]},
    "COCTIO": {"-": ["13"]},
    "COCTVRA": {"-": ["11"]},
    "COCTVRNIX": {"-": ["13"]},
    "COCVLVM": {"-": ["12"]},
    "CODEX": {"-": ["13"]},
    "CODICILLVS": {"-": ["12"]},
    "COELIA": {"-": ["17"]},
    "COELIACVS": {"-": ["21"]},
    "COELIVS": {"N": ["12"]},
    "COEMO": {"-": ["53"]},
    "COEO": {"1": ["56"]},
    "COEPIO": {"-": ["55"]},
    "COEPTO": {"-": ["51"]},
    "COEPTVM": {"-": ["12"]},
    "COEPTVS": {"-": ["14"]},
    "COERCEO": {"-": ["52"]},
    "COERCITIO": {"-": ["13"]},
    "COETVS": {"-": ["14"]},
    "COGITABILIS": {"-": ["24"]},
    "COGITATIO": {"-": ["13"]},
    "COGITATVM": {"-": ["12"]},
    "COGITO": {"-": ["51"]},
    "COGNATIO": {"-": ["13"]},
    "COGNATVS": {"1": ["12"], "2": ["21"]},
    "COGNITIO": {"-": ["13"]},
    "COGNITOR": {"-": ["13"]},
    "COGNITVS": {"1": ["14"], "2": ["21"]},
    "COGNOBILIS": {"-": ["24"]},
    "COGNOMEN": {"-": ["13"]},
    "COGNOMENTVM": {"-": ["12"]},
    "COGNOSCO": {"-": ["53"]},
    "COGO": {"-": ["53"]},
    "COHAERENTIA": {"-": ["11"]},
    "COHAEREO": {"-": ["52"]},
    "COHAERESCO": {"-": ["53"]},
    "COHERES": {"-": ["13"]},
    "COHIBEO": {"-": ["52"]},
    "COHONESTO": {"-": ["51"]},
    "COHORS": {"-": ["13"]},
    "COHORTATIO": {"-": ["13"]},
    "COHORTOR": {"-": ["5J"]},
    "COI": {"N": ["12"]},
    "COICIO": {"-": ["55"]},
    "COINQVINO": {"-": ["51"]},
    "COITVS": {"-": ["14"]},
    "COLAPHVS": {"-": ["12"], "N": ["12"]},
    "COLLABEFACIO": {"-": ["55"]},
    "COLLABEFIO": {"-": ["56"]},
    "COLLABOR": {"2": ["5L"]},
    "COLLACTEVS": {"-": ["21"]},
    "COLLAEVO": {"-": ["51"]},
    "COLLATIA": {"N": ["11"]},
    "COLLATICIVS": {"-": ["21"]},
    "COLLATINI": {"N": ["12"]},
    "COLLATINVS": {"A": ["21"], "N": ["12"]},
    "COLLATIO": {"-": ["13"]},
    "COLLATRO": {"-": ["51"]},
    "COLLATVS": {"-": ["14"]},
    "COLLAVDO": {"-": ["51"]},
    "COLLAXO": {"-": ["51"]},
    "COLLECTIO": {"-": ["13"]},
    "COLLECTVS": {"1": ["14"], "2": ["21"]},
    "COLLEGA": {"-": ["11"], "N": ["11"]},
    "COLLEGIVM": {"-": ["12"]},
    "COLLEVO": {"2": ["51"]},
    "COLLIBERTVS": {"-": ["12"]},
    "COLLIBET": {"-": ["52"]},
    "COLLIBRO": {"-": ["51"]},
    "COLLICIAE": {"-": ["11"]},
    "COLLICIARIS": {"-": ["24"]},
    "COLLIDO": {"-": ["53"]},
    "COLLIGO": {"2": ["51"], "3": ["53"]},
    "COLLINO": {"-": ["53"]},
    "COLLINVS": {"N": ["21"], "A": ["21"]},
    "COLLIQVIARIA": {"-": ["12"]},
    "COLLIS": {"-": ["13"]},
    "COLLOCATIO": {"-": ["13"]},
    "COLLOCO": {"-": ["51"]},
    "COLLOQVIVM": {"-": ["12"]},
    "COLLOQVOR": {"-": ["5L"]},
    "COLLVCEO": {"-": ["52"]},
    "COLLVCO": {"-": ["51"]},
    "COLLVCTATIO": {"-": ["13"]},
    "COLLVCTOR": {"-": ["5J"]},
    "COLLVDO": {"-": ["53"]},
    "COLLVM": {"-": ["12"]},
    "COLLVO": {"-": ["53"]},
    "COLLVS": {"-": ["12"]},
    "COLLVSIO": {"-": ["13"]},
    "COLLVSOR": {"-": ["13"]},
    "COLLVSTRO": {"2": ["51"]},
    "COLLVVIES": {"-": ["15"]},
    "COLLVVIO": {"-": ["13"]},
    "COLLYBVS": {"-": ["12"]},
    "COLLYRIVM": {"-": ["12"]},
    "COLO": {"1": ["51"], "2": ["53"]},
    "COLOCASIVM": {"-": ["12"]},
    "COLOCYNTHA": {"-": ["11"]},
    "COLOEPHIVM": {"-": ["12"]},
    "COLONIA": {"-": ["11"], "N": ["11"]},
    "COLONVS": {"1": ["12"], "2": ["21"]},
    "COLOR": {"-": ["13"]},
    "COLORATVS": {"C": ["21"], "-": ["21"]},
    "COLORO": {"-": ["51"]},
    "COLOSSICVS": {"-": ["2G"]},
    "COLOSSVS": {"-": ["12"]},
    "COLVBER": {"-": ["12"]},
    "COLVBRA": {"-": ["11"]},
    "COLVM": {"-": ["12"]},
    "COLVMBA": {"-": ["11"]},
    "COLVMBARIVM": {"-": ["12"]},
    "COLVMBINVS": {"-": ["21"]},
    "COLVMBOR": {"-": ["5J"]},
    "COLVMBVS": {"-": ["12"]},
    "COLVMELLA": {"-": ["11"]},
    "COLVMEN": {"-": ["13"]},
    "COLVMNA": {"-": ["11"]},
    "COLVMNARIVM": {"-": ["12"]},
    "COLVMNATVS": {"-": ["21"]},
    "COLVRNVS": {"-": ["21"]},
    "COLVS": {"-": ["16"]},
    "COMA": {"-": ["11"]},
    "COMANS": {"-": ["25"]},
    "COMATVS": {"-": ["21"], "A": ["21"]},
    "COMBIBO": {"1": ["13"], "2": ["53"]},
    "COMBVRO": {"-": ["53"]},
    "COMEDO": {"1": ["13"], "2": ["56"]},
    "COMES": {"-": ["13"]},
    "COMETES": {"-": ["17"]},
    "COMICE": {"-": ["60"]},
    "COMICVS": {"1": ["12"], "2": ["21"]},
    "COMINIVS": {"N": ["12"]},
    "COMINVS": {"-": ["60"]},
    "COMIS": {"-": ["24"]},
    "COMISSABVNDVS": {"-": ["21"]},
    "COMISSATIO": {"-": ["13"]},
    "COMISSOR": {"-": ["5J"]},
    "COMITAS": {"-": ["13"]},
    "COMITATVS": {"1": ["14"], "2": ["21"]},
    "COMITER": {"-": ["60"]},
    "COMITIA": {"-": ["12"]},
    "COMITIALIS": {"1": ["13"], "2": ["24"]},
    "COMITIATVS": {"-": ["14"]},
    "COMITIO": {"-": ["51"]},
    "COMITIVM": {"-": ["12"]},
    "COMITO": {"-": ["51"]},
    "COMITOR": {"-": ["5J"]},
    "COMMACVLO": {"-": ["51"]},
    "COMMADEO": {"-": ["52"]},
    "COMMEATVS": {"-": ["14"]},
    "COMMEDITOR": {"-": ["5J"]},
    "COMMEMORATIO": {"-": ["13"]},
    "COMMEMORO": {"-": ["51"]},
    "COMMENDATICIVS": {"-": ["21"]},
    "COMMENDATIO": {"-": ["13"]},
    "COMMENDO": {"-": ["51"]},
    "COMMENSVS": {"-": ["14"]},
    "COMMENTARIVM": {"-": ["12"]},
    "COMMENTARIVS": {"1": ["12"], "2": ["21"]},
    "COMMENTICIVS": {"-": ["21"]},
    "COMMENTOR": {"1": ["13"], "2": ["5J"]},
    "COMMENTVM": {"-": ["12"]},
    "COMMEO": {"-": ["51"]},
    "COMMERCIVM": {"-": ["12"]},
    "COMMERCOR": {"-": ["5J"]},
    "COMMEREO": {"-": ["52"]},
    "COMMEREOR": {"-": ["5K"]},
    "COMMERITVM": {"-": ["12"]},
    "COMMIGRATIO": {"-": ["13"]},
    "COMMIGRO": {"-": ["51"]},
    "COMMILITIVM": {"-": ["12"]},
    "COMMILITO": {"1": ["13"], "2": ["51"]},
    "COMMINATIO": {"-": ["13"]},
    "COMMINGO": {"-": ["53"]},
    "CO": {"-": ["5L"]},
    "COMMINOR": {"-": ["5J"]},
    "COMMINVO": {"-": ["53"]},
    "COMMINVS": {"-": ["60"]},
    "COMMIS": {"-": ["13"]},
    "COMMISCEO": {"-": ["52"]},
    "COMMISSIO": {"-": ["13"]},
    "COMMISSVM": {"-": ["12"]},
    "COMMISSVRA": {"-": ["11"]},
    "COMMITTO": {"-": ["53"]},
    "COMMODE": {"-": ["60"]},
    "COMMODITAS": {"-": ["13"]},
    "COMMODO": {"1": ["51"], "2": ["60"]},
    "COMMODVLATIO": {"-": ["13"]},
    "COMMODVM": {"1": ["12"], "2": ["60"]},
    "COMMODVS": {"-": ["21"]},
    "COMMOLIOR": {"-": ["5M"]},
    "COMMONEFACIO": {"-": ["55"]},
    "COMMONEFIO": {"-": ["56"]},
    "COMMONEO": {"-": ["52"]},
    "COMMONSTRO": {"-": ["51"]},
    "COMMORDEO": {"-": ["52"]},
    "COMMORIOR": {"-": ["5N"]},
    "COMMOROR": {"1": ["5J"]},
    "COMMOTIO": {"-": ["13"]},
    "COMMOTVS": {"-": ["21"]},
    "COMMOVEO": {"-": ["52"]},
    "COMMVNE": {"C": ["13"], "-": ["13"]},
    "COMMVNICATIO": {"-": ["13"]},
    "COMMVNICO": {"-": ["51"]},
    "COMMVNIO": {"1": ["13"], "2": ["54"]},
    "COMMVNIS": {"-": ["24"]},
    "COMMVNITAS": {"-": ["13"]},
    "COMMVNITER": {"-": ["60"]},
    "COMMVNITIO": {"-": ["13"]},
    "COMMVTABILIS": {"-": ["24"]},
    "COMMVTATIO": {"-": ["13"]},
    "COMMVTATVS": {"-": ["14"]},
    "COMMVTO": {"2": ["51"]},
    "COMO": {"1": ["51"], "2": ["53"]},
    "COMOEDIA": {"-": ["11"]},
    "COMOEDVS": {"1": ["12"], "2": ["21"]},
    "COMPACTIO": {"-": ["13"]},
    "COMPACTVRA": {"-": ["11"]},
    "COMPACTVS": {"-": ["21"]},
    "COMPAGES": {"-": ["13"]},
    "COMPAGO": {"-": ["13"]},
    "COMPAR": {"1": ["13"], "2": ["25"]},
    "COMPARATIO": {"1": ["13"], "2": ["13"]},
    "COMPARATVS": {"-": ["14"]},
    "COMPAREO": {"-": ["52"]},
    "COMPARO": {"2": ["51"], "3": ["51"]},
    "COMPEDIO": {"-": ["54"]},
    "COMPELLO": {"1": ["51"], "2": ["53"]},
    "COMPENDIARIA": {"-": ["11"]},
    "COMPENDIARIVM": {"-": ["12"]},
    "COMPENDIARIVS": {"-": ["21"]},
    "COMPENDIVM": {"-": ["12"]},
    "COMPENDO": {"-": ["53"]},
    "COMPENSATIO": {"-": ["13"]},
    "COMPENSO": {"-": ["51"]},
    "COMPERENDINATIO": {"-": ["13"]},
    "COMPERENDINATVS": {"-": ["14"]},
    "COMPERENDINO": {"-": ["51"]},
    "COMPERIO": {"-": ["54"]},
    "COMPERTVS": {"-": ["21"]},
    "COMPESCO": {"-": ["53"]},
    "COMPETITOR": {"-": ["13"]},
    "COMPETO": {"-": ["53"]},
    "COMPILATIO": {"-": ["13"]},
    "COMPILO": {"1": ["51"], "2": ["51"]},
    "COMPINGO": {"1": ["53"], "2": ["53"]},
    "COMPITVM": {"-": ["12"]},
    "COMPLANO": {"-": ["51"]},
    "COMPLECTO": {"2": ["53"]},
    "COMPLECTOR": {"-": ["5L"]},
    "COMPLEMENTVM": {"-": ["12"]},
    "COMPLEO": {"-": ["52"]},
    "COMPLETVS": {"-": ["21"]},
    "COMPLEXIO": {"-": ["13"]},
    "COMPLEXVS": {"-": ["14"]},
    "COMPLICO": {"-": ["51"]},
    "COMPLODO": {"-": ["53"]},
    "COMPLORATIO": {"-": ["13"]},
    "COMPLORATVS": {"-": ["14"]},
    "COMPLORO": {"-": ["51"]},
    "COMPLVRES": {"1": ["13"], "2": ["22"]},
    "COMPLVRIES": {"-": ["60"]},
    "COMPLVVIVM": {"-": ["12"]},
    "COMPONO": {"-": ["53"]},
    "COMPORTATIO": {"-": ["13"]},
    "COMPORTO": {"-": ["51"]},
    "COMPOS": {"-": ["22"]},
    "COMPOSITE": {"-": ["60"]},
    "COMPOSITIO": {"-": ["13"]},
    "COMPOSITVM": {"-": ["12"]},
    "COMPOSITVRA": {"-": ["11"]},
    "COMPOSITVS": {"O": ["21"], "-": ["21"]},
    "COMPRECOR": {"-": ["5J"]},
    "COMPREHENDO": {"-": ["53"]},
    "COMPREHENSIBILIS": {"-": ["24"]},
    "COMPREHENSIO": {"-": ["13"]},
    "COMPRESSIO": {"-": ["13"]},
    "COMPRESSO": {"-": ["51"]},
    "COMPRESSVS": {"1": ["14"], "2": ["21"]},
    "COMPRIMO": {"-": ["53"]},
    "COMPROBO": {"-": ["51"]},
    "COMPROMISSVM": {"-": ["12"]},
    "COMPSANI": {"N": ["12"]},
    "COMPSANVS": {"N": ["21"], "A": ["21"]},
    "COMPTE": {"-": ["60"]},
    "COMPTVLVS": {"-": ["21"]},
    "COMPTVS": {"1": ["14"], "2": ["21"]},
    "COMPVNGO": {"-": ["53"]},
    "COMPVTATIO": {"-": ["13"]},
    "COMPVTATOR": {"-": ["13"]},
    "COMPVTO": {"-": ["51"]},
    "COMPVTRESCO": {"-": ["53"]},
    "CONAMEN": {"-": ["13"]},
    "CONARACHNE": {"-": ["17"]},
    "CONATA": {"-": ["12"]},
    "CONATVS": {"-": ["14"]},
    "CONCAEDES": {"-": ["13"]},
    "CONCALEFACIO": {"-": ["55"]},
    "CONCALEO": {"-": ["52"]},
    "CONCALESCO": {"-": ["53"]},
    "CONCALLEO": {"-": ["52"]},
    "CONCAMERATIO": {"-": ["13"]},
    "CONCAMERO": {"-": ["51"]},
    "CONCAVO": {"1": ["51"]},
    "CONCAVVS": {"-": ["21"]},
    "CONCEDO": {"1": ["53"]},
    "CONCELEBRO": {"-": ["51"]},
    "CONCENTIO": {"-": ["13"]},
    "CONCENTVS": {"-": ["14"]},
    "CONCEPTIO": {"-": ["13"]},
    "CONCEPTO": {"-": ["55"]},
    "CONCEPTVS": {"-": ["14"], "2": ["14"], "1": ["12"]},
    "CONCERNO": {"-": ["53"]},
    "CONCERTATOR": {"-": ["13"]},
    "CONCERTO": {"1": ["51"]},
    "CONCESSIO": {"-": ["13"]},
    "CONCESSVM": {"-": ["12"]},
    "CONCESSVS": {"-": ["14"]},
    "CONCHA": {"-": ["11"]},
    "CONCHIS": {"-": ["13"]},
    "CONCHYLIATVS": {"-": ["21"]},
    "CONCHYLIVM": {"-": ["12"]},
    "CONCIDO": {"1": ["53"], "2": ["53"]},
    "CONCIEO": {"-": ["52"]},
    "CONCILIABVLVM": {"-": ["12"]},
    "CONCILIATIO": {"-": ["13"]},
    "CONCILIATOR": {"-": ["13"]},
    "CONCILIATRIX": {"-": ["13"]},
    "CONCILIATVRA": {"-": ["11"]},
    "CONCILIATVS": {"-": ["14"]},
    "CONCILIO": {"-": ["51"]},
    "CONCILIVM": {"-": ["12"]},
    "CONCINNITAS": {"-": ["13"]},
    "CONCINNO": {"-": ["51"]},
    "CONCINNVS": {"-": ["21"]},
    "CONCINO": {"-": ["53"]},
    "CONCIO": {"-": ["54"]},
    "CONCIPIO": {"-": ["55"]},
    "CONCISVRA": {"-": ["11"]},
    "CONCISVS": {"-": ["21"]},
    "CONCITAMENTVM": {"-": ["12"]},
    "CONCITATE": {"-": ["60"]},
    "CONCITATIO": {"-": ["13"]},
    "CONCITATOR": {"-": ["13"]},
    "CONCITATVS": {"-": ["21"]},
    "CONCITO": {"1": ["51"]},
    "CONCITOR": {"-": ["13"]},
    "CONCLAMATIO": {"-": ["13"]},
    "CONCLAMO": {"-": ["51"]},
    "CONCLAVE": {"-": ["16"]},
    "CONCLVDO": {"-": ["53"]},
    "CONCLVSIO": {"-": ["13"]},
    "CONCLVSVRA": {"-": ["11"]},
    "CONCLVSVS": {"1": ["14"], "2": ["21"]},
    "CONCOLOR": {"-": ["25"]},
    "CONCOQVO": {"-": ["53"]},
    "CONCORDIA": {"-": ["11"], "N": ["11"]},
    "CONCORDITER": {"-": ["6-"]},
    "CONCORDO": {"-": ["51"]},
    "CONCORS": {"-": ["25"]},
    "CONCREDO": {"-": ["53"]},
    "CONCREMO": {"-": ["51"]},
    "CONCREPITO": {"-": ["51"]},
    "CONCREPO": {"-": ["51"]},
    "CONCRESCENTIA": {"-": ["11"]},
    "CONCRESCO": {"-": ["53"]},
    "CONCRETVS": {"1": ["14"], "2": ["21"]},
    "CONCRISPO": {"-": ["51"]},
    "CONCRVCIO": {"-": ["51"]},
    "CONCVBINA": {"-": ["11"]},
    "CONCVBINVS": {"-": ["12"]},
    "CONCVBITVS": {"-": ["14"]},
    "CONCVBIVS": {"-": ["21"]},
    "CONCVLCO": {"-": ["51"]},
    "CONCVMBO": {"-": ["53"]},
    "CONCVPIO": {"-": ["55"]},
    "CONCVPISCO": {"-": ["53"]},
    "CONCVPITOR": {"-": ["13"]},
    "CONCVRRO": {"-": ["53"]},
    "CONCVRSATIO": {"-": ["13"]},
    "CONCVRSATOR": {"-": ["13"]},
    "CONCVRSIO": {"-": ["13"]},
    "CONCVRSO": {"-": ["51"]},
    "CONCVRSVS": {"-": ["14"]},
    "CONCVSSVS": {"-": ["14"]},
    "CONCVTIO": {"-": ["55"]},
    "CONDECORO": {"-": ["51"]},
    "CONDELIQVESCO": {"-": ["53"]},
    "CONDEMNATOR": {"-": ["13"]},
    "CONDEMNO": {"-": ["51"]},
    "CONDENSEO": {"-": ["52"]},
    "CONDENSO": {"-": ["51"]},
    "CONDENSVS": {"-": ["21"]},
    "CONDEPSO": {"-": ["53"]},
    "CONDICIO": {"-": ["13"]},
    "CONDICO": {"2": ["53"]},
    "CONDIGNVS": {"-": ["21"]},
    "CONDIMENTVM": {"-": ["12"]},
    "CONDIO": {"-": ["54"]},
    "CONDISCIPVLVS": {"-": ["12"]},
    "CONDISCO": {"-": ["53"]},
    "CONDITIVVM": {"-": ["12"]},
    "CONDITIVVS": {"-": ["21"]},
    "CONDITOR": {"1": ["13"], "2": ["13"]},
    "CONDITVRA": {"1": ["11"], "2": ["11"]},
    "CONDITVS": {"1": ["14"], "2": ["14"], "3": ["21"], "4": ["21"]},
    "CONDO": {"-": ["53"]},
    "CONDOCEFACIO": {"-": ["55"]},
    "CONDOLEO": {"-": ["52"]},
    "CONDOLESCO": {"-": ["53"]},
    "CONDONATIO": {"-": ["13"]},
    "CONDONO": {"-": ["51"]},
    "CONDORMIO": {"-": ["54"]},
    "CONDVCO": {"-": ["53"]},
    "CONDVCTIO": {"-": ["13"]},
    "CONDVCTOR": {"-": ["13"]},
    "CONDVCTVM": {"-": ["12"]},
    "CONDVPLICO": {"-": ["51"]},
    "CONDVRO": {"-": ["51"]},
    "CONECTO": {"-": ["53"]},
    "CONFARREATIO": {"-": ["13"]},
    "CONFARREO": {"-": ["51"]},
    "CONFECTIO": {"-": ["13"]},
    "CONFECTOR": {"-": ["13"]},
    "CONFERCIO": {"-": ["54"]},
    "CONFERO": {"-": ["56"]},
    "CONFERTIM": {"-": ["60"]},
    "CONFERTVS": {"-": ["21"]},
    "CONFERVEFACIO": {"-": ["55"]},
    "CONFERVESCO": {"-": ["53"]},
    "CONFESSIO": {"-": ["13"]},
    "CONFESSVM": {"-": ["12"]},
    "CONFESSVS": {"-": ["21"]},
    "CONFESTIM": {"-": ["60"]},
    "CONFICIENS": {"-": ["25"]},
    "CONFICIO": {"-": ["55"]},
    "CONFIDENS": {"-": ["25"]},
    "CONFIDENTIA": {"-": ["11"]},
    "CONFIDO": {"-": ["5T"]},
    "CONFIGO": {"-": ["53"]},
    "CONFINE": {"-": ["13"]},
    "CONFINGO": {"-": ["53"]},
    "CONFINIS": {"1": ["13"], "2": ["24"]},
    "CONFINIVM": {"-": ["12"]},
    "CONFIO": {"-": ["56"]},
    "CONFIRMATIO": {"-": ["13"]},
    "CONFIRMATVS": {"-": ["21"]},
    "CONFIRMO": {"-": ["51"]},
    "CONFITEOR": {"-": ["5K"]},
    "CONFLAGRO": {"-": ["51"]},
    "CONFLICTO": {"-": ["51"]},
    "CONFLICTVS": {"-": ["14"]},
    "CONFLIGO": {"-": ["53"]},
    "CONFLO": {"-": ["51"]},
    "CONFLVENS": {"-": ["13"]},
    "CONFLVENTES": {"N": ["13"]},
    "CONFLVO": {"-": ["53"]},
    "CONFODIO": {"-": ["55"]},
    "CONFORMATIO": {"-": ["13"]},
    "CONFORMO": {"-": ["51"]},
    "CONFORNICATIO": {"-": ["13"]},
    "CONFORNICO": {"-": ["51"]},
    "CONFRAGOSVM": {"-": ["12"]},
    "CONFRAGOSVS": {"-": ["21"]},
    "CONFRICO": {"-": ["51"]},
    "CONFRINGO": {"-": ["53"]},
    "CONFRIO": {"-": ["51"]},
    "CONFVGIO": {"-": ["55"]},
    "CONFVLCIO": {"-": ["54"]},
    "CONFVNDO": {"2": ["53"]},
    "CONFVSE": {"-": ["60"]},
    "CONFVSIO": {"-": ["13"]},
    "CONFVSVS": {"-": ["21"]},
    "CONFVTO": {"-": ["51"]},
    "CONFVTVO": {"-": ["53"]},
    "CONGELO": {"-": ["51"]},
    "CONGEMO": {"-": ["53"]},
    "CONGERIES": {"-": ["15"]},
    "CONGERO": {"1": ["13"], "2": ["53"]},
    "CONGESTICIVS": {"-": ["21"]},
    "CONGESTIO": {"-": ["13"]},
    "CONGESTVS": {"1": ["14"], "2": ["21"]},
    "CONGIARIVM": {"-": ["12"]},
    "CONGIARIVS": {"-": ["21"]},
    "CONGIVS": {"-": ["12"]},
    "CONGLOBATIO": {"-": ["13"]},
    "CONGLOBO": {"-": ["51"]},
    "CONGLOMERO": {"-": ["51"]},
    "CONGLVTINO": {"-": ["51"]},
    "CONGREDIOR": {"-": ["5N"]},
    "CONGREGATIO": {"-": ["13"]},
    "CONGREGO": {"-": ["51"]},
    "CONGRESSVS": {"-": ["14"]},
    "CONGRVENS": {"-": ["25"]},
    "CONGRVO": {"-": ["53"]},
    "CONGRVVS": {"-": ["21"]},
    "CONICIO": {"-": ["55"]},
    "CONICVS": {"-": ["21"]},
    "CONIECTIO": {"-": ["13"]},
    "CONIECTO": {"-": ["51"]},
    "CONIECTVRA": {"-": ["11"]},
    "CONIECTVS": {"-": ["14"]},
    "CONIFER": {"-": ["21"]},
    "CONIGER": {"-": ["21"]},
    "CONISTERIVM": {"-": ["12"]},
    "CONIVGALIS": {"-": ["24"]},
    "CONIVGATOR": {"-": ["13"]},
    "CONIVGIVM": {"-": ["12"]},
    "CONIVGVLVS": {"-": ["21"]},
    "CONIVNCTE": {"-": ["6-"]},
    "CONIVNCTIM": {"-": ["60"]},
    "CONIVNCTIO": {"-": ["13"]},
    "CONIVNCTVM": {"-": ["12"]},
    "CONIVNCTVS": {"-": ["21"]},
    "CONIVNGO": {"-": ["53"]},
    "CONIVOLVS": {"-": ["21"]},
    "CONIVRATI": {"-": ["12"]},
    "CONIVRATIM": {"-": ["60"]},
    "CONIVRATIO": {"-": ["13"]},
    "CONIVRATOR": {"-": ["13"]},
    "CONIVRATVS": {"-": ["21"]},
    "CONIVRO": {"-": ["51"]},
    "CONIVX": {"-": ["13"]},
    "CONNECTO": {"-": ["53"]},
    "CONNEXVS": {"1": ["14"], "2": ["21"]},
    "CONNITOR": {"2": ["5L"]},
    "CONNIVEO": {"-": ["52"]},
    "CONNVBIVM": {"-": ["12"]},
    "CONOPEVM": {"-": ["12"]},
    "CONOPIVM": {"-": ["12"]},
    "CONOR": {"-": ["5J"]},
    "CONQVASSO": {"-": ["51"]},
    "CONQVEROR": {"-": ["5L"]},
    "CONQVESTIO": {"-": ["13"]},
    "CONQVIESCO": {"-": ["53"]},
    "CONQVIRO": {"-": ["53"]},
    "CONQVISITIO": {"-": ["13"]},
    "CONQVISITOR": {"-": ["13"]},
    "CONQVISITVS": {"-": ["21"]},
    "CONSAEPTVM": {"-": ["12"]},
    "CONSAEPTVS": {"1": ["14"], "2": ["21"]},
    "CONSALVTATIO": {"-": ["13"]},
    "CONSALVTO": {"-": ["51"]},
    "CONSANGVINEA": {"-": ["11"]},
    "CONSANGVINEI": {"-": ["12"]},
    "CONSANGVINEVS": {"1": ["12"], "2": ["21"]},
    "CONSANGVINITAS": {"-": ["13"]},
    "CONSARIO": {"-": ["54"]},
    "CONSCELERATVS": {"-": ["21"]},
    "CONSCELERO": {"-": ["51"]},
    "CONSCENDO": {"-": ["53"]},
    "CONSCIENTIA": {"-": ["11"]},
    "CONSCIO": {"-": ["54"]},
    "CONSCISCO": {"-": ["53"]},
    "CONSCIVS": {"-": ["21"]},
    "CONSCRIBILLO": {"-": ["51"]},
    "CONSCRIBO": {"-": ["53"]},
    "CONSCRIPTIO": {"-": ["13"]},
    "CONSECO": {"-": ["51"]},
    "CONSECRATIO": {"-": ["13"]},
    "CONSECRATVS": {"-": ["21"]},
    "CONSECRO": {"-": ["51"]},
    "CONSECTATIO": {"-": ["13"]},
    "CONSECTOR": {"2": ["5J"]},
    "CONSECVE": {"-": ["60"]},
    "CONSEDO": {"-": ["51"]},
    "CONSENESCO": {"-": ["53"]},
    "CONSENSIO": {"-": ["13"]},
    "CONSENSVS": {"-": ["14"]},
    "CONSENTANEVS": {"-": ["21"]},
    "CONSENTIENS": {"-": ["25"]},
    "CONSENTIO": {"-": ["54"]},
    "CONSEQVENS": {"1": ["13"], "2": ["25"]},
    "CONSEQVOR": {"-": ["5L"]},
    "CONSERO": {"2": ["53"], "3": ["53"]},
    "CONSERVATIO": {"-": ["13"]},
    "CONSERVATOR": {"-": ["13"], "N": ["13"]},
    "CONSERVO": {"-": ["51"]},
    "CONSERVVS": {"1": ["12"], "2": ["21"]},
    "CONSESSVS": {"-": ["14"]},
    "CONSIDEO": {"-": ["52"]},
    "CONSIDERANTIA": {"-": ["11"]},
    "CONSIDERATE": {"-": ["6&"]},
    "CONSIDERATIO": {"-": ["13"]},
    "CONSIDERATVS": {"-": ["21"]},
    "CONSIDERO": {"-": ["51"]},
    "CONSIDIVS": {"N": ["12"]},
    "CONSIDO": {"-": ["53"]},
    "CONSIGNO": {"-": ["51"]},
    "CONSILIARIVS": {"1": ["12"], "2": ["21"]},
    "CONSILIOR": {"-": ["5J"]},
    "CONSILIVM": {"-": ["12"]},
    "CONSIMILIS": {"-": ["24"]},
    "CONSIPIO": {"-": ["55"]},
    "CONSISTO": {"-": ["53"]},
    "CONSITOR": {"-": ["13"]},
    "CONSOBRINVS": {"-": ["12"]},
    "CONSOCIO": {"-": ["51"]},
    "CONSOLATIO": {"-": ["13"]},
    "CONSOLATOR": {"-": ["13"]},
    "CONSOLIDO": {"-": ["51"]},
    "CONSOLOR": {"-": ["5J"]},
    "CONSONANTER": {"-": ["6-"]},
    "CONSONANTIA": {"-": ["11"]},
    "CONSONO": {"1": ["51"]},
    "CONSONVS": {"-": ["21"]},
    "CONSOPIO": {"-": ["54"]},
    "CONSORS": {"-": ["25"]},
    "CONSORTIVM": {"-": ["12"]},
    "CONSPECTVS": {"1": ["14"], "2": ["21"]},
    "CONSPERGO": {"-": ["53"]},
    "CONSPICIO": {"-": ["55"]},
    "CONSPICOR": {"-": ["5J"]},
    "CONSPICVVM": {"-": ["12"]},
    "CONSPICVVS": {"-": ["21"]},
    "CONSPIRATIO": {"-": ["13"]},
    "CONSPIRO": {"1": ["51"], "2": ["51"]},
    "CONSPVO": {"-": ["53"]},
    "CONSPVRCO": {"-": ["51"]},
    "CONSTABILIO": {"-": ["54"]},
    "CONSTANS": {"-": ["25"]},
    "CONSTANTER": {"-": ["60"]},
    "CONSTANTIA": {"-": ["11"]},
    "CONSTERNATIO": {"-": ["13"]},
    "CONSTERNO": {"1": ["51"], "2": ["53"]},
    "CONSTIBILIS": {"-": ["24"]},
    "CONSTIPO": {"-": ["51"]},
    "CONSTITVO": {"-": ["53"]},
    "CONSTITVTIO": {"-": ["13"]},
    "CONSTITVTVM": {"-": ["12"]},
    "CONSTITVTVS": {"1": ["14"], "2": ["21"]},
    "CONSTO": {"-": ["51"]},
    "CONSTRATVS": {"-": ["21"]},
    "CONSTRICTVS": {"-": ["21"]},
    "CONSTRINGO": {"-": ["53"]},
    "CONSTRVCTIO": {"-": ["13"]},
    "CONSTRVO": {"-": ["53"]},
    "CONSTVPRO": {"-": ["51"]},
    "CONSVALIA": {"N": ["16"]},
    "CONSVDO": {"-": ["51"]},
    "CONSVEFACIO": {"-": ["55"]},
    "CONSVESCO": {"-": ["53"]},
    "CONSVETVDO": {"-": ["13"]},
    "CONSVETVS": {"-": ["21"]},
    "CONSVL": {"-": ["13"]},
    "CONSVLARIS": {"1": ["13"], "2": ["24"]},
    "CONSVLATVS": {"-": ["14"]},
    "CONSVLO": {"-": ["53"]},
    "CONSVLTATIO": {"-": ["13"]},
    "CONSVLTO": {"1": ["51"], "2": ["60"]},
    "CONSVLTOR": {"-": ["13"]},
    "CONSVLTVM": {"-": ["12"]},
    "CONSVMMABILIS": {"-": ["24"]},
    "CONSVMMATIO": {"-": ["13"]},
    "CONSVMMATVS": {"-": ["21"]},
    "CONSVMMO": {"-": ["51"]},
    "CONSVMO": {"-": ["53"]},
    "CONSVO": {"-": ["53"]},
    "CONSVRGO": {"-": ["53"]},
    "CONTABVLATIO": {"-": ["13"]},
    "CONTABVLO": {"-": ["51"]},
    "CONTACTVS": {"-": ["14"]},
    "CONTAGES": {"-": ["13"]},
    "CONTAGIO": {"-": ["13"]},
    "CONTAGIVM": {"-": ["12"]},
    "CONTAMINATVS": {"-": ["21"]},
    "CONTAMINO": {"-": ["51"]},
    "CONTEGO": {"-": ["53"]},
    "CONTEMERO": {"-": ["51"]},
    "CONTEMNENDVS": {"-": ["21"]},
    "CONTEMNO": {"-": ["53"]},
    "CONTEMPLATIO": {"-": ["13"]},
    "CONTEMPLATIVVS": {"-": ["21"]},
    "CONTEMPLATOR": {"-": ["13"]},
    "CONTEMPLOR": {"-": ["5J"]},
    "CONTEMPTE": {"-": ["6&"]},
    "CONTEMPTIM": {"-": ["60"]},
    "CONTEMPTIO": {"-": ["13"]},
    "CONTEMPTOR": {"-": ["13"]},
    "CONTEMPTRIX": {"-": ["13"]},
    "CONTEMPTVS": {"1": ["14"], "2": ["21"]},
    "CONTENDO": {"-": ["53"]},
    "CONTENTE": {"1": ["60"], "2": ["60"]},
    "CONTENTIO": {"1": ["13"], "2": ["13"]},
    "CONTENTVS": {"1": ["21"], "2": ["21"]},
    "CONTERMINVM": {"-": ["12"]},
    "CONTERMINVS": {"1": ["12"], "2": ["21"]},
    "CONTERO": {"-": ["53"]},
    "CONTERREO": {"-": ["52"]},
    "CONTESTATE": {"-": ["60"]},
    "CONTESTOR": {"-": ["5J"]},
    "CONTEXO": {"-": ["53"]},
    "CONTEXTVS": {"1": ["14"], "2": ["21"]},
    "CONTICESCO": {"-": ["53"]},
    "CONTIGNATIO": {"-": ["13"]},
    "CONTIGNO": {"-": ["51"]},
    "CONTIGVITAS": {"-": ["13"]},
    "CONTIGVVS": {"-": ["21"]},
    "CONTINENS": {"1": ["13"], "2": ["13"], "3": ["25"]},
    "CONTINENTER": {"-": ["60"]},
    "CONTINENTIA": {"-": ["11"]},
    "CONTINEO": {"-": ["52"]},
    "CONTINGO": {"1": ["53"], "2": ["53"]},
    "CONTINVATIO": {"-": ["13"]},
    "CONTINVATVS": {"2": ["21"], "-": ["21"]},
    "CONTINVO": {"1": ["51"], "2": ["60"]},
    "CONTINVVS": {"-": ["21"]},
    "CONTIO": {"-": ["13"]},
    "CONTIONABVNDVS": {"-": ["21"]},
    "CONTIONATOR": {"-": ["13"]},
    "CONTIONOR": {"-": ["5J"]},
    "CONTORQVEO": {"-": ["52"]},
    "CONTRA": {"1": ["60"], "2": ["70"]},
    "CONTRACTABILITER": {"-": ["60"]},
    "CONTRACTE": {"-": ["6&"]},
    "CONTRACTIO": {"-": ["13"]},
    "CONTRACTO": {"-": ["51"]},
    "CONTRACTVRA": {"-": ["11"]},
    "CONTRACTVS": {"1": ["14"], "2": ["21"]},
    "CONTRADICO": {"2": ["53"]},
    "CONTRADICTIO": {"-": ["13"]},
    "CONTRAHO": {"-": ["53"]},
    "CONTRALEGO": {"2": ["53"]},
    "CONTRARIE": {"-": ["60"]},
    "CONTRARIETAS": {"-": ["13"]},
    "CONTRARIVM": {"C": ["12"], "-": ["12"]},
    "CONTRARIVS": {"-": ["21"]},
    "CONTRAVERSVM": {"-": ["6O"]},
    "CONTRAVERTO": {"-": ["53"]},
    "CONTRECTO": {"-": ["51"]},
    "CONTREMISCO": {"-": ["53"]},
    "CONTREMO": {"-": ["53"]},
    "CONTRIBVO": {"-": ["53"]},
    "CONTRICO": {"-": ["51"]},
    "CONTRISTO": {"-": ["51"]},
    "CONTROVERSIA": {"-": ["11"]},
    "CONTROVERSIOSVS": {"-": ["21"]},
    "CONTROVERSVS": {"1": ["21"], "2": ["60"]},
    "CONTRVCIDO": {"-": ["51"]},
    "CONTRVDO": {"-": ["53"]},
    "CONTVBERNALIS": {"-": ["13"]},
    "CONTVBERNIVM": {"-": ["12"]},
    "CONTVEOR": {"-": ["5K"]},
    "CONTVITVS": {"-": ["14"]},
    "CONTVMACIA": {"-": ["11"]},
    "CONTVMACITER": {"-": ["60"]},
    "CONTVMAX": {"-": ["25"]},
    "CONTVMELIA": {"-": ["11"]},
    "CONTVMELIOSE": {"-": ["60"]},
    "CONTVMELIOSVS": {"-": ["21"]},
    "CONTVNDO": {"-": ["53"]},
    "CONTVOR": {"-": ["5L"]},
    "CONTVRBO": {"2": ["51"]},
    "CONTVS": {"-": ["12"]},
    "CONVALESCO": {"-": ["53"]},
    "CONVALLIS": {"-": ["13"]},
    "CONVECTO": {"-": ["51"]},
    "CONVECTOR": {"-": ["13"]},
    "CONVEHO": {"-": ["53"]},
    "CONVELLO": {"-": ["53"]},
    "CONVENA": {"-": ["11"]},
    "CONVENIENS": {"-": ["25"]},
    "CONVENIENTER": {"-": ["60"]},
    "CONVENIENTIA": {"-": ["11"]},
    "CONVENIO": {"-": ["54"]},
    "CONVENTICVLVM": {"-": ["12"]},
    "CONVENTIO": {"-": ["13"]},
    "CONVENTO": {"-": ["51"]},
    "CONVENTVM": {"-": ["12"]},
    "CONVENTVS": {"-": ["14"]},
    "CONVERBERO": {"2": ["51"]},
    "CONVERRO": {"-": ["53"]},
    "CONVERSATIO": {"-": ["13"]},
    "CONVERSIO": {"-": ["13"]},
    "CONVERSO": {"-": ["51"]},
    "CONVERSOR": {"-": ["5J"]},
    "CONVERTO": {"-": ["53"]},
    "CONVESTIO": {"-": ["54"]},
    "CONVEXVM": {"-": ["12"]},
    "CONVEXVS": {"-": ["21"]},
    "CONVICIATOR": {"-": ["13"]},
    "CONVICIO": {"-": ["51"]},
    "CONVICIOR": {"-": ["5J"]},
    "CONVICIVM": {"-": ["12"]},
    "CONVICTOR": {"-": ["13"]},
    "CONVICTVS": {"-": ["14"]},
    "CONVINCIO": {"-": ["54"]},
    "CONVINCO": {"-": ["53"]},
    "CONVISO": {"-": ["53"]},
    "CONVIVA": {"-": ["11"]},
    "CONVIVALIS": {"-": ["24"]},
    "CONVIVATOR": {"-": ["13"]},
    "CONVIVIVM": {"-": ["12"]},
    "CONVIVO": {"1": ["51"], "2": ["53"]},
    "CONVIVOR": {"-": ["5J"]},
    "CONVOCO": {"-": ["51"]},
    "CONVOLO": {"2": ["51"]},
    "CONVOLVO": {"-": ["53"]},
    "CONVOLVTO": {"-": ["51"]},
    "CONVOLVVLVS": {"-": ["12"]},
    "CONVOMO": {"-": ["53"]},
    "CONVS": {"-": ["12"]},
    "CONVVLNERO": {"-": ["51"]},
    "COOPERIO": {"-": ["54"]},
    "COOPTATIO": {"-": ["13"]},
    "COOPTO": {"-": ["51"]},
    "COORIOR": {"-": ["5M"]},
    "COORTVS": {"-": ["14"]},
    "COPHINVS": {"-": ["12"]},
    "COPIA": {"-": ["11"], "N": ["11"]},
    "COPIOSE": {"-": ["60"]},
    "COPIOSVS": {"-": ["21"]},
    "COPVLA": {"-": ["11"]},
    "COPVLO": {"-": ["51"]},
    "COPVLOR": {"-": ["5J"]},
    "COQVO": {"-": ["53"]},
    "COQVVS": {"-": ["12"]},
    "COR": {"-": ["13"]},
    "CORA": {"1N": ["11"], "N": ["11"]},
    "CORACINVS": {"-": ["21"]},
    "CORALIVM": {"-": ["12"]},
    "CORAM": {"1": ["60"], "2": ["70"]},
    "CORAX": {"-": ["13"], "N": ["13"]},
    "CORBIO": {"N": ["13"]},
    "CORBIS": {"-": ["13"]},
    "CORBVLA": {"-": ["11"]},
    "CORCYRA": {"N": ["11"]},
    "CORCYRAEI": {"N": ["12"]},
    "CORCYRAEVS": {"N": ["21"], "A": ["21"]},
    "CORDAX": {"-": ["13"]},
    "CORDVBA": {"N": ["11"]},
    "CORDVS": {"N": ["12"]},
    "CORIANDRVM": {"-": ["12"]},
    "CORINTHII": {"N": ["12"]},
    "CORINTHIVM": {"N": ["12"]},
    "CORINTHIVS": {"N": ["21"], "A": ["21"]},
    "CORINTHVS": {"N": ["12"]},
    "CORIOLANVS": {"A": ["21"], "N": ["12"]},
    "CORIOLI": {"N": ["12"]},
    "CORIVS": {"-": ["12"]},
    "CORNELIA": {"N": ["11"]},
    "CORNELIVS": {"A": ["21"], "N": ["12"]},
    "CORNEVS": {"1": ["21"], "2": ["21"]},
    "CORNICEN": {"-": ["13"]},
    "CORNICOR": {"-": ["5J"]},
    "CORNICVLA": {"-": ["11"]},
    "CORNICVLVM": {"N": ["12"]},
    "CORNIFICIVS": {"N": ["12"]},
    "CORNIGER": {"-": ["21"]},
    "CORNIPES": {"1": ["13"], "2": ["25"]},
    "CORNIX": {"-": ["13"]},
    "CORNV": {"-": ["14"]},
    "CORNVM": {"1": ["12"], "2": ["12"]},
    "CORNVS": {"-": ["16"]},
    "COROLLA": {"-": ["11"]},
    "COROLLARIVM": {"-": ["12"]},
    "CORONA": {"-": ["11"], "N": ["11"]},
    "CORONAMENTVM": {"-": ["12"]},
    "CORONARIVS": {"-": ["21"]},
    "CORONO": {"-": ["51"]},
    "CORPORALIS": {"-": ["24"]},
    "CORPORATVRA": {"-": ["11"]},
    "CORPOREVS": {"-": ["21"]},
    "CORPVS": {"-": ["13"]},
    "CORPVSCVLVM": {"-": ["12"]},
    "CORRADO": {"-": ["53"]},
    "CORRECTIO": {"-": ["13"]},
    "CORRECTOR": {"-": ["13"]},
    "CORREPO": {"-": ["53"]},
    "CORREPTIO": {"-": ["13"]},
    "CORRIDEO": {"-": ["52"]},
    "CORRIGO": {"-": ["53"]},
    "CORRIPIO": {"-": ["55"]},
    "CORROBORO": {"-": ["51"]},
    "CORRODO": {"-": ["53"]},
    "CORROGO": {"-": ["51"]},
    "CORROTVNDO": {"-": ["51"]},
    "CORRVDA": {"-": ["11"]},
    "CORRVGO": {"-": ["53"]},
    "CORRVMPO": {"-": ["53"]},
    "CORRVO": {"-": ["53"]},
    "CORRVPTELA": {"-": ["11"]},
    "CORRVPTIO": {"-": ["13"]},
    "CORRVPTOR": {"-": ["13"]},
    "CORRVPTVS": {"C": ["21"], "-": ["21"]},
    "CORSA": {"-": ["11"]},
    "CORTEX": {"-": ["13"]},
    "CORTINA": {"-": ["11"]},
    "CORVSCO": {"-": ["51"]},
    "CORVSCVS": {"-": ["21"]},
    "CORVVS": {"-": ["12"], "N": ["12"]},
    "CORYCEVM": {"-": ["12"]},
    "CORYLVS": {"-": ["12"]},
    "CORYMBVS": {"-": ["12"]},
    "CORYTVS": {"-": ["12"]},
    "COS": {"-": ["13"]},
    "COSMETA": {"-": ["11"]},
    "COSSVRA": {"N": ["17"]},
    "COSSVS": {"N": ["12"]},
    "COSSVTIVS": {"N": ["12"]},
    "COSTA": {"-": ["11"], "N": ["11"]},
    "COSTVM": {"-": ["17"]},
    "COTES": {"-": ["13"]},
    "COTHON": {"-": ["17"]},
    "COTHVRNATVS": {"-": ["21"]},
    "COTHVRNVS": {"-": ["12"]},
    "COTIDIANO": {"-": ["60"]},
    "COTIDIANVS": {"-": ["21"]},
    "COTIDIE": {"-": ["60"]},
    "COTONEVS": {"-": ["21"]},
    "COTTA": {"N": ["11"]},
    "COTTANA": {"-": ["12"]},
    "COTTIVS": {"N": ["12"], "A": ["21"]},
    "COTVRNIX": {"-": ["13"]},
    "COTYLA": {"-": ["11"], "N": ["11"]},
    "COTYTIA": {"-": ["12"]},
    "COVINNARIVS": {"-": ["12"]},
    "COVM": {"-": ["12"]},
    "COXA": {"-": ["11"]},
    "COXENDIX": {"-": ["13"]},
    "COXVS": {"-": ["21"]},
    "CRABRO": {"-": ["13"]},
    "CRAMBE": {"-": ["17"]},
    "CRAPVLA": {"-": ["11"]},
    "CRAS": {"1": ["16"], "2": ["60"]},
    "CRASSE": {"-": ["60"]},
    "CRASSITVDO": {"-": ["13"]},
    "CRASSVS": {"-": ["21"], "N": ["12"]},
    "CRASTINVM": {"-": ["12"]},
    "CRASTINVS": {"-": ["21"], "N": ["12"]},
    "CRATER": {"-": ["17"], "N": ["12"]},
    "CRATERA": {"-": ["11"]},
    "CRATES": {"-": ["13"], "N": ["13"]},
    "CRATICIVS": {"-": ["21"]},
    "CRATICVLA": {"-": ["11"]},
    "CRATIPPVS": {"N": ["12"]},
    "CREATOR": {"-": ["13"]},
    "CREATRIX": {"-": ["13"]},
    "CREBER": {"-": ["21"]},
    "CREBRA": {"-": ["60"]},
    "CREBRE": {"-": ["60"]},
    "CREBRESCO": {"-": ["53"]},
    "CREBRITAS": {"-": ["13"]},
    "CREBRO": {"1": ["51"], "2": ["60"]},
    "CREDIBILIS": {"-": ["24"]},
    "CREDITOR": {"-": ["13"]},
    "CREDITVM": {"-": ["12"]},
    "CREDO": {"-": ["53"]},
    "CREDVLITAS": {"-": ["13"]},
    "CREDVLVS": {"-": ["21"]},
    "CREMERA": {"N": ["11"]},
    "CREMO": {"-": ["51"]},
    "CREMOR": {"-": ["13"]},
    "CREMVTIVS": {"N": ["12"]},
    "CREO": {"-": ["51"]},
    "CREPAX": {"-": ["25"]},
    "CREPER": {"-": ["21"]},
    "CREPEREIVS": {"N": ["12"]},
    "CREPIDA": {"-": ["11"]},
    "CREPIDATVS": {"-": ["21"]},
    "CREPIDO": {"-": ["13"]},
    "CREPITACILLVM": {"-": ["12"]},
    "CREPITO": {"-": ["51"]},
    "CREPITVS": {"-": ["14"]},
    "CREPO": {"-": ["51"]},
    "CREPVNDIA": {"-": ["12"]},
    "CREPVSCVLVM": {"-": ["12"]},
    "CRESCENTIA": {"-": ["11"]},
    "CRESCO": {"-": ["53"]},
    "CRETA": {"-": ["11"], "N": ["11"]},
    "CRETATVS": {"-": ["21"]},
    "CRETENSES": {"N": ["13"]},
    "CRETEVS": {"-": ["21"]},
    "CRETICVS": {"-": ["21"], "A": ["21"], "N": ["12"]},
    "CRETIO": {"-": ["13"]},
    "CRETOSVS": {"-": ["21"]},
    "CRETVLA": {"-": ["11"]},
    "CREVSA": {"1N": ["11"], "2N": ["11"], "N": ["11"]},
    "CRIBRVM": {"-": ["12"]},
    "CRIMEN": {"-": ["13"]},
    "CRIMINALIS": {"-": ["24"]},
    "CRIMINATIO": {"-": ["13"]},
    "CRIMINATOR": {"-": ["13"]},
    "CRIMINO": {"-": ["51"]},
    "CRIMINOR": {"-": ["5J"]},
    "CRIMINOSE": {"-": ["60"]},
    "CRIMINOSVS": {"-": ["21"]},
    "CRINIO": {"-": ["54"]},
    "CRINIS": {"-": ["13"]},
    "CRINITVS": {"-": ["21"]},
    "CRIODOCIS": {"-": ["17"]},
    "CRISIS": {"-": ["17"]},
    "CRISPO": {"-": ["51"]},
    "CRISPVLVS": {"-": ["21"]},
    "CRISPVS": {"-": ["21"], "N": ["12"]},
    "CRISSO": {"-": ["51"]},
    "CRISTA": {"-": ["11"]},
    "CRISTATVS": {"-": ["21"]},
    "CRITICVS": {"1": ["12"]},
    "CRITOLAVS": {"N": ["12"]},
    "CROCEVS": {"-": ["21"]},
    "CROCINVM": {"-": ["12"]},
    "CROCINVS": {"-": ["21"]},
    "CROCODILVS": {"-": ["17"]},
    "CROCVM": {"-": ["12"]},
    "CROCVS": {"-": ["12"]},
    "CROTO": {"N": ["17"]},
    "CRVCIABILITER": {"-": ["60"]},
    "CRVCIATVS": {"-": ["14"]},
    "CRVCIO": {"-": ["51"]},
    "CRVDELIS": {"-": ["24"]},
    "CRVDELITAS": {"-": ["13"]},
    "CRVDELITER": {"-": ["60"]},
    "CRVDESCO": {"-": ["53"]},
    "CRVDITAS": {"-": ["13"]},
    "CRVDVS": {"-": ["21"]},
    "CRVENTE": {"-": ["6&"]},
    "CRVENTO": {"-": ["51"]},
    "CRVENTVS": {"-": ["21"]},
    "CRVMENA": {"-": ["11"]},
    "CRVOR": {"-": ["13"]},
    "CRVPPELLARII": {"-": ["12"]},
    "CRVS": {"-": ["13"]},
    "CRVSTA": {"-": ["11"]},
    "CRVSTO": {"-": ["51"]},
    "CRVSTVLARIVS": {"-": ["12"]},
    "CRVSTVLVM": {"-": ["12"]},
    "CRVSTVM": {"-": ["12"]},
    "CRVSTVMERIA": {"N": ["11"]},
    "CRVSTVMERIVM": {"N": ["12"]},
    "CRVSTVMINI": {"N": ["12"]},
    "CRVSTVMINVS": {"N": ["21"]},
    "CRVX": {"-": ["13"]},
    "CRYPTA": {"-": ["11"]},
    "CRYSTALLINA": {"-": ["12"]},
    "CRYSTALLINVS": {"-": ["21"]},
    "CVBICVLARIS": {"-": ["24"]},
    "CVBICVLARIVS": {"1": ["12"], "2": ["21"]},
    "CVBICVLATVS": {"-": ["21"]},
    "CVBICVLVM": {"-": ["12"]},
    "CVBICVS": {"-": ["21"]},
    "CVBILE": {"-": ["13"]},
    "CVBITAL": {"-": ["13"]},
    "CVBITALIS": {"-": ["24"]},
    "CVBITO": {"-": ["51"]},
    "CVBITVM": {"-": ["12"]},
    "CVBITVS": {"1": ["14"], "2": ["14"]},
    "CVBO": {"-": ["51"]},
    "CVBVS": {"-": ["12"]},
    "CVCVLLIO": {"-": ["13"]},
    "CVCVLLVS": {"-": ["12"]},
    "CVCVLVS": {"-": ["12"]},
    "CVCVMIS": {"-": ["13"]},
    "CVCVRBITA": {"-": ["11"]},
    "CVCVRBITIVVS": {"-": ["21"]},
    "CVDO": {"-": ["53"]},
    "CVICVIMODI": {"1": ["66"], "2": ["67"]},
    "CVIVS": {"1": ["46"], "2": ["47"]},
    "CVIVSMODI": {"1": ["66"], "2": ["67"]},
    "CVIVSQVEMODI": {"-": ["60"]},
    "CVLCITA": {"-": ["11"]},
    "CVLCO": {"-": ["51"]},
    "CVLEARIS": {"-": ["24"]},
    "CVLEVM": {"-": ["12"]},
    "CVLEVS": {"-": ["12"]},
    "CVLEX": {"-": ["13"]},
    "CVLIGNA": {"-": ["11"]},
    "CVLINA": {"-": ["11"]},
    "CVLMEN": {"-": ["13"]},
    "CVLMVS": {"-": ["12"]},
    "CVLO": {"-": ["53"]},
    "CVLPA": {"-": ["11"]},
    "CVLPABILIS": {"-": ["24"]},
    "CVLPATVS": {"-": ["21"]},
    "CVLPO": {"-": ["51"]},
    "CVLTA": {"-": ["12"]},
    "CVLTE": {"-": ["6&"]},
    "CVLTELLVS": {"-": ["12"]},
    "CVLTER": {"-": ["12"]},
    "CVLTIO": {"-": ["13"]},
    "CVLTO": {"-": ["51"]},
    "CVLTOR": {"-": ["13"]},
    "CVLTRIX": {"-": ["13"]},
    "CVLTVRA": {"-": ["11"]},
    "CVLTVS": {"1": ["14"], "2": ["21"]},
    "CVLVLLVS": {"-": ["12"]},
    "CVLVS": {"-": ["12"]},
    "CVM": {"1": ["60"], "2": ["70"], "3": ["82"]},
    "CVMAE": {"N": ["11"]},
    "CVMANVS": {"N": ["21"], "A": ["21"]},
    "CVMBO": {"-": ["53"]},
    "CVMCVMQVE": {"-": ["60"]},
    "CVMERA": {"1": ["11"], "-": ["11"]},
    "CVMINVM": {"-": ["12"]},
    "CVMPRIMIS": {"-": ["60"]},
    "CVMQVE": {"-": ["60"]},
    "CVMVLATE": {"-": ["60"]},
    "CVMVLATVS": {"-": ["21"]},
    "CVMVLO": {"-": ["51"]},
    "CVMVLVS": {"-": ["12"]},
    "CVNABVLA": {"-": ["12"]},
    "CVNAE": {"-": ["11"]},
    "CVNCTABVNDVS": {"-": ["21"]},
    "CVNCTANS": {"-": ["25"]},
    "CVNCTANTER": {"-": ["60"]},
    "CVNCTATIO": {"-": ["13"]},
    "CVNCTATOR": {"-": ["13"]},
    "CVNCTOR": {"-": ["5J"]},
    "CVNCTVS": {"-": ["21"]},
    "CVNEATIM": {"-": ["60"]},
    "CVNEO": {"-": ["51"]},
    "CVNEOLVS": {"-": ["12"]},
    "CVNEVS": {"-": ["12"]},
    "CVNICA": {"-": ["11"]},
    "CVNICVLOSVS": {"-": ["21"]},
    "CVNICVLVS": {"-": ["12"]},
    "CVNNVS": {"-": ["12"]},
    "CVPA": {"1": ["11"], "2": ["11"]},
    "CVPEDIA": {"-": ["11"]},
    "CVPEDO": {"-": ["13"]},
    "CVPIDE": {"-": ["60"]},
    "CVPIDITAS": {"-": ["13"]},
    "CVPIDO": {"-": ["13"], "N": ["13"]},
    "CVPIDVS": {"-": ["21"]},
    "CVPIENS": {"-": ["25"]},
    "CVPIO": {"-": ["55"]},
    "CVPITOR": {"-": ["13"]},
    "CVPITVM": {"-": ["12"]},
    "CVPRESSETVM": {"-": ["12"]},
    "CVPRESSEVS": {"-": ["21"]},
    "CVPRESSIFER": {"-": ["21"]},
    "CVPRESSVS": {"-": ["16"]},
    "CVPVLA": {"2": ["11"]},
    "CVR": {"1": ["66"], "2": ["67"], "-": ["6-"]},
    "CVRA": {"-": ["11"], "N": ["11"]},
    "CVRABILIS": {"-": ["24"]},
    "CVRATE": {"-": ["6&"]},
    "CVRATIO": {"-": ["13"]},
    "CVRATOR": {"-": ["13"]},
    "CVRATVS": {"1": ["14"], "2": ["21"]},
    "CVRCVLIO": {"-": ["13"], "N": ["13"]},
    "CVRES": {"N": ["13"]},
    "CVRIA": {"-": ["11"], "N": ["11"]},
    "CVRIALIS": {"-": ["24"]},
    "CVRIATIVS": {"N": ["12"]},
    "CVRIATVS": {"-": ["21"]},
    "CVRIDIVS": {"N": ["12"]},
    "CVRIO": {"N": ["13"], "-": ["22"]},
    "CVRIOSE": {"-": ["60"]},
    "CVRIOSVS": {"-": ["21"]},
    "CVRO": {"-": ["51"]},
    "CVRRICVLVM": {"-": ["12"]},
    "CVRRO": {"-": ["53"]},
    "CVRRVS": {"-": ["14"]},
    "CVRSIM": {"-": ["60"]},
    "CVRSITO": {"-": ["51"]},
    "CVRSO": {"-": ["51"]},
    "CVRSOR": {"-": ["13"], "N": ["13"]},
    "CVRSVS": {"-": ["14"]},
    "CVRTIVS": {
        "N": ["12"],
        "1N": ["12"],
        "2N": ["12"],
        "1": ["12"],
        "2": ["12"],
    },
    "CVRTO": {"-": ["51"]},
    "CVRTVS": {"-": ["21"]},
    "CVRVAMEN": {"-": ["13"]},
    "CVRVATVRA": {"-": ["11"]},
    "CVRVLIS": {"-": ["24"]},
    "CVRVO": {"-": ["51"]},
    "CVRVVS": {"-": ["21"]},
    "CVSO": {"-": ["51"]},
    "CVSPIS": {"-": ["13"]},
    "CVSTODIA": {"-": ["11"]},
    "CVSTODIO": {"-": ["54"]},
    "CVSTOS": {"-": ["13"], "N": ["13"]},
    "CVTICVLA": {"-": ["11"]},
    "CVTIO": {"-": ["55"]},
    "CVTIS": {"-": ["13"]},
    "CYATHVS": {"-": ["12"]},
    "CYBAEA": {"-": ["11"]},
    "CYCLADES": {"N": ["17", "13"]},
    "CYCLAS": {"-": ["13"], "N": ["17"]},
    "CYCLICOS": {"-": ["27"]},
    "CYCLOPS": {"N": ["17"]},
    "CYCLOTOS": {"-": ["27"]},
    "CYCNEVS": {"-": ["21"]},
    "CYCNVS": {"-": ["12"], "N": ["12"]},
    "CYLINDRVS": {"-": ["12"]},
    "CYMA": {"1": ["11"], "2": ["13"]},
    "CYMATIVM": {"-": ["12"]},
    "CYMBA": {"-": ["11"]},
    "CYMBALVM": {"-": ["17"]},
    "CYMBIVM": {"-": ["12"]},
    "CYNICE": {"-": ["17"]},
    "CYNICI": {"N": ["12"]},
    "CYNICVS": {"N": ["21"], "A": ["21"]},
    "CYPARISSVS": {"-": ["12"]},
    "CYPRIVS": {"-": ["21"], "N": ["21"], "A": ["21"]},
    "CYPRVS": {"N": ["17"]},
    "CYRVS": {"N": ["12"]},
    "CYTISVS": {"-": ["12"]},
    "CYZICENI": {"N": ["12"]},
    "CYZICVS": {"N": ["17"]},
    "DAEDALVS": {"-": ["21"], "N": ["12"]},
    "DAEMON": {"-": ["17"]},
    "DAMA": {"-": ["11"], "N": ["11"]},
    "DAMNAS": {"-": ["26"]},
    "DAMNATIO": {"-": ["13"]},
    "DAMNATORIVS": {"-": ["21"]},
    "DAMNATVS": {"1": ["14"], "2": ["21"]},
    "DAMNO": {"-": ["51"]},
    "DAMNOSE": {"-": ["60"]},
    "DAMNOSVS": {"-": ["21"]},
    "DAMNVM": {"-": ["12"]},
    "DAPALIS": {"-": ["24"]},
    "DAPS": {"-": ["13"]},
    "DAREVS": {"N": ["12"]},
    "DATIO": {"-": ["13"]},
    "DATO": {"-": ["51"]},
    "DATOR": {"-": ["13"]},
    "DATVM": {"-": ["12"]},
    "DE": {"-": ["70"]},
    "DEA": {"-": ["11"], "N": ["11"]},
    "DEACINO": {"-": ["51"]},
    "DEALBO": {"-": ["51"]},
    "DEAMBVLO": {"-": ["51"]},
    "DEBACCHOR": {"-": ["5J"]},
    "DEBELLO": {"-": ["51"]},
    "DEBENS": {"-": ["13"]},
    "DEBEO": {"-": ["52"]},
    "DEBILIS": {"-": ["24"]},
    "DEBILITAS": {"-": ["13"]},
    "DEBILITATIO": {"-": ["13"]},
    "DEBILITO": {"-": ["51"]},
    "DEBITOR": {"-": ["13"]},
    "DEBITVM": {"-": ["12"]},
    "DECANTO": {"-": ["51"]},
    "DECARPO": {"-": ["53"]},
    "DECASTYLVS": {"-": ["27"]},
    "DECEDO": {"1": ["53"]},
    "DECEM": {"-": ["31"]},
    "DECEMBER": {"A": ["23"], "N": ["13"]},
    "DECEMPEDA": {"-": ["11"]},
    "DECEMVIR": {"-": ["12"]},
    "DECEMVIRALIS": {"-": ["24"]},
    "DECENNIS": {"-": ["24"]},
    "DECENS": {"-": ["25"]},
    "DECENTER": {"-": ["60"]},
    "DECENTIA": {"-": ["11"]},
    "DECEPTOR": {"-": ["13"]},
    "DECEPTORIVS": {"-": ["21"]},
    "DECERNO": {"-": ["53"]},
    "DECERPO": {"-": ["53"]},
    "DECERTO": {"1": ["51"]},
    "DECESSIO": {"-": ["13"]},
    "DECESSOR": {"-": ["13"]},
    "DECESSVS": {"-": ["14"]},
    "DECET": {"-": ["52"]},
    "DECIDO": {"1": ["53"], "2": ["53"]},
    "DECIES": {"-": ["36"]},
    "DECIMA": {"-": ["11"]},
    "DECIMANI": {"-": ["12"]},
    "DECIMANVS": {"1": ["12"], "2": ["21"]},
    "DECIMVS": {"-": ["32"], "N": ["12"]},
    "DECIPIO": {"-": ["55"]},
    "DECISIO": {"-": ["13"]},
    "DECLAMATIO": {"-": ["13"]},
    "DECLAMATOR": {"-": ["13"]},
    "DECLAMATORIVS": {"-": ["21"]},
    "DECLAMO": {"-": ["51"]},
    "DECLARATOR": {"-": ["13"]},
    "DECLARO": {"-": ["51"]},
    "DECLINATIO": {"-": ["13"]},
    "DECLINO": {"-": ["51"]},
    "DECLIVE": {"-": ["13"]},
    "DECLIVIS": {"-": ["24"]},
    "DECLIVITAS": {"-": ["13"]},
    "DECLIVVS": {"-": ["21"]},
    "DECOCTA": {"-": ["11"]},
    "DECOCTIOR": {"-": ["2&"]},
    "DECOCTOR": {"-": ["13"]},
    "DECOCTVM": {"-": ["12"], "D\u00e9coction": ["12"]},
    "DECOCTVS": {"-": ["21"]},
    "DECOLLO": {"-": ["51"]},
    "DECOLO": {"1": ["51"], "-": ["51"]},
    "DECOLOR": {"-": ["25"]},
    "DECOLORO": {"-": ["51"]},
    "DECONDO": {"-": ["53"]},
    "DECOQVO": {"-": ["53"]},
    "DECOR": {"-": ["13"], "N": ["13"]},
    "DECORATVS": {"O": ["21"], "-": ["21"]},
    "DECORE": {"-": ["60"]},
    "DECORO": {"-": ["51"]},
    "DECORVM": {"C": ["12"], "-": ["12"]},
    "DECORVS": {"-": ["21"]},
    "DECREPITVS": {"-": ["21"]},
    "DECRESCENTIA": {"-": ["11"]},
    "DECRESCO": {"-": ["53"]},
    "DECRETORIVS": {"-": ["21"]},
    "DECRETVM": {"-": ["12"]},
    "DECVMATES": {"-": ["24"]},
    "DECVMBO": {"-": ["53"]},
    "DECVRIA": {"-": ["11"]},
    "DECVRIALIS": {"-": ["24"]},
    "DECVRIATIO": {"-": ["13"]},
    "DECVRIATVS": {"-": ["14"]},
    "DECVRIO": {"1": ["13"], "2": ["51"]},
    "DECVRIONALIS": {"-": ["24"]},
    "DECVRIONATVS": {"-": ["14"]},
    "DECVRRO": {"-": ["53"]},
    "DECVRSIO": {"-": ["13"]},
    "DECVRSVS": {"-": ["14"]},
    "DECVRTO": {"-": ["51"]},
    "DECVS": {"-": ["13"]},
    "DECVSSATIM": {"-": ["60"]},
    "DECVSSATIO": {"-": ["13"]},
    "DECVSSIS": {"-": ["16"]},
    "DECVTIO": {"-": ["55"]},
    "DEDECET": {"-": ["52"]},
    "DEDECORO": {"-": ["51"]},
    "DEDECORVS": {"-": ["21"]},
    "DEDECVS": {"-": ["13"]},
    "DEDICATIO": {"-": ["13"]},
    "DEDICATVS": {"-": ["21"]},
    "DEDICO": {"1": ["51"]},
    "DEDIGNATIO": {"-": ["13"]},
    "DEDIGNOR": {"-": ["5J"]},
    "DEDISCO": {"-": ["53"]},
    "DEDITICIVS": {"1": ["12"], "2": ["21"]},
    "DEDITIO": {"-": ["13"]},
    "DEDITVS": {"-": ["21"]},
    "DEDO": {"-": ["53"]},
    "DEDOCEO": {"-": ["52"]},
    "DEDOLEO": {"-": ["52"]},
    "DEDVCO": {"-": ["53"]},
    "DEDVCTIO": {"-": ["13"]},
    "DEDVCTVS": {"1": ["14"], "2": ["21"]},
    "DEERRO": {"2": ["51"]},
    "DEFAECATVS": {"-": ["21"]},
    "DEFATIGATIO": {"-": ["13"]},
    "DEFATIGO": {"-": ["51"]},
    "DEFECTIO": {"-": ["13"]},
    "DEFECTOR": {"-": ["13"]},
    "DEFECTVS": {"1": ["14"], "2": ["21"]},
    "DEFENDO": {"-": ["53"]},
    "DEFENSIO": {"-": ["13"]},
    "DEFENSO": {"-": ["51"]},
    "DEFENSOR": {"-": ["13"]},
    "DEFERO": {"-": ["56"]},
    "DEFERVEFACIO": {"-": ["55"]},
    "DEFERVEO": {"-": ["52"]},
    "DEFESSVS": {"-": ["21"]},
    "DEFETISCOR": {"-": ["5L"]},
    "DEFICIO": {"-": ["55"]},
    "DEFIGO": {"-": ["53"]},
    "DEFINGO": {"-": ["53"]},
    "DEFINIO": {"-": ["54"]},
    "DEFINITE": {"-": ["60"]},
    "DEFINITIO": {"-": ["13"]},
    "DEFINITIVVS": {"-": ["21"]},
    "DEFINITVS": {"-": ["21"]},
    "DEFIO": {"-": ["56"]},
    "DEFLAGRO": {"-": ["51"]},
    "DEFLECTO": {"-": ["53"]},
    "DEFLEO": {"-": ["52"]},
    "DEFLORESCO": {"-": ["53"]},
    "DEFLVO": {"-": ["53"]},
    "DEFODIO": {"-": ["55"]},
    "DEFORMATIO": {"1": ["13"], "2": ["13"]},
    "DEFORMATVS": {"-": ["21"]},
    "DEFORMIS": {"-": ["24"]},
    "DEFORMITAS": {"-": ["13"]},
    "DEFORMO": {"1": ["51"], "2": ["51"]},
    "DEFRICO": {"-": ["51"]},
    "DEFRINGO": {"-": ["53"]},
    "DEFRVDO": {"-": ["51"]},
    "DEFRVTO": {"-": ["51"]},
    "DEFRVTVM": {"-": ["12"]},
    "DEFVGIO": {"-": ["55"]},
    "DEFVNCTVS": {"-": ["21"]},
    "DEFVNDO": {"2": ["53"]},
    "DEFVNGOR": {"-": ["5L"]},
    "DEFVTVO": {"-": ["53"]},
    "DEGENER": {"-": ["22"]},
    "DEGENERO": {"-": ["51"]},
    "DEGERO": {"-": ["53"]},
    "DEGLVBO": {"-": ["53"]},
    "DEGO": {"-": ["53"]},
    "DEGRAVO": {"-": ["51"]},
    "DEGREDIOR": {"-": ["5N"]},
    "DEGVSTO": {"-": ["51"]},
    "DEHAVRIO": {"-": ["54"]},
    "DEHINC": {"-": ["60"]},
    "DEHISCO": {"-": ["53"]},
    "DEHONESTAMENTVM": {"-": ["12"]},
    "DEHONESTO": {"-": ["51"]},
    "DEHORTOR": {"-": ["5J"]},
    "DEICIO": {"-": ["55"]},
    "DEIECTIO": {"-": ["13"]},
    "DEIECTVS": {"1": ["14"], "2": ["21"]},
    "DEILLIVS": {"N": ["12"]},
    "DEINCEPS": {"-": ["60"]},
    "DEINDE": {"-": ["60"]},
    "DEIVNGO": {"-": ["53"]},
    "DELABOR": {"2": ["5L"]},
    "DELAPIDO": {"-": ["51"]},
    "DELASSO": {"-": ["51"]},
    "DELATIO": {"-": ["13"]},
    "DELATOR": {"-": ["13"]},
    "DELECTABILIS": {"-": ["24"]},
    "DELECTATIO": {"-": ["13"]},
    "DELECTO": {"-": ["51"]},
    "DELECTVS": {"-": ["14"]},
    "DELEGATIO": {"-": ["13"]},
    "DELEGATVM": {"-": ["12"]},
    "DELEGO": {"1": ["51"]},
    "DELENIMENTVM": {"-": ["12"]},
    "DELENIO": {"-": ["54"]},
    "DELEO": {"-": ["52"]},
    "DELIACVS": {"N": ["21"], "A": ["21"]},
    "DELIBERATIO": {"-": ["13"]},
    "DELIBERATVS": {"-": ["21"]},
    "DELIBERO": {"-": ["51"]},
    "DELIBO": {"-": ["51"]},
    "DELIBRO": {"-": ["51"]},
    "DELIBVO": {"-": ["53"]},
    "DELICATA": {"-": ["11"]},
    "DELICATE": {"-": ["60"]},
    "DELICATVS": {"1": ["12"], "2": ["21"]},
    "DELICIA": {"1": ["11"], "2": ["11"]},
    "DELICIO": {"-": ["55"]},
    "DELICIOLVM": {"-": ["12"]},
    "DELICIVM": {"-": ["12"]},
    "DELICTVM": {"-": ["12"]},
    "DELICVLVS": {"-": ["21"]},
    "DELIGO": {"2": ["51"], "3": ["53"]},
    "DELINQVO": {"-": ["53"]},
    "DELIQVESCO": {"-": ["53"]},
    "DELIQVIAE": {"-": ["11"]},
    "DELIRATIO": {"-": ["13"]},
    "DELIRO": {"-": ["51"]},
    "DELIRVS": {"-": ["21"]},
    "DELITEO": {"-": ["52"]},
    "DELITESCO": {"-": ["53"]},
    "DELIVS": {"N": ["21"], "A": ["21"]},
    "DELPHI": {"N": ["12"]},
    "DELPHICVS": {"N": ["21"], "A": ["21"]},
    "DELPHIN": {"-": ["17"], "N": ["13"]},
    "DELPHINVS": {"-": ["12"]},
    "DELVBRVM": {"-": ["12"]},
    "DELVDO": {"-": ["53"]},
    "DELVMBIS": {"-": ["24"]},
    "DELVMBO": {"-": ["51"]},
    "DELVS": {"N": ["17"]},
    "DELVTO": {"-": ["51"]},
    "DEMANO": {"-": ["51"]},
    "DEMARATVS": {"N": ["12"]},
    "DEMENS": {"-": ["25"]},
    "DEMENTER": {"-": ["60"]},
    "DEMENTIA": {"-": ["11"]},
    "DEMENTIO": {"-": ["54"]},
    "DEMEREO": {"-": ["52"]},
    "DEMEREOR": {"-": ["5K"]},
    "DEMERGO": {"-": ["53"]},
    "DEMETO": {"2": ["53"]},
    "DEMETRIVS": {"N": ["12"]},
    "DEMIGRO": {"-": ["51"]},
    "DEMINVO": {"-": ["53"]},
    "DEMINVTIO": {"-": ["13"]},
    "DEMINVTVS": {"-": ["21"]},
    "DEMIROR": {"-": ["5J"]},
    "DEMISSE": {"-": ["60"]},
    "DEMISSIO": {"-": ["13"]},
    "DEMISSVS": {"-": ["21"]},
    "DEMITTO": {"-": ["53"]},
    "DEMIVRGVS": {"-": ["12"]},
    "DEMNO": {"-": ["51"]},
    "DEMO": {"-": ["53"]},
    "DEMOCRITVS": {"N": ["12"]},
    "DEMOLIOR": {"-": ["5M"]},
    "DEMOLITIO": {"-": ["13"]},
    "DEMOLITOR": {"-": ["13"]},
    "DEMONSTRATIO": {"-": ["13"]},
    "DEMONSTRO": {"-": ["51"]},
    "DEMORDEO": {"-": ["52"]},
    "DEMORIOR": {"-": ["5N"]},
    "DEMOROR": {"1": ["5J"]},
    "DEMOVEO": {"-": ["52"]},
    "DEMVM": {"-": ["60"]},
    "DEMVTO": {"2": ["51"]},
    "DENARIVM": {"-": ["12"]},
    "DENARIVS": {"1": ["12"], "2": ["21"]},
    "DENARRO": {"-": ["51"]},
    "DENATO": {"-": ["51"]},
    "DENEGO": {"-": ["51"]},
    "DENI": {"-": ["33"]},
    "DENIGRO": {"-": ["51"]},
    "DENIQVE": {"-": ["60"]},
    "DENOMINO": {"-": ["51"]},
    "DENORMO": {"-": ["51"]},
    "DENOTATVS": {"1": ["14"], "2": ["21"]},
    "DENOTO": {"-": ["51"]},
    "DENS": {"-": ["13"]},
    "DENSE": {"-": ["6-"]},
    "DENSEO": {"-": ["52"]},
    "DENSO": {"-": ["51"]},
    "DENSVS": {"-": ["21"], "N": ["12"]},
    "DENTALE": {"-": ["13"]},
    "DENTATVS": {"-": ["21"], "N": ["12"]},
    "DENTICVLVS": {"-": ["12"]},
    "DENVBO": {"-": ["53"]},
    "DENVDO": {"-": ["51"]},
    "DENVNTIATIO": {"-": ["13"]},
    "DENVNTIO": {"-": ["51"]},
    "DENVO": {"1": ["51"], "2": ["53"]},
    "DEORSVM": {"-": ["60"]},
    "DEPACISCOR": {"-": ["5L"]},
    "DEPACO": {"-": ["51"]},
    "DEPANGO": {"-": ["53"]},
    "DEPASCO": {"-": ["53"]},
    "DEPASCOR": {"-": ["5L"]},
    "DEPECTO": {"-": ["53"]},
    "DEPECVLATOR": {"-": ["13"]},
    "DEPECVLOR": {"-": ["5J"]},
    "DEPELLO": {"-": ["53"]},
    "DEPENDEO": {"-": ["52"]},
    "DEPENDO": {"-": ["53"]},
    "DEPERDITVS": {"-": ["21"]},
    "DEPERDO": {"-": ["53"]},
    "DEPEREO": {"1": ["56"]},
    "DEPETIGO": {"-": ["13"]},
    "DEPILO": {"1": ["51"]},
    "DEPINGO": {"-": ["53"]},
    "DEPLANGO": {"-": ["53"]},
    "DEPLEO": {"-": ["52"]},
    "DEPLEXVS": {"-": ["21"]},
    "DEPLORATIO": {"-": ["13"]},
    "DEPLORO": {"-": ["51"]},
    "DEPLVO": {"-": ["53"]},
    "DEPONO": {"-": ["53"]},
    "DEPOPVLATIO": {"-": ["13"]},
    "DEPOPVLO": {"-": ["51"]},
    "DEPOPVLOR": {"-": ["5J"]},
    "DEPORTATIO": {"-": ["13"]},
    "DEPORTO": {"-": ["51"]},
    "DEPOSCO": {"-": ["53"]},
    "DEPOSITVM": {"-": ["12"]},
    "DEPOSTVLO": {"-": ["51"]},
    "DEPRAVATIO": {"-": ["13"]},
    "DEPRAVO": {"-": ["51"]},
    "DEPRECABVNDVS": {"-": ["21"]},
    "DEPRECATIO": {"-": ["13"]},
    "DEPRECATOR": {"-": ["13"]},
    "DEPRECATRIX": {"-": ["13"]},
    "DEPRECOR": {"-": ["5J"]},
    "DEPREHENDO": {"-": ["53"]},
    "DEPRESSE": {"-": ["6&"]},
    "DEPRESSIO": {"-": ["13"]},
    "DEPRESSVS": {"-": ["21"]},
    "DEPRIMO": {"-": ["53"]},
    "DEPROELIOR": {"-": ["5J"]},
    "DEPROMO": {"-": ["53"]},
    "DEPROPERO": {"-": ["51"]},
    "DEPSO": {"-": ["53"]},
    "DEPSTICIVS": {"-": ["21"]},
    "DEPVDET": {"-": ["52"]},
    "DEPVGIS": {"-": ["24"]},
    "DEPVGNATIO": {"-": ["13"]},
    "DEPVGNO": {"-": ["51"]},
    "DEPVNGO": {"-": ["53"]},
    "DEPVRGO": {"-": ["51"]},
    "DEPVTO": {"-": ["51"]},
    "DERADO": {"-": ["53"]},
    "DERELINQVO": {"-": ["53"]},
    "DEREPENTE": {"-": ["60"]},
    "DEREPO": {"-": ["53"]},
    "DERIDEO": {"-": ["52"]},
    "DERIDICVLVM": {"-": ["12"]},
    "DERIDICVLVS": {"-": ["21"]},
    "DERIGESCO": {"-": ["53"]},
    "DERIGO": {"-": ["53"]},
    "DERIPIO": {"-": ["55"]},
    "DERISOR": {"-": ["13"]},
    "DERISVS": {"-": ["14"]},
    "DERIVO": {"-": ["51"]},
    "DEROGO": {"-": ["51"]},
    "DERVMPO": {"-": ["53"]},
    "DERVPTA": {"-": ["12"]},
    "DERVPTVS": {"-": ["21"]},
    "DESAEVIO": {"-": ["54"]},
    "DESCENDO": {"-": ["53"]},
    "DESCENSVS": {"-": ["14"]},
    "DESCISCO": {"-": ["53"]},
    "DESCRIBO": {"-": ["53"]},
    "DESCRIPTA": {"-": ["12"]},
    "DESCRIPTIO": {"-": ["13"]},
    "DESECO": {"-": ["51"]},
    "DESERO": {"2": ["53"], "3": ["53"]},
    "DESERTA": {"-": ["12"]},
    "DESERTO": {"-": ["51"]},
    "DESERTOR": {"-": ["13"]},
    "DESERTVM": {"-": ["12"]},
    "DESERTVS": {"-": ["21"]},
    "DESERVIO": {"-": ["54"]},
    "DESES": {"-": ["25"]},
    "DESIDEO": {"-": ["52"]},
    "DESIDERABILIS": {"-": ["24"]},
    "DESIDERATIO": {"-": ["13"]},
    "DESIDERIVM": {"-": ["12"]},
    "DESIDERO": {"-": ["51"]},
    "DESIDIA": {"1": ["11"]},
    "DESIDIOSE": {"-": ["60"]},
    "DESIDIOSVS": {"-": ["21"]},
    "DESIDO": {"-": ["53"]},
    "DESIGNATIO": {"-": ["13"]},
    "DESIGNATOR": {"-": ["13"]},
    "DESIGNO": {"-": ["51"]},
    "DESILIO": {"-": ["54"]},
    "DESINO": {"-": ["53"]},
    "DESIPIENTIA": {"-": ["11"]},
    "DESIPIO": {"-": ["55"]},
    "DESISTO": {"-": ["53"]},
    "DESOLO": {"-": ["51"]},
    "DESPECTATIO": {"-": ["13"]},
    "DESPECTO": {"-": ["51"]},
    "DESPECTVS": {"1": ["14"], "2": ["21"]},
    "DESPERATIO": {"-": ["13"]},
    "DESPERATVS": {"-": ["21"]},
    "DESPERO": {"-": ["51"]},
    "DESPICATVS": {"1": ["14"], "2": ["21"]},
    "DESPICIENTIA": {"-": ["11"]},
    "DESPICIO": {"-": ["55"]},
    "DESPOLIO": {"-": ["51"]},
    "DESPONDEO": {"-": ["52"]},
    "DESPVMO": {"-": ["51"]},
    "DESPVO": {"-": ["53"]},
    "DESTERTO": {"-": ["53"]},
    "DESTILLATIO": {"-": ["13"]},
    "DESTILLO": {"-": ["51"]},
    "DESTINA": {"-": ["11"]},
    "DESTINATIO": {"-": ["13"]},
    "DESTINATVM": {"-": ["12"]},
    "DESTINATVS": {"-": ["21"]},
    "DESTINEO": {"-": ["52"]},
    "DESTINO": {"-": ["51"]},
    "DESTITVO": {"-": ["53"]},
    "DESTRICTVS": {"-": ["21"]},
    "DESTRINGO": {"-": ["53"]},
    "DESTRVO": {"-": ["53"]},
    "DESVBITO": {"-": ["60"]},
    "DESVDO": {"-": ["51"]},
    "DESVESCO": {"-": ["53"]},
    "DESVETVDO": {"-": ["13"]},
    "DESVLTOR": {"-": ["13"]},
    "DESVM": {"1": ["56"]},
    "DESVMO": {"-": ["53"]},
    "DESVPER": {"-": ["60"]},
    "DESVRGO": {"-": ["53"]},
    "DETECTVS": {"-": ["21"]},
    "DETEGO": {"-": ["53"]},
    "DETENDO": {"-": ["53"]},
    "DETENTO": {"-": ["51"]},
    "DETERGEO": {"-": ["52"]},
    "DETERGO": {"-": ["53"]},
    "DETERIOR": {"-": ["2&"]},
    "DETERIVS": {"-": ["6&"]},
    "DETERMINO": {"-": ["51"]},
    "DETERO": {"-": ["53"]},
    "DETERREO": {"-": ["52"]},
    "DETESTABILIS": {"-": ["24"]},
    "DETESTATIO": {"-": ["13"]},
    "DETESTOR": {"-": ["5J"]},
    "DETEXO": {"-": ["53"]},
    "DETINEO": {"-": ["52"]},
    "DETONDEO": {"-": ["52"]},
    "DETORQVEO": {"-": ["52"]},
    "DETRACTIO": {"-": ["13"]},
    "DETRACTO": {"-": ["51"]},
    "DETRACTOR": {"-": ["13"]},
    "DETRAHO": {"-": ["53"]},
    "DETRECTO": {"-": ["51"]},
    "DETRIMENTOSVS": {"-": ["21"]},
    "DETRIMENTVM": {"-": ["12"]},
    "DETRVDO": {"-": ["53"]},
    "DETRVNCO": {"-": ["51"]},
    "DETVRBO": {"2": ["51"]},
    "DEVECTO": {"-": ["51"]},
    "DEVEHO": {"-": ["53"]},
    "DEVELLO": {"-": ["53"]},
    "DEVENEROR": {"-": ["5J"]},
    "DEVENIO": {"-": ["54"]},
    "DEVERRO": {"-": ["53"]},
    "DEVERSOR": {"1": ["13"], "2": ["5J"]},
    "DEVERSORIVM": {"-": ["12"]},
    "DEVERTICVLVM": {"-": ["12"]},
    "DEVERTO": {"-": ["53"]},
    "DEVEXVM": {"-": ["12"]},
    "DEVEXVS": {"-": ["21"]},
    "DEVIA": {"-": ["12"]},
    "DEVINCIO": {"-": ["54"]},
    "DEVINCO": {"-": ["53"]},
    "DEVINCTVS": {"-": ["21"]},
    "DEVITO": {"-": ["51"]},
    "DEVIVS": {"-": ["21"]},
    "DEVNX": {"-": ["13"]},
    "DEVOCO": {"-": ["51"]},
    "DEVOLO": {"2": ["51"]},
    "DEVOLVO": {"-": ["53"]},
    "DEVORO": {"-": ["51"]},
    "DEVOTI": {"-": ["12"]},
    "DEVOTIO": {"-": ["13"]},
    "DEVOTVS": {"-": ["21"]},
    "DEVOVEO": {"-": ["52"]},
    "DEVRO": {"-": ["53"]},
    "DEVS": {"-": ["12"], "N": ["12"]},
    "DEXO": {"N": ["17"]},
    "DEXTANS": {"-": ["13"]},
    "DEXTER": {"-": ["21"], "N": ["12"]},
    "DEXTERA": {"-": ["11"]},
    "DEXTERE": {"-": ["60"]},
    "DEXTRORSVM": {"-": ["60"]},
    "DEXTRORSVS": {"-": ["60"]},
    "DIADEMA": {"-": ["17"]},
    "DIADVMENOS": {"-": ["17"]},
    "DIAGONIOS": {"-": ["27"]},
    "DIAGRAMMA": {"-": ["17"]},
    "DIALECTICA": {"-": ["11"]},
    "DIALECTICVS": {"1": ["12"], "2": ["21"]},
    "DIALIS": {"N": ["24"], "A": ["24"]},
    "DIALOGVS": {"-": ["12"]},
    "DIAMETROS": {"-": ["17"]},
    "DIANA": {"N": ["11"]},
    "DIANIVM": {"N": ["12"]},
    "DIARIVM": {"-": ["12"]},
    "DIASTYLOS": {"-": ["27"]},
    "DIATHESIS": {"-": ["17"]},
    "DIATHYRA": {"-": ["17"]},
    "DIATONOS": {"-": ["27"]},
    "DIAVLOS": {"-": ["17"]},
    "DIAZOMA": {"-": ["17"]},
    "DICA": {"-": ["11"]},
    "DICACITAS": {"-": ["13"]},
    "DICAX": {"-": ["25"]},
    "DICHALCVM": {"-": ["12"]},
    "DICIO": {"-": ["13"]},
    "DICIS": {"-": ["16"]},
    "DICO": {"1": ["51"], "2": ["53"]},
    "DICROTA": {"-": ["11"]},
    "DICTATOR": {"-": ["13"]},
    "DICTATORIVS": {"-": ["21"]},
    "DICTATVRA": {"-": ["11"]},
    "DICTIO": {"-": ["13"]},
    "DICTITO": {"-": ["51"]},
    "DICTO": {"-": ["51"]},
    "DICTVM": {"-": ["12"]},
    "DIDO": {"-": ["53"], "N": ["13"]},
    "DIDVCO": {"-": ["53"]},
    "DIEQVINTI": {"-": ["60"]},
    "DIES": {"-": ["15"], "N": ["13"]},
    "DIESIS": {"-": ["13"]},
    "DIEZEVGMENOS": {"-": ["27"]},
    "DIFFAMO": {"-": ["51"]},
    "DIFFERCIO": {"-": ["54"]},
    "DIFFERENTIA": {"-": ["11"]},
    "DIFFERITAS": {"-": ["13"]},
    "DIFFERO": {"-": ["56"]},
    "DIFFERTVS": {"-": ["21"]},
    "DIFFICILE": {"-": ["60"]},
    "DIFFICILIS": {"-": ["24"]},
    "DIFFICILITER": {"-": ["60"]},
    "DIFFICVLTAS": {"-": ["13"]},
    "DIFFICVLTER": {"-": ["60"]},
    "DIFFIDENTIA": {"-": ["11"]},
    "DIFFIDO": {"-": ["5T"]},
    "DIFFINDO": {"-": ["53"]},
    "DIFFINGO": {"-": ["53"]},
    "DIFFITEOR": {"-": ["5K"]},
    "DIFFLVO": {"-": ["53"]},
    "DIFFRINGO": {"-": ["53"]},
    "DIFFVGIO": {"-": ["55"]},
    "DIFFVNDO": {"2": ["53"]},
    "DIFFVSILIS": {"-": ["24"]},
    "DIFFVSIO": {"-": ["13"]},
    "DIFFVSVS": {"-": ["21"]},
    "DIFFVTVO": {"-": ["53"]},
    "DIGERO": {"-": ["53"]},
    "DIGITVLVS": {"-": ["12"]},
    "DIGITVS": {"-": ["12"]},
    "DIGNATIO": {"-": ["13"]},
    "DIGNE": {"-": ["60"]},
    "DIGNITAS": {"-": ["13"]},
    "DIGNO": {"-": ["51"]},
    "DIGNOR": {"-": ["5J"]},
    "DIGNOSCO": {"-": ["53"]},
    "DIGNVS": {"-": ["21"]},
    "DIGREDIOR": {"-": ["5N"]},
    "DIGRESSVS": {"-": ["14"]},
    "DIIVDICO": {"-": ["51"]},
    "DIIVNCTVS": {"-": ["21"]},
    "DIIVNGO": {"-": ["53"]},
    "DILABOR": {"2": ["5L"]},
    "DILACERO": {"-": ["51"]},
    "DILANIO": {"-": ["51"]},
    "DILARGIOR": {"-": ["5M"]},
    "DILATATIO": {"-": ["13"]},
    "DILATIO": {"-": ["13"]},
    "DILATO": {"-": ["51"]},
    "DILECTVS": {"1": ["14"], "2": ["21"]},
    "DILIGENS": {"-": ["25"]},
    "DILIGENTER": {"-": ["60"]},
    "DILIGENTIA": {"-": ["11"]},
    "DILIGO": {"3": ["53"]},
    "DILVCEO": {"-": ["52"]},
    "DILVCESCO": {"-": ["53"]},
    "DILVCIDE": {"-": ["60"]},
    "DILVCIDVS": {"-": ["21"]},
    "DILVDIVM": {"-": ["12"]},
    "DILVO": {"-": ["53"]},
    "DILVTVM": {"-": ["12"]},
    "DILVTVS": {"-": ["21"]},
    "DILVVIES": {"-": ["15"]},
    "DILVVIO": {"-": ["51"]},
    "DILVVIVM": {"-": ["12"]},
    "DIMACHAE": {"-": ["11"]},
    "DIMENSVS": {"-": ["21"]},
    "DIMETIOR": {"-": ["5M"]},
    "DIMICATIO": {"-": ["13"]},
    "DIMICO": {"-": ["51"]},
    "DIMIDIATVS": {"-": ["21"]},
    "DIMIDIO": {"-": ["51"]},
    "DIMIDIVM": {"-": ["12"]},
    "DIMIDIVS": {"-": ["21"]},
    "DIMINVTIO": {"-": ["13"]},
    "DIMISSIO": {"-": ["13"]},
    "DIMITTO": {"-": ["53"]},
    "DIMOEROS": {"-": ["27"]},
    "DIMOVEO": {"-": ["52"]},
    "DINOSCO": {"-": ["53"]},
    "DINVMERO": {"1": ["51"]},
    "DIO": {"N": ["17"]},
    "DIOCLES": {"N": ["17"]},
    "DIODORVS": {"N": ["12"]},
    "DIOGENES": {"N": ["17"]},
    "DIOGNETVS": {"N": ["12"]},
    "DIONYSIA": {"1N": ["11"], "N": ["11"]},
    "DIONYSIARCHVS": {"N": ["12"]},
    "DIONYSIVS": {"N": ["12"]},
    "DIONYSODORVS": {"N": ["12"]},
    "DIOPTRA": {"-": ["11"]},
    "DIOTA": {"-": ["11"]},
    "DIPECHYAEOS": {"-": ["27"]},
    "DIPISCOR": {"-": ["5L"]},
    "DIPLASION": {"-": ["27"]},
    "DIPLINTHIVS": {"-": ["21"]},
    "DIPONDIVM": {"-": ["12"]},
    "DIPONDIVS": {"-": ["12"]},
    "DIPTEROS": {"-": ["27"]},
    "DIRA": {"-": ["12"], "N": ["11"]},
    "DIRAE": {"-": ["11"]},
    "DIRECTE": {"-": ["60"]},
    "DIRECTIO": {"-": ["13"]},
    "DIRECTO": {"-": ["60"]},
    "DIRECTVM": {"1": ["12"], "2": ["60"]},
    "DIRECTVRA": {"-": ["11"]},
    "DIRECTVS": {"-": ["21"]},
    "DIREPTIO": {"-": ["13"]},
    "DIREPTOR": {"-": ["13"]},
    "DIRIGO": {"-": ["53"]},
    "DIRIMO": {"-": ["53"]},
    "DIRIPIO": {"-": ["55"]},
    "DIRVM": {"-": ["60"]},
    "DIRVMPO": {"-": ["53"]},
    "DIRVO": {"-": ["53"]},
    "DIRVS": {"-": ["21"]},
    "DIS": {"-": ["25"], "N": ["13"]},
    "DISCEDO": {"1": ["53"]},
    "DISCENS": {"-": ["13"]},
    "DISCEPTATIO": {"-": ["13"]},
    "DISCEPTATOR": {"-": ["13"]},
    "DISCEPTO": {"-": ["51"]},
    "DISCERNO": {"-": ["53"]},
    "DISCERPO": {"-": ["53"]},
    "DISCESSIO": {"-": ["13"]},
    "DISCESSVS": {"-": ["14"]},
    "DISCIDIVM": {"-": ["12"]},
    "DISCIDO": {"2": ["53"]},
    "DISCINCTVS": {"-": ["21"]},
    "DISCINDO": {"-": ["53"]},
    "DISCINGO": {"-": ["53"]},
    "DISCIPLINA": {"-": ["11"]},
    "DISCIPLINOSVS": {"-": ["21"]},
    "DISCIPVLA": {"-": ["11"]},
    "DISCIPVLVS": {"-": ["12"]},
    "DISCITO": {"1": ["51"]},
    "DISCLVDO": {"-": ["53"]},
    "DISCO": {"-": ["53"]},
    "DISCOLOR": {"-": ["25"]},
    "DISCONVENIO": {"-": ["54"]},
    "DISCOQVO": {"-": ["53"]},
    "DISCORDIA": {"-": ["11"], "N": ["11"]},
    "DISCORDIOSVS": {"-": ["21"]},
    "DISCORDO": {"-": ["51"]},
    "DISCORS": {"-": ["25"]},
    "DISCREPANTIA": {"-": ["11"]},
    "DISCREPITO": {"-": ["51"]},
    "DISCREPO": {"-": ["51"]},
    "DISCRESCO": {"-": ["53"]},
    "DISCRIBO": {"-": ["53"]},
    "DISCRIMEN": {"-": ["13"]},
    "DISCRIMINO": {"-": ["51"]},
    "DISCRVCIO": {"-": ["51"]},
    "DISCVMBO": {"-": ["53"]},
    "DISCVPIO": {"-": ["55"]},
    "DISCVRRO": {"-": ["53"]},
    "DISCVRSATIO": {"-": ["13"]},
    "DISCVRSO": {"-": ["51"]},
    "DISCVRSVS": {"-": ["14"]},
    "DISCVS": {"-": ["12"]},
    "DISCVTIO": {"-": ["55"]},
    "DISDIAPENTE": {"-": ["17"]},
    "DISDIATESSARON": {"-": ["17"]},
    "DISDO": {"-": ["53"]},
    "DISERTE": {"-": ["60"]},
    "DISERTVS": {"-": ["21"]},
    "DISHIASCO": {"-": ["53"]},
    "DISICIO": {"-": ["55"]},
    "DISIECTO": {"-": ["51"]},
    "DISIECTVS": {"-": ["14"]},
    "DISPANDO": {"2": ["53"]},
    "DISPAR": {"-": ["25"]},
    "DISPARATIO": {"-": ["13"]},
    "DISPARO": {"2": ["51"]},
    "DISPECTO": {"-": ["51"]},
    "DISPECTVS": {"-": ["14"]},
    "DISPELLO": {"-": ["53"]},
    "DISPENDIVM": {"-": ["12"]},
    "DISPENDO": {"-": ["53"]},
    "DISPENSATIO": {"-": ["13"]},
    "DISPENSATOR": {"-": ["13"]},
    "DISPENSO": {"-": ["51"]},
    "DISPERDO": {"-": ["53"]},
    "DISPEREO": {"1": ["56"]},
    "DISPERGO": {"-": ["53"]},
    "DISPERSE": {"-": ["60"]},
    "DISPERTIO": {"-": ["54"]},
    "DISPESCO": {"-": ["53"]},
    "DISPICIO": {"-": ["55"]},
    "DISPLICENTIA": {"-": ["11"]},
    "DISPLICEO": {"-": ["52"]},
    "DISPLODO": {"-": ["53"]},
    "DISPLVVIATVS": {"-": ["21"]},
    "DISPONO": {"-": ["53"]},
    "DISPOSITE": {"-": ["60"]},
    "DISPOSITIO": {"-": ["13"]},
    "DISPOSITVM": {"-": ["12"]},
    "DISPOSITVRA": {"-": ["11"]},
    "DISPOSITVS": {"-": ["21"], "2": ["21"], "1": ["12"]},
    "DISPVNGO": {"-": ["53"]},
    "DISPVTABILIS": {"-": ["24"]},
    "DISPVTATIO": {"-": ["13"]},
    "DISPVTATIVNCVLA": {"-": ["11"]},
    "DISPVTO": {"-": ["51"]},
    "DISQVIRO": {"-": ["53"]},
    "DISQVISITIO": {"-": ["13"]},
    "DISSAEPIO": {"-": ["54"]},
    "DISSAEPTIO": {"-": ["13"]},
    "DISSAEPTVM": {"-": ["12"]},
    "DISSEMINO": {"-": ["51"]},
    "DISSENSIO": {"-": ["13"]},
    "DISSENTIO": {"-": ["54"]},
    "DISSERO": {"1": ["51"], "2": ["53"], "3": ["53"]},
    "DISSERPO": {"-": ["53"]},
    "DISSERTATIO": {"-": ["13"]},
    "DISSERTIO": {"-": ["13"]},
    "DISSERTO": {"-": ["51"]},
    "DISSICIO": {"-": ["55"]},
    "DISSIDEO": {"-": ["52"]},
    "DISSIGNATOR": {"-": ["13"]},
    "DISSIGNO": {"-": ["51"]},
    "DISSILIO": {"-": ["54"]},
    "DISSIMILIS": {"-": ["24"]},
    "DISSIMILITER": {"-": ["60"]},
    "DISSIMILITVDO": {"-": ["13"]},
    "DISSIMVLANTER": {"-": ["60"]},
    "DISSIMVLATIO": {"-": ["13"]},
    "DISSIMVLATOR": {"-": ["13"]},
    "DISSIMVLO": {"-": ["51"]},
    "DISSIPABILIS": {"-": ["24"]},
    "DISSIPO": {"-": ["51"]},
    "DISSOCIABILIS": {"-": ["24"]},
    "DISSOCIATIO": {"-": ["13"]},
    "DISSOCIO": {"-": ["51"]},
    "DISSOLVBILIS": {"-": ["24"]},
    "DISSOLVO": {"-": ["53"]},
    "DISSOLVTE": {"-": ["60"]},
    "DISSOLVTIO": {"-": ["13"]},
    "DISSOLVTVS": {"-": ["21"]},
    "DISSONO": {"1": ["51"]},
    "DISSONVS": {"-": ["21"]},
    "DISSVADEO": {"-": ["52"]},
    "DISSVASIO": {"-": ["13"]},
    "DISSVASOR": {"-": ["13"]},
    "DISSVLTO": {"-": ["51"]},
    "DISSVO": {"-": ["53"]},
    "DISTABESCO": {"-": ["53"]},
    "DISTANTIA": {"-": ["11"]},
    "DISTENDO": {"-": ["53"]},
    "DISTENTO": {"-": ["51"]},
    "DISTENTVS": {"1": ["14"], "2": ["21"], "3": ["21"]},
    "DISTERMINO": {"-": ["51"]},
    "DISTERO": {"-": ["53"]},
    "DISTICHON": {"-": ["17"]},
    "DISTINCTE": {"-": ["60"]},
    "DISTINCTIO": {"-": ["13"]},
    "DISTINCTVS": {"1": ["14"], "2": ["21"]},
    "DISTINEO": {"-": ["52"]},
    "DISTINGVO": {"-": ["53"]},
    "DISTO": {"-": ["51"]},
    "DISTORQVEO": {"-": ["52"]},
    "DISTORTVS": {"-": ["21"]},
    "DISTRACTIO": {"-": ["13"]},
    "DISTRACTVS": {"-": ["21"]},
    "DISTRAHO": {"-": ["53"]},
    "DISTRIBVO": {"-": ["53"]},
    "DISTRIBVTIO": {"-": ["13"]},
    "DISTRICTIO": {"-": ["13"]},
    "DISTRICTVS": {"-": ["21"]},
    "DISTRINGO": {"-": ["53"]},
    "DISTVRBO": {"2": ["51"]},
    "DITESCO": {"-": ["53"]},
    "DITHYRAMBVS": {"-": ["12"]},
    "DITO": {"-": ["51"]},
    "DIV": {"-": ["60"]},
    "DIVA": {"-": ["11"], "N": ["11"]},
    "DIVARICO": {"-": ["51"]},
    "DIVDICO": {"-": ["51"]},
    "DIVELLO": {"-": ["53"]},
    "DIVENDO": {"-": ["53"]},
    "DIVERBERO": {"2": ["51"]},
    "DIVERSE": {"-": ["60"]},
    "DIVERSITAS": {"-": ["13"]},
    "DIVERSORIVM": {"-": ["12"]},
    "DIVERSVM": {"-": ["12"]},
    "DIVERSVS": {"-": ["21"]},
    "DIVERTO": {"-": ["53"]},
    "DIVES": {"-": ["22"]},
    "DIVIDO": {"-": ["53"]},
    "DIVIDVVS": {"-": ["21"]},
    "DIVINA": {"1": ["11"], "2": ["12"]},
    "DIVINATIO": {"-": ["13"]},
    "DIVINITAS": {"-": ["13"]},
    "DIVINITVS": {"-": ["60"]},
    "DIVINO": {"-": ["51"]},
    "DIVINVM": {"-": ["12"]},
    "DIVINVS": {"1": ["12"], "2": ["21"]},
    "DIVISIO": {"-": ["13"]},
    "DIVISOR": {"-": ["13"]},
    "DIVISVS": {"1": ["14"], "2": ["21"]},
    "DIVITIAE": {"-": ["11"]},
    "DIVORTIVM": {"-": ["12"]},
    "DIVRNVM": {"N": ["12"], "-": ["12"]},
    "DIVRNVS": {"-": ["21"]},
    "DIVS": {"-": ["21"], "A": ["21"]},
    "DIVTINVS": {"-": ["21"]},
    "DIVTVRNITAS": {"-": ["13"]},
    "DIVTVRNVS": {"-": ["21"]},
    "DIVVLGO": {"1": ["51"]},
    "DIVVLSIO": {"-": ["13"]},
    "DIVVM": {"-": ["12"]},
    "DIVVS": {"1": ["12"], "2": ["21"], "A": ["21"]},
    "DO": {"-": ["53", "51"], "1": ["51"], "2": ["53"]},
    "DOCEO": {"-": ["52"]},
    "DOCIBILIS": {"-": ["24"]},
    "DOCILIS": {"-": ["24"], "N": ["13"]},
    "DOCIMVS": {"N": ["12"]},
    "DOCIS": {"-": ["13"]},
    "DOCTE": {"-": ["60"]},
    "DOCTOR": {"-": ["13"]},
    "DOCTRINA": {"-": ["11"]},
    "DOCTVS": {"-": ["21"]},
    "DOCVMEN": {"-": ["13"]},
    "DOCVMENTVM": {"-": ["12"]},
    "DODRA": {"-": ["11"]},
    "DODRANS": {"-": ["13"]},
    "DOGMA": {"-": ["17"]},
    "DOLABELLA": {"N": ["11"]},
    "DOLABRA": {"-": ["11"]},
    "DOLENS": {"-": ["25"]},
    "DOLEO": {"-": ["52"]},
    "DOLESCO": {"-": ["53"]},
    "DOLIOLVM": {"-": ["12"]},
    "DOLITO": {"-": ["51"]},
    "DOLIVM": {"-": ["12"]},
    "DOLO": {"-": ["51"]},
    "DOLOR": {"-": ["13"], "N": ["13"]},
    "DOLOSE": {"-": ["60"]},
    "DOLOSVS": {"-": ["21"]},
    "DOLVS": {"-": ["12"]},
    "DOMABILIS": {"-": ["24"]},
    "DOMESTICVS": {"1": ["12"], "2": ["21"]},
    "DOMICILIVM": {"-": ["12"]},
    "DOMINA": {"-": ["11"]},
    "DOMINANS": {"1": ["13"], "2": ["25"]},
    "DOMINATIO": {"-": ["13"]},
    "DOMINATOR": {"-": ["13"]},
    "DOMINATRIX": {"-": ["13"]},
    "DOMINATVS": {"-": ["14"]},
    "DOMINICVS": {"-": ["21"]},
    "DOMINIVM": {"-": ["12"]},
    "DOMINOR": {"-": ["5J"]},
    "DOMINVS": {"-": ["12"]},
    "DOMITIVS": {"N": ["12"], "A": ["21"]},
    "DOMITO": {"-": ["51"]},
    "DOMITOR": {"-": ["13"]},
    "DOMITRIX": {"-": ["13"]},
    "DOMO": {"-": ["51"]},
    "DOMVNCVLA": {"-": ["11"]},
    "DOMVS": {"-": ["16"]},
    "DONARIA": {"-": ["12"]},
    "DONATICVS": {"-": ["21"]},
    "DONATIO": {"-": ["13"]},
    "DONATIVVM": {"-": ["12"]},
    "DONATOR": {"-": ["13"]},
    "DONEC": {"-": ["82"]},
    "DONO": {"-": ["51"]},
    "DONVM": {"-": ["12"]},
    "DORCAS": {"-": ["13"]},
    "DORICE": {"-": ["60"]},
    "DORMIO": {"-": ["54"]},
    "DORMITO": {"-": ["51"]},
    "DORON": {"-": ["17"]},
    "DOROTHEVS": {"N": ["12"]},
    "DORSVM": {"-": ["12"]},
    "DORYPHORA": {"-": ["11"]},
    "DORYPHOROS": {"-": ["17"]},
    "DOS": {"-": ["13"]},
    "DOTALIS": {"-": ["24"]},
    "DOTATVS": {"-": ["21"]},
    "DOTO": {"-": ["51"], "N": ["11"]},
    "DRACHMA": {"-": ["11"]},
    "DRACO": {"-": ["13"]},
    "DREPANITANVS": {"N": ["21"], "A": ["21"]},
    "DROMAS": {"-": ["13"]},
    "DROPAX": {"-": ["13"]},
    "DRVIDAE": {"-": ["13"]},
    "DRVIDES": {"-": ["13"]},
    "DRVSILLA": {"N": ["11"]},
    "DRVSVS": {"N": ["12"]},
    "DVBIE": {"-": ["60"]},
    "DVBITABILIS": {"-": ["24"]},
    "DVBITANTER": {"-": ["60"]},
    "DVBITATIM": {"-": ["60"]},
    "DVBITATIO": {"-": ["13"]},
    "DVBITATIVVS": {"-": ["21"]},
    "DVBITATOR": {"-": ["13"]},
    "DVBITO": {"-": ["51"]},
    "DVBIVM": {"-": ["12"]},
    "DVBIVS": {"-": ["21"], "N": ["12"]},
    "DVCENI": {"-": ["33"]},
    "DVCENTESIMA": {"-": ["11"]},
    "DVCENTESIMVS": {"-": ["32"]},
    "DVCENTI": {"-": ["31"]},
    "DVCENTIES": {"-": ["36"]},
    "DVCO": {"-": ["53"]},
    "DVCTARIVS": {"-": ["21"]},
    "DVCTIO": {"-": ["13"]},
    "DVCTO": {"-": ["51"]},
    "DVCTOR": {"-": ["13"]},
    "DVCTVS": {"-": ["14"]},
    "DVDVM": {"-": ["60"]},
    "DVELLVM": {"-": ["12"]},
    "DVILIVS": {"N": ["12"]},
    "DVLCE": {"1": ["13"], "2": ["60"]},
    "DVLCEDO": {"-": ["13"]},
    "DVLCESCO": {"-": ["53"]},
    "DVLCIS": {"-": ["24"]},
    "DVM": {"1": ["60"], "2": ["82"]},
    "DVMETVM": {"-": ["12"]},
    "DVMMODO": {"-": ["82"]},
    "DVMOSVS": {"-": ["21"]},
    "DVMTAXAT": {"-": ["60"]},
    "DVMVS": {"-": ["12"]},
    "DVO": {"-": ["31"]},
    "DVODECIM": {"-": ["31"]},
    "DVODECIMVS": {"-": ["32"]},
    "DVODENI": {"-": ["33"]},
    "DVODEQVADRAGINTA": {"-": ["31"]},
    "DVODETRICESIMVS": {"-": ["32"]},
    "DVODETRICIES": {"-": ["36"]},
    "DVODETRIGINTA": {"-": ["31"]},
    "DVODEVICESIMVS": {"-": ["32"]},
    "DVODEVIGINTI": {"-": ["31"]},
    "DVOETVICESIMVS": {"-": ["32"]},
    "DVPLA": {"-": ["11"]},
    "DVPLEX": {"-": ["34", "22"]},
    "DVPLICATIO": {"-": ["13"]},
    "DVPLICITER": {"-": ["60"]},
    "DVPLICO": {"-": ["51"]},
    "DVPLO": {"1": ["51"], "2": ["60"]},
    "DVPLVM": {"-": ["12"]},
    "DVPLVS": {"-": ["21"]},
    "DVRABILIS": {"-": ["24"]},
    "DVRACINVS": {"-": ["21"]},
    "DVRAMEN": {"-": ["13"]},
    "DVRAMENTVM": {"-": ["12"]},
    "DVRATEVS": {"-": ["21"]},
    "DVRE": {"-": ["60"]},
    "DVREO": {"-": ["52"]},
    "DVRESCO": {"-": ["53"]},
    "DVRITER": {"-": ["60"]},
    "DVRITIA": {"-": ["11"]},
    "DVRITIES": {"-": ["15"]},
    "DVRITVDO": {"-": ["13"]},
    "DVRO": {"-": ["51"]},
    "DVRVM": {"-": ["12"]},
    "DVRVS": {"-": ["21"], "N": ["12"]},
    "DVVMVIR": {"-": ["12"]},
    "DVX": {"-": ["13"]},
    "DYNASTES": {"-": ["17"]},
    "DYRRACHIVM": {"N": ["12"]},
    "DYSEROS": {"-": ["17"]},
    "DYSPEPSIA": {"-": ["11"]},
    "EA": {"-": ["60"]},
    "EADEM": {"-": ["60"]},
    "EATENVS": {"-": ["60"]},
    "EBENVS": {"-": ["12"]},
    "EBIBO": {"2": ["53"]},
    "EBLANDIOR": {"-": ["5M"]},
    "EBRIETAS": {"-": ["13"]},
    "EBRIOSVS": {"-": ["21"]},
    "EBRIVS": {"-": ["21"]},
    "EBVLLIO": {"-": ["54"]},
    "EBVLVM": {"-": ["12"]},
    "EBVR": {"-": ["13"]},
    "EBVRNEVS": {"-": ["21"]},
    "ECCE": {"-": ["90"]},
    "ECCLESIA": {"-": ["11"]},
    "ECCLESIASTERION": {"-": ["17"]},
    "ECETRANI": {"N": ["12"]},
    "ECHEA": {"-": ["17"]},
    "ECHINVS": {"-": ["12"]},
    "ECHO": {"-": ["17"], "N": ["16"]},
    "ECNVBO": {"-": ["53"]},
    "ECPHORA": {"-": ["17"]},
    "ECQVANDO": {"-": ["67"]},
    "ECQVI": {"-": ["67"]},
    "ECQVID": {"-": ["67"]},
    "ECQVIS": {"-": ["47"]},
    "ECQVISNAM": {"-": ["47"]},
    "ECQVO": {"-": ["67"]},
    "ECTYPVS": {"-": ["21"]},
    "EDAX": {"-": ["25"]},
    "EDICO": {"2": ["53"]},
    "EDICTVM": {"-": ["12"]},
    "EDISCO": {"-": ["53"]},
    "EDISSERO": {"2": ["53"]},
    "EDITA": {"-": ["12"]},
    "EDITIO": {"-": ["13"]},
    "EDITVM": {"-": ["12"]},
    "EDITVS": {"1": ["14"], "2": ["21"]},
    "EDO": {"1": ["53"], "2": ["56"]},
    "EDOCEO": {"-": ["52"]},
    "EDOMO": {"-": ["51"]},
    "EDORMIO": {"-": ["54"]},
    "EDVCATIO": {"-": ["13"]},
    "EDVCATOR": {"-": ["13"]},
    "EDVCO": {"1": ["51"], "2": ["53"]},
    "EDVCTIO": {"-": ["13"]},
    "EDVLIS": {"-": ["24"]},
    "EDVRO": {"-": ["51"]},
    "EDVRVS": {"-": ["21"]},
    "EFFARCIO": {"-": ["54"]},
    "EFFATVM": {"-": ["12"]},
    "EFFECTVM": {"-": ["12"]},
    "EFFECTVS": {"1": ["14"], "2": ["21"]},
    "EFFEMINATE": {"-": ["60"]},
    "EFFEMINATVS": {"-": ["21"]},
    "EFFEMINO": {"-": ["51"]},
    "EFFERATVS": {"-": ["21"]},
    "EFFERCIO": {"-": ["54"]},
    "EFFERO": {"1": ["51"], "2": ["56"]},
    "EFFERVEO": {"-": ["52"]},
    "EFFERVESCO": {"-": ["53"]},
    "EFFERVO": {"-": ["53"]},
    "EFFERVS": {"-": ["21"]},
    "EFFETVS": {"-": ["21"]},
    "EFFICACITER": {"-": ["6&"]},
    "EFFICAX": {"-": ["25"]},
    "EFFICIENS": {"1": ["13"], "2": ["25"]},
    "EFFICIENTIA": {"-": ["11"]},
    "EFFICIO": {"-": ["55"]},
    "EFFIGIA": {"-": ["11"]},
    "EFFIGIES": {"-": ["15"]},
    "EFFINGO": {"-": ["53"]},
    "EFFLAGITATIO": {"-": ["13"]},
    "EFFLAGITATVS": {"-": ["14"]},
    "EFFLAGITO": {"-": ["51"]},
    "EFFLIGO": {"-": ["53"]},
    "EFFLO": {"-": ["51"]},
    "EFFLORESCO": {"-": ["53"]},
    "EFFLVO": {"-": ["53"]},
    "EFFLVVIVM": {"-": ["12"]},
    "EFFODIO": {"-": ["55"]},
    "EFFOR": {"-": ["5J"]},
    "EFFRACTARIVS": {"-": ["12"]},
    "EFFRENATVS": {"-": ["21"]},
    "EFFRENO": {"-": ["51"]},
    "EFFRENVS": {"-": ["21"]},
    "EFFRICO": {"-": ["51"]},
    "EFFRINGO": {"-": ["53"]},
    "EFFVGIO": {"-": ["55"]},
    "EFFVGIVM": {"-": ["12"]},
    "EFFVLGEO": {"-": ["52"]},
    "EFFVNDO": {"2": ["53"]},
    "EFFVNSVS": {"-": ["21"]},
    "EFFVSE": {"-": ["60"]},
    "EFFVSIO": {"-": ["13"]},
    "EFFVSVS": {"-": ["21"]},
    "EFFVTIO": {"-": ["54"]},
    "EFFVTVO": {"-": ["53"]},
    "EGELIDVS": {"-": ["21"]},
    "EGENS": {"C": ["25"], "-": ["25"]},
    "EGENVS": {"-": ["21"]},
    "EGEO": {"-": ["52"]},
    "EGERIA": {"N": ["11"]},
    "EGERIVS": {"N": ["12"]},
    "EGERO": {"-": ["53"]},
    "EGESTAS": {"-": ["13"], "N": ["13"]},
    "EGIGNO": {"-": ["53"]},
    "EGNATIVS": {"N": ["12"]},
    "EGO": {"-": ["41"]},
    "EGREDIOR": {"-": ["5N"]},
    "EGREGIE": {"-": ["60"]},
    "EGREGIVM": {"-": ["12"]},
    "EGREGIVS": {"-": ["21"]},
    "EGRESSVS": {"-": ["14"]},
    "EHEV": {"-": ["90"]},
    "EI": {"-": ["90"]},
    "EIA": {"-": ["90"]},
    "EIACVLOR": {"-": ["5J"]},
    "EICIO": {"-": ["55"]},
    "EIECTAMENTVM": {"-": ["12"]},
    "EIECTIO": {"-": ["13"]},
    "EIECTO": {"-": ["51"]},
    "EIECTVS": {"-": ["14"]},
    "EIVLATIO": {"-": ["13"]},
    "EIVLATVS": {"-": ["14"]},
    "EIVLO": {"-": ["51"]},
    "EIVRATIO": {"-": ["13"]},
    "EIVRO": {"-": ["51"]},
    "EIVSMODI": {"-": ["60"]},
    "ELABOR": {"2": ["5L"]},
    "ELABORO": {"-": ["51"]},
    "ELAEOTHESIVM": {"-": ["12"]},
    "ELANGVESCO": {"-": ["53"]},
    "ELARGIOR": {"-": ["5M"]},
    "ELATE": {"1": ["17"], "2": ["60"]},
    "ELATIO": {"-": ["13"]},
    "ELATRO": {"2": ["51"]},
    "ELATVS": {"-": ["21"]},
    "ELAVO": {"-": ["51"]},
    "ELECTIO": {"-": ["13"]},
    "ELECTO": {"1": ["51"], "2": ["51"]},
    "ELECTRVM": {"-": ["12"]},
    "ELECTVS": {"1": ["14"], "2": ["21"]},
    "ELEGANS": {"-": ["25"]},
    "ELEGANTER": {"-": ["60"]},
    "ELEGANTIA": {"-": ["11"]},
    "ELEGEON": {"-": ["17"]},
    "ELEGI": {"-": ["12"]},
    "ELEGIDION": {"-": ["17"]},
    "ELEGO": {"1": ["51"]},
    "ELEMENTA": {"-": ["12"]},
    "ELEMENTARIVS": {"-": ["21"]},
    "ELEMENTVM": {"-": ["12"]},
    "ELENCHVS": {"-": ["12"]},
    "ELEPHANTVS": {"-": ["12"]},
    "ELEPHAS": {"-": ["13"]},
    "ELEVO": {"1": ["51"]},
    "ELICIO": {"-": ["55"]},
    "ELICIVS": {"N": ["12"]},
    "ELIDO": {"-": ["53"]},
    "ELIGO": {"3": ["53"]},
    "ELIMINO": {"-": ["51"]},
    "ELINGVIS": {"-": ["24"]},
    "ELIQVESCO": {"-": ["53"]},
    "ELIQVO": {"-": ["51"]},
    "ELIS": {"N": ["17"]},
    "ELISIO": {"-": ["13"]},
    "ELIXVS": {"-": ["21"]},
    "ELLIPSIS": {"-": ["17"]},
    "ELLYCHNIVM": {"-": ["12"]},
    "ELOCO": {"-": ["51"]},
    "ELOGIVM": {"-": ["12"]},
    "ELOQVENS": {"-": ["25"]},
    "ELOQVENTIA": {"-": ["11"]},
    "ELOQVIVM": {"-": ["12"]},
    "ELOQVOR": {"-": ["5L"]},
    "ELORINI": {"N": ["12"]},
    "ELORINVS": {"N": ["21"], "A": ["21"]},
    "ELORVS": {"N": ["12"]},
    "ELVACRVM": {"-": ["12"]},
    "ELVCEO": {"-": ["52"]},
    "ELVCTOR": {"-": ["5J"]},
    "ELVCVBRO": {"-": ["51"]},
    "ELVDO": {"-": ["53"]},
    "ELVMBIS": {"-": ["24"]},
    "ELVO": {"-": ["53"]},
    "ELVTVS": {"-": ["21"]},
    "ELVVIES": {"-": ["15"]},
    "EMACERATVS": {"-": ["21"]},
    "EMANCIPO": {"-": ["51"]},
    "EMANO": {"-": ["51"]},
    "EMARCEO": {"-": ["52"]},
    "EMARCESCO": {"-": ["53"]},
    "EMAX": {"-": ["25"]},
    "EMBATER": {"-": ["17"]},
    "EMBLEMA": {"-": ["17"]},
    "EMBOLIVM": {"-": ["12"]},
    "EMBOLVS": {"-": ["12"]},
    "EMENDABILIS": {"-": ["24"]},
    "EMENDATE": {"-": ["60"]},
    "EMENDATIO": {"-": ["13"]},
    "EMENDATOR": {"-": ["13"]},
    "EMENDATVS": {"-": ["21"]},
    "EMENDICATICIVS": {"-": ["21"]},
    "EMENDO": {"-": ["51"]},
    "EMENTIOR": {"-": ["5M"]},
    "EMERCOR": {"-": ["5J"]},
    "EMEREO": {"-": ["52"]},
    "EMEREOR": {"-": ["5K"]},
    "EMERGO": {"-": ["53"]},
    "EMERITVM": {"-": ["12"]},
    "EMERITVS": {"-": ["21"]},
    "EMERSVS": {"-": ["14"]},
    "EMETIOR": {"-": ["5M"]},
    "EMETO": {"2": ["53"]},
    "EMICO": {"-": ["51"]},
    "EMIGRO": {"-": ["51"]},
    "EMINENS": {"-": ["25"]},
    "EMINEO": {"-": ["52"]},
    "EMINVS": {"-": ["60"]},
    "EMIROR": {"-": ["5J"]},
    "EMISSARIVS": {"-": ["12"]},
    "EMISSVS": {"-": ["14"]},
    "EMITTO": {"-": ["53"]},
    "EMO": {"-": ["53"]},
    "EMODEROR": {"-": ["5J"]},
    "EMODVLOR": {"-": ["5J"]},
    "EMOLIOR": {"-": ["5M"]},
    "EMOLLIO": {"-": ["54"]},
    "EMOLO": {"-": ["53"]},
    "EMOLVMENTVM": {"-": ["12"]},
    "EMORIOR": {"-": ["5N"]},
    "EMOVEO": {"-": ["52"]},
    "EMPLASTRVM": {"-": ["12"]},
    "EMPLECTOS": {"-": ["27"]},
    "EMPORIVM": {"-": ["12"]},
    "EMPTIO": {"-": ["13"]},
    "EMPTITO": {"-": ["51"]},
    "EMPTOR": {"-": ["13"]},
    "EMVLGEO": {"-": ["52"]},
    "EMVNDO": {"-": ["51"]},
    "EMVNGO": {"-": ["53"]},
    "EMVNIO": {"2": ["54"]},
    "EN": {"-": ["90"]},
    "ENARRABILIS": {"-": ["24"]},
    "ENARRATIO": {"-": ["13"]},
    "ENARRO": {"-": ["51"]},
    "ENASCOR": {"-": ["5L"]},
    "ENATO": {"-": ["51"]},
    "ENAVIGO": {"-": ["51"]},
    "ENCARPI": {"-": ["12"]},
    "ENCELADOS": {"-": ["17"]},
    "ENCHYTVM": {"-": ["12"]},
    "ENCLIMA": {"-": ["17"]},
    "ENCYCLIOS": {"-": ["27"]},
    "ENDROMIS": {"-": ["17"]},
    "ENECO": {"-": ["51"]},
    "ENERVATVS": {"A": ["21"], "-": ["21"]},
    "ENERVIS": {"-": ["24"]},
    "ENERVO": {"-": ["51"]},
    "ENGVINI": {"N": ["12"]},
    "ENGVINVS": {"N": ["21"], "A": ["21"]},
    "ENIM": {"1": ["60"], "2": ["81"]},
    "ENIMVERO": {"-": ["60"]},
    "ENITEO": {"-": ["52"]},
    "ENITESCO": {"-": ["53"]},
    "ENITOR": {"2": ["5L"]},
    "ENIXE": {"-": ["60"]},
    "ENIXVS": {"1": ["14"], "2": ["21"]},
    "ENNIVS": {"N": ["12"]},
    "ENO": {"-": ["51"]},
    "ENODATIO": {"-": ["13"]},
    "ENODIS": {"-": ["24"]},
    "ENODO": {"-": ["51"]},
    "ENORMIS": {"-": ["24"]},
    "ENORMITAS": {"-": ["13"]},
    "ENOTESCO": {"-": ["53"]},
    "ENS": {"-": ["13"]},
    "ENSIGER": {"-": ["21"]},
    "ENSIS": {"-": ["13"]},
    "ENTASIS": {"-": ["17"]},
    "ENTELLINI": {"N": ["12"]},
    "ENTELLINVS": {"N": ["21"], "A": ["21"]},
    "ENTHEVS": {"-": ["21"]},
    "ENVCLEATVS": {"-": ["21"]},
    "ENVCLEO": {"-": ["51"]},
    "ENVMERATIO": {"-": ["13"]},
    "ENVMERO": {"1": ["51"]},
    "ENVNTIATIVVS": {"-": ["21"]},
    "ENVNTIATVM": {"-": ["12"]},
    "ENVNTIO": {"-": ["51"]},
    "ENVTRIO": {"-": ["54"]},
    "EO": {"1": ["56"], "3": ["60"]},
    "EODEM": {"-": ["60"]},
    "EOVS": {"N": ["21"], "A": ["21"]},
    "EPAEETIS": {"-": ["17"]},
    "EPAGON": {"-": ["57"]},
    "EPHEBEVM": {"-": ["12"]},
    "EPHEBVS": {"-": ["12"]},
    "EPHECTOS": {"-": ["27"]},
    "EPHEMERIDAE": {"-": ["11"]},
    "EPHEMERIS": {"-": ["17"]},
    "EPHESII": {"N": ["12"]},
    "EPHESIVS": {"N": ["21"], "A": ["21"]},
    "EPHESVS": {"N": ["17"]},
    "EPHIPPIATVS": {"-": ["21"]},
    "EPHIPPIVM": {"-": ["12"]},
    "EPIBATA": {"-": ["11"]},
    "EPIBATHRA": {"-": ["17"]},
    "EPICRATES": {"N": ["17"]},
    "EPICVRVS": {"N": ["12"]},
    "EPIDIMOEROS": {"-": ["27"]},
    "EPIDROMVS": {"-": ["12"]},
    "EPIGRAMMA": {"-": ["17"]},
    "EPIGRVS": {"-": ["12"]},
    "EPIMENIVS": {"-": ["21"]},
    "EPIPEMPTOS": {"-": ["27"]},
    "EPIRHEDIVM": {"-": ["12"]},
    "EPIRVS": {"N": ["17"]},
    "EPISCAENOS": {"-": ["17"]},
    "EPISCOPVS": {"-": ["12"]},
    "EPISTATES": {"-": ["17"]},
    "EPISTOLA": {"-": ["11"], "N": ["11"]},
    "EPISTOLIVM": {"-": ["12"]},
    "EPISTOMIVM": {"-": ["12"]},
    "EPISTYLIVM": {"-": ["12"]},
    "EPITONIVM": {"-": ["12"]},
    "EPITOXIS": {"-": ["17"]},
    "EPITRITOS": {"-": ["27"]},
    "EPITYRVM": {"-": ["12"]},
    "EPIZYGIS": {"-": ["17"]},
    "EPOS": {"-": ["17"]},
    "EPOTO": {"-": ["51"]},
    "EPVLAE": {"-": ["11"]},
    "EPVLOR": {"-": ["5J"]},
    "EPVLVM": {"-": ["12"]},
    "EQVA": {"-": ["11"]},
    "EQVES": {"-": ["13"]},
    "EQVESTER": {"-": ["23"], "A": ["13"]},
    "EQVESTRIA": {"Si\u00e8ges": ["13"], "-": ["13"]},
    "EQVIDEM": {"-": ["60"]},
    "EQVILE": {"-": ["13"]},
    "EQVINVS": {"-": ["21"]},
    "EQVITABILIS": {"-": ["24"]},
    "EQVITATVS": {"1": ["14"], "2": ["14"]},
    "EQVITO": {"-": ["51"]},
    "EQVOR": {"-": ["5L"]},
    "EQVVLEVS": {"-": ["12"]},
    "EQVVS": {"-": ["12"]},
    "ERADO": {"-": ["53"]},
    "ERCEO": {"-": ["52"]},
    "ERECTIO": {"-": ["13"]},
    "ERECTVS": {"-": ["21"]},
    "EREPO": {"-": ["53"]},
    "EREPTIO": {"-": ["13"]},
    "EREPTOR": {"-": ["13"]},
    "ERGA": {"-": ["70"]},
    "ERGASTVLVM": {"-": ["12"]},
    "ERGATA": {"-": ["11"]},
    "ERGO": {"1": ["70"], "2": ["81"]},
    "ERICIVS": {"-": ["12"]},
    "ERIGO": {"-": ["53"]},
    "ERIPHYLA": {"N": ["11"]},
    "ERIPIO": {"-": ["55"]},
    "ERISMA": {"-": ["16"]},
    "ERNEVM": {"-": ["12"]},
    "ERODO": {"-": ["53"]},
    "EROGATIO": {"-": ["13"]},
    "EROGO": {"-": ["51"]},
    "EROMENION": {"-": ["17"]},
    "EROS": {"2N": ["17"], "N": ["12"]},
    "ERRABVNDVS": {"-": ["21"]},
    "ERRATICVS": {"-": ["21"]},
    "ERRATIO": {"-": ["13"]},
    "ERRATVM": {"-": ["12"]},
    "ERRO": {"1": ["13"], "2": ["51"]},
    "ERROR": {"-": ["13"], "N": ["13"]},
    "ERVBEO": {"-": ["52"]},
    "ERVBESCO": {"-": ["53"]},
    "ERVCA": {"-": ["11"]},
    "ERVCTO": {"-": ["51"]},
    "ERVDERO": {"-": ["51"]},
    "ERVDIO": {"-": ["54"]},
    "ERVDITI": {"-": ["12"]},
    "ERVDITIO": {"-": ["13"]},
    "ERVDITVLVS": {"-": ["21"]},
    "ERVDITVS": {"-": ["21"]},
    "ERVMPO": {"-": ["53"]},
    "ERVNCO": {"-": ["51"]},
    "ERVO": {"-": ["53"]},
    "ERVPTIO": {"-": ["13"]},
    "ERVVM": {"-": ["12"]},
    "ERYCINVS": {"N": ["21"], "A": ["21"]},
    "ERYCVS": {"N": ["12"]},
    "ERYMANTHIVS": {"N": ["21"], "A": ["21"]},
    "ERYTHRAE": {"N": ["11"]},
    "ESCA": {"-": ["11"]},
    "ESCARIVS": {"-": ["21"]},
    "ESCENDO": {"-": ["53"]},
    "ESCENSIO": {"-": ["13"]},
    "ESCENSVS": {"-": ["14"]},
    "ESCHARA": {"-": ["17"]},
    "ESITO": {"-": ["51"]},
    "ESQVILIAE": {"N": ["11"]},
    "ESQVILINVS": {"N": ["21"], "A": ["21"]},
    "ESSEDA": {"-": ["11"]},
    "ESSEDARIVS": {"1": ["12"], "2": ["21"]},
    "ESSEDVM": {"-": ["12"]},
    "ESSENTIA": {"-": ["11"]},
    "ESVRIO": {"1": ["13"], "2": ["54"]},
    "ESVRITIO": {"-": ["13"]},
    "ET": {"1": ["60"], "2": ["81"]},
    "ETENIM": {"-": ["81"]},
    "ETESIAE": {"-": ["11"]},
    "ETESIVS": {"-": ["21"], "N": ["12"]},
    "ETHOLOGIA": {"-": ["17"]},
    "ETIAM": {"-": ["60"]},
    "ETIAMNVNC": {"-": ["60"]},
    "ETIAMSI": {"-": ["82"]},
    "ETIAMTVM": {"-": ["60"]},
    "ETIAMTVNC": {"-": ["60"]},
    "ETRVRIA": {"N": ["11"]},
    "ETRVSCI": {"N": ["12"]},
    "ETRVSCVS": {"N": ["21"], "A": ["21"]},
    "ETSI": {"1": ["81"], "2": ["82"]},
    "EV": {"-": ["90"]},
    "EVADO": {"2": ["53"]},
    "EVAGATIO": {"-": ["13"]},
    "EVAGOR": {"2": ["5J"]},
    "EVALESCO": {"-": ["53"]},
    "EVANDER": {"N": ["12"]},
    "EVANESCO": {"-": ["53"]},
    "EVANIDVS": {"-": ["21"]},
    "EVANS": {"-": ["13"]},
    "EVBVLIDA": {"N": ["11"]},
    "EVBVLIDAS": {"N": ["17"]},
    "EVECTIO": {"-": ["13"]},
    "EVECTO": {"-": ["51"]},
    "EVECTVS": {"-": ["14"]},
    "EVEHO": {"-": ["53"]},
    "EVELLO": {"-": ["53"]},
    "EVENIO": {"-": ["54"]},
    "EVENTVM": {"-": ["12"]},
    "EVENTVRA": {"-": ["12"]},
    "EVENTVS": {"-": ["14"]},
    "EVERBERO": {"2": ["51"]},
    "EVERRICVLVM": {"-": ["12"]},
    "EVERRO": {"-": ["53"]},
    "EVERSIO": {"-": ["13"]},
    "EVERSOR": {"-": ["13"]},
    "EVERTO": {"-": ["53"]},
    "EVGANEI": {"N": ["12"]},
    "EVGE": {"-": ["90"]},
    "EVGENEVS": {"-": ["21"]},
    "EVHYIVS": {"-": ["17"]},
    "EVIDENS": {"-": ["25"]},
    "EVIDENTIA": {"-": ["11"]},
    "EVIGILO": {"-": ["51"]},
    "EVINCIO": {"-": ["54"]},
    "EVINCO": {"-": ["53"]},
    "EVIRO": {"-": ["51"]},
    "EVITABILIS": {"-": ["24"]},
    "EVITO": {"-": ["51"]},
    "EVMENIDAS": {"N": ["17"]},
    "EVNVCHVS": {"-": ["12"]},
    "EVOCATI": {"-": ["12"]},
    "EVOCATIO": {"-": ["13"]},
    "EVOCATOR": {"-": ["13"]},
    "EVOCATVS": {"-": ["21"]},
    "EVOCO": {"-": ["51"]},
    "EVOE": {"-": ["90"]},
    "EVOLO": {"2": ["51"]},
    "EVOLVO": {"-": ["53"]},
    "EVOMO": {"-": ["53"]},
    "EVPHRANOR": {"-": ["13"], "N": ["13"]},
    "EVPOLEMVS": {"N": ["12"]},
    "EVRIPVS": {"-": ["12"], "N": ["12"]},
    "EVROPA": {"N": ["11"]},
    "EVROVS": {"-": ["21"]},
    "EVRVS": {"-": ["12"], "N": ["12"]},
    "EVRYTHMIA": {"-": ["11"]},
    "EVSTYLOS": {"-": ["27"]},
    "EVTHVS": {"-": ["27"]},
    "EVTHYGRAMMOS": {"-": ["27"]},
    "EVTHYMIA": {"-": ["17"]},
    "EVVLGO": {"1": ["51"]},
    "EVVLSIO": {"-": ["13"]},
    "EX": {"-": ["70"]},
    "EXACERBO": {"-": ["51"]},
    "EXACTIO": {"-": ["13"]},
    "EXACTOR": {"-": ["13"]},
    "EXACTVS": {"1": ["14"], "2": ["21"]},
    "EXACVO": {"-": ["53"]},
    "EXADVERSVM": {"1": ["60"], "2": ["70"]},
    "EXAEDIFICO": {"-": ["51"]},
    "EXAEQVATIO": {"-": ["13"]},
    "EXAEQVO": {"-": ["51"]},
    "EXAESTVO": {"-": ["51"]},
    "EXAGGERO": {"1": ["51"]},
    "EXAGITO": {"-": ["51"]},
    "EXAMEN": {"-": ["13"]},
    "EXAMINATIO": {"-": ["13"]},
    "EXAMINO": {"-": ["51"]},
    "EXANCLO": {"-": ["51"]},
    "EXANIMIS": {"-": ["24"]},
    "EXANIMO": {"-": ["51"]},
    "EXANIMVS": {"-": ["21"]},
    "EXARDEO": {"-": ["52"]},
    "EXARDESCO": {"-": ["53"]},
    "EXAREO": {"-": ["52"]},
    "EXARESCO": {"-": ["53"]},
    "EXARMO": {"-": ["51"]},
    "EXARO": {"-": ["51"]},
    "EXASPERO": {"-": ["51"]},
    "EXAVCTORO": {"-": ["51"]},
    "EXAVDIO": {"-": ["54"]},
    "EXAVGEO": {"-": ["52"]},
    "EXAVGVRO": {"-": ["51"]},
    "EXCAECO": {"-": ["51"]},
    "EXCALCEO": {"-": ["52"]},
    "EXCANDEO": {"-": ["52"]},
    "EXCANDESCO": {"-": ["53"]},
    "EXCANTO": {"-": ["51"]},
    "EXCARNIFICO": {"-": ["51"]},
    "EXCAVO": {"1": ["51"]},
    "EXCEDO": {"1": ["53"]},
    "EXCELLENS": {"-": ["25"]},
    "EXCELLENTIA": {"-": ["11"]},
    "EXCELLO": {"-": ["53"]},
    "EXCELSE": {"-": ["6&"]},
    "EXCELSITAS": {"-": ["13"]},
    "EXCELSVM": {"-": ["12"]},
    "EXCELSVS": {"-": ["21"]},
    "EXCEPTIO": {"-": ["13"]},
    "EXCEPTIVNCVLA": {"-": ["11"]},
    "EXCEPTO": {"-": ["51"]},
    "EXCERNO": {"-": ["53"]},
    "EXCERPO": {"-": ["53"]},
    "EXCESSVS": {"-": ["14"]},
    "EXCIDIVM": {"1": ["12"], "2": ["12"]},
    "EXCIDO": {"1": ["53"], "2": ["53"]},
    "EXCIEO": {"-": ["52"]},
    "EXCIO": {"-": ["54"]},
    "EXCIPIO": {"-": ["55"]},
    "EXCISIO": {"-": ["13"]},
    "EXCITATVS": {"-": ["21"]},
    "EXCITO": {"1": ["51"]},
    "EXCLAMATIO": {"-": ["13"]},
    "EXCLAMO": {"-": ["51"]},
    "EXCLARO": {"-": ["51"]},
    "EXCLVDO": {"-": ["53"]},
    "EXCLVSIO": {"-": ["13"]},
    "EXCLVSVS": {"-": ["21"]},
    "EXCOGITO": {"-": ["51"]},
    "EXCOLO": {"1": ["51"], "2": ["53"]},
    "EXCOQVO": {"-": ["53"]},
    "EXCORS": {"-": ["25"]},
    "EXCREMENTVM": {"-": ["12"]},
    "EXCREO": {"-": ["51"]},
    "EXCRESCO": {"-": ["53"]},
    "EXCRETVS": {"-": ["21"]},
    "EXCRVCIO": {"-": ["51"]},
    "EXCVBIAE": {"-": ["11"]},
    "EXCVBITOR": {"-": ["13"]},
    "EXCVBITVS": {"-": ["14"]},
    "EXCVBO": {"-": ["51"]},
    "EXCVDO": {"-": ["53"]},
    "EXCVLCO": {"-": ["51"]},
    "EXCVRRO": {"-": ["53"]},
    "EXCVRSIO": {"-": ["13"]},
    "EXCVRSOR": {"-": ["13"]},
    "EXCVRSVS": {"-": ["14"]},
    "EXCVSATE": {"-": ["6-"]},
    "EXCVSATIO": {"-": ["13"]},
    "EXCVSATVS": {"-": ["21"]},
    "EXCVSO": {"-": ["51"]},
    "EXCVSSE": {"-": ["60"]},
    "EXCVSSVS": {"-": ["21"]},
    "EXCVTIO": {"-": ["55"]},
    "EXEDO": {"2": ["56"]},
    "EXEDRA": {"-": ["11"]},
    "EXEMPLAR": {"-": ["13"]},
    "EXEMPLVM": {"-": ["12"]},
    "EXEMPTVS": {"-": ["14"]},
    "EXEO": {"1": ["56"]},
    "EXERCEO": {"2": ["52"], "-": ["52"]},
    "EXERCITATE": {"-": ["6&"]},
    "EXERCITATIO": {"-": ["13"]},
    "EXERCITATVS": {"-": ["21"]},
    "EXERCITIO": {"-": ["13"]},
    "EXERCITIVM": {"-": ["12"]},
    "EXERCITO": {"-": ["51"]},
    "EXERCITVS": {"1": ["14"], "2": ["21"]},
    "EXESOR": {"-": ["13"]},
    "EXFLORESCO": {"-": ["53"]},
    "EXHALO": {"-": ["51"]},
    "EXHAVRIO": {"-": ["54"]},
    "EXHEREDO": {"-": ["51"]},
    "EXHIBEO": {"-": ["52"]},
    "EXHILARO": {"-": ["51"]},
    "EXHORREO": {"-": ["52"]},
    "EXHORRESCO": {"-": ["53"]},
    "EXHORTATIO": {"-": ["13"]},
    "EXHORTOR": {"-": ["5J"]},
    "EXIGO": {"-": ["53"]},
    "EXIGVE": {"-": ["60"]},
    "EXIGVITAS": {"-": ["13"]},
    "EXIGVO": {"-": ["60"]},
    "EXIGVVM": {"1": ["12"], "2": ["60"]},
    "EXIGVVS": {"-": ["21"]},
    "EXILIS": {"-": ["24"]},
    "EXILITAS": {"-": ["13"]},
    "EXILITER": {"-": ["60"]},
    "EXIMIE": {"-": ["60"]},
    "EXIMIVS": {"-": ["21"]},
    "EXIMO": {"-": ["53"]},
    "EXINANIO": {"-": ["54"]},
    "EXINDE": {"-": ["60"]},
    "EXISTIMATIO": {"-": ["13"]},
    "EXISTIMATOR": {"-": ["13"]},
    "EXISTIMO": {"-": ["51"]},
    "EXITIABILIS": {"-": ["24"]},
    "EXITIALIS": {"-": ["24"]},
    "EXITIOSVS": {"-": ["21"]},
    "EXITIVM": {"-": ["12"]},
    "EXITVS": {"-": ["14"]},
    "EXODIVM": {"-": ["12"]},
    "EXOLEO": {"-": ["52"]},
    "EXOLESCO": {"-": ["53"]},
    "EXOLETI": {"-": ["12"]},
    "EXOLETVS": {"-": ["21"]},
    "EXONERO": {"-": ["51"]},
    "EXOPTATVS": {"-": ["21"]},
    "EXOPTO": {"-": ["51"]},
    "EXORABILIS": {"-": ["24"]},
    "EXORDIOR": {"-": ["5M"]},
    "EXORDIVM": {"-": ["12"]},
    "EXORIOR": {"-": ["5M"]},
    "EXORNO": {"-": ["51"]},
    "EXORO": {"-": ["51"]},
    "EXORSA": {"-": ["12"]},
    "EXORSVS": {"-": ["14"]},
    "EXORTVS": {"-": ["14"]},
    "EXOS": {"-": ["25"]},
    "EXOSCVLOR": {"-": ["5J"]},
    "EXOSSO": {"-": ["51"]},
    "EXOSVS": {"-": ["21"]},
    "EXPALLEO": {"-": ["52"]},
    "EXPALLESCO": {"-": ["53"]},
    "EXPANDO": {"2": ["53"]},
    "EXPATRO": {"-": ["51"]},
    "EXPAVEO": {"-": ["52"]},
    "EXPAVESCO": {"-": ["53"]},
    "EXPEDIO": {"-": ["54"]},
    "EXPEDITE": {"-": ["60"]},
    "EXPEDITIO": {"-": ["13"]},
    "EXPEDITO": {"-": ["54"]},
    "EXPEDITVM": {"-": ["12"]},
    "EXPEDITVS": {"-": ["21"]},
    "EXPELLO": {"-": ["53"]},
    "EXPENDO": {"-": ["53"]},
    "EXPENSO": {"-": ["51"]},
    "EXPENSVM": {"-": ["12"]},
    "EXPERGEFACIO": {"-": ["55"]},
    "EXPERGISCOR": {"-": ["5L"]},
    "EXPERGO": {"-": ["53"]},
    "EXPERIENS": {"-": ["25"]},
    "EXPERIENTIA": {"-": ["11"]},
    "EXPERIMENTVM": {"-": ["12"]},
    "EXPERIOR": {"-": ["5M"]},
    "EXPERISCOR": {"-": ["5L"]},
    "EXPERS": {"-": ["25"]},
    "EXPERTIO": {"-": ["13"]},
    "EXPERTVS": {"-": ["21"]},
    "EXPETIBILIS": {"-": ["24"]},
    "EXPETO": {"-": ["53"]},
    "EXPILATIO": {"-": ["13"]},
    "EXPILO": {"2": ["51"]},
    "EXPINSO": {"-": ["53"]},
    "EXPIO": {"-": ["51"]},
    "EXPLANABILIS": {"-": ["24"]},
    "EXPLANATIO": {"-": ["13"]},
    "EXPLANATVS": {"-": ["21"]},
    "EXPLANO": {"-": ["51"]},
    "EXPLAVDO": {"-": ["53"]},
    "EXPLEO": {"-": ["52"]},
    "EXPLETVS": {"-": ["21"]},
    "EXPLICABILIS": {"-": ["24"]},
    "EXPLICATIO": {"-": ["13"]},
    "EXPLICATVS": {"1": ["14"], "2": ["21"]},
    "EXPLICITVS": {"2": ["21"], "-": ["21"]},
    "EXPLICO": {"-": ["51"]},
    "EXPLODO": {"-": ["53"]},
    "EXPLORATIO": {"-": ["13"]},
    "EXPLORATOR": {"-": ["13"]},
    "EXPLORATVS": {"-": ["21"]},
    "EXPLORO": {"-": ["51"]},
    "EXPOLIO": {"2": ["54"]},
    "EXPOLITIO": {"-": ["13"]},
    "EXPOLITVS": {"-": ["21"]},
    "EXPONO": {"-": ["53"]},
    "EXPORRIGO": {"2": ["53"]},
    "EXPORTATIO": {"-": ["13"]},
    "EXPORTO": {"-": ["51"]},
    "EXPOSCO": {"-": ["53"]},
    "EXPOSITVS": {"-": ["21"]},
    "EXPOSTVLATIO": {"-": ["13"]},
    "EXPOSTVLO": {"-": ["51"]},
    "EXPRESSIO": {"-": ["13"]},
    "EXPRESSVS": {"1": ["14"], "2": ["21"]},
    "EXPRIMO": {"-": ["53"]},
    "EXPROBRATIO": {"-": ["13"]},
    "EXPROBRATOR": {"-": ["13"]},
    "EXPROBRATRIX": {"-": ["13"]},
    "EXPROBRO": {"-": ["51"]},
    "EXPROMO": {"-": ["53"]},
    "EXPVGNABILIS": {"-": ["24"]},
    "EXPVGNATIO": {"-": ["13"]},
    "EXPVGNATOR": {"-": ["13"]},
    "EXPVGNO": {"-": ["51"]},
    "EXPVLTRIX": {"-": ["13"]},
    "EXPVNGO": {"-": ["53"]},
    "EXPVRGO": {"-": ["51"]},
    "EXQVIRO": {"-": ["53"]},
    "EXQVISITE": {"-": ["6&"]},
    "EXQVISITVS": {"-": ["21"]},
    "EXSANGVINATVS": {"-": ["21"]},
    "EXSANGVIS": {"-": ["24"]},
    "EXSANIO": {"-": ["51"]},
    "EXSATIO": {"2": ["51"]},
    "EXSATVRABILIS": {"-": ["24"]},
    "EXSATVRO": {"-": ["51"]},
    "EXSCALPO": {"-": ["51"]},
    "EXSCINDO": {"-": ["53"]},
    "EXSCREO": {"-": ["51"]},
    "EXSCRIBO": {"-": ["53"]},
    "EXSCVLPO": {"-": ["53"]},
    "EXSECO": {"-": ["51"]},
    "EXSECRABILIS": {"-": ["24"]},
    "EXSECRANDVS": {"-": ["21"]},
    "EXSECRATIO": {"-": ["13"]},
    "EXSECRATVS": {"-": ["21"]},
    "EXSECROR": {"-": ["5J"]},
    "EXSECVTIO": {"-": ["13"]},
    "EXSEQVIAE": {"-": ["11"]},
    "EXSEQVOR": {"-": ["5L"]},
    "EXSERO": {"2": ["53"]},
    "EXSERTO": {"-": ["51"]},
    "EXSERTVS": {"-": ["21"]},
    "EXSIBILO": {"-": ["51"]},
    "EXSICCESCO": {"-": ["53"]},
    "EXSICCO": {"-": ["51"]},
    "EXSILIO": {"-": ["54"]},
    "EXSILIVM": {"-": ["12"]},
    "EXSISTO": {"-": ["53"]},
    "EXSOLESCO": {"-": ["53"]},
    "EXSOLVO": {"-": ["53"]},
    "EXSOLVTIO": {"-": ["13"]},
    "EXSOMNIS": {"-": ["24"]},
    "EXSORBEO": {"-": ["52"]},
    "EXSORS": {"-": ["25"]},
    "EXSPARGO": {"2": ["53"]},
    "EXSPATIOR": {"-": ["5J"]},
    "EXSPECTATIO": {"-": ["13"]},
    "EXSPECTATVM": {"-": ["12"]},
    "EXSPECTATVS": {"-": ["21"]},
    "EXSPECTO": {"-": ["51"]},
    "EXSPES": {"-": ["26"]},
    "EXSPIRO": {"-": ["51"]},
    "EXSPOLIO": {"-": ["51"]},
    "EXSPVO": {"-": ["53"]},
    "EXSTANS": {"-": ["25"]},
    "EXSTANTIA": {"-": ["11"]},
    "EXSTERNO": {"-": ["51"]},
    "EXSTIMVLATOR": {"-": ["13"]},
    "EXSTIMVLO": {"-": ["51"]},
    "EXSTINCTOR": {"-": ["13"]},
    "EXSTINGVO": {"-": ["53"]},
    "EXSTIRPO": {"-": ["51"]},
    "EXSTO": {"-": ["51"]},
    "EXSTRVCTIO": {"-": ["13"]},
    "EXSTRVCTVS": {"-": ["21"]},
    "EXSTRVO": {"-": ["53"]},
    "EXSVCATVS": {"-": ["21"]},
    "EXSVCTVS": {"-": ["21"]},
    "EXSVDO": {"-": ["51"]},
    "EXSVGO": {"-": ["53"]},
    "EXSVL": {"-": ["13"]},
    "EXSVLO": {"-": ["51"]},
    "EXSVLTATIO": {"-": ["13"]},
    "EXSVLTIM": {"-": ["60"]},
    "EXSVLTO": {"-": ["51"]},
    "EXSVPERABILIS": {"-": ["24"]},
    "EXSVPERATIO": {"-": ["13"]},
    "EXSVPERO": {"-": ["51"]},
    "EXSVRDO": {"-": ["51"]},
    "EXSVRGO": {"-": ["53"]},
    "EXTA": {"-": ["12"]},
    "EXTEMPLO": {"-": ["60"]},
    "EXTEMPORALIS": {"-": ["24"]},
    "EXTENDO": {"-": ["53"]},
    "EXTENTIO": {"-": ["13"]},
    "EXTENTO": {"-": ["51"]},
    "EXTENTVS": {"1": ["14"], "2": ["21"]},
    "EXTENVATIO": {"-": ["13"]},
    "EXTENVO": {"-": ["51"]},
    "EXTER": {"-": ["21"]},
    "EXTERGEO": {"-": ["52"]},
    "EXTERMINO": {"-": ["51"]},
    "EXTERNI": {"-": ["12"]},
    "EXTERNVM": {"-": ["12"]},
    "EXTERNVS": {"-": ["21"]},
    "EXTERO": {"-": ["53"]},
    "EXTERREO": {"-": ["52"]},
    "EXTERRITVS": {"-": ["21"]},
    "EXTIMESCO": {"-": ["53"]},
    "EXTOLLO": {"-": ["53"]},
    "EXTORQVEO": {"-": ["52"]},
    "EXTORRIS": {"-": ["24"]},
    "EXTRA": {"1": ["60"], "2": ["70"]},
    "EXTRAHO": {"-": ["53"]},
    "EXTRANEVS": {"-": ["21"]},
    "EXTRAORDINARIVS": {"-": ["21"]},
    "EXTRARIVS": {"-": ["21"]},
    "EXTREMA": {"-": ["12"]},
    "EXTREMITAS": {"-": ["13"]},
    "EXTREMVM": {"A": ["12"], "-": ["12"]},
    "EXTRICO": {"-": ["51"]},
    "EXTRINSECVS": {"-": ["60"]},
    "EXTRVDO": {"-": ["53"]},
    "EXTVNDO": {"-": ["53"]},
    "EXTVRBO": {"2": ["51"]},
    "EXVBERO": {"-": ["51"]},
    "EXVLCERATIO": {"-": ["13"]},
    "EXVLCERO": {"-": ["51"]},
    "EXVLVLO": {"-": ["51"]},
    "EXVNDATIO": {"-": ["13"]},
    "EXVNDO": {"-": ["51"]},
    "EXVO": {"-": ["53"]},
    "EXVRGEO": {"-": ["52"]},
    "EXVRO": {"-": ["53"]},
    "EXVVIAE": {"-": ["11"]},
    "FABA": {"-": ["11"], "N": ["11"]},
    "FABAGINVS": {"-": ["21"]},
    "FABALIA": {"-": ["13"]},
    "FABARIVS": {"-": ["21"]},
    "FABELLA": {"-": ["11"]},
    "FABER": {"1": ["12"], "2": ["21"]},
    "FABIANVS": {"A": ["21"], "N": ["12"]},
    "FABIVS": {"A": ["21"], "N": ["12"]},
    "FABRICA": {"-": ["11"]},
    "FABRICATIO": {"-": ["13"]},
    "FABRICATOR": {"-": ["13"]},
    "FABRICO": {"-": ["51"]},
    "FABRICOR": {"-": ["5J"]},
    "FABRILIS": {"-": ["24"]},
    "FABVLA": {"1": ["11"], "2": ["11"]},
    "FABVLATOR": {"-": ["13"]},
    "FABVLOR": {"-": ["5J"]},
    "FABVLOSVS": {"-": ["21"]},
    "FABVLVS": {"-": ["12"]},
    "FACESSO": {"-": ["53"]},
    "FACETE": {"-": ["60"]},
    "FACETIAE": {"-": ["11"]},
    "FACETVS": {"-": ["21"]},
    "FACIES": {"-": ["15"]},
    "FACILE": {"-": ["60"]},
    "FACILIS": {"-": ["24"]},
    "FACILITAS": {"-": ["13"]},
    "FACINOROSVS": {"-": ["21"]},
    "FACINVS": {"-": ["13"]},
    "FACIO": {"-": ["55"]},
    "FACTIO": {"-": ["13"]},
    "FACTIOSVS": {"-": ["21"]},
    "FACTITO": {"-": ["51"]},
    "FACTOR": {"-": ["13"]},
    "FACTVM": {"-": ["12"]},
    "FACTVS": {"1": ["14"], "2": ["21"]},
    "FACVLA": {"-": ["11"]},
    "FACVLTAS": {"-": ["13"]},
    "FACVNDE": {"-": ["60"]},
    "FACVNDIA": {"-": ["11"]},
    "FACVNDVS": {"-": ["21"]},
    "FAECARIVS": {"-": ["21"]},
    "FAECATVS": {"-": ["21"]},
    "FAECVLA": {"-": ["11"]},
    "FAENVS": {"-": ["13"]},
    "FAEX": {"-": ["13"]},
    "FAGINEVS": {"-": ["21"]},
    "FAGINVS": {"-": ["21"]},
    "FAGVS": {"-": ["16"]},
    "FALA": {"-": ["11"]},
    "FALARICA": {"-": ["11"]},
    "FALCARIVS": {"-": ["12"]},
    "FALCATVS": {"-": ["21"]},
    "FALCIFER": {"-": ["21"]},
    "FALCVLA": {"-": ["11"], "N": ["11"]},
    "FALLACIA": {"-": ["11"]},
    "FALLAX": {"-": ["25"]},
    "FALLO": {"-": ["53"]},
    "FALSARIVS": {"-": ["12"]},
    "FALSIPARENS": {"-": ["25"]},
    "FALSO": {"1": ["51"], "2": ["60"]},
    "FALSVM": {"-": ["12"]},
    "FALSVS": {"-": ["21"]},
    "FALX": {"-": ["13"]},
    "FAMA": {"-": ["11"], "N": ["11"]},
    "FAMELICVS": {"1": ["12"], "2": ["21"]},
    "FAMES": {"-": ["13"], "N": ["13"]},
    "FAMILIA": {"-": ["11"]},
    "FAMILIARICVS": {"-": ["21"]},
    "FAMILIARIS": {"1": ["13"], "2": ["24"]},
    "FAMILIARITAS": {"-": ["13"]},
    "FAMILIARITER": {"-": ["60"]},
    "FAMOSVS": {"-": ["21"]},
    "FAMVL": {"-": ["12"]},
    "FAMVLA": {"-": ["11"]},
    "FAMVLARIS": {"-": ["24"]},
    "FAMVLATVS": {"-": ["14"]},
    "FAMVLOR": {"-": ["5J"]},
    "FAMVLVS": {"1": ["12"], "2": ["21"]},
    "FANATICVS": {"-": ["21"]},
    "FANDVS": {"-": ["21"]},
    "FANNIVS": {"N": ["12"]},
    "FANOR": {"-": ["5J"]},
    "FANVM": {"-": ["12"], "N": ["12"]},
    "FAR": {"-": ["13"]},
    "FARCIO": {"-": ["54"]},
    "FARINA": {"-": ["11"]},
    "FARINARIVS": {"-": ["21"]},
    "FARNVS": {"-": ["12"]},
    "FARRAGO": {"-": ["13"]},
    "FARRARIVM": {"-": ["12"]},
    "FARRARIVS": {"-": ["21"]},
    "FARRATVS": {"-": ["21"]},
    "FARREO": {"-": ["51"]},
    "FARTOR": {"-": ["13"]},
    "FARTVRA": {"-": ["11"]},
    "FAS": {"-": ["16"], "N": ["16"]},
    "FASCIA": {"-": ["11"]},
    "FASCICVLVS": {"-": ["12"]},
    "FASCINA": {"-": ["11"]},
    "FASCINO": {"-": ["51"]},
    "FASCIOLA": {"-": ["11"]},
    "FASCIS": {"-": ["13"]},
    "FASTI": {"-": ["12"]},
    "FASTIDIO": {"-": ["54"]},
    "FASTIDIOSE": {"-": ["60"]},
    "FASTIDIOSVS": {"-": ["21"]},
    "FASTIDIVM": {"-": ["12"]},
    "FASTIGATE": {"-": ["60"]},
    "FASTIGATVS": {"-": ["21"]},
    "FASTIGIVM": {"-": ["12"]},
    "FASTIGO": {"-": ["51"]},
    "FASTVS": {"1": ["14"], "2": ["14"], "3": ["21"]},
    "FATALIS": {"-": ["24"]},
    "FATEOR": {"-": ["5K"]},
    "FATIDICVS": {"1": ["12"], "2": ["21"]},
    "FATIFER": {"-": ["21"]},
    "FATIGATIO": {"-": ["13"]},
    "FATIGO": {"-": ["51"]},
    "FATISCO": {"-": ["53"]},
    "FATISCOR": {"-": ["5L"]},
    "FATVA": {"-": ["11"]},
    "FATVM": {"-": ["12"], "N": ["12"]},
    "FATVVS": {"1": ["12"], "2": ["21"]},
    "FAVEO": {"-": ["52"]},
    "FAVILLA": {"-": ["11"]},
    "FAVNVS": {"-": ["12"], "N": ["12"]},
    "FAVONIVS": {"-": ["12"], "N": ["12"]},
    "FAVOR": {"-": ["13"]},
    "FAVORABILIS": {"-": ["24"]},
    "FAVSTVLVS": {"N": ["12"]},
    "FAVSTVS": {"-": ["21"], "N": ["12"]},
    "FAVTOR": {"-": ["13"]},
    "FAVVS": {"-": ["12"]},
    "FAVX": {"-": ["13"]},
    "FAX": {"-": ["13"]},
    "FEBRICITO": {"-": ["51"]},
    "FEBRICVLA": {"-": ["11"]},
    "FEBRICVLOSVS": {"-": ["21"]},
    "FEBRIS": {"-": ["13"], "N": ["13"]},
    "FEBRVARIVS": {"A": ["21"], "N": ["12"]},
    "FECTO": {"-": ["51"]},
    "FECVNDITAS": {"-": ["13"]},
    "FECVNDO": {"-": ["51"]},
    "FECVNDVS": {"-": ["21"]},
    "FEL": {"-": ["13"]},
    "FELES": {"-": ["13"]},
    "FELICITAS": {"-": ["13"], "N": ["13"]},
    "FELICITER": {"-": ["60"]},
    "FELIX": {"-": ["25"], "N": ["13"]},
    "FELLATOR": {"-": ["13"]},
    "FELLO": {"-": ["51"]},
    "FEMELLA": {"-": ["11"]},
    "FEMINA": {"-": ["11"]},
    "FEMINEVS": {"-": ["21"]},
    "FEMININVS": {"-": ["21"]},
    "FEMINO": {"-": ["51"]},
    "FEMVR": {"-": ["13"]},
    "FENARIVS": {"-": ["21"]},
    "FENDO": {"-": ["53"]},
    "FENEBRIS": {"-": ["24"]},
    "FENERATIO": {"-": ["13"]},
    "FENERATOR": {"-": ["13"]},
    "FENERO": {"-": ["51"]},
    "FENEROR": {"-": ["5J"]},
    "FENESTRA": {"-": ["11"]},
    "FENESTRO": {"-": ["51"]},
    "FENICVLVM": {"-": ["12"]},
    "FENILIA": {"-": ["13"]},
    "FENISECA": {"-": ["11"]},
    "FENVM": {"-": ["12"]},
    "FENVS": {"-": ["13"]},
    "FERA": {"-": ["11"]},
    "FERALIA": {"-": ["13"], "N": ["11"]},
    "FERALIS": {"-": ["24"]},
    "FERAX": {"-": ["25"]},
    "FERCIO": {"-": ["54"]},
    "FERCVLVM": {"-": ["12"]},
    "FERE": {"-": ["60"]},
    "FERENTARIVS": {"-": ["12"]},
    "FERENTINVS": {"N": ["21"], "A": ["21"]},
    "FERETRIVS": {"N": ["12"], "A": ["21"]},
    "FERETRVM": {"-": ["12"]},
    "FERIAE": {"-": ["11"]},
    "FERIATVS": {"-": ["21"]},
    "FERINA": {"-": ["11"]},
    "FERINVS": {"-": ["21"]},
    "FERIO": {"-": ["54"]},
    "FERITAS": {"-": ["13"]},
    "FERME": {"-": ["60"]},
    "FERMENTVM": {"-": ["12"]},
    "FERO": {"-": ["56"]},
    "FEROCIA": {"-": ["11"]},
    "FEROCITAS": {"-": ["13"]},
    "FEROCITER": {"-": ["60"]},
    "FEROCVLVS": {"-": ["21"]},
    "FERONIA": {"1N": ["11"], "2N": ["11"], "N": ["11"]},
    "FEROX": {"-": ["25"], "N": ["13"]},
    "FERRAMENTVM": {"-": ["12"]},
    "FERRARIA": {"1": ["11"], "2": ["11"]},
    "FERRARIVS": {"1": ["12"], "2": ["21"]},
    "FERRATVS": {"-": ["21"]},
    "FERREA": {"-": ["11"]},
    "FERREVS": {"-": ["21"]},
    "FERRVGINEVS": {"-": ["21"]},
    "FERRVGO": {"-": ["13"]},
    "FERRVM": {"-": ["12"]},
    "FERTILIS": {"-": ["24"]},
    "FERTILITAS": {"-": ["13"]},
    "FERTVM": {"-": ["12"]},
    "FERVEFACIO": {"-": ["55"]},
    "FERVENS": {"-": ["25"], "-": ["22"]},
    "FERVEO": {"-": ["52"]},
    "FERVESCO": {"-": ["53"]},
    "FERVIDVS": {"-": ["21"]},
    "FERVLA": {"-": ["11"]},
    "FERVO": {"-": ["53"]},
    "FERVOR": {"-": ["13"]},
    "FERVS": {"1": ["12"], "2": ["21"]},
    "FESCENNINI": {"-": ["12"]},
    "FESCENNINVS": {"1": ["12"], "2": ["21"], "A": ["21"]},
    "FESSVS": {"-": ["21"]},
    "FESTINANTER": {"-": ["6&"]},
    "FESTINATIO": {"-": ["13"]},
    "FESTINO": {"-": ["51"]},
    "FESTIVVS": {"-": ["21"]},
    "FESTVCA": {"-": ["11"]},
    "FESTVCVLA": {"2": ["11"], "-": ["11"]},
    "FESTVM": {"-": ["12"]},
    "FESTVS": {"-": ["21"], "N": ["12"]},
    "FETIALIS": {"1": ["13"], "2": ["24"]},
    "FETVRA": {"-": ["11"]},
    "FETVS": {"1": ["14"], "2": ["21"]},
    "FIBER": {"-": ["12"]},
    "FIBRA": {"-": ["11"]},
    "FIBVLA": {"-": ["11"]},
    "FIBVLATIO": {"-": ["13"]},
    "FICANA": {"N": ["11"]},
    "FICARIVS": {"-": ["21"]},
    "FICEDVLA": {"-": ["11"]},
    "FICIO": {"-": ["55"]},
    "FICTILE": {"-": ["13"]},
    "FICTILIS": {"-": ["24"]},
    "FICTRIX": {"-": ["13"]},
    "FICTVM": {"1": ["12"], "2": ["60"]},
    "FICTVS": {"-": ["21"]},
    "FICVLEA": {"N": ["11"]},
    "FICVLNEVS": {"-": ["21"]},
    "FICVLNVS": {"-": ["21"]},
    "FICVS": {"1": ["16"], "2": ["16"]},
    "FIDELIA": {"-": ["11"]},
    "FIDELIS": {"1": ["13"], "2": ["24"]},
    "FIDELITAS": {"-": ["13"]},
    "FIDELITER": {"-": ["60"]},
    "FIDENAE": {"N": ["11"]},
    "FIDENAS": {"N": ["25"]},
    "FIDENATES": {"N": ["13"]},
    "FIDENS": {"-": ["25"]},
    "FIDENTER": {"-": ["60"]},
    "FIDES": {
        "1": ["13"],
        "2": ["15"],
        "1N": ["13"],
        "2N": ["15"],
        "3N": ["13"],
        "N": ["13"],
    },
    "FIDICEN": {"-": ["13"], "N": ["13"]},
    "FIDICVLA": {"-": ["11"]},
    "FIDO": {"-": ["5T"]},
    "FIDVCIA": {"-": ["11"]},
    "FIDVCIARIVS": {"-": ["21"]},
    "FIDVS": {"-": ["21"], "N": ["12"]},
    "FIGLINVS": {"-": ["21"]},
    "FIGO": {"-": ["53"]},
    "FIGVLVS": {"-": ["12"], "N": ["12"]},
    "FIGVRA": {"-": ["11"]},
    "FIGVRATIO": {"-": ["13"]},
    "FIGVRO": {"-": ["51"]},
    "FILATIM": {"-": ["60"]},
    "FILIA": {"-": ["11"]},
    "FILICVLA": {"-": ["11"]},
    "FILIOLA": {"-": ["11"]},
    "FILIOLVS": {"-": ["12"]},
    "FILIVS": {"-": ["12"]},
    "FILIX": {"-": ["13"]},
    "FILVM": {"-": ["12"]},
    "FIMBRIA": {"N": ["11"], "-": ["11"]},
    "FIMVS": {"-": ["12"]},
    "FINDO": {"-": ["53"]},
    "FINE": {"-": ["70"]},
    "FINGO": {"-": ["53"]},
    "FINIO": {"-": ["54"]},
    "FINIS": {"-": ["13"]},
    "FINITIMI": {"-": ["12"]},
    "FINITIMVS": {"-": ["21"]},
    "FINITIO": {"-": ["13"]},
    "FINITOR": {"-": ["13"]},
    "FINITVS": {"-": ["21"]},
    "FIO": {"-": ["56"]},
    "FIRMAMEN": {"-": ["13"]},
    "FIRMAMENTVM": {"-": ["12"]},
    "FIRMATOR": {"-": ["13"]},
    "FIRME": {"-": ["60"]},
    "FIRMITAS": {"-": ["13"]},
    "FIRMITER": {"-": ["60"]},
    "FIRMITVDO": {"-": ["13"]},
    "FIRMO": {"-": ["51"]},
    "FIRMVS": {"-": ["21"], "N": ["12"]},
    "FISCELLA": {"-": ["11"]},
    "FISCINA": {"-": ["11"]},
    "FISCVS": {"-": ["12"]},
    "FISSILIS": {"-": ["24"]},
    "FISSVM": {"-": ["12"]},
    "FISSVRA": {"-": ["11"]},
    "FISTVCA": {"-": ["11"]},
    "FISTVCATIO": {"-": ["13"]},
    "FISTVCO": {"-": ["51"]},
    "FISTVLA": {"-": ["11"]},
    "FISTVLOSVS": {"-": ["21"]},
    "FITEOR": {"-": ["5K"]},
    "FITILLA": {"-": ["11"]},
    "FIXVS": {"-": ["21", "14"]},
    "FLABELLVM": {"-": ["12"]},
    "FLABRVM": {"-": ["12"]},
    "FLACCEO": {"-": ["52"]},
    "FLACCESCO": {"-": ["53"]},
    "FLACCIDVS": {"-": ["21"]},
    "FLACCVS": {"-": ["21"], "N": ["12"]},
    "FLAGELLO": {"-": ["51"]},
    "FLAGELLVM": {"-": ["12"]},
    "FLAGITATIO": {"-": ["13"]},
    "FLAGITATOR": {"-": ["13"]},
    "FLAGITIOSE": {"-": ["60"]},
    "FLAGITIOSVS": {"-": ["21"]},
    "FLAGITIVM": {"-": ["12"]},
    "FLAGITO": {"-": ["51"]},
    "FLAGRANS": {"-": ["25"]},
    "FLAGRANTER": {"-": ["6-"]},
    "FLAGRANTIA": {"-": ["11"]},
    "FLAGRO": {"-": ["51"]},
    "FLAGRVM": {"-": ["12"]},
    "FLAMEN": {"1": ["13"], "2": ["13"]},
    "FLAMINICA": {"-": ["11"]},
    "FLAMININVS": {"N": ["12"]},
    "FLAMINIVM": {"-": ["12"]},
    "FLAMINIVS": {"N": ["12"], "A": ["21"]},
    "FLAMMA": {"-": ["11"], "N": ["11"]},
    "FLAMMEOLVM": {"-": ["12"]},
    "FLAMMESCO": {"-": ["53"]},
    "FLAMMEVM": {"-": ["12"]},
    "FLAMMEVS": {"-": ["21"]},
    "FLAMMIFER": {"-": ["21"]},
    "FLAMMO": {"-": ["51"]},
    "FLATVRA": {"-": ["11"]},
    "FLATVS": {"-": ["14"]},
    "FLAVEO": {"-": ["52"]},
    "FLAVESCO": {"-": ["53"]},
    "FLAVIVS": {"N": ["12"], "A": ["21"]},
    "FLAVOLEIVS": {"N": ["12"]},
    "FLAVVS": {"-": ["21"], "N": ["12"]},
    "FLEBILE": {"-": ["60"]},
    "FLEBILIS": {"-": ["24"]},
    "FLEBILITER": {"-": ["60"]},
    "FLECTO": {"-": ["53"]},
    "FLEO": {"-": ["52"]},
    "FLETVS": {"-": ["14"]},
    "FLEXANIMVS": {"-": ["21"]},
    "FLEXIBILIS": {"-": ["24"]},
    "FLEXILIS": {"-": ["24"]},
    "FLEXO": {"-": ["51"]},
    "FLEXVOSVS": {"-": ["21"]},
    "FLEXVRA": {"-": ["11"]},
    "FLEXVS": {"-": ["14"]},
    "FLIGO": {"-": ["53"]},
    "FLO": {"-": ["51"]},
    "FLOCCVS": {"-": ["12"]},
    "FLORA": {"N": ["11"]},
    "FLORENS": {"-": ["25"]},
    "FLOREO": {"-": ["52"]},
    "FLORESCO": {"-": ["53"]},
    "FLOREVS": {"-": ["21"]},
    "FLORIDVLVS": {"-": ["21"]},
    "FLORIDVS": {"-": ["21"]},
    "FLORIFER": {"-": ["21"]},
    "FLOS": {"-": ["13"]},
    "FLOSCVLVS": {"-": ["12"]},
    "FLVCTIFRAGVS": {"-": ["21"]},
    "FLVCTISONVS": {"-": ["21"]},
    "FLVCTVATIO": {"-": ["13"]},
    "FLVCTVO": {"-": ["51"]},
    "FLVCTVOR": {"-": ["5J"]},
    "FLVCTVS": {"-": ["14"]},
    "FLVENS": {"-": ["25"]},
    "FLVENTER": {"-": ["60"]},
    "FLVENTISONVS": {"-": ["21"]},
    "FLVENTVM": {"-": ["12"]},
    "FLVIDVS": {"-": ["21"]},
    "FLVITO": {"-": ["51"]},
    "FLVMEN": {"-": ["13"]},
    "FLVMINEVS": {"-": ["21"]},
    "FLVO": {"-": ["53"]},
    "FLVTO": {"-": ["51"]},
    "FLVVIALIS": {"-": ["24"]},
    "FLVVIATICVS": {"-": ["21"]},
    "FLVVIATILIS": {"-": ["24"]},
    "FLVVIDVS": {"-": ["21"]},
    "FLVVIVS": {"-": ["12"]},
    "FLVXVS": {"1": ["14"], "2": ["21", "14"]},
    "FOCALE": {"-": ["13"]},
    "FOCILLO": {"-": ["51"]},
    "FOCO": {"-": ["51"]},
    "FOCVLVS": {"-": ["12"]},
    "FOCVS": {"-": ["12"]},
    "FODICO": {"-": ["51"]},
    "FODINA": {"-": ["11"]},
    "FODIO": {"-": ["55"]},
    "FOEDE": {"-": ["60"]},
    "FOEDERATVS": {"-": ["21"]},
    "FOEDITAS": {"-": ["13"]},
    "FOEDO": {"-": ["51"]},
    "FOEDVS": {"1": ["13"], "2": ["21"]},
    "FOETIDVS": {"-": ["21"]},
    "FOLIATVM": {"-": ["12"]},
    "FOLIATVRA": {"-": ["11"]},
    "FOLIVM": {"-": ["12"]},
    "FOLLICVLVS": {"-": ["12"]},
    "FOLLIS": {"-": ["13"]},
    "FOMENTVM": {"-": ["12"]},
    "FOMES": {"-": ["13"]},
    "FONS": {"-": ["13"]},
    "FONTANVS": {"-": ["21"]},
    "FONTICVLVS": {"-": ["12"]},
    "FOR": {"-": ["5J"]},
    "FORAMEN": {"-": ["13"]},
    "FORAS": {"-": ["60"]},
    "FORCEPS": {"-": ["13"]},
    "FORENSIS": {"-": ["24"]},
    "FORFEX": {"-": ["13"]},
    "FORICA": {"-": ["11"]},
    "FORIS": {"1": ["13"], "2": ["60"]},
    "FORMA": {"-": ["11"]},
    "FORMAMENTVM": {"-": ["12"]},
    "FORMATIO": {"-": ["13"]},
    "FORMATOR": {"-": ["13"]},
    "FORMATVRA": {"-": ["11"]},
    "FORMICA": {"-": ["11"]},
    "FORMIDABILIS": {"-": ["24"]},
    "FORMIDO": {"1": ["13"], "2": ["51"], "N": ["11"]},
    "FORMIDOLOSE": {"-": ["6&"]},
    "FORMIDOLOSVS": {"-": ["21"]},
    "FORMO": {"-": ["51"]},
    "FORMOSE": {"-": ["6&"]},
    "FORMOSVS": {"-": ["21"]},
    "FORMVLA": {"-": ["11"]},
    "FORNACVLA": {"-": ["11"]},
    "FORNAX": {"-": ["13"], "N": ["13"]},
    "FORNICATIO": {"-": ["13"]},
    "FORNIX": {"-": ["13"]},
    "FORO": {"-": ["51"]},
    "FORS": {"1": ["13"], "2": ["60"], "N": ["16"]},
    "FORSAN": {"-": ["60"]},
    "FORSIT": {"-": ["60"]},
    "FORSITAN": {"-": ["60"]},
    "FORTASSE": {"-": ["60"]},
    "FORTAX": {"-": ["13"]},
    "FORTE": {"-": ["60"]},
    "FORTIS": {"-": ["24"], "A": ["24"]},
    "FORTITER": {"-": ["60"]},
    "FORTITVDO": {"-": ["13"]},
    "FORTVITA": {"-": ["12"]},
    "FORTVITO": {"-": ["60"]},
    "FORTVITV": {"-": ["60"]},
    "FORTVITVS": {"-": ["21"]},
    "FORTVNA": {"-": ["11"], "N": ["11"]},
    "FORTVNO": {"-": ["51"]},
    "FORVLI": {"-": ["12"], "N": ["12"]},
    "FORVM": {"1": ["12"], "2": ["12"], "N": ["12"]},
    "FORVS": {"-": ["12"]},
    "FOSSA": {"-": ["11"]},
    "FOSSICIVS": {"-": ["21"]},
    "FOSSILIS": {"1": ["13"], "2": ["24"]},
    "FOSSIO": {"-": ["13"]},
    "FOSSOR": {"-": ["13"]},
    "FOSSVLA": {"-": ["11"]},
    "FOSSVRA": {"-": ["11"]},
    "FOVEA": {"-": ["11"]},
    "FOVEO": {"-": ["52"]},
    "FRACES": {"-": ["13"]},
    "FRACESCO": {"-": ["53"]},
    "FRACIDVS": {"-": ["21"]},
    "FRACTVRA": {"-": ["11"]},
    "FRACTVS": {"-": ["21"]},
    "FRAGA": {"-": ["12"]},
    "FRAGILIS": {"-": ["24"]},
    "FRAGILITAS": {"-": ["13"]},
    "FRAGMEN": {"-": ["13"]},
    "FRAGMENTVM": {"-": ["12"]},
    "FRAGOR": {"-": ["13"]},
    "FRAGOSVS": {"-": ["21"]},
    "FRAGRO": {"-": ["51"]},
    "FRAGVM": {"-": ["12"]},
    "FRAMEA": {"-": ["11"]},
    "FRANGO": {"-": ["53"]},
    "FRATER": {"-": ["13"]},
    "FRATERCVLVS": {"-": ["21", "12"]},
    "FRATERNITAS": {"-": ["13"]},
    "FRATERNVS": {"-": ["21"]},
    "FRAVDATIO": {"-": ["13"]},
    "FRAVDATOR": {"-": ["13"]},
    "FRAVDO": {"-": ["51"]},
    "FRAVDVLENTER": {"-": ["60"]},
    "FRAVDVLENTIA": {"-": ["11"]},
    "FRAVDVLENTVS": {"-": ["21"]},
    "FRAVS": {"-": ["13"]},
    "FRAXINEVS": {"-": ["21"]},
    "FRAXINVS": {"-": ["12"]},
    "FREMITVS": {"-": ["14"]},
    "FREMO": {"-": ["53"]},
    "FRENATOR": {"-": ["13"]},
    "FRENDO": {"-": ["53"]},
    "FRENO": {"-": ["51"]},
    "FRENVM": {"-": ["12"]},
    "FREQVENS": {"-": ["25"]},
    "FREQVENTATVS": {"-": ["21"]},
    "FREQVENTER": {"-": ["60"]},
    "FREQVENTIA": {"-": ["11"]},
    "FREQVENTO": {"-": ["51"]},
    "FRETVM": {"-": ["12"]},
    "FRETVS": {"1": ["14"], "2": ["21"]},
    "FRICATVRA": {"-": ["11"]},
    "FRICO": {"-": ["51"]},
    "FRIGEO": {"-": ["52"]},
    "FRIGERO": {"-": ["51"]},
    "FRIGESCO": {"-": ["53"]},
    "FRIGIDA": {"-": ["11"]},
    "FRIGIDARIVS": {"-": ["21"]},
    "FRIGIDVLVS": {"-": ["21"]},
    "FRIGIDVS": {"-": ["21"]},
    "FRIGO": {"1": ["53"]},
    "FRIGVS": {"-": ["13"]},
    "FRINGO": {"-": ["53"]},
    "FRIO": {"-": ["51"]},
    "FRITILLVS": {"-": ["12"]},
    "FRIVOLA": {"-": ["12"]},
    "FRIVOLVS": {"-": ["21"]},
    "FRONDATOR": {"-": ["13"]},
    "FRONDEO": {"-": ["52"]},
    "FRONDESCO": {"-": ["53"]},
    "FRONDEVS": {"-": ["21"]},
    "FRONDIFER": {"-": ["21"]},
    "FRONDOSVS": {"-": ["21"]},
    "FRONS": {"1": ["13"], "2": ["13"]},
    "FRONTATI": {"-": ["21"]},
    "FRVCTIFER": {"-": ["21"]},
    "FRVCTVOSVS": {"-": ["21"]},
    "FRVCTVS": {"-": ["14"]},
    "FRVGALIS": {"-": ["24"]},
    "FRVGALITAS": {"-": ["13"]},
    "FRVGALITER": {"-": ["60"]},
    "FRVGI": {"-": ["26"], "C": ["16"], "N": ["12"]},
    "FRVGIFER": {"-": ["21"]},
    "FRVGIFERENS": {"-": ["25"]},
    "FRVGIPARVS": {"-": ["21"]},
    "FRVMENTARIVS": {"1": ["12"], "2": ["21"]},
    "FRVMENTATIO": {"-": ["13"]},
    "FRVMENTOR": {"-": ["5J"]},
    "FRVMENTVM": {"-": ["12"]},
    "FRVOR": {"-": ["5L"]},
    "FRVSTRA": {"-": ["60"]},
    "FRVSTRAMEN": {"-": ["13"]},
    "FRVSTRATIO": {"-": ["13"]},
    "FRVSTROR": {"-": ["5J"]},
    "FRVSTVM": {"-": ["12"]},
    "FRVTEX": {"-": ["13"]},
    "FRVTICETVM": {"-": ["12"]},
    "FRVTICOR": {"-": ["5J"]},
    "FRVTICOSVS": {"-": ["21"]},
    "FRVX": {"-": ["13"]},
    "FVCATVS": {"-": ["21"]},
    "FVCO": {"-": ["51"]},
    "FVCVS": {"1": ["12"], "2": ["12"]},
    "FVFETIVS": {"N": ["12"]},
    "FVFICIVS": {"N": ["12"]},
    "FVGA": {"-": ["11"], "N": ["11"]},
    "FVGAX": {"-": ["25"]},
    "FVGELA": {"-": ["11"]},
    "FVGIO": {"-": ["55"]},
    "FVGITIVVS": {"1": ["12"], "2": ["21"]},
    "FVGITO": {"-": ["51"]},
    "FVGO": {"-": ["51"]},
    "FVLCIMENTVM": {"-": ["12"]},
    "FVLCIO": {"-": ["54"]},
    "FVLCRVM": {"-": ["12"]},
    "FVLGENS": {"-": ["25"]},
    "FVLGEO": {"-": ["52"]},
    "FVLGIDVS": {"-": ["21"]},
    "FVLGO": {"-": ["53"]},
    "FVLGOR": {"-": ["13"]},
    "FVLGVR": {"-": ["13"]},
    "FVLGVRATOR": {"-": ["13"]},
    "FVLGVRIO": {"-": ["54"]},
    "FVLGVRO": {"-": ["51"]},
    "FVLICA": {"-": ["11"]},
    "FVLIGO": {"-": ["13"]},
    "FVLLO": {"-": ["13"]},
    "FVLLONICA": {"-": ["11"]},
    "FVLLONICVS": {"-": ["21"]},
    "FVLLONIVS": {"-": ["21"]},
    "FVLMEN": {"-": ["13"]},
    "FVLMENTA": {"-": ["11"]},
    "FVLMENTVM": {"-": ["12"]},
    "FVLMINEVS": {"-": ["21"]},
    "FVLMINO": {"-": ["51"]},
    "FVLTVRA": {"-": ["11"]},
    "FVLVIVS": {"N": ["12"]},
    "FVLVVS": {"-": ["21"], "N": ["12"]},
    "FVMEVS": {"-": ["21"]},
    "FVMIDVS": {"-": ["21"]},
    "FVMO": {"-": ["51"]},
    "FVMOSVS": {"-": ["21"]},
    "FVMVS": {"-": ["12"]},
    "FVNALE": {"-": ["13"]},
    "FVNALIS": {"-": ["24"]},
    "FVNCTIO": {"-": ["13"]},
    "FVNCTVS": {"-": ["12"]},
    "FVNDA": {"-": ["11"]},
    "FVNDAMEN": {"-": ["13"]},
    "FVNDAMENTVM": {"-": ["12"]},
    "FVNDATIO": {"-": ["13"]},
    "FVNDATVS": {"-": ["21"]},
    "FVNDITOR": {"-": ["13"]},
    "FVNDITVS": {"-": ["60"]},
    "FVNDO": {"1": ["51"], "2": ["53"]},
    "FVNDVLVS": {"-": ["12"]},
    "FVNDVS": {"-": ["12"]},
    "FVNEBRIS": {"-": ["24"]},
    "FVNEREVS": {"-": ["21"]},
    "FVNERO": {"-": ["51"]},
    "FVNESTO": {"-": ["51"]},
    "FVNESTVS": {"-": ["21"]},
    "FVNGOR": {"-": ["5L"]},
    "FVNGVS": {"-": ["12"]},
    "FVNICVLVS": {"-": ["12"]},
    "FVNIS": {"-": ["13"]},
    "FVNVS": {"-": ["13"], "N": ["12"]},
    "FVR": {"-": ["13"]},
    "FVRCA": {"-": ["11"]},
    "FVRCIFER": {"-": ["12"]},
    "FVRCILLA": {"-": ["11"]},
    "FVRENS": {"-": ["25"]},
    "FVRIA": {"-": ["11"], "N": ["11"]},
    "FVRIALIS": {"-": ["24"]},
    "FVRIBVNDVS": {"-": ["21"]},
    "FVRIO": {"1": ["51"], "2": ["54"]},
    "FVRIOSE": {"-": ["60"]},
    "FVRIOSVS": {"-": ["21"]},
    "FVRIVS": {"A": ["21"], "N": ["12"]},
    "FVRNVS": {"-": ["12"]},
    "FVRO": {"-": ["53"]},
    "FVROR": {"1": ["13"], "2": ["5J", "13"], "N": ["13"]},
    "FVRTIM": {"-": ["60"]},
    "FVRTIVE": {"-": ["60"]},
    "FVRTIVVS": {"-": ["21"]},
    "FVRTVM": {"-": ["12"]},
    "FVRVVS": {"-": ["21"]},
    "FVSCINA": {"-": ["11"]},
    "FVSCO": {"-": ["51"]},
    "FVSCVS": {"-": ["21"], "N": ["12"]},
    "FVSILIS": {"-": ["24"]},
    "FVSIO": {"-": ["13"]},
    "FVSIVS": {"A": ["21"], "N": ["12"]},
    "FVSTERNA": {"-": ["11"]},
    "FVSTIS": {"-": ["13"]},
    "FVSVS": {"1": ["12"], "2": ["14"], "3": ["21"]},
    "FVTILE": {"1": ["13"], "2": ["60"]},
    "FVTILIS": {"-": ["24"]},
    "FVTO": {"-": ["51"]},
    "FVTVO": {"-": ["53"]},
    "FVTVRI": {"-": ["12"]},
    "FVTVRVM": {"-": ["12"]},
    "FVTVRVS": {"-": ["21"]},
    "FVTVTIO": {"-": ["13"]},
    "GABII": {"N": ["12"]},
    "GABINI": {"N": ["12"]},
    "GABINVS": {"N": ["21"], "A": ["21"]},
    "GADES": {"N": ["13"]},
    "GAESVM": {"-": ["12"]},
    "GAGATES": {"-": ["13"]},
    "GAIVS": {"N": ["12"]},
    "GALBA": {"N": ["11"]},
    "GALBANEVS": {"-": ["21"]},
    "GALBEVS": {"-": ["12"]},
    "GALBINVM": {"-": ["12"]},
    "GALEA": {"-": ["11"]},
    "GALEATVS": {"-": ["21"]},
    "GALEO": {"-": ["51"]},
    "GALERVS": {"-": ["12"]},
    "GALLA": {"-": ["11"], "N": ["11"]},
    "GALLI": {"N": ["12"]},
    "GALLIA": {"N": ["11"]},
    "GALLICVS": {"N": ["21"], "A": ["21"]},
    "GALLINA": {"-": ["11"], "N": ["11"]},
    "GALLINACEVS": {"-": ["21"]},
    "GALLVS": {"-": ["12"], "A": ["21"], "N": ["12"]},
    "GANEA": {"-": ["11"]},
    "GANEO": {"-": ["13"]},
    "GANEVM": {"-": ["12"]},
    "GANGABA": {"-": ["11"]},
    "GANNEO": {"-": ["13"]},
    "GANNIO": {"-": ["54"]},
    "GANNITVS": {"-": ["14"]},
    "GANOSIS": {"-": ["17"]},
    "GANYMEDES": {"N": ["17"]},
    "GARRIO": {"-": ["54"]},
    "GARRVLITAS": {"-": ["13"]},
    "GARRVLVS": {"-": ["21"]},
    "GARVM": {"-": ["12"]},
    "GAVDEO": {"-": ["5S"]},
    "GAVDIVM": {"-": ["12"], "N": ["12"]},
    "GAVIVS": {"N": ["12"]},
    "GAVSAPATVS": {"-": ["21"]},
    "GAVSAPVM": {"-": ["16"]},
    "GAZA": {"-": ["11"], "N": ["11"]},
    "GEGANIVS": {"N": ["12"]},
    "GELENSES": {"N": ["13"]},
    "GELICIDIVM": {"-": ["12"]},
    "GELIDVS": {"-": ["21"]},
    "GELLIVS": {"N": ["12"], "A": ["21"]},
    "GELO": {"-": ["51"]},
    "GELV": {"-": ["14"]},
    "GELVM": {"-": ["12"]},
    "GELVS": {"-": ["14"]},
    "GEMELLVS": {"1": ["12"], "2": ["21"]},
    "GEMINI": {"-": ["12"], "N": ["12"]},
    "GEMINO": {"-": ["51"]},
    "GEMINVS": {"-": ["21"], "N": ["12"], "A": ["21"]},
    "GEMISCO": {"-": ["53"]},
    "GEMITVS": {"-": ["14"]},
    "GEMMA": {"-": ["11"]},
    "GEMMATVS": {"-": ["21"]},
    "GEMMEVS": {"-": ["21"]},
    "GEMMIFER": {"-": ["21"]},
    "GEMMO": {"-": ["51"]},
    "GEMO": {"-": ["53"]},
    "GENA": {"-": ["11"]},
    "GENEALOGVS": {"-": ["12"]},
    "GENER": {"-": ["12"]},
    "GENERALIS": {"-": ["24"]},
    "GENERALITER": {"-": ["60"]},
    "GENERASCO": {"-": ["53"]},
    "GENERATIM": {"-": ["60"]},
    "GENERATOR": {"-": ["13"]},
    "GENERO": {"-": ["51"]},
    "GENEROSE": {"-": ["60"]},
    "GENEROSVS": {"-": ["21"]},
    "GENESIS": {"-": ["13"]},
    "GENETHLIOLOGIA": {"-": ["11"]},
    "GENETRIX": {"-": ["13"], "N": ["13"]},
    "GENIALIS": {"-": ["24"]},
    "GENIALITER": {"-": ["60"]},
    "GENICVLVS": {"-": ["12"]},
    "GENISTA": {"-": ["11"]},
    "GENITABILIS": {"-": ["24"]},
    "GENITALIA": {"-": ["12"]},
    "GENITALIS": {"-": ["24"], "A": ["24"]},
    "GENITALITER": {"-": ["60"]},
    "GENITOR": {"-": ["13"], "N": ["13"]},
    "GENIVS": {"-": ["12"], "N": ["12"]},
    "GENO": {"-": ["53"]},
    "GENS": {"-": ["13"]},
    "GENTICVS": {"-": ["21"]},
    "GENTILICIVS": {"-": ["21"]},
    "GENTILIS": {"1": ["13"], "2": ["24"]},
    "GENTILITAS": {"-": ["13"]},
    "GENV": {"-": ["14"]},
    "GENVALIA": {"-": ["13"]},
    "GENVCIVS": {"N": ["12"]},
    "GENVINVS": {"-": ["21"]},
    "GENVS": {"1": ["13"], "2": ["14"]},
    "GEOMETRES": {"-": ["17"]},
    "GEOMETRIA": {"-": ["11"]},
    "GEOMETRICA": {"-": ["12"]},
    "GEOMETRICE": {"1": ["17"], "2": ["60"]},
    "GEOMETRICVS": {"1": ["12"], "2": ["21"]},
    "GERMANA": {"-": ["11"]},
    "GERMANIA": {"N": ["11"]},
    "GERMANICVS": {"N": ["21"], "A": ["21"]},
    "GERMANVS": {"1": ["12"], "2": ["21"], "N": ["21"], "A": ["21"]},
    "GERMEN": {"-": ["13"]},
    "GERMINO": {"-": ["51"]},
    "GERO": {"-": ["53"]},
    "GERVLVS": {"1": ["12"], "2": ["21"]},
    "GERVSIA": {"-": ["11"], "N": ["11"]},
    "GESTA": {"-": ["12"]},
    "GESTAMEN": {"-": ["13"]},
    "GESTATIO": {"-": ["13"]},
    "GESTIO": {"1": ["13"], "2": ["54"]},
    "GESTO": {"-": ["51"]},
    "GESTVS": {"-": ["14"]},
    "GETA": {"C": ["11"], "N": ["11"]},
    "GETAE": {"-": ["11"], "N": ["11"]},
    "GETICVS": {"N": ["21"], "A": ["21"]},
    "GIBBER": {"1": ["13"], "2": ["21"]},
    "GIBBVS": {"-": ["12"]},
    "GIGAS": {"-": ["17"], "N": ["17"]},
    "GIGNO": {"-": ["53"]},
    "GILVVS": {"-": ["21"]},
    "GINGIVA": {"-": ["11"]},
    "GLABER": {"1": ["12"], "2": ["21"]},
    "GLABRIO": {"N": ["13"]},
    "GLACIALIS": {"-": ["24"]},
    "GLACIES": {"-": ["15"]},
    "GLACIO": {"-": ["51"]},
    "GLADIATOR": {"-": ["13"]},
    "GLADIATORIVS": {"-": ["21"]},
    "GLADIATVRA": {"-": ["11"]},
    "GLADIOLVS": {"-": ["12"]},
    "GLADIVS": {"-": ["12"]},
    "GLANDARIVS": {"-": ["21"]},
    "GLANDIFER": {"-": ["21"]},
    "GLANS": {"-": ["13"]},
    "GLAREA": {"-": ["11"]},
    "GLAREOSVS": {"-": ["21"]},
    "GLAVCIA": {"N": ["11"]},
    "GLAVCVS": {"1": ["12"], "2": ["21"], "N": ["12"]},
    "GLEBA": {"-": ["11"]},
    "GLEBVLA": {"-": ["11"]},
    "GLESVM": {"-": ["12"]},
    "GLISCO": {"-": ["53"]},
    "GLOBO": {"-": ["51"]},
    "GLOBOSVS": {"-": ["21"]},
    "GLOBVS": {"-": ["12"]},
    "GLOMERAMEN": {"-": ["13"]},
    "GLOMERO": {"-": ["51"]},
    "GLOMVS": {"-": ["13"]},
    "GLORIA": {"-": ["11"], "N": ["11"]},
    "GLORIOR": {"-": ["5J"]},
    "GLORIOSE": {"-": ["60"]},
    "GLORIOSVS": {"-": ["21"]},
    "GLOSSA": {"-": ["11"]},
    "GLVBO": {"-": ["53"]},
    "GLVTEN": {"-": ["13"]},
    "GLVTINATOR": {"-": ["13"]},
    "GLVTINVM": {"-": ["12"]},
    "GLVTIO": {"-": ["54"]},
    "GLVTVS": {"1": ["14"], "2": ["21"]},
    "GNAEVS": {"N": ["12"]},
    "GNARVS": {"-": ["21"]},
    "GNOMON": {"-": ["17"]},
    "GNOMONICE": {"-": ["17"]},
    "GNOMONICVS": {"-": ["21"]},
    "GOBIO": {"-": ["13"]},
    "GORGO": {"N": ["17"]},
    "GRABATVS": {"-": ["12"]},
    "GRACCHVS": {"N": ["12"]},
    "GRACILIS": {"-": ["24"], "N": ["13"]},
    "GRACILITAS": {"-": ["13"]},
    "GRACVLVS": {"-": ["12"]},
    "GRADARIVS": {"-": ["21"]},
    "GRADATIM": {"-": ["60"]},
    "GRADATIO": {"-": ["13"]},
    "GRADIOR": {"-": ["5N"]},
    "GRADIVVS": {"N": ["12"]},
    "GRADVS": {"-": ["14"]},
    "GRAECE": {"-": ["60"]},
    "GRAECI": {"N": ["12"]},
    "GRAECIA": {"N": ["11"]},
    "GRAECOR": {"-": ["5J"]},
    "GRAECVLVS": {"N": ["21"], "A": ["21"]},
    "GRAECVS": {"N": ["21"], "A": ["21"]},
    "GRAII": {"N": ["12"]},
    "GRAIVS": {"N": ["21"], "A": ["21"]},
    "GRAMEN": {"-": ["13"]},
    "GRAMINEVS": {"-": ["21"]},
    "GRAMMATICA": {"1": ["12"], "2": ["17"]},
    "GRAMMATICVS": {"1": ["12"], "2": ["21"]},
    "GRAMMICVS": {"-": ["21"]},
    "GRANARIVM": {"-": ["12"]},
    "GRANATVS": {"-": ["14"]},
    "GRANDAEVVS": {"-": ["21"]},
    "GRANDE": {"-": ["60"]},
    "GRANDESCO": {"-": ["53"]},
    "GRANDIO": {"-": ["54"]},
    "GRANDIS": {"-": ["24"]},
    "GRANDISCAPIVS": {"-": ["21"]},
    "GRANDITER": {"-": ["6&"]},
    "GRANDO": {"-": ["13"]},
    "GRANEA": {"-": ["11"]},
    "GRANEVS": {"-": ["21"]},
    "GRANIFER": {"-": ["21"]},
    "GRANIVS": {"N": ["12"]},
    "GRANVM": {"-": ["12"]},
    "GRAPHICVS": {"-": ["2G", "21"]},
    "GRAPHIS": {"-": ["17"]},
    "GRAPHIVM": {"-": ["12"]},
    "GRASSATOR": {"-": ["13"]},
    "GRASSOR": {"-": ["5J"]},
    "GRATE": {"-": ["60"]},
    "GRATES": {"-": ["13"]},
    "GRATIA": {"-": ["11"], "N": ["11"]},
    "GRATIFICOR": {"-": ["5J"]},
    "GRATIOSVS": {"-": ["21"]},
    "GRATIS": {"-": ["60"]},
    "GRATOR": {"-": ["5J"]},
    "GRATVITO": {"-": ["60"]},
    "GRATVITVS": {"-": ["21"]},
    "GRATVLATIO": {"-": ["13"]},
    "GRATVLOR": {"-": ["5J"]},
    "GRATVS": {"-": ["21"], "N": ["12"]},
    "GRAVATIM": {"-": ["60"]},
    "GRAVE": {"-": ["60"]},
    "GRAVEDO": {"-": ["13"]},
    "GRAVEOLENS": {"-": ["25"]},
    "GRAVESCO": {"-": ["53"]},
    "GRAVIDVS": {"-": ["21"]},
    "GRAVIS": {"-": ["24"]},
    "GRAVITAS": {"-": ["13"]},
    "GRAVITER": {"-": ["60"]},
    "GRAVITVDO": {"-": ["13"]},
    "GRAVO": {"-": ["51"]},
    "GRAVOR": {"-": ["5J"]},
    "GREDIOR": {"-": ["5N"]},
    "GREGALES": {"-": ["13"]},
    "GREGALIS": {"-": ["24"]},
    "GREGARIVS": {"1": ["12"], "2": ["21"]},
    "GREGATIM": {"-": ["60"]},
    "GREGO": {"-": ["51"]},
    "GREMIVM": {"-": ["12"]},
    "GRESSVS": {"-": ["14"]},
    "GREX": {"-": ["13"]},
    "GROSPHVS": {"N": ["12"]},
    "GROSSVS": {"-": ["12"]},
    "GRVMVLVS": {"-": ["12"]},
    "GRVMVS": {"-": ["12"]},
    "GRVNNIO": {"-": ["54"]},
    "GRVO": {"-": ["53"]},
    "GRVS": {"-": ["13"]},
    "GRYPS": {"-": ["17"]},
    "GVBERNACVLVM": {"-": ["12"]},
    "GVBERNATIO": {"-": ["13"]},
    "GVBERNATOR": {"-": ["13"]},
    "GVBERNO": {"-": ["51"]},
    "GVLA": {"-": ["11"]},
    "GVLOSVS": {"-": ["21"]},
    "GVMMI": {"-": ["16"]},
    "GVRGES": {"-": ["13"], "N": ["13"]},
    "GVRGVLIO": {"-": ["13"]},
    "GVSTO": {"-": ["51"]},
    "GVSTVS": {"-": ["14"]},
    "GVTTA": {"-": ["11"], "N": ["11"]},
    "GVTTVLA": {"-": ["11"]},
    "GVTTVR": {"-": ["13"]},
    "GVTTVS": {"-": ["12"]},
    "GYMNASIARCHVS": {"-": ["12"]},
    "GYMNASIVM": {"-": ["12"]},
    "GYMNICVS": {"-": ["21"]},
    "GYNAECONITIS": {"-": ["17"]},
    "GYPSO": {"-": ["51"]},
    "GYPSVM": {"-": ["12"]},
    "GYRVS": {"-": ["12"]},
    "HABENA": {"-": ["11"]},
    "HABEO": {"-": ["52"]},
    "HABILIS": {"-": ["24"]},
    "HABILITAS": {"-": ["13"]},
    "HABITABILIS": {"-": ["24"]},
    "HABITANS": {"-": ["13"]},
    "HABITATIO": {"-": ["13"]},
    "HABITATOR": {"-": ["13"]},
    "HABITO": {"-": ["51"]},
    "HABITVS": {"1": ["14"], "2": ["21"], "N": ["12"]},
    "HABONIVS": {"N": ["12"]},
    "HAC": {"-": ["60"]},
    "HACTENVS": {"-": ["60"]},
    "HADRIANVS": {"N": ["12"], "A": ["21"]},
    "HADRIATICVS": {"N": ["21"], "A": ["21"]},
    "HADRVMENTVM": {"-": ["12"], "N": ["12"]},
    "HAEDILIA": {"-": ["11"]},
    "HAEDINVS": {"-": ["21"]},
    "HAEDVLVS": {"-": ["12"]},
    "HAEDVS": {"-": ["12"], "N": ["12"]},
    "HAEREO": {"-": ["52"]},
    "HAERESCO": {"-": ["53"]},
    "HAERESIS": {"-": ["17"]},
    "HAESITATIO": {"-": ["13"]},
    "HAESITO": {"-": ["51"]},
    "HALAESA": {"N": ["11"]},
    "HALAESINI": {"N": ["12"]},
    "HALAESINVS": {"N": ["21"], "A": ["21"]},
    "HALICARNASSVS": {"N": ["17"]},
    "HALICYENSES": {"N": ["13"]},
    "HALICYENSIS": {"N": ["24"], "A": ["24"]},
    "HALITVS": {"-": ["14"]},
    "HALO": {"-": ["51"]},
    "HAMA": {"-": ["11"]},
    "HAMATVS": {"-": ["21"]},
    "HAMAXOPODES": {"-": ["17"]},
    "HAMILCAR": {"N": ["13"]},
    "HAMVS": {"-": ["12"]},
    "HANNIBAL": {"N": ["13"]},
    "HAPHE": {"-": ["17"]},
    "HARA": {"-": ["11"]},
    "HARIOLVS": {"-": ["12"]},
    "HARMONIA": {"-": ["11"], "N": ["11"]},
    "HARMONICE": {"-": ["17"]},
    "HARPAGO": {"1": ["13"], "2": ["51"]},
    "HARVSPEX": {"-": ["13"]},
    "HARVSPICINA": {"-": ["11"]},
    "HARVSPICIVM": {"-": ["12"]},
    "HASDRVBAL": {"N": ["13"]},
    "HASTA": {"-": ["11"], "N": ["11"]},
    "HASTARIVS": {"1": ["12"], "2": ["21"]},
    "HASTATVS": {"1": ["12"], "2": ["21"]},
    "HASTILE": {"-": ["13"]},
    "HAVD": {"-": ["68"]},
    "HAVDDVM": {"-": ["68"]},
    "HAVDQVAQVAM": {"-": ["68"]},
    "HAVRIO": {"-": ["54"]},
    "HAVSTVS": {"-": ["14"]},
    "HEBEO": {"-": ["52"]},
    "HEBES": {"-": ["25"]},
    "HEBESCO": {"-": ["53"]},
    "HEBETATIO": {"-": ["13"]},
    "HEBETO": {"-": ["51"]},
    "HEBRVS": {"1N": ["12"], "2N": ["12"], "N": ["12"]},
    "HECATOMBE": {"-": ["17"]},
    "HEDERA": {"-": ["11"]},
    "HEDERACEVS": {"-": ["21"]},
    "HEDERIGER": {"-": ["21"]},
    "HEI": {"-": ["90"]},
    "HEIVS": {"N": ["12"]},
    "HELENA": {"N": ["11"]},
    "HELEPOLIS": {"-": ["17"]},
    "HELICON": {"-": ["17"], "N": ["17"]},
    "HELIX": {"-": ["13"]},
    "HELLEBORVM": {"-": ["12"]},
    "HELLEBORVS": {"-": ["12"]},
    "HELLESPONTVS": {"N": ["12"]},
    "HELO": {"-": ["51"]},
    "HELVOLVS": {"-": ["21"]},
    "HELVOR": {"-": ["5J"]},
    "HEM": {"-": ["90"]},
    "HEMICYCLIVM": {"-": ["12"]},
    "HEMINA": {"-": ["11"]},
    "HEMIOLIOS": {"-": ["27"]},
    "HEMISPHAERIVM": {"-": ["12"]},
    "HEMITONIVM": {"-": ["17"]},
    "HEMITRIGLYPHVS": {"-": ["12"]},
    "HENDECASYLLABVS": {"-": ["12"]},
    "HENNA": {"N": ["11"]},
    "HENNENSES": {"N": ["13"]},
    "HENNENSIS": {"N": ["24"], "A": ["24"]},
    "HERA": {"-": ["11"]},
    "HERACLEA": {"N": ["11"]},
    "HERACLEO": {"N": ["17"]},
    "HERACLEVS": {"N": ["12"], "A": ["21"]},
    "HERACLIENSES": {"N": ["13"]},
    "HERACLIENSIS": {"N": ["24"], "A": ["24"]},
    "HERACLIVS": {"N": ["12"]},
    "HERBA": {"-": ["11"]},
    "HERBIDVS": {"-": ["21"]},
    "HERBITA": {"N": ["11"]},
    "HERBITENSES": {"N": ["13"]},
    "HERBITENSIS": {"N": ["24"], "A": ["24"]},
    "HERBOSVS": {"-": ["21"]},
    "HERBVLA": {"-": ["11"]},
    "HERCISCO": {"-": ["53"]},
    "HERCVLES": {"-": ["90"], "N": ["16"]},
    "HEREDITARIVS": {"-": ["21"]},
    "HEREDITAS": {"-": ["13"]},
    "HEREDITO": {"-": ["51"]},
    "HERENNIVS": {"N": ["12"]},
    "HERES": {"-": ["13"]},
    "HERI": {"-": ["60"]},
    "HERIFVGA": {"-": ["21"]},
    "HERILIS": {"-": ["24"]},
    "HERMINIVS": {"N": ["12"]},
    "HERNICI": {"N": ["12"]},
    "HERNICVS": {"N": ["21"], "A": ["21"]},
    "HERODOTVS": {"N": ["12"]},
    "HEROICVS": {"-": ["21"]},
    "HEROINE": {"-": ["17"]},
    "HEROIS": {"-": ["17"]},
    "HEROS": {"-": ["17"]},
    "HEROVM": {"-": ["12"]},
    "HEROVS": {"-": ["21"]},
    "HERSILIA": {"N": ["11"]},
    "HERVS": {"-": ["12"]},
    "HESIODVS": {"N": ["12"]},
    "HESPERIA": {"N": ["11"]},
    "HESPERIVS": {"N": ["21"], "A": ["21"]},
    "HESTERNO": {"-": ["60"]},
    "HESTERNVS": {"-": ["21"]},
    "HEV": {"-": ["90"]},
    "HEVS": {"-": ["90"]},
    "HEXACHORDOS": {"-": ["27"]},
    "HEXAPHOROS": {"-": ["27"]},
    "HEXASTYLOS": {"-": ["27"]},
    "HIASCO": {"-": ["53"]},
    "HIATVS": {"-": ["14"]},
    "HIBEO": {"-": ["52"]},
    "HIBERNA": {"-": ["12"]},
    "HIBERNACVLVM": {"-": ["12"]},
    "HIBERNO": {"-": ["51"]},
    "HIBERNVM": {"-": ["12"]},
    "HIBERNVS": {"-": ["21"]},
    "HIBISCVM": {"-": ["12"]},
    "HIC": {"1": ["45"], "2": ["60"]},
    "HICCINE": {"-": ["47"]},
    "HIEMALIS": {"-": ["24"]},
    "HIEMO": {"-": ["51"]},
    "HIEMS": {"-": ["13"], "N": ["13"]},
    "HIERA": {"1": ["11"], "2": ["17"]},
    "HIERO": {"N": ["17"]},
    "HIERONICVS": {"N": ["21"], "A": ["21"]},
    "HILARE": {"-": ["60"]},
    "HILARICVLVS": {"-": ["21"]},
    "HILARIS": {"-": ["24"]},
    "HILARITAS": {"-": ["13"]},
    "HILARO": {"-": ["51"]},
    "HILARVS": {"-": ["21"], "N": ["12"]},
    "HILLA": {"-": ["11"]},
    "HILVM": {"-": ["12"]},
    "HIMERA": {"N": ["11"]},
    "HINC": {"-": ["60"]},
    "HINNIO": {"-": ["54"]},
    "HINNITVS": {"-": ["14"]},
    "HINNVLEVS": {"-": ["12"]},
    "HIO": {"-": ["51"]},
    "HIPPOCAMELVS": {"-": ["12"]},
    "HIPPOMANES": {"-": ["13"]},
    "HIPPOPERAE": {"-": ["11"]},
    "HIPPOPOTAMVS": {"-": ["12"]},
    "HIPPOTOTAE": {"-": ["11"]},
    "HIPPOTOXOTAE": {"-": ["11"]},
    "HIRCINVS": {"-": ["21"]},
    "HIRCOSVS": {"-": ["21"]},
    "HIRCVS": {"-": ["12"]},
    "HIRNEA": {"-": ["11"]},
    "HIRSVTVS": {"-": ["21"]},
    "HIRTVS": {"-": ["21"]},
    "HIRVNDO": {"-": ["13"]},
    "HISCO": {"-": ["53"]},
    "HISPANI": {"N": ["12"]},
    "HISPANIA": {"N": ["11"]},
    "HISPANVS": {"N": ["21"], "A": ["21"]},
    "HISPIDVS": {"-": ["21"]},
    "HISTORIA": {"-": ["11"], "N": ["11"]},
    "HISTORICVS": {"1": ["12"], "2": ["21"]},
    "HISTORVMENOS": {"-": ["57"]},
    "HISTRIO": {"-": ["13"]},
    "HISTRIONALIS": {"-": ["24"]},
    "HIVLCO": {"-": ["51"]},
    "HIVLCVS": {"-": ["21"]},
    "HODIE": {"-": ["60"]},
    "HODIERNVS": {"-": ["21"]},
    "HOMERVS": {"N": ["12"]},
    "HOMICIDA": {"-": ["11"]},
    "HOMICIDIVM": {"-": ["12"]},
    "HOMO": {"-": ["13"]},
    "HOMOEOMERIA": {"-": ["17"]},
    "HOMOTONVS": {"-": ["21"]},
    "HOMVLLVS": {"-": ["12"], "N": ["12"]},
    "HOMVNCIO": {"-": ["13"]},
    "HOMVNCVLVS": {"-": ["12"]},
    "HONESTAMENTVM": {"-": ["12"]},
    "HONESTAS": {"-": ["13"]},
    "HONESTE": {"-": ["60"]},
    "HONESTI": {"-": ["12"]},
    "HONESTO": {"-": ["51"]},
    "HONESTVM": {"-": ["12"]},
    "HONESTVS": {"-": ["21"]},
    "HONOR": {"-": ["13"], "N": ["13"]},
    "HONORARIVS": {"-": ["21"]},
    "HONORATVS": {"-": ["21"], "N": ["12"]},
    "HONORIFICE": {"-": ["60"]},
    "HONORIFICVS": {"-": ["21"]},
    "HONORO": {"-": ["51"]},
    "HONORVS": {"-": ["21"]},
    "HORA": {"-": ["11"], "N": ["11"]},
    "HORATIA": {"N": ["11"]},
    "HORATIVS": {"A": ["21"], "N": ["12"]},
    "HORDEACEVS": {"-": ["21"]},
    "HORDEVM": {"-": ["12"]},
    "HORIZON": {"-": ["17"]},
    "HORNOTINVS": {"-": ["21"]},
    "HORNVS": {"-": ["21"]},
    "HOROLOGIVM": {"-": ["12"]},
    "HOROSCOPVS": {"-": ["12"]},
    "HORRENDVM": {"-": ["60"]},
    "HORRENDVS": {"-": ["21"]},
    "HORRENS": {"-": ["25"]},
    "HORREO": {"-": ["52"]},
    "HORRESCO": {"-": ["53"]},
    "HORREVM": {"-": ["12"]},
    "HORRIBILIS": {"-": ["24"]},
    "HORRIDVLVS": {"-": ["21"]},
    "HORRIDVS": {"-": ["21"]},
    "HORRIFER": {"-": ["21"]},
    "HORRIFICE": {"-": ["60"]},
    "HORRIFICO": {"-": ["51"]},
    "HORRIFICVS": {"-": ["21"]},
    "HORRISONVS": {"-": ["21"]},
    "HORROR": {"-": ["13"], "N": ["13"]},
    "HORTAMEN": {"-": ["13"]},
    "HORTAMENTVM": {"-": ["12"]},
    "HORTATIO": {"-": ["13"]},
    "HORTATOR": {"-": ["13"]},
    "HORTATVS": {"-": ["14"]},
    "HORTENSIVS": {"N": ["12"]},
    "HORTENTIVS": {"N": ["12"]},
    "HORTOR": {"-": ["5J"]},
    "HORTVLVS": {"-": ["12"]},
    "HORTVS": {"-": ["12"]},
    "HOSPES": {"-": ["13"]},
    "HOSPITA": {"-": ["11"]},
    "HOSPITALIA": {"-": ["13"]},
    "HOSPITALIS": {"1": ["13"], "2": ["24"]},
    "HOSPITALITAS": {"-": ["13"]},
    "HOSPITALITER": {"-": ["60"]},
    "HOSPITIVM": {"-": ["12"]},
    "HOSPITOR": {"-": ["5J"]},
    "HOSPITVS": {"-": ["21"]},
    "HOSTES": {"-": ["13"]},
    "HOSTIA": {"-": ["11"]},
    "HOSTICVM": {"-": ["12"]},
    "HOSTICVS": {"-": ["21"]},
    "HOSTILE": {"-": ["13"]},
    "HOSTILIS": {"-": ["24"]},
    "HOSTILITER": {"-": ["60"]},
    "HOSTILIVS": {"A": ["21"], "N": ["12"]},
    "HOSTIS": {"-": ["13"]},
    "HOSTIVS": {"N": ["12"]},
    "HVC": {"-": ["60"]},
    "HVCCINE": {"-": ["67"]},
    "HVIVSMODI": {"-": ["60"]},
    "HVMANA": {"-": ["12"]},
    "HVMANE": {"-": ["60"]},
    "HVMANI": {"-": ["12"]},
    "HVMANITAS": {"-": ["13"]},
    "HVMANITER": {"-": ["60"]},
    "HVMANITVS": {"-": ["60"]},
    "HVMANVM": {"-": ["12"]},
    "HVMANVS": {"-": ["21"]},
    "HVMECTO": {"-": ["51"]},
    "HVMECTVS": {"-": ["21"]},
    "HVMEO": {"-": ["52"]},
    "HVMERVS": {"-": ["12"]},
    "HVMESCO": {"-": ["53"]},
    "HVMIDITAS": {"-": ["13"]},
    "HVMIDVLVS": {"-": ["21"]},
    "HVMIDVM": {"-": ["12"]},
    "HVMIDVS": {"-": ["21"]},
    "HVMILIS": {"-": ["24"]},
    "HVMILITAS": {"-": ["13"]},
    "HVMILITER": {"-": ["60"]},
    "HVMO": {"-": ["51"]},
    "HVMOR": {"-": ["13"]},
    "HVMVS": {"-": ["12"]},
    "HYACINTHINVS": {"-": ["21"]},
    "HYACINTHVS": {"-": ["12"]},
    "HYALVS": {"-": ["12"]},
    "HYBLENSES": {"N": ["13"]},
    "HYBRIDA": {"-": ["11"]},
    "HYDOR": {"-": ["17"]},
    "HYDRA": {"-": ["11"], "N": ["11"]},
    "HYDRALETES": {"-": ["17"]},
    "HYDRAVLICVS": {"-": ["21"]},
    "HYDRIA": {"-": ["11"]},
    "HYDRIAS": {"-": ["27"]},
    "HYDROPICVS": {"-": ["21"]},
    "HYDROPS": {"-": ["13"]},
    "HYDRVS": {"-": ["17"]},
    "HYMEN": {"-": ["13"], "N": ["13"]},
    "HYMENAEOS": {"-": ["17"]},
    "HYMENEIVS": {"-": ["21"]},
    "HYPAETROS": {"-": ["27"]},
    "HYPATE": {"-": ["17"]},
    "HYPATOS": {"-": ["27"]},
    "HYPERBOLAEOS": {"1": ["17"], "2": ["27"]},
    "HYPERBOLE": {"-": ["17"]},
    "HYPERTHYRVM": {"-": ["12"]},
    "HYPOCAVSIS": {"-": ["17"]},
    "HYPOGEVM": {"-": ["12"]},
    "HYPOMOCHLION": {"-": ["17"]},
    "HYPOTRACHELIVM": {"-": ["17"]},
    "HYSGINVM": {"-": ["12"]},
    "IACCHVS": {"N": ["12"]},
    "IACEO": {"-": ["52"]},
    "IACIO": {"-": ["55"]},
    "IACTANS": {"-": ["25"]},
    "IACTANTER": {"-": ["6&"]},
    "IACTANTIA": {"-": ["11"]},
    "IACTATIO": {"-": ["13"]},
    "IACTATVS": {"-": ["14"]},
    "IACTO": {"-": ["51"]},
    "IACTVRA": {"-": ["11"]},
    "IACTVS": {"-": ["14"]},
    "IACVLATOR": {"-": ["13"]},
    "IACVLATRIX": {"-": ["13"]},
    "IACVLOR": {"-": ["5J"]},
    "IACVLVM": {"-": ["12"]},
    "IACVLVS": {"1": ["12"], "2": ["21"]},
    "IALYSVS": {"N": ["12"]},
    "IAM": {"-": ["60"]},
    "IAMBVS": {"-": ["12"]},
    "IAMDIV": {"-": ["60"]},
    "IAMDVDVM": {"-": ["60"]},
    "IAMIAM": {"-": ["60"]},
    "IAMPRIDEM": {"-": ["60"]},
    "IANICVLVM": {"N": ["12"]},
    "IANITOR": {"-": ["13"], "N": ["13"]},
    "IANVA": {"-": ["11"]},
    "IANVARIVS": {"A": ["21"], "N": ["12"]},
    "IANVS": {"N": ["12"], "-": ["12"]},
    "IASPIS": {"-": ["17"]},
    "IBERVS": {"N": ["12"], "A": ["21"]},
    "IBI": {"-": ["60"]},
    "IBIDEM": {"-": ["60"]},
    "IBIS": {"-": ["17"], "N": ["13"]},
    "ICHNEVMON": {"-": ["17"]},
    "ICHNOGRAPHIA": {"-": ["11"]},
    "ICILIVS": {"N": ["12"]},
    "ICIO": {"-": ["55"]},
    "ICO": {"-": ["53"]},
    "ICONISMVS": {"-": ["12"]},
    "ICTERICVS": {"-": ["21"]},
    "ICTVS": {"-": ["14"]},
    "IDAEVS": {"A": ["21"], "N": ["12"]},
    "IDCIRCO": {"-": ["60"]},
    "IDEA": {"-": ["17"]},
    "IDEM": {"-": ["45"]},
    "IDENTIDEM": {"-": ["60"]},
    "IDEO": {"-": ["60"]},
    "IDIOTA": {"-": ["17"]},
    "IDONEVS": {"-": ["21"]},
    "IDOS": {"-": ["17"]},
    "IDVS": {"-": ["14"]},
    "IECTO": {"-": ["51"]},
    "IECVR": {"-": ["13"]},
    "IEIVNITAS": {"-": ["13"]},
    "IEIVNIVM": {"-": ["12"]},
    "IEIVNVS": {"-": ["21"]},
    "IETINI": {"N": ["12"]},
    "IGITVR": {"-": ["81"]},
    "IGNARVS": {"-": ["21"]},
    "IGNAVE": {"-": ["60"]},
    "IGNAVIA": {"-": ["11"]},
    "IGNAVITER": {"-": ["60"]},
    "IGNAVVS": {"1": ["12"], "2": ["21"]},
    "IGNESCO": {"-": ["53"]},
    "IGNEVS": {"-": ["21"]},
    "IGNIARIVM": {"-": ["12"]},
    "IGNICVLVS": {"-": ["12"]},
    "IGNIFER": {"-": ["21"]},
    "IGNIS": {"-": ["13"]},
    "IGNOBILIS": {"-": ["24"]},
    "IGNOBILITAS": {"-": ["13"]},
    "IGNOMINIA": {"-": ["11"]},
    "IGNOMINIOSVS": {"-": ["21"]},
    "IGNORANS": {"-": ["25"]},
    "IGNORANTIA": {"-": ["11"]},
    "IGNORATIO": {"-": ["13"]},
    "IGNORO": {"-": ["51"]},
    "IGNOSCO": {"-": ["53"]},
    "IGNOTITIA": {"-": ["11"]},
    "IGNOTVS": {"-": ["21"]},
    "IGO": {"-": ["53"]},
    "ILEX": {"-": ["13"]},
    "ILIA": {"-": ["16"], "N": ["11"]},
    "ILICET": {"-": ["60"]},
    "ILIGNEVS": {"-": ["21"]},
    "ILIGNVS": {"-": ["21"]},
    "ILIVM": {"-": ["12"], "N": ["12"]},
    "ILLA": {"-": ["60"]},
    "ILLABOR": {"2": ["5L"]},
    "ILLABORATVS": {"-": ["21"]},
    "ILLABORO": {"-": ["51"]},
    "ILLAC": {"-": ["60"]},
    "ILLACESSITVS": {"-": ["21"]},
    "ILLACRIMABILIS": {"-": ["24"]},
    "ILLACRIMO": {"-": ["51"]},
    "ILLACRIMOR": {"-": ["5J"]},
    "ILLAESVS": {"-": ["21"]},
    "ILLAETABILIS": {"-": ["24"]},
    "ILLAQVEO": {"2": ["51"]},
    "ILLARGIO": {"-": ["54"]},
    "ILLAVDATVS": {"-": ["21"]},
    "ILLAVTVS": {"-": ["21"]},
    "ILLE": {"-": ["45"]},
    "ILLECEBRA": {"-": ["11"]},
    "ILLECTO": {"-": ["51"]},
    "ILLECTVS": {"1": ["14"], "2": ["21"]},
    "ILLEPIDE": {"-": ["60"]},
    "ILLEPIDVS": {"-": ["21"]},
    "ILLI": {"-": ["60"]},
    "ILLIBATVS": {"-": ["21"]},
    "ILLIBERALIS": {"-": ["24"]},
    "ILLIC": {"1": ["45"], "2": ["60"]},
    "ILLICIO": {"-": ["55"]},
    "ILLICITE": {"1": ["60"], "2": ["60"]},
    "ILLICITVS": {"-": ["21"]},
    "ILLICO": {"-": ["60"]},
    "ILLIDO": {"-": ["53"]},
    "ILLIGO": {"2": ["51"]},
    "ILLIM": {"-": ["60"]},
    "ILLINC": {"-": ["60"]},
    "ILLINO": {"-": ["53"]},
    "ILLITTERATVS": {"-": ["21"]},
    "ILLIVSMODI": {"-": ["60"]},
    "ILLO": {"-": ["60"]},
    "ILLOC": {"-": ["60"]},
    "ILLOTVS": {"-": ["21"]},
    "ILLVC": {"-": ["60"]},
    "ILLVCEO": {"-": ["52"]},
    "ILLVCESCO": {"-": ["53"]},
    "ILLVDO": {"-": ["53"]},
    "ILLVMINO": {"-": ["51"]},
    "ILLVO": {"-": ["53"]},
    "ILLVSTRIS": {"-": ["24"]},
    "ILLVSTRO": {"2": ["51"]},
    "ILLVVIES": {"-": ["15"]},
    "ILLYRICVM": {"N": ["12"]},
    "IMACHARENSES": {"N": ["13"]},
    "IMACHARENSIS": {"N": ["24"], "A": ["24"]},
    "IMAGINARIVS": {"-": ["21"]},
    "IMAGINATIO": {"-": ["13"]},
    "IMAGINOR": {"-": ["5J"]},
    "IMAGINOSVS": {"-": ["21"]},
    "IMAGO": {"-": ["13"]},
    "IMBECILLI": {"-": ["12"]},
    "IMBECILLIS": {"-": ["24"]},
    "IMBECILLITAS": {"-": ["13"]},
    "IMBECILLVS": {"-": ["21"]},
    "IMBELLIS": {"-": ["24"]},
    "IMBER": {"-": ["13"]},
    "IMBERBIS": {"-": ["24"]},
    "IMBIBO": {"2": ["53"]},
    "IMBREX": {"-": ["13"]},
    "IMBRICATVS": {"-": ["21"]},
    "IMBRIFER": {"-": ["21"]},
    "IMBVO": {"-": ["53"]},
    "IMITABILIS": {"-": ["24"]},
    "IMITAMENTVM": {"-": ["12"]},
    "IMITATIO": {"-": ["13"]},
    "IMITATOR": {"-": ["13"]},
    "IMITATRIX": {"-": ["13"]},
    "IMITOR": {"-": ["5J"]},
    "IMMANE": {"-": ["60"]},
    "IMMANIS": {"-": ["24"]},
    "IMMANITAS": {"-": ["13"]},
    "IMMANSVETVS": {"-": ["21"]},
    "IMMATVRVS": {"-": ["21"]},
    "IMMEDIATVS": {"-": ["21"]},
    "IMMEIO": {"-": ["53"]},
    "IMMEMOR": {"-": ["26"]},
    "IMMEMORABILIS": {"-": ["24"]},
    "IMMENSVM": {"1": ["12"], "2": ["60"]},
    "IMMENSVS": {"-": ["21"]},
    "IMMERENS": {"-": ["25"]},
    "IMMERGO": {"-": ["53"]},
    "IMMERITO": {"-": ["60"]},
    "IMMERITVS": {"-": ["21"]},
    "IMMERSABILIS": {"-": ["24"]},
    "IMMETATVS": {"-": ["21"]},
    "IMMIGRO": {"-": ["51"]},
    "IMMINEO": {"-": ["52"]},
    "IMMINVO": {"-": ["53"]},
    "IMMISCEO": {"-": ["52"]},
    "IMMISERABILIS": {"-": ["24"]},
    "IMMISSARIVM": {"-": ["12"]},
    "IMMITIS": {"-": ["24"]},
    "IMMITTO": {"-": ["53"]},
    "IMMO": {"-": ["60"]},
    "IMMOBILIS": {"-": ["24"]},
    "IMMODERATE": {"-": ["60"]},
    "IMMODERATVS": {"-": ["21"]},
    "IMMODESTIA": {"-": ["11"]},
    "IMMODESTVS": {"-": ["21"]},
    "IMMODICE": {"-": ["60"]},
    "IMMODICVS": {"-": ["21"]},
    "IMMOLATIO": {"-": ["13"]},
    "IMMOLO": {"-": ["51"]},
    "IMMORDEO": {"-": ["52"]},
    "IMMORIOR": {"-": ["5N"]},
    "IMMOROR": {"-": ["5J"], "1": ["5J"]},
    "IMMORTALIS": {"-": ["24"]},
    "IMMORTALITAS": {"-": ["13"]},
    "IMMOTVS": {"-": ["21"]},
    "IMMVGIO": {"-": ["54"]},
    "IMMVNDVS": {"-": ["21"]},
    "IMMVNIO": {"2": ["54"]},
    "IMMVNIS": {"-": ["24"]},
    "IMMVNITAS": {"-": ["13"]},
    "IMMVNITVS": {"-": ["21"]},
    "IMMVRMVRO": {"-": ["51"]},
    "IMMVTABILIS": {"1": ["24"], "2": ["24"]},
    "IMMVTATIO": {"-": ["13"]},
    "IMMVTO": {"2": ["51"]},
    "IMO": {"-": ["53"]},
    "IMPACATVS": {"-": ["21"]},
    "IMPAGES": {"-": ["13"]},
    "IMPALLESCO": {"-": ["53"]},
    "IMPAR": {"-": ["25"]},
    "IMPARATVS": {"-": ["21"]},
    "IMPATIENS": {"-": ["25"]},
    "IMPATIENTER": {"-": ["60"]},
    "IMPATIENTIA": {"-": ["11"]},
    "IMPAVIDE": {"-": ["60"]},
    "IMPAVIDVS": {"-": ["21"]},
    "IMPEDIMENTVM": {"-": ["12"]},
    "IMPEDIO": {"-": ["54"]},
    "IMPEDITIO": {"-": ["13"]},
    "IMPEDITVM": {"-": ["12"]},
    "IMPEDITVS": {"-": ["21"]},
    "IMPELLO": {"-": ["53"]},
    "IMPENDEO": {"-": ["52"]},
    "IMPENDIVM": {"-": ["12"]},
    "IMPENDO": {"-": ["53"]},
    "IMPENETRABILIS": {"-": ["24"]},
    "IMPENSA": {"-": ["11"]},
    "IMPENSE": {"-": ["60"]},
    "IMPENSVS": {"-": ["21"]},
    "IMPERATOR": {"-": ["13"], "N": ["13"]},
    "IMPERATORIVS": {"-": ["21"]},
    "IMPERATVM": {"-": ["12"]},
    "IMPERCVSSVS": {"-": ["21"]},
    "IMPERFECTVS": {"-": ["21"]},
    "IMPERIOSVS": {"-": ["21"]},
    "IMPERITIA": {"-": ["11"]},
    "IMPERITO": {"-": ["51"]},
    "IMPERITVS": {"-": ["21"]},
    "IMPERIVM": {"-": ["12"]},
    "IMPERMISSVS": {"-": ["21"]},
    "IMPERO": {"-": ["51"]},
    "IMPERPETVVS": {"-": ["21"]},
    "IMPERTIO": {"-": ["54"]},
    "IMPERTVRBATVS": {"-": ["21"]},
    "IMPERVIVS": {"-": ["21"]},
    "IMPES": {"-": ["13"]},
    "IMPETRABILIS": {"-": ["24"]},
    "IMPETRO": {"-": ["51"]},
    "IMPETVS": {"-": ["14"]},
    "IMPEXVS": {"-": ["21"]},
    "IMPIE": {"-": ["60"]},
    "IMPIETAS": {"-": ["13"], "N": ["13"]},
    "IMPIGER": {"-": ["21"]},
    "IMPIGRE": {"-": ["60"]},
    "IMPINGO": {"-": ["53"]},
    "IMPIO": {"-": ["51"]},
    "IMPIVS": {"-": ["21"]},
    "IMPLACABILIS": {"-": ["24"]},
    "IMPLACABILITER": {"-": ["6&"]},
    "IMPLACATVS": {"-": ["21"]},
    "IMPLACIDVS": {"-": ["21"]},
    "IMPLECTO": {"2": ["53"]},
    "IMPLEO": {"-": ["52"]},
    "IMPLICATVS": {"-": ["21"]},
    "IMPLICITVS": {"-": ["21"]},
    "IMPLICO": {"-": ["51"]},
    "IMPLORATIO": {"-": ["13"]},
    "IMPLORO": {"-": ["51"]},
    "IMPLVMBO": {"-": ["51"]},
    "IMPLVMIS": {"-": ["24"]},
    "IMPLVO": {"-": ["53"]},
    "IMPLVVIVM": {"-": ["12"]},
    "IMPOLITVS": {"-": ["21"]},
    "IMPOLLVTVS": {"-": ["21"]},
    "IMPONO": {"-": ["53"]},
    "IMPORTATICIVS": {"-": ["21"]},
    "IMPORTO": {"-": ["51"]},
    "IMPORTVNITAS": {"-": ["13"]},
    "IMPORTVNVS": {"-": ["21"]},
    "IMPORTVOSVS": {"-": ["21"]},
    "IMPOS": {"-": ["22"]},
    "IMPOTENS": {"-": ["25"]},
    "IMPOTENTER": {"-": ["60"]},
    "IMPOTENTIA": {"-": ["11"]},
    "IMPRAESENTIARVM": {"-": ["60"]},
    "IMPRANSVS": {"-": ["21"]},
    "IMPRECATIO": {"-": ["13"]},
    "IMPRECOR": {"-": ["5J"]},
    "IMPRESSIO": {"-": ["13"]},
    "IMPRESSO": {"-": ["51"]},
    "IMPRIMIS": {"-": ["60"]},
    "IMPRIMO": {"-": ["53"]},
    "IMPROBABILIS": {"-": ["24"]},
    "IMPROBATIO": {"-": ["13"]},
    "IMPROBE": {"-": ["60"]},
    "IMPROBITAS": {"-": ["13"]},
    "IMPROBO": {"-": ["51"]},
    "IMPROBVLVS": {"-": ["21"]},
    "IMPROBVS": {"-": ["21"]},
    "IMPROCERVS": {"-": ["21"]},
    "IMPROMPTVS": {"-": ["21"]},
    "IMPROPERATVS": {"-": ["21"]},
    "IMPROPERO": {"-": ["51"]},
    "IMPROSPER": {"-": ["21"]},
    "IMPROSPERE": {"-": ["60"]},
    "IMPROVIDE": {"-": ["60"]},
    "IMPROVIDVS": {"-": ["21"]},
    "IMPROVISO": {"-": ["60"]},
    "IMPROVISVM": {"-": ["12"]},
    "IMPROVISVS": {"-": ["21"]},
    "IMPRVDENS": {"-": ["25"]},
    "IMPRVDENTIA": {"-": ["11"]},
    "IMPVBES": {"-": ["26"]},
    "IMPVDENS": {"-": ["25"]},
    "IMPVDENTER": {"-": ["60"]},
    "IMPVDENTIA": {"-": ["11"]},
    "IMPVDICITIA": {"-": ["11"]},
    "IMPVDICVS": {"-": ["21"]},
    "IMPVGNO": {"-": ["51"]},
    "IMPVLSOR": {"-": ["13"]},
    "IMPVLSVS": {"-": ["14"]},
    "IMPVNE": {"-": ["60"]},
    "IMPVNIS": {"-": ["24"]},
    "IMPVNITAS": {"-": ["13"]},
    "IMPVNITVS": {"-": ["21"]},
    "IMPVRVS": {"-": ["21"]},
    "IMPVTATOR": {"-": ["13"]},
    "IMPVTATVS": {"-": ["21"]},
    "IMPVTO": {"-": ["51"]},
    "IMVLVS": {"-": ["21"]},
    "IMVM": {"-": ["12"]},
    "IN": {"-": ["70"]},
    "INACCEDO": {"1": ["53"]},
    "INACCESSVS": {"-": ["21"]},
    "INACESCO": {"-": ["53"]},
    "INADVSTVS": {"-": ["21"]},
    "INAEDIFICO": {"-": ["51"]},
    "INAEQVABILIS": {"-": ["24"]},
    "INAEQVALIS": {"-": ["24"]},
    "INAEQVALITAS": {"-": ["13"]},
    "INAEQVALITER": {"-": ["60"]},
    "INAEQVATVS": {"-": ["21"]},
    "INAEQVO": {"-": ["51"]},
    "INAESTIMABILIS": {"-": ["24"]},
    "INAESTVO": {"-": ["51"]},
    "INAFFECTATVS": {"-": ["21"]},
    "INAFFECTO": {"-": ["51"]},
    "INAGITATVS": {"-": ["21"]},
    "INAMABILIS": {"-": ["24"]},
    "INAMARESCO": {"-": ["53"]},
    "INAMBVLATIO": {"-": ["13"]},
    "INAMBVLO": {"-": ["51"]},
    "INAMOENVS": {"-": ["21"]},
    "INANE": {"-": ["13"]},
    "INANIMALIS": {"-": ["24"]},
    "INANIMATVS": {"-": ["21"]},
    "INANIMIS": {"-": ["24"]},
    "INANIMVS": {"-": ["21"]},
    "INANIO": {"-": ["54"]},
    "INANIS": {"-": ["24"]},
    "INANITAS": {"-": ["13"]},
    "INANITER": {"-": ["60"]},
    "INARATVS": {"-": ["21"]},
    "INARDEO": {"-": ["52"]},
    "INARDESCO": {"-": ["53"]},
    "INARESCO": {"-": ["53"]},
    "INARO": {"-": ["51"]},
    "INASCENSVS": {"-": ["21"]},
    "INASSVETVS": {"-": ["21"]},
    "INAVDAX": {"-": ["25"]},
    "INAVDIO": {"-": ["54"]},
    "INAVDITVS": {"-": ["21"]},
    "INAVRATVS": {"-": ["21"]},
    "INAVRO": {"-": ["51"]},
    "INAVSPICATVS": {"-": ["21"]},
    "INAVSVS": {"-": ["21"]},
    "INCAEDVVS": {"-": ["21"]},
    "INCALESCO": {"-": ["53"]},
    "INCALLIDVS": {"-": ["21"]},
    "INCANDEO": {"-": ["52"]},
    "INCANDESCO": {"-": ["53"]},
    "INCANESCO": {"-": ["53"]},
    "INCANTO": {"-": ["51"]},
    "INCANVS": {"-": ["21"]},
    "INCASSVM": {"-": ["60"]},
    "INCASTIGO": {"-": ["51"]},
    "INCAVTE": {"-": ["60"]},
    "INCAVTVS": {"-": ["21"]},
    "INCEDO": {"1": ["53"]},
    "INCELEBRATVS": {"-": ["21"]},
    "INCENATVS": {"-": ["21"]},
    "INCENDIARIVS": {"1": ["12"], "2": ["21"]},
    "INCENDIVM": {"-": ["12"]},
    "INCENDO": {"-": ["53"]},
    "INCENSIO": {"-": ["13"]},
    "INCENSVM": {"-": ["12"]},
    "INCENSVS": {"1": ["21"], "2": ["21"]},
    "INCEPTVM": {"-": ["12"]},
    "INCEPTVS": {"1": ["14"], "2": ["21"]},
    "INCERNICVLVM": {"-": ["12"]},
    "INCERNO": {"-": ["53"]},
    "INCERO": {"-": ["51"]},
    "INCERTVM": {"-": ["12"]},
    "INCERTVS": {"-": ["21"]},
    "INCESSO": {"-": ["53"]},
    "INCESSVS": {"-": ["14"]},
    "INCESTE": {"-": ["60"]},
    "INCESTIFICVS": {"-": ["21"]},
    "INCESTO": {"-": ["51"]},
    "INCESTVM": {"-": ["12"]},
    "INCESTVS": {"1": ["14"], "2": ["21"]},
    "INCHOO": {"-": ["51"]},
    "INCIDENTIA": {"-": ["13", "11"]},
    "INCIDO": {"1": ["53"], "2": ["53"]},
    "INCIEO": {"-": ["52"]},
    "INCILE": {"-": ["13"]},
    "INCILIS": {"-": ["24"]},
    "INCILO": {"-": ["51"]},
    "INCINGO": {"-": ["53"]},
    "INCINO": {"-": ["53"]},
    "INCIO": {"-": ["54"]},
    "INCIPIO": {"-": ["55"]},
    "INCISVM": {"-": ["12"], "L": ["12"]},
    "INCITAMENTVM": {"-": ["12"]},
    "INCITATIO": {"-": ["13"]},
    "INCITATVS": {"1": ["14"], "2": ["21"]},
    "INCITO": {"1": ["51"]},
    "INCITVS": {"1": ["21"], "2": ["21"]},
    "INCLAMO": {"-": ["51"]},
    "INCLARESCO": {"-": ["53"]},
    "INCLEMENS": {"-": ["25"]},
    "INCLEMENTER": {"-": ["60"]},
    "INCLEMENTIA": {"-": ["11"]},
    "INCLINABILIS": {"-": ["24"]},
    "INCLINATIO": {"-": ["13"]},
    "INCLINATVS": {"1": ["14"], "2": ["21"]},
    "INCLINO": {"-": ["51"]},
    "INCLVDO": {"-": ["53"]},
    "INCLVTVS": {"-": ["21"]},
    "INCOACTVS": {"-": ["21"]},
    "INCOGITATVS": {"-": ["21"]},
    "INCOGITO": {"-": ["51"]},
    "INCOGNITVS": {"-": ["21"]},
    "INCOHIBEO": {"-": ["52"]},
    "INCOHO": {"-": ["51"]},
    "INCOLA": {"-": ["11"]},
    "INCOLO": {"1": ["51"], "2": ["53"]},
    "INCOLVMIS": {"-": ["24"]},
    "INCOLVMITAS": {"-": ["13"]},
    "INCOMITATVS": {"-": ["21"]},
    "INCOMMODE": {"-": ["60"]},
    "INCOMMODVM": {"-": ["12"]},
    "INCOMMODVS": {"-": ["21"]},
    "INCOMPERTVS": {"-": ["21"]},
    "INCOMPOSITVS": {"-": ["21"]},
    "INCOMPREHENSIBILIS": {"-": ["24"]},
    "INCOMPTVS": {"-": ["21"]},
    "INCONCESSVS": {"-": ["21"]},
    "INCONCINNVS": {"-": ["21"]},
    "INCONCVSSVS": {"-": ["21"]},
    "INCONDEMNATVS": {"-": ["21"]},
    "INCONDITVS": {"-": ["21"]},
    "INCONFVSVS": {"-": ["21"]},
    "INCONSIDERATVS": {"-": ["21"]},
    "INCONSTANS": {"-": ["25"]},
    "INCONSTANTER": {"-": ["60"]},
    "INCONSTANTIA": {"-": ["11"]},
    "INCONSVETVS": {"-": ["21"]},
    "INCONSVLTE": {"-": ["60"]},
    "INCONSVLTVS": {"1": ["14"], "2": ["21"]},
    "INCONTAMINATVS": {"-": ["21"]},
    "INCONTINENS": {"-": ["25"]},
    "INCONVENIENS": {"-": ["25"]},
    "INCOQVO": {"-": ["53"]},
    "INCORPORALIS": {"-": ["24"]},
    "INCORRVPTVS": {"-": ["21"]},
    "INCREBRESCO": {"-": ["53"]},
    "INCREDIBILIS": {"-": ["24"]},
    "INCREDIBILITER": {"-": ["60"]},
    "INCREDVLVS": {"-": ["21"]},
    "INCREMENTVM": {"-": ["12"]},
    "INCREPATVS": {"-": ["21"]},
    "INCREPITO": {"-": ["51"]},
    "INCREPITVS": {"1": ["14"], "2": ["21"]},
    "INCREPO": {"-": ["51"]},
    "INCRESCO": {"-": ["53"]},
    "INCRVENTVS": {"-": ["21"]},
    "INCRVSTO": {"-": ["51"]},
    "INCVBO": {"1": ["13"], "2": ["51"], "N": ["13"]},
    "INCVLCO": {"-": ["51"]},
    "INCVLTE": {"-": ["6&"]},
    "INCVLTVS": {"1": ["14"], "2": ["21"]},
    "INCVMBA": {"-": ["11"]},
    "INCVMBO": {"-": ["53"]},
    "INCVNABVLA": {"-": ["12"]},
    "INCVRIA": {"-": ["11"]},
    "INCVRIOSVS": {"-": ["21"]},
    "INCVRO": {"-": ["51"]},
    "INCVRRO": {"-": ["53"]},
    "INCVRSIO": {"-": ["13"]},
    "INCVRSITO": {"-": ["51"]},
    "INCVRSO": {"-": ["51"]},
    "INCVRSVS": {"-": ["14"]},
    "INCVRVO": {"-": ["51"]},
    "INCVRVVS": {"-": ["21"]},
    "INCVS": {"-": ["13"]},
    "INCVSATIO": {"-": ["13"]},
    "INCVSO": {"-": ["51"]},
    "INCVSSVS": {"-": ["14"]},
    "INCVSTODITVS": {"-": ["21"]},
    "INCVSVS": {"-": ["21"]},
    "INCVTIO": {"-": ["55"]},
    "INDAGATIO": {"-": ["13"]},
    "INDAGATOR": {"-": ["13"]},
    "INDAGATRIX": {"-": ["13"]},
    "INDAGO": {"1": ["13"], "2": ["51"]},
    "INDE": {"-": ["60"]},
    "INDEBITVS": {"-": ["21"]},
    "INDECENS": {"-": ["25"]},
    "INDECENTER": {"-": ["6&"]},
    "INDECENTIA": {"-": ["11"]},
    "INDECLINABILIS": {"-": ["24"]},
    "INDECORE": {"-": ["60"]},
    "INDECORO": {"-": ["51"]},
    "INDECORVS": {"-": ["21"]},
    "INDEFATIGABILIS": {"-": ["24"]},
    "INDEFATIGATVS": {"-": ["21"]},
    "INDEFENSVS": {"-": ["21"]},
    "INDEFESSVS": {"-": ["21"]},
    "INDEFLEXVS": {"-": ["21"]},
    "INDELEBILIS": {"-": ["24"]},
    "INDEMNIS": {"-": ["24"]},
    "INDEPRAVATVS": {"-": ["21"]},
    "INDEPREHENSVS": {"-": ["21"]},
    "INDESERTVS": {"-": ["21"]},
    "INDEX": {"-": ["13"]},
    "INDIA": {"N": ["11"]},
    "INDICIVM": {"-": ["12"]},
    "INDICO": {"1": ["51"], "2": ["53"]},
    "INDICTIO": {"-": ["13"]},
    "INDICTVS": {"-": ["21"]},
    "INDIDEM": {"-": ["60"]},
    "INDIFFERENS": {"-": ["25"]},
    "INDIGENA": {"-": ["11"]},
    "INDIGENS": {"1": ["13"], "2": ["25"]},
    "INDIGENTIA": {"-": ["11"]},
    "INDIGEO": {"-": ["52"]},
    "INDIGES": {"N": ["13"]},
    "INDIGESTVS": {"-": ["21"]},
    "INDIGNANDVS": {"-": ["21"]},
    "INDIGNATIO": {"-": ["13"]},
    "INDIGNE": {"-": ["60"]},
    "INDIGNITAS": {"-": ["13"]},
    "INDIGNOR": {"-": ["5J"]},
    "INDIGNVS": {"-": ["21"]},
    "INDIGVS": {"-": ["21"]},
    "INDILIGENS": {"-": ["25"]},
    "INDILIGENTER": {"-": ["6&"]},
    "INDILIGENTIA": {"-": ["11"]},
    "INDIPISCOR": {"-": ["5L"]},
    "INDISCRETVS": {"-": ["21"]},
    "INDISPOSITE": {"-": ["60"]},
    "INDISTINCTVS": {"-": ["21"]},
    "INDIVIDVVS": {"-": ["21"]},
    "INDIVISVS": {"-": ["21"]},
    "INDO": {"-": ["53"]},
    "INDOCILIS": {"-": ["24"]},
    "INDOCTVS": {"-": ["21"]},
    "INDOLENTIA": {"-": ["11"]},
    "INDOLES": {"-": ["13"]},
    "INDOLESCO": {"-": ["53"]},
    "INDOMITVS": {"-": ["21"]},
    "INDORMIO": {"-": ["54"]},
    "INDOTATVS": {"-": ["21"]},
    "INDVBITATVS": {"-": ["21"]},
    "INDVBIVS": {"-": ["21"]},
    "INDVCO": {"-": ["53"]},
    "INDVCTIO": {"-": ["13"]},
    "INDVCTVS": {"-": ["14"]},
    "INDVLGENS": {"-": ["25"]},
    "INDVLGENTER": {"-": ["6&"]},
    "INDVLGENTIA": {"-": ["11"]},
    "INDVLGEO": {"-": ["52"]},
    "INDVO": {"-": ["53"]},
    "INDVRATVS": {"-": ["21"]},
    "INDVRESCO": {"-": ["53"]},
    "INDVRO": {"-": ["51"]},
    "INDVS": {"A": ["21"], "N": ["12"]},
    "INDVSTRIA": {"-": ["11"]},
    "INDVSTRIE": {"-": ["60"]},
    "INDVSTRIOSE": {"-": ["60"]},
    "INDVSTRIVS": {"-": ["21"]},
    "INDVTIAE": {"-": ["11"]},
    "INDVTILIS": {"-": ["24"]},
    "INDVTVS": {"-": ["14"]},
    "INEBRIO": {"-": ["51"]},
    "INEDIA": {"-": ["11"]},
    "INEFFICAX": {"-": ["25"]},
    "INELABORATVS": {"-": ["21"]},
    "INELEGANS": {"-": ["25"]},
    "INELVCTABILIS": {"-": ["24"]},
    "INEMENDABILIS": {"-": ["24"]},
    "INEMORIOR": {"-": ["5J"]},
    "INEMPTVS": {"-": ["21"]},
    "INENSES": {"N": ["13"]},
    "INEO": {"1": ["56"]},
    "INEPTE": {"-": ["60"]},
    "INEPTIAE": {"-": ["11"]},
    "INEPTIO": {"-": ["54"]},
    "INEPTVS": {"-": ["21"]},
    "INERMIS": {"-": ["24"]},
    "INERMVS": {"-": ["21"]},
    "INERRANS": {"-": ["25"]},
    "INERRO": {"2": ["51"]},
    "INERS": {"-": ["25"]},
    "INERTIA": {"-": ["11"]},
    "INESCO": {"-": ["51"]},
    "INEVITABILIS": {"-": ["24"]},
    "INEXCITABILIS": {"-": ["24"]},
    "INEXCVSABILIS": {"-": ["24"]},
    "INEXERCITATVS": {"-": ["21"]},
    "INEXERCITO": {"-": ["51"]},
    "INEXHAVSTVS": {"-": ["21"]},
    "INEXORABILIS": {"-": ["24"]},
    "INEXPERTVS": {"-": ["21"]},
    "INEXPIABILIS": {"-": ["24"]},
    "INEXPLEBILIS": {"-": ["24"]},
    "INEXPLICABILIS": {"-": ["24"]},
    "INEXPLORATO": {"-": ["60"]},
    "INEXPLORATVS": {"-": ["21"]},
    "INEXPVGNABILIS": {"-": ["24"]},
    "INEXSPECTATVS": {"-": ["21"]},
    "INEXSVPERABILIS": {"-": ["24"]},
    "INEXTRICABILIS": {"-": ["24"]},
    "INFABRE": {"-": ["60"]},
    "INFABRICATVS": {"-": ["21"]},
    "INFACETIAE": {"-": ["11"]},
    "INFACETVS": {"-": ["21"]},
    "INFAMIA": {"-": ["11"]},
    "INFAMIS": {"-": ["24"]},
    "INFAMO": {"-": ["51"]},
    "INFANDVS": {"-": ["21"]},
    "INFANS": {"1": ["13"], "2": ["25"]},
    "INFANTIA": {"-": ["11"]},
    "INFATIGABILIS": {"-": ["24"]},
    "INFATVO": {"-": ["51"]},
    "INFAVSTVS": {"-": ["21"]},
    "INFECTIVVS": {"-": ["21"]},
    "INFECTVS": {"1": ["14"], "2": ["21"]},
    "INFECVNDITAS": {"-": ["13"]},
    "INFECVNDVS": {"-": ["21"]},
    "INFELICITAS": {"-": ["13"]},
    "INFELICITER": {"-": ["60"]},
    "INFELIX": {"-": ["25"]},
    "INFENDO": {"-": ["53"]},
    "INFENSE": {"-": ["6&"]},
    "INFENSO": {"-": ["51"]},
    "INFENSVS": {"-": ["21"]},
    "INFERI": {"-": ["12"], "N": ["12"]},
    "INFERIAE": {"-": ["11"]},
    "INFERIORA": {"-": ["13"]},
    "INFERIORES": {"-": ["13"]},
    "INFERIVS": {"-": ["21"]},
    "INFERNA": {"-": ["12"]},
    "INFERNAS": {"-": ["22"]},
    "INFERNE": {"-": ["60"]},
    "INFERNI": {"-": ["12"]},
    "INFERNVS": {"-": ["21"]},
    "INFERO": {"-": ["56"]},
    "INFERVEFACIO": {"-": ["55"]},
    "INFERVEO": {"-": ["52"]},
    "INFERVESCO": {"-": ["53"]},
    "INFERVS": {"-": ["21"]},
    "INFESTE": {"-": ["6&"]},
    "INFESTO": {"-": ["51"]},
    "INFESTVS": {"-": ["21"]},
    "INFEXIO": {"-": ["13"]},
    "INFICIO": {"-": ["55"]},
    "INFICISCO": {"-": ["53"]},
    "INFIDELIS": {"-": ["24"]},
    "INFIDELITAS": {"-": ["13"]},
    "INFIDVS": {"-": ["21"]},
    "INFIGO": {"-": ["53"]},
    "INFIMVM": {"-": ["12"]},
    "INFINDO": {"-": ["53"]},
    "INFINITVM": {"-": ["60"]},
    "INFINITVS": {"-": ["21"]},
    "INFIO": {"-": ["56"]},
    "INFIRMIS": {"-": ["24"]},
    "INFIRMITAS": {"-": ["13"]},
    "INFIRMO": {"-": ["51"]},
    "INFIRMVS": {"-": ["21"]},
    "INFITIAS": {"-": ["16"]},
    "INFITIATIO": {"-": ["13"]},
    "INFITIATOR": {"-": ["13"]},
    "INFITIOR": {"-": ["5J"]},
    "INFLAMMO": {"-": ["51"]},
    "INFLATE": {"-": ["6&"]},
    "INFLATIO": {"-": ["13"]},
    "INFLATVS": {"1": ["14"], "2": ["21"]},
    "INFLECTIO": {"-": ["13"]},
    "INFLECTO": {"-": ["53"]},
    "INFLEXIBILIS": {"-": ["24"]},
    "INFLEXILIS": {"-": ["24"]},
    "INFLEXIO": {"-": ["13"]},
    "INFLIGO": {"-": ["53"]},
    "INFLO": {"-": ["51"]},
    "INFLVO": {"-": ["53"]},
    "INFODIO": {"-": ["55"]},
    "INFORMIS": {"-": ["24"]},
    "INFORMO": {"-": ["51"]},
    "INFRA": {"1": ["60"], "2": ["70"]},
    "INFRACTVS": {"-": ["21"]},
    "INFRAGILIS": {"-": ["24"]},
    "INFREMO": {"-": ["53"]},
    "INFRENDEO": {"-": ["52"]},
    "INFRENVS": {"-": ["21"]},
    "INFREQVENS": {"-": ["25"]},
    "INFREQVENTIA": {"-": ["11"]},
    "INFRINGO": {"-": ["53"]},
    "INFRIO": {"-": ["51"]},
    "INFRVCTVOSVS": {"-": ["21"]},
    "INFRVNITVS": {"-": ["21"]},
    "INFVLA": {"-": ["11"]},
    "INFVLCIO": {"-": ["54"]},
    "INFVNDIBVLVM": {"-": ["12"]},
    "INFVNDO": {"2": ["53"]},
    "INFVSCO": {"-": ["51"]},
    "INGEMINO": {"-": ["51"]},
    "INGEMISCO": {"-": ["53"]},
    "INGEMO": {"-": ["53"]},
    "INGENERO": {"-": ["51"]},
    "INGENIOSE": {"-": ["60"]},
    "INGENIOSVS": {"-": ["21"]},
    "INGENITVS": {"1": ["21"], "2": ["21"]},
    "INGENIVM": {"-": ["12"]},
    "INGENS": {"-": ["25"]},
    "INGENVE": {"-": ["60"]},
    "INGENVITAS": {"-": ["13"]},
    "INGENVVS": {"-": ["21"]},
    "INGERO": {"-": ["53"]},
    "INGLORIVS": {"-": ["21"]},
    "INGLVVIES": {"-": ["15"]},
    "INGRATE": {"-": ["60"]},
    "INGRATIS": {"-": ["60"]},
    "INGRATVS": {"-": ["21"]},
    "INGRAVESCO": {"-": ["53"]},
    "INGREDIOR": {"-": ["5N"]},
    "INGRESSVS": {"-": ["14"]},
    "INGRVO": {"-": ["53"]},
    "INGVEN": {"-": ["13"]},
    "INGVSTATVS": {"-": ["21"]},
    "INHABILIS": {"-": ["24"]},
    "INHABITO": {"-": ["51"]},
    "INHAEREO": {"-": ["52"]},
    "INHIBEO": {"-": ["52"]},
    "INHIO": {"-": ["51"]},
    "INHONESTVS": {"-": ["21"]},
    "INHONORIFICVS": {"-": ["21"]},
    "INHORREO": {"-": ["52"]},
    "INHORRESCO": {"-": ["53"]},
    "INHOSPITALIS": {"-": ["24"]},
    "INHOSPITVS": {"-": ["21"]},
    "INHVMANE": {"-": ["60"]},
    "INHVMANITAS": {"-": ["13"]},
    "INHVMANITER": {"-": ["60"]},
    "INHVMANVS": {"-": ["21"]},
    "INHVMATVS": {"-": ["21"]},
    "INIBI": {"-": ["60"]},
    "INICIO": {"-": ["55"]},
    "INIECTIO": {"-": ["13"]},
    "INIECTVS": {"-": ["14"]},
    "INIMICITIA": {"-": ["11"]},
    "INIMICO": {"-": ["51"]},
    "INIMICVS": {"1": ["12"], "2": ["21"]},
    "INIMO": {"-": ["53"]},
    "INIQVE": {"-": ["60"]},
    "INIQVITAS": {"-": ["13"]},
    "INIQVVS": {"1": ["12"], "2": ["21"]},
    "INITIAMENTA": {"-": ["12"]},
    "INITIO": {"-": ["51"]},
    "INITIVM": {"-": ["12"]},
    "INITVS": {"-": ["14"]},
    "INIVCVNDVS": {"-": ["21"]},
    "INIVDICATVS": {"-": ["21"]},
    "INIVNGO": {"-": ["53"]},
    "INIVRATVS": {"-": ["21"]},
    "INIVRIA": {"-": ["11"]},
    "INIVRIOSE": {"-": ["6&"]},
    "INIVRIOSVS": {"-": ["21"]},
    "INIVROSVS": {"-": ["21"]},
    "INIVSSVS": {"1": ["14"], "2": ["21"]},
    "INIVSTE": {"-": ["60"]},
    "INIVSTITIA": {"-": ["11"]},
    "INIVSTVM": {"-": ["12"]},
    "INIVSTVS": {"-": ["21"]},
    "INLEXILIS": {"-": ["24"]},
    "INNABILIS": {"-": ["24"]},
    "INNASCOR": {"-": ["5L"]},
    "INNATO": {"-": ["51"]},
    "INNECTO": {"-": ["53"]},
    "INNITOR": {"2": ["5L"]},
    "INNO": {"-": ["51"]},
    "INNOCENS": {"-": ["25"]},
    "INNOCENTER": {"-": ["60"]},
    "INNOCENTIA": {"-": ["11"]},
    "INNOCVE": {"-": ["60"]},
    "INNOCVVS": {"-": ["21"]},
    "INNOTESCO": {"-": ["53"]},
    "INNOXIVS": {"-": ["21"]},
    "INNVBILVS": {"-": ["21"]},
    "INNVBIS": {"-": ["24"]},
    "INNVBVS": {"-": ["21"]},
    "INNVMERABILIS": {"-": ["24"]},
    "INNVMERABILITER": {"-": ["60"]},
    "INNVMERALIS": {"-": ["24"]},
    "INNVMERVS": {"-": ["21"]},
    "INNVO": {"-": ["53"]},
    "INNVPTA": {"-": ["11"]},
    "INNVPTVS": {"-": ["21"]},
    "INNVTRIO": {"-": ["54"]},
    "INOBSEQVENS": {"-": ["25"]},
    "INOBSERVABILIS": {"-": ["24"]},
    "INOBSERVATVS": {"-": ["21"]},
    "INODORVS": {"-": ["21"]},
    "INOFFENSE": {"-": ["60"]},
    "INOFFENSVS": {"-": ["21"]},
    "INOFFICIOSVS": {"-": ["21"]},
    "INOLENS": {"-": ["25"]},
    "INOLESCO": {"-": ["53"]},
    "INOMINATVS": {"-": ["21"]},
    "INOPERATVS": {"-": ["21"]},
    "INOPIA": {"-": ["11"]},
    "INOPINANS": {"-": ["25"]},
    "INOPINATVM": {"-": ["12"]},
    "INOPINATVS": {"-": ["21"]},
    "INOPINVS": {"-": ["21"]},
    "INOPS": {"-": ["26"]},
    "INORDINATVS": {"-": ["21"]},
    "INORNATVS": {"-": ["21"]},
    "INQVIES": {"1": ["13"], "2": ["25"]},
    "INQVIETO": {"-": ["51"]},
    "INQVIETVS": {"-": ["21"]},
    "INQVILINVS": {"-": ["12"]},
    "INQVINAMENTVM": {"-": ["12"]},
    "INQVINATVS": {"-": ["21"]},
    "INQVINO": {"-": ["51"]},
    "INQVIO": {"-": ["56"]},
    "INQVIRO": {"-": ["53"]},
    "INQVISITIO": {"-": ["13"]},
    "INQVISITOR": {"-": ["13"]},
    "INREGILLVM": {"N": ["12"]},
    "INSAEPTVS": {"1": ["21"], "2": ["21"]},
    "INSALVBRIS": {"-": ["24"]},
    "INSANABILIS": {"-": ["24"]},
    "INSANE": {"-": ["60"]},
    "INSANIA": {"-": ["11"]},
    "INSANIO": {"-": ["54"]},
    "INSANVM": {"-": ["60"]},
    "INSANVS": {"-": ["21"]},
    "INSAPIENS": {"-": ["25"]},
    "INSATIABILIS": {"-": ["24"]},
    "INSATIABILITER": {"-": ["60"]},
    "INSCENDO": {"-": ["53"]},
    "INSCIENS": {"1": ["25"], "2": ["25"]},
    "INSCIENTIA": {"-": ["11"]},
    "INSCITE": {"-": ["60"]},
    "INSCITIA": {"-": ["11"]},
    "INSCITVS": {"-": ["21"]},
    "INSCIVS": {"-": ["21"]},
    "INSCRIBO": {"-": ["53"]},
    "INSCRIPTIO": {"-": ["13"]},
    "INSCVLPO": {"-": ["53"]},
    "INSECABILIS": {"-": ["24"]},
    "INSECO": {"-": ["51"]},
    "INSECTATIO": {"-": ["13"]},
    "INSECTOR": {"2": ["5J"]},
    "INSEDABILITER": {"-": ["60"]},
    "INSEMINO": {"-": ["51"]},
    "INSENESCO": {"-": ["53"]},
    "INSENSIBILIS": {"-": ["24"]},
    "INSEPARABILIS": {"-": ["24"]},
    "INSEPVLTVS": {"1": ["21"], "2": ["21"]},
    "INSEQVOR": {"-": ["5L"]},
    "INSERO": {"2": ["53"], "3": ["53"]},
    "INSERTO": {"-": ["51"]},
    "INSERVIO": {"-": ["54"]},
    "INSIDEO": {"-": ["52"]},
    "INSIDIAE": {"-": ["11"], "N": ["11"]},
    "INSIDIATOR": {"-": ["13"]},
    "INSIDIOR": {"-": ["5J"]},
    "INSIDIOSE": {"-": ["6-"]},
    "INSIDIOSVS": {"-": ["21"]},
    "INSIDO": {"-": ["53"]},
    "INSIGNE": {"-": ["13"]},
    "INSIGNIO": {"-": ["54"]},
    "INSIGNIS": {"-": ["24"]},
    "INSIGNITA": {"-": ["12"]},
    "INSIGNITVS": {"-": ["21"]},
    "INSILIA": {"-": ["13"]},
    "INSILIO": {"-": ["54"]},
    "INSIMVLATIO": {"-": ["13"]},
    "INSIMVLO": {"-": ["51"]},
    "INSINCERVS": {"-": ["21"]},
    "INSINVO": {"-": ["51"]},
    "INSIPIENS": {"-": ["25"]},
    "INSIPIO": {"-": ["55"]},
    "INSISTO": {"-": ["53"]},
    "INSITICIVS": {"-": ["21"]},
    "INSITIO": {"-": ["13"]},
    "INSITIVVS": {"-": ["21"]},
    "INSITOR": {"-": ["13"]},
    "INSITVS": {"1": ["14"], "2": ["21"]},
    "INSOCIABILIS": {"-": ["24"]},
    "INSOLABILITER": {"-": ["60"]},
    "INSOLENS": {"-": ["25"]},
    "INSOLENTER": {"-": ["60"]},
    "INSOLENTIA": {"-": ["11"]},
    "INSOLESCO": {"-": ["53"]},
    "INSOLITVS": {"-": ["21"]},
    "INSOLVBILIS": {"-": ["24"]},
    "INSOMNIS": {"-": ["24"]},
    "INSOMNIVM": {"-": ["12"]},
    "INSONO": {"1": ["51"]},
    "INSONS": {"-": ["25"]},
    "INSOPOR": {"-": ["25"]},
    "INSPECTIO": {"-": ["13"]},
    "INSPECTO": {"-": ["51"]},
    "INSPECTOR": {"-": ["13"]},
    "INSPECTVS": {"-": ["14"]},
    "INSPERANS": {"-": ["25"]},
    "INSPERATVS": {"-": ["21"]},
    "INSPERGO": {"-": ["53"]},
    "INSPICIO": {"-": ["55"]},
    "INSPICO": {"-": ["51"]},
    "INSPIRO": {"-": ["51"]},
    "INSPVO": {"-": ["53"]},
    "INSPVRCO": {"-": ["51"]},
    "INSTABILIS": {"-": ["24"]},
    "INSTANS": {"-": ["25"]},
    "INSTANTER": {"-": ["6&"]},
    "INSTAR": {"-": ["16"]},
    "INSTAVRO": {"-": ["51"]},
    "INSTERNO": {"-": ["53"]},
    "INSTIGO": {"-": ["51"]},
    "INSTILLO": {"-": ["51"]},
    "INSTINCTVS": {"1": ["14"], "2": ["21"]},
    "INSTINGVO": {"-": ["53"]},
    "INSTIPO": {"-": ["51"]},
    "INSTITA": {"-": ["11"]},
    "INSTITOR": {"-": ["13"]},
    "INSTITVO": {"-": ["53"]},
    "INSTITVTIO": {"-": ["13"]},
    "INSTITVTVM": {"-": ["12"]},
    "INSTO": {"-": ["51"]},
    "INSTRAGVLVM": {"-": ["12"]},
    "INSTRATVM": {"-": ["12"]},
    "INSTREPO": {"-": ["53"]},
    "INSTRINGO": {"-": ["53"]},
    "INSTRVCTIO": {"-": ["13"]},
    "INSTRVCTVS": {"1": ["14"], "2": ["21"]},
    "INSTRVMENTVM": {"-": ["12"]},
    "INSTRVO": {"-": ["53"]},
    "INSVAVIS": {"-": ["24"]},
    "INSVDO": {"-": ["51"]},
    "INSVEFACTVS": {"-": ["21"]},
    "INSVESCO": {"-": ["53"]},
    "INSVETVS": {"-": ["21"]},
    "INSVLA": {"-": ["11"], "N": ["11"]},
    "INSVLANVS": {"-": ["12"]},
    "INSVLSVS": {"-": ["21"]},
    "INSVLTO": {"-": ["51"]},
    "INSVM": {"1": ["56"]},
    "INSVMO": {"-": ["53"]},
    "INSVO": {"-": ["53"]},
    "INSVPER": {"1": ["60"], "2": ["70"]},
    "INSVPERABILIS": {"-": ["24"]},
    "INSVRGO": {"-": ["53"]},
    "INSVSVRRO": {"-": ["51"]},
    "INTABESCO": {"-": ["53"]},
    "INTACTILIS": {"-": ["24"]},
    "INTACTVS": {"1": ["14"], "2": ["21"]},
    "INTAMINATVS": {"-": ["21"]},
    "INTECTVS": {"-": ["21"]},
    "INTEGELLVS": {"-": ["21"]},
    "INTEGER": {"-": ["21"]},
    "INTEGO": {"-": ["53"]},
    "INTEGRE": {"-": ["60"]},
    "INTEGRITAS": {"-": ["13"]},
    "INTEGRO": {"-": ["51"]},
    "INTELLECTIO": {"-": ["12"]},
    "INTELLECTVALIS": {"-": ["24"]},
    "INTELLECTVS": {"-": ["14"]},
    "INTELLIGENS": {"-": ["25"]},
    "INTELLIGENTIA": {"-": ["11"]},
    "INTELLIGIBILIS": {"-": ["24"]},
    "INTELLIGO": {"-": ["53"]},
    "INTEMERATVS": {"-": ["21"]},
    "INTEMPERANS": {"-": ["25"]},
    "INTEMPERANTER": {"-": ["60"]},
    "INTEMPERANTIA": {"-": ["11"]},
    "INTEMPERATVS": {"-": ["21"]},
    "INTEMPERIAE": {"-": ["11"]},
    "INTEMPERIES": {"-": ["15"]},
    "INTEMPESTIVE": {"-": ["60"]},
    "INTEMPESTIVVS": {"-": ["21"]},
    "INTEMPESTVS": {"-": ["21"]},
    "INTENDO": {"-": ["53"]},
    "INTENTATIO": {"-": ["13"]},
    "INTENTATVS": {"-": ["21"]},
    "INTENTE": {"-": ["6&"]},
    "INTENTIO": {"-": ["13"]},
    "INTENTO": {"-": ["51"]},
    "INTENTVS": {"1": ["14"], "2": ["21"]},
    "INTEPEO": {"-": ["52"]},
    "INTEPESCO": {"-": ["53"]},
    "INTER": {"-": ["70"]},
    "INTERARESCO": {"-": ["53"]},
    "INTERCALARIVS": {"-": ["21"]},
    "INTERCALO": {"-": ["51"]},
    "INTERCAPEDO": {"-": ["13"]},
    "INTERCARDINATVS": {"-": ["21"]},
    "INTERCEDO": {"1": ["53"]},
    "INTERCESSIO": {"-": ["13"]},
    "INTERCESSOR": {"-": ["13"]},
    "INTERCESSVS": {"-": ["14"]},
    "INTERCIDO": {"1": ["53"], "2": ["53"]},
    "INTERCIPIO": {"-": ["55"]},
    "INTERCLVDO": {"-": ["53"]},
    "INTERCOLVMNIVM": {"-": ["12"]},
    "INTERCVRRO": {"-": ["53"]},
    "INTERCVRSO": {"-": ["51"]},
    "INTERCVRSVS": {"-": ["14"]},
    "INTERCVS": {"-": ["22"]},
    "INTERDATVS": {"-": ["21"]},
    "INTERDICO": {"2": ["53"]},
    "INTERDICTVM": {"-": ["12"]},
    "INTERDIV": {"-": ["60"]},
    "INTERDO": {"-": ["51"]},
    "INTERDVM": {"-": ["60"]},
    "INTEREA": {"-": ["60"]},
    "INTEREMPTOR": {"-": ["13"]},
    "INTEREO": {"1": ["56"]},
    "INTEREQVITO": {"-": ["51"]},
    "INTERFECTOR": {"-": ["13"]},
    "INTERFECTRIX": {"-": ["13"]},
    "INTERFICIO": {"-": ["55"]},
    "INTERFIO": {"-": ["56"]},
    "INTERFLVO": {"-": ["53"]},
    "INTERFODIO": {"-": ["55"]},
    "INTERFOR": {"-": ["5J"]},
    "INTERFRINGO": {"-": ["53"]},
    "INTERFVGIO": {"-": ["55"]},
    "INTERFVNDO": {"2": ["53"]},
    "INTERFVSVS": {"-": ["21"]},
    "INTERIACEO": {"-": ["52"]},
    "INTERIACIO": {"-": ["55"]},
    "INTERICIO": {"-": ["55"]},
    "INTERIECTVS": {"-": ["14"]},
    "INTERIM": {"-": ["60"]},
    "INTERIMO": {"-": ["53"]},
    "INTERIOR": {"-": ["2&"], "2": ["2&"]},
    "INTERIORA": {"-": ["13"]},
    "INTERITIO": {"-": ["13"]},
    "INTERITVS": {"-": ["14"]},
    "INTERIVNGO": {"-": ["53"]},
    "INTERLABOR": {"2": ["5L"]},
    "INTERLEGO": {"2": ["53"]},
    "INTERLIGO": {"2": ["51"]},
    "INTERLINO": {"-": ["53"]},
    "INTERLOQVOR": {"-": ["5L"]},
    "INTERLVCEO": {"-": ["52"]},
    "INTERLVNIVM": {"-": ["12"]},
    "INTERLVO": {"-": ["53"]},
    "INTERMESTRIS": {"-": ["21"]},
    "INTERMINO": {"-": ["51"]},
    "INTERMISCEO": {"-": ["52"]},
    "INTERMISSIO": {"-": ["13"]},
    "INTERMITTO": {"-": ["53"]},
    "INTERMORIOR": {"-": ["5N"]},
    "INTERNASCOR": {"-": ["5L"]},
    "INTERNECIO": {"-": ["13"]},
    "INTERNECIVVS": {"-": ["21"]},
    "INTERNITEO": {"-": ["52"]},
    "INTERNODIVM": {"-": ["12"]},
    "INTERNOSCO": {"-": ["53"]},
    "INTERNVNTIA": {"-": ["11"]},
    "INTERNVNTIVS": {"-": ["12"]},
    "INTERNVS": {"-": ["21"]},
    "INTERO": {"-": ["53"]},
    "INTERPELLATIO": {"-": ["13"]},
    "INTERPELLATOR": {"-": ["13"]},
    "INTERPELLO": {"-": ["51"]},
    "INTERPENSIVA": {"-": ["12"]},
    "INTERPOLO": {"-": ["51"]},
    "INTERPONO": {"-": ["53"]},
    "INTERPOSITIO": {"-": ["13"]},
    "INTERPRES": {"-": ["13"]},
    "INTERPRETATIO": {"-": ["13"]},
    "INTERPRETOR": {"-": ["5J"]},
    "INTERPRIMO": {"-": ["53"]},
    "INTERPVNGO": {"-": ["53"]},
    "INTERPVTO": {"-": ["51"]},
    "INTERQVIESCO": {"-": ["53"]},
    "INTERRITVS": {"-": ["21"]},
    "INTERROGATIO": {"-": ["13"]},
    "INTERROGATIVNCVLA": {"-": ["11"]},
    "INTERROGO": {"-": ["51"]},
    "INTERRVMPO": {"-": ["53"]},
    "INTERSAEPIO": {"-": ["54"]},
    "INTERSCALMIVM": {"-": ["12"]},
    "INTERSCINDO": {"-": ["53"]},
    "INTERSECO": {"-": ["51"]},
    "INTERSECTIO": {"-": ["13"]},
    "INTERSERO": {"2": ["53"], "3": ["53"]},
    "INTERSPIRO": {"-": ["51"]},
    "INTERSTINCTVS": {"-": ["21"]},
    "INTERSTINGVO": {"-": ["53"]},
    "INTERSVM": {"1": ["56"]},
    "INTERTIGNIVM": {"-": ["12"]},
    "INTERTRIGO": {"-": ["13"]},
    "INTERTRIMENTVM": {"-": ["12"]},
    "INTERVALLVM": {"-": ["12"]},
    "INTERVECTVS": {"-": ["21"]},
    "INTERVELLO": {"-": ["53"]},
    "INTERVENIO": {"-": ["54"]},
    "INTERVENIVM": {"-": ["12"]},
    "INTERVENTVS": {"-": ["14"]},
    "INTERVERSOR": {"1": ["13"], "2": ["5J"], "-": ["5J"]},
    "INTERVERTO": {"-": ["53"]},
    "INTERVISO": {"-": ["53"]},
    "INTERVOMO": {"-": ["53"]},
    "INTERVTRASQVE": {"-": ["60"]},
    "INTESTABILIS": {"-": ["24"]},
    "INTESTATVS": {"-": ["21"]},
    "INTESTINVM": {"-": ["12"]},
    "INTESTINVS": {"-": ["21"]},
    "INTEXO": {"-": ["53"]},
    "INTEXTVS": {"-": ["14"]},
    "INTIME": {"-": ["6-"]},
    "INTIMVS": {"-": ["12"]},
    "INTINGO": {"-": ["53"]},
    "INTOLERABILIS": {"-": ["24"]},
    "INTOLERANDVS": {"-": ["21"]},
    "INTOLERANS": {"-": ["25"]},
    "INTOLERANTER": {"-": ["6&"]},
    "INTOLERANTIA": {"-": ["11"]},
    "INTONO": {"-": ["51"]},
    "INTONSVS": {"-": ["21"]},
    "INTORQVEO": {"-": ["52"]},
    "INTRA": {"1": ["60"], "2": ["70"]},
    "INTRACTABILIS": {"-": ["24"]},
    "INTRACTATVS": {"-": ["21"]},
    "INTREMO": {"-": ["53"]},
    "INTREPIDE": {"-": ["60"]},
    "INTREPIDVS": {"-": ["21"]},
    "INTRINSECVS": {"-": ["60"]},
    "INTRITVS": {"1": ["21"], "2": ["21"]},
    "INTRO": {"1": ["51"], "2": ["60"]},
    "INTROCEDO": {"1": ["53"]},
    "INTRODVCO": {"-": ["53"]},
    "INTRODVCTIO": {"-": ["13"]},
    "INTROEO": {"1": ["56"]},
    "INTROFERO": {"-": ["5F", "53"]},
    "INTROGREDIOR": {"-": ["5N"]},
    "INTROITVS": {"-": ["14"]},
    "INTROMITTO": {"-": ["53"]},
    "INTRORSVM": {"-": ["60"]},
    "INTRORVMPO": {"-": ["53"]},
    "INTROSPECTO": {"-": ["51"]},
    "INTROSPICIO": {"-": ["55"]},
    "INTROTRVDO": {"-": ["53"]},
    "INTROVERSVS": {"-": ["60"]},
    "INTVBVM": {"-": ["12"]},
    "INTVBVS": {"-": ["12"]},
    "INTVEOR": {"-": ["5K"]},
    "INTVMESCO": {"-": ["53"]},
    "INTVMVLATVS": {"-": ["21"]},
    "INTVOR": {"-": ["5L"]},
    "INTVRBIDVS": {"-": ["21"]},
    "INTVRBO": {"2": ["51"]},
    "INTVS": {"1": ["60"], "2": ["70"]},
    "INTVTVS": {"-": ["21"]},
    "INVADO": {"2": ["53"]},
    "INVALESCO": {"-": ["53"]},
    "INVALIDVS": {"-": ["21"]},
    "INVASIO": {"-": ["13"]},
    "INVECTICIVS": {"-": ["21"]},
    "INVECTOR": {"-": ["13"]},
    "INVECTVS": {"-": ["14"]},
    "INVEHO": {"-": ["53"]},
    "INVENIO": {"-": ["54"]},
    "INVENTIO": {"-": ["13"]},
    "INVENTOR": {"-": ["13"]},
    "INVENTRIX": {"-": ["13"]},
    "INVENTVM": {"-": ["12"]},
    "INVENTVS": {"-": ["14"]},
    "INVENVSTVS": {"-": ["21"]},
    "INVERECVNDE": {"-": ["60"]},
    "INVERECVNDVS": {"-": ["21"]},
    "INVERGO": {"-": ["53"]},
    "INVERSVRA": {"-": ["11"]},
    "INVERSVS": {"1": ["21"]},
    "INVERTO": {"-": ["53"]},
    "INVESPERASCO": {"-": ["53"]},
    "INVESTIGO": {"-": ["51"]},
    "INVESTIO": {"-": ["54"]},
    "INVETERASCO": {"-": ["53"]},
    "INVETERATVM": {"-": ["12"]},
    "INVETERATVS": {"-": ["21"]},
    "INVIA": {"-": ["12"]},
    "INVICEM": {"-": ["60"]},
    "INVICTVS": {"-": ["21"], "A": ["21"]},
    "INVIDEO": {"-": ["52"]},
    "INVIDIA": {"-": ["11"]},
    "INVIDIOSE": {"-": ["60"]},
    "INVIDIOSVS": {"-": ["21"]},
    "INVIDVS": {"1": ["12"], "2": ["21"]},
    "INVIGILO": {"-": ["51"]},
    "INVINCTVS": {"-": ["21"]},
    "INVIOLABILIS": {"-": ["24"]},
    "INVIOLATVS": {"-": ["21"]},
    "INVISITATVS": {"-": ["21"]},
    "INVISO": {"-": ["53"]},
    "INVISVS": {"1": ["21"], "2": ["21"]},
    "INVITAMENTVM": {"-": ["12"]},
    "INVITATIO": {"-": ["13"]},
    "INVITE": {"-": ["60"]},
    "INVITO": {"-": ["51"]},
    "INVITVS": {"-": ["21"]},
    "INVIVS": {"-": ["21"]},
    "INVLA": {"-": ["11"]},
    "INVLTVS": {"-": ["21"]},
    "INVMBRO": {"-": ["51"]},
    "INVNDATIO": {"-": ["13"]},
    "INVNDO": {"-": ["51"]},
    "INVNGO": {"-": ["53"]},
    "INVOCATVS": {"1": ["14"], "2": ["21"]},
    "INVOCO": {"-": ["51"]},
    "INVOLITO": {"-": ["51"]},
    "INVOLO": {"2": ["51"]},
    "INVOLVCRVM": {"-": ["12"]},
    "INVOLVO": {"-": ["53"]},
    "INVOLVTIO": {"-": ["13"]},
    "INVOLVTVS": {"-": ["21"]},
    "INVRBANE": {"-": ["60"]},
    "INVRGEO": {"-": ["52"]},
    "INVRO": {"-": ["53"]},
    "INVSITATE": {"-": ["60"]},
    "INVSITATVS": {"-": ["21"]},
    "INVTILIS": {"-": ["24"]},
    "INVTILITAS": {"-": ["13"]},
    "INVTILITER": {"-": ["60"]},
    "INVVLNERABILIS": {"-": ["24"]},
    "INVVS": {"N": ["12"]},
    "IO": {"-": ["90"], "N": ["13"]},
    "IOCATIO": {"-": ["13"]},
    "IOCOR": {"-": ["5J"]},
    "IOCOSE": {"-": ["60"]},
    "IOCOSVS": {"-": ["21"]},
    "IOCVLARIS": {"-": ["24"]},
    "IOCVS": {"-": ["12"], "N": ["12"]},
    "IONIA": {"N": ["11"]},
    "IONICE": {"-": ["60"]},
    "IPSE": {"-": ["45"]},
    "IRA": {"-": ["11"], "N": ["11"]},
    "IRACVNDIA": {"-": ["11"]},
    "IRACVNDVS": {"-": ["21"]},
    "IRASCOR": {"-": ["5L"]},
    "IRATVS": {"-": ["21"]},
    "IRIS": {"-": ["17"], "N": ["13"]},
    "IRONIA": {"-": ["11"]},
    "IRPEX": {"-": ["13"]},
    "IRRADO": {"-": ["53"]},
    "IRRATIONALIS": {"-": ["24"]},
    "IRRELIGATVS": {"-": ["21"]},
    "IRRELIGIOSE": {"-": ["60"]},
    "IRREMEABILIS": {"-": ["24"]},
    "IRREMEDIABILIS": {"-": ["24"]},
    "IRREPARABILIS": {"-": ["24"]},
    "IRREPERTVS": {"-": ["21"]},
    "IRREPO": {"-": ["53"]},
    "IRREQVIETVS": {"-": ["21"]},
    "IRRESECTVS": {"-": ["21"]},
    "IRRETIO": {"-": ["54"]},
    "IRRETORTVS": {"-": ["21"]},
    "IRREVERENTIA": {"-": ["11"]},
    "IRREVOCABILIS": {"-": ["24"]},
    "IRREVOCATVS": {"-": ["21"]},
    "IRRIDEO": {"-": ["52"]},
    "IRRIDICVLE": {"-": ["60"]},
    "IRRIGIVVS": {"-": ["21"]},
    "IRRIGO": {"-": ["51"]},
    "IRRIGVVS": {"-": ["21"]},
    "IRRISIO": {"-": ["13"]},
    "IRRISOR": {"-": ["13"]},
    "IRRISVS": {"-": ["14"]},
    "IRRITABILIS": {"-": ["24"]},
    "IRRITAMENTVM": {"-": ["12"]},
    "IRRITATIO": {"-": ["13"]},
    "IRRITATOR": {"-": ["13"]},
    "IRRITATVS": {"-": ["21"]},
    "IRRITO": {"-": ["51"]},
    "IRRITVM": {"-": ["12"]},
    "IRRITVS": {"-": ["21"]},
    "IRROGATIO": {"-": ["13"]},
    "IRROGO": {"-": ["51"]},
    "IRRORO": {"-": ["51"]},
    "IRRVMATIO": {"-": ["13"]},
    "IRRVMATOR": {"-": ["13"]},
    "IRRVMO": {"-": ["51"]},
    "IRRVMPO": {"-": ["53"]},
    "IRRVO": {"-": ["53"]},
    "IRRVPTIO": {"-": ["13"]},
    "IRRVPTVS": {"-": ["21"]},
    "IS": {"-": ["45"]},
    "ISATIS": {"-": ["17"]},
    "ISCHIACVS": {"-": ["21"]},
    "ISCHNOS": {"-": ["27"]},
    "ISIDORVS": {"N": ["12"]},
    "ISMENIAS": {"N": ["17"]},
    "ISODOMVS": {"-": ["21"]},
    "ISTAC": {"-": ["60"]},
    "ISTE": {"-": ["45"]},
    "ISTHMVS": {"-": ["17"], "N": ["12"]},
    "ISTI": {"-": ["60"]},
    "ISTIC": {"1": ["45"], "2": ["60"]},
    "ISTINC": {"-": ["60"]},
    "ISTIVSMODI": {"-": ["60"]},
    "ISTO": {"-": ["60"]},
    "ISTOC": {"-": ["60"]},
    "ISTVC": {"-": ["60"]},
    "ITA": {"-": ["60"]},
    "ITALIA": {"N": ["11"]},
    "ITALICVS": {"N": ["21"], "A": ["21"]},
    "ITALVS": {"A": ["21"], "N": ["12"]},
    "ITAQVE": {"-": ["60"]},
    "ITEM": {"-": ["60"]},
    "ITER": {"-": ["13"]},
    "ITERATIO": {"-": ["13"]},
    "ITERO": {"-": ["51"]},
    "ITERVM": {"-": ["60"]},
    "ITIDEM": {"-": ["60"]},
    "ITVS": {"-": ["14"]},
    "IVBA": {"-": ["11"], "N": ["11"]},
    "IVBAR": {"-": ["13"]},
    "IVBATVS": {"-": ["21"]},
    "IVBEO": {"-": ["52"]},
    "IVCVNDE": {"-": ["60"]},
    "IVCVNDITAS": {"-": ["13"]},
    "IVCVNDVS": {"-": ["21"]},
    "IVDEX": {"-": ["13"]},
    "IVDICATVM": {"-": ["12"]},
    "IVDICATVS": {"1": ["14"], "2": ["21"]},
    "IVDICIALIS": {"-": ["24"]},
    "IVDICIARIVS": {"-": ["21"]},
    "IVDICIVM": {"-": ["12"]},
    "IVDICO": {"-": ["51"]},
    "IVGALES": {"-": ["13"]},
    "IVGALIS": {"-": ["24"]},
    "IVGERVM": {"-": ["16"]},
    "IVGIS": {"-": ["24"]},
    "IVGLANS": {"-": ["13"]},
    "IVGO": {"-": ["51"]},
    "IVGOSVS": {"-": ["21"]},
    "IVGVLATIO": {"-": ["13"]},
    "IVGVLO": {"-": ["51"]},
    "IVGVLVM": {"-": ["12"]},
    "IVGVLVS": {"-": ["12"]},
    "IVGVM": {"-": ["12"]},
    "IVGVMENTO": {"-": ["51"]},
    "IVGVMENTVM": {"-": ["12"]},
    "IVGVS": {"-": ["21"]},
    "IVLIA": {"N": ["11"]},
    "IVLIVS": {"A": ["21"], "N": ["12"]},
    "IVLVS": {"N": ["12"]},
    "IVMENTVM": {"-": ["12"]},
    "IVNCTVRA": {"-": ["11"]},
    "IVNCTVS": {"-": ["21"]},
    "IVNCVS": {"-": ["12"], "N": ["12"]},
    "IVNGO": {"-": ["53"]},
    "IVNIANVS": {"N": ["21"], "A": ["21"]},
    "IVNIO": {"N": ["13"]},
    "IVNIOR": {"-": ["13"], "N": ["13"]},
    "IVNIPERVS": {"-": ["12"]},
    "IVNIVS": {"A": ["21"], "N": ["12"]},
    "IVNIX": {"-": ["13"]},
    "IVNO": {"N": ["13"]},
    "IVPPITER": {"N": ["13"]},
    "IVRAMENTVM": {"-": ["12"]},
    "IVRATVS": {"-": ["21"]},
    "IVRE": {"-": ["60"]},
    "IVRGIVM": {"-": ["12"]},
    "IVRGO": {"-": ["51"]},
    "IVRIDICVS": {"1": ["12"], "2": ["21"]},
    "IVRISCONSVLTVS": {"-": ["12"]},
    "IVRISDICTIO": {"-": ["13"]},
    "IVRO": {"-": ["51"]},
    "IVROR": {"-": ["5J"]},
    "IVS": {"1": ["13"], "2": ["13"]},
    "IVSCVLVM": {"-": ["12"]},
    "IVSIVRANDVM": {"-": ["13"]},
    "IVSSVM": {"-": ["12"]},
    "IVSSVS": {"-": ["14"]},
    "IVSTA": {"C": ["12"], "-": ["12"]},
    "IVSTE": {"-": ["60"]},
    "IVSTIFICVS": {"-": ["21"]},
    "IVSTITIA": {"-": ["11"], "N": ["11"]},
    "IVSTITIVM": {"-": ["12"]},
    "IVSTVM": {"-": ["12"]},
    "IVSTVS": {"-": ["21"], "N": ["12"]},
    "IVTO": {"-": ["51"]},
    "IVVAMEN": {"-": ["13"]},
    "IVVENALIS": {"-": ["24"], "N": ["13"], "A": ["21"]},
    "IVVENALITER": {"-": ["60"]},
    "IVVENCA": {"-": ["11"]},
    "IVVENCVS": {"1": ["12"], "2": ["21"]},
    "IVVENESCO": {"-": ["53"]},
    "IVVENILIS": {"-": ["24"]},
    "IVVENILITER": {"-": ["60"]},
    "IVVENIS": {"1": ["13"], "2": ["26"]},
    "IVVENTA": {"-": ["11"]},
    "IVVENTAS": {"-": ["13"], "N": ["13"]},
    "IVVENTVS": {"-": ["13"]},
    "IVVO": {"-": ["51"]},
    "IVXTA": {"1": ["60"], "2": ["70"]},
    "IVXTIM": {"-": ["60"]},
    "KALENDAE": {"-": ["11"]},
    "KALENDARIVM": {"-": ["12"]},
    "LABASCO": {"-": ["53"]},
    "LABASCOR": {"-": ["5J"]},
    "LABEFACIO": {"-": ["55"]},
    "LABEFACTO": {"-": ["51"]},
    "LABEFIO": {"-": ["56"]},
    "LABELLVM": {"1": ["12"], "2": ["12"]},
    "LABEOSVS": {"-": ["21"]},
    "LABES": {"-": ["13"]},
    "LABIA": {"-": ["11"]},
    "LABICI": {"N": ["12"]},
    "LABIDVS": {"-": ["21"]},
    "LABO": {"-": ["51"]},
    "LABOR": {"1": ["13"], "2": ["5L"], "N": ["13"]},
    "LABORIOSE": {"-": ["60"]},
    "LABORIOSVS": {"-": ["21"]},
    "LABORO": {"-": ["51"]},
    "LABRVM": {"1": ["12"], "2": ["12"]},
    "LABRVSCA": {"-": ["11"]},
    "LABYRINTHEVS": {"-": ["21"]},
    "LABYRINTHVS": {"-": ["12"]},
    "LAC": {"-": ["13"]},
    "LACEDAEMON": {"N": ["17"]},
    "LACEDAEMONII": {"N": ["12"]},
    "LACEDAEMONIVS": {"N": ["21"], "A": ["21"]},
    "LACER": {"-": ["21"]},
    "LACERATIO": {"-": ["13"]},
    "LACERNA": {"-": ["11"]},
    "LACERNATVS": {"-": ["21"]},
    "LACERO": {"-": ["51"]},
    "LACERTA": {"-": ["11"], "N": ["11"]},
    "LACERTOSVS": {"-": ["21"]},
    "LACERTVS": {"1": ["12"], "2": ["12"]},
    "LACESSO": {"-": ["53"]},
    "LACRIMA": {"-": ["11"]},
    "LACRIMABILIS": {"-": ["24"]},
    "LACRIMO": {"-": ["51"]},
    "LACRIMOSVS": {"-": ["21"]},
    "LACRIMVLA": {"-": ["11"]},
    "LACTEO": {"-": ["52"]},
    "LACTEOLVS": {"-": ["21"]},
    "LACTES": {"-": ["13"]},
    "LACTEVS": {"-": ["21"]},
    "LACTO": {"-": ["51"]},
    "LACTVCA": {"-": ["11"]},
    "LACVNA": {"-": ["11"]},
    "LACVNAR": {"-": ["13"]},
    "LACVNOSVS": {"-": ["21"]},
    "LACVS": {"-": ["14"]},
    "LAEDO": {"-": ["53"]},
    "LAELIVS": {"N": ["12"]},
    "LAENA": {"-": ["11"]},
    "LAESIO": {"-": ["13"]},
    "LAETABILIS": {"-": ["24"]},
    "LAETANS": {"-": ["25"]},
    "LAETE": {"-": ["6&"]},
    "LAETIFICO": {"-": ["51"]},
    "LAETIFICVS": {"-": ["21"]},
    "LAETILIVS": {"N": ["12"]},
    "LAETITIA": {"-": ["11"], "N": ["11"]},
    "LAETOR": {"-": ["5J"]},
    "LAETORIVS": {"N": ["12"]},
    "LAETVS": {"1": ["12"], "2": ["21"]},
    "LAEVA": {"1": ["11"], "2": ["12"], "3": ["60"]},
    "LAEVE": {"-": ["60"]},
    "LAEVINVS": {"N": ["12"]},
    "LAEVVM": {"1": ["12"], "2": ["60"], "-": ["60"]},
    "LAGANVM": {"-": ["12"]},
    "LAGENA": {"-": ["11"]},
    "LAGEVS": {"-": ["17"]},
    "LAGOIS": {"-": ["13"]},
    "LALLO": {"-": ["51"]},
    "LAMA": {"-": ["11"]},
    "LAMBO": {"-": ["53"]},
    "LAMELLA": {"-": ["11"]},
    "LAMENTA": {"-": ["12"]},
    "LAMENTABILIS": {"-": ["24"]},
    "LAMENTATIO": {"-": ["13"]},
    "LAMENTOR": {"-": ["5J"]},
    "LAMENTVM": {"-": ["12"]},
    "LAMIA": {"N": ["11"]},
    "LAMINA": {"-": ["11"]},
    "LAMPAS": {"-": ["17"]},
    "LAMPSACENI": {"N": ["12"]},
    "LAMPSACENVS": {"N": ["21"], "A": ["21"]},
    "LAMPSACVS": {"N": ["17"]},
    "LANA": {"-": ["11"]},
    "LANATVS": {"-": ["21"]},
    "LANCEA": {"-": ["11"]},
    "LANCINATIO": {"-": ["13"]},
    "LANCINO": {"-": ["51"]},
    "LANCVLA": {"-": ["11"]},
    "LANEVS": {"-": ["21"]},
    "LANGOR": {"-": ["13"]},
    "LANGVENS": {"-": ["25"]},
    "LANGVEO": {"-": ["52"]},
    "LANGVESCO": {"-": ["53"]},
    "LANGVIDE": {"-": ["60"]},
    "LANGVIDVLVS": {"-": ["21"]},
    "LANGVIDVS": {"-": ["21"]},
    "LANGVOR": {"-": ["13"]},
    "LANIATVS": {"-": ["14"]},
    "LANIFICVS": {"-": ["21"]},
    "LANIGER": {"1": ["12"], "2": ["21"]},
    "LANIGERA": {"-": ["11"]},
    "LANIO": {"-": ["51"], "2": ["13"], "1": ["21"]},
    "LANISTA": {"-": ["11"]},
    "LANITIVM": {"-": ["12"]},
    "LANVGO": {"-": ["13"]},
    "LANX": {"-": ["13"]},
    "LAODICIA": {"N": ["11"]},
    "LAPATHVM": {"-": ["12"]},
    "LAPATHVS": {"-": ["12"]},
    "LAPHIRO": {"N": ["13"]},
    "LAPICIDINAE": {"-": ["11"]},
    "LAPIDATIO": {"-": ["13"]},
    "LAPIDEVS": {"-": ["21"]},
    "LAPIDO": {"-": ["51"]},
    "LAPIDOSA": {"-": ["21"]},
    "LAPIDOSVS": {"-": ["21"]},
    "LAPILLVS": {"-": ["12"]},
    "LAPIS": {"-": ["13"]},
    "LAPPA": {"-": ["11"], "N": ["11"]},
    "LAPSO": {"-": ["51"]},
    "LAPSVS": {"-": ["14"]},
    "LAQVEAR": {"-": ["13"]},
    "LAQVEO": {"1": ["51"], "2": ["51"]},
    "LAQVEVS": {"-": ["12"]},
    "LAR": {"N": ["13"]},
    "LARCIVS": {"N": ["12"]},
    "LARDVM": {"-": ["12"]},
    "LARENTIA": {"N": ["11"]},
    "LARGE": {"-": ["60"]},
    "LARGIFICVS": {"-": ["21"]},
    "LARGIFLVVS": {"-": ["21"]},
    "LARGIOR": {"-": ["5M"]},
    "LARGITAS": {"-": ["13"]},
    "LARGITER": {"-": ["60"]},
    "LARGITIO": {"-": ["13"]},
    "LARGITOR": {"1": ["13"], "2": ["5J"], "-": ["5J"]},
    "LARGITVDO": {"-": ["13"]},
    "LARGIVS": {"N": ["12"]},
    "LARGO": {"-": ["60"]},
    "LARGVS": {"-": ["21"], "N": ["12"]},
    "LARIDVM": {"-": ["12"]},
    "LARIX": {"-": ["13"]},
    "LARVA": {"-": ["11"], "N": ["11"]},
    "LARVALIS": {"-": ["24"]},
    "LASANVM": {"-": ["12"]},
    "LASCIVE": {"-": ["60"]},
    "LASCIVIA": {"-": ["11"]},
    "LASCIVIO": {"-": ["54"]},
    "LASCIVVS": {"-": ["21"]},
    "LASER": {"-": ["13"]},
    "LASERPICIATVS": {"-": ["21"]},
    "LASERPICIVM": {"-": ["12"]},
    "LASERPITIFER": {"-": ["21"]},
    "LASSITVDO": {"-": ["13"]},
    "LASSO": {"-": ["51"]},
    "LASSVLVS": {"-": ["21"]},
    "LASSVS": {"-": ["21"]},
    "LATE": {"-": ["60"]},
    "LATEBRA": {"-": ["11"]},
    "LATEBROSVS": {"-": ["21"]},
    "LATENS": {"-": ["25"]},
    "LATEO": {"-": ["52"]},
    "LATER": {"-": ["13"]},
    "LATERAMEN": {"-": ["13"]},
    "LATERARIVS": {"-": ["21"]},
    "LATERCVLVS": {"-": ["12"]},
    "LATERICIVS": {"-": ["21"]},
    "LATERNA": {"-": ["11"]},
    "LATEX": {"-": ["13"]},
    "LATIBVLVM": {"-": ["12"]},
    "LATIFVNDIVM": {"-": ["12"]},
    "LATINE": {"-": ["60"]},
    "LATINI": {"N": ["12"]},
    "LATINIVS": {"N": ["12"]},
    "LATINVS": {"A": ["21"], "N": ["12"]},
    "LATIO": {"-": ["13"]},
    "LATITO": {"1": ["51"], "2": ["51"]},
    "LATITVDO": {"-": ["13"]},
    "LATIVM": {"N": ["12"]},
    "LATONA": {"N": ["11"]},
    "LATOR": {"-": ["13"]},
    "LATRATVS": {"-": ["14"]},
    "LATRO": {"1": ["13"], "2": ["51"]},
    "LATROCINIVM": {"-": ["12"]},
    "LATROCINOR": {"-": ["5J"]},
    "LATRVNCVLARIVS": {"-": ["21"]},
    "LATRVNCVLVS": {"-": ["12"]},
    "LATVS": {"1": ["13"], "2": ["21"]},
    "LATVSCVLVM": {"-": ["12"]},
    "LAVABRVM": {"-": ["12"]},
    "LAVACRVM": {"-": ["12"]},
    "LAVATIO": {"-": ["13"]},
    "LAVDABILIS": {"-": ["24"]},
    "LAVDABILITER": {"-": ["60"]},
    "LAVDATIO": {"-": ["13"]},
    "LAVDATOR": {"-": ["13"]},
    "LAVDATRIX": {"-": ["13"]},
    "LAVDATVS": {"-": ["21"]},
    "LAVDO": {"-": ["51"]},
    "LAVINIA": {"N": ["11"]},
    "LAVINIVM": {"N": ["12"]},
    "LAVO": {"1": ["51"], "2": ["53"]},
    "LAVREA": {"-": ["11"]},
    "LAVREATVS": {"-": ["21"]},
    "LAVRENS": {"N": ["25"], "A": ["25"]},
    "LAVRENTES": {"N": ["13"]},
    "LAVRENTINVS": {"N": ["21"], "A": ["21"]},
    "LAVREVS": {"-": ["21"]},
    "LAVRICOMVS": {"-": ["21"]},
    "LAVRIGER": {"-": ["21"]},
    "LAVRVS": {"-": ["16"]},
    "LAVS": {"-": ["13"]},
    "LAVTE": {"-": ["60"]},
    "LAVTITIA": {"-": ["11"]},
    "LAVTVMIAE": {"-": ["11"]},
    "LAVTVS": {"C": ["21"], "-": ["21"]},
    "LAXAMENTVM": {"-": ["12"]},
    "LAXATIO": {"-": ["13"]},
    "LAXATVS": {"-": ["21"]},
    "LAXE": {"-": ["60"]},
    "LAXITAS": {"-": ["13"]},
    "LAXO": {"-": ["51"]},
    "LAXVS": {"-": ["21"]},
    "LEA": {"-": ["11"]},
    "LEAENA": {"-": ["11"], "N": ["11"]},
    "LEBES": {"-": ["17"]},
    "LECTICA": {"-": ["11"]},
    "LECTICARIVS": {"-": ["12"]},
    "LECTICVLVS": {"-": ["12"]},
    "LECTIO": {"-": ["13"]},
    "LECTISTERNIVM": {"-": ["12"]},
    "LECTITO": {"-": ["51"]},
    "LECTO": {"-": ["51"]},
    "LECTOR": {"-": ["13"]},
    "LECTVLVS": {"-": ["12"]},
    "LECTVS": {"1": ["12"], "2": ["14"], "3": ["21"]},
    "LEGALIS": {"-": ["24"]},
    "LEGATIO": {"-": ["13"]},
    "LEGATVM": {"-": ["12"]},
    "LEGATVS": {"-": ["12"]},
    "LEGENS": {"-": ["13"]},
    "LEGIFER": {"-": ["21"]},
    "LEGIO": {"-": ["13"]},
    "LEGIONARIVS": {"-": ["21"]},
    "LEGITIME": {"-": ["60"]},
    "LEGITIMVS": {"-": ["21"]},
    "LEGO": {"1": ["51"], "2": ["53"]},
    "LEGVLVS": {"-": ["12"]},
    "LEGVMEN": {"-": ["13"]},
    "LEMBVS": {"-": ["12"]},
    "LEMVRES": {"-": ["13"]},
    "LENA": {"-": ["11"]},
    "LENE": {"1": ["13"], "2": ["60"]},
    "LENIMEN": {"-": ["13"]},
    "LENIO": {"-": ["54"]},
    "LENIS": {"1": ["13"], "2": ["24"]},
    "LENITAS": {"-": ["13"]},
    "LENITER": {"-": ["60"]},
    "LENITVDO": {"-": ["13"]},
    "LENO": {"1": ["13"], "2": ["51"]},
    "LENOCINIVM": {"-": ["12"]},
    "LENOCINOR": {"-": ["5J"]},
    "LENS": {"1": ["13"], "2": ["13"]},
    "LENTE": {"-": ["60"]},
    "LENTESCO": {"-": ["53"]},
    "LENTISCVS": {"-": ["12"]},
    "LENTITVDO": {"-": ["13"]},
    "LENTO": {"-": ["51"], "N": ["11"]},
    "LENTVLVS": {"N": ["12"]},
    "LENTVS": {"-": ["21"]},
    "LENVNCVLVS": {"-": ["12"]},
    "LEO": {"-": ["13"], "N": ["17"]},
    "LEONIDAS": {"N": ["17"]},
    "LEONINVS": {"-": ["21"]},
    "LEONTINI": {"1N": ["12"], "2N": ["12"], "N": ["12"]},
    "LEONTINVS": {"N": ["21"], "A": ["21"]},
    "LEPIDE": {"-": ["60"]},
    "LEPIDVS": {"-": ["21"], "N": ["12"]},
    "LEPOR": {"-": ["13"]},
    "LEPTIS": {"N": ["13"]},
    "LEPVS": {"-": ["13"]},
    "LEPVSCVLVS": {"-": ["12"]},
    "LETALIS": {"-": ["24"]},
    "LETALITER": {"-": ["60"]},
    "LETHARGICVS": {"-": ["12"]},
    "LETHARGVS": {"-": ["12"]},
    "LETIFER": {"-": ["21"]},
    "LETIFICVS": {"-": ["21"]},
    "LETO": {"-": ["51"]},
    "LETVM": {"-": ["12"], "N": ["12"]},
    "LEVAMEN": {"-": ["13"]},
    "LEVAMENTVM": {"-": ["12"]},
    "LEVATIO": {"-": ["13"]},
    "LEVATVS": {"-": ["21"]},
    "LEVCOPHAEVS": {"-": ["21"]},
    "LEVIGATIO": {"-": ["13"]},
    "LEVIGO": {"1": ["51"], "2": ["51"]},
    "LEVIS": {"1": ["24"], "2": ["24"]},
    "LEVISOMNVS": {"2": ["21"], "-": ["21"]},
    "LEVITAS": {"1": ["13"], "2": ["13"]},
    "LEVITER": {"-": ["60"]},
    "LEVO": {"1": ["51"], "2": ["51"]},
    "LEVOR": {"-": ["13"]},
    "LEX": {"-": ["13"]},
    "LIACVLVM": {"-": ["12"]},
    "LIBAMEN": {"-": ["13"]},
    "LIBAMENTVM": {"-": ["12"]},
    "LIBARIVS": {"-": ["12"]},
    "LIBELLA": {"-": ["11"]},
    "LIBELLVS": {"-": ["12"]},
    "LIBENS": {"-": ["25"]},
    "LIBENTER": {"-": ["60"]},
    "LIBER": {"1": ["12"], "2": ["21"], "N": ["12"]},
    "LIBERA": {"N": ["11"]},
    "LIBERALIS": {"-": ["24"], "N": ["13"]},
    "LIBERALITAS": {"-": ["13"]},
    "LIBERALITER": {"-": ["60"]},
    "LIBERATIO": {"-": ["13"]},
    "LIBERATOR": {"-": ["13"]},
    "LIBERE": {"-": ["60"]},
    "LIBERI": {"-": ["12"]},
    "LIBERO": {"-": ["51"]},
    "LIBERTA": {"-": ["11"]},
    "LIBERTAS": {"-": ["13"], "N": ["13"]},
    "LIBERTINA": {"-": ["11"]},
    "LIBERTINVS": {"1": ["12"], "2": ["21"]},
    "LIBERTVS": {"-": ["12"]},
    "LIBET": {"-": ["52"]},
    "LIBIDINOSE": {"-": ["60"]},
    "LIBIDINOSVS": {"-": ["21"]},
    "LIBIDO": {"-": ["13"], "N": ["13"]},
    "LIBITA": {"-": ["12"]},
    "LIBITINARIVS": {"-": ["12"]},
    "LIBO": {"-": ["51"], "N": ["13"]},
    "LIBRA": {"-": ["11"], "N": ["11"]},
    "LIBRAMENTVM": {"-": ["12"]},
    "LIBRARIA": {"-": ["11"]},
    "LIBRARIVM": {"-": ["12"]},
    "LIBRARIVS": {"1": ["12"], "2": ["12"], "3": ["21"], "4": ["21"]},
    "LIBRATIO": {"-": ["13"]},
    "LIBRATOR": {"-": ["13"]},
    "LIBRILE": {"-": ["13"]},
    "LIBRO": {"-": ["51"]},
    "LIBVM": {"-": ["12"]},
    "LIBVRNA": {"-": ["11"]},
    "LIBVRNICA": {"-": ["11"], "N": ["11"]},
    "LICENS": {"-": ["25"]},
    "LICENTER": {"-": ["60"]},
    "LICENTIA": {"-": ["11"], "N": ["11"]},
    "LICEO": {"-": ["52"]},
    "LICEOR": {"-": ["5K"]},
    "LICET": {"1": ["52"], "2": ["60"], "3": ["82"]},
    "LICHANOS": {"-": ["17"]},
    "LICINIVS": {"N": ["12"], "A": ["21"]},
    "LICIO": {"-": ["55"]},
    "LICITATIO": {"-": ["13"]},
    "LICITVS": {"-": ["21"]},
    "LICIVM": {"-": ["12"]},
    "LICTOR": {"-": ["13"]},
    "LIDO": {"-": ["53"]},
    "LIEN": {"-": ["13"]},
    "LIENOSVS": {"-": ["21"]},
    "LIGAMEN": {"-": ["13"]},
    "LIGAMENTVM": {"-": ["12"]},
    "LIGELLVM": {"-": ["12"]},
    "LIGNATIO": {"-": ["13"]},
    "LIGNATOR": {"-": ["13"]},
    "LIGNEVS": {"-": ["21"]},
    "LIGNOR": {"-": ["5J"]},
    "LIGNVM": {"-": ["12"]},
    "LIGO": {"1": ["13"], "2": ["51"], "3": ["53"]},
    "LIGVLA": {"-": ["11"]},
    "LIGVR": {"C": ["13"], "N": ["25"], "A": ["25"]},
    "LIGVRES": {"N": ["13"]},
    "LIGVRRIO": {"-": ["54"]},
    "LIGVRRITOR": {"-": ["13"]},
    "LIGVSTRVM": {"-": ["12"]},
    "LILIVM": {"-": ["12"]},
    "LILYBAETANVS": {"N": ["21"], "A": ["21"]},
    "LILYBAEVM": {"N": ["17"]},
    "LIMA": {"-": ["11"]},
    "LIMATVRA": {"-": ["11"]},
    "LIMATVS": {"-": ["21"]},
    "LIMBVS": {"-": ["12"]},
    "LIMEN": {"-": ["13"]},
    "LIMES": {"-": ["13"]},
    "LIMNE": {"-": ["17"]},
    "LIMO": {"-": ["51"]},
    "LIMOSVS": {"-": ["21"]},
    "LIMPIDVS": {"-": ["21"]},
    "LIMVS": {"1": ["12"], "2": ["12"], "3": ["21"]},
    "LINEA": {"-": ["11"]},
    "LINEAMENTVM": {"-": ["12"]},
    "LINEATIO": {"-": ["13"]},
    "LINEO": {"-": ["51"]},
    "LINEVS": {"-": ["21"]},
    "LINGO": {"-": ["53"]},
    "LINGVA": {"-": ["11"]},
    "LINGVARIVM": {"-": ["12"]},
    "LINGVLA": {"-": ["11"]},
    "LINGVLATVS": {"-": ["21"]},
    "LINIGER": {"-": ["21"]},
    "LINIO": {"-": ["54"]},
    "LINO": {"-": ["53"]},
    "LINQVO": {"-": ["53"]},
    "LINTEATVS": {"-": ["21"]},
    "LINTER": {"-": ["13"]},
    "LINTEVM": {"-": ["12"]},
    "LINTEVS": {"-": ["21"]},
    "LINVM": {"-": ["12"]},
    "LIPARENSES": {"N": ["13"]},
    "LIPARENSIS": {"N": ["24"], "A": ["24"]},
    "LIPPIO": {"-": ["54"]},
    "LIPPITVDO": {"-": ["13"]},
    "LIPPVS": {"-": ["21"]},
    "LIQVEFACIO": {"-": ["55"]},
    "LIQVEFIO": {"-": ["56"]},
    "LIQVEO": {"-": ["52"]},
    "LIQVESCO": {"-": ["53"]},
    "LIQVIDO": {"-": ["60"]},
    "LIQVIDVM": {"1": ["12"], "2": ["60"]},
    "LIQVIDVS": {"-": ["21"]},
    "LIQVOR": {"1": ["13"], "2": ["5L"]},
    "LIS": {"-": ["13"]},
    "LITIGANS": {"-": ["13"]},
    "LITIGATOR": {"-": ["13"]},
    "LITIGIOSVS": {"-": ["21"]},
    "LITIGO": {"-": ["51"]},
    "LITO": {"-": ["51"]},
    "LITORALIS": {"-": ["24"]},
    "LITOREVS": {"-": ["21"]},
    "LITTERA": {"-": ["11"]},
    "LITTERARIVS": {"-": ["21"]},
    "LITTERATOR": {"-": ["13"]},
    "LITTERATVRA": {"-": ["11"]},
    "LITTERATVS": {"-": ["21"]},
    "LITTERVLA": {"-": ["11"]},
    "LITVRA": {"-": ["11"]},
    "LITVS": {"-": ["13"], "N": ["12"]},
    "LITVVS": {"-": ["12"]},
    "LIVENS": {"-": ["25"]},
    "LIVEO": {"-": ["52"]},
    "LIVESCO": {"-": ["53"]},
    "LIVIA": {"N": ["11"]},
    "LIVIDVLVS": {"-": ["21"]},
    "LIVIDVS": {"-": ["21"]},
    "LIVIVS": {"N": ["12"], "A": ["21"]},
    "LIVOR": {"-": ["13"], "N": ["13"]},
    "LIXA": {"1": ["11"], "-": ["11"]},
    "LIXIVVS": {"-": ["21"]},
    "LOCALIS": {"-": ["24"]},
    "LOCATIO": {"-": ["13"]},
    "LOCATOR": {"-": ["13"]},
    "LOCO": {"-": ["51"]},
    "LOCRENSES": {"N": ["13"]},
    "LOCVLAMENTVM": {"-": ["12"]},
    "LOCVLVS": {"-": ["12"]},
    "LOCVPLES": {"-": ["25"]},
    "LOCVPLETO": {"-": ["51"]},
    "LOCVS": {"-": ["12"]},
    "LOCVSTA": {"-": ["11"], "N": ["11"]},
    "LOCVTIO": {"-": ["13"]},
    "LODIX": {"-": ["13"]},
    "LOGEION": {"-": ["17"]},
    "LOGOS": {"-": ["17"]},
    "LOLIGO": {"-": ["13"]},
    "LOLIVM": {"-": ["12"]},
    "LOLLIVS": {"N": ["12"]},
    "LONGAEVVS": {"-": ["21"]},
    "LONGE": {"-": ["60"]},
    "LONGINQVITAS": {"-": ["13"]},
    "LONGINQVVM": {"1": ["12"], "2": ["60"]},
    "LONGINQVVS": {"-": ["21"]},
    "LONGITVDO": {"-": ["13"]},
    "LONGVLA": {"N": ["11"], "N12": ["11"]},
    "LONGVM": {"2": ["60", "6-"], "1": ["12"]},
    "LONGVRIVS": {"-": ["12"]},
    "LONGVS": {"-": ["21"], "A": ["21"], "N": ["12"]},
    "LOQVACITER": {"-": ["60"]},
    "LOQVACVLVS": {"-": ["21"]},
    "LOQVAX": {"-": ["25"]},
    "LOQVELA": {"-": ["11"]},
    "LOQVOR": {"-": ["5L"]},
    "LORA": {"-": ["11"]},
    "LOREVS": {"-": ["21"]},
    "LORICA": {"-": ["11"]},
    "LORICATIO": {"-": ["13"]},
    "LORICATVS": {"-": ["21"]},
    "LORICO": {"-": ["51"]},
    "LORICVLA": {"-": ["11"]},
    "LORIPES": {"-": ["13"]},
    "LORVM": {"-": ["12"]},
    "LOTIO": {"-": ["13"]},
    "LOTIVM": {"-": ["12"]},
    "LOTOS": {"-": ["17"]},
    "LVBRICO": {"-": ["51"]},
    "LVBRICVM": {"-": ["12"]},
    "LVBRICVS": {"-": ["21"]},
    "LVCA": {"-": ["11"], "N": ["11"], "A": ["21"]},
    "LVCAR": {"-": ["13"]},
    "LVCCEIVS": {"N": ["12"]},
    "LVCELLVM": {"-": ["12"]},
    "LVCENS": {"-": ["25"]},
    "LVCEO": {"-": ["52"]},
    "LVCERES": {"N": ["13"]},
    "LVCERNA": {"-": ["11"]},
    "LVCESCO": {"-": ["53"]},
    "LVCIDE": {"-": ["6&"]},
    "LVCIDVM": {"-": ["60"]},
    "LVCIDVS": {"-": ["21"]},
    "LVCIFER": {"1": ["12"], "2": ["21"], "N": ["12"]},
    "LVCIFVGA": {"-": ["11"]},
    "LVCIFVGVS": {"-": ["21"]},
    "LVCILIVS": {"N": ["12"]},
    "LVCIVS": {"N": ["12"]},
    "LVCRETIA": {"N": ["11"]},
    "LVCRETIVS": {"N": ["12"]},
    "LVCRIFICO": {"-": ["51"]},
    "LVCRINVS": {"N": ["21"], "A": ["21"]},
    "LVCROR": {"-": ["5J"]},
    "LVCROSVS": {"-": ["21"]},
    "LVCRVM": {"-": ["12"]},
    "LVCTA": {"-": ["11"]},
    "LVCTATIO": {"-": ["13"]},
    "LVCTATOR": {"-": ["13"]},
    "LVCTIFER": {"-": ["21"]},
    "LVCTIFICABILIS": {"-": ["23"]},
    "LVCTIFICVS": {"-": ["21"]},
    "LVCTOR": {"-": ["5J"]},
    "LVCTVOSVS": {"-": ["21"]},
    "LVCTVS": {"-": ["14"], "N": ["12"]},
    "LVCVBRATIO": {"-": ["13"]},
    "LVCVBRO": {"-": ["51"]},
    "LVCVLENTVS": {"-": ["21"]},
    "LVCVLLVS": {"N": ["12"]},
    "LVCVMO": {"N": ["13"]},
    "LVCVS": {"-": ["12"], "A": ["21"], "N": ["12"]},
    "LVDIA": {"-": ["11"]},
    "LVDIBRIVM": {"-": ["12"]},
    "LVDIBVNDVS": {"-": ["21"]},
    "LVDICER": {"-": ["21"]},
    "LVDICRVM": {"-": ["12"]},
    "LVDICRVS": {"-": ["21"]},
    "LVDIFICATIO": {"-": ["13"]},
    "LVDIFICO": {"-": ["51"]},
    "LVDIFICOR": {"-": ["5J"]},
    "LVDIVS": {"-": ["12"]},
    "LVDO": {"-": ["53"]},
    "LVDVS": {"-": ["12"], "N": ["12"]},
    "LVELA": {"-": ["11"]},
    "LVES": {"-": ["13"], "N": ["13"]},
    "LVGEO": {"-": ["52"]},
    "LVGVBRE": {"-": ["60"]},
    "LVGVBRIA": {"-": ["13"]},
    "LVGVBRIS": {"-": ["24"]},
    "LVMBRICVS": {"-": ["12"]},
    "LVMBVS": {"-": ["12"]},
    "LVMEN": {"-": ["13"]},
    "LVMINARE": {"-": ["12"]},
    "LVMINO": {"-": ["51"]},
    "LVMINOSVS": {"-": ["21"]},
    "LVNA": {"-": ["11"], "N": ["11"]},
    "LVNARIS": {"-": ["24"], "A": ["24"]},
    "LVNATVS": {"-": ["21"]},
    "LVNO": {"-": ["51"]},
    "LVO": {"-": ["53"]},
    "LVPA": {"-": ["11"]},
    "LVPANAR": {"-": ["13"]},
    "LVPATA": {"-": ["12"]},
    "LVPATVS": {"-": ["21"]},
    "LVPERCAL": {"N": ["13"]},
    "LVPINARIVS": {"-": ["21"]},
    "LVPINVM": {"-": ["12"]},
    "LVPINVS": {"1": ["12"], "2": ["21"]},
    "LVPVS": {"-": ["12"], "N": ["12"]},
    "LVRIDVS": {"-": ["21"]},
    "LVROR": {"-": ["13"]},
    "LVSCINIA": {"-": ["11"]},
    "LVSCINIVS": {"-": ["12"]},
    "LVSCITIO": {"-": ["13"]},
    "LVSCITIOSVS": {"-": ["21"]},
    "LVSCVS": {"-": ["21"], "N": ["12"]},
    "LVSOR": {"-": ["13"]},
    "LVSORIAE": {"-": ["11"]},
    "LVSORIVM": {"-": ["12"]},
    "LVSORIVS": {"-": ["21"]},
    "LVSTRALIS": {"-": ["24"]},
    "LVSTRATIO": {"-": ["13"]},
    "LVSTRO": {"1": ["13"], "2": ["51"]},
    "LVSTRVM": {"1": ["12"], "2": ["12"]},
    "LVSVS": {"-": ["14"]},
    "LVTAMENTVM": {"-": ["12"]},
    "LVTATIVS": {"N": ["12"], "A": ["21"]},
    "LVTEOLVS": {"-": ["21"]},
    "LVTEVM": {"-": ["12"]},
    "LVTEVS": {"1": ["21"], "2": ["21"]},
    "LVTO": {"1": ["51"]},
    "LVTOSVS": {"-": ["21"]},
    "LVTRON": {"-": ["17"]},
    "LVTVLENTVS": {"-": ["21"]},
    "LVTVM": {"1": ["12"], "2": ["12"]},
    "LVX": {"-": ["13"], "N": ["13"]},
    "LVXO": {"-": ["51"]},
    "LVXVM": {"-": ["12"]},
    "LVXVRIA": {"-": ["11"], "N": ["11"]},
    "LVXVRIES": {"-": ["15"]},
    "LVXVRIO": {"-": ["51"]},
    "LVXVRIOR": {"-": ["5J"]},
    "LVXVRIOSE": {"-": ["60"]},
    "LVXVRIOSVS": {"-": ["21"]},
    "LVXVS": {"1": ["14"], "2": ["14"], "3": ["21"]},
    "LYCAEVS": {"A": ["21"], "N": ["12"]},
    "LYCHNOBIVS": {"-": ["12"]},
    "LYCHNVS": {"-": ["12"]},
    "LYCIA": {"N": ["11"]},
    "LYCII": {"N": ["12"]},
    "LYCISCVS": {"-": ["12"]},
    "LYMPHA": {"-": ["11"], "N": ["11"]},
    "LYMPHATICVS": {"1": ["12"], "2": ["21"]},
    "LYMPHATVS": {"1": ["14"], "2": ["21"]},
    "LYMPHO": {"-": ["51"]},
    "LYNX": {"-": ["17"]},
    "LYRA": {"-": ["11"], "N": ["11"]},
    "LYRICI": {"-": ["12"], "2": ["12"], "1": ["12"]},
    "LYRICVS": {"-": ["21"]},
    "LYSIMACHVS": {"N": ["12"]},
    "LYSIS": {"-": ["13"], "N": ["13"]},
    "LYSO": {"N": ["17"]},
    "MACEDONES": {"N": ["17"]},
    "MACEDONIA": {"N": ["11"]},
    "MACEDONICVS": {"N": ["21"], "A": ["21"]},
    "MACELLVM": {"-": ["12"]},
    "MACER": {"-": ["21"], "N": ["12"]},
    "MACERATIO": {"-": ["13"]},
    "MACERESCO": {"-": ["53"]},
    "MACERIA": {"-": ["11"]},
    "MACERO": {"-": ["51"]},
    "MACHAERA": {"-": ["11"], "N": ["11"]},
    "MACHINA": {"-": ["11"]},
    "MACHINAMENTVM": {"-": ["12"]},
    "MACHINATIO": {"-": ["13"]},
    "MACHINATOR": {"-": ["13"]},
    "MACHINATRIX": {"-": ["13"]},
    "MACHINOR": {"-": ["5J"]},
    "MACIES": {"-": ["15"], "N": ["13"]},
    "MACRESCO": {"-": ["53"]},
    "MACRITAS": {"-": ["13"]},
    "MACRO": {"N": ["13"]},
    "MACTABILIS": {"-": ["24"]},
    "MACTATOR": {"-": ["13"]},
    "MACTATVS": {"-": ["14"]},
    "MACTE": {"-": ["90"]},
    "MACTO": {"-": ["51"]},
    "MACTVS": {"-": ["21"]},
    "MACVLA": {"-": ["11"]},
    "MACVLO": {"-": ["51"]},
    "MACVLOSVS": {"-": ["21"]},
    "MADEFACIO": {"-": ["55"]},
    "MADEFIO": {"-": ["56"]},
    "MADEO": {"-": ["52"]},
    "MADESCO": {"-": ["53"]},
    "MADIDVS": {"-": ["21"]},
    "MAECENAS": {"N": ["13"]},
    "MAECILIVS": {"N": ["12"]},
    "MAENA": {"-": ["11"]},
    "MAENAS": {"-": ["17"], "N": ["17"]},
    "MAERENS": {"-": ["25"]},
    "MAEREO": {"-": ["52"]},
    "MAEROR": {"-": ["13"]},
    "MAESTITIA": {"-": ["11"]},
    "MAESTVS": {"-": ["21"]},
    "MAEVIVS": {"N": ["12"]},
    "MAGA": {"-": ["11"]},
    "MAGALIA": {"-": ["13"]},
    "MAGE": {"-": ["6&"]},
    "MAGICVS": {"-": ["21"]},
    "MAGIS": {"1": ["13"], "2": ["6&"]},
    "MAGISTER": {"-": ["12"]},
    "MAGISTERIVM": {"-": ["12"]},
    "MAGISTRA": {"-": ["11"]},
    "MAGISTRATVS": {"-": ["14"]},
    "MAGIVS": {"N": ["12"]},
    "MAGNANIMITAS": {"-": ["13"]},
    "MAGNANIMVS": {"-": ["21"]},
    "MAGNAS": {"-": ["13"]},
    "MAGNES": {"-": ["17"], "N": ["13"]},
    "MAGNETICVS": {"-": ["21"]},
    "MAGNIFICE": {"-": ["60"]},
    "MAGNIFICENS": {"-": ["25"]},
    "MAGNIFICENTER": {"-": ["60"]},
    "MAGNIFICENTIA": {"-": ["11"]},
    "MAGNIFICO": {"-": ["51"]},
    "MAGNIFICVS": {"-": ["21"]},
    "MAGNILOQVVS": {"-": ["21"]},
    "MAGNITVDO": {"-": ["13"]},
    "MAGNOPERE": {"-": ["60"]},
    "MAGNVM": {"1": ["12"], "2": ["60"]},
    "MAGNVS": {"-": ["21"], "A": ["21"], "N": ["12"]},
    "MAGVS": {"1": ["12"], "2": ["21"], "N": ["12"]},
    "MAIESTAS": {"-": ["13"], "N": ["13"]},
    "MAIORES": {"-": ["13"]},
    "MAIVS": {"A": ["21"], "N": ["12"]},
    "MAIVSCVLVS": {"-": ["21"]},
    "MALA": {"-": ["11"]},
    "MALACIA": {"-": ["11"]},
    "MALAXO": {"-": ["51"]},
    "MALE": {"-": ["60"]},
    "MALEDICENS": {"-": ["25"]},
    "MALEDICO": {"2": ["53"]},
    "MALEDICTVM": {"-": ["12"]},
    "MALEDICVS": {"-": ["21"]},
    "MALEFACIO": {"-": ["55"]},
    "MALEFACTVM": {"-": ["12"]},
    "MALEFICIVM": {"-": ["12"]},
    "MALEFICVS": {"-": ["21"]},
    "MALESVADVS": {"-": ["21"]},
    "MALEVOLENTIA": {"-": ["11"]},
    "MALEVOLVS": {"-": ["21"]},
    "MALI": {"-": ["12"]},
    "MALIGNE": {"-": ["60"]},
    "MALIGNITAS": {"-": ["13"]},
    "MALIGNVS": {"-": ["21"]},
    "MALITIA": {"-": ["11"]},
    "MALITIOSA": {"N": ["11"]},
    "MALITIOSE": {"-": ["60"]},
    "MALLEOLVS": {"-": ["12"], "N": ["12"]},
    "MALLEVS": {"-": ["12"]},
    "MALLVS": {"-": ["12"], "N": ["12"]},
    "MALO": {"-": ["56"]},
    "MALOBATHRON": {"-": ["17"]},
    "MALVA": {"-": ["11"]},
    "MALVM": {"1": ["12"], "2": ["12"], "3": ["90"]},
    "MALVS": {"1": ["12"], "2": ["12"], "3": ["21"]},
    "MAMERTINI": {"N": ["12"]},
    "MAMERTINVS": {"A": ["21"]},
    "MAMILIVS": {"N": ["12"], "A": ["21"]},
    "MAMILLA": {"-": ["11"]},
    "MAMMA": {"-": ["11"]},
    "MAMMOSVS": {"-": ["21"]},
    "MANABILIS": {"-": ["24"]},
    "MANAEVS": {"-": ["21"]},
    "MANCEPS": {"-": ["13"]},
    "MANCIPIVM": {"-": ["12"]},
    "MANCIPO": {"-": ["51"]},
    "MANCVS": {"-": ["21"]},
    "MANDATVM": {"-": ["12"]},
    "MANDATVS": {"-": ["14"]},
    "MANDELA": {"-": ["11"]},
    "MANDO": {"1": ["13"], "2": ["51"], "3": ["53"]},
    "MANDRA": {"-": ["11"]},
    "MANDVCO": {"1": ["13"], "2": ["51"]},
    "MANE": {"1": ["16"], "2": ["60"]},
    "MANEO": {"-": ["52"]},
    "MANES": {"-": ["13"], "N": ["13"]},
    "MANGO": {"-": ["13"]},
    "MANICA": {"-": ["11"]},
    "MANICATVS": {"-": ["21"]},
    "MANICVLVS": {"-": ["12"]},
    "MANIFESTE": {"-": ["60"]},
    "MANIFESTO": {"1": ["51"], "2": ["60"]},
    "MANIFESTVS": {"-": ["21"]},
    "MANILIVS": {"N": ["12"], "A": ["21"]},
    "MANIPVLARIS": {"-": ["24"], "2": ["24"], "1": ["21"]},
    "MANIPVLVS": {"-": ["12"]},
    "MANIVS": {"N": ["12"]},
    "MANLIVS": {"N": ["12"], "A": ["21"]},
    "MANNVS": {"-": ["12"], "N": ["12"]},
    "MANO": {"-": ["51"]},
    "MANSITO": {"-": ["51"]},
    "MANSVEFACIO": {"-": ["55"]},
    "MANSVEFIO": {"-": ["56"]},
    "MANSVES": {"-": ["22"]},
    "MANSVESCO": {"-": ["53"]},
    "MANSVETE": {"-": ["60"]},
    "MANSVETVDO": {"-": ["13"]},
    "MANSVETVS": {"-": ["21"], "N": ["12"]},
    "MANTELE": {"-": ["13"]},
    "MANTICA": {"-": ["11"]},
    "MANVALIS": {"-": ["24"]},
    "MANVBIAE": {"-": ["11"]},
    "MANVBRIVM": {"-": ["12"]},
    "MANVCLA": {"-": ["11"]},
    "MANVLEATVS": {"-": ["21"]},
    "MANVMISSIO": {"-": ["13"]},
    "MANVMITTO": {"-": ["53"]},
    "MANVPRETIOSVS": {"-": ["21"]},
    "MANVPRETIVM": {"-": ["12"]},
    "MANVS": {"1": ["14"], "2": ["21"]},
    "MAPALIA": {"-": ["13"]},
    "MAPPA": {"-": ["11"]},
    "MARATHVS": {"-": ["12"], "N": ["12"]},
    "MARCELLIA": {"N": ["12"]},
    "MARCELLINVS": {"N": ["12"]},
    "MARCELLVS": {"N": ["12"]},
    "MARCENS": {"-": ["25"]},
    "MARCEO": {"-": ["52"]},
    "MARCESCO": {"-": ["53"]},
    "MARCIA": {"N": ["11"]},
    "MARCIDVS": {"-": ["21"]},
    "MARCIVS": {"N": ["12"], "A": ["21"]},
    "MARCOR": {"-": ["13"]},
    "MARCVS": {"N": ["12"]},
    "MARE": {"-": ["13"], "N": ["13"]},
    "MARGARITA": {"-": ["11"], "N": ["11"]},
    "MARGARITVM": {"-": ["12"]},
    "MARGO": {"-": ["13"]},
    "MARINVS": {"-": ["21"], "N": ["12"]},
    "MARISCA": {"-": ["11"]},
    "MARITA": {"-": ["11"]},
    "MARITALIS": {"-": ["24"]},
    "MARITIMI": {"-": ["12"]},
    "MARITIMVS": {"-": ["21"]},
    "MARITO": {"-": ["51"]},
    "MARITVS": {"1": ["12"], "2": ["21"]},
    "MARIVS": {"N": ["12"]},
    "MARMOR": {"-": ["13"]},
    "MARMORARIVS": {"1": ["12"], "2": ["21"]},
    "MARMOREVS": {"-": ["21"]},
    "MARRA": {"-": ["11"]},
    "MARS": {"N": ["13"]},
    "MARTIALIS": {"-": ["24"], "N": ["13"], "A": ["12"]},
    "MARTIVS": {"A": ["21"], "N": ["12"]},
    "MAS": {"-": ["13"]},
    "MASCVLINVS": {"-": ["21"]},
    "MASCVLVS": {"-": ["21"]},
    "MASINISSA": {"N": ["11"]},
    "MASSA": {"-": ["11"], "N": ["11"]},
    "MASSILIA": {"N": ["11"]},
    "MATARA": {"-": ["11"]},
    "MATELLA": {"-": ["11"]},
    "MATELLIO": {"-": ["13"]},
    "MATEOLA": {"-": ["11"]},
    "MATER": {"-": ["13"], "N": ["13"]},
    "MATERCVLA": {"-": ["11"]},
    "MATERIA": {"-": ["11"]},
    "MATERIALIS": {"-": ["24"]},
    "MATERIATIO": {"-": ["13"]},
    "MATERIATVRA": {"-": ["11"]},
    "MATERIES": {"-": ["15"]},
    "MATERINVS": {"-": ["21"]},
    "MATERIO": {"-": ["51"]},
    "MATERIOR": {"-": ["5J"]},
    "MATERNVS": {"-": ["21"], "N": ["12"]},
    "MATERTERA": {"-": ["11"]},
    "MATHEMATICA": {"-": ["17"]},
    "MATHEMATICVS": {"1": ["12"], "2": ["21"]},
    "MATRIMONIVM": {"-": ["12"]},
    "MATRIMVS": {"-": ["21"]},
    "MATRINIVS": {"N": ["12"]},
    "MATRONA": {"-": ["11"], "N": ["11"]},
    "MATRONALIA": {"-": ["13"]},
    "MATRONALIS": {"-": ["24"]},
    "MATVRE": {"-": ["60"]},
    "MATVRESCO": {"-": ["53"]},
    "MATVRITAS": {"-": ["13"]},
    "MATVRO": {"-": ["51"]},
    "MATVRVS": {"-": ["21"], "N": ["12"]},
    "MATVTINVM": {"-": ["12"]},
    "MATVTINVS": {"-": ["21"], "A": ["21"]},
    "MAVRETANIA": {"N": ["11"]},
    "MAXILLA": {"-": ["11"]},
    "MAXIMITAS": {"-": ["13"]},
    "MAXIMVM": {"-": ["6-"]},
    "MAXIMVS": {"N": ["12"]},
    "MAZONOMVS": {"-": ["12"]},
    "MEATVS": {"-": ["14"]},
    "MECHANICOS": {"-": ["60"]},
    "MEDEA": {"N": ["11"]},
    "MEDENS": {"-": ["13"]},
    "MEDEOR": {"-": ["5K"]},
    "MEDIANVS": {"-": ["21"]},
    "MEDIASTINVS": {"-": ["12"]},
    "MEDICA": {"1": ["11"], "2": ["11"]},
    "MEDICABILIS": {"-": ["24"]},
    "MEDICAMEN": {"-": ["13"]},
    "MEDICAMENTOSVS": {"-": ["21"]},
    "MEDICAMENTVM": {"-": ["12"]},
    "MEDICATVS": {"1": ["14"], "2": ["21"]},
    "MEDICINA": {"-": ["11"]},
    "MEDICO": {"-": ["51"]},
    "MEDICOR": {"-": ["5J"]},
    "MEDICVS": {"1": ["12"], "2": ["21"], "N": ["21"], "A": ["21"]},
    "MEDIETAS": {"-": ["13"]},
    "MEDIMNVM": {"-": ["12"]},
    "MEDIO": {"-": ["51"]},
    "MEDIOCRICVLVS": {"-": ["21"]},
    "MEDIOCRIS": {"-": ["24"]},
    "MEDIOCRITAS": {"-": ["13"]},
    "MEDIOCRITER": {"-": ["60"]},
    "MEDIPONTVS": {"-": ["12"]},
    "MEDITAMENTVM": {"-": ["12"]},
    "MEDITATE": {"-": ["60"]},
    "MEDITATIO": {"-": ["13"]},
    "MEDITERRANEVM": {"-": ["12"]},
    "MEDITERRANEVS": {"-": ["21"]},
    "MEDITOR": {"-": ["5J"]},
    "MEDIVM": {"-": ["12"]},
    "MEDIVS": {"-": ["21"]},
    "MEDIVSFIDIVS": {"-": ["90"]},
    "MEDVLLA": {"-": ["11"]},
    "MEDVLLIA": {"N": ["11"]},
    "MEDVSA": {"-": ["11"], "N": ["11"]},
    "MEGALENSIS": {"-": ["24"], "A": ["24"]},
    "MEGALOGRAPHIA": {"-": ["11"]},
    "MEGARIS": {"N": ["17"]},
    "MEGISTANES": {"-": ["17"]},
    "MEHERCVLE": {"-": ["90"]},
    "MEI": {"-": ["42"]},
    "MEIO": {"-": ["53"]},
    "MEL": {"-": ["13"]},
    "MELA": {"N": ["11"]},
    "MELANTHIVM": {"-": ["12"]},
    "MELICHROS": {"-": ["17"]},
    "MELICVS": {"1": ["12"], "2": ["21"]},
    "MELIMELVM": {"-": ["12"]},
    "MELIORO": {"-": ["51"]},
    "MELISPHYLLVM": {"-": ["12"]},
    "MELITA": {"N": ["17"]},
    "MELITENSES": {"N": ["13"]},
    "MELITENSIA": {"N": ["13"]},
    "MELITENSIS": {"N": ["24"], "A": ["24"]},
    "MELIVSCVLVS": {"-": ["21"]},
    "MELLITVS": {"-": ["21"]},
    "MELOGRAPHIA": {"-": ["11"]},
    "MELOS": {"-": ["17"]},
    "MEMBRANA": {"-": ["11"]},
    "MEMBRATIM": {"-": ["60"]},
    "MEMBRATVRA": {"-": ["11"]},
    "MEMBRVM": {"-": ["12"]},
    "MEMINI": {"-": ["53"]},
    "MEMOR": {"-": ["26"]},
    "MEMORABILIS": {"-": ["24"]},
    "MEMORANDVS": {"-": ["21"]},
    "MEMORATOR": {"-": ["13"]},
    "MEMORIA": {"-": ["11"]},
    "MEMORITER": {"-": ["60"]},
    "MEMORO": {"-": ["51"]},
    "MENAENI": {"N": ["12"]},
    "MENAENVS": {"N": ["21"], "A": ["21"]},
    "MENDA": {"-": ["11"]},
    "MENDACIVM": {"-": ["12"]},
    "MENDAX": {"-": ["25"]},
    "MENDICATIO": {"-": ["13"]},
    "MENDICE": {"-": ["60"]},
    "MENDICITAS": {"-": ["13"]},
    "MENDICO": {"-": ["51"]},
    "MENDICVS": {"-": ["21"], "2": ["12"], "1": ["12"]},
    "MENDO": {"-": ["51"]},
    "MENDOSE": {"-": ["60"]},
    "MENDOSVS": {"-": ["21"]},
    "MENDVM": {"-": ["12"]},
    "MENENIANVS": {"N": ["21"]},
    "MENENIVS": {"N": ["12"]},
    "MENISCVS": {"N": ["12"]},
    "MENS": {"-": ["13"], "N": ["13"]},
    "MENSA": {"-": ["11"]},
    "MENSIS": {"-": ["13"]},
    "MENSOR": {"-": ["13"]},
    "MENSTRVA": {"-": ["12"]},
    "MENSTRVALIS": {"-": ["24"]},
    "MENSTRVVM": {"-": ["12"]},
    "MENSTRVVS": {"-": ["21"]},
    "MENSVRA": {"-": ["11"]},
    "MENTA": {"-": ["11"]},
    "MENTIO": {"1": ["13"], "2": ["54"]},
    "MENTIOR": {"-": ["5M"]},
    "MENTOR": {"N": ["13"]},
    "MENTVLA": {"-": ["11"], "N": ["11"]},
    "MENTVM": {"-": ["12"]},
    "MEO": {"-": ["51"]},
    "MEPHITIS": {"-": ["13"]},
    "MERACE": {"-": ["6&"]},
    "MERACVS": {"-": ["21"]},
    "MERCABILIS": {"-": ["24"]},
    "MERCATOR": {"-": ["13"]},
    "MERCATVRA": {"-": ["11"]},
    "MERCATVS": {"-": ["14"]},
    "MERCEDVLA": {"-": ["11"]},
    "MERCENARIVS": {"1": ["12"], "2": ["21"]},
    "MERCES": {"-": ["13"]},
    "MERCIMONIVM": {"-": ["12"]},
    "MERCOR": {"-": ["5J"]},
    "MERCVRIALIS": {"-": ["21"], "A": ["24"]},
    "MERCVRIVS": {"N": ["12"]},
    "MERDA": {"-": ["11"]},
    "MERENS": {"-": ["25"]},
    "MEREO": {"-": ["52"]},
    "MEREOR": {"-": ["5K"]},
    "MERETRICIVS": {"-": ["21"]},
    "MERETRICVLA": {"-": ["11"]},
    "MERETRIX": {"-": ["13"]},
    "MEREVS": {"-": ["12"]},
    "MERGA": {"-": ["11"]},
    "MERGES": {"-": ["13"]},
    "MERGO": {"-": ["53"]},
    "MERGVS": {"-": ["12"], "N": ["12"]},
    "MERIDIANO": {"-": ["60"]},
    "MERIDIANVM": {"-": ["12"]},
    "MERIDIANVS": {"-": ["21"]},
    "MERIDIES": {"-": ["15"]},
    "MERIDIO": {"-": ["51"]},
    "MERITO": {"1": ["51"], "2": ["60"]},
    "MERITORIVM": {"-": ["12"]},
    "MERITORIVS": {"-": ["21"]},
    "MERITVM": {"-": ["12"]},
    "MERITVS": {"-": ["21"]},
    "MEROPS": {"-": ["13"], "N": ["13"]},
    "MEROS": {"-": ["17"], "N": ["17"]},
    "MERSO": {"-": ["51"]},
    "MERVLA": {"-": ["11"], "N": ["11"]},
    "MERVM": {"-": ["12"]},
    "MERVS": {"-": ["21"]},
    "MERX": {"-": ["13"]},
    "MESAVLOE": {"-": ["17"]},
    "MESE": {"-": ["17"]},
    "MESIA": {"N": ["11"]},
    "MESOLABIVM": {"-": ["12"]},
    "MESOS": {"-": ["27"]},
    "MESSALA": {"N": ["11"]},
    "MESSANA": {"N": ["11"]},
    "MESSIS": {"-": ["13"]},
    "MESSOR": {"-": ["13"]},
    "META": {"-": ["11"], "N": ["11"]},
    "METALLVM": {"-": ["12"]},
    "METAPONTVM": {"N": ["12"]},
    "METAXA": {"-": ["11"]},
    "METELLVS": {"N": ["12"]},
    "METHODVS": {"-": ["17"]},
    "METILIVS": {"N": ["12"]},
    "METIOR": {"-": ["5M"]},
    "METO": {"1": ["51"], "2": ["53"]},
    "METOPA": {"-": ["17"]},
    "METOR": {"-": ["5J"]},
    "METRETA": {"-": ["11"]},
    "METRODORVS": {"N": ["12"]},
    "METROPOLIS": {"N": ["17"]},
    "METROPOLITAE": {"N": ["17"]},
    "METRVM": {"-": ["12"]},
    "METTIVS": {"N": ["12"]},
    "METVENS": {"-": ["25"]},
    "METVO": {"-": ["53"]},
    "METVS": {"-": ["14"], "N": ["12"]},
    "MEVM": {"-": ["12"]},
    "MEVS": {"-": ["42"]},
    "MEZENTIVS": {"N": ["12"]},
    "MIA": {"-": ["11"]},
    "MICA": {"-": ["11"]},
    "MICANS": {"-": ["25"]},
    "MICO": {"-": ["51"]},
    "MICTVRIO": {"-": ["54"]},
    "MIGRATIO": {"-": ["13"]},
    "MIGRO": {"-": ["51"]},
    "MILES": {"-": ["13"]},
    "MILESII": {"N": ["12"]},
    "MILESIVS": {"N": ["21"], "A": ["21"]},
    "MILETVS": {"N": ["12"]},
    "MILIA": {"-": ["13"]},
    "MILIARIVS": {"-": ["21"]},
    "MILITARIS": {"-": ["24"]},
    "MILITIA": {"-": ["11"]},
    "MILITO": {"1": ["13"], "2": ["51"]},
    "MILIVM": {"-": ["12"]},
    "MILLE": {"-": ["31"]},
    "MILLESIMVS": {"-": ["32"]},
    "MILLIARIVM": {"-": ["12"]},
    "MILLIARIVS": {"-": ["21"]},
    "MILLIES": {"-": ["36"]},
    "MILVVS": {"-": ["12"], "N": ["12"]},
    "MILYAE": {"N": ["17"]},
    "MIMA": {"-": ["11"]},
    "MIMICE": {"-": ["60"]},
    "MIMICVS": {"-": ["21"]},
    "MIMVS": {"-": ["12"]},
    "MINA": {"-": ["11"]},
    "MINACITER": {"-": ["60"]},
    "MINAE": {"-": ["11"], "N": ["11"]},
    "MINANTER": {"-": ["60"]},
    "MINAX": {"-": ["25"]},
    "MINEO": {"-": ["52"]},
    "MINERVA": {"N": ["11"]},
    "MINGO": {"-": ["53"]},
    "MINIACEVS": {"-": ["21"]},
    "MINIMVM": {"1": ["12"], "2": ["6-", "12"]},
    "MINISCOR": {"-": ["5L"]},
    "MINISTER": {"1": ["12"], "2": ["21"]},
    "MINISTERIVM": {"-": ["12"]},
    "MINISTRA": {"-": ["11"]},
    "MINISTRATIO": {"-": ["13"]},
    "MINISTRATOR": {"-": ["13"]},
    "MINISTRO": {"-": ["51"]},
    "MINITABVNDVS": {"-": ["21"]},
    "MINITOR": {"-": ["5J"]},
    "MINIVM": {"-": ["12"]},
    "MINOR": {"-": ["5J"]},
    "MINVCIVS": {"A": ["21"], "N": ["12"]},
    "MINVO": {"-": ["53"]},
    "MINVS": {"-": ["12"]},
    "MINVSCVLVS": {"-": ["21"]},
    "MINVTAL": {"-": ["13"]},
    "MINVTATIM": {"-": ["60"]},
    "MINVTE": {"-": ["60"]},
    "MINVTIA": {"-": ["11"]},
    "MINVTIM": {"-": ["60"]},
    "MINVTVM": {"-": ["12"]},
    "MINVTVS": {"1": ["12"], "2": ["21"]},
    "MIRABILE": {"-": ["60"]},
    "MIRABILIS": {"-": ["24"]},
    "MIRABVNDVS": {"-": ["21"]},
    "MIRACVLVM": {"-": ["12"]},
    "MIRANDE": {"-": ["60"]},
    "MIRANDVS": {"-": ["21"]},
    "MIRATOR": {"-": ["13"]},
    "MIRATRIX": {"-": ["13"]},
    "MIRE": {"-": ["60"]},
    "MIRIFICE": {"-": ["60"]},
    "MIRIFICVS": {"-": ["21"]},
    "MIRMILLO": {"-": ["13"]},
    "MIROR": {"-": ["5J"]},
    "MIRVM": {"-": ["12"]},
    "MIRVS": {"-": ["21"]},
    "MISCELLANEA": {"-": ["12"]},
    "MISCELLVS": {"-": ["21"]},
    "MISCEO": {"-": ["52"]},
    "MISELLVS": {"-": ["21"]},
    "MISER": {"-": ["21"]},
    "MISERABILE": {"-": ["60"]},
    "MISERABILIS": {"-": ["24"]},
    "MISERABILITER": {"-": ["60"]},
    "MISERANDVS": {"-": ["21"]},
    "MISERATIO": {"-": ["13"]},
    "MISERE": {"-": ["60"]},
    "MISEREO": {"-": ["52"]},
    "MISEREOR": {"-": ["5K"]},
    "MISERESCO": {"-": ["53"]},
    "MISERIA": {"-": ["11"]},
    "MISERICORDIA": {"-": ["11"]},
    "MISERICORS": {"-": ["25"]},
    "MISERITER": {"-": ["60"]},
    "MISEROR": {"-": ["5J"]},
    "MISSILE": {"-": ["13"]},
    "MISSILIA": {"-": ["13"]},
    "MISSILIS": {"-": ["24"]},
    "MISSIO": {"-": ["13"]},
    "MISSITO": {"-": ["51"]},
    "MISSVS": {"-": ["14"]},
    "MITE": {"-": ["6&"]},
    "MITESCO": {"-": ["53"]},
    "MITHRIDATES": {"N": ["17"]},
    "MITIGO": {"-": ["51"]},
    "MITIS": {"-": ["24"]},
    "MITRA": {"-": ["11"]},
    "MITRATVS": {"-": ["21"]},
    "MITTO": {"-": ["53"]},
    "MITVLVS": {"-": ["12"]},
    "MITYLENAE": {"N": ["11"]},
    "MIXTIM": {"-": ["60"]},
    "MIXTIO": {"-": ["13"]},
    "MIXTVRA": {"-": ["11"]},
    "MIXTVS": {"-": ["21"]},
    "MNASISTRATVS": {"N": ["12"]},
    "MNEMOSYNON": {"-": ["17"]},
    "MOBILIS": {"-": ["24"]},
    "MOBILITAS": {"-": ["13"]},
    "MOBILITER": {"-": ["60"]},
    "MOBILITO": {"-": ["51"]},
    "MODERABILIS": {"-": ["24"]},
    "MODERAMEN": {"-": ["13"]},
    "MODERANTER": {"-": ["60"]},
    "MODERATE": {"-": ["60"]},
    "MODERATIM": {"-": ["60"]},
    "MODERATIO": {"-": ["13"]},
    "MODERATOR": {"-": ["13"]},
    "MODERATRIX": {"-": ["13"]},
    "MODERATVS": {"-": ["21"]},
    "MODEROR": {"-": ["5J"]},
    "MODESTE": {"-": ["60"]},
    "MODESTIA": {"-": ["11"]},
    "MODESTVS": {"-": ["21"], "N": ["12"]},
    "MODICE": {"-": ["60"]},
    "MODICVS": {"-": ["21"]},
    "MODIFICATIO": {"-": ["13"]},
    "MODIOLVS": {"-": ["12"]},
    "MODIVM": {"-": ["12"]},
    "MODIVS": {"-": ["12"], "N": ["12"]},
    "MODO": {"1": ["60"], "2": ["82"]},
    "MODVLATIO": {"-": ["13"]},
    "MODVLATOR": {"-": ["13"]},
    "MODVLATVS": {"1": ["14"], "2": ["21"]},
    "MODVLOR": {"-": ["5J"]},
    "MODVLVS": {"-": ["12"]},
    "MODVS": {"-": ["12"]},
    "MOECHA": {"-": ["11"]},
    "MOECHOR": {"-": ["5J"]},
    "MOECHVS": {"-": ["12"]},
    "MOENIA": {"-": ["13"]},
    "MOLA": {"-": ["11"]},
    "MOLARIS": {"1": ["13"], "2": ["24"]},
    "MOLARIVS": {"-": ["21"]},
    "MOLES": {"-": ["13"]},
    "MOLESTE": {"-": ["60"]},
    "MOLESTIA": {"-": ["11"]},
    "MOLESTO": {"-": ["51"]},
    "MOLESTVS": {"-": ["21"]},
    "MOLETRINA": {"-": ["11"]},
    "MOLILE": {"-": ["13"]},
    "MOLIMEN": {"-": ["13"]},
    "MOLIMENTVM": {"-": ["12"]},
    "MOLIOR": {"-": ["5M"]},
    "MOLITIO": {"-": ["13"], "1": ["13"]},
    "MOLITOR": {"-": ["13"]},
    "MOLLESCO": {"-": ["53"]},
    "MOLLICELLVS": {"-": ["21"]},
    "MOLLICVLVS": {"-": ["21"]},
    "MOLLIMENTVM": {"-": ["12"]},
    "MOLLIO": {"-": ["54"]},
    "MOLLIS": {"-": ["24"]},
    "MOLLITER": {"-": ["60"]},
    "MOLLITIA": {"-": ["11"]},
    "MOLLITIES": {"-": ["15"]},
    "MOLLITVDO": {"-": ["13"]},
    "MOLO": {"-": ["53"]},
    "MOMEN": {"-": ["13"]},
    "MOMENTVM": {"-": ["12"]},
    "MONACHA": {"-": ["11"]},
    "MONACHVS": {"-": ["12"]},
    "MONAS": {"-": ["17"]},
    "MONASTERIVM": {"-": ["12"]},
    "MONEO": {"-": ["52"]},
    "MONETA": {"-": ["11"], "N": ["11"]},
    "MONETALIS": {"-": ["24"]},
    "MONILE": {"-": ["13"]},
    "MONIMENTVM": {"-": ["12"]},
    "MONITIO": {"-": ["13"]},
    "MONITOR": {"-": ["13"]},
    "MONITVM": {"-": ["12"]},
    "MONITVS": {"-": ["14"]},
    "MONOPTEROS": {"-": ["27"]},
    "MONOTRIGLYPHVS": {"-": ["27"]},
    "MONS": {"-": ["13"], "N": ["13"]},
    "MONSTRATIO": {"-": ["13"]},
    "MONSTRATOR": {"-": ["13"]},
    "MONSTRO": {"-": ["51"]},
    "MONSTRVM": {"-": ["12"]},
    "MONSTRVOSVS": {"-": ["21"]},
    "MONTANI": {"-": ["12"]},
    "MONTANVS": {"-": ["21"], "N": ["12"]},
    "MONTIFER": {"-": ["21"]},
    "MONTIVAGVS": {"-": ["21"]},
    "MONTVOSVS": {"-": ["21"]},
    "MONVMENTVM": {"-": ["12"]},
    "MORA": {"-": ["11"]},
    "MORALIS": {"-": ["24"]},
    "MORATIO": {"-": ["13"]},
    "MORATOR": {"-": ["13"]},
    "MORATVS": {"-": ["21"]},
    "MORBIDVS": {"-": ["21"]},
    "MORBOSVS": {"-": ["21"]},
    "MORBVS": {"-": ["12"], "N": ["12"]},
    "MORDAX": {"-": ["25"]},
    "MORDEO": {"-": ["52"]},
    "MORDICVS": {"-": ["60"], "2": ["21"]},
    "MORIBVNDVS": {"-": ["21"]},
    "MORIGEROR": {"-": ["5J"]},
    "MORIGERVS": {"-": ["21"]},
    "MORIOR": {"-": ["5N"]},
    "MOROR": {"1": ["5J"], "2": ["5J"]},
    "MOROSVS": {"-": ["21"]},
    "MORS": {"-": ["13"], "N": ["13"]},
    "MORSVS": {"-": ["14"]},
    "MORTALIA": {"-": ["13"]},
    "MORTALIS": {"1": ["13"], "2": ["24"]},
    "MORTALITAS": {"-": ["13"]},
    "MORTARIVM": {"-": ["12"]},
    "MORTICINVS": {"-": ["21"]},
    "MORTIFER": {"-": ["21"]},
    "MORTVVS": {"1": ["12"], "2": ["21"]},
    "MORVM": {"-": ["12"]},
    "MORVS": {"-": ["12"]},
    "MOS": {"-": ["13"]},
    "MOTIO": {"-": ["13"]},
    "MOTIVNCVLA": {"-": ["11"]},
    "MOTO": {"-": ["51"]},
    "MOTVS": {"-": ["14"]},
    "MOVEO": {"-": ["52"]},
    "MOX": {"-": ["60"]},
    "MVCEO": {"-": ["52"]},
    "MVCIA": {"N": ["12"]},
    "MVCIDVS": {"-": ["21"]},
    "MVCIVS": {"A": ["21"], "N": ["12"]},
    "MVCRO": {"-": ["13"]},
    "MVCVS": {"-": ["12"]},
    "MVGILIS": {"-": ["13"]},
    "MVGILLA": {"N": ["11"]},
    "MVGIO": {"-": ["54"]},
    "MVGITVS": {"-": ["14"]},
    "MVLA": {"-": ["11"]},
    "MVLCEO": {"-": ["52"]},
    "MVLCO": {"-": ["51"]},
    "MVLCTRA": {"-": ["11"]},
    "MVLCTRARIVM": {"-": ["12"]},
    "MVLCTRVM": {"-": ["12"]},
    "MVLGEO": {"-": ["52"]},
    "MVLIEBRIA": {"-": ["13"]},
    "MVLIEBRIS": {"-": ["24"], "N": ["24"]},
    "MVLIEBRITER": {"-": ["60"]},
    "MVLIER": {"-": ["13"]},
    "MVLIERARIVS": {"1": ["12"], "2": ["21"]},
    "MVLIERCVLA": {"-": ["11"]},
    "MVLINVS": {"-": ["21"]},
    "MVLIO": {"-": ["13"]},
    "MVLLEVS": {"-": ["21"]},
    "MVLLVS": {"-": ["12"]},
    "MVLSVM": {"-": ["12"]},
    "MVLSVS": {"-": ["21"]},
    "MVLTA": {"1": ["11"], "2": ["60"]},
    "MVLTANGVLVS": {"-": ["21"]},
    "MVLTATIO": {"-": ["13"]},
    "MVLTESIMVS": {"-": ["21"]},
    "MVLTI": {"-": ["12"]},
    "MVLTICIA": {"-": ["12"]},
    "MVLTIFACIO": {"-": ["55"]},
    "MVLTIFARIAM": {"-": ["60"]},
    "MVLTIFIDVS": {"-": ["21"]},
    "MVLTIFORMIS": {"-": ["24"]},
    "MVLTIFORVS": {"-": ["21"]},
    "MVLTIGENVS": {"-": ["21"]},
    "MVLTIMODIS": {"-": ["60"]},
    "MVLTIMODVS": {"-": ["21"]},
    "MVLTIPLEX": {"-": ["25"]},
    "MVLTIPLICATIO": {"-": ["13"]},
    "MVLTIPLICO": {"-": ["51"]},
    "MVLTITO": {"-": ["51"]},
    "MVLTITVDO": {"-": ["13"]},
    "MVLTIVAGVS": {"-": ["21"]},
    "MVLTIVOLVS": {"-": ["21"]},
    "MVLTO": {"1": ["51"], "2": ["60"]},
    "MVLTVM": {"1": ["12"], "2": ["60"]},
    "MVLTVS": {"-": ["21"]},
    "MVLVS": {"-": ["12"]},
    "MVMMIVS": {"N": ["12"]},
    "MVNATIVS": {"-": ["12"], "N": ["12"]},
    "MVNDANVS": {"-": ["21"]},
    "MVNDE": {"-": ["60"]},
    "MVNDITIA": {"-": ["11"], "N": ["11"]},
    "MVNDITIES": {"-": ["15"]},
    "MVNDVS": {"1": ["12"], "2": ["21"]},
    "MVNERO": {"-": ["51"]},
    "MVNEROR": {"-": ["5J"]},
    "MVNIA": {"-": ["16"]},
    "MVNICEPS": {"-": ["13"]},
    "MVNICIPALIS": {"-": ["24"]},
    "MVNICIPIVM": {"-": ["12"]},
    "MVNIFICE": {"-": ["60"]},
    "MVNIFICENTIA": {"-": ["11"]},
    "MVNIFICO": {"-": ["51"]},
    "MVNIFICVS": {"-": ["21"]},
    "MVNIMEN": {"-": ["13"]},
    "MVNIMENTVM": {"-": ["12"]},
    "MVNIO": {"1": ["13"], "2": ["54"]},
    "MVNITIO": {"-": ["13"]},
    "MVNITOR": {"-": ["13"]},
    "MVNITVS": {"-": ["21"]},
    "MVNVS": {"-": ["13"]},
    "MVNVSCVLVM": {"-": ["12"]},
    "MVRAENA": {"-": ["11"]},
    "MVRALIS": {"-": ["24"]},
    "MVRCIA": {"N": ["11"]},
    "MVRENA": {"N": ["11"]},
    "MVREX": {"-": ["13"]},
    "MVRGANTINI": {"N": ["12"]},
    "MVRGANTINVS": {"N": ["21"], "A": ["21"]},
    "MVRIA": {"-": ["11"]},
    "MVRIES": {"-": ["15"]},
    "MVRMVR": {"-": ["13"]},
    "MVRMVRATIO": {"-": ["13"]},
    "MVRMVRO": {"-": ["51"]},
    "MVRVS": {"-": ["12"]},
    "MVS": {"-": ["13"], "N": ["12"]},
    "MVSA": {"-": ["11"], "N": ["11"]},
    "MVSAEVS": {"-": ["21"], "N": ["12"]},
    "MVSCA": {"-": ["11"]},
    "MVSCARIVS": {"-": ["21"]},
    "MVSCIPVLVM": {"-": ["12"]},
    "MVSCOSVS": {"-": ["21"]},
    "MVSCVLVS": {"-": ["12"]},
    "MVSCVS": {"-": ["12"]},
    "MVSICA": {"1": ["12", "11"], "2": ["17"]},
    "MVSICVS": {"1": ["12"], "2": ["21"]},
    "MVSIMO": {"-": ["13"]},
    "MVSSO": {"-": ["51"]},
    "MVSTACEVM": {"-": ["12"]},
    "MVSTACEVS": {"-": ["12"]},
    "MVSTARIVS": {"-": ["21"]},
    "MVSTELA": {"-": ["11"], "N": ["11"]},
    "MVSTEVS": {"-": ["21"]},
    "MVSTIVS": {"N": ["12"]},
    "MVSTVM": {"-": ["12"]},
    "MVSTVS": {"-": ["21"]},
    "MVTABILIS": {"-": ["24"]},
    "MVTABILITAS": {"-": ["13"]},
    "MVTATIO": {"-": ["13"]},
    "MVTILO": {"-": ["51"]},
    "MVTILVS": {"-": ["21"], "N": ["12"]},
    "MVTIO": {"-": ["54"]},
    "MVTO": {"1": ["13"], "2": ["51"]},
    "MVTVLVS": {"-": ["12"]},
    "MVTVO": {"1": ["51"], "2": ["60"]},
    "MVTVOR": {"-": ["5J"]},
    "MVTVS": {"-": ["21"], "N": ["12"], "A": ["21"]},
    "MVTVVM": {"-": ["12"]},
    "MVTVVS": {"-": ["21"]},
    "MVTYCENSIS": {"N": ["24"], "A": ["24"]},
    "MYNDVS": {"N": ["12"]},
    "MYOPARO": {"-": ["17"]},
    "MYRICE": {"-": ["17"]},
    "MYRO": {"N": ["17"]},
    "MYRRHA": {"-": ["11"], "N": ["11"]},
    "MYRRHEVS": {"-": ["21"]},
    "MYRRHINA": {"1": ["11"], "2": ["12"], "N": ["11"]},
    "MYRTA": {"-": ["11"]},
    "MYRTETVM": {"-": ["12"]},
    "MYRTEVS": {"-": ["21"]},
    "MYRTVM": {"-": ["12"]},
    "MYRTVS": {"-": ["16"]},
    "MYSTA": {"-": ["11"]},
    "MYSTAGOGVS": {"-": ["12"]},
    "MYSTERIVM": {"-": ["12"]},
    "MYSTICVS": {"-": ["21"]},
    "MYTILENAE": {"N": ["17"]},
    "NABLIA": {"-": ["16"]},
    "NAEVIVS": {"A": ["11"], "N": ["12"]},
    "NAEVVS": {"-": ["12"]},
    "NAM": {"-": ["81"]},
    "NAMQVE": {"-": ["81"]},
    "NANCISCOR": {"-": ["5L"]},
    "NANVS": {"-": ["12"]},
    "NAOS": {"-": ["17"]},
    "NARCISSVS": {"-": ["12"], "N": ["12"]},
    "NARDVM": {"-": ["12"]},
    "NARDVS": {"-": ["12"]},
    "NARIS": {"-": ["13"]},
    "NARRATIO": {"-": ["13"]},
    "NARRO": {"-": ["51"]},
    "NASCENTIA": {"-": ["11"]},
    "NASCOR": {"-": ["5L"]},
    "NASITERNA": {"-": ["11"]},
    "NASSA": {"-": ["11"]},
    "NASVS": {"-": ["12"]},
    "NASVTE": {"-": ["60"]},
    "NASVTVS": {"-": ["21"]},
    "NATA": {"-": ["11"]},
    "NATALES": {"-": ["13"]},
    "NATALICIVS": {"-": ["21"]},
    "NATALIS": {"1": ["13"], "2": ["24"], "N": ["13"]},
    "NATATOR": {"-": ["13"]},
    "NATI": {"-": ["12"]},
    "NATINOR": {"-": ["5J"]},
    "NATIO": {"1": ["13"], "2": ["13"]},
    "NATIS": {"-": ["13"]},
    "NATIVVS": {"-": ["21"]},
    "NATO": {"-": ["51"]},
    "NATRIX": {"-": ["13"]},
    "NATV": {"-": ["14"]},
    "NATVRA": {"-": ["11"], "N": ["11"]},
    "NATVRALIA": {"-": ["13"]},
    "NATVRALIS": {"-": ["24"]},
    "NATVRALITER": {"-": ["60"]},
    "NATVS": {"1": ["12"], "2": ["21"]},
    "NAVALE": {"-": ["13"]},
    "NAVALIA": {"-": ["13"]},
    "NAVALIS": {"-": ["24"], "A": ["24"]},
    "NAVARCHVS": {"-": ["12"]},
    "NAVE": {"-": ["60"]},
    "NAVFRAGIVM": {"-": ["12"]},
    "NAVFRAGVS": {"1": ["12"], "2": ["21"]},
    "NAVICVLA": {"-": ["11"]},
    "NAVICVLARIA": {"-": ["11"]},
    "NAVICVLARIVS": {"1": ["12"], "2": ["21"]},
    "NAVIFRAGVS": {"-": ["21"]},
    "NAVIGABILIS": {"-": ["24"]},
    "NAVIGANS": {"-": ["13"]},
    "NAVIGATIO": {"-": ["13"]},
    "NAVIGER": {"-": ["21"]},
    "NAVIGIOLVM": {"-": ["12"]},
    "NAVIGIVM": {"-": ["12"]},
    "NAVIGO": {"-": ["51"]},
    "NAVIS": {"-": ["13"]},
    "NAVITA": {"-": ["11"]},
    "NAVITER": {"-": ["60"]},
    "NAVIVS": {"N": ["12"]},
    "NAVLVM": {"-": ["12"]},
    "NAVMACHIA": {"-": ["11"]},
    "NAVO": {"-": ["51"]},
    "NAVPACTVS": {"N": ["17", "12"]},
    "NAVSEA": {"-": ["11"]},
    "NAVSEABVNDVS": {"-": ["21"]},
    "NAVSEATOR": {"-": ["13"]},
    "NAVSEO": {"-": ["51"]},
    "NAVTA": {"-": ["11"]},
    "NAVTICI": {"-": ["12"]},
    "NAVTICVS": {"-": ["21"]},
    "NAVTIVS": {"N": ["12"]},
    "NAVVS": {"-": ["21"]},
    "NE": {"1": ["60"], "2": ["67"], "3": ["68"], "4": ["82"]},
    "NEAPOLIS": {"N": ["17"]},
    "NEBVLA": {"-": ["11"]},
    "NEBVLO": {"-": ["13"]},
    "NEBVLOSVS": {"-": ["21"]},
    "NEC": {"1": ["68"], "2": ["81"]},
    "NECDVM": {"1": ["68"], "2": ["81"]},
    "NECESSARIA": {"-": ["11"]},
    "NECESSARIO": {"-": ["60"]},
    "NECESSARIVS": {"1": ["12"], "2": ["21"]},
    "NECESSE": {"-": ["26"]},
    "NECESSIS": {"-": ["13"]},
    "NECESSITAS": {"-": ["13"], "N": ["13"]},
    "NECESSITVDO": {"-": ["13"]},
    "NECNE": {"-": ["69", "6-"]},
    "NECNON": {"-": ["81"]},
    "NECO": {"-": ["51"]},
    "NECOPINANS": {"-": ["25"]},
    "NECOPINATO": {"-": ["60"]},
    "NECOPINATVS": {"-": ["21"]},
    "NECTAR": {"-": ["13"]},
    "NECTO": {"-": ["53"]},
    "NECVBI": {"-": ["82"]},
    "NEDVM": {"1": ["60"], "2": ["82"]},
    "NEFANDVS": {"-": ["21"]},
    "NEFARIE": {"-": ["60"]},
    "NEFARIVM": {"-": ["12"]},
    "NEFARIVS": {"-": ["21"]},
    "NEFAS": {"-": ["16"]},
    "NEFASTVM": {"-": ["12"]},
    "NEFASTVS": {"-": ["21"]},
    "NEGITO": {"-": ["51"]},
    "NEGLECTVS": {"1": ["14"], "2": ["21"]},
    "NEGLIGENS": {"-": ["25"]},
    "NEGLIGENTER": {"-": ["60"]},
    "NEGLIGENTIA": {"-": ["11"]},
    "NEGLIGO": {"-": ["53"]},
    "NEGO": {"-": ["51"]},
    "NEGOTIALIS": {"-": ["24"]},
    "NEGOTIATIO": {"-": ["13"]},
    "NEGOTIATOR": {"-": ["13"]},
    "NEGOTIOR": {"-": ["5J"]},
    "NEGOTIOSVS": {"-": ["21"]},
    "NEGOTIVM": {"-": ["12"]},
    "NEMO": {"-": ["48"]},
    "NEMORALIS": {"-": ["24"]},
    "NEMORIVAGVS": {"-": ["21"]},
    "NEMOROSVS": {"-": ["21"]},
    "N": {"-": ["60"]},
    "NEMVS": {"-": ["13"]},
    "NENIA": {"-": ["11"]},
    "NENVM": {"-": ["68"]},
    "NEO": {"-": ["52"]},
    "NEPOS": {"1": ["13"], "2": ["13"], "N": ["12"]},
    "NEPOTOR": {"-": ["5J"]},
    "NEPTIS": {"-": ["13"]},
    "NEPTVNVS": {"N": ["12"]},
    "NEQVAM": {"1": ["16"], "2": ["26"]},
    "NEQVAQVAM": {"-": ["68"]},
    "NEQVE": {"-": ["81"]},
    "NEQVEO": {"-": ["56"]},
    "NEQVIQVAM": {"-": ["60"]},
    "NEQVIS": {"-": ["48"]},
    "NEQVITER": {"-": ["60"]},
    "NEQVITIA": {"-": ["11"]},
    "NEQVITIES": {"-": ["15"]},
    "NEQVO": {"-": ["60"]},
    "NERO": {"N": ["13"]},
    "NERVICVS": {"-": ["21"], "A": ["21"]},
    "NERVOSVS": {"-": ["21"]},
    "NERVVLVS": {"-": ["12"]},
    "NERVVS": {"-": ["12"]},
    "NESCIO": {"-": ["54"]},
    "NESCIOQVA": {"-": ["60"]},
    "NESCIOQVANTVS": {"-": ["60"]},
    "NESCIOQVIS": {"-": ["48"]},
    "NESCIOQVOT": {"-": ["48"]},
    "NESCIVS": {"-": ["21"]},
    "NESTOR": {"-": ["17"], "N": ["13"]},
    "NETE": {"-": ["17"]},
    "NETINI": {"N": ["12"]},
    "NETVM": {"N": ["12"]},
    "NEVE": {"1": ["81"], "2": ["82"]},
    "NEVTER": {"-": ["48"]},
    "NEX": {"-": ["13"]},
    "NEXILIS": {"-": ["24"]},
    "NEXVS": {"-": ["14"]},
    "NI": {"1": ["68"], "2": ["82"]},
    "NICASIO": {"N": ["17"]},
    "NICE": {"N": ["17"]},
    "NICETERIVM": {"-": ["12"]},
    "NICO": {"N": ["13"]},
    "NICOMEDES": {"N": ["17"]},
    "NICTO": {"-": ["51"]},
    "NIDEO": {"-": ["52"]},
    "NIDESCO": {"-": ["53"]},
    "NIDIFICVS": {"-": ["21"]},
    "NIDOR": {"-": ["13"]},
    "NIDVS": {"-": ["12"]},
    "NIGER": {"-": ["21"], "N": ["12"]},
    "NIGRANS": {"-": ["25"]},
    "NIGRESCO": {"-": ["53"]},
    "NIGRO": {"-": ["51"]},
    "NIGROR": {"-": ["13"]},
    "NIHIL": {"-": ["48", "16"]},
    "NIHILDVM": {"-": ["60"]},
    "NIHILOMINVS": {"-": ["68"]},
    "NIHILOSETIVS": {"-": ["68"]},
    "NIHILVM": {"-": ["12"]},
    "NILVS": {"N": ["12"]},
    "NIMBOSVS": {"-": ["21"]},
    "NIMBVS": {"-": ["12"]},
    "NIMIO": {"-": ["60"]},
    "NIMIOPERE": {"-": ["60"]},
    "NIMIRVM": {"-": ["60"]},
    "NIMIS": {"-": ["60"]},
    "NIMIVM": {"1": ["12"], "2": ["60"]},
    "NIMIVS": {"-": ["21"]},
    "NINGO": {"-": ["53"]},
    "NINGVIS": {"-": ["13"]},
    "NISI": {"-": ["82"]},
    "NISVS": {"-": ["14"], "N": ["12"]},
    "NITENS": {"-": ["25"]},
    "NITEO": {"-": ["52"]},
    "NITESCO": {"-": ["53"]},
    "NITIDVS": {"-": ["21"]},
    "NITOR": {"1": ["13"], "2": ["5L"]},
    "NITROSVS": {"-": ["21"]},
    "NITRVM": {"-": ["12"]},
    "NIVALIS": {"-": ["24"]},
    "NIVENDIALIS": {"-": ["24"]},
    "NIVEVS": {"-": ["21"]},
    "NIVOSVS": {"-": ["21"]},
    "NIX": {"-": ["13"]},
    "NIXOR": {"-": ["5J"]},
    "NO": {"-": ["51"]},
    "NOBILIS": {"1": ["13"], "2": ["24"]},
    "NOBILITAS": {"-": ["13"]},
    "NOBILITER": {"-": ["60"]},
    "NOBILITO": {"-": ["51"]},
    "NOCENS": {"-": ["25"]},
    "NOCEO": {"-": ["52"]},
    "NOCTE": {"-": ["60"]},
    "NOCTIVAGVS": {"-": ["21"]},
    "NOCTV": {"-": ["60"]},
    "NOCTVA": {"-": ["11"]},
    "NOCTVRNVS": {"-": ["21"], "N": ["12"], "A": ["21"]},
    "NODATIO": {"-": ["13"]},
    "NODO": {"-": ["51"]},
    "NODOSVS": {"-": ["21"]},
    "NODVS": {"-": ["12"]},
    "NOLO": {"-": ["56"]},
    "NOMEN": {"-": ["13"]},
    "NOMENCLATIO": {"-": ["13"]},
    "NOMENCLATOR": {"-": ["13"]},
    "NOMENCLATVRA": {"-": ["11"]},
    "NOMENTVM": {"N": ["12"]},
    "NOMINATIM": {"-": ["60"]},
    "NOMINATIO": {"-": ["13"]},
    "NOMINATVS": {"1": ["14"], "2": ["21"], "N": ["12"]},
    "NOMINITO": {"-": ["51"]},
    "NOMINO": {"-": ["51"]},
    "NOMISMA": {"-": ["17"]},
    "NON": {"-": ["68"]},
    "NONAE": {"-": ["11"]},
    "NONAGENI": {"-": ["33"]},
    "NONAGESIMVS": {"-": ["32"]},
    "NONAGIES": {"-": ["36"]},
    "NONAGINTA": {"-": ["31"]},
    "NONANVS": {"-": ["21"]},
    "NONARIVS": {"-": ["21"]},
    "NONDVM": {"-": ["68"]},
    "NONGENTI": {"-": ["31"]},
    "NONNE": {"-": ["69", "6-"]},
    "NONNVLLVS": {"-": ["48"]},
    "NONNVMQVAM": {"-": ["60"]},
    "NONVS": {"-": ["32"]},
    "NORBA": {"N": ["11"]},
    "NORBANVS": {"N": ["12"]},
    "NORMA": {"-": ["11"]},
    "NOS": {"-": ["41"]},
    "NOSCITO": {"-": ["51"]},
    "NOSCO": {"-": ["53"]},
    "NOSTER": {"-": ["42"]},
    "NOSTRAS": {"-": ["25"]},
    "NOSTRI": {"-": ["42", "12"]},
    "NOTA": {"-": ["11"]},
    "NOTABILIS": {"-": ["24"]},
    "NOTATIO": {"-": ["13"]},
    "NOTESCO": {"-": ["53"]},
    "NOTHVS": {"-": ["21"], "N": ["12"]},
    "NOTI": {"-": ["12"], "N": ["12"]},
    "NOTIO": {"-": ["13"]},
    "NOTITIA": {"-": ["11"]},
    "NOTITIES": {"-": ["15"]},
    "NOTO": {"-": ["51"]},
    "NOTOR": {"-": ["13"]},
    "NOTVS": {"1": ["17"], "2": ["21"], "N": ["12"]},
    "NOVALE": {"-": ["13"]},
    "NOVALIS": {"-": ["24"]},
    "NOVATILLA": {"N": ["11"]},
    "NOVE": {"-": ["60"]},
    "NOVELLA": {"-": ["11"]},
    "NOVELLVS": {"-": ["21"], "N": ["12"]},
    "NOVEM": {"-": ["31"], "N": ["16"]},
    "NOVEMBER": {"A": ["23"], "N": ["13"]},
    "NOVENDIALIS": {"-": ["24"]},
    "NOVENVS": {"-": ["33"]},
    "NOVERCA": {"-": ["11"]},
    "NOVERCALIS": {"-": ["24"]},
    "NOVICIVS": {"-": ["21"]},
    "NOVIES": {"-": ["36"]},
    "NOVITAS": {"-": ["13"]},
    "NOVO": {"-": ["51"]},
    "NOVVMCOMVM": {"-": ["21"], "N": ["12"]},
    "NOVVS": {"-": ["21"], "A": ["21"]},
    "NOX": {"-": ["13"], "N": ["16"]},
    "NOXA": {"-": ["11"]},
    "NOXIA": {"-": ["11"]},
    "NOXIOSVS": {"-": ["21"]},
    "NOXIVS": {"-": ["21"]},
    "NVBES": {"-": ["13"]},
    "NVBIFER": {"-": ["21"]},
    "NVBILA": {"-": ["12"]},
    "NVBILIS": {"-": ["24"]},
    "NVBILO": {"-": ["51"]},
    "NVBILVM": {"-": ["12"]},
    "NVBILVS": {"-": ["21"]},
    "NVBO": {"-": ["53"]},
    "NVCEVS": {"-": ["21"]},
    "NVCLEVS": {"-": ["12"]},
    "NVDIVS": {"-": ["16"]},
    "NVDIVSTERTIVS": {"-": ["60"]},
    "NVDO": {"-": ["51"]},
    "NVDVS": {"-": ["21"]},
    "NVGAE": {"-": ["11"]},
    "NVGATOR": {"-": ["13"]},
    "NVGATORIVS": {"-": ["21"]},
    "NVGAX": {"-": ["25"]},
    "NVGOR": {"-": ["5J"]},
    "NVLLATENVS": {"-": ["68"]},
    "NVLLVS": {"-": ["48"]},
    "NVM": {"-": ["67"]},
    "NVMA": {"N": ["11"]},
    "NVMANTIA": {"N": ["11"]},
    "NVMARIVS": {"-": ["21"]},
    "NVMATVS": {"-": ["21"]},
    "NVMEN": {"-": ["13"]},
    "NVMENIVS": {"N": ["12"]},
    "NVMERATIO": {"-": ["13"]},
    "NVMERATVM": {"-": ["12"]},
    "NVMERATVS": {"C": ["21"], "-": ["21"]},
    "NVMERO": {"1": ["51"], "2": ["60"]},
    "NVMEROSVS": {"-": ["21"]},
    "NVMERVS": {"-": ["12"]},
    "NVMICIVS": {"N": ["12"]},
    "NVMICVS": {"N": ["12"]},
    "NVMIDICVS": {"C": ["12"], "A": ["21"], "N": ["12"]},
    "NVMITORIVS": {"N": ["12"]},
    "NVMNE": {"-": ["67"]},
    "NVMQVAM": {"-": ["68"]},
    "NVMQVANDO": {"-": ["67"]},
    "NVMQVI": {"-": ["67"]},
    "NVMQVID": {"-": ["67"]},
    "NVMQVIDNAM": {"-": ["67"]},
    "NVMQVIS": {"-": ["47"]},
    "NVMVLVS": {"-": ["12"]},
    "NVMVS": {"-": ["12"]},
    "NVNC": {"-": ["60"]},
    "NVNCVPATIO": {"-": ["13"]},
    "NVNCVPO": {"-": ["51"]},
    "NVNDINAE": {"-": ["11"]},
    "NVNDINATIO": {"-": ["13"]},
    "NVNDINOR": {"-": ["5J"]},
    "NVNTIA": {"-": ["11"]},
    "NVNTIO": {"-": ["51"]},
    "NVNTIVM": {"-": ["12"]},
    "NVNTIVS": {"1": ["12"], "2": ["21"]},
    "NVO": {"-": ["53"]},
    "NVPER": {"-": ["60"]},
    "NVPTA": {"-": ["11"]},
    "NVPTIAE": {"-": ["11"]},
    "NVPTIALIS": {"-": ["24"]},
    "NVRVS": {"-": ["14"]},
    "NVSQVAM": {"-": ["68"]},
    "NVTATIO": {"-": ["13"]},
    "NVTO": {"-": ["51"]},
    "NVTRICIA": {"1": ["11"], "2": ["12"]},
    "NVTRICIVM": {"-": ["12"]},
    "NVTRICIVS": {"1": ["12"], "2": ["21"]},
    "NVTRICVLA": {"-": ["11"]},
    "NVTRIMENTVM": {"-": ["12"]},
    "NVTRIO": {"-": ["54"]},
    "NVTRIOR": {"-": ["5M"]},
    "NVTRIX": {"-": ["13"]},
    "NVTVS": {"-": ["14"]},
    "NVX": {"-": ["13"]},
    "NYMPHA": {"-": ["11"], "N": ["11"]},
    "NYMPHAEVM": {"N": ["12"]},
    "NYMPHE": {"-": ["17"]},
    "NYMPHO": {"N": ["17"]},
    "NYMPHODORVS": {"N": ["12"]},
    "NYSA": {"-": ["17"], "N": ["11"]},
    "O": {"-": ["90"]},
    "OB": {"-": ["70"]},
    "OBAERATVS": {"1": ["12"], "2": ["21"]},
    "OBAMBVLO": {"-": ["51"]},
    "OBARMO": {"-": ["51"]},
    "OBBA": {"-": ["11"]},
    "OBBRVTESCO": {"-": ["53"]},
    "OBDO": {"-": ["53"]},
    "OBDVCO": {"-": ["53"]},
    "OBDVCTIO": {"-": ["13"]},
    "OBDVRESCO": {"-": ["53"]},
    "OBDVRO": {"-": ["51"]},
    "OBEDIENS": {"-": ["25"]},
    "OBEDIENTER": {"-": ["60"]},
    "OBEDIO": {"-": ["54"]},
    "OBEO": {"1": ["56"]},
    "OBEQVITO": {"-": ["51"]},
    "OBERRO": {"2": ["51"]},
    "OBESVS": {"-": ["21"]},
    "OBEX": {"-": ["13"]},
    "OBHAEREO": {"-": ["52"]},
    "OBIACEO": {"-": ["52"]},
    "OBICIO": {"-": ["55"]},
    "OBIECTA": {"-": ["12"]},
    "OBIECTATIO": {"-": ["13"]},
    "OBIECTO": {"-": ["51"]},
    "OBIECTVM": {"-": ["12"]},
    "OBIECTVS": {"-": ["14"]},
    "OBIRASCOR": {"-": ["5L"]},
    "OBITER": {"-": ["60"]},
    "OBITVS": {"-": ["14"]},
    "OBIVRGATIO": {"-": ["13"]},
    "OBIVRGATOR": {"-": ["13"]},
    "OBIVRGO": {"-": ["51"]},
    "OBLATRO": {"-": ["51"]},
    "OBLECTAMENTVM": {"-": ["12"]},
    "OBLECTATIO": {"-": ["13"]},
    "OBLECTO": {"-": ["51"]},
    "OBLENIO": {"-": ["54"]},
    "OBLIDO": {"-": ["53"]},
    "OBLIGATVS": {"-": ["21"]},
    "OBLIGO": {"2": ["51"]},
    "OBLIMO": {"-": ["51"]},
    "OBLINO": {"-": ["53"]},
    "OBLIQVE": {"-": ["60"]},
    "OBLIQVO": {"1": ["51"], "2": ["60"]},
    "OBLIQVVM": {"1": ["12"], "2": ["60"]},
    "OBLIQVVS": {"-": ["21"]},
    "OBLITESCO": {"-": ["53"]},
    "OBLITTERO": {"-": ["51"]},
    "OBLIVIO": {"-": ["13"], "N": ["13"]},
    "OBLIVIOSVS": {"-": ["21"]},
    "OBLIVISCOR": {"-": ["5L"]},
    "OBLIVIVM": {"-": ["12"]},
    "OBLONGVS": {"-": ["21"]},
    "OBLOQVOR": {"-": ["5L"]},
    "OBLVCTOR": {"-": ["5J"]},
    "OBMOLIOR": {"-": ["5M"]},
    "OBMOVEO": {"-": ["52"]},
    "OBMVRMVRO": {"-": ["51"]},
    "OBMVTESCO": {"-": ["53"]},
    "OBNITOR": {"2": ["5L"]},
    "OBNIXVM": {"-": ["60"]},
    "OBNIXVS": {"-": ["21"]},
    "OBNOXIVS": {"-": ["21"]},
    "OBNVBO": {"-": ["53"]},
    "OBOLVS": {"-": ["12"]},
    "OBORIOR": {"-": ["5M"]},
    "OBREPO": {"-": ["53"]},
    "OBRETIO": {"-": ["54"]},
    "OBRIGESCO": {"-": ["53"]},
    "OBRVO": {"-": ["53"]},
    "OBRVSSA": {"-": ["11"]},
    "OBSAEPIO": {"-": ["54"]},
    "OBSCENA": {"-": ["12"]},
    "OBSCENE": {"-": ["6&"]},
    "OBSCENVM": {"-": ["12"]},
    "OBSCENVS": {"-": ["21"]},
    "OBSCVRATIO": {"-": ["13"]},
    "OBSCVRE": {"-": ["60"]},
    "OBSCVRITAS": {"-": ["13"]},
    "OBSCVRO": {"-": ["51"]},
    "OBSCVRVM": {"1": ["12"], "2": ["60"]},
    "OBSCVRVS": {"-": ["21"]},
    "OBSECRO": {"-": ["51"]},
    "OBSECVNDO": {"-": ["51"]},
    "OBSEDEO": {"-": ["52"]},
    "OBSEQVENS": {"O": ["25"], "A": ["22"], "-": ["21"]},
    "OBSEQVENTIA": {"-": ["11"]},
    "OBSEQVIVM": {"-": ["12"]},
    "OBSEQVOR": {"-": ["5L"]},
    "OBSERO": {"1": ["51"], "3": ["53", "51"]},
    "OBSERVABILIS": {"-": ["24"]},
    "OBSERVATIO": {"-": ["13"]},
    "OBSERVATOR": {"-": ["13"]},
    "OBSERVO": {"-": ["51"]},
    "OBSES": {"-": ["13"]},
    "OBSESSIO": {"-": ["13"]},
    "OBSESSOR": {"-": ["13"]},
    "OBSIDEO": {"-": ["52"]},
    "OBSIDIO": {"-": ["13"]},
    "OBSIDIVM": {"1": ["12"], "2": ["12"]},
    "OBSIDO": {"-": ["53"]},
    "OBSIGNO": {"-": ["51"]},
    "OBSISTO": {"-": ["53"]},
    "OBSOLEFACIO": {"-": ["55"]},
    "OBSOLEFIO": {"-": ["56"]},
    "OBSOLESCO": {"-": ["53"]},
    "OBSOLETE": {"-": ["60"]},
    "OBSOLETVS": {"-": ["21"]},
    "OBSOLIDATVS": {"-": ["21"]},
    "OBSONATOR": {"-": ["13"]},
    "OBSONITO": {"-": ["51"]},
    "OBSONIVM": {"-": ["12"]},
    "OBSONO": {"1": ["51"]},
    "OBSTACVLVM": {"-": ["12"]},
    "OBSTANTIA": {"1": ["11"], "2": ["13"]},
    "OBSTETRIX": {"-": ["13"]},
    "OBSTINATE": {"-": ["60"]},
    "OBSTINATIO": {"-": ["13"]},
    "OBSTINATVS": {"-": ["21"]},
    "OBSTIPVS": {"-": ["21"]},
    "OBSTO": {"-": ["51"]},
    "OBSTREPO": {"-": ["53"]},
    "OBSTRIGILLO": {"-": ["51"]},
    "OBSTRINGO": {"-": ["53"]},
    "OBSTRVO": {"-": ["53"]},
    "OBSTVPEFACIO": {"-": ["55"]},
    "OBSTVPEFIO": {"-": ["56"]},
    "OBSTVPESCO": {"-": ["53"]},
    "OBSVM": {"1": ["56"]},
    "OBSVRDESCO": {"-": ["53"]},
    "OBTEGO": {"-": ["53"]},
    "OBTEMPERO": {"-": ["51"]},
    "OBTENDO": {"-": ["53"]},
    "OBTENTVS": {"-": ["14"]},
    "OBTERO": {"-": ["53"]},
    "OBTESTATIO": {"-": ["13"]},
    "OBTESTOR": {"-": ["5J"]},
    "OBTINEO": {"-": ["52"]},
    "OBTINGO": {"-": ["53"]},
    "OBTINGVO": {"-": ["53"]},
    "OBTORPESCO": {"-": ["53"]},
    "OBTORQVEO": {"-": ["52"]},
    "OBTRECTATIO": {"-": ["13"]},
    "OBTRECTATOR": {"-": ["13"]},
    "OBTRECTO": {"-": ["51"]},
    "OBTRVDO": {"-": ["53"]},
    "OBTRVNCO": {"-": ["51"]},
    "OBTVNDO": {"-": ["53"]},
    "OBTVRBO": {"2": ["51"]},
    "OBTVRGESCO": {"-": ["53"]},
    "OBTVRO": {"-": ["51"]},
    "OBTVSVS": {"-": ["21"]},
    "OBTVTVS": {"-": ["14"]},
    "OBVCVLA": {"-": ["11"], "N": ["11"]},
    "OBVENIO": {"-": ["54"]},
    "OBVERSO": {"-": ["5J"]},
    "OBVERSOR": {"-": ["5J"]},
    "OBVERSVS": {"-": ["21"]},
    "OBVERTO": {"-": ["53"]},
    "OBVIAM": {"-": ["60"]},
    "OBVIVS": {"-": ["21"]},
    "OBVMBRO": {"-": ["51"]},
    "OBVNCVS": {"-": ["21"]},
    "OBVOLVO": {"-": ["53"]},
    "OCCAECO": {"-": ["51"]},
    "OCCALLESCO": {"-": ["53"]},
    "OCCANO": {"-": ["53"]},
    "OCCASIO": {"-": ["13"]},
    "OCCASVS": {"-": ["14"]},
    "OCCEDO": {"1": ["53"]},
    "OCCIDENS": {"N": ["13"], "-": ["13"]},
    "OCCIDIO": {"-": ["13"]},
    "OCCIDO": {"1": ["53"], "2": ["53"]},
    "OCCIDVVS": {"-": ["21"]},
    "OCCIPIO": {"-": ["55"]},
    "OCCIPITIVM": {"-": ["12"]},
    "OCCIPVT": {"-": ["13"]},
    "OCCISIO": {"-": ["13"]},
    "OCCISVS": {"-": ["21"]},
    "OCCLVDO": {"-": ["53"]},
    "OCCO": {"-": ["51"]},
    "OCCVBO": {"-": ["51"]},
    "OCCVLCO": {"-": ["51"]},
    "OCCVLO": {"-": ["53"]},
    "OCCVLTATIO": {"-": ["13"]},
    "OCCVLTE": {"-": ["60"]},
    "OCCVLTO": {"-": ["51"]},
    "OCCVLTVM": {"-": ["12"]},
    "OCCVLTVS": {"-": ["21"]},
    "OCCVMBO": {"-": ["53"]},
    "OCCVPATIO": {"-": ["13"]},
    "OCCVPATVS": {"-": ["21"]},
    "OCCVPO": {"1": ["13"], "2": ["51"], "N": ["13"]},
    "OCCVRO": {"-": ["51"]},
    "OCCVRRO": {"-": ["53"]},
    "OCCVRSO": {"-": ["51"]},
    "OCCVRSVS": {"-": ["14"]},
    "OCEANVS": {"1": ["12"], "2": ["21"], "N": ["12"], "A": ["21"]},
    "OCELLVS": {"-": ["12"]},
    "OCHRA": {"-": ["17"]},
    "OCIMVM": {"-": ["12"]},
    "OCINVM": {"-": ["12"]},
    "OCIOR": {"-": ["2&"]},
    "OCITER": {"-": ["60"]},
    "OCLIFERIVS": {"-": ["21"]},
    "OCREA": {"-": ["11"], "N": ["11"]},
    "OCREATVS": {"-": ["21"]},
    "OCTACHORDOS": {"-": ["27"]},
    "OCTANS": {"-": ["13"]},
    "OCTASTYLOS": {"-": ["27"]},
    "OCTAVAE": {"-": ["11"]},
    "OCTAVIA": {"N": ["11"]},
    "OCTAVIANVS": {"C": ["12"], "N": ["21"], "A": ["21"]},
    "OCTAVIVS": {"N": ["12"]},
    "OCTAVVS": {"-": ["32"], "N": ["12"]},
    "OCTINGENTESIMVS": {"-": ["32"]},
    "OCTINGENTI": {"-": ["31"]},
    "OCTO": {"-": ["31"]},
    "OCTOBER": {"A": ["23"], "N": ["13"]},
    "OCTOGENARIVS": {"-": ["21"]},
    "OCTOGESIMVS": {"-": ["32"]},
    "OCTOGINTA": {"-": ["31"]},
    "OCTOGONOS": {"-": ["27"]},
    "OCTONI": {"-": ["33"]},
    "OCTOPHORON": {"-": ["17"]},
    "OCTVPLVM": {"-": ["12"]},
    "OCTVSSIS": {"-": ["13"]},
    "OCVLVS": {"-": ["12"]},
    "ODEVM": {"-": ["12"]},
    "ODI": {"-": ["53"]},
    "ODIOSVS": {"-": ["21"]},
    "ODIVM": {"-": ["12"]},
    "ODOR": {"-": ["13"]},
    "ODORATVS": {"1": ["14"], "2": ["21"]},
    "ODORIFER": {"-": ["21"]},
    "ODORO": {"-": ["51"]},
    "ODOROR": {"-": ["5J"]},
    "ODORVS": {"-": ["21"]},
    "ODYSSEA": {"N": ["11"]},
    "OEAX": {"-": ["17"]},
    "OECONOMIA": {"-": ["17"]},
    "OECVS": {"-": ["17"]},
    "OENOPHORVM": {"-": ["12"]},
    "OESTRVS": {"-": ["17"]},
    "OESYPVM": {"-": ["12"]},
    "OETA": {"N": ["17"]},
    "OFELLA": {"-": ["11"]},
    "OFFA": {"-": ["11"]},
    "OFFENDO": {"-": ["53"]},
    "OFFENSA": {"-": ["11"]},
    "OFFENSATIO": {"-": ["13"]},
    "OFFENSIO": {"-": ["13"]},
    "OFFENSO": {"-": ["51"]},
    "OFFENSVM": {"-": ["12"]},
    "OFFENSVS": {"1": ["14"], "2": ["21"]},
    "OFFERO": {"-": ["56"]},
    "OFFICINA": {"-": ["11"]},
    "OFFICINATOR": {"-": ["13"]},
    "OFFICIO": {"-": ["55"]},
    "OFFICIOSE": {"-": ["60"]},
    "OFFICIOSVS": {"1": ["12"], "2": ["21"]},
    "OFFICIVM": {"-": ["12"]},
    "OFFIGO": {"-": ["53"]},
    "OFFIRMO": {"-": ["51"]},
    "OFFOCO": {"-": ["51"]},
    "OFFVNDO": {"2": ["53"]},
    "OH": {"-": ["90"]},
    "OHE": {"-": ["90"]},
    "OLEA": {"-": ["11"]},
    "OLEAGINEVS": {"-": ["21"]},
    "OLEAGINVS": {"-": ["21"]},
    "OLEARIVS": {"1": ["12"], "2": ["21"]},
    "OLEASTER": {"-": ["12"]},
    "OLEITAS": {"-": ["13"]},
    "OLENS": {"-": ["25"]},
    "OLEO": {"-": ["52"]},
    "OLESCO": {"-": ["53"]},
    "OLETVM": {"1": ["12"], "2": ["12"]},
    "OLEVM": {"-": ["12"]},
    "OLFACIO": {"-": ["55"]},
    "OLFACTVS": {"-": ["14"]},
    "OLIDVS": {"-": ["21"]},
    "OLIM": {"-": ["60"]},
    "OLITIO": {"-": ["13"]},
    "OLITOR": {"-": ["13"]},
    "OLITORIVS": {"-": ["21"]},
    "OLIVA": {"-": ["11"]},
    "OLIVETVM": {"-": ["12"]},
    "OLIVVM": {"-": ["12"]},
    "OLOR": {"1": ["13"], "2": ["13"]},
    "OLORINVS": {"-": ["21"]},
    "OLVS": {"-": ["13"]},
    "OLVSCVLVM": {"-": ["12"]},
    "OLYMPICVS": {"-": ["21"], "A": ["21"]},
    "OLYMPIVS": {"A": ["21"]},
    "OLYMPVS": {"1N": ["12"], "2N": ["12"], "3N": ["12"], "N": ["12"]},
    "OMASVM": {"-": ["12"]},
    "OMEN": {"-": ["13"]},
    "OMENTVM": {"-": ["12"]},
    "OMINOR": {"-": ["5J"]},
    "OMITTO": {"-": ["53"]},
    "OMNIGENVS": {"1": ["21"], "2": ["60"]},
    "OMNIMODIS": {"-": ["60"]},
    "OMNIMODVS": {"-": ["21"]},
    "OMNINO": {"-": ["60"]},
    "OMNIPARENS": {"-": ["25"]},
    "OMNIPATENS": {"-": ["25"]},
    "OMNIPOTENS": {"-": ["25"], "A": ["22"]},
    "OMNIS": {"-": ["48"]},
    "OMNITVENS": {"-": ["25"]},
    "OMNIVOLVS": {"-": ["21"]},
    "ONAGER": {"-": ["12"]},
    "ONASVS": {"N": ["12"]},
    "ONERARIA": {"-": ["11"]},
    "ONERARIVS": {"-": ["21"]},
    "ONERATVS": {"-": ["21"]},
    "ONERO": {"-": ["51"]},
    "ONEROSVS": {"-": ["21"]},
    "ONVS": {"-": ["13"]},
    "ONVSTVS": {"-": ["21"]},
    "ONYX": {"-": ["17"]},
    "OPACITAS": {"-": ["13"]},
    "OPACO": {"-": ["51"]},
    "OPACVM": {"-": ["12"]},
    "OPACVS": {"-": ["21"]},
    "OPE": {"-": ["17"]},
    "OPELLA": {"-": ["11"]},
    "OPERA": {"-": ["11"]},
    "OPERANS": {"-": ["25"]},
    "OPERARIVS": {"1": ["12"], "2": ["12"]},
    "OPERATIO": {"-": ["13"]},
    "OPERATVS": {"-": ["21"]},
    "OPERCVLVM": {"-": ["12"]},
    "OPERIMENTVM": {"-": ["12"]},
    "OPERIO": {"-": ["54"]},
    "OPEROR": {"-": ["5J"]},
    "OPEROSE": {"-": ["60"]},
    "OPEROSVS": {"-": ["21"]},
    "OPERTORIVM": {"-": ["12"]},
    "OPERTVM": {"-": ["12"]},
    "OPERTVS": {"1": ["14"], "2": ["21"]},
    "OPICVS": {"-": ["21"]},
    "OPIFEX": {"-": ["13"]},
    "OPILIO": {"-": ["13"]},
    "OPIMIVS": {"N": ["12"]},
    "OPIMVS": {"-": ["21"]},
    "OPINATVS": {"-": ["14"]},
    "OPINIO": {"-": ["13"]},
    "OPINOR": {"-": ["5J"]},
    "OPITER": {"N": ["13"]},
    "OPITVLOR": {"-": ["5J"]},
    "OPOBALSAMVM": {"-": ["12"]},
    "OPORTET": {"-": ["52"]},
    "OPPEDO": {"-": ["53"]},
    "OPPERIOR": {"-": ["5M"]},
    "OPPETO": {"-": ["53"]},
    "OPPIA": {"N": ["11"]},
    "OPPICO": {"-": ["51"]},
    "OPPIDANI": {"-": ["12"]},
    "OPPIDANVS": {"1": ["12"], "2": ["21"]},
    "OPPIDO": {"-": ["60"]},
    "OPPIDVLVM": {"-": ["12"]},
    "OPPIDVM": {"-": ["12"]},
    "OPPIGNERO": {"-": ["51"]},
    "OPPILO": {"-": ["51"]},
    "OPPLEO": {"-": ["52"]},
    "OPPONO": {"-": ["53"]},
    "OPPORTVNE": {"-": ["60"]},
    "OPPORTVNITAS": {"-": ["13"]},
    "OPPORTVNVS": {"-": ["21"]},
    "OPPOSITVS": {"1": ["14"], "2": ["21"]},
    "OPPRESSIO": {"-": ["13"]},
    "OPPRESSVS": {"-": ["14"]},
    "OPPRIMO": {"-": ["53"]},
    "OPPROBRIVM": {"-": ["12"]},
    "OPPVGNATIO": {"-": ["13"]},
    "OPPVGNATOR": {"-": ["13"]},
    "OPPVGNATORIVS": {"-": ["21"]},
    "OPPVGNO": {"-": ["51"]},
    "OPS": {"-": ["13"], "N": ["13"]},
    "OPTABILIS": {"-": ["24"]},
    "OPTATVM": {"-": ["12"]},
    "OPTATVS": {"-": ["21"]},
    "OPTICE": {"-": ["17"]},
    "OPTICOS": {"-": ["27"]},
    "OPTICVS": {"-": ["21"]},
    "OPTIMAS": {"-": ["25"]},
    "OPTIMATES": {"-": ["13"]},
    "OPTIO": {"1": ["13"], "2": ["13"]},
    "OPTIONATVS": {"-": ["14"]},
    "OPTIVVS": {"-": ["21"]},
    "OPTO": {"-": ["51"]},
    "OPVLENS": {"-": ["25"]},
    "OPVLENTER": {"-": ["60"]},
    "OPVLENTIA": {"-": ["11"]},
    "OPVLENTO": {"-": ["51"]},
    "OPVLENTVS": {"-": ["21"]},
    "OPVNTIVS": {"N": ["21"], "A": ["21"]},
    "OPVS": {"1": ["13"], "2": ["16"]},
    "OPVSCVLVM": {"-": ["12"]},
    "ORA": {"-": ["11"], "N": ["11"]},
    "ORACVLVM": {"-": ["12"]},
    "ORATIO": {"-": ["13"]},
    "ORATIVNCVLA": {"-": ["11"]},
    "ORATOR": {"-": ["13"]},
    "ORATORIE": {"-": ["60"]},
    "ORATORIVS": {"-": ["21"]},
    "ORATVS": {"-": ["14"]},
    "ORBATIO": {"-": ["13"]},
    "ORBICVLVS": {"-": ["12"]},
    "ORBIS": {"-": ["13"]},
    "ORBITA": {"-": ["11"]},
    "ORBITAS": {"-": ["13"]},
    "ORBO": {"-": ["51"]},
    "ORBVS": {"-": ["21"]},
    "ORCA": {"-": ["11"]},
    "ORCHESTRA": {"-": ["11"]},
    "ORCHITIS": {"-": ["13"]},
    "ORCHOMENOS": {"N": ["17"]},
    "ORCVLA": {"-": ["11"]},
    "ORCVS": {"N": ["12"]},
    "ORDINARIVS": {"-": ["21"]},
    "ORDINATE": {"-": ["60"]},
    "ORDINATIM": {"-": ["60"]},
    "ORDINATIO": {"-": ["13"]},
    "ORDINATOR": {"-": ["13"]},
    "ORDINATVS": {"-": ["21"]},
    "ORDINO": {"-": ["51"]},
    "ORDIOR": {"-": ["5M"]},
    "ORDO": {"-": ["13"]},
    "OREAE": {"-": ["11"]},
    "OREXIS": {"-": ["13"]},
    "ORGANICOS": {"-": ["60"]},
    "ORGANICVS": {"1": ["12"], "2": ["21"]},
    "ORGANVM": {"-": ["12"]},
    "ORGIA": {"-": ["12"]},
    "ORIENS": {"N": ["13"], "-": ["22"]},
    "ORIFICIVM": {"-": ["12"]},
    "ORIGANITVS": {"-": ["21"]},
    "ORIGO": {"-": ["13"], "N": ["13"]},
    "ORIOR": {"-": ["5M"]},
    "ORIVNDVS": {"-": ["21"]},
    "ORNAMENTVM": {"-": ["12"]},
    "ORNATE": {"-": ["60"]},
    "ORNATIO": {"-": ["13"]},
    "ORNATRIX": {"-": ["13"]},
    "ORNATVS": {"1": ["14"], "2": ["21"]},
    "ORNO": {"-": ["51"]},
    "ORNVS": {"-": ["12"]},
    "ORO": {"-": ["51"]},
    "ORPHEVS": {"N": ["17"], "A": ["21"]},
    "ORSA": {"-": ["12"]},
    "ORTHOGONIVS": {"-": ["21"]},
    "ORTHOGRAPHIA": {"-": ["11"]},
    "ORTHOS": {"-": ["27"]},
    "ORTHOSTATA": {"-": ["11"]},
    "ORTONA": {"N": ["11"]},
    "ORTVS": {"-": ["14"]},
    "ORYX": {"-": ["13"]},
    "ORYZA": {"-": ["11"]},
    "OS": {"1": ["13"], "2": ["13"]},
    "OSCEN": {"-": ["13"]},
    "OSCILLVM": {"1": ["12"], "2": ["12"]},
    "OSCITANS": {"-": ["25"]},
    "OSCITATIO": {"-": ["13"]},
    "OSCITO": {"-": ["51"]},
    "OSCITOR": {"-": ["5J"]},
    "OSCVLATIO": {"-": ["13"]},
    "OSCVLOR": {"-": ["5J"]},
    "OSCVLVM": {"-": ["12"]},
    "OSSEVS": {"-": ["21"]},
    "OSSICVLVM": {"-": ["12"]},
    "OSSIFRAGA": {"-": ["11"]},
    "OSSIFRAGVS": {"-": ["21"]},
    "OSTENDO": {"-": ["53"]},
    "OSTENTATIO": {"-": ["13"]},
    "OSTENTATOR": {"-": ["13"]},
    "OSTENTO": {"-": ["51"]},
    "OSTENTVM": {"-": ["12"]},
    "OSTENTVS": {"-": ["14"]},
    "OSTIA": {"N": ["11"]},
    "OSTIARIVM": {"-": ["12"]},
    "OSTIARIVS": {"-": ["12"]},
    "OSTIATIM": {"-": ["60"]},
    "OSTIVM": {"-": ["12"]},
    "OSTREA": {"-": ["11"]},
    "OSTREOSVS": {"-": ["21"]},
    "OSTREVM": {"-": ["12"]},
    "OSTRIFER": {"-": ["21"]},
    "OSTRINVS": {"-": ["21"]},
    "OSTRVM": {"-": ["12"]},
    "OTIOR": {"-": ["5J"]},
    "OTIOSE": {"-": ["60"]},
    "OTIOSVS": {"-": ["21"]},
    "OTIVM": {"-": ["12"]},
    "OVATIO": {"-": ["13"]},
    "OVILE": {"-": ["13"]},
    "OVILLVS": {"-": ["21"]},
    "OVIS": {"-": ["13"]},
    "OVO": {"-": ["51"]},
    "OVVM": {"-": ["12"]},
    "OXYMELI": {"-": ["13"]},
    "PABVLATIO": {"-": ["13"]},
    "PABVLATOR": {"-": ["13"]},
    "PABVLOR": {"-": ["5J"]},
    "PABVLVM": {"-": ["12"]},
    "PACATOR": {"-": ["13"]},
    "PACATVS": {"-": ["21"]},
    "PACHYNVM": {"N": ["17"]},
    "PACIFICO": {"-": ["51"]},
    "PACILIVS": {"N": ["12"]},
    "PACISCOR": {"-": ["5L"]},
    "PACO": {"-": ["51"]},
    "PACTIO": {"-": ["13"]},
    "PACTVM": {"-": ["12"]},
    "PACTVS": {"1": ["14"], "2": ["21"]},
    "PAEAN": {"-": ["17"], "N": ["17"]},
    "PAEDAGOGIVM": {"-": ["12"]},
    "PAEDAGOGVS": {"-": ["12"]},
    "PAEDICO": {"1": ["13"], "2": ["51"]},
    "PAEDOR": {"-": ["13"]},
    "PAENE": {"-": ["60"]},
    "PAENITENTIA": {"-": ["11"]},
    "PAENITEO": {"-": ["52"]},
    "PAENVLA": {"-": ["11"]},
    "PAENVLATVS": {"-": ["21"]},
    "PAETVS": {"-": ["21"], "N": ["12"]},
    "PAGANVS": {"1": ["12"], "2": ["21"]},
    "PAGINA": {"-": ["11"]},
    "PAGVS": {"-": ["12"]},
    "PALA": {"-": ["11"]},
    "PALAESTRA": {"-": ["11"]},
    "PALAESTRICVS": {"-": ["21"]},
    "PALAESTRITA": {"-": ["11"]},
    "PALAM": {"1": ["60"], "2": ["70"]},
    "PALATINVS": {"N": ["21"], "A": ["21"]},
    "PALATIO": {"-": ["13"]},
    "PALATIVM": {"-": ["12"], "N": ["12"]},
    "PALATVM": {"-": ["12"]},
    "PALEA": {"-": ["11"]},
    "PALEAR": {"-": ["13"]},
    "PALICANVS": {"N": ["12"]},
    "PALIMPSESTVS": {"-": ["17"]},
    "PALIVRVS": {"-": ["12"]},
    "PALLA": {"-": ["11"], "N": ["11"]},
    "PALLADIVM": {"-": ["12"], "N": ["12"]},
    "PALLANTEVM": {"N": ["12"]},
    "PALLANTIVM": {"N": ["12"]},
    "PALLAS": {"-": ["17"], "O": ["13"], "N": ["21"]},
    "PALLENS": {"-": ["25"]},
    "PALLEO": {"-": ["52"]},
    "PALLESCO": {"-": ["53"]},
    "PALLIATVS": {"-": ["21"]},
    "PALLIDVLVS": {"-": ["21"]},
    "PALLIDVS": {"-": ["21"]},
    "PALLIOLVM": {"-": ["12"]},
    "PALLIVM": {"-": ["12"]},
    "PALLOR": {"-": ["13"], "N": ["13"]},
    "PALMA": {"-": ["11"]},
    "PALMES": {"-": ["13"]},
    "PALMETVM": {"-": ["12"]},
    "PALMEVS": {"-": ["21"]},
    "PALMIFER": {"-": ["21"]},
    "PALMOSVS": {"-": ["21"]},
    "PALMVLA": {"-": ["11"]},
    "PALMVS": {"-": ["12"], "N": ["12"]},
    "PALOR": {"-": ["5J"]},
    "PALPEBRAE": {"-": ["11"]},
    "PALPITATIO": {"-": ["13"]},
    "PALPITO": {"-": ["51"]},
    "PALPO": {"1": ["13"], "2": ["51"]},
    "PALPOR": {"-": ["5J"]},
    "PALPVM": {"-": ["12"]},
    "PALVDAMENTVM": {"-": ["12"]},
    "PALVDATVS": {"-": ["21"]},
    "PALVMBES": {"-": ["13"]},
    "PALVMBVS": {"-": ["12"]},
    "PALVS": {"1": ["12"], "2": ["13"]},
    "PALVSTER": {"-": ["23"]},
    "PAMPHILVS": {"N": ["12"]},
    "PAMPHYLIA": {"N": ["11"]},
    "PAMPINEVS": {"-": ["21"]},
    "PAMPINO": {"-": ["51"]},
    "PAMPINVS": {"-": ["12"]},
    "PAN": {"N": ["17"]},
    "PANACEA": {"-": ["11"]},
    "PANAX": {"-": ["13"]},
    "PANCHRESTVS": {"-": ["21"]},
    "PANCRATION": {"-": ["17"]},
    "PANDATIO": {"-": ["13"]},
    "PANDO": {"1": ["51"], "2": ["53"]},
    "PANDVS": {"-": ["21"]},
    "PANGO": {"-": ["53"]},
    "PANHORMITANVS": {"N": ["21"], "A": ["21"]},
    "PANHORMVS": {"N": ["12"]},
    "PANICVM": {"-": ["12"]},
    "PANIS": {"-": ["13"]},
    "PANNICVLVS": {"-": ["12"]},
    "PANNOSVS": {"-": ["21"]},
    "PANNVCIVS": {"-": ["21"]},
    "PANNVS": {"-": ["12"]},
    "PANTHERA": {"1": ["11"]},
    "PANTOMIMA": {"-": ["11"]},
    "PANTOMIMICVS": {"-": ["21"]},
    "PANTOMIMVS": {"-": ["12"]},
    "PANVS": {"-": ["12"]},
    "PAPAE": {"-": ["90"]},
    "PAPAS": {"-": ["17"]},
    "PAPAVER": {"-": ["13"]},
    "PAPHLAGONIA": {"N": ["11"]},
    "PAPILLA": {"-": ["11"]},
    "PAPINIVS": {"N": ["12"]},
    "PAPIRIVS": {"N": ["12"], "A": ["21"]},
    "PAPPO": {"-": ["51"]},
    "PAPPVS": {"-": ["12"]},
    "PAPVLA": {"-": ["11"]},
    "PAPYRVS": {"-": ["12"]},
    "PAR": {"1": ["13"], "2": ["25"]},
    "PARABILIS": {"-": ["24"]},
    "PARABOLA": {"-": ["17"]},
    "PARADOXVM": {"-": ["17"]},
    "PARADROMIS": {"-": ["17"]},
    "PARAENETICOS": {"-": ["27"]},
    "PARALVS": {"N": ["12"]},
    "PARALYSIS": {"-": ["17"]},
    "PARAMESE": {"-": ["17"]},
    "PARANETE": {"-": ["17"]},
    "PARAPEGMA": {"-": ["17"]},
    "PARARIVS": {"-": ["12"]},
    "PARASITA": {"-": ["11"]},
    "PARASITVS": {"-": ["12"]},
    "PARASTAS": {"-": ["17"]},
    "PARASTATA": {"-": ["11"]},
    "PARASTATICA": {"-": ["11"]},
    "PARATA": {"-": ["12"]},
    "PARATE": {"-": ["60"]},
    "PARATIO": {"-": ["13"]},
    "PARATVS": {"1": ["14"], "2": ["21"]},
    "PARCA": {"N": ["11"]},
    "PARCE": {"-": ["60"]},
    "PARCENS": {"-": ["25"]},
    "PARCITAS": {"-": ["13"]},
    "PARCO": {"-": ["53"]},
    "PARCVS": {"-": ["21"]},
    "PARDALIS": {"-": ["13"]},
    "PARDVS": {"-": ["12"]},
    "PARENS": {"1": ["13"], "2": ["25", "22"]},
    "PARENTES": {"-": ["13"]},
    "PARENTO": {"-": ["51"]},
    "PAREO": {"-": ["52"]},
    "PARERGON": {"-": ["17"]},
    "PARGO": {"-": ["53"]},
    "PARHYPATE": {"-": ["17"]},
    "PARIES": {"-": ["13"]},
    "PARILIS": {"-": ["24"], "A": ["24"]},
    "PARIO": {"1": ["51"], "2": ["55"]},
    "PARITER": {"-": ["60"]},
    "PARMA": {"-": ["11"], "N": ["11"]},
    "PARMVLA": {"-": ["11"]},
    "PARO": {"1": ["13"], "2": ["51"]},
    "PAROCHVS": {"-": ["12"]},
    "PAROPSIS": {"-": ["13"]},
    "PAROTIS": {"-": ["17"]},
    "PARRA": {"-": ["11"]},
    "PARRICIDA": {"-": ["11"]},
    "PARRICIDIVM": {"-": ["12"]},
    "PARS": {"-": ["13"]},
    "PARSIMONIA": {"-": ["11"]},
    "PARTHENICE": {"-": ["17"]},
    "PARTHICVS": {"N": ["21"], "A": ["21"]},
    "PARTHVS": {"N": ["21"], "A": ["21"]},
    "PARTIARIVS": {"-": ["21"]},
    "PARTICEPS": {"1": ["13"], "2": ["22"]},
    "PARTICIPO": {"-": ["51"]},
    "PARTICVLA": {"-": ["11"]},
    "PARTICVLARIS": {"-": ["24"]},
    "PARTICVLATIM": {"-": ["60"]},
    "PARTIM": {"-": ["60"]},
    "PARTIO": {"1": ["13"], "2": ["54"]},
    "PARTIOR": {"-": ["5M"]},
    "PARTITIO": {"-": ["13"]},
    "PARTVRIO": {"-": ["54"]},
    "PARTVS": {"1": ["14"], "2": ["21"]},
    "PARVE": {"-": ["60"]},
    "PARVICAX": {"-": ["25"]},
    "PARVM": {"1": ["12"], "2": ["60"]},
    "PARVMPER": {"-": ["60"]},
    "PARVVLVS": {"-": ["21"]},
    "PARVVM": {"-": ["12"]},
    "PARVVS": {"1": ["12"], "2": ["21"]},
    "PASCALIS": {"-": ["24"]},
    "PASCEOLVS": {"-": ["12"]},
    "PASCO": {"-": ["53"]},
    "PASCOR": {"-": ["5L"]},
    "PASCVVM": {"-": ["12"]},
    "PASCVVS": {"-": ["21"]},
    "PASSER": {"-": ["13"]},
    "PASSIM": {"-": ["60"]},
    "PASSVM": {"-": ["12"]},
    "PASSVS": {"1": ["14"], "2": ["21"]},
    "PASTAS": {"-": ["17"]},
    "PASTILLVS": {"-": ["12"]},
    "PASTIO": {"-": ["13"]},
    "PASTOR": {"-": ["13"], "N": ["13"]},
    "PASTVS": {"-": ["14"]},
    "PATAVIVM": {"N": ["12"]},
    "PATEFACIO": {"-": ["55"]},
    "PATEFIO": {"-": ["56"]},
    "PATELLA": {"-": ["11"]},
    "PATENS": {"-": ["25"]},
    "PATEO": {"-": ["52"]},
    "PATER": {"-": ["13"], "N": ["12"]},
    "PATERA": {"-": ["11"]},
    "PATERNVS": {"-": ["21"]},
    "PATESCO": {"-": ["53"]},
    "PATHICVS": {"-": ["21"]},
    "PATIBILIS": {"-": ["24"]},
    "PATIBVLVM": {"-": ["12"]},
    "PATIENS": {"-": ["25"]},
    "PATIENTER": {"-": ["60"]},
    "PATIENTIA": {"-": ["11"]},
    "PATINA": {"-": ["11"], "N": ["11"]},
    "PATIOR": {"-": ["5N"]},
    "PATRATOR": {"-": ["13"]},
    "PATRIA": {"-": ["11"], "N": ["11"]},
    "PATRICIVS": {"1": ["12"], "2": ["21"]},
    "PATRIMONIVM": {"-": ["12"]},
    "PATRIVS": {"-": ["21"]},
    "PATRO": {"-": ["51"]},
    "PATROCINIVM": {"-": ["12"]},
    "PATROCINOR": {"-": ["5J"]},
    "PATRONA": {"-": ["11"]},
    "PATRONVS": {"-": ["12"]},
    "PATRVELIS": {"1": ["13"], "2": ["24"]},
    "PATRVVS": {"1": ["12"], "2": ["21"]},
    "PATVLVS": {"-": ["21"]},
    "PAVA": {"-": ["11"]},
    "PAVCA": {"-": ["12"]},
    "PAVCI": {"-": ["12"]},
    "PAVCITAS": {"-": ["13"]},
    "PAVCVLVS": {"-": ["21"]},
    "PAVCVS": {"-": ["21"]},
    "PAVEFACIO": {"-": ["55"]},
    "PAVEO": {"-": ["52"]},
    "PAVESCO": {"-": ["53"]},
    "PAVICVLA": {"-": ["11"]},
    "PAVIDE": {"-": ["60"]},
    "PAVIDVS": {"-": ["21"]},
    "PAVIMENTATVS": {"-": ["21"]},
    "PAVIMENTVM": {"-": ["12"]},
    "PAVIO": {"-": ["54"]},
    "PAVITO": {"-": ["51"]},
    "PAVLATIM": {"-": ["60", "6-"]},
    "PAVLISPER": {"-": ["60"]},
    "PAVLO": {"-": ["60", "6-"]},
    "PAVLVLVM": {"1": ["12"], "2": ["60"]},
    "PAVLVLVS": {"-": ["21"]},
    "PAVLVM": {"1": ["12"], "2": ["60"]},
    "PAVLVS": {"-": ["21"], "N": ["12"]},
    "PAVO": {"-": ["13"]},
    "PAVOR": {"-": ["13"], "N": ["13"]},
    "PAVPER": {"-": ["22"]},
    "PAVPERCVLVS": {"-": ["21"]},
    "PAVPERIES": {"-": ["15"]},
    "PAVPERO": {"-": ["51"]},
    "PAVPERTAS": {"-": ["13"]},
    "PAVSA": {"-": ["11"]},
    "PAVSARIVS": {"-": ["12"]},
    "PAVSEA": {"-": ["11"]},
    "PAVVS": {"-": ["12"]},
    "PAVXILLVM": {"1": ["12"], "2": ["12"]},
    "PAVXILLVS": {"-": ["21"]},
    "PAX": {"-": ["13"], "N": ["13"]},
    "PECCANS": {"-": ["13"]},
    "PECCATVM": {"-": ["12"]},
    "PECCO": {"-": ["51"]},
    "PECTEN": {"-": ["13"]},
    "PECTINATIM": {"-": ["60"]},
    "PECTO": {"-": ["53"]},
    "PECTVS": {"-": ["13"]},
    "PECV": {"-": ["14"]},
    "PECVARIA": {"-": ["12"]},
    "PECVARIVS": {"1": ["12"], "2": ["21"]},
    "PECVLATVS": {"-": ["14"]},
    "PECVLIARIS": {"-": ["24"]},
    "PECVLIVM": {"-": ["12"]},
    "PECVNIA": {"-": ["11"], "N": ["11"]},
    "PECVNIARIVS": {"1": ["12"], "2": ["21"]},
    "PECVNIOSVS": {"-": ["21"]},
    "PECVS": {"1": ["13"], "2": ["13"]},
    "PEDALIS": {"-": ["24"]},
    "PEDARIVS": {"-": ["21"]},
    "PEDATVM": {"-": ["12"]},
    "PEDATVS": {"1": ["14"]},
    "PEDES": {"-": ["13"]},
    "PEDESTER": {"-": ["23"]},
    "PEDETENTIM": {"-": ["60"]},
    "PEDICA": {"-": ["11"]},
    "PEDICINVS": {"-": ["12"]},
    "PEDICVLVS": {"1": ["12"]},
    "PEDIO": {"-": ["54"]},
    "PEDITATVS": {"1": ["14"], "2": ["21"]},
    "PEDITO": {"-": ["51"]},
    "PEDITVM": {"-": ["12"]},
    "PEDO": {"1": ["13"], "2": ["51"], "3": ["53"], "N": ["13"]},
    "PEDVCAEANVS": {"N": ["21"], "A": ["21"]},
    "PEDVCAEVS": {"N": ["12"]},
    "PEDVM": {"-": ["12"], "N": ["12"]},
    "PEGMA": {"-": ["17"]},
    "PEIERO": {"-": ["51"]},
    "PELAGVS": {"-": ["17"]},
    "PELAMIS": {"-": ["13"]},
    "PELASTES": {"-": ["13"]},
    "PELECINVS": {"-": ["12"]},
    "PELLACIA": {"-": ["11"]},
    "PELLAX": {"-": ["25"]},
    "PELLEX": {"-": ["13"]},
    "PELLICIO": {"-": ["55"]},
    "PELLICVLA": {"-": ["11"]},
    "PELLIS": {"-": ["13"]},
    "PELLITVS": {"-": ["21"]},
    "PELLO": {"-": ["53"]},
    "PELLVCEO": {"-": ["52"]},
    "PELLVCIDVLVS": {"-": ["21"]},
    "PELLVCIDVS": {"-": ["21"]},
    "PELOPONNESVS": {"N": ["12"]},
    "PELORIS": {"-": ["17"], "N": ["13"]},
    "PELTA": {"-": ["11"]},
    "PELTATVS": {"-": ["21"]},
    "PELVIS": {"-": ["13"]},
    "PELVSIVM": {"N": ["12"]},
    "PENARIVS": {"-": ["21"]},
    "PENATES": {"-": ["13"], "N": ["13"]},
    "PENATOR": {"-": ["13"]},
    "PENDEO": {"-": ["52"]},
    "PENDO": {"-": ["53"]},
    "PENDVLVS": {"-": ["21"]},
    "PENES": {"-": ["70"]},
    "PENETRABILIS": {"-": ["24"]},
    "PENETRALE": {"-": ["13"]},
    "PENETRALIS": {"-": ["24"]},
    "PENETRO": {"-": ["51"]},
    "PENIS": {"-": ["13"]},
    "PENITE": {"-": ["60"]},
    "PENITVS": {"1": ["21"], "2": ["60"]},
    "PENNA": {"-": ["11"]},
    "PENNATVS": {"-": ["21"]},
    "PENNIPES": {"-": ["25"]},
    "PENNIPOTENTES": {"-": ["13"]},
    "PENSILIS": {"-": ["24"]},
    "PENSIO": {"-": ["13"]},
    "PENSITATIO": {"-": ["13"]},
    "PENSITO": {"-": ["51"]},
    "PENSO": {"-": ["51"]},
    "PENSVM": {"-": ["12"]},
    "PENSVS": {"-": ["21"]},
    "PENTADOROS": {"1": ["27"], "2": ["60"]},
    "PENTASPASTON": {"-": ["17"]},
    "PENTEMOEROS": {"-": ["27"]},
    "PENTEREMIS": {"-": ["13"]},
    "PENTERES": {"-": ["13"]},
    "PENVRIA": {"-": ["11"]},
    "PENVS": {"-": ["16"]},
    "PEPLVS": {"-": ["12"]},
    "PER": {"-": ["70"]},
    "PERA": {"-": ["11"]},
    "PERACVTVS": {"-": ["21"]},
    "PERADOLESCENS": {"-": ["25"]},
    "PERAEQVE": {"-": ["60"]},
    "PERAEQVO": {"-": ["51"]},
    "PERAGITO": {"-": ["51"]},
    "PERAGO": {"-": ["53"]},
    "PERAGRO": {"-": ["51"]},
    "PERAMBVLO": {"-": ["51"]},
    "PERAMOENVS": {"-": ["21"]},
    "PERAMPLVS": {"-": ["21"]},
    "PERANGVSTVS": {"-": ["21"]},
    "PERANTIQVVS": {"-": ["21"]},
    "PERARDVVS": {"-": ["21"]},
    "PERARIDVS": {"-": ["21"]},
    "PERARMO": {"-": ["51"]},
    "PERARO": {"-": ["51"]},
    "PERATTENTVS": {"-": ["21"]},
    "PERBIBO": {"2": ["53"]},
    "PERBONVS": {"-": ["21"]},
    "PERBREVI": {"-": ["60"]},
    "PERBREVIS": {"-": ["24"]},
    "PERCALEFACIO": {"-": ["55"]},
    "PERCALESCO": {"-": ["53"]},
    "PERCANDEFACIO": {"-": ["54"]},
    "PERCARVS": {"-": ["21"]},
    "PERCELEBRO": {"-": ["51"]},
    "PERCELLO": {"-": ["53"]},
    "PERCENNIVS": {"N": ["12"]},
    "PERCENSEO": {"-": ["52"]},
    "PERCEPTIO": {"-": ["13"]},
    "PERCIDO": {"2": ["53"]},
    "PERCIEO": {"-": ["52"]},
    "PERCIO": {"-": ["54"]},
    "PERCIPIO": {"-": ["55"]},
    "PERCITO": {"1": ["51"]},
    "PERCOLATIO": {"-": ["13"]},
    "PERCOLO": {"1": ["51"], "2": ["53"]},
    "PERCOMMODE": {"-": ["60"]},
    "PERCOMPERIO": {"-": ["54"]},
    "PERCONTATIO": {"-": ["13"]},
    "PERCONTATOR": {"-": ["13"]},
    "PERCOQVO": {"-": ["53"]},
    "PERCREBRESCO": {"-": ["53"]},
    "PERCREPO": {"-": ["51"]},
    "PERCRESCO": {"-": ["53"]},
    "PERCRVDVS": {"-": ["21"]},
    "PERCVLSVS": {"-": ["14"]},
    "PERCVNCTATIO": {"-": ["13"]},
    "PERCVNCTO": {"-": ["51"]},
    "PERCVNCTOR": {"-": ["5J"]},
    "PERCVRO": {"-": ["51"]},
    "PERCVRRO": {"-": ["53"]},
    "PERCVRSO": {"-": ["51"]},
    "PERCVSSIO": {"-": ["13"]},
    "PERCVSSOR": {"-": ["13"]},
    "PERCVSSVS": {"-": ["14"]},
    "PERCVTIO": {"-": ["55"]},
    "PERDELIRVS": {"-": ["21"]},
    "PERDEPSO": {"-": ["53"]},
    "PERDIFFICILIS": {"-": ["24"]},
    "PERDISCO": {"-": ["53"]},
    "PERDITE": {"-": ["60"]},
    "PERDITOR": {"-": ["13"]},
    "PERDITVS": {"-": ["21"]},
    "PERDIVES": {"-": ["22"]},
    "PERDIVTVRNVS": {"-": ["21"]},
    "PERDO": {"-": ["53"]},
    "PERDOCEO": {"-": ["52"]},
    "PERDOLO": {"-": ["51"]},
    "PERDOMO": {"-": ["51"]},
    "PERDVCO": {"-": ["53"]},
    "PERDVCTIO": {"-": ["13"]},
    "PERDVCTOR": {"-": ["13"]},
    "PERDVELLIO": {"-": ["13"]},
    "PERDVELLIS": {"-": ["13"]},
    "PERDVLCIS": {"-": ["24"]},
    "PERDVRO": {"-": ["51"]},
    "PEREDO": {"1": ["53"], "2": ["56"]},
    "PEREGRE": {"-": ["60"]},
    "PEREGRINATIO": {"-": ["13"]},
    "PEREGRINOR": {"-": ["5J"]},
    "PEREGRINVS": {"1": ["12"], "2": ["21"]},
    "PEREMPTOR": {"-": ["13"]},
    "PERENDINVS": {"-": ["21"]},
    "PERENNE": {"-": ["60"]},
    "PERENNIS": {"-": ["24"]},
    "PERENNO": {"-": ["51"]},
    "PEREO": {"1": ["56"]},
    "PEREQVITO": {"-": ["51"]},
    "PERERRO": {"2": ["51"]},
    "PEREXCELSVS": {"-": ["21"]},
    "PEREXIGVVS": {"-": ["21"]},
    "PERFACETE": {"-": ["60"]},
    "PERFACILE": {"-": ["60"]},
    "PERFACILIS": {"-": ["24"]},
    "PERFAMILIARIS": {"-": ["24"]},
    "PERFECTE": {"-": ["60"]},
    "PERFECTIO": {"-": ["13"]},
    "PERFECTVS": {"1": ["14"], "2": ["21"]},
    "PERFERO": {"-": ["56"]},
    "PERFICIO": {"-": ["55"]},
    "PERFICVS": {"-": ["21"]},
    "PERFIDE": {"-": ["60"]},
    "PERFIDIA": {"-": ["11"], "N": ["11"]},
    "PERFIDIOSVS": {"-": ["21"]},
    "PERFIDVM": {"-": ["60"]},
    "PERFIDVS": {"1": ["21"], "2": ["21"]},
    "PERFIGO": {"-": ["53"]},
    "PERFLATVS": {"-": ["14"]},
    "PERFLO": {"-": ["51"]},
    "PERFLVCTVO": {"-": ["51"]},
    "PERFLVO": {"-": ["53"]},
    "PERFODIO": {"-": ["55"]},
    "PERFORO": {"-": ["51"]},
    "PERFRICO": {"-": ["51"]},
    "PERFRIGESCO": {"-": ["53"]},
    "PERFRIGIDVS": {"-": ["21"]},
    "PERFRINGO": {"-": ["53"]},
    "PERFRVOR": {"-": ["5L"]},
    "PERFVGA": {"-": ["11"]},
    "PERFVGIO": {"-": ["55"]},
    "PERFVGIVM": {"-": ["12"]},
    "PERFVNCTIO": {"-": ["13"]},
    "PERFVNDO": {"2": ["53"]},
    "PERFVNGOR": {"-": ["5L"]},
    "PERFVRO": {"-": ["53"]},
    "PERFVSORIVS": {"-": ["21"]},
    "PERGA": {"N": ["11"]},
    "PERGAEVS": {"N": ["21"], "A": ["21"]},
    "PERGAMVS": {"N": ["17"]},
    "PERGO": {"-": ["53"]},
    "PERGRANDIS": {"-": ["24"]},
    "PERGRATVS": {"-": ["21"]},
    "PERGVLA": {"-": ["11"]},
    "PERHIBEO": {"-": ["52"]},
    "PERHILVM": {"-": ["60"]},
    "PERHONORIFICVS": {"-": ["21"]},
    "PERHORRESCO": {"-": ["53"]},
    "PERIACTOS": {"-": ["27"]},
    "PERICLES": {"N": ["17"]},
    "PERICLITOR": {"-": ["5J"]},
    "PERICVLOR": {"-": ["5J"]},
    "PERICVLOSE": {"-": ["60"]},
    "PERICVLOSVS": {"-": ["21"]},
    "PERICVLVM": {"-": ["12"]},
    "PERIDONEVS": {"-": ["21"]},
    "PERIECHON": {"-": ["57"]},
    "PERIMETROS": {"-": ["17"]},
    "PERIMO": {"-": ["53"]},
    "PERIMPEDITVS": {"-": ["21"]},
    "PERINDE": {"-": ["60"]},
    "PERINGRATVS": {"-": ["21"]},
    "PERINIQVVS": {"-": ["21"]},
    "PERINIVRIVS": {"-": ["21"]},
    "PERINTEMPESTIVVS": {"-": ["21"]},
    "PERINVALIDVS": {"-": ["21"]},
    "PERIPATETICI": {"N": ["12"]},
    "PERIPATETICVS": {"1": ["12"], "N": ["21"], "A": ["21"]},
    "PERIPETASMA": {"-": ["17"]},
    "PERIPTEROS": {"-": ["27"]},
    "PERISCELIS": {"-": ["13"]},
    "PERISTYLIVM": {"-": ["12"]},
    "PERISTYLVM": {"-": ["12"]},
    "PERITE": {"-": ["60"]},
    "PERITHECIOS": {"-": ["27"]},
    "PERITIA": {"-": ["11"]},
    "PERITRETOS": {"-": ["27"]},
    "PERITVS": {"-": ["21"]},
    "PERIVRIVM": {"-": ["12"]},
    "PERIVRO": {"-": ["51"]},
    "PERIVRVS": {"-": ["21"]},
    "PERLABOR": {"2": ["5L"]},
    "PERLATEO": {"-": ["52"]},
    "PERLEGO": {"2": ["53"]},
    "PERLIBRATIO": {"-": ["13"]},
    "PERLIBRO": {"-": ["51"]},
    "PERLIMO": {"-": ["51"]},
    "PERLVO": {"-": ["53"]},
    "PERLVSTRO": {"2": ["51"]},
    "PERMACERO": {"-": ["51"]},
    "PERMADESCO": {"-": ["53"]},
    "PERMAGNVS": {"-": ["21"]},
    "PERMANANTER": {"-": ["60"]},
    "PERMANEO": {"-": ["52"]},
    "PERMANO": {"-": ["51"]},
    "PERMATVRESCO": {"-": ["53"]},
    "PERMEO": {"-": ["51"]},
    "PERMEREO": {"-": ["52"]},
    "PERMETIOR": {"-": ["5M"]},
    "PERMINGO": {"-": ["53"]},
    "PERMISCEO": {"-": ["52"]},
    "PERMISSVS": {"1": ["14"], "2": ["21"]},
    "PERMITIALIS": {"-": ["24"]},
    "PERMITTO": {"-": ["53"]},
    "PERMIXTIO": {"-": ["13"]},
    "PERMIXTVS": {"-": ["21"]},
    "PERMODESTVS": {"-": ["21"]},
    "PERMOLESTE": {"-": ["60"]},
    "PERMOLO": {"-": ["53"]},
    "PERMOVEO": {"-": ["52"]},
    "PERMVLCEO": {"-": ["52"]},
    "PERMVLTO": {"-": ["60"]},
    "PERMVLTVM": {"-": ["60"]},
    "PERMVLTVS": {"-": ["21"]},
    "PERMVNIO": {"2": ["54"]},
    "PERMVTATIO": {"-": ["13"]},
    "PERMVTO": {"2": ["51"]},
    "PERNA": {"-": ["11"]},
    "PERNEGO": {"-": ["51"]},
    "PERNICIABILIS": {"-": ["24"]},
    "PERNICIES": {"-": ["15"]},
    "PERNICIOSVS": {"-": ["21"]},
    "PERNICITAS": {"-": ["13"]},
    "PERNICITER": {"-": ["60"]},
    "PERNIX": {"-": ["25"]},
    "PERNOBILIS": {"-": ["24"]},
    "PERNOCTO": {"-": ["51"]},
    "PERNOSCO": {"-": ["53"]},
    "PERNOTESCO": {"-": ["53"]},
    "PERNOX": {"-": ["25"]},
    "PERNVMERO": {"1": ["51"]},
    "PERO": {"-": ["13"], "N": ["13"]},
    "PEROLEO": {"-": ["52"]},
    "PERONATVS": {"-": ["21"]},
    "PEROPPORTVNE": {"-": ["60"]},
    "PEROPPORTVNVS": {"-": ["21"]},
    "PERORNO": {"-": ["51"]},
    "PERORO": {"-": ["51"]},
    "PEROSVS": {"-": ["21"]},
    "PERPARVVLVS": {"-": ["21"]},
    "PERPARVVS": {"-": ["21"]},
    "PERPAVCI": {"-": ["12"]},
    "PERPAVCVS": {"-": ["21"]},
    "PERPELLO": {"-": ["53"]},
    "PERPENDEO": {"-": ["52"]},
    "PERPENDICVLVM": {"-": ["12"]},
    "PERPENDO": {"-": ["53"]},
    "PERPENSO": {"-": ["51"]},
    "PERPERAM": {"-": ["60"]},
    "PERPERE": {"-": ["60"]},
    "PERPERNA": {"N": ["11"]},
    "PERPERVS": {"-": ["21"]},
    "PERPES": {"-": ["25"]},
    "PERPESSICIVS": {"-": ["21"]},
    "PERPESSIO": {"-": ["13"]},
    "PERPETIOR": {"-": ["5N"]},
    "PERPETO": {"-": ["53"]},
    "PERPETRO": {"-": ["51"]},
    "PERPETVE": {"-": ["6&"]},
    "PERPETVITAS": {"-": ["13"]},
    "PERPETVO": {"1": ["51"], "2": ["60"]},
    "PERPETVVM": {"-": ["60"]},
    "PERPETVVS": {"-": ["21"]},
    "PERPLEXVS": {"-": ["21"]},
    "PERPLICATVS": {"-": ["21"]},
    "PERPLVO": {"-": ["53"]},
    "PERPOPVLOR": {"-": ["5J"]},
    "PERPOTO": {"-": ["51"]},
    "PERPRIMO": {"-": ["53"]},
    "PERPVRGO": {"-": ["51"]},
    "PERQVADRATVS": {"-": ["21"]},
    "PERQVAM": {"-": ["60"]},
    "PERQVIRO": {"-": ["53"]},
    "PERRARO": {"-": ["60"]},
    "PERREPO": {"-": ["53"]},
    "PERRIDICVLE": {"-": ["60"]},
    "PERROGO": {"-": ["51"]},
    "PERRVMPO": {"-": ["53"]},
    "PERSA": {"N": ["11"]},
    "PERSAE": {"N": ["11"]},
    "PERSAEPE": {"-": ["60"]},
    "PERSALVTO": {"-": ["51"]},
    "PERSANO": {"-": ["51"]},
    "PERSANVS": {"-": ["21"]},
    "PERSCINDO": {"-": ["53"]},
    "PERSCRIBO": {"-": ["53"]},
    "PERSCRIPTOR": {"-": ["13"]},
    "PERSCRVTATIO": {"-": ["13"]},
    "PERSCRVTOR": {"-": ["5J"]},
    "PERSECO": {"-": ["51"]},
    "PERSECTOR": {"2": ["5J"]},
    "PERSEDEO": {"-": ["52"]},
    "PERSENTIO": {"-": ["54"]},
    "PERSENTISCO": {"-": ["53"]},
    "PERSEQVOR": {"-": ["5L"]},
    "PERSERO": {"2": ["53"], "3": ["53"]},
    "PERSES": {"N": ["17"], "A": ["25"]},
    "PERSEVERANS": {"-": ["25"]},
    "PERSEVERANTER": {"-": ["60"]},
    "PERSEVERANTIA": {"-": ["11"]},
    "PERSEVERO": {"-": ["51"]},
    "PERSEVERVS": {"-": ["21"]},
    "PERSIDO": {"-": ["53"]},
    "PERSIMPLEX": {"-": ["25"]},
    "PERSISTO": {"-": ["53"]},
    "PERSOLVO": {"-": ["53"]},
    "PERSONA": {"-": ["11"]},
    "PERSONATVS": {"-": ["21"]},
    "PERSONO": {"1": ["51"]},
    "PERSPECTO": {"-": ["51"]},
    "PERSPECTVS": {"-": ["21"]},
    "PERSPERGO": {"-": ["53"]},
    "PERSPICIO": {"-": ["55"]},
    "PERSPICVE": {"-": ["60"]},
    "PERSPICVITAS": {"-": ["13"]},
    "PERSPICVVS": {"-": ["21"]},
    "PERSPIRO": {"-": ["51"]},
    "PERSTERNO": {"-": ["53"]},
    "PERSTIMVLO": {"-": ["51"]},
    "PERSTO": {"-": ["51"]},
    "PERSTREPO": {"-": ["53"]},
    "PERSTRINGO": {"-": ["53"]},
    "PERSTRVO": {"-": ["53"]},
    "PERSTVDIOSVS": {"-": ["21"]},
    "PERSVADEO": {"-": ["52"]},
    "PERSVASIO": {"-": ["13"]},
    "PERSVBTILIS": {"-": ["24"]},
    "PERSVLTO": {"-": ["51"]},
    "PERTAEDESCO": {"-": ["53"]},
    "PERTAEDET": {"-": ["52"]},
    "PERTEGO": {"-": ["53"]},
    "PERTENDO": {"-": ["53"]},
    "PERTENTO": {"-": ["51"]},
    "PERTENVIS": {"-": ["24"]},
    "PERTEREBRO": {"-": ["51"]},
    "PERTERGEO": {"-": ["52"]},
    "PERTERREO": {"-": ["52"]},
    "PERTERRICREPVS": {"-": ["21"]},
    "PERTEXO": {"-": ["53"]},
    "PERTICA": {"-": ["11"]},
    "PERTIMEO": {"-": ["52"]},
    "PERTIMESCO": {"-": ["53"]},
    "PERTINACIA": {"-": ["11"]},
    "PERTINACITER": {"-": ["60"]},
    "PERTINAX": {"-": ["25"]},
    "PERTINEO": {"-": ["52"]},
    "PERTINGO": {"-": ["53"]},
    "PERTIO": {"-": ["54"]},
    "PERTOLERO": {"-": ["51"]},
    "PERTORQVEO": {"-": ["52"]},
    "PERTRACTO": {"-": ["51"]},
    "PERTRAHO": {"-": ["53"]},
    "PERTRITVS": {"-": ["21"]},
    "PERTVNDO": {"-": ["53"]},
    "PERTVRBATE": {"-": ["60"]},
    "PERTVRBATIO": {"-": ["13"]},
    "PERTVRBATVS": {"-": ["21"]},
    "PERTVRBO": {"2": ["51"]},
    "PERVADO": {"2": ["53"]},
    "PERVAGOR": {"2": ["5J"]},
    "PERVAGVS": {"-": ["21"]},
    "PERVALEO": {"-": ["52"]},
    "PERVASTO": {"-": ["51"]},
    "PERVECTOR": {"-": ["13"]},
    "PERVEHO": {"-": ["53"]},
    "PERVELLO": {"-": ["53"]},
    "PERVENIO": {"-": ["54"]},
    "PERVERSE": {"-": ["60"]},
    "PERVERSITAS": {"-": ["13"]},
    "PERVERSVM": {"-": ["12"]},
    "PERVERSVS": {"-": ["21"]},
    "PERVERTO": {"-": ["53"]},
    "PERVESTIGO": {"-": ["51"]},
    "PERVETVS": {"-": ["22"]},
    "PERVICACIA": {"-": ["11"]},
    "PERVICACITER": {"-": ["6&"]},
    "PERVICAX": {"-": ["25"]},
    "PERVIDEO": {"-": ["52"]},
    "PERVIGEO": {"-": ["52"]},
    "PERVIGIL": {"-": ["25"]},
    "PERVIGILIVM": {"-": ["12"]},
    "PERVIGILO": {"-": ["51"]},
    "PERVINCO": {"-": ["53"]},
    "PERVIVM": {"-": ["12"]},
    "PERVIVS": {"-": ["21"]},
    "PERVLA": {"-": ["11"]},
    "PERVNGO": {"-": ["53"]},
    "PERVOLITANTIA": {"-": ["11"]},
    "PERVOLITO": {"-": ["51"]},
    "PERVOLO": {"2": ["51"], "3": ["56"]},
    "PERVOLVO": {"-": ["53"]},
    "PERVRO": {"-": ["53"]},
    "PERVSINVS": {"N": ["21"], "A": ["21"]},
    "PERVVLGO": {"1": ["51"]},
    "PES": {"-": ["13"]},
    "PESSVM": {"1": ["12"], "2": ["60"]},
    "PESSVMDO": {"-": ["51"]},
    "PESTIFER": {"-": ["21"]},
    "PESTILENS": {"-": ["25"]},
    "PESTILENTIA": {"-": ["11"]},
    "PESTILITAS": {"-": ["13"]},
    "PESTIS": {"-": ["13"], "N": ["13"]},
    "PETASVNCVLVS": {"-": ["12"]},
    "PETAVRVM": {"-": ["12"]},
    "PETESSO": {"-": ["53"]},
    "PETILIVS": {"N": ["12"]},
    "PETIOR": {"-": ["5N"]},
    "PETITIO": {"-": ["13"]},
    "PETITOR": {"-": ["13"]},
    "PETITVS": {"-": ["14"]},
    "PETO": {"-": ["53"]},
    "PETORITVM": {"-": ["12"]},
    "PETRA": {"-": ["11"], "N": ["11"]},
    "PETRINI": {"N": ["12"]},
    "PETRO": {"-": ["51"]},
    "PETVLANS": {"-": ["25"]},
    "PETVLANTER": {"-": ["6&"]},
    "PETVLANTIA": {"-": ["11"]},
    "PETVLCVS": {"-": ["21"]},
    "PEXVS": {"-": ["21"]},
    "PHAECASIATVS": {"-": ["21"]},
    "PHAECASIVM": {"-": ["12"]},
    "PHALACRVS": {"N": ["12"]},
    "PHALANGA": {"-": ["11"]},
    "PHALANGARIVS": {"-": ["12"]},
    "PHALANX": {"-": ["13"]},
    "PHALARIS": {"N": ["17"]},
    "PHALERAE": {"-": ["11"]},
    "PHARETRA": {"-": ["11"]},
    "PHARETRATVS": {"-": ["21"]},
    "PHARMACOPOLA": {"-": ["11"]},
    "PHAROS": {"-": ["17"], "N": ["17"]},
    "PHARVS": {"-": ["17"], "N": ["12"]},
    "PHASELIS": {"N": ["13"]},
    "PHELLOS": {"-": ["17"]},
    "PHIALA": {"-": ["11"]},
    "PHILEMA": {"-": ["17"]},
    "PHILINVS": {"N": ["12"]},
    "PHILIPPVS": {"1N": ["12"], "2N": ["12"], "A": ["21"], "N": ["12"]},
    "PHILO": {"N": ["17"]},
    "PHILOCRATES": {"N": ["17"]},
    "PHILODAMVS": {"N": ["12"]},
    "PHILOLOGIA": {"-": ["11"]},
    "PHILOLOGVS": {"1": ["12"], "2": ["21"]},
    "PHILOMELIENSES": {"N": ["13"]},
    "PHILOMELIVM": {"N": ["12"]},
    "PHILOSOPHIA": {"-": ["11"]},
    "PHILOSOPHOR": {"-": ["5J"]},
    "PHILOSOPHVS": {"1": ["12"], "2": ["21"]},
    "PHILOTECHNVS": {"-": ["21"]},
    "PHILTRVM": {"-": ["12"]},
    "PHILYRA": {"-": ["11"]},
    "PHIMES": {"N": ["17"]},
    "PHIMVS": {"-": ["12"]},
    "PHINTIA": {"N": ["11"]},
    "PHOCA": {"-": ["11"]},
    "PHOCIS": {"N": ["17"]},
    "PHOEBVS": {"N": ["12"]},
    "PHOENICE": {"N": ["17"]},
    "PHOENICES": {"N": ["17", "13"]},
    "PHOENICOPTERVS": {"-": ["12"]},
    "PHOENIX": {"-": ["13"], "N": ["13"]},
    "PHONASCVS": {"-": ["12"]},
    "PHRENESIS": {"-": ["17"]},
    "PHRENETICVS": {"-": ["21"]},
    "PHRYGIA": {"N": ["11"]},
    "PHRYX": {"A": ["27"], "N": ["17"]},
    "PHTHISICVS": {"-": ["21"]},
    "PHTHISIS": {"-": ["17"]},
    "PHTHONGOS": {"-": ["17"]},
    "PHYLARCHVS": {"N": ["12"]},
    "PHYSETER": {"-": ["17"]},
    "PHYSICA": {"1": ["12"], "2": ["17"]},
    "PHYSICVS": {"1": ["12"], "2": ["21"]},
    "PHYSIOLOGIA": {"-": ["11"]},
    "PIACVLARIS": {"-": ["21"]},
    "PIACVLO": {"-": ["51"]},
    "PIACVLVM": {"-": ["12"]},
    "PIAMENTVM": {"-": ["12"]},
    "PICA": {"-": ["11"]},
    "PICEA": {"-": ["11"]},
    "PICEVS": {"-": ["21"]},
    "PICO": {"-": ["51"]},
    "PICTOR": {"-": ["13"], "N": ["13"]},
    "PICTVRA": {"-": ["11"]},
    "PICTVRATVS": {"-": ["21"]},
    "PICTVS": {"-": ["21"]},
    "PICVS": {"-": ["12"], "N": ["12"]},
    "PIE": {"-": ["60"]},
    "PIETAS": {"-": ["13"], "N": ["13"]},
    "PIGER": {"-": ["21"]},
    "PIGET": {"-": ["52"]},
    "PIGNERATOR": {"-": ["13"]},
    "PIGNERO": {"-": ["51"]},
    "PIGNVS": {"-": ["13"]},
    "PIGRE": {"-": ["60"]},
    "PIGRITIA": {"-": ["11"]},
    "PIGROR": {"-": ["5J"]},
    "PILA": {"1": ["11"], "2": ["11"], "3": ["11"]},
    "PILATIM": {"-": ["60"]},
    "PILATVS": {"-": ["21"], "N": ["12"]},
    "PILEATVS": {"-": ["21"]},
    "PILENTVM": {"-": ["12"]},
    "PILEOLVS": {"-": ["12"]},
    "PILEVM": {"-": ["12"]},
    "PILEVS": {"-": ["12"]},
    "PILICREPVS": {"-": ["12"]},
    "PILO": {"1": ["51"], "2": ["51"]},
    "PILOSVS": {"-": ["21"]},
    "PILVM": {"-": ["12"], "N": ["12"]},
    "PILVS": {"1": ["12"], "2": ["12"]},
    "PINACOTHECA": {"-": ["17"]},
    "PINARIVS": {"A": ["21"], "N": ["12"]},
    "PINAX": {"-": ["17"]},
    "PINETVM": {"-": ["12"]},
    "PINEVS": {"-": ["21"]},
    "PINGO": {"-": ["53"]},
    "PINGVE": {"-": ["13"]},
    "PINGVESCO": {"-": ["53"]},
    "PINGVIS": {"-": ["24"]},
    "PINGVITVDO": {"-": ["13"]},
    "PINIFER": {"-": ["21"]},
    "PINNA": {"-": ["11"]},
    "PINNIGER": {"-": ["21"]},
    "PINNIRAPVS": {"-": ["12"]},
    "PINSATIO": {"-": ["13"]},
    "PINSO": {"-": ["56"]},
    "PINVS": {"-": ["16"]},
    "PIO": {"-": ["51"]},
    "PIPA": {"N": ["11"]},
    "PIPER": {"-": ["13"]},
    "PIPIO": {"1": ["13"], "2": ["51"], "3": ["54"]},
    "PIRAEEVS": {"N": ["12"]},
    "PIRATA": {"-": ["11"]},
    "PIRATICA": {"-": ["11"]},
    "PIRATICVS": {"-": ["21"]},
    "PIRVM": {"-": ["12"]},
    "PIRVS": {"-": ["12"]},
    "PISATIO": {"-": ["13"]},
    "PISCARIVS": {"-": ["21"]},
    "PISCATOR": {"-": ["13"]},
    "PISCATORIVS": {"-": ["21"]},
    "PISCATVS": {"-": ["14"]},
    "PISCICVLVS": {"-": ["12"]},
    "PISCINA": {"-": ["11"]},
    "PISCIS": {"-": ["13"], "N": ["13"]},
    "PISCOR": {"-": ["5J"]},
    "PISCOSVS": {"-": ["21"]},
    "PISCVLENTVS": {"-": ["21"]},
    "PISIDIA": {"N": ["11"]},
    "PISO": {"N": ["13"]},
    "PISTOR": {"-": ["13"], "N": ["13"]},
    "PISTRINVM": {"-": ["12"]},
    "PISTRIS": {"-": ["13"]},
    "PISTRIX": {"-": ["13"]},
    "PITVITA": {"-": ["11"]},
    "PIVS": {"-": ["21"], "N": ["12"]},
    "PIX": {"-": ["13"], "2": ["13"], "1": ["12"]},
    "PLACABILIS": {"-": ["24"]},
    "PLACAMENTVM": {"-": ["12"]},
    "PLACATIO": {"-": ["13"]},
    "PLACATVS": {"-": ["21"]},
    "PLACENTA": {"-": ["11"]},
    "PLACENTINVS": {"N": ["21"], "A": ["21"]},
    "PLACEO": {"-": ["52"]},
    "PLACIDE": {"-": ["60"]},
    "PLACIDVS": {"-": ["21"], "N": ["12"]},
    "PLACITO": {"-": ["51"]},
    "PLACITVM": {"-": ["12"]},
    "PLACITVS": {"-": ["21"]},
    "PLACO": {"-": ["51"]},
    "PLAGA": {"1": ["11"], "2": ["11"], "3": ["11"]},
    "PLAGIARIVS": {"-": ["12"]},
    "PLAGOSVS": {"-": ["21"]},
    "PLANATVS": {"-": ["21"]},
    "PLANCTVS": {"-": ["14"]},
    "PLANE": {"-": ["60"]},
    "PLANGO": {"-": ["53"]},
    "PLANGOR": {"-": ["13"]},
    "PLANIPES": {"-": ["13"]},
    "PLANITAS": {"-": ["13"]},
    "PLANITIA": {"-": ["11"]},
    "PLANITIES": {"-": ["15"]},
    "PLANTA": {"-": ["11"], "N": ["11"]},
    "PLANTARIA": {"-": ["13"]},
    "PLANTARIS": {"-": ["24"]},
    "PLANTO": {"-": ["51"]},
    "PLANVM": {"-": ["12"]},
    "PLANVS": {"1": ["12"], "2": ["21"]},
    "PLASMA": {"-": ["17"]},
    "PLASTES": {"-": ["17"]},
    "PLASTICVS": {"-": ["21"]},
    "PLATANON": {"-": ["17"]},
    "PLATANVS": {"-": ["16"]},
    "PLATEA": {"1": ["11"]},
    "PLATO": {"N": ["17"]},
    "PLAVDO": {"-": ["53"]},
    "PLAVSIBILIS": {"-": ["24"]},
    "PLAVSOR": {"-": ["13"]},
    "PLAVSTELLVM": {"-": ["12"]},
    "PLAVSTRARIVS": {"2": ["21"]},
    "PLAVSTRVM": {"-": ["12"], "N": ["12"]},
    "PLAVSVS": {"-": ["14"]},
    "PLEBECVLA": {"-": ["11"]},
    "PLEBEIVS": {"1": ["12"], "2": ["21"]},
    "PLEBICOLA": {"-": ["11"]},
    "PLEBISCITVM": {"-": ["12"]},
    "PLEBITAS": {"-": ["13"]},
    "PLEBS": {"-": ["16"]},
    "PLECTO": {"1": ["53"], "2": ["53"]},
    "PLECTOR": {"-": ["5L"]},
    "PLECTRVM": {"-": ["12"]},
    "PLENE": {"-": ["60"]},
    "PLENITAS": {"-": ["13"]},
    "PLENVM": {"-": ["12"]},
    "PLENVS": {"-": ["21"]},
    "PLEO": {"-": ["52"]},
    "PLERAQVE": {"-": ["12"]},
    "PLERIQVE": {"-": ["12"]},
    "PLERVMQVE": {"-": ["60"]},
    "PLERVS": {"-": ["21"]},
    "PLERVSQVE": {"-": ["21"]},
    "PLEVRITIS": {"-": ["17"]},
    "PLICEO": {"-": ["52"]},
    "PLICO": {"-": ["51"]},
    "PLINTHIS": {"-": ["17"]},
    "PLINTHIVM": {"-": ["12"]},
    "PLINTHVS": {"-": ["17"]},
    "PLODO": {"-": ["53"]},
    "PLORABILIS": {"-": ["24"]},
    "PLORATVS": {"-": ["14"]},
    "PLORO": {"-": ["51"]},
    "PLOXEMVM": {"-": ["12"]},
    "PLVMA": {"-": ["11"]},
    "PLVMARIVS": {"-": ["12"]},
    "PLVMBARIVS": {"-": ["21"]},
    "PLVMBEVM": {"-": ["12"]},
    "PLVMBEVS": {"-": ["21"]},
    "PLVMBO": {"-": ["51"]},
    "PLVMBVM": {"-": ["12"]},
    "PLVMIPES": {"-": ["27"]},
    "PLVMO": {"-": ["51"]},
    "PLVMVLA": {"-": ["11"]},
    "PLVO": {"-": ["53"]},
    "PLVRIMVM": {"-": ["12"]},
    "PLVSCVLVM": {"1": ["12"], "2": ["60"]},
    "PLVTEVM": {"-": ["12"]},
    "PLVTEVS": {"-": ["12"]},
    "PLVVIA": {"-": ["11"]},
    "PLVVIALIS": {"-": ["24"]},
    "PLVVIVS": {"-": ["21"]},
    "PNEVMATICOS": {"-": ["27"]},
    "PNIGEVS": {"-": ["17"]},
    "POCILLVM": {"-": ["12"]},
    "POCVLVM": {"-": ["12"]},
    "PODAGRA": {"-": ["11"]},
    "PODAGRICVS": {"-": ["21"]},
    "PODEX": {"-": ["13"]},
    "PODIVM": {"-": ["12"]},
    "POEMA": {"-": ["16"]},
    "POENA": {"-": ["11"], "N": ["11"]},
    "POENI": {"N": ["12"]},
    "POENVS": {"N": ["21"], "A": ["21"]},
    "POETA": {"-": ["11"]},
    "POETICA": {"-": ["11"], "2": ["11"]},
    "POETICVS": {"-": ["21"]},
    "POETRIA": {"-": ["11"]},
    "POETRIS": {"-": ["17"]},
    "POL": {"-": ["90"]},
    "POLEAS": {"N": ["17"]},
    "POLEMARCHVS": {"N": ["12"]},
    "POLENTA": {"-": ["11"]},
    "POLIO": {"1": ["13"], "2": ["54"]},
    "POLITIO": {"-": ["13"]},
    "POLITOR": {"-": ["13"]},
    "POLITORIVM": {"N": ["12"]},
    "POLITVRA": {"-": ["11"]},
    "POLITVS": {"-": ["21"]},
    "POLLEN": {"-": ["13"]},
    "POLLENS": {"-": ["25"]},
    "POLLEO": {"-": ["52"]},
    "POLLEX": {"-": ["13"]},
    "POLLICEOR": {"-": ["5K"]},
    "POLLICITATIO": {"-": ["13"]},
    "POLLICITOR": {"-": ["5J"]},
    "POLLICITVM": {"-": ["12"]},
    "POLLINGO": {"-": ["53"]},
    "POLLIO": {"N": ["13"]},
    "POLLVCEO": {"-": ["52"]},
    "POLLVO": {"-": ["53"]},
    "POLLVTVS": {"-": ["21"]},
    "POLLVX": {"N": ["13"]},
    "POLVS": {"-": ["12"]},
    "POLVSCA": {"N": ["11"]},
    "POLYBIVS": {"N": ["12"]},
    "POLYCLITVS": {"N": ["12"]},
    "POLYGONIVS": {"-": ["21"]},
    "POLYPVS": {"-": ["12"]},
    "POLYSPASTON": {"-": ["17"]},
    "POMARIVM": {"-": ["12"]},
    "POMARIVS": {"1": ["12"], "2": ["21"]},
    "POMERIDIANVS": {"-": ["21"]},
    "POMERIVM": {"-": ["12"]},
    "POMETIA": {"N": ["11"]},
    "POMETINVS": {"N": ["21"], "A": ["21"]},
    "POMIFER": {"-": ["21"]},
    "POMOSVS": {"-": ["21"]},
    "POMPA": {"-": ["11"]},
    "POMPEIVS": {"A": ["21"], "N": ["12"]},
    "POMPILIVS": {"N": ["12"], "A": ["21"]},
    "POMPONIVS": {"N": ["12"]},
    "POMPTINVS": {"A": ["21"], "N": ["12"]},
    "POMVM": {"-": ["12"]},
    "POMVS": {"-": ["12"]},
    "PONDERATIO": {"-": ["13"]},
    "PONDERATVS": {"-": ["21"]},
    "PONDERO": {"-": ["51"]},
    "PONDEROSVS": {"-": ["21"]},
    "PONDO": {"-": ["16"]},
    "PONDVS": {"-": ["13"]},
    "PONE": {"1": ["60"], "2": ["70"]},
    "PONEVERSVS": {"-": ["70"]},
    "PONO": {"-": ["53"]},
    "PONS": {"-": ["13"], "N": ["13"]},
    "PONTICVLVS": {"-": ["12"]},
    "PONTICVS": {"N": ["21"], "A": ["21"]},
    "PONTIFEX": {"-": ["13"]},
    "PONTIFICALIS": {"-": ["24"]},
    "PONTIFICATVS": {"-": ["14"]},
    "PONTIFICIVS": {"-": ["21"], "N": ["12"]},
    "PONTO": {"-": ["13"]},
    "PONTVS": {"-": ["12"], "N": ["12"]},
    "POPA": {"-": ["11"]},
    "POPANVM": {"-": ["12"]},
    "POPELLVS": {"-": ["12"]},
    "POPILIVS": {"N": ["12"]},
    "POPINA": {"-": ["11"]},
    "POPINO": {"-": ["13"]},
    "POPLES": {"-": ["13"]},
    "POPVLARIS": {"1": ["13"], "2": ["24"]},
    "POPVLARITAS": {"-": ["13"]},
    "POPVLARITER": {"-": ["60"]},
    "POPVLATIO": {"-": ["13"]},
    "POPVLATOR": {"-": ["13"]},
    "POPVLEVS": {"-": ["21"]},
    "POPVLIFER": {"-": ["21"]},
    "POPVLNEVS": {"-": ["21"]},
    "POPVLO": {"-": ["51"]},
    "POPVLOR": {"-": ["5J"]},
    "POPVLVS": {"1": ["12"], "2": ["12"]},
    "PORCA": {"-": ["11"]},
    "PORCINVS": {"-": ["21"]},
    "PORCIVS": {"-": ["12"], "A": ["21"], "N": ["12"]},
    "PORCVLVS": {"-": ["12"]},
    "PORCVS": {"-": ["12"]},
    "PORRECTIO": {"-": ["13"]},
    "PORRECTVS": {"-": ["21"]},
    "PORRIGO": {"1": ["13"], "2": ["53"]},
    "PORRINA": {"-": ["11"]},
    "PORRO": {"-": ["60"]},
    "PORRVM": {"-": ["12"]},
    "PORRVS": {"-": ["12"]},
    "PORSENA": {"N": ["11"]},
    "PORTA": {"-": ["11"]},
    "PORTATIO": {"-": ["13"]},
    "PORTENDO": {"-": ["53"]},
    "PORTENTOSVS": {"-": ["21"]},
    "PORTENTVM": {"-": ["12"]},
    "PORTHMEVS": {"-": ["17"], "N": ["12"]},
    "PORTICVS": {"-": ["14"], "N": ["12"]},
    "PORTIO": {"-": ["13"]},
    "PORTISCVLVS": {"-": ["12"]},
    "PORTITOR": {"1": ["13"], "2": ["13"]},
    "PORTIVS": {"N": ["12"]},
    "PORTO": {"-": ["51"]},
    "PORTORIVM": {"-": ["12"]},
    "PORTVLA": {"-": ["11"]},
    "PORTVOSVS": {"-": ["21"]},
    "PORTVS": {"-": ["14"]},
    "PORVS": {"-": ["12"], "N": ["12"]},
    "POSCIMVS": {"-": ["53"]},
    "POSCO": {"-": ["53"]},
    "POSIDES": {"N": ["17"]},
    "POSIDORVS": {"N": ["12"]},
    "POSITIO": {"-": ["13"]},
    "POSITVRA": {"-": ["11"]},
    "POSITVS": {"-": ["14"]},
    "POSOTES": {"-": ["17"]},
    "POSSESSIO": {"-": ["13"]},
    "POSSESSOR": {"-": ["13"]},
    "POSSIDEO": {"-": ["52"]},
    "POSSIDO": {"-": ["53"]},
    "POSSVM": {"1": ["56"]},
    "POST": {"1": ["60"], "2": ["70"]},
    "POSTEA": {"-": ["60"]},
    "POSTEAQVAM": {"-": ["82"]},
    "POSTERGA": {"-": ["60"]},
    "POSTERGVM": {"-": ["60"]},
    "POSTERI": {"-": ["12"]},
    "POSTERITAS": {"-": ["13"]},
    "POSTERIVS": {"-": ["6&"]},
    "POSTERVS": {"-": ["21"]},
    "POSTFERO": {"-": ["56"]},
    "POSTGENITVS": {"-": ["21"]},
    "POSTHABEO": {"-": ["52"]},
    "POSTHAC": {"-": ["60"]},
    "POSTICVM": {"-": ["12"]},
    "POSTICVS": {"-": ["21"]},
    "POSTILLA": {"-": ["60"]},
    "POSTIS": {"-": ["13"]},
    "POSTMERIDIANVS": {"-": ["21"]},
    "POSTMODO": {"-": ["60"]},
    "POSTMODVM": {"-": ["60"]},
    "POSTPONO": {"-": ["53"]},
    "POSTQVAM": {"-": ["82"]},
    "POSTREMO": {"-": ["60"]},
    "POSTREMVM": {"-": ["6-"]},
    "POSTRIDIE": {"-": ["60"]},
    "POSTSCENIVM": {"-": ["12"]},
    "POSTSCRIBO": {"-": ["53"]},
    "POSTVLATICIVS": {"-": ["21"]},
    "POSTVLATIO": {"-": ["13"]},
    "POSTVLATVM": {"-": ["12"]},
    "POSTVLO": {"-": ["51"]},
    "POSTVMIVS": {"N": ["12"], "A": ["21"]},
    "POSTVMVS": {"N": ["12"]},
    "POTA": {"N": ["11"]},
    "POTAMO": {"N": ["13"]},
    "POTATIO": {"-": ["13"]},
    "POTENS": {"-": ["25"]},
    "POTENTATVS": {"-": ["14"]},
    "POTENTER": {"-": ["60"]},
    "POTENTIA": {"-": ["11"], "N": ["11"]},
    "POTESTAS": {"-": ["13"]},
    "POTIO": {"1": ["13"], "2": ["54"]},
    "POTIOR": {"-": ["5M"]},
    "POTIS": {"-": ["26"]},
    "POTISSIME": {"-": ["6-"]},
    "POTISSIMVM": {"-": ["6-"]},
    "POTITII": {"N": ["12"]},
    "POTIVS": {"-": ["6&"]},
    "POTO": {"-": ["51"]},
    "POTOR": {"-": ["13"]},
    "POTORO": {"-": ["51"]},
    "POTVS": {"1": ["12"], "2": ["14"]},
    "PRAE": {"1": ["60"], "2": ["70"]},
    "PRAEACVO": {"-": ["53"]},
    "PRAEALTVS": {"-": ["21"]},
    "PRAEBEO": {"-": ["52"]},
    "PRAEBITA": {"-": ["12"]},
    "PRAECALIDVS": {"-": ["21"]},
    "PRAECANVS": {"-": ["21"]},
    "PRAECAVEO": {"-": ["52"]},
    "PRAECEDO": {"1": ["53"]},
    "PRAECELLENS": {"-": ["25"]},
    "PRAECELLO": {"-": ["53"]},
    "PRAECELSVS": {"-": ["21"]},
    "PRAECEPS": {"1": ["13"], "2": ["25"], "3": ["60"]},
    "PRAECEPTIO": {"-": ["13"]},
    "PRAECEPTIVVS": {"-": ["21"]},
    "PRAECEPTOR": {"-": ["13"]},
    "PRAECEPTRIX": {"-": ["13"]},
    "PRAECEPTVM": {"-": ["12"]},
    "PRAECERPO": {"-": ["53"]},
    "PRAECIDANEVS": {"-": ["21"]},
    "PRAECIDO": {"2": ["53"]},
    "PRAECINCTIO": {"-": ["13"]},
    "PRAECINCTVRA": {"-": ["11"]},
    "PRAECINGO": {"-": ["53"]},
    "PRAECINO": {"-": ["53"]},
    "PRAECIPIO": {"-": ["55"]},
    "PRAECIPITANTER": {"-": ["60"]},
    "PRAECIPITATIO": {"-": ["13"]},
    "PRAECIPITO": {"-": ["51"]},
    "PRAECIPVA": {"-": ["12"]},
    "PRAECIPVE": {"-": ["60"]},
    "PRAECIPVVS": {"-": ["21"]},
    "PRAECISIO": {"-": ["13"]},
    "PRAECISVS": {"-": ["21"]},
    "PRAECLARE": {"-": ["60"]},
    "PRAECLARVS": {"-": ["21"]},
    "PRAECLVDO": {"-": ["53"]},
    "PRAECLVSIO": {"-": ["13"]},
    "PRAECO": {"-": ["13"]},
    "PRAECOGITO": {"-": ["51"]},
    "PRAECOLO": {"2": ["53"]},
    "PRAECONIVM": {"-": ["12"]},
    "PRAECOQVO": {"-": ["53"]},
    "PRAECORDIA": {"-": ["12"]},
    "PRAECOX": {"-": ["25"]},
    "PRAECVRRO": {"-": ["53"]},
    "PRAECVRSOR": {"-": ["13"]},
    "PRAECVTIO": {"-": ["55"]},
    "PRAEDA": {"-": ["11"]},
    "PRAEDABVNDVS": {"-": ["21"]},
    "PRAEDATIO": {"-": ["13"]},
    "PRAEDATOR": {"-": ["13"]},
    "PRAEDATORIVS": {"-": ["21"]},
    "PRAEDICABILIS": {"-": ["24"]},
    "PRAEDICATIO": {"-": ["13"]},
    "PRAEDICO": {"1": ["51"], "2": ["53"]},
    "PRAEDICTIO": {"-": ["13"]},
    "PRAEDICTVM": {"-": ["12"]},
    "PRAEDISCO": {"-": ["53"]},
    "PRAEDITVS": {"-": ["21"]},
    "PRAEDIVES": {"-": ["22"]},
    "PRAEDIVM": {"-": ["12"]},
    "PRAEDO": {"-": ["13"]},
    "PRAEDOCEO": {"-": ["52"]},
    "PRAEDOMO": {"-": ["51"]},
    "PRAEDONVLVS": {"-": ["12"]},
    "PRAEDOR": {"-": ["5J"]},
    "PRAEDVCO": {"-": ["53"]},
    "PRAEDVCTORIVS": {"-": ["21"]},
    "PRAEDVRVS": {"-": ["21"]},
    "PRAEEO": {"1": ["56"]},
    "PRAEFATIO": {"-": ["13"]},
    "PRAEFATVS": {"1": ["14"], "2": ["21"]},
    "PRAEFECTVRA": {"-": ["11"]},
    "PRAEFECTVS": {"-": ["12"]},
    "PRAEFERO": {"-": ["56"]},
    "PRAEFEROX": {"-": ["25"]},
    "PRAEFERRATVS": {"-": ["21"]},
    "PRAEFERVIDVS": {"-": ["21"]},
    "PRAEFESTINO": {"-": ["51"]},
    "PRAEFICIO": {"-": ["55"]},
    "PRAEFIGO": {"-": ["53"]},
    "PRAEFINIO": {"-": ["54"]},
    "PRAEFLORO": {"-": ["51"]},
    "PRAEFLVO": {"-": ["53"]},
    "PRAEFOR": {"-": ["5J"]},
    "PRAEFRACTVS": {"-": ["21"]},
    "PRAEFRINGO": {"-": ["53"]},
    "PRAEFVLGEO": {"-": ["52"]},
    "PRAEFVRNIVM": {"-": ["12"]},
    "PRAEGERO": {"-": ["53"]},
    "PRAEGESTIO": {"2": ["54"]},
    "PRAEGNANS": {"-": ["25"]},
    "PRAEGRACILIS": {"-": ["24"]},
    "PRAEGRANDIS": {"-": ["23"]},
    "PRAEGRAVIS": {"-": ["24"]},
    "PRAEGRAVO": {"-": ["51"]},
    "PRAEGREDIOR": {"-": ["5N"]},
    "PRAEGVSTO": {"-": ["51"]},
    "PRAEIACEO": {"-": ["52"]},
    "PRAEIVDICIVM": {"-": ["12"]},
    "PRAEIVDICO": {"-": ["51"]},
    "PRAELABOR": {"2": ["5L"]},
    "PRAELAMBO": {"-": ["53"]},
    "PRAELARGVS": {"-": ["21"]},
    "PRAELEGO": {"1": ["51"], "2": ["53"]},
    "PRAELIGANEVS": {"-": ["21"]},
    "PRAELIGO": {"2": ["51"]},
    "PRAELONGVS": {"-": ["21"]},
    "PRAELVCEO": {"-": ["52"]},
    "PRAEMATVRVS": {"-": ["21"]},
    "PRAEMEDICATVS": {"-": ["21"]},
    "PRAEMEDITATVS": {"-": ["21"]},
    "PRAEMEDITOR": {"-": ["5J"]},
    "PRAEMETVENTER": {"-": ["60"]},
    "PRAEMETVO": {"-": ["53"]},
    "PRAEMINEO": {"-": ["52"]},
    "PRAEMIOSVS": {"-": ["21"]},
    "PRAEMITTO": {"-": ["53"]},
    "PRAEMIVM": {"-": ["12"]},
    "PRAEMONEO": {"-": ["52"]},
    "PRAEMONSTRO": {"-": ["51"]},
    "PRAEMORDEO": {"-": ["52"]},
    "PRAEMORIOR": {"-": ["5N"]},
    "PRAEMVNIO": {"2": ["54"]},
    "PRAENATO": {"-": ["51"]},
    "PRAENAVIGO": {"-": ["51"]},
    "PRAENESTE": {"N": ["13"]},
    "PRAENITEO": {"-": ["52"]},
    "PRAENOMEN": {"-": ["13"]},
    "PRAENOSCO": {"-": ["53"]},
    "PRAENOTIO": {"-": ["13"]},
    "PRAENVBILVS": {"-": ["21"]},
    "PRAENVNTIO": {"-": ["51"]},
    "PRAENVNTIVS": {"1": ["12"], "2": ["21"]},
    "PRAEOBTVRO": {"-": ["51"]},
    "PRAEOCCVPO": {"2": ["51"]},
    "PRAEOPTO": {"-": ["51"]},
    "PRAEPANDO": {"2": ["53"]},
    "PRAEPARATIO": {"-": ["13"]},
    "PRAEPARATVS": {"-": ["21"]},
    "PRAEPARO": {"2": ["51"]},
    "PRAEPEDIO": {"-": ["54"]},
    "PRAEPENDEO": {"-": ["52"]},
    "PRAEPES": {"1": ["13"], "2": ["25"]},
    "PRAEPETO": {"-": ["53"]},
    "PRAEPILATVS": {"1": ["21"]},
    "PRAEPINGVIS": {"-": ["24"]},
    "PRAEPOLLENS": {"-": ["25"]},
    "PRAEPOLLEO": {"-": ["52"]},
    "PRAEPONDERO": {"-": ["51"]},
    "PRAEPONO": {"-": ["53"]},
    "PRAEPORTO": {"-": ["51"]},
    "PRAEPOSITIO": {"-": ["13"]},
    "PRAEPOSITVS": {"1": ["12"]},
    "PRAEPOSTERO": {"1": ["51"], "2": ["60"]},
    "PRAEPOSTERVS": {"-": ["21"]},
    "PRAEPOTENS": {"-": ["25"]},
    "PRAEPOTENTES": {"-": ["13"]},
    "PRAEPROPERANTER": {"-": ["60"]},
    "PRAEPROPERE": {"-": ["60"]},
    "PRAEPVTIVM": {"-": ["12"]},
    "PRAERADIO": {"-": ["51"]},
    "PRAERAPIDVS": {"-": ["21"]},
    "PRAERIGEO": {"-": ["52"]},
    "PRAERIPIO": {"-": ["55"]},
    "PRAERODO": {"-": ["53"]},
    "PRAEROGATIVA": {"-": ["11"]},
    "PRAEROGATIVVS": {"-": ["21"]},
    "PRAERVMPO": {"-": ["53"]},
    "PRAERVPTA": {"-": ["12"]},
    "PRAERVPTVM": {"-": ["12"]},
    "PRAERVPTVS": {"-": ["21"]},
    "PRAES": {"1": ["13"], "2": ["60"]},
    "PRAESAEPE": {"-": ["13"]},
    "PRAESAEPES": {"-": ["13"]},
    "PRAESAEPIO": {"-": ["54"]},
    "PRAESAGIO": {"-": ["54"]},
    "PRAESAGIVM": {"-": ["12"]},
    "PRAESAGVS": {"-": ["21"]},
    "PRAESCISCO": {"-": ["53"]},
    "PRAESCIVS": {"-": ["21"]},
    "PRAESCRIBO": {"-": ["53"]},
    "PRAESCRIPTIO": {"-": ["13"]},
    "PRAESCRIPTVM": {"-": ["12"]},
    "PRAESECO": {"-": ["51"]},
    "PRAESEMINATIO": {"-": ["13"]},
    "PRAESENS": {"-": ["25"]},
    "PRAESENSIO": {"-": ["13"]},
    "PRAESENTANEVS": {"-": ["21"]},
    "PRAESENTIA": {"1": ["11"], "2": ["13"]},
    "PRAESENTIO": {"-": ["54"]},
    "PRAESENTO": {"-": ["51"]},
    "PRAESERO": {"3": ["53"]},
    "PRAESERTIM": {"-": ["60"]},
    "PRAESES": {"-": ["13"]},
    "PRAESIDENS": {"-": ["13"]},
    "PRAESIDEO": {"-": ["52"]},
    "PRAESIDIVM": {"-": ["12"]},
    "PRAESIGNIS": {"-": ["24"]},
    "PRAESONO": {"1": ["51"]},
    "PRAESPARGO": {"2": ["53"]},
    "PRAESPICIO": {"-": ["55"]},
    "PRAESTABILIS": {"-": ["24"]},
    "PRAESTANS": {"-": ["25"]},
    "PRAESTANTIA": {"-": ["11"]},
    "PRAESTATIO": {"-": ["13"]},
    "PRAESTERNO": {"-": ["53"]},
    "PRAESTIGIAE": {"-": ["11"]},
    "PRAESTIGIATOR": {"-": ["13"]},
    "PRAESTITVO": {"-": ["53"]},
    "PRAESTO": {"1": ["51"], "2": ["60"]},
    "PRAESTOLOR": {"-": ["5J"]},
    "PRAESTRIGIAE": {"-": ["11"]},
    "PRAESTRINGO": {"-": ["53"]},
    "PRAESTRVO": {"-": ["53"]},
    "PRAESVM": {"1": ["56"]},
    "PRAESVMO": {"-": ["53"]},
    "PRAESVMPTIO": {"-": ["13"]},
    "PRAESVMPTIOSVS": {"-": ["21"]},
    "PRAESVMPTOR": {"-": ["13"]},
    "PRAESVMPTVS": {"-": ["21"]},
    "PRAETEGO": {"-": ["53"]},
    "PRAETEMPTO": {"-": ["51"]},
    "PRAETENDO": {"-": ["53"]},
    "PRAETENTO": {"-": ["51"]},
    "PRAETEPESCO": {"-": ["53"]},
    "PRAETER": {"1": ["60"], "2": ["70"]},
    "PRAETERAGO": {"-": ["53"]},
    "PRAETEREA": {"-": ["60"]},
    "PRAETEREO": {"1": ["56"]},
    "PRAETERFERO": {"-": ["56"]},
    "PRAETERFLVO": {"-": ["53"]},
    "PRAETERGREDIOR": {"-": ["5N"]},
    "PRAETERITA": {"-": ["12"]},
    "PRAETERITVM": {"-": ["12"]},
    "PRAETERITVS": {"-": ["21"]},
    "PRAETERLABOR": {"2": ["5L"]},
    "PRAETERMEO": {"-": ["51"]},
    "PRAETERMITTO": {"-": ["53"]},
    "PRAETERO": {"-": ["53"]},
    "PRAETERQVAM": {"-": ["60"]},
    "PRAETERVECTIO": {"-": ["13"]},
    "PRAETERVEHOR": {"-": ["5L"]},
    "PRAETERVERTO": {"-": ["53"]},
    "PRAETERVOLO": {"2": ["51"]},
    "PRAETEXO": {"-": ["53"]},
    "PRAETEXTA": {"-": ["11"]},
    "PRAETEXTATVS": {"-": ["21"]},
    "PRAETEXTVM": {"-": ["12"]},
    "PRAETEXTVS": {"1": ["14"], "2": ["21"]},
    "PRAETIMEO": {"-": ["52"]},
    "PRAETOR": {"-": ["13"]},
    "PRAETORIANVS": {"-": ["21"]},
    "PRAETORIVM": {"-": ["12"]},
    "PRAETORIVS": {"1": ["12"], "2": ["21"]},
    "PRAETRACTO": {"-": ["51"]},
    "PRAETREPIDO": {"-": ["51"]},
    "PRAETREPIDVS": {"-": ["21"]},
    "PRAETVRA": {"-": ["11"]},
    "PRAEVALENS": {"-": ["25"]},
    "PRAEVALEO": {"-": ["52"]},
    "PRAEVALIDVS": {"-": ["21"]},
    "PRAEVALLO": {"-": ["51"]},
    "PRAEVARICATIO": {"-": ["13"]},
    "PRAEVARICOR": {"-": ["5J"]},
    "PRAEVEHOR": {"-": ["5L"]},
    "PRAEVENIO": {"-": ["54"]},
    "PRAEVERRO": {"-": ["53"]},
    "PRAEVERTO": {"-": ["53"]},
    "PRAEVERTOR": {"-": ["5L"]},
    "PRAEVIDEO": {"-": ["52"]},
    "PRAEVINCO": {"-": ["53"]},
    "PRAEVIVS": {"-": ["21"]},
    "PRAEVMBRO": {"-": ["51"]},
    "PRAEVRO": {"-": ["53"]},
    "PRAGMATICVS": {"1": ["12"], "2": ["21"]},
    "PRANDEO": {"-": ["52"]},
    "PRANDIVM": {"-": ["12"]},
    "PRANSITO": {"-": ["51"]},
    "PRATENSIS": {"-": ["24"]},
    "PRATVM": {"-": ["12"]},
    "PRAVE": {"-": ["60"]},
    "PRAVITAS": {"-": ["13"]},
    "PRAVVS": {"-": ["21"]},
    "PRAXITELES": {"N": ["17"]},
    "PRECARIO": {"-": ["60"]},
    "PRECARIVS": {"-": ["21"]},
    "PRECATIO": {"-": ["13"]},
    "PRECIAE": {"-": ["11"]},
    "PRECOR": {"-": ["5J"]},
    "PREHENDO": {"-": ["53"]},
    "PREHENSO": {"-": ["51"]},
    "PRELVM": {"-": ["12"]},
    "PREMO": {"-": ["53"]},
    "PRESBYTER": {"-": ["12"]},
    "PRESBYTERIVM": {"-": ["12"]},
    "PRESSE": {"-": ["60"]},
    "PRESSIO": {"-": ["13"]},
    "PRESSO": {"-": ["51"]},
    "PRESSVS": {"1": ["14"], "2": ["21"]},
    "PRESTER": {"-": ["17"]},
    "PRETIOSE": {"-": ["6&"]},
    "PRETIOSVS": {"-": ["21"]},
    "PRETIVM": {"-": ["12"]},
    "PREX": {"-": ["13"]},
    "PRIAMVS": {"N": ["12"]},
    "PRIAPVS": {"-": ["12"], "N": ["12"]},
    "PRIDEM": {"-": ["60"]},
    "PRIDIE": {"-": ["60"]},
    "PRIMA": {"1": ["12"], "2": ["60"], "N": ["11"]},
    "PRIMAEVVS": {"-": ["21"]},
    "PRIMANI": {"-": ["12"]},
    "PRIMARIVS": {"-": ["21"]},
    "PRIMIGENVS": {"-": ["21"]},
    "PRIMIPILARIS": {"-": ["13"]},
    "PRIMIPILVS": {"-": ["12"]},
    "PRIMITVS": {"-": ["60"]},
    "PRIMO": {"-": ["53", "60"], "1": ["3N"], "2": ["53"]},
    "PRIMORDIVM": {"-": ["12"]},
    "PRIMORES": {"-": ["13"]},
    "PRIMORIS": {"-": ["26"]},
    "PRIMVM": {"-": ["3N", "6-"]},
    "PRIMVS": {"-": ["3K", "13"], "N": ["12"]},
    "PRINCEPS": {"1": ["13"], "2": ["22"]},
    "PRINCIPALIS": {"-": ["24"]},
    "PRINCIPALITER": {"-": ["60"]},
    "PRINCIPATVS": {"-": ["14"]},
    "PRINCIPIALIS": {"-": ["24"]},
    "PRINCIPIVM": {"-": ["12"]},
    "PRIOR": {"-": ["3B", "2&"]},
    "PRIORES": {"-": ["13"]},
    "PRISCI": {"-": ["12"], "N": ["12"]},
    "PRISCVS": {"-": ["21"], "N": ["12"]},
    "PRISTINVS": {"-": ["21"]},
    "PRIVATIM": {"-": ["60"]},
    "PRIVATO": {"-": ["60"]},
    "PRIVATVS": {"-": ["21"]},
    "PRIVIENVS": {"-": ["12"]},
    "PRIVIGNA": {"-": ["11"]},
    "PRIVIGNVS": {"-": ["12"]},
    "PRIVILEGIVM": {"-": ["12"]},
    "PRIVO": {"-": ["51"]},
    "PRIVS": {"-": ["3E", "6-"]},
    "PRIVSQVAM": {"-": ["82"]},
    "PRIVVS": {"-": ["21"]},
    "PRO": {"1": ["70"], "2": ["90"], "3": ["70"]},
    "PROAGORVS": {"-": ["12"]},
    "PROAVIA": {"-": ["11"]},
    "PROAVVS": {"-": ["12"]},
    "PROBABILIS": {"-": ["24"]},
    "PROBABILITAS": {"-": ["13"]},
    "PROBATIO": {"-": ["13"]},
    "PROBATOR": {"-": ["13"]},
    "PROBATVS": {"-": ["21"]},
    "PROBE": {"-": ["60"]},
    "PROBITAS": {"-": ["13"]},
    "PROBO": {"-": ["51"]},
    "PROBOSCIS": {"-": ["17"]},
    "PROBRO": {"-": ["51"]},
    "PROBROSVS": {"-": ["21"]},
    "PROBRVM": {"-": ["12"]},
    "PROBVS": {"-": ["21"], "N": ["12"]},
    "PROCACITER": {"-": ["6&"]},
    "PROCAS": {"N": ["17"]},
    "PROCAX": {"-": ["25"]},
    "PROCEDO": {"1": ["53"]},
    "PROCELLA": {"-": ["11"]},
    "PROCELLO": {"-": ["53"]},
    "PROCELLOSVS": {"-": ["21"]},
    "PROCER": {"-": ["13"]},
    "PROCERES": {"-": ["13"]},
    "PROCERITAS": {"-": ["13"]},
    "PROCERVS": {"-": ["21"]},
    "PROCESSIO": {"-": ["13"]},
    "PROCESSVS": {"-": ["14"]},
    "PROCIDO": {"1": ["53"]},
    "PROCINCTVS": {"1": ["14"], "2": ["21"]},
    "PROCLAMO": {"-": ["51"]},
    "PROCLINATIO": {"-": ["13"]},
    "PROCLINO": {"-": ["51"]},
    "PROCLIVE": {"1": ["13"], "2": ["60"]},
    "PROCLIVIS": {"-": ["24"]},
    "PROCLIVITAS": {"-": ["13"]},
    "PROCLIVITER": {"-": ["60"]},
    "PROCLIVVS": {"-": ["21"]},
    "PROCONSVL": {"-": ["13"]},
    "PROCONSVLARIS": {"-": ["24"]},
    "PROCONSVLATVS": {"-": ["14"]},
    "PROCRASTINO": {"-": ["51"]},
    "PROCREATIO": {"-": ["13"]},
    "PROCREO": {"-": ["51"]},
    "PROCRESCO": {"-": ["53"]},
    "PROCVBO": {"-": ["51"]},
    "PROCVDO": {"-": ["53"]},
    "PROCVL": {"-": ["60"]},
    "PROCVLCATIO": {"-": ["13"]},
    "PROCVLCO": {"-": ["51"]},
    "PROCVLVS": {"N": ["12"]},
    "PROCVMBO": {"-": ["53"]},
    "PROCVRATIO": {"-": ["13"]},
    "PROCVRATIVNCVLA": {"-": ["11"]},
    "PROCVRATOR": {"-": ["13"]},
    "PROCVRO": {"-": ["51"]},
    "PROCVRRO": {"-": ["53"]},
    "PROCVRSVS": {"-": ["14"]},
    "PROCVRVVS": {"-": ["21"]},
    "PROCVS": {"-": ["12"]},
    "PRODEO": {"1": ["56"]},
    "PRODIGE": {"-": ["60"]},
    "PRODIGENTIA": {"-": ["11"]},
    "PRODIGIOSVS": {"-": ["21"]},
    "PRODIGIVM": {"-": ["12"]},
    "PRODIGO": {"-": ["53"]},
    "PRODIGVS": {"-": ["21"]},
    "PRODITIO": {"1": ["13"], "2": ["13"]},
    "PRODITOR": {"-": ["13"]},
    "PRODO": {"-": ["53"]},
    "PRODOCEO": {"-": ["52"]},
    "PRODVCO": {"-": ["53"]},
    "PRODVCTA": {"-": ["12"]},
    "PRODVCTIO": {"-": ["13"]},
    "PRODVCTVS": {"-": ["21"]},
    "PROELIATOR": {"-": ["13"]},
    "PROELIOR": {"-": ["5J"]},
    "PROELIVM": {"-": ["12"]},
    "PROFANO": {"1": ["51"], "2": ["51"]},
    "PROFANVS": {"-": ["21"]},
    "PROFECTIO": {"-": ["13"]},
    "PROFECTO": {"-": ["60"]},
    "PROFECTVS": {"-": ["14"]},
    "PROFERO": {"-": ["56"]},
    "PROFESSIO": {"-": ["13"]},
    "PROFESSOR": {"-": ["13"]},
    "PROFESSORIVS": {"-": ["21"]},
    "PROFESSVS": {"-": ["21"]},
    "PROFESTVS": {"-": ["21"]},
    "PROFICIO": {"-": ["55"]},
    "PROFICISCOR": {"-": ["5L"]},
    "PROFITEOR": {"-": ["5K"]},
    "PROFLIGATOR": {"-": ["13"]},
    "PROFLIGATVS": {"-": ["21"]},
    "PROFLIGO": {"-": ["51"]},
    "PROFLO": {"-": ["51"]},
    "PROFLVENS": {"-": ["25"]},
    "PROFLVENTER": {"-": ["60"]},
    "PROFLVO": {"-": ["53"]},
    "PROFLVVIVM": {"-": ["12"]},
    "PROFLVVIVS": {"-": ["21"]},
    "PROFOR": {"-": ["5J"]},
    "PROFVGIO": {"-": ["55"]},
    "PROFVGVS": {"1": ["12"], "2": ["21"]},
    "PROFVNDO": {"2": ["53"]},
    "PROFVNDVM": {"-": ["12"]},
    "PROFVNDVS": {"-": ["21"]},
    "PROFVSE": {"-": ["6&"]},
    "PROFVSIO": {"-": ["13"]},
    "PROFVSVS": {"-": ["21"]},
    "PROGENER": {"-": ["12"]},
    "PROGENERO": {"-": ["51"]},
    "PROGENIES": {"-": ["15"]},
    "PROGERO": {"-": ["53"]},
    "PROGIGNO": {"-": ["53"]},
    "PROGNATVS": {"-": ["21"]},
    "PROGNE": {"-": ["17"], "N": ["11"]},
    "PROGREDIOR": {"-": ["5N"]},
    "PROGRESSIO": {"-": ["13"]},
    "PROGRESSVS": {"1": ["14"], "2": ["21"]},
    "PROGYMNASTES": {"-": ["17"]},
    "PROHIBEO": {"-": ["52"]},
    "PROHIBITA": {"-": ["12"]},
    "PROHIBITIO": {"-": ["13"]},
    "PROICIO": {"-": ["55"]},
    "PROIECTVRA": {"-": ["11"]},
    "PROIECTVS": {"1": ["14"], "2": ["21"]},
    "PROINDE": {"-": ["60"]},
    "PROLABOR": {"2": ["5L"]},
    "PROLATIO": {"-": ["13"]},
    "PROLATO": {"-": ["51"]},
    "PROLECTO": {"-": ["51"]},
    "PROLES": {"-": ["13"]},
    "PROLETARIVS": {"-": ["21"]},
    "PROLEVO": {"1": ["51"]},
    "PROLICEO": {"-": ["52"]},
    "PROLICIO": {"-": ["55"]},
    "PROLIXE": {"-": ["60"]},
    "PROLIXVS": {"-": ["21"]},
    "PROLOGVS": {"-": ["12"]},
    "PROLONGO": {"-": ["51"]},
    "PROLOQVOR": {"-": ["5L"]},
    "PROLVBIVM": {"-": ["12"]},
    "PROLVDO": {"-": ["53"]},
    "PROLVO": {"-": ["53"]},
    "PROLVVIES": {"-": ["15"]},
    "PROMEREO": {"-": ["52"]},
    "PROMEREOR": {"-": ["5K"]},
    "PROMERITVM": {"-": ["12"]},
    "PROMETHEVS": {"N": ["17"], "A": ["21"]},
    "PROMINENS": {"1": ["13"], "2": ["25"]},
    "PROMINENTIA": {"-": ["11"]},
    "PROMINEO": {"-": ["52"]},
    "PROMISCVE": {"-": ["60"]},
    "PROMISCVVM": {"-": ["12"]},
    "PROMISCVVS": {"-": ["21"]},
    "PROMISSVM": {"-": ["12"]},
    "PROMISSVS": {"-": ["21"]},
    "PROMITTO": {"-": ["53"]},
    "PROMO": {"-": ["53"]},
    "PROMONTORIVM": {"-": ["12"]},
    "PROMOVEO": {"-": ["52"]},
    "PROMPTE": {"-": ["60"]},
    "PROMPTVARIVS": {"-": ["21"]},
    "PROMPTVS": {"1": ["14"], "2": ["21"]},
    "PROMVLGO": {"-": ["51"]},
    "PROMVS": {"-": ["12"]},
    "PROMVTVVS": {"-": ["21"]},
    "PRONAVS": {"-": ["17"]},
    "PRONE": {"-": ["60"]},
    "PRONEPOS": {"-": ["13"]},
    "PRONEPTIS": {"-": ["13"]},
    "PRONVBA": {"-": ["11"]},
    "PRONVBVS": {"-": ["21"]},
    "PRONVNTIATIO": {"-": ["13"]},
    "PRONVNTIO": {"-": ["51"]},
    "PRONVRVS": {"-": ["14"]},
    "PRONVS": {"-": ["21"]},
    "PROOEMIVM": {"-": ["12"]},
    "PROPAGATIO": {"-": ["13"]},
    "PROPAGATOR": {"-": ["13"]},
    "PROPAGO": {"1": ["13"], "2": ["51"]},
    "PROPALAM": {"-": ["60"]},
    "PROPATVLVM": {"-": ["12"]},
    "PROPATVLVS": {"-": ["21"]},
    "PROPE": {"1": ["60"], "2": ["70"]},
    "PROPEDIEM": {"-": ["60"]},
    "PROPELLO": {"-": ["53"]},
    "PROPEMODVM": {"-": ["60"]},
    "PROPENDEO": {"-": ["52"]},
    "PROPENSIO": {"-": ["13"]},
    "PROPENSVS": {"-": ["21"]},
    "PROPERANS": {"-": ["25"]},
    "PROPERANTER": {"-": ["60"]},
    "PROPERANTIA": {"-": ["11"]},
    "PROPERATIO": {"-": ["13"]},
    "PROPERATO": {"-": ["60"]},
    "PROPERATVS": {"-": ["21"]},
    "PROPERE": {"-": ["60"]},
    "PROPERIPES": {"-": ["25"]},
    "PROPERO": {"-": ["51"]},
    "PROPERVS": {"-": ["21"]},
    "PROPINATIO": {"-": ["13"]},
    "PROPINO": {"-": ["51"]},
    "PROPINQVI": {"-": ["12"]},
    "PROPINQVITAS": {"-": ["13"]},
    "PROPINQVO": {"-": ["51"]},
    "PROPINQVVM": {"-": ["12"]},
    "PROPINQVVS": {"-": ["21"], "N": ["12"]},
    "PROPIOR": {"-": ["2&"]},
    "PROPITIO": {"-": ["51"]},
    "PROPITIVS": {"-": ["21"]},
    "PROPIVS": {"-": ["70"]},
    "PROPNIGEVM": {"-": ["17"]},
    "PROPOLA": {"-": ["11"]},
    "PROPOLLVO": {"-": ["53"]},
    "PROPONO": {"-": ["53"]},
    "PROPORRO": {"-": ["60"]},
    "PROPORTIO": {"-": ["13"]},
    "PROPOSITIO": {"-": ["13"]},
    "PROPOSITVM": {"-": ["12"]},
    "PROPRAETOR": {"-": ["13"]},
    "PROPRIATIM": {"-": ["60"]},
    "PROPRIE": {"-": ["60"]},
    "PROPRIETAS": {"-": ["13"]},
    "PROPRITIM": {"-": ["60"]},
    "PROPRIVM": {"-": ["12"]},
    "PROPRIVS": {"-": ["21"]},
    "PROPTER": {"1": ["60"], "2": ["70"]},
    "PROPTEREA": {"-": ["60"]},
    "PROPVGNACVLVM": {"-": ["12"]},
    "PROPVGNATOR": {"-": ["13"]},
    "PROPVGNO": {"-": ["51"]},
    "PROPVLSO": {"-": ["51"]},
    "PROQVAESTOR": {"-": ["13"]},
    "PROQVAM": {"-": ["66"]},
    "PRORA": {"-": ["11"]},
    "PROREPO": {"-": ["53"]},
    "PRORIPIO": {"-": ["55"]},
    "PRORIS": {"-": ["13"]},
    "PRORITO": {"-": ["51"]},
    "PROROGO": {"-": ["51"]},
    "PRORSVM": {"-": ["60"]},
    "PRORSVS": {"1": ["21"], "2": ["60"]},
    "PRORVMPO": {"-": ["53"]},
    "PRORVO": {"-": ["53"]},
    "PROSAPIA": {"-": ["11"]},
    "PROSCAENIVM": {"-": ["12"]},
    "PROSCENIVM": {"-": ["12"]},
    "PROSCINDO": {"-": ["53"]},
    "PROSCRIBO": {"-": ["53"]},
    "PROSCRIPTIO": {"-": ["13"]},
    "PROSECO": {"-": ["51"]},
    "PROSEMINO": {"-": ["51"]},
    "PROSEQVOR": {"-": ["5L"]},
    "PROSERO": {"2": ["53"], "3": ["53"]},
    "PROSERPINA": {"N": ["11"]},
    "PROSEVCHA": {"-": ["11"]},
    "PROSILIO": {"-": ["54"]},
    "PROSITVS": {"-": ["21"]},
    "PROSLAMBANOMENOS": {"-": ["17"]},
    "PROSOCER": {"-": ["12"]},
    "PROSPECTO": {"-": ["51"]},
    "PROSPECTVS": {"-": ["14"]},
    "PROSPECVLOR": {"-": ["5J"]},
    "PROSPER": {"-": ["21"]},
    "PROSPERA": {"-": ["12"]},
    "PROSPERE": {"-": ["60"]},
    "PROSPERITAS": {"-": ["13"]},
    "PROSPERO": {"-": ["51"]},
    "PROSPICIO": {"-": ["55"]},
    "PROSTAS": {"-": ["17"]},
    "PROSTERNO": {"-": ["53"]},
    "PROSTITVO": {"-": ["53"]},
    "PROSTITVTA": {"-": ["11"]},
    "PROSTITVTVS": {"-": ["21"]},
    "PROSTO": {"-": ["51"]},
    "PROSTYLOS": {"-": ["27"]},
    "PROSVBIGO": {"-": ["51"]},
    "PROSVM": {"1": ["56"]},
    "PROTECTVRA": {"-": ["11"]},
    "PROTEGO": {"-": ["53"]},
    "PROTELVM": {"-": ["12"]},
    "PROTENDO": {"-": ["53"]},
    "PROTENTVS": {"-": ["21"]},
    "PROTERO": {"-": ["53"]},
    "PROTERREO": {"-": ["52"]},
    "PROTERVE": {"-": ["60"]},
    "PROTERVITAS": {"-": ["13"]},
    "PROTERVVS": {"-": ["21"]},
    "PROTHYRVM": {"-": ["17"]},
    "PROTINVS": {"-": ["60"]},
    "PROTOLLO": {"-": ["53"]},
    "PROTRAHO": {"-": ["53"]},
    "PROTRITVS": {"-": ["21"]},
    "PROTROPVM": {"-": ["12"]},
    "PROTRVDO": {"-": ["53"]},
    "PROTVRBO": {"2": ["51"]},
    "PROVECTVS": {"1": ["14"], "2": ["21"]},
    "PROVEHO": {"-": ["53"]},
    "PROVENIO": {"-": ["54"]},
    "PROVENTVS": {"-": ["14"]},
    "PROVERBIVM": {"-": ["12"]},
    "PROVERTO": {"-": ["53"]},
    "PROVIDENS": {"-": ["25"]},
    "PROVIDENTER": {"-": ["60"]},
    "PROVIDENTIA": {"-": ["11"]},
    "PROVIDEO": {"-": ["52"]},
    "PROVIDVS": {"-": ["21"]},
    "PROVINCIA": {"-": ["11"]},
    "PROVINCIALIS": {"1": ["13"], "2": ["24"]},
    "PROVINCO": {"-": ["53"]},
    "PROVISIO": {"-": ["13"]},
    "PROVISO": {"1": ["53"], "2": ["60"]},
    "PROVISOR": {"-": ["13"]},
    "PROVISVS": {"-": ["14"]},
    "PROVIVO": {"-": ["53"]},
    "PROVOCATIO": {"-": ["13"]},
    "PROVOCO": {"-": ["51"]},
    "PROVOLO": {"2": ["51"]},
    "PROVOLVO": {"-": ["53"]},
    "PROVOMO": {"-": ["53"]},
    "PROVT": {"-": ["66"]},
    "PROXENETA": {"-": ["11"]},
    "PROXIME": {"-": ["70"]},
    "PROXIMITAS": {"-": ["13"]},
    "PROXIMVM": {"-": ["12"]},
    "PRVDENS": {"-": ["25"]},
    "PRVDENTER": {"-": ["60"]},
    "PRVDENTIA": {"-": ["11"]},
    "PRVINA": {"-": ["11"]},
    "PRVINOSVS": {"-": ["21"]},
    "PRVNA": {"-": ["11"]},
    "PRVNVM": {"-": ["12"]},
    "PRVNVS": {"-": ["12"]},
    "PRVRIGO": {"-": ["13"]},
    "PRVRIO": {"-": ["54"]},
    "PRYTANEVM": {"-": ["12"]},
    "PRYTANIS": {"-": ["13"], "N": ["13"]},
    "PSALLO": {"-": ["53"]},
    "PSALTRIA": {"-": ["11"]},
    "PSEVDOTHYRVM": {"-": ["12"]},
    "PSITTACVS": {"-": ["12"]},
    "PSYLLVS": {"-": ["12"]},
    "PSYTHIA": {"-": ["11"]},
    "PTEROMA": {"-": ["17"]},
    "PTERYGOMA": {"-": ["17"]},
    "PTISANARIVM": {"-": ["12"]},
    "PTOLEMAEVS": {"N": ["12"]},
    "PVBENS": {"-": ["25"]},
    "PVBERTAS": {"-": ["13"]},
    "PVBES": {"1": ["13"], "2": ["26"]},
    "PVBESCO": {"-": ["53"]},
    "PVBLICA": {"-": ["11"]},
    "PVBLICANVS": {"1": ["12"], "2": ["21"]},
    "PVBLICATIO": {"-": ["13"]},
    "PVBLICE": {"-": ["60"]},
    "PVBLICO": {"-": ["51"]},
    "PVBLICOLA": {"N": ["11"]},
    "PVBLICVM": {"-": ["12"]},
    "PVBLICVS": {"1": ["12"], "2": ["21"], "N": ["12"], "A": ["21"]},
    "PVBLILIVS": {"N": ["12"]},
    "PVBLIVS": {"N": ["12"]},
    "PVDENDA": {"-": ["12"]},
    "PVDENDVM": {"-": ["90"]},
    "PVDENDVS": {"-": ["21"]},
    "PVDENS": {"-": ["25"], "N": ["13"]},
    "PVDENTER": {"-": ["60"]},
    "PVDEO": {"-": ["52"]},
    "PVDIBVNDVS": {"-": ["21"]},
    "PVDICE": {"-": ["60"]},
    "PVDICITIA": {"-": ["11"], "N": ["11"]},
    "PVDICVS": {"-": ["21"]},
    "PVDOR": {"-": ["13"], "N": ["13"]},
    "PVELLA": {"-": ["11"], "N": ["11"]},
    "PVELLARIS": {"-": ["24"]},
    "PVELLVLA": {"-": ["11"]},
    "PVELLVS": {"-": ["12"]},
    "PVER": {"-": ["12"]},
    "PVERILIS": {"-": ["24"]},
    "PVERILITAS": {"-": ["13"]},
    "PVERILITER": {"-": ["60"]},
    "PVERITIA": {"-": ["11"]},
    "PVERPERA": {"-": ["11"]},
    "PVERPERIVM": {"-": ["12"]},
    "PVERPERVS": {"-": ["21"]},
    "PVERVLVS": {"-": ["12"]},
    "PVGA": {"-": ["11"]},
    "PVGIL": {"-": ["13"]},
    "PVGILLAR": {"-": ["13"]},
    "PVGILLARES": {"-": ["13"]},
    "PVGILLARIS": {"-": ["24"]},
    "PVGILLVS": {"-": ["12"]},
    "PVGIO": {"-": ["13"]},
    "PVGNA": {"-": ["11"]},
    "PVGNACITAS": {"-": ["13"]},
    "PVGNACITER": {"-": ["6&"]},
    "PVGNANS": {"-": ["25"]},
    "PVGNAX": {"-": ["25"]},
    "PVGNO": {"-": ["51"]},
    "PVGNVS": {"-": ["12"]},
    "PVLCHER": {"-": ["21"], "C": ["12"], "N": ["12"]},
    "PVLCHRE": {"-": ["60"]},
    "PVLCHRITVDO": {"-": ["13"]},
    "PVLEX": {"-": ["13"]},
    "PVLLARIVS": {"1": ["12"], "2": ["21"]},
    "PVLLATVS": {"-": ["21"]},
    "PVLLVLO": {"-": ["51"]},
    "PVLLVM": {"2": ["12"], "-": ["12"]},
    "PVLLVS": {"1": ["12"], "2": ["21"]},
    "PVLMENTARIVM": {"-": ["12"]},
    "PVLMENTVM": {"-": ["12"]},
    "PVLMO": {"-": ["13"]},
    "PVLPA": {"-": ["11"]},
    "PVLPITVM": {"-": ["12"]},
    "PVLS": {"-": ["13"]},
    "PVLSO": {"-": ["51"]},
    "PVLSVS": {"-": ["14"]},
    "PVLTO": {"-": ["51"]},
    "PVLVEREVS": {"-": ["21"]},
    "PVLVERVLENTVS": {"-": ["21"]},
    "PVLVILLVS": {"N": ["12"], "-": ["12"]},
    "PVLVINAR": {"-": ["13"]},
    "PVLVINATVS": {"-": ["21"]},
    "PVLVINVS": {"-": ["12"]},
    "PVLVIS": {"-": ["13"]},
    "PVLVISCVLVS": {"-": ["12"]},
    "PVMEX": {"-": ["13"]},
    "PVMICO": {"-": ["51"]},
    "PVMICOSVS": {"-": ["21"]},
    "PVMILIO": {"-": ["13"]},
    "PVNCTIVNCVLA": {"-": ["11"]},
    "PVNCTVM": {"-": ["12"]},
    "PVNCTVS": {"1": ["14"], "2": ["21"]},
    "PVNGO": {"-": ["53"]},
    "PVNICEVS": {"-": ["21"], "A": ["21"]},
    "PVNICIVS": {"-": ["21"]},
    "PVNICVM": {"-": ["12"]},
    "PVNICVS": {"-": ["21"], "N": ["21"], "A": ["21"]},
    "PVNIO": {"-": ["54"]},
    "PVNIOR": {"-": ["5M"]},
    "PVPA": {"-": ["11"]},
    "PVPILLA": {"-": ["11"]},
    "PVPILLARIS": {"-": ["24"]},
    "PVPILLVS": {"-": ["12"]},
    "PVPPIS": {"-": ["13"]},
    "PVPVLA": {"-": ["11"]},
    "PVPVLVS": {"-": ["12"]},
    "PVRE": {"-": ["60"]},
    "PVRGAMENTVM": {"-": ["12"]},
    "PVRGATIO": {"-": ["13"]},
    "PVRGATVS": {"-": ["21"]},
    "PVRGO": {"-": ["51"]},
    "PVRITER": {"-": ["60"]},
    "PVRPVRA": {"-": ["11"]},
    "PVRPVRATVS": {"1": ["12"], "2": ["21"]},
    "PVRPVREVS": {"-": ["21"]},
    "PVRVLENTVS": {"-": ["21"]},
    "PVRVS": {"-": ["21"]},
    "PVS": {"-": ["13"]},
    "PVSILLVM": {"1": ["12"], "2": ["60"]},
    "PVSILLVS": {"-": ["21"]},
    "PVSIO": {"-": ["13"], "N": ["13"]},
    "PVSTVLA": {"-": ["11"]},
    "PVSVLA": {"-": ["11"]},
    "PVTAMEN": {"-": ["13"]},
    "PVTATOR": {"-": ["13"]},
    "PVTEAL": {"-": ["13"], "N": ["13"]},
    "PVTEALIS": {"-": ["24"]},
    "PVTEO": {"-": ["52"]},
    "PVTEOLI": {"N": ["12"]},
    "PVTESCO": {"-": ["53"]},
    "PVTEVS": {"-": ["12"]},
    "PVTIDE": {"-": ["60"]},
    "PVTIDVS": {"-": ["21"]},
    "PVTO": {"-": ["51"]},
    "PVTOR": {"-": ["13"]},
    "PVTREFACIO": {"-": ["55"]},
    "PVTREO": {"-": ["52"]},
    "PVTRESCO": {"-": ["53"]},
    "PVTRIDVS": {"-": ["21"]},
    "PVTRIS": {"-": ["24"]},
    "PYCNOSTYLOS": {"-": ["27"]},
    "PYGARGVS": {"-": ["12"]},
    "PYLAEMENES": {"N": ["13"]},
    "PYRA": {"-": ["11"]},
    "PYRAGRO": {"N": ["17"]},
    "PYRAMIS": {"-": ["17"], "N": ["13"]},
    "PYRENAEVS": {"N": ["21"], "A": ["21"]},
    "PYRETHRVM": {"-": ["12"]},
    "PYROPVS": {"-": ["12"]},
    "PYTHAGORAS": {"N": ["17"]},
    "PYTHAVLA": {"-": ["17"]},
    "PYTHIA": {"N": ["12", "11"], "O": ["11"]},
    "PYTHICVS": {"N": ["21"], "A": ["21"]},
    "PYTISMA": {"-": ["17"]},
    "PYXIS": {"-": ["17"]},
    "QVA": {"1": ["66"], "2": ["67"], "3": ["60"]},
    "QVACVMQVE": {"1": ["66"], "2": ["60"]},
    "QVADRA": {"-": ["11"]},
    "QVADRAGENARIVS": {"-": ["21"]},
    "QVADRAGENI": {"-": ["33"]},
    "QVADRAGESIMA": {"-": ["11"]},
    "QVADRAGESIMVS": {"-": ["32"]},
    "QVADRAGIES": {"-": ["36"]},
    "QVADRAGINTA": {"-": ["31"]},
    "QVADRANS": {"-": ["13"]},
    "QVADRANTAL": {"-": ["13"]},
    "QVADRANTARIVS": {"-": ["21"]},
    "QVADRATIO": {"-": ["13"]},
    "QVADRATVM": {"-": ["12"]},
    "QVADRATVS": {"1": ["12"], "2": ["21"], "N": ["12"]},
    "QVADRIANGVLVS": {"-": ["21"]},
    "QVADRIDENS": {"-": ["25"]},
    "QVADRIDVVM": {"-": ["12"]},
    "QVADRIENNIVM": {"-": ["12"]},
    "QVADRIFARIAM": {"-": ["60"]},
    "QVADRIFIDVS": {"-": ["21"]},
    "QVADRIFLVVIVM": {"-": ["12"]},
    "QVADRIGA": {"-": ["11"]},
    "QVADRIIVGI": {"-": ["12"]},
    "QVADRIIVGIS": {"-": ["24"]},
    "QVADRIIVGVS": {"-": ["21"]},
    "QVADRIMVS": {"-": ["21"]},
    "QVADRINGENI": {"-": ["34", "21"]},
    "QVADRINGENTESIMVS": {"-": ["32"]},
    "QVADRINGENTI": {"-": ["31"]},
    "QVADRINGENTIES": {"-": ["36"]},
    "QVADRINI": {"-": ["33"]},
    "QVADRIPARTITVS": {"-": ["21"]},
    "QVADRIREMIS": {"1": ["13"], "2": ["24"]},
    "QVADRIVIVM": {"-": ["12"]},
    "QVADRO": {"-": ["51"]},
    "QVADRVPEDVS": {"-": ["21"]},
    "QVADRVPES": {"1": ["13"], "2": ["25"]},
    "QVADRVPLATOR": {"-": ["13"]},
    "QVADRVPLEX": {"-": ["34", "13"]},
    "QVADRVPLO": {"1": ["51"], "2": ["60"]},
    "QVADRVPLVM": {"-": ["12"]},
    "QVAERITO": {"-": ["51"]},
    "QVAERO": {"-": ["53"]},
    "QVAESITOR": {"-": ["13"]},
    "QVAESITVM": {"-": ["12"]},
    "QVAESITVS": {"1": ["14"], "2": ["21"]},
    "QVAESO": {"-": ["56"]},
    "QVAESTIO": {"-": ["13"]},
    "QVAESTIVNCVLA": {"-": ["11"]},
    "QVAESTOR": {"-": ["13"]},
    "QVAESTORIVM": {"-": ["12"]},
    "QVAESTORIVS": {"1": ["12"], "2": ["21"]},
    "QVAESTVARIA": {"-": ["11"]},
    "QVAESTVOSE": {"-": ["6-"]},
    "QVAESTVOSVS": {"-": ["21"]},
    "QVAESTVRA": {"-": ["11"]},
    "QVAESTVS": {"-": ["14"]},
    "QVALIBET": {"-": ["60"]},
    "QVALIS": {"1": ["46"], "2": ["47"]},
    "QVALISCVMQVE": {"1": ["46"], "2": ["48"]},
    "QVALITAS": {"-": ["13"]},
    "QVALITER": {"1": ["66"], "2": ["67"]},
    "QVALVM": {"-": ["12"]},
    "QVALVS": {"-": ["12"]},
    "QVAM": {"1": ["66"], "2": ["67"]},
    "QVAMDIV": {"1": ["66"], "2": ["67"]},
    "QVAMDVDVM": {"-": ["67"]},
    "QVAMLIBET": {"1": ["66"], "2": ["60"]},
    "QVAMOBREM": {"1": ["66"], "2": ["67"]},
    "QVAMPLVRES": {"-": ["22"]},
    "QVAMPRIMVM": {"-": ["60"]},
    "QVAMQVAM": {"1": ["66"], "2": ["82"], "3": ["60"]},
    "QVAMVIS": {"1": ["66"], "2": ["82"]},
    "QVANAM": {"-": ["67"]},
    "QVANDANTENVS": {"-": ["60"]},
    "QVANDO": {"1": ["66"], "2": ["67"], "3": ["82"], "4": ["60"]},
    "QVANDOCVMQVE": {"1": ["66"], "2": ["60"]},
    "QVANDOQVE": {"1": ["60"], "2": ["82"]},
    "QVANDOQVIDEM": {"-": ["82"]},
    "QVANTITAS": {"-": ["13"]},
    "QVANTO": {"1": ["66"], "2": ["67"]},
    "QVANTOPERE": {"1": ["66"], "2": ["67"]},
    "QVANTVLVM": {"1": ["46"], "2": ["47"], "3": ["66"], "4": ["67"]},
    "QVANTVLVS": {"1": ["46"], "2": ["47"]},
    "QVANTVLVSCVMQVE": {"1": ["46"], "2": ["48"]},
    "QVANTVLVSLIBET": {"1": ["46"], "2": ["48"]},
    "QVANTVM": {"1": ["46"], "2": ["47"], "3": ["66"], "4": ["67"]},
    "QVANTVMCVMQVE": {"-": ["66"]},
    "QVANTVMLIBET": {"1": ["48"], "2": ["60"]},
    "QVANTVMVIS": {"1": ["60"], "2": ["82"]},
    "QVANTVS": {"1": ["46"], "2": ["47"]},
    "QVANTVSCVMQVE": {"1": ["46"], "2": ["48"]},
    "QVANTVSLIBET": {"-": ["48"]},
    "QVANTVSVIS": {"-": ["48"]},
    "QVAPROPTER": {"1": ["66", "6-"], "2": ["67"]},
    "QVAQVAVERSVM": {"-": ["60"]},
    "QVAQVE": {"-": ["60"]},
    "QVARE": {"1": ["66"], "2": ["67"]},
    "QVARTADECIMANI": {"-": ["12"]},
    "QVARTANA": {"-": ["11"]},
    "QVARTANVS": {"-": ["21"]},
    "QVARTARIVS": {"-": ["12"]},
    "QVARTO": {"-": ["35"]},
    "QVARTVM": {"1": ["12"], "2": ["36"]},
    "QVARTVS": {"-": ["32"]},
    "QVASI": {"1": ["66"], "2": ["82"]},
    "QVASILLVS": {"-": ["12"]},
    "QVASSATIO": {"-": ["13"]},
    "QVASSO": {"-": ["51"]},
    "QVASSVS": {"1": ["14"], "2": ["21"]},
    "QVATENVS": {"1": ["66"], "2": ["67"], "3": ["82"]},
    "QVATER": {"-": ["36"]},
    "QVATERNI": {"-": ["33"]},
    "QVATIO": {"-": ["55"]},
    "QVATRIDVVM": {"-": ["12"]},
    "QVATVOR": {"-": ["31"]},
    "QVATVORDECIM": {"-": ["31"]},
    "QVATVORVIR": {"-": ["12"]},
    "QVAVIS": {"-": ["60"]},
    "QVE": {"-": ["81"]},
    "QVEMADMODVM": {"1": ["66"], "2": ["67"]},
    "QVEO": {"-": ["56"]},
    "QVERCANEVS": {"-": ["21"]},
    "QVERCEVS": {"-": ["21"]},
    "QVERCVS": {"-": ["16"]},
    "QVERELA": {"-": ["11"]},
    "QVERIMONIA": {"-": ["11"]},
    "QVERITOR": {"-": ["5J"]},
    "QVERNEVS": {"-": ["21"]},
    "QVERNVS": {"-": ["21"]},
    "QVEROR": {"-": ["5L"]},
    "QVERQVETVM": {"-": ["12"]},
    "QVERVLVS": {"-": ["21"]},
    "QVESTVS": {"-": ["14"]},
    "QVI": {"1": ["46"], "2": ["66"], "3": ["67"], "4": ["60"]},
    "QVIA": {"-": ["82"]},
    "QVIANAM": {"-": ["67"]},
    "QVICVMQVE": {"1": ["46"], "2": ["48"]},
    "QVIDAM": {"-": ["48"]},
    "QVIDEM": {"-": ["60"]},
    "QVIDNI": {"-": ["69", "6-"]},
    "QVIES": {"-": ["13"]},
    "QVIESCO": {"-": ["53"]},
    "QVIETE": {"-": ["60"]},
    "QVIETO": {"-": ["51"]},
    "QVIETVM": {"-": ["12"]},
    "QVIETVS": {"-": ["21"], "N": ["12"]},
    "QVILIBET": {"-": ["48"]},
    "QVIN": {"1": ["60"], "2": ["82"], "3": ["69", "6-"]},
    "QVINARIVS": {"-": ["21"]},
    "QVINCVNX": {"-": ["13"]},
    "QVINDECIES": {"-": ["36"]},
    "QVINDECIM": {"-": ["31"]},
    "QVINDECIMVIRALIS": {"-": ["24"]},
    "QVINDECIMVIRI": {"-": ["12"]},
    "QVINGENARIVS": {"-": ["21"]},
    "QVINGENI": {"-": ["33"]},
    "QVINGENTESIMVS": {"-": ["32"]},
    "QVINGENTI": {"-": ["31"]},
    "QVINI": {"-": ["33"]},
    "QVINIDENI": {"-": ["33"]},
    "QVINQVAGENARIVS": {"-": ["21"]},
    "QVINQVAGENI": {"-": ["33"]},
    "QVINQVAGESIMA": {"-": ["11"]},
    "QVINQVAGESIMVS": {"-": ["32"]},
    "QVINQVAGINTA": {"-": ["31"]},
    "QVINQVATRIA": {"-": ["13"]},
    "QVINQVE": {"-": ["31"]},
    "QVINQVENNALIS": {"-": ["24"]},
    "QVINQVENNIS": {"-": ["24"]},
    "QVINQVENNIVM": {"-": ["12"]},
    "QVINQVEREMIS": {"1": ["13"], "2": ["24"]},
    "QVINQVEVIR": {"-": ["12"]},
    "QVINQVIES": {"-": ["36"]},
    "QVINQVIPLICO": {"-": ["51"]},
    "QVINTADECIMANI": {"-": ["12"]},
    "QVINTANI": {"-": ["12"]},
    "QVINTARIVS": {"-": ["21"]},
    "QVINTILIS": {"A": ["24"], "N": ["13"]},
    "QVINTIVS": {"N": ["12"]},
    "QVINTVM": {"-": ["36"]},
    "QVINTVS": {"-": ["32"], "N": ["12"]},
    "QVIPPE": {"1": ["60"], "2": ["81"]},
    "QVIQVI": {"-": ["48"]},
    "QVIRINALIS": {"N": ["24"], "A": ["24"]},
    "QVIRINVS": {"N": ["12"], "A": ["21"]},
    "QVIRITES": {"N": ["13"]},
    "QVIRITO": {"-": ["51"]},
    "QVIRO": {"-": ["53"]},
    "QVIS": {"1": ["47"], "2": ["48"]},
    "QVISNAM": {"-": ["47"]},
    "QVISPIAM": {"-": ["48"]},
    "QVISQVAM": {"-": ["48"]},
    "QVISQVE": {"1": ["46"], "2": ["48"]},
    "QVISQVIS": {"1": ["46"], "2": ["48"]},
    "QVIVIS": {"-": ["48"]},
    "QVIVISCVMQVE": {"-": ["48"]},
    "QVO": {
        "1": ["66"],
        "2": ["66"],
        "3": ["67"],
        "4": ["60"],
        "5": ["82", "6-"],
    },
    "QVOAD": {"1": ["66"], "2": ["67"]},
    "QVOCIRCA": {"-": ["66"]},
    "QVOCVMQVE": {"1": ["66"], "2": ["60"]},
    "QVOD": {"1": ["81"], "2": ["82"]},
    "QVODAMMODO": {"-": ["60"]},
    "QVOLIBET": {"-": ["60"]},
    "QVOMINVS": {"-": ["82"]},
    "QVOMODO": {"1": ["66"], "2": ["67"]},
    "QVOMODOCVMQVE": {"1": ["66"], "2": ["60"]},
    "QVONAM": {"-": ["67"]},
    "QVONDAM": {"-": ["60"]},
    "QVONIAM": {"-": ["82"]},
    "QVOQVAM": {"-": ["60"]},
    "QVOQVE": {"-": ["60"]},
    "QVOQVO": {"-": ["66"]},
    "QVOQVOVERSVS": {"-": ["60"]},
    "QVORSVM": {"-": ["67"]},
    "QVOT": {"1": ["46"], "2": ["47"], "3": ["48"]},
    "QVOTANNIS": {"-": ["60"]},
    "QVOTCVMQVE": {"-": ["46"]},
    "QVOTIENS": {"1": ["66"], "2": ["67"]},
    "QVOTIENSCVMQVE": {"-": ["66"]},
    "QVOTQVOT": {"1": ["46"], "2": ["48"]},
    "QVOTVIS": {"-": ["48"]},
    "QVOTVS": {"-": ["47"]},
    "QVOTVSCVMQVE": {"-": ["46"]},
    "QVOTVSQVISQVE": {"-": ["47"]},
    "QVOVIS": {"-": ["66"]},
    "QVOVSQVE": {"-": ["67"], "2": ["6-"]},
    "R": {"-": ["16"]},
    "RABIDE": {"-": ["60"]},
    "RABIDVS": {"-": ["21"]},
    "RABIES": {"-": ["15"]},
    "RABIOSVS": {"-": ["21"]},
    "RABO": {"2": ["53"]},
    "RACEMVS": {"-": ["12"]},
    "RACILIVS": {"N": ["12"]},
    "RADIANS": {"-": ["25"]},
    "RADIATVS": {"-": ["21"]},
    "RADICESCO": {"-": ["53"]},
    "RADICITVS": {"-": ["60"]},
    "RADICVLA": {"-": ["11"]},
    "RADIO": {"-": ["51"]},
    "RADIVS": {"-": ["12"]},
    "RADIX": {"-": ["13"]},
    "RADO": {"-": ["53"]},
    "RAECIVS": {"N": ["12"]},
    "RAEDA": {"-": ["11"]},
    "RAMALE": {"-": ["13"]},
    "RAMENTVM": {"-": ["12"]},
    "RAMEVS": {"-": ["21"]},
    "RAMEX": {"-": ["13"]},
    "RAMNENSES": {"N": ["13"]},
    "RAMNES": {"N": ["13"]},
    "RAMOSVS": {"-": ["21"]},
    "RAMVLVS": {"-": ["12"]},
    "RAMVS": {"-": ["12"]},
    "RANA": {"-": ["11"]},
    "RANCENS": {"-": ["25"]},
    "RANCIDVLVS": {"-": ["21"]},
    "RANCIDVS": {"-": ["21"]},
    "RAPACITAS": {"-": ["13"]},
    "RAPAX": {"-": ["25"], "A": ["13"]},
    "RAPHANVS": {"-": ["12"]},
    "RAPICIVS": {"-": ["21"]},
    "RAPIDE": {"-": ["60"]},
    "RAPIDITAS": {"-": ["13"]},
    "RAPIDVS": {"-": ["21"]},
    "RAPINA": {"1": ["11"], "2": ["11"]},
    "RAPIO": {"-": ["55"]},
    "RAPSO": {"-": ["51"]},
    "RAPTIM": {"-": ["60"]},
    "RAPTO": {"-": ["51"]},
    "RAPTOR": {"-": ["13"]},
    "RAPTVM": {"-": ["12"]},
    "RAPTVS": {"-": ["14"]},
    "RAPVLVM": {"-": ["12"]},
    "RAPVM": {"-": ["12"]},
    "RARE": {"-": ["60"]},
    "RAREFACIO": {"-": ["55"]},
    "RAREFIO": {"-": ["56"]},
    "RARENTER": {"-": ["60"]},
    "RARESCO": {"-": ["53"]},
    "RARITAS": {"-": ["13"]},
    "RARO": {"-": ["60"]},
    "RARVS": {"-": ["21"]},
    "RASILIS": {"-": ["24"]},
    "RASTER": {"-": ["12"]},
    "RASTRI": {"-": ["12"]},
    "RASTRVM": {"-": ["12"]},
    "RATIO": {"-": ["13"]},
    "RATIOCINATIO": {"-": ["13"]},
    "RATIOCINATOR": {"-": ["13"]},
    "RATIOCINO": {"-": ["51"]},
    "RATIOCINOR": {"-": ["5J"]},
    "RATIONALIS": {"1": ["13"], "2": ["24"]},
    "RATIONALITER": {"-": ["60"]},
    "RATIS": {"-": ["13"]},
    "RATIVNCVLA": {"-": ["11"]},
    "RATVS": {"-": ["21"]},
    "RAVCISONVS": {"-": ["21"]},
    "RAVCVS": {"-": ["21"]},
    "RAVVS": {"-": ["21"]},
    "REA": {"-": ["11"]},
    "REALIS": {"-": ["24"]},
    "REAPSE": {"-": ["60"]},
    "REBELLATIO": {"-": ["13"]},
    "REBELLIO": {"-": ["13"]},
    "REBELLIS": {"-": ["24"]},
    "REBELLO": {"-": ["51"]},
    "REBOO": {"-": ["51"]},
    "RECALCITRO": {"-": ["51"]},
    "RECALESCO": {"-": ["53"]},
    "RECALFACIO": {"-": ["55"]},
    "RECANTO": {"-": ["51"]},
    "RECEDO": {"1": ["53"]},
    "RECELLO": {"-": ["53"]},
    "RECENS": {"1": ["25"], "2": ["60"]},
    "RECENSEO": {"-": ["52"]},
    "RECEPIO": {"-": ["55"]},
    "RECEPTACVLVM": {"-": ["12"]},
    "RECEPTICIVS": {"-": ["21"]},
    "RECEPTO": {"-": ["51"]},
    "RECEPTOR": {"-": ["13"]},
    "RECEPTRIX": {"-": ["13"]},
    "RECEPTVM": {"-": ["12"]},
    "RECEPTVS": {"1": ["14"], "2": ["21"], "N": ["12"]},
    "RECESSIO": {"-": ["13"]},
    "RECESSVS": {"1": ["14"], "2": ["21"]},
    "RECHAMVS": {"-": ["12"]},
    "RECIDIVVS": {"-": ["21"]},
    "RECIDO": {"1": ["53"], "2": ["53"]},
    "RECINGO": {"-": ["53"]},
    "RECINO": {"-": ["53"]},
    "RECIPIO": {"-": ["55"]},
    "RECIPROCO": {"-": ["51"]},
    "RECIPROCVS": {"-": ["21"]},
    "RECISVS": {"-": ["21"]},
    "RECITATIO": {"-": ["13"]},
    "RECITATOR": {"-": ["13"]},
    "RECITO": {"1": ["51"]},
    "RECLAMO": {"-": ["51"]},
    "RECLINIS": {"-": ["24"]},
    "RECLINO": {"-": ["51"]},
    "RECLVDO": {"-": ["53"]},
    "RECOGITO": {"-": ["51"]},
    "RECOGNITIO": {"-": ["13"]},
    "RECOGNOSCO": {"-": ["53"]},
    "RECOLLIGO": {"-": ["53"], "3": ["53"]},
    "RECOLO": {"1": ["51"], "2": ["53"]},
    "RECOMPONO": {"-": ["53"]},
    "RECONCILIATIO": {"-": ["13"]},
    "RECONCILIO": {"-": ["51"]},
    "RECONDITVS": {"-": ["21"]},
    "RECONDO": {"-": ["53"]},
    "RECONFLO": {"-": ["51"]},
    "RECOQVO": {"-": ["53"]},
    "RECORDATIO": {"-": ["13"]},
    "RECORDOR": {"-": ["5J"]},
    "RECORRIGO": {"-": ["53"]},
    "RECREO": {"-": ["51"]},
    "RECREPO": {"-": ["51"]},
    "RECRESCO": {"-": ["53"]},
    "RECRVDESCO": {"-": ["53"]},
    "RECTA": {"-": ["60"]},
    "RECTE": {"-": ["60"]},
    "RECTOR": {"-": ["13"]},
    "RECTRIX": {"-": ["13"]},
    "RECTVM": {"-": ["12"]},
    "RECTVS": {"-": ["21"]},
    "RECVBO": {"-": ["51"]},
    "RECVMBO": {"-": ["53"]},
    "RECVPERATOR": {"-": ["13"]},
    "RECVPERATORIVS": {"-": ["21"]},
    "RECVPERO": {"-": ["51"]},
    "RECVRO": {"-": ["51"]},
    "RECVRRO": {"-": ["53"]},
    "RECVRSO": {"-": ["51"]},
    "RECVRSVS": {"-": ["14"]},
    "RECVRVO": {"-": ["51"]},
    "RECVRVVS": {"-": ["21"]},
    "RECVSATIO": {"-": ["13"]},
    "RECVSO": {"-": ["51"]},
    "RECVTIO": {"-": ["55"]},
    "RECVTITVS": {"-": ["21"]},
    "REDAMO": {"-": ["51"]},
    "REDARDESCO": {"-": ["53"]},
    "REDARGVO": {"-": ["53"]},
    "REDDO": {"-": ["53"]},
    "REDDVCO": {"-": ["53"]},
    "REDEMPTIO": {"-": ["13"]},
    "REDEMPTITO": {"-": ["51"]},
    "REDEMPTOR": {"-": ["13"]},
    "REDEO": {"1": ["56"]},
    "REDHALO": {"-": ["51"]},
    "REDIGO": {"-": ["53"]},
    "REDIMICVLVM": {"-": ["12"]},
    "REDIMIO": {"-": ["54"]},
    "REDIMO": {"-": ["53"]},
    "REDINTEGRO": {"-": ["51"]},
    "REDITIO": {"-": ["13"]},
    "REDITVS": {"-": ["14"]},
    "REDIVIVVS": {"-": ["21"]},
    "REDOLEO": {"-": ["52"]},
    "REDONO": {"-": ["51"]},
    "REDVCO": {"-": ["53"]},
    "REDVCTIO": {"-": ["13"]},
    "REDVCTVS": {"-": ["21"]},
    "REDVNDANTIA": {"-": ["11"]},
    "REDVNDATIO": {"-": ["13"]},
    "REDVNDO": {"-": ["51"]},
    "REDVRESCO": {"-": ["53"]},
    "REDVX": {"-": ["25"]},
    "REFECTIO": {"-": ["13"]},
    "REFELLO": {"-": ["53"]},
    "REFERCIO": {"-": ["54"]},
    "REFERIO": {"-": ["54"]},
    "REFERO": {"-": ["56"]},
    "REFERT": {"-": ["56"]},
    "REFERTVS": {"-": ["21"]},
    "REFICIO": {"-": ["55"]},
    "REFIGO": {"-": ["53"]},
    "REFINGO": {"-": ["53"]},
    "REFLAGITO": {"-": ["51"]},
    "REFLECTO": {"-": ["53"]},
    "REFLO": {"-": ["51"]},
    "REFLVO": {"-": ["53"]},
    "REFLVVS": {"-": ["21"]},
    "REFORMATIO": {"-": ["13"]},
    "REFORMIDO": {"2": ["51"]},
    "REFORMO": {"-": ["51"]},
    "REFOVEO": {"-": ["52"]},
    "REFRACTARIVS": {"-": ["21"]},
    "REFRAGOR": {"-": ["5J"]},
    "REFRENATIO": {"-": ["13"]},
    "REFRENO": {"-": ["51"]},
    "REFRICO": {"-": ["51"]},
    "REFRIGERATIO": {"-": ["13"]},
    "REFRIGERO": {"-": ["51"]},
    "REFRIGESCO": {"-": ["53"]},
    "REFRINGO": {"-": ["53"]},
    "REFVGIO": {"-": ["55"]},
    "REFVGVS": {"1": ["12"], "2": ["21"]},
    "REFVLGEO": {"-": ["52"]},
    "REFVNDO": {"2": ["53"]},
    "REFVTATVS": {"-": ["14"]},
    "REFVTO": {"-": ["51"]},
    "REGALIS": {"-": ["24"]},
    "REGELO": {"-": ["51"]},
    "REGENS": {"-": ["13"]},
    "REGERO": {"-": ["53"]},
    "REGIA": {"-": ["11"], "N": ["11"]},
    "REGIE": {"-": ["60"]},
    "REGIFICVS": {"-": ["21"]},
    "REGIGNO": {"-": ["53"]},
    "REGILLVS": {"1N": ["12"], "2N": ["12"], "1": ["12"], "2": ["12"]},
    "REGIMEN": {"-": ["13"]},
    "REGINA": {"-": ["11"], "N": ["11"]},
    "REGINI": {"N": ["12"]},
    "REGIO": {"-": ["13"]},
    "REGIVM": {"N": ["12"]},
    "REGIVS": {"-": ["21"], "N": ["12"]},
    "REGLVTINO": {"-": ["51"]},
    "REGNATOR": {"-": ["13"]},
    "REGNATRIX": {"-": ["25"]},
    "REGNO": {"-": ["51"]},
    "REGNVM": {"-": ["12"]},
    "REGO": {"-": ["53"]},
    "REGREDIOR": {"-": ["5N"]},
    "REGRESSVS": {"-": ["14"]},
    "REGVLA": {"-": ["11"]},
    "REGVLVS": {"N": ["12"], "-": ["12"]},
    "REGVSTO": {"-": ["51"]},
    "REICIO": {"-": ["55"]},
    "REICVLVS": {"-": ["21"]},
    "REIECTIO": {"-": ["13"]},
    "REIECTO": {"-": ["51"]},
    "RELABOR": {"2": ["5L"]},
    "RELANGVESCO": {"-": ["53"]},
    "RELATIO": {"-": ["13"]},
    "RELATVS": {"-": ["14"]},
    "RELAXO": {"-": ["51"]},
    "RELEGATIO": {"-": ["13"]},
    "RELEGO": {"1": ["51"], "2": ["53"]},
    "RELENTESCO": {"-": ["53"]},
    "RELEVATVS": {"C": ["21"], "-": ["21"]},
    "RELEVO": {"1": ["51"]},
    "RELICTIO": {"-": ["13"]},
    "RELICTVM": {"-": ["12"]},
    "RELICTVS": {"1": ["14"], "2": ["21"]},
    "RELIGIO": {"-": ["13"]},
    "RELIGIOSE": {"-": ["60"]},
    "RELIGIOSVS": {"-": ["21"]},
    "RELIGO": {"2": ["51"]},
    "RELINO": {"-": ["53"]},
    "RELINQVO": {"-": ["53"]},
    "RELIQVA": {"-": ["12"]},
    "RELIQVI": {"-": ["12"]},
    "RELIQVIAE": {"-": ["11"]},
    "RELIQVVM": {"-": ["12"]},
    "RELIQVVS": {"-": ["21"]},
    "RELVCEO": {"-": ["52"]},
    "RELVCTOR": {"-": ["5J"]},
    "REMANEO": {"-": ["52"]},
    "REMANO": {"-": ["51"]},
    "REMEDIABILIS": {"-": ["24"]},
    "REMEDIVM": {"-": ["12"]},
    "REMEO": {"-": ["51"]},
    "REMETIOR": {"-": ["5M"]},
    "REMEX": {"-": ["13"]},
    "REMIGIVM": {"-": ["12"]},
    "REMIGO": {"-": ["51"]},
    "REMIGRO": {"-": ["51"]},
    "REMINISCO": {"-": ["53"]},
    "REMINISCOR": {"-": ["5L"]},
    "REMISCEO": {"-": ["52"]},
    "REMISSARIVS": {"-": ["21"]},
    "REMISSE": {"-": ["6&"]},
    "REMISSIO": {"-": ["13"]},
    "REMISSVS": {"-": ["21"]},
    "REMITTO": {"-": ["53"]},
    "REMOLIOR": {"-": ["5M"]},
    "REMOLLESCO": {"-": ["53"]},
    "REMOLLIO": {"-": ["54"]},
    "REMORDEO": {"-": ["52"]},
    "REMOROR": {"1": ["5J"]},
    "REMOTA": {"-": ["12"]},
    "REMOTVS": {"-": ["21"]},
    "REMOVEO": {"-": ["52"]},
    "REMVGIO": {"-": ["54"]},
    "REMVLCVM": {"-": ["12"]},
    "REMVNERATIO": {"-": ["13"]},
    "REMVNERO": {"-": ["51"]},
    "REMVNEROR": {"-": ["5J"]},
    "REMVS": {"-": ["12"], "N": ["12"], "A": ["21"]},
    "RENARRO": {"-": ["51"]},
    "RENASCOR": {"-": ["5L"]},
    "RENAVIGO": {"-": ["51"]},
    "RENES": {"-": ["13"]},
    "RENIDEO": {"-": ["52"]},
    "RENIDESCO": {"-": ["53"]},
    "RENO": {"-": ["51"]},
    "RENODO": {"-": ["51"]},
    "RENOVO": {"-": ["51"]},
    "RENVMERO": {"1": ["51"]},
    "RENVNTIATIO": {"-": ["13"]},
    "RENVNTIO": {"-": ["51"]},
    "RENVO": {"-": ["53"]},
    "RENVTO": {"-": ["51"]},
    "REOR": {"-": ["5K"]},
    "REPAGVLA": {"-": ["12"]},
    "REPARABILIS": {"-": ["24"]},
    "REPARCO": {"-": ["53"]},
    "REPARO": {"2": ["51"]},
    "REPASTINO": {"-": ["51"]},
    "REPECTO": {"-": ["53"]},
    "REPELLO": {"-": ["53"]},
    "REPENDO": {"-": ["53"]},
    "REPENS": {"1": ["25"], "2": ["60"]},
    "REPENSO": {"-": ["51"]},
    "REPENTE": {"-": ["60"]},
    "REPENTINO": {"-": ["60"]},
    "REPENTINVS": {"-": ["21"], "N": ["12"]},
    "REPERCVSSVS": {"-": ["14"]},
    "REPERCVTIO": {"-": ["55"]},
    "REPERIO": {"-": ["54"]},
    "REPERTOR": {"-": ["13"]},
    "REPERTVM": {"-": ["12"]},
    "REPERTVS": {"1": ["14"], "2": ["21"]},
    "REPETENTIA": {"-": ["11"]},
    "REPETITIO": {"-": ["13"]},
    "REPETITOR": {"-": ["13"]},
    "REPETO": {"-": ["53"]},
    "REPETVNDAE": {"-": ["11"]},
    "REPLEO": {"-": ["52"]},
    "REPLETVS": {"-": ["21"]},
    "REPLICO": {"-": ["51"]},
    "REPLVM": {"-": ["12"]},
    "REPO": {"-": ["53"]},
    "REPONO": {"-": ["53"]},
    "REPORTO": {"-": ["51"]},
    "REPOSCO": {"-": ["53"]},
    "REPOSITORIVM": {"-": ["12"]},
    "REPOSITVS": {"-": ["21"]},
    "REPOTIA": {"-": ["12"]},
    "REPRAESENTATIO": {"-": ["13"]},
    "REPRAESENTO": {"-": ["51"]},
    "REPREHENDO": {"-": ["53"]},
    "REPREHENSIO": {"-": ["13"]},
    "REPREHENSOR": {"-": ["13"]},
    "REPRIMO": {"-": ["53"]},
    "REPTABVNDVS": {"-": ["21"]},
    "REPTO": {"-": ["51"]},
    "REPVDIATIO": {"-": ["13"]},
    "REPVDIO": {"-": ["51"]},
    "REPVDIVM": {"-": ["12"]},
    "REPVGNANS": {"-": ["25"]},
    "REPVGNANTER": {"-": ["60"]},
    "REPVGNATORIVS": {"-": ["21"]},
    "REPVGNO": {"-": ["51"]},
    "REPVLLVLO": {"-": ["51"]},
    "REPVLSA": {"-": ["11"]},
    "REPVLSO": {"-": ["51"]},
    "REPVLSVS": {"1": ["14"], "2": ["21"]},
    "REPVRGO": {"-": ["51"]},
    "REPVTATIO": {"-": ["13"]},
    "REPVTO": {"-": ["51"]},
    "REQVIES": {"-": ["16"]},
    "REQVIESCO": {"-": ["53"]},
    "REQVIRO": {"-": ["53"]},
    "RES": {"-": ["15"]},
    "RESALVTO": {"-": ["51"]},
    "RESANESCO": {"-": ["53"]},
    "RESARCIO": {"-": ["54"]},
    "RESCINDO": {"-": ["53"]},
    "RESCIO": {"-": ["54"]},
    "RESCISCO": {"-": ["53"]},
    "RESCRIBO": {"-": ["53"]},
    "RESCRIPTVM": {"-": ["12"]},
    "RESECO": {"-": ["51"]},
    "RESEQVOR": {"-": ["5L"]},
    "RESERO": {"1": ["51"], "3": ["53"]},
    "RESERVO": {"-": ["51"]},
    "RESES": {"-": ["25"]},
    "RESIDEO": {"-": ["52"]},
    "RESIDO": {"-": ["53"]},
    "RESIDVVS": {"-": ["21"]},
    "RESIGNO": {"-": ["51"]},
    "RESILIO": {"-": ["54"]},
    "RESINA": {"-": ["11"]},
    "RESINATVS": {"-": ["21"]},
    "RESIPISCO": {"-": ["53"]},
    "RESISTO": {"-": ["53"]},
    "RESOLVO": {"-": ["53"]},
    "RESOLVTVS": {"-": ["21"]},
    "RESONABILIS": {"-": ["24"]},
    "RESONANTIA": {"-": ["11"]},
    "RESONO": {"1": ["51"]},
    "RESORBEO": {"-": ["52"]},
    "RESPECTO": {"-": ["51"]},
    "RESPECTVS": {"-": ["14"]},
    "RESPERGO": {"-": ["53"]},
    "RESPICIO": {"-": ["55"]},
    "RESPIRATIO": {"-": ["13"]},
    "RESPIRO": {"-": ["51"]},
    "RESPLENDEO": {"-": ["52"]},
    "RESPONDEO": {"-": ["52"]},
    "RESPONSIO": {"-": ["13"]},
    "RESPONSO": {"-": ["51"]},
    "RESPONSVM": {"-": ["12"]},
    "RESPONSVS": {"-": ["14"]},
    "RESPVBLICA": {"-": ["15"]},
    "RESPVO": {"-": ["53"]},
    "RESTAGNO": {"1": ["51"]},
    "RESTAVRO": {"-": ["51"]},
    "RESTIBILIS": {"-": ["24"]},
    "RESTICVLA": {"-": ["11"]},
    "RESTINGVO": {"-": ["53"]},
    "RESTIS": {"-": ["13"]},
    "RESTITVO": {"-": ["53"]},
    "RESTO": {"-": ["51"]},
    "RESTRICTE": {"-": ["60"]},
    "RESTRICTVS": {"-": ["21"]},
    "RESTRINGO": {"-": ["53"]},
    "RESVDO": {"-": ["51"]},
    "RESVLTO": {"-": ["51"]},
    "RESVMO": {"-": ["53"]},
    "RESVPINO": {"-": ["51"]},
    "RESVPINVS": {"-": ["21"]},
    "RESVRGO": {"-": ["53"]},
    "RETARDATIO": {"-": ["13"]},
    "RETARDO": {"-": ["51"]},
    "RETE": {"-": ["13"]},
    "RETEGO": {"-": ["53"]},
    "RETENDO": {"-": ["53"]},
    "RETENTIO": {"-": ["13"]},
    "RETENTO": {"1": ["51"], "2": ["51"]},
    "RETERO": {"-": ["53"]},
    "RETEXO": {"-": ["53"]},
    "RETICENTIA": {"-": ["11"]},
    "RETICEO": {"-": ["52"]},
    "RETICESCO": {"-": ["53"]},
    "RETICVLATVS": {"-": ["21"]},
    "RETICVLVM": {"-": ["12"]},
    "RETINA": {"-": ["11"]},
    "RETINACVLVM": {"-": ["12"]},
    "RETINENS": {"-": ["25"]},
    "RETINENTIA": {"-": ["11"]},
    "RETINEO": {"-": ["52"]},
    "RETINGO": {"1": ["53"], "2": ["53"]},
    "RETIO": {"-": ["54"]},
    "RETONO": {"-": ["51"]},
    "RETORQVEO": {"-": ["52"]},
    "RETORRIDVS": {"-": ["21"]},
    "RETRACTATIO": {"-": ["13"]},
    "RETRACTIO": {"-": ["13"]},
    "RETRACTO": {"-": ["51"]},
    "RETRAHO": {"-": ["53"]},
    "RETRIBVO": {"-": ["53"]},
    "RETRO": {"1": ["60"], "2": ["70"]},
    "RETROCEDO": {"1": ["53"]},
    "RETRODVCO": {"-": ["53"]},
    "RETROFLECTO": {"-": ["53"]},
    "RETROLEGO": {"2": ["53"]},
    "RETRORSVM": {"-": ["60"]},
    "RETROVERTO": {"-": ["53"]},
    "RETRVDO": {"-": ["53"]},
    "RETVNDO": {"-": ["53"]},
    "RETVSVS": {"-": ["21"]},
    "REVALESCO": {"-": ["53"]},
    "REVECTO": {"-": ["51"]},
    "REVEHO": {"-": ["53"]},
    "REVELLO": {"-": ["53"]},
    "REVELO": {"-": ["51"]},
    "REVENIO": {"-": ["54"]},
    "REVERA": {"-": ["60"]},
    "REVERBERO": {"2": ["51"]},
    "REVERENS": {"-": ["25"]},
    "REVERENTER": {"-": ["60"]},
    "REVERENTIA": {"-": ["11"], "N": ["11"]},
    "REVEREOR": {"-": ["5K"]},
    "REVERSIO": {"-": ["13"]},
    "REVERSO": {"-": ["51"]},
    "REVERTO": {"-": ["53"]},
    "REVERTOR": {"-": ["5L"]},
    "REVILESCO": {"-": ["53"]},
    "REVINCIO": {"-": ["54"]},
    "REVINCO": {"-": ["53"]},
    "REVISO": {"-": ["53"]},
    "REVIVISCO": {"-": ["53"]},
    "REVIVO": {"-": ["53"]},
    "REVOCABILIS": {"-": ["24"]},
    "REVOCAMEN": {"-": ["13"]},
    "REVOCATIO": {"-": ["13"]},
    "REVOCO": {"-": ["51"]},
    "REVOLO": {"2": ["51"]},
    "REVOLVO": {"-": ["53"]},
    "REVOMO": {"-": ["53"]},
    "REVS": {"-": ["12"]},
    "REX": {"-": ["13"], "N": ["13"]},
    "RHADINE": {"-": ["17"]},
    "RHEA": {"N": ["11"]},
    "RHEDA": {"-": ["11"]},
    "RHENO": {"-": ["13"]},
    "RHETOR": {"-": ["13"]},
    "RHETORICVS": {"-": ["21"]},
    "RHINOCEROS": {"-": ["17"]},
    "RHODII": {"N": ["12"]},
    "RHODIVS": {"N": ["21"], "A": ["21"]},
    "RHODOS": {"N": ["17"]},
    "RHOMBOS": {"-": ["17"]},
    "RHYTHMVS": {"-": ["12"]},
    "RICINVS": {"-": ["12"]},
    "RICTVM": {"-": ["12"]},
    "RICTVS": {"-": ["14"]},
    "RIDENDVS": {"-": ["21"]},
    "RIDEO": {"-": ["52"]},
    "RIDIBVNDVS": {"-": ["21"]},
    "RIDICA": {"-": ["11"]},
    "RIDICVLARIA": {"-": ["12"]},
    "RIDICVLE": {"-": ["60"]},
    "RIDICVLVM": {"-": ["12"]},
    "RIDICVLVS": {"1": ["12"], "2": ["21"]},
    "RIGENS": {"-": ["25"]},
    "RIGEO": {"-": ["52"]},
    "RIGESCO": {"-": ["53"]},
    "RIGIDA": {"-": ["11"]},
    "RIGIDE": {"-": ["60"]},
    "RIGIDITAS": {"-": ["13"]},
    "RIGIDO": {"-": ["51"]},
    "RIGIDVS": {"-": ["21"]},
    "RIGO": {"-": ["53", "51"]},
    "RIGOR": {"-": ["13"]},
    "RIGVVS": {"-": ["21"]},
    "RIMA": {"-": ["11"]},
    "RIMOR": {"-": ["5J"]},
    "RIMOSVS": {"-": ["21"]},
    "RIMVLA": {"-": ["11"]},
    "RINGOR": {"-": ["5L"]},
    "RIPA": {"-": ["11"]},
    "RIPIO": {"-": ["55"]},
    "RISVS": {"-": ["14"]},
    "RITE": {"-": ["60"]},
    "RITVS": {"-": ["14"]},
    "RIVALIS": {"1": ["13"], "2": ["24"]},
    "RIVO": {"-": ["53"], "1": ["51"], "2": ["53"]},
    "RIVVS": {"-": ["12"]},
    "RIXA": {"-": ["11"]},
    "RIXOR": {"-": ["5J"]},
    "ROBIGALIA": {"-": ["13"]},
    "ROBIGO": {"-": ["13"], "N": ["13"]},
    "ROBORATVS": {"-": ["21"]},
    "ROBORO": {"-": ["51"]},
    "ROBVR": {"-": ["13"]},
    "ROBVS": {"-": ["21"]},
    "ROBVSTEVS": {"-": ["21"]},
    "ROBVSTVS": {"-": ["21"], "N": ["12"]},
    "RODEVS": {"-": ["21"]},
    "RODO": {"-": ["53"]},
    "ROGALIS": {"-": ["24"]},
    "ROGATIO": {"-": ["13"]},
    "ROGATVS": {"-": ["14"]},
    "ROGITO": {"-": ["51"]},
    "ROGO": {"-": ["51"]},
    "ROGVS": {"-": ["12"]},
    "ROMA": {"N": ["11"]},
    "ROMANI": {"N": ["12"]},
    "ROMANVS": {"N": ["21"], "A": ["21"]},
    "ROMVLARIS": {"N": ["24"]},
    "ROMVLEVS": {"N": ["21"], "A": ["21"]},
    "ROMVLVS": {"N": ["12"], "A": ["21"]},
    "RORIDVS": {"-": ["21"]},
    "RORIFER": {"-": ["21"]},
    "RORO": {"-": ["51"]},
    "RORVLENTVS": {"-": ["21"]},
    "ROS": {"-": ["13"], "N": ["13"]},
    "ROSA": {"-": ["11"]},
    "ROSARIVM": {"-": ["12"]},
    "ROSCIDVS": {"-": ["21"]},
    "ROSETVM": {"-": ["12"]},
    "ROSEVS": {"-": ["21"], "A": ["21"]},
    "ROSTRATVS": {"-": ["21"]},
    "ROSTRVM": {"-": ["12"]},
    "ROTA": {"-": ["11"]},
    "ROTATIO": {"-": ["13"]},
    "ROTO": {"-": ["51"]},
    "ROTVLA": {"-": ["11"]},
    "ROTVNDATIO": {"-": ["13"]},
    "ROTVNDITAS": {"-": ["13"]},
    "ROTVNDO": {"-": ["51"]},
    "ROTVNDVS": {"-": ["21"]},
    "RVBELLVS": {"-": ["21"]},
    "RVBENS": {"-": ["25"]},
    "RVBEO": {"-": ["52"]},
    "RVBER": {"-": ["21"], "N": ["21"], "A": ["23"]},
    "RVBESCO": {"-": ["53"]},
    "RVBETA": {"-": ["11"]},
    "RVBETVM": {"-": ["12"]},
    "RVBEVS": {"-": ["21"]},
    "RVBIA": {"-": ["11"]},
    "RVBICVNDVLVS": {"-": ["21"]},
    "RVBICVNDVS": {"-": ["21"]},
    "RVBOR": {"-": ["13"]},
    "RVBRICA": {"-": ["11"]},
    "RVBRICOSVS": {"-": ["21"]},
    "RVBRIVS": {"N": ["12"]},
    "RVBVS": {"-": ["12"]},
    "RVCTABVNDVS": {"-": ["21"]},
    "RVCTO": {"-": ["51"]},
    "RVCTVS": {"-": ["14"]},
    "RVDECTVS": {"-": ["21"]},
    "RVDENS": {"-": ["13"]},
    "RVDERATIO": {"-": ["13"]},
    "RVDICVLA": {"-": ["11"]},
    "RVDIMENTVM": {"-": ["12"]},
    "RVDIS": {"1": ["13"], "2": ["24"]},
    "RVDO": {"-": ["53"]},
    "RVDVS": {"-": ["13"]},
    "RVFVS": {"-": ["21"], "N": ["21"]},
    "RVGA": {"-": ["11"]},
    "RVGOSVS": {"-": ["21"]},
    "RVINA": {"-": ["11"]},
    "RVINOSVS": {"-": ["21"]},
    "RVMINALIS": {"N": ["24"], "A": ["24"]},
    "RVMINO": {"-": ["51"]},
    "RVMOR": {"-": ["13"], "N": ["13"]},
    "RVMPO": {"-": ["53"]},
    "RVNCO": {"-": ["51"]},
    "RVO": {"-": ["53"]},
    "RVPES": {"-": ["13"]},
    "RVPILIVS": {"A": ["21"], "N": ["12"]},
    "RVPTOR": {"-": ["13"]},
    "RVRICOLA": {"-": ["11"]},
    "RVRSVS": {"-": ["60"]},
    "RVS": {"-": ["13"]},
    "RVSCEVS": {"-": ["21"]},
    "RVSCVM": {"-": ["12"]},
    "RVSCVS": {"-": ["12"]},
    "RVSSATVS": {"-": ["21"]},
    "RVSSEVS": {"-": ["21"]},
    "RVSSVS": {"-": ["21"]},
    "RVSTARIVS": {"-": ["21"]},
    "RVSTICA": {"-": ["11"]},
    "RVSTICANVS": {"1": ["12"], "2": ["21"]},
    "RVSTICE": {"-": ["6&"]},
    "RVSTICITAS": {"-": ["13"]},
    "RVSTICVS": {"1": ["12"], "2": ["21"], "N": ["12"]},
    "RVTA": {"-": ["11"]},
    "RVTABVLVM": {"-": ["12"]},
    "RVTILANS": {"-": ["25"]},
    "RVTILIA": {"N": ["11"]},
    "RVTILIVS": {"N": ["12"]},
    "RVTILO": {"-": ["51"]},
    "RVTILVS": {"-": ["21"], "N": ["12"]},
    "RVTRVM": {"-": ["12"]},
    "RVTVLI": {"N": ["12"]},
    "SABBATVM": {"-": ["12"], "N": ["12"]},
    "SABINI": {"N": ["12"]},
    "SABINVS": {"N": ["21"], "A": ["21"]},
    "SABVLO": {"-": ["13"]},
    "SABVLOSVS": {"-": ["21"]},
    "SABVLVM": {"-": ["12"]},
    "SABVRRA": {"-": ["11"], "N": ["11"]},
    "SACCATVS": {"-": ["21"]},
    "SACCO": {"-": ["51"]},
    "SACCVLVS": {"-": ["12"]},
    "SACCVS": {"-": ["12"]},
    "SACELLVM": {"-": ["12"]},
    "SACER": {"-": ["26"], "N": ["21"], "A": ["23"]},
    "SACERDOS": {"-": ["13"], "N": ["13"]},
    "SACERDOTIVM": {"-": ["12"]},
    "SACRAMENTALIS": {"-": ["24"]},
    "SACRAMENTVM": {"-": ["12"]},
    "SACRARIVM": {"-": ["12"]},
    "SACRATIVIR": {"N": ["12"]},
    "SACRATVS": {"-": ["21"]},
    "SACRIFICALIS": {"-": ["24"]},
    "SACRIFICIVM": {"-": ["12"]},
    "SACRIFICO": {"-": ["51"]},
    "SACRIFICVS": {"-": ["21"]},
    "SACRILEGA": {"-": ["11"]},
    "SACRILEGIVM": {"-": ["12"]},
    "SACRILEGVS": {"1": ["12"], "2": ["21"]},
    "SACRO": {"-": ["51"]},
    "SACROSANCTVS": {"-": ["21"]},
    "SACRVM": {"-": ["12"]},
    "SADALA": {"N": ["11"]},
    "SAECVLARIS": {"-": ["24"]},
    "SAECVLVM": {"-": ["12"]},
    "SAEPE": {"-": ["60"]},
    "SAEPENVMERO": {"-": ["60"]},
    "SAEPES": {"-": ["13"]},
    "SAEPIA": {"-": ["11"]},
    "SAEPIO": {"-": ["54"]},
    "SAEPTA": {"-": ["12"], "L": ["12"]},
    "SAEPTIO": {"-": ["13"]},
    "SAEPTVM": {"-": ["12"]},
    "SAETA": {"-": ["11"]},
    "SAETIGER": {"1": ["12"], "2": ["21"]},
    "SAETOSVS": {"-": ["21"]},
    "SAEVE": {"-": ["6&"]},
    "SAEVIO": {"-": ["54"]},
    "SAEVITIA": {"-": ["11"]},
    "SAEVVS": {"-": ["21"]},
    "SAGA": {"1": ["11"], "2": ["11"]},
    "SAGACITAS": {"-": ["13"]},
    "SAGACITER": {"-": ["6&"]},
    "SAGAX": {"-": ["25"]},
    "SAGINA": {"-": ["11"]},
    "SAGINATVS": {"-": ["21"]},
    "SAGINO": {"-": ["51"]},
    "SAGIO": {"-": ["54"]},
    "SAGITTA": {"-": ["11"], "N": ["11"]},
    "SAGITTARIVS": {"1": ["12"], "2": ["21"]},
    "SAGITTIFER": {"1": ["12"], "2": ["21"]},
    "SAGVLVM": {"-": ["12"]},
    "SAGVM": {"-": ["12"]},
    "SAGVNTINI": {"-": ["12"], "N": ["12"]},
    "SAGVS": {"1": ["12"], "2": ["21"]},
    "SAL": {"-": ["13"]},
    "SALAPVTIVM": {"-": ["12"]},
    "SALARIVM": {"-": ["12"]},
    "SALAX": {"-": ["25"]},
    "SALEBRA": {"-": ["11"]},
    "SALEBROSVS": {"-": ["21"]},
    "SALGAMA": {"-": ["12"]},
    "SALIARIS": {"1": ["24"], "2": ["24"], "A": ["24"]},
    "SALICTARIVS": {"1": ["12"]},
    "SALICTVM": {"-": ["12"]},
    "SALIFODINA": {"-": ["11"]},
    "SALIGNEVS": {"-": ["21"]},
    "SALIGNVS": {"-": ["21"]},
    "SALII": {"-": ["12"], "N": ["12"]},
    "SALILLVM": {"-": ["12"]},
    "SALINAE": {"-": ["11"]},
    "SALINATOR": {"-": ["13"], "N": ["13"]},
    "SALINVM": {"-": ["12"]},
    "SALIO": {"1": ["54"], "2": ["54"]},
    "SALIVA": {"-": ["11"]},
    "SALIVNCA": {"-": ["11"]},
    "SALIX": {"-": ["13"]},
    "SALLVSTIVS": {"N": ["12"]},
    "SALSAMENTVM": {"-": ["12"]},
    "SALSITVDO": {"-": ["13"]},
    "SALSVGO": {"-": ["13"]},
    "SALSVM": {"-": ["12"], "N": ["12"]},
    "SALSVS": {"-": ["21"]},
    "SALTATIO": {"-": ["13"]},
    "SALTATOR": {"-": ["13"]},
    "SALTATVS": {"-": ["14"]},
    "SALTEM": {"-": ["60"]},
    "SALTO": {"-": ["51"]},
    "SALTVOSVS": {"-": ["21"]},
    "SALTVS": {"1": ["14"], "2": ["14"]},
    "SALVBER": {"-": ["23"]},
    "SALVBRITAS": {"-": ["13"]},
    "SALVBRITER": {"-": ["60"]},
    "SALVEO": {"-": ["52"]},
    "SALVIDIENVS": {"N": ["12"]},
    "SALVM": {"-": ["12"]},
    "SALVS": {"-": ["13"], "N": ["13"]},
    "SALVTARIS": {"-": ["24"]},
    "SALVTARITER": {"-": ["60"]},
    "SALVTATIO": {"-": ["13"]},
    "SALVTATOR": {"-": ["13"]},
    "SALVTATRIX": {"-": ["13"]},
    "SALVTIFER": {"-": ["21"]},
    "SALVTO": {"-": ["51"]},
    "SALVVS": {"-": ["21"], "N": ["12"]},
    "SAMBVCA": {"-": ["17"]},
    "SAMII": {"N": ["12"]},
    "SAMIVS": {"N": ["21"], "A": ["21"]},
    "SAMNIS": {"N": ["25"], "A": ["24"]},
    "SAMOS": {"N": ["17"]},
    "SANABILIS": {"-": ["24"]},
    "SANCIO": {"-": ["54"]},
    "SANCTE": {"-": ["60"]},
    "SANCTIMONIA": {"-": ["11"]},
    "SANCTIO": {"-": ["13"]},
    "SANCTITAS": {"-": ["13"]},
    "SANCTOR": {"-": ["13"]},
    "SANCTVM": {"-": ["12"]},
    "SANCTVS": {"-": ["21"], "A": ["21"], "N": ["12"]},
    "SANCVS": {"-": ["16"], "N": ["12"]},
    "SANDAPILA": {"-": ["11"]},
    "SANDARACA": {"-": ["11"]},
    "SANDYX": {"-": ["13"]},
    "SANE": {"-": ["60"]},
    "SANGVINARIVS": {"-": ["21"]},
    "SANGVINEVS": {"-": ["21"]},
    "SANGVINO": {"-": ["51"]},
    "SANGVINOLENTVS": {"-": ["21"]},
    "SANGVIS": {"-": ["13"]},
    "SANIES": {"-": ["15"]},
    "SANITAS": {"-": ["13"]},
    "SANNA": {"-": ["11"]},
    "SANO": {"-": ["51"]},
    "SANVS": {"-": ["21"]},
    "SAPA": {"-": ["11"]},
    "SAPERDA": {"-": ["11"]},
    "SAPIENS": {"1": ["13"], "2": ["25"], "C": ["13"], "N": ["13"]},
    "SAPIENTER": {"-": ["60"]},
    "SAPIENTIA": {"-": ["11"]},
    "SAPINEVS": {"-": ["21"]},
    "SAPINVS": {"-": ["12"]},
    "SAPIO": {"-": ["55"]},
    "SAPOR": {"-": ["13"]},
    "SAPPHO": {"N": ["17"]},
    "SARCINA": {"-": ["11"]},
    "SARCINARIVS": {"-": ["21"]},
    "SARCINVLA": {"-": ["11"]},
    "SARCIO": {"-": ["54"]},
    "SARCOPHAGVS": {"-": ["12"]},
    "SARCVLVM": {"-": ["12"]},
    "SARDA": {"-": ["11"]},
    "SARDINIA": {"N": ["11"]},
    "SARDONYX": {"-": ["17"]},
    "SARIO": {"-": ["54"]},
    "SARISA": {"-": ["11"]},
    "SARISOPHORVS": {"-": ["12"]},
    "SARMENTVM": {"-": ["12"]},
    "SARRACVM": {"-": ["12"]},
    "SARTAGO": {"-": ["13"]},
    "SARTVRA": {"-": ["11"]},
    "SARTVS": {"-": ["21"]},
    "SATA": {"-": ["12"]},
    "SATAGEVS": {"-": ["21"]},
    "SATAGO": {"-": ["53"]},
    "SATARIVS": {"-": ["21"]},
    "SATELLES": {"-": ["13"]},
    "SATIAS": {"-": ["13"]},
    "SATIETAS": {"-": ["13"]},
    "SATIO": {"1": ["13"], "2": ["51"]},
    "SATIS": {"1": ["26"], "2": ["60"]},
    "SATISACCIPIO": {"-": ["55"]},
    "SATISDO": {"-": ["51"]},
    "SATISFACIO": {"-": ["55"]},
    "SATISFACTIO": {"-": ["13"]},
    "SATISFIO": {"-": ["56"]},
    "SATOR": {"-": ["13"], "N": ["13"]},
    "SATRAPEA": {"-": ["11"]},
    "SATRAPES": {"-": ["16"]},
    "SATRICVM": {"N": ["12"]},
    "SATRIVS": {"N": ["12"]},
    "SATVR": {"-": ["21"]},
    "SATVRA": {"-": ["11"], "N": ["11"]},
    "SATVRATVS": {"-": ["21"]},
    "SATVREIVM": {"-": ["12"]},
    "SATVRITAS": {"-": ["13"], "N": ["13"]},
    "SATVRNALIA": {"N": ["13"]},
    "SATVRNINVS": {"N": ["12"]},
    "SATVRNVS": {"N": ["12"]},
    "SATVRO": {"-": ["51"]},
    "SATVS": {"1": ["14"], "2": ["21"]},
    "SATYRICVS": {"-": ["21"]},
    "SATYRVS": {"-": ["12"], "N": ["12"]},
    "SAVCIATIO": {"-": ["13"]},
    "SAVCIO": {"-": ["51"]},
    "SAVCIVS": {"-": ["21"]},
    "SAVILLVM": {"-": ["12"]},
    "SAVIOR": {"-": ["5J"]},
    "SAVIVM": {"-": ["12"]},
    "SAXEVS": {"-": ["21"]},
    "SAXIFICVS": {"-": ["21"]},
    "SAXOSA": {"-": ["12"]},
    "SAXOSVS": {"-": ["21"]},
    "SAXVM": {"-": ["12"], "N": ["12"]},
    "SCABELLVM": {"-": ["12"]},
    "SCABER": {"-": ["21"]},
    "SCABIES": {"-": ["15"]},
    "SCABIOSVS": {"-": ["21"]},
    "SCABO": {"-": ["53"]},
    "SCAEVOLA": {"N": ["11"]},
    "SCAEVVS": {"-": ["21"]},
    "SCALAE": {"-": ["11"]},
    "SCALMVS": {"-": ["12"]},
    "SCALPELLVM": {"-": ["12"]},
    "SCALPO": {"-": ["53"]},
    "SCALPRVM": {"-": ["12"]},
    "SCALPTVRA": {"-": ["11"]},
    "SCAMILLVS": {"-": ["12"]},
    "SCAMMONIVM": {"-": ["12"]},
    "SCAMNVM": {"-": ["12"]},
    "SCANDENS": {"-": ["25"]},
    "SCANDILIANVS": {"N": ["21"], "A": ["21"]},
    "SCANDILIVS": {"N": ["12"]},
    "SCANDO": {"-": ["53"]},
    "SCANDVLA": {"-": ["11"]},
    "SCANSIO": {"-": ["13"]},
    "SCANSOR": {"-": ["13"]},
    "SCANSORIVS": {"-": ["21"]},
    "SCAPHE": {"-": ["17"]},
    "SCAPHIVM": {"-": ["12"]},
    "SCAPVLA": {"-": ["11"], "N": ["11"]},
    "SCAPVS": {"-": ["12"]},
    "SCARABEVS": {"-": ["12"]},
    "SCARVS": {"-": ["12"]},
    "SCATEBRA": {"-": ["11"]},
    "SCATEO": {"-": ["52"]},
    "SCATO": {"-": ["53"], "N": ["13"]},
    "SCATVRIGO": {"-": ["13"]},
    "SCAVRVS": {"-": ["21"], "N": ["12"]},
    "SCELERATE": {"-": ["60"]},
    "SCELERATI": {"-": ["12"]},
    "SCELERATVS": {"-": ["21"], "N": ["21"], "A": ["21"]},
    "SCELERO": {"1": ["13"], "2": ["51"]},
    "SCELEROSVS": {"-": ["21"]},
    "SCELESTVS": {"-": ["21"]},
    "SCELVS": {"-": ["13"], "N": ["12"]},
    "SCENA": {"-": ["11"]},
    "SCENICA": {"-": ["11"]},
    "SCENICVS": {"1": ["12"], "2": ["21"]},
    "SCEPTRIFER": {"-": ["21"]},
    "SCEPTRVM": {"-": ["12"]},
    "SCEPTVCHVS": {"-": ["12"]},
    "SCHEDA": {"-": ["11"]},
    "SCHEMA": {"-": ["16"]},
    "SCHERINI": {"N": ["12"]},
    "SCHIDIA": {"-": ["11"]},
    "SCHOENOBATES": {"-": ["17"]},
    "SCHOENVS": {"-": ["12"]},
    "SCHOLA": {"-": ["11"], "N": ["11"]},
    "SCHOLASTICA": {"-": ["12"]},
    "SCHOLASTICVS": {"1": ["12"], "2": ["21"]},
    "SCIATHVS": {"N": ["12"]},
    "SCIENS": {"-": ["25"]},
    "SCIENTER": {"-": ["60"]},
    "SCIENTIA": {"-": ["11"]},
    "SCILICET": {"-": ["60"]},
    "SCILLA": {"-": ["11"]},
    "SCINDO": {"-": ["53"]},
    "SCINTILLA": {"-": ["11"], "N": ["11"]},
    "SCINTILLO": {"-": ["51"]},
    "SCIO": {"-": ["54"]},
    "SCIOTHERES": {"-": ["17"]},
    "SCIPIO": {"-": ["13"], "N": ["13"]},
    "SCIRPEA": {"-": ["11"]},
    "SCIRPICVLVS": {"1": ["12"], "2": ["21"]},
    "SCISCITOR": {"-": ["5J"]},
    "SCISCO": {"-": ["53"]},
    "SCISSVRA": {"-": ["11"]},
    "SCITE": {"-": ["60"]},
    "SCITOR": {"-": ["5J"]},
    "SCITVM": {"-": ["12"]},
    "SCITVS": {"1": ["14"], "2": ["21"]},
    "SCOMBER": {"-": ["12"]},
    "SCOPAE": {"-": ["11"]},
    "SCOPIO": {"-": ["13"]},
    "SCOPVLA": {"-": ["11"]},
    "SCOPVLVS": {"-": ["12"]},
    "SCORDALVS": {"-": ["12"]},
    "SCORPIO": {"-": ["13"]},
    "SCORPIVS": {"-": ["17"], "N": ["12"]},
    "SCORTATOR": {"-": ["13"]},
    "SCORTILLVM": {"-": ["12"]},
    "SCORTVM": {"-": ["12"], "N": ["12"]},
    "SCOTEINOS": {"-": ["27"]},
    "SCOTIA": {"-": ["11"]},
    "SCREO": {"-": ["51"]},
    "SCRIBA": {"-": ["11"]},
    "SCRIBLITA": {"-": ["11"]},
    "SCRIBO": {"-": ["53"]},
    "SCRINIVM": {"-": ["12"]},
    "SCRIPTA": {"-": ["12"]},
    "SCRIPTOR": {"-": ["13"]},
    "SCRIPTVLVM": {"-": ["12"]},
    "SCRIPTVM": {"-": ["12"]},
    "SCRIPTVRA": {"-": ["11"]},
    "SCRIPVLVM": {"-": ["12"]},
    "SCROBIS": {"-": ["13"]},
    "SCROFA": {"-": ["11"]},
    "SCRVPEVS": {"-": ["21"]},
    "SCRVPOSVS": {"-": ["21"]},
    "SCRVPVLOSVS": {"-": ["21"]},
    "SCRVPVLVS": {"-": ["12"]},
    "SCRVTA": {"-": ["12"]},
    "SCRVTANS": {"-": ["25"]},
    "SCRVTATIO": {"-": ["13"]},
    "SCRVTOR": {"-": ["5J"]},
    "SCVLPO": {"-": ["53"]},
    "SCVLPONEAE": {"-": ["11"]},
    "SCVLPTVRA": {"-": ["11"]},
    "SCVRRA": {"-": ["11"]},
    "SCVRRILIS": {"-": ["24"]},
    "SCVRRILITAS": {"-": ["13"]},
    "SCVTATVS": {"-": ["21"]},
    "SCVTICA": {"-": ["11"]},
    "SCVTRA": {"-": ["11"]},
    "SCVTRISCVM": {"-": ["12"]},
    "SCVTVLA": {"-": ["11"]},
    "SCVTVLATA": {"-": ["12"]},
    "SCVTVM": {"-": ["12"]},
    "SCYLLA": {"N": ["11"]},
    "SCYMNVS": {"-": ["12"]},
    "SCYPHVS": {"-": ["12"]},
    "SCYTHAE": {"N": ["11"]},
    "SCYTHIA": {"N": ["11"]},
    "SEBVM": {"-": ["12"]},
    "SECEDO": {"1": ["53"]},
    "SECENTI": {"-": ["31"]},
    "SECERNO": {"-": ["53"]},
    "SECESSIO": {"-": ["13"]},
    "SECESSVS": {"-": ["14"]},
    "SECLVDO": {"-": ["53"]},
    "SECLVSVS": {"-": ["21"]},
    "SECO": {"-": ["51"]},
    "SECRETE": {"-": ["60"]},
    "SECRETO": {"-": ["60"]},
    "SECRETVM": {"-": ["12"]},
    "SECRETVS": {"-": ["21"]},
    "SECRO": {"-": ["51"]},
    "SECROR": {"-": ["5J"]},
    "SECTA": {"-": ["11"]},
    "SECTATOR": {"-": ["13"]},
    "SECTILIS": {"-": ["24"]},
    "SECTIO": {"-": ["13"]},
    "SECTIVVS": {"-": ["21"]},
    "SECTOR": {"1": ["13"], "2": ["5J"]},
    "SECTVRA": {"-": ["11"]},
    "SECVBITVS": {"-": ["14"]},
    "SECVBO": {"-": ["51"]},
    "SECVNDAE": {"-": ["11"]},
    "SECVNDANI": {"-": ["12"]},
    "SECVNDANVS": {"-": ["21"]},
    "SECVNDARIVS": {"-": ["21"]},
    "SECVNDE": {"-": ["60"]},
    "SECVNDO": {"1": ["35"], "2": ["51"]},
    "SECVNDVM": {"1": ["60"], "2": ["70"]},
    "SECVNDVS": {"1": ["21"], "2": ["32"], "N": ["12"]},
    "SECVRE": {"-": ["60"]},
    "SECVRICLATVS": {"-": ["21"]},
    "SECVRICVLA": {"-": ["11"]},
    "SECVRIGER": {"-": ["21"]},
    "SECVRIS": {"-": ["13"]},
    "SECVRITAS": {"-": ["13"]},
    "SECVRVS": {"-": ["21"]},
    "SECVS": {"1": ["16"], "2": ["60"], "3": ["70"]},
    "SECVTOR": {"-": ["13"]},
    "SED": {"-": ["81"]},
    "SEDAMEN": {"-": ["13"]},
    "SEDATIO": {"-": ["13"]},
    "SEDATVS": {"-": ["21"]},
    "SEDECIM": {"-": ["31"]},
    "SEDENTARIVS": {"-": ["21"]},
    "SEDEO": {"-": ["52"]},
    "SEDES": {"-": ["13"]},
    "SEDILE": {"-": ["13"]},
    "SEDITIO": {"-": ["13"]},
    "SEDITIOSE": {"-": ["60"]},
    "SEDITIOSVS": {"-": ["21"]},
    "SEDO": {"-": ["51"]},
    "SEDVCO": {"-": ["53"]},
    "SEDVCTVS": {"-": ["21"]},
    "SEDVLITAS": {"-": ["13"]},
    "SEDVLO": {"-": ["60"]},
    "SEDVLVS": {"-": ["21"]},
    "SEGES": {"-": ["13"]},
    "SEGESTA": {"N": ["11"]},
    "SEGESTANI": {"N": ["12"]},
    "SEGESTANVS": {"N": ["21"], "A": ["21"]},
    "SEGMEN": {"-": ["13"]},
    "SEGMENTATVS": {"-": ["21"]},
    "SEGMENTVM": {"-": ["12"]},
    "SEGNE": {"-": ["60"]},
    "SEGNIPES": {"-": ["25"]},
    "SEGNIS": {"-": ["24"]},
    "SEGNITER": {"-": ["60"]},
    "SEGNITIA": {"-": ["11"]},
    "SEGNITIES": {"-": ["15"]},
    "SEGREGO": {"-": ["51"]},
    "SEGREX": {"-": ["25"]},
    "SEIANIANVS": {"N": ["21"], "A": ["21"]},
    "SEIANVS": {"N": ["12"]},
    "SEIVNCTIM": {"-": ["60"]},
    "SEIVNCTVS": {"-": ["21"]},
    "SEIVNGO": {"-": ["53"]},
    "SELECTVS": {"C": ["21"], "-": ["21"]},
    "SELENE": {"N": ["17"]},
    "SELIBRA": {"-": ["11"]},
    "SELIGO": {"3": ["53"]},
    "SELLA": {"-": ["11"]},
    "SELLARIVS": {"-": ["12"]},
    "SELLISTERNIVM": {"-": ["12"]},
    "SEMEL": {"-": ["36"]},
    "SEMEN": {"1": ["13"], "2": ["16"]},
    "SEMENTIS": {"-": ["13"]},
    "SEMENTIVVS": {"-": ["21"], "A": ["21"]},
    "SEMESTRIS": {"-": ["24"]},
    "SEMESVS": {"-": ["21"]},
    "SEMIADAPERTVS": {"-": ["21"]},
    "SEMIANIMIS": {"-": ["24"]},
    "SEMIANIMVS": {"-": ["21"]},
    "SEMIBOS": {"-": ["25"]},
    "SEMICANALICVLVS": {"-": ["12"]},
    "SEMIDEA": {"-": ["11"]},
    "SEMIDEVS": {"1": ["12"], "2": ["21"]},
    "SEMIERMIS": {"-": ["24"]},
    "SEMIERMVS": {"-": ["21"]},
    "SEMIFACTVS": {"-": ["21"]},
    "SEMIFASTIGIVM": {"-": ["12"]},
    "SEMIFER": {"-": ["21"]},
    "SEMIFVNIVM": {"-": ["12"]},
    "SEMIGERMANVS": {"-": ["21"]},
    "SEMIGRAVIS": {"-": ["24"]},
    "SEMIHIANS": {"-": ["25"]},
    "SEMIHOMO": {"-": ["13"]},
    "SEMIHORA": {"-": ["11"]},
    "SEMILATER": {"-": ["16"]},
    "SEMILAVTVS": {"-": ["21"]},
    "SEMILIBER": {"-": ["21"]},
    "SEMIMARINVS": {"-": ["21"]},
    "SEMIMAS": {"-": ["25"]},
    "SEMIMETOPIA": {"-": ["12"]},
    "SEMIMORTVVS": {"-": ["21"]},
    "SEMINARIVM": {"-": ["12"]},
    "SEMINARIVS": {"-": ["21"]},
    "SEMINATOR": {"-": ["13"]},
    "SEMINEX": {"-": ["25"]},
    "SEMINIVM": {"-": ["12"]},
    "SEMINO": {"-": ["51"]},
    "SEMINVDVS": {"-": ["21"]},
    "SEMIPAGANVS": {"-": ["21"]},
    "SEMIPES": {"-": ["13"]},
    "SEMIPLENVS": {"-": ["21"]},
    "SEMIPVTATVS": {"-": ["21"]},
    "SEMIRASVS": {"-": ["21"]},
    "SEMIREDVCTVS": {"-": ["21"]},
    "SEMIREFECTVS": {"-": ["21"]},
    "SEMIRVTVS": {"-": ["21"]},
    "SEMIS": {"-": ["16"], "2": ["13"], "1": ["21"]},
    "SEMISEPVLTVS": {"-": ["21"]},
    "SEMISOMNIS": {"-": ["24"]},
    "SEMISOMNVS": {"-": ["21"]},
    "SEMISVPINVS": {"-": ["21"]},
    "SEMITA": {"-": ["11"]},
    "SEMITARIVS": {"-": ["21"]},
    "SEMIVIR": {"-": ["12"]},
    "SEMIVIVVS": {"-": ["21"]},
    "SEMIVSTVS": {"-": ["21"]},
    "SEMODIALIS": {"-": ["24"]},
    "SEMODIVS": {"-": ["12"]},
    "SEMOTVS": {"-": ["21"]},
    "SEMOVEO": {"-": ["52"]},
    "SEMPER": {"-": ["60"]},
    "SEMPITERNO": {"-": ["60"]},
    "SEMPITERNVM": {"-": ["60"]},
    "SEMPITERNVS": {"-": ["21"]},
    "SEMPRONIVS": {"A": ["21"], "N": ["12"]},
    "SEMVNCIA": {"-": ["11"]},
    "SEMVSTVLATVS": {"-": ["21"]},
    "SENARIOLVS": {"-": ["12"]},
    "SENATOR": {"-": ["13"]},
    "SENATORIVS": {"-": ["21"]},
    "SENATVS": {"-": ["14"]},
    "SENECA": {"N": ["11"]},
    "SENECTA": {"-": ["11"]},
    "SENECTVS": {"1": ["13"], "2": ["21"], "N": ["12"]},
    "SENEO": {"-": ["52"]},
    "SENESCO": {"-": ["53"]},
    "SENEX": {"1": ["13"], "2": ["26"]},
    "SENI": {"-": ["33"]},
    "SENILIS": {"-": ["24"]},
    "SENIO": {"-": ["13"]},
    "SENIOR": {"-": ["13"]},
    "SENIOSVS": {"-": ["21"]},
    "SENIVM": {"-": ["12"]},
    "SENSA": {"-": ["12"]},
    "SENSATIO": {"-": ["13"]},
    "SENSIBILIS": {"-": ["24"]},
    "SENSIFER": {"-": ["21"]},
    "SENSILIS": {"-": ["24"]},
    "SENSIM": {"-": ["60"]},
    "SENSVS": {"-": ["14"]},
    "SENTENTIA": {"-": ["11"]},
    "SENTINA": {"-": ["11"]},
    "SENTINOSVS": {"-": ["21"]},
    "SENTIO": {"-": ["54"]},
    "SENTIOR": {"-": ["5M"]},
    "SENTIS": {"-": ["13"]},
    "SENTISCO": {"-": ["53"]},
    "SENTIVS": {"N": ["12"]},
    "SENTVS": {"-": ["21"]},
    "SEORSVM": {"1": ["60"], "2": ["70"]},
    "SEPARATE": {"-": ["60"]},
    "SEPARATIM": {"-": ["60"]},
    "SEPARATVM": {"-": ["12"]},
    "SEPARATVS": {"1": ["14"], "2": ["21"]},
    "SEPARO": {"2": ["51"]},
    "SEPELIO": {"-": ["54"]},
    "SEPES": {"-": ["13"]},
    "SEPONO": {"-": ["53"]},
    "SEPOSITVS": {"-": ["21"]},
    "SEPSE": {"-": ["43"]},
    "SEPTEM": {"-": ["31"], "N": ["16"]},
    "SEPTEMBER": {"A": ["23"], "N": ["13"]},
    "SEPTEMGEMINVS": {"-": ["21"]},
    "SEPTEMPLEX": {"-": ["34", "22"]},
    "SEPTEMREMIS": {"-": ["25"]},
    "SEPTEMVIR": {"-": ["12"]},
    "SEPTENDECIM": {"-": ["31"]},
    "SEPTENDECIMVS": {"-": ["32"]},
    "SEPTENTRIO": {"-": ["13"], "N": ["13"]},
    "SEPTENTRIONALIA": {"-": ["13"]},
    "SEPTENTRIONALIS": {"-": ["24"]},
    "SEPTENVS": {"-": ["33"]},
    "SEPTICIVS": {"N": ["12"]},
    "SEPTIMIVS": {"N": ["12"]},
    "SEPTIMVM": {"-": ["36"]},
    "SEPTIMVS": {"-": ["32"]},
    "SEPTINGENTESIMVS": {"-": ["32"]},
    "SEPTINGENTI": {"-": ["31"]},
    "SEPTVAGESIMVS": {"-": ["32"]},
    "SEPTVAGINTA": {"-": ["31"]},
    "SEPVLCRALIS": {"-": ["24"]},
    "SEPVLCRETVM": {"-": ["12"]},
    "SEPVLCRVM": {"-": ["12"]},
    "SEPVLTVRA": {"-": ["11"]},
    "SEQVAX": {"-": ["25"]},
    "SEQVENS": {"-": ["25"]},
    "SEQVESTER": {"1": ["16"], "2": ["26"]},
    "SEQVESTRO": {"2": ["60"]},
    "SEQVOR": {"-": ["5L"]},
    "SERA": {"1": ["11"], "2": ["60"]},
    "SERAPIS": {"N": ["17"]},
    "SERARIVS": {"-": ["21"]},
    "SERENITAS": {"-": ["13"]},
    "SERENO": {"-": ["51"]},
    "SERENVM": {"-": ["12"]},
    "SERENVS": {"-": ["21"], "N": ["12"]},
    "SERESCO": {"1": ["53"]},
    "SERGIVS": {"N": ["12"], "A": ["21"]},
    "SERIA": {"-": ["11"]},
    "SERICA": {"-": ["12"]},
    "SERICVS": {"-": ["21"], "A": ["21"]},
    "SERIES": {"-": ["15"]},
    "SERIO": {"-": ["60"]},
    "SERIOLA": {"-": ["11"]},
    "SERIPHVS": {"N": ["12"]},
    "SERIVM": {"-": ["12"]},
    "SERIVS": {"-": ["21"]},
    "SERMO": {"-": ["13"], "N": ["13"]},
    "SERMOCINOR": {"-": ["5J"]},
    "SERO": {"1": ["51"], "2": ["53"], "3": ["53"], "4": ["60"]},
    "SERPENS": {"-": ["13"]},
    "SERPO": {"-": ["53"]},
    "SERPYLLVM": {"-": ["12"]},
    "SERRA": {"-": ["11"]},
    "SERRARIVS": {"-": ["12"]},
    "SERRATIM": {"-": ["60"]},
    "SERRATVS": {"-": ["21"]},
    "SERTA": {"-": ["11"]},
    "SERTIVS": {"N": ["12"]},
    "SERTORIANVS": {"N": ["21"], "A": ["21"]},
    "SERTORIVS": {"N": ["12"]},
    "SERTVM": {"-": ["12"]},
    "SERVA": {"-": ["11"]},
    "SERVANS": {"-": ["25"]},
    "SERVATOR": {"-": ["13"]},
    "SERVIENS": {"-": ["13"]},
    "SERVILIA": {"N": ["11"]},
    "SERVILIS": {"-": ["24"]},
    "SERVILIVS": {"N": ["12"], "A": ["21"]},
    "SERVIO": {"-": ["54"]},
    "SERVITIVM": {"-": ["12"]},
    "SERVITVS": {"-": ["13"], "N": ["12"]},
    "SERVIVS": {"N": ["12"]},
    "SERVM": {"1": ["12"], "2": ["12"]},
    "SERVO": {"-": ["51"]},
    "SERVS": {"-": ["21"]},
    "SERVVLVS": {"-": ["12"]},
    "SERVVS": {"1": ["12"], "2": ["21"]},
    "SESAMA": {"-": ["11"]},
    "SESCENI": {"-": ["33"]},
    "SESCENTESIMVS": {"-": ["32"]},
    "SESCENTI": {"-": ["31"]},
    "SESCENTIES": {"-": ["36"]},
    "SESQVIALTER": {"-": ["48"]},
    "SESQVIDIGITVS": {"-": ["12"]},
    "SESQVILIBRA": {"-": ["11"]},
    "SESQVIMODIVS": {"-": ["12"]},
    "SESQVIPEDALIS": {"-": ["24"]},
    "SESQVIPES": {"-": ["13"]},
    "SESQVIPLAGA": {"-": ["11"]},
    "SESSILIS": {"-": ["24"]},
    "SESSIO": {"-": ["13"]},
    "SESSOR": {"-": ["13"]},
    "SESTERTIVM": {"-": ["12"]},
    "SESTERTIVS": {"1": ["12"], "2": ["21"]},
    "SEV": {"1": ["81"], "2": ["82"]},
    "SEVEHO": {"-": ["53"]},
    "SEVERA": {"-": ["12"]},
    "SEVERE": {"-": ["60"]},
    "SEVERI": {"-": ["12"]},
    "SEVERITAS": {"-": ["13"]},
    "SEVERO": {"-": ["51"]},
    "SEVERVS": {"-": ["21"], "N": ["12"]},
    "SEVOCO": {"-": ["51"]},
    "SEX": {"-": ["31"]},
    "SEXAGENI": {"-": ["33"]},
    "SEXAGESIMVS": {"-": ["32"]},
    "SEXAGINTA": {"-": ["31"]},
    "SEXCENARIVS": {"-": ["21"]},
    "SEXENNIS": {"-": ["24"]},
    "SEXIES": {"-": ["36"]},
    "SEXIS": {"-": ["13"]},
    "SEXTANS": {"-": ["13"]},
    "SEXTARIVS": {"-": ["12"]},
    "SEXTILIS": {"-": ["13"], "A": ["24"], "N": ["13"]},
    "SEXTIVS": {"N": ["12"], "A": ["21"]},
    "SEXTVLA": {"-": ["11"]},
    "SEXTVM": {"-": ["35"]},
    "SEXTVS": {"-": ["32"], "N": ["12"]},
    "SEXVS": {"-": ["14"]},
    "SI": {"1": ["67"], "2": ["82"]},
    "SIBILO": {"-": ["51"]},
    "SIBILVM": {"-": ["12"]},
    "SIBILVS": {"1": ["12"], "2": ["21"]},
    "SIBYLLA": {"N": ["11"]},
    "SIBYLLINVS": {"-": ["21"], "A": ["21"]},
    "SIC": {"-": ["60"]},
    "SICA": {"-": ["11"]},
    "SICARIVS": {"-": ["12"]},
    "SICCESCO": {"-": ["51"]},
    "SICCITAS": {"-": ["13"]},
    "SICCIVS": {"N": ["12"]},
    "SICCO": {"-": ["51"]},
    "SICCVM": {"-": ["12"]},
    "SICCVS": {"-": ["21"]},
    "SICILIA": {"N": ["11"]},
    "SICILICVS": {"-": ["12"]},
    "SICILIENSIS": {"N": ["24"], "A": ["24"]},
    "SICILIMENTVM": {"-": ["12"]},
    "SICINE": {"-": ["67"]},
    "SICINIVS": {"N": ["12"]},
    "SICVBI": {"-": ["82"]},
    "SICVLA": {"-": ["11"]},
    "SICVLI": {"N": ["12"]},
    "SICVLVS": {"N": ["21"], "A": ["21"]},
    "SICVT": {"1": ["66"], "2": ["82"]},
    "SICYONIVS": {"N": ["21"], "A": ["21"]},
    "SIDEO": {"-": ["52"]},
    "SIDEREVS": {"-": ["21"]},
    "SIDERO": {"-": ["51"]},
    "SIDO": {"-": ["53"], "N": ["13"]},
    "SIDVS": {"-": ["13"]},
    "SIGILLARIA": {"-": ["16"]},
    "SIGILLATVS": {"-": ["21"]},
    "SIGILLVM": {"-": ["12"]},
    "SIGNATOR": {"-": ["13"]},
    "SIGNATVS": {"-": ["21"]},
    "SIGNIA": {"N": ["11"]},
    "SIGNIFER": {"1": ["12"], "2": ["21"]},
    "SIGNIFICANS": {"-": ["25"]},
    "SIGNIFICATIO": {"-": ["13"]},
    "SIGNIFICATVM": {"-": ["12"]},
    "SIGNIFICATVS": {"-": ["14"]},
    "SIGNIFICO": {"-": ["51"]},
    "SIGNO": {"-": ["51"]},
    "SIGNVM": {"-": ["12"]},
    "SIL": {"-": ["13"]},
    "SILACEVS": {"-": ["21"]},
    "SILANIO": {"N": ["13"]},
    "SILANVS": {"N": ["12"], "-": ["12"]},
    "SILENS": {"-": ["25"]},
    "SILENTIVM": {"-": ["12"]},
    "SILEO": {"-": ["52"]},
    "SILER": {"-": ["13"]},
    "SILESCO": {"-": ["53"]},
    "SILEX": {"-": ["13"]},
    "SILICEVS": {"-": ["21"]},
    "SILICVLVS": {"-": ["12"]},
    "SILIGINEVS": {"1": ["12"], "2": ["21"]},
    "SILIGO": {"-": ["13"]},
    "SILIO": {"-": ["54"]},
    "SILIQVA": {"-": ["11"]},
    "SILPHIVM": {"-": ["12"]},
    "SILVA": {"-": ["11"]},
    "SILVATICVS": {"-": ["21"]},
    "SILVESTER": {"-": ["23"]},
    "SILVESTRIA": {"-": ["13"]},
    "SILVESTRIS": {"-": ["24"]},
    "SILVIA": {"N": ["11"]},
    "SILVICOLA": {"-": ["11"]},
    "SILVICVLTRIX": {"-": ["13"]},
    "SILVIFRAGVS": {"-": ["21"]},
    "SILVIVS": {"A": ["21"], "N": ["12"]},
    "SILVOSVS": {"-": ["21"]},
    "SILVRVS": {"-": ["12"]},
    "SIMA": {"-": ["11"]},
    "SIMIA": {"-": ["11"]},
    "SIMILAGO": {"-": ["13"]},
    "SIMILE": {"1": ["13"], "2": ["60"]},
    "SIMILIA": {"-": ["13"]},
    "SIMILIS": {"-": ["24"]},
    "SIMILITAS": {"-": ["13"]},
    "SIMILITER": {"-": ["60"]},
    "SIMILITVDO": {"-": ["13"]},
    "SIMIVS": {"-": ["12"]},
    "SIMO": {"-": ["51"], "N": ["11"]},
    "SIMPLEX": {"-": ["34", "22"], "N": ["13"]},
    "SIMPLICITAS": {"-": ["13"]},
    "SIMPLICITER": {"-": ["60"]},
    "SIMPLVM": {"-": ["12"]},
    "SIMPVLVM": {"-": ["12"]},
    "SIMPVVIVM": {"-": ["12"]},
    "SIMVL": {"1": ["60"], "2": ["70"], "3": ["82"]},
    "SIMVLAC": {"1": ["60"], "2": ["82"], "-": ["6-"]},
    "SIMVLACRVM": {"-": ["12"]},
    "SIMVLAMEN": {"-": ["13"]},
    "SIMVLANS": {"-": ["25"], "N": ["13"]},
    "SIMVLATIO": {"-": ["13"]},
    "SIMVLATOR": {"-": ["13"]},
    "SIMVLATQVE": {"1": ["60"], "2": ["82"], "-": ["6-"]},
    "SIMVLO": {"-": ["51"]},
    "SIMVLTAS": {"-": ["13"]},
    "SIMVLVS": {"-": ["21"]},
    "SIMVS": {"-": ["21"]},
    "SIN": {"-": ["82"]},
    "SINCERE": {"-": ["60"]},
    "SINCERITAS": {"-": ["13"]},
    "SINCERVS": {"-": ["21"]},
    "SINCIPVT": {"-": ["13"]},
    "SINE": {"-": ["70"]},
    "SINGILLARITER": {"-": ["60"]},
    "SINGILLATIM": {"-": ["60"]},
    "SINGVLA": {"1": ["11"], "2": ["12"]},
    "SINGVLARES": {"-": ["13"], "N": ["13"]},
    "SINGVLARIS": {"-": ["24"]},
    "SINGVLARITER": {"-": ["60"]},
    "SINGVLATIM": {"-": ["60"]},
    "SINGVLTIM": {"-": ["60"]},
    "SINGVLTIO": {"-": ["54"]},
    "SINGVLTVS": {"-": ["14"]},
    "SINGVLVM": {"-": ["12"]},
    "SINGVLVS": {"-": ["33"]},
    "SINISTER": {"-": ["21"]},
    "SINISTRA": {"1": ["11"], "2": ["12"]},
    "SINISTRE": {"-": ["60"]},
    "SINISTRORSVM": {"-": ["60"]},
    "SINISTRORSVS": {"-": ["60"]},
    "SINISTRVM": {"-": ["12"]},
    "SINO": {"-": ["53"]},
    "SINOPE": {"N": ["17"]},
    "SINVO": {"-": ["51"]},
    "SINVOSVS": {"-": ["21"]},
    "SINVS": {"-": ["14"]},
    "SIPARIVM": {"-": ["12"]},
    "SIPHO": {"-": ["13"]},
    "SIPIO": {"-": ["55"]},
    "SIQVIDEM": {"-": ["82"]},
    "SIREMPS": {"-": ["48"]},
    "SIRPICIVM": {"-": ["12"]},
    "SIRVS": {"-": ["12"]},
    "SISENNA": {"N": ["11"]},
    "SISER": {"-": ["13"]},
    "SISTO": {"-": ["53"]},
    "SISTRVM": {"-": ["12"]},
    "SISYPHVS": {"N": ["17"]},
    "SITICVLOSVS": {"-": ["21"]},
    "SITIO": {"-": ["54"]},
    "SITIS": {"-": ["13"]},
    "SITVLVS": {"-": ["12"]},
    "SITVS": {"1": ["14"], "2": ["21"]},
    "SIVE": {"1": ["81", "82"], "2": ["82"]},
    "SMARAGDVS": {"-": ["12"]},
    "SMYRNA": {"-": ["11"], "N": ["11"]},
    "SMYRNION": {"-": ["12"]},
    "SOBRIETAS": {"-": ["13"]},
    "SOBRINA": {"-": ["11"]},
    "SOBRIVS": {"-": ["21"]},
    "SOCCATVS": {"-": ["21"]},
    "SOCCVLVS": {"-": ["12"]},
    "SOCCVS": {"-": ["12"]},
    "SOCER": {"-": ["12"]},
    "SOCIA": {"-": ["11"]},
    "SOCIABILIS": {"-": ["24"]},
    "SOCIALIS": {"-": ["24"]},
    "SOCIALITAS": {"-": ["13"]},
    "SOCIETAS": {"-": ["13"]},
    "SOCIO": {"-": ["51"]},
    "SOCIVS": {"1": ["12"], "2": ["21"]},
    "SOCORDIA": {"-": ["11"]},
    "SOCORDITER": {"-": ["6&"]},
    "SOCORS": {"-": ["25"]},
    "SOCRATES": {"N": ["17"]},
    "SOCRVS": {"-": ["14"]},
    "SODALICIVM": {"-": ["12"]},
    "SODALIS": {"1": ["13"], "2": ["24"]},
    "SODALITAS": {"-": ["13"]},
    "SODALITIVM": {"-": ["12"]},
    "SODES": {"-": ["60"]},
    "SOL": {"-": ["13"], "N": ["13"]},
    "SOLACIOLVM": {"-": ["12"]},
    "SOLACIVM": {"-": ["12"]},
    "SOLAMEN": {"-": ["13"]},
    "SOLATOR": {"-": ["13"]},
    "SOLDVRII": {"-": ["12"]},
    "SOLEA": {"-": ["11"]},
    "SOLEATVS": {"-": ["21"]},
    "SOLENNE": {"-": ["13"]},
    "SOLENS": {"-": ["25"]},
    "SOLEO": {"-": ["5S"]},
    "SOLICISMVS": {"-": ["12"]},
    "SOLIDATIO": {"-": ["13"]},
    "SOLIDESCO": {"-": ["53"]},
    "SOLIDITAS": {"-": ["13"]},
    "SOLIDO": {"-": ["51"]},
    "SOLIDVM": {"1": ["12"], "2": ["60"]},
    "SOLIDVS": {"1": ["12"], "2": ["21"]},
    "SOLIFER": {"-": ["21"]},
    "SOLITARIVS": {"-": ["21"]},
    "SOLITVDO": {"-": ["13"]},
    "SOLITVM": {"-": ["12"]},
    "SOLITVS": {"-": ["21"]},
    "SOLIVAGVS": {"-": ["21"]},
    "SOLIVM": {"-": ["12"]},
    "SOLLEMNE": {"-": ["13"]},
    "SOLLEMNIS": {"-": ["24"]},
    "SOLLERS": {"-": ["25"], "N": ["13"]},
    "SOLLERTER": {"-": ["60"]},
    "SOLLERTIA": {"-": ["11"]},
    "SOLLICITATIO": {"-": ["13"]},
    "SOLLICITE": {"-": ["60"]},
    "SOLLICITO": {"-": ["51"]},
    "SOLLICITVDO": {"-": ["13"]},
    "SOLLICITVS": {"-": ["21"]},
    "SOLO": {"1": ["51"], "2": ["51"]},
    "SOLOECISMVS": {"-": ["12"]},
    "SOLOR": {"-": ["5J"]},
    "SOLSTITIALIS": {"-": ["24"]},
    "SOLSTITIVM": {"-": ["12"]},
    "SOLVM": {"1": ["12"], "2": ["60"]},
    "SOLVNTINI": {"N": ["12"]},
    "SOLVNTINVS": {"N": ["21"], "A": ["21"]},
    "SOLVO": {"-": ["53"]},
    "SOLVS": {"-": ["48"]},
    "SOLVTE": {"-": ["60"]},
    "SOLVTIO": {"-": ["13"]},
    "SOLVTVM": {"-": ["12"]},
    "SOLVTVS": {"-": ["21"]},
    "SOMNIO": {"-": ["51"]},
    "SOMNIVM": {"-": ["12"], "N": ["12"]},
    "SOMNVS": {"-": ["12"], "N": ["12"]},
    "SONANS": {"-": ["25"]},
    "SONAX": {"-": ["25"]},
    "SONIPES": {"1": ["13"], "2": ["25"]},
    "SONITVS": {"-": ["14"]},
    "SONIVIVS": {"-": ["21"]},
    "SONO": {"1": ["51"], "2": ["53"]},
    "SONOR": {"-": ["13"]},
    "SONORVS": {"-": ["21"]},
    "SONS": {"1": ["13"], "2": ["25"]},
    "SONTICVS": {"-": ["21"]},
    "SONVS": {"-": ["16"]},
    "SOPATER": {"N": ["12"]},
    "SOPHIA": {"-": ["11"]},
    "SOPHISMA": {"-": ["17"]},
    "SOPHISTES": {"-": ["17"]},
    "SOPIO": {"-": ["54"]},
    "SOPITIO": {"-": ["13"]},
    "SOPOR": {"-": ["13"], "N": ["13"]},
    "SOPORATVS": {"-": ["21"]},
    "SOPORIFER": {"-": ["21"]},
    "SOPORO": {"-": ["51"]},
    "SOPORVS": {"-": ["21"]},
    "SORBEO": {"-": ["52"]},
    "SORBITIO": {"-": ["13"]},
    "SORBVM": {"-": ["12"]},
    "SORBVS": {"-": ["12"]},
    "SORDEO": {"-": ["52"]},
    "SORDES": {"-": ["13"]},
    "SORDESCO": {"-": ["53"]},
    "SORDIDATVS": {"-": ["21"]},
    "SORDIDE": {"-": ["60"]},
    "SORDIDVLVS": {"-": ["21"]},
    "SORDIDVS": {"-": ["21"]},
    "SORITES": {"-": ["17"]},
    "SOROR": {"-": ["13"], "N": ["13"]},
    "SORS": {"-": ["13"]},
    "SORTIOR": {"-": ["5M"]},
    "SORTITIO": {"-": ["13"]},
    "SORTITO": {"-": ["60"]},
    "SORTITOR": {"-": ["13"]},
    "SORTITVS": {"-": ["14"]},
    "SOSIPPVS": {"N": ["12"]},
    "SOSITHEVS": {"N": ["12"]},
    "SOSPES": {"-": ["22"]},
    "SOSPITO": {"-": ["51"]},
    "SOSTRATVS": {"N": ["12"]},
    "SPADIX": {"-": ["13"]},
    "SPADO": {"-": ["13"]},
    "SPARGO": {"1": ["13"], "2": ["53"]},
    "SPARSVS": {"-": ["21"]},
    "SPARTA": {"N": ["17"]},
    "SPARTEVS": {"-": ["21"]},
    "SPARTVM": {"-": ["12"]},
    "SPARVS": {"-": ["12"]},
    "SPATHA": {"-": ["11"]},
    "SPATIATOR": {"-": ["13"]},
    "SPATIOR": {"-": ["5J"]},
    "SPATIOSE": {"-": ["60"]},
    "SPATIOSVM": {"-": ["12"]},
    "SPATIOSVS": {"-": ["21"]},
    "SPATIVM": {"-": ["12"]},
    "SPECIALIS": {"-": ["24"]},
    "SPECIATIM": {"-": ["60"]},
    "SPECIES": {"-": ["15"]},
    "SPECILLVM": {"-": ["12"]},
    "SPECIMEN": {"-": ["13"]},
    "SPECIO": {"-": ["55"]},
    "SPECIOSE": {"-": ["60"]},
    "SPECIOSVS": {"-": ["21"]},
    "SPECTABILIS": {"-": ["24"]},
    "SPECTACVLVM": {"-": ["12"]},
    "SPECTATIO": {"-": ["13"]},
    "SPECTATOR": {"-": ["13"]},
    "SPECTATRIX": {"-": ["13"]},
    "SPECTATVS": {"-": ["21"]},
    "SPECTO": {"-": ["51"]},
    "SPECVLA": {"1": ["11"], "2": ["11"]},
    "SPECVLARE": {"-": ["13"]},
    "SPECVLARIA": {"-": ["16"]},
    "SPECVLARIS": {"1": ["24"], "2": ["24"], "-": ["24"]},
    "SPECVLATOR": {"-": ["13"]},
    "SPECVLATORIVS": {"-": ["21"]},
    "SPECVLATRIX": {"-": ["13"]},
    "SPECVLOR": {"-": ["5J"]},
    "SPECVLVM": {"-": ["12"]},
    "SPECVS": {"-": ["14"]},
    "SPELAEVM": {"-": ["12"]},
    "SPELVNCA": {"-": ["11"], "N": ["11"]},
    "SPERATA": {"-": ["11"]},
    "SPERATVM": {"-": ["12"]},
    "SPERATVS": {"1": ["12"], "2": ["21"]},
    "SPERGO": {"-": ["53"]},
    "SPERNENDVS": {"-": ["21"]},
    "SPERNENS": {"-": ["25"]},
    "SPERNO": {"-": ["53"]},
    "SPERNOR": {"-": ["5J"]},
    "SPERO": {"-": ["51"]},
    "SPES": {"-": ["15"], "N": ["15"]},
    "SPHAERA": {"-": ["11"]},
    "SPHAERITA": {"-": ["11"]},
    "SPHAEROIDES": {"-": ["27"]},
    "SPHAEROMACHIA": {"-": ["17"]},
    "SPICA": {"-": ["11"]},
    "SPICESCO": {"-": ["53"]},
    "SPICEVS": {"-": ["21"]},
    "SPICIFER": {"-": ["21"]},
    "SPICIO": {"-": ["55"]},
    "SPICO": {"-": ["51"]},
    "SPICVLVM": {"-": ["12"]},
    "SPINA": {"-": ["11"]},
    "SPINETVM": {"-": ["12"]},
    "SPINOSVS": {"-": ["21"]},
    "SPINTRIA": {"-": ["11"]},
    "SPINVS": {"-": ["12"]},
    "SPIRA": {"-": ["11"]},
    "SPIRABILIS": {"-": ["24"]},
    "SPIRACVLVM": {"-": ["12"]},
    "SPIRAMEN": {"-": ["13"]},
    "SPIRAMENTVM": {"-": ["12"]},
    "SPIRITVS": {"-": ["14"]},
    "SPIRO": {"-": ["51"]},
    "SPISSAMENTVM": {"-": ["12"]},
    "SPISSE": {"-": ["60"]},
    "SPISSITAS": {"-": ["13"]},
    "SPISSO": {"-": ["51"]},
    "SPISSVS": {"-": ["21"]},
    "SPLEN": {"-": ["17"]},
    "SPLENDENS": {"-": ["25"]},
    "SPLENDEO": {"-": ["52"]},
    "SPLENDESCO": {"-": ["53"]},
    "SPLENDIDE": {"-": ["60"]},
    "SPLENDIDVS": {"-": ["21"]},
    "SPLENDOR": {"-": ["13"]},
    "SPOLIARIVM": {"-": ["12"]},
    "SPOLIATIO": {"-": ["13"]},
    "SPOLIATOR": {"-": ["13"]},
    "SPOLIATVS": {"1": ["21"], "2": ["21"]},
    "SPOLIO": {"-": ["51"]},
    "SPOLIVM": {"-": ["12"]},
    "SPONDA": {"-": ["11"]},
    "SPONDEO": {"-": ["52"]},
    "SPONDYLVS": {"-": ["12"]},
    "SPONGIA": {"-": ["11"]},
    "SPONS": {"-": ["13"]},
    "SPONSA": {"-": ["11"]},
    "SPONSALIA": {"-": ["16"]},
    "SPONSIO": {"-": ["13"]},
    "SPONSOR": {"-": ["13"]},
    "SPONSVM": {"-": ["12"]},
    "SPONSVS": {"1": ["12"], "2": ["14"], "N": ["12"]},
    "SPONTE": {"-": ["60"]},
    "SPORTA": {"-": ["11"]},
    "SPORTVLA": {"-": ["11"]},
    "SPRETOR": {"-": ["13"]},
    "SPRETVS": {"1": ["14"], "2": ["21"]},
    "SPVMA": {"-": ["11"]},
    "SPVMANS": {"-": ["25"]},
    "SPVMESCO": {"-": ["53"]},
    "SPVMEVS": {"-": ["21"]},
    "SPVMIGER": {"-": ["21"]},
    "SPVMO": {"-": ["51"]},
    "SPVMOSVS": {"-": ["21"]},
    "SPVO": {"-": ["53"]},
    "SPVRCE": {"-": ["6&"]},
    "SPVRCITIA": {"-": ["11"]},
    "SPVRCITIES": {"-": ["15"]},
    "SPVRCO": {"-": ["51"]},
    "SPVRCVS": {"-": ["21"]},
    "SPVRIVS": {"1": ["12"], "2": ["21"], "N": ["12"]},
    "SPVTVM": {"-": ["12"]},
    "SPVTVS": {"-": ["14"]},
    "SQVALENS": {"-": ["25"]},
    "SQVALEO": {"-": ["52"]},
    "SQVALIDVS": {"-": ["21"]},
    "SQVALOR": {"-": ["13"]},
    "SQVAMA": {"-": ["11"]},
    "SQVAMEVS": {"-": ["21"]},
    "SQVAMIFER": {"-": ["21"]},
    "SQVAMIGER": {"-": ["21"]},
    "SQVAMOSVS": {"-": ["21"]},
    "SQVILLA": {"-": ["11"]},
    "STABILIA": {"-": ["13"]},
    "STABILIMEN": {"-": ["13"]},
    "STABILIO": {"-": ["54"]},
    "STABILIS": {"-": ["24"]},
    "STABILITAS": {"-": ["13"]},
    "STABILITER": {"-": ["60"]},
    "STABILITOR": {"-": ["13"]},
    "STABVLARIVS": {"-": ["12"]},
    "STABVLO": {"-": ["51"]},
    "STABVLOR": {"-": ["5J"]},
    "STABVLVM": {"-": ["12"]},
    "STACTA": {"-": ["11"]},
    "STADIATVS": {"-": ["21"]},
    "STADIVM": {"-": ["12"]},
    "STAGNANS": {"-": ["25"]},
    "STAGNO": {"1": ["51"], "2": ["51"]},
    "STAGNVM": {"1": ["12"], "2": ["12"]},
    "STAIENVS": {"N": ["12"]},
    "STAMEN": {"-": ["13"]},
    "STAMINEVS": {"-": ["21"]},
    "STATERA": {"-": ["11"]},
    "STATICVLVS": {"-": ["12"]},
    "STATIM": {"-": ["60"]},
    "STATIO": {"-": ["13"]},
    "STATIVA": {"-": ["12"]},
    "STATIVS": {"N": ["12"]},
    "STATIVVS": {"-": ["21"]},
    "STATOR": {"-": ["13"], "N": ["13"]},
    "STATVA": {"-": ["11"]},
    "STATVARIVS": {"1": ["12"], "2": ["21"]},
    "STATVMEN": {"-": ["13"]},
    "STATVMINATIO": {"-": ["13"]},
    "STATVMINO": {"-": ["51"]},
    "STATVO": {"-": ["53"]},
    "STATVRA": {"-": ["11"]},
    "STATVS": {"1": ["14"], "2": ["21"]},
    "STATVTIO": {"-": ["13"]},
    "STATVTVM": {"-": ["12"]},
    "STATVTVS": {"-": ["21"]},
    "STELLA": {"-": ["11"], "N": ["11"]},
    "STELLIFER": {"-": ["21"]},
    "STELLIGER": {"-": ["21"]},
    "STELLIO": {"-": ["13"]},
    "STELLO": {"-": ["51"]},
    "STEMMA": {"-": ["17"]},
    "STERCORARIVS": {"-": ["21"]},
    "STERCORO": {"-": ["51"]},
    "STERCOROSVS": {"-": ["21"]},
    "STERCVLINVM": {"-": ["12"]},
    "STERCVS": {"-": ["13"]},
    "STEREOBATES": {"-": ["17"]},
    "STERILIS": {"-": ["24"]},
    "STERILITAS": {"-": ["13"]},
    "STERILVS": {"-": ["21"]},
    "STERNO": {"-": ["53"]},
    "STERNVMENTVM": {"-": ["12"]},
    "STERNVO": {"-": ["53"]},
    "STERNVTAMENTVM": {"-": ["12"]},
    "STERTO": {"-": ["53"]},
    "STESICHORVS": {"N": ["12"]},
    "STHENIANVS": {"N": ["21"], "A": ["21"]},
    "STHENIVS": {"N": ["12"]},
    "STIGMA": {"-": ["17"]},
    "STILLA": {"-": ["11"]},
    "STILLARIVM": {"-": ["12"]},
    "STILLICIDIVM": {"-": ["12"]},
    "STILLO": {"-": ["51"]},
    "STILVS": {"-": ["12"]},
    "STIMVLO": {"-": ["51"]},
    "STIMVLVS": {"-": ["12"]},
    "STINGVO": {"-": ["53"]},
    "STIPATOR": {"-": ["13"]},
    "STIPATVS": {"-": ["21"]},
    "STIPENDIALIS": {"-": ["24"]},
    "STIPENDIARIVS": {"1": ["12"], "2": ["21"]},
    "STIPENDIOR": {"-": ["5J"]},
    "STIPENDIOSVS": {"-": ["21"]},
    "STIPENDIVM": {"-": ["12"]},
    "STIPES": {"-": ["13"]},
    "STIPO": {"-": ["51"]},
    "STIPS": {"-": ["13"]},
    "STIPVLA": {"-": ["11"]},
    "STIPVLATIO": {"-": ["13"]},
    "STIPVLOR": {"-": ["5J"]},
    "STIRIA": {"-": ["11"]},
    "STIRPS": {"-": ["13"]},
    "STITVO": {"-": ["53"]},
    "STIVA": {"-": ["11"]},
    "STLATARIVS": {"-": ["12"]},
    "STLOPPVS": {"-": ["12"]},
    "STO": {"-": ["51"]},
    "STOECHION": {"-": ["17"]},
    "STOICA": {"1": ["12"], "2": ["17"]},
    "STOICI": {"N": ["12"]},
    "STOICVS": {"N": ["21"], "A": ["21"]},
    "STOLA": {"-": ["11"]},
    "STOLATVS": {"-": ["21"]},
    "STOLIDE": {"-": ["60"]},
    "STOLIDVS": {"-": ["21"]},
    "STOMACHICVS": {"-": ["12"]},
    "STOMACHOR": {"-": ["5J"]},
    "STOMACHOSVS": {"-": ["21"]},
    "STOMACHVS": {"-": ["12"]},
    "STOREA": {"-": ["11"]},
    "STRABO": {"-": ["13"], "N": ["13"]},
    "STRAGES": {"-": ["13"]},
    "STRAGVLA": {"-": ["11"]},
    "STRAGVLVM": {"-": ["12"]},
    "STRAGVLVS": {"-": ["21"]},
    "STRAMEN": {"-": ["13"]},
    "STRAMENTARIVS": {"-": ["21"]},
    "STRAMENTICIVS": {"-": ["21"]},
    "STRAMENTVM": {"-": ["12"]},
    "STRAMINEVS": {"-": ["21"]},
    "STRANGVLO": {"-": ["51"]},
    "STRANGVRIA": {"-": ["11"]},
    "STRATVM": {"-": ["12"]},
    "STRENVE": {"-": ["60"]},
    "STRENVVS": {"-": ["21"]},
    "STREPITO": {"-": ["51"]},
    "STREPITVS": {"-": ["14"]},
    "STREPO": {"-": ["53"]},
    "STRIA": {"-": ["11"]},
    "STRICTIM": {"-": ["60"]},
    "STRICTIVVS": {"-": ["21"]},
    "STRICTOR": {"-": ["13"]},
    "STRICTVS": {"-": ["21"]},
    "STRIDENS": {"-": ["25"]},
    "STRIDEO": {"-": ["52"]},
    "STRIDO": {"-": ["53"]},
    "STRIDOR": {"-": ["13"]},
    "STRIDVLVS": {"-": ["21"]},
    "STRIGILIS": {"-": ["13"]},
    "STRIGO": {"-": ["51"]},
    "STRINGO": {"-": ["53"]},
    "STRINGOR": {"-": ["13"]},
    "STRIO": {"-": ["51"]},
    "STRIX": {"-": ["13"]},
    "STROPHA": {"-": ["17"]},
    "STROPHIVM": {"-": ["12"]},
    "STROPHVS": {"-": ["12"]},
    "STRVCTILIS": {"-": ["21"]},
    "STRVCTOR": {"-": ["13"]},
    "STRVCTVRA": {"-": ["11"]},
    "STRVES": {"-": ["13"]},
    "STRVMA": {"-": ["11"]},
    "STRVMOSVS": {"-": ["21"]},
    "STRVO": {"-": ["53"]},
    "STRVTHEVS": {"-": ["21"]},
    "STRVTHOCAMELVS": {"-": ["12"]},
    "STVDENS": {"-": ["25"]},
    "STVDEO": {"-": ["52"]},
    "STVDIOSE": {"-": ["60"]},
    "STVDIOSVS": {"-": ["21"]},
    "STVDIVM": {"-": ["12"]},
    "STVLTE": {"-": ["60"]},
    "STVLTI": {"-": ["12"], "N": ["12"]},
    "STVLTITIA": {"-": ["11"]},
    "STVLTVS": {"-": ["21"]},
    "STVPEFACIO": {"-": ["55"]},
    "STVPEFIO": {"-": ["56"]},
    "STVPEO": {"-": ["52"]},
    "STVPESCO": {"-": ["53"]},
    "STVPIDVS": {"-": ["21"]},
    "STVPOR": {"-": ["13"]},
    "STVPPA": {"-": ["11"]},
    "STVPPEVS": {"-": ["21"]},
    "STVPRATOR": {"-": ["13"]},
    "STVPRO": {"-": ["51"]},
    "STVPRVM": {"-": ["12"]},
    "STYGIVS": {"-": ["21"], "A": ["21"]},
    "STYLOBATA": {"-": ["17"]},
    "STYX": {"N": ["17"]},
    "SUPERINICIO": {"-": ["55"]},
    "SVADEO": {"-": ["52"]},
    "SVASIO": {"-": ["13"]},
    "SVASOR": {"-": ["13"]},
    "SVASORIA": {"-": ["11"]},
    "SVASVS": {"-": ["14"]},
    "SVAVE": {"-": ["60"]},
    "SVAVIA": {"-": ["13"]},
    "SVAVIDICVS": {"-": ["21"]},
    "SVAVILOQVENS": {"-": ["25"]},
    "SVAVIOLVM": {"-": ["12"]},
    "SVAVIOR": {"-": ["5J"]},
    "SVAVIS": {"-": ["24"]},
    "SVAVITAS": {"-": ["13"]},
    "SVAVITER": {"-": ["60"]},
    "SVAVIVM": {"-": ["12"]},
    "SVB": {"-": ["70"]},
    "SVBACIDVS": {"-": ["21"]},
    "SVBACTIO": {"-": ["13"]},
    "SVBAERATVS": {"-": ["21"]},
    "SVBARESCO": {"-": ["53"]},
    "SVBCAVVS": {"-": ["21"]},
    "SVBCERNO": {"-": ["53"]},
    "SVBCRISPVS": {"-": ["21"]},
    "SVBCRVDVS": {"-": ["21"]},
    "SVBDEFICIO": {"-": ["55"]},
    "SVBDIFFICILIS": {"-": ["24"]},
    "SVBDITIVVS": {"-": ["21"]},
    "SVBDIV": {"-": ["60"]},
    "SVBDO": {"-": ["53"]},
    "SVBDOLE": {"-": ["60"]},
    "SVBDOLVS": {"-": ["21"]},
    "SVBDVCO": {"-": ["53"]},
    "SVBDVCTARIVS": {"-": ["21"]},
    "SVBDVCTIO": {"-": ["13"]},
    "SVBEDO": {"2": ["56"]},
    "SVBEO": {"1": ["56"], "-": ["52"]},
    "SVBER": {"-": ["13"]},
    "SVBFVSCVS": {"-": ["21"]},
    "SVBIACEO": {"-": ["52"]},
    "SVBICIO": {"-": ["55"]},
    "SVBIECTE": {"-": ["6-"]},
    "SVBIECTIO": {"-": ["13"]},
    "SVBIECTO": {"-": ["51"]},
    "SVBIECTOR": {"-": ["13"]},
    "SVBIECTVM": {"-": ["12"]},
    "SVBIECTVS": {"1": ["12"], "2": ["14"], "3": ["21"]},
    "SVBIGO": {"-": ["53"]},
    "SVBINDE": {"-": ["60"]},
    "SVBITARIVS": {"-": ["21"]},
    "SVBITO": {"-": ["60"]},
    "SVBITVM": {"-": ["12"]},
    "SVBITVS": {"-": ["21"]},
    "SVBIVGIVS": {"-": ["21"]},
    "SVBIVNGO": {"-": ["53"]},
    "SVBLABOR": {"2": ["5L"]},
    "SVBLECTO": {"-": ["51"]},
    "SVBLEGO": {"2": ["53"]},
    "SVBLEVO": {"1": ["51"]},
    "SVBLICA": {"-": ["11"]},
    "SVBLICIVS": {"-": ["21"], "N": ["21"], "A": ["21"]},
    "SVBLIDO": {"-": ["53"]},
    "SVBLIGAR": {"-": ["13"]},
    "SVBLIGO": {"2": ["51"], "3": ["53"]},
    "SVBLIME": {"1": ["13"], "2": ["60"]},
    "SVBLIMIS": {"-": ["24"]},
    "SVBLIMITAS": {"-": ["13"]},
    "SVBLIMITER": {"-": ["60"]},
    "SVBLIMO": {"-": ["51"]},
    "SVBLIMVS": {"1": ["21"]},
    "SVBLINO": {"-": ["53"]},
    "SVBLVCEO": {"-": ["52"]},
    "SVBLVO": {"-": ["53"]},
    "SVBLVSTRIS": {"-": ["24"]},
    "SVBMANO": {"-": ["51"]},
    "SVBMERGO": {"-": ["53"]},
    "SVBMINISTRATOR": {"-": ["13"]},
    "SVBMINISTRO": {"-": ["51"]},
    "SVBMISSE": {"-": ["60"]},
    "SVBMISSVS": {"-": ["21"]},
    "SVBMITTO": {"-": ["53"]},
    "SVBMOVEO": {"-": ["52"]},
    "SVBMVGIO": {"-": ["54"]},
    "SVBNASCOR": {"-": ["5L"]},
    "SVBNECTO": {"-": ["53"]},
    "SVBNITOR": {"2": ["5L"], "-": ["5L"]},
    "SVBNIXVS": {"-": ["21"]},
    "SVBNOTO": {"-": ["51"]},
    "SVBNVBA": {"-": ["11"]},
    "SVBNVBILVS": {"-": ["21"]},
    "SVBO": {"-": ["51"]},
    "SVBOLEO": {"-": ["52"]},
    "SVBOLES": {"-": ["13"]},
    "SVBORIOR": {"-": ["5M"]},
    "SVBORNO": {"-": ["51"]},
    "SVBORTVS": {"-": ["14"]},
    "SVBRADO": {"-": ["53"]},
    "SVBRECTITO": {"-": ["51"]},
    "SVBRECTVS": {"-": ["21"]},
    "SVBREPO": {"-": ["53"]},
    "SVBRIDEO": {"-": ["52"]},
    "SVBRIGO": {"-": ["53"]},
    "SVBRIPIO": {"-": ["55"]},
    "SVBROGO": {"-": ["51"]},
    "SVBROTATVS": {"-": ["21"]},
    "SVBRVBEO": {"-": ["52"]},
    "SVBRVBICVNDVS": {"-": ["21"]},
    "SVBRVFVS": {"-": ["21"]},
    "SVBRVO": {"-": ["53"]},
    "SVBSCRIBO": {"-": ["53"]},
    "SVBSCRIPTIO": {"-": ["13"]},
    "SVBSCVS": {"-": ["13"]},
    "SVBSECIVVS": {"-": ["21"]},
    "SVBSECO": {"-": ["51"]},
    "SVBSELLIVM": {"-": ["12"]},
    "SVBSEQVOR": {"-": ["5L"]},
    "SVBSERO": {"2": ["53"], "3": ["53"]},
    "SVBSESSOR": {"-": ["13"]},
    "SVBSIDENTIA": {"-": ["11"]},
    "SVBSIDEO": {"-": ["52"]},
    "SVBSIDIARII": {"-": ["12"]},
    "SVBSIDIARIVS": {"-": ["21"]},
    "SVBSIDIOR": {"-": ["5J"]},
    "SVBSIDIVM": {"-": ["12"]},
    "SVBSIDO": {"-": ["53"]},
    "SVBSILIO": {"-": ["54"]},
    "SVBSILVS": {"-": ["21"]},
    "SVBSISTO": {"-": ["53"]},
    "SVBSORTIOR": {"-": ["5M"]},
    "SVBSORTITIO": {"-": ["13"]},
    "SVBSTANTIA": {"-": ["11"]},
    "SVBSTERNO": {"-": ["53"]},
    "SVBSTILLVS": {"-": ["21"]},
    "SVBSTITVO": {"-": ["53"]},
    "SVBSTO": {"-": ["51"]},
    "SVBSTRINGO": {"-": ["53"]},
    "SVBSTRVCTIO": {"-": ["13"]},
    "SVBSTRVCTVM": {"-": ["12"]},
    "SVBSTRVO": {"-": ["53"]},
    "SVBSVM": {"1": ["56"]},
    "SVBSVO": {"-": ["53"]},
    "SVBTEGO": {"-": ["53"]},
    "SVBTEMEN": {"-": ["13"]},
    "SVBTENDO": {"-": ["53"]},
    "SVBTER": {"1": ["60"], "2": ["70"]},
    "SVBTERFLVO": {"-": ["53"]},
    "SVBTERFVGIO": {"-": ["55"]},
    "SVBTERLABOR": {"2": ["5L"]},
    "SVBTERO": {"-": ["53"]},
    "SVBTERRANEVM": {"-": ["12"]},
    "SVBTERRANEVS": {"-": ["21"]},
    "SVBTEXO": {"-": ["53"]},
    "SVBTILE": {"-": ["13"]},
    "SVBTILIS": {"-": ["24"]},
    "SVBTILITAS": {"-": ["13"]},
    "SVBTILITER": {"-": ["60"]},
    "SVBTRAHO": {"-": ["53"]},
    "SVBTVS": {"1": ["60"], "2": ["70"]},
    "SVBTVSVS": {"-": ["21"]},
    "SVBVCVLA": {"-": ["11"]},
    "SVBVECTIO": {"-": ["13"]},
    "SVBVECTO": {"-": ["51"]},
    "SVBVECTOR": {"-": ["13"]},
    "SVBVECTVS": {"-": ["14"]},
    "SVBVEHO": {"-": ["53"]},
    "SVBVENIO": {"-": ["54"]},
    "SVBVERSO": {"-": ["51"]},
    "SVBVERSOR": {"-": ["13"]},
    "SVBVERTO": {"-": ["53"]},
    "SVBVLA": {"-": ["11"]},
    "SVBVLCVS": {"-": ["12"]},
    "SVBVOLVO": {"-": ["53"]},
    "SVBVRBANI": {"-": ["12"]},
    "SVBVRBANITAS": {"-": ["13"]},
    "SVBVRBANVM": {"-": ["12"]},
    "SVBVRBANVS": {"-": ["21"]},
    "SVBVRGEO": {"-": ["52"]},
    "SVCCEDENS": {"-": ["13"]},
    "SVCCEDO": {"1": ["53"]},
    "SVCCENDO": {"-": ["53"]},
    "SVCCENSEO": {"-": ["52"]},
    "SVCCESSIO": {"-": ["13"]},
    "SVCCESSOR": {"-": ["13"]},
    "SVCCESSVS": {"-": ["14"]},
    "SVCCIDIA": {"-": ["11"]},
    "SVCCIDO": {"1": ["53"], "2": ["53"]},
    "SVCCIDVVS": {"-": ["21"]},
    "SVCCINCTVS": {"-": ["21"]},
    "SVCCINGO": {"-": ["53"]},
    "SVCCINO": {"-": ["53"]},
    "SVCCIPIO": {"-": ["55"]},
    "SVCCRESCO": {"-": ["53"]},
    "SVCCVBO": {"2": ["51"]},
    "SVCCVMBO": {"-": ["53"]},
    "SVCCVRRO": {"-": ["53"]},
    "SVCCVTIO": {"-": ["55"]},
    "SVCIDVS": {"-": ["21"]},
    "SVCINVM": {"-": ["12"]},
    "SVCINVS": {"1": ["12"], "2": ["21"]},
    "SVCVLA": {"-": ["11"]},
    "SVCVS": {"-": ["12"]},
    "SVDARIVM": {"-": ["12"]},
    "SVDATIO": {"-": ["13"]},
    "SVDATORIVM": {"-": ["12"]},
    "SVDIS": {"-": ["13"]},
    "SVDO": {"-": ["51"]},
    "SVDOR": {"-": ["13"]},
    "SVDVS": {"-": ["21"]},
    "SVESCO": {"-": ["53"]},
    "SVESSA": {"N": ["11"]},
    "SVETTIVS": {"N": ["12"]},
    "SVETVS": {"-": ["21"]},
    "SVFES": {"-": ["13"]},
    "SVFFERO": {"-": ["56"]},
    "SVFFICIENS": {"-": ["25"]},
    "SVFFICIO": {"-": ["55"]},
    "SVFFIGO": {"-": ["53"]},
    "SVFFIO": {"-": ["54"]},
    "SVFFLAMEN": {"-": ["13"]},
    "SVFFLO": {"-": ["51"]},
    "SVFFOCATIO": {"-": ["13"]},
    "SVFFOCO": {"1": ["51"], "2": ["51"]},
    "SVFFODIO": {"-": ["55"]},
    "SVFFOSSIO": {"-": ["13"]},
    "SVFFRAGATIO": {"-": ["13"]},
    "SVFFRAGATOR": {"-": ["13"]},
    "SVFFRAGIVM": {"-": ["12"]},
    "SVFFRAGOR": {"-": ["5J"]},
    "SVFFVGIO": {"-": ["55"]},
    "SVFFVGIVM": {"-": ["12"]},
    "SVFFVLCIO": {"-": ["54"]},
    "SVFFVNDO": {"1": ["51"], "2": ["53"]},
    "SVFFVSIO": {"-": ["13"]},
    "SVGGERO": {"-": ["53"]},
    "SVGGESTVM": {"-": ["12"]},
    "SVGGESTVS": {"-": ["14"]},
    "SVGGREDIOR": {"-": ["5N"]},
    "SVGGRVNDA": {"-": ["11"]},
    "SVGGRVNDATIO": {"-": ["13"]},
    "SVGGRVNDIVM": {"-": ["12"]},
    "SVGILLATIO": {"-": ["13"]},
    "SVGILLO": {"-": ["51"]},
    "SVGO": {"-": ["53"]},
    "SVI": {"1": ["43"], "2": ["44"]},
    "SVILLVS": {"-": ["21"]},
    "SVLCO": {"-": ["51"]},
    "SVLCVS": {"-": ["12"]},
    "SVLFVR": {"-": ["13"]},
    "SVLFVRATVS": {"-": ["21"]},
    "SVLFVREVS": {"-": ["21"]},
    "SVLFVROSVS": {"-": ["21"]},
    "SVLLA": {"N": ["11"]},
    "SVLPICIVS": {"N": ["12"], "A": ["21"]},
    "SVLTO": {"-": ["51"]},
    "SVM": {"1": ["56"], "2": ["E6"]},
    "SVMEN": {"-": ["13"]},
    "SVMMA": {"-": ["11"]},
    "SVMMARIVM": {"-": ["12"]},
    "SVMMARIVS": {"1": ["12"], "2": ["21"]},
    "SVMMATIM": {"-": ["60"]},
    "SVMMATVS": {"-": ["14"]},
    "SVMME": {"-": ["60"]},
    "SVMMOPERE": {"-": ["60"]},
    "SVMMVLA": {"-": ["11"]},
    "SVMMVM": {"1": ["12"], "2": ["60"]},
    "SVMO": {"-": ["53"]},
    "SVMPTIO": {"-": ["13"]},
    "SVMPTVARIVS": {"-": ["21"]},
    "SVMPTVOSE": {"-": ["60"]},
    "SVMPTVOSVS": {"-": ["21"]},
    "SVMPTVS": {"1": ["12"], "2": ["14"]},
    "SVO": {"-": ["53"]},
    "SVOVETAVRILIA": {"-": ["13"]},
    "SVPELLEX": {"-": ["13"]},
    "SVPER": {"1": ["60"], "2": ["70"]},
    "SVPERA": {"1": ["60"], "2": ["70"]},
    "SVPERABILIS": {"-": ["24"]},
    "SVPERADDO": {"-": ["53"]},
    "SVPERANS": {"-": ["25"]},
    "SVPERATIO": {"-": ["13"]},
    "SVPERATOR": {"-": ["13"]},
    "SVPERBE": {"-": ["60"]},
    "SVPERBIA": {"-": ["11"]},
    "SVPERBIFICVS": {"-": ["21"]},
    "SVPERBIO": {"-": ["54"]},
    "SVPERBVM": {"-": ["12"]},
    "SVPERBVS": {"1": ["12"], "2": ["21"], "N": ["12"]},
    "SVPERCADO": {"-": ["53"]},
    "SVPERCERNO": {"-": ["53"]},
    "SVPERCILIOSVS": {"-": ["21"]},
    "SVPERCILIVM": {"-": ["12"]},
    "SVPERCRESCO": {"-": ["53"]},
    "SVPERCVRRO": {"-": ["53"]},
    "SVPEREMINEO": {"-": ["52"]},
    "SVPERFICIARIVS": {"1": ["12"], "2": ["21"]},
    "SVPERFIO": {"-": ["56"]},
    "SVPERFLVENS": {"-": ["25"]},
    "SVPERFLVO": {"1": ["53"], "2": ["60"]},
    "SVPERFVNDO": {"2": ["53"]},
    "SVPERGREDIOR": {"-": ["5N"]},
    "SVPERI": {"-": ["12"]},
    "SVPERIACIO": {"-": ["55"]},
    "SVPERIMPONO": {"-": ["53"]},
    "SVPERINCVMBO": {"-": ["53"]},
    "SVPERIOR": {"-": ["13", "2&"]},
    "SVPERIORA": {"-": ["13"]},
    "SVPERIVS": {"-": ["6&"]},
    "SVPERLABOR": {"2": ["5L"]},
    "SVPERLIGO": {"2": ["51"], "-": ["51"]},
    "SVPERNAS": {"2": ["22"]},
    "SVPERNE": {"-": ["60"]},
    "SVPERNVM": {"-": ["12"]},
    "SVPERNVS": {"-": ["21"]},
    "SVPERO": {"-": ["51"]},
    "SVPERPENDENS": {"-": ["25"]},
    "SVPERPONO": {"-": ["53"]},
    "SVPERSCANDO": {"-": ["53"]},
    "SVPERSEDEO": {"-": ["52"]},
    "SVPERSERO": {"3": ["53"]},
    "SVPERSIDO": {"-": ["53"]},
    "SVPERSPICIO": {"-": ["55"]},
    "SVPERSTAGNO": {"-": ["51"]},
    "SVPERSTES": {"-": ["22"]},
    "SVPERSTITIO": {"-": ["13"]},
    "SVPERSTITIOSE": {"-": ["60"]},
    "SVPERSTO": {"-": ["51"]},
    "SVPERSTRVO": {"-": ["53"]},
    "SVPERSVM": {"1": ["56"]},
    "SVPERTEGO": {"-": ["53"]},
    "SVPERVACANEVS": {"-": ["21"]},
    "SVPERVACVVM": {"-": ["12"]},
    "SVPERVACVVS": {"-": ["21"]},
    "SVPERVADO": {"2": ["53"]},
    "SVPERVECTO": {"-": ["51"]},
    "SVPERVEHO": {"-": ["53"]},
    "SVPERVEHOR": {"-": ["5L"]},
    "SVPERVENIO": {"-": ["54"]},
    "SVPERVINCO": {"-": ["53"]},
    "SVPERVIVO": {"-": ["53"]},
    "SVPERVM": {"-": ["12"]},
    "SVPERVOLO": {"2": ["51"]},
    "SVPERVRGEO": {"-": ["52"]},
    "SVPERVS": {"-": ["21"]},
    "SVPINE": {"-": ["60"]},
    "SVPINO": {"-": ["51"]},
    "SVPINVS": {"-": ["21"]},
    "SVPPARVM": {"-": ["12"]},
    "SVPPEDITO": {"-": ["51"]},
    "SVPPERNATVS": {"-": ["21"]},
    "SVPPETIAE": {"-": ["11"]},
    "SVPPETO": {"-": ["53"]},
    "SVPPILO": {"2": ["51"]},
    "SVPPLANTO": {"-": ["51"]},
    "SVPPLAVDO": {"-": ["53"]},
    "SVPPLEMENTVM": {"-": ["12"]},
    "SVPPLEO": {"-": ["52"]},
    "SVPPLEX": {"1": ["13"], "2": ["25"]},
    "SVPPLICATIO": {"-": ["13"]},
    "SVPPLICITER": {"-": ["60"]},
    "SVPPLICIVM": {"-": ["12"]},
    "SVPPLICO": {"-": ["51"]},
    "SVPPONO": {"-": ["53"]},
    "SVPPORTO": {"-": ["51"]},
    "SVPPRESSIO": {"-": ["13"]},
    "SVPPRESSVS": {"-": ["21"]},
    "SVPPRIMO": {"-": ["53"]},
    "SVPPVRATIO": {"-": ["13"]},
    "SVPPVRO": {"-": ["51"]},
    "SVPPVS": {"-": ["21"]},
    "SVPPVTO": {"-": ["51"]},
    "SVPRA": {"1": ["60"], "2": ["70"]},
    "SVPRADICO": {"2": ["53"]},
    "SVPRASCANDO": {"-": ["53"]},
    "SVPRAVIVO": {"-": ["53"]},
    "SVPREMO": {"-": ["6-"], "1": ["6-"]},
    "SVPREMVM": {"1": ["12"], "2": ["6-"]},
    "SVRA": {"-": ["11"], "N": ["11"]},
    "SVRCVLVS": {"-": ["12"]},
    "SVRDVS": {"-": ["21"]},
    "SVRGO": {"-": ["53"]},
    "SVRRIPIO": {"-": ["55"]},
    "SVRSVM": {"-": ["60"]},
    "SVRSVMVORSVM": {"-": ["60"]},
    "SVS": {"-": ["13"]},
    "SVSCENSEO": {"-": ["52"]},
    "SVSCEPTIO": {"-": ["13"]},
    "SVSCIPIO": {"-": ["55"]},
    "SVSCITO": {"1": ["51"]},
    "SVSPECTO": {"1": ["51"], "2": ["60"]},
    "SVSPECTVS": {"1": ["14"], "2": ["21"]},
    "SVSPENDIVM": {"-": ["12"]},
    "SVSPENDO": {"-": ["53"]},
    "SVSPENSIO": {"-": ["13"]},
    "SVSPENSVM": {"-": ["12"]},
    "SVSPENSVRA": {"-": ["11"]},
    "SVSPENSVS": {"-": ["21"]},
    "SVSPICAX": {"-": ["25"]},
    "SVSPICIO": {"1": ["13"], "2": ["55"]},
    "SVSPICIOSVS": {"-": ["21"]},
    "SVSPICOR": {"-": ["5J"]},
    "SVSPIRIVM": {"-": ["12"]},
    "SVSPIRO": {"-": ["51"]},
    "SVSTENTO": {"-": ["51"]},
    "SVSTINEO": {"-": ["52"]},
    "SVSTOLLO": {"-": ["53"]},
    "SVSVRRO": {"-": ["51"]},
    "SVSVRRVS": {"-": ["12"]},
    "SVTILIS": {"-": ["24"]},
    "SVTOR": {"-": ["13"]},
    "SVTORIVS": {"-": ["21"]},
    "SVTRINA": {"-": ["11"]},
    "SVTRINVM": {"-": ["12"]},
    "SVTRINVS": {"-": ["21"]},
    "SVVM": {"-": ["12"]},
    "SVVS": {"-": ["44"]},
    "SYLLA": {"N": ["11"]},
    "SYLLABA": {"-": ["11"]},
    "SYLLANVS": {"N": ["21"], "A": ["21"]},
    "SYLLOGISMVS": {"-": ["12"]},
    "SYMBOLVS": {"-": ["12"]},
    "SYMMACHVS": {"N": ["12"]},
    "SYMMETRIA": {"-": ["11"]},
    "SYMMETROS": {"-": ["27"]},
    "SYMPATHIA": {"-": ["11"]},
    "SYMPHONIA": {"-": ["11"]},
    "SYMPHONIACVS": {"1": ["12"], "2": ["21"]},
    "SYNEMMENOS": {"-": ["27"]},
    "SYNGRAPHA": {"-": ["17"]},
    "SYNGRAPHE": {"-": ["17"]},
    "SYNHECHON": {"-": ["57"]},
    "SYRACVSAE": {"N": ["11"]},
    "SYRACVSANI": {"N": ["12"]},
    "SYRACVSANVS": {"N": ["21"], "A": ["21"]},
    "SYRI": {"N": ["12"]},
    "SYRIA": {"N": ["11"]},
    "SYRINX": {"-": ["17"]},
    "SYRMA": {"-": ["17"]},
    "SYRTIS": {"N": ["17"], "-": ["13"]},
    "SYSTYLOS": {"-": ["27"]},
    "TABELLA": {"-": ["11"]},
    "TABELLARIVS": {"1": ["12"], "2": ["21"]},
    "TABEO": {"-": ["52"]},
    "TABERNA": {"-": ["11"]},
    "TABERNACVLVM": {"-": ["12"]},
    "TABERNARIVS": {"1": ["12"], "2": ["21"]},
    "TABES": {"-": ["13"], "N": ["13"]},
    "TABESCO": {"-": ["53"]},
    "TABIDVS": {"-": ["21"]},
    "TABIFICVS": {"-": ["21"]},
    "TABVLA": {"-": ["11"]},
    "TABVLARIVM": {"-": ["12"]},
    "TABVLARIVS": {"1": ["12"], "2": ["21"]},
    "TABVLATIO": {"-": ["13"]},
    "TABVLATVM": {"-": ["12"]},
    "TABVLINVM": {"-": ["12"]},
    "TABVM": {"-": ["12"]},
    "TACEO": {"-": ["52"]},
    "TACITE": {"-": ["60"]},
    "TACITVM": {"-": ["12"]},
    "TACITVRNITAS": {"-": ["13"]},
    "TACITVRNVS": {"-": ["21"]},
    "TACITVS": {"-": ["21"], "N": ["12"], "A": ["21"]},
    "TACTILIS": {"-": ["24"]},
    "TACTIO": {"-": ["13"]},
    "TACTVS": {"-": ["14"]},
    "TADIVS": {"N": ["12"]},
    "TAEDA": {"-": ["11"]},
    "TAEDET": {"-": ["52"]},
    "TAEDIFER": {"-": ["21"]},
    "TAEDIVM": {"-": ["12"]},
    "TAENIA": {"-": ["11"]},
    "TALARIA": {"-": ["13"]},
    "TALARIS": {"-": ["24"]},
    "TALEA": {"-": ["11"]},
    "TALENTVM": {"-": ["12"]},
    "TALIO": {"-": ["13"]},
    "TALIS": {"-": ["45"]},
    "TALPA": {"-": ["11"]},
    "TALVS": {"-": ["12"], "N": ["12"]},
    "TAM": {"-": ["60"]},
    "TAMDIV": {"-": ["60"]},
    "TAMEN": {"-": ["60"]},
    "TAMENETSI": {"-": ["82"]},
    "TAMETSI": {"1": ["81"], "2": ["82"]},
    "TAMINO": {"-": ["51"]},
    "TAMQVAM": {"1": ["66"], "2": ["82"]},
    "TAMTVMDEM": {"1": ["45"]},
    "TANAQVIL": {"N": ["13"]},
    "TANDEM": {"-": ["60"]},
    "TANGO": {"-": ["53"]},
    "TANTILLVM": {"1": ["45"], "2": ["60"]},
    "TANTILLVS": {"-": ["45"]},
    "TANTISPER": {"-": ["60"]},
    "TANTO": {"-": ["60"]},
    "TANTOPERE": {"-": ["60"]},
    "TANTVLVM": {"1": ["45"], "2": ["60"]},
    "TANTVLVS": {"-": ["45"]},
    "TANTVM": {"1": ["45"], "2": ["60"]},
    "TANTVMDEM": {"1": ["45"], "2": ["60"]},
    "TANTVMMODO": {"-": ["60"]},
    "TANTVS": {"-": ["45"]},
    "TANTVSDEM": {"-": ["45"]},
    "TAPETVM": {"-": ["12"]},
    "TARDE": {"-": ["60"]},
    "TARDESCO": {"-": ["53"]},
    "TARDILOQVVS": {"-": ["21"]},
    "TARDIPES": {"-": ["25"]},
    "TARDITAS": {"-": ["13"]},
    "TARDO": {"-": ["51"]},
    "TARDVS": {"-": ["21"]},
    "TARENTINI": {"N": ["12"]},
    "TARIVS": {"N": ["12"]},
    "TARPEIVS": {"A": ["21"], "N": ["12"]},
    "TARQVINIA": {"N": ["11"]},
    "TARQVINIANVS": {"N": ["21"]},
    "TARQVINIENSES": {"N": ["13"]},
    "TARQVINIENSIS": {"N": ["24"], "A": ["24"]},
    "TARQVINII": {"N": ["12"]},
    "TARQVINIVS": {"A": ["21"], "N": ["12"]},
    "TARTARA": {"N": ["12"]},
    "TARTAREVS": {"-": ["21"], "A": ["21"]},
    "TARTARVS": {"N": ["12"]},
    "TATIVS": {"N": ["12"], "A": ["21"]},
    "TAVREA": {"-": ["11"], "N": ["11"]},
    "TAVREVS": {"-": ["21"]},
    "TAVRIFORMIS": {"-": ["24"]},
    "TAVRINVS": {"-": ["21"], "A": ["21"]},
    "TAVRIS": {"N": ["13"]},
    "TAVROBOLIVM": {"-": ["12"]},
    "TAVROMENITANI": {"N": ["12"]},
    "TAVROMENITANVS": {"N": ["21"], "A": ["21"]},
    "TAVRVS": {"-": ["12"], "N": ["12"]},
    "TAXATIO": {"-": ["13"]},
    "TAXILLVS": {"-": ["12"]},
    "TAXIS": {"-": ["17"]},
    "TAXO": {"-": ["51"]},
    "TAXVS": {"-": ["12"]},
    "TECTE": {"-": ["6&"]},
    "TECTINATVS": {"-": ["21"]},
    "TECTOR": {"-": ["13"]},
    "TECTORIVM": {"-": ["12"]},
    "TECTORIVS": {"-": ["21"]},
    "TECTVM": {"-": ["12"]},
    "TECTVS": {"-": ["21"], "A": ["21"]},
    "TEGES": {"-": ["13"]},
    "TEGIMEN": {"-": ["13"]},
    "TEGIMENTVM": {"-": ["12"]},
    "TEGO": {"-": ["53"]},
    "TEGVLA": {"-": ["11"]},
    "TELA": {"-": ["11"]},
    "TELEOS": {"-": ["27"]},
    "TELLENA": {"N": ["12"]},
    "TELLVS": {"-": ["13"], "N": ["13"]},
    "TELVM": {"-": ["12"]},
    "TEMENITES": {"N": ["17"]},
    "TEMERARIVS": {"-": ["21"]},
    "TEMERE": {"-": ["60"]},
    "TEMERITAS": {"-": ["13"]},
    "TEMERO": {"-": ["51"]},
    "TEMESA": {"N": ["17"]},
    "TEMETVM": {"-": ["12"]},
    "TEMNO": {"-": ["53"]},
    "TEMO": {"-": ["13"]},
    "TEMPE": {"-": ["17"], "N": ["11"]},
    "TEMPERAMENTVM": {"-": ["12"]},
    "TEMPERANS": {"-": ["25"]},
    "TEMPERANTER": {"-": ["60"]},
    "TEMPERANTIA": {"-": ["11"]},
    "TEMPERATE": {"-": ["60"]},
    "TEMPERATIO": {"-": ["13"]},
    "TEMPERATOR": {"-": ["13"]},
    "TEMPERATVRA": {"-": ["11"]},
    "TEMPERATVS": {"1": ["14"], "2": ["21"]},
    "TEMPERIES": {"-": ["15"]},
    "TEMPERO": {"-": ["51"]},
    "TEMPESTAS": {"-": ["13"], "N": ["13"]},
    "TEMPESTIVE": {"-": ["60"]},
    "TEMPESTIVO": {"-": ["60"]},
    "TEMPESTIVVS": {"-": ["21"]},
    "TEMPLVM": {"-": ["12"]},
    "TEMPORARIVM": {"-": ["12"]},
    "TEMPORARIVS": {"-": ["21"]},
    "TEMPSANVS": {"N": ["21"], "A": ["21"]},
    "TEMPTATOR": {"-": ["13"]},
    "TEMPVS": {"1": ["13"], "2": ["13"]},
    "TEMVLENTVS": {"1": ["12"], "2": ["21"]},
    "TENACITER": {"-": ["60"]},
    "TENAX": {"-": ["25"]},
    "TENDICVLA": {"-": ["11"]},
    "TENDO": {"-": ["53"]},
    "TENEBRAE": {"-": ["11"]},
    "TENEBRICOSVS": {"-": ["21"]},
    "TENEBROSVS": {"-": ["21"]},
    "TENEDII": {"N": ["12"]},
    "TENEDVS": {"N": ["17"]},
    "TENELLVLVS": {"-": ["21"]},
    "TENEO": {"-": ["52"]},
    "TENER": {"-": ["21"]},
    "TENERASCO": {"-": ["53"]},
    "TENERE": {"-": ["60"]},
    "TENERITAS": {"-": ["13"]},
    "TENERVM": {"-": ["12"]},
    "TENES": {"N": ["17"]},
    "TENOR": {"-": ["13"]},
    "TENSA": {"-": ["11"]},
    "TENSVS": {"-": ["21"]},
    "TENTABVNDVS": {"-": ["21"]},
    "TENTIGO": {"-": ["13"]},
    "TENTIO": {"-": ["13"]},
    "TENTO": {"-": ["51"]},
    "TENTORIOLVM": {"-": ["12"]},
    "TENTORIVM": {"-": ["12"]},
    "TENTVM": {"-": ["12"]},
    "TENVE": {"-": ["13"]},
    "TENVIS": {"-": ["24"]},
    "TENVITAS": {"-": ["13"]},
    "TENVITER": {"-": ["60"]},
    "TENVO": {"-": ["51"]},
    "TENVS": {"1": ["16"], "2": ["70"], "N": ["12"]},
    "TEPEFACIO": {"-": ["55"]},
    "TEPEFACTO": {"-": ["51"]},
    "TEPEFIO": {"-": ["56"]},
    "TEPEO": {"-": ["52"]},
    "TEPESCO": {"-": ["53"]},
    "TEPIDARIVS": {"-": ["21"]},
    "TEPIDVS": {"-": ["21"]},
    "TEPOR": {"-": ["13"]},
    "TER": {"-": ["36"]},
    "TERCENTVM": {"-": ["31"]},
    "TERDECIES": {"-": ["36"]},
    "TEREBINTHVS": {"-": ["12"]},
    "TEREBRA": {"-": ["11"]},
    "TEREBRATIO": {"-": ["13"]},
    "TEREBRO": {"-": ["51"]},
    "TEREDO": {"-": ["13"]},
    "TERENTIVS": {"A": ["21"], "N": ["12"]},
    "TERES": {"-": ["25"]},
    "TERGEMINVS": {"-": ["21"]},
    "TERGEO": {"-": ["52"]},
    "TERGIVERSATIO": {"-": ["13"]},
    "TERGIVERSOR": {"-": ["5J"]},
    "TERGO": {"-": ["53"]},
    "TERGVM": {"-": ["12"]},
    "TERGVS": {"-": ["13"]},
    "TERMES": {"-": ["13"]},
    "TERMINALIA": {"-": ["13"], "N": ["11"]},
    "TERMINALIS": {"-": ["24"]},
    "TERMINATIO": {"-": ["13"]},
    "TERMINO": {"-": ["51"]},
    "TERMINVS": {"-": ["12"], "N": ["12"]},
    "TERNI": {"-": ["33"]},
    "TERO": {"-": ["53"]},
    "TERRA": {"-": ["11"], "N": ["11"]},
    "TERRENA": {"-": ["12"]},
    "TERRENI": {"-": ["12"]},
    "TERRENVM": {"-": ["12"]},
    "TERRENVS": {"-": ["21"]},
    "TERREO": {"-": ["52"]},
    "TERRESTRIS": {"-": ["24"]},
    "TERREVS": {"-": ["21"]},
    "TERRIBILIS": {"-": ["24"]},
    "TERRIFICO": {"-": ["51"]},
    "TERRIFICVS": {"-": ["21"]},
    "TERRIGENA": {"-": ["11"]},
    "TERRILOQVVS": {"-": ["21"]},
    "TERRITO": {"-": ["51"]},
    "TERRITORIVM": {"-": ["12"]},
    "TERROR": {"-": ["13"]},
    "TERROSVS": {"-": ["21"]},
    "TERSVS": {"1": ["14"], "2": ["21"]},
    "TERTIA": {"N": ["11"]},
    "TERTIANVS": {"1": ["12"], "2": ["21"]},
    "TERTIARIVS": {"1": ["12"], "2": ["21"]},
    "TERTIO": {"-": ["35"]},
    "TERTIVM": {"-": ["35"]},
    "TERTIVS": {"-": ["32"]},
    "TERTIVSDECIMVS": {"-": ["32"]},
    "TESCA": {"-": ["12"]},
    "TESSELA": {"-": ["11"]},
    "TESSERA": {"-": ["11"]},
    "TESSERVLA": {"-": ["11"]},
    "TESTA": {"-": ["11"]},
    "TESTACEVS": {"-": ["21"]},
    "TESTAMENTARIVS": {"2": ["21"], "1": ["12"]},
    "TESTAMENTVM": {"-": ["12"]},
    "TESTATVS": {"-": ["21"]},
    "TESTICVLVS": {"-": ["12"]},
    "TESTIFICATIO": {"-": ["13"]},
    "TESTIFICO": {"-": ["51"]},
    "TESTIFICOR": {"-": ["5J"]},
    "TESTIMONIVM": {"-": ["12"]},
    "TESTIS": {"1": ["13"], "2": ["13"]},
    "TESTO": {"-": ["51"]},
    "TESTOR": {"-": ["5J"]},
    "TESTV": {"-": ["16"]},
    "TESTVDINEATVS": {"-": ["21"]},
    "TESTVDINEVS": {"-": ["21"]},
    "TESTVDO": {"-": ["13"]},
    "TESTVM": {"-": ["12"]},
    "TETER": {"-": ["21"]},
    "TETRADOROS": {"1": ["27"], "2": ["60"]},
    "TETRANS": {"-": ["13"]},
    "TETRARCHES": {"-": ["17"], "N": ["13"]},
    "TETRASTYLVS": {"-": ["27"]},
    "TETRE": {"-": ["60"]},
    "TETRICVS": {"-": ["21"]},
    "TETTIVS": {"N": ["12"]},
    "TEXO": {"-": ["53"]},
    "TEXTILE": {"-": ["13"]},
    "TEXTILIS": {"-": ["24"]},
    "TEXTOR": {"-": ["13"]},
    "TEXTORIVS": {"-": ["21"]},
    "TEXTRINVM": {"-": ["12"]},
    "TEXTRIX": {"-": ["13"]},
    "TEXTVM": {"-": ["12"]},
    "TEXTVRA": {"-": ["11"]},
    "TEXTVS": {"-": ["14"]},
    "THALAMVS": {"-": ["12"]},
    "THALASSINVS": {"-": ["21"]},
    "THALASSIVS": {"N": ["12"]},
    "THAPSVS": {"-": ["12"], "N": ["12"]},
    "THARSVS": {"N": ["17"]},
    "THEATRALIS": {"-": ["24"]},
    "THEATRVM": {"-": ["12"]},
    "THEBANI": {"N": ["12"]},
    "THEBANVS": {"N": ["21"], "A": ["21"]},
    "THECA": {"-": ["11"]},
    "THEMATISMOS": {"-": ["17"]},
    "THEMISTAGORAS": {"N": ["17"]},
    "THEODORVS": {"N": ["12"]},
    "THEOLOGVS": {"-": ["12"]},
    "THEOMNASTVS": {"N": ["12"]},
    "THEOPANES": {"N": ["17"]},
    "THEORACTVS": {"N": ["12"]},
    "THERICLEVS": {"N": ["21"], "A": ["21"]},
    "THERMAE": {"-": ["11"]},
    "THERMITANI": {"N": ["12"]},
    "THERMITANVS": {"N": ["21"], "A": ["21"]},
    "THESAVRVS": {"-": ["12"]},
    "THESPIADES": {"N": ["13"]},
    "THESPIAE": {"N": ["11"]},
    "THESPIENSES": {"N": ["13"]},
    "THESSALVS": {"N": ["12"], "A": ["21"]},
    "THETA": {"-": ["17"]},
    "THIASVS": {"-": ["12"]},
    "THOLVS": {"-": ["12"]},
    "THOMIX": {"-": ["13"]},
    "THORAX": {"-": ["17"]},
    "THRACES": {"N": ["17"]},
    "THRACIA": {"N": ["11"]},
    "THRACIVS": {"N": ["21"], "A": ["21"]},
    "THRASO": {"N": ["17"]},
    "THRAX": {"N": ["27"], "A": ["25"]},
    "THVRINVS": {"N": ["21"], "A": ["21"]},
    "THYAS": {"-": ["17"], "N": ["17"]},
    "THYIVS": {"-": ["21"]},
    "THYMBRA": {"-": ["11"]},
    "THYMELICVS": {"-": ["21"]},
    "THYMVM": {"-": ["12"]},
    "THYNNVS": {"-": ["12"]},
    "THYROMA": {"-": ["17"]},
    "THYRORON": {"-": ["17"]},
    "THYRSIGER": {"-": ["21"]},
    "THYRSVS": {"-": ["12"]},
    "TIARA": {"-": ["11"]},
    "TIARAS": {"-": ["17"]},
    "TIBERINVS": {"N": ["12"], "A": ["21"]},
    "TIBERIS": {"N": ["13"]},
    "TIBERIVS": {"N": ["12"]},
    "TIBIA": {"-": ["11"]},
    "TIBICEN": {"-": ["13"]},
    "TIBICINA": {"-": ["11"]},
    "TICESCO": {"-": ["53"]},
    "TIGILLVM": {"-": ["12"]},
    "TIGNARIVS": {"-": ["21"]},
    "TIGNVM": {"-": ["12"]},
    "TIGRIS": {"-": ["17"], "N": ["13"]},
    "TILIA": {"-": ["11"]},
    "TIMARCHIDES": {"N": ["17"]},
    "TIMEFACIO": {"-": ["55"]},
    "TIMENS": {"-": ["25"]},
    "TIMEO": {"-": ["52"]},
    "TIMESCO": {"-": ["53"]},
    "TIMIDE": {"-": ["60"]},
    "TIMIDITAS": {"-": ["13"]},
    "TIMIDVS": {"-": ["21"]},
    "TIMOR": {"-": ["13"], "N": ["13"]},
    "TINEA": {"-": ["11"]},
    "TINEO": {"-": ["52"]},
    "TINGO": {"-": ["53"]},
    "TINNIO": {"-": ["54"]},
    "TINNITVS": {"-": ["14"]},
    "TINNVLVS": {"-": ["21"]},
    "TINTINNABVLVM": {"-": ["12"]},
    "TINTINNO": {"-": ["51"]},
    "TIRO": {"-": ["13"], "N": ["13"]},
    "TIROCINIVM": {"-": ["12"]},
    "TIRVNCVLVS": {"-": ["12"]},
    "TISSENSES": {"N": ["13"]},
    "TITIENSES": {"N": ["13"]},
    "TITILLATIO": {"-": ["13"]},
    "TITILLO": {"-": ["51"]},
    "TITINIVS": {"N": ["12"]},
    "TITIVS": {"N": ["12"], "A": ["21"]},
    "TITVBATIO": {"-": ["13"]},
    "TITVBO": {"-": ["51"]},
    "TITVLVS": {"-": ["12"]},
    "TITVS": {"N": ["12"], "-": ["12"]},
    "TLEPOLEMVS": {"N": ["12"]},
    "TOFVS": {"-": ["12"]},
    "TOGA": {"-": ["11"]},
    "TOGATA": {"-": ["11"]},
    "TOGATI": {"-": ["12"]},
    "TOGATVS": {"-": ["21"]},
    "TOLERABILIS": {"-": ["24"]},
    "TOLERABILITER": {"-": ["60"]},
    "TOLERANDVS": {"-": ["21"]},
    "TOLERANS": {"-": ["25"]},
    "TOLERANTIA": {"-": ["11"]},
    "TOLERATVS": {"1": ["14"], "2": ["21"]},
    "TOLERO": {"-": ["51"]},
    "TOLLO": {"-": ["53"]},
    "TOLVTARIVS": {"-": ["21"]},
    "TOMACVLVM": {"-": ["12"]},
    "TOMENTVM": {"-": ["12"]},
    "TONANS": {"-": ["25"], "A": ["22"], "N": ["13"]},
    "TONDEO": {"-": ["52"]},
    "TONITRALIS": {"-": ["24"]},
    "TONITRVS": {"-": ["14"]},
    "TONITRVVM": {"-": ["12"]},
    "TONO": {"-": ["51"]},
    "TONSA": {"1": ["11"], "2": ["11"]},
    "TONSOR": {"-": ["13"]},
    "TONSTRICVLA": {"-": ["11"]},
    "TONSVRA": {"-": ["11"]},
    "TONVS": {"-": ["12"]},
    "TOPEODVS": {"-": ["17"]},
    "TOPIA": {"-": ["12"]},
    "TORAL": {"-": ["13"]},
    "TORCVLAR": {"-": ["13"]},
    "TORCVLARIVM": {"-": ["12"]},
    "TORCVLVS": {"-": ["21"]},
    "TOREVMA": {"-": ["17"]},
    "TORMENTVM": {"-": ["12"]},
    "TORMINA": {"-": ["13"]},
    "TORNATVS": {"-": ["21"]},
    "TORNO": {"-": ["51"]},
    "TORNVS": {"-": ["12"]},
    "TOROSVS": {"-": ["21"]},
    "TORPEDO": {"-": ["13"]},
    "TORPEO": {"-": ["52"]},
    "TORPESCO": {"-": ["53"]},
    "TORPIDVS": {"-": ["21"]},
    "TORPOR": {"-": ["13"]},
    "TORQVATVS": {"-": ["21"], "N": ["12"]},
    "TORQVEO": {"-": ["52"]},
    "TORQVIS": {"-": ["13"]},
    "TORRENS": {"1": ["13"], "2": ["25"]},
    "TORREO": {"-": ["52"]},
    "TORRESCO": {"-": ["53"]},
    "TORRIDVS": {"-": ["21"]},
    "TORRO": {"-": ["53"]},
    "TORSIO": {"-": ["13"]},
    "TORTE": {"-": ["60"]},
    "TORTILIS": {"-": ["24"]},
    "TORTIO": {"-": ["13"]},
    "TORTIVVS": {"-": ["21"]},
    "TORTO": {"-": ["51"]},
    "TORTOR": {"-": ["13"]},
    "TORTVOSVS": {"-": ["21"]},
    "TORTVS": {"-": ["14"]},
    "TORVA": {"-": ["60"]},
    "TORVITAS": {"-": ["13"]},
    "TORVLVS": {"-": ["12"]},
    "TORVS": {"-": ["12"]},
    "TORVVM": {"-": ["60"]},
    "TORVVS": {"-": ["21"]},
    "TOT": {"-": ["45"]},
    "TOTIDEM": {"1": ["45"], "2": ["60"]},
    "TOTIENS": {"-": ["60"]},
    "TOTVM": {"-": ["12"]},
    "TOTVS": {"-": ["48"]},
    "TOXICVM": {"-": ["12"]},
    "TRABALIS": {"-": ["24"]},
    "TRABEA": {"-": ["11"]},
    "TRABEATVS": {"-": ["21"]},
    "TRABECVLA": {"-": ["11"]},
    "TRABS": {"-": ["13"]},
    "TRACTA": {"-": ["12"]},
    "TRACTABILIS": {"-": ["24"]},
    "TRACTABILITAS": {"-": ["13"]},
    "TRACTATIO": {"-": ["13"]},
    "TRACTATOR": {"-": ["13"]},
    "TRACTATVS": {"-": ["14"]},
    "TRACTIM": {"-": ["60"]},
    "TRACTIO": {"-": ["13"]},
    "TRACTO": {"-": ["51"]},
    "TRACTORIVS": {"-": ["21"]},
    "TRACTVS": {"-": ["14"]},
    "TRADITIO": {"-": ["13"]},
    "TRADO": {"-": ["53"]},
    "TRADVCO": {"-": ["53"]},
    "TRADVCTIO": {"-": ["13"]},
    "TRAGICE": {"-": ["60"]},
    "TRAGICVS": {"1": ["12"], "2": ["21"]},
    "TRAGOEDIA": {"-": ["11"], "N": ["11"]},
    "TRAGOEDVS": {"-": ["12"]},
    "TRAGVLA": {"-": ["11"]},
    "TRAHA": {"-": ["11"]},
    "TRAHO": {"-": ["53"]},
    "TRAICIO": {"-": ["55"]},
    "TRAIECTIO": {"-": ["13"]},
    "TRAIECTVRA": {"-": ["11"]},
    "TRAIECTVS": {"-": ["14"]},
    "TRAMA": {"-": ["11"]},
    "TRAMES": {"-": ["13"]},
    "TRANGRESSVS": {"-": ["14"]},
    "TRANO": {"-": ["51"]},
    "TRANQVILLE": {"-": ["6&"]},
    "TRANQVILLITAS": {"-": ["13"]},
    "TRANQVILLO": {"-": ["51"]},
    "TRANQVILLVM": {"-": ["12"]},
    "TRANQVILLVS": {"-": ["21"], "N": ["12"]},
    "TRANS": {"1": ["60"], "2": ["70"]},
    "TRANSACTOR": {"-": ["13"]},
    "TRANSALPINVS": {"N": ["21"], "A": ["21"]},
    "TRANSCENDO": {"-": ["53"]},
    "TRANSCOLO": {"1": ["51"]},
    "TRANSCRIBO": {"-": ["53"]},
    "TRANSCVRRO": {"-": ["53"]},
    "TRANSCVRSVS": {"-": ["14"]},
    "TRANSDVCO": {"-": ["53"]},
    "TRANSEO": {"1": ["56"]},
    "TRANSERO": {"2": ["53"]},
    "TRANSFERO": {"-": ["56"]},
    "TRANSFIGO": {"-": ["53"]},
    "TRANSFIGVRO": {"-": ["51"]},
    "TRANSFODIO": {"-": ["55"]},
    "TRANSFORMO": {"-": ["51"]},
    "TRANSFORO": {"-": ["51"]},
    "TRANSFVGA": {"-": ["11"]},
    "TRANSFVGIO": {"-": ["55"]},
    "TRANSFVGIVM": {"-": ["12"]},
    "TRANSFVNDO": {"2": ["53"]},
    "TRANSGERO": {"-": ["53"]},
    "TRANSGREDIOR": {"-": ["5N"]},
    "TRANSGRESSIO": {"-": ["13"]},
    "TRANSGRESSVS": {"-": ["14"]},
    "TRANSIGO": {"-": ["53"]},
    "TRANSILIO": {"-": ["54"]},
    "TRANSITIO": {"-": ["13"]},
    "TRANSITVS": {"-": ["14"]},
    "TRANSLATICIVS": {"-": ["21"]},
    "TRANSLATIO": {"-": ["13"]},
    "TRANSLATOR": {"-": ["13"]},
    "TRANSLATVS": {"-": ["14"]},
    "TRANSLEGO": {"1": ["51"], "2": ["53"]},
    "TRANSLVCEO": {"-": ["52"]},
    "TRANSMARINVS": {"-": ["21"]},
    "TRANSMEO": {"-": ["51"]},
    "TRANSMIGRO": {"-": ["51"]},
    "TRANSMISSVS": {"-": ["14"]},
    "TRANSMITTO": {"-": ["53"]},
    "TRANSMOVEO": {"-": ["52"]},
    "TRANSMVTO": {"2": ["51"]},
    "TRANSNO": {"-": ["51"]},
    "TRANSPARENS": {"-": ["25"]},
    "TRANSPECTVS": {"-": ["14"]},
    "TRANSPICIO": {"-": ["55"]},
    "TRANSPONO": {"-": ["53"]},
    "TRANSPORTATIO": {"-": ["13"]},
    "TRANSPORTO": {"-": ["51"]},
    "TRANSSCRIBO": {"-": ["53"]},
    "TRANSTRVM": {"-": ["12"]},
    "TRANSVECTIO": {"-": ["13"]},
    "TRANSVECTO": {"-": ["51"]},
    "TRANSVEHO": {"-": ["53"]},
    "TRANSVERBERO": {"2": ["51"]},
    "TRANSVERSE": {"-": ["60"]},
    "TRANSVERSO": {"-": ["51"]},
    "TRANSVERSVM": {"1": ["12"], "2": ["60"]},
    "TRANSVERSVS": {"-": ["21"]},
    "TRANSVERTO": {"-": ["53"]},
    "TRANSVIO": {"-": ["51"]},
    "TRANSVO": {"-": ["53"]},
    "TRANSVOLITO": {"-": ["51"]},
    "TRANSVOLO": {"2": ["51"]},
    "TRAPES": {"-": ["13"]},
    "TRAPETVM": {"-": ["12"]},
    "TRAPETVS": {"-": ["12"]},
    "TRAVLIZO": {"-": ["53"]},
    "TREBIVM": {"N": ["12"]},
    "TREBONIVS": {"N": ["12"]},
    "TRECENI": {"-": ["33"]},
    "TRECENTENI": {"-": ["33"]},
    "TRECENTESIMVS": {"-": ["32"]},
    "TRECENTI": {"-": ["31"], "N": ["12"]},
    "TRECENTIES": {"-": ["36"]},
    "TRECHEDIPNVM": {"-": ["12"]},
    "TRECTO": {"-": ["51"]},
    "TREDECIM": {"-": ["31"]},
    "TREMEBVNDVS": {"-": ["21"]},
    "TREMEFACIO": {"-": ["55"]},
    "TREMELLIVS": {"N": ["12"]},
    "TREMENDVS": {"-": ["21"]},
    "TREMESCO": {"-": ["53"]},
    "TREMO": {"-": ["53"]},
    "TREMOR": {"-": ["13"]},
    "TREMVLVS": {"1": ["12"], "2": ["12"], "3": ["21"], "N": ["12"]},
    "TREPIDANTER": {"-": ["6&"]},
    "TREPIDATIO": {"-": ["13"]},
    "TREPIDE": {"-": ["60"]},
    "TREPIDO": {"-": ["51"]},
    "TREPIDVS": {"-": ["21"]},
    "TRES": {"-": ["31"], "N": ["13"]},
    "TRESSIS": {"-": ["13"]},
    "TRESVIRI": {"-": ["12"]},
    "TRIARIVS": {"1": ["12"], "2": ["21"], "N": ["12"]},
    "TRIBON": {"-": ["17"]},
    "TRIBVLIS": {"-": ["24"]},
    "TRIBVLO": {"-": ["51"]},
    "TRIBVLVM": {"-": ["12"]},
    "TRIBVLVS": {"-": ["12"]},
    "TRIBVNAL": {"-": ["13"]},
    "TRIBVNATVS": {"-": ["14"]},
    "TRIBVNICIVS": {"1": ["12"], "2": ["21"]},
    "TRIBVNVS": {"-": ["12"]},
    "TRIBVO": {"-": ["53"]},
    "TRIBVS": {"-": ["14"]},
    "TRIBVTARIVS": {"-": ["21"]},
    "TRIBVTIM": {"-": ["60"]},
    "TRIBVTVM": {"-": ["12"]},
    "TRIBVTVS": {"1": ["14"], "2": ["21"]},
    "TRICENARIVS": {"-": ["21"]},
    "TRICENI": {"-": ["33"]},
    "TRICEPS": {"-": ["25"]},
    "TRICESIMVS": {"-": ["32"]},
    "TRICHALCVM": {"-": ["12"]},
    "TRICHILA": {"-": ["11"]},
    "TRICHO": {"N": ["13"]},
    "TRICIENS": {"-": ["36"]},
    "TRICIPITINVS": {"N": ["12"]},
    "TRICLINIVM": {"-": ["12"]},
    "TRICOR": {"-": ["5J"]},
    "TRICORPOR": {"-": ["25"]},
    "TRIDENS": {"1": ["13"], "2": ["25"]},
    "TRIDVVM": {"-": ["12"]},
    "TRIENNIVM": {"-": ["12"]},
    "TRIERARCHVS": {"-": ["12"]},
    "TRIERIS": {"2": ["24"]},
    "TRIETERICA": {"-": ["12"]},
    "TRIETERICVS": {"-": ["21"]},
    "TRIFARIAM": {"-": ["60"]},
    "TRIFAVX": {"-": ["25"]},
    "TRIFIDVS": {"-": ["21"]},
    "TRIFISSILIS": {"-": ["24"]},
    "TRIFORMIS": {"-": ["24"]},
    "TRIGINTA": {"-": ["31"]},
    "TRIGLYPHVS": {"-": ["12"]},
    "TRIGON": {"-": ["13"]},
    "TRIGONIVM": {"-": ["12"]},
    "TRIGONOS": {"-": ["27"]},
    "TRIGONVM": {"-": ["12"]},
    "TRILIBRIS": {"-": ["24"]},
    "TRILINGVIS": {"-": ["24"]},
    "TRILIX": {"-": ["25"]},
    "TRIMESTRIS": {"-": ["24"]},
    "TRIMVS": {"-": ["21"]},
    "TRINODIS": {"-": ["24"]},
    "TRINVS": {"-": ["33"]},
    "TRIOBOLVS": {"-": ["12"]},
    "TRIOCALINVS": {"N": ["21"], "A": ["21"]},
    "TRIONES": {"-": ["13"]},
    "TRIPARTIO": {"-": ["54", "13"]},
    "TRIPARTITO": {"-": ["60", "6-"]},
    "TRIPECTORVS": {"-": ["21"]},
    "TRIPEDANEVS": {"-": ["21"]},
    "TRIPERTITO": {"-": ["60"]},
    "TRIPES": {"1": ["13"], "2": ["25"]},
    "TRIPLEX": {"-": ["34", "13"]},
    "TRIPLINTHIVS": {"-": ["21"]},
    "TRIPODO": {"-": ["51"]},
    "TRIPTOLEMVS": {"N": ["12"]},
    "TRIPVDIO": {"-": ["51"]},
    "TRIPVDIVM": {"-": ["12"]},
    "TRIPVS": {"1": ["17"], "2": ["25"]},
    "TRIQVETRVS": {"-": ["21"], "A": ["21"]},
    "TRIREMIS": {"1": ["13"], "2": ["24"]},
    "TRISCVRRIVM": {"-": ["12"]},
    "TRISPASTOS": {"-": ["27"]},
    "TRISTE": {"1": ["13"], "2": ["60"]},
    "TRISTIMONIA": {"-": ["11"]},
    "TRISTIS": {"-": ["24"]},
    "TRISTITIA": {"-": ["11"]},
    "TRISTOR": {"-": ["5J"]},
    "TRISVLCVS": {"-": ["21"]},
    "TRITE": {"-": ["17"]},
    "TRITICEVS": {"-": ["21"]},
    "TRITICVM": {"-": ["12"]},
    "TRITVRA": {"-": ["11"]},
    "TRITVS": {"1": ["14"], "2": ["21"]},
    "TRIVIALIS": {"-": ["24"]},
    "TRIVIVM": {"-": ["12"]},
    "TRIVIVS": {"-": ["21"]},
    "TRIVMPHALIA": {"-": ["13"]},
    "TRIVMPHALIS": {"1": ["13"], "2": ["24"]},
    "TRIVMPHO": {"-": ["51"]},
    "TRIVMPHVS": {"-": ["12"], "N": ["12"]},
    "TRIVMVIR": {"-": ["12"]},
    "TRIVMVIRALIS": {"-": ["24"]},
    "TRIVMVIRATVS": {"-": ["14"]},
    "TROCHILVS": {"-": ["17"]},
    "TROCHLEA": {"-": ["11"]},
    "TROCHVS": {"-": ["12"]},
    "TROIA": {"N": ["11"]},
    "TROIANI": {"N": ["12"]},
    "TROIANVS": {"A": ["21"], "N": ["12"]},
    "TROPAEVM": {"-": ["12"]},
    "TROSSVLVS": {"-": ["12"]},
    "TRVCIDATIO": {"-": ["13"]},
    "TRVCIDO": {"-": ["51"]},
    "TRVCVLENTIA": {"-": ["11"]},
    "TRVCVLENTVS": {"-": ["21"], "N": ["12"]},
    "TRVDIS": {"-": ["13"]},
    "TRVDO": {"-": ["53"]},
    "TRVLLA": {"-": ["11"]},
    "TRVLLEVS": {"-": ["12"]},
    "TRVLLISSATIO": {"-": ["13"]},
    "TRVLLISSO": {"-": ["51"]},
    "TRVNCO": {"-": ["51"]},
    "TRVNCVS": {"1": ["12"], "2": ["21"]},
    "TRVSATILIS": {"-": ["24"]},
    "TRVSO": {"-": ["51"]},
    "TRVTINA": {"-": ["11"]},
    "TRVTINOR": {"-": ["5J"]},
    "TRVX": {"-": ["25"]},
    "TV": {"-": ["41"]},
    "TVBA": {"-": ["11"]},
    "TVBER": {"-": ["13"]},
    "TVBICEN": {"-": ["13"]},
    "TVBVLA": {"-": ["11"]},
    "TVBVLVS": {"-": ["12"], "N": ["12"]},
    "TVBVS": {"-": ["12"]},
    "TVCCETVM": {"-": ["12"]},
    "TVDITO": {"-": ["51"]},
    "TVEOR": {"-": ["5K"]},
    "TVGVRIVM": {"-": ["12"]},
    "TVI": {"-": ["42", "12"]},
    "TVLLIA": {"N": ["11"]},
    "TVLLIVS": {"N": ["12"]},
    "TVLLVS": {"N": ["12"]},
    "TVM": {"-": ["60"]},
    "TVMEFACIO": {"-": ["55"]},
    "TVMEO": {"-": ["52"]},
    "TVMESCO": {"-": ["53"]},
    "TVMIDVS": {"-": ["21"]},
    "TVMOR": {"-": ["13"]},
    "TVMVLO": {"-": ["51"]},
    "TVMVLOSVS": {"-": ["21"]},
    "TVMVLTVARIVS": {"-": ["21"]},
    "TVMVLTVOR": {"-": ["5J"]},
    "TVMVLTVOSE": {"-": ["6&"]},
    "TVMVLTVOSVS": {"-": ["21"]},
    "TVMVLTVS": {"-": ["14"]},
    "TVMVLVS": {"-": ["12"]},
    "TVNC": {"-": ["60"]},
    "TVNDO": {"-": ["53"]},
    "TVNICA": {"-": ["11"]},
    "TVNICATVS": {"-": ["21"]},
    "TVOR": {"-": ["5L"]},
    "TVRBA": {"-": ["11"]},
    "TVRBATE": {"-": ["60"]},
    "TVRBATOR": {"-": ["13"]},
    "TVRBATVS": {"-": ["21"]},
    "TVRBEN": {"-": ["13"]},
    "TVRBIDE": {"-": ["60"]},
    "TVRBIDVM": {"-": ["60"]},
    "TVRBIDVS": {"-": ["21"]},
    "TVRBO": {"1": ["13"], "2": ["51"], "N": ["11"]},
    "TVRBVLENTVS": {"-": ["21"]},
    "TVRDA": {"-": ["11"]},
    "TVRDVS": {"-": ["12"]},
    "TVREVS": {"-": ["21"]},
    "TVRGEO": {"-": ["52"]},
    "TVRGESCO": {"-": ["53"]},
    "TVRGIDVLVS": {"-": ["21"]},
    "TVRGIDVS": {"-": ["21"]},
    "TVRIBVLVM": {"-": ["12"]},
    "TVRICREMVS": {"-": ["21"]},
    "TVRIFER": {"-": ["21"]},
    "TVRMA": {"-": ["11"]},
    "TVRMATIM": {"-": ["60"]},
    "TVRNVS": {"N": ["12"]},
    "TVRPE": {"1": ["13"], "2": ["60"]},
    "TVRPICVLVS": {"-": ["21"]},
    "TVRPIO": {"N": ["13"]},
    "TVRPIS": {"-": ["24"]},
    "TVRPITER": {"-": ["60"]},
    "TVRPITVDO": {"-": ["13"]},
    "TVRPO": {"-": ["51"]},
    "TVRRICVLA": {"-": ["11"]},
    "TVRRIGER": {"-": ["21"]},
    "TVRRIS": {"-": ["13"]},
    "TVRRITVS": {"-": ["21"]},
    "TVRTVR": {"-": ["13"]},
    "TVRTVRILLA": {"-": ["11"]},
    "TVRVNDA": {"-": ["11"]},
    "TVS": {"-": ["13"]},
    "TVSCI": {"N": ["12"]},
    "TVSCVLANVM": {"N": ["12"]},
    "TVSCVLANVS": {"N": ["21"], "A": ["21"]},
    "TVSCVLVM": {"N": ["12"], "-": ["12"]},
    "TVSCVS": {"N": ["21"], "A": ["21"]},
    "TVSSIO": {"-": ["54"]},
    "TVSSIS": {"-": ["13"]},
    "TVTAMEN": {"-": ["13"]},
    "TVTAMENTVM": {"-": ["12"]},
    "TVTE": {"-": ["60"]},
    "TVTELA": {"-": ["11"], "N": ["11"]},
    "TVTO": {"1": ["51"], "2": ["60"], "-": ["6-"]},
    "TVTOR": {"1": ["13"], "2": ["5J"], "N": ["11"]},
    "TVTVM": {"-": ["12"]},
    "TVTVS": {"-": ["21"]},
    "TVVM": {"-": ["12"]},
    "TVVS": {"-": ["42"]},
    "TYCHA": {"N": ["11"]},
    "TYMPANVM": {"-": ["12"]},
    "TYNDARIS": {"N": ["17"]},
    "TYNDARITANI": {"N": ["12"]},
    "TYNDARITANVS": {"N": ["21"], "A": ["21"]},
    "TYRACINVS": {"N": ["12"]},
    "TYRANNICE": {"-": ["60"]},
    "TYRANNICIDA": {"-": ["11"]},
    "TYRANNICVS": {"-": ["21"]},
    "TYRANNIS": {"-": ["17"]},
    "TYRANNVS": {"-": ["12"]},
    "TYRIVS": {"N": ["21"], "A": ["21"]},
    "TYROS": {"N": ["17"]},
    "VACATIO": {"-": ["13"]},
    "VACCA": {"-": ["11"], "N": ["11"]},
    "VACCINIVM": {"-": ["12"]},
    "VACEFIO": {"-": ["56"]},
    "VACILLO": {"-": ["51"]},
    "VACO": {"-": ["51"]},
    "VACVEFACIO": {"-": ["55"]},
    "VACVITAS": {"-": ["13"]},
    "VACVO": {"1": ["51"], "2": ["60"]},
    "VACVVM": {"-": ["12"]},
    "VACVVS": {"-": ["21"]},
    "VADIMONIVM": {"-": ["12"]},
    "VADO": {"1": ["51"], "2": ["53"]},
    "VADOR": {"-": ["5J"]},
    "VADOSVS": {"-": ["21"]},
    "VADVM": {"-": ["12"], "N": ["12"]},
    "VAE": {"-": ["90"]},
    "VAFER": {"-": ["21"]},
    "VAFRE": {"-": ["60"]},
    "VAFRITIA": {"-": ["11"]},
    "VAGABVNDVS": {"-": ["21"]},
    "VAGATIO": {"-": ["13"]},
    "VAGE": {"-": ["60"]},
    "VAGINA": {"-": ["11"]},
    "VAGIO": {"-": ["54"]},
    "VAGITVS": {"-": ["14"]},
    "VAGOR": {"1": ["13"], "2": ["5J"]},
    "VAGVS": {"-": ["21"]},
    "VALDE": {"-": ["60"]},
    "VALEDICO": {"2": ["53"], "-": ["53"]},
    "VALENS": {"-": ["25"], "N": ["13"]},
    "VALENTINI": {"N": ["12"]},
    "VALENTIVS": {"N": ["12"]},
    "VALEO": {"-": ["52"]},
    "VALERIVS": {"N": ["12"], "A": ["21"]},
    "VALESCO": {"-": ["53"]},
    "VALETVDINARIVM": {"-": ["12"]},
    "VALETVDINARIVS": {"1": ["12"], "2": ["21"]},
    "VALETVDO": {"-": ["13"]},
    "VALIDE": {"-": ["6&"]},
    "VALIDVS": {"-": ["21"]},
    "VALLIS": {"-": ["13"]},
    "VALLO": {"-": ["51"]},
    "VALLVM": {"-": ["12"]},
    "VALLVS": {"1": ["12"], "2": ["12"]},
    "VALOR": {"-": ["13"]},
    "VALVA": {"-": ["11"]},
    "VALVAE": {"-": ["11"]},
    "VALVATVS": {"-": ["21"]},
    "VANA": {"-": ["12"]},
    "VANE": {"-": ["60"]},
    "VANESCO": {"-": ["53"]},
    "VANILOQVENTIA": {"-": ["11"]},
    "VANILOQVVS": {"-": ["21"]},
    "VANITAS": {"-": ["13"]},
    "VANNVS": {"-": ["12"]},
    "VANVM": {"-": ["12"]},
    "VANVS": {"-": ["21"]},
    "VAPIDVS": {"-": ["21"]},
    "VAPOR": {"-": ["13"]},
    "VAPORO": {"-": ["51"]},
    "VAPPA": {"-": ["11"]},
    "VAPVLO": {"-": ["51"]},
    "VARA": {"-": ["11"]},
    "VARIANTIA": {"-": ["11"]},
    "VARIATVS": {"-": ["21"]},
    "VARICOSVS": {"-": ["21"]},
    "VARICVS": {"-": ["21"]},
    "VARIE": {"-": ["60"]},
    "VARIETAS": {"-": ["13"]},
    "VARIO": {"-": ["51"]},
    "VARIVM": {"-": ["12"]},
    "VARIVS": {"-": ["21"], "N": ["12"], "A": ["21"]},
    "VARIX": {"-": ["13"]},
    "VARRO": {"N": ["13"]},
    "VARVS": {"1": ["12"], "2": ["21"], "N": ["12"]},
    "VAS": {"1": ["13"], "2": ["16"]},
    "VASARIVM": {"-": ["12"]},
    "VASCVLARIVS": {"-": ["12"]},
    "VASCVLVM": {"-": ["12"]},
    "VASCVS": {"-": ["21"]},
    "VASTATIO": {"-": ["13"]},
    "VASTATOR": {"-": ["13"]},
    "VASTATRIX": {"-": ["13"]},
    "VASTE": {"-": ["60"]},
    "VASTITAS": {"-": ["13"]},
    "VASTITVDO": {"-": ["13"]},
    "VASTO": {"-": ["51"]},
    "VASTVS": {"-": ["21"]},
    "VATES": {"-": ["13"]},
    "VATICINATIO": {"-": ["13"]},
    "VATICINOR": {"-": ["5J"]},
    "VATILLVM": {"-": ["12"]},
    "VBER": {"1": ["13"], "2": ["26"]},
    "VBERE": {"-": ["60"]},
    "VBERO": {"-": ["51"]},
    "VBERTAS": {"-": ["13"]},
    "VBERTIM": {"-": ["60"]},
    "VBI": {"1": ["66"], "2": ["67"], "3": ["82"]},
    "VBICVMQVE": {"1": ["66"], "2": ["60"]},
    "VBILIBET": {"-": ["60"]},
    "VBINAM": {"1": ["66"], "2": ["67"]},
    "VBIQVE": {"-": ["60"]},
    "VBIVIS": {"-": ["60"]},
    "VDVS": {"-": ["21"]},
    "VE": {"-": ["81"]},
    "VECORDIA": {"-": ["11"]},
    "VECORS": {"-": ["25"]},
    "VECTATIO": {"-": ["13"]},
    "VECTIARIVS": {"-": ["12"]},
    "VECTICVLARIVS": {"-": ["21"]},
    "VECTIGAL": {"-": ["13"]},
    "VECTIGALIS": {"-": ["24"]},
    "VECTIS": {"-": ["13"]},
    "VECTITO": {"-": ["51"]},
    "VECTO": {"-": ["51"]},
    "VECTOR": {"1": ["13"], "2": ["5J", "13"]},
    "VECTORIVS": {"-": ["21"]},
    "VECTVRA": {"-": ["11"]},
    "VEDIVS": {"N": ["12"]},
    "VEGEO": {"-": ["52"]},
    "VEGETO": {"-": ["51"]},
    "VEGETVS": {"-": ["21"]},
    "VEGRANDIS": {"-": ["24"]},
    "VEHEMENS": {"-": ["25"]},
    "VEHEMENTER": {"-": ["60"]},
    "VEHEMENTIA": {"-": ["11"]},
    "VEHES": {"-": ["13"]},
    "VEHICVLVM": {"-": ["12"]},
    "VEHO": {"-": ["53"]},
    "VEHOR": {"-": ["5L"]},
    "VEIENS": {"N": ["25"], "A": ["25"]},
    "VEIENTES": {"N": ["13"]},
    "VEII": {"N": ["12"]},
    "VEL": {"1": ["60"], "2": ["81"]},
    "VELAMEN": {"-": ["13"]},
    "VELAMENTVM": {"-": ["12"]},
    "VELARIVM": {"-": ["12"]},
    "VELES": {"-": ["13"]},
    "VELIA": {"N": ["11"]},
    "VELIFER": {"-": ["21"]},
    "VELIFICO": {"-": ["51"]},
    "VELIFICOR": {"-": ["5J"]},
    "VELITARIS": {"-": ["24"]},
    "VELITERNVS": {"N": ["21"]},
    "VELITRAE": {"N": ["11"]},
    "VELIVOLVS": {"-": ["21"]},
    "VELLICATIO": {"-": ["13"]},
    "VELLICO": {"-": ["51"]},
    "VELLO": {"-": ["53"]},
    "VELLVS": {"-": ["13"]},
    "VELO": {"-": ["51"]},
    "VELOCITAS": {"-": ["13"]},
    "VELOCITER": {"-": ["60"]},
    "VELOX": {"-": ["25"]},
    "VELVM": {"-": ["12"]},
    "VELVT": {"1": ["66"], "2": ["82"]},
    "VELVTSI": {"-": ["82"]},
    "VENA": {"-": ["11"]},
    "VENABVLVM": {"-": ["12"]},
    "VENALES": {"-": ["13"]},
    "VENALIA": {"-": ["13"]},
    "VENALIS": {"-": ["24"]},
    "VENANS": {"-": ["13"]},
    "VENATICVS": {"-": ["21"]},
    "VENATIO": {"-": ["13"]},
    "VENATOR": {"-": ["13"]},
    "VENATORIVS": {"-": ["21"]},
    "VENATRIX": {"-": ["13"]},
    "VENATVS": {"-": ["14"]},
    "VENDAX": {"-": ["25"]},
    "VENDIBILIS": {"-": ["24"]},
    "VENDITATIO": {"-": ["13"]},
    "VENDITIO": {"-": ["13"]},
    "VENDITO": {"-": ["51"]},
    "VENDITOR": {"-": ["13"]},
    "VENDITVM": {"-": ["12"]},
    "VENDO": {"-": ["53"]},
    "VENEFICA": {"-": ["11"]},
    "VENEFICIVM": {"-": ["12"]},
    "VENEFICVS": {"1": ["12"], "2": ["21"]},
    "VENENATVM": {"-": ["12"]},
    "VENENATVS": {"-": ["21"]},
    "VENENO": {"-": ["51"]},
    "VENENVM": {"-": ["12"]},
    "VENEO": {"-": ["56"]},
    "VENERABILIS": {"-": ["24"]},
    "VENERABVNDVS": {"-": ["21"]},
    "VENERANDVS": {"-": ["21"]},
    "VENERATIO": {"-": ["13"]},
    "VENEREVS": {"N": ["21"], "A": ["21"], "-": ["21"]},
    "VENERIA": {"-": ["11"]},
    "VENEROR": {"-": ["5J"]},
    "VENETI": {"1N": ["12"], "N": ["12"]},
    "VENIA": {"-": ["11"]},
    "VENIO": {"-": ["54"]},
    "VENNONIVS": {"N": ["12"]},
    "VENOR": {"-": ["5J"]},
    "VENOSVS": {"-": ["21"]},
    "VENTER": {"-": ["13"]},
    "VENTILO": {"-": ["51"]},
    "VENTITO": {"-": ["51"]},
    "VENTO": {"-": ["51"]},
    "VENTOSVS": {"-": ["21"]},
    "VENTRICVLVM": {"-": ["12"]},
    "VENTRICVLVS": {"-": ["12"]},
    "VENTVRVM": {"-": ["12"]},
    "VENTVS": {"-": ["12"]},
    "VENVCVLVS": {"-": ["21"]},
    "VENVLEIVS": {"N": ["12"]},
    "VENVM": {"-": ["12"]},
    "VENVNDO": {"-": ["51"]},
    "VENVS": {"1": ["13"], "2": ["14"], "N": ["13"]},
    "VENVSTAS": {"-": ["13"], "N": ["13"]},
    "VENVSTE": {"-": ["60"]},
    "VENVSTVS": {"-": ["21"]},
    "VEPALLIDVS": {"-": ["21"]},
    "VEPRES": {"-": ["13"]},
    "VER": {"-": ["13"]},
    "VERATRVM": {"-": ["12"]},
    "VERAX": {"-": ["25"], "N": ["13"]},
    "VERBENA": {"-": ["11"]},
    "VERBER": {"-": ["13"]},
    "VERBERATVS": {"-": ["14"]},
    "VERBERO": {"1": ["13"], "2": ["51"]},
    "VERBOSVS": {"-": ["21"]},
    "VERBVM": {"-": ["12"]},
    "VERE": {"-": ["60"]},
    "VERECVNDE": {"-": ["60"]},
    "VERECVNDIA": {"-": ["11"]},
    "VERECVNDVS": {"-": ["21"]},
    "VERENDVS": {"-": ["21"]},
    "VEREOR": {"-": ["5K"]},
    "VERGENS": {"-": ["25"]},
    "VERGO": {"-": ["53"]},
    "VERGOBRETVS": {"-": ["12"]},
    "VERIDICVS": {"-": ["21"]},
    "VERISIMILIS": {"-": ["24"]},
    "VERISIMILITVDO": {"-": ["13"]},
    "VERITAS": {"-": ["13"], "N": ["13"]},
    "VERMICVLVS": {"-": ["12"]},
    "VERMINA": {"-": ["13"]},
    "VERMINATIO": {"-": ["13"]},
    "VERMINO": {"-": ["51"]},
    "VERMIS": {"-": ["13"]},
    "VERNA": {"-": ["11"]},
    "VERNACVLA": {"-": ["11"]},
    "VERNACVLVS": {"1": ["12"], "2": ["21"]},
    "VERNILITAS": {"-": ["13"]},
    "VERNILITER": {"-": ["60"]},
    "VERNO": {"1": ["51"], "2": ["60"]},
    "VERNVLA": {"-": ["11"]},
    "VERNVM": {"-": ["12"]},
    "VERNVS": {"-": ["21"]},
    "VERO": {"1": ["13"], "2": ["51"], "3": ["60"], "4": ["81"]},
    "VERPA": {"-": ["11"]},
    "VERPVS": {"-": ["12"]},
    "VERRES": {"-": ["13"], "N": ["13"]},
    "VERRIA": {"N": ["12"]},
    "VERRINVS": {"N": ["21"], "-": ["21"], "A": ["21"]},
    "VERRIVS": {"A": ["21"]},
    "VERRO": {"-": ["53"]},
    "VERRVCA": {"-": ["11"]},
    "VERRVCIVS": {"N": ["12"]},
    "VERRVCOSVS": {"-": ["21"], "N": ["12"]},
    "VERSABILIS": {"-": ["24"]},
    "VERSABVNDVS": {"-": ["21"]},
    "VERSATILIS": {"-": ["24"]},
    "VERSATIO": {"-": ["13"]},
    "VERSATVS": {"-": ["21"]},
    "VERSICOLOR": {"-": ["25"]},
    "VERSICVLVS": {"-": ["12"]},
    "VERSO": {"-": ["51"]},
    "VERSOR": {"-": ["5J"]},
    "VERSVRA": {"-": ["11"]},
    "VERSVS": {"1": ["14"], "2": ["21"], "3": ["60"], "4": ["70"]},
    "VERSVTVS": {"-": ["21"]},
    "VERTEBRA": {"-": ["11"]},
    "VERTENS": {"-": ["25"]},
    "VERTEX": {"-": ["13"]},
    "VERTICVLA": {"-": ["11"]},
    "VERTIGO": {"-": ["13"]},
    "VERTO": {"-": ["53"]},
    "VERTVMNVS": {"N": ["12"]},
    "VERV": {"-": ["14"]},
    "VERVACTVM": {"-": ["12"]},
    "VERVEX": {"-": ["13"]},
    "VERVM": {"1": ["12"], "2": ["12"], "3": ["60"], "4": ["81"], "-": ["12"]},
    "VERVMTAMEN": {"-": ["60"]},
    "VERVMVERO": {"-": ["60"]},
    "VERVS": {"-": ["21"], "N": ["12"]},
    "VERVTVM": {"-": ["12"]},
    "VERVTVS": {"-": ["21"]},
    "VESANIA": {"-": ["11"]},
    "VESANIO": {"-": ["54"]},
    "VESANVS": {"-": ["21"]},
    "VESCOR": {"-": ["5L"]},
    "VESCVS": {"-": ["21"]},
    "VESICA": {"-": ["11"]},
    "VESICVLA": {"-": ["11"]},
    "VESPER": {"-": ["16"], "N": ["12"]},
    "VESPERA": {"-": ["11"]},
    "VESPERASCO": {"-": ["53"]},
    "VESPERTINVS": {"-": ["21"]},
    "VESTA": {"N": ["11"]},
    "VESTALIA": {"-": ["13"]},
    "VESTALIS": {"1": ["13"], "2": ["24"], "A": ["24"], "N": ["13"]},
    "VESTER": {"-": ["42"]},
    "VESTIARIVM": {"-": ["12"]},
    "VESTIARIVS": {"-": ["21"], "2": ["21"], "1": ["12"]},
    "VESTIBVLVM": {"-": ["12"]},
    "VESTIGATOR": {"-": ["13"]},
    "VESTIGIVM": {"-": ["12"]},
    "VESTIGO": {"-": ["51"]},
    "VESTIMENTVM": {"-": ["12"]},
    "VESTIO": {"-": ["54"]},
    "VESTIS": {"-": ["13"]},
    "VESTITVS": {"1": ["14"], "2": ["21"]},
    "VESTRI": {"-": ["42"]},
    "VETELIA": {"N": ["11"]},
    "VETERA": {"-": ["13"], "N": ["11"]},
    "VETERANVS": {"1": ["12"], "2": ["21"]},
    "VETERARIVM": {"-": ["12"]},
    "VETERATOR": {"-": ["13"]},
    "VETERATORIVS": {"-": ["21"]},
    "VETERES": {"-": ["13"]},
    "VETERINVS": {"-": ["21"]},
    "VETERNOSVS": {"-": ["21"]},
    "VETERNVS": {"1": ["12"], "2": ["21"]},
    "VETITVM": {"-": ["12"]},
    "VETO": {"-": ["51"]},
    "VETTIVS": {"N": ["12"]},
    "VETVLA": {"-": ["11"]},
    "VETVLVS": {"1": ["12"], "2": ["21"]},
    "VETVRIA": {"N": ["11"]},
    "VETVRIVS": {"N": ["12"]},
    "VETVS": {"-": ["22"], "N": ["22"], "A": ["21"]},
    "VETVSTAS": {"-": ["13"]},
    "VETVSTVS": {"-": ["21"]},
    "VEXAMEN": {"-": ["13"]},
    "VEXATIO": {"-": ["13"]},
    "VEXATOR": {"-": ["13"]},
    "VEXILLARIVS": {"-": ["12"]},
    "VEXILLVM": {"-": ["12"]},
    "VEXO": {"-": ["51"]},
    "VIA": {"-": ["11"], "N": ["11"]},
    "VIATICVM": {"-": ["12"]},
    "VIATOR": {"-": ["13"]},
    "VIATORIVS": {"-": ["21"]},
    "VIBEX": {"-": ["13"]},
    "VIBIVS": {"N": ["12"]},
    "VIBO": {"N": ["13"]},
    "VIBRANS": {"-": ["25", "35"], "V\u00e9h\u00e9ment": ["25"]},
    "VIBRATVS": {"1": ["14"], "2": ["21"]},
    "VIBRO": {"-": ["51"]},
    "VIBVRNVM": {"-": ["12"]},
    "VICA": {"N": ["11"]},
    "VICARIVS": {"1": ["12"], "2": ["21"]},
    "VICATIM": {"-": ["60"]},
    "VICECOMES": {"-": ["13"]},
    "VICENARIVS": {"-": ["21"]},
    "VICENVS": {"-": ["33"]},
    "VICESIMA": {"-": ["11"]},
    "VICESIMANVS": {"-": ["12"]},
    "VICESIMVS": {"-": ["32"]},
    "VICIA": {"1": ["11"], "2": ["11"]},
    "VICIES": {"-": ["36"]},
    "VICINA": {"-": ["11"]},
    "VICINALIS": {"-": ["24"]},
    "VICINIA": {"-": ["11"]},
    "VICINITAS": {"-": ["13"]},
    "VICINVM": {"-": ["12"]},
    "VICINVS": {"1": ["12"], "2": ["21"]},
    "VICIS": {"-": ["13"]},
    "VICISSIM": {"-": ["60"]},
    "VICISSITVDO": {"-": ["13"]},
    "VICTIMA": {"-": ["11"]},
    "VICTOR": {"-": ["13"], "N": ["13"]},
    "VICTORIA": {"-": ["11"], "N": ["11"]},
    "VICTORIATVS": {"-": ["12"]},
    "VICTRIX": {"-": ["13"], "2": ["22"], "1": ["21"]},
    "VICTVS": {"-": ["14"]},
    "VICVLVS": {"-": ["12"]},
    "VICVS": {"-": ["12"], "N": ["12"]},
    "VIDELICET": {"-": ["60"]},
    "VIDENS": {"-": ["25"]},
    "VIDEO": {"-": ["52"]},
    "VIDVA": {"-": ["11"]},
    "VIDVERTAS": {"-": ["13"]},
    "VIDVITAS": {"-": ["13"]},
    "VIDVO": {"-": ["51"]},
    "VIDVVS": {"-": ["21"]},
    "VIEO": {"-": ["52"]},
    "VIESCO": {"-": ["53"]},
    "VIETVS": {"-": ["21"]},
    "VIGENS": {"-": ["25"]},
    "VIGEO": {"-": ["52"]},
    "VIGESCO": {"-": ["53"]},
    "VIGIL": {"-": ["25"], "2": ["13"], "1": ["12"]},
    "VIGILANS": {"-": ["25"]},
    "VIGILANTER": {"-": ["60"]},
    "VIGILANTIA": {"-": ["11"]},
    "VIGILES": {"-": ["13"]},
    "VIGILIA": {"-": ["11"]},
    "VIGILIARIVM": {"-": ["12"]},
    "VIGILO": {"-": ["51"]},
    "VIGINTI": {"-": ["31"]},
    "VIGINTIVIRATVS": {"-": ["14"]},
    "VIGOR": {"-": ["13"]},
    "VILIS": {"-": ["24"]},
    "VILITAS": {"-": ["13"]},
    "VILLA": {"-": ["11"]},
    "VILLICA": {"-": ["11"]},
    "VILLICO": {"2": ["51"]},
    "VILLICVS": {"1": ["12"], "2": ["21"]},
    "VILLOSVS": {"-": ["21"]},
    "VILLVLA": {"-": ["11"]},
    "VILLVS": {"-": ["12"]},
    "VIMEN": {"-": ["13"]},
    "VIMENTVM": {"-": ["12"]},
    "VIMINALIS": {"-": ["24"], "N": ["24"]},
    "VIMINEVS": {"-": ["21"]},
    "VINACEA": {"-": ["11"]},
    "VINACEVS": {"-": ["12"]},
    "VINARIVM": {"-": ["12"]},
    "VINARIVS": {"2": ["21"], "1": ["12"]},
    "VINCIO": {"-": ["54"]},
    "VINCO": {"-": ["53"]},
    "VINCTVS": {"-": ["12"]},
    "VINCVLVM": {"-": ["12"]},
    "VINDEMIA": {"-": ["11"]},
    "VINDEMIATOR": {"-": ["13"]},
    "VINDEX": {"-": ["13"], "N": ["13"]},
    "VINDICIAE": {"-": ["11"]},
    "VINDICIVS": {"N": ["12"]},
    "VINDICO": {"-": ["51"]},
    "VINDICTA": {"-": ["11"]},
    "VINEA": {"-": ["11"]},
    "VINEATICVS": {"-": ["21"]},
    "VINETVM": {"-": ["12"]},
    "VINITOR": {"-": ["13"]},
    "VINOLENTIA": {"-": ["11"]},
    "VINOLENTVS": {"1": ["12"], "2": ["21"]},
    "VINOSVS": {"-": ["21"]},
    "VINVM": {"-": ["12"]},
    "VIOLA": {"-": ["11"]},
    "VIOLABILIS": {"-": ["24"]},
    "VIOLACEVS": {"-": ["21"]},
    "VIOLARIVM": {"-": ["12"]},
    "VIOLATIO": {"-": ["13"]},
    "VIOLATOR": {"-": ["13"]},
    "VIOLENS": {"-": ["25"]},
    "VIOLENTER": {"-": ["60"]},
    "VIOLENTIA": {"-": ["11"]},
    "VIOLENTVS": {"-": ["21"]},
    "VIOLO": {"-": ["51"]},
    "VIPERA": {"-": ["11"]},
    "VIPEREVS": {"-": ["21"]},
    "VIPERINVS": {"-": ["21"]},
    "VIR": {"-": ["12"]},
    "VIRAGO": {"-": ["13"]},
    "VIRENS": {"-": ["25"]},
    "VIRENTIA": {"-": ["13"]},
    "VIREO": {"1": ["13"], "2": ["52"]},
    "VIRESCO": {"1": ["53"], "2": ["53"], "-": ["53"]},
    "VIRETVM": {"-": ["12"]},
    "VIRGA": {"-": ["11"]},
    "VIRGATVS": {"-": ["21"]},
    "VIRGEVS": {"-": ["21"]},
    "VIRGILIAE": {"-": ["11"]},
    "VIRGILIVS": {"N": ["12"]},
    "VIRGINALIS": {"-": ["24"]},
    "VIRGINEVS": {"-": ["21"], "A": ["21"]},
    "VIRGINITAS": {"-": ["13"]},
    "VIRGINIVS": {"N": ["12"]},
    "VIRGO": {"-": ["13"], "N": ["13"]},
    "VIRGVLA": {"-": ["11"]},
    "VIRGVLTVM": {"-": ["12"]},
    "VIRGVNCVLA": {"-": ["11"]},
    "VIRIDANS": {"-": ["25"]},
    "VIRIDE": {"1": ["13"], "2": ["60"], "-": ["6-"]},
    "VIRIDIS": {"-": ["24"]},
    "VIRIDITAS": {"-": ["13"]},
    "VIRIDO": {"-": ["51"]},
    "VIRILIS": {"-": ["24"], "A": ["24"]},
    "VIRILITAS": {"-": ["13"]},
    "VIRILITER": {"-": ["60"]},
    "VIRITIM": {"-": ["60"]},
    "VIROSVS": {"-": ["21"]},
    "VIRTVS": {"-": ["13"], "N": ["13"]},
    "VIRVS": {"-": ["16"]},
    "VIS": {"-": ["13"]},
    "VISCERATIO": {"-": ["13"]},
    "VISCO": {"-": ["51"]},
    "VISCVM": {"-": ["12"]},
    "VISCVS": {"1": ["12"], "2": ["13"], "N": ["12"]},
    "VISITATIO": {"-": ["13"]},
    "VISITO": {"-": ["51"]},
    "VISO": {"-": ["53"]},
    "VISVM": {"-": ["12"]},
    "VISVS": {"-": ["14"]},
    "VITA": {"-": ["11"]},
    "VITABVNDVS": {"-": ["21"]},
    "VITALIA": {"-": ["13"]},
    "VITALIS": {"-": ["24"]},
    "VITALITER": {"-": ["60"]},
    "VITELLIVS": {"N": ["12"]},
    "VITELLVS": {"-": ["12"]},
    "VITEVS": {"-": ["21"]},
    "VITEX": {"-": ["13"]},
    "VITIARIVM": {"-": ["12"]},
    "VITICVLA": {"-": ["11"]},
    "VITIGENVS": {"-": ["21"]},
    "VITIGINEVS": {"-": ["21"]},
    "VITILIS": {"-": ["24"]},
    "VITILITIGO": {"-": ["51"]},
    "VITIO": {"1": ["51"], "2": ["54"]},
    "VITIOSITAS": {"-": ["13"]},
    "VITIOSVS": {"-": ["21"]},
    "VITIS": {"-": ["13"]},
    "VITIVM": {"-": ["12"]},
    "VITO": {"-": ["51"]},
    "VITRARIVS": {"1": ["12"], "2": ["21"]},
    "VITREVM": {"-": ["12"]},
    "VITREVS": {"-": ["21"]},
    "VITRICVS": {"-": ["12"]},
    "VITRVM": {"-": ["12"]},
    "VITTA": {"-": ["11"]},
    "VITTATVS": {"-": ["21"]},
    "VITVLA": {"-": ["11"]},
    "VITVLVS": {"-": ["12"]},
    "VITVPERATIO": {"-": ["13"]},
    "VITVPERIVM": {"-": ["12"]},
    "VITVPERO": {"1": ["13"], "2": ["51"]},
    "VIVARIVM": {"-": ["12"]},
    "VIVATVS": {"-": ["21"]},
    "VIVAX": {"-": ["25"]},
    "VIVENS": {"-": ["13"]},
    "VIVESCO": {"-": ["53"]},
    "VIVIDVS": {"-": ["21"]},
    "VIVIRADIX": {"-": ["13"]},
    "VIVO": {"-": ["53"]},
    "VIVVM": {"-": ["12"]},
    "VIVVS": {"1": ["12"], "2": ["21"]},
    "VIX": {"-": ["60"]},
    "VIXDVM": {"-": ["60"]},
    "VLCERATIO": {"-": ["13"]},
    "VLCERO": {"-": ["51"]},
    "VLCEROSVS": {"-": ["21"]},
    "VLCISCOR": {"-": ["5L"]},
    "VLCVS": {"-": ["13"]},
    "VLCVSCVLVM": {"-": ["12"]},
    "VLIGO": {"-": ["13"]},
    "VLIXES": {"N": ["17"]},
    "VLIXEVS": {"N": ["21"]},
    "VLLVS": {"-": ["48"]},
    "VLMEVS": {"-": ["21"]},
    "VLMVS": {"-": ["12"]},
    "VLNA": {"-": ["11"]},
    "VLPICVM": {"-": ["12"]},
    "VLTERIOR": {"-": ["2&"], "A": ["2&"]},
    "VLTERIORA": {"-": ["13"]},
    "VLTIMVM": {"1": ["12"], "2": ["6-", "12"]},
    "VLTIMVS": {"-": ["2-", "21"]},
    "VLTIO": {"-": ["13"], "N": ["13"]},
    "VLTOR": {"-": ["13"], "N": ["13"]},
    "VLTRA": {"1": ["60"], "2": ["70"]},
    "VLTRIX": {"-": ["13"]},
    "VLTRO": {"-": ["60"]},
    "VLVA": {"-": ["11"]},
    "VLVLA": {"-": ["11"]},
    "VLVLATVS": {"-": ["14"]},
    "VLVLO": {"-": ["51"]},
    "VMBELLA": {"-": ["11"]},
    "VMBILICVS": {"-": ["12"]},
    "VMBO": {"-": ["13"]},
    "VMBRA": {"-": ["11"]},
    "VMBRACVLVM": {"-": ["12"]},
    "VMBRATICVS": {"-": ["21"]},
    "VMBRIFER": {"-": ["21"]},
    "VMBRO": {"-": ["51"], "N": ["11"]},
    "VMBROSA": {"-": ["12"]},
    "VMBROSVS": {"-": ["21"]},
    "VMQVAM": {"-": ["60"]},
    "VNA": {"-": ["60"]},
    "VNANIMVS": {"-": ["21"]},
    "VNCIA": {"-": ["11"]},
    "VNCIARIVS": {"-": ["21"]},
    "VNCINVS": {"-": ["12"]},
    "VNCIOLA": {"-": ["11"]},
    "VNCTIO": {"-": ["13"]},
    "VNCTOR": {"-": ["13"]},
    "VNCTVM": {"-": ["12"]},
    "VNCTVS": {"1": ["14"], "2": ["21"]},
    "VNCVS": {"1": ["12"], "2": ["21"]},
    "VNDA": {"-": ["11"]},
    "VNDE": {"1": ["66"], "2": ["67"]},
    "VNDECIM": {"-": ["31"]},
    "VNDECIMVS": {"-": ["32"]},
    "VNDECVMQVE": {"1": ["66"], "2": ["60"]},
    "VNDENI": {"-": ["33"]},
    "VNDEOCTOGINTA": {"-": ["31"]},
    "VNDEQVINQVAGESIMVS": {"-": ["31"]},
    "VNDEQVINQVAGINTA": {"-": ["31"]},
    "VNDETRICESIMVS": {"-": ["32"]},
    "VNDETRIGINTA": {"-": ["31"]},
    "VNDEVICESIMVS": {"-": ["32"]},
    "VNDEVIGINTI": {"-": ["31"]},
    "VNDIQVE": {"-": ["60"]},
    "VNDISONVS": {"-": ["21"]},
    "VNDO": {"-": ["51"]},
    "VNDOSVS": {"-": ["21"]},
    "VNETVICESIMANI": {"-": ["12"]},
    "VNETVICESIMVS": {"-": ["32"]},
    "VNGO": {"-": ["53"]},
    "VNGVEN": {"-": ["13"]},
    "VNGVENTARIA": {"-": ["11"]},
    "VNGVENTARIVM": {"-": ["12"]},
    "VNGVENTARIVS": {"1": ["12"], "2": ["21"]},
    "VNGVENTO": {"-": ["51"]},
    "VNGVENTVM": {"-": ["12"]},
    "VNGVICVLVS": {"-": ["12"]},
    "VNGVIS": {"-": ["13"]},
    "VNGVITO": {"-": ["51"]},
    "VNGVLA": {"-": ["11"]},
    "VNICE": {"-": ["60"]},
    "VNICVS": {"-": ["21"]},
    "VNIFORMIS": {"-": ["24"]},
    "VNIGENA": {"-": ["21"]},
    "VNIO": {"1": ["13"], "2": ["54"]},
    "VNITAS": {"-": ["13"]},
    "VNITER": {"-": ["60"]},
    "VNIVERSALIS": {"-": ["24"]},
    "VNIVERSE": {"-": ["60"]},
    "VNIVERSI": {"-": ["12"]},
    "VNIVERSITAS": {"-": ["13"]},
    "VNIVERSVM": {"-": ["12"]},
    "VNIVERSVS": {"-": ["21"]},
    "VNVM": {"-": ["12"]},
    "VNVS": {"-": ["31"]},
    "VNVSQVISQVE": {"-": ["48"]},
    "VNVSQVISQVIS": {"-": ["48"]},
    "VOCABVLVM": {"-": ["12"]},
    "VOCALES": {"-": ["13"]},
    "VOCALIS": {"1": ["13"], "2": ["24"]},
    "VOCAMEN": {"-": ["13"]},
    "VOCATIO": {"-": ["13"]},
    "VOCATOR": {"-": ["13"]},
    "VOCIFERATIO": {"-": ["13"]},
    "VOCIFEROR": {"-": ["5J"]},
    "VOCITO": {"-": ["51"]},
    "VOCO": {"-": ["51"]},
    "VOCONIVS": {"A": ["21"], "N": ["12"]},
    "VOCVLA": {"-": ["11"], "N": ["11"]},
    "VOLANS": {"-": ["13"]},
    "VOLANTES": {"-": ["13"]},
    "VOLATICVS": {"-": ["21"]},
    "VOLATILIS": {"-": ["24"]},
    "VOLATVS": {"-": ["14"]},
    "VOLEMVS": {"-": ["21"]},
    "VOLENS": {"-": ["25"]},
    "VOLERO": {"N": ["13"]},
    "VOLESVS": {"N": ["12"]},
    "VOLITO": {"-": ["51"]},
    "VOLO": {"1": ["13"], "2": ["51"], "3": ["56"]},
    "VOLSCI": {"N": ["12"]},
    "VOLSCVS": {"N": ["21"], "A": ["21"]},
    "VOLTEIVS": {"N": ["12"]},
    "VOLVBILIS": {"-": ["24"]},
    "VOLVBILITAS": {"-": ["13"]},
    "VOLVCER": {"-": ["23"]},
    "VOLVCRIS": {"-": ["13"]},
    "VOLVMEN": {"-": ["13"]},
    "VOLVMNIA": {"N": ["11"]},
    "VOLVNTARII": {"-": ["12"]},
    "VOLVNTARIVS": {"-": ["21"]},
    "VOLVNTAS": {"-": ["13"]},
    "VOLVO": {"-": ["53"]},
    "VOLVPTARIVS": {"-": ["21"]},
    "VOLVPTAS": {"-": ["13"], "N": ["13"]},
    "VOLVSIVS": {"N": ["12"]},
    "VOLVTA": {"-": ["11"]},
    "VOLVTABRVM": {"-": ["12"]},
    "VOLVTATIO": {"-": ["13"]},
    "VOLVTO": {"-": ["51"]},
    "VOMER": {"-": ["13"]},
    "VOMICA": {"-": ["11"]},
    "VOMIS": {"-": ["13"]},
    "VOMITIO": {"-": ["13"]},
    "VOMITO": {"-": ["51"]},
    "VOMITOR": {"-": ["13"]},
    "VOMITORIVM": {"-": ["12"]},
    "VOMITORIVS": {"-": ["21"]},
    "VOMITVS": {"-": ["14"]},
    "VOMO": {"-": ["53"]},
    "VORAGINOSVS": {"-": ["21"]},
    "VORAGO": {"-": ["13"]},
    "VORAX": {"-": ["25"]},
    "VORO": {"-": ["51"]},
    "VOS": {"-": ["41"]},
    "VOTIVVM": {"-": ["12"]},
    "VOTIVVS": {"-": ["21"]},
    "VOTO": {"-": ["51"]},
    "VOTVM": {"-": ["12"]},
    "VOVEO": {"-": ["52"]},
    "VOX": {"-": ["13"]},
    "VRBANI": {"-": ["12"]},
    "VRBANITAS": {"-": ["13"]},
    "VRBANVS": {"-": ["21"]},
    "VRBIVS": {"N": ["21"]},
    "VRBS": {"-": ["13"], "N": ["13"]},
    "VRCEOLVS": {"-": ["12"]},
    "VRCEVS": {"-": ["12"]},
    "VREDO": {"-": ["13"]},
    "VRGENS": {"-": ["25"]},
    "VRGEO": {"-": ["52"]},
    "VRINA": {"-": ["11"]},
    "VRIVS": {"N": ["17"]},
    "VRNA": {"-": ["11"]},
    "VRNALIS": {"-": ["24"]},
    "VRO": {"-": ["53"]},
    "VRSA": {"-": ["11"], "N": ["11"]},
    "VRSVS": {"-": ["12"]},
    "VRTICA": {"-": ["11"]},
    "VRVCA": {"-": ["11"]},
    "VRVS": {"-": ["12"]},
    "VSIO": {"-": ["13"]},
    "VSITATVS": {"-": ["21"]},
    "VSPIAM": {"-": ["60"]},
    "VSQVAM": {"-": ["60"]},
    "VSQVE": {"-": ["60"]},
    "VSQVEQVAQVE": {"-": ["60"]},
    "VSTA": {"-": ["11"]},
    "VSTOR": {"-": ["13"]},
    "VSTVLO": {"-": ["51"]},
    "VSVALIS": {"-": ["24"]},
    "VSVCAPIO": {"1": ["13"], "2": ["55"]},
    "VSVRA": {"-": ["11"]},
    "VSVRPATIO": {"-": ["13"]},
    "VSVRPO": {"-": ["51"]},
    "VSVS": {"-": ["14"]},
    "VSVSFRVCTVS": {"-": ["14"]},
    "VSVVENIO": {"-": ["54"]},
    "VT": {"1": ["66"], "2": ["67"], "3": ["60"], "4": ["82"]},
    "VTCVMQVE": {"1": ["66"], "2": ["60"]},
    "VTENS": {"-": ["25"]},
    "VTENSILIA": {"-": ["13"]},
    "VTENSILIS": {"-": ["24"]},
    "VTER": {"1": ["13"], "2": ["46"], "3": ["47"], "4": ["48"]},
    "VTERCVMQVE": {"1": ["46"], "2": ["48"]},
    "VTERLIBET": {"-": ["48"]},
    "VTERQVE": {"-": ["48"]},
    "VTERVIS": {"-": ["48"]},
    "VTERVS": {"-": ["12"]},
    "VTICA": {"N": ["11"]},
    "VTICENSIS": {"N": ["24"], "A": ["24"]},
    "VTILE": {"-": ["13"]},
    "VTILIS": {"-": ["24"]},
    "VTILITAS": {"-": ["13"]},
    "VTILITER": {"-": ["60"]},
    "VTINAM": {"-": ["60"]},
    "VTIQVE": {"-": ["60"]},
    "VTOR": {"-": ["5L"]},
    "VTPOTE": {"-": ["60"]},
    "VTQVI": {"-": ["82"]},
    "VTRAQVE": {"-": ["60"]},
    "VTRICVS": {"-": ["12"]},
    "VTRIMQVE": {"-": ["60"]},
    "VTRIMQVESECVS": {"-": ["60"]},
    "VTRINDE": {"-": ["60"]},
    "VTRO": {"1": ["60"], "2": ["67"]},
    "VTROBI": {"-": ["60"]},
    "VTROBIQVE": {"-": ["60"]},
    "VTROQVE": {"-": ["60"]},
    "VTRVM": {"-": ["67"]},
    "VTRVMNE": {"-": ["67"]},
    "VVA": {"-": ["11"]},
    "VVESCO": {"-": ["53"]},
    "VVIDVLVS": {"-": ["21"]},
    "VVIDVS": {"-": ["21"]},
    "VVLCANVS": {"N": ["12"]},
    "VVLCATIVS": {"N": ["12"]},
    "VVLGARIA": {"-": ["13"]},
    "VVLGARIS": {"-": ["24"]},
    "VVLGATOR": {"-": ["13"]},
    "VVLGATVS": {"-": ["21"]},
    "VVLGIVAGVS": {"-": ["21"]},
    "VVLGO": {"1": ["51"], "2": ["60"]},
    "VVLGVS": {"-": ["12"]},
    "VVLNERABILIS": {"-": ["24"]},
    "VVLNERATIO": {"-": ["13"]},
    "VVLNERO": {"-": ["51"]},
    "VVLNIFICVS": {"-": ["21"]},
    "VVLNVS": {"-": ["13"]},
    "VVLPECVLA": {"-": ["11"]},
    "VVLPES": {"-": ["13"]},
    "VVLTVR": {"-": ["13"], "N": ["13"]},
    "VVLTVRIVS": {"-": ["12"]},
    "VVLTVRNVS": {"-": ["12"], "N": ["12"]},
    "VVLTVS": {"-": ["14"]},
    "VVLVA": {"-": ["11"]},
    "VXOR": {"-": ["13"]},
    "VXORIVS": {"-": ["21"]},
    "XENIVM": {"-": ["12"]},
    "XENO": {"N": ["17"]},
    "XENOCRATES": {"N": ["17"]},
    "XERAMPELINVS": {"-": ["21"]},
    "XYSTVS": {"-": ["17"]},
    "ZAMA": {"N": ["11"]},
    "ZELOTYPVS": {"-": ["21"]},
    "ZELVS": {"-": ["12"]},
    "ZENO": {"N": ["17"]},
    "ZEPHYRVS": {"N": ["12"], "-": ["12"]},
    "ZMARAGDVS": {"-": ["12"]},
    "ZONA": {"-": ["11"]},
    "ZONVLA": {"-": ["11"]},
    "ZOPHORVS": {"-": ["17"]},
    "ZOSIPPVS": {"N": ["12"]},
    "ZYGIOS": {"-": ["25"]},
    "ZYGON": {"-": ["17"]},
    "ALTOR": {"-": ["13"]},
    "AMFRACTVS": {"1": ["14"], "2": ["21"]},
    "BVTHROTOS": {"N": ["17"]},
    "CHERSONENSVS": {"N": ["12"]},
    "COMMINGO5L": {"COMMINISCOR": ["53"]},
    "CONSVLTVS": {"-": ["21"]},
    "CORIVM": {"-": ["12"]},
    "DISIVNCTVS": {"-": ["21"]},
    "DIVNGO": {"-": ["53"]},
    "DISIVNGO": {"-": ["53"]},
    "FVFIVS": {"A": ["21"], "N": ["12"]},
    "GERMANA11": {"GERMANIA": ["11"]},
    "HOSTVS": {"-": ["12"]},
    "LAEVVS": {"-": ["221"]},
    "MVSCVS": {"-": ["12"]},
    "NEMPE": {"-": ["60"]},
    "PAVLLO": {"-": ["60"]},
    "PETREIVS": {"N": ["12"]},
    "PROMVNTVRIVM": {"-": ["12"]},
    "RECERNO": {"-": ["53"]},
    "RVSTICATIO": {"-": ["213"]},
    "SCOBIS": {"-": ["13"]},
    "SVPERINICIO": {"-": ["55"]},
    "_SVM": {"1": ["56"], "2": ["E6"]},
    "VENERIVS": {"N": ["21"], "-": ["21"]},
    "#": {"-": ["--"]},
    "BABYLONICVS": {"A": ["21"]},
    "MAMMIVS": {"N": ["12"]},
    "MICIPSA": {"N": ["11"]},
    "SVCRONENSIS": {"A": ["24"]},
    "CANDAVIA": {"N": ["11"]},
    "CECROPIA": {"N": ["11"]},
    "LVCCIVS": {"N": ["12"]},
    "SEGIMERVS": {"N": ["12"]},
    "BEDRIACVM": {"N": ["12"]},
    "PYLIVS": {"A": ["21"]},
    "PERILLEVS": {"A": ["21"]},
    "REMMIVS": {"A": ["21"], "N": ["12"]},
    "ICARIVM": {"N": ["12"]},
    "LATIARIS": {"A": ["24"], "N": ["16"]},
    "ACHILLEVS": {"A": ["21"]},
    "LIBANVS": {"N": ["12"]},
    "APAMEA": {"N": ["11"]},
    "CRISPINILLA": {"N": ["11"]},
    "LOLLVS": {"N": ["12"]},
    "CAIVS": {"N": ["12"]},
    "MARCIANVS": {"N": ["12"], "A": ["21"]},
    "SABELLVS": {"A": ["21"]},
    "ZOSIMVS": {"N": ["12"]},
    "ILLYRIVS": {"A": ["21"]},
    "FVCINVS": {"N": ["12"]},
    "ARGYRIPPVS": {"N": ["12"]},
    "PRIAMEIVS": {"A": ["21"]},
    "ORTYGIVS": {"A": ["21"], "N": ["12"]},
    "VALA": {"N": ["11"]},
    "THYBRIS": {"A": ["24"], "N": ["16"]},
    "AVERNA": {"N": ["11"]},
    "SCAEA": {"N": ["11"]},
    "PAPIVS": {"N": ["12"], "A": ["21"]},
    "CRVSTVMIVS": {"A": ["21"]},
    "SVLMONENSIS": {"A": ["24"]},
    "EPYTVS": {"N": ["12"]},
    "RAVOLA": {"N": ["11"]},
    "SALAMINIVS": {"A": ["21"]},
    "APOPHORETA": {"2": ["12"], "1": ["12"]},
    "VESTRICIVS": {"N": ["12"]},
    "CEPHEIVS": {"A": ["21"]},
    "ACVTIA": {"N": ["11"]},
    "CATVLLA": {"N": ["11"]},
    "COLCHVS": {"A": ["21"]},
    "CVPIENNIVS": {"N": ["12"]},
    "PHEMIVS": {"A": ["21"]},
    "ASCRA": {"N": ["11"]},
    "DVCENIVS": {"N": ["12"]},
    "CARVILIVS": {"N": ["12"]},
    "CHERVSCVS": {"A": ["21"]},
    "TESSERARIVS": {"1": ["12"]},
    "LEVCTRA": {"N": ["11"]},
    "SAENIVS": {"A": ["21"], "N": ["12"]},
    "OBARITVS": {"N": ["12"]},
    "VIBVLENVS": {"N": ["12"]},
    "TARRACINENSIS": {"A": ["24"], "N": ["16"]},
    "HIBERVS": {"A": ["21"], "N": ["12"]},
    "ORCHIVIVS": {"N": ["12"]},
    "TALAVS": {"N": ["12"]},
    "CLONIVS": {"N": ["12"]},
    "TIMOTHEVS": {"N": ["12"]},
    "LATERANVS": {"N": ["12"]},
    "STATILIVS": {"N": ["12"]},
    "GABINIVS": {"N": ["12"], "A": ["21"]},
    "FLAVIANVS": {"A": ["21"], "N": ["12"]},
    "HICETAONIVS": {"A": ["21"]},
    "CLARIVS": {"N": ["12"], "A": ["21"]},
    "ACILIANVS": {"N": ["12"]},
    "CINNAMVS": {"N": ["12"]},
    "COLMINIANVS": {"A": ["21"]},
    "CARATACVS": {"N": ["12"]},
    "VOPISCVS": {"N": ["12"]},
    "ACTAEVS": {"A": ["21"]},
    "SIMBRVINVS": {"A": ["21"]},
    "ARIOVISTVS": {"N": ["12"]},
    "SOCRATICVS": {"A": ["21"]},
    "BIBVLA": {"N": ["11"]},
    "OEENSIS": {"A": ["24"]},
    "MEGAREA": {"N": ["11"]},
    "SOPHOCLEVS": {"A": ["21"]},
    "ALBINTIMILIVM": {"N": ["12"]},
    "HIPPARCHVS": {"N": ["12"]},
    "SILANA": {"N": ["11"]},
    "AMALTHEA": {"N": ["11"]},
    "CASTALIVS": {"A": ["21"]},
    "CAPHAREVS": {"A": ["21"], "N": ["12"]},
    "ELYSIVS": {"A": ["21"]},
    "OTOGESA": {"N": ["11"]},
    "THAMYRVS": {"N": ["12"]},
    "NERITIVS": {"A": ["21"]},
    "ALCATHOVS": {"N": ["12"]},
    "TOGONIVS": {"N": ["12"]},
    "VOLTVRCIVS": {"N": ["12"]},
    "SALENTINVS": {"A": ["21"]},
    "TIRYNTHIVS": {"N": ["12"], "A": ["21"]},
    "FVSCINVS": {"N": ["12"]},
    "PHENEVM": {"N": ["12"]},
    "TOLVMNIVS": {"N": ["12"]},
    "LATIVS": {"A": ["21"]},
    "SEMIPLACENTINVS": {"A": ["21"]},
    "AMORGVS": {"N": ["12"]},
    "PASITHEA": {"N": ["11"]},
    "VOLVMNIVS": {"N": ["12"]},
    "ANCHARIVS": {"N": ["12"]},
    "PHALEREVS": {"A": ["21"]},
    "SORICARIA": {"N": ["11"]},
    "BAEBIVS": {"N": ["12"]},
    "MODIA": {"N": ["11"]},
    "LYCVS": {"N": ["12"]},
    "MASSICVM": {"N": ["12"]},
    "CAVARILLVS": {"N": ["12"]},
    "LABDACVS": {"N": ["12"]},
    "PALINVRVS": {"N": ["12"]},
    "LARONIA": {"N": ["11"]},
    "VTRICVLVS": {"1": ["12"]},
    "TRAIANVS": {"A": ["21"], "N": ["12"]},
    "GABOLVS": {"N": ["12"]},
    "SVMMANVS": {"N": ["12"]},
    "HYRIEVS": {"A": ["21"], "N": ["12"]},
    "STYMPHALA": {"N": ["11"]},
    "ILIONEVS": {"N": ["12"]},
    "CEPHEVS": {"N": ["12"], "A": ["21"]},
    "PEDIVS": {"N": ["12"]},
    "CAPVA": {"N": ["11"]},
    "VOLSINIVS": {"A": ["21"]},
    "LAESTRYGONIVS": {"A": ["21"]},
    "AETHIOPIA": {"N": ["11"]},
    "CAMPANVS": {"A": ["21"], "N": ["12"]},
    "AMPSANCTVS": {"N": ["12"]},
    "HORVS": {"N": ["12"]},
    "TRAVSIVS": {"N": ["12"]},
    "VITELLIA": {"N": ["11"]},
    "PATVLCIVS": {"N": ["12"]},
    "CVRTILIVS": {"N": ["12"]},
    "SCANTIVS": {"A": ["21"]},
    "MEMNONIVS": {"A": ["21"]},
    "SATYRA": {"N": ["11"]},
    "PERSICVS": {"N": ["12"], "A": ["21"]},
    "TERPOLIVS": {"N": ["12"]},
    "POMPONIA": {"N": ["11"]},
    "CAMVLOGENVS": {"N": ["12"]},
    "AGNALIA": {"N": ["11"]},
    "COLONEVS": {"A": ["21"]},
    "SVADELA": {"N": ["11"]},
    "TRYPHAENA": {"N": ["11"]},
    "SARDANAPALVS": {"N": ["12"]},
    "IASIVS": {"N": ["12"]},
    "DECRIVS": {"N": ["12"]},
    "PENTHESILEA": {"N": ["11"]},
    "CESTIVS": {"N": ["12"]},
    "TRINACRIVS": {"A": ["21"]},
    "ADIATVANVS": {"N": ["12"]},
    "MENA": {"N": ["11"]},
    "FIRMINVS": {"N": ["12"]},
    "CLITVMNVS": {"N": ["12"]},
    "TIBVRTVS": {"N": ["12"]},
    "DAMASIPPVS": {"N": ["12"]},
    "PEGASEVS": {"A": ["21"]},
    "VOSEGVS": {"N": ["12"]},
    "CHLOREVS": {"N": ["12"]},
    "NARNIA": {"N": ["11"]},
    "RHOETVS": {"N": ["12"]},
    "LYCAONIVS": {"A": ["21"]},
    "CHAONIVS": {"A": ["21"]},
    "PEDANVS": {"A": ["21"]},
    "TRIFOLINVS": {"A": ["21"]},
    "TARENTVM": {"N": ["12"]},
    "CYLLARVS": {"N": ["12"]},
    "CLARANVS": {"N": ["12"]},
    "SALIVS": {"A": ["21"], "N": ["12"]},
    "LYSIPPVS": {"N": ["12"]},
    "OFONIVS": {"N": ["12"]},
    "GYGAEVS": {"A": ["21"]},
    "HYPERBOREVS": {"A": ["21"]},
    "ORESTILLA": {"N": ["11"]},
    "POLYPHEMVS": {"N": ["12"]},
    "ATRIVS": {"N": ["12"]},
    "LANVVIVM": {"N": ["12"]},
    "GALLOGRAECIA": {"N": ["11"]},
    "ARQVITIVS": {"N": ["12"]},
    "DELIA": {"N": ["11"]},
    "VBIVS": {"A": ["21"]},
    "NONACRIVS": {"A": ["21"]},
    "NERVLLINVS": {"N": ["12"]},
    "QVADRATILLA": {"N": ["11"]},
    "IVLIANVS": {"A": ["21"], "N": ["12"]},
    "ANAGNIA": {"N": ["11"]},
    "ZALACVS": {"N": ["12"]},
    "MIMALLONEVS": {"A": ["21"]},
    "ARGIVVS": {"A": ["21"]},
    "TARRACINA": {"N": ["11"]},
    "VIBELLIVS": {"N": ["12"]},
    "TISAMENVS": {"N": ["12"]},
    "TANTALEVS": {"A": ["21"]},
    "BIONEVS": {"A": ["21"]},
    "CYDNVS": {"N": ["12"]},
    "LARINA": {"N": ["11"]},
    "CAMARINA": {"N": ["11"]},
    "SEBOSIANVS": {"A": ["21"]},
    "NERTHVS": {"N": ["12"]},
    "CALVINIVS": {"N": ["12"]},
    "CLEOPATRA": {"N": ["11"]},
    "PISANVS": {"N": ["12"], "A": ["21"]},
    "RVLLVS": {"N": ["12"]},
    "SILARVS": {"N": ["12"]},
    "BVTA": {"N": ["11"]},
    "BAETICA": {"N": ["11"]},
    "FORTVNATVS": {"-": ["21"], "N": ["12"]},
    "HYPSIPYLAEVS": {"A": ["21"]},
    "ATHENAEVS": {"A": ["21"]},
    "FADIVS": {"N": ["12"]},
    "CLAVDIALIS": {"A": ["24"]},
    "MEGABOCCVS": {"N": ["12"]},
    "CINCIVS": {"A": ["21"]},
    "ATELLA": {"N": ["11"]},
    "PERGAMVM": {"N": ["12"]},
    "ANXVRVS": {"A": ["21"]},
    "MASTANESOSVS": {"N": ["12"]},
    "SVETONIVS": {"N": ["12"]},
    "CINARA": {"N": ["11"]},
    "CATONIANVS": {"A": ["21"]},
    "ITHACA": {"N": ["11"]},
    "ZACYNTHVS": {"N": ["12"]},
    "TILLIVS": {"N": ["12"]},
    "STICHVS": {"N": ["12"]},
    "NEMETOCENNA": {"N": ["11"]},
    "ALEXANDRINVS": {"A": ["21"]},
    "GRAECINA": {"N": ["11"]},
    "HERODVS": {"N": ["12"]},
    "CAMVRIVS": {"N": ["12"]},
    "MARTIGENA": {"N": ["11"]},
    "HECTOREVS": {"A": ["21"]},
    "LVGDVNENSIS": {"A": ["24"]},
    "CORELLIA": {"N": ["11"]},
    "TVRANNIVS": {"N": ["12"]},
    "HERMIPPVS": {"N": ["12"]},
    "LILYBAEA": {"N": ["11"]},
    "ENTELLVS": {"N": ["12"]},
    "PHOLVS": {"N": ["12"]},
    "TROILVS": {"N": ["12"]},
    "VENETVS": {"A": ["21"], "N": ["12"]},
    "PHTHIVS": {"A": ["21"]},
    "OPPIDIVS": {"N": ["12"]},
    "DACICVS": {"N": ["12"], "A": ["21"]},
    "BALEARICVS": {"N": ["12"]},
    "HELVIDIVS": {"A": ["21"], "N": ["12"]},
    "SILVS": {"N": ["12"]},
    "NISEVS": {"A": ["21"]},
    "THESSALIS": {"A": ["24"]},
    "ACRAGANTINVS": {"A": ["21"]},
    "IANALIS": {"A": ["24"]},
    "LVCTERIVS": {"N": ["12"]},
    "LITERNINVS": {"A": ["21"]},
    "PHERONACTVS": {"N": ["12"]},
    "IVNONICOLA": {"N": ["11"]},
    "MEDIA": {"N": ["11"]},
    "PARAETONICVS": {"A": ["21"]},
    "MINIVS": {"N": ["12"]},
    "SALONINVS": {"N": ["12"]},
    "CARPATHIVS": {"A": ["21"]},
    "CAENINA": {"N": ["11"]},
    "THYONIANVS": {"A": ["21"]},
    "LENNIVS": {"N": ["12"]},
    "CAECILIANVS": {"N": ["12"], "A": ["21"]},
    "BEDRIACVS": {"A": ["21"]},
    "PANDATERIA": {"N": ["11"]},
    "TARCONDARIVS": {"N": ["12"]},
    "CANOPIVS": {"A": ["21"]},
    "SALERNVM": {"N": ["12"]},
    "BATAVVS": {"A": ["21"]},
    "CVRTILLVS": {"N": ["12"]},
    "QVARTILLA": {"N": ["11"]},
    "SVEBVS": {"A": ["21"]},
    "PACENVS": {"N": ["12"]},
    "IAPETVS": {"N": ["12"]},
    "NASIDIENVS": {"N": ["12"]},
    "FAVSTA": {"N": ["11"]},
    "MOESIA": {"N": ["11"]},
    "HARPAGVS": {"N": ["12"]},
    "FAESIDIVS": {"N": ["12"]},
    "HYBLA": {"N": ["11"]},
    "MESSALLA": {"N": ["11"]},
    "MESSIVS": {"N": ["12"]},
    "ILISSVS": {"N": ["12"]},
    "AMBIVIVS": {"N": ["12"]},
    "NAVPACTOVS": {"A": ["21"]},
    "LAECA": {"N": ["11"]},
    "TYPHOEVS": {"A": ["21"], "N": ["12"]},
    "IPHITVS": {"N": ["12"]},
    "ARGYNNVS": {"N": ["12"]},
    "BAIVS": {"A": ["21"], "N": ["12"]},
    "HYPSAEVS": {"N": ["12"]},
    "HARPALVS": {"N": ["12"]},
    "DECIANVS": {"N": ["12"]},
    "ANDROGEVS": {"N": ["12"]},
    "LYCOMEDIVS": {"N": ["12"]},
    "PHOCVS": {"N": ["12"]},
    "BETACEVS": {"1": ["12"]},
    "SATRACHVS": {"N": ["12"]},
    "TVRIENSIS": {"A": ["24"]},
    "CROTONIENSIS": {"A": ["24"]},
    "PHIDIPPVS": {"N": ["12"]},
    "MAELIVS": {"N": ["12"]},
    "OEBALVS": {"N": ["12"]},
    "PAVLLVS": {"N": ["12"]},
    "LAVREOLVS": {"N": ["12"]},
    "TAENARIVS": {"A": ["21"]},
    "LICINVS": {"N": ["12"]},
    "OENOTRIVS": {"A": ["21"]},
    "ANICIVS": {"N": ["12"]},
    "LVSTRICIVS": {"N": ["12"]},
    "VEIA": {"N": ["11"]},
    "VACCENSIS": {"A": ["24"]},
    "DINIA": {"N": ["11"]},
    "METHYMNAEVS": {"A": ["21"]},
    "PAGYDA": {"N": ["11"]},
    "LAMVS": {"N": ["12"]},
    "PAGASA": {"N": ["11"]},
    "VOLAGINIVS": {"N": ["12"]},
    "REATINVS": {"A": ["21"]},
    "MELISSA": {"N": ["11"]},
    "FLEVVM": {"N": ["12"]},
    "DECIBALVS": {"N": ["12"]},
    "SINOPA": {"N": ["11"]},
    "GORDYAEVS": {"A": ["21"]},
    "RABOCENTVS": {"N": ["12"]},
    "LIBYA": {"N": ["11"]},
    "CORDOBA": {"N": ["11"]},
    "ANTEIVS": {"N": ["12"]},
    "PARIVM": {"N": ["12"]},
    "SALVIANVS": {"N": ["12"]},
    "NEAPOLITANVS": {"A": ["21"]},
    "TRITONIS": {"A": ["24"], "N": ["13"]},
    "CRINISVS": {"N": ["12"]},
    "MACCVS": {"N": ["12"]},
    "CLAVDIA": {"N": ["11"]},
    "SCYLACEVM": {"N": ["12"]},
    "THEMISCYRA": {"N": ["11"]},
    "AGERMVS": {"N": ["12"]},
    "ACYLLA": {"N": ["11"]},
    "OECHALIA": {"N": ["11"]},
    "VIRGINIA": {"N": ["11"]},
    "TITIDIVS": {"N": ["12"]},
    "ILIONA": {"N": ["11"]},
    "SOPHRONISCVS": {"N": ["12"]},
    "NEAERA": {"N": ["11"]},
    "TRANSPADANVS": {"A": ["21"]},
    "HIMELLA": {"N": ["11"]},
    "PARRHASIS": {"A": ["24"]},
    "APSVS": {"N": ["12"]},
    "CAPRICORNVS": {"N": ["12"], "-": ["12"]},
    "PVLCHRA": {"N": ["11"]},
    "PACHYNVS": {"N": ["12"]},
    "COSYRA": {"N": ["11"]},
    "LVGDVNVM": {"N": ["12"]},
    "ACADEMICA": {"N": ["11"]},
    "ZOILVS": {"N": ["12"]},
    "ORATA": {"N": ["11"]},
    "PATROBIVS": {"N": ["12"]},
    "MARSICVS": {"A": ["21"]},
    "ONAGVS": {"N": ["12"]},
    "SIMONIDEVS": {"A": ["21"]},
    "CADIVS": {"N": ["12"]},
    "CLVVIVS": {"N": ["12"]},
    "TRASIMENVS": {"N": ["12"], "A": ["21"]},
    "SVLGIVS": {"N": ["12"]},
    "THEOPHRASTVS": {"N": ["12"]},
    "PESSINVNTIVS": {"A": ["21"]},
    "CHALYBEIVS": {"A": ["21"]},
    "PSEVDOPHILIPPVS": {"N": ["12"]},
    "HIRTIVS": {"A": ["21"], "N": ["12"]},
    "CORFIDIVS": {"N": ["12"]},
    "ARRIA": {"N": ["11"]},
    "MECVLONIVS": {"N": ["12"]},
    "MARSAEVS": {"N": ["12"]},
    "FALANIVS": {"N": ["12"]},
    "VICTIVS": {"N": ["12"]},
    "ORTINVS": {"A": ["21"]},
    "LOLLIANVS": {"A": ["21"]},
    "PANNONICVS": {"A": ["21"]},
    "THRAESSA": {"N": ["11"]},
    "CONTEREBROMNIVS": {"A": ["21"]},
    "MYSIA": {"N": ["11"]},
    "OLYNTHIVS": {"A": ["21"]},
    "DARDANIVS": {"A": ["21"]},
    "MONOBAZVS": {"N": ["12"]},
    "EVTHYNICVS": {"N": ["12"]},
    "ARATVS": {"N": ["12"]},
    "CINNANVS": {"A": ["21"]},
    "EVANDRIVS": {"A": ["21"]},
    "VORTVMNVS": {"N": ["12"]},
    "GRAECANICVS": {"A": ["21"]},
    "CYRENAEVS": {"A": ["21"]},
    "MELIBOEVS": {"A": ["21"], "N": ["12"]},
    "FVNDANIVS": {"N": ["12"]},
    "CATIVS": {"N": ["12"]},
    "VOLVSVS": {"N": ["12"]},
    "TELEVS": {"N": ["12"]},
    "PRISCA": {"N": ["11"]},
    "METABVS": {"N": ["12"]},
    "LIGVRINVS": {"N": ["12"]},
    "VERVLAMIVM": {"N": ["12"]},
    "OGVLNIA": {"N": ["11"]},
    "VINIVS": {"N": ["12"]},
    "SIGNINVS": {"A": ["21"]},
    "BACTRA": {"N": ["11"]},
    "GONOESSA": {"N": ["11"]},
    "LEGERDA": {"N": ["11"]},
    "NEMEAEVS": {"A": ["21"]},
    "NYCTEVS": {"N": ["12"]},
    "MAGNESIVS": {"A": ["21"]},
    "RVSPINA": {"N": ["11"]},
    "LIBYCVS": {"A": ["21"]},
    "PHARNABAZVS": {"N": ["12"]},
    "AVRVNCVLEIVS": {"N": ["12"]},
    "BERECYNTIA": {"N": ["11"]},
    "PHERAEVS": {"A": ["21"]},
    "HELYMVS": {"N": ["12"]},
    "AMBRACIVS": {"A": ["21"]},
    "TIBERIOLVS": {"N": ["12"]},
    "METISCVS": {"N": ["12"]},
    "PHILADELPHVS": {"N": ["12"]},
    "PRIVERNVS": {"N": ["12"]},
    "AESONIVS": {"A": ["21"]},
    "CLYTAEMESTRA": {"N": ["11"]},
    "DINDYMENA": {"N": ["11"]},
    "ACCIVS": {"N": ["12"]},
    "DANAVS": {"A": ["21"], "N": ["12"]},
    "BOIOCALVS": {"N": ["12"]},
    "AQVILIA": {"N": ["11"]},
    "CASINVM": {"N": ["12"]},
    "VICESIMARIVS": {"1": ["12"]},
    "ALFIVS": {"N": ["12"]},
    "ARIADNAEVS": {"A": ["21"]},
    "VELLAVNODVNVM": {"N": ["12"]},
    "SPOLETINVS": {"A": ["21"]},
    "POLYHYMNIA": {"N": ["11"]},
    "POMEIA": {"N": ["11"]},
    "CAMERINVM": {"N": ["12"]},
    "NVCERINVS": {"N": ["12"], "A": ["21"]},
    "LAPITHA": {"N": ["11"]},
    "PAVLINVS": {"N": ["12"]},
    "TINGITANVS": {"A": ["21"]},
    "THABENA": {"N": ["11"]},
    "BLVCIVM": {"N": ["12"]},
    "HERMARCHVS": {"N": ["12"]},
    "AESCHYLEVS": {"A": ["21"]},
    "PELETHRONIVM": {"N": ["12"]},
    "CORREVS": {"N": ["12"]},
    "BADVHENNA": {"N": ["11"]},
    "COEVS": {"N": ["12"]},
    "CALLISTRATVS": {"N": ["12"]},
    "CAENEVS": {"N": ["12"]},
    "LEDAEVS": {"A": ["21"]},
    "APICATA": {"N": ["11"]},
    "CERCINA": {"N": ["11"]},
    "QVINTILIVS": {"N": ["12"]},
    "CARYSTIVS": {"A": ["21"]},
    "MALTINVS": {"N": ["12"]},
    "MAZIPPA": {"N": ["11"]},
    "DARDANIS": {"A": ["24"]},
    "BITHYNICVS": {"A": ["21"], "N": ["12"]},
    "GARGARA": {"N": ["11"]},
    "ELIDENSIS": {"A": ["24"]},
    "STRATONICA": {"N": ["11"]},
    "HELVETICVS": {"A": ["21"]},
    "IVNONIVS": {"A": ["21"]},
    "LAVRENTIVS": {"A": ["21"]},
    "CHALDAEVS": {"N": ["12"], "A": ["21"]},
    "CROESVS": {"N": ["12"]},
    "AQVILIANVS": {"A": ["21"]},
    "SCIRONIVS": {"A": ["21"]},
    "CARTIMANDVA": {"N": ["11"]},
    "ARCESILAVS": {"N": ["12"]},
    "VANNIVS": {"N": ["12"]},
    "PLINIVS": {"N": ["12"]},
    "ERGENNA": {"N": ["11"]},
    "NIOBEVS": {"A": ["21"]},
    "ARIMIENSIS": {"A": ["24"]},
    "SVTRIVM": {"N": ["12"]},
    "ARVPINVS": {"A": ["21"]},
    "MOCHVS": {"N": ["12"]},
    "CERBEREVS": {"A": ["21"]},
    "IBYCVS": {"N": ["12"]},
    "NABDALSA": {"N": ["11"]},
    "SORANVS": {"N": ["12"], "A": ["21"]},
    "AMPHOTERVS": {"N": ["12"]},
    "VENVSTINVS": {"A": ["21"]},
    "PALAEPHARSALVS": {"N": ["12"]},
    "QVADRIGARIVS": {"1": ["12"], "N": ["12"]},
    "VINNIVS": {"N": ["12"]},
    "MAGIA": {"N": ["11"]},
    "COMMIVS": {"N": ["12"]},
    "THERAEVS": {"A": ["21"]},
    "PERIBOMIVS": {"N": ["12"]},
    "HESVS": {"N": ["12"]},
    "CYTHEREA": {"N": ["11"]},
    "LARGA": {"N": ["11"]},
    "VRBICVS": {"N": ["12"]},
    "CALABRIA": {"N": ["11"]},
    "LAVSVS": {"N": ["12"]},
    "IVVENALIA": {"N": ["11"]},
    "HIPPOMENEIVS": {"A": ["21"]},
    "AMPHAEVS": {"N": ["12"]},
    "DAVNVS": {"N": ["12"]},
    "ASYLVS": {"N": ["12"]},
    "FAVENTINVS": {"N": ["12"]},
    "DVRIVS": {"N": ["12"]},
    "GESSIVS": {"N": ["12"]},
    "ILITHYIA": {"N": ["11"]},
    "PLAGVLEIVS": {"N": ["12"]},
    "OCELVM": {"N": ["12"]},
    "CEIVS": {"N": ["12"]},
    "SABELLICVS": {"A": ["21"]},
    "HECATAEVS": {"N": ["12"]},
    "GANGETIS": {"A": ["24"]},
    "RVFVLVS": {"-": ["21"], "N": ["12"]},
    "SEIVS": {"N": ["12"]},
    "SINVESSA": {"N": ["11"]},
    "HORMVS": {"N": ["12"]},
    "VERCASSIVELLAVNVS": {"N": ["12"]},
    "MAECIA": {"N": ["11"]},
    "EPIDAVRVS": {"N": ["12"]},
    "LVCILIANVS": {"A": ["21"]},
    "PHILIPPEVS": {"A": ["21"]},
    "CALATINVS": {"N": ["12"]},
    "TREVIRVS": {"A": ["21"]},
    "CHALINVS": {"N": ["12"]},
    "NICOMEDIA": {"N": ["11"]},
    "DACIA": {"N": ["11"]},
    "DAMASCVS": {"N": ["12"]},
    "ATACINVS": {"A": ["21"]},
    "LAERTIVS": {"A": ["21"], "N": ["12"]},
    "CEBALINVS": {"N": ["12"]},
    "ABVDIVS": {"N": ["12"]},
    "PLOTINA": {"N": ["11"]},
    "AQVINVS": {"N": ["12"]},
    "MVNIVS": {"N": ["12"]},
    "ANDRVS": {"N": ["12"]},
    "HEBRAEVS": {"A": ["21"]},
    "VENVSINVS": {"A": ["21"]},
    "CLEOBVLA": {"N": ["11"]},
    "PANATHENAICVS": {"N": ["12"]},
    "ARELLIVS": {"N": ["12"]},
    "CHIA": {"N": ["11"]},
    "FONTEIANVS": {"A": ["21"]},
    "QVIRINIVS": {"N": ["12"]},
    "TVTICANVS": {"N": ["12"]},
    "BABYLONIVS": {"A": ["21"]},
    "DIDIVS": {"N": ["12"], "A": ["21"]},
    "FABRICIVS": {"A": ["21"], "N": ["12"]},
    "MAEOTIS": {"A": ["24"], "N": ["16"]},
    "IDYIA": {"N": ["11"]},
    "OSTIENSIS": {"A": ["24"]},
    "QVINQVATRVS": {"N": ["12"]},
    "LACAENA": {"N": ["11"]},
    "APONIVS": {"N": ["12"]},
    "PRAECONINVS": {"N": ["12"]},
    "AVGAEVS": {"N": ["12"]},
    "MAEONIA": {"N": ["11"]},
    "LILYBEIVS": {"A": ["21"]},
    "AMISVS": {"N": ["12"]},
    "CASMILLA": {"N": ["11"]},
    "ASVVIVS": {"N": ["12"]},
    "SALARIVS": {"N": ["12"], "A": ["21"]},
    "ABYDENVS": {"A": ["21"]},
    "CAVDIVM": {"N": ["12"]},
    "HYMETTVS": {"N": ["12"]},
    "PARAPANISVS": {"N": ["12"]},
    "PORTICANVS": {"N": ["12"]},
    "DIANIVS": {"A": ["21"]},
    "ICCIVS": {"N": ["12"]},
    "MARICA": {"N": ["11"]},
    "LVSCIVS": {"N": ["12"]},
    "BROGITARVS": {"N": ["12"]},
    "TAMPIVS": {"N": ["12"]},
    "ADIABENVS": {"A": ["21"]},
    "CLVENTIVS": {"N": ["12"]},
    "PLAVTIVS": {"N": ["12"], "A": ["21"]},
    "TIFERNIVM": {"N": ["12"]},
    "BONONIA": {"N": ["11"]},
    "MAEONIVS": {"A": ["21"]},
    "NERIVS": {"N": ["12"]},
    "CHAERIPPVS": {"N": ["12"]},
    "ABDERA": {"N": ["11"]},
    "METTVS": {"N": ["12"]},
    "ZALEVCVS": {"N": ["12"]},
    "ELEGIA": {"N": ["11"]},
    "TYNDAREVS": {"N": ["12"]},
    "COMETA": {"N": ["11"]},
    "SEDVLLVS": {"N": ["12"]},
    "TANTALVS": {"N": ["12"]},
    "SCYLLAEVS": {"A": ["21"]},
    "BITHYNIVS": {"A": ["21"]},
    "VELEDA": {"N": ["11"]},
    "HELIODORVS": {"N": ["12"]},
    "SCAEVA": {"2": ["11"], "N": ["12"]},
    "ORICVM": {"N": ["12"]},
    "PANAETIVS": {"N": ["12"]},
    "HISPVLLA": {"N": ["11"]},
    "TMARVS": {"N": ["12"]},
    "THERAPNAEVS": {"A": ["21"]},
    "CLVSIVM": {"N": ["12"]},
    "APPVLEIVS": {"A": ["21"], "N": ["12"]},
    "GRATIANVS": {"N": ["12"]},
    "ERYTHEIS": {"A": ["24"]},
    "TVDICIVS": {"N": ["12"]},
    "EPHYREIACVS": {"A": ["21"]},
    "CASTRICIANVS": {"A": ["21"]},
    "ATALANTA": {"N": ["11"]},
    "BOSPORANVS": {"A": ["21"]},
    "TELESPHORVS": {"N": ["12"]},
    "PLANESIVM": {"N": ["12"]},
    "AMISSVS": {"N": ["12"]},
    "CRITOGNATVS": {"N": ["12"]},
    "PRVSA": {"N": ["11"]},
    "LITERNVM": {"N": ["12"]},
    "OSSA": {"N": ["11"]},
    "LETOVS": {"A": ["21"]},
    "TARENTINVS": {"A": ["21"]},
    "IVGVRTHA": {"N": ["11"]},
    "CYNOSVRA": {"N": ["11"]},
    "NASIDIANVS": {"A": ["21"]},
    "FENESTELLA": {"N": ["11"]},
    "CREMERENSIS": {"A": ["24"]},
    "CASCA": {"N": ["11"]},
    "CLYMENVS": {"N": ["12"]},
    "STAIVS": {"N": ["12"]},
    "GIGANTEVS": {"A": ["21"]},
    "DIPHILVS": {"N": ["12"]},
    "PYTHIVS": {"N": ["12"], "A": ["21"]},
    "GORGONEVS": {"A": ["21"]},
    "BESTIVS": {"N": ["12"]},
    "MALCHVS": {"N": ["12"]},
    "HIERONYMVS": {"N": ["12"]},
    "PYTAGOREVS": {"A": ["21"]},
    "LONDINIVM": {"N": ["12"]},
    "VINALIA": {"N": ["11"]},
    "ALFENVS": {"N": ["12"]},
    "TRYPHERVS": {"N": ["12"]},
    "MEGARVS": {"A": ["21"]},
    "PARTHENOPAEVS": {"N": ["12"]},
    "LECHAEVM": {"N": ["12"]},
    "HERIVS": {"N": ["12"]},
    "OCNVS": {"N": ["12"]},
    "NEREIVS": {"A": ["21"]},
    "SARNVS": {"N": ["12"]},
    "BAZAIRA": {"N": ["11"]},
    "THASVS": {"N": ["12"]},
    "ROMILIVS": {"N": ["12"]},
    "FORMIANVM": {"N": ["12"]},
    "TRISANTONA": {"N": ["11"]},
    "EPIMACHVS": {"N": ["12"]},
    "CIRTA": {"N": ["11"]},
    "ANDROMACHA": {"N": ["11"]},
    "CYRENENSIS": {"A": ["24"]},
    "VERANIVS": {"N": ["12"]},
    "LARINVM": {"N": ["12"]},
    "GANNASCVS": {"N": ["12"]},
    "ABNOBA": {"N": ["11"]},
    "SERGESTVS": {"N": ["12"]},
    "EPPONINA": {"N": ["11"]},
    "EVRYSTHEVS": {"N": ["12"]},
    "CYTHERIACVS": {"A": ["21"]},
    "CALLIOPEA": {"N": ["11"]},
    "MARVS": {"N": ["12"]},
    "MARRVCINVS": {"A": ["21"]},
    "MYTILENENSIS": {"A": ["24"]},
    "ARISTIPPVS": {"N": ["12"]},
    "NOREIA": {"N": ["11"]},
    "TRANSRHENANVS": {"A": ["21"]},
    "STATILIA": {"N": ["11"]},
    "OCELLA": {"N": ["11"]},
    "SARDONIVS": {"A": ["21"]},
    "TITARESVS": {"N": ["12"]},
    "FAENIVS": {"N": ["12"]},
    "OENOTRVS": {"A": ["21"]},
    "COSSINIVS": {"N": ["12"]},
    "NEMORENSIS": {"A": ["24"]},
    "VIRBIVS": {"N": ["12"]},
    "PANSA": {"N": ["11"]},
    "FORTVNATA": {"N": ["11"]},
    "SISYPHIVS": {"A": ["21"]},
    "HATERIVS": {"N": ["12"]},
    "VMBRIA": {"N": ["11"]},
    "GAVDA": {"N": ["11"]},
    "PROTESILAEVS": {"A": ["21"]},
    "CYTHEREIVS": {"A": ["21"]},
    "LAECANIVS": {"N": ["12"]},
    "RECENTORICVS": {"A": ["21"]},
    "CONSVS": {"N": ["12"]},
    "ARGANTHONIVS": {"N": ["12"]},
    "CAVILLONVM": {"N": ["12"]},
    "LESBIVS": {"N": ["12"], "A": ["21"]},
    "CLEONAEVS": {"A": ["21"]},
    "DAVVS": {"N": ["12"]},
    "HASTENSIS": {"A": ["24"]},
    "COMPSA": {"N": ["11"]},
    "PYLVS": {"N": ["12"]},
    "DARDANIA": {"N": ["11"]},
    "THERMODONTIVS": {"A": ["21"]},
    "GALLINARIVS": {"A": ["21"]},
    "CAESIANVS": {"N": ["12"]},
    "ARISTOBVLVS": {"N": ["12"]},
    "ADRASTVS": {"N": ["12"]},
    "GALATIA": {"N": ["11"]},
    "CLOTA": {"N": ["11"]},
    "QVINTIANVS": {"A": ["21"], "N": ["12"]},
    "VESTRIVS": {"N": ["12"]},
    "PAVSIACVS": {"A": ["21"]},
    "ARABVS": {"A": ["21"], "N": ["12"]},
    "APRILIS": {"N": ["13"], "A": ["12"]},
    "TVSCVLVS": {"A": ["21"], "N": ["12"]},
    "PATISIVS": {"N": ["12"]},
    "CAMILLVS": {"N": ["12"]},
    "GRAVPIVS": {"N": ["12"]},
    "ACERRONIA": {"N": ["11"]},
    "SOSPITA": {"N": ["11"]},
    "LESBIS": {"A": ["24"]},
    "EPIDAMNVS": {"N": ["12"]},
    "THESSANDRVS": {"N": ["12"]},
    "TORONAEVS": {"A": ["21"]},
    "BELVS": {"N": ["12"]},
    "TELEGONVS": {"N": ["12"]},
    "TEMNVS": {"N": ["12"]},
    "BAGRADA": {"N": ["11"]},
    "TELAMONIVS": {"N": ["12"]},
    "STABERIVS": {"N": ["12"]},
    "CYTORIVS": {"A": ["21"]},
    "MINOTAVRVS": {"N": ["12"]},
    "ANGITIA": {"N": ["11"]},
    "CRETANVS": {"N": ["12"]},
    "MAEOTICVS": {"A": ["21"]},
    "ANAPVS": {"N": ["12"]},
    "ITYRAEVS": {"A": ["21"]},
    "TARRACINIENSIS": {"A": ["24"]},
    "SAMAXVS": {"N": ["12"]},
    "LEMANNVS": {"N": ["12"]},
    "EVODVS": {"N": ["12"]},
    "DONVSA": {"N": ["11"]},
    "NILIACVS": {"A": ["21"]},
    "DVRATIVS": {"N": ["12"]},
    "FLORALIS": {"A": ["24"]},
    "ZONARIVS": {"1": ["12"]},
    "EPRIVS": {"N": ["12"]},
    "FABVLLA": {"N": ["11"]},
    "COTTIANVS": {"A": ["21"]},
    "NIREVS": {"N": ["12"]},
    "ARDEATINVM": {"N": ["12"]},
    "SILA": {"N": ["11"]},
    "TYRRHENVS": {"N": ["12"], "A": ["21"]},
    "MENELAVS": {"N": ["12"]},
    "MALIVS": {"A": ["21"]},
    "VERATIANVS": {"A": ["21"]},
    "GYTHEVM": {"N": ["12"]},
    "NEAPOLITANVM": {"N": ["12"]},
    "ZELA": {"N": ["11"]},
    "ATLANTEVS": {"A": ["21"]},
    "CARRVCA": {"N": ["11"]},
    "LYGDAMVS": {"N": ["12"]},
    "SAEVINVS": {"N": ["12"]},
    "AREOPAGVS": {"N": ["12"]},
    "DEMODOCVS": {"N": ["12"]},
    "NIGRINVS": {"N": ["12"]},
    "MAROBODVVS": {"N": ["12"]},
    "VELABRA": {"N": ["11"]},
    "ALTINVS": {"N": ["12"]},
    "HIPPONOVS": {"N": ["12"]},
    "PHILOPOLEMVS": {"N": ["12"]},
    "LAVINVS": {"A": ["21"]},
    "VENILIA": {"N": ["11"]},
    "VFENTINA": {"N": ["11"]},
    "LICYMNIA": {"N": ["11"]},
    "MARONEVS": {"A": ["21"]},
    "SIDICINVS": {"A": ["21"]},
    "PRIMIGENIVS": {"N": ["12"]},
    "AMATHVSIVS": {"A": ["21"]},
    "BASSARICVS": {"A": ["21"]},
    "LEVCADIVS": {"A": ["21"]},
    "LYCTIVS": {"A": ["21"]},
    "LYNCEVS": {"N": ["12"], "A": ["21"]},
    "TORQVATA": {"N": ["11"]},
    "ELATEA": {"N": ["11"]},
    "VENTIDIVS": {"N": ["12"]},
    "THRIASIVS": {"A": ["21"]},
    "NICEPHORIVM": {"N": ["12"]},
    "ERGASILVS": {"N": ["12"]},
    "SAPPHICVS": {"A": ["21"]},
    "GALLICANVS": {"A": ["21"]},
    "PHOENISSVS": {"A": ["21"]},
    "THIA": {"N": ["11"]},
    "ASTYPALAEA": {"N": ["11"]},
    "VENAFRVM": {"N": ["12"]},
    "DAVCIVS": {"A": ["21"]},
    "PASSIENVS": {"N": ["12"]},
    "CABERA": {"N": ["11"]},
    "TROVCILLVS": {"N": ["12"]},
    "IAPYDIA": {"N": ["11"]},
    "DRVSIANVS": {"A": ["21"]},
    "STYMPHALIVS": {"A": ["21"]},
    "SERAPA": {"N": ["11"]},
    "CREMONA": {"N": ["11"]},
    "TONGILIVS": {"N": ["12"]},
    "PACARIVS": {"N": ["12"]},
    "LYCINNA": {"N": ["11"]},
    "SATTIA": {"N": ["11"]},
    "TELESINVS": {"N": ["12"]},
    "SACROVIRIANVS": {"A": ["21"]},
    "CERAVNVS": {"A": ["21"]},
    "SCAEVINVS": {"N": ["12"]},
    "ADAMASTVS": {"N": ["12"]},
    "COCYTIVS": {"A": ["21"]},
    "CARFINIA": {"N": ["11"]},
    "PIPLEIVS": {"A": ["21"]},
    "HYMETTIVS": {"A": ["21"]},
    "BIBACVLVS": {"N": ["12"]},
    "SEQVANA": {"N": ["11"]},
    "SVBVRANVS": {"A": ["21"], "N": ["12"]},
    "EPPIA": {"N": ["11"]},
    "ERETVM": {"N": ["12"]},
    "SVEVVS": {"A": ["21"]},
    "PAELIGNVS": {"A": ["21"], "N": ["12"]},
    "SVFFICIVS": {"N": ["12"]},
    "CORYCIVS": {"A": ["21"]},
    "HYRTACVS": {"N": ["12"]},
    "LATONIVS": {"A": ["21"]},
    "CAMERIVM": {"N": ["12"]},
    "GORTYNIS": {"A": ["24"]},
    "SICYONIA": {"N": ["11"]},
    "LYCAONIA": {"N": ["11"]},
    "BAREA": {"N": ["11"]},
    "PAGASAEVS": {"A": ["21"]},
    "CHARIOVALDA": {"N": ["11"]},
    "MOLORCHVS": {"N": ["12"]},
    "VELINVS": {"A": ["21"], "N": ["12"]},
    "AVTOBVLVS": {"N": ["12"]},
    "PANDARVS": {"N": ["12"]},
    "CATHEGVS": {"N": ["12"]},
    "TERMESTINVS": {"A": ["21"]},
    "PYRRHVS": {"N": ["12"]},
    "NERITVS": {"N": ["12"]},
    "ILIACVS": {"A": ["21"]},
    "OPHIVCHVS": {"N": ["12"]},
    "ERYMANTHVS": {"N": ["12"]},
    "ARTEMONA": {"N": ["11"]},
    "ALPHESIBOEA": {"N": ["11"]},
    "PAMPHILA": {"N": ["11"]},
    "GALERIVS": {"N": ["12"]},
    "ARVLENVS": {"N": ["12"]},
    "TERTVLLINVS": {"N": ["12"]},
    "OLYMPISCVS": {"N": ["12"]},
    "HESIONA": {"N": ["11"]},
    "DAEDALEVS": {"A": ["21"]},
    "ARCHESTRATA": {"N": ["11"]},
    "GARVMNA": {"N": ["11"]},
    "MATINVS": {"N": ["12"], "A": ["21"]},
    "MNESTHEVS": {"N": ["12"]},
    "MATVTA": {"N": ["11"]},
    "PHTHIOTICVS": {"A": ["21"]},
    "CATIENA": {"N": ["11"]},
    "MIMNERMVS": {"N": ["12"]},
    "SVILLIVS": {"N": ["12"]},
    "COMMAGENVS": {"A": ["21"]},
    "AVENIENSIS": {"A": ["24"]},
    "ANTONIA": {"N": ["11"]},
    "ARPINVM": {"N": ["12"]},
    "PVNICANVS": {"A": ["21"]},
    "SVRRENTINVS": {"A": ["21"]},
    "TERENTIANVS": {"A": ["21"]},
    "ANCONITANVS": {"A": ["21"]},
    "VLIA": {"N": ["11"]},
    "THAPSACVS": {"N": ["12"]},
    "MAENIVS": {"A": ["21"], "N": ["12"]},
    "VACVNALIS": {"A": ["24"]},
    "AENEIVS": {"A": ["21"]},
    "EVRYLOCHVS": {"N": ["12"]},
    "ASSYRIA": {"N": ["11"]},
    "TMOLIVS": {"A": ["21"]},
    "TEGEAEVS": {"A": ["21"]},
    "VELABRVM": {"N": ["12"]},
    "RVFRIVS": {"N": ["12"]},
    "COERANVS": {"N": ["12"]},
    "CLEVSTRATA": {"N": ["11"]},
    "PHILOCTETA": {"N": ["11"]},
    "NABALIA": {"N": ["11"]},
    "CHRISTVS": {"N": ["12"]},
    "PILVMNVS": {"N": ["12"]},
    "SANQVINIVS": {"N": ["12"]},
    "ANDANIA": {"N": ["11"]},
    "HERBESVS": {"N": ["12"]},
    "CRISPINVS": {"N": ["12"]},
    "ARCTVS": {"N": ["12"]},
    "TYRTAEVS": {"N": ["12"]},
    "CARYSTVS": {"N": ["12"]},
    "VERVLANA": {"N": ["11"]},
    "MAVORTIVS": {"A": ["21"]},
    "LESBOVS": {"A": ["21"]},
    "ICARIA": {"N": ["11"]},
    "BOLA": {"N": ["11"]},
    "LVCANIA": {"N": ["11"]},
    "CANTIVM": {"N": ["12"]},
    "HIBERINA": {"N": ["11"]},
    "PERIDIA": {"N": ["11"]},
    "EPHORVS": {"-": ["12"], "N": ["12"]},
    "PROTHOVS": {"N": ["12"]},
    "THEONINVS": {"A": ["21"]},
    "CABVRVS": {"N": ["12"]},
    "DICAEARCHVS": {"N": ["12"]},
    "PETVSIVS": {"N": ["12"]},
    "NYCTELIVS": {"A": ["21"], "N": ["12"]},
    "ANTAEVS": {"N": ["12"]},
    "TEGEA": {"N": ["11"]},
    "NYMPHIDIVS": {"N": ["12"]},
    "ATTIANVS": {"A": ["21"], "N": ["12"]},
    "IAERA": {"N": ["11"]},
    "POSIDONIVS": {"N": ["12"]},
    "BISTONIS": {"A": ["24"]},
    "PALILIS": {"A": ["24"]},
    "THYNVS": {"A": ["21"]},
    "PHOCEVS": {"A": ["21"], "N": ["12"]},
    "DALMATICVS": {"A": ["21"]},
    "AVGVSTALIS": {"A": ["24"]},
    "ARENACVM": {"N": ["12"]},
    "ORCHOMENVS": {"N": ["12"]},
    "LEDA": {"N": ["11"]},
    "ARGVS": {"A": ["21"], "N": ["12"]},
    "NAXIVS": {"A": ["21"]},
    "CORNVTVS": {"N": ["12"]},
    "OLYMPIACVS": {"A": ["21"]},
    "AMPHILOCHIA": {"N": ["11"]},
    "SAETABVS": {"A": ["21"]},
    "GABINIANVS": {"A": ["21"], "N": ["12"]},
    "ARPINIVS": {"N": ["12"]},
    "SEXTILIVS": {"N": ["12"]},
    "REGINVS": {"N": ["12"], "A": ["21"]},
    "EVRYALVS": {"N": ["12"]},
    "PESSINVS": {"N": ["12"]},
    "SVEBIA": {"N": ["11"]},
    "DALMATIA": {"N": ["11"]},
    "PISISTRATVS": {"N": ["12"]},
    "PICENVM": {"N": ["12"]},
    "ATINA": {"N": ["11"]},
    "DINDYMVS": {"A": ["21"], "N": ["12"]},
    "PICENTIVS": {"A": ["21"]},
    "THESEVS": {"A": ["21"], "N": ["12"]},
    "NORENSIS": {"A": ["24"]},
    "DYMNVS": {"N": ["12"]},
    "BATAVODVRVM": {"N": ["12"]},
    "FALERNVM": {"N": ["12"]},
    "ICARIVS": {"A": ["21"], "N": ["12"]},
    "PETELIA": {"N": ["11"]},
    "CATVLLVS": {"N": ["12"]},
    "RVBRENVS": {"N": ["12"]},
    "SARMENTVS": {"N": ["12"]},
    "PROXVMVS": {"N": ["12"]},
    "VITIA": {"N": ["11"]},
    "CRESSIVS": {"A": ["21"]},
    "ACADEMVS": {"N": ["12"]},
    "NERONEVS": {"A": ["21"]},
    "PATAREVS": {"A": ["21"]},
    "NVMISIVS": {"N": ["12"]},
    "DVELLIVS": {"N": ["12"]},
    "OPTIMVS": {"N": ["12"]},
    "ZENOBIA": {"N": ["11"]},
    "PARILIA": {"N": ["11"]},
    "OPPIVS": {"N": ["12"], "A": ["21"]},
    "BVLLATIVS": {"N": ["12"]},
    "SILLIVS": {"N": ["12"]},
    "OLLIVS": {"N": ["12"]},
    "CANVTIVS": {"N": ["12"]},
    "MAVSOLEVS": {"A": ["21"]},
    "CLASSICIANVS": {"N": ["12"]},
    "LATINIENSIS": {"A": ["24"], "N": ["16"]},
    "FRONTINVS": {"N": ["12"]},
    "GAETVLVS": {"A": ["21"]},
    "ASPARAGIVM": {"N": ["12"]},
    "AMATHVSIA": {"N": ["11"]},
    "MVRCVS": {"N": ["12"]},
    "LYDIVS": {"A": ["21"]},
    "CVTIVS": {"N": ["12"]},
    "OLEARVS": {"N": ["12"]},
    "NEARCHVS": {"N": ["12"]},
    "NASAMONIVS": {"A": ["21"]},
    "TYPHOIS": {"A": ["24"]},
    "BISTONIVS": {"A": ["21"]},
    "PLOCAMVS": {"N": ["12"]},
    "CEPHALVS": {"N": ["12"]},
    "PHORCVS": {"N": ["12"]},
    "SCANTIANVS": {"A": ["21"]},
    "ANAXARCHVS": {"N": ["12"]},
    "LVCAGVS": {"N": ["12"]},
    "TIMONEVS": {"A": ["21"]},
    "POSTVMIANVS": {"A": ["21"]},
    "BOVIANVM": {"N": ["12"]},
    "ROVCILLVS": {"N": ["12"]},
    "FOLIA": {"N": ["11"]},
    "TVCCIVS": {"N": ["12"]},
    "CAECVBVS": {"A": ["21"]},
    "DANAEIVS": {"A": ["21"]},
    "CRASSICIVS": {"N": ["12"]},
    "APONIANA": {"N": ["11"]},
    "HERCVLEIVS": {"N": ["12"]},
    "BRENNVS": {"N": ["12"]},
    "FVRFANIVS": {"N": ["12"]},
    "ADONEVS": {"N": ["12"]},
    "BRVNDISINVS": {"A": ["21"]},
    "CENTONARIVS": {"1": ["12"]},
    "TEVTHRANTIVS": {"A": ["21"]},
    "MOPSOPIA": {"N": ["11"]},
    "PLATENIVS": {"A": ["21"]},
    "CALIGVLA": {"N": ["11"]},
    "ENIPEVS": {"N": ["12"]},
    "PERMESSVS": {"N": ["12"]},
    "THESSALA": {"N": ["11"]},
    "ACMONENSIS": {"A": ["24"]},
    "STHENEBOEA": {"N": ["11"]},
    "TIRESIA": {"N": ["11"]},
    "SCYTHICVS": {"A": ["21"]},
    "SCANTIA": {"N": ["11"]},
    "BOCVLVS": {"N": ["12"]},
    "DIOMEDEVS": {"A": ["21"]},
    "PETRINVM": {"N": ["12"]},
    "ATIVS": {"N": ["12"]},
    "ILVA": {"N": ["11"]},
    "DOSSENVS": {"N": ["12"]},
    "SQVILLVS": {"N": ["12"]},
    "TARSVS": {"N": ["12"]},
    "CVRETIS": {"A": ["24"]},
    "GORTYNIVS": {"A": ["21"]},
    "ACCA": {"N": ["11"]},
    "CECROPIS": {"A": ["24"]},
    "BELLIENVS": {"N": ["12"]},
    "TECMESSA": {"N": ["11"]},
    "PHERECLEVS": {"A": ["21"]},
    "SAMOTHRACIA": {"N": ["11"]},
    "MAENALIS": {"A": ["24"]},
    "PAPHIVS": {"A": ["21"]},
    "LVCETIVS": {"N": ["12"]},
    "DEXITHEA": {"N": ["11"]},
    "MEMPHITIS": {"A": ["24"]},
    "PHARIVS": {"N": ["12"], "A": ["21"]},
    "ISTHMIVS": {"A": ["21"]},
    "SVBRIVS": {"N": ["12"]},
    "MEGALESIACVS": {"A": ["21"]},
    "SALIENVS": {"N": ["12"]},
    "ALBESIVS": {"N": ["12"]},
    "XENOPHILVS": {"N": ["12"]},
    "DAEDALA": {"N": ["11"]},
    "ALBIVS": {"N": ["12"]},
    "ACVTIVS": {"N": ["12"]},
    "MANNICIVS": {"N": ["12"]},
    "MAECILIA": {"N": ["11"]},
    "BYZANTIVS": {"A": ["21"]},
    "VAGELLIVS": {"N": ["12"]},
    "ALLOBROGICVS": {"N": ["12"]},
    "CORNICVLANVS": {"A": ["21"]},
    "ARISBA": {"N": ["11"]},
    "PROCILLA": {"N": ["11"]},
    "TYDEVS": {"N": ["12"]},
    "BATTVS": {"N": ["12"]},
    "MEDVS": {"N": ["12"], "A": ["21"]},
    "LEONIDA": {"N": ["11"]},
    "MARDVS": {"A": ["21"]},
    "AEQVICVLVS": {"A": ["21"]},
    "PEDANIVS": {"N": ["12"]},
    "AEGIMVRVS": {"N": ["12"]},
    "ACTIVM": {"N": ["12"]},
    "SERGIANVS": {"A": ["21"]},
    "CRETAEVS": {"A": ["21"]},
    "PADVSA": {"N": ["11"]},
    "GORGOBINA": {"N": ["11"]},
    "ENNIA": {"N": ["11"]},
    "TEMESAEVS": {"A": ["21"]},
    "CASSIVELLAVNVS": {"N": ["12"]},
    "PVPINIA": {"N": ["11"]},
    "NARBONENSIS": {"A": ["24"]},
    "APPENNINVS": {"N": ["12"]},
    "VIRDOMARVS": {"N": ["12"]},
    "CONGVS": {"N": ["12"]},
    "CAESELLIVS": {"N": ["12"]},
    "OSINIVS": {"N": ["12"]},
    "SCRIBONIANVS": {"N": ["12"]},
    "DIONYSVS": {"N": ["12"]},
    "RVMINVS": {"A": ["21"]},
    "THEMISTA": {"N": ["11"]},
    "HYMENAEVS": {"-": ["12"], "N": ["12"]},
    "NASIDIVS": {"N": ["12"]},
    "CAICVS": {"N": ["12"]},
    "DVLICHIA": {"N": ["11"]},
    "ECBOLIMA": {"N": ["11"]},
    "AEGEVS": {"N": ["12"]},
    "THEODAMANTEVS": {"A": ["21"]},
    "OPITERGIVM": {"N": ["12"]},
    "VESPASIANVS": {"N": ["12"]},
    "AVIENVS": {"N": ["12"]},
    "PISTORIENSIS": {"A": ["24"]},
    "VALENTINVS": {"N": ["12"]},
    "THRAECIVS": {"A": ["21"]},
    "IONIS": {"A": ["24"]},
    "FADVS": {"N": ["12"]},
    "VERVLANVS": {"N": ["12"]},
    "VESVVIVS": {"N": ["12"]},
    "LABERIANVS": {"A": ["21"]},
    "CLEOMBROTVS": {"N": ["12"]},
    "TVLLINVS": {"N": ["12"]},
    "ALINDIENSIS": {"A": ["24"]},
    "DANVVIVS": {"N": ["12"]},
    "CASTICVS": {"N": ["12"]},
    "AVRVNCVLEIA": {"N": ["11"]},
    "ADRANA": {"N": ["11"]},
    "CALPVRNIANVS": {"N": ["12"]},
    "CERVARIVS": {"N": ["12"]},
    "ZETA": {"N": ["11"]},
    "GALBIANVS": {"A": ["21"]},
    "ITHACVS": {"A": ["21"], "N": ["12"]},
    "ARPINVS": {"A": ["21"]},
    "ACHAEMENIVS": {"A": ["21"]},
    "PANGAEVS": {"N": ["12"]},
    "AEGIENSIS": {"A": ["24"]},
    "CLYTIVS": {"N": ["12"]},
    "THREISSA": {"N": ["11"]},
    "NORICA": {"N": ["11"]},
    "COVS": {"A": ["21"], "N": ["12"]},
    "FALISCVS": {"A": ["21"]},
    "HYPSAEA": {"N": ["11"]},
    "ALLIARIA": {"N": ["11"]},
    "ANICETVS": {"N": ["12"]},
    "MORITASGVS": {"N": ["12"]},
    "FONTEIVS": {"A": ["21"], "N": ["12"]},
    "OLELVS": {"N": ["12"]},
    "GRATILLA": {"N": ["11"]},
    "COROEBVS": {"N": ["12"]},
    "CANTHARA": {"N": ["11"]},
    "HADRIVS": {"A": ["21"]},
    "PORRIMA": {"N": ["11"]},
    "EREBEVS": {"A": ["21"]},
    "PARNETHVS": {"N": ["12"]},
    "NICOMEDENSIS": {"A": ["24"]},
    "LEONNATVS": {"N": ["12"]},
    "PSEVDOLVS": {"N": ["12"]},
    "COPTVS": {"N": ["12"]},
    "OFELLVS": {"N": ["12"]},
    "PARENTALIS": {"A": ["24"]},
    "HIPPONENSIS": {"A": ["24"]},
    "BELLOVACVS": {"A": ["21"]},
    "DREPANVM": {"N": ["12"]},
    "ACRISIONEVS": {"A": ["21"]},
    "RVTVBA": {"N": ["11"]},
    "LIBITINA": {"N": ["11"]},
    "LYSIDICVS": {"N": ["12"]},
    "ASCONIVS": {"N": ["12"]},
    "PHILOMELA": {"N": ["11"]},
    "VESCVLARIVS": {"N": ["12"]},
    "SOLYMVS": {"A": ["21"]},
    "TARQVITVS": {"N": ["12"]},
    "AMPHIONIVS": {"A": ["21"]},
    "MARGIANA": {"N": ["11"]},
    "DVMNACVS": {"N": ["12"]},
    "PYGMAEVS": {"A": ["21"]},
    "SOZA": {"N": ["11"]},
    "CAESENNIVS": {"N": ["12"]},
    "MOSELLA": {"N": ["11"]},
    "PONTIVS": {"N": ["12"]},
    "CISSEVS": {"N": ["12"]},
    "PHOEBEVS": {"A": ["21"]},
    "SATICVLVS": {"N": ["12"]},
    "APENNINVS": {"N": ["12"]},
    "VELLEIVS": {"N": ["12"]},
    "ANEMVRIENSIS": {"A": ["24"]},
    "NEPTVNIVS": {"A": ["21"]},
    "EGNATIA": {"N": ["11"]},
    "SCISSA": {"N": ["11"]},
    "TEVTONICVS": {"A": ["21"]},
    "ISMARA": {"N": ["11"]},
    "AMVLLIVS": {"N": ["12"]},
    "LVCERIA": {"N": ["11"]},
    "SELINVS": {"N": ["12"]},
    "LIBERALIA": {"N": ["11"]},
    "MALLIVS": {"N": ["12"]},
    "TROXOBORVS": {"N": ["12"]},
    "CORFINIENSIS": {"A": ["24"]},
    "TMARIVS": {"A": ["21"]},
    "HIRRVS": {"N": ["12"]},
    "MINTVRNENSIS": {"A": ["24"]},
    "FABATVS": {"N": ["12"]},
    "FLACCILLA": {"N": ["11"]},
    "PANTHVS": {"N": ["12"]},
    "BACCHEIVS": {"A": ["21"]},
    "ANDRIA": {"N": ["11"]},
    "LAMYRVS": {"N": ["12"]},
    "AVARICENSIS": {"A": ["24"]},
    "AERIA": {"N": ["11"]},
    "IDRVS": {"N": ["12"]},
    "MILICHVS": {"N": ["12"]},
    "DIVODVRVM": {"N": ["12"]},
    "CAPHARIS": {"A": ["24"]},
    "CORYBANTIVS": {"A": ["21"]},
    "FIMBRIANVS": {"A": ["21"]},
    "LEMNIVS": {"A": ["21"]},
    "SELEVCVS": {"N": ["12"]},
    "ARGOLICVS": {"A": ["21"], "N": ["12"]},
    "ALBINA": {"N": ["11"]},
    "POLYPLVSIVS": {"A": ["21"]},
    "AMATHVS": {"N": ["12"]},
    "SPEVSIPPVS": {"N": ["12"]},
    "PIERIVS": {"A": ["21"]},
    "CORVNCANIVS": {"N": ["12"]},
    "QVINTILIANVS": {"N": ["12"]},
    "CLVVIDIENVS": {"N": ["12"]},
    "LEVCIPPVS": {"N": ["12"]},
    "POMPEIANVS": {"N": ["12"], "A": ["21"]},
    "AGRIVS": {"N": ["12"]},
    "FVLVIA": {"N": ["11"]},
    "DORIVS": {"A": ["21"]},
    "THRASYLLVS": {"N": ["12"]},
    "DEIANIRA": {"N": ["11"]},
    "SABRINA": {"N": ["11"]},
    "BAIANVS": {"A": ["21"]},
    "ARTORIVS": {"N": ["12"]},
    "TRALLIANVS": {"A": ["21"]},
    "MAVRICVS": {"N": ["12"]},
    "NERATIVS": {"N": ["12"]},
    "BERYTVS": {"N": ["12"]},
    "POPVLIANA": {"N": ["11"]},
    "TORANIVS": {"N": ["12"]},
    "CAMPANICVS": {"A": ["21"]},
    "LVGIVS": {"A": ["21"]},
    "CHAERESTRATVS": {"N": ["12"]},
    "TANFANA": {"N": ["11"]},
    "CATILVS": {"N": ["12"]},
    "DODONAEVS": {"A": ["21"]},
    "DIVITIACVS": {"N": ["12"]},
    "MARSVS": {"N": ["12"], "A": ["21"]},
    "BLOSSIVS": {"N": ["12"]},
    "THERMVS": {"N": ["12"]},
    "ALBANA": {"N": ["11"]},
    "GALATEA": {"N": ["11"]},
    "TABRACA": {"N": ["11"]},
    "LAODICEA": {"N": ["11"]},
    "SYRACOSIVS": {"A": ["21"]},
    "SYMAETHIVS": {"A": ["21"]},
    "AEACIDINVS": {"A": ["21"]},
    "PROCILLVS": {"N": ["12"]},
    "ANICIANVS": {"A": ["21"]},
    "AQVITANVS": {"A": ["21"], "N": ["12"]},
    "CIMBERIVS": {"N": ["12"]},
    "MOLORCHAEVS": {"A": ["21"]},
    "CISPIANVS": {"A": ["21"]},
    "ADGANDESTRIVS": {"N": ["12"]},
    "PARDALISCA": {"N": ["11"]},
    "CARPVS": {"N": ["12"]},
    "PLANCINA": {"N": ["11"]},
    "BERECYNTIVS": {"A": ["21"]},
    "CETHEGVS": {"N": ["12"]},
    "STROBILVS": {"N": ["12"]},
    "DINDYMIS": {"A": ["24"]},
    "ILLVRGAVONENSIS": {"A": ["24"]},
    "SOHAEMVS": {"N": ["12"]},
    "STROPHIVS": {"N": ["12"]},
    "FVNDANVS": {"N": ["12"], "A": ["21"]},
    "MEGARA": {"N": ["11"], "O": ["11"]},
    "AEOLVS": {"N": ["12"]},
    "LIBVRNICVS": {"A": ["21"]},
    "MEGALOBVLVS": {"N": ["12"]},
    "ALLIENVS": {"N": ["12"]},
    "THARSIMACHVS": {"N": ["12"]},
    "NONACRINVS": {"A": ["21"]},
    "AGYIEVS": {"N": ["12"]},
    "RIPHAEVS": {"A": ["21"]},
    "MEDOBRIGA": {"N": ["11"]},
    "CAESARIANVS": {"A": ["21"]},
    "MACRINVS": {"N": ["12"]},
    "BELIVS": {"N": ["12"]},
    "VOLVSENVS": {"N": ["12"]},
    "SAFINIVS": {"N": ["12"]},
    "HIBERIA": {"N": ["11"]},
    "PAPVS": {"N": ["12"]},
    "ILVS": {"N": ["12"]},
    "GNOSIS": {"A": ["24"], "N": ["16"]},
    "AGYLLINVS": {"A": ["21"]},
    "VERGILIVS": {"N": ["12"]},
    "RAETIA": {"N": ["11"]},
    "ELAPTONIVS": {"N": ["12"]},
    "LASERPICIARIVS": {"A": ["21"]},
    "SABAEA": {"N": ["11"]},
    "RVTVPINVS": {"N": ["12"]},
    "PISONIANVS": {"A": ["21"]},
    "SAXA": {"N": ["11"]},
    "THALIARCHVS": {"N": ["12"]},
    "AGRIPPINA": {"N": ["11"]},
    "TVCCIA": {"N": ["11"]},
    "APONIANVS": {"N": ["12"]},
    "PAXAEA": {"N": ["11"]},
    "PATROCLVS": {"N": ["12"]},
    "BROMIVS": {"N": ["12"]},
    "CARTEINENSIS": {"A": ["24"]},
    "CINGA": {"N": ["11"]},
    "BALBINVS": {"N": ["12"]},
    "AEGISTHVS": {"N": ["12"]},
    "ELISSA": {"N": ["11"]},
    "PERRHAEBIA": {"N": ["11"]},
    "OEOBAZVS": {"N": ["12"]},
    "SAGVNTINVS": {"A": ["21"]},
    "ATHAMANVS": {"A": ["21"]},
    "MELANIPPA": {"N": ["11"]},
    "STYMPHALIS": {"A": ["24"], "N": ["16"]},
    "IVNIA": {"N": ["11"]},
    "VENAFRANVS": {"A": ["21"]},
    "SYCHAEVS": {"A": ["21"], "N": ["12"]},
    "HOMERICVS": {"A": ["21"]},
    "ALSIENSIS": {"A": ["24"]},
    "ILLYRICVS": {"A": ["21"]},
    "APHOBETVS": {"N": ["12"]},
    "CAVCASEVS": {"A": ["21"]},
    "ONESICRITVS": {"N": ["12"]},
    "ACADEMIA": {"N": ["11"]},
    "VALERIA": {"N": ["11"]},
    "ARTABAZVS": {"N": ["12"]},
    "VNOMAMMIVS": {"A": ["21"]},
    "POLYCLETVS": {"N": ["12"]},
    "HIBERNIA": {"N": ["11"]},
    "FLAVINIVS": {"A": ["21"]},
    "PENELOPEVS": {"A": ["21"]},
    "CVRENSIS": {"A": ["24"]},
    "TERETINA": {"N": ["11"]},
    "COMANA": {"N": ["11"]},
    "HYACINTVS": {"N": ["12"]},
    "AMISIA": {"N": ["11"]},
    "CALLIMACHVS": {"N": ["12"]},
    "VISTILIVS": {"N": ["12"]},
    "CYNTHIVS": {"A": ["21"], "N": ["12"]},
    "PEVCOLAVS": {"N": ["12"]},
    "SOTERICVS": {"N": ["12"]},
    "FVNISVLANVS": {"N": ["12"]},
    "BITHVS": {"N": ["12"]},
    "AMINNEVS": {"A": ["21"]},
    "PHENEVS": {"N": ["12"]},
    "FIDENA": {"N": ["11"]},
    "MYRTILVS": {"N": ["12"]},
    "SVRVS": {"N": ["12"]},
    "ASVDIVS": {"N": ["12"]},
    "ANAGNINVS": {"A": ["21"]},
    "CADMEIVS": {"A": ["21"]},
    "CYDONIVM": {"N": ["12"]},
    "VALGIVS": {"N": ["12"]},
    "LAIVS": {"N": ["12"]},
    "SYRVS": {"N": ["12"], "A": ["21"]},
    "LYCIVS": {"N": ["12"], "A": ["21"]},
    "NOVARIA": {"N": ["11"]},
    "MINCIVS": {"N": ["12"]},
    "PACIDEIANVS": {"N": ["12"]},
    "THESSALONICA": {"N": ["11"]},
    "AFRICANA": {"N": ["11"]},
    "MASSICVS": {"A": ["21"], "N": ["12"]},
    "PYTHAGOREVS": {"A": ["21"]},
    "ASIVS": {"A": ["21"], "N": ["12"]},
    "CARMANIA": {"N": ["11"]},
    "MANTVA": {"N": ["11"]},
    "GARVTIANVS": {"N": ["12"]},
    "ITVRIVS": {"N": ["12"]},
    "BRIAREVS": {"N": ["12"]},
    "BRVCTERVS": {"A": ["21"]},
    "TITIANVS": {"N": ["12"]},
    "TREBELLIVS": {"N": ["12"]},
    "PACVVIVS": {"N": ["12"]},
    "BVTRA": {"N": ["11"]},
    "CHOBVS": {"N": ["12"]},
    "VINDONISSA": {"N": ["11"]},
    "MENNIVS": {"N": ["12"]},
    "AEGIALVS": {"N": ["12"]},
    "ACESTA": {"N": ["11"]},
    "DYRRACHINVS": {"A": ["21"]},
    "LYCA": {"N": ["11"]},
    "PAPIA": {"N": ["11"]},
    "RHESVS": {"N": ["12"]},
    "ORFITVS": {"N": ["12"]},
    "CISALPINVS": {"A": ["21"]},
    "GRANICVS": {"N": ["12"]},
    "COSA": {"N": ["11"]},
    "RHOETEVM": {"N": ["12"]},
    "VATICANVS": {"A": ["21"], "N": ["12"]},
    "FIRMANVS": {"A": ["21"]},
    "CADMEVS": {"A": ["21"]},
    "CHARAXVS": {"N": ["12"]},
    "FERENTINVM": {"N": ["12"]},
    "MINICIANVS": {"N": ["12"]},
    "SISOCOSTVS": {"N": ["12"]},
    "TREBELLENVS": {"N": ["12"]},
    "GAVRANVS": {"A": ["21"]},
    "PLANCIVS": {"N": ["12"]},
    "FRISIVS": {"A": ["21"]},
    "APERENSIS": {"A": ["24"]},
    "EVMOLPVS": {"N": ["12"]},
    "NAVPLIVS": {"N": ["12"]},
    "FADIA": {"N": ["11"]},
    "ASCLA": {"N": ["11"]},
    "SITHONIS": {"A": ["24"]},
    "LICINIA": {"N": ["11"]},
    "PIRITHOVS": {"N": ["12"]},
    "ARRIANVS": {"N": ["12"]},
    "TVRSELIVS": {"N": ["12"]},
    "CORMA": {"N": ["11"]},
    "SOGDIANVS": {"A": ["21"]},
    "VIRGILIANVS": {"N": ["12"]},
    "VOLCACIVS": {"N": ["12"]},
    "NAXVS": {"N": ["12"]},
    "MNESILOCHVS": {"N": ["12"]},
    "CINYPHIVS": {"A": ["21"]},
    "ANTEIA": {"N": ["11"]},
    "PACONIVS": {"N": ["12"]},
    "ACHERVSIVS": {"A": ["21"]},
    "VINICIVS": {"N": ["12"]},
    "MARVLLVS": {"N": ["12"]},
    "ENCOLPIVS": {"N": ["12"]},
    "PICENVS": {"A": ["21"]},
    "POENINVS": {"A": ["21"]},
    "BRIXIA": {"N": ["11"]},
    "OSCVS": {"A": ["21"]},
    "MVLVIVS": {"A": ["21"], "N": ["12"]},
    "PALPELLIVS": {"N": ["12"]},
    "NOVATVS": {"N": ["12"]},
    "SYBILLA": {"N": ["11"]},
    "REBILVS": {"N": ["12"]},
    "FVRVNCVLVS": {"1": ["12"]},
    "CARPOPHORVS": {"N": ["12"]},
    "ARIADNEVS": {"A": ["21"]},
    "TVRVLLIVS": {"N": ["12"]},
    "SESTVS": {"N": ["12"]},
    "AEOLIVS": {"A": ["21"]},
    "TVRRANIVS": {"N": ["12"]},
    "OTACILIVS": {"N": ["12"]},
    "OGYGIVS": {"A": ["21"]},
    "ELIVS": {"N": ["12"]},
    "SETINVM": {"N": ["12"]},
    "VXELLODVNVM": {"N": ["12"]},
    "RHADAMANTHVS": {"N": ["12"]},
    "BYRSA": {"N": ["11"]},
    "HAEMONIVS": {"A": ["21"]},
    "PANVRGVS": {"N": ["12"]},
    "ATELLANA": {"N": ["11"]},
    "EXAERAMBVS": {"N": ["12"]},
    "SAFINIANVS": {"A": ["21"]},
    "SVESSANVS": {"A": ["21"]},
    "APINIVS": {"N": ["12"]},
    "NVMITORIA": {"N": ["11"]},
    "PARADA": {"N": ["11"]},
    "PERIMEDEVS": {"A": ["21"]},
    "CHRYSALVS": {"N": ["12"]},
    "SIBYRTIVS": {"N": ["12"]},
    "SOSIBIVS": {"N": ["12"]},
    "SEQVANVS": {"A": ["21"]},
    "ITHACENSIS": {"A": ["24"]},
    "TIMAVVS": {"N": ["12"]},
    "CIRRHAEVS": {"A": ["21"]},
    "STALAGMVS": {"N": ["12"]},
    "ENNOSIGAEVS": {"N": ["12"]},
    "PARTHIA": {"N": ["11"]},
    "COSCONIVS": {"N": ["12"]},
    "SCIRONIS": {"A": ["24"]},
    "APPIANVS": {"A": ["21"], "N": ["12"]},
    "AGONALIS": {"A": ["24"]},
    "COA": {"N": ["11"]},
    "VICETIA": {"N": ["11"]},
    "ATHENA": {"N": ["11"]},
    "PATALIA": {"N": ["11"]},
    "AVENTICVM": {"N": ["12"]},
    "CALVENTIVS": {"A": ["21"], "N": ["12"]},
    "TEANVM": {"N": ["12"]},
    "BRIXIANVS": {"A": ["21"]},
    "PHALANTHVS": {"N": ["12"]},
    "CYPRIA": {"N": ["11"]},
    "MYTILENAEVS": {"A": ["21"]},
    "IVDA": {"N": ["11"]},
    "CODRVS": {"N": ["12"]},
    "TITYRVS": {"N": ["12"]},
    "MOLOSSVS": {"N": ["12"], "A": ["21"]},
    "BRISAEVS": {"N": ["12"]},
    "DODONA": {"N": ["11"]},
    "FVRNIVS": {"N": ["12"]},
    "ATTICINVS": {"N": ["12"]},
    "AVTRONIVS": {"N": ["12"]},
    "ROSIVS": {"N": ["12"]},
    "BALBVTIVS": {"N": ["12"]},
    "VRBINIA": {"N": ["11"]},
    "AVERNALIS": {"A": ["24"]},
    "ALBANVM": {"N": ["12"]},
    "SYRIACVS": {"A": ["21"]},
    "CINYRA": {"N": ["11"]},
    "PINARVS": {"N": ["12"]},
    "GAZABA": {"N": ["11"]},
    "INOVS": {"A": ["21"]},
    "PAETINA": {"N": ["11"]},
    "CHAONIA": {"N": ["11"]},
    "MARIANVS": {"A": ["21"]},
    "TANAGRAEVS": {"A": ["21"]},
    "PARTHENIVS": {"N": ["12"], "A": ["21"]},
    "ERYTHRVS": {"N": ["12"]},
    "CONCANVS": {"A": ["21"]},
    "BESSICVS": {"A": ["21"]},
    "APPVLEIA": {"N": ["11"]},
    "MAMERCVS": {"N": ["12"]},
    "BARBIVS": {"N": ["12"]},
    "MAVRVSIVS": {"A": ["21"]},
    "MILONIVS": {"N": ["12"]},
    "ACTVMERVS": {"N": ["12"]},
    "TIBERINIS": {"A": ["24"]},
    "CAYCVS": {"N": ["12"]},
    "MALEA": {"N": ["11"]},
    "ISIACVS": {"A": ["21"]},
    "CENTAVREVS": {"A": ["21"]},
    "VORANVS": {"N": ["12"]},
    "MVTVSCA": {"N": ["11"]},
    "PRVSIADA": {"N": ["11"]},
    "APVLEIVS": {"N": ["12"]},
    "OLENNIVS": {"N": ["12"]},
    "ALPINVS": {"N": ["12"], "A": ["21"]},
    "GORDIVS": {"N": ["12"]},
    "ARCETIVS": {"N": ["12"]},
    "LVSITANVS": {"A": ["21"], "N": ["12"]},
    "ALSVS": {"N": ["12"]},
    "ANIENVS": {"A": ["21"]},
    "PHLEGRAEVS": {"A": ["21"]},
    "ALBINGAVNVM": {"N": ["12"]},
    "GALVS": {"N": ["12"]},
    "COTVATVS": {"N": ["12"]},
    "ARVERNVS": {"A": ["21"]},
    "AETHALVS": {"N": ["12"]},
    "TANAGRA": {"N": ["11"]},
    "DINDYMA": {"N": ["11"]},
    "IVGVRTHINVS": {"A": ["21"]},
    "VOTIENVS": {"N": ["12"]},
    "SEMPRONIA": {"N": ["11"]},
    "PVLFENIVS": {"N": ["12"]},
    "STRYMONIVS": {"A": ["21"], "N": ["12"]},
    "STLOGA": {"N": ["11"]},
    "PLANCVS": {"N": ["12"]},
    "HYDAEPA": {"N": ["11"]},
    "HERACLITVS": {"N": ["12"]},
    "ACRATVS": {"N": ["12"]},
    "REMVRIA": {"N": ["11"]},
    "DEMOLEVS": {"N": ["12"]},
    "ARCHIPPVS": {"N": ["12"]},
    "SARDOVS": {"A": ["21"]},
    "LEVCOPETRA": {"N": ["11"]},
    "CAMPANEVS": {"A": ["21"]},
    "CYBELVS": {"N": ["12"]},
    "NABATHAEVS": {"A": ["21"]},
    "SCETANIVS": {"N": ["12"]},
    "CATIA": {"N": ["11"]},
    "PYTHAGORICVS": {"A": ["21"]},
    "CHOERILVS": {"N": ["12"]},
    "PVLLIA": {"N": ["11"]},
    "CERINTHVS": {"N": ["12"]},
    "TYNDARVS": {"N": ["12"]},
    "PAEONIVS": {"A": ["21"]},
    "RHAEBVS": {"N": ["12"]},
    "PELOPEIVS": {"A": ["21"]},
    "POLYTIMETVS": {"N": ["12"]},
    "LEBEDVS": {"N": ["12"]},
    "CETRIVS": {"N": ["12"]},
    "ARVSEIVS": {"N": ["12"]},
    "NIGIDIVS": {"N": ["12"]},
    "CHAONIS": {"A": ["24"]},
    "GERGOVIA": {"N": ["11"]},
    "MOSA": {"N": ["11"]},
    "CENTAVROMACHIA": {"N": ["11"]},
    "LVCANIVS": {"A": ["21"], "N": ["12"]},
    "AMANVS": {"N": ["12"]},
    "LERNA": {"N": ["11"]},
    "EVRYTHMVS": {"N": ["12"]},
    "MARMARICVS": {"A": ["21"]},
    "MATTIVM": {"N": ["12"]},
    "HELLESPONTIVS": {"A": ["21"]},
    "CORINTHIA": {"N": ["11"]},
    "SELEVCEA": {"N": ["11"]},
    "DECIVS": {"N": ["12"]},
    "AEGALEVS": {"N": ["12"]},
    "ANCONA": {"N": ["11"]},
    "PITVANIVS": {"N": ["12"]},
    "GALLONIVS": {"N": ["12"]},
    "SYBYLLA": {"N": ["11"]},
    "ANTEMNA": {"N": ["11"]},
    "RAEDARIVS": {"1": ["12"]},
    "EVBOEA": {"N": ["11"]},
    "ACIDALIA": {"N": ["11"]},
    "HYRCANVS": {"A": ["21"]},
    "IDALIVS": {"A": ["21"]},
    "ASPAVIA": {"N": ["11"]},
    "PHOCAEVS": {"A": ["21"]},
    "HERCVLANEVS": {"N": ["12"], "A": ["21"]},
    "FIDVLIVS": {"N": ["12"]},
    "TVSCILIVS": {"N": ["12"]},
    "ANA": {"N": ["11"]},
    "CLVSIVS": {"N": ["12"]},
    "PERGAMA": {"N": ["11"]},
    "AEQVIMAELIVM": {"N": ["12"]},
    "APENNINICOLA": {"N": ["11"]},
    "SORA": {"N": ["11"]},
    "IOLAVS": {"N": ["12"]},
    "RADAMISTVS": {"N": ["12"]},
    "INSTEIVS": {"N": ["12"]},
    "TANVSIVS": {"N": ["12"]},
    "VEHILIVS": {"N": ["12"]},
    "MACAREVS": {"N": ["12"]},
    "COMPITALIA": {"-": ["11"], "N": ["12"]},
    "CAESERNIVS": {"N": ["12"]},
    "PANCHAIA": {"N": ["11"]},
    "CARSIDIVS": {"N": ["12"]},
    "EVNOMIA": {"N": ["11"]},
    "MEMMIVS": {"N": ["12"]},
    "HECVBA": {"N": ["11"]},
    "FERENTVM": {"N": ["12"]},
    "CLVENTIA": {"N": ["11"]},
    "PYRRHA": {"N": ["11"]},
    "COLCHIS": {"N": ["13"], "A": ["21"]},
    "HYRCANIA": {"N": ["11"]},
    "METLOSEDVM": {"N": ["12"]},
    "PANOPEA": {"N": ["11"]},
    "IDALIA": {"N": ["11"]},
    "HEGELOCHVS": {"N": ["12"]},
    "PELOPONNESIACVS": {"A": ["21"]},
    "ALBVNEA": {"N": ["11"]},
    "IASONIVS": {"A": ["21"]},
    "VISELLIVS": {"N": ["12"]},
    "CRODVNVM": {"N": ["12"]},
    "QVINTILIA": {"N": ["11"]},
    "PERGAMENVS": {"A": ["21"]},
    "DOMITIANVS": {"N": ["12"], "A": ["21"]},
    "MARACANDA": {"N": ["11"]},
    "AECVS": {"N": ["12"]},
    "AVRVNCA": {"N": ["11"]},
    "EPPIVS": {"N": ["12"]},
    "PARRHASIVS": {"A": ["21"], "N": ["12"]},
    "DEMIANVS": {"N": ["12"]},
    "ARRETINVS": {"A": ["21"]},
    "TIBVRNVS": {"A": ["21"], "N": ["12"]},
    "AEMILIA": {"N": ["11"]},
    "FORNACALIA": {"N": ["11"]},
    "CLINIA": {"N": ["11"]},
    "NVMIDVS": {"A": ["21"]},
    "LAVINIVS": {"A": ["21"]},
    "VRGVLANIA": {"N": ["11"]},
    "PHRIXVS": {"N": ["12"]},
    "VIBIDIA": {"N": ["11"]},
    "LVPIA": {"N": ["11"]},
    "ALBIANVS": {"A": ["21"]},
    "BACTRVS": {"N": ["12"]},
    "TOLOSA": {"N": ["11"]},
    "TAMPILVS": {"N": ["12"]},
    "AMPHRYSVS": {"N": ["12"]},
    "GLOBVLVS": {"-": ["12"], "N": ["12"]},
    "FIDICVLANIVS": {"N": ["12"]},
    "MESOPOTAMIA": {"N": ["11"]},
    "QVINTA": {"N": ["11"]},
    "ANIENSIS": {"A": ["24"]},
    "BACCHIVS": {"N": ["12"]},
    "ELEGEIA": {"N": ["11"]},
    "SIDONICVS": {"A": ["21"]},
    "LABIENVS": {"N": ["12"]},
    "THERODAMANTEVS": {"A": ["21"]},
    "LAPPIVS": {"N": ["12"]},
    "BODOTRIA": {"N": ["11"]},
    "ROMANIENSIS": {"A": ["24"]},
    "CAECILIA": {"N": ["11"]},
    "ARIONIVS": {"A": ["21"]},
    "THRONIVS": {"N": ["12"]},
    "DAVNIVS": {"A": ["21"]},
    "BALBILLVS": {"N": ["12"]},
    "STHENELVS": {"N": ["12"]},
    "MINOVS": {"A": ["21"]},
    "TREVIRICVS": {"A": ["21"]},
    "PHINEIVS": {"A": ["21"]},
    "SESCENTOPLAGVS": {"N": ["12"]},
    "QVERQVETVLANVS": {"N": ["12"]},
    "VECELLINVS": {"N": ["12"]},
    "TARRACONENSIS": {"A": ["24"]},
    "COCCEIANVS": {"N": ["12"]},
    "AEFVLA": {"N": ["11"]},
    "CARNVTIS": {"A": ["24"]},
    "SALLVSTIANVS": {"A": ["21"]},
    "PODALIRIVS": {"N": ["12"]},
    "INDVTIOMARVS": {"N": ["12"]},
    "SANGARIVS": {"N": ["12"]},
    "MOENVS": {"N": ["12"]},
    "LIPARAEVS": {"A": ["21"]},
    "INTERAMNA": {"N": ["11"]},
    "CAESENNIA": {"N": ["11"]},
    "ARTEMITA": {"N": ["11"]},
    "LVCANVS": {"A": ["21"], "N": ["12"]},
    "ALEXANDREA": {"N": ["11"]},
    "CASTALIA": {"N": ["11"]},
    "ERYTHRAEVS": {"A": ["21"]},
    "RVFRIVM": {"N": ["12"]},
    "THERMOPOLIVM": {"N": ["12"]},
    "SIRPICVS": {"N": ["12"]},
    "ERECHTHEVS": {"N": ["12"]},
    "APOLLONIDENSIS": {"A": ["24"]},
    "TEIVS": {"A": ["21"]},
    "ILLYRIA": {"N": ["11"]},
    "CYNTHVS": {"N": ["12"]},
    "MVSONIVS": {"N": ["12"]},
    "PACTOLVS": {"N": ["12"]},
    "AQVILEIENSIS": {"A": ["24"]},
    "SVSIS": {"A": ["24"]},
    "TROICA": {"N": ["11"]},
    "HELLESPONTIACVS": {"A": ["21"]},
    "TIGELLIVS": {"N": ["12"]},
    "RHEGINVS": {"A": ["21"]},
    "CISPIVS": {"N": ["12"]},
    "TERENTIA": {"N": ["11"]},
    "SICANVS": {"A": ["21"]},
    "IARBA": {"N": ["11"]},
    "COMVM": {"N": ["12"]},
    "PHARMACVS": {"N": ["12"]},
    "ANDRONICVS": {"N": ["12"]},
    "NELEIVS": {"A": ["21"]},
    "ABYDVS": {"N": ["12"]},
    "RIGODVLVM": {"N": ["12"]},
    "METHYMNA": {"N": ["11"]},
    "PACIVS": {"N": ["12"]},
    "GYARVS": {"N": ["12"]},
    "BOVILLANVS": {"A": ["21"]},
    "CERAVNIA": {"N": ["11"]},
    "IONICVS": {"A": ["21"]},
    "GARGONIVS": {"N": ["12"]},
    "NICAEA": {"N": ["11"]},
    "CENCHREVS": {"N": ["12"]},
    "CYLLENA": {"N": ["11"]},
    "IVDAEVS": {"A": ["21"]},
    "HEGESIMACHVS": {"N": ["12"]},
    "TYCHIVS": {"N": ["12"]},
    "SENTICVS": {"A": ["21"]},
    "CELADVS": {"N": ["12"]},
    "TARPA": {"N": ["11"]},
    "HELVINVS": {"A": ["21"]},
    "BABYLONIA": {"N": ["11"]},
    "ZEVGMA": {"N": ["11"]},
    "VESVLVS": {"N": ["12"]},
    "VRSIDIVS": {"N": ["12"]},
    "PLAETORIVS": {"A": ["21"], "N": ["12"]},
    "HAEMVS": {"N": ["12"]},
    "RHENVS": {"N": ["12"]},
    "XENIPPA": {"N": ["11"]},
    "VENALICIVS": {"1": ["12"]},
    "AMASTRVS": {"N": ["12"]},
    "HISTRIA": {"N": ["11"]},
    "MACRINA": {"N": ["11"]},
    "LICINIANVS": {"A": ["21"], "N": ["12"]},
    "COENVS": {"N": ["12"]},
    "ASINARIA": {"N": ["11"]},
    "VALERIANVS": {"A": ["21"]},
    "IGVIVM": {"N": ["12"]},
    "HELENVS": {"N": ["12"]},
    "CVRIVS": {"N": ["12"]},
    "SICELIS": {"A": ["24"]},
    "ALEIVS": {"A": ["21"]},
    "PENTHEVS": {"N": ["12"]},
    "AHALA": {"N": ["11"]},
    "ERILLVS": {"N": ["12"]},
    "PASSENNVS": {"N": ["12"]},
    "CENABENSIS": {"A": ["24"]},
    "DIOXENVS": {"N": ["12"]},
    "POSTVERTA": {"N": ["11"]},
    "CISRHENANVS": {"A": ["21"]},
    "SANGA": {"N": ["11"]},
    "NORTIA": {"N": ["11"]},
    "THEOPHILVS": {"N": ["12"]},
    "ELECTRVS": {"N": ["12"]},
    "BITHYNVS": {"A": ["21"]},
    "VENETIA": {"N": ["11"]},
    "PALLADIVS": {"A": ["21"]},
    "RHINOCOLVRA": {"N": ["11"]},
    "PERSEVS": {"N": ["12"], "A": ["21"]},
    "SCYTHA": {"N": ["11"]},
    "PROTESILAVS": {"N": ["12"]},
    "BABA": {"N": ["11"]},
    "DIA": {"N": ["11"]},
    "GRATIDIVS": {"N": ["12"]},
    "OROANDICVS": {"A": ["21"]},
    "PAPHLAGONA": {"N": ["11"]},
    "PRISCINVS": {"N": ["12"]},
    "BELGICVS": {"A": ["21"]},
    "TREBVLANVS": {"A": ["21"]},
    "STERTINIVS": {"N": ["12"], "A": ["21"]},
    "DECETIA": {"N": ["11"]},
    "LYDVS": {"A": ["21"], "N": ["12"]},
    "PALILIA": {"N": ["11"]},
    "BIRRIVS": {"N": ["12"]},
    "PHRYGIVS": {"A": ["21"]},
    "AMEANA": {"N": ["11"]},
    "LYGDVS": {"N": ["12"]},
    "FANNIA": {"N": ["11"]},
    "ERIGONEIVS": {"A": ["21"]},
    "COPONIVS": {"N": ["12"]},
    "VIRIATVS": {"N": ["12"]},
    "EVNAEVS": {"N": ["12"]},
    "ARGOVS": {"A": ["21"]},
    "SARRANVS": {"A": ["21"]},
    "FALCIDIANVS": {"A": ["21"]},
    "ABASCANTVS": {"N": ["12"]},
    "ANDROGEONEVS": {"A": ["21"]},
    "HAMILLVS": {"N": ["12"]},
    "LENAEVS": {"A": ["21"], "N": ["12"]},
    "HERMVS": {"N": ["12"]},
    "CAEPASIVS": {"N": ["12"]},
    "DARDANVS": {"A": ["21"], "N": ["12"]},
    "VARIA": {"N": ["11"]},
    "ARCADIVS": {"A": ["21"]},
    "BETILIENVS": {"N": ["12"]},
    "HIRPINVS": {"A": ["21"], "N": ["12"]},
    "CATVVOLCVS": {"N": ["12"]},
    "DEIDAMIA": {"N": ["11"]},
    "SACRANVS": {"A": ["21"]},
    "CLOACINA": {"N": ["11"]},
    "RIPAEVS": {"A": ["21"]},
    "RVFA": {"N": ["11"]},
    "VIBIENVS": {"N": ["12"]},
    "MAZAEVS": {"N": ["12"]},
    "CIRRHA": {"N": ["11"]},
    "TAVROMENVM": {"N": ["12"]},
    "PERRHAEBVS": {"A": ["21"]},
    "MEGADORVS": {"N": ["12"]},
    "VIBIA": {"N": ["11"]},
    "EPEVS": {"N": ["12"]},
    "PHIDIACVS": {"A": ["21"]},
    "ECBATANA": {"N": ["11"]},
    "PETRAEVS": {"N": ["12"]},
    "MEMORATVS": {"1": ["12"]},
    "SASERNA": {"N": ["11"]},
    "CRVSTVMINVM": {"N": ["12"]},
    "DORIS": {"N": ["13"], "A": ["12"]},
    "VERANIA": {"N": ["11"]},
    "ICELVS": {"N": ["12"]},
    "MAENALA": {"N": ["11"]},
    "CLITARCHVS": {"N": ["12"]},
    "GERYONACEVS": {"A": ["21"]},
    "ALCESIMVS": {"N": ["12"]},
    "SILIANVS": {"A": ["21"]},
    "PARNASSIVS": {"A": ["21"]},
    "PICENTINVS": {"A": ["21"]},
    "BONNA": {"N": ["11"]},
    "MENOPHILA": {"N": ["11"]},
    "MONOECVS": {"N": ["12"]},
    "PISTOCLERVS": {"N": ["12"]},
    "NAVTACA": {"N": ["11"]},
    "MAMVRRA": {"N": ["11"]},
    "CONCONNETODVMNVS": {"N": ["12"]},
    "ANDROMACHVS": {"N": ["12"]},
    "VOLTINIENSIS": {"A": ["24"]},
    "ITIVS": {"N": ["12"]},
    "CLEONICVS": {"N": ["12"]},
    "SITTIVS": {"N": ["12"]},
    "VARILLVS": {"N": ["12"]},
    "CNEVS": {"N": ["12"]},
    "APOLLINEVS": {"A": ["21"]},
    "DEIPHOBVS": {"N": ["12"]},
    "HERMOBIVS": {"N": ["12"]},
    "FREGELLANVS": {"A": ["21"], "N": ["12"]},
    "PHAEACIVS": {"A": ["21"]},
    "ITYLVS": {"N": ["12"]},
    "CVPENCVS": {"N": ["12"]},
    "ORBIVS": {"N": ["12"]},
    "VARGVNTEIVS": {"N": ["12"]},
    "COSANVM": {"N": ["12"]},
    "CRISPINA": {"N": ["11"]},
    "CYBELEIVS": {"A": ["21"]},
    "PYRAGRVS": {"N": ["12"]},
    "TAXIMAGVLVS": {"N": ["12"]},
    "ASSACANVS": {"N": ["12"]},
    "SESTIANVS": {"A": ["21"]},
    "PVTEOLANVS": {"A": ["21"]},
    "GEMELLINVS": {"N": ["12"]},
    "GLAVCIPPVS": {"N": ["12"]},
    "ASIANVS": {"A": ["21"]},
    "DELLIVS": {"N": ["12"]},
    "VOLOGAESVS": {"N": ["12"]},
    "TRICCA": {"N": ["11"]},
    "LOLLIA": {"N": ["11"]},
    "CEA": {"N": ["11"]},
    "TIBVRTINVS": {"A": ["21"]},
    "HISPANIENSIS": {"A": ["24"]},
    "EDESSA": {"N": ["11"]},
    "TREBVLA": {"N": ["11"]},
    "PACTVMEVS": {"N": ["12"]},
    "SYMAETHEVS": {"A": ["21"]},
    "SYMPOSIVM": {"N": ["12"]},
    "AQVILEIA": {"N": ["11"]},
    "ACHERVNTIA": {"N": ["11"]},
    "MARICCVS": {"N": ["12"]},
    "OXVS": {"N": ["12"]},
    "CALVINVS": {"N": ["12"]},
    "MYRTOVS": {"A": ["21"]},
    "TELANVS": {"A": ["21"]},
    "MOSCHVS": {"N": ["12"]},
    "HOMERISTA": {"N": ["11"]},
    "LYRNESIVS": {"A": ["21"]},
    "PLATAGIDORVS": {"N": ["12"]},
    "IGILIVM": {"N": ["12"]},
    "SCYRIVS": {"A": ["21"]},
    "DORYLENSIS": {"A": ["24"]},
    "TAVNVS": {"N": ["12"]},
    "CRATINVS": {"N": ["12"]},
    "SERVAEVS": {"N": ["12"]},
    "HYLLEVS": {"A": ["21"]},
    "PLISTHENIVS": {"A": ["21"]},
    "LETHAEVS": {"A": ["21"]},
    "SVEBICVS": {"A": ["21"]},
    "ORICIVS": {"A": ["21"]},
    "CAECINA": {"N": ["11"]},
    "THRASEA": {"N": ["11"]},
    "ERICHTHONIVS": {"N": ["12"], "A": ["21"]},
    "PLACENTIA": {"N": ["11"]},
    "SAMVS": {"N": ["12"]},
    "VERONENSIS": {"A": ["24"]},
    "VICANVS": {"1": ["12"]},
    "CAMVRTIVS": {"N": ["12"]},
    "NVMANTINA": {"N": ["11"]},
    "IRVS": {"N": ["12"]},
    "METANIRA": {"N": ["11"]},
    "DEIOTARVS": {"N": ["12"]},
    "HAGNA": {"N": ["11"]},
    "PEGASVS": {"N": ["12"]},
    "AQVITANIA": {"N": ["11"]},
    "VIENNA": {"N": ["11"]},
    "HYDROCHOEVS": {"N": ["12"]},
    "DESVLTORIVS": {"1": ["12"]},
    "OBVLTRONIVS": {"N": ["12"]},
    "NASVA": {"N": ["11"]},
    "GLAPHYRVS": {"N": ["12"]},
    "TAYGETVS": {"N": ["12"]},
    "TRIPHALLVS": {"N": ["12"]},
    "ALCAEVS": {"N": ["12"]},
    "PETRONIA": {"N": ["11"]},
    "ORCYNIA": {"N": ["11"]},
    "PETRIANVS": {"A": ["21"]},
    "FORMIANVS": {"A": ["21"]},
    "THREICIVS": {"A": ["21"]},
    "CAELIANVS": {"N": ["12"], "A": ["21"]},
    "TROIVS": {"A": ["21"]},
    "EVROPAEVS": {"A": ["21"]},
    "VATIA": {"N": ["11"]},
    "PELIVS": {"A": ["21"]},
    "SEGOVIA": {"N": ["11"]},
    "FALCIDIVS": {"N": ["12"]},
    "RVFILLA": {"N": ["11"]},
    "SYRTICVS": {"A": ["21"]},
    "NASICA": {"N": ["11"]},
    "LEVCTRICVS": {"A": ["21"]},
    "ORITHYIA": {"N": ["11"]},
    "BRATVSPANTIVM": {"N": ["12"]},
    "CHIONIVS": {"A": ["21"]},
    "TENTYRA": {"N": ["11"]},
    "PYRRICHVS": {"N": ["12"]},
    "GELOVS": {"A": ["21"]},
    "THESPROTVS": {"N": ["12"]},
    "OETAEVS": {"A": ["21"]},
    "CAEPARIVS": {"N": ["12"]},
    "AVARICVM": {"N": ["12"]},
    "ORICVS": {"N": ["12"]},
    "THYNIA": {"N": ["11"]},
    "GALERIA": {"N": ["11"]},
    "MEGALOPOLITANVS": {"A": ["21"]},
    "AMPIVS": {"N": ["12"]},
    "AVFIDIVS": {"N": ["12"]},
    "THYREATIS": {"A": ["24"]},
    "ARIONVS": {"A": ["21"]},
    "PANDIONIVS": {"A": ["21"]},
    "LIBENTIA": {"N": ["11"]},
    "CORVINVS": {"N": ["12"]},
    "CENTAVRICVS": {"A": ["21"]},
    "MAMVRRIVS": {"N": ["12"]},
    "GNOSIVS": {"A": ["21"]},
    "GRYPVS": {"N": ["12"]},
    "MYCONVS": {"N": ["12"]},
    "ANTIVS": {"N": ["12"]},
    "MVLVCHA": {"N": ["11"]},
    "PTOLOMAEVS": {"N": ["12"]},
    "ACIDINVS": {"N": ["12"]},
    "ALPINIVS": {"N": ["12"]},
    "SANTONICVS": {"A": ["21"]},
    "ASINA": {"N": ["11"]},
    "AXONA": {"N": ["11"]},
    "TIGELLINVS": {"N": ["12"]},
    "TETRARCHIA": {"-": ["11"], "N": ["12"]},
    "NOTARIVS": {"1": ["12"]},
    "PVSILLA": {"N": ["11"]},
    "NVMERIVS": {"N": ["12"]},
    "MAEOTIVS": {"A": ["21"]},
    "PERILAVS": {"N": ["12"]},
    "HELVETIVS": {"A": ["21"]},
    "MOPSOPIVS": {"A": ["21"]},
    "VENVTIVS": {"N": ["12"]},
    "HYLAEVS": {"N": ["12"], "A": ["21"]},
    "CAMVLODVNVM": {"N": ["12"]},
    "LAELIANVS": {"A": ["21"]},
    "SVLLANVS": {"A": ["21"]},
    "HALOTVS": {"N": ["12"]},
    "SALONINA": {"N": ["11"]},
    "INTIMILIVM": {"N": ["12"]},
    "CNIDVS": {"N": ["12"]},
    "ROSIANVS": {"N": ["12"]},
    "OLENVS": {"N": ["12"]},
    "BOVDICCA": {"N": ["11"]},
    "DORYCLVS": {"N": ["12"]},
    "GRVMENTVM": {"N": ["12"]},
    "HYBLAEVS": {"A": ["21"]},
    "CYLLENIVS": {"A": ["21"], "N": ["12"]},
    "OEBALIA": {"N": ["11"]},
    "HIPPOLYTA": {"N": ["11"]},
    "ARSAMOSATA": {"N": ["11"]},
    "MAECIVS": {"N": ["12"]},
    "LONGARENVS": {"N": ["12"]},
    "LONGINVS": {"N": ["12"]},
    "CARA": {"N": ["11"]},
    "STHENELEIVS": {"A": ["21"]},
    "MASSICA": {"N": ["11"]},
    "GRYLLVS": {"N": ["12"]},
    "DORVS": {"N": ["12"], "A": ["21"]},
    "VISIDIVS": {"N": ["12"]},
    "IONIVS": {"A": ["21"]},
    "RVSTIANVS": {"N": ["12"]},
    "HERCEVS": {"A": ["21"]},
    "PROCHYTA": {"N": ["11"]},
    "AVSONIA": {"N": ["11"]},
    "ELYSIVM": {"N": ["12"]},
    "SEGIMVNDVS": {"N": ["12"]},
    "EPOREDIA": {"N": ["11"]},
    "CORANVS": {"N": ["12"], "A": ["21"]},
    "PANNONIVS": {"A": ["21"]},
    "AMITERNVS": {"A": ["21"]},
    "LAODAMIA": {"N": ["11"]},
    "CAYSTRVS": {"N": ["12"]},
    "LEPIDA": {"N": ["11"]},
    "NYSIVS": {"N": ["12"]},
    "NOVAESIVM": {"N": ["12"]},
    "THEODOTVS": {"N": ["12"]},
    "DIADEMATVS": {"N": ["12"]},
    "LENTIDIVS": {"N": ["12"]},
    "FISCELLVS": {"N": ["12"]},
    "IPHIANASSA": {"N": ["11"]},
    "ARCTVRVS": {"N": ["12"]},
    "BRVTTIDIVS": {"N": ["12"]},
    "COTVS": {"N": ["12"]},
    "LIMETANVS": {"N": ["12"]},
    "PELVSIACVS": {"A": ["21"]},
    "AXIA": {"N": ["11"]},
    "PTERELA": {"N": ["11"]},
    "CHAERIBVLVS": {"N": ["12"]},
    "COBIAMACHVS": {"N": ["12"]},
    "TAENARVS": {"N": ["12"]},
    "GENVSVS": {"N": ["12"]},
    "BLESAMIVS": {"N": ["12"]},
    "LYBIA": {"N": ["11"]},
    "AVIOLA": {"N": ["11"]},
    "CARYSTEVS": {"A": ["21"]},
    "MISENVS": {"N": ["12"], "A": ["21"]},
    "ACARNANIA": {"N": ["11"]},
    "PAVLINA": {"N": ["11"]},
    "CYRSILVS": {"N": ["12"]},
    "OCTAVIANIVS": {"A": ["21"]},
    "QVINCTIVS": {"N": ["12"]},
    "CLVSTVMINA": {"N": ["11"]},
    "ARPVS": {"N": ["12"]},
    "DILLIVS": {"N": ["12"]},
    "EPIGONVS": {"N": ["12"]},
    "PERGAMEVS": {"A": ["21"]},
    "STATIRA": {"N": ["11"]},
    "ARARICVS": {"N": ["12"]},
    "HERMOLAVS": {"N": ["12"]},
    "AMBRACIA": {"N": ["11"]},
    "MENTOREVS": {"A": ["21"]},
    "CRESSVS": {"A": ["21"]},
    "ISMARIVS": {"A": ["21"]},
    "HORTALVS": {"N": ["12"]},
    "ISMENVS": {"N": ["12"]},
    "AVGVSTODVNVM": {"N": ["12"]},
    "PALLANTIVS": {"A": ["21"]},
    "AEDVVS": {"A": ["21"]},
    "MVTINA": {"N": ["11"]},
    "EVRYTVS": {"N": ["12"]},
    "SVADA": {"N": ["11"]},
    "IVLLVS": {"N": ["12"]},
    "VIRIOLA": {"N": ["11"]},
    "AVFIDIENVS": {"N": ["12"]},
    "DORYPHORVS": {"N": ["12"]},
    "CAESONINVS": {"N": ["12"]},
    "LYAEVS": {"N": ["12"]},
    "ACONTIVS": {"N": ["12"]},
    "CALENVM": {"N": ["12"]},
    "SASSIA": {"N": ["11"]},
    "MESSALINA": {"N": ["11"]},
    "PETREVS": {"N": ["12"]},
    "CELEMNA": {"N": ["11"]},
    "AELIVS": {"A": ["21"], "N": ["12"]},
    "TOLBIACVM": {"N": ["12"]},
    "VETILIVS": {"N": ["12"]},
    "CAVSINIVS": {"N": ["12"]},
    "PINDARVS": {"N": ["12"]},
    "CORINTHIENSIS": {"A": ["24"]},
    "PHYLLEIS": {"A": ["24"]},
    "CASTA": {"N": ["11"]},
    "IOLCIACVS": {"A": ["21"]},
    "LIGVSTICVS": {"A": ["21"]},
    "LABIENIANVS": {"A": ["21"]},
    "ANTINOVS": {"N": ["12"]},
    "RVSTICVLVS": {"1": ["12"]},
    "PHARSALIA": {"N": ["11"]},
    "MENOECEVS": {"N": ["12"]},
    "CINYRVS": {"N": ["12"]},
    "OECONOMICVS": {"N": ["12"]},
    "HARMODIVS": {"N": ["12"]},
    "SVNIVM": {"N": ["12"]},
    "ITVRAEVS": {"A": ["21"]},
    "IONIACVS": {"A": ["21"]},
    "AGANIPPAEVS": {"A": ["21"]},
    "NEOPTOLEMVS": {"N": ["12"]},
    "CEPHALIVS": {"A": ["21"]},
    "IVVENTIVS": {"N": ["12"], "A": ["21"]},
    "DONATIVS": {"N": ["12"]},
    "BOSPORVS": {"N": ["12"]},
    "LATONIA": {"N": ["11"]},
    "CHALCIDICVS": {"A": ["21"]},
    "GANGETICVS": {"A": ["21"]},
    "CANINIVS": {"N": ["12"]},
    "POMPONIANVS": {"N": ["12"], "A": ["21"]},
    "TANTALICVS": {"A": ["21"]},
    "LABERIVS": {"N": ["12"]},
    "HELIVS": {"N": ["12"]},
    "BOLANVS": {"N": ["12"]},
    "AVRIVS": {"N": ["12"]},
    "MYGDONIVS": {"A": ["21"]},
    "CAECVBVM": {"N": ["12"]},
    "PRIVERNVM": {"N": ["12"]},
    "CABALLINVS": {"A": ["21"]},
    "ALTINVM": {"N": ["12"]},
    "CARENVS": {"A": ["21"]},
    "GAIANVS": {"A": ["21"]},
    "THRASIVS": {"N": ["12"]},
    "EVRYMACHVS": {"N": ["12"]},
    "CASINVS": {"N": ["12"]},
    "NINVS": {"N": ["12"]},
    "SENIENSIS": {"A": ["24"]},
    "BOIHAEMVM": {"N": ["12"]},
    "PAEGNIVM": {"N": ["12"]},
    "ARGEVS": {"A": ["21"]},
    "XANTHIPPVS": {"N": ["12"]},
    "VELANIVS": {"N": ["12"]},
    "CAECVLVS": {"N": ["12"]},
    "HERCVLEVS": {"A": ["21"]},
    "AMPHRYSIVS": {"A": ["21"]},
    "CARFVLENVS": {"N": ["12"]},
    "CATAMITVS": {"N": ["12"], "-": ["12"]},
    "AMPELVS": {"N": ["12"]},
    "CORYNAEVS": {"N": ["12"]},
    "LACONICVS": {"A": ["21"]},
    "ALBVCIVS": {"N": ["12"]},
    "LAOMEDONTEVS": {"A": ["21"]},
    "AQVICVLVS": {"N": ["12"]},
    "CALYDONIVS": {"A": ["21"]},
    "POPVLONIA": {"N": ["11"]},
    "MATRALIA": {"N": ["11"]},
    "IOLLA": {"N": ["11"]},
    "PHILIPPENSIS": {"A": ["24"]},
    "PHILODEMVS": {"N": ["12"]},
    "EVDOXVS": {"N": ["12"]},
    "GRYNEVS": {"A": ["21"]},
    "HESPERVS": {"N": ["12"]},
    "BERENICEVS": {"A": ["21"]},
    "LACETANIA": {"N": ["11"]},
    "OPHIONIVS": {"A": ["21"]},
    "ANCYRA": {"N": ["11"]},
    "CREMONENSIS": {"A": ["24"]},
    "ATABVLVS": {"N": ["12"]},
    "NICEPHORIVS": {"N": ["12"]},
    "CYTHERA": {"N": ["11"]},
    "ARTACANA": {"N": ["11"]},
    "ANCHEMOLVS": {"N": ["12"]},
    "SVLPICIA": {"N": ["11"]},
    "TYRRHVS": {"N": ["12"]},
    "MENVLLA": {"N": ["11"]},
    "ISAVRICVS": {"N": ["12"]},
    "BARIVM": {"N": ["12"]},
    "METELLA": {"N": ["11"]},
    "PHAEDRIA": {"N": ["11"]},
    "ANTISTIA": {"N": ["11"]},
    "ARMILLATVS": {"-": ["21"], "N": ["12"]},
    "ANDROMEDA": {"N": ["11"]},
    "PISEVS": {"A": ["21"]},
    "VARISIDIVS": {"N": ["12"]},
    "LINGONVS": {"N": ["12"]},
    "ARNIENSIS": {"A": ["24"]},
    "AGRIPPINVS": {"N": ["12"]},
    "MEGALESIA": {"N": ["11"]},
    "EQVIRRIA": {"N": ["11"]},
    "LANARIVS": {"1": ["12"], "N": ["12"]},
    "GRATTIVS": {"N": ["12"]},
    "TOXEVS": {"N": ["12"]},
    "ARECTEVS": {"A": ["21"]},
    "MVTINENSIS": {"A": ["24"]},
    "OEAGRVS": {"N": ["12"]},
    "VSTICA": {"N": ["11"]},
    "CALVISIVS": {"N": ["12"]},
    "ORFIDIVS": {"N": ["12"]},
    "CEDROSIA": {"N": ["11"]},
    "PTOLEMAEEVS": {"A": ["21"]},
    "TITANIVS": {"A": ["21"]},
    "EPIDAVRIVS": {"A": ["21"]},
    "NIPHAEVS": {"N": ["12"]},
    "FABRICIANVS": {"A": ["21"]},
    "MEGYLLA": {"N": ["11"]},
    "EXECESTVS": {"N": ["12"]},
    "TICIDA": {"N": ["11"]},
    "SATVREIANVS": {"A": ["21"]},
    "CANIVS": {"N": ["12"]},
    "VOLANERIVS": {"N": ["12"]},
    "LARIVS": {"N": ["12"]},
    "CHARVS": {"N": ["12"]},
    "MESSAPVS": {"N": ["12"]},
    "PANTILIVS": {"N": ["12"]},
    "PHLEGRA": {"N": ["11"]},
    "VACCVS": {"N": ["12"]},
    "TALLVS": {"N": ["12"]},
    "HERCVLIVS": {"A": ["21"]},
    "PROPONTIACVS": {"A": ["21"]},
    "CORFINIVM": {"N": ["12"]},
    "PLANASIA": {"N": ["11"]},
    "CALVSIDIVS": {"N": ["12"]},
    "VITELLIANVS": {"A": ["21"]},
    "ATRATINVS": {"N": ["12"]},
    "MAEANDRIVS": {"N": ["12"], "A": ["21"]},
    "EPIDICVS": {"N": ["12"]},
    "CARSEOLANVS": {"A": ["21"]},
    "MEDEIS": {"A": ["24"]},
    "OCRICVLANVS": {"A": ["21"]},
    "RESTITVTVS": {"N": ["12"]},
    "PADVA": {"N": ["11"]},
    "VIRDIVS": {"N": ["12"]},
    "BRVTTEDIVS": {"N": ["12"]},
    "ANIVS": {"N": ["12"]},
    "AVNVS": {"N": ["12"]},
    "PHRIXEVS": {"A": ["21"]},
    "CLAVDIANVS": {"A": ["21"]},
    "ENNIANVS": {"A": ["21"]},
    "QVINTIA": {"N": ["11"]},
    "CRATERVS": {"N": ["12"]},
    "ATIA": {"N": ["11"]},
    "APVLVS": {"A": ["21"]},
    "ASELLIVS": {"N": ["12"]},
    "ILIPA": {"N": ["11"]},
    "BARBATIVS": {"N": ["12"]},
    "GERELLANVS": {"N": ["12"]},
    "ASTVRICVS": {"N": ["12"]},
    "IMBRASVS": {"N": ["12"]},
    "DVLICHIVS": {"A": ["21"]},
    "ATTIVS": {"N": ["12"]},
    "CADVRCVS": {"A": ["21"]},
    "CIMINVS": {"N": ["12"]},
    "TABVRNVS": {"N": ["12"]},
    "VIBENNA": {"N": ["11"]},
    "HAEMONIA": {"N": ["11"]},
    "PHTHIA": {"N": ["11"]},
    "CRAGVS": {"N": ["12"]},
    "MVLLINVS": {"N": ["12"]},
    "CRESIVS": {"A": ["21"]},
    "VERBIGENVS": {"N": ["12"]},
    "PANDA": {"N": ["11"]},
    "ONOMASTVS": {"N": ["12"]},
    "THEOTIMVS": {"N": ["12"]},
    "OPIMIANVM": {"N": ["12"]},
    "BOXVS": {"N": ["12"]},
    "AEACVS": {"N": ["12"]},
    "EVRYPYLVS": {"N": ["12"]},
    "DOMNILAVS": {"N": ["12"]},
    "ANCAEVS": {"N": ["12"]},
    "FAESVLANVS": {"A": ["21"]},
    "HYPERMESTRA": {"N": ["11"]},
    "CARTEIA": {"N": ["11"]},
    "ELEAZARVS": {"N": ["12"]},
    "PHIGALESIVS": {"A": ["21"]},
    "SIRIVS": {"N": ["12"], "A": ["21"]},
    "CEREALIA": {"N": ["11"]},
    "RHIPEVS": {"N": ["12"]},
    "LESBIVM": {"N": ["12"]},
    "PHARSALIVS": {"A": ["21"]},
    "ASTERIA": {"N": ["11"]},
    "VARENVS": {"N": ["12"]},
    "PROPERTIVS": {"N": ["12"]},
    "SICCA": {"N": ["11"]},
    "ATREVS": {"N": ["12"]},
    "ACHARNEVS": {"N": ["12"]},
    "VOLCANVS": {"N": ["12"]},
    "EVCAERVS": {"N": ["12"]},
    "PAEONIA": {"N": ["11"]},
    "ROMVLIVS": {"N": ["12"]},
    "ACILIA": {"N": ["11"]},
    "SOLIMVS": {"N": ["12"]},
    "PRAETEXTATA": {"N": ["11"]},
    "PLAVTINVS": {"A": ["21"]},
    "NESSEVS": {"A": ["21"]},
    "GRAECINVS": {"N": ["12"]},
    "DORICVS": {"A": ["21"]},
    "DIRCAEVS": {"A": ["21"]},
    "FALERNVS": {"A": ["21"]},
    "BOEOTIVS": {"A": ["21"]},
    "GAVILLA": {"N": ["11"]},
    "PERILLVS": {"N": ["12"]},
    "BESSA": {"N": ["11"]},
    "ACRISIVS": {"N": ["12"]},
    "XANTHVS": {"N": ["12"]},
    "MAREOTICVM": {"N": ["12"]},
    "MINOIS": {"N": ["13"], "A": ["12"]},
    "BVPALVS": {"N": ["12"]},
    "SIGEVM": {"N": ["12"]},
    "NOVIA": {"N": ["11"]},
    "VOLTINIA": {"N": ["11"]},
    "RVTVLVS": {"A": ["21"]},
    "POSTVMINA": {"N": ["11"]},
    "CARCHEDONIVS": {"A": ["21"]},
    "GRATIDIANVS": {"N": ["12"]},
    "PAESTVM": {"N": ["12"]},
    "HAEDVVS": {"A": ["21"]},
    "ARGOLIS": {"A": ["24"]},
    "SILENVS": {"N": ["12"]},
    "SICANIVS": {"A": ["21"]},
    "SEMVRIVM": {"N": ["12"]},
    "PHAEACIA": {"N": ["11"]},
    "POLYXENIVS": {"A": ["21"]},
    "THASIVS": {"A": ["21"]},
    "TRINACRIA": {"N": ["11"]},
    "SCANTINIVS": {"A": ["21"]},
    "TEDIA": {"N": ["11"]},
    "SIGEVS": {"A": ["21"]},
    "TVRPILIANVS": {"N": ["12"]},
    "BANTINVS": {"A": ["21"]},
    "IVDAICVS": {"A": ["21"]},
    "BALVENTIVS": {"N": ["12"]},
    "MAENALIVS": {"A": ["21"]},
    "GARAMANTIS": {"A": ["24"]},
    "IAVOLENVS": {"N": ["12"]},
    "PHAEACVS": {"A": ["21"]},
    "ACHELOVS": {"N": ["12"]},
    "LITAVICCVS": {"N": ["12"]},
    "SASSVRA": {"N": ["11"]},
    "THIRMIDA": {"N": ["11"]},
    "EPICHARNVS": {"N": ["12"]},
    "DANDARICA": {"N": ["11"]},
    "MAGNESSVS": {"A": ["21"]},
    "PHARSALICVS": {"A": ["21"]},
    "SELEVCENSIS": {"A": ["24"]},
    "FIRMIDIVS": {"N": ["12"]},
    "THYSDRA": {"N": ["11"]},
    "ASCYLTVS": {"N": ["12"]},
    "PISAEVS": {"A": ["21"]},
    "SPARTANVS": {"A": ["21"]},
    "ASCLEPIODOTVS": {"N": ["12"]},
    "TROICVS": {"A": ["21"]},
    "MASSIVA": {"N": ["11"]},
    "SERVIANVS": {"N": ["12"]},
    "VEIANIVS": {"N": ["12"]},
    "ASCIBVRGIVM": {"N": ["12"]},
    "ERVCIVS": {"N": ["12"]},
    "VOLANDVM": {"N": ["12"]},
    "SAVREA": {"N": ["11"]},
    "ACROCERAVNIA": {"N": ["11"]},
    "ATLANTICVS": {"A": ["21"]},
    "INACHIA": {"N": ["11"]},
    "MENELAEVS": {"A": ["21"]},
    "SVBVRA": {"N": ["11"]},
    "ANNA": {"N": ["11"]},
    "POLYXENA": {"N": ["11"]},
    "BOEOTVS": {"A": ["21"]},
    "CVRICTA": {"N": ["11"]},
    "CORSICVS": {"A": ["21"]},
    "TIBVRTINVM": {"N": ["12"]},
    "METIVS": {"N": ["12"]},
    "ASCVRVM": {"N": ["12"]},
    "CILLA": {"N": ["11"]},
    "SORRENTVM": {"N": ["12"]},
    "ERASINVS": {"N": ["12"]},
    "ORBILIVS": {"N": ["12"]},
    "VCVBENSIS": {"A": ["24"]},
    "FVLGINIVS": {"N": ["12"]},
    "ARABIVS": {"A": ["21"]},
    "PHYLLEIVS": {"A": ["21"]},
    "THYMBRAEVS": {"A": ["21"], "N": ["12"]},
    "CLEOPHANTVS": {"N": ["12"]},
    "ARRETIVM": {"N": ["12"]},
    "CHAEREA": {"N": ["11"]},
    "NYSAEVS": {"A": ["21"]},
    "ARRHIDAEVS": {"N": ["12"]},
    "VENETICVS": {"A": ["21"]},
    "PHAEDROMVS": {"N": ["12"]},
    "MESSALINVS": {"N": ["12"]},
    "TIMAEVS": {"N": ["12"]},
    "PERINTHVS": {"N": ["12"]},
    "GEDVSANVS": {"A": ["21"]},
    "SEXTIA": {"N": ["11"]},
    "NEREVS": {"N": ["12"]},
    "RAVIDVS": {"N": ["12"]},
    "AEOLIA": {"N": ["11"]},
    "PHILODORVS": {"N": ["12"]},
    "VIRIDOMARVS": {"N": ["12"]},
    "PHOSPHORVS": {"N": ["12"]},
    "CAMERTINVS": {"A": ["21"]},
    "EVHENVS": {"N": ["12"]},
    "MANDVBRACIVS": {"N": ["12"]},
    "NOCTILVCA": {"N": ["11"]},
    "ERYCINA": {"N": ["11"]},
    "PELOPEIS": {"A": ["24"]},
    "PANGAEA": {"N": ["11"]},
    "CLODIANVS": {"A": ["21"], "N": ["12"]},
    "IMBRVS": {"N": ["12"]},
    "BASILICVS": {"2": ["21"], "1": ["12"]},
    "CRVCIARIVS": {"1": ["12"]},
    "NICOSTRATVS": {"N": ["12"]},
    "PLOTIVS": {"A": ["21"], "N": ["12"]},
    "NVRSIA": {"N": ["11"]},
    "THERSIPPVS": {"N": ["12"]},
    "CARISTIA": {"N": ["11"]},
    "SITHONIVS": {"A": ["21"]},
    "ROSCIVS": {"A": ["21"], "N": ["12"]},
    "ALESIA": {"N": ["11"]},
    "CELTIBERIA": {"N": ["11"]},
    "SPVRINNA": {"N": ["11"]},
    "MATTIACVS": {"A": ["21"]},
    "VELIENSIS": {"A": ["24"]},
    "FVLCINIVS": {"N": ["12"]},
    "ARCHILOCHVS": {"N": ["12"]},
    "EVGANEVS": {"A": ["21"]},
    "SMYRNAEVS": {"A": ["21"]},
    "ASCVLVM": {"N": ["12"]},
    "ARMENIVS": {"N": ["12"], "A": ["21"]},
    "QVINTILLA": {"N": ["11"]},
    "ARISTONVS": {"N": ["12"]},
    "SATVRNALICIVS": {"A": ["21"]},
    "PELIACVS": {"A": ["21"]},
    "TIBVRTIVS": {"A": ["21"], "N": ["12"]},
    "ARCHEMORVS": {"N": ["12"]},
    "GADITANVS": {"A": ["21"]},
    "PROTEVS": {"N": ["12"]},
    "TITHONVS": {"N": ["12"]},
    "FOROIVLIENSIS": {"A": ["24"]},
    "PANCHAEVS": {"A": ["21"]},
    "POMPVLENVS": {"N": ["12"]},
    "ALCMAEVS": {"N": ["12"]},
    "CALPETVS": {"N": ["12"]},
    "MOGONTIACVM": {"N": ["12"]},
    "CILICENSIS": {"A": ["24"]},
    "ZETHVS": {"N": ["12"]},
    "CASPERIA": {"N": ["11"]},
    "BRVTTIANVS": {"N": ["12"]},
    "APVLA": {"N": ["11"]},
    "SESTIVS": {"N": ["12"]},
    "VERTISCVS": {"N": ["12"]},
    "FABVLLVS": {"N": ["12"]},
    "PLATONICVS": {"A": ["21"]},
    "VADA": {"N": ["11"]},
    "MVRRANVS": {"N": ["12"]},
    "ASOPVS": {"N": ["12"]},
    "VERGINIVS": {"N": ["12"]},
    "NESTOREVS": {"A": ["21"]},
    "FVFIDIVS": {"N": ["12"]},
    "POLYBVS": {"N": ["12"]},
    "SALTIVS": {"N": ["12"]},
    "VERRITVS": {"N": ["12"]},
    "ATRIA": {"N": ["11"]},
    "TRAPEZVS": {"N": ["12"]},
    "THERSILOCHVS": {"N": ["12"]},
    "PAVCVVIVS": {"N": ["12"]},
    "TREBIVS": {"N": ["12"]},
    "AMASTRIACVS": {"A": ["21"]},
    "PAPIRVS": {"N": ["12"]},
    "MEDIOLANVM": {"N": ["12"]},
    "VELIVS": {"N": ["12"]},
    "BROCHVBELVS": {"N": ["12"]},
    "CAELICVLVS": {"N": ["12"]},
    "ISTHMIA": {"N": ["11"]},
    "LARENTALIA": {"N": ["11"]},
    "PROCVLEIVS": {"N": ["12"]},
    "NVCERIA": {"N": ["11"]},
    "SINCAIVS": {"N": ["12"]},
    "MINICIVS": {"N": ["12"]},
    "AVZEA": {"N": ["11"]},
    "THYESTVS": {"A": ["21"]},
    "ALBVRNVS": {"N": ["12"]},
    "METAVRVM": {"N": ["12"]},
    "SATVRNIA": {"N": ["11"]},
    "VENERARIVS": {"2": ["21"], "1": ["12"]},
    "PRODICVS": {"N": ["12"]},
    "CATIENVS": {"N": ["12"]},
    "HORDEONIVS": {"N": ["12"]},
    "TRAVLVS": {"N": ["12"]},
    "LAPIDARIVS": {"2": ["21"], "1": ["12"]},
    "PARNASVS": {"N": ["12"]},
    "CALVINA": {"N": ["11"]},
    "LABICANVS": {"A": ["21"]},
    "MVCIANVS": {"N": ["12"]},
    "ANTEGVA": {"N": ["11"]},
    "SCRIBONIVS": {"N": ["12"]},
    "CANVSIVM": {"N": ["12"]},
    "DIODOTVS": {"N": ["12"]},
    "VRANIA": {"N": ["11"]},
    "DELTA": {"N": ["11"]},
    "HADRIA": {"N": ["11"]},
    "CEBENNA": {"N": ["11"]},
    "MEVIVS": {"N": ["12"]},
    "DECIDIANA": {"N": ["11"]},
    "AMPHIARAEVS": {"A": ["21"]},
    "VETTIDIVS": {"N": ["12"]},
    "CAESONIA": {"N": ["11"]},
    "ATHAMANIA": {"N": ["11"]},
    "MAXIMILLA": {"N": ["11"]},
    "ISELASTICVS": {"2": ["21"], "1": ["12"]},
    "GELA": {"N": ["11"]},
    "CERCIVS": {"N": ["12"]},
    "IDA": {"N": ["11"]},
    "POPPAEANVS": {"A": ["21"]},
    "ERYTHEA": {"N": ["11"]},
    "MANLIANVS": {"A": ["21"]},
    "AFRANIVS": {"N": ["12"]},
    "DEXAMENVS": {"N": ["12"]},
    "CASTABALVS": {"N": ["12"]},
    "CITA": {"N": ["11"]},
    "CAMERINVS": {"N": ["12"], "A": ["21"]},
    "VARINIVS": {"N": ["12"]},
    "CLOANTHVS": {"N": ["12"]},
    "APPVLEIANVS": {"A": ["21"]},
    "TEVTOBVRGIENSIS": {"A": ["24"]},
    "BONONIENSIS": {"A": ["24"]},
    "ALPHESIBOEVS": {"N": ["12"]},
    "BENACVS": {"N": ["12"]},
    "RAVENNA": {"N": ["11"]},
    "VOLATERRANVS": {"A": ["21"]},
    "ITALICENSIS": {"A": ["24"]},
    "FABRATERIA": {"N": ["11"]},
    "CLYTAEMNESTRA": {"N": ["11"]},
    "COCYTVS": {"N": ["12"]},
    "HIEROSOLYMVS": {"N": ["12"]},
    "ELEVSIVM": {"N": ["12"]},
    "VETTONIANVS": {"N": ["12"]},
    "POPPAEVS": {"N": ["12"], "A": ["21"]},
    "PIRAEVS": {"N": ["12"]},
    "PHINEVS": {"A": ["21"], "N": ["12"]},
    "CAVLIVS": {"N": ["12"]},
    "BRVTTIVS": {"A": ["21"]},
    "POLYSTRATVS": {"N": ["12"]},
    "THOANTEVS": {"A": ["21"]},
    "GLITIVS": {"N": ["12"]},
    "LVCINA": {"N": ["11"]},
    "ABGARVS": {"N": ["12"]},
    "SCYTHVS": {"A": ["21"]},
    "PYRAMVS": {"N": ["12"]},
    "VORENVS": {"N": ["12"]},
    "SARDANAPALLVS": {"N": ["12"]},
    "AQVINIVS": {"N": ["12"]},
    "HELICONIVS": {"A": ["21"]},
    "NONIANVS": {"N": ["12"]},
    "OCCIA": {"N": ["11"]},
    "VINICIANVS": {"N": ["12"]},
    "CECROPIVS": {"A": ["21"]},
    "ALLIA": {"N": ["11"]},
    "ILERDA": {"N": ["11"]},
    "RVDINVS": {"A": ["21"]},
    "BALEARIS": {"A": ["24"]},
    "TERTVLLVS": {"N": ["12"]},
    "LYRNESVS": {"N": ["12"]},
    "SERESTVS": {"N": ["12"]},
    "SEXTILIA": {"N": ["11"]},
    "PHILETAEVS": {"A": ["21"]},
    "ASTRAEA": {"N": ["11"]},
    "NINNIVS": {"N": ["12"]},
    "ATRACIVS": {"A": ["21"]},
    "LATVM": {"N": ["12"]},
    "AESOPVS": {"N": ["12"]},
    "DICTYNNA": {"N": ["11"]},
    "ICARVS": {"N": ["12"]},
    "AVSONIVS": {"A": ["21"]},
    "CALLISTVS": {"N": ["12"]},
    "VERANIOLVS": {"N": ["12"]},
    "SERVILIANVS": {"A": ["21"]},
    "NAVPORTVM": {"N": ["12"]},
    "ITONVS": {"N": ["12"]},
    "CVMAEVS": {"A": ["21"]},
    "BATHYLLVS": {"N": ["12"]},
    "CYRRVS": {"N": ["12"]},
    "CLVTORIVS": {"N": ["12"]},
    "CHRYSANTHVS": {"N": ["12"]},
    "MASSVGRADA": {"N": ["11"]},
    "TARTESIVS": {"A": ["21"]},
    "LARISAEVS": {"A": ["21"]},
    "AMERIA": {"N": ["11"]},
    "OLENIVS": {"A": ["21"]},
    "FIDIVS": {"N": ["12"]},
    "ABELLA": {"N": ["11"]},
    "PONTIA": {"N": ["11"]},
    "NERVA": {"N": ["11"]},
    "SINOPENSIS": {"A": ["24"]},
    "AMAZONIVS": {"A": ["21"]},
    "ECHIONIVS": {"A": ["21"]},
    "ZANCLAEVS": {"A": ["21"]},
    "ASTRAEVS": {"N": ["12"]},
    "AFRANIANVS": {"A": ["21"]},
    "POTITIVS": {"N": ["12"]},
    "THRAECIDICA": {"N": ["11"]},
    "VESTINVS": {"A": ["21"], "N": ["12"]},
    "MELITENA": {"N": ["11"]},
    "APPIA": {"N": ["11"]},
    "CASSANDREVS": {"N": ["12"]},
    "ATEIVS": {"N": ["12"]},
    "VIPSTANVS": {"N": ["12"]},
    "PHASMA": {"N": ["11"]},
    "DACVS": {"A": ["21"]},
    "TITANIS": {"A": ["24"], "N": ["16"]},
    "ALPHEVS": {"N": ["12"], "A": ["21"]},
    "CARMELVS": {"N": ["12"]},
    "PALAESTINVS": {"A": ["21"]},
    "TVRONVS": {"A": ["21"]},
    "SEPTIMILLVS": {"N": ["12"]},
    "MAIA": {"N": ["11"]},
    "OPPIANICVS": {"N": ["12"]},
    "TELEMVS": {"N": ["12"]},
    "DERCENNVS": {"N": ["12"]},
    "CVRTISIVS": {"N": ["12"]},
    "BEDRIACENSIS": {"A": ["24"]},
    "ARRECINVS": {"N": ["12"]},
    "NATTA": {"N": ["11"]},
    "NASTA": {"N": ["11"]},
    "PINDARICVS": {"A": ["21"]},
    "AQVILONIVS": {"A": ["21"]},
    "PRILIVS": {"N": ["12"]},
    "BOCCHVS": {"N": ["12"]},
    "PALLACINVS": {"A": ["21"]},
    "LEMONIA": {"N": ["11"]},
    "MAVRA": {"N": ["11"]},
    "FVFIANVS": {"A": ["21"]},
    "MISENENSIS": {"A": ["24"]},
    "XENOPHANTVS": {"N": ["12"]},
    "PROETVS": {"N": ["12"]},
    "CVSVS": {"N": ["12"]},
    "PARSAGADA": {"N": ["11"]},
    "CAPANEVS": {"N": ["12"]},
    "SVEDIVS": {"N": ["12"]},
    "LEARCHVS": {"N": ["12"]},
    "GRACCHANVS": {"A": ["21"]},
    "ARCHIDEMVS": {"N": ["12"]},
    "TVNNIVS": {"N": ["12"]},
    "TOLENVM": {"N": ["12"]},
    "AVFIDVS": {"N": ["12"]},
    "DICTAEVS": {"A": ["21"]},
    "BRIGANTICVS": {"N": ["12"]},
    "BELLICA": {"N": ["11"]},
    "LAENIVS": {"N": ["12"]},
    "MOPSVS": {"N": ["12"]},
    "CLATERNA": {"N": ["11"]},
    "ALERNVM": {"N": ["12"]},
    "MENSARIVS": {"1": ["12"]},
    "CATVLINVS": {"N": ["12"]},
    "ABDVCILLVS": {"N": ["12"]},
    "VATINIVS": {"N": ["12"]},
    "PITTHEIVS": {"A": ["21"]},
    "SYRIVS": {"A": ["21"]},
    "GARGILIVS": {"N": ["12"]},
    "BARDAICVS": {"A": ["21"]},
    "VSSETA": {"N": ["11"]},
    "VIPSANIVS": {"A": ["21"], "N": ["12"]},
    "ANTANDRVS": {"N": ["12"]},
    "ISAEVS": {"N": ["12"]},
    "GAVRVS": {"N": ["12"]},
    "AMYCLAEVS": {"A": ["21"]},
    "AREMORICVS": {"A": ["21"]},
    "EPICVREVS": {"A": ["21"]},
    "PACONIANVS": {"N": ["12"]},
    "HEGESARETVS": {"N": ["12"]},
    "ANTICYRA": {"N": ["11"]},
    "BOSPHORVS": {"N": ["12"]},
    "DORYLAVS": {"N": ["12"]},
    "PADAEVS": {"A": ["21"]},
    "TVLLA": {"N": ["11"]},
    "CASPERIVS": {"N": ["12"]},
    "AEETAEVS": {"A": ["21"]},
    "ABDALONYMVS": {"N": ["12"]},
    "CANTABRICVS": {"A": ["21"]},
    "AELIANVS": {"N": ["12"]},
    "QVARINVS": {"N": ["12"]},
    "SALEIVS": {"N": ["12"]},
    "SIPYLVS": {"N": ["12"]},
    "SAVFEIA": {"N": ["11"]},
    "ARGYRIPA": {"N": ["11"]},
    "AVXIMVM": {"N": ["12"]},
    "PHAEDRA": {"N": ["11"]},
    "SAMOTHRACIVS": {"A": ["21"]},
    "STOICIDA": {"N": ["11"]},
    "SILVANVS": {"N": ["12"]},
    "CASSANDRA": {"N": ["11"]},
    "SELEVCIA": {"N": ["11"]},
    "THEAETETVS": {"N": ["12"]},
    "OSTORIVS": {"N": ["12"]},
    "THALLVS": {"N": ["12"]},
    "VOLSINIENSIS": {"A": ["24"]},
    "PENTHEIVS": {"A": ["21"]},
    "SARMATIA": {"N": ["11"]},
    "CALGACVS": {"N": ["12"]},
    "ATELLANICVS": {"A": ["21"]},
    "MAVRVS": {"A": ["21"]},
    "SALMONEVS": {"N": ["12"]},
    "MELANTHIVS": {"N": ["12"]},
    "PHANAEVS": {"A": ["21"]},
    "TEIA": {"N": ["11"]},
    "VISTILIA": {"N": ["11"]},
    "VOLOGESVS": {"N": ["12"]},
    "IXIONIVS": {"A": ["21"]},
    "PELOPIVS": {"A": ["21"]},
    "MASSYLVS": {"A": ["21"]},
    "BETVVS": {"N": ["12"]},
    "ALLIENSIS": {"A": ["24"]},
    "CIMBRICVS": {"A": ["21"]},
    "CIMMERIVS": {"A": ["21"]},
    "RHOETEVS": {"A": ["21"], "N": ["12"]},
    "CORDALVS": {"N": ["12"]},
    "ERICTHONIVS": {"N": ["12"]},
    "LEMONVM": {"N": ["12"]},
    "FERENTIVM": {"N": ["12"]},
    "ARCHIDAMVS": {"N": ["12"]},
    "LAGVS": {"N": ["12"]},
    "PHOENISSA": {"N": ["11"]},
    "TRIVICVM": {"N": ["12"]},
    "APHIDNA": {"N": ["11"]},
    "ISMARVS": {"N": ["12"]},
    "TRITONIA": {"N": ["11"]},
    "LIBVRNVS": {"A": ["21"]},
    "IDOMENEVS": {"N": ["12"]},
    "PHASIACVS": {"A": ["21"]},
    "CAVARINVS": {"N": ["12"]},
    "FLORALIA": {"N": ["11"]},
    "PALLANTEVS": {"A": ["21"]},
    "NAEVA": {"N": ["11"]},
    "DEMAENETVS": {"N": ["12"]},
    "SAVFEIVS": {"N": ["12"]},
    "MANCIA": {"N": ["11"]},
    "LEBINTHVS": {"N": ["12"]},
    "THVBVRSICVS": {"A": ["21"]},
    "GENAVA": {"N": ["11"]},
    "NVCVLA": {"N": ["11"]},
    "ARETHVSIS": {"A": ["24"]},
    "PERENNA": {"N": ["11"]},
    "TALAIONIVS": {"A": ["21"]},
    "PARMENSIS": {"A": ["24"]},
    "TITYVS": {"N": ["12"]},
    "CILICIVS": {"A": ["21"]},
    "CORYTHVS": {"N": ["12"]},
    "CELEVS": {"N": ["12"]},
    "SERIPHIVS": {"A": ["21"]},
    "ARACINTHVS": {"N": ["12"]},
    "MARPESIVS": {"A": ["21"]},
    "HELORIVS": {"A": ["21"]},
    "EVDEMVS": {"N": ["12"]},
    "SILIA": {"N": ["11"]},
    "DINAEA": {"N": ["11"]},
    "SPARTACVS": {"N": ["12"]},
    "OVIVS": {"N": ["12"]},
    "OEBALIVS": {"A": ["21"]},
    "ISAVRVS": {"A": ["21"]},
    "SAGANA": {"N": ["11"]},
    "VETTIANVS": {"A": ["21"]},
    "IPHIGENIA": {"N": ["11"]},
    "PELOPEA": {"N": ["11"]},
    "ZMINTHEVS": {"A": ["21"]},
    "ENNA": {"N": ["11"]},
    "MEGAERA": {"N": ["11"]},
    "MYTILENA": {"N": ["11"]},
    "VALETIACVS": {"N": ["12"]},
    "MASVRIVS": {"N": ["12"]},
    "NOVIODVNVM": {"N": ["12"]},
    "ILIVS": {"A": ["21"]},
    "PAPHVS": {"N": ["12"]},
    "PANTOLABVS": {"N": ["12"]},
    "CYDONEVS": {"A": ["21"]},
    "BVRRIENVS": {"N": ["12"]},
    "PLVTONIVS": {"A": ["21"]},
    "LATONIGENA": {"N": ["11"]},
    "CORSICA": {"N": ["11"]},
    "CRETENSIS": {"A": ["24"]},
    "ADVATVCA": {"N": ["11"]},
    "ARMINIVS": {"N": ["12"]},
    "ARIVSIVS": {"A": ["21"]},
    "RHODOPEIVS": {"A": ["21"]},
    "THRACA": {"N": ["11"]},
    "SVSIANVS": {"A": ["21"]},
    "IBERICVS": {"A": ["21"]},
    "AMATA": {"N": ["11"]},
    "VIBIDIVS": {"N": ["12"]},
    "SARIOLENVS": {"N": ["12"]},
    "THRAESSVS": {"A": ["21"]},
    "CHIMAERA": {"N": ["11"]},
    "BALACRVS": {"N": ["12"]},
    "TVBILVSTRIA": {"N": ["11"]},
    "EVBOICVS": {"A": ["21"]},
    "BOVMELVS": {"N": ["12"]},
    "AEOLIS": {"A": ["24"], "N": ["13"]},
    "PADVS": {"N": ["12"]},
    "DAVARA": {"N": ["11"]},
    "IVDAEA": {"N": ["11"]},
    "TELEMACHVS": {"N": ["12"]},
    "EVTRAPELVS": {"N": ["12"]},
    "MAMMAEA": {"N": ["11"]},
    "CAVCASVS": {"N": ["12"]},
    "APOLLINARIS": {"N": ["13"], "A": ["12"]},
    "MYRINA": {"N": ["11"]},
    "STILONIVS": {"N": ["12"]},
    "TRIGEMINA": {"N": ["11"]},
    "VOCETIVS": {"N": ["12"]},
    "VVLCANIVS": {"A": ["21"]},
    "BASSVS": {"N": ["12"]},
    "CHAERONEA": {"N": ["11"]},
    "SATELLIVS": {"N": ["12"]},
    "AEAEVS": {"A": ["21"]},
    "GAETVLICVS": {"N": ["12"]},
    "RVFILLVS": {"N": ["12"]},
    "NORICVS": {"A": ["21"]},
    "ANYTVS": {"N": ["12"]},
    "LIGVRIA": {"N": ["11"]},
    "CANNVTIVS": {"N": ["12"]},
    "ARGVETIVS": {"N": ["12"]},
    "BASSAREVS": {"N": ["12"]},
    "TEVCRIA": {"N": ["11"]},
    "SEMELEIVS": {"A": ["21"]},
    "SERRANA": {"N": ["11"]},
    "MARDONIVS": {"N": ["12"]},
    "CAESAREVS": {"A": ["21"]},
    "ALBINOVANVS": {"N": ["12"]},
    "CLODIA": {"N": ["11"]},
    "CAMERINA": {"N": ["11"]},
    "FIBVLVS": {"N": ["12"]},
    "IDISTAVISVS": {"N": ["12"]},
    "APIDIVS": {"N": ["12"]},
    "CALEDONIA": {"N": ["11"]},
    "CAESAREA": {"N": ["11"]},
    "CLEARETA": {"N": ["11"]},
    "ADRAMYTTEVM": {"N": ["12"]},
    "CERIALIA": {"N": ["11"]},
    "MITHRIDATICVS": {"A": ["21"]},
    "SERRANVS": {"N": ["12"]},
    "CORELLIVS": {"N": ["12"]},
    "GNOSIACVS": {"A": ["21"]},
    "TARVLA": {"N": ["11"]},
    "CAMERIVS": {"N": ["12"]},
    "NARYCIVS": {"A": ["21"]},
    "PARVS": {"N": ["12"]},
    "ISEVM": {"N": ["12"]},
    "DECIMIVS": {"N": ["12"]},
    "PELASGVS": {"N": ["12"], "A": ["21"]},
    "LVSIVS": {"N": ["12"]},
    "EXTITIVS": {"N": ["12"]},
    "ATIMETVS": {"N": ["12"]},
    "CLEANTHEVS": {"A": ["21"]},
    "BELICA": {"N": ["11"]},
    "DIGENTIA": {"N": ["11"]},
    "NOVIVS": {"N": ["12"]},
    "PACALIS": {"A": ["24"]},
    "POSTVMIA": {"N": ["11"]},
    "ORONTEVS": {"A": ["21"]},
    "AVFILENVS": {"N": ["12"]},
    "AMMIVS": {"N": ["12"]},
    "CYTHEREVS": {"A": ["21"]},
    "PHILOXENVS": {"N": ["12"]},
    "ISINDIVS": {"A": ["21"]},
    "VARIANVS": {"A": ["21"]},
    "SVRENA": {"N": ["11"]},
    "SINVESSANVS": {"A": ["21"]},
    "RAMNVSIVS": {"A": ["21"]},
    "MONTANIVS": {"N": ["12"]},
    "ROMANICVS": {"A": ["21"]},
    "RVBELLIVS": {"N": ["12"]},
    "MOLOSSICVS": {"A": ["21"]},
    "CAVDINVS": {"N": ["12"], "A": ["21"]},
    "RHODVS": {"N": ["12"]},
    "SCRIBONIA": {"N": ["11"]},
    "NEMAEVS": {"A": ["21"]},
    "PAGASVS": {"N": ["12"]},
    "TVDITANVS": {"N": ["12"]},
    "IDALIVM": {"N": ["12"]},
    "PERSIVS": {"N": ["12"]},
    "OCTODVRVS": {"N": ["12"]},
    "ALBRVNA": {"N": ["11"]},
    "MENEDEMVS": {"N": ["12"]},
    "EVENVS": {"N": ["12"]},
    "ARRIONILLA": {"N": ["11"]},
    "PALICVS": {"N": ["12"]},
    "GABBA": {"N": ["11"]},
    "CLONVS": {"N": ["12"]},
    "ERSILAVS": {"N": ["12"]},
    "SVEBA": {"N": ["11"]},
    "LACAENVS": {"A": ["21"]},
    "MATIVS": {"N": ["12"]},
    "ARIADNA": {"N": ["11"]},
    "CAESILIANVS": {"N": ["12"]},
    "ACADIRA": {"N": ["11"]},
    "DORYLAEVM": {"N": ["12"]},
    "PARNASIVS": {"A": ["21"]},
    "IPHICLVS": {"N": ["12"]},
    "GELONVS": {"N": ["12"]},
    "LISCVS": {"N": ["12"]},
    "ARDA": {"N": ["11"]},
    "ALCINOVS": {"N": ["12"]},
    "CAESARIENSIS": {"A": ["24"]},
    "MELANTHEVS": {"A": ["21"]},
    "SOSIVS": {"N": ["12"]},
    "ORSILOCHVS": {"N": ["12"]},
    "THESPIVS": {"A": ["21"]},
    "PATRVITVS": {"N": ["12"]},
    "MENARCHVS": {"N": ["12"]},
    "TARPEIA": {"N": ["11"]},
    "SABINA": {"N": ["11"]},
    "BONNENSIS": {"A": ["24"]},
    "SVBLAQVEVM": {"N": ["12"]},
    "VALENTIA": {"N": ["11"]},
    "OLYNTHVS": {"N": ["12"]},
    "SAMAROBRIVA": {"N": ["11"]},
    "SCIRTVS": {"N": ["12"]},
    "GALLIVS": {"N": ["12"]},
    "ADRAMYTTENVS": {"A": ["21"]},
    "HADRVMETVM": {"N": ["12"]},
    "TVRTA": {"N": ["11"]},
    "CERCYONEVS": {"A": ["21"]},
    "CATONIVS": {"N": ["12"]},
    "MARTINA": {"N": ["11"]},
    "POEANTIVS": {"A": ["21"]},
    "LAMPADIVM": {"N": ["12"]},
    "MEGARICVS": {"A": ["21"]},
    "PHILVS": {"N": ["12"]},
    "BACTRIANVS": {"A": ["21"]},
    "TEREVS": {"N": ["12"]},
    "CICIRRVS": {"N": ["12"]},
    "MAREOTIS": {"A": ["24"], "N": ["16"]},
    "DIOXIPPVS": {"N": ["12"]},
    "LIBYSTINVS": {"A": ["21"]},
    "VIBILIVS": {"N": ["12"]},
    "EPASNACTVS": {"N": ["12"]},
    "EVMELVS": {"N": ["12"]},
    "LEMNVS": {"N": ["12"]},
    "AVIDIENVS": {"N": ["12"]},
    "OENOTHEA": {"N": ["11"]},
    "MISENA": {"N": ["11"]},
    "PITTACVS": {"N": ["12"]},
    "PVBLICIVS": {"A": ["21"], "N": ["12"]},
    "RVTILA": {"N": ["11"]},
    "AHENOBARBVS": {"N": ["12"]},
    "ALLIVS": {"N": ["12"]},
    "AMPHILOCHVS": {"N": ["12"]},
    "PERSEIS": {"A": ["24"], "N": ["16"]},
    "DEIOPEA": {"N": ["11"]},
    "HERMVNDVRVS": {"A": ["21"]},
    "BAETICVS": {"A": ["21"]},
    "AVIDIVS": {"N": ["12"]},
    "ODRYSIVS": {"A": ["21"]},
    "PHLYVS": {"N": ["12"]},
    "PIRENIDVS": {"A": ["21"]},
    "CLVVIA": {"N": ["11"]},
    "NAMMEIVS": {"N": ["12"]},
    "CYNTHIA": {"N": ["11"]},
    "FONTEIA": {"N": ["11"]},
    "ANQVILLARIA": {"N": ["11"]},
    "THALA": {"N": ["11"]},
    "TRACHALVS": {"N": ["12"]},
    "SVRRENTVM": {"N": ["12"]},
    "VMBRENVS": {"N": ["12"]},
    "ADMETVS": {"N": ["12"]},
    "DOMITIA": {"N": ["11"]},
    "HIRTVLEIVS": {"N": ["12"]},
    "PALFVRIVS": {"N": ["12"]},
    "ALEXA": {"N": ["11"]},
    "COGIDVMNVS": {"N": ["12"]},
    "ARTEMISIVS": {"N": ["12"]},
    "DANVBIVS": {"N": ["12"]},
    "SILIVS": {"N": ["12"]},
    "MARCODVRVM": {"N": ["12"]},
    "EVTYCHVS": {"N": ["12"]},
    "ENCELADVS": {"N": ["12"]},
    "CNIDIVS": {"A": ["21"]},
    "LAVRENTINVM": {"N": ["12"]},
    "PINTHIA": {"N": ["11"]},
    "MYSVS": {"A": ["21"]},
    "NONIVS": {"N": ["12"]},
    "AGONIA": {"-": ["12"], "N": ["12"]},
    "CASILINVM": {"N": ["12"]},
    "NESSVS": {"N": ["12"]},
    "MYCENA": {"N": ["11"]},
    "CAPRIVS": {"N": ["12"]},
    "ANTONINVS": {"N": ["12"]},
    "NVMANVS": {"N": ["12"]},
    "RVFINVS": {"N": ["12"]},
    "LYCVRGVS": {"N": ["12"]},
    "PETROSIDIVS": {"N": ["12"]},
    "ANTHEVS": {"N": ["12"]},
    "CELTILLVS": {"N": ["12"]},
    "PERSIS": {"A": ["24"], "N": ["16"]},
    "SEPLASIA": {"N": ["11"]},
    "BODVOGNATVS": {"N": ["12"]},
    "OLYMPIA": {"O": ["12"], "N": ["12"]},
    "PHAEDRVS": {"N": ["12"]},
    "TENCTERVS": {"A": ["21"]},
    "CEVS": {"A": ["21"]},
    "AVRIANVS": {"A": ["21"]},
    "PORSENNA": {"N": ["11"]},
    "TANAVS": {"N": ["12"]},
    "LIBYSTIS": {"A": ["24"]},
    "LVPERCALIA": {"N": ["11"]},
    "IONIVM": {"N": ["12"]},
    "LVTETIA": {"N": ["11"]},
    "NOMENTANVM": {"N": ["12"]},
    "MANILIA": {"N": ["11"]},
    "CORINNA": {"N": ["11"]},
    "ARTORIA": {"N": ["11"]},
    "ERIGYIVS": {"N": ["12"]},
    "PASIPPVS": {"N": ["12"]},
    "MAECENATIANVS": {"N": ["12"]},
    "CAMILLIANVS": {"A": ["21"]},
    "NERONIANVS": {"A": ["21"]},
    "PANDVSA": {"N": ["11"]},
    "OPSIVS": {"N": ["12"]},
    "VERVCLOETIVS": {"N": ["12"]},
    "ASVVIANVS": {"A": ["21"]},
    "TAVRIANVS": {"A": ["21"]},
    "CENABVM": {"N": ["12"]},
    "CRETHEVS": {"N": ["12"]},
    "CRABRA": {"N": ["11"]},
    "MASSILIENSIS": {"A": ["24"]},
    "PERELLIVS": {"N": ["12"]},
    "OCRESIA": {"N": ["11"]},
    "CINGONIVS": {"N": ["12"]},
    "SALVIVS": {"N": ["12"]},
    "CRESSA": {"N": ["11"], "A": ["21"]},
    "SAMNIVM": {"N": ["12"]},
    "THYESTEVS": {"A": ["21"]},
    "HELVIVS": {"N": ["12"]},
    "HADRIACVS": {"A": ["21"]},
    "PERVSIA": {"N": ["11"]},
    "DEMOPHILVS": {"N": ["12"]},
    "INDICVS": {"A": ["21"]},
    "SVFFENVS": {"N": ["12"]},
    "POLLITTA": {"N": ["11"]},
    "ALTHAEA": {"N": ["11"]},
    "AVGVRINVS": {"N": ["12"]},
    "ELECTRA": {"N": ["11"]},
    "GALERIANVS": {"N": ["12"]},
    "EVCHADIA": {"N": ["11"]},
    "LINVS": {"N": ["12"]},
    "ANDECVMBORIVS": {"N": ["12"]},
    "CASCVS": {"N": ["12"]},
    "TROIVGENA": {"N": ["11"]},
    "TASGETIVS": {"N": ["12"]},
    "THERMAEVS": {"A": ["21"]},
    "LARA": {"N": ["11"]},
    "SARDVS": {"A": ["21"]},
    "EVHIVS": {"-": ["21"], "N": ["12"]},
    "GALESVS": {"N": ["12"]},
    "PLEMVRIVM": {"N": ["12"]},
    "TECTA": {"N": ["11"]},
    "PHILARGYRVS": {"N": ["12"]},
    "AMYCVS": {"N": ["12"]},
    "BACTRIANA": {"N": ["11"]},
    "ASCVLANVS": {"A": ["21"]},
    "CVSINIVS": {"N": ["12"]},
    "SAMIARIVS": {"N": ["12"]},
    "IVNONALIS": {"A": ["24"]},
    "ORMENVS": {"N": ["12"]},
    "ERIDANVS": {"N": ["12"]},
    "FORNACALIS": {"A": ["24"]},
    "BEBRYCIVS": {"A": ["21"]},
    "ARISTONICVS": {"N": ["12"]},
    "ARMENIA": {"N": ["11"]},
    "TORIVS": {"N": ["12"]},
    "TAGVS": {"N": ["12"]},
    "GALLOGRAECVS": {"A": ["21"]},
    "VMMIDIVS": {"N": ["12"]},
    "GLYCERA": {"N": ["11"]},
    "AMYTHAONIVS": {"A": ["21"]},
    "MEVANIA": {"N": ["11"]},
    "PHILAENIVM": {"N": ["12"]},
    "DIABOLVS": {"N": ["12"]},
    "GOMPHENSIS": {"A": ["24"]},
    "CASCELLIVS": {"N": ["12"]},
    "AETIDEMVS": {"N": ["12"]},
    "MAEVIA": {"N": ["11"]},
    "PHILIPPICVS": {"A": ["21"]},
    "GRAPTVS": {"N": ["12"]},
    "VERGILIANVS": {"A": ["21"]},
    "GELDVBA": {"N": ["11"]},
    "FIRMIVS": {"N": ["12"]},
    "CIRCEVS": {"A": ["21"]},
    "NAVA": {"N": ["11"]},
    "TYRVS": {"N": ["12"]},
    "BINGIVM": {"N": ["12"]},
    "PANNONIA": {"N": ["11"]},
    "LAOMEDONTIVS": {"A": ["21"]},
    "CESTILIVS": {"N": ["12"]},
    "GORDIVM": {"N": ["12"]},
    "ALCMENA": {"N": ["11"]},
    "ANTILOCHVS": {"N": ["12"]},
    "PACIDEIVS": {"N": ["12"]},
    "HIPPOLYTVS": {"N": ["12"]},
    "DVLICHIVM": {"N": ["12"]},
    "APIDANVS": {"N": ["12"]},
    "PHOEBIGENA": {"N": ["11"]},
    "CANVSINVS": {"A": ["21"]},
    "GVLVSA": {"N": ["11"]},
    "CILNIVS": {"N": ["12"]},
    "EVXINVS": {"N": ["12"]},
    "TIGRANOCERTA": {"N": ["11"]},
    "APELLEVS": {"A": ["21"]},
    "CAVCILIVS": {"N": ["12"]},
    "THYBRINVS": {"A": ["21"]},
    "XVTHVS": {"N": ["12"]},
    "PHEGEVS": {"N": ["12"]},
    "SARMATA": {"N": ["11"]},
    "PHRYXIANVS": {"A": ["21"]},
    "OEAGRIVS": {"A": ["21"]},
    "CALLIPHANA": {"N": ["11"]},
    "APVLIA": {"N": ["11"]},
    "GAIA": {"N": ["11"]},
    "TRIVIA": {"N": ["11"]},
    "CAELIMONTANVS": {"A": ["21"]},
    "TIBERIANVS": {"A": ["21"]},
    "SAMBVS": {"N": ["12"]},
    "GEMINIVS": {"N": ["12"]},
    "ALCMAEONIVS": {"A": ["21"]},
    "PETILLIVS": {"N": ["12"]},
    "MAMVRIVS": {"N": ["12"]},
    "IDVMAEVS": {"A": ["21"]},
    "PENEIVS": {"A": ["21"]},
    "VZITTA": {"N": ["11"]},
    "AVLERCVS": {"A": ["21"]},
    "MEDIOLANENSIS": {"A": ["24"]},
    "CLITVS": {"N": ["12"]},
    "TIMARCHVS": {"N": ["12"]},
    "SINGILIENSIS": {"A": ["24"]},
    "ITALIS": {"A": ["24"]},
    "MELLA": {"N": ["11"]},
    "PEREDIA": {"N": ["11"]},
    "NVMMVLARIVS": {"1": ["12"]},
    "VIBENNIVS": {"N": ["12"]},
    "INTERAMNIVM": {"N": ["12"]},
    "ACASTVS": {"N": ["12"]},
    "ALLEDIVS": {"N": ["12"]},
    "INACHIVS": {"A": ["21"]},
    "SICANIA": {"N": ["11"]},
    "TITHONIVS": {"A": ["21"]},
    "BRITANNICVS": {"A": ["21"], "N": ["12"]},
    "THERMODONTIACVS": {"A": ["21"]},
    "ABELLANVS": {"A": ["21"]},
    "NVMIDA": {"N": ["11"]},
    "TALASSIVS": {"N": ["12"]},
    "PANCHARIA": {"N": ["11"]},
    "AMANTIA": {"N": ["11"]},
    "CYCNEIVS": {"A": ["21"]},
    "HYLLVS": {"N": ["12"]},
    "STAPHYLA": {"N": ["11"]},
    "ACHERONTICVS": {"A": ["21"]},
    "OENEVS": {"N": ["12"]},
    "EMATHIA": {"N": ["11"]},
    "HOSTILIA": {"N": ["11"]},
    "MENESTHEVS": {"N": ["12"]},
    "CINCINNATVS": {"-": ["21"], "N": ["12"]},
    "AGENOREVS": {"A": ["21"]},
    "THYONEVS": {"N": ["12"]},
    "RHODIA": {"N": ["11"]},
    "CLEOMACHVS": {"N": ["12"]},
    "CYTHNVS": {"N": ["12"]},
    "REMVLVS": {"N": ["12"]},
    "HADRIATICVM": {"N": ["12"]},
    "SEPTIMINVS": {"N": ["12"]},
    "PROMOLVS": {"N": ["12"]},
    "BAVIVS": {"N": ["12"]},
    "LATAGVS": {"N": ["12"]},
    "CALPVRNIA": {"N": ["11"]},
    "MINOIVS": {"A": ["21"]},
    "CYCLOPIVS": {"A": ["21"]},
    "CATVALDA": {"N": ["11"]},
    "BOIVS": {"A": ["21"]},
    "VACVNA": {"N": ["11"]},
    "DOMNOTAVRVS": {"N": ["12"]},
    "MVTILIA": {"N": ["11"]},
    "PETTIVS": {"N": ["12"]},
    "ERYLVS": {"N": ["12"]},
    "LEARCHEVS": {"A": ["21"]},
    "THESSALIA": {"N": ["11"]},
    "RAETICVS": {"A": ["21"]},
    "EPONA": {"N": ["11"]},
    "LERNAEVS": {"A": ["21"]},
    "SYGAMBRA": {"N": ["11"]},
    "COSSVTIANVS": {"N": ["12"]},
    "TRIARIA": {"N": ["11"]},
    "LIVINEIVS": {"N": ["12"]},
    "DOSSENNVS": {"N": ["12"]},
    "LVSCINVS": {"N": ["12"]},
    "NERVIVS": {"A": ["21"]},
    "HARPYIA": {"N": ["11"]},
    "NERINVS": {"A": ["21"]},
    "SABAEVS": {"A": ["21"]},
    "PHALAECEVS": {"A": ["21"]},
    "LYRNESIS": {"A": ["24"]},
    "TALTHYBIVS": {"N": ["12"]},
    "TARSA": {"N": ["11"]},
    "BEIRA": {"N": ["11"]},
    "HELLANICVS": {"N": ["12"]},
    "CASINA": {"N": ["11"]},
    "FLAMINIA": {"N": ["11"]},
    "STIMVLA": {"N": ["11"]},
    "MELETEVS": {"A": ["21"]},
    "AELIA": {"N": ["11"]},
    "ARISTIVS": {"N": ["12"]},
    "SCAPTENSVLA": {"N": ["11"]},
    "PRASVTAGVS": {"N": ["12"]},
    "AEGAEVM": {"N": ["12"]},
    "PERPENNA": {"N": ["11"]},
    "PERVSINVM": {"N": ["12"]},
    "SALONA": {"N": ["11"]},
    "FVLCINIANVS": {"A": ["21"]},
    "BELGIVM": {"N": ["12"]},
    "MAEONIS": {"A": ["24"], "N": ["16"]},
    "MEMPHITICVS": {"A": ["21"]},
    "CASTVLONENSIS": {"A": ["24"]},
    "POPPAEA": {"N": ["11"]},
    "ARBELA": {"N": ["11"]},
    "PARIVS": {"A": ["21"]},
    "IVRA": {"N": ["11"]},
    "CAEDICVS": {"N": ["12"]},
    "LANVVINVS": {"A": ["21"]},
    "BATVLVM": {"N": ["12"]},
    "HILAIRA": {"N": ["11"]},
    "CORDVBENSIS": {"A": ["24"]},
    "OILEVS": {"N": ["12"]},
    "TARBELLVS": {"A": ["21"]},
    "POLYDORVS": {"N": ["12"]},
    "MONA": {"N": ["11"]},
    "PHILIPPA": {"N": ["11"]},
    "PELOPEVS": {"A": ["21"]},
    "COLCHICVS": {"A": ["21"]},
    "TMOLVS": {"N": ["12"]},
    "POLLENTIA": {"-": ["11"], "N": ["12"]},
    "DIONAEVS": {"A": ["21"]},
    "HALAESVS": {"N": ["12"]},
    "VARILLA": {"N": ["11"]},
    "VIPSANIA": {"N": ["11"]},
    "AEGINA": {"N": ["11"]},
    "MOESICVS": {"A": ["21"]},
    "LYCAMBEVS": {"A": ["21"]},
    "TROGVS": {"N": ["12"]},
    "AENOBARBVS": {"N": ["12"]},
    "MANCINVS": {"N": ["12"]},
    "TITVRIVS": {"N": ["12"]},
    "VENVLVS": {"N": ["12"]},
    "PACIDEVS": {"N": ["12"]},
    "MEVVLANVS": {"N": ["12"]},
    "BROMIA": {"N": ["11"]},
    "HARPALYCVS": {"N": ["12"]},
    "CENTVMALVS": {"N": ["12"]},
    "CHRYSIPPVS": {"N": ["12"]},
    "MYCENAEVS": {"A": ["21"]},
    "PACORVS": {"N": ["12"]},
    "MARRVVIVS": {"A": ["21"]},
    "CAETRONIVS": {"N": ["12"]},
    "AVRELIA": {"N": ["11"]},
    "CALAVIVS": {"N": ["12"]},
    "TETRICA": {"N": ["11"]},
    "AVSONIS": {"A": ["24"]},
    "OCRICVLVM": {"N": ["12"]},
    "BRIXELLVM": {"N": ["12"]},
    "PIRAEA": {"N": ["11"]},
    "ADIMANTVS": {"N": ["12"]},
    "APPVS": {"N": ["12"]},
    "OVIDIVS": {"N": ["12"]},
    "OEDIPVS": {"N": ["12"]},
    "HALIVS": {"N": ["12"]},
    "ELEVSINVS": {"A": ["21"]},
    "PENEVS": {"N": ["12"]},
    "PYLADVS": {"N": ["12"]},
    "NORICVM": {"N": ["12"]},
    "LVSITANIA": {"N": ["11"]},
    "PENELEVS": {"N": ["12"]},
    "AGRIANVS": {"A": ["21"]},
    "AGRIPPINENSIS": {"A": ["24"]},
    "MELINVS": {"N": ["12"]},
    "TARQVITIVS": {"N": ["12"]},
    "CALLAICVS": {"A": ["21"]},
    "HADRIACVM": {"N": ["12"]},
    "HIEROCAESARIA": {"N": ["11"]},
    "VIBVLLIVS": {"N": ["12"]},
    "LEVCADIA": {"N": ["11"]},
    "SANCIA": {"N": ["11"]},
    "PHRVGIA": {"N": ["11"]},
    "MONYCHVS": {"N": ["12"]},
    "SATRIA": {"N": ["11"]},
    "ATEGVA": {"N": ["11"]},
    "ISSA": {"N": ["11"]},
    "CENAEVS": {"A": ["21"]},
    "TRIMERVS": {"N": ["12"]},
    "NOMENTANVS": {"N": ["12"]},
    "CARBONARIVS": {"1": ["12"]},
    "PERBIBESIA": {"N": ["11"]},
    "HALVS": {"N": ["12"]},
    "SIPONTINVS": {"A": ["21"]},
    "CAETRA": {"N": ["11"]},
    "PELLAEVS": {"A": ["21"]},
    "VELLOCATVS": {"N": ["12"]},
    "NVMANTINVS": {"A": ["21"]},
    "MAGNETIS": {"A": ["24"]},
    "AMYCLANVS": {"A": ["21"]},
    "PHILYREIVS": {"A": ["21"]},
    "FLORVS": {"N": ["12"]},
    "CVRIANVS": {"N": ["12"]},
    "DVROCORTORVM": {"N": ["12"]},
    "ARADVS": {"N": ["12"]},
    "APELLA": {"N": ["11"]},
    "LAVERNA": {"N": ["11"]},
    "OTHONIANVS": {"A": ["21"]},
    "HYRCANIVS": {"A": ["21"]},
    "MEDVLLINA": {"N": ["11"]},
    "SAGVNTVM": {"N": ["12"]},
    "ISMENIS": {"A": ["24"]},
    "VLLEXODVNVM": {"N": ["12"]},
    "ASCLEPIODORVS": {"N": ["12"]},
    "BAETVRIA": {"N": ["11"]},
    "ESQVILIVS": {"A": ["21"]},
    "LATINA": {"N": ["11"]},
    "CARNA": {"N": ["11"]},
    "LEVCOTHEA": {"N": ["11"]},
    "LYSANIA": {"N": ["11"]},
    "TERRASIDIVS": {"N": ["12"]},
    "SANGVNNIVS": {"N": ["12"]},
    "POLYAENVS": {"N": ["12"]},
    "IVTVRNA": {"N": ["11"]},
    "GRAIVGENA": {"N": ["11"]},
    "ATHENODORVS": {"N": ["12"]},
    "ARTABANVS": {"N": ["12"]},
    "ELEVS": {"A": ["21"]},
    "SACRATA": {"N": ["11"]},
    "PRAENESTINVS": {"A": ["21"]},
    "ANCHARIANVS": {"A": ["21"]},
    "ARNVS": {"N": ["12"]},
    "GYLIPPVS": {"N": ["12"]},
    "ASCRAEVS": {"A": ["21"]},
    "LIGEA": {"N": ["11"]},
    "ALBVCILLA": {"N": ["11"]},
    "LENTICVLA": {"N": ["11"]},
    "BELLEROPHONTEVS": {"A": ["21"]},
    "NICOBVLVS": {"N": ["12"]},
    "SIDONIVS": {"A": ["21"]},
    "AVFILENA": {"N": ["11"]},
    "POTITVS": {"N": ["12"]},
    "VESAEVVS": {"N": ["12"]},
    "AEGIVM": {"N": ["12"]},
    "CARSVLANVM": {"N": ["12"]},
    "IOLCVS": {"N": ["12"]},
    "CVRIONIANVS": {"A": ["21"]},
    "PHOTINVS": {"N": ["12"]},
    "ALLIFANVS": {"A": ["21"]},
    "CYTORVS": {"N": ["12"]},
    "SCYRVS": {"N": ["12"]},
    "HIPPODAMIA": {"N": ["11"]},
    "CVPIDINEVS": {"A": ["21"]},
    "EPAPHRODITVS": {"N": ["12"]},
    "TVTILIVS": {"N": ["12"]},
    "PHILOSITVS": {"N": ["12"]},
    "RHODANVS": {"N": ["12"]},
    "AMENANVS": {"A": ["21"]},
    "PLAVTVS": {"N": ["12"]},
    "AMOEBEVS": {"N": ["12"]},
    "CALLIDROMVS": {"N": ["12"]},
    "INTERAMNANVS": {"A": ["21"]},
    "TEGEEVS": {"A": ["21"]},
    "SATVRNIVS": {"A": ["21"]},
    "STROLVS": {"N": ["12"]},
    "ERECTHEVS": {"A": ["21"]},
    "HERCYNIVS": {"A": ["21"]},
    "CAEDICIA": {"N": ["11"]},
    "BARCANVS": {"A": ["21"]},
    "CADRA": {"N": ["11"]},
    "LISSVS": {"N": ["12"]},
    "SOSIA": {"N": ["11"]},
    "PARAETONIVM": {"N": ["12"]},
    "RECTINA": {"N": ["11"]},
    "TITHONIA": {"N": ["11"]},
    "BITTIVS": {"N": ["12"]},
    "AGESILAVS": {"N": ["12"]},
    "VALERVS": {"N": ["12"]},
    "ARABIA": {"N": ["11"]},
    "CLANIVS": {"N": ["12"]},
    "LACINIVS": {"A": ["21"]},
    "POENIVS": {"N": ["12"]},
    "ACHELOIVS": {"A": ["21"]},
    "CYMAEVS": {"A": ["21"]},
    "AGAMEMNONIVS": {"A": ["21"]},
    "QVIRINIANVS": {"A": ["21"]},
    "VILLIVS": {"N": ["12"]},
    "IOCASTA": {"N": ["11"]},
    "THEODOSIVS": {"N": ["12"]},
    "BEROEA": {"N": ["11"]},
    "CAIETA": {"N": ["11"]},
    "VMMIDIA": {"N": ["11"]},
    "TVBERTVS": {"N": ["12"]},
    "SETINVS": {"A": ["21"]},
    "MARSYA": {"N": ["11"]},
    "LVCVLLIANVS": {"A": ["21"]},
    "ITALICA": {"N": ["11"]},
    "ARVIRAGVS": {"N": ["12"]},
    "ARCHIBVLVS": {"N": ["12"]},
    "SATVRIVS": {"N": ["12"]},
    "AVILLIVS": {"N": ["12"]},
    "CIRCAEVS": {"A": ["21"]},
    "LATMIVS": {"A": ["21"]},
    "VAGA": {"N": ["11"]},
    "EDONVS": {"A": ["21"]},
    "POSTVMINVS": {"N": ["12"]},
    "AFINIVS": {"N": ["12"]},
    "AVTOLYCVS": {"N": ["12"]},
    "DINOMACHVS": {"N": ["12"]},
    "VELLEVS": {"N": ["12"]},
    "TIGVRINVS": {"A": ["21"]},
    "ANTIOPA": {"N": ["11"]},
    "SIGEIVS": {"A": ["21"]},
    "NVMIDIA": {"N": ["11"]},
    "SERGIOLVS": {"N": ["12"]},
    "ORTYGIA": {"N": ["11"]},
    "BVCCILLVS": {"N": ["12"]},
    "CALESTRIVS": {"N": ["12"]},
    "LENTINVS": {"N": ["12"]},
    "LVPERCVS": {"N": ["12"]},
    "BVCEPHALA": {"N": ["11"]},
    "INGVIOMERVS": {"N": ["12"]},
    "LYCISCA": {"N": ["11"]},
    "ASSARACVS": {"N": ["12"]},
    "EREBVS": {"N": ["12"]},
    "MELAMPVS": {"N": ["12"]},
    "ALABANDA": {"N": ["11"]},
    "CORYPHAEVS": {"N": ["12"]},
    "LATOVS": {"N": ["12"]},
    "LAELIA": {"N": ["11"]},
    "ETHYMANTVS": {"N": ["12"]},
    "CYDONIVS": {"A": ["21"]},
    "PELORVS": {"N": ["12"]},
    "RABIRIVS": {"N": ["12"]},
    "PENNVS": {"N": ["12"]},
    "EPHYRAEVS": {"A": ["21"]},
    "APRONIA": {"N": ["11"]},
    "SVLPICIANVS": {"A": ["21"]},
    "ARTAXATA": {"N": ["11"]},
    "AGEDINCVM": {"N": ["12"]},
    "PROCVLA": {"N": ["11"]},
    "GALAESVS": {"N": ["12"]},
    "CASPIVS": {"A": ["21"]},
    "MENANDREVS": {"A": ["21"]},
    "VASIVS": {"N": ["12"]},
    "SVRRENTINVM": {"N": ["12"]},
    "EBVSVS": {"N": ["12"]},
    "MISENVM": {"N": ["12"]},
    "ANCHISEVS": {"A": ["21"]},
    "POMPEIA": {"N": ["11"]},
    "CEPARIVS": {"N": ["12"]},
    "CAMILLA": {"N": ["11"]},
    "PACCIVS": {"N": ["12"]},
    "CYRNEVS": {"A": ["21"]},
    "MEDVSAEVS": {"A": ["21"]},
    "AMASENVS": {"N": ["12"]},
    "GAVIANVS": {"N": ["12"]},
    "CRVCISALVS": {"N": ["12"]},
    "TVRPILIVS": {"N": ["12"]},
    "PEDIATIA": {"N": ["11"]},
    "TISIDIVM": {"N": ["12"]},
    "DIDYMVS": {"N": ["12"]},
    "TAVRICVS": {"A": ["21"]},
    "PYRRIA": {"N": ["11"]},
    "NEMEA": {"N": ["11"]},
    "BIVIVS": {"-": ["21"], "N": ["12"]},
    "THESSALICVS": {"A": ["21"]},
    "CRITOBVLVS": {"N": ["12"]},
    "PINDVS": {"N": ["12"]},
    "TRITONIVS": {"A": ["21"]},
    "IVLEVS": {"A": ["21"]},
    "PACIAECVS": {"N": ["12"]},
    "NAEVOLVS": {"N": ["12"]},
    "CAPENVS": {"A": ["21"]},
    "COSMVS": {"N": ["12"]},
    "APHIDNVS": {"N": ["12"]},
    "MVNDENSIS": {"A": ["24"]},
    "THALIA": {"N": ["11"]},
    "BELGICA": {"N": ["11"]},
    "BANDVSIA": {"N": ["11"]},
    "BLITIVS": {"N": ["12"]},
    "NEAEREA": {"N": ["11"]},
    "SABINVM": {"N": ["12"]},
    "PVPIVS": {"N": ["12"]},
    "MAGETOBRIGA": {"N": ["11"]},
    "PISA": {"N": ["11"]},
    "VESCINVS": {"A": ["21"]},
    "VRBINVM": {"N": ["12"]},
    "MNASYLVS": {"N": ["12"]},
    "MAZACA": {"N": ["11"]},
    "ANTIMACHVS": {"N": ["12"]},
    "TREBATIVS": {"N": ["12"]},
    "NOLANVS": {"A": ["21"]},
    "IPSITHILLA": {"N": ["11"]},
    "ARRVNTIVS": {"N": ["12"]},
    "CARANVS": {"N": ["12"]},
    "EGNATVLEIVS": {"N": ["12"]},
    "NICOMACHVS": {"N": ["12"]},
    "TVLLIANVM": {"N": ["12"]},
    "BASILVS": {"N": ["12"]},
    "PORTVNVS": {"N": ["12"]},
    "CORSVS": {"A": ["21"]},
    "TVRIVS": {"N": ["12"]},
    "THERAPONTIGONVS": {"N": ["12"]},
    "ASICIVS": {"N": ["12"]},
    "RHIDAGNVS": {"N": ["12"]},
    "GYARA": {"N": ["11"]},
    "THEOPOMPVS": {"N": ["12"]},
    "SVSA": {"N": ["11"]},
    "CLVENTIANVS": {"A": ["21"]},
    "VERONA": {"N": ["11"]},
    "CYLLENEVS": {"A": ["21"]},
    "SVSAGVS": {"N": ["12"]},
    "VEIENTANVS": {"A": ["21"]},
    "IBERIA": {"N": ["11"]},
    "ACERRONIVS": {"N": ["12"]},
    "EVPHOLIVS": {"N": ["12"]},
    "PELEVS": {"N": ["12"]},
    "PETRONIVS": {"N": ["12"]},
    "ATTA": {"N": ["11"]},
    "SIPHNVS": {"N": ["12"]},
    "LARISA": {"N": ["11"]},
    "MALACA": {"N": ["11"]},
    "AQVINVM": {"N": ["12"]},
    "SARMATICVS": {"A": ["21"]},
    "LESBIA": {"N": ["11"]},
    "ABDVS": {"N": ["12"]},
    "SILENA": {"N": ["11"]},
    "TIBVLLVS": {"N": ["12"]},
    "PYTHODORVS": {"N": ["12"]},
    "ANTIOCHINVS": {"A": ["21"]},
    "PANTVLEIVS": {"N": ["12"]},
    "MYLASA": {"N": ["11"]},
    "THRESSVS": {"A": ["21"]},
    "ARABICVS": {"A": ["21"]},
    "SENIVS": {"A": ["21"]},
    "GNATIA": {"N": ["11"]},
    "ACONTEVS": {"N": ["12"]},
    "INACHVS": {"N": ["12"]},
    "VATINIANVS": {"A": ["21"]},
    "CAVCASIVS": {"A": ["21"]},
    "PLEVRATVS": {"N": ["12"]},
    "CARIA": {"N": ["11"]},
    "ARDVENNA": {"N": ["11"]},
    "TEVTOMATVS": {"N": ["12"]},
    "SMINTHEVS": {"N": ["12"]},
    "ASSYRIVS": {"A": ["21"]},
    "LIGARIVS": {"N": ["12"]},
    "STRATVS": {"N": ["12"]},
    "GNOSIA": {"N": ["11"]},
    "NOLA": {"N": ["11"]},
    "CANOPVS": {"N": ["12"]},
    "TELEPHVS": {"N": ["12"]},
    "TICINVM": {"N": ["12"]},
    "AMPHIARAVS": {"N": ["12"]},
    "EPIDAVRVM": {"N": ["12"]},
    "VACALVS": {"N": ["12"]},
    "CYMODOCEA": {"N": ["11"]},
    "BESSVS": {"N": ["12"]},
    "GARGANVS": {"N": ["12"]},
    "OCRICVLANVM": {"N": ["12"]},
    "SOSIANVS": {"N": ["12"]},
    "CENSORINVS": {"N": ["12"]},
    "DECIDIVS": {"N": ["12"]},
    "ATTIA": {"N": ["11"]},
    "AVRIA": {"N": ["11"]},
    "MACHAONIVS": {"A": ["21"]},
    "AMERINVS": {"A": ["21"]},
    "ARGIVS": {"N": ["12"]},
    "VMBRICIVS": {"N": ["12"]},
    "CLVVIENVS": {"N": ["12"]},
    "ARISTARCHVS": {"N": ["12"]},
    "MALLOVENDVS": {"N": ["12"]},
    "MVNDA": {"N": ["11"]},
    "ARCHIACVS": {"A": ["21"]},
    "CRETIS": {"A": ["24"]},
    "LYDIA": {"N": ["11"]},
    "ARGANTHVS": {"N": ["12"]},
    "MAENALVS": {"N": ["12"]},
    "AONIVS": {"A": ["21"]},
    "LEMVRIA": {"N": ["11"]},
    "CELTAE": {"N": ["11"]},
    "NVMIDAE": {"N": ["11"]},
    "FORMIAE": {"N": ["11"]},
    "SVIONAE": {"N": ["11"]},
    "VOLCAE": {"N": ["11"]},
    "ARSACIDAE": {"N": ["11"]},
    "VERGILIAE": {"N": ["11"]},
    "CYRENAE": {"N": ["11"]},
    "SAVROMATAE": {"N": ["11"]},
    "RVFRAE": {"N": ["11"]},
    "PHARITAE": {"N": ["11"]},
    "ARVAE": {"N": ["11"]},
    "LAPITHAE": {"N": ["11"]},
    "SALONAE": {"N": ["11"]},
    "NERSAE": {"N": ["11"]},
    "CARSVLAE": {"N": ["11"]},
    "CLVVIAE": {"N": ["11"]},
    "CYANEAE": {"N": ["11"]},
    "AENEADAE": {"N": ["11"]},
    "SYBARITAE": {"N": ["11"]},
    "STABIAE": {"N": ["11"]},
    "AEGEAE": {"N": ["11"]},
    "ORNYTVS": {"N": ["12"]},
    "CARINAE": {"N": ["11"]},
    "DAHAE": {"N": ["11"]},
    "BORYSTHENIDAE": {"N": ["11"]},
    "GORTVAE": {"N": ["11"]},
    "NAPAEAE": {"N": ["11"]},
    "BOVILLAE": {"N": ["11"]},
    "ANTENORIDAE": {"N": ["11"]},
    "IVGVLAE": {"N": ["11"]},
    "MINYAE": {"N": ["11"]},
    "MYCENAE": {"N": ["11"]},
    "VLVBRAE": {"N": ["11"]},
    "BISALTAE": {"N": ["11"]},
    "PHASELITAE": {"N": ["11"]},
    "BASTARNAE": {"N": ["11"]},
    "LYNCESTAE": {"N": ["11"]},
    "CELAENAE": {"N": ["11"]},
    "CIETAE": {"N": ["11"]},
    "AVSONIDAE": {"N": ["11"]},
    "PARAPAMISADAE": {"N": ["11"]},
    "SARMATAE": {"N": ["11"]},
    "BRANCHIDAE": {"N": ["11"]},
    "ORESTAE": {"N": ["11"]},
    "LIBVRNAE": {"N": ["11"]},
    "FAESVLAE": {"N": ["11"]},
    "BELITAE": {"N": ["11"]},
    "THESIDAE": {"N": ["11"]},
    "VOLATERRAE": {"N": ["11"]},
    "TEGEATAE": {"N": ["11"]},
    "MAZAGAE": {"N": ["11"]},
    "EVMOLPIDAE": {"N": ["11"]},
    "BELGAE": {"N": ["11"]},
    "ARGINVSAE": {"N": ["11"]},
    "OXIONAE": {"N": ["11"]},
    "OCHVS": {"N": ["12"]},
    "CLVPEAE": {"N": ["11"]},
    "TELEBOAE": {"N": ["11"]},
    "THERMOPYLAE": {"N": ["11"]},
    "CLEONAE": {"N": ["11"]},
    "DANAE": {"N": ["11"]},
    "ARCHELAVS": {"N": ["12"]},
    "SVDRACAE": {"N": ["11"]},
    "PLATAEAE": {"N": ["11"]},
    "PYLAE": {"N": ["11"]},
    "SABARCAE": {"N": ["11"]},
    "CANNAE": {"N": ["11"]},
    "SAMARITAE": {"N": ["11"]},
    "ODRYSAE": {"N": ["11"]},
    "PISAVRVM": {"N": ["12"]},
    "EVERGETAE": {"N": ["11"]},
    "PIRVSTAE": {"N": ["11"]},
    "AEGINETAE": {"N": ["11"]},
    "TYRRHIDAE": {"N": ["11"]},
    "MINTVRNAE": {"N": ["11"]},
    "PHOENISSAE": {"N": ["11"]},
    "SCAEAE": {"N": ["11"]},
    "GORNEAE": {"N": ["11"]},
    "DALMATAE": {"N": ["11"]},
    "OEBALIDAE": {"N": ["11"]},
    "PETREIANVS": {"A": ["21"]},
    "ARGONAVTAE": {"N": ["11"]},
    "THEBAE": {"N": ["11"]},
    "HELVECONAE": {"N": ["11"]},
    "HERACLEOTAE": {"N": ["11"]},
    "CECROPIDAE": {"N": ["11"]},
    "ANTEMNAE": {"N": ["11"]},
    "TABAE": {"N": ["11"]},
    "DENSELETAE": {"N": ["11"]},
    "SABINAE": {"N": ["11"]},
    "ALOIDAE": {"N": ["11"]},
    "BAPTAE": {"N": ["11"]},
    "GRAVISCAE": {"N": ["11"]},
    "MYLAE": {"N": ["11"]},
    "CERCETAE": {"N": ["11"]},
    "CRANAE": {"N": ["11"]},
    "ALAVDAE": {"N": ["11"]},
    "MASSAGETAE": {"N": ["11"]},
    "LVRIVS": {"N": ["12"]},
    "GEMONIAE": {"N": ["11"]},
    "ACERRAE": {"N": ["11"]},
    "HORITAE": {"N": ["11"]},
    "ARABITAE": {"N": ["11"]},
    "CENTVMCELLAE": {"N": ["11"]},
    "CALYDNAE": {"N": ["11"]},
    "GANGARIDAE": {"N": ["11"]},
    "PISAE": {"N": ["11"]},
    "APHIDNAE": {"N": ["11"]},
    "COELALETAE": {"N": ["11"]},
    "TYNDARIDAE": {"N": ["11"]},
    "VERCELLAE": {"N": ["11"]},
    "DRANGAE": {"N": ["11"]},
    "ROMVLIDAE": {"N": ["11"]},
    "BIAE": {"N": ["11"]},
    "CANOPITAE": {"N": ["11"]},
    "SACAE": {"N": ["11"]},
    "GALATAE": {"N": ["11"]},
    "CASPIAE": {"N": ["11"]},
    "CAPREAE": {"N": ["11"]},
    "CLAZOMENAE": {"N": ["11"]},
    "ONCHAE": {"N": ["11"]},
    "PASIPHAE": {"N": ["11"]},
    "PISIDAE": {"N": ["11"]},
    "MAEONIDAE": {"N": ["11"]},
    "AMYCLAE": {"N": ["11"]},
    "DANDARIDAE": {"N": ["11"]},
    "DANAIDAE": {"N": ["11"]},
    "PHOCAEI": {"N": ["12"]},
    "TRIBALLI": {"N": ["12"]},
    "SYCAMBRI": {"N": ["12"]},
    "VBII": {"N": ["12"]},
    "TERMESTINI": {"N": ["12"]},
    "ALPINI": {"N": ["12"]},
    "HIBERI": {"N": ["12"]},
    "FAVNI": {"N": ["12"]},
    "ARGOLICI": {"N": ["12"]},
    "TIANI": {"N": ["12"]},
    "CHAMAVI": {"N": ["12"]},
    "ARECOMICI": {"N": ["12"]},
    "HERMVNDVRI": {"N": ["12"]},
    "BRITANNI": {"N": ["12"]},
    "POMPEII": {"N": ["12"]},
    "DII": {"N": ["12"]},
    "AVSCI": {"N": ["12"]},
    "COMMAGENI": {"N": ["12"]},
    "THVRII": {"N": ["12"]},
    "MELDI": {"N": ["12"]},
    "SEDOCHEZI": {"N": ["12"]},
    "CALETI": {"N": ["12"]},
    "EPICVREI": {"N": ["12"]},
    "GRAI": {"N": ["12"]},
    "HAMMONII": {"N": ["12"]},
    "SAPAEI": {"N": ["12"]},
    "AGATHYRSI": {"N": ["12"]},
    "SABELLI": {"N": ["12"]},
    "PARTHIENI": {"N": ["12"]},
    "CINITHII": {"N": ["12"]},
    "VXII": {"N": ["12"]},
    "NOLANI": {"N": ["12"]},
    "HELVII": {"N": ["12"]},
    "BRVNDISINI": {"N": ["12"]},
    "ADVATICI": {"N": ["12"]},
    "MALLI": {"N": ["12"]},
    "TAVRINI": {"N": ["12"]},
    "OSISMI": {"N": ["12"]},
    "MATTIACI": {"N": ["12"]},
    "TVRONI": {"N": ["12"]},
    "SEGVSIAVI": {"N": ["12"]},
    "AVLERCI": {"N": ["12"]},
    "SVLCITANI": {"N": ["12"]},
    "TVLINGI": {"N": ["12"]},
    "VELLAVII": {"N": ["12"]},
    "SITHONII": {"N": ["12"]},
    "DARDANI": {"N": ["12"]},
    "PERGAMENI": {"N": ["12"]},
    "PRAESTI": {"N": ["12"]},
    "DECANGI": {"N": ["12"]},
    "MARSACI": {"N": ["12"]},
    "CARALITANI": {"N": ["12"]},
    "TVNGRI": {"N": ["12"]},
    "BOTTIAEI": {"N": ["12"]},
    "TEVTONI": {"N": ["12"]},
    "MAGI": {"N": ["12"]},
    "RIPAEI": {"N": ["12"]},
    "PHILADELPHENI": {"N": ["12"]},
    "ACYLLITANI": {"N": ["12"]},
    "PARTHINI": {"N": ["12"]},
    "AFRANIANI": {"N": ["12"]},
    "CADVRCI": {"N": ["12"]},
    "SVEVI": {"N": ["12"]},
    "GRAECVLI": {"N": ["12"]},
    "ARACHOSII": {"N": ["12"]},
    "CHASVARII": {"N": ["12"]},
    "VSIPI": {"N": ["12"]},
    "MARSI": {"N": ["12"]},
    "MARRVCINI": {"N": ["12"]},
    "TELEBOI": {"N": ["12"]},
    "VATINIANI": {"N": ["12"]},
    "SANTONI": {"N": ["12"]},
    "THERAEI": {"N": ["12"]},
    "GELONI": {"N": ["12"]},
    "MAGYNI": {"N": ["12"]},
    "CASSI": {"N": ["12"]},
    "REATINI": {"N": ["12"]},
    "MENAPII": {"N": ["12"]},
    "FABRATERNI": {"N": ["12"]},
    "CHALDEI": {"N": ["12"]},
    "TERGESTINI": {"N": ["12"]},
    "MAVRI": {"N": ["12"]},
    "MARSIGNI": {"N": ["12"]},
    "CENOMANI": {"N": ["12"]},
    "PTIANII": {"N": ["12"]},
    "TREVIRI": {"N": ["12"]},
    "ARIMASPI": {"N": ["12"]},
    "MOSCHI": {"N": ["12"]},
    "LEMOVII": {"N": ["12"]},
    "POMETII": {"N": ["12"]},
    "SYRII": {"N": ["12"]},
    "MEGARICI": {"N": ["12"]},
    "PYRRHONEI": {"N": ["12"]},
    "ELEVTETI": {"N": ["12"]},
    "ANGRIVARII": {"N": ["12"]},
    "PRAESTVTIANI": {"N": ["12"]},
    "MANII": {"N": ["12"]},
    "LARISAEI": {"N": ["12"]},
    "ASSYRII": {"N": ["12"]},
    "VINDELICI": {"N": ["12"]},
    "SARDIANI": {"N": ["12"]},
    "VELIOCASSI": {"N": ["12"]},
    "LANGOBARDI": {"N": ["12"]},
    "HAEDVI": {"N": ["12"]},
    "ALEXANDRINI": {"N": ["12"]},
    "GEIDVMNI": {"N": ["12"]},
    "ARVERNI": {"N": ["12"]},
    "AEGYPTII": {"N": ["12"]},
    "CHALCEDONII": {"N": ["12"]},
    "GALLOGRAECI": {"N": ["12"]},
    "ILII": {"N": ["12"]},
    "MEDIOMATRICI": {"N": ["12"]},
    "HESPERII": {"N": ["12"]},
    "DACI": {"N": ["12"]},
    "AVSETANI": {"N": ["12"]},
    "PELASGI": {"N": ["12"]},
    "GORDYAEI": {"N": ["12"]},
    "COLOPHONII": {"N": ["12"]},
    "PLATONICI": {"N": ["12"]},
    "PETELINI": {"N": ["12"]},
    "VERAGRI": {"N": ["12"]},
    "AMANTINI": {"N": ["12"]},
    "SIRACI": {"N": ["12"]},
    "CAMPANI": {"N": ["12"]},
    "ITALI": {"N": ["12"]},
    "OPICI": {"N": ["12"]},
    "MARDI": {"N": ["12"]},
    "MOSSYNI": {"N": ["12"]},
    "VOLATERRANI": {"N": ["12"]},
    "SABAEI": {"N": ["12"]},
    "NVMANTINI": {"N": ["12"]},
    "BOEOTI": {"N": ["12"]},
    "LIBVRNI": {"N": ["12"]},
    "ITVRAEI": {"N": ["12"]},
    "FENNI": {"N": ["12"]},
    "FIRMANI": {"N": ["12"]},
    "OLYMPENI": {"N": ["12"]},
    "CONDRVSI": {"N": ["12"]},
    "CIRCEII": {"N": ["12"]},
    "CLAVDIOPOLITANI": {"N": ["12"]},
    "TIGVRINI": {"N": ["12"]},
    "ESSVBII": {"N": ["12"]},
    "COLCHI": {"N": ["12"]},
    "ANNICERII": {"N": ["12"]},
    "SALAPINI": {"N": ["12"]},
    "ALLOBROGICI": {"N": ["12"]},
    "CEDROSII": {"N": ["12"]},
    "SVEBI": {"N": ["12"]},
    "OSI": {"N": ["12"]},
    "MORINI": {"N": ["12"]},
    "BOII": {"N": ["12"]},
    "PANNONII": {"N": ["12"]},
    "BAETICI": {"N": ["12"]},
    "BESSI": {"N": ["12"]},
    "CVGERNI": {"N": ["12"]},
    "LVSITANI": {"N": ["12"]},
    "AFRI": {"N": ["12"]},
    "GABALI": {"N": ["12"]},
    "SEDVNI": {"N": ["12"]},
    "CORSI": {"N": ["12"]},
    "LEVCOSYRI": {"N": ["12"]},
    "LANVVINI": {"N": ["12"]},
    "INDI": {"N": ["12"]},
    "MOSTENI": {"N": ["12"]},
    "CAEROSI": {"N": ["12"]},
    "PARENTANI": {"N": ["12"]},
    "SOLONI": {"N": ["12"]},
    "IVDAEI": {"N": ["12"]},
    "SOLYMI": {"N": ["12"]},
    "ANAGNINI": {"N": ["12"]},
    "NABATAEI": {"N": ["12"]},
    "CERAVNII": {"N": ["12"]},
    "SVSIANI": {"N": ["12"]},
    "SICANI": {"N": ["12"]},
    "SOGDIANI": {"N": ["12"]},
    "TYRII": {"N": ["12"]},
    "VRII": {"N": ["12"]},
    "ACHAICI": {"N": ["12"]},
    "LEVCI": {"N": ["12"]},
    "LVGII": {"N": ["12"]},
    "REMI": {"N": ["12"]},
    "IBERI": {"N": ["12"]},
    "LEPTITANI": {"N": ["12"]},
    "CHATTI": {"N": ["12"]},
    "CIMBRI": {"N": ["12"]},
    "GARVMNI": {"N": ["12"]},
    "THESVNTI": {"N": ["12"]},
    "IADESTINI": {"N": ["12"]},
    "SOLI": {"N": ["12"]},
    "PECOLI": {"N": ["12"]},
    "FRENTRANI": {"N": ["12"]},
    "APAMENI": {"N": ["12"]},
    "SARDI": {"N": ["12"]},
    "HYPAEPENI": {"N": ["12"]},
    "SIMVI": {"N": ["12"]},
    "TEMNII": {"N": ["12"]},
    "SIBI": {"N": ["12"]},
    "AEDVI": {"N": ["12"]},
    "NERVII": {"N": ["12"]},
    "CALAGVRRITANI": {"N": ["12"]},
    "SALENTINI": {"N": ["12"]},
    "CENIMAGNI": {"N": ["12"]},
    "IVLIOPOLITANI": {"N": ["12"]},
    "GAETVLI": {"N": ["12"]},
    "HELISII": {"N": ["12"]},
    "QVADI": {"N": ["12"]},
    "CAEMANI": {"N": ["12"]},
    "AMBARRI": {"N": ["12"]},
    "PVTEOLANI": {"N": ["12"]},
    "BRVTTII": {"N": ["12"]},
    "SEDVSII": {"N": ["12"]},
    "PHILIPPI": {"N": ["12"]},
    "AMBIVARETI": {"N": ["12"]},
    "LIBVI": {"N": ["12"]},
    "CHAVCI": {"N": ["12"]},
    "SALAMINII": {"N": ["12"]},
    "AMBIBARII": {"N": ["12"]},
    "MACROBII": {"N": ["12"]},
    "SEQVANI": {"N": ["12"]},
    "RVGII": {"N": ["12"]},
    "LATOBRIGI": {"N": ["12"]},
    "MOESICI": {"N": ["12"]},
    "RHOXOLANI": {"N": ["12"]},
    "ATII": {"N": ["12"]},
    "DOMITIANI": {"N": ["12"]},
    "MYNDII": {"N": ["12"]},
    "OSCI": {"N": ["12"]},
    "TENCTHERI": {"N": ["12"]},
    "VENETHI": {"N": ["12"]},
    "BELLOVACI": {"N": ["12"]},
    "MVSICANI": {"N": ["12"]},
    "GOMPHI": {"N": ["12"]},
    "CYRENAICI": {"N": ["12"]},
    "CERCINITANI": {"N": ["12"]},
    "PLEVMOXII": {"N": ["12"]},
    "STYMPHALIDAE": {"N": ["11"]},
    "TAVRI": {"N": ["12"]},
    "PYTHAGOREI": {"N": ["12"]},
    "VOLSINII": {"N": ["12"]},
    "FERENTANI": {"N": ["12"]},
    "ADIABENI": {"N": ["12"]},
    "VNELLI": {"N": ["12"]},
    "LEXOVII": {"N": ["12"]},
    "TAPVRI": {"N": ["12"]},
    "LEPIDANI": {"N": ["12"]},
    "GALILAEI": {"N": ["12"]},
    "IDAEI": {"N": ["12"]},
    "SVNVCI": {"N": ["12"]},
    "DELIACI": {"N": ["12"]},
    "CRVSTVMERI": {"N": ["12"]},
    "GABINIANI": {"N": ["12"]},
    "ANGLII": {"N": ["12"]},
    "ARGIVI": {"N": ["12"]},
    "IVLIANI": {"N": ["12"]},
    "LOCRI": {"N": ["12"]},
    "CHERVSCI": {"N": ["12"]},
    "CASSIANI": {"N": ["12"]},
    "AETOLI": {"N": ["12"]},
    "AVSONII": {"N": ["12"]},
    "GOLGI": {"N": ["12"]},
    "GAMBRIVII": {"N": ["12"]},
    "TRANSRHENANI": {"N": ["12"]},
    "COSSAEI": {"N": ["12"]},
    "HYRCANI": {"N": ["12"]},
    "AMERINI": {"N": ["12"]},
    "PANHORMITANI": {"N": ["12"]},
    "TYRRHENI": {"N": ["12"]},
    "POMEIANI": {"N": ["12"]},
    "ADELPHI": {"N": ["12"]},
    "COTINI": {"N": ["12"]},
    "IGVINI": {"N": ["12"]},
    "MEDI": {"N": ["12"]},
    "MANIMI": {"N": ["12"]},
    "TVSCVLANI": {"N": ["12"]},
    "CAESARIANI": {"N": ["12"]},
    "VANDILII": {"N": ["12"]},
    "VARINI": {"N": ["12"]},
    "SMYRNAEI": {"N": ["12"]},
    "ICENI": {"N": ["12"]},
    "GEMELLI": {"N": ["12"]},
    "PANNONICI": {"N": ["12"]},
    "TENII": {"N": ["12"]},
    "TAPSITANI": {"N": ["12"]},
    "CIMMERII": {"N": ["12"]},
    "HELVETII": {"N": ["12"]},
    "AFRICANI": {"N": ["12"]},
    "SOSII": {"N": ["12"]},
    "CELTIBERI": {"N": ["12"]},
    "CARMANI": {"N": ["12"]},
    "AVERNI": {"N": ["12"]},
    "VOCONTII": {"N": ["12"]},
    "CHORASMII": {"N": ["12"]},
    "ANDECAVI": {"N": ["12"]},
    "HELLVSII": {"N": ["12"]},
    "SIDICINI": {"N": ["12"]},
    "EOI": {"N": ["12"]},
    "TRANSPADANI": {"N": ["12"]},
    "BRVCTERI": {"N": ["12"]},
    "MASSYLI": {"N": ["12"]},
    "NARISTI": {"N": ["12"]},
    "MEMACENI": {"N": ["12"]},
    "POMPEI": {"N": ["12"]},
    "PEVCINI": {"N": ["12"]},
    "TARBELLI": {"N": ["12"]},
    "DANAI": {"N": ["12"]},
    "MYTILENAEI": {"N": ["12"]},
    "CAVNII": {"N": ["12"]},
    "GADITANI": {"N": ["12"]},
    "PAELIGNI": {"N": ["12"]},
    "CVMANI": {"N": ["12"]},
    "BREVNI": {"N": ["12"]},
    "CADVSII": {"N": ["12"]},
    "MARCOMANI": {"N": ["12"]},
    "PHILAENI": {"N": ["12"]},
    "MAEDI": {"N": ["12"]},
    "LAODICENI": {"N": ["12"]},
    "PATAVINI": {"N": ["12"]},
    "HADRVMETINI": {"N": ["12"]},
    "PRASII": {"N": ["12"]},
    "AQVITANI": {"N": ["12"]},
    "TRALLIANI": {"N": ["12"]},
    "HALICARNASSII": {"N": ["12"]},
    "DYRRACHINI": {"N": ["12"]},
    "MESSENII": {"N": ["12"]},
    "VIROMANDVI": {"N": ["12"]},
    "SEBAGNINI": {"N": ["12"]},
    "AESTII": {"N": ["12"]},
    "AMPSIVARII": {"N": ["12"]},
    "BABYLONII": {"N": ["12"]},
    "AORSI": {"N": ["12"]},
    "LVCANI": {"N": ["12"]},
    "SYNEPHEBI": {"N": ["12"]},
    "ELYMAEI": {"N": ["12"]},
    "EDONI": {"N": ["12"]},
    "BAETASII": {"N": ["12"]},
    "HYPERBOREI": {"N": ["12"]},
    "SEGNI": {"N": ["12"]},
    "BARCAEI": {"N": ["12"]},
    "HENIOCHI": {"N": ["12"]},
    "ARGI": {"N": ["12"]},
    "EPIGONI": {"N": ["12"]},
    "VMBRI": {"N": ["12"]},
    "ASCVLANI": {"N": ["12"]},
    "OMBI": {"N": ["12"]},
    "AMASTRIANI": {"N": ["12"]},
    "LOTOPHAGI": {"N": ["12"]},
    "ARAVISCI": {"N": ["12"]},
    "PELOPONNESII": {"N": ["12"]},
    "PHOCII": {"N": ["12"]},
    "VACCAEI": {"N": ["12"]},
    "RAETI": {"N": ["12"]},
    "ALANI": {"N": ["12"]},
    "THYSDRITANI": {"N": ["12"]},
    "MEGALOPOLITANI": {"N": ["12"]},
    "SEGONTIACI": {"N": ["12"]},
    "ERETRICI": {"N": ["12"]},
    "HENETI": {"N": ["12"]},
    "ARMENII": {"N": ["12"]},
    "ABII": {"N": ["12"]},
    "CARSEOLI": {"N": ["12"]},
    "FRISII": {"N": ["12"]},
    "PYRGI": {"N": ["12"]},
    "GRVDII": {"N": ["12"]},
    "FVNDI": {"N": ["12"]},
    "DVLICHII": {"N": ["12"]},
    "RVBI": {"N": ["12"]},
    "NAHANARVALI": {"N": ["12"]},
    "RVTENI": {"N": ["12"]},
    "TRIBOCI": {"N": ["12"]},
    "ARRETINI": {"N": ["12"]},
    "MANDVBII": {"N": ["12"]},
    "AMISENI": {"N": ["12"]},
    "MVSVLAMII": {"N": ["12"]},
    "MYSI": {"N": ["12"]},
    "FOSI": {"N": ["12"]},
    "FALISCI": {"N": ["12"]},
    "BITHYNI": {"N": ["12"]},
    "BIBROCI": {"N": ["12"]},
    "ACHAEI": {"N": ["12"]},
    "VICETINI": {"N": ["12"]},
    "REVDIGNI": {"N": ["12"]},
    "PLACENTINI": {"N": ["12"]},
    "SPARTANI": {"N": ["12"]},
    "PANICI": {"N": ["12"]},
    "MOESI": {"N": ["12"]},
    "TENCTERI": {"N": ["12"]},
    "POMPEIANI": {"N": ["12"]},
    "MINYI": {"N": ["12"]},
    "TEVCRI": {"N": ["12"]},
    "BAVLI": {"N": ["12"]},
    "ATTICI": {"N": ["12"]},
    "CASPII": {"N": ["12"]},
    "LEVACI": {"N": ["12"]},
    "HARII": {"N": ["12"]},
    "TVRDETANI": {"N": ["12"]},
    "CHII": {"N": ["12"]},
    "THESSALI": {"N": ["12"]},
    "ELESIODVLI": {"N": ["12"]},
    "AMBILIATI": {"N": ["12"]},
    "METHYMNAEI": {"N": ["12"]},
    "RAVRACI": {"N": ["12"]},
    "OCTAVIANI": {"N": ["12"]},
    "AVGVSTIANI": {"N": ["12"]},
    "ALBICI": {"N": ["12"]},
    "GERMANI": {"N": ["12"]},
    "TROGODYTAE": {"N": ["11"]},
    "NEAPOLITANI": {"N": ["12"]},
    "NORICI": {"N": ["12"]},
    "AGRIANI": {"N": ["12"]},
    "BOSPORANI": {"N": ["12"]},
    "TARTESSII": {"N": ["12"]},
    "FLORENTINI": {"N": ["12"]},
    "ILLYRII": {"N": ["12"]},
    "ORDOVICES": {"N": ["13"]},
    "LAERTES": {"N": ["13"]},
    "ACESINES": {"N": ["13"]},
    "HEDYMELES": {"N": ["13"]},
    "MITHRES": {"N": ["13"]},
    "NICARCHIDES": {"N": ["13"]},
    "EPIPHANES": {"N": ["13"]},
    "CINYRADES": {"N": ["13"]},
    "CORIOSOLITES": {"N": ["13"]},
    "PRAEXASPES": {"N": ["13"]},
    "AMPHICTYONES": {"N": ["13"]},
    "ARSACIDES": {"N": ["13"]},
    "TERIOLTES": {"N": ["13"]},
    "LEMNIADES": {"N": ["13"]},
    "ONOMASTORIDES": {"N": ["13"]},
    "MENETES": {"N": ["13"]},
    "CALES": {"N": ["13"]},
    "NARNIENSES": {"N": ["13"]},
    "TANTALIDES": {"N": ["13"]},
    "LADES": {"N": ["13"]},
    "COCOSATES": {"N": ["13"]},
    "THAPSITANI": {"N": ["12"]},
    "MYRMIDONES": {"N": ["13"]},
    "SOPHITES": {"N": ["13"]},
    "THESSALONICENSES": {"N": ["13"]},
    "CIRTENSES": {"N": ["13"]},
    "SVLMONENSES": {"N": ["13"]},
    "CAMERTES": {"N": ["13"]},
    "CARMONENSES": {"N": ["13"]},
    "GOTONES": {"N": ["13"]},
    "ACMONENSES": {"N": ["13"]},
    "HERACLIDES": {"N": ["13"]},
    "DEMOCRATES": {"N": ["13"]},
    "PONTICI": {"N": ["12"]},
    "VRSAONENSES": {"N": ["13"]},
    "PHERETIADES": {"N": ["13"]},
    "METHYMNIADES": {"N": ["13"]},
    "MEDATES": {"N": ["13"]},
    "PHTHIOTES": {"N": ["13"]},
    "ATRIDES": {"N": ["13"]},
    "AONIDES": {"N": ["13"]},
    "PARRACES": {"N": ["13"]},
    "IXIONIDES": {"N": ["13"]},
    "ANTICLES": {"N": ["13"]},
    "DRYADES": {"N": ["13"]},
    "GARAMANTES": {"N": ["13"]},
    "DROPIDES": {"N": ["13"]},
    "ATTALENSES": {"N": ["13"]},
    "ZAMENSES": {"N": ["13"]},
    "ORSINES": {"N": ["13"]},
    "PYRRHIADES": {"N": ["13"]},
    "LYCVRGIDES": {"N": ["13"]},
    "ANCHISIADES": {"N": ["13"]},
    "INTERAMNATES": {"N": ["13"]},
    "ARII": {"N": ["12"]},
    "ERINDES": {"N": ["13"]},
    "EVNONES": {"N": ["13"]},
    "MINTVRNENSES": {"N": ["13"]},
    "LEPTINES": {"N": ["13"]},
    "POLYBOETES": {"N": ["13"]},
    "DRANCES": {"N": ["13"]},
    "ARSAMES": {"N": ["13"]},
    "EVANTHES": {"N": ["13"]},
    "LAOMEDONTIADES": {"N": ["13"]},
    "BRONTES": {"N": ["13"]},
    "NAMNETES": {"N": ["13"]},
    "OILIADES": {"N": ["13"]},
    "NAVSIPHANES": {"N": ["13"]},
    "HYDASPES": {"N": ["13"]},
    "ERINNYES": {"N": ["13"]},
    "EBVRONES": {"N": ["13"]},
    "SAMNITES": {"N": ["13"]},
    "GATES": {"N": ["13"]},
    "IACETANI": {"N": ["12"]},
    "SOTIATES": {"N": ["13"]},
    "AEOLIDES": {"N": ["13"]},
    "HIPPOMENES": {"N": ["13"]},
    "SAVROMATES": {"N": ["13"]},
    "ARPI": {"N": ["12"]},
    "ORODES": {"N": ["13"]},
    "PISCES": {"N": ["13"]},
    "MERIONES": {"N": ["13"]},
    "BORYSTHENES": {"N": ["13"]},
    "OAXES": {"N": ["13"]},
    "MALIENSES": {"N": ["13"]},
    "PHRADATES": {"N": ["13"]},
    "CATVRIGES": {"N": ["13"]},
    "AENIDES": {"N": ["13"]},
    "HELICONIDES": {"N": ["13"]},
    "AESCHINES": {"N": ["13"]},
    "LELEGES": {"N": ["13"]},
    "PLEIADES": {"N": ["13"]},
    "ECHECRATIDES": {"N": ["13"]},
    "ASBYTES": {"N": ["13"]},
    "CORDVBENSES": {"N": ["13"]},
    "BVLIDENSES": {"N": ["13"]},
    "TAVRINITES": {"N": ["13"]},
    "TREVERI": {"N": ["12"]},
    "ARTAVASDES": {"N": ["13"]},
    "SABACES": {"N": ["13"]},
    "HAMADRYADES": {"N": ["13"]},
    "TYDIDES": {"N": ["13"]},
    "RHOSACES": {"N": ["13"]},
    "PARMENIDES": {"N": ["13"]},
    "AMBRACIOTES": {"N": ["13"]},
    "MEHERDATES": {"N": ["13"]},
    "NAVTES": {"N": ["13"]},
    "ILIENSES": {"N": ["13"]},
    "IGVVINATES": {"N": ["13"]},
    "GOMPHENSES": {"N": ["13"]},
    "BATAVI": {"N": ["12"]},
    "ORESTES": {"N": ["13"]},
    "LACIADES": {"N": ["13"]},
    "DRYOPES": {"N": ["13"]},
    "SPITAMENES": {"N": ["13"]},
    "HERMOGENES": {"N": ["13"]},
    "THYESTES": {"N": ["13"]},
    "SINNACES": {"N": ["13"]},
    "REDONES": {"N": ["13"]},
    "CATANES": {"N": ["13"]},
    "MOYSES": {"N": ["13"]},
    "VELIOCASSES": {"N": ["13"]},
    "BARZAENTES": {"N": ["13"]},
    "AGRIPPINENSES": {"N": ["13"]},
    "PANTHOIDES": {"N": ["13"]},
    "HARVDES": {"N": ["13"]},
    "ACHAEMENIDES": {"N": ["13"]},
    "TYRIOTES": {"N": ["13"]},
    "IORDANES": {"N": ["13"]},
    "AEGIDES": {"N": ["13"]},
    "PRVSENSES": {"N": ["13"]},
    "PENEIDES": {"N": ["13"]},
    "THEOPHANES": {"N": ["13"]},
    "ARCHIGENES": {"N": ["13"]},
    "APOLLONIDENSES": {"N": ["13"]},
    "VOCATES": {"N": ["13"]},
    "CHOASPES": {"N": ["13"]},
    "CEPHALLENES": {"N": ["13"]},
    "TROADES": {"N": ["13"]},
    "STROPHADES": {"N": ["13"]},
    "ASTASPES": {"N": ["13"]},
    "AQVINATES": {"N": ["13"]},
    "BRVTTIANI": {"N": ["12"]},
    "ARGOLIDES": {"N": ["13"]},
    "PROETIDES": {"N": ["13"]},
    "OENIDES": {"N": ["13"]},
    "ZORSINES": {"N": ["13"]},
    "IAZYGES": {"N": ["13"]},
    "HESPERIDES": {"N": ["13"]},
    "CHARICLES": {"N": ["13"]},
    "TRINOBANTES": {"N": ["13"]},
    "THESIDES": {"N": ["13"]},
    "NASAMONES": {"N": ["13"]},
    "APOECIDES": {"N": ["13"]},
    "IMBRASIDES": {"N": ["13"]},
    "TROIADES": {"N": ["13"]},
    "LAERTIADES": {"N": ["13"]},
    "BELIDES": {"N": ["13"]},
    "SELEVCENSES": {"N": ["13"]},
    "CROTOPIADES": {"N": ["13"]},
    "ABISARES": {"N": ["13"]},
    "CROTONIATES": {"N": ["13"]},
    "THABENENSES": {"N": ["13"]},
    "SITALCES": {"N": ["13"]},
    "CVRETES": {"N": ["13"]},
    "ACARNANES": {"N": ["13"]},
    "BERONES": {"N": ["13"]},
    "INDIGETES": {"N": ["13"]},
    "ARAXES": {"N": ["13"]},
    "STEROPES": {"N": ["13"]},
    "PITTHIDES": {"N": ["13"]},
    "PHAETHONTIADES": {"N": ["13"]},
    "LEOPREPIDES": {"N": ["13"]},
    "TMOLITES": {"N": ["13"]},
    "SOPHOCLES": {"N": ["13"]},
    "AMYNTIADES": {"N": ["13"]},
    "ACOETES": {"N": ["13"]},
    "ELVSATES": {"N": ["13"]},
    "MINDYRIDES": {"N": ["13"]},
    "NEALCES": {"N": ["13"]},
    "GYNDES": {"N": ["13"]},
    "ORNOSPADES": {"N": ["13"]},
    "OXYARTES": {"N": ["13"]},
    "BITVRIGES": {"N": ["13"]},
    "NAVCRATES": {"N": ["13"]},
    "ATARNITES": {"N": ["13"]},
    "EVMEDES": {"N": ["13"]},
    "HARMERIDES": {"N": ["13"]},
    "SYRTES": {"N": ["13"]},
    "EVRYMIDES": {"N": ["13"]},
    "TIMOCLES": {"N": ["13"]},
    "ORIGINES": {"N": ["13"]},
    "MITHRENES": {"N": ["13"]},
    "BATTIADES": {"N": ["13"]},
    "ALYATTES": {"N": ["13"]},
    "ALETRINATES": {"N": ["13"]},
    "TARVSATES": {"N": ["13"]},
    "PIERIDES": {"N": ["13"]},
    "CEVTRONES": {"N": ["13"]},
    "MENOETIADES": {"N": ["13"]},
    "BOOTES": {"N": ["13"]},
    "SALMACIDES": {"N": ["13"]},
    "AGYLLINI": {"N": ["12"]},
    "ASTACIDES": {"N": ["13"]},
    "NAVPLIADES": {"N": ["13"]},
    "DATAPHERNES": {"N": ["13"]},
    "LIBETHRIDES": {"N": ["13"]},
    "DORYLENSES": {"N": ["13"]},
    "GERYONES": {"N": ["13"]},
    "MEDOBRIGENSES": {"N": ["13"]},
    "NORENSES": {"N": ["13"]},
    "ANTISTHENES": {"N": ["13"]},
    "HYSTASPES": {"N": ["13"]},
    "PETROCORII": {"N": ["12"]},
    "CHREMES": {"N": ["13"]},
    "VONONES": {"N": ["13"]},
    "SEVTHES": {"N": ["13"]},
    "PHRAATES": {"N": ["13"]},
    "STOECHADES": {"N": ["13"]},
    "EPIMENES": {"N": ["13"]},
    "MASSILIENSES": {"N": ["13"]},
    "ADRYADES": {"N": ["13"]},
    "RHEOMITHRES": {"N": ["13"]},
    "CARES": {"N": ["13"]},
    "CHALYBES": {"N": ["13"]},
    "CATAMANTALOES": {"N": ["13"]},
    "NIPHATES": {"N": ["13"]},
    "VSPENSES": {"N": ["13"]},
    "XERXES": {"N": ["13"]},
    "EVPHRATES": {"N": ["13"]},
    "PEVCESTES": {"N": ["13"]},
    "ELYSII": {"N": ["12"]},
    "PHAEACES": {"N": ["13"]},
    "GENAVNI": {"N": ["12"]},
    "CYCLOPES": {"N": ["13"]},
    "EVDOSES": {"N": ["13"]},
    "VTICENSES": {"N": ["13"]},
    "TEMNITES": {"N": ["13"]},
    "SEMNONES": {"N": ["13"]},
    "IAPYDES": {"N": ["13"]},
    "PEGASIDES": {"N": ["13"]},
    "PARMENSES": {"N": ["13"]},
    "IASIDES": {"N": ["13"]},
    "TROES": {"N": ["13"]},
    "MAZACES": {"N": ["13"]},
    "BRIGANTES": {"N": ["13"]},
    "CAMBYSES": {"N": ["13"]},
    "MASSAGETES": {"N": ["13"]},
    "LVPERCI": {"N": ["12"]},
    "THEODOROMEDES": {"N": ["13"]},
    "IOANNES": {"N": ["13"]},
    "PERIPHANES": {"N": ["13"]},
    "MONAESES": {"N": ["13"]},
    "ARIMAZES": {"N": ["13"]},
    "ONITES": {"N": ["13"]},
    "HIPPOCRATES": {"N": ["13"]},
    "PISTORENSES": {"N": ["13"]},
    "ACHILLES": {"N": ["13"]},
    "THALES": {"N": ["13"]},
    "INSVBRES": {"N": ["13"]},
    "CEPHENES": {"N": ["13"]},
    "IPHICRATES": {"N": ["13"]},
    "AMEDINES": {"N": ["13"]},
    "GRAIOCELI": {"N": ["12"]},
    "OTHRYADES": {"N": ["13"]},
    "ACHATES": {"N": ["13"]},
    "PICTONES": {"N": ["13"]},
    "PHILIPPENSES": {"N": ["13"]},
    "BVRI": {"N": ["12"]},
    "LIBONES": {"N": ["13"]},
    "ATIZYES": {"N": ["13"]},
    "HISPELLATES": {"N": ["13"]},
    "TERPSICHORES": {"N": ["13"]},
    "PENTHELIDES": {"N": ["13"]},
    "ACESTES": {"N": ["13"]},
    "SANTONES": {"N": ["13"]},
    "SISYPHIDES": {"N": ["13"]},
    "DITAMENES": {"N": ["13"]},
    "VELIENSES": {"N": ["13"]},
    "VIENNENSES": {"N": ["13"]},
    "PHYLACIDES": {"N": ["13"]},
    "PANES": {"N": ["13"]},
    "VELLIOCASSES": {"N": ["13"]},
    "NICOCLES": {"N": ["13"]},
    "PLISTHENES": {"N": ["13"]},
    "ANTIPHATES": {"N": ["13"]},
    "ATINATES": {"N": ["13"]},
    "AETHIOPES": {"N": ["13"]},
    "ANDROMENES": {"N": ["13"]},
    "EBVROVICES": {"N": ["13"]},
    "SVGAMBRI": {"N": ["12"]},
    "HERODES": {"N": ["13"]},
    "SARDES": {"N": ["13"]},
    "AVXIMATES": {"N": ["13"]},
    "NICOMEDENSES": {"N": ["13"]},
    "SALLVES": {"N": ["13"]},
    "DEMOSTHENES": {"N": ["13"]},
    "CLINIADES": {"N": ["13"]},
    "OZINES": {"N": ["13"]},
    "AMYNTORIDES": {"N": ["13"]},
    "TIRIDATES": {"N": ["13"]},
    "THVCYDIDES": {"N": ["13"]},
    "HYRTACIDES": {"N": ["13"]},
    "ABORIGINES": {"N": ["13"]},
    "APOLLONIATES": {"N": ["13"]},
    "HARPOCRATES": {"N": ["13"]},
    "POTNIADES": {"N": ["13"]},
    "VSIPETES": {"N": ["13"]},
    "SVARDONES": {"N": ["13"]},
    "SORTIENTES": {"N": ["13"]},
    "SIDONII": {"N": ["12"]},
    "ACHAEMENES": {"N": ["13"]},
    "PHILYRIDES": {"N": ["13"]},
    "FOROIVLIENSES": {"N": ["13"]},
    "RHOEMETALCES": {"N": ["13"]},
    "HELICONIADES": {"N": ["13"]},
    "EMERITENSES": {"N": ["13"]},
    "ARGYRASPIDES": {"N": ["13"]},
    "NITIOBROGES": {"N": ["13"]},
    "TVBANTES": {"N": ["13"]},
    "HISPALENSES": {"N": ["13"]},
    "ETEOCLES": {"N": ["13"]},
    "ANARTES": {"N": ["13"]},
    "PETRAITES": {"N": ["13"]},
    "APHRODISIENSES": {"N": ["13"]},
    "CAPSENSES": {"N": ["13"]},
    "SERES": {"N": ["13"]},
    "TARRACINENSES": {"N": ["13"]},
    "AEGEATES": {"N": ["13"]},
    "ACMONIDES": {"N": ["13"]},
    "SARRASTES": {"N": ["13"]},
    "BASILIDES": {"N": ["13"]},
    "PHINIDES": {"N": ["13"]},
    "ERATOSTHENES": {"N": ["13"]},
    "ANTORES": {"N": ["13"]},
    "LEVCIPPIDES": {"N": ["13"]},
    "DARDANIDES": {"N": ["13"], "A": ["25"]},
    "GRINNES": {"N": ["13"]},
    "CRETES": {"N": ["13"]},
    "AREOPAGITES": {"N": ["13"]},
    "AGGRAMMES": {"N": ["13"]},
    "CENABENSES": {"N": ["13"]},
    "ORCADES": {"N": ["13"]},
    "CHALDAEI": {"N": ["12"]},
    "PARTHYAEI": {"N": ["12"]},
    "HERMES": {"N": ["13"]},
    "ARETES": {"N": ["13"]},
    "TECTOSAGES": {"N": ["13"]},
    "BORESTI": {"N": ["12"]},
    "MITHRACENES": {"N": ["13"]},
    "APOLLONIDES": {"N": ["13"]},
    "MENOETES": {"N": ["13"]},
    "ATREBATES": {"N": ["13"]},
    "LYSIADES": {"N": ["13"]},
    "BAGISTANES": {"N": ["13"]},
    "PHILOCTETES": {"N": ["13"]},
    "NVITONES": {"N": ["13"]},
    "GYGES": {"N": ["13"]},
    "TEVTONES": {"N": ["13"]},
    "MYSTES": {"N": ["13"]},
    "ARISTOPHONTES": {"N": ["13"]},
    "CORONIDES": {"N": ["13"]},
    "ANTIGENES": {"N": ["13"]},
    "GANGARIDES": {"N": ["13"]},
    "NICETES": {"N": ["13"]},
    "PALES": {"N": ["13"]},
    "STRATOCLES": {"N": ["13"]},
    "ELEI": {"N": ["12"]},
    "ALETES": {"N": ["13"]},
    "THYMOETES": {"N": ["13"]},
    "BARGATES": {"N": ["13"]},
    "ABANTIADES": {"N": ["13"]},
    "IDARNES": {"N": ["13"]},
    "CARTEIENSES": {"N": ["13"]},
    "CALLICRATES": {"N": ["13"]},
    "SVIONES": {"N": ["13"]},
    "SISENES": {"N": ["13"]},
    "LYNCESTES": {"N": ["13"]},
    "DEMADES": {"N": ["13"]},
    "PHRYGES": {"N": ["13"]},
    "LEMOVICES": {"N": ["13"]},
    "AMBIANI": {"N": ["12"]},
    "ASPRENATES": {"N": ["13"]},
    "CALETES": {"N": ["13"]},
    "DARES": {"N": ["13"]},
    "PRIAMIDES": {"N": ["13"]},
    "PALAMEDES": {"N": ["13"]},
    "THIMODES": {"N": ["13"]},
    "VASACES": {"N": ["13"]},
    "CAPPADOCES": {"N": ["13"]},
    "SICCENSES": {"N": ["13"]},
    "ORONTES": {"N": ["13"]},
    "MEMMIADES": {"N": ["13"]},
    "ABVLITES": {"N": ["13"]},
    "ERICHAETES": {"N": ["13"]},
    "HAVSTANES": {"N": ["13"]},
    "TRINOVANTES": {"N": ["13"]},
    "BAGOPHANES": {"N": ["13"]},
    "CREMONENSES": {"N": ["13"]},
    "CARNVTES": {"N": ["13"]},
    "HELIADES": {"N": ["13"]},
    "MILTIADES": {"N": ["13"]},
    "SATROPATES": {"N": ["13"]},
    "AVLESTES": {"N": ["13"]},
    "LAESTRYGONES": {"N": ["13"]},
    "SISIMITHRES": {"N": ["13"]},
    "SEGESTES": {"N": ["13"]},
    "PYTHOCLES": {"N": ["13"]},
    "BVRSAVONENSES": {"N": ["13"]},
    "ARTAXERXES": {"N": ["13"]},
    "TAXILES": {"N": ["13"]},
    "SINDES": {"N": ["13"]},
    "ARISTOMEDES": {"N": ["13"]},
    "BYLLIDENSES": {"N": ["13"]},
    "IMMORTALES": {"N": ["13"]},
    "ABDAGAESES": {"N": ["13"]},
    "PHARNACES": {"N": ["13"]},
    "MYGDONIDES": {"N": ["13"]},
    "EPIDAPHNES": {"N": ["13"]},
    "BACTRIANI": {"N": ["12"]},
    "BERECYNTIADES": {"N": ["13"]},
    "EPIRENSES": {"N": ["13"]},
    "VETTONES": {"N": ["13"]},
    "LYCAMBES": {"N": ["13"]},
    "STRATIPPOCLES": {"N": ["13"]},
    "OECLIDES": {"N": ["13"]},
    "CYRENENSES": {"N": ["13"]},
    "AGRIANES": {"N": ["13"]},
    "NABATHES": {"N": ["13"]},
    "SATIBARZANES": {"N": ["13"]},
    "VARDANES": {"N": ["13"]},
    "PAPHLAGONES": {"N": ["13"]},
    "CATAONES": {"N": ["13"]},
    "LEVCATES": {"N": ["13"]},
    "ARIARATHES": {"N": ["13"]},
    "NEREIDES": {"N": ["13"]},
    "RHODIENSES": {"N": ["13"]},
    "LARES": {"N": ["13"]},
    "POLEMOCRATES": {"N": ["13"]},
    "ISOCRATES": {"N": ["13"]},
    "AMPHIAREIADES": {"N": ["13"]},
    "IONES": {"N": ["13"]},
    "SAMOTHRACES": {"N": ["13"]},
    "PHILLYRIDES": {"N": ["13"]},
    "CARALES": {"N": ["13"]},
    "ORONTOBATES": {"N": ["13"]},
    "ZETES": {"N": ["13"]},
    "MIMALLONIDES": {"N": ["13"]},
    "LABEATES": {"N": ["13"]},
    "TIGRANES": {"N": ["13"]},
    "ALCIBIADES": {"N": ["13"]},
    "NVCERINI": {"N": ["12"]},
    "HYPERIDES": {"N": ["13"]},
    "MVNDENSES": {"N": ["13"]},
    "STRATONICENSES": {"N": ["13"]},
    "ZALACES": {"N": ["13"]},
    "AMPHITRYONIADES": {"N": ["13"]},
    "TIBVRTES": {"N": ["13"]},
    "ABISTAMENES": {"N": ["13"]},
    "NICAEENSES": {"N": ["13"]},
    "ANCALITES": {"N": ["13"]},
    "ILIADES": {"N": ["13"], "O": ["17"]},
    "MELICERTES": {"N": ["13"]},
    "SINOPENSES": {"N": ["13"]},
    "HYADES": {"N": ["13"]},
    "LACONES": {"N": ["13"]},
    "NABARZANES": {"N": ["13"]},
    "CYPRII": {"N": ["12"]},
    "OLYMPIONICES": {"N": ["13"]},
    "ERICES": {"N": ["13"]},
    "AEETES": {"N": ["13"]},
    "PANOPES": {"N": ["13"]},
    "RAMESSES": {"N": ["13"]},
    "EVMENIDES": {"N": ["13"]},
    "PICENTES": {"N": ["13"]},
    "PELOPONNENSES": {"N": ["13"]},
    "DEMOCHARES": {"N": ["13"]},
    "CANNINEFATES": {"N": ["13"]},
    "SITHONES": {"N": ["13"]},
    "AVGVSTALES": {"N": ["13"]},
    "SCIPIADES": {"N": ["13"]},
    "EMPEDOCLES": {"N": ["13"]},
    "GOBARES": {"N": ["13"]},
    "CHARES": {"N": ["13"]},
    "DIORES": {"N": ["13"]},
    "BVTES": {"N": ["13"]},
    "TIMAGENES": {"N": ["13"]},
    "SYMPLEGADES": {"N": ["13"]},
    "APHELIOTES": {"N": ["13"]},
    "GANGES": {"N": ["13"]},
    "PARTHI": {"N": ["12"]},
    "PERIPLANES": {"N": ["13"]},
    "NEMETES": {"N": ["13"]},
    "PAMMENES": {"N": ["13"]},
    "POLIORCETES": {"N": ["13"]},
    "LVGDVNENSES": {"N": ["13"]},
    "MAEONIDES": {"N": ["13"]},
    "CALLICRATIDES": {"N": ["13"]},
    "PROTOGENES": {"N": ["13"]},
    "COBARES": {"N": ["13"]},
    "DRAPPES": {"N": ["13"]},
    "OXYDATES": {"N": ["13"]},
    "LINGONES": {"N": ["13"]},
    "LARINATES": {"N": ["13"]},
    "TYRES": {"N": ["13"]},
    "ELELEIDES": {"N": ["13"]},
    "HOMONADENSES": {"N": ["13"]},
    "AGAMEMNONIDES": {"N": ["13"]},
    "ARISTOPHANES": {"N": ["13"]},
    "ACTORIDES": {"N": ["13"]},
    "EVMENES": {"N": ["13"]},
    "HYPSIDES": {"N": ["13"]},
    "DIABLINTES": {"N": ["13"]},
    "ANTIPHANES": {"N": ["13"]},
    "ALTINATES": {"N": ["13"]},
    "NANTVATES": {"N": ["13"]},
    "THEMISTOCLES": {"N": ["13"]},
    "OEENSES": {"N": ["13"]},
    "CICONES": {"N": ["13"]},
    "THYESTIADES": {"N": ["13"]},
    "ISTAEVONES": {"N": ["13"]},
    "PHRATAPHERNES": {"N": ["13"]},
    "OPHELTES": {"N": ["13"]},
    "SVESSIONES": {"N": ["13"]},
    "LYCONIDES": {"N": ["13"]},
    "CLEANTHES": {"N": ["13"]},
    "LYDI": {"N": ["12"]},
    "SOCRATICI": {"N": ["12"]},
    "IZATES": {"N": ["13"]},
    "FICEDVLENSES": {"N": ["13"]},
    "ARSACES": {"N": ["13"]},
    "APAMENSES": {"N": ["13"]},
    "VERONENSES": {"N": ["13"]},
    "ITALICENSES": {"N": ["13"]},
    "ILLVRGAVONENSES": {"N": ["13"]},
    "INACHIDES": {"N": ["13"]},
    "DISCORIDES": {"N": ["13"]},
    "ARISTOMENES": {"N": ["13"]},
    "TITIES": {"N": ["13"]},
    "AVIONES": {"N": ["13"]},
    "ATLANTIDES": {"N": ["13"]},
    "ANDES": {"N": ["13"]},
    "ARABES": {"N": ["13"]},
    "DRYANTIADES": {"N": ["13"]},
    "MARTIALES": {"N": ["13"]},
    "DOLOPES": {"N": ["13"]},
    "PAEONES": {"N": ["13"]},
    "VACCENSES": {"N": ["13"]},
    "SENIENSES": {"N": ["13"]},
    "ORSES": {"N": ["13"]},
    "ARCHIDEMIDES": {"N": ["13"]},
    "ATHAMANES": {"N": ["13"]},
    "BISTONES": {"N": ["13"]},
    "GAVDIMONIVM": {"-": ["12"]},
    "VNISVBSELLIVM": {"-": ["12"]},
    "PIPVLVM": {"-": ["12"]},
    "LVMBIFRAGIVM": {"-": ["12"]},
    "PVGNVM": {"-": ["12"]},
    "PROPYLAEVM": {"-": ["12"]},
    "PRAEDIOLVM": {"-": ["12"]},
    "CONSTRATVM": {"-": ["12"]},
    "PERMAGNVM": {"-": ["12"]},
    "PVLPAMENTVM": {"-": ["12"]},
    "HARPASTES": {"N": ["13"]},
    "HAEMONIDES": {"N": ["13"]},
    "CLOSTELLVM": {"-": ["12"]},
    "PATARACINVM": {"-": ["12"]},
    "AVRICHALCVM": {"-": ["12"]},
    "IDOLVM": {"-": ["12"]},
    "GVRGVSTIVM": {"-": ["12"]},
    "TRIOBOLVM": {"-": ["12"]},
    "EPYTIDES": {"N": ["13"]},
    "AVTOMATVM": {"-": ["12"]},
    "PERPAVXILLVM": {"-": ["12"]},
    "BARBARVM": {"-": ["12"]},
    "BICLINIVM": {"-": ["12"]},
    "STIBADIVM": {"-": ["12"]},
    "LYCOMEDES": {"N": ["13"]},
    "CORYLETVM": {"-": ["12"]},
    "PARRHESIASTES": {"N": ["13"]},
    "SILVRES": {"N": ["13"]},
    "DVLGVBINI": {"N": ["12"]},
    "CONTIGNATVM": {"-": ["12"]},
    "PORCINVM": {"-": ["12"]},
    "ALIORSVM": {"-": ["12"]},
    "LINTEOLVM": {"-": ["12"]},
    "TORCVLVM": {"-": ["12"]},
    "CARNEADES": {"N": ["13"]},
    "PATROCLES": {"N": ["13"]},
    "SYNOECIVM": {"-": ["12"]},
    "BAPTISTERIVM": {"-": ["12"]},
    "PSALTERIVM": {"-": ["12"]},
    "POLITES": {"N": ["13"]},
    "ACHILLIDES": {"N": ["13"]},
    "PELIDES": {"N": ["13"]},
    "CATINVM": {"-": ["12"]},
    "MVSCARIVM": {"-": ["12"]},
    "INGAEVONES": {"N": ["13"]},
    "INTERPRETAMENTVM": {"-": ["12"]},
    "CLAVARIVM": {"-": ["12"]},
    "OEBALIDES": {"N": ["13"]},
    "VIRIDARIVM": {"-": ["12"]},
    "SIBVLATES": {"N": ["13"]},
    "NEXVM": {"-": ["12"]},
    "PENICILLVM": {"-": ["12"]},
    "SIMONIDES": {"N": ["13"]},
    "DEPRAESENTIARVM": {"-": ["12"]},
    "COMPACTVM": {"-": ["12"]},
    "DELIRAMENTVM": {"-": ["12"]},
    "DELECTAMENTVM": {"-": ["12"]},
    "PHERES": {"N": ["13"]},
    "PERSICVM": {"-": ["12"]},
    "HAVSTRVM": {"-": ["12"]},
    "CAVAEDIVM": {"-": ["12"]},
    "REFVGIVM": {"-": ["12"]},
    "CORCILLVM": {"-": ["12"]},
    "AONES": {"N": ["13"]},
    "PRATVLVM": {"-": ["12"]},
    "TRISTIMONIVM": {"-": ["12"]},
    "GALLINICIVM": {"-": ["12"]},
    "FRIGIDVM": {"-": ["12"]},
    "PIGMENTVM": {"-": ["12"]},
    "PANARIVM": {"-": ["12"]},
    "SPHAERISTERIVM": {"-": ["12"]},
    "SVDVM": {"-": ["12"]},
    "PARTVM": {"-": ["12"]},
    "CHIRAMAXIVM": {"-": ["12"]},
    "OSTIOLVM": {"-": ["12"]},
    "SPITACES": {"N": ["13"]},
    "DESIDIABVLVM": {"-": ["12"]},
    "SEXENNIVM": {"-": ["12"]},
    "BELLEROPHONTES": {"N": ["13"]},
    "PISVM": {"-": ["12"]},
    "STERQVILINVM": {"-": ["12"]},
    "METHODIVM": {"-": ["12"]},
    "TERMENTVM": {"-": ["12"]},
    "POSTLIMINIVM": {"-": ["12"]},
    "ARCISELLIVM": {"-": ["12"]},
    "VANGIONES": {"N": ["13"]},
    "VVLTVM": {"-": ["12"]},
    "ALCIDES": {"N": ["13"]},
    "CATILLVM": {"-": ["12"]},
    "SESAMVM": {"-": ["12"]},
    "QVASILLVM": {"-": ["12"]},
    "ALIQVOVORSVM": {"-": ["12"]},
    "MAGNETES": {"N": ["13"]},
    "SEMICINCTIVM": {"-": ["12"]},
    "EVRIPIDES": {"N": ["13"]},
    "LEPONTII": {"N": ["12"]},
    "PYLADES": {"N": ["13"]},
    "OSCENSES": {"N": ["13"]},
    "DIVERBIVM": {"-": ["12"]},
    "TRALLES": {"N": ["13"]},
    "TVRBAMENTVM": {"-": ["12"]},
    "AEACIDES": {"N": ["13"]},
    "CINGILLVM": {"-": ["12"]},
    "IMMERITVM": {"-": ["12"]},
    "LARIDES": {"N": ["13"]},
    "HORAEVM": {"-": ["12"]},
    "DIFFVGIVM": {"-": ["12"]},
    "SVBLIGACVLVM": {"-": ["12"]},
    "NASVM": {"-": ["12"]},
    "SENONES": {"N": ["13"]},
    "BIFVRCVM": {"-": ["12"]},
    "SEPLASIVM": {"-": ["12"]},
    "BLITEVM": {"-": ["12"]},
    "OXATHRES": {"N": ["13"]},
    "CORCVLVM": {"-": ["12"]},
    "VESTICONTVBERNIVM": {"-": ["12"]},
    "LENIMENTVM": {"-": ["12"]},
    "PITTACIVM": {"-": ["12"]},
    "VIRECTVM": {"-": ["12"]},
    "SAXETVM": {"-": ["12"]},
    "NAVCVM": {"-": ["12"]},
    "BASILICVM": {"-": ["12"]},
    "PROPVDIVM": {"-": ["12"]},
    "VTERVM": {"-": ["12"]},
    "CONTICINIVM": {"-": ["12"]},
    "MANTELVM": {"-": ["12"]},
    "CONSIDIVM": {"-": ["12"]},
    "EXVVIVM": {"-": ["12"]},
    "VTROQVEVERSVM": {"-": ["12"]},
    "RHAMNES": {"N": ["13"]},
    "VNCTORIVM": {"-": ["12"]},
    "LITIGIVM": {"-": ["12"]},
    "PISTILLVM": {"-": ["12"]},
    "CLIVVM": {"-": ["12"]},
    "FVSTVARIVM": {"-": ["12"]},
    "LATINIENSES": {"N": ["13"]},
    "CLIPEVM": {"-": ["12"]},
    "ORICALCHVM": {"-": ["12"]},
    "EVRYTIDES": {"N": ["13"]},
    "IENTACVLVM": {"-": ["12"]},
    "STABILIMENTVM": {"-": ["12"]},
    "HIPPOTADES": {"N": ["13"]},
    "PRAEIVDICATVM": {"-": ["12"]},
    "GLANDIVM": {"-": ["12"]},
    "OENOPOLIVM": {"-": ["12"]},
    "VERIVERBIVM": {"-": ["12"]},
    "BVBACES": {"N": ["13"]},
    "CALLISTHENES": {"N": ["13"]},
    "TOLOSATES": {"N": ["13"]},
    "LESBIADES": {"N": ["13"]},
    "CORFINIENSES": {"N": ["13"]},
    "ARIOBARZANES": {"N": ["13"]},
    "CALAMISTRVM": {"-": ["12"]},
    "POEMATIVM": {"-": ["12"]},
    "CVPEDIVM": {"-": ["12"]},
    "DENTIFRANGIBVLVM": {"-": ["12"]},
    "CECROPIDES": {"N": ["13"]},
    "GVSTATORIVM": {"-": ["12"]},
    "EXPLEMENTVM": {"-": ["12"]},
    "MYROPOLIVM": {"-": ["12"]},
    "NASTVRTIVM": {"-": ["12"]},
    "INTEGVMENTVM": {"-": ["12"]},
    "THESTIADES": {"N": ["13"]},
    "IRRIDICVLVM": {"-": ["12"]},
    "ODARIVM": {"-": ["12"]},
    "HOSTIMENTVM": {"-": ["12"]},
    "CRYSTALLVM": {"-": ["12"]},
    "NVCIFRANGIBVLVM": {"-": ["12"]},
    "ASCLEPIADES": {"N": ["13"]},
    "TRIFOLIVM": {"-": ["12"]},
    "MARSVPIVM": {"-": ["12"]},
    "THESAVRVM": {"-": ["12"]},
    "EMBOLVM": {"-": ["12"]},
    "BACCIBALLVM": {"-": ["12"]},
    "PROCRVSTES": {"N": ["13"]},
    "PROMONTVRIVM": {"-": ["12"]},
    "NOMADES": {"N": ["13"]},
    "AMPHITEATRVM": {"-": ["12"]},
    "CORYBANTES": {"N": ["13"]},
    "ILLORSVM": {"-": ["12"]},
    "BISTONIDES": {"N": ["13"]},
    "SVBVRBIVM": {"-": ["12"]},
    "MENDICABVLVM": {"-": ["12"]},
    "SPECVM": {"-": ["12"]},
    "MALEICORIVM": {"-": ["12"]},
    "EVRYBATES": {"N": ["13"]},
    "TELINVM": {"-": ["12"]},
    "GOTARZES": {"N": ["13"]},
    "VASCONES": {"N": ["13"]},
    "LABDACIDES": {"N": ["13"]},
    "CLEOCHARES": {"N": ["13"]},
    "BRANNOVICES": {"N": ["13"]},
    "IDYLLIVM": {"-": ["12"]},
    "GYNAECEVM": {"-": ["12"]},
    "CONSERVITIVM": {"-": ["12"]},
    "DILVCVLVM": {"-": ["12"]},
    "OREADES": {"N": ["13"]},
    "MELCVLVM": {"-": ["12"]},
    "MANVCIOLVM": {"-": ["12"]},
    "STATVNCVLVM": {"-": ["12"]},
    "SISYMBRIVM": {"-": ["12"]},
    "BASIOLVM": {"-": ["12"]},
    "ZARIASPES": {"N": ["13"]},
    "CHARITES": {"N": ["13"]},
    "THERSITES": {"N": ["13"]},
    "AESONIDES": {"N": ["13"]},
    "COLIPHIVM": {"-": ["12"]},
    "OFFENDICVLVM": {"-": ["12"]},
    "ARABARCHES": {"N": ["13"]},
    "MELLICVLVM": {"-": ["12"]},
    "CRVCIAMENTVM": {"-": ["12"]},
    "ALTROVORSVM": {"-": ["12"]},
    "VENALICIVM": {"-": ["12"]},
    "MILIARIVM": {"-": ["12"]},
    "BIGERRIONES": {"N": ["13"]},
    "APODYTERIVM": {"-": ["12"]},
    "CONDITORIVM": {"-": ["12"]},
    "LATERICIVM": {"-": ["12"]},
    "PROSTIBVLVM": {"-": ["12"]},
    "TENTAMENTVM": {"-": ["12"]},
    "OXARTES": {"N": ["13"]},
    "SVBNIMIVM": {"-": ["12"]},
    "FASCINVM": {"-": ["12"]},
    "OXYCOMINVM": {"-": ["12"]},
    "EXOTICVM": {"-": ["12"]},
    "DIOMEDES": {"N": ["13"]},
    "CALIDVM": {"-": ["12"]},
    "BLANDVM": {"-": ["12"]},
    "LIBYES": {"N": ["13"]},
    "CHRYSES": {"N": ["13"]},
    "LESBIDES": {"N": ["13"]},
    "CORYMBIVM": {"-": ["12"]},
    "HERMINONES": {"N": ["13"]},
    "DIARDINES": {"N": ["13"]},
    "MORETVM": {"-": ["12"]},
    "GASTRVM": {"-": ["12"]},
    "TONSTRINVM": {"-": ["12"]},
    "ARPINATES": {"N": ["13"]},
    "PARISII": {"N": ["12"]},
    "INFORTVNIVM": {"-": ["12"]},
    "TRINVNDINVM": {"-": ["12"]},
    "CEREBELLVM": {"-": ["12"]},
    "ELLVM": {"-": ["12"]},
    "FEBRVVM": {"-": ["12"]},
    "LIBRVM": {"-": ["12"]},
    "PHARASMANES": {"N": ["13"]},
    "COPHES": {"N": ["13"]},
    "SENTICETVM": {"-": ["12"]},
    "OFFVLGEO": {"-": ["52"]},
    "SACES": {"N": ["13"]},
    "CAERACATES": {"N": ["13"]},
    "ARGESTES": {"N": ["13"]},
    "DIRIBEO": {"-": ["52"]},
    "SESSORIVM": {"-": ["12"]},
    "MACEO": {"-": ["52"]},
    "PRAEHIBEO": {"-": ["52"]},
    "VERCVLVM": {"-": ["12"]},
    "TITTIBILICIVM": {"-": ["12"]},
    "EXSVGEO": {"-": ["52"]},
    "ALLEVAMENTVM": {"-": ["12"]},
    "ATTORQVEO": {"-": ["52"]},
    "THVCIDIDES": {"N": ["13"]},
    "SPICVM": {"-": ["12"]},
    "ISELASTICVM": {"-": ["12"]},
    "OBOLEO": {"-": ["52"]},
    "AFFLEO": {"-": ["52"]},
    "ORICHALCVM": {"-": ["12"]},
    "FRENDEO": {"-": ["52"]},
    "REFERVEO": {"-": ["52"]},
    "SPLENIVM": {"-": ["12"]},
    "CONSVADEO": {"-": ["52"]},
    "ADDENSEO": {"-": ["52"]},
    "LINTEO": {"-": ["52"]},
    "FOETEO": {"-": ["52"]},
    "RECALEO": {"-": ["52"]},
    "TYRIDATES": {"N": ["13"]},
    "CONFVLGEO": {"-": ["52"]},
    "CIRCVMSIDEO": {"-": ["52"]},
    "SVBTIMEO": {"-": ["52"]},
    "IMMVLGEO": {"-": ["52"]},
    "SVPERIMMINEO": {"-": ["52"]},
    "SCHEDIVM": {"-": ["12"]},
    "INSCRIPTVM": {"-": ["12"]},
    "COMPLACEO": {"-": ["52"]},
    "REDHIBEO": {"-": ["52"]},
    "REMVLCEO": {"-": ["52"]},
    "OBSORBEO": {"-": ["52"]},
    "SEPARATIO": {"-": ["13"]},
    "PERPOTATIO": {"-": ["13"]},
    "PERSALVTATIO": {"-": ["13"]},
    "PARASITATIO": {"-": ["13"]},
    "COLLAVDATIO": {"-": ["13"]},
    "LIBATIO": {"-": ["13"]},
    "DISTVRBATIO": {"-": ["13"]},
    "CONSOCIATIO": {"-": ["13"]},
    "PELLICVLATIO": {"-": ["13"]},
    "CONCERTATIO": {"-": ["13"]},
    "OCCVRSATIO": {"-": ["13"]},
    "OCCATIO": {"-": ["13"]},
    "EGESTIO": {"-": ["13"]},
    "RECEPTIO": {"-": ["13"]},
    "SPECTIO": {"-": ["13"]},
    "OSCILLATIO": {"-": ["13"]},
    "PERSCRIPTIO": {"-": ["13"]},
    "IMMODERATIO": {"-": ["13"]},
    "OBSECRATIO": {"-": ["13"]},
    "ASSECTATIO": {"-": ["13"]},
    "PLACABILITAS": {"-": ["13"]},
    "INSTAVRATIO": {"-": ["13"]},
    "RESTITVTIO": {"-": ["13"]},
    "MVTITIO": {"-": ["13"]},
    "NEGATIO": {"-": ["13"]},
    "BLATIO": {"-": ["13"]},
    "EXANIMATIO": {"-": ["13"]},
    "SINISTERITAS": {"-": ["13"]},
    "VERBIVELITATIO": {"-": ["13"]},
    "COEMPTIO": {"-": ["13"]},
    "CONDEMNATIO": {"-": ["13"]},
    "AEQVABILITAS": {"-": ["13"]},
    "INVECTIO": {"-": ["13"]},
    "DERELICTIO": {"-": ["13"]},
    "EMANCIPATIO": {"-": ["13"]},
    "PERCVRSATIO": {"-": ["13"]},
    "VILICATIO": {"-": ["13"]},
    "IVGATIO": {"-": ["13"]},
    "DIIVNCTIO": {"-": ["13"]},
    "AFFABILITAS": {"-": ["13"]},
    "EMBASICOETAS": {"-": ["13"]},
    "AMANDATIO": {"-": ["13"]},
    "STIMVLATIO": {"-": ["13"]},
    "PVLSATIO": {"-": ["13"]},
    "AMATIO": {"-": ["13"]},
    "LOQVACITAS": {"-": ["13"]},
    "OBSTRVCTIO": {"-": ["13"]},
    "EXPVRGATIO": {"-": ["13"]},
    "EMINATIO": {"-": ["13"]},
    "CONSORTIO": {"-": ["13"]},
    "INFLAMMATIO": {"-": ["13"]},
    "CONVOCATIO": {"-": ["13"]},
    "INVESTIGATIO": {"-": ["13"]},
    "PRAECENTIO": {"-": ["13"]},
    "OBROGATIO": {"-": ["13"]},
    "COLLIGATIO": {"-": ["13"]},
    "ADMIRABILITAS": {"-": ["13"]},
    "DOCILITAS": {"-": ["13"]},
    "PENSATIO": {"-": ["13"]},
    "FRIGVTIO": {"-": ["13"]},
    "PROCACITAS": {"-": ["13"]},
    "SORTIO": {"-": ["13"]},
    "DECLARATIO": {"-": ["13"]},
    "FESTIVITAS": {"-": ["13"]},
    "INTERFATIO": {"-": ["13"]},
    "GVSTATIO": {"-": ["13"]},
    "DEFLAGRATIO": {"-": ["13"]},
    "ANNOTATIO": {"-": ["13"]},
    "COMPELLATIO": {"-": ["13"]},
    "MALEDICTIO": {"-": ["13"]},
    "INCOMMODITAS": {"-": ["13"]},
    "CONGLVTINATIO": {"-": ["13"]},
    "CONSITIO": {"-": ["13"]},
    "EXPLICEO": {"-": ["52"]},
    "DISCRIPTIO": {"-": ["13"]},
    "ADMIXTIO": {"-": ["13"]},
    "COITIO": {"-": ["13"]},
    "GRANDITAS": {"-": ["13"]},
    "RELIGATIO": {"-": ["13"]},
    "NVNTIATIO": {"-": ["13"]},
    "PERACTIO": {"-": ["13"]},
    "IACVLATIO": {"-": ["13"]},
    "DIRITAS": {"-": ["13"]},
    "EFFRENATIO": {"-": ["13"]},
    "APPROPINQVATIO": {"-": ["13"]},
    "CELLATIO": {"-": ["13"]},
    "DESTITVTIO": {"-": ["13"]},
    "OBSCENITAS": {"-": ["13"]},
    "RECVPERATIO": {"-": ["13"]},
    "DICATIO": {"-": ["13"]},
    "INTERPVNCTIO": {"-": ["13"]},
    "INTERCEPTIO": {"-": ["13"]},
    "VIVACITAS": {"-": ["13"]},
    "STIPATIO": {"-": ["13"]},
    "QVAESITIO": {"-": ["13"]},
    "VITATIO": {"-": ["13"]},
    "EMACITAS": {"-": ["13"]},
    "INFIRMATIO": {"-": ["13"]},
    "EXPIATIO": {"-": ["13"]},
    "SVBIGITATIO": {"-": ["13"]},
    "APRICATIO": {"-": ["13"]},
    "CONFICTIO": {"-": ["13"]},
    "IRRIGATIO": {"-": ["13"]},
    "STVPIDITAS": {"-": ["13"]},
    "COMPOTATIO": {"-": ["13"]},
    "SCISCITATIO": {"-": ["13"]},
    "HELVATIO": {"-": ["13"]},
    "GERMANITAS": {"-": ["13"]},
    "ABIECTIO": {"-": ["13"]},
    "PVTATIO": {"-": ["13"]},
    "SEDVCTIO": {"-": ["13"]},
    "CONCENATIO": {"-": ["13"]},
    "INCOMITIO": {"-": ["13"]},
    "PRAEROGATIO": {"-": ["13"]},
    "SERVATIO": {"-": ["13"]},
    "DECERTATIO": {"-": ["13"]},
    "EXSECTIO": {"-": ["13"]},
    "PERAGRATIO": {"-": ["13"]},
    "RESTIPVLATIO": {"-": ["13"]},
    "EDICTIO": {"-": ["13"]},
    "AMPVTATIO": {"-": ["13"]},
    "SCRIPTIO": {"-": ["13"]},
    "SVPPOSITIO": {"-": ["13"]},
    "SATISDATIO": {"-": ["13"]},
    "PROPVLSATIO": {"-": ["13"]},
    "OPIMITAS": {"-": ["13"]},
    "INSVLSITAS": {"-": ["13"]},
    "PROCRASTINATIO": {"-": ["13"]},
    "DISPERTITIO": {"-": ["13"]},
    "DIRIBITIO": {"-": ["13"]},
    "NEGLECTIO": {"-": ["13"]},
    "PROPVGNATIO": {"-": ["13"]},
    "ILLIBERALITAS": {"-": ["13"]},
    "TEMPESTIVITAS": {"-": ["13"]},
    "OBNVNTIO": {"-": ["13"]},
    "CONSVETIO": {"-": ["13"]},
    "EPVLATIO": {"-": ["13"]},
    "IMPVRITAS": {"-": ["13"]},
    "RECLAMATIO": {"-": ["13"]},
    "MVTVATIO": {"-": ["13"]},
    "AEQVATIO": {"-": ["13"]},
    "FORMOSITAS": {"-": ["13"]},
    "VINDICATIO": {"-": ["13"]},
    "GRATIFICATIO": {"-": ["13"]},
    "DIVITATIO": {"-": ["13"]},
    "REPASTINATIO": {"-": ["13"]},
    "DEVEXITAS": {"-": ["13"]},
    "COLLOCVTIO": {"-": ["13"]},
    "MOROSITAS": {"-": ["13"]},
    "DISPERCVTIO": {"-": ["13"]},
    "DISSIPATIO": {"-": ["13"]},
    "INSATIETAS": {"-": ["13"]},
    "IMMATVRITAS": {"-": ["13"]},
    "DETRECTATIO": {"-": ["13"]},
    "COACTIO": {"-": ["13"]},
    "ASSIGNATIO": {"-": ["13"]},
    "PROMVLGATIO": {"-": ["13"]},
    "IMPLICATIO": {"-": ["13"]},
    "ASSERTIO": {"-": ["13"]},
    "INTERDICTIO": {"-": ["13"]},
    "DERIVATIO": {"-": ["13"]},
    "DEBITIO": {"-": ["13"]},
    "OPTATIO": {"-": ["13"]},
    "OBLIQVITAS": {"-": ["13"]},
    "ELOCVTIO": {"-": ["13"]},
    "INSPECTATIO": {"-": ["13"]},
    "ROGITATIO": {"-": ["13"]},
    "CONSVRRECTIO": {"-": ["13"]},
    "CALVIA": {"N": ["11"]},
    "GALLITTA": {"N": ["11"]},
    "IARBITAS": {"N": ["13"]},
    "AMYNTAS": {"N": ["13"]},
    "DAMOETAS": {"N": ["13"]},
    "LYCOTAS": {"N": ["13"]},
    "GORGATAS": {"N": ["13"]},
    "PHILETAS": {"N": ["13"]},
    "HORRATAS": {"N": ["13"]},
    "ARCHYTAS": {"N": ["13"]},
    "EVROTAS": {"N": ["13"]},
    "FAVSTITAS": {"N": ["13"]},
    "PHILOTAS": {"N": ["13"]},
    "AERIFER": {"-": ["12"]},
    "DIFFIDENTER": {"-": ["12"]},
    "MELEAGER": {"N": ["12"]},
    "FATALITER": {"-": ["12"]},
    "INSIPIENTER": {"-": ["12"]},
    "GRANDIFER": {"-": ["12"]},
    "ELOQVENTER": {"-": ["12"]},
    "TVRRIFER": {"-": ["12"]},
    "SCAMANDER": {"N": ["12"]},
    "INCONTINENTER": {"-": ["12"]},
    "SOTER": {"N": ["12"]},
    "IRREVERENTER": {"-": ["12"]},
    "THYMBER": {"N": ["12"]},
    "HISTER": {"N": ["12"]},
    "AFER": {"N": ["12"], "A": ["23"]},
    "RACEMIFER": {"-": ["12"]},
    "CAPITALITER": {"-": ["12"]},
    "AMPHITHEATER": {"-": ["12"]},
    "NEGLEGENTER": {"-": ["12"]},
    "FVMIFER": {"-": ["12"]},
    "CONFIDENTER": {"-": ["12"]},
    "FVRIALITER": {"-": ["12"]},
    "FVRACITER": {"-": ["12"]},
    "INSIGNITER": {"-": ["12"]},
    "PVELLARITER": {"-": ["12"]},
    "LATENTER": {"-": ["12"]},
    "DENTER": {"N": ["12"]},
    "TEVCER": {"N": ["12"], "A": ["23"]},
    "PERSAPIENTER": {"-": ["12"]},
    "CIMBER": {"N": ["12"], "A": ["23"]},
    "SOCIALITER": {"-": ["12"]},
    "ASSIMILITER": {"-": ["12"]},
    "INTELLIGENTER": {"-": ["12"]},
    "DVLCITER": {"-": ["12"]},
    "VILITER": {"-": ["12"]},
    "ELAVER": {"N": ["12"]},
    "LYSANDER": {"N": ["12"]},
    "OLIVIFER": {"-": ["12"]},
    "TANAGER": {"N": ["12"]},
    "MONSTRIFER": {"-": ["12"]},
    "NOTABILITER": {"-": ["12"]},
    "PRODIGIALITER": {"-": ["12"]},
    "TITVBANTER": {"-": ["12"]},
    "TRISTITER": {"-": ["12"]},
    "EXCELLENTER": {"-": ["12"]},
    "ALCANDER": {"N": ["12"]},
    "OBSEQVENTER": {"-": ["12"]},
    "PINIGER": {"-": ["12"]},
    "PERBREVITER": {"-": ["12"]},
    "PERCELER": {"-": ["12"]},
    "ANTONIASTER": {"N": ["12"]},
    "SPHINTER": {"N": ["12"]},
    "AMPLITER": {"-": ["12"]},
    "BLANDITER": {"-": ["12"]},
    "CONGER": {"-": ["12"]},
    "EXSVLTANTER": {"-": ["12"]},
    "ALABASTER": {"-": ["12"]},
    "ASANDER": {"N": ["12"]},
    "SACRIFER": {"-": ["12"]},
    "SPINTHER": {"N": ["12"]},
    "ENSIFER": {"-": ["12"]},
    "DOLENTER": {"-": ["12"]},
    "CREDIBILITER": {"-": ["12"]},
    "CASSANDER": {"N": ["12"]},
    "ANTIPATER": {"N": ["12"]},
    "SERVILITER": {"-": ["12"]},
    "SOMNIFER": {"-": ["12"]},
    "MILITARITER": {"-": ["12"]},
    "MVLCIBER": {"N": ["12"]},
    "CLEANDER": {"N": ["12"]},
    "HERBIFER": {"-": ["12"]},
    "CORYMBIFER": {"-": ["12"]},
    "ARISTANDER": {"N": ["12"]},
    "SCVRRILITER": {"-": ["12"]},
    "IMPERABILITER": {"-": ["12"]},
    "PISANDER": {"N": ["12"]},
    "VNGVLATER": {"-": ["12"]},
    "LEANDER": {"N": ["12"]},
    "FALLACITER": {"-": ["12"]},
    "PACIFER": {"-": ["12"]},
    "PRAESTANTER": {"-": ["12"]},
    "SEMICAPER": {"-": ["12"]},
    "TRVCVLENTER": {"-": ["12"]},
    "DIESPITER": {"N": ["12"]},
    "TVDER": {"N": ["12"]},
    "REDVNDANTER": {"-": ["12"]},
    "IMPARITER": {"-": ["12"]},
    "PHAGER": {"-": ["12"]},
    "CORPORALITER": {"-": ["12"]},
    "MNESTER": {"N": ["12"]},
    "TRIFVRCIFER": {"-": ["12"]},
    "MAEANDER": {"N": ["12"], "-": ["12"]},
    "APPETENTER": {"-": ["12"]},
    "BICORNIGER": {"N": ["12"]},
    "MALIFER": {"-": ["12"]},
    "ARVNDIFER": {"-": ["12"]},
    "MENANDER": {"N": ["12"]},
    "ANTHEMVSIAS": {"N": ["17"]},
    "ARSANIAS": {"N": ["17"]},
    "ARCHIAS": {"N": ["17"]},
    "ADARRHIAS": {"N": ["17"]},
    "ATARRHIAS": {"N": ["17"]},
    "HARPOCRAS": {"N": ["17"]},
    "HABINNAS": {"N": ["17"]},
    "TIRESIAS": {"N": ["17"]},
    "ATTINAS": {"N": ["17"]},
    "PNYTAGORAS": {"N": ["17"]},
    "SCIPIADAS": {"N": ["17"]},
    "MIMAS": {"N": ["17"]},
    "CASPIAS": {"N": ["17"]},
    "ONOMAS": {"N": ["17"]},
    "MIDAS": {"N": ["17"]},
    "DERDAS": {"N": ["17"]},
    "LYCORMAS": {"N": ["17"]},
    "DRYAS": {"N": ["17"]},
    "GNOSIAS": {"N": ["17"], "A": ["25"]},
    "LYSIAS": {"N": ["17"]},
    "PHIDIAS": {"N": ["17"]},
    "ERYMAS": {"N": ["17"]},
    "TROAS": {"N": ["17"], "A": ["25"]},
    "NAIAS": {"N": ["17"]},
    "SYMPLEGAS": {"N": ["17"]},
    "IOLLAS": {"N": ["17"]},
    "LADAS": {"N": ["17"]},
    "CYDAS": {"N": ["17"]},
    "ANAXAGORAS": {"N": ["17"]},
    "SIMIAS": {"N": ["17"]},
    "ABAS": {"N": ["17"]},
    "ARCAS": {"N": ["17"], "A": ["25"]},
    "PYTHIAS": {"N": ["17"]},
    "ATHENAGORAS": {"N": ["17"]},
    "DYMAS": {"N": ["17"]},
    "SCOPAS": {"N": ["17"]},
    "HYLAS": {"N": ["17"]},
    "ACHILLAS": {"N": ["17"]},
    "THESTIAS": {"N": ["17"]},
    "PERDICCAS": {"N": ["17"]},
    "HIERAS": {"N": ["17"]},
    "ATYANAS": {"N": ["17"]},
    "CALLICRATIDAS": {"N": ["17"]},
    "PTERELAS": {"N": ["17"]},
    "THOAS": {"N": ["17"]},
    "PRVSIAS": {"N": ["17"]},
    "GYAS": {"N": ["17"]},
    "CORYBAS": {"N": ["17"]},
    "ASILAS": {"N": ["17"]},
    "EVHYAS": {"N": ["17"]},
    "PAVSANIAS": {"N": ["17"]},
    "MENIDAS": {"N": ["17"]},
    "PERIPHAS": {"N": ["17"]},
    "PHORBAS": {"N": ["17"]},
    "ATHARRIAS": {"N": ["17"]},
    "AVGEAS": {"N": ["17"]},
    "THAVMANTIAS": {"N": ["17"]},
    "PELIAS": {"N": ["17"], "A": ["25"]},
    "PANTAGIAS": {"N": ["17"]},
    "GELAS": {"N": ["17"]},
    "MAENIDAS": {"N": ["17"]},
    "THAMYRAS": {"N": ["17"]},
    "IARBAS": {"N": ["17"]},
    "PHINTIAS": {"N": ["17"]},
    "ORCAS": {"N": ["17"]},
    "ELIAS": {"N": ["17"]},
    "PALLANTIAS": {"N": ["17"]},
    "LICHAS": {"N": ["17"]},
    "PHOEBAS": {"N": ["17"]},
    "PLEIAS": {"N": ["17"]},
    "ACAMAS": {"N": ["17"]},
    "ATHANAGORAS": {"N": ["17"]},
    "BELIAS": {"N": ["17"]},
    "PHTHIAS": {"N": ["17"]},
    "PROTAGORAS": {"N": ["17"]},
    "ALEVAS": {"N": ["17"]},
    "AERIAS": {"N": ["17"]},
    "HYAS": {"N": ["17"]},
    "CINEAS": {"N": ["17"]},
    "CINYRAS": {"N": ["17"]},
    "PELORIAS": {"N": ["17"]},
    "CARRINAS": {"N": ["17"]},
    "TACFARINAS": {"N": ["17"]},
    "ARTAXIAS": {"N": ["17"]},
    "ARPINAS": {"N": ["17"], "A": ["25"]},
    "THOANTIAS": {"N": ["17"]},
    "DIPSAS": {"N": ["17"]},
    "THIODAMAS": {"N": ["17"]},
    "LYCIDAS": {"N": ["17"]},
    "APPIAS": {"N": ["17"]},
    "EVRYDAMAS": {"N": ["17"]},
    "BITIAS": {"N": ["17"]},
    "PASAS": {"N": ["17"]},
    "IOPAS": {"N": ["17"]},
    "LAENAS": {"N": ["17"]},
    "PHASIAS": {"N": ["17"]},
    "BROTEAS": {"N": ["17"]},
    "NOMAS": {"N": ["17"]},
    "MENALCAS": {"N": ["17"]},
    "MARSYAS": {"N": ["17"]},
    "HELLAS": {"N": ["17"]},
    "CORAS": {"N": ["17"]},
    "CLVSINAS": {"N": ["17"]},
    "POEAS": {"N": ["17"]},
    "PLIAS": {"N": ["17"]},
    "PSECAS": {"N": ["17"]},
    "BIAS": {"N": ["17"]},
    "CALCHAS": {"N": ["17"]},
    "TEVTHRAS": {"N": ["17"]},
    "ATHAMAS": {"N": ["17"]},
    "PHLEGYAS": {"N": ["17"]},
    "IPHIAS": {"N": ["17"]},
    "BAGOAS": {"N": ["17"]},
    "TAMIRAS": {"N": ["17"]},
    "THEMILLAS": {"N": ["17"]},
    "PERLIBERALITER": {"-": ["12"]},
    "PRAETERPROPTER": {"-": ["12"]},
    "ARELAS": {"N": ["17"]},
    "IDAS": {"N": ["17"]},
    "FELGINAS": {"N": ["17"]},
    "ATLAS": {"N": ["17"]},
    "THYIAS": {"N": ["17"]},
    "LYCORIAS": {"N": ["17"]},
    "GORGIAS": {"N": ["17"]},
    "OLYMPIAS": {"N": ["17"]},
    "EPAMINONDAS": {"N": ["17"]},
    "HIPPIAS": {"N": ["17"]},
    "ACRAGAS": {"N": ["17"]},
    "XANTHIAS": {"N": ["17"]},
    "BVCEPHALAS": {"N": ["17"]},
    "ILIAS": {"N": ["17"]},
    "HERMAGORAS": {"N": ["17"]},
    "ATHENION": {"N": ["17"]},
    "LYCON": {"N": ["17"]},
    "STIMICHON": {"N": ["17"]},
    "EMATHION": {"N": ["17"]},
    "OROMEDON": {"N": ["17"]},
    "SYNODON": {"-": ["17"]},
    "MEDON": {"N": ["17"]},
    "CALCHEDON": {"N": ["17"]},
    "NICASION": {"N": ["17"]},
    "PORPHYRION": {"N": ["17"]},
    "AVEON": {"N": ["17"]},
    "MARATHON": {"N": ["17"]},
    "LYCAON": {"N": ["17"]},
    "HAEMON": {"N": ["17"]},
    "LEVCON": {"N": ["17"]},
    "AVTOMEDON": {"N": ["17"]},
    "PELAGON": {"N": ["17"]},
    "BITON": {"N": ["17"]},
    "PYGMALION": {"N": ["17"]},
    "CHIRON": {"N": ["17"]},
    "TIMON": {"N": ["17"]},
    "IXION": {"N": ["17"]},
    "SASON": {"N": ["17"]},
    "HYPERION": {"N": ["17"]},
    "TAVRON": {"N": ["17"]},
    "BABYLON": {"N": ["17"]},
    "AGATHON": {"N": ["17"]},
    "CALYDON": {"N": ["17"]},
    "PLEVRON": {"N": ["17"]},
    "XENOPHON": {"N": ["17"]},
    "ALCIMEDON": {"N": ["17"]},
    "OLYMPION": {"N": ["17"]},
    "PALAEMON": {"N": ["17"]},
    "AMAZON": {"N": ["17"]},
    "PHITOLEON": {"N": ["17"]},
    "DEVCALION": {"N": ["17"]},
    "CREON": {"N": ["17"]},
    "MAEON": {"N": ["17"]},
    "ENDYMION": {"N": ["17"]},
    "IASON": {"N": ["17"]},
    "AGAMEMNON": {"N": ["17"]},
    "POTAMON": {"N": ["17"]},
    "POLEMON": {"N": ["17"]},
    "PATRON": {"N": ["17"]},
    "ARISTION": {"N": ["17"]},
    "AMYDON": {"N": ["17"]},
    "PROCOETON": {"-": ["17"]},
    "HECATON": {"N": ["17"]},
    "ARTAMON": {"N": ["17"]},
    "COLOPHON": {"N": ["17"]},
    "EVRYTION": {"N": ["17"]},
    "ELEGIDARION": {"-": ["17"]},
    "SILANION": {"N": ["17"]},
    "HERACLEON": {"N": ["17"]},
    "APION": {"N": ["17"]},
    "ZENON": {"N": ["17"]},
    "GERYON": {"N": ["17"]},
    "THEON": {"N": ["17"]},
    "LEON": {"N": ["17"]},
    "XENON": {"N": ["17"]},
    "DEMOPHOON": {"N": ["17"]},
    "PARMENON": {"N": ["17"]},
    "TELAMON": {"N": ["17"]},
    "MACHAON": {"N": ["17"]},
    "HIPPON": {"N": ["17"]},
    "LYCMON": {"N": ["17"]},
    "HEGION": {"N": ["17"]},
    "BOLON": {"N": ["17"]},
    "CLEON": {"N": ["17"]},
    "LADON": {"N": ["17"]},
    "METRON": {"N": ["17"]},
    "DEMOPHON": {"N": ["17"]},
    "CRANNON": {"N": ["17"]},
    "HAMMON": {"N": ["17"]},
    "HISBON": {"N": ["17"]},
    "PHORMION": {"N": ["17"]},
    "GITON": {"N": ["17"]},
    "PYRRHON": {"N": ["17"]},
    "ALISON": {"N": ["17"]},
    "VRION": {"N": ["17"]},
    "STRYMON": {"N": ["17"]},
    "PHAON": {"N": ["17"]},
    "NICON": {"N": ["17"]},
    "CTESIPHON": {"N": ["17"]},
    "CITHAERON": {"N": ["17"]},
    "THERON": {"N": ["17"]},
    "PHAEDON": {"N": ["17"]},
    "AMMON": {"N": ["17"]},
    "POLYPEMON": {"N": ["17"]},
    "TELON": {"N": ["17"]},
    "DIDYMEON": {"N": ["17"]},
    "ACHERON": {"N": ["17"]},
    "LACON": {"N": ["17"], "A": ["25"]},
    "HIERON": {"N": ["17"]},
    "CONON": {"N": ["17"]},
    "PYTHON": {"N": ["17"]},
    "MATHON": {"N": ["17"]},
    "LYSON": {"N": ["17"]},
    "SIMON": {"N": ["17"]},
    "ANACREON": {"N": ["17"]},
    "MELON": {"N": ["17"]},
    "EVDAEMON": {"N": ["17"]},
    "LYCOPHRON": {"N": ["17"]},
    "AVLON": {"N": ["17"]},
    "DRACON": {"N": ["17"]},
    "LEOMEDON": {"N": ["17"]},
    "DAMASICTHON": {"N": ["17"]},
    "SCIRON": {"N": ["17"]},
    "PLVTON": {"N": ["17"]},
    "DAMON": {"N": ["17"]},
    "NYMPHON": {"N": ["17"]},
    "TARCHON": {"N": ["17"]},
    "BLEPHARON": {"N": ["17"]},
    "VCALEGON": {"N": ["17"]},
    "DIDYMAON": {"N": ["17"]},
    "ACRON": {"N": ["17"]},
    "THRASON": {"N": ["17"]},
    "ECHION": {"N": ["17"]},
    "PHARON": {"N": ["17"]},
    "CROTON": {"N": ["17"]},
    "MENON": {"N": ["17"]},
    "CHAON": {"N": ["17"]},
    "SERAPION": {"N": ["17"]},
    "DOLICHAON": {"N": ["17"]},
    "HERACON": {"N": ["17"]},
    "ANDROGEON": {"N": ["17"]},
    "CAVLON": {"N": ["17"]},
    "GNATHON": {"N": ["17"]},
    "PROCYON": {"N": ["17"]},
    "HEPHAESTION": {"N": ["17"]},
    "EVCTEMON": {"N": ["17"]},
    "PHAETHON": {"N": ["17"]},
    "HIPPOCOON": {"N": ["17"]},
    "AMPHION": {"N": ["17"]},
    "LAOMEDON": {"N": ["17"]},
    "ORION": {"N": ["17"]},
    "OARION": {"N": ["17"]},
    "SOCRATION": {"N": ["17"]},
    "POLYPERCON": {"N": ["17"]},
    "ARISTON": {"N": ["17"]},
    "SOLON": {"N": ["17"]},
    "THAEMON": {"N": ["17"]},
    "IMAON": {"N": ["17"]},
    "STRABON": {"N": ["17"]},
    "NOEMON": {"N": ["17"]},
    "SOTION": {"N": ["17"]},
    "PANDION": {"N": ["17"]},
    "MILANION": {"N": ["17"]},
    "ARION": {"N": ["17"]},
    "AEGON": {"N": ["17"]},
    "ATHLON": {"-": ["17"]},
    "DAPHNON": {"-": ["17"]},
    "ARTEMON": {"N": ["17"]},
    "CORYDON": {"N": ["17"]},
    "HYPERBATON": {"-": ["17"]},
    "AESCHRION": {"N": ["17"]},
    "MICON": {"N": ["17"]},
    "MYRON": {"N": ["17"]},
    "ACTAEON": {"N": ["17"]},
    "MILON": {"N": ["17"]},
    "PARMENION": {"N": ["17"]},
    "SARPEDON": {"N": ["17"]},
    "CHARON": {"N": ["17"]},
    "STRATON": {"N": ["17"]},
    "IDMON": {"N": ["17"]},
    "DION": {"N": ["17"]},
    "PHILON": {"N": ["17"]},
    "MYOPARON": {"-": ["17"]},
    "AMPHITRYON": {"N": ["17"]},
    "PYRACMON": {"N": ["17"]},
    "CYDON": {"N": ["17"]},
    "SINON": {"N": ["17"]},
    "THERMODON": {"N": ["17"]},
    "SICYON": {"N": ["17"]},
    "SVNION": {"N": ["17"]},
    "HYPOCAVSTON": {"-": ["17"]},
    "CIMON": {"N": ["17"]},
    "CALLIPHON": {"N": ["17"]},
    "SATYRION": {"-": ["17"]},
    "AETHON": {"N": ["17"]},
    "ACMON": {"N": ["17"]},
    "THESPRION": {"N": ["17"]},
    "MEMNON": {"N": ["17"]},
    "PLATON": {"N": ["17"]},
    "THEMISON": {"N": ["17"]},
    "EETION": {"N": ["17"]},
    "ALIACMON": {"N": ["17"]},
    "EVCLION": {"N": ["17"]},
    "PELION": {"N": ["17"]},
    "ALCON": {"N": ["17"]},
    "TETRASTYLON": {"-": ["17"]},
    "STILBON": {"N": ["17"]},
    "GLYCON": {"N": ["17"]},
    "DOLON": {"-": ["13"], "N": ["21"]},
    "TRICHON": {"N": ["17"]},
    "PRASINVS": {"-": ["12"]},
    "PRISCE": {"-": ["6-"]},
    "TRICENVS": {"-": ["21"]},
    "AVSVS": {"-": ["21"]},
    "SEPTIENS": {"-": ["6-"]},
    "SVLMO": {"N": ["13"]},
    "ADBITO": {"-": ["53"]},
    "MALITIOSVS": {"-": ["21"]},
    "PHAECASIA": {"-": ["11"]},
    "HECTOR": {"N": ["13"]},
    "FVSTITVDINVS": {"-": ["21"]},
    "PETAVRISTARIVS": {"-": ["12"]},
    "PROMISSOR": {"-": ["13"]},
    "ENTHYMEMA": {"-": ["13"]},
    "HAESITABVNDVS": {"-": ["21"]},
    "CANTAMEN": {"-": ["13"]},
    "DISPOLIO": {"-": ["51"]},
    "CONGREGABILIS": {"-": ["24"]},
    "EVRYNOME": {"N": ["11"]},
    "LYDE": {"N": ["11"]},
    "VLTIMO": {"2": ["51"]},
    "EXAVSPICO": {"-": ["51"]},
    "HALOPHANTA": {"-": ["11"]},
    "DELIBERATOR": {"-": ["13"]},
    "ARVNDINEVS": {"-": ["21"]},
    "PHYLACA": {"-": ["11"]},
    "LECTICVLA": {"-": ["11"]},
    "SVBIGITO": {"-": ["51"]},
    "INTESTINA": {"-": ["12"]},
    "PERFIDIOSE": {"-": ["6-"]},
    "CATEIA": {"-": ["11"]},
    "STEGA": {"-": ["11"]},
    "CAVTOR": {"-": ["13"]},
    "IMPASTVS": {"-": ["21"]},
    "SORORIVS": {"-": ["12"]},
    "SPATALOCINAEDVS": {"-": ["12"]},
    "POSTILIO": {"-": ["13"]},
    "FESTINVS": {"-": ["21"]},
    "INDIGNATIVNCVLA": {"-": ["11"]},
    "PRAETERBITO": {"-": ["53"]},
    "INANIAE": {"-": ["11"]},
    "AETATVLA": {"-": ["11"]},
    "PRODIGIALIS": {"-": ["24"]},
    "INFORO": {"-": ["51"]},
    "CROTALISTRIA": {"-": ["11"]},
    "STIGMOSVS": {"-": ["21"]},
    "TONSORIVS": {"-": ["21"]},
    "SVPERVOLITO": {"-": ["51"]},
    "LAMELLVLA": {"-": ["11"]},
    "RIDEOR": {"-": ["5K"]},
    "CORNEOLVS": {"-": ["21"]},
    "AEQVIPARABILIS": {"-": ["24"]},
    "CAVSIFICOR": {"-": ["13"]},
    "DELIQVVS": {"-": ["21"]},
    "DEPVLSOR": {"-": ["13"]},
    "PASTORICIVS": {"-": ["21"]},
    "OPTEMPERO": {"-": ["51"]},
    "CIRCVMSPECTATRIX": {"-": ["13"]},
    "CRITICA": {"-": ["12"]},
    "LAVDICENVS": {"-": ["12"]},
    "LATERNARIVS": {"-": ["12"]},
    "OBSCAEVO": {"-": ["51"]},
    "ELEVSIN": {"N": ["13"]},
    "SISAPO": {"N": ["13"]},
    "NEREIS": {"N": ["13"]},
    "SCYLAX": {"N": ["13"]},
    "ROGATIVNCVLA": {"-": ["11"]},
    "BELLIPOTENS": {"-": ["22"]},
    "MERCATORIVS": {"-": ["21"]},
    "LATERENSIS": {"N": ["13"]},
    "DVELLATOR": {"-": ["13"]},
    "REPVGNANTIA": {"-": ["13"]},
    "COPIS": {"-": ["24"]},
    "HAEDILLVS": {"-": ["12"]},
    "MISCIX": {"-": ["22"]},
    "PASTORALIS": {"-": ["24"]},
    "LVCRETILIS": {"N": ["13"]},
    "CENATORIA": {"-": ["12"]},
    "PROPINQVE": {"-": ["6-"]},
    "SVPPINGO": {"-": ["53"]},
    "POLYDAMAS": {"N": ["13"]},
    "NEVTIQVAM": {"-": ["6-"]},
    "VASVS": {"-": ["12"]},
    "INTERNECO": {"-": ["51"]},
    "AVTOPYROS": {"-": ["17"]},
    "QVATERNVS": {"-": ["21"]},
    "ARIETINVS": {"-": ["21"]},
    "LENONIVS": {"-": ["21"]},
    "MORIGERO": {"-": ["51"]},
    "EMINOR": {"-": ["5J"]},
    "TAPES": {"-": ["13"]},
    "COMPITALICIVS": {"-": ["21"]},
    "SOLARIVM": {"2": ["12"]},
    "BALISCVS": {"-": ["12"]},
    "VIVE": {"-": ["6-"]},
    "INDVSIATVS": {"-": ["21"]},
    "PALMVLVS": {"-": ["12"]},
    "IMPERATRIX": {"-": ["13"]},
    "SEPTIFLVVS": {"-": ["21"]},
    "ABITO": {"-": ["53"]},
    "INSECTO": {"-": ["51"]},
    "FIDICINA": {"-": ["11"]},
    "PEREGRI": {"-": ["6-"]},
    "LIMAX": {"-": ["13"]},
    "CROCOTVLA": {"-": ["11"]},
    "LINGVLACA": {"-": ["11"]},
    "CENTIENS": {"-": ["6-"]},
    "CISTVLA": {"-": ["11"]},
    "PACTA": {"-": ["11"]},
    "LIBIDINOR": {"-": ["5J"]},
    "STRIGA": {"2": ["11"]},
    "AMASIS": {"N": ["13"]},
    "DAPHNIS": {"N": ["13"]},
    "LVSITO": {"-": ["51"]},
    "LIMONE": {"N": ["11"]},
    "DEVIRGINO": {"-": ["51"]},
    "RVSTICOR": {"-": ["5J"]},
    "INSPVTO": {"-": ["51"]},
    "CANTVRIO": {"-": ["54"]},
    "COQVINO": {"-": ["51"]},
    "APPINGO": {"-": ["53"]},
    "SALAMIS": {"N": ["13"]},
    "SENVS": {"-": ["21"]},
    "PVGNATOR": {"-": ["13"]},
    "ARIETILLVS": {"-": ["21"]},
    "DVODECIENS": {"-": ["6-"]},
    "PARENTALIA": {"-": ["13"]},
    "TRITON": {"N": ["13"]},
    "PROPONTIS": {"N": ["13"]},
    "ATYS": {"N": ["13"]},
    "FALSE": {"-": ["6-"]},
    "PERFVNCTORIE": {"-": ["6-"]},
    "INSTRVCTOR": {"-": ["13"]},
    "SIMITV": {"-": ["6-"]},
    "ECLOGA": {"-": ["11"]},
    "PARTHENOPE": {"N": ["11"]},
    "VESERIS": {"N": ["13"]},
    "FLAMMARIVS": {"-": ["12"]},
    "ARBITRARIVS": {"-": ["21"]},
    "PANNVCIA": {"-": ["11"]},
    "INFIRME": {"-": ["6-"]},
    "CONTVMVLO": {"-": ["51"]},
    "DIFFLAGITO": {"-": ["51"]},
    "MYTILENE": {"N": ["11"]},
    "PATINARIVS": {"-": ["21"]},
    "CASTIGATORIVS": {"-": ["21"]},
    "CELAENO": {"N": ["13"]},
    "ZEPHYRITIS": {"N": ["13"]},
    "OENONE": {"N": ["11"]},
    "SCVRROR": {"-": ["5J"]},
    "CONIVGO": {"-": ["51"]},
    "STERNAX": {"-": ["22"]},
    "CAPO": {"-": ["13"]},
    "HIPPVRVS": {"-": ["12"]},
    "HOSTIO": {"1": ["54"]},
    "INSIPIENTIA": {"-": ["11"]},
    "ORACVLARIVS": {"-": ["21"]},
    "FVNDATOR": {"1": ["13"]},
    "SCALPVRRIO": {"-": ["54"]},
    "IRREVERENS": {"-": ["22"]},
    "INDIPISCO": {"-": ["53"]},
    "SITIENS": {"-": ["2&"]},
    "CONGRAECO": {"-": ["51"]},
    "DIARIA": {"-": ["12"]},
    "ASSESSVS": {"-": ["14"]},
    "MELILOTOS": {"-": ["17"]},
    "COACESCO": {"-": ["53"]},
    "AVELLANA": {"-": ["11"]},
    "EXPETESSO": {"-": ["53"]},
    "OBSVO": {"-": ["53"]},
    "COTYS": {"N": ["13"]},
    "MEROBIBVS": {"-": ["21"]},
    "BEATITVDO": {"-": ["13"]},
    "INTERSCRIBO": {"-": ["53"]},
    "CONSVLTE": {"-": ["6-"]},
    "GAVSAPE": {"-": ["13"]},
    "HEREDIPETA": {"-": ["11"]},
    "PERNECESSARIVS": {"-": ["21"]},
    "HECALE": {"N": ["11"]},
    "BVLEVTA": {"-": ["11"]},
    "DEAMO": {"-": ["51"]},
    "DESOMNIS": {"-": ["24"]},
    "HYDRAVLES": {"-": ["11"]},
    "VIATICVS": {"-": ["12"]},
    "ACLYS": {"-": ["13"]},
    "VITIOSE": {"-": ["6-"]},
    "SEDECIES": {"-": ["6-"]},
    "EXSVSCITO": {"1": ["51"]},
    "CILO": {"N": ["13"]},
    "MVLTIPOTENS": {"-": ["22"]},
    "NOCTIVIGILVS": {"-": ["21"]},
    "ZOTHECA": {"-": ["11"]},
    "QVINDECIENS": {"-": ["6-"]},
    "PRAEDVLCIS": {"-": ["24"]},
    "TARRACO": {"N": ["13"]},
    "COLLABEFACTO": {"-": ["51"]},
    "COCIO": {"-": ["13"]},
    "INSALVTATVS": {"-": ["21"]},
    "DEFAECO": {"-": ["51"]},
    "TITAN": {"N": ["13"]},
    "ALARIS": {"-": ["24"]},
    "DEONERO": {"-": ["51"]},
    "LAGVNCVLA": {"-": ["11"]},
    "PERSVASVS": {"-": ["14"]},
    "BAIVLO": {"-": ["51"]},
    "ERADICO": {"-": ["51"]},
    "APOCVLO": {"-": ["51"]},
    "INTERCEPTOR": {"-": ["13"]},
    "SERMONOR": {"-": ["5J"]},
    "AQVILEX": {"-": ["13"]},
    "COMVLA": {"-": ["11"]},
    "SVCOSVS": {"-": ["21"]},
    "PERFLAGITIOSVS": {"-": ["21"]},
    "ORESTIS": {"N": ["13"]},
    "DIGNITOSVS": {"-": ["21"]},
    "SVPEREGREDIOR": {"-": ["5N"]},
    "EXPVLSIO": {"-": ["13"]},
    "PERPLEXOR": {"-": ["5J"]},
    "SYBARIS": {"N": ["13"]},
    "OPPESSVLATVS": {"-": ["21"]},
    "DISSENSVS": {"-": ["14"]},
    "REDIVIA": {"-": ["11"]},
    "BYLLIS": {"N": ["13"]},
    "CENATICVS": {"-": ["21"]},
    "CATAPIRATERIA": {"-": ["11"]},
    "HELLE": {"N": ["11"]},
    "AGRICVLTVRA": {"-": ["11"]},
    "INDELECTATVS": {"-": ["21"]},
    "CONSAEPIO": {"-": ["54"]},
    "TRACHIN": {"N": ["13"]},
    "NOVACVLA": {"-": ["11"]},
    "CONTRVNCO": {"-": ["51"]},
    "INDEPLORATVS": {"-": ["21"]},
    "INCONTINENTIA": {"-": ["11"]},
    "CARINVS": {"-": ["21"]},
    "DAMMA": {"-": ["11"]},
    "CIRCVMVISO": {"-": ["53"]},
    "OTHRYS": {"N": ["13"]},
    "SENTENTIOLA": {"-": ["11"]},
    "OGGANNIO": {"-": ["54"]},
    "INEXSTINCTVS": {"-": ["21"]},
    "MANNVLVS": {"-": ["12"]},
    "HIPPODROMVS": {"-": ["12"]},
    "HERMIONE": {"N": ["11"]},
    "BVRDVBASTA": {"-": ["11"]},
    "PARALYTICVS": {"-": ["21"]},
    "CONFLICTOR": {"2": ["5J"]},
    "DEOSCVLOR": {"-": ["5J"]},
    "TOLVTIM": {"-": ["6-"]},
    "MALEDICE": {"-": ["6-"]},
    "PLECTILIS": {"-": ["24"]},
    "VIOLARIVS": {"-": ["12"]},
    "REVIRESCO": {"-": ["53"]},
    "EVISCERO": {"-": ["51"]},
    "EBVRATVS": {"-": ["21"]},
    "FALSIMONIA": {"-": ["11"]},
    "PERISTROMA": {"-": ["13"]},
    "MODERO": {"-": ["51"]},
    "IMPERTERRITVS": {"-": ["21"]},
    "SEMIGRO": {"-": ["51"]},
    "CHORAVLA": {"-": ["11"]},
    "VIDVLVS": {"-": ["12"]},
    "LVSIO": {"-": ["13"]},
    "EXTERSVS": {"-": ["21"]},
    "FABARIS": {"N": ["13"]},
    "POLYMITVS": {"-": ["21"]},
    "SERMVNCVLVS": {"-": ["12"]},
    "OPINO": {"-": ["51"]},
    "DELACERO": {"-": ["51"]},
    "NATALICIA": {"-": ["11"]},
    "CERASINVS": {"-": ["21"]},
    "EXPRESSE": {"-": ["6-"]},
    "ALMO": {"N": ["13"]},
    "AMBESTRIX": {"-": ["13"]},
    "QVINQVEVIRATVS": {"-": ["14"]},
    "EPVLARIS": {"-": ["24"]},
    "DEVERSITOR": {"-": ["13"]},
    "SEXCENTIENS": {"-": ["6-"]},
    "SCROFIPASCVS": {"-": ["21"]},
    "INCERTO": {"1": ["6-"]},
    "ANXVR": {"N": ["13"]},
    "SYCOPHANTIA": {"-": ["11"]},
    "FVLIGINEVS": {"-": ["21"]},
    "POETICE": {"2": ["17"], "-": ["12"], "1": ["12"]},
    "PROMPTO": {"-": ["51"]},
    "PRISTIS": {"-": ["13"], "N": ["13"]},
    "BINOMINIS": {"-": ["24"]},
    "QVISQVILIA": {"-": ["12"]},
    "SVPRASCRIBO": {"-": ["53"]},
    "COMMENDATVS": {"-": ["21"]},
    "SPERABILIS": {"-": ["24"]},
    "BVCINVS": {"-": ["12"]},
    "PRANSVS": {"2": ["14"]},
    "TRANSFVSIO": {"-": ["13"]},
    "PERELEGANS": {"-": ["22"]},
    "VANITVDO": {"-": ["13"]},
    "PRAESERVIO": {"-": ["54"]},
    "HORDEIA": {"-": ["11"]},
    "SVBAVSCVLTO": {"-": ["51"]},
    "THYONE": {"N": ["11"]},
    "EDICTO": {"-": ["51"]},
    "GORGE": {"N": ["11"]},
    "DEMANDO": {"2": ["51"]},
    "ERATO": {"N": ["13"]},
    "INVIOLATE": {"-": ["6-"]},
    "DIVEXO": {"-": ["51"]},
    "VENERO": {"-": ["51"]},
    "EXSONO": {"1": ["51"]},
    "CORONIS": {"N": ["13"]},
    "IMPVRE": {"-": ["6-"]},
    "PVGILICE": {"-": ["6-"]},
    "MANTISCINOR": {"-": ["5J"]},
    "HELOPS": {"-": ["13"]},
    "MALEDICTOR": {"-": ["13"]},
    "SVCRO": {"N": ["13"]},
    "CVRSVRA": {"-": ["11"]},
    "MANTO": {"N": ["13"]},
    "EXHERES": {"-": ["13"]},
    "PERTRACTATE": {"-": ["6-"]},
    "SALTATRIX": {"-": ["13"]},
    "BARBVLA": {"-": ["11"]},
    "AFFIRMATE": {"-": ["6-"]},
    "AMASIVNCVLVS": {"-": ["12"]},
    "RESTITO": {"-": ["51"]},
    "IRRELIGIOSVS": {"-": ["21"]},
    "ARMAMENTA": {"-": ["12"]},
    "PERCOPIOSVS": {"-": ["21"]},
    "EXHORTO": {"-": ["51"]},
    "MATVLA": {"-": ["11"]},
    "CALCITROSVS": {"-": ["21"]},
    "NERVIA": {"-": ["12"]},
    "RECENSIO": {"-": ["13"]},
    "EXOSTRA": {"-": ["11"]},
    "AMPLEXO": {"-": ["51"]},
    "PERLIBET": {"-": ["52"]},
    "TRAPEZITA": {"-": ["11"]},
    "AVLOEDVS": {"-": ["12"]},
    "SAEPICVLE": {"-": ["6-"]},
    "CHRYSOPHRYS": {"-": ["17"]},
    "INCLVSIO": {"-": ["13"]},
    "INSTIMVLO": {"-": ["51"]},
    "SPONDEVS": {"-": ["21"]},
    "ACTIVNCVLA": {"-": ["11"]},
    "OPERARIA": {"-": ["11"]},
    "STENTOR": {"N": ["13"]},
    "ABSTINAX": {"-": ["22"]},
    "OBSEQVELA": {"-": ["11"]},
    "OSOR": {"-": ["13"]},
    "CARYOTA": {"-": ["11"]},
    "POLLINCTOR": {"-": ["13"]},
    "MEDVLLVLA": {"-": ["11"]},
    "DISTAEDET": {"-": ["52"]},
    "IAMIAMQVE": {"-": ["6-"]},
    "PERSAPIENS": {"-": ["22"]},
    "DEBLATERO": {"-": ["51"]},
    "CENSIO": {"-": ["13"]},
    "PYRENE": {"N": ["11"]},
    "VENTRIOSVS": {"-": ["21"]},
    "AMOENE": {"-": ["6-"]},
    "LEVATOR": {"-": ["13"]},
    "HISPALIS": {"N": ["13"]},
    "TARPEZITA": {"-": ["11"]},
    "HERBEVS": {"-": ["21"]},
    "PERSPICIENTIA": {"-": ["11"]},
    "CVBITORIVS": {"-": ["21"]},
    "PERPLEXE": {"-": ["6-"]},
    "REDAMBVLO": {"-": ["51"]},
    "NVTRICO": {"-": ["51"]},
    "PANOPE": {"N": ["11"]},
    "REPOSTOR": {"-": ["13"]},
    "CRINALIS": {"-": ["24"]},
    "MANIPVLATIM": {"-": ["6-"]},
    "ACCIPITRINA": {"-": ["11"]},
    "COLLARE": {"-": ["13"]},
    "ANAGNOSTES": {"-": ["17"]},
    "BVXEVS": {"-": ["21"]},
    "OCTOGIENS": {"-": ["6-"]},
    "ARISTO": {"N": ["13"]},
    "BARBATORIA": {"-": ["11"]},
    "GEMELLIPARVS": {"-": ["21"]},
    "FAVSTE": {"-": ["6-"]},
    "QVANTILLVS": {"-": ["21"]},
    "NVMVLARIOLVS": {"-": ["12"]},
    "PATAGIATVS": {"-": ["21"]},
    "MORTIFERE": {"-": ["6-"]},
    "HELIOPOLIS": {"N": ["13"]},
    "CANTHARIS": {"-": ["13"]},
    "ATHLETICE": {"-": ["6-"]},
    "INDVSIARIVS": {"-": ["12"]},
    "CIRRATVS": {"-": ["21"]},
    "MARO": {"N": ["13"]},
    "SVPPLOSIO": {"-": ["13"]},
    "ISTACTENVS": {"-": ["6-"]},
    "CORPVLENTVS": {"-": ["21"]},
    "PERVTILIS": {"-": ["24"]},
    "NEMESIS": {"N": ["13"]},
    "NAR": {"N": ["13"]},
    "LANISTICIVS": {"-": ["21"]},
    "INDVCTOR": {"-": ["13"]},
    "IMPLVVIATVS": {"-": ["21"]},
    "ARABS": {"N": ["13"], "A": ["21"]},
    "PENELOPE": {"N": ["11"]},
    "DECEMPEDATOR": {"-": ["13"]},
    "LVNVLA": {"-": ["11"]},
    "CIRCVMSCRIPTVS": {"-": ["21"]},
    "SEMICREMATVS": {"-": ["21"]},
    "ALCIS": {"N": ["13"]},
    "TONSTRINVS": {"-": ["21"]},
    "SOBRIE": {"-": ["6-"]},
    "NIVATVS": {"-": ["21"]},
    "CASTERIA": {"-": ["11"]},
    "ADONIS": {"N": ["13"]},
    "PHALERATVS": {"-": ["21"]},
    "RETROVERSVS": {"-": ["21"]},
    "ALLACRIMO": {"-": ["51"]},
    "COENO": {"-": ["51"]},
    "PERSVASTRIX": {"-": ["13"]},
    "VIRGATOR": {"-": ["13"]},
    "BLENNVS": {"-": ["21"]},
    "DVCENVS": {"-": ["21"]},
    "PORCINA": {"-": ["11"]},
    "GOBIVS": {"-": ["12"]},
    "BABAECALVS": {"-": ["21"]},
    "PATAGIARIVS": {"-": ["12"]},
    "CALVOR": {"-": ["5L"]},
    "PROTOPRAXIA": {"-": ["11"]},
    "APPOTVS": {"-": ["21"]},
    "INGRATIFICVS": {"-": ["21"]},
    "PERINVISVS": {"-": ["21"]},
    "INEXCITVS": {"-": ["21"]},
    "CINCTVTVS": {"-": ["21"]},
    "RAPACIDA": {"-": ["11"]},
    "MELANVRVS": {"-": ["12"]},
    "TRITICEIVS": {"-": ["21"]},
    "SCOPVLOSVS": {"-": ["21"]},
    "SIRPICVLVS": {"-": ["21"]},
    "SOPOLIS": {"N": ["13"]},
    "AMYNTOR": {"N": ["13"]},
    "PRAEGRESSVS": {"-": ["21"]},
    "LAPIDATOR": {"-": ["13"]},
    "PERSONVS": {"-": ["21"]},
    "INTERCINO": {"-": ["53"]},
    "INAVGVRO": {"-": ["51"]},
    "LATRATOR": {"-": ["13"]},
    "PEDALE": {"-": ["13"]},
    "TONSITO": {"-": ["51"]},
    "CAPSELLA": {"-": ["11"]},
    "TVBVRCINABVNDVS": {"-": ["21"]},
    "CRATHIS": {"N": ["13"]},
    "SVBRANCIDVS": {"-": ["21"]},
    "VOLVPTVOSVS": {"-": ["21"]},
    "COMPLVRIENS": {"-": ["6-"]},
    "LATRINA": {"-": ["11"]},
    "SEXIENS": {"-": ["6-"]},
    "FOETOR": {"-": ["13"]},
    "DODRANTARIVS": {"-": ["21"]},
    "HYPASIS": {"N": ["13"]},
    "ISIS": {"N": ["13"]},
    "PETOSIRIS": {"N": ["13"]},
    "APPOSCO": {"-": ["53"]},
    "XIPHIAS": {"-": ["17"]},
    "CVRATVRA": {"-": ["11"]},
    "SVBINVISVS": {"-": ["21"]},
    "LABECVLA": {"-": ["11"]},
    "SIRMIO": {"N": ["13"]},
    "NERVOSE": {"-": ["6-"]},
    "SOLIDE": {"-": ["6-"]},
    "PACIFICVS": {"-": ["21"]},
    "MISSICVLO": {"-": ["51"]},
    "MENSVLA": {"-": ["11"]},
    "ELPENOR": {"N": ["13"]},
    "PARENTICIDA": {"-": ["11"]},
    "LATVRA": {"-": ["11"]},
    "POTITO": {"-": ["51"]},
    "PELOPS": {"N": ["13"]},
    "DESTIMVLO": {"-": ["51"]},
    "NARES": {"-": ["13"]},
    "RISOR": {"-": ["13"]},
    "SITELLA": {"-": ["11"]},
    "ASTITVO": {"-": ["12"]},
    "MVSSITO": {"-": ["51"]},
    "POLITE": {"-": ["6-"]},
    "IATRALIPTA": {"-": ["11"]},
    "COLLATIVVS": {"-": ["21"]},
    "SIREN": {"N": ["13"]},
    "CLIO": {"N": ["13"]},
    "OMPHALE": {"N": ["11"]},
    "PATRICE": {"-": ["6-"]},
    "NITEDVLA": {"-": ["11"]},
    "CONNVBIALIS": {"-": ["24"]},
    "NESIS": {"N": ["13"]},
    "QVERIBVNDVS": {"-": ["21"]},
    "CONCVRO": {"-": ["51"]},
    "SEVIRATVS": {"-": ["14"]},
    "SINGVLARIVS": {"-": ["21"]},
    "ECHIDNA": {"-": ["11"]},
    "OBRODO": {"-": ["53"]},
    "AVSTERE": {"-": ["6-"]},
    "APOPHORETVS": {"-": ["21"]},
    "PERSEQVENS": {"-": ["22"]},
    "PIPERATVS": {"-": ["21"]},
    "LITTERATE": {"-": ["6-"]},
    "SVTVRA": {"-": ["11"]},
    "AHA": {"-": ["11"]},
    "CARCERARIVS": {"2": ["12"]},
    "AVLIS": {"N": ["13"]},
    "CYMATILIS": {"-": ["24"]},
    "HERO": {"N": ["13"]},
    "SVBREMIGO": {"-": ["51"]},
    "SALTVARIVS": {"-": ["12"]},
    "TERDECIENS": {"-": ["13"]},
    "COMMVRMVRO": {"-": ["51"]},
    "PERTVRPIS": {"-": ["24"]},
    "VBIVBI": {"-": ["6-"]},
    "PERSOLLA": {"-": ["11"]},
    "GRACILIPES": {"-": ["22"]},
    "SOMNICVLOSVS": {"-": ["21"]},
    "CISTOPHORVS": {"-": ["12"]},
    "FERRICREPINVS": {"-": ["21"]},
    "ADIACENS": {"-": ["13"]},
    "CANDIDE": {"-": ["6-"]},
    "QVINQVIENS": {"-": ["6-"]},
    "AGRICVLTOR": {"-": ["13"]},
    "FIDELE": {"-": ["6-"]},
    "DENVS": {"-": ["21"]},
    "SOBRINVS": {"-": ["12"]},
    "LOQVITOR": {"-": ["5J"]},
    "ECHENEIS": {"-": ["13"]},
    "LVDIFICABILIS": {"-": ["24"]},
    "COCTOR": {"-": ["13"]},
    "MESSORIVS": {"-": ["21"]},
    "LATICLAVIVS": {"2": ["21"]},
    "HERMEROS": {"N": ["12"]},
    "CYME": {"N": ["11"]},
    "AGGEMO": {"-": ["13"]},
    "INVENVSTE": {"-": ["6-"]},
    "PERREPTO": {"-": ["51"]},
    "NVMELLA": {"-": ["11"]},
    "HISTORICE": {"-": ["17"]},
    "MVGIL": {"-": ["13"]},
    "PRAENVNTIA": {"-": ["11"]},
    "THAIS": {"N": ["13"]},
    "OBVIGILO": {"-": ["51"]},
    "TERNVS": {"-": ["21"]},
    "COTISO": {"N": ["13"]},
    "NINIVE": {"N": ["11"]},
    "CAPREIDA": {"-": ["11"]},
    "PERCITVS": {"-": ["21"]},
    "BILYCHNIS": {"-": ["24"]},
    "INCOMMODO": {"1": ["51"]},
    "PRAEFIDENS": {"-": ["22"]},
    "MONSTRABILIS": {"-": ["24"]},
    "PERLITTERATVS": {"-": ["21"]},
    "INSTINCTOR": {"-": ["13"]},
    "SERVATVS": {"-": ["21"]},
    "TONSTRINA": {"-": ["11"]},
    "CALCEOLARIVS": {"-": ["12"]},
    "CHIONE": {"N": ["11"]},
    "SPONSIVNCVLA": {"-": ["11"]},
    "QVOQVOMODO": {"1": ["6-"], "2": ["21"]},
    "QVADRAGENVS": {"-": ["21"]},
    "ALTARE": {"-": ["13"]},
    "OCCVLTATOR": {"-": ["13"]},
    "SEPTEMPEDALIS": {"-": ["24"]},
    "QVADRIMVLVS": {"-": ["21"]},
    "PERBACCHOR": {"-": ["5J"]},
    "BIMARITVS": {"-": ["21"]},
    "COMISSATOR": {"-": ["13"]},
    "BIFRONS": {"-": ["22"]},
    "SPECVLABVNDVS": {"-": ["21"]},
    "MVRICIDVS": {"-": ["21"]},
    "CAPERO": {"-": ["51"]},
    "COMPACISCOR": {"-": ["5L"]},
    "PERLEPIDE": {"-": ["6-"]},
    "EROGITO": {"-": ["51"]},
    "ARCTOPHYLAX": {"N": ["13"]},
    "DELECTOR": {"-": ["5J"]},
    "SVFFIMEN": {"-": ["13"]},
    "DORMITORIVS": {"-": ["21"]},
    "PRECATOR": {"-": ["13"]},
    "XANTHIPPE": {"N": ["11"]},
    "STRICTVRA": {"-": ["11"]},
    "CALLISTO": {"N": ["13"]},
    "PERIVCVNDVS": {"-": ["21"]},
    "SCAZON": {"-": ["13"]},
    "ANACHARSIS": {"N": ["13"]},
    "VMBER": {"A": ["23"]},
    "ACACIA": {"-": ["11"]},
    "FORDA": {"-": ["11"]},
    "CONDVCTICIVS": {"-": ["21"]},
    "PORCELLVS": {"-": ["12"]},
    "INEXPLETVS": {"-": ["21"]},
    "GESTITO": {"-": ["51"]},
    "INFECTO": {"-": ["51"]},
    "DESQVAMO": {"-": ["51"]},
    "VRBANE": {"-": ["6-"]},
    "PSAMATHE": {"N": ["11"]},
    "PERSEPHONE": {"N": ["11"]},
    "INTERTEXO": {"-": ["53"]},
    "MASTRVCA": {"-": ["11"]},
    "EPILOGVS": {"-": ["12"]},
    "SALIATVS": {"-": ["14"]},
    "BARDVS": {"2": ["21"]},
    "LVCTAMEN": {"-": ["13"]},
    "MAIALIS": {"-": ["13"]},
    "CATAPHRACTA": {"-": ["11"]},
    "COMMANIPVLARIS": {"-": ["24"]},
    "AGAVE": {"N": ["11"]},
    "TRIPOLIS": {"N": ["13"]},
    "CAPIS": {"-": ["13"]},
    "HAESITATOR": {"-": ["13"]},
    "FVCOSVS": {"-": ["21"]},
    "ECCERE": {"-": ["6-"]},
    "BVLE": {"-": ["17"]},
    "GRAVATE": {"-": ["6-"]},
    "REPROMITTO": {"-": ["53"]},
    "NEPHELE": {"N": ["11"]},
    "FATVS": {"-": ["21"]},
    "SOMNICVLOSE": {"-": ["6-"]},
    "INCONCILIO": {"-": ["51"]},
    "PERMIRVS": {"-": ["21"]},
    "PAVONINVS": {"-": ["21"]},
    "SPOLIATRIX": {"-": ["13"]},
    "VIRILIA": {"-": ["13"]},
    "IAPYX": {"A": ["22"], "N": ["21"]},
    "OMINOSVS": {"-": ["21"]},
    "GALBINVS": {"-": ["21"]},
    "INEQVITABILIS": {"-": ["24"]},
    "SERISAPIA": {"-": ["11"]},
    "AVGE": {"N": ["11"]},
    "OBSIGNATOR": {"-": ["13"]},
    "EXPISCOR": {"-": ["5J"]},
    "COEMPTIONALIS": {"-": ["24"]},
    "FALSIDICVS": {"-": ["21"]},
    "REPRESSOR": {"-": ["13"]},
    "SAME": {"N": ["11"]},
    "CICARO": {"-": ["13"]},
    "LANIENA": {"-": ["11"]},
    "MVLIONIVS": {"-": ["21"]},
    "FIMBRIAE": {"-": ["11"]},
    "GRAVASTELLVS": {"-": ["21"]},
    "ANAS": {"1": ["13"]},
    "PRAESTIGIATRIX": {"-": ["13"]},
    "CONFOSSVS": {"-": ["21"]},
    "RALLVS": {"-": ["21"]},
    "ATTHIS": {"N": ["13"]},
    "EXPVLSOR": {"-": ["13"]},
    "SCALDIS": {"N": ["13"]},
    "COMINCOMMODVS": {"-": ["21"]},
    "CHARONDAS": {"N": ["13"]},
    "HALYS": {"N": ["13"]},
    "DELITIGO": {"-": ["51"]},
    "REPROMISSIO": {"-": ["13"]},
    "NICEROS": {"N": ["12"]},
    "INIMICE": {"-": ["6-"]},
    "CATVLA": {"-": ["11"]},
    "CHLORIS": {"N": ["13"]},
    "VOLVPE": {"-": ["13"]},
    "EBRIOLVS": {"-": ["21"]},
    "TRANSFORMIS": {"-": ["24"]},
    "AMPHITRITE": {"N": ["17"]},
    "SORORICIDA": {"-": ["11"]},
    "VRCEATIM": {"-": ["6-"]},
    "EPVLO": {"1": ["51"], "N": ["11"], "-": ["12"]},
    "IVNCEVS": {"-": ["21"]},
    "INTERNECTO": {"-": ["53"]},
    "DESVDASCO": {"-": ["53"]},
    "CALIGARIVS": {"-": ["21"]},
    "PEDISEQVA": {"-": ["11"]},
    "PARAETACENE": {"N": ["11"]},
    "PRIMITIAE": {"-": ["11"]},
    "OBTVEOR": {"-": ["5K"]},
    "FLEXIO": {"-": ["13"]},
    "INCOMPOSITE": {"-": ["6-"]},
    "DESPICOR": {"-": ["5J"]},
    "MIMIAMBI": {"-": ["12"]},
    "PACENSIS": {"N": ["13"]},
    "NOMENCVLATOR": {"-": ["13"]},
    "SVBLATVS": {"-": ["21"]},
    "MYCALE": {"N": ["11"]},
    "ACOR": {"-": ["13"]},
    "ILLVSTRE": {"-": ["6-"]},
    "PERNICIOSE": {"-": ["6-"]},
    "SELLVLA": {"-": ["11"]},
    "MECASTOR": {"-": ["6-"]},
    "FVRCIFERA": {"-": ["11"]},
    "SVBBASILICANVS": {"-": ["12"]},
    "SCRIBILITA": {"-": ["11"]},
    "OCTIPES": {"-": ["22"]},
    "CRANTOR": {"N": ["13"]},
    "IMPETRIO": {"-": ["54"]},
    "NISIBIS": {"N": ["13"]},
    "QVADRVPEDANS": {"1": ["13"], "2": ["12"]},
    "EXLEX": {"-": ["22"]},
    "DETRACTATOR": {"-": ["13"]},
    "PERCONTOR": {"-": ["5J"]},
    "DIAETA": {"-": ["11"]},
    "LABEO": {"N": ["13"]},
    "RECESSIM": {"-": ["6-"]},
    "FICTOR": {"-": ["13"]},
    "COLLIQVEFACIO": {"-": ["55"]},
    "AGEMA": {"-": ["13"]},
    "SANGVEN": {"-": ["13"]},
    "ROGATOR": {"-": ["13"]},
    "SEXTADECIMANI": {"-": ["12"]},
    "CVMBVLA": {"-": ["11"]},
    "PROMINVLVS": {"-": ["21"]},
    "ASPELLO": {"-": ["53"]},
    "PARASITICVS": {"-": ["21"]},
    "INSPECIOSVS": {"-": ["21"]},
    "LIMVLVS": {"-": ["21"]},
    "PHOLOE": {"N": ["11"]},
    "PHLEGETHON": {"N": ["13"]},
    "BIBLIOPOLA": {"-": ["11"]},
    "CONCINNE": {"-": ["6-"]},
    "PLVMOSVS": {"-": ["21"]},
    "NARO": {"N": ["13"]},
    "EPIDIPNIS": {"-": ["13"]},
    "ABSTO": {"-": ["51"]},
    "RICA": {"-": ["11"]},
    "MARMORO": {"-": ["51"]},
    "COMARCHVS": {"-": ["12"]},
    "DEDVCTOR": {"-": ["13"]},
    "SAXATILIS": {"-": ["24"]},
    "PRAEFODIO": {"-": ["55"]},
    "RHETORICA": {"-": ["11"]},
    "THYMELE": {"N": ["11"]},
    "OCTONVS": {"-": ["21"]},
    "FOLLITVS": {"-": ["21"]},
    "LOLIGVNCVLA": {"-": ["11"]},
    "CLVO": {"-": ["53"]},
    "ANIMADVERSOR": {"-": ["13"]},
    "HIPPOLYTE": {"N": ["11"]},
    "AVTONOE": {"N": ["11"]},
    "DIVINE": {"-": ["6-"]},
    "APOLACTIZO": {"-": ["51"]},
    "VERBEREVS": {"-": ["21"]},
    "PRAECONIVS": {"-": ["21"]},
    "NOCTVINVS": {"-": ["21"]},
    "COMMENDATOR": {"-": ["13"]},
    "INGIGNO": {"-": ["53"]},
    "PVBLICITVS": {"-": ["6-"]},
    "INSTIMVLATOR": {"-": ["13"]},
    "DVCENARIVS": {"-": ["12"]},
    "BABAE": {"-": ["11"]},
    "PHANTASIA": {"-": ["11"]},
    "EXENTERO": {"-": ["51"]},
    "HORRIDE": {"-": ["6-"]},
    "PYGISIACVS": {"-": ["21"]},
    "MESOCHORVS": {"-": ["12"]},
    "ACCLINO": {"-": ["51"]},
    "SMARIS": {"-": ["13"]},
    "BELLATVLVS": {"-": ["21"]},
    "EXEMPLARES": {"-": ["13"]},
    "RVBIDVS": {"-": ["21"]},
    "PERDECORVS": {"-": ["21"]},
    "TVBEROSVS": {"-": ["21"]},
    "VOLVPTABILIS": {"-": ["24"]},
    "INDEMNATVS": {"-": ["21"]},
    "HELVO": {"-": ["13"]},
    "CALYPSO": {"N": ["13"]},
    "CALLIOPE": {"N": ["11"]},
    "TISIPHONE": {"N": ["11"]},
    "CONSVASOR": {"-": ["13"]},
    "OCVLISSIMVS": {"-": ["21"]},
    "COLVMBVLVS": {"-": ["12"]},
    "SALVTIGERVLVS": {"-": ["21"]},
    "LESBOS": {"N": ["12"]},
    "ALICVLA": {"-": ["11"]},
    "QVADRINGENTIENS": {"-": ["6-"]},
    "PERSIMILIS": {"-": ["24"]},
    "ANCORARIVS": {"-": ["12"]},
    "PRAESTINO": {"-": ["51"]},
    "HECATE": {"N": ["11"]},
    "CYLLENE": {"N": ["11"]},
    "MENDICVLVS": {"-": ["21"]},
    "PARTITVDO": {"-": ["13"]},
    "ODIOSE": {"-": ["6-"]},
    "DECLAMITO": {"-": ["51"]},
    "CALDICEREBRIVS": {"-": ["12"]},
    "CIRCVMFORANEVS": {"-": ["21"]},
    "SCABITVDO": {"-": ["13"]},
    "ANTEGREDIOR": {"-": ["5N"]},
    "DIPONDIARIVS": {"-": ["21"]},
    "MISCELLANEVS": {"-": ["21"]},
    "AGORANOMVS": {"-": ["12"]},
    "METICVLOSVS": {"-": ["21"]},
    "TRANSMISSIO": {"-": ["13"]},
    "IGNIPOTENS": {"-": ["22"]},
    "OFFENSIVNCVLA": {"-": ["11"]},
    "HEPATIA": {"-": ["12"]},
    "GENICVLATVS": {"-": ["21"]},
    "REMVRMVRO": {"-": ["51"]},
    "TRADITOR": {"-": ["13"]},
    "ERANVS": {"-": ["12"]},
    "ILLICITATOR": {"-": ["13"]},
    "PALAESTE": {"N": ["11"]},
    "BERENICE": {"N": ["11"]},
    "DEFENSITO": {"-": ["51"]},
    "DEGRANDINAT": {"-": ["51"]},
    "PANCRATICE": {"-": ["6-"]},
    "MANTISSA": {"-": ["11"]},
    "DVCTIM": {"-": ["6-"]},
    "AVGVRIVS": {"-": ["21"]},
    "SPHINX": {"N": ["13"]},
    "DRYOPE": {"N": ["11"]},
    "SCORTOR": {"-": ["5J"]},
    "INTERTVRBO": {"2": ["51"]},
    "BOTVLVS": {"-": ["12"]},
    "TOPIARIVS": {"-": ["12"]},
    "QVIRITATVS": {"-": ["14"]},
    "LIMBOLARIVS": {"-": ["12"]},
    "POLENTARIVS": {"-": ["21"]},
    "PROCRIS": {"N": ["13"]},
    "SVSQVEDEQVE": {"-": ["6-"]},
    "ADVENTOR": {"-": ["13"]},
    "ATESTE": {"N": ["11"]},
    "QVIRIS": {"N": ["13"]},
    "VIGILAX": {"-": ["13"]},
    "APICVLA": {"-": ["11"]},
    "RVCTOR": {"-": ["5J"]},
    "SAEVITVDO": {"-": ["13"]},
    "HELICE": {"N": ["17"]},
    "QVINVS": {"-": ["21"]},
    "NEPA": {"-": ["11"]},
    "PRAEPOSTERE": {"-": ["6-"]},
    "PRAEBITOR": {"-": ["13"]},
    "TRECENTIENS": {"-": ["6-"]},
    "SVFFLATVS": {"-": ["14"]},
    "MVLTIBIBVS": {"-": ["21"]},
    "COPTOPLACENTA": {"-": ["11"]},
    "GELIDE": {"-": ["6-"]},
    "PRAETERVEHO": {"-": ["53"]},
    "HETAERIA": {"-": ["11"]},
    "TRIENS": {"-": ["22"]},
    "VTIBILIS": {"-": ["24"]},
    "DETONO": {"-": ["51"]},
    "DESTRICTE": {"-": ["6-"]},
    "PROSPICIENTIA": {"-": ["11"]},
    "PIRENE": {"N": ["11"]},
    "COMMEMORABILIS": {"-": ["24"]},
    "PAVXILLATIM": {"-": ["6-"]},
    "DIGLADIOR": {"-": ["5J"]},
    "MANIFESTARIVS": {"-": ["21"]},
    "DEFVNCTORIVS": {"-": ["21"]},
    "NEBRIS": {"-": ["13"]},
    "PELAGIVS": {"-": ["21"]},
    "INHONORVS": {"-": ["21"]},
    "CALVITIES": {"-": ["15"]},
    "MOECHISSO": {"-": ["51"]},
    "INDOMABILIS": {"-": ["24"]},
    "SCRIPTITO": {"-": ["51"]},
    "CONIECTOR": {"-": ["13"]},
    "CONDORMISCO": {"-": ["53"]},
    "LEVCAS": {"N": ["13"]},
    "ANTHRAX": {"N": ["13"]},
    "TRICLINIARCHES": {"-": ["11"]},
    "CYCLICVS": {"-": ["21"]},
    "PERBASIO": {"-": ["51"]},
    "EXDORSVO": {"-": ["51"]},
    "EFFVLTVS": {"-": ["21"]},
    "PERPLEXABILIS": {"-": ["24"]},
    "CROCOTARIVS": {"-": ["21"]},
    "IMBERBVS": {"-": ["21"]},
    "MEMPHIS": {"N": ["13"]},
    "BESSALIS": {"-": ["24"]},
    "BONATVS": {"-": ["21"]},
    "HISTRIONIA": {"-": ["11"]},
    "VERBOSE": {"-": ["6-"]},
    "OBSONATVS": {"-": ["14"]},
    "FVLCIPEDIA": {"-": ["11"]},
    "AMATRIX": {"-": ["22"]},
    "OBNOXIOSE": {"-": ["6-"]},
    "NESAPIVS": {"-": ["21"]},
    "INTERDIVS": {"-": ["6-"]},
    "RETRICES": {"-": ["13"]},
    "DENTIFRANGIBVLVS": {"-": ["21"]},
    "PANTHERINVS": {"-": ["21"]},
    "INGRESSIO": {"-": ["13"]},
    "PERISTASIS": {"-": ["13"]},
    "PERINDVLGENS": {"-": ["22"]},
    "QVADRILIBRIS": {"-": ["24"]},
    "NIDVLVS": {"-": ["12"]},
    "PERCALLESCO": {"-": ["53"]},
    "PELLENE": {"N": ["11"]},
    "CANTABVNDVS": {"-": ["21"]},
    "NAVTEA": {"-": ["11"]},
    "DAVLIS": {"N": ["13"]},
    "PRIMVLVS": {"-": ["21"]},
    "VICTITO": {"-": ["51"]},
    "INFECTOR": {"-": ["13"]},
    "PRAEVARICATOR": {"-": ["13"]},
    "GESTATOR": {"-": ["13"]},
    "ESCVLENTVS": {"-": ["21"]},
    "BARCALA": {"-": ["11"]},
    "APPROMITTO": {"-": ["53"]},
    "CALAMISTRATVS": {"-": ["21"]},
    "MELINA": {"-": ["11"]},
    "ACTE": {"N": ["17"]},
    "PILANVS": {"-": ["12"]},
    "DENASO": {"-": ["51"]},
    "COAEQVALIS": {"-": ["24"]},
    "PHILOPATOR": {"N": ["13"]},
    "INFRENO": {"-": ["51"]},
    "APELLO": {"-": ["53"]},
    "CALYMNE": {"N": ["11"]},
    "PERTVSVS": {"-": ["21"]},
    "CVRIS": {"-": ["13"]},
    "PVLVINARIS": {"-": ["24"]},
    "DIPLOMA": {"-": ["13"]},
    "IMPENDIOSVS": {"-": ["21"]},
    "DISPALESCO": {"-": ["53"]},
    "BILIX": {"-": ["22"]},
    "CRVRALIS": {"-": ["24"]},
    "FABRE": {"-": ["6-"]},
    "AFFORMIDO": {"2": ["51"]},
    "TVSSICVLA": {"-": ["11"]},
    "SCRVPVS": {"-": ["12"]},
    "PERDOCTVS": {"-": ["21"]},
    "CELOX": {"-": ["13"]},
    "CONCVSTODIO": {"-": ["54"]},
    "PAVXILLVLVM": {"2": ["12"]},
    "LEPOS": {"-": ["13"], "N": ["13"]},
    "ADMOLIOR": {"-": ["5M"]},
    "SCISSOR": {"-": ["13"]},
    "LODICVLA": {"-": ["11"]},
    "DECIMO": {"-": ["6-"]},
    "SACERDOTALIS": {"-": ["13"]},
    "VNOCVLVS": {"-": ["12"]},
    "NIXVS": {"-": ["21"]},
    "VICIENS": {"-": ["6-"]},
    "CLYMENE": {"N": ["11"]},
    "PROFATVS": {"-": ["21"]},
    "QVARTANI": {"-": ["12"]},
    "FRATRICIDA": {"-": ["11"]},
    "DESCENSIO": {"-": ["13"]},
    "PHANTASMA": {"-": ["13"]},
    "TAPANTA": {"-": ["16"]},
    "HIPPODAME": {"N": ["11"]},
    "INCORRVPTE": {"-": ["6-"]},
    "TEPIDE": {"-": ["6-"]},
    "AGNELLVS": {"-": ["12"]},
    "ARCESSITOR": {"-": ["13"]},
    "INTRIBVO": {"-": ["53"]},
    "SPVRCATVS": {"-": ["21"]},
    "CENATVS": {"-": ["21"]},
    "TOGVLA": {"-": ["11"]},
    "SIMVLATE": {"-": ["6-"]},
    "CALAIS": {"N": ["13"]},
    "ASTIPVLATOR": {"-": ["13"]},
    "HIPPOCRENE": {"N": ["11"]},
    "PRASINATVS": {"-": ["21"]},
    "RVBICO": {"N": ["13"]},
    "BALAENACEVS": {"-": ["21"]},
    "IMMVTATVS": {"-": ["21"]},
    "SCALPTOR": {"-": ["13"]},
    "TVRBVLENTE": {"-": ["6-"]},
    "SVBBLANDIOR": {"-": ["5M"]},
    "DIRCE": {"N": ["11"]},
    "INTERIBI": {"-": ["6-"]},
    "CONCESSO": {"-": ["51"]},
    "CONVADOR": {"-": ["5J"]},
    "TVNICVLA": {"-": ["11"]},
    "TVBERO": {"N": ["13"]},
    "PHOEBE": {"N": ["11"]},
    "AVRARIVS": {"2": ["21"]},
    "SVBDOMO": {"-": ["51"]},
    "PRAEQVAM": {"-": ["6-"]},
    "HEPATARIVS": {"-": ["21"]},
    "CINCTICVLVS": {"-": ["12"]},
    "BOVILE": {"-": ["13"]},
    "CVIATIS": {"-": ["2&"]},
    "ANTIPOLIS": {"N": ["13"]},
    "RENVTVS": {"-": ["14"]},
    "HAESITANTIA": {"-": ["11"]},
    "CVSTODITE": {"-": ["6-"]},
    "INTESTATO": {"-": ["6-"]},
    "MALEDICAX": {"-": ["22"]},
    "PERFACETVS": {"-": ["21"]},
    "RESECRO": {"-": ["51"]},
    "OFFRENATVS": {"-": ["21"]},
    "SOCIENNVS": {"-": ["12"]},
    "DERVNCINO": {"-": ["51"]},
    "STROPHIARIVS": {"-": ["12"]},
    "RESTIPVLOR": {"-": ["5J"]},
    "MILLESIMA": {"-": ["6-"]},
    "BERBEX": {"-": ["13"]},
    "ASSVO": {"-": ["53"]},
    "GILLO": {"N": ["13"]},
    "SCVTARIVS": {"-": ["21"]},
    "REPVERASCO": {"-": ["53"]},
    "CAVILLA": {"-": ["11"]},
    "RVSCINO": {"N": ["13"]},
    "BENEMORIVS": {"-": ["21"]},
    "CONCILIATRICVLA": {"-": ["11"]},
    "LIBRARIOLVS": {"-": ["12"]},
    "INSATVRABILIS": {"-": ["24"]},
    "INCIPISSO": {"-": ["53"]},
    "PACIFICATORIVS": {"-": ["21"]},
    "PRAETRVNCO": {"-": ["51"]},
    "DEPVLSIO": {"-": ["13"]},
    "GVSTATVS": {"-": ["21"]},
    "MELICA": {"-": ["11"]},
    "VEPRECVLA": {"-": ["11"]},
    "PERSPICAX": {"-": ["22"]},
    "MYROPOLA": {"-": ["11"]},
    "PERCVPIO": {"-": ["55"]},
    "FIGVLARIS": {"-": ["24"]},
    "CARINARIVS": {"-": ["12"]},
    "ANNOTATOR": {"-": ["13"]},
    "INCERTE": {"-": ["6-"]},
    "LIQVO": {"-": ["51"]},
    "PROTINAM": {"-": ["6-"]},
    "COCHLEARE": {"-": ["13"]},
    "PERCRVCIO": {"-": ["51"]},
    "IMPERSPICVVS": {"-": ["21"]},
    "PERPETVARIVS": {"-": ["12"]},
    "QVISQVILIAE": {"-": ["11"]},
    "TECHNA": {"-": ["11"]},
    "AMASTRIS": {"N": ["13"]},
    "OBSERVANS": {"-": ["22"]},
    "ACERBO": {"-": ["51"]},
    "PROOEMIOR": {"-": ["5J"]},
    "PHILIPPOPOLIS": {"N": ["13"]},
    "INCONSIDERATE": {"-": ["6-"]},
    "VERENDA": {"-": ["11"]},
    "DIOECETES": {"-": ["17"]},
    "QVADRAGIENS": {"-": ["6-"]},
    "DEFLOCCO": {"-": ["51"]},
    "CYMOTHOE": {"N": ["11"]},
    "POSTIDEA": {"-": ["6-"]},
    "CORYCVS": {"-": ["12"]},
    "NVGIGERVLVS": {"-": ["12"]},
    "MILVA": {"-": ["11"]},
    "IAMBEVS": {"-": ["21"]},
    "MALVGINENSIS": {"N": ["13"]},
    "DEBELLATOR": {"-": ["13"]},
    "DISSORS": {"-": ["22"]},
    "PLVSSCIVS": {"-": ["21"]},
    "FVLLONIA": {"-": ["11"]},
    "MILO": {"N": ["13"]},
    "PELLICATVS": {"-": ["14"]},
    "FLEMINA": {"-": ["13"]},
    "PRAECANTO": {"-": ["51"]},
    "PRAETERMISSIO": {"-": ["13"]},
    "SVBSICIVVS": {"-": ["21"]},
    "REVOLVBILIS": {"-": ["24"]},
    "LAMENTARIVS": {"-": ["21"]},
    "TROEZEN": {"N": ["13"]},
    "ACIDE": {"-": ["6-"]},
    "PVRPVRISSATVS": {"-": ["21"]},
    "OTHO": {"N": ["13"]},
    "DAPSILIS": {"-": ["24"]},
    "PERIODVS": {"-": ["12"]},
    "ADAEQVE": {"-": ["6-"]},
    "TIBVR": {"N": ["13"]},
    "TRIMETER": {"1": ["12"]},
    "INSTIGATOR": {"-": ["13"]},
    "VXORCVLA": {"-": ["11"]},
    "NVMERABILIS": {"-": ["24"]},
    "QVIPPINI": {"-": ["6-"]},
    "OBSEQVIOSVS": {"-": ["21"]},
    "MALACISSO": {"-": ["51"]},
    "VENVSTVLVS": {"-": ["21"]},
    "OBTEXO": {"-": ["53"]},
    "SVBLESTVS": {"-": ["21"]},
    "ALIQVANTILLVS": {"-": ["21"]},
    "IATRALIPTES": {"-": ["17"]},
    "PSEVDOMENOS": {"-": ["17"]},
    "EFFRENATE": {"-": ["6-"]},
    "TEMPORI": {"-": ["6-"]},
    "CIRCVMMINGO": {"-": ["6-"]},
    "CVIAS": {"-": ["13"]},
    "CLIBANVS": {"-": ["12"]},
    "VITVLINA": {"-": ["11"]},
    "VINVS": {"-": ["12"]},
    "BOSTAR": {"N": ["13"]},
    "LVRCHINABVNDVS": {"-": ["21"]},
    "PATRIMVS": {"-": ["21"]},
    "TRANSNATO": {"-": ["51"]},
    "VASA": {"-": ["12"]},
    "TERVNCIVS": {"-": ["12"]},
    "SARCINATOR": {"-": ["13"]},
    "CROCOTA": {"-": ["11"]},
    "MACRITVDO": {"-": ["13"]},
    "OBSTINO": {"-": ["51"]},
    "CVNCTO": {"-": ["51"]},
    "RVMVSCVLVS": {"-": ["12"]},
    "PROMVLSIS": {"-": ["13"]},
    "ALEC": {"-": ["13"]},
    "ABNVTO": {"-": ["51"]},
    "ACIS": {"N": ["13"]},
    "NOVIENS": {"-": ["6-"]},
    "INEPTIA": {"-": ["11"]},
    "ILLOCABILIS": {"-": ["24"]},
    "SPVTO": {"-": ["51"]},
    "SESQVIHORA": {"-": ["11"]},
    "SEPIOLA": {"-": ["11"]},
    "QVAQVA": {"-": ["6-"]},
    "PIAMEN": {"-": ["13"]},
    "EXOBSECRO": {"-": ["51"]},
    "VPVPA": {"-": ["11"]},
    "DOMVSIO": {"-": ["13"]},
    "DEPECVLATVS": {"-": ["21"]},
    "PERPOLIO": {"-": ["54"], "2": ["21"]},
    "INTERSTRINGO": {"-": ["53"]},
    "EVAN": {"N": ["16"]},
    "ZAMIA": {"-": ["11"]},
    "AVDIENTIA": {"-": ["11"]},
    "ANNEXVS": {"-": ["14"]},
    "STERILICVLA": {"-": ["11"]},
    "MORMYR": {"-": ["13"]},
    "FERICVLVS": {"-": ["21"]},
    "LARIFVGA": {"-": ["11"]},
    "PASTORIVS": {"-": ["21"]},
    "ANATHYMIASIS": {"-": ["13"]},
    "PRAELOQVOR": {"-": ["5L"]},
    "ELVVIO": {"-": ["13"]},
    "ANTIDHAC": {"-": ["6-"]},
    "IMMEDICABILIS": {"-": ["24"]},
    "FENEVS": {"-": ["21"]},
    "TALARIVS": {"-": ["21"]},
    "CATILLO": {"2": ["13"]},
    "DANISTA": {"-": ["11"]},
    "SPARVLVS": {"-": ["12"]},
    "EFFLICTIM": {"-": ["6-"]},
    "IEIVNIOSVS": {"-": ["21"]},
    "INCVRIOSE": {"-": ["6-"]},
    "ABORTIO": {"1": ["54"]},
    "BALLIO": {"N": ["13"]},
    "HOMOLE": {"N": ["11"]},
    "COMMENDATRIX": {"-": ["13"]},
    "TETHYS": {"N": ["13"]},
    "SAGARIS": {"N": ["13"]},
    "VINNVLVS": {"-": ["21"]},
    "IMPLICISCO": {"-": ["53"]},
    "PRAEFRACTE": {"-": ["6-"]},
    "LORVS": {"-": ["12"]},
    "PARRICIDALIS": {"-": ["24"]},
    "DILATOR": {"-": ["51"]},
    "NVPERVS": {"-": ["21"]},
    "METATOR": {"-": ["13"]},
    "INDISPOSITVS": {"-": ["21"]},
    "SYNGRAPHVS": {"-": ["12"]},
    "FRENI": {"-": ["12"]},
    "RHODOPE": {"N": ["11"]},
    "GIGERIA": {"-": ["12"]},
    "LEMNISCATVS": {"-": ["21"]},
    "ADOLESCENTVLA": {"-": ["11"]},
    "EXCETRA": {"-": ["11"]},
    "CONSERVA": {"-": ["11"]},
    "IMPERDITVS": {"-": ["21"]},
    "BION": {"N": ["13"]},
    "MELLILLA": {"-": ["11"]},
    "INVVLNERATVS": {"-": ["21"]},
    "THESAVRARIVS": {"2": ["12"]},
    "MASCARPIO": {"-": ["13"]},
    "OCVLEVS": {"-": ["21"]},
    "COPS": {"-": ["22"]},
    "HISPO": {"N": ["13"]},
    "MALEFACTOR": {"-": ["13"]},
    "PRAEDICATOR": {"-": ["13"]},
    "COLLARIS": {"-": ["24"]},
    "PRAXIS": {"-": ["13"]},
    "RETRACTVS": {"-": ["14"]},
    "TRIFVR": {"-": ["13"]},
    "LAODICE": {"N": ["11"]},
    "DIIVNCTE": {"-": ["6-"]},
    "CATAPHAGAS": {"-": ["17"]},
    "DIERECTVS": {"-": ["21"]},
    "SVTELA": {"-": ["11"]},
    "INSVLARIVS": {"-": ["12"]},
    "CONTEMPLO": {"-": ["51"]},
    "FRONTO": {"N": ["13"]},
    "ANTESCHOLANVS": {"-": ["12"]},
    "STERNVTO": {"-": ["51"]},
    "PLVMATILIS": {"-": ["24"]},
    "SECVTVLEIVS": {"-": ["21"]},
    "CORBITA": {"-": ["11"]},
    "CLAVIGER": {"2": ["23"]},
    "OCCALESCO": {"-": ["53"]},
    "VOCATVS": {"-": ["14"]},
    "DISPVTATOR": {"-": ["13"]},
    "RECLAMITO": {"-": ["51"]},
    "GAVSAPA": {"-": ["12"]},
    "PROSERPO": {"-": ["53"]},
    "CISTELLVLA": {"-": ["11"]},
    "APPENDICVLA": {"-": ["11"]},
    "POPPYSMA": {"-": ["13"]},
    "NICANOR": {"N": ["13"]},
    "CIVITATVLA": {"-": ["11"]},
    "MINITO": {"-": ["51"]},
    "ALCITHOE": {"N": ["17"]},
    "INFRENIS": {"-": ["24"]},
    "SACRICOLA": {"-": ["11"]},
    "DEPOPVLATOR": {"-": ["13"]},
    "ALCIMEDE": {"N": ["17"]},
    "COLEVS": {"-": ["12"]},
    "COLLATOR": {"-": ["13"]},
    "LOCELLVS": {"-": ["12"]},
    "ELEPHANTINE": {"N": ["11"]},
    "SEVIR": {"-": ["12"]},
    "LOQVENTIA": {"-": ["11"]},
    "RVBRICATVS": {"-": ["21"]},
    "MINOS": {"N": ["12"]},
    "DEMARCHVS": {"-": ["12"]},
    "ARENARIA": {"-": ["11"]},
    "PASSERCVLVS": {"-": ["12"]},
    "PHALANGAE": {"-": ["11"]},
    "CONCVSSIO": {"-": ["13"]},
    "CONTONAT": {"-": ["51"]},
    "EXPOSITICIVS": {"-": ["21"]},
    "LACTICVLOSVS": {"-": ["12"]},
    "LAMPETIE": {"N": ["11"]},
    "SYROPHOENIX": {"N": ["13"]},
    "TENELLVS": {"-": ["21"]},
    "PRAESEPIS": {"-": ["13"]},
    "PRAEVOLO": {"2": ["51"]},
    "STELA": {"-": ["11"]},
    "CHRYSIS": {"N": ["13"]},
    "DENVMERO": {"1": ["51"]},
    "PROVOCATOR": {"-": ["13"]},
    "ADVERSARIA": {"-": ["12"]},
    "MVROBATHARIVS": {"-": ["12"]},
    "PSEPHISMA": {"-": ["13"]},
    "AREOLA": {"-": ["11"]},
    "HIRA": {"-": ["11"]},
    "TERTIANA": {"-": ["11"]},
    "INTERBIBO": {"2": ["53"]},
    "SVBLATE": {"-": ["6-"]},
    "PRAESTES": {"A": ["13"]},
    "QVINTVSDECIMVS": {"-": ["21"]},
    "TVATIM": {"-": ["6-"]},
    "PERLONGINQVVS": {"-": ["21"]},
    "CONTRITVS": {"-": ["21"]},
    "PALMARIS": {"2": ["24"]},
    "OBSONOR": {"-": ["5J"]},
    "EXPVTRESCO": {"-": ["53"]},
    "INGRAVO": {"-": ["51"]},
    "CALAMITOSE": {"-": ["6-"]},
    "MELANIPPE": {"N": ["11"]},
    "RVGO": {"-": ["51"]},
    "COALO": {"-": ["53"]},
    "REDORMIO": {"-": ["54"]},
    "AMPVLLA": {"-": ["11"]},
    "AGINO": {"-": ["51"]},
    "MAVORS": {"N": ["13"]},
    "CONDICTOR": {"-": ["13"]},
    "OLLA": {"-": ["11"]},
    "SINGVLTO": {"-": ["51"]},
    "FABALIS": {"-": ["24"]},
    "AFFLICTOR": {"-": ["13"]},
    "MEROE": {"N": ["11"]},
    "IMBITO": {"-": ["53"]},
    "BATVO": {"-": ["53"]},
    "LATERICVLVS": {"-": ["12"]},
    "ASCRIPTOR": {"-": ["13"]},
    "SACCELLVS": {"-": ["12"]},
    "ANAPAESTVS": {"-": ["12"]},
    "SYENE": {"N": ["11"]},
    "CANACE": {"N": ["11"]},
    "SVBREPTICIVS": {"-": ["21"]},
    "TRIBVARIVS": {"-": ["21"]},
    "CERNVVS": {"2": ["21"]},
    "OFFLA": {"-": ["11"]},
    "GERVLIFIGVLVS": {"-": ["12"]},
    "PERACESCO": {"-": ["53"]},
    "CAVPONVLA": {"-": ["11"]},
    "INTERTRAHO": {"-": ["53"]},
    "INVESTIGATOR": {"-": ["13"]},
    "FORENSIA": {"-": ["13"]},
    "AVGVRO": {"-": ["51"]},
    "EXPALLIATVS": {"-": ["21"]},
    "BVCCO": {"-": ["13"]},
    "PVMICEVS": {"-": ["21"]},
    "GINGILIPHVS": {"-": ["12"]},
    "RECVLA": {"-": ["11"]},
    "IVLIS": {"-": ["13"]},
    "INFLETVS": {"-": ["21"]},
    "AVDACVLVS": {"-": ["21"]},
    "PVRGAMEN": {"-": ["13"]},
    "CATAPVLTARIVS": {"-": ["21"]},
    "DETVMESCO": {"-": ["53"]},
    "MVNITO": {"-": ["51"]},
    "ELINGVO": {"-": ["51"]},
    "LVCVLENTE": {"-": ["6-"]},
    "MOLOCHINARIVS": {"-": ["12"]},
    "PHYLLIS": {"N": ["13"]},
    "HONORABILIS": {"-": ["24"]},
    "SPECTANS": {"-": ["22"]},
    "PECVLATOR": {"-": ["13"]},
    "TONSVS": {"-": ["21"]},
    "MALACVS": {"-": ["21"]},
    "NOCVVS": {"-": ["21"]},
    "COPVLATVS": {"-": ["14"]},
    "VRSINA": {"-": ["11"]},
    "IVRATOR": {"-": ["13"]},
    "SYCOPHANTA": {"-": ["11"]},
    "PETASVS": {"-": ["12"]},
    "DEFRAVDO": {"-": ["51"]},
    "PVGIVNCVLVS": {"-": ["12"]},
    "PERSEPOLIS": {"N": ["13"]},
    "AGANIPPE": {"N": ["11"]},
    "ZEVXIS": {"N": ["13"]},
    "PARONYCHIA": {"-": ["11"]},
    "COCCINEVS": {"-": ["21"]},
    "LACHESIS": {"N": ["13"]},
    "PSYCHROLVTA": {"-": ["11"]},
    "EXCATARISSO": {"-": ["51"]},
    "LVPATRIA": {"-": ["11"]},
    "INARIME": {"N": ["11"]},
    "LONGIVSCVLVS": {"-": ["21"]},
    "VACIVVS": {"-": ["21"]},
    "PLEIONE": {"N": ["11"]},
    "VITISATOR": {"-": ["13"]},
    "EDITICIVS": {"-": ["21"]},
    "OBTICESCO": {"-": ["53"]},
    "MELIE": {"N": ["11"]},
    "FVRFVR": {"-": ["13"]},
    "PINNVLA": {"-": ["11"]},
    "COMATORIVS": {"-": ["21"]},
    "COMPOTOR": {"-": ["13"]},
    "DELIQVIO": {"-": ["13"]},
    "ARCVLARIVS": {"-": ["12"]},
    "ILLECEBROSVS": {"-": ["21"]},
    "SCAPHA": {"-": ["11"]},
    "INCOMMODESTICVS": {"-": ["21"]},
    "PVGNEVS": {"-": ["21"]},
    "FORTVNATE": {"-": ["6-"]},
    "SARGVS": {"-": ["12"]},
    "LEMMA": {"-": ["13"]},
    "FVGIENS": {"2": ["22"]},
    "COLONICVS": {"-": ["21"]},
    "ARIS": {"N": ["13"]},
    "PERNECESSE": {"-": ["6-"]},
    "COGNOMINIS": {"-": ["24"]},
    "CRYPTOPORTICVS": {"-": ["14"]},
    "EXTEXO": {"-": ["53"]},
    "PERBENE": {"-": ["6-"]},
    "PROPEXVS": {"-": ["21"]},
    "PROSECTA": {"-": ["11"]},
    "FASTOSVS": {"-": ["21"]},
    "IMPERCO": {"-": ["53"]},
    "VINEVS": {"-": ["21"]},
    "FVRAX": {"-": ["22"]},
    "FREMOR": {"-": ["13"]},
    "DEFERVESCO": {"-": ["53"]},
    "FRATERNE": {"-": ["6-"]},
    "OCCLAMITO": {"-": ["51"]},
    "HECATOMPYLOS": {"N": ["17"]},
    "VERSIPELLIS": {"-": ["13"]},
    "DESVEFACIO": {"-": ["55"]},
    "EBOREVS": {"-": ["21"]},
    "OCCENTO": {"-": ["51"]},
    "SORICINVS": {"-": ["21"]},
    "EVILESCO": {"-": ["53"]},
    "ABSVMEDO": {"-": ["13"]},
    "ELECEBRA": {"-": ["11"]},
    "NVNDINALIS": {"-": ["24"]},
    "ADMVTILO": {"-": ["51"]},
    "ANTICATO": {"N": ["13"]},
    "ESVRIALIS": {"-": ["24"]},
    "AMASIVS": {"-": ["12"]},
    "THEMIS": {"N": ["13"]},
    "VENDITATOR": {"-": ["13"]},
    "TEREBINTHINVS": {"-": ["21"]},
    "GYPSATVS": {"-": ["21"]},
    "ARTVTVS": {"-": ["21"]},
    "DVOETVICESIMANI": {"-": ["12"]},
    "DELETRIX": {"-": ["22"]},
    "IMITAMEN": {"-": ["13"]},
    "NVBECVLA": {"-": ["11"]},
    "MATTVS": {"-": ["21"]},
    "MOERIS": {"N": ["13"]},
    "DECEMVIRATVS": {"-": ["14"]},
    "PACISCO": {"-": ["53"]},
    "QVINQVAGIENS": {"-": ["6-"]},
    "CATOMIDIO": {"-": ["51"]},
    "PAVPERVS": {"-": ["21"]},
    "TERVENEFICVS": {"-": ["12"]},
    "AMOENA": {"-": ["12"]},
    "SVPERSTITIOSVS": {"-": ["21"]},
    "TVBVLATVS": {"-": ["21"]},
    "SVFFLAMINO": {"-": ["51"]},
    "INGRATIA": {"-": ["11"]},
    "VALENTVLVS": {"-": ["21"]},
    "VTVT": {"-": ["6-"]},
    "BOGVD": {"N": ["13"]},
    "SVAVITVDO": {"-": ["13"]},
    "TAVRVLVS": {"-": ["12"]},
    "ORARIVS": {"-": ["21"]},
    "DEPREHENSIO": {"-": ["13"]},
    "MASTIGIA": {"-": ["11"]},
    "PISTORIVS": {"-": ["21"]},
    "TVGIO": {"N": ["13"]},
    "MACILENTVS": {"-": ["21"]},
    "SCORDALIA": {"-": ["11"]},
    "MONEDVLA": {"-": ["11"]},
    "PVLCHRALIA": {"-": ["13"]},
    "PERVAGATVS": {"-": ["21"]},
    "AGAGA": {"-": ["11"]},
    "AMATORIE": {"-": ["6-"]},
    "INDECOR": {"-": ["22"]},
    "OPTATO": {"-": ["6-"]},
    "DEBATVO": {"-": ["53"]},
    "CERCYRVS": {"-": ["12"]},
    "TYRO": {"N": ["13"]},
    "DILVTE": {"-": ["6-"]},
    "OVILLA": {"-": ["11"]},
    "OPIPARE": {"-": ["6-"]},
    "EDENTVLVS": {"-": ["21"]},
    "SVFFRINGO": {"-": ["53"]},
    "SEMIDOCTVS": {"-": ["21"]},
    "VSVRARIVS": {"-": ["21"]},
    "CRYSTALLVS": {"-": ["12"]},
    "IVVENOR": {"-": ["5J"]},
    "REFORMATOR": {"-": ["13"]},
    "ROBOREVS": {"-": ["21"]},
    "RESTITVTOR": {"-": ["13"]},
    "ALOGVS": {"-": ["21"]},
    "CLIMACTERICVS": {"-": ["21"]},
    "ADLVO": {"-": ["53"]},
    "TROPICA": {"-": ["12"]},
    "SVBMISSIO": {"-": ["13"]},
    "CYMBALISTRIA": {"-": ["11"]},
    "SEVERITVDO": {"-": ["13"]},
    "LYRICA": {"-": ["12"]},
    "DISCOLORIVS": {"-": ["21"]},
    "ADMISSARIVS": {"-": ["21"]},
    "SOMNIOR": {"-": ["5J"]},
    "CONTINVE": {"-": ["6-"]},
    "GERRAE": {"-": ["11"]},
    "PVNITOR": {"-": ["13"]},
    "MEROPE": {"N": ["11"]},
    "INCRETO": {"-": ["51"]},
    "EPASTVS": {"-": ["21"]},
    "LITICEN": {"-": ["13"]},
    "MORIO": {"-": ["13"]},
    "INCALLIDE": {"-": ["6-"]},
    "STEROPE": {"N": ["11"]},
    "BAETIS": {"N": ["13"]},
    "DIRIBITOR": {"-": ["13"]},
    "CENATIVNCVLA": {"-": ["11"]},
    "FESTIVE": {"-": ["6-"]},
    "FEMEN": {"-": ["13"]},
    "MIMVLA": {"-": ["11"]},
    "EMISSICIVS": {"-": ["21"]},
    "SVBSCRIPTOR": {"-": ["13"]},
    "CAESICIVS": {"-": ["21"]},
    "REBITO": {"-": ["53"]},
    "EFFERTVS": {"-": ["21"]},
    "OBSERVANTIA": {"-": ["11"]},
    "TRANSENNA": {"-": ["11"]},
    "MATTEA": {"-": ["11"]},
    "SEPTEMVIRI": {"-": ["12"]},
    "REDEMPTO": {"-": ["51"]},
    "ELECTA": {"-": ["12"]},
    "SEXAGIENS": {"-": ["6-"]},
    "POSITOR": {"-": ["13"]},
    "MAGMENTARIVS": {"-": ["21"]},
    "CECROPS": {"N": ["13"]},
    "OBVALLO": {"-": ["51"]},
    "RHOMBVS": {"-": ["12"]},
    "PIETATICVLTRIX": {"-": ["13"]},
    "SECTARIVS": {"-": ["21"]},
    "SEPTEMVIRATVS": {"-": ["14"]},
    "CYRENE": {"N": ["11"]},
    "PERDOLESCO": {"-": ["53"]},
    "ITYS": {"N": ["13"]},
    "INIVRIVS": {"-": ["21"]},
    "ANATICVLA": {"-": ["11"]},
    "OLO": {"-": ["53"]},
    "ARMENTALIS": {"-": ["24"]},
    "CELERE": {"-": ["6-"]},
    "CVCVMVLA": {"-": ["11"]},
    "SVBALBVS": {"-": ["21"]},
    "ENYO": {"N": ["13"]},
    "ALCYONE": {"N": ["17"]},
    "DVCTITO": {"-": ["51"]},
    "TVRPIFICATVS": {"-": ["21"]},
    "DISCORDABILIS": {"-": ["24"]},
    "BYBLOS": {"N": ["17"]},
    "PHASIS": {"N": ["13"]},
    "EXAMBVLO": {"-": ["51"]},
    "DRVIDA": {"-": ["11"]},
    "DEIERO": {"-": ["51"]},
    "CYDIPPE": {"N": ["11"]},
    "CONSECTATRIX": {"-": ["13"]},
    "CLOACALIS": {"-": ["24"]},
    "CONCACO": {"-": ["51"]},
    "VAVATO": {"-": ["13"]},
    "SVBTERDVCO": {"-": ["53"]},
    "FRVSTILLATIM": {"-": ["6-"]},
    "ADAMBVLO": {"-": ["51"]},
    "EDISSERTO": {"-": ["51"]},
    "RHOEA": {"-": ["11"]},
    "CHANE": {"-": ["17"]},
    "DELINGO": {"-": ["53"]},
    "CONGREDIO": {"-": ["54"]},
    "PLVMEVS": {"-": ["21"]},
    "OBVSTVS": {"-": ["21"]},
    "ASSVLATIM": {"-": ["6-"]},
    "COMMISEROR": {"-": ["5J"]},
    "ENOTO": {"-": ["51"]},
    "FERRVMEN": {"-": ["13"]},
    "CORBVLO": {"N": ["13"]},
    "HARIOLOR": {"-": ["5J"]},
    "COMPRANSOR": {"-": ["13"]},
    "OCVLATVS": {"-": ["21"]},
    "OBEDIENTIA": {"-": ["11"]},
    "SALAMANDRA": {"-": ["11"]},
    "VERVINA": {"-": ["11"]},
    "CONTOLLO": {"-": ["53"]},
    "CAPPARIS": {"-": ["13"]},
    "DVVMVIRATVS": {"-": ["14"]},
    "DESIPIENS": {"-": ["22"]},
    "REDVCTOR": {"-": ["13"]},
    "COLLACRIMO": {"-": ["51"]},
    "IMPVRATVS": {"-": ["21"]},
    "MVFRIVS": {"-": ["12"]},
    "OPIME": {"-": ["6-"]},
    "PVTILLVS": {"-": ["21"]},
    "AEGLE": {"N": ["17"]},
    "CIRCITOR": {"-": ["13"]},
    "LVSCINIOLA": {"-": ["11"]},
    "SITVLA": {"-": ["11"]},
    "MODICO": {"-": ["6-"]},
    "BALITANS": {"-": ["22"]},
    "INCLITVS": {"-": ["21"]},
    "AXICIA": {"-": ["11"]},
    "ERVILIA": {"-": ["11"]},
    "INVOLVCRE": {"-": ["13"]},
    "CACCABVS": {"-": ["12"]},
    "NOCTIFER": {"A": ["12"]},
    "SVSPICIOSE": {"-": ["6-"]},
    "CIRCVMLVO": {"-": ["53"]},
    "TVRBELLAE": {"-": ["11"]},
    "AMYMONE": {"N": ["11"]},
    "STIGMATIAS": {"-": ["11"]},
    "DRAPETA": {"-": ["11"]},
    "NARRATIVNCVLA": {"-": ["11"]},
    "SALTATORIVS": {"-": ["21"]},
    "SVBOLFACIO": {"-": ["55"]},
    "CARIS": {"-": ["13"]},
    "PERTRECTO": {"-": ["51"]},
    "APICATVS": {"-": ["21"]},
    "DISPVDET": {"-": ["52"]},
    "LAOCOON": {"N": ["13"]},
    "ALOGIA": {"-": ["11"]},
    "EVRYDICE": {"N": ["11"]},
    "SVBSVLTO": {"-": ["51"]},
    "LARVO": {"-": ["51"]},
    "LYMPHOR": {"-": ["13"]},
    "INCONGRVENS": {"-": ["22"]},
    "ZOTHECVLA": {"-": ["11"]},
    "CONCERPO": {"-": ["53"]},
    "DEVOTO": {"-": ["51"]},
    "INERVDITVS": {"-": ["21"]},
    "PERTRIBVO": {"-": ["53"]},
    "BARCE": {"N": ["11"]},
    "BAVCIS": {"N": ["13"]},
    "PROLVSIO": {"-": ["13"]},
    "APODIXIS": {"-": ["13"]},
    "SVBSIGNO": {"-": ["51"]},
    "COMPETITRIX": {"-": ["13"]},
    "AVSPICATO": {"-": ["6-"]},
    "SEMIVSTVLO": {"-": ["51"]},
    "SVBFODIO": {"-": ["55"]},
    "CHROMIS": {"-": ["13"], "N": ["13"]},
    "MACEDO": {"N": ["13"], "A": ["25"]},
    "PRAESEGMEN": {"-": ["13"]},
    "DIONE": {"N": ["11"]},
    "OCIVS": {"-": ["6-"]},
    "POTIVNCVLA": {"-": ["11"]},
    "CVCVMA": {"-": ["11"]},
    "INSPOLIATVS": {"-": ["21"]},
    "PERMAESTVS": {"-": ["21"]},
    "PVELLARIVS": {"-": ["12"]},
    "SEXCENTI": {"-": ["21"]},
    "IMPERITE": {"-": ["6-"]},
    "INFACVNDVS": {"-": ["21"]},
    "PRAECVRSORIVS": {"-": ["21"]},
    "AGIS": {"N": ["13"]},
    "PRAEGVSTATOR": {"-": ["13"]},
    "EBVRNVS": {"-": ["21"]},
    "REDAVSPICO": {"-": ["51"]},
    "ERYTHINVS": {"-": ["12"]},
    "LETHE": {"N": ["11"]},
    "MAESTITVDO": {"-": ["13"]},
    "INFORTVNATVS": {"-": ["21"]},
    "SORTILEGVS": {"2": ["12"]},
    "ANSERINVS": {"-": ["21"]},
    "COMPES": {"-": ["22"]},
    "CALAMIS": {"N": ["13"]},
    "SVSPICO": {"-": ["51"]},
    "INSTIGATRIX": {"-": ["13"]},
    "IEIVNE": {"-": ["6-"]},
    "CONGIALIS": {"-": ["24"]},
    "CONDECET": {"-": ["52"]},
    "CERINVS": {"-": ["21"]},
    "LOTIS": {"N": ["13"]},
    "CATAPLVS": {"-": ["12"]},
    "EXCVRATVS": {"-": ["21"]},
    "SOPHOS": {"-": ["17"]},
    "PERGRAVIS": {"-": ["24"]},
    "SYMBOLA": {"-": ["11"]},
    "VISCATVS": {"-": ["21"]},
    "SAGATVS": {"-": ["21"]},
    "FRVSTRO": {"-": ["51"]},
    "CORDATVS": {"-": ["21"]},
    "COGITATE": {"-": ["6-"]},
    "SITICEN": {"-": ["13"]},
    "PROLAPSIO": {"-": ["13"]},
    "DECLINIS": {"-": ["24"]},
    "SERRVLA": {"-": ["11"]},
    "PESSVLVS": {"-": ["12"]},
    "EXASCIATVS": {"-": ["21"]},
    "OPPANGO": {"-": ["53"]},
    "INTERMINOR": {"-": ["5J"]},
    "BELLVLVS": {"-": ["21"]},
    "MORDEX": {"-": ["22"]},
    "TRAGVS": {"-": ["12"]},
    "SARDIS": {"N": ["13"]},
    "LIRIS": {"N": ["13"]},
    "AVTHEPSA": {"-": ["11"]},
    "CAPREAGINVS": {"-": ["21"]},
    "SARTOR": {"2": ["13"]},
    "SVBSIGNANVS": {"-": ["21"]},
    "NAPE": {"N": ["11"]},
    "PERCVRIOSVS": {"-": ["21"]},
    "PERACERBVS": {"-": ["21"]},
    "EOS": {"N": ["12"]},
    "SIPHVNCVLVS": {"-": ["12"]},
    "CALTHVLA": {"-": ["11"]},
    "TRAGICOMOEDIA": {"-": ["11"]},
    "SEPTEMVIRALIS": {"2": ["24"], "1": ["12"]},
    "ADRADO": {"-": ["53"]},
    "THETIS": {"N": ["13"]},
    "VERBERITO": {"-": ["51"]},
    "PECOROSVS": {"-": ["21"]},
    "FVNDITO": {"-": ["51"]},
    "PERLEVIS": {"-": ["24"]},
    "SINAPE": {"-": ["13"]},
    "FRVSTVLENTVS": {"-": ["21"]},
    "ECDICVS": {"-": ["12"]},
    "SOLITO": {"-": ["51"]},
    "PERCOLOPO": {"-": ["51"]},
    "ASSIDVO": {"2": ["51"]},
    "PALVDOSVS": {"-": ["21"]},
    "ANTHIAS": {"-": ["17"]},
    "RVRO": {"-": ["51"]},
    "COMMETO": {"1": ["51"]},
    "SEMIRAMIS": {"N": ["13"]},
    "SESTERTIARIVS": {"-": ["21"]},
    "LACINIA": {"-": ["11"]},
    "ADVESPERASCIT": {"-": ["53"]},
    "HEDEROSVS": {"-": ["21"]},
    "IMPRODICTVS": {"-": ["21"]},
    "ZAPLVTVS": {"-": ["21"]},
    "COMMENTARIOLVS": {"-": ["12"]},
    "OFFVCIA": {"-": ["11"]},
    "PERATVS": {"-": ["12"]},
    "INTROVOCO": {"-": ["51"]},
    "METVLA": {"-": ["11"]},
    "CLANDESTINO": {"-": ["6-"]},
    "MACESCO": {"-": ["53"]},
    "MERACVLVS": {"-": ["21"]},
    "CYPEROS": {"-": ["17"]},
    "FATVOR": {"1": ["5J"]},
    "INFELICO": {"-": ["51"]},
    "VINDEMITOR": {"-": ["13"], "N": ["13"]},
    "NAVFRAGO": {"-": ["51"]},
    "LVSTROR": {"-": ["5J"]},
    "DEPLORABVNDVS": {"-": ["21"]},
    "PEDICELLVS": {"-": ["12"]},
    "VENTVLVS": {"-": ["12"]},
    "ARQVATVS": {"-": ["12"]},
    "ANIMVLVS": {"-": ["12"]},
    "ATTACTVS": {"-": ["14"]},
    "DAPINO": {"-": ["51"]},
    "AMASIVNCVLA": {"-": ["11"]},
    "BEROE": {"N": ["11"]},
    "BIFARIAM": {"-": ["6-"]},
    "FOEDIFRAGVS": {"-": ["21"]},
    "BACALVSIAE": {"-": ["11"]},
    "ALCESTIS": {"N": ["13"]},
    "OCCILLO": {"-": ["51"]},
    "ANTIQVO": {"-": ["51"]},
    "PEDISEQVVS": {"-": ["21"]},
    "VERNILIS": {"-": ["24"]},
    "OBROGO": {"-": ["51"]},
    "SALVE": {"-": ["13"]},
    "VNAETVICESIMANI": {"-": ["12"]},
    "REMELIGO": {"-": ["13"]},
    "MILVINVS": {"-": ["21"]},
    "PRAEFISCINE": {"-": ["6-"]},
    "MVNDO": {"-": ["51"]},
    "ASTYANAX": {"N": ["13"]},
    "ACCVSATRIX": {"-": ["13"]},
    "PRAEDIATORIVS": {"-": ["21"]},
    "AGENOR": {"N": ["13"]},
    "INCONDITE": {"-": ["6-"]},
    "FVRTIFICVS": {"-": ["21"]},
    "CORRVPTE": {"-": ["6-"]},
    "HIRVDO": {"-": ["13"]},
    "PERTRISTIS": {"-": ["24"]},
    "LANIVS": {"-": ["12"]},
    "FAVTRIX": {"-": ["13"]},
    "PVGILATVS": {"-": ["14"]},
    "STRATEGVS": {"-": ["12"]},
    "MANVLEARIVS": {"-": ["12"]},
    "INDVCVLA": {"-": ["11"]},
    "SOLEARIVS": {"-": ["12"]},
    "FLICTVS": {"-": ["14"]},
    "POMPILVS": {"-": ["12"]},
    "TORRIS": {"-": ["13"]},
    "PHASELVS": {"-": ["12"]},
    "APPARATE": {"-": ["6-"]},
    "EGVRGITO": {"-": ["51"]},
    "RETICVLVS": {"-": ["12"]},
    "ARVINA": {"-": ["11"]},
    "IRACVNDE": {"-": ["6-"]},
    "PROMISSIO": {"-": ["13"]},
    "CANDELABRVS": {"-": ["12"]},
    "OFFIRMATVS": {"-": ["21"]},
    "DVODETRICIENS": {"-": ["6-"]},
    "PERATTENTE": {"-": ["6-"]},
    "PRAESAGIOR": {"-": ["5M"]},
    "CONSILIATOR": {"-": ["13"]},
    "PACATE": {"-": ["6-"]},
    "FVLCIMEN": {"-": ["13"]},
    "RAPA": {"-": ["11"]},
    "PIGNEROR": {"-": ["5J"]},
    "POSTILENA": {"-": ["11"]},
    "CALCITRO": {"1": ["13"], "2": ["12"]},
    "SILVESCO": {"-": ["53"]},
    "EXANIMALIS": {"-": ["24"]},
    "GALERITVS": {"-": ["12"]},
    "PERLECEBRA": {"-": ["11"]},
    "PRAEPROPERVS": {"-": ["21"]},
    "MALEVOLENS": {"-": ["13"]},
    "OMINATOR": {"-": ["13"]},
    "DECIENS": {"-": ["6-"]},
    "PERSCRVTO": {"-": ["51"]},
    "CHALCIOPE": {"N": ["11"]},
    "SVPELLECTICARIVS": {"-": ["21"]},
    "CAMELLA": {"-": ["11"]},
    "LEPAS": {"-": ["13"]},
    "MODIALIS": {"-": ["24"]},
    "SORBILIS": {"-": ["24"]},
    "INCLYTVS": {"-": ["21"]},
    "ANATINA": {"-": ["11"]},
    "DEARTVO": {"-": ["51"]},
    "REPORRIGO": {"2": ["53"]},
    "LINGVOSVS": {"-": ["21"]},
    "MORTVALIA": {"-": ["13"]},
    "TANAIS": {"N": ["13"]},
    "PERIVCVNDE": {"-": ["6-"]},
    "SEMELE": {"N": ["11"]},
    "ERIGONE": {"N": ["11"]},
    "FRVNISCOR": {"-": ["5L"]},
    "NEVTRO": {"-": ["6-"]},
    "ADVENTICIA": {"-": ["11"]},
    "NAVCLERICVS": {"-": ["21"]},
    "FERRVMINO": {"-": ["51"]},
    "RAMENTA": {"-": ["11"]},
    "CALPE": {"N": ["11"]},
    "PROELIARIS": {"-": ["24"]},
    "PIABILIS": {"-": ["24"]},
    "ARGVTOR": {"-": ["5J"]},
    "SCORTEVS": {"-": ["21"]},
    "PROMVLSIDARE": {"-": ["13"]},
    "ANTIGONE": {"N": ["17"]},
    "GERENS": {"-": ["22"]},
    "SVPPARASITOR": {"-": ["5J"]},
    "TRYGONVS": {"-": ["12"]},
    "EXAGGERATVS": {"-": ["21"]},
    "OCLOPETA": {"-": ["11"]},
    "CONDVCIBILIS": {"-": ["24"]},
    "PRAECVRSIO": {"-": ["13"]},
    "INDIREPTVS": {"-": ["21"]},
    "SPORTELLA": {"-": ["11"]},
    "PEROPTATVS": {"-": ["21"]},
    "EMPTICIVS": {"-": ["21"]},
    "ILLEX": {"1": ["13"]},
    "TRICAE": {"-": ["11"]},
    "BASILICE": {"-": ["17"]},
    "DVRIVSCVLVS": {"-": ["21"]},
    "LYRISTES": {"-": ["17"]},
    "PVLTARIVS": {"-": ["12"]},
    "INTERAESTVO": {"-": ["51"]},
    "FALSILOQVVS": {"-": ["12"]},
    "PEDIS": {"-": ["13"]},
    "QVADRIPARTITO": {"-": ["6-"]},
    "NONAGIENS": {"-": ["6-"]},
    "SVBACCVSO": {"-": ["51"]},
    "SCELESTE": {"-": ["6-"]},
    "MICARIVS": {"-": ["21"]},
    "SEXTVSDECIMVS": {"-": ["21"]},
    "OPHTHALMIA": {"-": ["11"]},
    "INDVBITO": {"-": ["51"]},
    "INSTANTIA": {"-": ["11"]},
    "STRABONVS": {"-": ["21"]},
    "NVPTVS": {"-": ["21"]},
    "INTERCALARIS": {"-": ["24"]},
    "PRAEFOCO": {"-": ["51"]},
    "OBDORMIO": {"-": ["54"]},
    "SEPTIMANI": {"-": ["12"]},
    "REMANSIO": {"-": ["13"]},
    "PERENDIE": {"-": ["6-"]},
    "GRADATVS": {"2": ["14"]},
    "CONSESSOR": {"-": ["13"]},
    "DVELLICVS": {"-": ["21"]},
    "CALAVTICA": {"-": ["11"]},
    "PERCA": {"-": ["11"]},
    "SVBRVMPO": {"-": ["53"]},
    "LACO": {"N": ["13"]},
    "LICITATOR": {"-": ["13"]},
    "DIABATHRARIVS": {"-": ["12"]},
    "IMPERIVRATVS": {"-": ["21"]},
    "STAMINATVS": {"-": ["21"]},
    "MASTRVCATVS": {"-": ["21"]},
    "PALLENE": {"N": ["11"]},
    "MVLTIGENERIS": {"-": ["24"]},
    "CYANE": {"N": ["11"]},
    "TABVLARIS": {"-": ["24"]},
    "INCONSVMPTVS": {"-": ["21"]},
    "PAPAVEREVS": {"-": ["21"]},
    "TRICLINIA": {"-": ["11"]},
    "AMICVLA": {"-": ["11"]},
    "SVFFARCINO": {"-": ["51"]},
    "IANITRIX": {"-": ["13"]},
    "TROS": {"N": ["12"], "A": ["25"]},
    "INHALO": {"-": ["51"]},
    "EXAMVSSIM": {"-": ["6-"]},
    "OPIPARVS": {"-": ["21"]},
    "ALCEDONIA": {"-": ["12"]},
    "DENTILEGVS": {"-": ["21"]},
    "MENDICOR": {"-": ["5J"]},
    "COMBARDVS": {"-": ["21"]},
    "INVRBANVS": {"-": ["21"]},
    "ERVDITE": {"-": ["6-"]},
    "PYTHO": {"N": ["13"]},
    "CATASTROPHA": {"-": ["11"]},
    "ERYX": {"N": ["13"]},
    "BOLVS": {"-": ["12"]},
    "PRAEDIATOR": {"-": ["13"]},
    "NEVTRVBI": {"-": ["6-"]},
    "TRANSADIGO": {"-": ["53"]},
    "PARIS": {"N": ["13"]},
    "INSIGNITE": {"-": ["6-"]},
    "CROCIO": {"-": ["53"]},
    "PRAEMERCOR": {"-": ["5J"]},
    "PERBLANDVS": {"-": ["21"]},
    "CASSIOPE": {"N": ["11"]},
    "DATATIM": {"-": ["6-"]},
    "APOSPHRAGISMA": {"-": ["13"]},
    "SIDON": {"N": ["13"]},
    "INGVRGITO": {"-": ["51"]},
    "INORATVS": {"-": ["21"]},
    "INCEPTO": {"-": ["51"]},
    "NITIDE": {"-": ["6-"]},
    "VOLSELLA": {"-": ["11"]},
    "COMMARITVS": {"-": ["12"]},
    "HYPSIPYLE": {"N": ["11"]},
    "OCCATOR": {"-": ["13"]},
    "SCIRPEVS": {"-": ["21"]},
    "LEVCOTHOE": {"N": ["11"]},
    "CARDVELIS": {"-": ["13"]},
    "SVBHORRIDVS": {"-": ["21"]},
    "PINSITO": {"-": ["51"]},
    "BOIA": {"-": ["11"]},
    "DOMEFACTVS": {"-": ["21"]},
    "TRADVX": {"-": ["13"]},
    "CONSOBRINA": {"-": ["11"]},
    "AENEOLVS": {"-": ["21"]},
    "MATRICIDA": {"-": ["11"]},
    "CONDIGNE": {"-": ["6-"]},
    "MASO": {"N": ["13"]},
    "ALEATORIVS": {"-": ["12"]},
    "HOSTIFICVS": {"-": ["21"]},
    "COMMINISCOR": {"-": ["5L"]},
    "HERBESCO": {"-": ["53"]},
    "IMMISSIO": {"-": ["13"]},
    "PRAELVSIO": {"-": ["13"]},
    "CAERVLVS": {"-": ["21"]},
    "CYPASSIS": {"N": ["13"]},
    "COMMEMINI": {"-": ["53"]},
    "SVCCEDANEVS": {"-": ["12"]},
    "PORCINARIVS": {"-": ["12"]},
    "TRANSVERSARIVS": {"-": ["21"]},
    "INCLAMITO": {"-": ["51"]},
    "IOLE": {"N": ["11"]},
    "ANTECESSIO": {"-": ["13"]},
    "CONCASTIGO": {"-": ["51"]},
    "RAVIS": {"-": ["13"]},
    "NOTATVS": {"-": ["21"]},
    "NARBO": {"N": ["13"]},
    "PERGRAECOR": {"-": ["5J"]},
    "FENERATO": {"-": ["6-"]},
    "PERPROPINQVVS": {"-": ["21"]},
    "HELIOCAMINVS": {"-": ["12"]},
    "NYCTIMENE": {"N": ["11"]},
    "SALPA": {"-": ["11"]},
    "GESTICVLOR": {"-": ["5J"]},
    "SALSVRA": {"-": ["11"]},
    "FRVMENTATOR": {"-": ["51"]},
    "EVADNE": {"N": ["11"]},
    "DELOS": {"N": ["12"]},
    "EXCIDIO": {"-": ["13"]},
    "INTERREX": {"-": ["13"]},
    "PERLONGVS": {"-": ["21"]},
    "SVBAVRATVS": {"-": ["21"]},
    "QVARTVSDECIMVS": {"-": ["21"]},
    "OBDORMISCO": {"-": ["53"]},
    "CONGEMINO": {"-": ["51"]},
    "BELLATORIVS": {"-": ["21"]},
    "FEBRVA": {"-": ["12"]},
    "GLIS": {"-": ["13"]},
    "PRODICO": {"2": ["53"]},
    "CALIDE": {"-": ["6-"]},
    "CONTENTIOSVS": {"-": ["21"]},
    "SCVTIGERVLVS": {"-": ["12"]},
    "RIVVLVS": {"-": ["12"]},
    "CONGRESSIO": {"-": ["13"]},
    "POESIS": {"-": ["13"]},
    "DICACVLVS": {"-": ["21"]},
    "TERTIADECIMANI": {"-": ["12"]},
    "SEPTVENNIS": {"-": ["24"]},
    "ORPHVS": {"-": ["12"]},
    "SPVRCIDICVS": {"-": ["21"]},
    "CROTALIA": {"-": ["12"]},
    "DVCENTIENS": {"-": ["6-"]},
    "PERVIVO": {"-": ["53"]},
    "VEIENTO": {"N": ["13"]},
    "FICTE": {"-": ["6-"]},
    "VSVFACIO": {"-": ["55"]},
    "SVPERVENTVS": {"-": ["14"]},
    "BADIZO": {"-": ["51"]},
    "CONGRVENTIA": {"-": ["13"]},
    "CHALCIS": {"N": ["13"]},
    "EDORMISCO": {"-": ["53"]},
    "MYS": {"N": ["13"]},
    "FRVSINO": {"N": ["13"]},
    "EXPVDORATVS": {"-": ["21"]},
    "CONFIRMATOR": {"-": ["13"]},
    "BETO": {"-": ["53"]},
    "DICTAMNVS": {"-": ["12"]},
    "TVRILEGVS": {"-": ["21"]},
    "VFENS": {"N": ["13"]},
    "DROMO": {"N": ["13"]},
    "HALOSIS": {"N": ["13"]},
    "VRANIE": {"N": ["17"]},
    "TYPHON": {"N": ["13"]},
    "RAPO": {"N": ["13"]},
    "INACHIS": {"N": ["13"]},
    "SENECIO": {"N": ["13"]},
    "LVCRO": {"N": ["11"]},
    "TRIO": {"N": ["13"]},
    "CEYX": {"N": ["13"]},
    "LVRCO": {"N": ["13"]},
    "PHALERIS": {"N": ["13"]},
    "SACRATOR": {"N": ["13"]},
    "BETIS": {"N": ["13"]},
    "LEVCASPIS": {"N": ["13"]},
    "ACANTHIS": {"N": ["13"]},
    "AEGOCEROS": {"N": ["17"]},
    "PELLIO": {"N": ["13"]},
    "MINIO": {"N": ["11"]},
    "SVAVISAVIATIO": {"N": ["13"]},
    "SILO": {"N": ["13"]},
    "POLLIS": {"N": ["13"]},
    "INO": {"N": ["13"]},
    "PVLLO": {"N": ["11"]},
    "MEFITIS": {"N": ["13"]},
    "PROSELENOS": {"N": ["17"]},
    "ARGITIS": {"N": ["13"]},
    "VESPERVGO": {"N": ["13"]},
    "VESPILLO": {"N": ["13"]},
    "DAPHNE": {"N": ["17"]},
    "METAMORPHOSIS": {"N": ["13"]},
    "MVTTO": {"N": ["13"]},
    "BVRDO": {"N": ["13"]},
    "LATOIS": {"N": ["16"]},
    "DAVNIAS": {"A": ["25"]},
    "ASOPIS": {"N": ["16"]},
    "DODONIS": {"N": ["16"]},
    "OMPHIS": {"N": ["16"]},
    "LARINAS": {"A": ["25"]},
    "ILIBERRIS": {"N": ["16"]},
    "BACCHIS": {"N": ["16"]},
    "AMARYLLIS": {"N": ["16"]},
    "PITTHEIS": {"N": ["16"]},
    "ZIOBETIS": {"N": ["16"]},
    "TOMYRIS": {"N": ["16"]},
    "THYRSIS": {"N": ["16"]},
    "GARAMAS": {"A": ["25"]},
    "STRATONICIS": {"N": ["16"]},
    "LYCORIS": {"N": ["16"]},
    "IGVVINAS": {"A": ["25"]},
    "BARDYLIS": {"N": ["16"]},
    "SABIS": {"N": ["16"]},
    "POTNIAS": {"A": ["25"]},
    "CADMEIS": {"N": ["16"]},
    "ABARIS": {"N": ["16"]},
    "SAGARITIS": {"N": ["16"]},
    "HIAROTIS": {"N": ["16"]},
    "ATREBAS": {"A": ["25"]},
    "ALETRINAS": {"A": ["25"]},
    "EPICHARIS": {"N": ["16"]},
    "VENAFER": {"A": ["23"]},
    "AQVINAS": {"A": ["25"]},
    "PHEGIS": {"N": ["16"]},
    "TRINACRIS": {"N": ["16"]},
    "AMAZONIS": {"N": ["16"]},
    "ATLANTIS": {"N": ["16"]},
    "CEPHEIS": {"N": ["16"]},
    "LIMNATIS": {"N": ["16"]},
    "CERIALIS": {"N": ["16"]},
    "BASSARIS": {"N": ["16"]},
    "BOEBEIS": {"N": ["16"]},
    "ACHAIAS": {"A": ["25"]},
    "INSVBER": {"A": ["23"]},
    "VADIMONIS": {"N": ["16"]},
    "PALLANTIS": {"N": ["16"]},
    "ERECHTHIS": {"N": ["16"]},
    "AGANIPPIS": {"N": ["16"]},
    "CISSEIS": {"N": ["16"]},
    "RHESCVPORIS": {"N": ["16"]},
    "TAENARIS": {"N": ["16"]},
    "THERMVTIS": {"N": ["16"]},
    "BYBLIS": {"N": ["16"]},
    "SIDONIS": {"N": ["16"]},
    "LOCRIS": {"N": ["16"]},
    "NISIAS": {"A": ["25"]},
    "THYNIAS": {"A": ["25"]},
    "SESOSIS": {"N": ["16"]},
    "CARTHASIS": {"N": ["16"]},
    "POMPEIOPOLIS": {"N": ["16"]},
    "DANAIS": {"N": ["16"]},
    "BELIS": {"N": ["16"]},
    "STELLAS": {"A": ["25"]},
    "OSIRIS": {"N": ["16"]},
    "ATRACIS": {"N": ["16"]},
    "PANNYCHIS": {"N": ["16"]},
    "SCYRIAS": {"A": ["25"]},
    "SALMONIS": {"N": ["16"]},
    "THEBAIS": {"N": ["16"]},
    "AETOLIS": {"N": ["16"]},
    "AZANIS": {"N": ["16"]},
    "ATHESIS": {"N": ["16"]},
    "OENEIS": {"N": ["16"]},
    "ACTIAS": {"A": ["25"]},
    "SEBETHIS": {"N": ["16"]},
    "ARARIS": {"N": ["16"]},
    "THYMBRIS": {"N": ["16"]},
    "TRITONIDIS": {"N": ["16"]},
    "CANTABER": {"A": ["23"]},
    "CYTAEIS": {"N": ["16"]},
    "PASITIGRIS": {"N": ["16"]},
    "THALESTRIS": {"N": ["16"]},
    "VRBINAS": {"A": ["25"]},
    "TAMESIS": {"N": ["16"]},
    "ICARIS": {"N": ["16"]},
    "SOTERIS": {"N": ["16"]},
    "MANAPIS": {"N": ["16"]},
    "VEIOVIS": {"N": ["16"]},
    "DENTHALIAS": {"A": ["25"]},
    "SISIGAMBIS": {"N": ["16"]},
    "CALABER": {"A": ["23"]},
    "LAIS": {"N": ["16"]},
    "LYRNESSIS": {"N": ["16"]},
    "RAVENNAS": {"A": ["25"]},
    "CANINEFAS": {"A": ["25"]},
    "ICARIOTIS": {"N": ["16"]},
    "EVRYTIS": {"N": ["16"]},
    "APOLLONIS": {"N": ["16"]},
    "DICARCHIS": {"N": ["16"]},
    "NYSIAS": {"A": ["25"]},
    "NAIS": {"N": ["16"]},
    "CRAVGIS": {"N": ["16"]},
    "INACHIOTIS": {"N": ["16"]},
    "SINIS": {"N": ["16"]},
    "SCHOENEIS": {"N": ["16"]},
    "SICORIS": {"N": ["16"]},
    "ARCHEPOLIS": {"N": ["16"]},
    "MYCENIS": {"N": ["16"]},
    "AORNIS": {"N": ["16"]},
    "ANTIAS": {"A": ["25"]},
    "BOCCHORIS": {"N": ["16"]},
    "HAEMONIS": {"N": ["16"]},
    "ARDEAS": {"A": ["25"]},
    "PRIVERNAS": {"A": ["25"]},
    "CHRYSEIS": {"N": ["16"]},
    "ALEXIS": {"N": ["16"]},
    "CYROPOLIS": {"N": ["16"]},
    "IBER": {"A": ["23"]},
    "MENECRATIS": {"N": ["16"]},
    "DAMALIS": {"N": ["16"]},
    "CELENDRIS": {"N": ["16"]},
    "VCVBIS": {"N": ["16"]},
    "DINIS": {"N": ["16"]},
    "CLEOPHIS": {"N": ["16"]},
    "SIMOIS": {"N": ["16"]},
    "SVGAMBER": {"A": ["23"]},
    "DITIS": {"N": ["16"]},
    "TRVCCVLENSIS": {"N": ["16"]},
    "BITHYNIS": {"N": ["16"]},
    "PHARIS": {"N": ["16"]},
    "PTOLEMAIS": {"N": ["16"]},
    "TANAITIS": {"N": ["16"]},
    "AGONIS": {"N": ["16"]},
    "MENNIS": {"N": ["16"]},
    "SAMIRAMIS": {"N": ["16"]},
    "OPIS": {"N": ["16"]},
    "PELOPEIAS": {"A": ["25"]},
    "PIMPLEIS": {"N": ["16"]},
    "ATINAS": {"A": ["25"], "N": ["13"]},
    "PHORONIS": {"N": ["16"]},
    "THESPIS": {"N": ["16"]},
    "NYCTEIS": {"N": ["16"]},
    "LEVCIPPIS": {"N": ["16"]},
    "EDONIS": {"N": ["16"]},
    "CELTIBER": {"A": ["23"]},
    "ANVBIS": {"N": ["16"]},
    "PELASGIS": {"N": ["16"]},
    "AMANICAE": {"A": ["25"]},
    "BACENIS": {"N": ["16"]},
    "LATRIS": {"N": ["16"]},
    "PEDASIS": {"N": ["16"]},
    "IASIS": {"N": ["16"]},
    "SITHON": {"A": ["25"]},
    "TELESTIS": {"N": ["16"]},
    "CAPPADOX": {"A": ["25"]},
    "HYPANIS": {"N": ["16"]},
    "TANTALIS": {"N": ["16"]},
    "CASINAS": {"A": ["25"]},
    "DYSPARIS": {"N": ["16"]},
    "ALBIS": {"N": ["16"]},
    "TRALLIS": {"N": ["16"]},
    "CAR": {"A": ["25"]},
    "ACARNAN": {"A": ["25"]},
    "CLANIS": {"N": ["16"]},
    "TVNGER": {"A": ["23"]},
    "CILIX": {"A": ["25"]},
    "LYCAONIS": {"N": ["16"]},
    "CANNINEFAS": {"A": ["25"]},
    "CRES": {"A": ["25"]},
    "SCYDROTHEMIS": {"N": ["16"]},
    "FVLGINAS": {"A": ["25"]},
    "THESTYLIS": {"N": ["16"]},
    "ASIS": {"N": ["16"]},
    "OEBALIS": {"N": ["16"]},
    "CILISSA": {"A": ["21"]},
    "DAVLIAS": {"A": ["25"]},
    "BRISEIS": {"N": ["16"]},
    "RHACOTIS": {"N": ["16"]},
    "NYSIGENA": {"A": ["21"]},
    "PELASGIAS": {"A": ["25"]},
    "THESEIS": {"N": ["16"]},
    "CALIS": {"N": ["16"]},
    "VISVRGIS": {"N": ["16"]},
    "CONVICTOLITAVIS": {"N": ["16"]},
    "SCYTHES": {"A": ["25"]},
    "SVDANS": {"A": ["25"]},
    "ATHAMANTIS": {"N": ["16"]},
    "DVBIS": {"N": ["16"]},
    "LIBYS": {"A": ["25"]},
    "PRIAMEIS": {"N": ["16"]},
    "RHASCYPOLIS": {"N": ["16"]},
    "CAERES": {"A": ["25"]},
    "PICENS": {"A": ["25"], "N": ["13"]},
    "TIBVRS": {"A": ["25"]},
    "EVPOLIS": {"N": ["16"]},
    "SPALIS": {"N": ["16"]},
    "OCEANITIS": {"N": ["16"]},
    "TVRESIS": {"N": ["16"]},
    "THRAEX": {"A": ["25"], "N": ["13"]},
    "AETHIOPS": {"A": ["25"]},
    "NISEIS": {"N": ["16"]},
    "PHORCIS": {"N": ["16"]},
    "PIERIS": {"N": ["16"]},
    "PHAEAX": {"A": ["25"]},
    "MEMPHITES": {"A": ["25"]},
    "ELEATES": {"A": ["25"]},
    "ACROPOLISTIS": {"N": ["16"]},
    "IAPYS": {"A": ["25"]},
    "CAMERS": {"A": ["25"], "N": ["13"]},
    "THESPIAS": {"A": ["25"]},
    "ALPIS": {"N": ["16"]},
    "NICOPOLIS": {"N": ["16"]},
    "STRYMONIS": {"N": ["16"]},
    "PHYLLODOCE": {"N": ["17"]},
    "BVBACENE": {"N": ["17"]},
    "BIBRACTE": {"N": ["17"]},
    "HELENE": {"N": ["17"]},
    "EVANTHE": {"N": ["17"]},
    "BARINE": {"N": ["17"]},
    "PHIDYLE": {"N": ["17"]},
    "TRICRENE": {"N": ["17"]},
    "ARTACIE": {"N": ["17"]},
    "PROCNE": {"N": ["17"]},
    "CHLOE": {"N": ["17"]},
    "PSYCHE": {"N": ["17"]},
    "ALCIPPE": {"N": ["17"]},
    "ACME": {"N": ["17"]},
    "ANDROMEDE": {"N": ["17"]},
    "NEPTVNINE": {"N": ["17"]},
    "LEVCOPHRYNE": {"N": ["17"]},
    "CEYCE": {"N": ["17"]},
    "ALSIENSE": {"N": ["17"]},
    "CAMERE": {"N": ["17"]},
    "CHLIDE": {"N": ["17"]},
    "SOPHENE": {"N": ["17"]},
    "EVTERPE": {"N": ["17"]},
    "COELE": {"N": ["17"]},
    "THEBE": {"N": ["17"]},
    "VSPE": {"N": ["17"]},
    "HELLANICE": {"N": ["17"]},
    "CARENE": {"N": ["17"]},
    "ALCMENE": {"N": ["17"]},
    "RHODE": {"N": ["17"]},
    "PHIALE": {"N": ["17"]},
    "EPHYRE": {"N": ["17"]},
    "SITTACENE": {"N": ["17"]},
    "SORACTE": {"N": ["17"]},
    "CYMODOCE": {"N": ["17"]},
    "AMYCLE": {"N": ["17"]},
    "MYRTALE": {"N": ["17"]},
    "LIPARE": {"N": ["17"]},
    "DIDYME": {"N": ["17"]},
    "ORTYGIE": {"N": ["17"]},
    "HARPALYCE": {"N": ["17"]},
    "LIBYE": {"N": ["17"]},
    "CYBELE": {"N": ["17"]},
    "HEBE": {"N": ["17"]},
    "DEIPHOBE": {"N": ["17"]},
    "NIOBE": {"N": ["17"]},
    "ILIONE": {"N": ["17"]},
    "EVROPE": {"N": ["17"]},
    "ATALANTE": {"N": ["17"]},
    "NEOBVLE": {"N": ["17"]},
    "LEVCONOE": {"N": ["17"]},
    "ROXANE": {"N": ["17"]},
    "ISCHOMACHE": {"N": ["17"]},
    "BARSINE": {"N": ["17"]},
    "CALYBE": {"N": ["17"]},
    "MOTHONE": {"N": ["17"]},
    "NISAEE": {"N": ["17"]},
    "IDE": {"N": ["17"]},
    "CALLIRHOE": {"N": ["17"]},
    "MELPOMENE": {"N": ["17"]},
    "PARTHIENE": {"N": ["17"]},
    "PARTHENIE": {"N": ["17"]},
    "THVLE": {"N": ["17"]},
    "PEGE": {"N": ["17"]},
    "SCARPHE": {"N": ["17"]},
    "HYPERMESTRE": {"N": ["17"]},
    "DINOMACHE": {"N": ["17"]},
    "HEDYLE": {"N": ["17"]},
    "ARELATE": {"N": ["17"]},
    "ANTIOPE": {"N": ["17"]},
    "NESAEE": {"N": ["17"]},
    "PALLACINE": {"N": ["17"]},
    "PHRYNE": {"N": ["17"]},
    "HERCVLANENSE": {"N": ["17"]},
    "PETALE": {"N": ["17"]},
    "CYBEBE": {"N": ["17"]},
    "HEROPHILE": {"N": ["17"]},
    "AEETINE": {"N": ["17"]},
    "CYTAEINE": {"N": ["17"]},
    "EVPHROSYNE": {"N": ["17"]},
    "ERIPHYLE": {"N": ["17"]},
    "PANTAGIE": {"N": ["17"]},
    "TAYGETE": {"N": ["17"]},
    "ANDROMACHE": {"N": ["17"]},
    "LALAGE": {"N": ["17"]},
    "ENISPE": {"N": ["17"]},
    "ANACTORIE": {"N": ["17"]},
    "LEDE": {"N": ["17"]},
    "THRACE": {"N": ["17"]},
    "CATABATHMOS": {"N": ["17"]},
    "ORSILOS": {"N": ["17"]},
    "ATHOS": {"N": ["17"]},
    "ILIOS": {"N": ["17"]},
    "CROCOS": {"N": ["17"]},
    "PHILEROS": {"N": ["17"]},
    "ISSOS": {"N": ["17"]},
    "SANBVLOS": {"N": ["17"]},
    "PEPARETHOS": {"N": ["17"]},
    "ARGOS": {"N": ["17"]},
    "SPERCHIOS": {"N": ["17"]},
    "PALAETYROS": {"N": ["17"]},
    "EVXENIPPOS": {"N": ["17"]},
    "OROPS": {"N": ["13"]},
    "PYRGO": {"N": ["13"]},
    "CONGRIO": {"N": ["13"]},
    "ALCANOR": {"N": ["13"]},
    "ARISTOGITON": {"N": ["13"]},
    "MACHAERIO": {"N": ["13"]},
    "ARGO": {"N": ["13"]},
    "CLOTHO": {"N": ["13"]},
    "EPOREDORIX": {"N": ["13"]},
    "ASPRENAS": {"N": ["13"]},
    "METRONAX": {"N": ["13"]},
    "HYLAX": {"N": ["13"]},
    "NVMITOR": {"N": ["13"]},
    "CVPAVO": {"N": ["13"]},
    "BRIMO": {"N": ["13"]},
    "THESAVROCHRYSONICOCH.": {"N": ["13"]},
    "RVFIO": {"N": ["13"]},
    "SER": {"N": ["13"]},
    "ASPAR": {"N": ["13"]},
    "CINGETORIX": {"N": ["13"]},
    "VOLVX": {"N": ["13"]},
    "TREVIR": {"N": ["13"]},
    "BOCCAR": {"N": ["13"]},
    "ATHO": {"N": ["13"]},
    "LVGOTORIX": {"N": ["13"]},
    "ARCENS": {"N": ["13"]},
    "AGGAR": {"N": ["13"]},
    "VANGIO": {"N": ["13"]},
    "AMBIORIX": {"N": ["13"]},
    "VOLCENS": {"N": ["13"]},
    "DIVICO": {"N": ["13"]},
    "DRYOPS": {"N": ["13"]},
    "BVRSAVO": {"N": ["13"]},
    "MASTANABAL": {"N": ["13"]},
    "ADHERBAL": {"N": ["13"]},
    "VRSAO": {"N": ["13"]},
    "GOBANNITIO": {"N": ["13"]},
    "HIEROCAESARIEN.": {"N": ["13"]},
    "CRVPTORIX": {"N": ["13"]},
    "DVMNORIX": {"N": ["13"]},
    "TIMO": {"N": ["13"]},
    "SACROVIR": {"N": ["13"]},
    "OLLOVICO": {"N": ["13"]},
    "SENO": {"N": ["13"]},
    "SEGOVAX": {"N": ["13"]},
    "STASANOR": {"N": ["13"]},
    "TAVROENS": {"N": ["13"]},
    "PVLEIO": {"N": ["13"]},
    "SPIO": {"N": ["13"]},
    "ZOPYRIO": {"N": ["13"]},
    "ANIEN": {"N": ["13"]},
    "TAPPO": {"N": ["13"]},
    "SVTHVL": {"N": ["13"]},
    "BRITTO": {"N": ["13"]},
    "VOCCIO": {"N": ["13"]},
    "POLYMESTOR": {"N": ["13"]},
    "CRESCENS": {"N": ["13"]},
    "COTYTO": {"N": ["13"]},
    "XANTHO": {"N": ["13"]},
    "BIANOR": {"N": ["13"]},
    "CARIO": {"N": ["13"]},
    "COTYLO": {"N": ["13"]},
    "DEMONAX": {"N": ["13"]},
    "RVSO": {"N": ["13"]},
    "MALORIX": {"N": ["13"]},
    "MATISCO": {"N": ["13"]},
    "PLATOR": {"N": ["13"]},
    "VIRIDOVIX": {"N": ["13"]},
    "IERVSALEM": {"N": ["13"]},
    "TIPHYS": {"N": ["13"]},
    "AGATHO": {"N": ["13"]},
    "NOBILIOR": {"N": ["13"]},
    "DRYMO": {"N": ["13"]},
    "CALAS": {"N": ["13"]},
    "MVTHVL": {"N": ["13"]},
    "TVISTO": {"N": ["13"]},
    "VERGILIO": {"N": ["13"]},
    "ARAR": {"N": ["13"]},
    "ORGETORIX": {"N": ["13"]},
    "ERINNYS": {"N": ["13"]},
    "VVLCHALO": {"N": ["13"]},
    "DOLOPS": {"N": ["13"]},
    "LATO": {"N": ["13"]},
    "BRINNO": {"N": ["13"]},
    "PITHON": {"N": ["13"]},
    "MACELO": {"N": ["13"]},
    "CAFO": {"N": ["13"]},
    "TVLLIO": {"N": ["13"]},
    "THEANO": {"N": ["13"]},
    "SYPHAX": {"N": ["13"]},
    "VENTIPO": {"N": ["13"]},
    "BAMBALIO": {"N": ["13"]},
    "GRAECVLIO": {"N": ["13"]},
    "PITHO": {"N": ["13"]},
    "VESONTIO": {"N": ["13"]},
    "TYPHO": {"N": ["13"]},
    "MERCELLO": {"N": ["13"]},
    "QVERCENS": {"N": ["13"]},
    "POETOVIO": {"N": ["13"]},
    "LIBAS": {"N": ["13"]},
    "LIGER": {"N": ["13"]},
    "ISION": {"N": ["13"]},
    "LEPISO": {"N": ["13"]},
    "HIEMPSAL": {"N": ["13"]},
    "CYDRO": {"N": ["13"]},
    "VERCINGETORIX": {"N": ["13"]},
    "NASO": {"N": ["13"]},
    "HELENOR": {"N": ["13"]},
    "ACCO": {"N": ["13"]},
    "CARCASO": {"N": ["13"]},
    "BOMILCAR": {"N": ["13"]},
    "BIBRAX": {"N": ["13"]},
    "ATAX": {"N": ["13"]},
    "DABAR": {"N": ["13"]},
    "TRIMALCHIO": {"N": ["13"]},
    "CORDALIO": {"N": ["13"]},
    "GALLIO": {"N": ["13"]},
    "SEMO": {"N": ["13"]},
    "VERTICO": {"N": ["13"]},
    "ASTYR": {"N": ["13"]},
    "FELICIO": {"N": ["13"]},
    "VIRRO": {"N": ["13"]},
    "COCOCOCO": {"-": ["--"]},
    "EXPRETVS": {"-": ["21"]},
    "DISMARITVS": {"-": ["21"]},
    "NONAGENVS": {"-": ["32"]},
    "MILITATVRA": {"-": ["11"]},
    "OPTESTOR": {"-": ["5J"]},
    "EM": {"-": ["90"]},
    "PTHISICVS": {"-": ["21"]},
    "STOICE": {"-": ["6&"]},
    "ODIOSSICVS": {"-": ["21"]},
    "NVBIGENA": {"-": ["27"]},
    "THYIADES": {"-": ["17"]},
    "QVADRIGENTI": {"-": ["31"]},
    "PERMEDITOR": {"-": ["5J"]},
    "LARS": {"-": ["13"]},
    "FIDVCARIVS": {"-": ["21"]},
    "INEXSTINGVO": {"-": ["53"]},
    "ANTECIPIO": {"-": ["55"]},
    "CIRCVMVEHO": {"-": ["53"]},
    "BAT": {"-": ["--"]},
    "EDEPOL": {"-": ["90"]},
    "ABSENTIVVS": {"-": ["21"]},
    "VNDENVS": {"-": ["32"]},
    "EVGEPAE": {"-": ["90"]},
    "DVODENVS": {"-": ["32"]},
    "APAGE": {"-": ["90"]},
    "EHEM": {"-": ["90"]},
    "VAH": {"-": ["90"]},
    "ARRADO": {"-": ["53"]},
    "ECASTOR": {"-": ["90"]},
    "QVINQVAGENVS": {"-": ["32"]},
    "EHO": {"-": ["90"]},
    "SESCENTIENS": {"-": ["33"]},
    "CONCILIOR": {"-": ["5J"]},
    "IMMODVLOR": {"-": ["5J"]},
    "TRICENTIENS": {"-": ["33"]},
    "SEXCENVS": {"-": ["31"]},
    "BALINEVS": {"-": ["21"]},
    "DRACONIGENVS": {"-": ["21"]},
    "INVERENCVNDVS": {"-": ["21"]},
    "_EO": {"2": ["56"]},
    "QVASSILARIA": {"-": ["11"]},
    "SVDES": {"-": ["13"]},
    "BENEDICE": {"-": ["6&"]},
    "QVADRINVS": {"-": ["32"]},
    "QVINGENVS": {"-": ["31"]},
    "PHYRGIO": {"-": ["13"]},
    "AVERTOR": {"-": ["5L"]},
    "ST": {"-": ["--"]},
    "MOSCILLVS": {"-": ["12"]},
    "HEDIA": {"-": ["11"]},
    "VVLCANALIS": {"-": ["24"]},
    "AVCTORATI": {"-": ["12"]},
    "SESCENVS": {"-": ["32"]},
    "VERBERABILIS": {"-": ["24"]},
    "MV": {"-": ["--"]},
    "MA": {"-": ["--"]},
    "VNVSETVICESIMVS": {"-": ["32"]},
    "MILLIENS": {"-": ["33"]},
    "FVI": {"-": ["56"]},
    "ATTATAE": {"-": ["90"]},
    "INCALEFACIO": {"-": ["55"]},
    "OCCLVSVS": {"-": ["21"]},
    "TRECENVS": {"-": ["31"]},
    "PRAEPOLLVO": {"-": ["53"]},
    "REGILLA": {"-": ["11"]},
    "SOTADICVS": {"-": ["21"]},
    "TARRACONENSES": {"-": ["13"]},
    "TRIVENEFICVS": {"-": ["21"]},
    "TRECENTENVS": {"-": ["31"]},
    "SEXAGENVS": {"-": ["32"]},
    "SEPTVAGIENS": {"-": ["33"]},
    "SVMMANIS": {"-": ["24"]},
    "APOCALO": {"-": ["51"]},
    "PRAECOMPONO": {"-": ["53"]},
    "EXIVRO": {"-": ["51"]},
    "THYLACISTA": {"-": ["11"]},
    "INDEMNO": {"-": ["51"]},
    "RVMIFERO": {"-": ["51"]},
    "HAMADRYAS": {"-": ["17"]},
    "VENERIAE": {"-": ["11"]},
    "ACIA": {"-": ["11"]},
    "THIMIAMAE": {"-": ["11"]},
    "OFFERRO": {"-": ["53"]},
    "COMPEDES": {"-": ["25"]},
    "INNOXIE": {"-": ["6&"]},
    "PVNCTATORIOLA": {"-": ["11"]},
    "ILLVNIS": {"-": ["34"]},
    "PLAGIPATIDA": {"-": ["11"]},
    "ANDROGYNOS": {"-": ["17"]},
    "PRASINIANVS": {"-": ["21"]},
    "OCCENDO": {"-": ["53"]},
    "PERNVLA": {"-": ["11"]},
    "EXCELLENTIS": {"-": ["24"]},
    "STRIGOR": {"-": ["13"]},
    "SESQVEOPVS": {"-": ["13"]},
    "CINIPES": {"-": ["13"]},
    "INGVRGITOR": {"-": ["5J"]},
    "CAVNEA": {"-": ["11"]},
    "INEXSPECTO": {"-": ["51"]},
    "LOQVO": {"-": ["53"]},
    "IMMEMORO": {"-": ["51"]},
    "CALVAE": {"-": ["11"]},
    "NESCIOQVID": {"-": ["48"]},
    "NOMVS": {"-": ["12"]},
    "DVETVICESIMVS": {"-": ["32"]},
    "LOGVS": {"-": ["12"]},
    "GRAMA": {"-": ["11"]},
    "CLATHRA": {"-": ["12"]},
    "VETVO": {"-": ["51"]},
    "SANGVNCVLVS": {"-": ["12"]},
    "BISSACIVS": {"-": ["12"]},
    "HIPPODROMOS": {"-": ["17"]},
    "INTERVIAS": {"-": ["26"]},
    "LAMYROS": {"-": ["17"]},
    "CLASSIA": {"-": ["11"]},
    "DIVIDIA": {"-": ["11"]},
    "CHIRAGRICI": {"-": ["12"]},
    "FIVO": {"-": ["53"]},
    "CHRYSOS": {"-": ["17"]},
    "PRAEVT": {"-": ["69"]},
    "THEBAICAE": {"-": ["11"]},
    "PENTHIACVS": {"-": ["12"]},
    "APSIS": {"-": ["17"]},
    "VESTES": {"-": ["13"]},
    "COMMISSVS": {"-": ["21"]},
    "D": {"-": ["--"]},
    "IMPOLLITVS": {"-": ["21"]},
    "VELANDA": {"-": ["12"]},
    "QVAQVAM": {"-": ["68"]},
    "AV": {"-": ["90"]},
    "STERTEIVS": {"-": ["21"]},
    "CAELVS": {"-": ["12"]},
    "VLS": {"-": ["7-"]},
    "POTICIVS": {"-": ["12"]},
    "OPISTOGRAPHVS": {"-": ["21"]},
    "MARTIOLVS": {"-": ["12"]},
    "STOLATA": {"-": ["11"]},
    "ASSVDASCO": {"-": ["53"]},
    "INDECET": {"-": ["52"]},
    "CEDRA": {"-": ["11"]},
    "NONNIHIL": {"-": ["26"]},
    "EVAX": {"-": ["90"]},
    "TAMIAS": {"-": ["13"]},
    "NVMNAM": {"-": ["67"]},
    "LASANVS": {"-": ["12"]},
    "NVMQVO": {"-": ["67"]},
    "NVMQVISNAM": {"-": ["47"]},
    "ADVSQVE": {"2": ["70"]},
    "CANALE": {"-": ["13"]},
    "NESCIOQVOMODO": {"-": ["67"]},
    "PORGO": {"-": ["53"]},
    "SERE": {"-": ["60"]},
    "EXOPINISSO": {"-": ["21"]},
}
mapping = {
    "AB": {"-": {"morpho": ["p---------"], "uri": ["a0001"],},},
    "ABACTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a9982"],},},
    "ABACVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a0006"]}},
    "ABALIENO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0015"]}},
    "ABATOS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "ABAVVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a0026"]}},
    "ABBAS": {"-": {"morpho": ["n-s---mn3-"], "uri": ["41588"]}},
    "ABDICO": {"1": {"morpho": ["v1spia--1-"], "uri": ["a0034"]}},
    "ABDITVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["46918"]}},
    "ABDITVS": {
        "C": {"morpho": ["aps---mn1-"], "uri": ["34180"]},
        "-": {"morpho": ["aps---mn1-"], "uri": ["34180"]},
    },
    "ABDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0038"]}},
    "ABDOMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["a0039"]}},
    "ABDVCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0040"]}},
    "ABEO": {"1": {"morpho": ["v1spia--4-"], "uri": ["a0043"]}},
    "ABERRO": {"2": {"morpho": ["v1spia--1-"], "uri": ["a0046"]}},
    "ABHINC": {"-": {"morpho": ["rp--------"], "uri": ["a0049"]}},
    "ABHORREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a0050"]}},
    "ABHORRIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0051"]}},
    "ABICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["a0053"]}},
    "ABIECTE": {
        "-": {
            "morpho": ["rp--------", "rp--------"],
            "uri": ["44049", "44049"],
        }
    },
    "ABIECTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["43027"]}},
    "ABIEGNEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["51153"]}},
    "ABIEGNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0055"]}},
    "ABIES": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0057"]}},
    "ABIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0063"]}},
    "ABITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a9984"]}},
    "ABIVDICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0067"]}},
    "ABIVNGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0069"]}},
    "ABIVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0073"]}},
    "ABLAQVEO": {"1": {"morpho": ["v1spia--1-"], "uri": ["a0077"]}},
    "ABLEGO": {"1": {"morpho": ["v1spia--1-"], "uri": ["$0149"]}},
    "ABLEVO": {"1": {"morpho": ["v1spia--1-"], "uri": ["a0085"]}},
    "ABLVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0091"]}},
    "ABLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a9091"]}},
    "ABNEGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0101"]}},
    "ABNEPOS": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a0102"]}},
    "ABNOCTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0104"]}},
    "ABNORMIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["a0106"]}},
    "ABNVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0111"]}},
    "ABOLEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a0116"]}},
    "ABOLESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a1752"]}},
    "ABOLITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0117"]}},
    "ABOLLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a0119"]}},
    "ABOMINOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["a0122"]}},
    "ABORIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["a0125"]}},
    "ABORISCOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["a1874"]}},
    "ABORTIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0129"]}},
    "ABORTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a0128"]}},
    "ABRADO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0133"]}},
    "ABRIPIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["a0140"]}},
    "ABROGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0143"]}},
    "ABROTONVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "ABROTONVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["49773"]}},
    "ABRVMPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0145"]}},
    "ABRVPTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["46275"]}},
    "ABRVPTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["54093"]}},
    "ABSCEDO": {"1": {"morpho": ["v1spia--3-"], "uri": ["a0147"]}},
    "ABSCESSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a0160"]}},
    "ABSCIDO": {"2": {"morpho": ["v1spia--3-"], "uri": ["a0149"]}},
    "ABSCINDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0150"]}},
    "ABSCISIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0151"]}},
    "ABSCISVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["29057"]}},
    "ABSCONDITVS": {
        "C": {"morpho": ["aps---mn1-"], "uri": ["57127"]},
        "-": {"morpho": ["aps---mn1-"], "uri": ["57127"]},
    },
    "ABSCONDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0155"]}},
    "ABSCONSVS": {
        "C": {"morpho": ["aps---mn1-"], "uri": ["47827"]},
        "-": {"morpho": ["aps---mn1-"], "uri": ["47827"]},
    },
    "ABSENS": {"-": {"morpho": ["aps---an3i"], "uri": ["97491"]}},
    "ABSENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a0184"]}},
    "ABSILIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["a0165"]}},
    "ABSIMILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["a0166"]}},
    "ABSINTHIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["19167"]}},
    "ABSISTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0169"]}},
    "ABSOLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0178"]}},
    "ABSOLVTE": {"-": {"morpho": ["rp--------"], "uri": ["19170"]}},
    "ABSOLVTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0174"]}},
    "ABSOLVTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["19172"]}},
    "ABSONVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0180"]}},
    "ABSORBEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a0181"]}},
    "ABSQVE": {
        "1": {"morpho": ["rp--------"], "uri": []},
        "2": {"morpho": ["p---------"], "uri": ["a0183"]},
    },
    "ABSTANTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a0186"]}},
    "ABSTEMIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0185"]}},
    "ABSTERGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a0187"]}},
    "ABSTERREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a0188"]}},
    "ABSTINENS": {"-": {"morpho": ["aps---an3i"], "uri": ["97494"]}},
    "ABSTINENTER": {"-": {"morpho": ["rp--------"], "uri": ["54689"]}},
    "ABSTINENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a0190"]}},
    "ABSTINEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a0191"]}},
    "ABSTRAHO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0195"]}},
    "ABSTRVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0196"]}},
    "ABSTRVSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["19184"]}},
    "ABSVM": {"1": {"morpho": ["v1spia--2-"], "uri": ["a0200"]}},
    "ABSVMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0202"]}},
    "ABSVRDE": {"-": {"morpho": ["rp--------"], "uri": ["19187"]}},
    "ABSVRDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0205"]}},
    "ABVNDANS": {
        "1": {"morpho": ["n-s----n3-"], "uri": []},
        "2": {"morpho": ["aps---an3i"], "uri": ["19189"]},
    },
    "ABVNDANTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a0210"]}},
    "ABVNDE": {"-": {"morpho": ["rp--------"], "uri": ["19193"]}},
    "ABVNDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0212"]}},
    "ABVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0213"]}},
    "ABVSQVE": {"-": {"morpho": ["p---------"], "uri": []}},
    "ABVTOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["a0217"]}},
    "AC": {
        "1": {"morpho": ["----------"], "uri": []},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "ACADEMICI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ACADEMICVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "ACALANTHIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0219"]}},
    "ACANTHVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a0225"]}},
    "ACCEDO": {"1": {"morpho": ["v1spia--3-"], "uri": ["a0268"]}},
    "ACCELERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0243"]}},
    "ACCENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0245"]}},
    "ACCENSEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a0247"]}},
    "ACCENSI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["54755"]}},
    "ACCENSVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["a8916"]},
        "2": {"morpho": ["n-s---mn4-"], "uri": ["a0277"]},
    },
    "ACCEPTATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["31591"]}},
    "ACCEPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0256"]}},
    "ACCEPTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0258"]}},
    "ACCEPTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a8917"]}},
    "ACCEPTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["19213"]}},
    "ACCESSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0265"]}},
    "ACCESSITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0266"]}},
    "ACCESSO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "ACCESSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a0239"]}},
    "ACCIDENS": {"-": {"morpho": ["n-s---nn3-"], "uri": ["19218"]}},
    "ACCIDO": {
        "1": {"morpho": ["v1spia--3-"], "uri": ["a0269", "a0268"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["a0269", "a0268"]},
    },
    "ACCIEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a9270"]}},
    "ACCINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0272"]}},
    "ACCIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["a0274"]}},
    "ACCIPIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["a0275"]}},
    "ACCIPITER": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a0276"]}},
    "ACCITO": {"1": {"morpho": ["v1spia--1-"], "uri": ["a0281"]}},
    "ACCITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a9989"]}},
    "ACCLAMATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0282"]}},
    "ACCLAMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0283"]}},
    "ACCLINIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["a0287"]}},
    "ACCLIVIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["a0289"]}},
    "ACCLIVITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0290"]}},
    "ACCLIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0289"]}},
    "ACCOGNOSCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0291"]}},
    "ACCOLA": {"-": {"morpho": ["n-s---cn1-"], "uri": ["a0292"]}},
    "ACCOLO": {"2": {"morpho": ["v1spia--3-"], "uri": ["a0294"]}},
    "ACCOMMODATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0295"]}},
    "ACCOMMODATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["32391"]}},
    "ACCOMMODO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0298"]}},
    "ACCOMMODVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0299"]}},
    "ACCREDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0301"]}},
    "ACCREMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a0302"]}},
    "ACCRESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0303"]}},
    "ACCVBITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0308"]}},
    "ACCVBO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0311"]}},
    "ACCVMBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0314"]}},
    "ACCVMVLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a0316"]}},
    "ACCVMVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0317"]}},
    "ACCVRATE": {"-": {"morpho": ["rp--------"], "uri": ["49145"]}},
    "ACCVRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["19234"]}},
    "ACCVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0320"]}},
    "ACCVRRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0321"]}},
    "ACCVRSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a0364"]}},
    "ACCVSATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0323"]}},
    "ACCVSATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a0325"]}},
    "ACCVSATORIE": {"-": {"morpho": ["rp--------"], "uri": ["41463"]}},
    "ACCVSATORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0326"]}},
    "ACCVSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0329"]}},
    "ACEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a0333"]}},
    "ACER": {
        "1": {"morpho": ["n-s---nn3-"], "uri": ["a0335"]},
        "2": {"morpho": ["aps---mn3-"], "uri": []},
    },
    "ACERBE": {"-": {"morpho": ["rp--------"], "uri": ["19242"]}},
    "ACERBITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0340"]}},
    "ACERBVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["43823"]}},
    "ACERBVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0343"]}},
    "ACERNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0344"]}},
    "ACERRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a0346"]}},
    "ACERSECOMES": {"-": {"morpho": ["n-s---mn1g"], "uri": ["a0347"]}},
    "ACERVATIM": {"-": {"morpho": ["rp--------"], "uri": ["a0350"]}},
    "ACERVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0352"]}},
    "ACERVVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a0353"]}},
    "ACESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0333"]}},
    "ACESTENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "ACETABVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a0356"]}},
    "ACETVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a0360"]}},
    "ACHAEVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "ACHAIA": {"N": {"morpho": ["n-s---fn1-"], "uri": ["101203"]}},
    "ACHAICVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "ACHIVI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ACHIVVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "ACHRADINA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ACIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0377"]}},
    "ACIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["a0379"]}},
    "ACILIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": ["97498"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "ACINA": {
        "1": {"morpho": ["n-s---fn1-"], "uri": ["a0385"]},
        "-": {"morpho": ["n-s---fn1-"], "uri": ["a0385"]},
    },
    "ACINACES": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "ACINVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a0385"]}},
    "ACINVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a0385"]}},
    "ACIPENSER": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a0386"]}},
    "ACOENONOETVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a0393"]}},
    "ACONITVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a0398"]}},
    "ACOSMOS": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "ACQVIESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0406"]}},
    "ACQVIRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0407"]}},
    "ACQVISITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0408"]}},
    "ACRIMONIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a0417"]}},
    "ACRITER": {"-": {"morpho": ["rp--------"], "uri": ["30866"]}},
    "ACRITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0418"]}},
    "ACROAMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["a0422"]}},
    "ACROASIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["a0423"]}},
    "ACROBATICOS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "ACROLITHVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["39093"]}},
    "ACROTERIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a0435"]}},
    "ACTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a0436"]}},
    "ACTIACVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "ACTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0441"]}},
    "ACTITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0445"]}},
    "ACTIVITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a9446"]}},
    "ACTIVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["97502"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["97502"]},
    },
    "ACTIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0447"]}},
    "ACTOR": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["a0448"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "ACTVARIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a0453"]}},
    "ACTVARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["a0453"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["a0453"]},
    },
    "ACTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a9991"]}},
    "ACTVOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0454"]}},
    "ACTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["a9991"]},
        "2": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "ACTVTVM": {"-": {"morpho": ["rp--------"], "uri": ["a0455"]}},
    "ACVLEATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0459"]}},
    "ACVLEVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a0460"]}},
    "ACVMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["a0462"]}},
    "ACVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0464"]}},
    "ACVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["a0468"]},
        "2": {"morpho": ["n-s---nn3-"], "uri": ["a0468"]},
        "3": {"morpho": ["n-s---fn4-"], "uri": ["a0468"]},
    },
    "ACVTE": {"-": {"morpho": ["rp--------"], "uri": ["19282"]}},
    "ACVTVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0476"]}},
    "ACVTVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "ACVTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["19283"]}},
    "AD": {
        "1": {"morpho": ["rp--------"], "uri": []},
        "2": {"morpho": ["p---------"], "uri": ["a0479"]},
    },
    "ADACTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a9993"]}},
    "ADAEQVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0483"]}},
    "ADAGGERO": {"1": {"morpho": ["v1spia--1-"], "uri": ["a0487"]}},
    "ADAMANTINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0494"]}},
    "ADAMAS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["19289"]}},
    "ADAMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0500"]}},
    "ADAPERIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["a0502"]}},
    "ADAREO": {"-": {"morpho": ["v1spia--2-"], "uri": []}},
    "ADAVCTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a0477"]}},
    "ADAVGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a0512"]}},
    "ADAVGESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0513"]}},
    "ADAVGMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["a0514"]}},
    "ADBIBO": {"2": {"morpho": ["v1spia--3-"], "uri": ["a0516"]}},
    "ADDECET": {"-": {"morpho": ["v3spia--2-"], "uri": ["a0521"]}},
    "ADDICO": {"2": {"morpho": ["v1spia--3-"], "uri": ["a0527"]}},
    "ADDICTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0528"]}},
    "ADDISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0530"]}},
    "ADDITAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a0531"]}},
    "ADDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0535"]}},
    "ADDOCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a0536"]}},
    "ADDVBITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0541"]}},
    "ADDVCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0542"]}},
    "ADDVCTE": {"-": {"morpho": ["rp--------"], "uri": ["37971"]}},
    "ADDVCTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["30018"]}},
    "ADEDO": {"2": {"morpho": ["v1spia--3-"], "uri": ["a0546"]}},
    "ADEMO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "ADEMPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0550"]}},
    "ADEO": {
        "1": {"morpho": ["v1spia--2-"], "uri": ["a0553"]},
        "2": {"morpho": ["rp--------"], "uri": ["a0552"]},
    },
    "ADEPS": {"-": {"morpho": ["n-s---mn3i"], "uri": []}},
    "ADEQVITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0556"]}},
    "ADHAEREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a0561"]}},
    "ADHAERESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a1878"]}},
    "ADHAESVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a9994"]}},
    "ADHIBEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a0565"]}},
    "ADHINNIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["a0567"]}},
    "ADHORTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0569"]}},
    "ADHORTATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a0571"]}},
    "ADHORTOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["a0573"]}},
    "ADHVC": {"-": {"morpho": ["rp--------"], "uri": ["a0575"]}},
    "ADIACEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a0577"]}},
    "ADICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["a0586"]}},
    "ADICO": {"2": {"morpho": ["v1spia--3-"], "uri": []}},
    "ADIECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0584"]}},
    "ADIECTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a9996"]}},
    "ADIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0585"]}},
    "ADIMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0588"]}},
    "ADINDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0593"]}},
    "ADIPATVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a0605"]}},
    "ADIPES": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "ADIPISCOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["a0607"]}},
    "ADITIALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["a0611"]}},
    "ADITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a9997"]}},
    "ADIVDICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0617"]}},
    "ADIVMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a0619"]}},
    "ADIVNCTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0620"]}},
    "ADIVNCTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a8919"]}},
    "ADIVNCTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["43482"]}},
    "ADIVNGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0623"]}},
    "ADIVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0628"]}},
    "ADIVTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0630"]}},
    "ADIVTOR": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["a0631"]},
        "2": {"morpho": ["v1spid--1-"], "uri": []},
    },
    "ADIVTORIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a0632"]}},
    "ADIVTRIX": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["a0633"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "ADIVVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0639"]}},
    "ADMATVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0643"]}},
    "ADMETIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["a0647"]}},
    "ADMINICVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a0857"]}},
    "ADMINISTER": {"-": {"morpho": ["n-s---mn2r"], "uri": ["a0655"]}},
    "ADMINISTRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a0655"]}},
    "ADMINISTRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0657"]}},
    "ADMINISTRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0662"]}},
    "ADMIRABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["a0663"]}},
    "ADMIRANDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["19402"]}},
    "ADMIRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0665"]}},
    "ADMIRATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a0667"]}},
    "ADMIROR": {"-": {"morpho": ["v1spid--1-"], "uri": ["a0668"]}},
    "ADMISCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a0669"]}},
    "ADMISSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0671"]}},
    "ADMISSVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a9998"]}},
    "ADMITTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0676"]}},
    "ADMODERATE": {"-": {"morpho": ["rp--------"], "uri": ["a0678"]}},
    "ADMODVM": {"-": {"morpho": ["rp--------"], "uri": ["a0681"]}},
    "ADMONEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a0684"]}},
    "ADMONITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0685"]}},
    "ADMONITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a0687"]}},
    "ADMONITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a0711"]}},
    "ADMORDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a0690"]}},
    "ADMOVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a0692"]}},
    "ADMVGIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["a0693"]}},
    "ADMVRMVRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0696"]}},
    "ADMVRMVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0697"]}},
    "ADOLEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a1875"]}},
    "ADOLESCENS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["97520"]},
        "2": {"morpho": ["aps---an3-"], "uri": ["38568"]},
    },
    "ADOLESCENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["45968"]}},
    "ADOLESCENTVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["42136"]}},
    "ADOLESCO": {
        "1": {"morpho": ["v1spia--3-"], "uri": ["a0727"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["a0727"]},
    },
    "ADOPERIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["a0729"]}},
    "ADOPINOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["a0732"]}},
    "ADOPTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0736"]}},
    "ADOPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0738"]}},
    "ADOPTIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0739"]}},
    "ADOPTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0740"]}},
    "ADOR": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "ADOREA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["41906"]}},
    "ADOREVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2263"]}},
    "ADORIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["a0751"]}},
    "ADORNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0752"]}},
    "ADORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0753"]}},
    "ADVECTICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0762"]}},
    "ADVECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0764"]}},
    "ADVECTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a0765"]}},
    "ADVECTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a8885"]}},
    "ADVEHO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0766"]}},
    "ADVEHOR": {"-": {"morpho": ["v1spid--3-"], "uri": []}},
    "ADVELO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0768"]}},
    "ADVENA": {"-": {"morpho": ["n-s---cn1-"], "uri": ["a0769"]}},
    "ADVENIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["a0772"]}},
    "ADVENTICIVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "ADVENTICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0773"]}},
    "ADVENTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0774"]}},
    "ADVENTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a8886"]}},
    "ADVERRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0781"]}},
    "ADVERSARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["a0964"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["a0749"]},
    },
    "ADVERSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0792"]}},
    "ADVERSOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["a0793"]}},
    "ADVERSVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a2245"]}},
    "ADVERSVS": {
        "1": {"morpho": ["aps---mn1-"], "uri": ["19548"]},
        "2": {"morpho": ["rp--------"], "uri": ["a0795"]},
        "-": {"morpho": ["p---------"], "uri": []},
        "3": {"morpho": ["p---------"], "uri": []},
    },
    "ADVERTO": {"-": {"morpho": ["rp--------"], "uri": []}},
    "ADVESPERASCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["97540"]}},
    "ADVIGILO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0799"]}},
    "ADVIVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0800"]}},
    "ADVLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0802"]}},
    "ADVLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a0803"]}},
    "ADVLATORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0804"]}},
    "ADVLESCENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a9973"]}},
    "ADVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0810"]}},
    "ADVLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132213"]}},
    "ADVLTER": {
        "1": {"morpho": ["n-s---mn2r"], "uri": ["a0812"]},
        "2": {"morpho": ["aps---mn1r"], "uri": ["a0812"]},
    },
    "ADVLTERA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a0812"]}},
    "ADVLTERINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0816"]}},
    "ADVLTERIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a0819"]}},
    "ADVLTERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0820"]}},
    "ADVLTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["37013"]}},
    "ADVMBRATIM": {"-": {"morpho": ["rp--------"], "uri": ["a0821"]}},
    "ADVMBRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a9821"]}},
    "ADVMBRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["19526"]}},
    "ADVMBRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0822"]}},
    "ADVNCVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0826"]}},
    "ADVOCATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0828"]}},
    "ADVOCATVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a0794"]}},
    "ADVOCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0829"]}},
    "ADVOLO": {"2": {"morpho": ["v1spia--1-"], "uri": ["a0832"]}},
    "ADVOLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0833"]}},
    "ADVRGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a0835"]}},
    "ADVRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0836"]}},
    "ADVSTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["19532"]}},
    "ADYTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a0839"]}},
    "AEBVTIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": ["97543"]},
    },
    "AEDES": {"-": {"morpho": ["n-s---fn3i"], "uri": ["a0842"]}},
    "AEDICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a0843"]}},
    "AEDIFICATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0845"]}},
    "AEDIFICATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a0847"]}},
    "AEDIFICIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a0851"]}},
    "AEDIFICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0852"]}},
    "AEDILICIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["a2143"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["a2095"]},
    },
    "AEDILIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["a0855"]}},
    "AEDILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0856"]}},
    "AEDITVENS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["a0859"]}},
    "AEDITVMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a0861"]}},
    "AEDITVVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a0863"]}},
    "AEDON": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0864"]}},
    "AEGAEVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "AEGER": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["aps---mn1r"], "uri": ["a0866"]},
    },
    "AEGINIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "AEGIS": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["a0870"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "AEGRE": {"-": {"morpho": ["rp--------"], "uri": ["19565"]}},
    "AEGREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a0880"]}},
    "AEGRESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0880"]}},
    "AEGRIMONIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a0881"]}},
    "AEGRITOMARVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AEGRITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0885"]}},
    "AEGROR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a0886"]}},
    "AEGROTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0888"]}},
    "AEGROTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0890"]}},
    "AEGROTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0891"]}},
    "AEGRVM": {
        "C": {"morpho": ["n-s----n2-"], "uri": []},
        "-": {"morpho": ["n-s----n2-"], "uri": []},
    },
    "AEGYPTIVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "AEGYPTVS": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "AELINOS": {"-": {"morpho": ["n-s---mn2g"], "uri": ["97548"]}},
    "AELVRVS": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "AEMILIANVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "AEMILIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": ["97549"]},
    },
    "AEMVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a0901"]}},
    "AEMVLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0897"]}},
    "AEMVLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a0898"]}},
    "AEMVLATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a8887"]}},
    "AEMVLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["a0900"]}},
    "AEMVLVS": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": ["a0901"]},
    },
    "AENEAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "AENEATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a0902"]}},
    "AENEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a2271"]}},
    "AENEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2286"]}},
    "AENIGMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["a0905"]}},
    "AENIPES": {"-": {"morpho": ["aps---an3-"], "uri": ["a0908"]}},
    "AENVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "AENVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["32574"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["37635"]},
    },
    "AEQVABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["a0912"]}},
    "AEQVABILITER": {"-": {"morpho": ["rp--------"], "uri": ["19582"]}},
    "AEQVAEVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0914"]}},
    "AEQVALIS": {
        "1": {"morpho": ["n-s----n3-"], "uri": []},
        "2": {"morpho": ["aps---cn3i"], "uri": ["a0915"]},
    },
    "AEQVALITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0916"]}},
    "AEQVALITER": {"-": {"morpho": ["rp--------"], "uri": ["19585"]}},
    "AEQVANIMITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0920"]}},
    "AEQVE": {"-": {"morpho": ["rp--------"], "uri": ["19589"]}},
    "AEQVI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "AEQVICOLI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "AEQVICVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "AEQVILATATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0933"]}},
    "AEQVINOCTIALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["a0942"]}},
    "AEQVINOCTIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a0943"]}},
    "AEQVIPARO": {"-": {"morpho": ["v1spia--1-"], "uri": ["56187"]}},
    "AEQVIPOLLENS": {"-": {"morpho": ["aps---an3i"], "uri": ["a0953"]}},
    "AEQVIPONDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a0954"]}},
    "AEQVITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0957"]}},
    "AEQVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0961"]}},
    "AEQVOR": {
        "-": {"morpho": ["n-s---nn3-"], "uri": ["a0962"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "AEQVOREVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0963"]}},
    "AEQVVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a0868"]}},
    "AEQVVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["a0867"]},
        "N": {"morpho": ["aps---mn1-"], "uri": ["a0867"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["a0867"]},
    },
    "AER": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a0965"]}},
    "AERACEVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "AERARIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a0970"]}},
    "AERARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a0970"]}},
    "AERARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["a0970"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["a0970"]},
    },
    "AERATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0972"]}},
    "AEREVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["a0974"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["a0974"]},
    },
    "AERIPES": {"-": {"morpho": ["aps---an3-"], "uri": ["a0979"]}},
    "AERIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0981"]}},
    "AERO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a0986"]}},
    "AERVCA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a0992"]}},
    "AERVGINOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0993"]}},
    "AERVGO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0995"]}},
    "AERVMNA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a0997"]}},
    "AERVMNABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["a0998"]}},
    "AERVMNOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0999"]}},
    "AES": {"-": {"morpho": ["n-s---nn3-"], "uri": ["a1004"]}},
    "AESCHRIO": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "AESCHYLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AESCVLAPIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AESCVLETVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a1009"]}},
    "AESCVLINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1012"]}},
    "AESCVLVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["a1013"]}},
    "AESERNINVS": {
        "C": {"morpho": ["n-s----n2-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "AESOPEVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "AESTAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1017"]}},
    "AESTIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["a1018"]}},
    "AESTIMATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1020"]}},
    "AESTIMATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a1021"]}},
    "AESTIMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a1025"]}},
    "AESTIVA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["a1030"]}},
    "AESTIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1030"]}},
    "AESTVARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a1032"]}},
    "AESTVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a1034"]}},
    "AESTVOSE": {"-": {"morpho": ["rp--------"], "uri": ["57061"]}},
    "AESTVOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1035"]}},
    "AESTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a1036"]}},
    "AETAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1037"]}},
    "AETERNITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1041"]}},
    "AETERNO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["a1042"]},
        "2": {"morpho": ["rp--------"], "uri": ["19627"]},
    },
    "AETERNVM": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["rp--------"], "uri": ["19628"]},
    },
    "AETERNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1043"]}},
    "AETHER": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a1046"]}},
    "AETHERIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1047"]}},
    "AETHRA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["a1049"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["a1049"]},
    },
    "AETIOLOGIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a1050"]}},
    "AETNA": {
        "1N": {"morpho": ["n-s----n1-"], "uri": []},
        "2N": {"morpho": ["n-s----n1-"], "uri": []},
        "N": {"morpho": ["n-s---fn1-"], "uri": []},
    },
    "AETNAEVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "AETNENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "AETNENSIS": {
        "N": {"morpho": ["aps---cn3i"], "uri": []},
        "A": {"morpho": ["aps---cn3i"], "uri": []},
    },
    "AETOLIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "AETOLICVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "AETOLVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "AEVVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a1057"]}},
    "AEVVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["28365"]}},
    "AFFABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["a1061"]}},
    "AFFABRE": {"-": {"morpho": ["rp--------"], "uri": ["55422"]}},
    "AFFATIM": {"-": {"morpho": ["rp--------"], "uri": ["a3705"]}},
    "AFFATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a8888"]}},
    "AFFECTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1069"]}},
    "AFFECTATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a1070"]}},
    "AFFECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1072"]}},
    "AFFECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a1076"]}},
    "AFFECTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["a1055"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["37417"]},
    },
    "AFFERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a1079"]}},
    "AFFICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["a1081"]}},
    "AFFIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a1084"]}},
    "AFFINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a1087"]}},
    "AFFINIS": {
        "1": {"morpho": ["n-s----n3-"], "uri": []},
        "2": {"morpho": ["aps---cn3i"], "uri": ["a1088"]},
    },
    "AFFINITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1089"]}},
    "AFFIRMATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1090"]}},
    "AFFIRMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a1093"]}},
    "AFFLATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a8889"]}},
    "AFFLICTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1100"]}},
    "AFFLICTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a1101"]}},
    "AFFLICTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["a8891"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["36969"]},
    },
    "AFFLIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a1104"]}},
    "AFFLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a1105"]}},
    "AFFLVENS": {"-": {"morpho": ["aps---an3-"], "uri": ["31487"]}},
    "AFFLVENTER": {"-": {"morpho": ["rp--------"], "uri": ["54931"]}},
    "AFFLVENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a1106"]}},
    "AFFLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a1108"]}},
    "AFFOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["a1112"]}},
    "AFFRICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a1116"]}},
    "AFFVLGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a1118"]}},
    "AFFVNDO": {"2": {"morpho": ["v1spia--3-"], "uri": ["a1119"]}},
    "AFRICA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "AFRICANVS": {
        "C": {"morpho": ["n-s----n2-"], "uri": []},
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "AFRICVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "AGASO": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "AGATHINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AGATHOCLES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "AGE": {"-": {"morpho": ["----------"], "uri": []}},
    "AGEDVM": {"-": {"morpho": ["----------"], "uri": []}},
    "AGELLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a1135"]}},
    "AGENS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["97562"]},
        "2": {"morpho": ["aps---an3i"], "uri": ["102047"]},
    },
    "AGER": {
        "-": {"morpho": ["n-s---mn2r"], "uri": ["a1137"]},
        "N": {"morpho": ["n-s----n2-"], "uri": []},
    },
    "AGGER": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a1145"]}},
    "AGGERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1147"]}},
    "AGGERO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["a1150"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["a1149"]},
    },
    "AGGESTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a8893"]}},
    "AGGLOMERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a1152"]}},
    "AGGLVTINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a1154"]}},
    "AGGRAVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a1157"]}},
    "AGGREDIOR": {"-": {"morpho": ["v1spid--3i"], "uri": ["a1158"]}},
    "AGGREGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a9967"]}},
    "AGILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["a1165"]}},
    "AGILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1166"]}},
    "AGITABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["a1171"]}},
    "AGITATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1172"]}},
    "AGITATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a1173"]}},
    "AGITATVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["a1191"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["19653"]},
    },
    "AGITE": {"-": {"morpho": ["----------"], "uri": []}},
    "AGITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a1176"]}},
    "AGMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["a1179"]}},
    "AGNA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a1207"]}},
    "AGNASCOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["a1183"]}},
    "AGNATVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a1163"]}},
    "AGNINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a1194"]}},
    "AGNINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1194"]}},
    "AGNOSCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a1204"]}},
    "AGNVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a1207"]}},
    "AGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a1208"]}},
    "AGON": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "AGRAMMATVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["a1223"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["a1223"]},
    },
    "AGRARII": {"-": {"morpho": ["n-p---mn2-"], "uri": ["a1226"]}},
    "AGRARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1226"]}},
    "AGRESTIS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["97566"]},
        "2": {"morpho": ["aps---cn3i"], "uri": ["a1230"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "AGRICOLA": {
        "-": {"morpho": ["n-s---mn1-"], "uri": ["a1233"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "AGRIGENTINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "AGRIGENTINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "AGRIGENTVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "AGRIPPA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "AGRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["102102"]}},
    "AGYRINENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "AGYRINENSIS": {
        "N": {"morpho": ["aps---cn3i"], "uri": []},
        "A": {"morpho": ["aps---cn3i"], "uri": []},
    },
    "AGYRIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "AH": {"-": {"morpho": ["----------"], "uri": []}},
    "AIAX": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "AIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["a1250"]}},
    "ALA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a1256"]}},
    "ALACER": {"-": {"morpho": ["aps---mn3-"], "uri": ["a1262"]}},
    "ALACRITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1264"]}},
    "ALACRITER": {"-": {"morpho": ["rp--------"], "uri": ["33446"]}},
    "ALAPA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a1265"]}},
    "ALARII": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "ALARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1268"]}},
    "ALATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1272"]}},
    "ALBA": {
        "1N": {"morpho": ["n-s---fn1-"], "uri": ["a1311"]},
        "2N": {"morpho": ["n-s---fn1-"], "uri": ["a1311"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["a1311"]},
    },
    "ALBANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ALBANVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "ALBARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1277"]}},
    "ALBATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1280"]}},
    "ALBEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a9041"]}},
    "ALBESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a1282"]}},
    "ALBICERIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["a2294"]}},
    "ALBICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a1290"]}},
    "ALBIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1294"]}},
    "ALBINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["a1296"]}},
    "ALBVLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ALBVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1306"]}},
    "ALBVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a1311"]}},
    "ALBVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["a1311"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["a1311"]},
    },
    "ALCES": {"-": {"morpho": ["n-s---fn3i"], "uri": ["30109"]}},
    "ALCYON": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1319"]}},
    "ALCYONEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a2253"]}},
    "ALEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a1324"]}},
    "ALEATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a1327"]}},
    "ALEBRIA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["102142"]}},
    "ALEO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a1331"]}},
    "ALES": {
        "1": {"morpho": ["n-s---cn3-"], "uri": ["a1332"]},
        "2": {"morpho": ["aps---an3-"], "uri": ["19691"]},
        "N": {"morpho": ["n-s---mn1g"], "uri": []},
    },
    "ALESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a1333"]}},
    "ALEXANDER": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "ALEXANDRIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ALGA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a1339"]}},
    "ALGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a1337"]}},
    "ALGIDVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["a1340"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": ["a1340"]},
    },
    "ALGOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a1343"]}},
    "ALGVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a1345"]}},
    "ALIA": {"-": {"morpho": ["rp--------"], "uri": ["44793"]}},
    "ALIAS": {"-": {"morpho": ["rp--------"], "uri": ["19698"]}},
    "ALIBI": {"-": {"morpho": ["rp--------"], "uri": ["a1348"]}},
    "ALICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a1350"]}},
    "ALICVBI": {"-": {"morpho": ["rp--------"], "uri": ["a1354"]}},
    "ALICVNDE": {"-": {"morpho": ["rp--------"], "uri": ["a1355"]}},
    "ALIENATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1356"]}},
    "ALIENIGENA": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "ALIENIGENVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1358"]}},
    "ALIENO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a1362"]}},
    "ALIENVM": {
        "C": {"morpho": ["n-s---nn2-"], "uri": ["a1363"]},
        "-": {"morpho": ["n-s---nn2-"], "uri": ["a1363"]},
    },
    "ALIENVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["a1363"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["a1363"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["a1363"]},
    },
    "ALIGER": {"-": {"morpho": ["aps---mn1r"], "uri": ["a1365"]}},
    "ALIMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a1367"]}},
    "ALIMONIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a1370"]}},
    "ALIO": {"-": {"morpho": ["rp--------"], "uri": ["19709"]}},
    "ALIOQVI": {"-": {"morpho": ["rp--------"], "uri": ["a1371"]}},
    "ALIPES": {"-": {"morpho": ["aps---an3-"], "uri": ["a9956", "a1374"]}},
    "ALIPILVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a1376"]}},
    "ALIPTES": {"-": {"morpho": ["n-s---mn1g"], "uri": ["97577"]}},
    "ALIQVA": {"-": {"morpho": ["rp--------"], "uri": ["45057"]}},
    "ALIQVAM": {"-": {"morpho": ["rp--------"], "uri": ["49199"]}},
    "ALIQVAMDIV": {"-": {"morpho": ["rp--------"], "uri": ["a9205"]}},
    "ALIQVAMMVLTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a9212"]}},
    "ALIQVANDO": {"-": {"morpho": ["rp--------"], "uri": ["a9206"]}},
    "ALIQVANTO": {"-": {"morpho": ["rp--------"], "uri": ["19713"]}},
    "ALIQVANTVLVM": {
        "1": {"morpho": ["----------"], "uri": []},
        "2": {"morpho": ["rp--------"], "uri": ["54977"]},
    },
    "ALIQVANTVM": {
        "1": {"morpho": ["----------"], "uri": []},
        "2": {"morpho": ["rp--------"], "uri": ["19714"]},
    },
    "ALIQVANTVS": {"-": {"morpho": ["----------"], "uri": []}},
    "ALIQVATENVS": {"-": {"morpho": ["rp--------"], "uri": ["a9210"]}},
    "ALIQVIS": {"-": {"morpho": ["----------"], "uri": []}},
    "ALIQVO": {"-": {"morpho": ["rp--------"], "uri": ["29384"]}},
    "ALIQVOT": {"-": {"morpho": ["----------"], "uri": []}},
    "ALIQVOTFARIAM": {"-": {"morpho": ["rp--------"], "uri": ["33662"]}},
    "ALIQVOTIENS": {"-": {"morpho": ["rp--------"], "uri": ["a9211"]}},
    "ALIS": {"-": {"morpho": ["----------"], "uri": []}},
    "ALITER": {"-": {"morpho": ["rp--------"], "uri": ["19715"]}},
    "ALIVBI": {"-": {"morpho": ["rp--------"], "uri": ["a9214"]}},
    "ALIVNDE": {"-": {"morpho": ["rp--------"], "uri": ["a9213"]}},
    "ALIVS": {"-": {"morpho": ["----------"], "uri": []}},
    "ALIVSMODI": {"-": {"morpho": ["rp--------"], "uri": ["a8504"]}},
    "ALLABOR": {"2": {"morpho": ["v1spid--3-"], "uri": ["a1384"]}},
    "ALLABORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a1385"]}},
    "ALLAPSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a1401"]}},
    "ALLAVDABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["a1391"]}},
    "ALLEC": {"-": {"morpho": ["n-s---nn3-"], "uri": ["31101"]}},
    "ALLECTO": {
        "-": {"morpho": ["v1spia--1-"], "uri": ["a1398"]},
        "N": {"morpho": ["n-s----n1-"], "uri": []},
    },
    "ALLEGATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1402"]}},
    "ALLEGO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["a1403"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["a1404"]},
    },
    "ALLEVO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["a1415", "a1378"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["a1415", "a1378"]},
    },
    "ALLEX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a9937"]}},
    "ALLICEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["a1416"]}},
    "ALLICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["a1417"]}},
    "ALLIDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a1418"]}},
    "ALLIGATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1421"]}},
    "ALLIGO": {"2": {"morpho": ["v1spia--1-"], "uri": ["a1424"]}},
    "ALLINO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a1426"]}},
    "ALLIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["57803"]}},
    "ALLOBROGES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ALLOBROX": {
        "N": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "ALLOCVTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1429"]}},
    "ALLOQVIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a1433"]}},
    "ALLOQVOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["a1434"]}},
    "ALLVCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a1437"]}},
    "ALLVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a1441"]}},
    "ALLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a1442"]}},
    "ALMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1454"]}},
    "ALNEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1455"]}},
    "ALNVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["a1456"]}},
    "ALO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a1457"]}},
    "ALOE": {"-": {"morpho": ["n-s---fn1g"], "uri": ["a1459"]}},
    "ALPES": {"N": {"morpho": ["n-p---fn3i"], "uri": []}},
    "ALPHA": {"-": {"morpho": ["n-s---nn--"], "uri": ["a1468"]}},
    "ALSIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1477"]}},
    "ALTAR": {"-": {"morpho": ["n-s---nn3i"], "uri": ["97583"]}},
    "ALTE": {"-": {"morpho": ["rp--------"], "uri": ["19730"]}},
    "ALTER": {"-": {"morpho": ["----------"], "uri": []}},
    "ALTERAS": {"-": {"morpho": ["rp--------"], "uri": ["44525"]}},
    "ALTERCATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1485"]}},
    "ALTERCOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["a1486"]}},
    "ALTERNIS": {"-": {"morpho": ["rp--------"], "uri": ["19738"]}},
    "ALTERNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a1493"]}},
    "ALTERNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1494"]}},
    "ALTERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a1495"]}},
    "ALTERVTER": {"-": {"morpho": ["----------"], "uri": []}},
    "ALTILIS": {
        "1": {"morpho": ["n-s----n3-"], "uri": []},
        "2": {"morpho": ["aps---cn3i"], "uri": ["a1513"]},
    },
    "ALTISONVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1519"]}},
    "ALTITONANS": {"-": {"morpho": ["aps---an3i"], "uri": ["a1498"]}},
    "ALTITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1522"]}},
    "ALTIVOLANS": {"-": {"morpho": ["aps---an3i"], "uri": ["a1525"]}},
    "A": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "ALTRINSECVS": {"-": {"morpho": ["rp--------"], "uri": ["a1529"]}},
    "ALTRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1530"]}},
    "ALTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a8929"]}},
    "ALTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["19744"]}},
    "ALVARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a1534"]}},
    "ALVCINATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1535"]}},
    "ALVEATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1541"]}},
    "ALVEOLATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1542"]}},
    "ALVEOLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a1543"]}},
    "ALVEVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a1544"]}},
    "ALVMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["a1548"]}},
    "ALVMINOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1551"]}},
    "ALVMNA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a1554"]}},
    "ALVMNVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["a1554"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["a1554"]},
    },
    "ALVNTINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ALVNTINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "ALVNTIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "ALVTA": {
        "1": {"morpho": ["n-s----n1-"], "uri": []},
        "2": {"morpho": ["n-s----n1-"], "uri": []},
    },
    "ALVVS": {"-": {"morpho": ["n-s---cn2-"], "uri": ["a1546"]}},
    "AM": {"-": {"morpho": ["p---------"], "uri": ["a1564"]}},
    "AMABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["a1565"]}},
    "AMABILITER": {"-": {"morpho": ["rp--------"], "uri": ["48420"]}},
    "AMANDO": {
        "2": {"morpho": ["v1spia--1-"], "uri": ["a1570"]},
        "-": {"morpho": ["v1spia--1-"], "uri": ["a1570"]},
    },
    "AMANS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["97589"]},
        "2": {"morpho": ["aps---an3i"], "uri": ["19755"]},
    },
    "AMANTER": {"-": {"morpho": ["rp--------"], "uri": ["97590"]}},
    "AMARA": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "AMARACINVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a1573"]}},
    "AMARACVS": {"-": {"morpho": ["n-s---cn2-"], "uri": ["a1574"]}},
    "AMARANTVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a1575"]}},
    "AMARE": {"-": {"morpho": ["rp--------"], "uri": ["43810"]}},
    "AMARITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1580"]}},
    "AMARITIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["a1581"]}},
    "AMARITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1584"]}},
    "AMAROR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a1587"]}},
    "AMARVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1589"]}},
    "AMATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a1596"]}},
    "AMATORIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a1598"]}},
    "AMATORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1598"]}},
    "AMBACTVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a1603"]}},
    "AMBAGES": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1605"]}},
    "AMBARVALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["a1608"]}},
    "AMBEDO": {"2": {"morpho": ["v1spia--3-"], "uri": ["a1612"]}},
    "AMBEO": {"1": {"morpho": ["v1spia--3-"], "uri": []}},
    "AMBIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a1622"]}},
    "AMBIGVE": {"-": {"morpho": ["rp--------"], "uri": ["19768"]}},
    "AMBIGVITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1623"]}},
    "AMBIGVVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a1624"]}},
    "AMBIGVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1624"]}},
    "AMBIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["a1626"]}},
    "AMBITIO": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["a1627"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "AMBITIOSE": {"-": {"morpho": ["rp--------"], "uri": ["56671"]}},
    "AMBITIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1628"]}},
    "AMBITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a8896"]}},
    "AMBO": {"2": {"morpho": ["----------"], "uri": []}},
    "AMBROSIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a2297"]}},
    "AMBROSIVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["a2370"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["97593"]},
    },
    "AMBVBAIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a1640"]}},
    "AMBVLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1644"]}},
    "AMBVLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a1647"]}},
    "AMBVLATORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1648"]}},
    "AMBVLATRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1649"]}},
    "AMBVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a1650"]}},
    "AMBVRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a1653"]}},
    "AMELLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a1657"]}},
    "AMENS": {"-": {"morpho": ["aps---an3i"], "uri": ["a1660"]}},
    "AMENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a1662"]}},
    "AMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["58292"]}},
    "AMERIOLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "AMES": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a1664"]}},
    "AMESTRATINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "AMESTRATINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "AMESTRATVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AMETHYSTINA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["a1666"]}},
    "AMETHYSTVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["a1668"]}},
    "AMICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a0808"]}},
    "AMICE": {"-": {"morpho": ["rp--------"], "uri": ["40694"]}},
    "AMICIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["a1679"]}},
    "AMICITIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a2373"]}},
    "AMICITIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["a2383"]}},
    "AMICTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a8898"]}},
    "AMICVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a1654"]}},
    "AMICVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a1685"]}},
    "AMICVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["a1687"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["a0783"]},
    },
    "AMISSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1689"]}},
    "AMITA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a1690"]}},
    "AMITTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a1723"]}},
    "AMMONIACVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["34825"]}},
    "AMNIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["a1704"]}},
    "AMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a1705"]}},
    "AMOENITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1708"]}},
    "AMOENVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1710"]}},
    "AMOLIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["a1712"]}},
    "AMOMON": {"-": {"morpho": ["n-s---nn2g"], "uri": ["102364"]}},
    "AMOMVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a1715"]}},
    "AMOR": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["a1716"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "AMOVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a1724"]}},
    "AMPHERES": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "AMPHILOCHI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "AMPHIO": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "AMPHIPOLIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "AMPHIPROSTYLOS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "AMPHITHALAMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["102388"]}},
    "AMPHITHEATRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": []}},
    "AMPHORA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a1757"]}},
    "AMPLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a1761"]}},
    "AMPLE": {"-": {"morpho": ["rp--------"], "uri": ["36310"]}},
    "AMPLECTOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["a1762"]}},
    "AMPLEXOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["a1766"]}},
    "AMPLEXVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a1780"]}},
    "AMPLIFICATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1769"]}},
    "AMPLIFICATOR": {"-": {"morpho": ["n-s---fn3-"], "uri": ["43416"]}},
    "AMPLIFICE": {"-": {"morpho": ["rp--------"], "uri": ["19819"]}},
    "AMPLIFICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a1772"]}},
    "AMPLIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a1774"]}},
    "AMPLITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1775"]}},
    "AMPLIVS": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "AMPLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1778"]}},
    "AMPVLLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["a1786"]}},
    "AMPVTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a1792"]}},
    "AMVLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AMVRCA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a1797"]}},
    "AMVRCARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1798"]}},
    "AMVSOS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "AMVSSIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a1803"]}},
    "AMYGDALVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a1806"]}},
    "AMYLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a1813"]}},
    "AMYSTIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1815"]}},
    "AN": {"-": {"morpho": ["rp--------"], "uri": ["45604"]}},
    "ANABATHRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a1821"]}},
    "ANADEMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["a1841"]}},
    "ANALECTA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["a1859"]}},
    "ANALEMMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["a1861"]}},
    "ANALEMPTRIS": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "ANALOGIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a1864"]}},
    "ANAPHORICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a8879"]}},
    "ANATONVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1905"]}},
    "ANCEPS": {
        "1": {"morpho": ["n-s----n3-"], "uri": []},
        "2": {"morpho": ["aps---an3-"], "uri": ["a1912"]},
    },
    "ANCHISES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ANCILE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["a1915"]}},
    "ANCILLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a1916"]}},
    "ANCILLARIOLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a1917"]}},
    "ANCILLARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["a1918"]}},
    "ANCILLVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a1920"]}},
    "ANCISVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1923"]}},
    "ANCON": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["a1926"]},
        "N": {"morpho": ["n-s----n-g"], "uri": []},
    },
    "ANCORA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a1927"]}},
    "ANCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["30758"]}},
    "ANDRO": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "ANDRON": {
        "-": {"morpho": ["n-s----n-g"], "uri": []},
        "N": {"morpho": ["n-s---mn3-"], "uri": []},
    },
    "ANDRONITIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1947"]}},
    "ANDROSTHENES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ANELLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a1955"]}},
    "ANETHVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a1961"]}},
    "ANFRACTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["a1968"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["a1968"]},
    },
    "ANGELLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a1977"]}},
    "ANGELVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a1979"]}},
    "ANGINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a1980"]}},
    "ANGIPORTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a1981"]}},
    "ANGIPORTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a1981"]}},
    "ANGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a1983"]}},
    "ANGOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a1985"]}},
    "ANGVBATAE": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "ANGVICOMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1987"]}},
    "ANGVIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["a1989"]}},
    "ANGVILLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a1991"]}},
    "ANGVIMANVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["44511"]}},
    "ANGVINEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1993"]}},
    "ANGVINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1994"]}},
    "ANGVIS": {
        "-": {"morpho": ["n-s---mn3i"], "uri": ["a1996"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "ANGVLARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["a1997"]}},
    "ANGVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a2004"]}},
    "ANGVSTE": {"-": {"morpho": ["rp--------"], "uri": ["39562"]}},
    "ANGVSTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a2006"]}},
    "ANGVSTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a2009"]}},
    "ANGVSTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a2010"]}},
    "ANGVSTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2010"]}},
    "ANHELITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a2014"]}},
    "ANHELO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a2015"]}},
    "ANHELVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2017"]}},
    "ANIATROLOGETVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2019"]}},
    "ANICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a2022"]}},
    "ANILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["a2024"]}},
    "ANILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a2025"]}},
    "ANILITER": {"-": {"morpho": ["rp--------"], "uri": ["46414"]}},
    "ANIMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a2028"]}},
    "ANIMADVERSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a2030"]}},
    "ANIMADVERTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a1944"]}},
    "ANIMAL": {"-": {"morpho": ["n-s---nn3i"], "uri": ["a2034"]}},
    "ANIMALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["a2034"]}},
    "ANIMANS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["97619"]},
        "2": {"morpho": ["aps---an3-"], "uri": ["42505"]},
    },
    "ANIMATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a2036"]}},
    "ANIMATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["19882"]}},
    "ANIMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a2041"]}},
    "ANIMOSE": {"-": {"morpho": ["rp--------"], "uri": ["19885"]}},
    "ANIMOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2043"]}},
    "ANIMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a2046"]}},
    "ANIO": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "ANISOCYCLI": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "ANISVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["51473"]}},
    "ANNAEA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ANNAEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ANNALIS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["102569"]},
        "2": {"morpho": ["aps---cn3i"], "uri": ["a2050"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "ANNATO": {"-": {"morpho": ["v1spia--1-"], "uri": ["47472"]}},
    "ANNE": {"-": {"morpho": ["rp--------"], "uri": ["97622"]}},
    "ANNECTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["33632"]}},
    "ANNIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ANNIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ANNICVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2054"]}},
    "ANNITOR": {"2": {"morpho": ["v1spid--3-"], "uri": ["132243"]}},
    "ANNIVERSARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2061"]}},
    "ANNIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ANNO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["a7061"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["a7061"]},
    },
    "ANNONA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a2063"]}},
    "ANNOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2067"]}},
    "ANNOTINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2068"]}},
    "ANNOTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["29083"]}},
    "ANNVMERO": {"1": {"morpho": ["v1spia--1-"], "uri": ["43132"]}},
    "ANNVNTIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["44727"]}},
    "ANNVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["29188"]}},
    "ANNVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["a2077"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["a2077"]},
    },
    "ANNVVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a0854"]}},
    "ANNVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2080"]}},
    "ANQVIRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a2090"]}},
    "ANSA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a2092"]}},
    "ANSER": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["a0919"]},
        "N": {"morpho": ["n-s----n2-"], "uri": []},
    },
    "ANTAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["a2102"]}},
    "ANTARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2677"]}},
    "ANTE": {
        "1": {"morpho": ["rp--------"], "uri": ["19899"]},
        "2": {"morpho": ["p---------"], "uri": ["a2109"]},
    },
    "ANTEA": {"-": {"morpho": ["rp--------"], "uri": ["a2110"]}},
    "ANTEACTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a7110"]}},
    "ANTEAGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["48884"]}},
    "ANTEAQVAM": {"-": {"morpho": ["----------"], "uri": []}},
    "ANTECAPIO": {"2": {"morpho": ["v1spia--3i"], "uri": ["a2117"]}},
    "ANTECEDENS": {"-": {"morpho": ["aps---an3i"], "uri": ["19903"]}},
    "ANTECEDENTIA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["102622"]}},
    "ANTECEDO": {"1": {"morpho": ["v1spia--3-"], "uri": ["a2118"]}},
    "ANTECELLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a7118"]}},
    "ANTECESSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a2122"]}},
    "ANTECHON": {"-": {"morpho": ["----------"], "uri": []}},
    "ANTECVRRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a2124"]}},
    "ANTECVRSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a2125"]}},
    "ANTEEO": {"1": {"morpho": ["v1spia--4-"], "uri": ["a1533"]}},
    "ANTEFERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a2128"]}},
    "ANTEFIXVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "ANTEFIXVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2129"]}},
    "ANTEHABEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a2135"]}},
    "ANTEHAC": {"-": {"morpho": ["rp--------"], "uri": ["a2136"]}},
    "ANTELVCANVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0952"]}},
    "ANTEMERIDIANVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2148"]}},
    "ANTEMNATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ANTENNA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["43966"]}},
    "ANTENOR": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ANTEPAGMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a2158"]}},
    "ANTEPONO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a2165"]}},
    "ANTEQVAM": {
        "-": {"morpho": ["----------"], "uri": []},
        "1": {"morpho": ["rp--------"], "uri": ["32023"]},
    },
    "ANTERIS": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "ANTES": {"-": {"morpho": ["n-s---mn3i"], "uri": ["a2172"]}},
    "ANTESIGNANVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["a2175"]},
        "2": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "ANTESTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a2176"]}},
    "ANTESTOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["a2177"]}},
    "ANTEVENIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["a2180"]}},
    "ANTEVERTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a2183"]}},
    "ANTHROPINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ANTIATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ANTIBASIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["a2216"]}},
    "ANTIBOREVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a2217"]}},
    "ANTICIPO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a2225"]}},
    "ANTIDOTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a2229"]}},
    "ANTIGONVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ANTIOCHENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "ANTIOCHIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ANTIOCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ANTIPODES": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a2244"]}},
    "ANTIQVARIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a2247"]}},
    "ANTIQVARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["a2247"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["a2247"]},
    },
    "ANTIQVE": {"-": {"morpho": ["rp--------"], "uri": ["45736"]}},
    "ANTIQVI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["a1063"]}},
    "ANTIQVITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a2250"]}},
    "ANTIQVITVS": {"-": {"morpho": ["rp--------"], "uri": ["a2251"]}},
    "ANTIQVVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a1228"]}},
    "ANTIQVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1056"]}},
    "ANTISTES": {"-": {"morpho": ["n-s---cn3-"], "uri": ["a2264"]}},
    "ANTISTITA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a2265"]}},
    "ANTISTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ANTITHETON": {"-": {"morpho": ["n-s---nn2g"], "uri": ["a2273"]}},
    "ANTIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "ANTONIANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ANTONIANVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "ANTONIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": ["97633"]},
    },
    "ANTRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a2282"]}},
    "ANVLARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2979"]}},
    "ANVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a2290"]}},
    "ANVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["a2292"]},
        "2": {"morpho": ["n-s---fn4-"], "uri": ["a2293"]},
    },
    "ANXIE": {"-": {"morpho": ["rp--------"], "uri": ["19940"]}},
    "ANXIETAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a2296"]}},
    "ANXIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2983"]}},
    "AONIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "APELLES": {"N": {"morpho": ["n-s---mn3i"], "uri": []}},
    "APER": {
        "-": {"morpho": ["n-s---mn2r"], "uri": ["a2320"]},
        "N": {"morpho": ["n-s----n2-"], "uri": []},
    },
    "APERIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["a2322"]}},
    "APERTE": {"-": {"morpho": ["rp--------"], "uri": ["19947"]}},
    "APERTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a2349"]}},
    "APERTVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a2330"]}},
    "APERTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["19951"]}},
    "APEX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a2331"]}},
    "APIACVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2350"]}},
    "APICIVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["a2359"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["97637"]},
    },
    "APIOLAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "APIS": {
        "-": {"morpho": ["n-s---fn3i"], "uri": ["a2369"]},
        "N": {"morpho": ["n-s---mn3g"], "uri": []},
    },
    "APISCOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["a1899"]}},
    "APIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a2371"]}},
    "APLVSTRE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["a2375"]}},
    "APOLLO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "APOLLODORVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "APOLLONIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "APOLLONIATAE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "APOLLONIENSIS": {
        "N": {"morpho": ["aps---cn3i"], "uri": []},
        "A": {"morpho": ["aps---cn3i"], "uri": []},
    },
    "APOLLONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "APOLOGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a2412"]}},
    "APOPHYSIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["a2419"]}},
    "APOTHECA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a2451"]}},
    "APOTHESIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["a2455"]}},
    "APPAGINECVLI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["102861"]}},
    "APPARATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a2463"]}},
    "APPARATVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["a2502"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["19985"]},
    },
    "APPAREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a2468"]}},
    "APPARIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["a2469"]}},
    "APPARITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a2470"]}},
    "APPARITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a2471"]}},
    "APPARO": {"2": {"morpho": ["v1spia--1-"], "uri": ["a2473"]}},
    "APPELLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a7476"]}},
    "APPELLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a2478"]}},
    "APPELLITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a2480"]}},
    "APPELLO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["a2481"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["a2482"]},
    },
    "APPENDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a2485"]}},
    "APPENDIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a2487"]}},
    "APPENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a2488"]}},
    "APPERIOR": {"-": {"morpho": ["v1spid--4-"], "uri": []}},
    "APPETENS": {"-": {"morpho": ["aps---an3i"], "uri": ["97648"]}},
    "APPETITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a2495"]}},
    "APPETITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a8901"]}},
    "APPETO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a2498"]}},
    "APPIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": ["97651"]},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "APPLAVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a2504"]}},
    "APPLICATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a2507"]}},
    "APPLICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a2508"]}},
    "APPLORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a2510"]}},
    "APPONO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a2513"]}},
    "APPORTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a2515"]}},
    "APPORTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a2516"]}},
    "APPOSITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["20007"]}},
    "APPRECOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["a2522"]}},
    "APPREHENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a2523"]}},
    "APPREHENSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a2525"]}},
    "APPRIME": {"-": {"morpho": ["rp--------"], "uri": ["42618"]}},
    "APPRIMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a2528"]}},
    "APPRIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2529"]}},
    "APPROBATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a2532"]}},
    "APPROBO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a2535"]}},
    "APPROPERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a2540"]}},
    "APPROPINQVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a2542"]}},
    "APPVGNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a2547"]}},
    "APPVLSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a8903"]}},
    "APRICVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a2555"]}},
    "APRICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2555"]}},
    "APRONIANVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "APRONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "APTE": {"-": {"morpho": ["rp--------"], "uri": ["49254"]}},
    "APTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a2575"]}},
    "APTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["20021"]}},
    "APVD": {"-": {"morpho": ["p---------"], "uri": ["a2579"]}},
    "APYRENVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a1287"]}},
    "AQVA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a2585"]}},
    "AQVALICVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a2589"]}},
    "AQVALIS": {
        "2": {"morpho": ["aps---cn3i"], "uri": ["a2590"]},
        "1": {"morpho": ["aps---cn3i"], "uri": ["a2590"]},
    },
    "AQVARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a1316"]}},
    "AQVARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["a1309"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["a1289"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["a1309"]},
    },
    "AQVATICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2594"]}},
    "AQVATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a2596"]}},
    "AQVATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a2597"]}},
    "AQVATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2598"]}},
    "AQVIFOLIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1326"]}},
    "AQVILA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["a1531"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["a1531"]},
    },
    "AQVILIFER": {"-": {"morpho": ["n-s---mn2r"], "uri": ["a2609"]}},
    "AQVILIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": ["a2611"]},
    },
    "AQVILO": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["a2612"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "AQVOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["a2622"]}},
    "AQVOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2624"]}},
    "AQVVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["43253"]}},
    "ARA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["a2625"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "ARACHNE": {
        "-": {"morpho": ["n-s---fn1g"], "uri": ["a2630"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": []},
    },
    "ARAEOSTYLOS": {"-": {"morpho": ["aps---mn1g"], "uri": ["a2633"]}},
    "ARANEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a2639"]}},
    "ARANEOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2638"]}},
    "ARANEVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["a2639"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["a2639"]},
    },
    "ARATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a2641"]}},
    "ARATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a2644"]}},
    "ARATRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a2647"]}},
    "ARBITER": {"-": {"morpho": ["n-s---mn2r"], "uri": ["a2649"]}},
    "ARBITRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a2649"]}},
    "ARBITRATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a8904"]}},
    "ARBITRIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a2657"]}},
    "ARBITROR": {"-": {"morpho": ["v1spid--1-"], "uri": ["a2658"]}},
    "ARBOR": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a2659"]}},
    "ARBORARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2661"]}},
    "ARBOREVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2664"]}},
    "ARBVSCVLA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["a2665"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["a2665"]},
    },
    "ARBVSTIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2668"]}},
    "ARBVSTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a2671"]}},
    "ARBVTEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2672"]}},
    "ARBVTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a2673"]}},
    "ARBVTVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["48234"]}},
    "ARCA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a2674"]}},
    "ARCADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ARCADIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ARCADICVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "ARCANO": {"-": {"morpho": ["rp--------"], "uri": ["41000"]}},
    "ARCANVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a1321"]}},
    "ARCANVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2675"]}},
    "ARCATVS": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "ARCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a2683"]}},
    "ARCESSITVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["a2676"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["97663"]},
    },
    "ARCESSO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a2688"]}},
    "ARCHAGATHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARCHETYPVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2696"]}},
    "ARCHIMAGIRVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a2712"]}},
    "ARCHIMEDES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ARCHIPIRATA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["a2717"]}},
    "ARCHITECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a2724"]}},
    "ARCHITECTONICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2726"]}},
    "ARCHITECTOR": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["a2728"]},
        "2": {"morpho": ["v1spid--1-"], "uri": ["132255"]},
        "-": {"morpho": ["v1spid--1-"], "uri": ["132255"]},
    },
    "ARCHITECTVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a2730"]}},
    "ARCHITECTVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a2733"]}},
    "ARCHONIDAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ARCITENENS": {
        "1": {"morpho": ["n-s----n3-"], "uri": []},
        "2": {"morpho": ["aps---an3i"], "uri": ["a2905"]},
    },
    "ARCTOS": {"-": {"morpho": ["n-s---fn2g"], "uri": ["a8930"]}},
    "ARCTOVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["20063"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["20063"]},
    },
    "ARCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a2759"]}},
    "ARCVS": {
        "-": {"morpho": ["n-s---mn4-"], "uri": ["a2764"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "ARDEA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["a2766"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "ARDENS": {"-": {"morpho": ["aps---an3i"], "uri": ["20068"]}},
    "ARDENTER": {"-": {"morpho": ["rp--------"], "uri": ["20069"]}},
    "ARDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a2767"]}},
    "ARDESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a2270"]}},
    "ARDOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a2771"]}},
    "ARDVVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a2774"]}},
    "ARDVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2774"]}},
    "AREA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a2775"]}},
    "AREFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["a3715"]}},
    "ARENA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["47288"]}},
    "ARENARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["34368"]}},
    "ARENATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["46929"]}},
    "ARENATVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["54448"]}},
    "ARENATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["37199"]}},
    "ARENOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["56634"]}},
    "AREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a2778"]}},
    "ARESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a2778"]}},
    "ARETALOGVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a7779"]}},
    "ARETHVSA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "AREVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": ["97670"]},
    },
    "ARGEI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ARGENTARIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a2788"]}},
    "ARGENTARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["a2788"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["a2788"]},
    },
    "ARGENTEVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["a2390"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["a2578"]},
    },
    "ARGENTIFODINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a3713"]}},
    "ARGENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a2795"]}},
    "ARGILETVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "ARGILLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a2797"]}},
    "ARGVMENTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a2802"]}},
    "ARGVMENTOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["a2806"]}},
    "ARGVMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a2808"]}},
    "ARGVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a2809"]}},
    "ARGVTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a2810"]}},
    "ARGVTE": {"-": {"morpho": ["rp--------"], "uri": ["20089"]}},
    "ARGVTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a2813"]}},
    "ARGVTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a2816"]}},
    "ARGVTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["20092"]}},
    "ARICIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ARICINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ARICINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "ARIDA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a1568"]}},
    "ARIDVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2826"]}},
    "ARIDVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a1594"]}},
    "ARIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1540"]}},
    "ARIES": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["a2829"]},
        "N": {"morpho": ["n-s---mn1g"], "uri": []},
    },
    "ARIETARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2830"]}},
    "ARIETO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a2834"]}},
    "ARIMINENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "ARIMINVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "ARISTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a2845"]}},
    "ARISTAEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARISTEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARISTIDES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ARISTODAMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARISTODEMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARISTOTELES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "ARISTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARITHMETICA": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "ARMA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["a2855"]}},
    "ARMAMAXA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a2856"]}},
    "ARMAMENTARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a1618"]}},
    "ARMAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a2857"]}},
    "ARMARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a2860"]}},
    "ARMATI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["97677"]}},
    "ARMATVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a2864"]}},
    "ARMATVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["a2909"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["45112"]},
    },
    "ARMENTARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["a2866"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["a2866"]},
    },
    "ARMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a2870"]}},
    "ARMIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["a2876"]}},
    "ARMIGER": {
        "1": {"morpho": ["n-s---mn2r"], "uri": ["a2877"]},
        "2": {"morpho": ["aps---mn1r"], "uri": ["a2877"]},
    },
    "ARMILLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a2879"]}},
    "ARMIPOTENS": {
        "-": {"morpho": ["aps---an3i"], "uri": ["a2885"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "ARMISONVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2886"]}},
    "ARMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a2888"]}},
    "ARMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a2891"]}},
    "ARO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a2895"]}},
    "ARRECTARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a2911"]}},
    "ARRECTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["34104"]}},
    "ARREPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a2912"]}},
    "ARRIDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a2918"]}},
    "ARRIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a2919"]}},
    "ARRIPIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["a2921"]}},
    "ARRISOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a2923"]}},
    "ARRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARRODO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "ARROGANS": {
        "A": {"morpho": ["aps---an3-"], "uri": ["32325"]},
        "-": {"morpho": ["aps---an3-"], "uri": ["32325"]},
    },
    "ARROGANTER": {"-": {"morpho": ["rp--------"], "uri": ["57933"]}},
    "ARROGANTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a2925"]}},
    "ARROGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a2928"]}},
    "ARROSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a2930"]}},
    "ARRVNS": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "ARS": {
        "-": {"morpho": ["n-s---fn3i"], "uri": ["a2934"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "ARSENICON": {"-": {"morpho": ["n-s---nn2-"], "uri": ["40294"]}},
    "ARSIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ARSINEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a2937"]}},
    "ARSINOE": {"N": {"morpho": ["n-s---fn1g"], "uri": []}},
    "ARTE": {"-": {"morpho": ["rp--------"], "uri": ["43992"]}},
    "ARTEMIDORVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARTEMO": {
        "-": {"morpho": ["n-s----n3-"], "uri": []},
        "N": {"morpho": ["n-s---mn3-"], "uri": []},
    },
    "ARTERIA": {"-": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ARTERIACE": {"-": {"morpho": ["n-s---fn1g"], "uri": ["a2944"]}},
    "ARTICVLARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2952"]}},
    "ARTICVLATIM": {"-": {"morpho": ["rp--------"], "uri": ["a2953"]}},
    "ARTICVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a2956"]}},
    "ARTICVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a2958"]}},
    "ARTIFEX": {"-": {"morpho": ["n-s---cn3-"], "uri": ["a2959"]}},
    "ARTIFICIOSE": {"-": {"morpho": ["rp--------"], "uri": ["52957"]}},
    "ARTIFICIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2963"]}},
    "ARTIFICIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a2964"]}},
    "ARTIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["a2967"]}},
    "ARTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a2971"]}},
    "ARTOCREAS": {"-": {"morpho": ["n-s---nn3-"], "uri": ["a2973"]}},
    "ARTOPTA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["a2975"]}},
    "ARTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a1973"]}},
    "ARTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["a1967"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["a1891"]},
    },
    "ARVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a2992"]}},
    "ARVNDINETVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["97689"]}},
    "ARVNDINOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["97690"]}},
    "ARVNDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["45524"]}},
    "ARVVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a3006"]}},
    "ARX": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["20128"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "AS": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a3009"]}},
    "ASCANIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ASCENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a3021"]}},
    "ASCENSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a3023"]}},
    "ASCENSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a2969"]}},
    "ASCIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a3026"]}},
    "ASCIO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["a2285"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["a2285"]},
    },
    "ASCISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a3028"]}},
    "ASCITVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["55283"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["33308"]},
    },
    "ASCRIBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a3034"]}},
    "ASCRIPTICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a3035"]}},
    "ASCRIPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a3036"]}},
    "ASELLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a3045"]}},
    "ASELLVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["a3045"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["a3045"]},
    },
    "ASIA": {"N": {"morpho": ["n-s---fn1-"], "uri": ["a3049"]}},
    "ASIATICVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": ["97696"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["97696"]},
    },
    "ASILVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a3051"]}},
    "ASINARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["a3055"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["a3055"]},
    },
    "ASININVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a3057"]}},
    "ASINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ASINVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a3052"]}},
    "ASOTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a3063"]}},
    "ASPARAGVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a3065"]}},
    "ASPECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a3068"]}},
    "ASPECTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a3106"]}},
    "ASPENDIVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "ASPENDVS": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "ASPER": {
        "-": {"morpho": ["aps---mn1r"], "uri": ["a3072"]},
        "N": {"morpho": ["n-s----n2-"], "uri": []},
    },
    "ASPERA": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "ASPERE": {"-": {"morpho": ["rp--------"], "uri": ["20152"]}},
    "ASPERGO": {
        "1": {"morpho": ["n-s---fn3-"], "uri": ["a3076"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["a3075"]},
    },
    "ASPERITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a3077"]}},
    "ASPERNATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a3081"]}},
    "ASPERNOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["a3083"]}},
    "ASPERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a3084"]}},
    "ASPERVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["56941"]}},
    "ASPHALTITIS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "ASPICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["a3091"]}},
    "ASPIRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a3096"]}},
    "ASPIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a3097"]}},
    "ASPLENOS": {"-": {"morpho": ["n-s---mn2g"], "uri": ["a3099"]}},
    "ASPORTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a3100"]}},
    "ASPORTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a3101"]}},
    "ASSARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a3113"]}},
    "ASSECLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["34815"]}},
    "ASSECTATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a3115"]}},
    "ASSECTOR": {"2": {"morpho": ["v1spid--1-"], "uri": ["a3116"]}},
    "ASSENSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a3125"]}},
    "ASSENSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a3126"]}},
    "ASSENSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a3151"]}},
    "ASSENTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a3128"]}},
    "ASSENTATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a3130"]}},
    "ASSENTIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["a3134"]}},
    "ASSENTIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["132261"]}},
    "ASSENTOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["a3135"]}},
    "ASSEQVOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["a3137"]}},
    "ASSER": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a3138"]}},
    "ASSERCVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["36947"]}},
    "ASSERO": {
        "2": {"morpho": ["v1spia--3-"], "uri": ["a3141"]},
        "3": {"morpho": ["v1spia--3-"], "uri": ["a3141"]},
    },
    "ASSERTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a6142"]}},
    "ASSERVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a3147"]}},
    "ASSESSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a3149"]}},
    "ASSEVERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a3153"]}},
    "ASSEVERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a3154"]}},
    "ASSICCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a3157"]}},
    "ASSIDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a3159"]}},
    "ASSIDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a3160"]}},
    "ASSIDVE": {"-": {"morpho": ["rp--------"], "uri": ["40859"]}},
    "ASSIDVITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a3161"]}},
    "ASSIDVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a3163"]}},
    "ASSIGNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a3169"]}},
    "ASSILIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["a3171"]}},
    "ASSIMILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["a3173"]}},
    "ASSIMVLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a3172"]}},
    "ASSIMVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a3175"]}},
    "ASSISTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a2418"]}},
    "ASSOLEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a3186"]}},
    "ASSONO": {"1": {"morpho": ["v1spia--1-"], "uri": ["a3190"]}},
    "ASSORINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ASSORINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "ASSORVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "ASSVEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["a3193"]}},
    "ASSVESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a3194"]}},
    "ASSVETVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a3195"]}},
    "ASSVETVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["35174"]}},
    "ASSVGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a3196"]}},
    "ASSVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a3197"]}},
    "ASSVLTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a3201"]}},
    "ASSVLTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a8907"]}},
    "ASSVM": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["36160"]},
        "2": {"morpho": ["n-s---nn2-"], "uri": ["a3108"]},
    },
    "ASSVMO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "ASSVRGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a3210"]}},
    "ASSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a3108"]}},
    "ASTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a3239"]}},
    "ASTRAGALVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a3245"]}},
    "ASTREPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a3249"]}},
    "ASTRICTE": {"-": {"morpho": ["rp--------"], "uri": ["56905"]}},
    "ASTRICTVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": ["43555"]},
        "-": {"morpho": ["aps---mn1-"], "uri": ["43555"]},
    },
    "ASTRINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a3259"]}},
    "ASTROLOGIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a3265"]}},
    "ASTROLOGVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a3267"]}},
    "ASTRONOMIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a3268"]}},
    "ASTRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a3275"]}},
    "ASTRVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a3276"]}},
    "ASTV": {"-": {"morpho": ["n-s---nn--"], "uri": ["a3277"]}},
    "ASTVPEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a3280"]}},
    "ASTVRCO": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "ASTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a3282"]}},
    "ASTVTE": {"-": {"morpho": ["rp--------"], "uri": ["20187"]}},
    "ASTVTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a3283"]}},
    "ASTVTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a3285"]}},
    "ASYLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a3290"]}},
    "AT": {
        "1": {"morpho": ["rp--------"], "uri": ["46556"]},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "ATAT": {"-": {"morpho": ["----------"], "uri": []}},
    "ATAVI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["97704"]}},
    "ATAVVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a3298"]}},
    "ATER": {"-": {"morpho": ["aps---mn1r"], "uri": ["a6288"]}},
    "ATHENAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "ATHENIENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "ATHENIENSIS": {
        "N": {"morpho": ["aps---cn3-"], "uri": []},
        "A": {"morpho": ["aps---cn3-"], "uri": []},
    },
    "ATHENIO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ATHLETA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["a6295"]}},
    "ATIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ATILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ATINIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "ATLANTES": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "ATOMVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["a6300"]}},
    "ATQVE": {
        "1": {"morpho": ["----------"], "uri": []},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "ATQVI": {"-": {"morpho": ["----------"], "uri": []}},
    "ATRAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a3305"]}},
    "ATRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a3307"]}},
    "ATRIENSIS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["a3312"]},
        "2": {"morpho": ["aps---cn3i"], "uri": ["97709"]},
    },
    "ATRIVM": {
        "-": {"morpho": ["n-s---nn2-"], "uri": ["a3318"]},
        "N": {"morpho": ["n-s---nn2-"], "uri": ["a3318"]},
    },
    "ATROCITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a3319"]}},
    "ATROCITER": {"-": {"morpho": ["rp--------"], "uri": ["20206"]}},
    "ATROX": {"-": {"morpho": ["aps---an3i"], "uri": ["a3323"]}},
    "ATTAGEN": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a1984"]}},
    "ATTALICVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "ATTALVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ATTAMEN": {"-": {"morpho": ["rp--------"], "uri": ["28550"]}},
    "ATTEGIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a3333"]}},
    "ATTEMPERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a3337"]}},
    "ATTENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a3339"]}},
    "ATTENTE": {"-": {"morpho": ["rp--------"], "uri": ["20216"]}},
    "ATTENTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["20218"]}},
    "ATTENTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["a3327"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["20219"]},
    },
    "ATTENVATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["20221"]}},
    "ATTENVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a3344"]}},
    "ATTERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a3346"]}},
    "ATTEXO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a3352"]}},
    "ATTICA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ATTICVS": {
        "C": {"morpho": ["n-s----n2-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "ATTINEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a3361"]}},
    "ATTINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a3400"]}},
    "ATTIS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ATTOLLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a3368"]}},
    "ATTONDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a3369"]}},
    "ATTONITVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["a8910"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["54232"]},
    },
    "ATTONO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a3370"]}},
    "ATTRACTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a3374"]}},
    "ATTRACTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["51775"]}},
    "ATTRAHO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a3376"]}},
    "ATTRECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a3378"]}},
    "ATTRIBVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a3381"]}},
    "ATTRITVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["a8913"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["20243"]},
    },
    "ATTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AVARA": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "AVARE": {"-": {"morpho": ["rp--------"], "uri": ["20320"]}},
    "AVARITER": {"-": {"morpho": ["rp--------"], "uri": ["20321"]}},
    "AVARITIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["a3395"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "AVARITIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["a3395"]}},
    "AVARVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["97737"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["a3396"]},
    },
    "AVCEPS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["a3401"]}},
    "AVCTE": {"-": {"morpho": ["rp--------"], "uri": ["39877"]}},
    "AVCTIFICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a3407"]}},
    "AVCTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a3408"]}},
    "AVCTIONARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a3410"]}},
    "AVCTIONOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["a3412"]}},
    "AVCTITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a3413"]}},
    "AVCTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a3416"]}},
    "AVCTOR": {"-": {"morpho": ["n-s---cn3-"], "uri": ["a3417"]}},
    "AVCTORAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a3419"]}},
    "AVCTORITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a3422"]}},
    "AVCTORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a3423"]}},
    "AVCTVMNVS": {"1": {"morpho": ["n-s---mn2-"], "uri": ["33721"]}},
    "AVCTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["a3499"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["97715"]},
    },
    "AVCVPIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a3431"]}},
    "AVCVPO": {"-": {"morpho": ["v1spia--1-"], "uri": ["97716"]}},
    "AVCVPOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["a3432"]}},
    "AVDACIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a3433"]}},
    "AVDACITER": {"-": {"morpho": ["rp--------"], "uri": ["20257"]}},
    "AVDAX": {"-": {"morpho": ["aps---an3i"], "uri": ["a3434"]}},
    "AVDENS": {"-": {"morpho": ["aps---an3i"], "uri": ["20260"]}},
    "AVDENTER": {"-": {"morpho": ["rp--------"], "uri": ["46574"]}},
    "AVDENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a3403"]}},
    "AVDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a3437"]}},
    "AVDIENS": {"-": {"morpho": ["aps---an3i"], "uri": ["97718"]}},
    "AVDIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["a3440"]}},
    "AVDITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a3441"]}},
    "AVDITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a3444"]}},
    "AVDITORIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a3446"]}},
    "AVDITVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["97719"]}},
    "AVDITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a3449"]}},
    "AVE": {"-": {"morpho": ["----------"], "uri": []}},
    "AVEHO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a3450"]}},
    "AVELLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a3451"]}},
    "AVENA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a3452"]}},
    "AVENTINVS": {
        "1N": {"morpho": ["n-s----n2-"], "uri": []},
        "2N": {"morpho": ["n-s----n--"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": ["97739"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "AVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a3465"]}},
    "AVERNVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": ["97740"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["97740"]},
    },
    "AVERRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a3456"]}},
    "AVERRVNCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a3457"]}},
    "AVERSABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["a3459"]}},
    "AVERSATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a3460"]}},
    "AVERSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a3462"]}},
    "AVERSOR": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["a3464"]},
        "2": {"morpho": ["v1spid--1-"], "uri": ["a3463"]},
    },
    "AVERSVM": {
        "-": {
            "morpho": ["n-s---nn2-", "n-s---nn2-"],
            "uri": ["47031", "47031"],
        }
    },
    "AVERSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["52297"]}},
    "AVERTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a3468"]}},
    "AVFERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a3469"]}},
    "AVFVGIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["a3470"]}},
    "AVGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a3471"]}},
    "AVGESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a6471"]}},
    "AVGMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["a3475"]}},
    "AVGVR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a3480"]}},
    "AVGVRALE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["a3482"]}},
    "AVGVRALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["a3482"]}},
    "AVGVRATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a3519"]}},
    "AVGVRIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a3489"]}},
    "AVGVROR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132271"]}},
    "AVGVSTA": {
        "1N": {"morpho": ["n-s----n1-"], "uri": []},
        "2N": {"morpho": ["n-s----n1-"], "uri": []},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "AVGVSTE": {"-": {"morpho": ["rp--------"], "uri": ["20280"]}},
    "AVGVSTVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["a3492"]},
        "C": {"morpho": ["n-s----n2-"], "uri": []},
        "N": {"morpho": ["aps---mn1-"], "uri": ["a3492"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["a3492"]},
    },
    "AVIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a3628"]}},
    "AVIARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a3494"]}},
    "AVIDE": {"-": {"morpho": ["rp--------"], "uri": ["20331"]}},
    "AVIDITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a3498"]}},
    "AVIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a6490"]}},
    "AVIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["a3502"]}},
    "AVITVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["a3504"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["97743"]},
    },
    "AVIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a3505"]}},
    "AVIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a3505"]}},
    "AVLA": {
        "1": {"morpho": ["n-s---fn1-"], "uri": ["a3672"]},
        "2": {"morpho": ["n-s---fn1-"], "uri": ["a3672"]},
    },
    "AVLAEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a3507"]}},
    "AVLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AVOCAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a3521"]}},
    "AVOCATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a3522"]}},
    "AVOCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a3525"]}},
    "AVOLO": {"2": {"morpho": ["v1spia--1-"], "uri": ["a3527"]}},
    "AVRA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["a3529"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "AVRARIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a3531"]}},
    "AVRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["$0164"]}},
    "AVRELIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": ["97723"]},
    },
    "AVREOLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2035"]}},
    "AVREVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["a3541"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["a3541"]},
    },
    "AVRICILLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["33303"]}},
    "AVRICOMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a3546"]}},
    "AVRICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a3547"]}},
    "AVRIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["a3550"]}},
    "AVRIFERVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "AVRIFEX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a3551"]}},
    "AVRIGA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["a3556"]}},
    "AVRION": {"-": {"morpho": ["rp--------"], "uri": []}},
    "AVRIPIGMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a3568"]}},
    "AVRIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["a3569"]}},
    "AVRITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a3572"]}},
    "AVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a3574"]}},
    "AVRORA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["a3576"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "AVRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a3585"]}},
    "AVRVNCI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "AVRVNCVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "AVSCVLTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a3586"]}},
    "AVSCVLTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a3588"]}},
    "AVSPEX": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["a3589"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "AVSPICATVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["97727"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["20300"]},
        "-": {"morpho": ["n-s---mn4-"], "uri": ["97727"]},
    },
    "AVSPICIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a3592"]}},
    "AVSPICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["97728"]}},
    "AVSPICOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["a3593"]}},
    "AVSTER": {
        "-": {"morpho": ["n-s---mn2r"], "uri": ["a3595"]},
        "N": {"morpho": ["n-s---mn2r"], "uri": ["a3595"]},
    },
    "AVSTERITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a3597"]}},
    "AVSTERVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a3599"]}},
    "AVSTRALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["a3600"]}},
    "AVSTRINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a3602"]}},
    "AVSVM": {
        "A": {"morpho": ["n-s---nn2-"], "uri": ["a3436"]},
        "-": {"morpho": ["n-s---nn2-"], "uri": ["a3436"]},
    },
    "AVT": {"-": {"morpho": ["----------"], "uri": []}},
    "AVTEM": {"-": {"morpho": ["----------"], "uri": []}},
    "AVTIONOR": {"-": {"morpho": ["v1spid--1-"], "uri": []}},
    "AVTOMATOPOEETVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["103429"]}},
    "AVTVMNALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["a3623"]}},
    "AVTVMNITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a3625"]}},
    "AVTVMNVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["a3668"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["a3667"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["a3668"]},
    },
    "AVTVMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a3630"]}},
    "AVVNCVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a3633"]}},
    "AVVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a3634"]}},
    "AVXILIARES": {"-": {"morpho": ["n-p---mn3i"], "uri": ["a3637"]}},
    "AVXILIARII": {"-": {"morpho": ["n-p---mn2-"], "uri": ["a3638"]}},
    "AVXILIARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["a3637"]}},
    "AVXILIARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1733"]}},
    "AVXILIATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a3640"]}},
    "AVXILIATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a3629"]}},
    "AVXILIOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["a3642"]}},
    "AVXILIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a3643"]}},
    "AXICVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a3659"]}},
    "AXIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["a3666"]}},
    "AXON": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a3658"]}},
    "BABYLONICVM": {
        "-": {"morpho": ["n-s----n2-"], "uri": []},
        "N": {"morpho": ["n-s---nn2-"], "uri": []},
    },
    "BACCA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["b0038"]}},
    "BACCAR": {"-": {"morpho": ["n-s---nn3i"], "uri": ["b0014"]}},
    "BACCATVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "BACCHA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["103458"]},
        "N": {"morpho": ["n-s---fn3-"], "uri": []},
    },
    "BACCHABVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["b0016"]}},
    "BACCHANAL": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "BACCHANALIA": {
        "-": {"morpho": ["n-s----n--"], "uri": []},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "BACCHATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["b0018"]}},
    "BACCHOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["b0022"]}},
    "BACCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["b0023"]}},
    "BACCIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["33914"]}},
    "BACILLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["b0028"]}},
    "BACVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["b0034"]}},
    "BACVLVS": {
        "-": {"morpho": ["n-s----n2-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "BAIAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "BALAENA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["50968"]}},
    "BALANATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["b0049"]}},
    "BALANTES": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "BALANVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["b0053"]}},
    "BALATRO": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["b0054"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "BALATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["b0060"]}},
    "BALBE": {"-": {"morpho": ["rp--------"], "uri": ["20350"]}},
    "BALBVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["b0056"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["54747"]},
    },
    "BALBVTIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["b0058"]}},
    "BALEARES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "BALINEAE": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "BALINEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": []}},
    "BALLISTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["b0070"]}},
    "BALNEARIA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["b0333"]}},
    "BALNEARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["b0286"]}},
    "BALNEATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["b0078"]}},
    "BALNEOLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["b0084"]}},
    "BALNEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["b0085"]}},
    "BALO": {"-": {"morpho": ["v1spia--1-"], "uri": ["b0086"]}},
    "BALSAMVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["b0092"]}},
    "BALTEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["97750"]}},
    "BALTEVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["b0096"]}},
    "BAPTIZO": {"-": {"morpho": ["v1spia--1-"], "uri": ["b0112"]}},
    "BARATHRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["b0113"]}},
    "BARBA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["b0114"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["b0114"]},
    },
    "BARBARE": {"-": {"morpho": ["rp--------"], "uri": ["46789"]}},
    "BARBARIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["b0115"]}},
    "BARBARICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["b0117"]}},
    "BARBARISMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["b0118"]}},
    "BARBARVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["b0121"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["b0121"]},
    },
    "BARBATVS": {"2": {"morpho": ["aps---mn1-"], "uri": ["b0125"]}},
    "BARBIGER": {"-": {"morpho": ["aps---mn1r"], "uri": ["b0127"]}},
    "BARBITON": {"-": {"morpho": ["n-s---nn2-"], "uri": ["36859"]}},
    "BARBITOS": {"-": {"morpho": ["n-s---mn2g"], "uri": ["b0131"]}},
    "BARDITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["103531"]}},
    "BARIOBAL": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "BARIS": {"-": {"morpho": ["n-s---fn3g"], "uri": ["b0143"]}},
    "BARO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["b0445"]}},
    "BARRVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["b0148"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["b0148"]},
    },
    "BARVLCOS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "BARYCEPHALVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["b0149"]}},
    "BASCAVDA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["b0152"]}},
    "BASIATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["b0154"]}},
    "BASILICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["b0157"]}},
    "BASILISCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["b0165"]}},
    "BASIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["b0168"]}},
    "BASIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["b0170"]}},
    "BASIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["b0171"]}},
    "BAVBOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["b0193"]}},
    "BEATE": {"-": {"morpho": ["rp--------"], "uri": ["53889"]}},
    "BEATVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["b0207"]}},
    "BEATVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["41561"]}},
    "BEATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["20387"]}},
    "BELLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["b0220"]}},
    "BELLATRIX": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "BELLE": {"-": {"morpho": ["rp--------"], "uri": ["20393"]}},
    "BELLICOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["b0227"]}},
    "BELLICVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["b0082"]}},
    "BELLICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["b0015"]}},
    "BELLIGER": {"-": {"morpho": ["aps---mn1r"], "uri": ["b0232"]}},
    "BELLIGERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["b0234"]}},
    "BELLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["b0241"]}},
    "BELLONA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "BELLVA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["20399"]}},
    "BELLVM": {
        "-": {"morpho": ["n-s---nn2-"], "uri": ["b0244"]},
        "N": {"morpho": ["n-s---nn2-"], "uri": ["b0244"]},
    },
    "BELLVOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["103587"]}},
    "BELLVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["b0245"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "BELVA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["b0164"]}},
    "BENE": {"-": {"morpho": ["rp--------"], "uri": ["b0256"]}},
    "BENEDICO": {"2": {"morpho": ["v1spia--3-"], "uri": ["b9256"]}},
    "BENEDICTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["103597"]}},
    "BENEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["b9257"]}},
    "BENEFACTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["b9259"]}},
    "BENEFICENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["b0262"]}},
    "BENEFICIARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["b0216"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["b0198"]},
    },
    "BENEFICIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["b0266"]}},
    "BENEFICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["b0268"]}},
    "BENEVENTANVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "BENEVENTVM": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "BENEVOLE": {"-": {"morpho": ["rp--------"], "uri": ["32459"]}},
    "BENEVOLENS": {"-": {"morpho": ["aps---an3i"], "uri": ["b0272"]}},
    "BENEVOLENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["b0795"]}},
    "BENEVOLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["b0905"]}},
    "BENIGNE": {"-": {"morpho": ["rp--------"], "uri": ["20416"]}},
    "BENIGNITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["b0274"]}},
    "BENIGNVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["b0275"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "BEO": {"-": {"morpho": ["v1spia--1-"], "uri": ["b0278"]}},
    "BERYLLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["b0280"]}},
    "BES": {"-": {"morpho": ["n-s---mn3i"], "uri": ["b0281"]}},
    "BESTIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["b0491"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["b0491"]},
    },
    "BESTIARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["b0285"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["b0285"]},
    },
    "BESTIOLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["b0254"]}},
    "BETA": {
        "1": {"morpho": ["n-s---fn1-"], "uri": ["b0287"]},
        "2": {"morpho": ["n-s---fn1-"], "uri": ["b0287"]},
    },
    "BIBERARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["b0300"]}},
    "BIBLIOTHECA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["b0307"]}},
    "BIBO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["b0312"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["b0311"]},
    },
    "BIBVLVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["b0315"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["34019"]},
    },
    "BICEPS": {"-": {"morpho": ["aps---an3-"], "uri": ["b0317"]}},
    "BICOLOR": {"-": {"morpho": ["aps---an3-"], "uri": ["b0322"]}},
    "BICORNES": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "BICORNIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["b0325"]}},
    "BIDENS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["b0332"]},
        "2": {"morpho": ["aps---an3i"], "uri": ["b0332"]},
    },
    "BIDENTAL": {"-": {"morpho": ["n-s---nn3i"], "uri": ["b0341"]}},
    "BIDINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "BIDINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "BIDIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "BIDVVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["b0338"]}},
    "BIENNIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["b0342"]}},
    "BIENNIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["b0342"]}},
    "BIFARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["b0346"]}},
    "BIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["b0344"]}},
    "BIFORIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["b0351"]}},
    "BIFORMIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["b0353"]}},
    "BIGA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["b0356"]}},
    "BIGATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["b0359"]}},
    "BIIVGI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["97766"]}},
    "BIIVGIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["b0798"]}},
    "BIIVGVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["b0797"]}},
    "BILIBRIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["b0370"]}},
    "BILINGVIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["b0371"]}},
    "BILIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["b0373"]}},
    "BILVSTRIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["b0376"]}},
    "BIMARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["b0379"]}},
    "BIMEMBRES": {
        "-": {"morpho": ["n-p---mn3i"], "uri": ["97773"]},
        "N": {"morpho": ["n-s---mn1g"], "uri": []},
    },
    "BIMEMBRIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["b0383"]}},
    "BIMESTRIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["103680"]}},
    "BIMVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["b0387"]}},
    "BIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["b0388"]}},
    "BINOCTIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["b0393"]}},
    "BINVS": {"-": {"morpho": ["rp--------"], "uri": []}},
    "BIPALIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["b0400"]}},
    "BIPARTIOR": {"-": {"morpho": ["v1spid--4-"], "uri": []}},
    "BIPARTITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["36619"]}},
    "BIPARTITO": {"-": {"morpho": ["rp--------"], "uri": ["40021"]}},
    "BIPATENS": {"-": {"morpho": ["aps---an3i"], "uri": ["b0403"]}},
    "BIPEDALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["b0404"]}},
    "BIPENNIS": {
        "1": {"morpho": ["n-s---fn3i"], "uri": ["58143"]},
        "2": {"morpho": ["aps---cn3i"], "uri": ["b0410"]},
    },
    "BIPES": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["97780"]},
        "2": {"morpho": ["aps---an3-"], "uri": ["b0413"]},
    },
    "BIREMIS": {
        "1": {"morpho": ["n-s---fn3i"], "uri": ["97782"]},
        "2": {"morpho": ["aps---cn3i"], "uri": ["b0415"]},
    },
    "BIS": {"-": {"morpho": ["rp--------"], "uri": ["b0418"]}},
    "BISON": {"-": {"morpho": ["n-s---mn3i"], "uri": ["b0428"]}},
    "BISVLCVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["b0433"]}},
    "BITHYNIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "BITVMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["b0438"]}},
    "BITVMINOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["b0441"]}},
    "BIVGVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "BIVIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["b0424"]}},
    "BLAESVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["b0450"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["47928"]},
    },
    "BLANDE": {"-": {"morpho": ["rp--------"], "uri": ["37042"]}},
    "BLANDILOQVENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["b0457"]}},
    "BLANDILOQVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["b0460"]}},
    "BLANDIMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["b0461"]}},
    "BLANDIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["b0462"]}},
    "BLANDITIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["b0463"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "BLANDVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["b0466"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "BLATERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["b0477"]}},
    "BLATTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["b0496"]}},
    "BLATTARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["b0482"]}},
    "BOARIVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["b0666"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["b0666"]},
    },
    "BOEOTIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "BOEOTII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "BOETHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["b0498"]}},
    "BOLETATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["b0504"]}},
    "BOLETVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["b0505"]}},
    "BOMBVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["b0522"]}},
    "BOMBYCINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["b0524"]}},
    "BOMBYX": {"-": {"morpho": ["n-s---cn3-"], "uri": ["b0527"]}},
    "BONI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["97788"]}},
    "BONITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["b0533"]}},
    "BONVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["b0144"]}},
    "BONVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["b0537"]},
        "N": {"morpho": ["aps---mn1-"], "uri": ["b0537"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["b0537"]},
    },
    "BOO": {"-": {"morpho": ["v1spia--1-"], "uri": ["b0539"]}},
    "BOREAS": {
        "-": {"morpho": ["n-s----n-g"], "uri": []},
        "N": {"morpho": ["n-s---mn1g"], "uri": ["b0611"]},
    },
    "BOS": {
        "-": {"morpho": ["n-s---cn3-"], "uri": ["b0551"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "BOTVLARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["b0568"]}},
    "BRACA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["b0582"]}},
    "BRACATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["b0584"]}},
    "BRACHIOLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["40790"]}},
    "BRACHIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["49386"]}},
    "BRACTEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["36625"]}},
    "BRACTEATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["55710"]}},
    "BRACTEOLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["35033"]}},
    "BRASSICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["b0607"]}},
    "BREVE": {
        "1": {"morpho": ["n-s---nn3i"], "uri": ["97793"]},
        "2": {"morpho": ["rp--------"], "uri": []},
    },
    "BREVI": {"-": {"morpho": ["rp--------"], "uri": ["51970"]}},
    "BREVIARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["b0621"]}},
    "BREVIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["b0630"]}},
    "BREVITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["b0631"]}},
    "BREVITER": {"-": {"morpho": ["rp--------"], "uri": ["20483"]}},
    "BRITANNIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "BRITANNVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "BROCCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BRVMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["b0646"]}},
    "BRVMALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["b0647"]}},
    "BRVNDISIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "BRVTESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["b0653"]}},
    "BRVTVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["b0651"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "BVBILE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["30069"]}},
    "BVBO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["b0664"]}},
    "BVBVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["b0703"]}},
    "BVBVLCVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["b0670"]}},
    "BVBVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["b0697"]}},
    "BVCCA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["b0677"]}},
    "BVCCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["b0687"]}},
    "BVCERIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["34960"]}},
    "BVCERVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["b0691"]}},
    "BVCINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["b0804"]}},
    "BVCINATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["b0695"]}},
    "BVCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["b0707"]}},
    "BVFO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["b0710"]}},
    "BVLBVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["b0720"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["b0720"]},
    },
    "BVLLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["b0728"]}},
    "BVLLATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["b0729"]}},
    "BVLLIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["b0731"]}},
    "BVLLITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["b0688"]}},
    "BVLLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["b0733"]}},
    "BVMASTVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["b0736"]}},
    "BVRIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["b0744"]}},
    "BVRRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BVSIRIS": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "BVSTVARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["b0768"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["b0768"]},
    },
    "BVSTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["b0769"]}},
    "BVTHROTVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "BVXIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["b0785"]}},
    "BVXVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["b0788"]}},
    "BVXVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["b0788"]}},
    "BYZANTII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "BYZANTIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "CABALLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c0002"]}},
    "CACHINNO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["c0030"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["c0029"]},
    },
    "CACHINNVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c0032"]}},
    "CACO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c0036"]}},
    "CACOETHES": {"-": {"morpho": ["n-s---nn3-"], "uri": ["c0040"]}},
    "CACVMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["c0050"]}},
    "CACVRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CACVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["103996"]}},
    "CADAVER": {"-": {"morpho": ["n-s---nn3-"], "uri": ["c0052"]}},
    "CADMEA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CADMVS": {
        "-": {"morpho": ["n-s----n2-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "CADO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c0061"]}},
    "CADVCEATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c0063"]}},
    "CADVCEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c0065"]}},
    "CADVCEVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["43410"]}},
    "CADVCIFER": {
        "-": {"morpho": ["aps---mn1r"], "uri": ["97799"]},
        "N": {"morpho": ["n-s----n2-"], "uri": []},
    },
    "CADVCVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c0068"]}},
    "CADVCVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0068"]}},
    "CADVRCVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c0069"]}},
    "CADVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c0070"]}},
    "CAECIGENVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0074"]}},
    "CAECILIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "CAECITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c0076"]}},
    "CAECO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c0078"]}},
    "CAECVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["c0080"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["c0080"]},
    },
    "CAEDES": {"-": {"morpho": ["n-s---fn3i"], "uri": ["c0082"]}},
    "CAEDICIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CAEDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c0083"]}},
    "CAEDVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0084"]}},
    "CAELATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c0086"]}},
    "CAELATVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0130"]}},
    "CAELEBS": {"-": {"morpho": ["aps---an3-"], "uri": ["c0087"]}},
    "CAELES": {
        "1": {"morpho": ["n-s----n3-"], "uri": []},
        "2": {"morpho": ["aps---an3-"], "uri": ["c0088"]},
        "N": {"morpho": ["n-s---mn1g"], "uri": []},
    },
    "CAELESTES": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "CAELESTIA": {"-": {"morpho": ["n-p---nn3i"], "uri": ["97801"]}},
    "CAELESTIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c0090"]}},
    "CAELI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["104013"]}},
    "CAELIBATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c4698"]}},
    "CAELICOLA": {"-": {"morpho": ["n-s---cn1-"], "uri": ["c0095"]}},
    "CAELIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["c0097"]}},
    "CAELITES": {"-": {"morpho": ["n-p---mn3-"], "uri": ["48504"]}},
    "CAELIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CAELO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c0104"]}},
    "CAELVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["c9105", "c0105"]},
        "2": {"morpho": ["n-s---nn2-"], "uri": ["c9105", "c0105"]},
    },
    "CAEMENTICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0107"]}},
    "CAEMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c0149"]}},
    "CAENINENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "CAENINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "CAENO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "CAENOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0110"]}},
    "CAENVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c0112"]}},
    "CAEPA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0317"]}},
    "CAEPIO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "CAERE": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "CAERIMONIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0000"]}},
    "CAERVLA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["97805"]}},
    "CAERVLEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c0726"]}},
    "CAERVLEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0626"]}},
    "CAESAR": {"N": {"morpho": ["n-s---mn3-"], "uri": ["104039"]}},
    "CAESARIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["c0132"]}},
    "CAESETIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CAESIVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["c0137"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["38093"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["c0137"]},
    },
    "CAESO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "CAESONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CAESPES": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c0139"]}},
    "CAESTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c0145"]}},
    "CALACTE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CALACTINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CALACTINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "CALAMISTER": {"-": {"morpho": ["n-s---mn2r"], "uri": ["c0162"]}},
    "CALAMITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c0164"]}},
    "CALAMITOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0166"]}},
    "CALAMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c0168"]}},
    "CALATHISCVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c0170"]}},
    "CALATHVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c0172"]}},
    "CALCAR": {"-": {"morpho": ["n-s---nn3i"], "uri": ["c0178"]}},
    "CALCARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["c0182"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["c0182"]},
    },
    "CALCATVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0213"]}},
    "CALCEAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c0190"]}},
    "CALCEO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c0193"]}},
    "CALCEVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c0198"]}},
    "CALCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c0203"]}},
    "CALCVLOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0211"]}},
    "CALCVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c0212"]}},
    "CALDA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["54763"]}},
    "CALDARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["41332"]}},
    "CALDARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["45236"]}},
    "CALEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["c0214"]}},
    "CALEFACTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c0218"]}},
    "CALEFIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c0215"]}},
    "CALENVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "CALEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c0222"]}},
    "CALESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c1101"]}},
    "CALICVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c0228"]}},
    "CALIDA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c2516"]}},
    "CALIDIANVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "CALIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CALIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c2491"]}},
    "CALIENDRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["36566"]}},
    "CALIGA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0232"]}},
    "CALIGATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c9235"]}},
    "CALIGINOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0238"]}},
    "CALIGO": {
        "1": {"morpho": ["n-s---fn3-"], "uri": ["c0239"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["c0240"]},
    },
    "CALIX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c0245"]}},
    "CALLEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c0250"]}},
    "CALLIDAMA": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "CALLIDE": {"-": {"morpho": ["rp--------"], "uri": ["45369"]}},
    "CALLIDITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c0253"]}},
    "CALLIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0255"]}},
    "CALLIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["c0263"]}},
    "CALLISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c9704"]}},
    "CALLOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0269"]}},
    "CALLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c0270"]}},
    "CALO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["c1152"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["c0272"]},
    },
    "CALOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c0276"]}},
    "CALPVRNIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "CALTHA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0281"]}},
    "CALVARIA": {
        "1": {"morpho": ["n-s---fn1-"], "uri": ["c9284"]},
        "2": {"morpho": ["n-s---fn1-"], "uri": ["c9284"]},
    },
    "CALVITIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c0290"]}},
    "CALVMNIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0292"]}},
    "CALVMNIATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c0294"]}},
    "CALVMNIOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["c0296"]}},
    "CALVVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["c2611"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["40642"]},
    },
    "CALX": {
        "1": {"morpho": ["n-s---fn3-"], "uri": ["c0291", "c0302"]},
        "2": {"morpho": ["n-s---fn3-"], "uri": ["c0291", "c0302"]},
    },
    "CAMELVS": {"-": {"morpho": ["n-s---cn2-"], "uri": ["c3037"]}},
    "CAMENA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["34882"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "CAMENAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "CAMERA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["$0155"]}},
    "CAMERIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CAMINVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c0325"]}},
    "CAMMARVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c0327"]}},
    "CAMPANIA": {"N": {"morpho": ["n-s---fn1-"], "uri": ["104154"]}},
    "CAMPESTER": {"-": {"morpho": ["aps---mn3-"], "uri": ["51934"]}},
    "CAMPESTRE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["97815"]}},
    "CAMPESTRIA": {"-": {"morpho": ["n-p---nn3i"], "uri": ["c0334"]}},
    "CAMPVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["c1593"]},
        "2": {"morpho": ["n-s---mn2-"], "uri": ["c1593"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["c1593"]},
    },
    "CAMVRVS": {"-": {"morpho": ["aps---mn1-"], "uri": []}},
    "CANALICVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c0361"]}},
    "CANALIS": {
        "1": {"morpho": ["n-s---fn3i"], "uri": []},
        "2": {"morpho": ["aps---cn3i"], "uri": ["104171"]},
    },
    "CANCELLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c0373"]}},
    "CANCER": {
        "-": {"morpho": ["n-s---mn2r"], "uri": ["c0375"]},
        "N": {"morpho": ["n-s----n2-"], "uri": []},
    },
    "CANDELA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0383"]}},
    "CANDELABRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c0385"]}},
    "CANDENS": {
        "-": {
            "morpho": ["aps---an3i", "aps---an3i"],
            "uri": ["97819", "97819"],
        }
    },
    "CANDENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c9917"]}},
    "CANDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c0386"]}},
    "CANDESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c1245"]}},
    "CANDIDATVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["c9393"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["c9722"]},
    },
    "CANDIDVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0393"]}},
    "CANDIDVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["c0394"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "CANDOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c0397"]}},
    "CANEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c0400"]}},
    "CANEPHORA": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "CANESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c0400"]}},
    "CANI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["c0472"]}},
    "CANICVLA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["c0401"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "CANIDIA": {
        "-": {"morpho": ["n-s----n1-"], "uri": []},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "CANINVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["c0407"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["41544"]},
    },
    "CANIS": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["20611"]},
        "N": {"morpho": ["n-s---fn3-"], "uri": []},
    },
    "CANISTRA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["97822"]}},
    "CANISTRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["41731"]}},
    "CANITIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["c0415"]}},
    "CANNA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c2412"]}},
    "CANNABIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["c0421"]}},
    "CANNENSIS": {
        "N": {"morpho": ["aps---cn3i"], "uri": []},
        "A": {"morpho": ["aps---cn3i"], "uri": []},
    },
    "CANNVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0427"]}},
    "CANO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c0428"]}},
    "CANON": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c0429"]}},
    "CANONICVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["c0431"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["c0431"]},
    },
    "CANOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c0433"]}},
    "CANORVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0434"]}},
    "CANTABRI": {
        "-": {"morpho": ["n-s----n2-"], "uri": []},
        "N": {"morpho": ["n-p---mn2-"], "uri": []},
    },
    "CANTERIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["42432"]}},
    "CANTHARVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c0451"]}},
    "CANTHVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c0456"]}},
    "CANTICVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c0458"]}},
    "CANTILENA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0460"]}},
    "CANTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c0463"]}},
    "CANTITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c0464"]}},
    "CANTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c0466"]}},
    "CANTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c0467"]}},
    "CANTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c9983"]}},
    "CANVLEIANVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "CANVLEIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CANVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["c0472"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "CAPACITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c0476"]}},
    "CAPAX": {
        "-": {"morpho": ["aps---an3i"], "uri": ["c0477"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "CAPEDVNCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0479"]}},
    "CAPELLA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["c0480"]},
        "N": {"morpho": ["n-p---fn1-"], "uri": []},
    },
    "CAPENA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CAPENAS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "CAPER": {"-": {"morpho": ["n-s---mn2r"], "uri": ["c0483"]}},
    "CAPESSO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c0485"]}},
    "CAPETVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CAPILLAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c0490"]}},
    "CAPILLATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0495"]}},
    "CAPILLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c0499"]}},
    "CAPIO": {
        "1": {"morpho": ["n-s---fn3-"], "uri": ["c0501"]},
        "2": {"morpho": ["v1spia--3i"], "uri": ["c0500"]},
    },
    "CAPISTRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c0505"]}},
    "CAPISTRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c0506"]}},
    "CAPITAL": {"-": {"morpho": ["n-s---nn3i"], "uri": ["c0507"]}},
    "CAPITALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c0507"]}},
    "CAPITATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c0511"]}},
    "CAPITATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0512"]}},
    "CAPITINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "CAPITO": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["c0516"]},
        "N": {"morpho": ["n-s---mn3-"], "uri": ["c0516"]},
    },
    "CAPITOLINVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["aps---mn1-", "n-s---mn2-"], "uri": []},
    },
    "CAPITOLIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "CAPITVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c0522"]}},
    "CAPPADOCIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CAPRA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["c0483"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "CAPREA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0537"]}},
    "CAPREOLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c0518"]}},
    "CAPRIFICVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["c0546"]}},
    "CAPRIGENVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0547"]}},
    "CAPRILE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["c0548"]}},
    "CAPRIMVLGVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c0549"]}},
    "CAPRINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0551"]}},
    "CAPRIPES": {"-": {"morpho": ["aps---an3-"], "uri": ["c0553"]}},
    "CAPSA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["c0557"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "CAPSVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0563"]}},
    "CAPSVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c0564"]}},
    "CAPTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c0566"]}},
    "CAPTATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c0567"]}},
    "CAPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c0571"]}},
    "CAPTIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0572"]}},
    "CAPTIVA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0581"]}},
    "CAPTIVITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c0574"]}},
    "CAPTIVVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["c0581"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["c0581"]},
    },
    "CAPTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c0582"]}},
    "CAPTVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c9918"]}},
    "CAPTVS": {
        "-": {"morpho": ["n-s---mn4-"], "uri": ["c9984"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["104282"]},
    },
    "CAPVLATOR": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "CAPVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c0590"]}},
    "CAPVT": {"-": {"morpho": ["n-s---nn3-"], "uri": ["c0592"]}},
    "CAPYS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "CARBASA": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "CARBASEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c9597"]}},
    "CARBASVS": {
        "1": {"morpho": ["n-s---fn2-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": ["c0598"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["c0598"]},
    },
    "CARBO": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["c0599"]},
        "N": {"morpho": ["n-s---mn3-"], "uri": ["c0599"]},
    },
    "CARBVNCVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c0606"]}},
    "CARCER": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c0608"]}},
    "CARCHESIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c0614"]}},
    "CARCINOMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["c0617"]}},
    "CARDIACVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["28973"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["c0519"]},
    },
    "CARDINALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c0629"]}},
    "CARDINATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["$0170"]}},
    "CARDO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c0633"]}},
    "CARDVVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c0635"]}},
    "CARE": {"-": {"morpho": ["rp--------"], "uri": ["44377"]}},
    "CARECTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c9636"]}},
    "CAREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c0637"]}},
    "CAREX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c0639"]}},
    "CARICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c9921"]}},
    "CARIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["c0642"]}},
    "CARINA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["c0643"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["c0643"]},
    },
    "CARIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0649"]}},
    "CARITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c0655"]}},
    "CARMEN": {
        "1": {"morpho": ["n-s---nn3-"], "uri": ["c4060"]},
        "2": {"morpho": ["n-s---nn3-"], "uri": ["c4060"]},
    },
    "CARMENTALIS": {
        "N": {"morpho": ["aps---cn3i"], "uri": []},
        "A": {"morpho": ["aps---cn3i"], "uri": []},
    },
    "CARMENTIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "CARMO": {
        "-": {"morpho": ["n-s----n3-"], "uri": []},
        "N": {"morpho": ["n-s---mn3-"], "uri": []},
    },
    "CARNARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c0663"]}},
    "CARNIFEX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c0670"]}},
    "CARNIFICINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0673"]}},
    "CARNIFICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c0675"]}},
    "CARO": {
        "1": {"morpho": ["n-s---fn3-"], "uri": ["c0681"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["c0680"]},
    },
    "CARPATINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["97836"]}},
    "CARPENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c0692"]}},
    "CARPINATIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CARPINVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["c0696"]}},
    "CARPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c0698"]}},
    "CARPTIM": {"-": {"morpho": ["rp--------"], "uri": ["c9698"]}},
    "CARPTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c0700"]}},
    "CARRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["53849"]}},
    "CARRVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c0713"]}},
    "CARTHAGINIENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "CARTHAGINIENSIS": {
        "N": {"morpho": ["aps---cn3-"], "uri": []},
        "A": {"morpho": ["aps---cn3-"], "uri": []},
    },
    "CARTHAGO": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "CARVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["c0721"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "CARYATIS": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "CASA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0728"]}},
    "CASEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["31950"]}},
    "CASEVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c0740"]}},
    "CASIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0741"]}},
    "CASSIANVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "CASSIDA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0749"]}},
    "CASSIS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["20712"]},
        "2": {"morpho": ["n-s---mn3i"], "uri": ["20712"]},
    },
    "CASSIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "CASSVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c0760"]}},
    "CASSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0760"]}},
    "CASTANEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0761"]}},
    "CASTANEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["97843"]}},
    "CASTE": {"-": {"morpho": ["rp--------"], "uri": ["20718"]}},
    "CASTELLANI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["c0764"]}},
    "CASTELLANVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["47581"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["c0764"]},
    },
    "CASTELLARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c0765"]}},
    "CASTELLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c0767"]}},
    "CASTIFICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0771"]}},
    "CASTIGATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c0773"]}},
    "CASTIGATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c0774"]}},
    "CASTIGATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["56329"]}},
    "CASTIGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c0776"]}},
    "CASTIMONIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0778"]}},
    "CASTITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c0781"]}},
    "CASTOR": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["c0783"]},
        "N": {"morpho": ["n-s---mn3-"], "uri": ["c0783"]},
    },
    "CASTOREVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c0784"]}},
    "CASTOREVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": ["c0784"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["c0784"]},
    },
    "CASTRA": {
        "1": {"morpho": ["n-s---fn1-"], "uri": ["c0787"]},
        "2": {"morpho": ["n-s---fn1-"], "uri": ["c0787"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "CASTRENSIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c0793"]}},
    "CASTRICIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CASTRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c0796"]}},
    "CASTRVM": {
        "-": {"morpho": ["n-s---nn2-"], "uri": ["c0787"]},
        "N": {"morpho": ["n-s---nn2-"], "uri": ["c0787"]},
    },
    "CASTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["c0798"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["c0798"]},
    },
    "CASV": {"-": {"morpho": ["rp--------"], "uri": ["45470"]}},
    "CASVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0800"]}},
    "CASVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c9985"]}},
    "CATAGRAPHVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0820"]}},
    "CATAPHRACTVS": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "CATAPLASMA": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "CATAPLEXIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["c0838"]}},
    "CATAPVLTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0841"]}},
    "CATARACTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0843"]}},
    "CATASCOPVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c9850"]}},
    "CATASTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0851"]}},
    "CATATEXITECHNOS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "CATATONVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0859"]}},
    "CATECHON": {"-": {"morpho": ["----------"], "uri": []}},
    "CATELLA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["c0871"]},
        "2": {"morpho": ["n-s---fn1-"], "uri": ["c0871"]},
        "1": {"morpho": ["n-s---fn1-"], "uri": ["c0871"]},
    },
    "CATELLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c0871"]}},
    "CATENA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0872"]}},
    "CATENARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0873"]}},
    "CATENATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c0874"]}},
    "CATENO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c0877"]}},
    "CATERVA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0880"]}},
    "CATERVATIM": {"-": {"morpho": ["rp--------"], "uri": ["c0882"]}},
    "CATHEDRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0885"]}},
    "CATHEDRARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0888"]}},
    "CATHETVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["c0893"]}},
    "CATILINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CATILLVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["c0900"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["c0900"]},
    },
    "CATINA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CATINENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "CATINENSIS": {
        "N": {"morpho": ["aps---cn3i"], "uri": []},
        "A": {"morpho": ["aps---cn3i"], "uri": []},
    },
    "CATINVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c0902"]}},
    "CATLASTER": {"-": {"morpho": ["n-s---mn2r"], "uri": ["c0903"]}},
    "CATO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "CATVLVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["c0923"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["c0923"]},
    },
    "CATVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["c0925"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["97850"]},
    },
    "CAVATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["46262"]}},
    "CAVDA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0935"]}},
    "CAVDEX": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["c0937"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "CAVDICARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0940"]}},
    "CAVEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0945"]}},
    "CAVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c0949"]}},
    "CAVERNA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0952"]}},
    "CAVILLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c0961"]}},
    "CAVILLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c0962"]}},
    "CAVILLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["c0966"]}},
    "CAVLA": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "CAVLICVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c0974"]}},
    "CAVLIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["c0976"]}},
    "CAVO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["c0980"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["c0980"]},
    },
    "CAVPO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c0982"]}},
    "CAVPONA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0983"]}},
    "CAVPONOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["c0987"]}},
    "CAVRVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["c0994"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["c0994"]},
    },
    "CAVSA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0995"]}},
    "CAVSARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0999"]}},
    "CAVSIDICVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c1008"]}},
    "CAVSOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["c1012"]}},
    "CAVSVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c1014"]}},
    "CAVTE": {"-": {"morpho": ["rp--------"], "uri": ["20765"]}},
    "CAVTES": {"-": {"morpho": ["n-s---fn3i"], "uri": ["c1020"]}},
    "CAVTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c1022"]}},
    "CAVTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["20769"]}},
    "CAVVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c1028"]}},
    "CAVVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["20779"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["c1028"]},
    },
    "CEDO": {
        "1": {"morpho": ["v1spia--3-"], "uri": ["c1031"]},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "CEDRIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c1037"]}},
    "CEDRVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["c1039"]}},
    "CELATVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["57522"]}},
    "CELEBER": {"-": {"morpho": ["aps---mn3-"], "uri": ["c1043"]}},
    "CELEBRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c1045"]}},
    "CELEBRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["20785"]}},
    "CELEBRITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c1047"]}},
    "CELEBRITER": {"-": {"morpho": ["rp--------"], "uri": ["31070"]}},
    "CELEBRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c1049"]}},
    "CELER": {
        "-": {"morpho": ["aps---an3-"], "uri": ["c1051"]},
        "N": {"morpho": ["n-s---mn3-"], "uri": []},
    },
    "CELERIPES": {"-": {"morpho": ["aps---an3-"], "uri": ["c1054"]}},
    "CELERITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c1055"]}},
    "CELERITER": {"-": {"morpho": ["rp--------"], "uri": ["20792"]}},
    "CELERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c1059"]}},
    "CELLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c1067"]}},
    "CELLARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["c1073"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["c1073"]},
    },
    "CELLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["104533"]}},
    "CELLVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c1076"]}},
    "CELO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c1078"]}},
    "CELSVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["c1085"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "CENA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c1089"]}},
    "CENACVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c1094"]}},
    "CENATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c1098"]}},
    "CENDO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "CENO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c1108"]}},
    "CENSEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c1119"]}},
    "CENSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c1123"]}},
    "CENSORIVS": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": ["c1125"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "CENSVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c1118"]}},
    "CENSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c1130"]}},
    "CENTAVREVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c3981"]}},
    "CENTAVRVS": {
        "-": {"morpho": ["n-s----n2-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "CENTENARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c1129"]}},
    "CENTENVS": {"-": {"morpho": ["rp--------"], "uri": []}},
    "CENTESIMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c3987"]}},
    "CENTESIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c3985"]}},
    "CENTICEPS": {"-": {"morpho": ["aps---an3-"], "uri": ["c1136"]}},
    "CENTIES": {"-": {"morpho": ["rp--------"], "uri": []}},
    "CENTIMANVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c1142"]}},
    "CENTO": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["c1150"]},
        "N": {"morpho": ["n-s---mn3-"], "uri": ["c1150"]},
    },
    "CENTRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c1158"]}},
    "CENTVM": {"-": {"morpho": ["----------"], "uri": []}},
    "CENTVMGEMINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c1160"]}},
    "CENTVMPONDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c8163"]}},
    "CENTVMVIR": {"-": {"morpho": ["n-s---mn2r"], "uri": ["35152"]}},
    "CENTVMVIRALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c1164"]}},
    "CENTVNCVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c1167"]}},
    "CENTVRIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c8165"]}},
    "CENTVRIATIM": {"-": {"morpho": ["rp--------"], "uri": ["c8167"]}},
    "CENTVRIATVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["c4468"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["57877"]},
    },
    "CENTVRIO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["c1171"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["c1169"]},
    },
    "CENTVRIONATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c1172"]}},
    "CENTVRIPINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CENTVRIPINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "CENTVSSIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["c1175"]}},
    "CENVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c1176"]}},
    "CEPHALOEDITANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CEPHALOEDITANVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "CEPHALOEDIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "CEPTO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "CERA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c1197"]}},
    "CERARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c1201"]}},
    "CERASTES": {"-": {"morpho": ["n-s---mn1g"], "uri": ["c1205"]}},
    "CERASVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c1206"]}},
    "CERASVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["c1206"]}},
    "CERBERVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CERCEI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CERCOPITHECVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c1222"]}},
    "CERDO": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["c1225"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "CEREALIS": {
        "-": {"morpho": ["aps---cn3i"], "uri": ["104612"]},
        "A": {"morpho": ["aps---cn3i"], "uri": ["104612"]},
    },
    "CEREBROSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c1229"]}},
    "CEREBRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c1230"]}},
    "CERES": {
        "N": {"morpho": ["n-s---fn3-"], "uri": []},
        "-": {"morpho": ["aps---mn1r"], "uri": []},
        "A": {"morpho": ["aps---mn1r"], "uri": []},
    },
    "CEREVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["c1233"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["c1233"]},
    },
    "CERINTHA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c1238"]}},
    "CERNO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c1244"]}},
    "CERNVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c1246"]}},
    "CERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c1250"]}},
    "CEROMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["c1252", "c1252"]}},
    "CEROMATICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c1253"]}},
    "CERPO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "CERRITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c1261"]}},
    "CERRVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["c1262"]}},
    "CERTAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["c1264"]}},
    "CERTATIM": {"-": {"morpho": ["rp--------"], "uri": ["c1265"]}},
    "CERTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c1266"]}},
    "CERTE": {"-": {"morpho": ["rp--------"], "uri": ["20829"]}},
    "CERTO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["c1274"]},
        "2": {"morpho": ["rp--------"], "uri": ["20833"]},
    },
    "CERTVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["c1275"]},
        "2": {"morpho": ["rp--------"], "uri": ["97861"]},
    },
    "CERTVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["c1275"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "CERVA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c1997"]}},
    "CERVICAL": {"-": {"morpho": ["n-s---nn3i"], "uri": ["20837"]}},
    "CERVICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c1286"]}},
    "CERVINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0975"]}},
    "CERVIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CERVIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c1289"]}},
    "CERVSSA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c1293"]}},
    "CERVVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c2004"]}},
    "CERYX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c1296"]}},
    "CESSATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c1298"]}},
    "CESSIM": {"-": {"morpho": ["rp--------"], "uri": ["c1301"]}},
    "CESSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c1303"]}},
    "CETARINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CETARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c1309"]}},
    "CETARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["c1309"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["c1309"]},
        "-": {"morpho": ["aps---mn1-"], "uri": ["c1309"]},
    },
    "CETERA": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["rp--------"], "uri": ["33172"]},
    },
    "CETERI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["56063"]}},
    "CETERVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["97866"]},
        "2": {"morpho": ["rp--------"], "uri": ["20846"]},
    },
    "CETERVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c1312"]}},
    "CETOS": {"-": {"morpho": ["n-s---nn2-"], "uri": ["57276"]}},
    "CETRA": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "CETRATI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["104663"]}},
    "CETRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["104664"]}},
    "CETVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["40981"]}},
    "CEV": {
        "1": {"morpho": ["rp--------"], "uri": ["20847"]},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "CEVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c1320"]}},
    "CHAERE": {"-": {"morpho": ["----------"], "uri": []}},
    "CHALCIDICVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c1337"]}},
    "CHALYBS": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c1343"]}},
    "CHAOS": {
        "-": {"morpho": ["n-s---nn2g"], "uri": ["c1382"]},
        "N": {"morpho": ["n-s---nn2-"], "uri": []},
    },
    "CHARA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c1383"]}},
    "CHARACTERISMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["48576"]}},
    "CHARIDEMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CHARINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CHARIS": {
        "-": {"morpho": ["n-s----n-g"], "uri": []},
        "N": {"morpho": ["n-s----n--"], "uri": []},
    },
    "CHARTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c1396"]}},
    "CHARYBDIS": {"N": {"morpho": ["n-s---fn3i"], "uri": []}},
    "CHEIROTOMETOS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "CHELE": {
        "-": {"morpho": ["n-s---fn1g"], "uri": ["c1412"]},
        "N": {"morpho": ["n-p---fn1g"], "uri": []},
    },
    "CHELIDON": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["c1413"]},
        "N": {"morpho": ["n-s---nn3-"], "uri": []},
    },
    "CHELONIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c1420"]}},
    "CHELYDRVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c1421"]}},
    "CHELYS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c1422"]}},
    "CHERAGRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c1429"]}},
    "CHERSONESVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CHILIARCHES": {"-": {"morpho": ["n-s---mn1g"], "uri": ["c1440"]}},
    "CHILO": {"N": {"morpho": ["n-s---mn3-"], "uri": ["104752"]}},
    "CHIRAGRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["37825"]}},
    "CHIROGRAPHVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c1451"]}},
    "CHIRONOMON": {"-": {"morpho": ["n-s---mn3i"], "uri": ["c1454"]}},
    "CHIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s----n-g"], "uri": []},
    },
    "CHLAMYDATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c1462"]}},
    "CHLAMYS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c1463"]}},
    "CHLORVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CHOO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "CHORAGIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c1484"]}},
    "CHORAVLES": {"-": {"morpho": ["n-s---mn1g"], "uri": ["c1487"]}},
    "CHORDA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c1489"]}},
    "CHORDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["58352"]}},
    "CHOREA": {"-": {"morpho": ["n-s---fn1g"], "uri": ["c1493"]}},
    "CHOROBATES": {"-": {"morpho": ["n-s---mn1g"], "uri": ["c1502"]}},
    "CHOROGRAPHIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c1504"]}},
    "CHORVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c1506"]}},
    "CHRIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c1509"]}},
    "CHRISTIANVS": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "CHROMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["c1515"]}},
    "CHROMATICE": {"-": {"morpho": ["n-s---fn1g"], "uri": ["c1517"]}},
    "CHROMATICOS": {"-": {"morpho": ["aps---mn1g"], "uri": ["c1517"]}},
    "CHRYSAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "CHRYSOCOLLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c1547"]}},
    "CHRYSOGONVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CHRYSOLITHVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["56039"]}},
    "CIBARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c1569"]}},
    "CIBARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c1569"]}},
    "CIBATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c9987"]}},
    "CIBORIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c1575"]}},
    "CIBVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c1576"]}},
    "CIBYRA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CIBYRATA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CIBYRATICVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "CICADA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c1577"]}},
    "CICATRICOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c1580"]}},
    "CICATRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c1582"]}},
    "CICER": {"-": {"morpho": ["n-s---nn3-"], "uri": ["c1584"]}},
    "CICERO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "CICHORIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["55952"]}},
    "CICONIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c1595"]}},
    "CICVR": {"-": {"morpho": ["aps---an3-"], "uri": ["c1598"]}},
    "CICVTA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["c1600"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "CIDARIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["c1602"]}},
    "CIDO": {
        "1": {"morpho": ["v1spia--3-"], "uri": []},
        "2": {"morpho": ["v1spia--3-"], "uri": []},
    },
    "CIEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c1604"]}},
    "CILICES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CILICIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CILICIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c1606"]}},
    "CILIO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "CIMEX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c1621"]}},
    "CINAEDVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["c1631"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["c1631"]},
    },
    "CINCINNVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c1637"]}},
    "CINEFACTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c1642"]}},
    "CINERARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["c1644"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["c1644"]},
    },
    "CINESIS": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "CINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c1651"]}},
    "CINGVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c1652"]}},
    "CINGVLVM": {
        "-": {"morpho": ["n-s---nn2-"], "uri": ["c1652"]},
        "N": {"morpho": ["n-s---nn2-"], "uri": ["c1652"]},
    },
    "CINIFLO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c1653"]}},
    "CINIS": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c1654"]}},
    "CINNA": {"N": {"morpho": ["n-s---mn1-"], "uri": []}},
    "CINNAMVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c1662"]}},
    "CINNO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "CINNVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c1663"]}},
    "CINO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "CIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["35726"]}},
    "CIPIO": {"-": {"morpho": ["v1spia--3i"], "uri": []}},
    "CIPPVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c1666"]}},
    "CIRCA": {
        "1": {"morpho": ["rp--------"], "uri": ["20899"]},
        "2": {"morpho": ["p---------"], "uri": ["c1668"]},
    },
    "CIRCE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CIRCENSES": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "CIRCENSIS": {
        "-": {"morpho": ["aps---cn3i"], "uri": ["c1674"]},
        "A": {"morpho": ["aps---cn3i"], "uri": ["c1674"]},
    },
    "CIRCINATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c1676"]}},
    "CIRCINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c1677"]}},
    "CIRCINVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c1678"]}},
    "CIRCITER": {
        "1": {"morpho": ["rp--------"], "uri": ["c1681"]},
        "2": {"morpho": ["p---------"], "uri": []},
    },
    "CIRCITO": {
        "1": {
            "morpho": ["v1spia--1-", "v1spia--1-"],
            "uri": ["c1682", "c1682"],
        }
    },
    "CIRCVITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["20905"]}},
    "CIRCVITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["20906"]}},
    "CIRCVLARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c1688"]}},
    "CIRCVLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c1690"]}},
    "CIRCVLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132295"]}},
    "CIRCVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c1694"]}},
    "CIRCVM": {
        "1": {"morpho": ["rp--------"], "uri": ["20909"]},
        "2": {"morpho": ["p---------"], "uri": ["c1695"]},
    },
    "CIRCVMACTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c1696"]}},
    "CIRCVMACTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c1716"]}},
    "CIRCVMAGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c1698"]}},
    "CIRCVMAMPLECTOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["c1701"]}},
    "CIRCVMCAESVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c1713"]}},
    "CIRCVMCIDANEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c1711"]}},
    "CIRCVMCIDO": {
        "2": {"morpho": ["v1spia--3-"], "uri": ["c1712"]},
        "1": {"morpho": ["v1spia--3-"], "uri": ["c1712"]},
    },
    "CIRCVMCINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c1714"]}},
    "CIRCVMCIRCA": {"-": {"morpho": ["rp--------"], "uri": ["c1718"]}},
    "CIRCVMCISVS": {
        "C": {"morpho": ["aps---mn1-"], "uri": ["20917"]},
        "-": {"morpho": ["aps---mn1-"], "uri": ["20917"]},
    },
    "CIRCVMCLVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c1725"]}},
    "CIRCVMCOLO": {"2": {"morpho": ["v1spia--3-"], "uri": ["c1727"]}},
    "CIRCVMCVRRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c1730"]}},
    "CIRCVMCVRSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c1732"]}},
    "CIRCVMDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c1734"]}},
    "CIRCVMDVCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c1737"]}},
    "CIRCVMDVCTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c1738"]}},
    "CIRCVMEO": {"1": {"morpho": ["v1spia--4-"], "uri": ["c1741"]}},
    "CIRCVMFERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c1745"]}},
    "CIRCVMFIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c1746"]}},
    "CIRCVMFLECTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c1750"]}},
    "CIRCVMFLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c1753"]}},
    "CIRCVMFLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c1754"]}},
    "CIRCVMFLVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c1755"]}},
    "CIRCVMFODIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["c1756"]}},
    "CIRCVMFREMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c1762"]}},
    "CIRCVMFRICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c1763"]}},
    "CIRCVMFVNDO": {"2": {"morpho": ["v1spia--3-"], "uri": ["c8040"]}},
    "CIRCVMGEMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c1769"]}},
    "CIRCVMGREDIOR": {"-": {"morpho": ["v1spid--3i"], "uri": ["c1773"]}},
    "CIRCVMIACEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c1777"]}},
    "CIRCVMICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["c1778"]}},
    "CIRCVMLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c1789"]}},
    "CIRCVMLATRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c1792"]}},
    "CIRCVMLEGO": {"2": {"morpho": ["v1spia--3-"], "uri": ["c1795"]}},
    "CIRCVMLEVIGO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "CIRCVMLEVO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["c8796"]},
        "-": {"morpho": ["v1spia--1-"], "uri": ["c8796"]},
    },
    "CIRCVMLIGO": {"2": {"morpho": ["v1spia--1-"], "uri": ["c1797"]}},
    "CIRCVMLINO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c1798"]}},
    "CIRCVMLITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c1800"]}},
    "CIRCVMLOQVOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["c1802"]}},
    "CIRCVMLVCENS": {"-": {"morpho": ["aps---an3i"], "uri": ["97883"]}},
    "CIRCVMMETIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["c1809"]}},
    "CIRCVMMITTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c1811"]}},
    "CIRCVMMVGIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["c1813"]}},
    "CIRCVMMVNIO": {"2": {"morpho": ["v1spia--4-"], "uri": ["c1815"]}},
    "CIRCVMMVNITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c1816"]}},
    "CIRCVMPENDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c1828"]}},
    "CIRCVMPLECTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["20938"]}},
    "CIRCVMPLECTOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["c1832"]}},
    "CIRCVMPLVMBO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c1834"]}},
    "CIRCVMPONO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c1835"]}},
    "CIRCVMQVAQVE": {"-": {"morpho": ["rp--------"], "uri": ["c1841"]}},
    "CIRCVMRETIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["c1845"]}},
    "CIRCVMRODO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c1846"]}},
    "CIRCVMSAEPIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["c1850"]}},
    "CIRCVMSCINDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c1855"]}},
    "CIRCVMSCRIBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c1856"]}},
    "CIRCVMSCRIPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c1857"]}},
    "CIRCVMSCRIPTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c1858"]}},
    "CIRCVMSECO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c1859"]}},
    "CIRCVMSEDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c1861"]}},
    "CIRCVMSERO": {"3": {"morpho": ["v1spia--3-"], "uri": ["c1863"]}},
    "CIRCVMSESSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c1864"]}},
    "CIRCVMSIDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c1861"]}},
    "CIRCVMSILIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["c1866"]}},
    "CIRCVMSISTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c1867"]}},
    "CIRCVMSONO": {"1": {"morpho": ["v1spia--1-"], "uri": ["c1870"]}},
    "CIRCVMSONVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c1871"]}},
    "CIRCVMSPECTE": {"-": {"morpho": ["rp--------"], "uri": ["44688"]}},
    "CIRCVMSPECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c1875"]}},
    "CIRCVMSPECTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["c1885"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["20952"]},
    },
    "CIRCVMSPICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["c1878"]}},
    "CIRCVMSTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c1867"]}},
    "CIRCVMSTREPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c1888"]}},
    "CIRCVMTENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c1897"]}},
    "CIRCVMTENEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c1898"]}},
    "CIRCVMTERGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c1899"]}},
    "CIRCVMTERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c1901"]}},
    "CIRCVMTEXO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "CIRCVMTONDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["37254"]}},
    "CIRCVMTONO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c1905"]}},
    "CIRCVMVADO": {"2": {"morpho": ["v1spia--3-"], "uri": ["c1911"]}},
    "CIRCVMVAGOR": {
        "2": {"morpho": ["v1spid--1-"], "uri": ["c1912"]},
        "-": {"morpho": ["v1spid--1-"], "uri": ["c1912"]},
    },
    "CIRCVMVAGVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c1913"]}},
    "CIRCVMVALLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c1914"]}},
    "CIRCVMVECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["44425"]}},
    "CIRCVMVECTOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["c1917"]}},
    "CIRCVMVEHOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["c1918"]}},
    "CIRCVMVENIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["c1921"]}},
    "CIRCVMVERRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c1925"]}},
    "CIRCVMVERSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["105035"]}},
    "CIRCVMVERTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c1928"]}},
    "CIRCVMVOLITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c1934"]}},
    "CIRCVMVOLO": {"2": {"morpho": ["v1spia--1-"], "uri": ["c1935"]}},
    "CIRCVMVOLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c1936"]}},
    "CIRCVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["c1940"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["c1940"]},
    },
    "CIRRATI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["c1942"]}},
    "CIRRVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c1946"]}},
    "CIS": {"-": {"morpho": ["p---------"], "uri": ["c1949"]}},
    "CISIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c1953"]}},
    "CISTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c1964"]}},
    "CISTERNA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c1970"]}},
    "CISTERNINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c1971"]}},
    "CITATIM": {"-": {"morpho": ["rp--------"], "uri": ["c1977"]}},
    "CITATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["97888"]}},
    "CITE": {"-": {"morpho": ["rp--------"], "uri": ["51678"]}},
    "CITER": {"-": {"morpho": ["aps---mn1r"], "uri": ["c1980"]}},
    "CITERIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c1981"]}},
    "CITERIOR": {
        "-": {"morpho": ["aps---cn3-"], "uri": ["40853"]},
        "A": {"morpho": ["aps---cn3-"], "uri": ["40853"]},
    },
    "CITHARA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c1982"]}},
    "CITHARISTA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["c1984"]}},
    "CITHAROEDVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c1987"]}},
    "CITO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["c1991"]},
        "2": {"morpho": ["rp--------"], "uri": ["20971"]},
    },
    "CITRA": {
        "1": {"morpho": ["rp--------"], "uri": ["c1993"]},
        "2": {"morpho": ["p---------"], "uri": []},
    },
    "CITREVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c1999"]}},
    "CITRO": {"-": {"morpho": ["rp--------"], "uri": ["c2001"]}},
    "CITRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c2003"]}},
    "CITRVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["c2003"]}},
    "CITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["53278"]}},
    "CIVICA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["c2005"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "CIVICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c2005"]}},
    "CIVILIS": {
        "-": {"morpho": ["aps---cn3i"], "uri": ["c2006"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "CIVILITER": {"-": {"morpho": ["rp--------"], "uri": ["43724"]}},
    "CIVIS": {"-": {"morpho": ["n-s---cn3i"], "uri": ["c2008"]}},
    "CIVITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2009"]}},
    "CLADES": {"-": {"morpho": ["n-s---fn3i"], "uri": ["c2012"]}},
    "CLAM": {
        "1": {"morpho": ["rp--------"], "uri": ["32681"]},
        "2": {"morpho": ["p---------"], "uri": ["c2013"]},
    },
    "CLAMITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2018"]}},
    "CLAMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2019"]}},
    "CLAMOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c2020"]}},
    "CLAMOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c2023"]}},
    "CLANCVLVM": {"-": {"morpho": ["rp--------"], "uri": ["c7025"]}},
    "CLANDESTINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c2026"]}},
    "CLANGOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c2028"]}},
    "CLARE": {"-": {"morpho": ["rp--------"], "uri": ["20984"]}},
    "CLAREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c2030"]}},
    "CLARESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c2030"]}},
    "CLARICITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["97894"]}},
    "CLARIGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2038"]}},
    "CLARISONVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c2039"]}},
    "CLARITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2041"]}},
    "CLARITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2042"]}},
    "CLARO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2044"]}},
    "CLARVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["c2046"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "CLASSIARII": {"-": {"morpho": ["n-p---mn2-"], "uri": ["c2048"]}},
    "CLASSIARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c2048"]}},
    "CLASSICI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["97897"]}},
    "CLASSICVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c2050"]}},
    "CLASSICVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["c2050"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["c2050"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["c2050"]},
    },
    "CLASSIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["c2051"]}},
    "CLATHRATVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "CLATHRI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["105088"]}},
    "CLAVA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c2071"]}},
    "CLAVDICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2060"]}},
    "CLAVDIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "CLAVDO": {
        "1": {"morpho": ["v1spia--3-"], "uri": ["c2058", "c2058"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["c2058", "c2058"]},
    },
    "CLAVDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c2063"]}},
    "CLAVICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c9930", "c2067"]}},
    "CLAVIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["c2071"]}},
    "CLAVSTRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c2077"]}},
    "CLAVSVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c2078"]}},
    "CLAVSVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["39905"]}},
    "CLAVSVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CLAVVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c2080"]}},
    "CLAVVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c2071"]}},
    "CLEMENS": {
        "-": {"morpho": ["aps---an3i"], "uri": ["c2085"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "CLEMENTER": {"-": {"morpho": ["rp--------"], "uri": ["21009"]}},
    "CLEMENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c2086"]}},
    "CLEOMENES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CLEPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c2089"]}},
    "CLEPSYDRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c2090"]}},
    "CLIENS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["c2103"]}},
    "CLIENTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c2104"]}},
    "CLIENTELA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c2105"]}},
    "CLIENTVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c2106"]}},
    "CLIMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["c2107"]}},
    "CLIMACHIAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CLIMACIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2108"]}},
    "CLINAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["c2113"]}},
    "CLINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2117"]}},
    "CLIPEATVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["97906"]},
        "2": {"morpho": ["n-s---mn2-"], "uri": ["97906"]},
        "1": {"morpho": ["n-s---mn2-"], "uri": ["97906"]},
    },
    "CLIPEO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2121"]}},
    "CLIPEVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c2123"]}},
    "CLITELLAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["c2124"]}},
    "CLITELLARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c2125"]}},
    "CLIVA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["105131"]}},
    "CLIVOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c2131"]}},
    "CLIVVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c2132"]}},
    "CLOACA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c2133"]}},
    "CLODIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "CLOELIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CLOELIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "CLVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["49269"]}},
    "CLVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c2145"]}},
    "CLVNIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["c2149"]}},
    "CLVPEA": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "CLVSINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "CNEMIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2165"]}},
    "CNIDII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CNODAX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c2172"]}},
    "COACERVATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2175"]}},
    "COACERVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2176"]}},
    "COACTA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["c9997"]}},
    "COACTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2185"]}},
    "COACTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c2186"]}},
    "COACTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c9997"]}},
    "COADDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c2189"]}},
    "COAEDIFICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2198"]}},
    "COAEQVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2201"]}},
    "COAGMENTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2211"]}},
    "COAGMENTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2212"]}},
    "COAGMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c2213"]}},
    "COAGVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2217"]}},
    "COAGVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c2218"]}},
    "COALESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c2219"]}},
    "COANGVSTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2224"]}},
    "COARCTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["49840"]}},
    "COARCTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["30538"]}},
    "COAREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["105208"]}},
    "COARGVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c2229"]}},
    "COARTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2233"]}},
    "COARTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2235"]}},
    "COAXATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2239"]}},
    "COAXO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2241"]}},
    "COCCEIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "COCCINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c2245"]}},
    "COCCVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c2247"]}},
    "COCHLEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c2252"]}},
    "COCLES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "COCTILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c7248"]}},
    "COCTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c7249"]}},
    "COCTVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c9933"]}},
    "COCTVRNIX": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "COCVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c7254"]}},
    "CODEX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["21050"]}},
    "CODICILLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c7260"]}},
    "COELIA": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "COELIACVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c2187"]}},
    "COELIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "COEMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c2268"]}},
    "COEO": {"1": {"morpho": ["v1spia--1-"], "uri": ["c2280"]}},
    "COEPIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["c2281"]}},
    "COEPTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2285"]}},
    "COEPTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c9999"]}},
    "COEPTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c9999"]}},
    "COERCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c2287"]}},
    "COERCITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2288"]}},
    "COETVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c2292"]}},
    "COGITABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c2296"]}},
    "COGITATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2301"]}},
    "COGITATVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c9859"]}},
    "COGITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c4657"]}},
    "COGNATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2306"]}},
    "COGNATVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["c2307"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["c2307"]},
    },
    "COGNITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2308"]}},
    "COGNITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c2311"]}},
    "COGNITVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["c9860"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["21066"]},
    },
    "COGNOBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c2314"]}},
    "COGNOMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["c2315"]}},
    "COGNOMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c2316"]}},
    "COGNOSCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c2322"]}},
    "COGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c2323"]}},
    "COHAERENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c9935"]}},
    "COHAEREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c2327"]}},
    "COHAERESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c1535"]}},
    "COHERES": {"-": {"morpho": ["n-s---cn3-"], "uri": ["97920"]}},
    "COHIBEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c2330"]}},
    "COHONESTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2333"]}},
    "COHORS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["c2335"]}},
    "COHORTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2338"]}},
    "COHORTOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["c2341"]}},
    "COI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "COICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["54161"]}},
    "COINQVINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2349"]}},
    "COITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c2292"]}},
    "COLAPHVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["c2354"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["c2354"]},
    },
    "COLLABEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": []}},
    "COLLABEFIO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c2367"]}},
    "COLLABOR": {"2": {"morpho": ["v1spid--3-"], "uri": ["c2369"]}},
    "COLLACTEVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "COLLAEVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["105348"]}},
    "COLLATIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "COLLATICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c2382"]}},
    "COLLATINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "COLLATINVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "COLLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2383"]}},
    "COLLATRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2386"]}},
    "COLLATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c9863"]}},
    "COLLAVDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2390"]}},
    "COLLAXO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2391"]}},
    "COLLECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2397"]}},
    "COLLECTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["c9936"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["97926"]},
    },
    "COLLEGA": {
        "-": {"morpho": ["n-s---mn1-"], "uri": ["c2400"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "COLLEGIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c2405"]}},
    "COLLEVO": {"2": {"morpho": ["v1spia--1-"], "uri": ["c2411"]}},
    "COLLIBERTVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c4482"]}},
    "COLLIBET": {"-": {"morpho": ["v1spia--2-"], "uri": []}},
    "COLLIBRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2413"]}},
    "COLLICIAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["c2416"]}},
    "COLLICIARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c2417"]}},
    "COLLIDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c2421"]}},
    "COLLIGO": {
        "2": {"morpho": ["v1spia--1-"], "uri": ["c2425"]},
        "3": {"morpho": ["v1spia--3-"], "uri": ["c2426"]},
    },
    "COLLINO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c2432"]}},
    "COLLINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": ["c2433"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["c2433"]},
    },
    "COLLIQVIARIA": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "COLLIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["c2438"]}},
    "COLLOCATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2440"]}},
    "COLLOCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2441"]}},
    "COLLOQVIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c2445"]}},
    "COLLOQVOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["c2446"]}},
    "COLLVCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c2447"]}},
    "COLLVCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2448"]}},
    "COLLVCTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2449"]}},
    "COLLVCTOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["c2451"]}},
    "COLLVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c2453"]}},
    "COLLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c2456"]}},
    "COLLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c2458"]}},
    "COLLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["46373"]}},
    "COLLVSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2461"]}},
    "COLLVSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c2462"]}},
    "COLLVSTRO": {"2": {"morpho": ["v1spia--1-"], "uri": ["c2465"]}},
    "COLLVVIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["c2469"]}},
    "COLLVVIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2470"]}},
    "COLLYBVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c2473"]}},
    "COLLYRIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c2477"]}},
    "COLO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["c2479"]},
        "2": {"morpho": ["v1spia--3-"], "uri": []},
    },
    "COLOCASIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c2486"]}},
    "COLOCYNTHA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c2487"]}},
    "COLOEPHIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["52733"]}},
    "COLONIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["c2494"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "COLONVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["c2497"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["97931"]},
    },
    "COLOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c2500"]}},
    "COLORATVS": {
        "C": {"morpho": ["aps---mn1-"], "uri": ["21131"]},
        "-": {"morpho": ["aps---mn1-"], "uri": ["21131"]},
    },
    "COLORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2507"]}},
    "COLOSSICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c2509"]}},
    "COLOSSVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c2510"]}},
    "COLVBER": {"-": {"morpho": ["n-s---mn2r"], "uri": ["c2515"]}},
    "COLVBRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c2515"]}},
    "COLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c2520"]}},
    "COLVMBA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c2521"]}},
    "COLVMBARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c2523"]}},
    "COLVMBINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c2526"]}},
    "COLVMBOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["c2527"]}},
    "COLVMBVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c2521"]}},
    "COLVMELLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c2531"]}},
    "COLVMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["c2533"]}},
    "COLVMNA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c2535"]}},
    "COLVMNARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c2538"]}},
    "COLVMNATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c2540"]}},
    "COLVRNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c2543"]}},
    "COLVS": {"-": {"morpho": ["n-s---cn2-"], "uri": []}},
    "COMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c2551"]}},
    "COMANS": {"-": {"morpho": ["aps---an3i"], "uri": ["c2553"]}},
    "COMATVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["29020"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["29020"]},
    },
    "COMBIBO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["c2561"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["c2560"]},
    },
    "COMBVRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c2566"]}},
    "COMEDO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["c2570"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["c2569"]},
    },
    "COMES": {"-": {"morpho": ["n-s---cn3-"], "uri": ["c2572"]}},
    "COMETES": {"-": {"morpho": ["n-s---mn1g"], "uri": ["c2576"]}},
    "COMICE": {"-": {"morpho": ["rp--------"], "uri": ["29733"]}},
    "COMICVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["c2577"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["c2577"]},
    },
    "COMINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "COMINVS": {"-": {"morpho": ["rp--------"], "uri": ["46052"]}},
    "COMIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c2579"]}},
    "COMISSABVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c2580"]}},
    "COMISSATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2581"]}},
    "COMISSOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["c2584"]}},
    "COMITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2586"]}},
    "COMITATVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["c2588"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["45884"]},
    },
    "COMITER": {"-": {"morpho": ["rp--------"], "uri": ["21157"]}},
    "COMITIA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["50262"]}},
    "COMITIALIS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["105464"]},
        "2": {"morpho": ["aps---cn3i"], "uri": ["c2590"]},
    },
    "COMITIATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c9866"]}},
    "COMITIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2592"]}},
    "COMITIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c2593"]}},
    "COMITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["21160"]}},
    "COMITOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["c2595"]}},
    "COMMACVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2600"]}},
    "COMMADEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c2601"]}},
    "COMMEATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c9867"]}},
    "COMMEDITOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["c2627"]}},
    "COMMEMORATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2634"]}},
    "COMMEMORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2637"]}},
    "COMMENDATICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c2639"]}},
    "COMMENDATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2640"]}},
    "COMMENDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c4699"]}},
    "COMMENSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c9868"]}},
    "COMMENTARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c2651"]}},
    "COMMENTARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["c2651"]},
        "2": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "COMMENTICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c2654"]}},
    "COMMENTOR": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["c2657"]},
        "2": {"morpho": ["v1spid--1-"], "uri": ["c2656"]},
    },
    "COMMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c2636"]}},
    "COMMEO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2658"]}},
    "COMMERCIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c2661"]}},
    "COMMERCOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["c2662"]}},
    "COMMEREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c2663"]}},
    "COMMEREOR": {"-": {"morpho": ["v1spid--2-"], "uri": ["132321"]}},
    "COMMERITVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "COMMIGRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2667"]}},
    "COMMIGRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2668"]}},
    "COMMILITIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c2670"]}},
    "COMMILITO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["c2672"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["c2671"]},
    },
    "COMMINATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2674"]}},
    "COMMINGO": {"-": {"morpho": ["v1spia--3-"], "uri": [None, "c2628"]}},
    "CO": {"-": {"morpho": ["v1spid--3-"], "uri": []}},
    "COMMINOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["c2678"]}},
    "COMMINVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c2679"]}},
    "COMMINVS": {"-": {"morpho": ["rp--------"], "uri": ["c2680"]}},
    "COMMIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["105521"]}},
    "COMMISCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c2682"]}},
    "COMMISSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2690"]}},
    "COMMISSVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c2692"]}},
    "COMMISSVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c2687"]}},
    "COMMITTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c2695"]}},
    "COMMODE": {"-": {"morpho": ["rp--------"], "uri": ["21206"]}},
    "COMMODITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2703"]}},
    "COMMODO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["c2704"]},
        "2": {"morpho": ["rp--------"], "uri": ["97946"]},
    },
    "COMMODVLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2705"]}},
    "COMMODVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["c2708"]},
        "2": {"morpho": ["rp--------"], "uri": ["21210"]},
    },
    "COMMODVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c2708"]}},
    "COMMOLIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["c2710"]}},
    "COMMONEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["c2712"]}},
    "COMMONEFIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2713"]}},
    "COMMONEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c2714"]}},
    "COMMONSTRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2719"]}},
    "COMMORDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c2722"]}},
    "COMMORIOR": {"-": {"morpho": ["v1spid--3i"], "uri": ["c2723"]}},
    "COMMOROR": {"1": {"morpho": ["v1spid--1-"], "uri": ["c2724"]}},
    "COMMOTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2729"]}},
    "COMMOTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["97947"]}},
    "COMMOVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c2733"]}},
    "COMMVNE": {
        "C": {"morpho": ["n-s---nn3i"], "uri": ["c4577"]},
        "-": {"morpho": ["n-s---nn3i"], "uri": ["c4577"]},
    },
    "COMMVNICATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2741"]}},
    "COMMVNICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2745"]}},
    "COMMVNIO": {
        "1": {"morpho": ["n-s---fn3-"], "uri": ["c2747"]},
        "2": {"morpho": ["v1spia--4-"], "uri": ["c2746"]},
    },
    "COMMVNIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c4552"]}},
    "COMMVNITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2749"]}},
    "COMMVNITER": {"-": {"morpho": ["rp--------"], "uri": ["36011"]}},
    "COMMVNITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2750"]}},
    "COMMVTABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c2754"]}},
    "COMMVTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2755"]}},
    "COMMVTATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c9871"]}},
    "COMMVTO": {"2": {"morpho": ["v1spia--1-"], "uri": ["c2756"]}},
    "COMO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["c2758"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["c2757"]},
    },
    "COMOEDIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c2759"]}},
    "COMOEDVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["c2762"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["c2762"]},
    },
    "COMPACTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2768"]}},
    "COMPACTVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c9941"]}},
    "COMPACTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["21238"]}},
    "COMPAGES": {"-": {"morpho": ["n-s---fn3i"], "uri": ["c2773"]}},
    "COMPAGO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c7776"]}},
    "COMPAR": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["97952"]},
        "2": {"morpho": ["aps---an3i"], "uri": ["c2779"]},
    },
    "COMPARATIO": {
        "1": {"morpho": ["n-s---fn3-"], "uri": ["c4786"]},
        "2": {"morpho": ["n-s---fn3-"], "uri": ["c4786"]},
    },
    "COMPARATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c9872"]}},
    "COMPAREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c2785"]}},
    "COMPARO": {
        "2": {"morpho": ["v1spia--1-"], "uri": ["c4812"]},
        "3": {"morpho": ["v1spia--1-"], "uri": ["c4812"]},
    },
    "COMPEDIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["c2806"]}},
    "COMPELLO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["c2811"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["c2810"]},
    },
    "COMPENDIARIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c2812"]}},
    "COMPENDIARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c2812"]}},
    "COMPENDIARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c2812"]}},
    "COMPENDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c2815"]}},
    "COMPENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c2816"]}},
    "COMPENSATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2817"]}},
    "COMPENSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2819"]}},
    "COMPERENDINATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2823"]}},
    "COMPERENDINATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c9873"]}},
    "COMPERENDINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2824"]}},
    "COMPERIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["c2827"]}},
    "COMPERTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["41846"]}},
    "COMPESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c2832"]}},
    "COMPETITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c2834"]}},
    "COMPETO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c2836"]}},
    "COMPILATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2837"]}},
    "COMPILO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["c4816"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["c4816"]},
    },
    "COMPINGO": {
        "1": {"morpho": ["v1spia--3-"], "uri": ["c2841"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["c2841"]},
    },
    "COMPITVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c2846"]}},
    "COMPLANO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2851"]}},
    "COMPLECTO": {"2": {"morpho": ["v1spia--3-"], "uri": ["44296"]}},
    "COMPLECTOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["c2854"]}},
    "COMPLEMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c2855"]}},
    "COMPLEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c2856"]}},
    "COMPLETVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["51622"]}},
    "COMPLEXIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2863"]}},
    "COMPLEXVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c9874"]}},
    "COMPLICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2868"]}},
    "COMPLODO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c2869"]}},
    "COMPLORATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2870"]}},
    "COMPLORATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c9875"]}},
    "COMPLORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2871"]}},
    "COMPLVRES": {
        "1": {"morpho": ["n-s----n3-"], "uri": []},
        "2": {"morpho": ["acp---cn3i"], "uri": ["c2873"]},
    },
    "COMPLVRIES": {"-": {"morpho": ["rp--------"], "uri": ["105626"]}},
    "COMPLVVIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c2879"]}},
    "COMPONO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c2881"]}},
    "COMPORTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2882"]}},
    "COMPORTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2885"]}},
    "COMPOS": {"-": {"morpho": ["aps---an3-"], "uri": ["c2886"]}},
    "COMPOSITE": {"-": {"morpho": ["rp--------"], "uri": ["50393"]}},
    "COMPOSITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2888"]}},
    "COMPOSITVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c2860"]}},
    "COMPOSITVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c2910"]}},
    "COMPOSITVS": {
        "O": {"morpho": ["aps---mn1-"], "uri": ["57352"]},
        "-": {"morpho": ["aps---mn1-"], "uri": ["57352"]},
    },
    "COMPRECOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["c2894"]}},
    "COMPREHENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c2895"]}},
    "COMPREHENSIBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c2896"]}},
    "COMPREHENSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2897"]}},
    "COMPRESSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2901"]}},
    "COMPRESSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2902"]}},
    "COMPRESSVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["c9876"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["21291"]},
    },
    "COMPRIMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c2904"]}},
    "COMPROBO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2909"]}},
    "COMPROMISSVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c2985"]}},
    "COMPSANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "COMPSANVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "COMPTE": {"-": {"morpho": ["rp--------"], "uri": ["45468"]}},
    "COMPTVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c2917"]}},
    "COMPTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["c9877"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["21296"]},
    },
    "COMPVNGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c2925"]}},
    "COMPVTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2929"]}},
    "COMPVTATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c2930"]}},
    "COMPVTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2932"]}},
    "COMPVTRESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c2933"]}},
    "CONAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["c2937"]}},
    "CONARACHNE": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "CONATA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["97963"]}},
    "CONATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c9878"]}},
    "CONCAEDES": {"-": {"morpho": ["n-p---fn3i"], "uri": ["c2945"]}},
    "CONCALEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["c2946"]}},
    "CONCALEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c2949"]}},
    "CONCALESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c2949"]}},
    "CONCALLEO": {"-": {"morpho": ["v1spia--2-"], "uri": []}},
    "CONCAMERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2953"]}},
    "CONCAMERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2954"]}},
    "CONCAVO": {"1": {"morpho": ["v1spia--1-"], "uri": ["c2966"]}},
    "CONCAVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c2967"]}},
    "CONCEDO": {"1": {"morpho": ["v1spia--3-"], "uri": ["c2968"]}},
    "CONCELEBRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2969"]}},
    "CONCENTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2975"]}},
    "CONCENTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c9880"]}},
    "CONCEPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2980"]}},
    "CONCEPTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2983"]}},
    "CONCEPTVS": {
        "-": {"morpho": ["n-s---mn4-"], "uri": ["c9881"]},
        "2": {"morpho": ["n-s---mn4-"], "uri": ["c9881"]},
        "1": {"morpho": ["n-s---mn4-"], "uri": ["c9881"]},
    },
    "CONCERNO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c2986"]}},
    "CONCERTATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c2990"]}},
    "CONCERTO": {"1": {"morpho": ["v1spia--1-"], "uri": ["c2992"]}},
    "CONCESSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2994"]}},
    "CONCESSVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["97966"]}},
    "CONCESSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c9882"]}},
    "CONCHA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c2998"]}},
    "CONCHIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["c3002"]}},
    "CONCHYLIATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c3007"]}},
    "CONCHYLIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c3008"]}},
    "CONCIDO": {
        "1": {"morpho": ["v1spia--3-"], "uri": ["c3010", "c3011", "c3010"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["c3010", "c3011", "c3010"]},
    },
    "CONCIEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c3013"]}},
    "CONCILIABVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c3014"]}},
    "CONCILIATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3015"]}},
    "CONCILIATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c3016"]}},
    "CONCILIATRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3018"]}},
    "CONCILIATVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c9944"]}},
    "CONCILIATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c9883"]}},
    "CONCILIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3020"]}},
    "CONCILIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c3021"]}},
    "CONCINNITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3028"]}},
    "CONCINNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3030"]}},
    "CONCINNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c3031"]}},
    "CONCINO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3032"]}},
    "CONCIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["c3013"]}},
    "CONCIPIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["c3034"]}},
    "CONCISVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c9946"]}},
    "CONCISVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["21350"]}},
    "CONCITAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c3038"]}},
    "CONCITATE": {"-": {"morpho": ["rp--------"], "uri": ["97969"]}},
    "CONCITATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3039"]}},
    "CONCITATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c3040"]}},
    "CONCITATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["21355"]}},
    "CONCITO": {"1": {"morpho": ["v1spia--1-"], "uri": ["c3042"]}},
    "CONCITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c3043"]}},
    "CONCLAMATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3045"]}},
    "CONCLAMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3047"]}},
    "CONCLAVE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["c3049"]}},
    "CONCLVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3051"]}},
    "CONCLVSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3052"]}},
    "CONCLVSVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c9947"]}},
    "CONCLVSVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["c9885"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["21365"]},
    },
    "CONCOLOR": {"-": {"morpho": ["aps---an3-"], "uri": ["c3057"]}},
    "CONCOQVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3060"]}},
    "CONCORDIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["c3063"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["c3063"]},
    },
    "CONCORDITER": {"-": {"morpho": ["rp--------"], "uri": ["32568"]}},
    "CONCORDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3066"]}},
    "CONCORS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c3071"]}},
    "CONCREDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3074"]}},
    "CONCREMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3077"]}},
    "CONCREPITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3080"]}},
    "CONCREPO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3081"]}},
    "CONCRESCENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c9949"]}},
    "CONCRESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3082"]}},
    "CONCRETVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["c9886"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["21377"]},
    },
    "CONCRISPO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3085"]}},
    "CONCRVCIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3088"]}},
    "CONCVBINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c1050"]}},
    "CONCVBINVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c1052"]}},
    "CONCVBITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c9887"]}},
    "CONCVBIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c3098"]}},
    "CONCVLCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3101"]}},
    "CONCVMBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3103"]}},
    "CONCVPIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["c2343"]}},
    "CONCVPISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3105"]}},
    "CONCVPITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c3109"]}},
    "CONCVRRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3113"]}},
    "CONCVRSATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3114"]}},
    "CONCVRSATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c3115"]}},
    "CONCVRSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3117"]}},
    "CONCVRSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3119"]}},
    "CONCVRSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c9889"]}},
    "CONCVSSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c9890"]}},
    "CONCVTIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["c3125"]}},
    "CONDECORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3132"]}},
    "CONDELIQVESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3137"]}},
    "CONDEMNATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c3140"]}},
    "CONDEMNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3141"]}},
    "CONDENSEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c6143"]}},
    "CONDENSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3143"]}},
    "CONDENSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c3145"]}},
    "CONDEPSO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3146"]}},
    "CONDICIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3151"]}},
    "CONDICO": {"2": {"morpho": ["v1spia--3-"], "uri": ["c3154"]}},
    "CONDIGNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c3159"]}},
    "CONDIMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c3161"]}},
    "CONDIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["c3162"]}},
    "CONDISCIPVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c3163"]}},
    "CONDISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3165"]}},
    "CONDITIVVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c3170"]}},
    "CONDITIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c3170"]}},
    "CONDITOR": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["c3172", "c3171"]},
        "2": {"morpho": ["n-s---mn3-"], "uri": ["c3172", "c3171"]},
    },
    "CONDITVRA": {
        "1": {"morpho": ["n-s---fn1-"], "uri": ["c3169"]},
        "2": {"morpho": ["n-s---fn1-"], "uri": ["c3169"]},
    },
    "CONDITVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["c9892", "c9891"]},
        "2": {"morpho": ["n-s---mn4-"], "uri": ["c9892", "c9891"]},
        "3": {"morpho": ["aps---mn1-"], "uri": ["105823"]},
        "4": {"morpho": ["aps---mn1-"], "uri": ["105823"]},
    },
    "CONDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3174"]}},
    "CONDOCEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["c3175"]}},
    "CONDOLEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c3178"]}},
    "CONDOLESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3178"]}},
    "CONDONATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3180"]}},
    "CONDONO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3182"]}},
    "CONDORMIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["c3183"]}},
    "CONDVCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3186"]}},
    "CONDVCTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3189"]}},
    "CONDVCTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c3191"]}},
    "CONDVCTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c9893"]}},
    "CONDVPLICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3196"]}},
    "CONDVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3198"]}},
    "CONECTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3201"]}},
    "CONFARREATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3212"]}},
    "CONFARREO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3213"]}},
    "CONFECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3215"]}},
    "CONFECTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c3216"]}},
    "CONFERCIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["c3220"]}},
    "CONFERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3222"]}},
    "CONFERTIM": {"-": {"morpho": ["rp--------"], "uri": ["c3224"]}},
    "CONFERTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["45375"]}},
    "CONFERVEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["c3226"]}},
    "CONFERVESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3227"]}},
    "CONFESSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3228"]}},
    "CONFESSVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c3250"]}},
    "CONFESSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["32047"]}},
    "CONFESTIM": {"-": {"morpho": ["rp--------"], "uri": ["c3231"]}},
    "CONFICIENS": {"-": {"morpho": ["aps---an3i"], "uri": ["97982"]}},
    "CONFICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["c3234"]}},
    "CONFIDENS": {"-": {"morpho": ["aps---an3i"], "uri": ["97983"]}},
    "CONFIDENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c9953"]}},
    "CONFIDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3241"]}},
    "CONFIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3242"]}},
    "CONFINE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["21448"]}},
    "CONFINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3247"]}},
    "CONFINIS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["105866"]},
        "2": {"morpho": ["aps---cn3i"], "uri": ["c3249"]},
    },
    "CONFINIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c3249"]}},
    "CONFIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["c3235"]}},
    "CONFIRMATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3251"]}},
    "CONFIRMATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["21454"]}},
    "CONFIRMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3256"]}},
    "CONFITEOR": {"-": {"morpho": ["v1spid--2-"], "uri": ["c3260"]}},
    "CONFLAGRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3267"]}},
    "CONFLICTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3278"]}},
    "CONFLICTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c9896"]}},
    "CONFLIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3281"]}},
    "CONFLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3282"]}},
    "CONFLVENS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["21470"]}},
    "CONFLVENTES": {"N": {"morpho": ["n-p---mn3i"], "uri": []}},
    "CONFLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3286"]}},
    "CONFODIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["c3289"]}},
    "CONFORMATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3298"]}},
    "CONFORMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3301"]}},
    "CONFORNICATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3302"]}},
    "CONFORNICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3303"]}},
    "CONFRAGOSVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["35298"]}},
    "CONFRAGOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c3310"]}},
    "CONFRICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3316"]}},
    "CONFRINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3318"]}},
    "CONFRIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3319"]}},
    "CONFVGIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["c3323"]}},
    "CONFVLCIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["c3326"]}},
    "CONFVNDO": {"2": {"morpho": ["v1spia--3-"], "uri": ["c3328"]}},
    "CONFVSE": {"-": {"morpho": ["rp--------"], "uri": ["21486"]}},
    "CONFVSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3333"]}},
    "CONFVSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["21488"]}},
    "CONFVTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3337", "c3273"]}},
    "CONFVTVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3338"]}},
    "CONGELO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3343"]}},
    "CONGEMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3346"]}},
    "CONGERIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["c3353"]}},
    "CONGERO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["105933"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["c3360"]},
    },
    "CONGESTICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c3363"]}},
    "CONGESTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3365"]}},
    "CONGESTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["c9898"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["97988"]},
    },
    "CONGIARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c3368"]}},
    "CONGIARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c3368"]}},
    "CONGIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c3369"]}},
    "CONGLOBATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3373"]}},
    "CONGLOBO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3374"]}},
    "CONGLOMERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3376"]}},
    "CONGLVTINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3379"]}},
    "CONGREDIOR": {"-": {"morpho": ["v1spid--3i"], "uri": ["c3386"]}},
    "CONGREGATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3390"]}},
    "CONGREGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3393"]}},
    "CONGRESSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c9900"]}},
    "CONGRVENS": {"-": {"morpho": ["aps---an3i"], "uri": ["21512"]}},
    "CONGRVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3400"]}},
    "CONGRVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c3401"]}},
    "CONICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["c3404"]}},
    "CONICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c3405"]}},
    "CONIECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3410"]}},
    "CONIECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3411"]}},
    "CONIECTVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c3394"]}},
    "CONIECTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c9901"]}},
    "CONIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["c3415"]}},
    "CONIGER": {"-": {"morpho": ["aps---mn1r"], "uri": ["c3416"]}},
    "CONISTERIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c3417"]}},
    "CONIVGALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c3422"]}},
    "CONIVGATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c3424"]}},
    "CONIVGIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c3426"]}},
    "CONIVGVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c3429"]}},
    "CONIVNCTE": {"-": {"morpho": ["rp--------"], "uri": ["54593"]}},
    "CONIVNCTIM": {"-": {"morpho": ["rp--------"], "uri": ["c3431"]}},
    "CONIVNCTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3432"]}},
    "CONIVNCTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c9902"]}},
    "CONIVNCTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["37590"]}},
    "CONIVNGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3435"]}},
    "CONIVOLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c3437"]}},
    "CONIVRATI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["97995"]}},
    "CONIVRATIM": {"-": {"morpho": ["rp--------"], "uri": []}},
    "CONIVRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3438"]}},
    "CONIVRATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c3439"]}},
    "CONIVRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["41900"]}},
    "CONIVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3440"]}},
    "CONIVX": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "CONNECTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["53428"]}},
    "CONNEXVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["30385"]},
        "2": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "CONNITOR": {"2": {"morpho": ["v1spid--3-"], "uri": ["c3418"]}},
    "CONNIVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c3420"]}},
    "CONNVBIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["43457"]}},
    "CONOPEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["39667"]}},
    "CONOPIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c3445"]}},
    "CONOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["c3446"]}},
    "CONQVASSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3451"]}},
    "CONQVEROR": {"-": {"morpho": ["v1spid--3-"], "uri": ["c3452"]}},
    "CONQVESTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3453"]}},
    "CONQVIESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3454"]}},
    "CONQVIRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3456"]}},
    "CONQVISITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3457"]}},
    "CONQVISITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c3458"]}},
    "CONQVISITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["35206"]}},
    "CONSAEPTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c3230"]}},
    "CONSAEPTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["49387"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["105991"]},
    },
    "CONSALVTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3462"]}},
    "CONSALVTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3463"]}},
    "CONSANGVINEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c3465"]}},
    "CONSANGVINEI": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "CONSANGVINEVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": ["c3465"]},
    },
    "CONSANGVINITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3466"]}},
    "CONSARIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["c3470"]}},
    "CONSCELERATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["21533"]}},
    "CONSCELERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3472"]}},
    "CONSCENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3473"]}},
    "CONSCIENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c9958"]}},
    "CONSCIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["c2403"]}},
    "CONSCISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3476"]}},
    "CONSCIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c3478"]}},
    "CONSCRIBILLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3480"]}},
    "CONSCRIBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3481"]}},
    "CONSCRIPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3482"]}},
    "CONSECO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3484"]}},
    "CONSECRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3486"]}},
    "CONSECRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["46278"]}},
    "CONSECRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3489"]}},
    "CONSECTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3492"]}},
    "CONSECTOR": {"2": {"morpho": ["v1spid--1-"], "uri": ["c3496"]}},
    "CONSECVE": {"-": {"morpho": ["rp--------"], "uri": ["40243"]}},
    "CONSEDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3500"]}},
    "CONSENESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3506"]}},
    "CONSENSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3508"]}},
    "CONSENSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c9905"]}},
    "CONSENTANEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c3509"]}},
    "CONSENTIENS": {"-": {"morpho": ["aps---an3i"], "uri": ["98003"]}},
    "CONSENTIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["c3510"]}},
    "CONSEQVENS": {
        "1": {"morpho": ["n-s---nn3-"], "uri": ["98004"]},
        "2": {"morpho": ["aps---an3-"], "uri": ["45714"]},
    },
    "CONSEQVOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["c3514"]}},
    "CONSERO": {
        "2": {"morpho": ["v1spia--3-"], "uri": ["c3517"]},
        "3": {"morpho": ["v1spia--3-"], "uri": ["c3517"]},
    },
    "CONSERVATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3522"]}},
    "CONSERVATOR": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["c3523"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "CONSERVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3526"]}},
    "CONSERVVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["c4652"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["98005"]},
    },
    "CONSESSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c6523"]}},
    "CONSIDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["106040"]}},
    "CONSIDERANTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c9960"]}},
    "CONSIDERATE": {"-": {"morpho": ["rp--------"], "uri": ["21564"]}},
    "CONSIDERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3530"]}},
    "CONSIDERATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["21565"]}},
    "CONSIDERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3532"]}},
    "CONSIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CONSIDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c6524"]}},
    "CONSIGNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c6528"]}},
    "CONSILIARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["c6530"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["c6530"]},
    },
    "CONSILIOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["c3536"]}},
    "CONSILIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c3538"]}},
    "CONSIMILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c3539"]}},
    "CONSIPIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["c3541"]}},
    "CONSISTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3543"]}},
    "CONSITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c3548"]}},
    "CONSOBRINVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c4654"]}},
    "CONSOCIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3555"]}},
    "CONSOLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3560"]}},
    "CONSOLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c3562"]}},
    "CONSOLIDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3568"]}},
    "CONSOLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132373"]}},
    "CONSONANTER": {"-": {"morpho": ["rp--------"], "uri": ["21586"]}},
    "CONSONANTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c9962"]}},
    "CONSONO": {"1": {"morpho": ["v1spia--1-"], "uri": ["c3573"]}},
    "CONSONVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c3574"]}},
    "CONSOPIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["c3575"]}},
    "CONSORS": {"-": {"morpho": ["aps---an3-"], "uri": ["c3577"]}},
    "CONSORTIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c3580"]}},
    "CONSPECTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["c9906"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["21596"]},
    },
    "CONSPERGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3583"]}},
    "CONSPICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["c3588"]}},
    "CONSPICOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["c3590"]}},
    "CONSPICVVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "CONSPICVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c3591"]}},
    "CONSPIRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3592"]}},
    "CONSPIRO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["c4835"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["c4835"]},
    },
    "CONSPVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3601"]}},
    "CONSPVRCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3602"]}},
    "CONSTABILIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["c3604"]}},
    "CONSTANS": {"-": {"morpho": ["aps---an3i"], "uri": ["21607"]}},
    "CONSTANTER": {"-": {"morpho": ["rp--------"], "uri": ["21608"]}},
    "CONSTANTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c9964"]}},
    "CONSTERNATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3609"]}},
    "CONSTERNO": {
        "1": {"morpho": ["v1spia--3-"], "uri": ["c3610"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["c3610"]},
    },
    "CONSTIBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c3612"]}},
    "CONSTIPO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3615"]}},
    "CONSTITVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3616"]}},
    "CONSTITVTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3617"]}},
    "CONSTITVTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c9908"]}},
    "CONSTITVTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["c9908"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["45373"]},
    },
    "CONSTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3620"]}},
    "CONSTRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["53194"]}},
    "CONSTRICTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["21619"]}},
    "CONSTRINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3626"]}},
    "CONSTRVCTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3627"]}},
    "CONSTRVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3628"]}},
    "CONSTVPRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3632"]}},
    "CONSVALIA": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "CONSVDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3642"]}},
    "CONSVEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["c3643"]}},
    "CONSVESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c2511"]}},
    "CONSVETVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3647"]}},
    "CONSVETVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["21630"]}},
    "CONSVL": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c3648"]}},
    "CONSVLARIS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["98012"]},
        "2": {"morpho": ["aps---cn3i"], "uri": ["c3649"]},
    },
    "CONSVLATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c3651"]}},
    "CONSVLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3652"]}},
    "CONSVLTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3653"]}},
    "CONSVLTO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["c3656"]},
        "2": {"morpho": ["rp--------"], "uri": ["58339"]},
    },
    "CONSVLTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c3657"]}},
    "CONSVLTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c9966"]}},
    "CONSVMMABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c6653"]}},
    "CONSVMMATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c6654"]}},
    "CONSVMMATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["21637"]}},
    "CONSVMMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c6657"]}},
    "CONSVMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3658"]}},
    "CONSVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3662"]}},
    "CONSVRGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3664"]}},
    "CONTABVLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3672"]}},
    "CONTABVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3673"]}},
    "CONTACTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c9909"]}},
    "CONTAGES": {"-": {"morpho": ["n-s---fn3i"], "uri": ["c3674"]}},
    "CONTAGIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3675"]}},
    "CONTAGIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c3677"]}},
    "CONTAMINATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["30320"]}},
    "CONTAMINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3683"]}},
    "CONTEGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3687"]}},
    "CONTEMERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3688"]}},
    "CONTEMNENDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["56947"]}},
    "CONTEMNO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3690"]}},
    "CONTEMPLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3696"]}},
    "CONTEMPLATIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c3697"]}},
    "CONTEMPLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c3698"]}},
    "CONTEMPLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132376"]}},
    "CONTEMPTE": {"-": {"morpho": ["rp--------"], "uri": ["28494"]}},
    "CONTEMPTIM": {"-": {"morpho": ["rp--------"], "uri": ["c3708"]}},
    "CONTEMPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3709"]}},
    "CONTEMPTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c3710"]}},
    "CONTEMPTRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3711"]}},
    "CONTEMPTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["c9910"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["21659"]},
    },
    "CONTENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3712"]}},
    "CONTENTE": {
        "1": {"morpho": ["rp--------"], "uri": ["21661"]},
        "2": {"morpho": ["rp--------"], "uri": ["21661"]},
    },
    "CONTENTIO": {
        "1": {"morpho": ["n-s---fn3-"], "uri": ["c3717"]},
        "2": {"morpho": ["n-s---fn3-"], "uri": ["c3717"]},
    },
    "CONTENTVS": {
        "1": {"morpho": ["aps---mn1-"], "uri": ["21663"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["21663"]},
    },
    "CONTERMINVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c3723"]}},
    "CONTERMINVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["c3723"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["c3723"]},
    },
    "CONTERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3726"]}},
    "CONTERREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c3728"]}},
    "CONTESTATE": {"-": {"morpho": ["rp--------"], "uri": ["106177"]}},
    "CONTESTOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["c3735"]}},
    "CONTEXO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3736"]}},
    "CONTEXTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["c9824"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["21667"]},
    },
    "CONTICESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3740"]}},
    "CONTIGNATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3745"]}},
    "CONTIGNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3746"]}},
    "CONTIGVITAS": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "CONTIGVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c3749"]}},
    "CONTINENS": {
        "1": {"morpho": ["n-s---nn3-"], "uri": ["98017"]},
        "2": {"morpho": ["n-s---nn3-"], "uri": ["98017"]},
        "3": {"morpho": ["aps---an3i"], "uri": ["21669"]},
    },
    "CONTINENTER": {"-": {"morpho": ["rp--------"], "uri": ["21670"]}},
    "CONTINENTIA": {"-": {"morpho": ["n-p---nn3i"], "uri": ["c9967"]}},
    "CONTINEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c3750"]}},
    "CONTINGO": {
        "1": {"morpho": ["v1spia--3-"], "uri": ["c3751"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["c3751"]},
    },
    "CONTINVATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3754"]}},
    "CONTINVATVS": {
        "2": {"morpho": ["aps---mn1-"], "uri": ["21676"]},
        "-": {"morpho": ["aps---mn1-"], "uri": ["21676"]},
    },
    "CONTINVO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["c3757"]},
        "2": {"morpho": ["rp--------"], "uri": ["21677"]},
    },
    "CONTINVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c3758"]}},
    "CONTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3759"]}},
    "CONTIONABVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c3760"]}},
    "CONTIONATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c3764"]}},
    "CONTIONOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["c3765"]}},
    "CONTORQVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c3774"]}},
    "CONTRA": {
        "1": {"morpho": ["rp--------"], "uri": ["37284"]},
        "2": {"morpho": ["p---------"], "uri": ["c3780"]},
    },
    "CONTRACTABILITER": {"-": {"morpho": ["rp--------"], "uri": ["c3782"]}},
    "CONTRACTE": {"-": {"morpho": ["rp--------"], "uri": ["47295"]}},
    "CONTRACTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3783"]}},
    "CONTRACTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["98022"]}},
    "CONTRACTVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c3796"]}},
    "CONTRACTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["c9825"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["21682"]},
    },
    "CONTRADICO": {"2": {"morpho": ["v1spia--3-"], "uri": ["c3787"]}},
    "CONTRADICTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3788"]}},
    "CONTRAHO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3794"]}},
    "CONTRALEGO": {"2": {"morpho": ["v1spia--3-"], "uri": ["106212"]}},
    "CONTRARIE": {"-": {"morpho": ["rp--------"], "uri": ["98024"]}},
    "CONTRARIETAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3798"]}},
    "CONTRARIVM": {
        "C": {"morpho": ["n-s---nn2-"], "uri": ["c3800"]},
        "-": {"morpho": ["n-s---nn2-"], "uri": ["c3800"]},
    },
    "CONTRARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c3800"]}},
    "CONTRAVERSVM": {"-": {"morpho": ["rp--------"], "uri": ["45181"]}},
    "CONTRAVERTO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "CONTRECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3811"]}},
    "CONTREMISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3813"]}},
    "CONTREMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3813"]}},
    "CONTRIBVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3818"]}},
    "CONTRICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3820"]}},
    "CONTRISTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3823"]}},
    "CONTROVERSIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c3825"]}},
    "CONTROVERSIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c3828"]}},
    "CONTROVERSVS": {
        "1": {"morpho": ["aps---mn1-"], "uri": ["c3830"]},
        "2": {"morpho": ["rp--------"], "uri": ["98026"]},
    },
    "CONTRVCIDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3831"]}},
    "CONTRVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3832"]}},
    "CONTVBERNALIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["98027"]}},
    "CONTVBERNIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c3835"]}},
    "CONTVEOR": {"-": {"morpho": ["v1spid--2-"], "uri": ["c3836"]}},
    "CONTVITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c9826"]}},
    "CONTVMACIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c3837"]}},
    "CONTVMACITER": {"-": {"morpho": ["rp--------"], "uri": ["21710"]}},
    "CONTVMAX": {"-": {"morpho": ["aps---an3-"], "uri": ["c3838"]}},
    "CONTVMELIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c3839"]}},
    "CONTVMELIOSE": {"-": {"morpho": ["rp--------"], "uri": ["29320"]}},
    "CONTVMELIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c3841"]}},
    "CONTVNDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3844"]}},
    "CONTVOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["c3836"]}},
    "CONTVRBO": {"2": {"morpho": ["v1spia--1-"], "uri": ["c3847"]}},
    "CONTVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c3850"]}},
    "CONVALESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3855"]}},
    "CONVALLIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["c3857"]}},
    "CONVECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3864"]}},
    "CONVECTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c4841"]}},
    "CONVEHO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3868"]}},
    "CONVELLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3869"]}},
    "CONVENA": {"-": {"morpho": ["n-s---cn1-"], "uri": ["c3871"]}},
    "CONVENIENS": {"-": {"morpho": ["aps---an3i"], "uri": ["21732"]}},
    "CONVENIENTER": {"-": {"morpho": ["rp--------"], "uri": ["41286"]}},
    "CONVENIENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c3893"]}},
    "CONVENIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["c3872"]}},
    "CONVENTICVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c3874"]}},
    "CONVENTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3875"]}},
    "CONVENTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3878"]}},
    "CONVENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c9969"]}},
    "CONVENTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c9969"]}},
    "CONVERBERO": {"2": {"morpho": ["v1spia--1-"], "uri": ["c3880"]}},
    "CONVERRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3883"]}},
    "CONVERSATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c4848"]}},
    "CONVERSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3887"]}},
    "CONVERSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3889"]}},
    "CONVERSOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132383"]}},
    "CONVERTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3892"]}},
    "CONVESTIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["c3895"]}},
    "CONVEXVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c3900"]}},
    "CONVEXVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c3900"]}},
    "CONVICIATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c3903"]}},
    "CONVICIO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "CONVICIOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["c3905"]}},
    "CONVICIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c3906"]}},
    "CONVICTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c3908"]}},
    "CONVICTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c3925"]}},
    "CONVINCIO": {"-": {"morpho": ["v1spia--4-"], "uri": []}},
    "CONVINCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3909"]}},
    "CONVISO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3915"]}},
    "CONVIVA": {"-": {"morpho": ["n-s---cn1-"], "uri": ["c3916"]}},
    "CONVIVALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["98031"]}},
    "CONVIVATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c3919"]}},
    "CONVIVIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c3923"]}},
    "CONVIVO": {
        "1": {"morpho": ["v1spia--3-"], "uri": ["c3924"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["c3924"]},
    },
    "CONVIVOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["c3926"]}},
    "CONVOCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3931"]}},
    "CONVOLO": {"2": {"morpho": ["v1spia--1-"], "uri": ["c3932"]}},
    "CONVOLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3934"]}},
    "CONVOLVTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["35311"]}},
    "CONVOLVVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["106306"]}},
    "CONVOMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3937"]}},
    "CONVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c3940"]}},
    "CONVVLNERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3941"]}},
    "COOPERIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["c3951"]}},
    "COOPTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3954"]}},
    "COOPTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3955"]}},
    "COORIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["c3956"]}},
    "COORTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c9828"]}},
    "COPHINVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c3959"]}},
    "COPIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["c3960"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "COPIOSE": {"-": {"morpho": ["rp--------"], "uri": ["21770"]}},
    "COPIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c3964"]}},
    "COPVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c1053"]}},
    "COPVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3978"]}},
    "COPVLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132392"]}},
    "COQVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3991"]}},
    "COQVVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c3992"]}},
    "COR": {"-": {"morpho": ["n-s---nn3-"], "uri": ["c3993"]}},
    "CORA": {
        "1N": {"morpho": ["n-s---fn1-"], "uri": ["c3994"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["c3994"]},
    },
    "CORACINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c3998"]}},
    "CORALIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["31988"]}},
    "CORAM": {
        "1": {"morpho": ["rp--------"], "uri": ["21786"]},
        "2": {"morpho": ["p---------"], "uri": ["c4004"]},
    },
    "CORAX": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["c4006"]},
        "N": {"morpho": ["n-s---mn3-"], "uri": ["c4006"]},
    },
    "CORBIO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "CORBIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["c4008"]}},
    "CORBVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4012"]}},
    "CORCYRA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CORCYRAEI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CORCYRAEVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "CORDAX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c4022"]}},
    "CORDVBA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CORDVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CORIANDRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c4032"]}},
    "CORINTHII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CORINTHIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "CORINTHIVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "CORINTHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CORIOLANVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "CORIOLI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CORIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c4037"]}},
    "CORNELIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CORNELIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "CORNEVS": {
        "1": {"morpho": ["aps---mn1-"], "uri": ["c4862"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["c4862"]},
    },
    "CORNICEN": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c4042"]}},
    "CORNICOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["c4044"]}},
    "CORNICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4045"]}},
    "CORNICVLVM": {"N": {"morpho": ["n-s---nn2-"], "uri": ["c1276"]}},
    "CORNIFICIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CORNIGER": {"-": {"morpho": ["aps---mn1r"], "uri": ["c4051"]}},
    "CORNIPES": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["53987"]},
        "2": {"morpho": ["aps---an3-"], "uri": ["c2357"]},
    },
    "CORNIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c4054"]}},
    "CORNV": {"-": {"morpho": ["n-s---nn4-"], "uri": ["c4055"]}},
    "CORNVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["c4062"]},
        "2": {"morpho": ["n-s---nn2-"], "uri": ["c4062"]},
    },
    "CORNVS": {"-": {"morpho": ["n-s---nn2-"], "uri": []}},
    "COROLLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4065"]}},
    "COROLLARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c4066"]}},
    "CORONA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["c4068"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "CORONAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c4071"]}},
    "CORONARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4073"]}},
    "CORONO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c4078"]}},
    "CORPORALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c4080"]}},
    "CORPORATVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c9970"]}},
    "CORPOREVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4085"]}},
    "CORPVS": {"-": {"morpho": ["n-s---nn3-"], "uri": ["c4091"]}},
    "CORPVSCVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c4092"]}},
    "CORRADO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c4093"]}},
    "CORRECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c4095"]}},
    "CORRECTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c4096"]}},
    "CORREPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c4101"]}},
    "CORREPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c4102"]}},
    "CORRIDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c4109"]}},
    "CORRIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c4111"]}},
    "CORRIPIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["c4112"]}},
    "CORROBORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c4120"]}},
    "CORRODO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c4122"]}},
    "CORROGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c4124"]}},
    "CORROTVNDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c4125"]}},
    "CORRVDA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4126"]}},
    "CORRVGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c4128"]}},
    "CORRVMPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c4130"]}},
    "CORRVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c4131"]}},
    "CORRVPTELA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4132"]}},
    "CORRVPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c4136"]}},
    "CORRVPTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c1344"]}},
    "CORRVPTVS": {
        "C": {"morpho": ["aps---mn1-"], "uri": ["21825"]},
        "-": {"morpho": ["aps---mn1-"], "uri": ["21825"]},
    },
    "CORSA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4142"]}},
    "CORTEX": {"-": {"morpho": ["n-s---cn3-"], "uri": ["c4144"]}},
    "CORTINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4863"]}},
    "CORVSCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c4162"]}},
    "CORVSCVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4163"]}},
    "CORVVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["c4164"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["c4164"]},
    },
    "CORYCEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c4166"]}},
    "CORYLVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["c4171"]}},
    "CORYMBVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c4177"]}},
    "CORYTVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["98042"]}},
    "COS": {"-": {"morpho": ["n-s---fn3-"], "uri": []}},
    "COSMETA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["44640"]}},
    "COSSVRA": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "COSSVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["c4192"]}},
    "COSSVTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "COSTA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["c4194"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "COSTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c4199"]}},
    "COTES": {"-": {"morpho": ["n-s---fn3i"], "uri": ["98043"]}},
    "COTHON": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "COTHVRNATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4203"]}},
    "COTHVRNVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c4205"]}},
    "COTIDIANO": {"-": {"morpho": ["rp--------"], "uri": ["37877"]}},
    "COTIDIANVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["34068"]}},
    "COTIDIE": {"-": {"morpho": ["rp--------"], "uri": ["39974"]}},
    "COTONEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c2471"]}},
    "COTTA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "COTTANA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["c4211"]}},
    "COTTIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "COTVRNIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c4218"]}},
    "COTYLA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["53593"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["53593"]},
    },
    "COTYTIA": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "COVINNARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c4221"]}},
    "COVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["44447"]}},
    "COXA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4228"]}},
    "COXENDIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c4225"]}},
    "COXVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4228"]}},
    "CRABRO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c4229"]}},
    "CRAMBE": {"-": {"morpho": ["n-s---fn1g"], "uri": ["c4232"]}},
    "CRAPVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4233"]}},
    "CRAS": {
        "1": {"morpho": ["n-s----n--"], "uri": []},
        "2": {"morpho": ["rp--------"], "uri": ["c4240"]},
    },
    "CRASSE": {"-": {"morpho": ["rp--------"], "uri": ["21850"]}},
    "CRASSITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c4252"]}},
    "CRASSVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["c4256"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "CRASTINVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c4258"]}},
    "CRASTINVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["c4258"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["56223"]},
    },
    "CRATER": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["c4264"]},
        "N": {"morpho": ["n-s----n2-"], "uri": []},
    },
    "CRATERA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4264"]}},
    "CRATES": {
        "-": {"morpho": ["n-s----n3-"], "uri": []},
        "N": {"morpho": ["n-p---mn3-"], "uri": []},
    },
    "CRATICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4265"]}},
    "CRATICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4266"]}},
    "CRATIPPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CREATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c4275"]}},
    "CREATRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c4276"]}},
    "CREBER": {"-": {"morpho": ["aps---mn1r"], "uri": ["c4277"]}},
    "CREBRA": {"-": {"morpho": ["rp--------"], "uri": ["49217"]}},
    "CREBRE": {"-": {"morpho": ["rp--------"], "uri": ["21865"]}},
    "CREBRESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c4279"]}},
    "CREBRITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c4282"]}},
    "CREBRO": {
        "1": {"morpho": ["v1spia--1-"], "uri": []},
        "2": {"morpho": ["rp--------"], "uri": ["21870"]},
    },
    "CREDIBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c4285"]}},
    "CREDITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c4288"]}},
    "CREDITVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c4284"]}},
    "CREDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c4290"]}},
    "CREDVLITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c4292"]}},
    "CREDVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4293"]}},
    "CREMERA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CREMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c4304"]}},
    "CREMOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c4305"]}},
    "CREMVTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CREO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c4307"]}},
    "CREPAX": {"-": {"morpho": ["aps---an3i"], "uri": ["c4311"]}},
    "CREPER": {"-": {"morpho": ["aps---mn1r"], "uri": ["c4312"]}},
    "CREPEREIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CREPIDA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4313"]}},
    "CREPIDATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4315"]}},
    "CREPIDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c4316"]}},
    "CREPITACILLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c4319"]}},
    "CREPITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c4322"]}},
    "CREPITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c9830"]}},
    "CREPO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c4324"]}},
    "CREPVNDIA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["c4327"]}},
    "CREPVSCVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c4330"]}},
    "CRESCENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c9973"]}},
    "CRESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c4331"]}},
    "CRETA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["c4332"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["c4332"]},
    },
    "CRETATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4335"]}},
    "CRETENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "CRETEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4337"]}},
    "CRETICVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "CRETIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c4340"]}},
    "CRETOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4341"]}},
    "CRETVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4342"]}},
    "CREVSA": {
        "1N": {"morpho": ["n-s----n1-"], "uri": []},
        "2N": {"morpho": ["n-s----n1-"], "uri": []},
        "N": {"morpho": ["n-s---fn1-"], "uri": []},
    },
    "CRIBRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c4348"]}},
    "CRIMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["c4349"]}},
    "CRIMINALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c4350"]}},
    "CRIMINATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c4351"]}},
    "CRIMINATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c4352"]}},
    "CRIMINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["33466"]}},
    "CRIMINOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["c4354"]}},
    "CRIMINOSE": {"-": {"morpho": ["rp--------"], "uri": ["47452"]}},
    "CRIMINOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4355"]}},
    "CRINIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["c4359"]}},
    "CRINIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["c4360"]}},
    "CRINITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4386"]}},
    "CRIODOCIS": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "CRISIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["c4366"]}},
    "CRISPO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c4373"]}},
    "CRISPVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4374"]}},
    "CRISPVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["c4375"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "CRISSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["30228"]}},
    "CRISTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4376"]}},
    "CRISTATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4377"]}},
    "CRITICVS": {"1": {"morpho": ["n-s---mn2-"], "uri": ["c4384"]}},
    "CRITOLAVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CROCEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4391"]}},
    "CROCINVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c4394"]}},
    "CROCINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4394"]}},
    "CROCODILVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c4403"]}},
    "CROCVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c4411"]}},
    "CROCVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c4411"]}},
    "CROTO": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "CRVCIABILITER": {"-": {"morpho": ["rp--------"], "uri": ["34493"]}},
    "CRVCIATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c4434"]}},
    "CRVCIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c4433"]}},
    "CRVDELIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c4440"]}},
    "CRVDELITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c4441"]}},
    "CRVDELITER": {"-": {"morpho": ["rp--------"], "uri": ["37160"]}},
    "CRVDESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c4442"]}},
    "CRVDITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c4443"]}},
    "CRVDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4447"]}},
    "CRVENTE": {"-": {"morpho": ["rp--------"], "uri": ["21925"]}},
    "CRVENTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c4449"]}},
    "CRVENTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4450"]}},
    "CRVMENA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["21928"]}},
    "CRVOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c4454"]}},
    "CRVPPELLARII": {"-": {"morpho": ["n-p---mn2-"], "uri": ["28308"]}},
    "CRVS": {"-": {"morpho": ["n-s---nn3-"], "uri": ["c4459"]}},
    "CRVSTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4463"]}},
    "CRVSTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c4465"]}},
    "CRVSTVLARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c4467"]}},
    "CRVSTVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c4690"]}},
    "CRVSTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c4463"]}},
    "CRVSTVMERIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CRVSTVMERIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "CRVSTVMINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CRVSTVMINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "CRVX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c4470"]}},
    "CRYPTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4471"]}},
    "CRYSTALLINA": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "CRYSTALLINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c1351"]}},
    "CVBICVLARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c7325"]}},
    "CVBICVLARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["c7326"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["c4483"]},
    },
    "CVBICVLATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4484"]}},
    "CVBICVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c4485"]}},
    "CVBICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4486"]}},
    "CVBILE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["c4487"]}},
    "CVBITAL": {"-": {"morpho": ["n-s---nn3i"], "uri": ["c4488"]}},
    "CVBITALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c4489"]}},
    "CVBITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c4491"]}},
    "CVBITVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c9833"]}},
    "CVBITVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["c9833", "c9833"]},
        "2": {"morpho": ["n-s---mn4-"], "uri": ["c9833", "c9833"]},
    },
    "CVBO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c4494"]}},
    "CVBVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c4495"]}},
    "CVCVLLIO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c4502"]}},
    "CVCVLLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c4505"]}},
    "CVCVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c4506"]}},
    "CVCVMIS": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c4511"]}},
    "CVCVRBITA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4513"]}},
    "CVCVRBITIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4518"]}},
    "CVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c4524"]}},
    "CVICVIMODI": {
        "1": {"morpho": ["rp--------"], "uri": ["c4528"]},
        "2": {"morpho": ["rp--------"], "uri": ["c4528"]},
    },
    "CVIVS": {
        "1": {"morpho": ["aps---mn1-"], "uri": []},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "CVIVSMODI": {
        "1": {"morpho": ["rp--------"], "uri": ["c4532"]},
        "2": {"morpho": ["rp--------"], "uri": ["c4532"]},
    },
    "CVIVSQVEMODI": {"-": {"morpho": ["rp--------"], "uri": ["c4534"]}},
    "CVLCITA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4536"]}},
    "CVLCO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "CVLEARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["106627"]}},
    "CVLEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["106629"]}},
    "CVLEVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["49249"]}},
    "CVLEX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c4541"]}},
    "CVLIGNA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4545"]}},
    "CVLINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4547"]}},
    "CVLMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["c4556"]}},
    "CVLMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c4560"]}},
    "CVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c4561"]}},
    "CVLPA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4562"]}},
    "CVLPABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c4563"]}},
    "CVLPATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["29728"]}},
    "CVLPO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c4566"]}},
    "CVLTA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["c9831"]}},
    "CVLTE": {"-": {"morpho": ["rp--------"], "uri": ["49628"]}},
    "CVLTELLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c4570"]}},
    "CVLTER": {"-": {"morpho": ["n-s---mn2r"], "uri": ["c4571"]}},
    "CVLTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c4572"]}},
    "CVLTO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "CVLTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c4573"]}},
    "CVLTRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c4576"]}},
    "CVLTVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c9975"]}},
    "CVLTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["c9831"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["21960"]},
    },
    "CVLVLLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c1356"]}},
    "CVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c4578"]}},
    "CVM": {
        "1": {"morpho": ["rp--------"], "uri": ["c4584"]},
        "2": {"morpho": ["p---------"], "uri": ["c4579"]},
        "3": {"morpho": ["----------"], "uri": []},
    },
    "CVMAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "CVMANVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "CVMBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["106643"]}},
    "CVMCVMQVE": {"-": {"morpho": ["rp--------"], "uri": []}},
    "CVMERA": {
        "1": {"morpho": ["n-s---fn1-"], "uri": ["c4585"]},
        "-": {"morpho": ["n-s---fn1-"], "uri": ["c4585"]},
    },
    "CVMINVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c4588"]}},
    "CVMPRIMIS": {"-": {"morpho": ["rp--------"], "uri": ["c4598"]}},
    "CVMQVE": {"-": {"morpho": ["rp--------"], "uri": ["c4599"]}},
    "CVMVLATE": {"-": {"morpho": ["rp--------"], "uri": ["21967"]}},
    "CVMVLATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["21968"]}},
    "CVMVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c4604"]}},
    "CVMVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c4606"]}},
    "CVNABVLA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["98066"]}},
    "CVNAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["98067"]}},
    "CVNCTABVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4610"]}},
    "CVNCTANS": {"-": {"morpho": ["aps---an3i"], "uri": ["98068"]}},
    "CVNCTANTER": {"-": {"morpho": ["rp--------"], "uri": ["21971"]}},
    "CVNCTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c4614"]}},
    "CVNCTATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c4616"]}},
    "CVNCTOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["c4622"]}},
    "CVNCTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4623"]}},
    "CVNEATIM": {"-": {"morpho": ["rp--------"], "uri": ["c4624"]}},
    "CVNEO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c4626"]}},
    "CVNEOLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c4627"]}},
    "CVNEVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c4628"]}},
    "CVNICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4635"]}},
    "CVNICVLOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4634"]}},
    "CVNICVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c4636"]}},
    "CVNNVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c4643"]}},
    "CVPA": {
        "1": {"morpho": ["n-s---fn1-"], "uri": ["c4650"]},
        "2": {"morpho": ["n-s---fn1-"], "uri": ["c4650"]},
    },
    "CVPEDIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["42349"]}},
    "CVPEDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["37494"]}},
    "CVPIDE": {"-": {"morpho": ["rp--------"], "uri": ["37813"]}},
    "CVPIDITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c4649"]}},
    "CVPIDO": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["c4655"]},
        "N": {"morpho": ["n-s---mn3-"], "uri": []},
    },
    "CVPIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4651"]}},
    "CVPIENS": {"-": {"morpho": ["aps---an3i"], "uri": ["98071"]}},
    "CVPIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["c4656"]}},
    "CVPITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c4658"]}},
    "CVPITVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["44261"]}},
    "CVPRESSETVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c4663"]}},
    "CVPRESSEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4664"]}},
    "CVPRESSIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["c4665"]}},
    "CVPRESSVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["c2489"]}},
    "CVPVLA": {"2": {"morpho": ["n-s---fn1-"], "uri": ["c4677"]}},
    "CVR": {
        "1": {"morpho": ["rp--------"], "uri": ["c4669"]},
        "2": {"morpho": ["rp--------"], "uri": ["c4669"]},
        "-": {"morpho": ["rp--------"], "uri": ["c4669"]},
    },
    "CVRA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["c4670"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "CVRABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c4671"]}},
    "CVRATE": {"-": {"morpho": ["rp--------"], "uri": ["39785"]}},
    "CVRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c4675"]}},
    "CVRATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c4676"]}},
    "CVRATVS": {
        "1": {"morpho": ["n-s----n4-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": ["36518"]},
    },
    "CVRCVLIO": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["c4681"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "CVRES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CVRIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["c4683"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "CVRIALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c4684"]}},
    "CVRIATIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CVRIATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4686"]}},
    "CVRIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CVRIO": {
        "N": {"morpho": ["n-s----n3-"], "uri": []},
        "-": {"morpho": ["aps---an3-"], "uri": ["55992"]},
        "A": {"morpho": ["aps---an3-"], "uri": ["55992"]},
    },
    "CVRIOSE": {"-": {"morpho": ["rp--------"], "uri": ["28778"]}},
    "CVRIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4702"]}},
    "CVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c4706"]}},
    "CVRRICVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c4710"]}},
    "CVRRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c4712"]}},
    "CVRRVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c5714"]}},
    "CVRSIM": {"-": {"morpho": ["rp--------"], "uri": ["c4717"]}},
    "CVRSITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c4720"]}},
    "CVRSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c4721"]}},
    "CVRSOR": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["c4722"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "CVRSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c9832"]}},
    "CVRTIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "1N": {"morpho": ["n-s----n2-"], "uri": []},
        "2N": {"morpho": ["n-s----n2-"], "uri": []},
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["n-s----n2-"], "uri": []},
    },
    "CVRTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c4726"]}},
    "CVRTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4727"]}},
    "CVRVAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["c4728"]}},
    "CVRVATVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4741"]}},
    "CVRVLIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c4735"]}},
    "CVRVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c4736"]}},
    "CVRVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4738"]}},
    "CVSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c4743"]}},
    "CVSPIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c4748"]}},
    "CVSTODIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4753"]}},
    "CVSTODIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["c4755"]}},
    "CVSTOS": {
        "-": {"morpho": ["n-s---cn3-"], "uri": ["22011"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "CVTICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4760"]}},
    "CVTIO": {"-": {"morpho": ["v1spia--3-"], "uri": ["106715"]}},
    "CVTIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["c4762"]}},
    "CYATHVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c4770"]}},
    "CYBAEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4771"]}},
    "CYCLADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CYCLAS": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["c4779"]},
        "N": {"morpho": ["n-s---mn1g"], "uri": []},
    },
    "CYCLICOS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "CYCLOPS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "CYCLOTOS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "CYCNEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4782"]}},
    "CYCNVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["c4784"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["c4784"]},
    },
    "CYLINDRVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c4790"]}},
    "CYMA": {
        "1": {"morpho": ["n-s---nn3-"], "uri": ["c4795"]},
        "2": {"morpho": ["n-s---nn3-"], "uri": ["c4795"]},
    },
    "CYMATIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c4796"]}},
    "CYMBA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["35905"]}},
    "CYMBALVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c4802"]}},
    "CYMBIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c4803"]}},
    "CYNICE": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "CYNICI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CYNICVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": ["c4815"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["c4815"]},
    },
    "CYPARISSVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c4840"]}},
    "CYPRIVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["c4850"]},
        "N": {"morpho": ["aps---mn1-"], "uri": ["c4850"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["c4850"]},
    },
    "CYPRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["55957"]}},
    "CYRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CYTISVS": {"-": {"morpho": ["n-s---cn2-"], "uri": ["c4856"]}},
    "CYZICENI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CYZICVS": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "DAEDALVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["d0010"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "DAEMON": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d0011"]}},
    "DAMA": {
        "-": {"morpho": ["n-s----n1-"], "uri": []},
        "N": {"morpho": ["n-s---fn1-"], "uri": []},
    },
    "DAMNAS": {"-": {"morpho": ["aps---an--"], "uri": ["d0029"]}},
    "DAMNATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0031"]}},
    "DAMNATORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d0032"]}},
    "DAMNATVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["34187"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["35474"]},
    },
    "DAMNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0038"]}},
    "DAMNOSE": {"-": {"morpho": ["rp--------"], "uri": ["51438"]}},
    "DAMNOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d0040"]}},
    "DAMNVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d0039"]}},
    "DAPALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["d0043"]}},
    "DAPS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0052"]}},
    "DAREVS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "DATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0064"]}},
    "DATO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0066"]}},
    "DATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d0067"]}},
    "DATVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d2049"]}},
    "DE": {"-": {"morpho": ["p---------"], "uri": ["d0071"]}},
    "DEA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["d0072"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "DEACINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0073"]}},
    "DEALBO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0080"]}},
    "DEAMBVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0084"]}},
    "DEBACCHOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["d0095"]}},
    "DEBELLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0100"]}},
    "DEBENS": {"-": {"morpho": ["n-s---cn3-"], "uri": ["39647"]}},
    "DEBEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["d0101"]}},
    "DEBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["d0103"]}},
    "DEBILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0104"]}},
    "DEBILITATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0105"]}},
    "DEBILITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0106"]}},
    "DEBITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d0108"]}},
    "DEBITVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d0070"]}},
    "DECANTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0130"]}},
    "DECARPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["39554"]}},
    "DECASTYLVS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "DECEDO": {"1": {"morpho": ["v1spia--3-"], "uri": ["d0142"]}},
    "DECEM": {"-": {"morpho": ["----------"], "uri": []}},
    "DECEMBER": {
        "A": {"morpho": ["aps---mn1r"], "uri": []},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "DECEMPEDA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d0149"]}},
    "DECEMVIR": {"-": {"morpho": ["n-s---mn2r"], "uri": ["d0159"]}},
    "DECEMVIRALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["98087"]}},
    "DECENNIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["d0164"]}},
    "DECENS": {"-": {"morpho": ["aps---an3i"], "uri": ["22063"]}},
    "DECENTER": {"-": {"morpho": ["rp--------"], "uri": ["34725"]}},
    "DECENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d0190"]}},
    "DECEPTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d0172"]}},
    "DECEPTORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d0172"]}},
    "DECERNO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0177"]}},
    "DECERPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0178"]}},
    "DECERTO": {"1": {"morpho": ["v1spia--1-"], "uri": ["d0182"]}},
    "DECESSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0184"]}},
    "DECESSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d0185"]}},
    "DECESSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["d0153"]}},
    "DECET": {"-": {"morpho": ["v3spia--2-"], "uri": ["d0186"]}},
    "DECIDO": {
        "1": {"morpho": ["v1spia--3-"], "uri": ["d0189", "d0188", "d0188"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["d0189", "d0188", "d0188"]},
    },
    "DECIES": {"-": {"morpho": ["rp--------"], "uri": ["98089"]}},
    "DECIMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d1048"]}},
    "DECIMANI": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "DECIMANVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": ["d1181"]},
    },
    "DECIMVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["d0935"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["d1059"]},
    },
    "DECIPIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["d0202"]}},
    "DECISIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0205"]}},
    "DECLAMATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0207"]}},
    "DECLAMATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d0209"]}},
    "DECLAMATORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d0209"]}},
    "DECLAMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0212"]}},
    "DECLARATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d0215"]}},
    "DECLARO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0216"]}},
    "DECLINATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0218"]}},
    "DECLINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0220"]}},
    "DECLIVE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["d0221"]}},
    "DECLIVIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["d0221"]}},
    "DECLIVITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0222"]}},
    "DECLIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d0221"]}},
    "DECOCTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d0240"]}},
    "DECOCTIOR": {
        "-": {"morpho": ["acs---cn3-"], "uri": []},
        "A": {"morpho": ["acs---cn3-"], "uri": []},
    },
    "DECOCTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d0226"]}},
    "DECOCTVM": {
        "-": {"morpho": ["n-s---nn2-"], "uri": ["d0240"]},
        "D\u00e9coction": {"morpho": ["n-s---nn2-"], "uri": ["d0240"]},
    },
    "DECOCTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["56347"]}},
    "DECOLLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0228"]}},
    "DECOLO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["d0229"]},
        "-": {"morpho": ["v1spia--1-"], "uri": ["d0229"]},
    },
    "DECOLOR": {"-": {"morpho": ["aps---an3-"], "uri": ["d0230"]}},
    "DECOLORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0232"]}},
    "DECONDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0236"]}},
    "DECOQVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0238"]}},
    "DECOR": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["d0239"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "DECORATVS": {
        "O": {"morpho": ["aps---mn1-"], "uri": ["106921"]},
        "-": {"morpho": ["aps---mn1-"], "uri": ["106921"]},
    },
    "DECORE": {"-": {"morpho": ["rp--------"], "uri": ["47731"]}},
    "DECORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0245"]}},
    "DECORVM": {
        "C": {"morpho": ["n-s---nn2-"], "uri": ["36901"]},
        "-": {"morpho": ["n-s---nn2-"], "uri": ["36901"]},
    },
    "DECORVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d0239"]}},
    "DECREPITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d0253"]}},
    "DECRESCENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d0244"]}},
    "DECRESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0254"]}},
    "DECRETORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d0258"]}},
    "DECRETVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d0173"]}},
    "DECVMATES": {
        "-": {"morpho": ["aps---cn3i"], "uri": []},
        "A": {"morpho": ["aps---cn3i"], "uri": []},
    },
    "DECVMBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0264"]}},
    "DECVRIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d0270"]}},
    "DECVRIALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["d0271"]}},
    "DECVRIATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0273"]}},
    "DECVRIATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["d0288"]}},
    "DECVRIO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["d0275"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["d0274"]},
    },
    "DECVRIONALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["d0276"]}},
    "DECVRIONATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["d0277"]}},
    "DECVRRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0279"]}},
    "DECVRSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0280"]}},
    "DECVRSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["d0290"]}},
    "DECVRTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0283"]}},
    "DECVS": {"-": {"morpho": ["n-s---nn3-"], "uri": ["d0285"]}},
    "DECVSSATIM": {"-": {"morpho": ["rp--------"], "uri": ["d0286"]}},
    "DECVSSATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0287"]}},
    "DECVSSIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["d0291"]}},
    "DECVTIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["d0295"]}},
    "DEDECET": {"-": {"morpho": ["v1spia--2-"], "uri": []}},
    "DEDECORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0303"]}},
    "DEDECORVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d0305"]}},
    "DEDECVS": {"-": {"morpho": ["n-s---nn3-"], "uri": ["d0306"]}},
    "DEDICATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0308"]}},
    "DEDICATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["44017"]}},
    "DEDICO": {"1": {"morpho": ["v1spia--1-"], "uri": ["d0311"]}},
    "DEDIGNATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0312"]}},
    "DEDIGNOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["d0313"]}},
    "DEDISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0314"]}},
    "DEDITICIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["48529"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["d0315"]},
    },
    "DEDITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0317"]}},
    "DEDITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["28738"]}},
    "DEDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0318"]}},
    "DEDOCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["d0319"]}},
    "DEDOLEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["d0320"]}},
    "DEDVCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0323"]}},
    "DEDVCTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0324"]}},
    "DEDVCTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["d9521"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["32801"]},
    },
    "DEERRO": {"2": {"morpho": ["v1spia--1-"], "uri": ["d0330"]}},
    "DEFAECATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["43977"]}},
    "DEFATIGATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0340"]}},
    "DEFATIGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0341"]}},
    "DEFECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0343"]}},
    "DEFECTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d0345"]}},
    "DEFECTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["d9522"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["98101"]},
    },
    "DEFENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0347"]}},
    "DEFENSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0354"]}},
    "DEFENSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0358"]}},
    "DEFENSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d0359"]}},
    "DEFERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0362"]}},
    "DEFERVEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["d0363"]}},
    "DEFERVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["d9892"]}},
    "DEFESSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["22132"]}},
    "DEFETISCOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["d0366"]}},
    "DEFICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["d0367"]}},
    "DEFIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0368"]}},
    "DEFINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0371"]}},
    "DEFINIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["d0372"]}},
    "DEFINITE": {"-": {"morpho": ["rp--------"], "uri": ["22136"]}},
    "DEFINITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0373"]}},
    "DEFINITIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d0375"]}},
    "DEFINITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["22138"]}},
    "DEFIO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0377"]}},
    "DEFLAGRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0380"]}},
    "DEFLECTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0382"]}},
    "DEFLEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["d0383"]}},
    "DEFLORESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d9875"]}},
    "DEFLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0392"]}},
    "DEFODIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["d0396"]}},
    "DEFORMATIO": {
        "1": {"morpho": ["n-s---fn3-"], "uri": ["d1467"]},
        "2": {"morpho": ["n-s---fn3-"], "uri": ["d1467"]},
    },
    "DEFORMATVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "DEFORMIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["d0401"]}},
    "DEFORMITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0402"]}},
    "DEFORMO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["d0403"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["d0403"]},
    },
    "DEFRICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0413"]}},
    "DEFRINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0415"]}},
    "DEFRVDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["98105"]}},
    "DEFRVTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0422"]}},
    "DEFRVTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d0423"]}},
    "DEFVGIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["d0425"]}},
    "DEFVNCTVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "DEFVNDO": {"2": {"morpho": ["v1spia--3-"], "uri": ["d0430"]}},
    "DEFVNGOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["d0431"]}},
    "DEFVTVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["57104"]}},
    "DEGENER": {"-": {"morpho": ["aps---an3-"], "uri": ["d0437"]}},
    "DEGENERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0438"]}},
    "DEGERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0440"]}},
    "DEGLVBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0442"]}},
    "DEGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0445"]}},
    "DEGRAVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0450"]}},
    "DEGREDIOR": {"-": {"morpho": ["v1spid--3i"], "uri": ["d0451"]}},
    "DEGVSTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0460"]}},
    "DEHAVRIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["d0462"]}},
    "DEHINC": {"-": {"morpho": ["rp--------"], "uri": ["d0463"]}},
    "DEHISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0464"]}},
    "DEHONESTAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d0465"]}},
    "DEHONESTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0468"]}},
    "DEHORTOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["d0475"]}},
    "DEICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["d0478"]}},
    "DEIECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0479"]}},
    "DEIECTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["d9528"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["48126"]},
    },
    "DEILLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DEINCEPS": {"-": {"morpho": ["rp--------"], "uri": ["d0489"]}},
    "DEINDE": {"-": {"morpho": ["rp--------"], "uri": ["d0490"]}},
    "DEIVNGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0500"]}},
    "DELABOR": {"2": {"morpho": ["v1spid--3-"], "uri": ["d0503"]}},
    "DELAPIDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0512"]}},
    "DELASSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0514"]}},
    "DELATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0515"]}},
    "DELATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d0516"]}},
    "DELECTABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["d9512"]}},
    "DELECTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d9514"]}},
    "DELECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0518"]}},
    "DELECTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["d9530"]}},
    "DELEGATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0520"]}},
    "DELEGATVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["107098"]}},
    "DELEGO": {"1": {"morpho": ["v1spia--1-"], "uri": ["d0523"]}},
    "DELENIMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d0525"]}},
    "DELENIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["d0526"]}},
    "DELEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["d0529"]}},
    "DELIACVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "DELIBERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0540"]}},
    "DELIBERATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["22201"]}},
    "DELIBERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0543"]}},
    "DELIBO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0544"]}},
    "DELIBRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0546"]}},
    "DELIBVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0547"]}},
    "DELICATA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["41966"]}},
    "DELICATE": {"-": {"morpho": ["rp--------"], "uri": ["51734"]}},
    "DELICATVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["98112"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["d0548"]},
    },
    "DELICIA": {
        "1": {"morpho": ["n-s---fn1-"], "uri": ["d0550"]},
        "2": {"morpho": ["n-s---fn1-"], "uri": ["d0550"]},
    },
    "DELICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["d0553"]}},
    "DELICIOLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d0554"]}},
    "DELICIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d0549"]}},
    "DELICTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d0522"]}},
    "DELICVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d0560"]}},
    "DELIGO": {
        "2": {"morpho": ["v1spia--1-"], "uri": ["d0564"]},
        "3": {"morpho": ["v1spia--3-"], "uri": ["d0563"]},
    },
    "DELINQVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0573"]}},
    "DELIQVESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0574"]}},
    "DELIQVIAE": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "DELIRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0579"]}},
    "DELIRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0582"]}},
    "DELIRVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d0583"]}},
    "DELITEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["107126"]}},
    "DELITESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["22224"]}},
    "DELIVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "DELPHI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "DELPHICVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "DELPHIN": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["22226"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "DELPHINVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["d0591"]}},
    "DELVBRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d0594"]}},
    "DELVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0598"]}},
    "DELVMBIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["d0599"]}},
    "DELVMBO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0600"]}},
    "DELVS": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "DELVTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0605"]}},
    "DEMANO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0611"]}},
    "DEMARATVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DEMENS": {"-": {"morpho": ["aps---an3i"], "uri": ["d0617"]}},
    "DEMENTER": {"-": {"morpho": ["rp--------"], "uri": ["22233"]}},
    "DEMENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d0619"]}},
    "DEMENTIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["d0620"]}},
    "DEMEREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["d0623"]}},
    "DEMEREOR": {"-": {"morpho": ["v1spid--2-"], "uri": ["132406"]}},
    "DEMERGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0624"]}},
    "DEMETO": {"2": {"morpho": ["v1spia--3-"], "uri": ["d0628"]}},
    "DEMETRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DEMIGRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0630"]}},
    "DEMINVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0633"]}},
    "DEMINVTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0634"]}},
    "DEMINVTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["22243"]}},
    "DEMIROR": {"-": {"morpho": ["v1spid--1-"], "uri": ["d0636"]}},
    "DEMISSE": {"-": {"morpho": ["rp--------"], "uri": ["22244"]}},
    "DEMISSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0638"]}},
    "DEMISSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["22246"]}},
    "DEMITTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0640"]}},
    "DEMIVRGVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["42855"]}},
    "DEMNO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "DEMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0641"]}},
    "DEMOCRITVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DEMOLIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["d0644"]}},
    "DEMOLITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0645"]}},
    "DEMOLITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d0646"]}},
    "DEMONSTRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0648"]}},
    "DEMONSTRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0653"]}},
    "DEMORDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["d0655"]}},
    "DEMORIOR": {"-": {"morpho": ["v1spid--3i"], "uri": ["d0656"]}},
    "DEMOROR": {"1": {"morpho": ["v1spid--1-"], "uri": ["d0657"]}},
    "DEMOVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["d0660"]}},
    "DEMVM": {"-": {"morpho": ["rp--------"], "uri": ["d0664"]}},
    "DEMVTO": {"2": {"morpho": ["v1spia--1-"], "uri": ["d0672"]}},
    "DENARIVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "DENARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": ["d0194"]},
    },
    "DENARRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0677"]}},
    "DENATO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0680"]}},
    "DENEGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0690"]}},
    "DENI": {"-": {"morpho": ["rp--------"], "uri": []}},
    "DENIGRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0694"]}},
    "DENIQVE": {"-": {"morpho": ["rp--------"], "uri": ["d0695"]}},
    "DENOMINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0699"]}},
    "DENORMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0701"]}},
    "DENOTATVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["d9533"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["107198"]},
    },
    "DENOTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0704"]}},
    "DENS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["d0705"]}},
    "DENSE": {"-": {"morpho": ["rp--------"], "uri": ["22274"]}},
    "DENSEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["d0708"]}},
    "DENSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0711"]}},
    "DENSVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["d0712"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "DENTALE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["d0713"]}},
    "DENTATVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["d0716"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "DENTICVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["d0721"]}},
    "DENVBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0731"]}},
    "DENVDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0734"]}},
    "DENVNTIATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0738"]}},
    "DENVNTIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0741"]}},
    "DENVO": {
        "1": {"morpho": ["v1spia--3-"], "uri": ["d0743"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["d0743"]},
    },
    "DEORSVM": {"-": {"morpho": ["rp--------"], "uri": ["d0753"]}},
    "DEPACISCOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["132901"]}},
    "DEPACO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "DEPANGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0760"]}},
    "DEPASCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0762"]}},
    "DEPASCOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["132415"]}},
    "DEPECTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0767"]}},
    "DEPECVLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d0770"]}},
    "DEPECVLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["d0771"]}},
    "DEPELLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0772"]}},
    "DEPENDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["d0774"]}},
    "DEPENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0775"]}},
    "DEPERDITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["38690"]}},
    "DEPERDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0780"]}},
    "DEPEREO": {"1": {"morpho": ["v1spia--2-"], "uri": ["d0782"]}},
    "DEPETIGO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0783"]}},
    "DEPILO": {"1": {"morpho": ["v1spia--1-"], "uri": ["d0789"]}},
    "DEPINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0790"]}},
    "DEPLANGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0793"]}},
    "DEPLEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["d0799"]}},
    "DEPLEXVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["38856"]}},
    "DEPLORATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0805"]}},
    "DEPLORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0807"]}},
    "DEPLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0810"]}},
    "DEPONO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0818"]}},
    "DEPOPVLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0821"]}},
    "DEPOPVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["52078"]}},
    "DEPOPVLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["d0824"]}},
    "DEPORTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0825"]}},
    "DEPORTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0828"]}},
    "DEPOSCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0829"]}},
    "DEPOSITVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d0827"]}},
    "DEPOSTVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0836"]}},
    "DEPRAVATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0844"]}},
    "DEPRAVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0846"]}},
    "DEPRECABVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d0848"]}},
    "DEPRECATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0849"]}},
    "DEPRECATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d0852"]}},
    "DEPRECATRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0854"]}},
    "DEPRECOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["d0855"]}},
    "DEPREHENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0856"]}},
    "DEPRESSE": {"-": {"morpho": ["rp--------"], "uri": ["29677"]}},
    "DEPRESSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0859"]}},
    "DEPRESSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["48447"]}},
    "DEPRIMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0862"]}},
    "DEPROELIOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["d0863"]}},
    "DEPROMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0864"]}},
    "DEPROPERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0865"]}},
    "DEPSO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0871"]}},
    "DEPSTICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d0872"]}},
    "DEPVDET": {"-": {"morpho": ["v3spia--2-"], "uri": ["d9874"]}},
    "DEPVGIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["d0876"]}},
    "DEPVGNATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0877"]}},
    "DEPVGNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0878"]}},
    "DEPVNGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0884"]}},
    "DEPVRGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0888"]}},
    "DEPVTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0890"]}},
    "DERADO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0893"]}},
    "DERELINQVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0897"]}},
    "DEREPENTE": {"-": {"morpho": ["rp--------"], "uri": ["38789"]}},
    "DEREPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0898"]}},
    "DERIDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["d0900"]}},
    "DERIDICVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d0901"]}},
    "DERIDICVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d0901"]}},
    "DERIGESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0902"]}},
    "DERIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["44975"]}},
    "DERIPIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["d0903"]}},
    "DERISOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d0905"]}},
    "DERISVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["d9536"]}},
    "DERIVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0912"]}},
    "DEROGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0918"]}},
    "DERVMPO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "DERVPTA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["98139"]}},
    "DERVPTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d0922"]}},
    "DESAEVIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["d0925"]}},
    "DESCENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0930"]}},
    "DESCENSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["d9537"]}},
    "DESCISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0934"]}},
    "DESCRIBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0937"]}},
    "DESCRIPTA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["98141"]}},
    "DESCRIPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0938"]}},
    "DESECO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0945"]}},
    "DESERO": {
        "2": {"morpho": ["v1spia--3-"], "uri": []},
        "3": {"morpho": ["v1spia--3-"], "uri": []},
    },
    "DESERTA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["107340"]}},
    "DESERTO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "DESERTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d0955"]}},
    "DESERTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["46129"]}},
    "DESERTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["22369"]}},
    "DESERVIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["d0957"]}},
    "DESES": {"-": {"morpho": ["aps---an3-"], "uri": ["d0959"]}},
    "DESIDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["d0963"]}},
    "DESIDERABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["d0964"]}},
    "DESIDERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0965"]}},
    "DESIDERIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d0967"]}},
    "DESIDERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0968"]}},
    "DESIDIA": {"1": {"morpho": ["n-s---fn1-"], "uri": ["d9979", "d0970"]}},
    "DESIDIOSE": {"-": {"morpho": ["rp--------"], "uri": ["35057"]}},
    "DESIDIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d0972"]}},
    "DESIDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0973"]}},
    "DESIGNATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0975"]}},
    "DESIGNATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["35355"]}},
    "DESIGNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0976"]}},
    "DESILIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["d0977"]}},
    "DESINO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0978"]}},
    "DESIPIENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d9980"]}},
    "DESIPIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["d0980"]}},
    "DESISTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0981"]}},
    "DESOLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0988"]}},
    "DESPECTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0993"]}},
    "DESPECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0996"]}},
    "DESPECTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["d9539"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["22389"]},
    },
    "DESPERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1001"]}},
    "DESPERATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["22390"]}},
    "DESPERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1003"]}},
    "DESPICATVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["d8010"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["107368"]},
    },
    "DESPICIENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d9981"]}},
    "DESPICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["d1007"]}},
    "DESPOLIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1011"]}},
    "DESPONDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["d1012"]}},
    "DESPVMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1017"]}},
    "DESPVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1018"]}},
    "DESTERTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1022"]}},
    "DESTILLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1024"]}},
    "DESTILLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1025"]}},
    "DESTINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d1027"]}},
    "DESTINATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1028"]}},
    "DESTINATVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d0917"]}},
    "DESTINATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["22409"]}},
    "DESTINEO": {"-": {"morpho": ["v1spia--2-"], "uri": []}},
    "DESTINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1030"]}},
    "DESTITVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1032"]}},
    "DESTRICTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["22414"]}},
    "DESTRINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1038"]}},
    "DESTRVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1043"]}},
    "DESVBITO": {"-": {"morpho": ["rp--------"], "uri": ["d1046"]}},
    "DESVDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1050"]}},
    "DESVESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1053"]}},
    "DESVETVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1054"]}},
    "DESVLTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d1058"]}},
    "DESVM": {"1": {"morpho": ["v1spia--3-"], "uri": ["d1061"]}},
    "DESVMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1062"]}},
    "DESVPER": {"-": {"morpho": ["rp--------"], "uri": ["d1064"]}},
    "DESVRGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1065"]}},
    "DETECTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["98149"]}},
    "DETEGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1071"]}},
    "DETENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1073"]}},
    "DETENTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1077"]}},
    "DETERGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["d1080"]}},
    "DETERGO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "DETERIOR": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "DETERIVS": {"-": {"morpho": ["rc--------"], "uri": ["22435"]}},
    "DETERMINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1088"]}},
    "DETERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1089"]}},
    "DETERREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["d1091"]}},
    "DETESTABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["d1093"]}},
    "DETESTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1470"]}},
    "DETESTOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["d1097"]}},
    "DETEXO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1098"]}},
    "DETINEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["d1100"]}},
    "DETONDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["d1101"]}},
    "DETORQVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["d1107"]}},
    "DETRACTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1109"]}},
    "DETRACTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1115"]}},
    "DETRACTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d1110"]}},
    "DETRAHO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1111"]}},
    "DETRECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1115"]}},
    "DETRIMENTOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d1117"]}},
    "DETRIMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d1118"]}},
    "DETRVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1120"]}},
    "DETRVNCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1122"]}},
    "DETVRBO": {"2": {"morpho": ["v1spia--1-"], "uri": ["d1128"]}},
    "DEVECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1134"]}},
    "DEVEHO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1135"]}},
    "DEVELLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1136"]}},
    "DEVENEROR": {"-": {"morpho": ["v1spid--1-"], "uri": ["d1138"]}},
    "DEVENIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["d1139"]}},
    "DEVERRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1144"]}},
    "DEVERSOR": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["d1148"]},
        "2": {"morpho": ["v1spid--1-"], "uri": ["d1147"]},
    },
    "DEVERSORIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d1209"]}},
    "DEVERTICVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d1152"]}},
    "DEVERTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1153"]}},
    "DEVEXVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["57140"]}},
    "DEVEXVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d1158"]}},
    "DEVIA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["d1171"]}},
    "DEVINCIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["d1161"]}},
    "DEVINCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1162"]}},
    "DEVINCTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["52838"]}},
    "DEVITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1170"]}},
    "DEVIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d1171"]}},
    "DEVNX": {"-": {"morpho": ["n-s---mn3i"], "uri": ["d1173"]}},
    "DEVOCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1174"]}},
    "DEVOLO": {"2": {"morpho": ["v1spia--1-"], "uri": ["d1175"]}},
    "DEVOLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1176"]}},
    "DEVORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1183"]}},
    "DEVOTI": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "DEVOTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1187"]}},
    "DEVOTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["53381"]}},
    "DEVOVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["d1190"]}},
    "DEVRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1191"]}},
    "DEVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["d1193"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["d1193"]},
    },
    "DEXO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "DEXTANS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["d1198"]}},
    "DEXTER": {
        "-": {"morpho": ["aps---mn1r"], "uri": ["d0000"]},
        "N": {"morpho": ["n-s----n2-"], "uri": []},
    },
    "DEXTERA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d0027"]}},
    "DEXTERE": {"-": {"morpho": ["rp--------"], "uri": ["33809"]}},
    "DEXTRORSVM": {"-": {"morpho": ["rp--------"], "uri": ["d1208"]}},
    "DEXTRORSVS": {"-": {"morpho": ["rp--------"], "uri": ["57573"]}},
    "DIADEMA": {"-": {"morpho": ["rp--------"], "uri": []}},
    "DIADVMENOS": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "DIAGONIOS": {"-": {"morpho": ["aps---mn1g"], "uri": ["d1254"]}},
    "DIAGRAMMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["d1255"]}},
    "DIALECTICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["52429"]}},
    "DIALECTICVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["98163"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["d1258"]},
    },
    "DIALIS": {
        "N": {"morpho": ["aps---cn3i"], "uri": ["$0025"]},
        "A": {"morpho": ["aps---cn3i"], "uri": ["$0025"]},
    },
    "DIALOGVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["d1268"]}},
    "DIAMETROS": {"-": {"morpho": ["n-s---fn2g"], "uri": ["d1276"]}},
    "DIANA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "DIANIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "DIARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d1296"]}},
    "DIASTYLOS": {"-": {"morpho": ["aps---mn1g"], "uri": ["d1308"]}},
    "DIATHESIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["107554"]}},
    "DIATHYRA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["107555"]}},
    "DIATONOS": {"-": {"morpho": ["aps---mn1g"], "uri": ["d1315"]}},
    "DIAVLOS": {"-": {"morpho": ["n-s---mn2g"], "uri": ["d1321"]}},
    "DIAZOMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["d1324"]}},
    "DICA": {"-": {"morpho": ["n-s---fn1-"], "uri": []}},
    "DICACITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1331"]}},
    "DICAX": {"-": {"morpho": ["aps---an3-"], "uri": ["d1335"]}},
    "DICHALCVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d1338"]}},
    "DICIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1347"]}},
    "DICIS": {"-": {"morpho": ["n-s----n--"], "uri": []}},
    "DICO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["d1349"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["d1350"]},
    },
    "DICROTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d1351"]}},
    "DICTATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d1355"]}},
    "DICTATORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d1355"]}},
    "DICTATVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d9986"]}},
    "DICTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1361"]}},
    "DICTITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1363"]}},
    "DICTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1365"]}},
    "DICTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d9542"]}},
    "DIDO": {
        "-": {"morpho": ["v1spia--3-"], "uri": ["d1372"]},
        "N": {"morpho": ["n-s---fn3-"], "uri": []},
    },
    "DIDVCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1374"]}},
    "DIEQVINTI": {"-": {"morpho": ["rp--------"], "uri": ["98167"]}},
    "DIES": {
        "-": {"morpho": ["n-s---fn5-"], "uri": []},
        "N": {"morpho": ["n-s---mn1g"], "uri": []},
    },
    "DIESIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["29553"]}},
    "DIEZEVGMENOS": {"-": {"morpho": ["aps---mn1g"], "uri": ["d1465"]}},
    "DIFFAMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1383"]}},
    "DIFFERCIO": {"-": {"morpho": ["v1spia--4-"], "uri": []}},
    "DIFFERENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d9987"]}},
    "DIFFERITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1385"]}},
    "DIFFERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1386"]}},
    "DIFFERTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d1387"]}},
    "DIFFICILE": {"-": {"morpho": ["rp--------"], "uri": ["41142"]}},
    "DIFFICILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["d1389"]}},
    "DIFFICILITER": {"-": {"morpho": ["rp--------"], "uri": ["29157"]}},
    "DIFFICVLTAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1391"]}},
    "DIFFICVLTER": {"-": {"morpho": ["rp--------"], "uri": ["49976"]}},
    "DIFFIDENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d9988"]}},
    "DIFFIDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1392"]}},
    "DIFFINDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1393"]}},
    "DIFFINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1394"]}},
    "DIFFITEOR": {"-": {"morpho": ["v1spid--2-"], "uri": ["d1399"]}},
    "DIFFLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1404"]}},
    "DIFFRINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1408"]}},
    "DIFFVGIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["d1409"]}},
    "DIFFVNDO": {"2": {"morpho": ["v1spia--3-"], "uri": ["d1414"]}},
    "DIFFVSILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["d1415"]}},
    "DIFFVSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1416"]}},
    "DIFFVSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["38077"]}},
    "DIFFVTVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1418"]}},
    "DIGERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1423"]}},
    "DIGITVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["d1433"]}},
    "DIGITVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["d1434"]}},
    "DIGNATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1439"]}},
    "DIGNE": {"-": {"morpho": ["rp--------"], "uri": ["22530"]}},
    "DIGNITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1440"]}},
    "DIGNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1442"]}},
    "DIGNOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132433"]}},
    "DIGNOSCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["22532"]}},
    "DIGNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d1444"]}},
    "DIGREDIOR": {"-": {"morpho": ["v1spid--3i"], "uri": ["d1445"]}},
    "DIGRESSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["d9545"]}},
    "DIIVDICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1451"]}},
    "DIIVNCTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["34295"]}},
    "DIIVNGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["43933"]}},
    "DILABOR": {"2": {"morpho": ["v1spid--3-"], "uri": ["d1556"]}},
    "DILACERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1558"]}},
    "DILANIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1561"]}},
    "DILARGIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["d1566"]}},
    "DILATATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1569"]}},
    "DILATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1571"]}},
    "DILATO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1572"]}},
    "DILECTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["d9530"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["98172"]},
    },
    "DILIGENS": {"-": {"morpho": ["aps---an3i"], "uri": ["98173"]}},
    "DILIGENTER": {"-": {"morpho": ["rp--------"], "uri": ["22545"]}},
    "DILIGENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d9990"]}},
    "DILIGO": {"3": {"morpho": ["v1spia--3-"], "uri": ["d1581"]}},
    "DILVCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["d1587"]}},
    "DILVCESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1587"]}},
    "DILVCIDE": {"-": {"morpho": ["rp--------"], "uri": ["51544"]}},
    "DILVCIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d1589"]}},
    "DILVDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d1592"]}},
    "DILVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1593"]}},
    "DILVTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d1574"]}},
    "DILVTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["98175"]}},
    "DILVVIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["d1595"]}},
    "DILVVIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1596"]}},
    "DILVVIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d1595"]}},
    "DIMACHAE": {"-": {"morpho": ["n-p---mn1-"], "uri": ["d1599"]}},
    "DIMENSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["50304"]}},
    "DIMETIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["d1606"]}},
    "DIMICATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1609"]}},
    "DIMICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1610"]}},
    "DIMIDIATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d8612"]}},
    "DIMIDIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1613"]}},
    "DIMIDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d1611"]}},
    "DIMIDIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d1611"]}},
    "DIMINVTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["29264"]}},
    "DIMISSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1617"]}},
    "DIMITTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1620"]}},
    "DIMOEROS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "DIMOVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["d1621"]}},
    "DINOSCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1623"]}},
    "DINVMERO": {"1": {"morpho": ["v1spia--1-"], "uri": ["d1626"]}},
    "DIO": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "DIOCLES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "DIODORVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DIOGENES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "DIOGNETVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DIONYSIA": {
        "1N": {"morpho": ["n-s---fn1-"], "uri": ["107709"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["107709"]},
    },
    "DIONYSIARCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DIONYSIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DIONYSODORVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DIOPTRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d1638"]}},
    "DIOTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d1640"]}},
    "DIPECHYAEOS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "DIPISCOR": {"-": {"morpho": ["v1spid--3-"], "uri": []}},
    "DIPLASION": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "DIPLINTHIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d1649"]}},
    "DIPONDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["98178"]}},
    "DIPONDIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["40496"]}},
    "DIPTEROS": {"-": {"morpho": ["aps---mn1g"], "uri": ["d1656"]}},
    "DIRA": {
        "-": {"morpho": ["n-p---nn2-"], "uri": ["d1662"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "DIRAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["d1662"]}},
    "DIRECTE": {"-": {"morpho": ["rp--------"], "uri": ["50661"]}},
    "DIRECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1669"]}},
    "DIRECTO": {"-": {"morpho": ["rp--------"], "uri": ["49101"]}},
    "DIRECTVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["45425"]},
        "2": {"morpho": ["rp--------"], "uri": []},
    },
    "DIRECTVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d1689"]}},
    "DIRECTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["22562"]}},
    "DIREPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1673"]}},
    "DIREPTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d1674"]}},
    "DIRIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1680"]}},
    "DIRIMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1681"]}},
    "DIRIPIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["d1682"]}},
    "DIRVM": {"-": {"morpho": ["rp--------"], "uri": ["98180"]}},
    "DIRVMPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1684"]}},
    "DIRVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1685"]}},
    "DIRVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d1662"]}},
    "DIS": {
        "-": {"morpho": ["aps---an3-"], "uri": ["d2023"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "DISCEDO": {"1": {"morpho": ["v1spia--3-"], "uri": ["d1695"]}},
    "DISCENS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["98181"]}},
    "DISCEPTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1696"]}},
    "DISCEPTATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d1697"]}},
    "DISCEPTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1700"]}},
    "DISCERNO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1703"]}},
    "DISCERPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1704"]}},
    "DISCESSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1709"]}},
    "DISCESSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["d9549"]}},
    "DISCIDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d1711"]}},
    "DISCIDO": {"2": {"morpho": ["v1spia--3-"], "uri": ["d1712"]}},
    "DISCINCTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["41136"]}},
    "DISCINDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1713"]}},
    "DISCINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1714"]}},
    "DISCIPLINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d1715"]}},
    "DISCIPLINOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d1717"]}},
    "DISCIPVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d1718"]}},
    "DISCIPVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["d1718"]}},
    "DISCITO": {"1": {"morpho": ["v1spia--1-"], "uri": ["d1721"]}},
    "DISCLVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1722"]}},
    "DISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1724"]}},
    "DISCOLOR": {"-": {"morpho": ["aps---an3-"], "uri": ["d1727"]}},
    "DISCONVENIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["d1732"]}},
    "DISCOQVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1735"]}},
    "DISCORDIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["d1737"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "DISCORDIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d1738"]}},
    "DISCORDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1741"]}},
    "DISCORS": {"-": {"morpho": ["aps---an3-"], "uri": ["d1745"]}},
    "DISCREPANTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d9994"]}},
    "DISCREPITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1749"]}},
    "DISCREPO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1750"]}},
    "DISCRESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1751"]}},
    "DISCRIBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1757"]}},
    "DISCRIMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["d1758"]}},
    "DISCRIMINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1764"]}},
    "DISCRVCIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1766"]}},
    "DISCVMBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1769"]}},
    "DISCVPIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["d1771"]}},
    "DISCVRRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1772"]}},
    "DISCVRSATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1773"]}},
    "DISCVRSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1777"]}},
    "DISCVRSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["d9552"]}},
    "DISCVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["d1778"]}},
    "DISCVTIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["d1782"]}},
    "DISDIAPENTE": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "DISDIATESSARON": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "DISDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["98191"]}},
    "DISERTE": {"-": {"morpho": ["rp--------"], "uri": ["22596"]}},
    "DISERTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d1888"]}},
    "DISHIASCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1794"]}},
    "DISICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["d1795"]}},
    "DISIECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1796"]}},
    "DISIECTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["d1803"]}},
    "DISPANDO": {"2": {"morpho": ["v1spia--3-"], "uri": ["d1807"]}},
    "DISPAR": {"-": {"morpho": ["aps---an3-"], "uri": ["d1809"]}},
    "DISPARATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1811"]}},
    "DISPARO": {"2": {"morpho": ["v1spia--1-"], "uri": ["d1815"]}},
    "DISPECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["107866"]}},
    "DISPECTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["d9553"]}},
    "DISPELLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1821"]}},
    "DISPENDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d1823"]}},
    "DISPENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1463"]}},
    "DISPENSATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1825"]}},
    "DISPENSATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d1827"]}},
    "DISPENSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1831"]}},
    "DISPERDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1833"]}},
    "DISPEREO": {"1": {"morpho": ["v1spia--1-"], "uri": ["d1834"]}},
    "DISPERGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1835"]}},
    "DISPERSE": {"-": {"morpho": ["rp--------"], "uri": ["54688"]}},
    "DISPERTIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["d1839"]}},
    "DISPESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1841"]}},
    "DISPICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["d1842"]}},
    "DISPLICENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d1824"]}},
    "DISPLICEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["d1844"]}},
    "DISPLODO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1846"]}},
    "DISPLVVIATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d1847"]}},
    "DISPONO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1850"]}},
    "DISPOSITE": {"-": {"morpho": ["rp--------"], "uri": ["32171"]}},
    "DISPOSITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1851"]}},
    "DISPOSITVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "DISPOSITVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d9995"]}},
    "DISPOSITVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["98205"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["98205"]},
        "1": {"morpho": ["n-s---mn4-"], "uri": []},
    },
    "DISPVNGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1858"]}},
    "DISPVTABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["d1859"]}},
    "DISPVTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1860"]}},
    "DISPVTATIVNCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d1861"]}},
    "DISPVTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1865"]}},
    "DISQVIRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1866"]}},
    "DISQVISITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1867"]}},
    "DISSAEPIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["d1870"]}},
    "DISSAEPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1871"]}},
    "DISSAEPTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d1863"]}},
    "DISSEMINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1876"]}},
    "DISSENSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1877"]}},
    "DISSENTIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["d1879"]}},
    "DISSERO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["d1886"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["d1885"]},
        "3": {"morpho": ["v1spia--3-"], "uri": ["d1885"]},
    },
    "DISSERPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1887"]}},
    "DISSERTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1889"]}},
    "DISSERTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1891"]}},
    "DISSERTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1893"]}},
    "DISSICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["98210"]}},
    "DISSIDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["d1895"]}},
    "DISSIGNATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d1901"]}},
    "DISSIGNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1902"]}},
    "DISSILIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["d1903"]}},
    "DISSIMILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["d1904"]}},
    "DISSIMILITER": {"-": {"morpho": ["rp--------"], "uri": ["44731"]}},
    "DISSIMILITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1905"]}},
    "DISSIMVLANTER": {"-": {"morpho": ["rp--------"], "uri": ["56603"]}},
    "DISSIMVLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1908"]}},
    "DISSIMVLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d1909"]}},
    "DISSIMVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1910"]}},
    "DISSIPABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["d1911"]}},
    "DISSIPO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1915"]}},
    "DISSOCIABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["d1917"]}},
    "DISSOCIATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1919"]}},
    "DISSOCIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1920"]}},
    "DISSOLVBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["d1921"]}},
    "DISSOLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1922"]}},
    "DISSOLVTE": {"-": {"morpho": ["rp--------"], "uri": ["46460"]}},
    "DISSOLVTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1923"]}},
    "DISSOLVTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["22633"]}},
    "DISSONO": {"1": {"morpho": ["v1spia--1-"], "uri": ["d1926"]}},
    "DISSONVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d1928"]}},
    "DISSVADEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["d1931"]}},
    "DISSVASIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1932"]}},
    "DISSVASOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d1933"]}},
    "DISSVLTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1937"]}},
    "DISSVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1938"]}},
    "DISTABESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1939"]}},
    "DISTANTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d9998"]}},
    "DISTENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1946"]}},
    "DISTENTO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "DISTENTVS": {
        "1": {"morpho": ["n-s----n4-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": ["22641"]},
        "3": {"morpho": ["aps---mn1-"], "uri": ["22641"]},
    },
    "DISTERMINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1951"]}},
    "DISTERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1954"]}},
    "DISTICHON": {"-": {"morpho": ["n-s---nn2g"], "uri": ["d1956"]}},
    "DISTINCTE": {"-": {"morpho": ["rp--------"], "uri": ["32806"]}},
    "DISTINCTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1958"]}},
    "DISTINCTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["d9558"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["22644"]},
    },
    "DISTINEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["d1963"]}},
    "DISTINGVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1961"]}},
    "DISTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1964"]}},
    "DISTORQVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["d1965"]}},
    "DISTORTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["36766"]}},
    "DISTRACTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1969"]}},
    "DISTRACTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["57050"]}},
    "DISTRAHO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1971"]}},
    "DISTRIBVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1973"]}},
    "DISTRIBVTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1975"]}},
    "DISTRICTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1978"]}},
    "DISTRICTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["22653"]}},
    "DISTRINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1981"]}},
    "DISTVRBO": {"2": {"morpho": ["v1spia--1-"], "uri": ["d1985"]}},
    "DITESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1992"]}},
    "DITHYRAMBVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["d1995"]}},
    "DITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d2000"]}},
    "DIV": {"-": {"morpho": ["rp--------"], "uri": ["d1998"]}},
    "DIVA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["d0157"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "DIVARICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d2007"]}},
    "DIVDICO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "DIVELLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d2010"]}},
    "DIVENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d2011"]}},
    "DIVERBERO": {"2": {"morpho": ["v1spia--1-"], "uri": ["d7012"]}},
    "DIVERSE": {"-": {"morpho": ["rp--------"], "uri": ["40501"]}},
    "DIVERSITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d2018"]}},
    "DIVERSORIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["43892"]}},
    "DIVERSVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "DIVERSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["22661"]}},
    "DIVERTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d2022"]}},
    "DIVES": {"-": {"morpho": ["aps---an3-"], "uri": ["d2023"]}},
    "DIVIDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d2029"]}},
    "DIVIDVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d2031"]}},
    "DIVINA": {
        "1": {"morpho": ["n-s---fn1-"], "uri": ["d1457"]},
        "2": {"morpho": ["n-s---fn1-"], "uri": ["d1457"]},
    },
    "DIVINATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d2033"]}},
    "DIVINITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d2037"]}},
    "DIVINITVS": {"-": {"morpho": ["rp--------"], "uri": ["d2038"]}},
    "DIVINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d2039"]}},
    "DIVINVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d1458"]}},
    "DIVINVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["d1150"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["d2040"]},
    },
    "DIVISIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d2043"]}},
    "DIVISOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d2044"]}},
    "DIVISVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["d2003"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["98223"]},
    },
    "DIVITIAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["d1996"]}},
    "DIVORTIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d2050"]}},
    "DIVRNVM": {
        "N": {"morpho": ["n-s---nn2-"], "uri": ["d2054"]},
        "-": {"morpho": ["n-s----n2-"], "uri": []},
    },
    "DIVRNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d2054"]}},
    "DIVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["d1461"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["d1461"]},
    },
    "DIVTINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d2057"]}},
    "DIVTVRNITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d2059"]}},
    "DIVTVRNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d2060"]}},
    "DIVVLGO": {"1": {"morpho": ["v1spia--1-"], "uri": ["d2063"]}},
    "DIVVLSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d2064"]}},
    "DIVVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d0161"]}},
    "DIVVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["d0160"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["d0126"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["d0126"]},
    },
    "DO": {
        "-": {
            "morpho": ["v1spia--1-", "v1spia--1-"],
            "uri": ["d2066", "d2066"],
        },
        "1": {"morpho": ["v1spia--1-"], "uri": ["d2066"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["d2066"]},
    },
    "DOCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["d2067"]}},
    "DOCIBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["d2069"]}},
    "DOCILIS": {
        "-": {"morpho": ["aps---cn3i"], "uri": ["d2071"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "DOCIMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DOCIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d2073"]}},
    "DOCTE": {"-": {"morpho": ["rp--------"], "uri": ["45704"]}},
    "DOCTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d2081"]}},
    "DOCTRINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d2082"]}},
    "DOCTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["22684"]}},
    "DOCVMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["d2086"]}},
    "DOCVMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d2089"]}},
    "DODRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d2093"]}},
    "DODRANS": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d2094"]}},
    "DOGMA": {"-": {"morpho": ["n-s---mn3-"], "uri": []}},
    "DOLABELLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "DOLABRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d2104"]}},
    "DOLENS": {"-": {"morpho": ["aps---an3i"], "uri": ["98232"]}},
    "DOLEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["d2113"]}},
    "DOLESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["108119"]}},
    "DOLIOLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d2121"]}},
    "DOLITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d2122"]}},
    "DOLIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d2123", "d2120"]}},
    "DOLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d2124"]}},
    "DOLOR": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["d2126"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "DOLOSE": {"-": {"morpho": ["rp--------"], "uri": ["29456"]}},
    "DOLOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d2128"]}},
    "DOLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["d2129"]}},
    "DOMABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["d2132"]}},
    "DOMESTICVS": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": ["d2138"]},
    },
    "DOMICILIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d2141"]}},
    "DOMINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d0165"]}},
    "DOMINANS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["98239"]},
        "2": {"morpho": ["aps---an3i"], "uri": ["98240"]},
    },
    "DOMINATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d2145"]}},
    "DOMINATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d2146"]}},
    "DOMINATRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d2147"]}},
    "DOMINATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["d9562"]}},
    "DOMINICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d0223"]}},
    "DOMINIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d2150"]}},
    "DOMINOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["d2151"]}},
    "DOMINVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["d0210"]}},
    "DOMITIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": ["108152"]},
    },
    "DOMITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d2156"]}},
    "DOMITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d2157"]}},
    "DOMITRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d2158"]}},
    "DOMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d2161"]}},
    "DOMVNCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d2163"]}},
    "DOMVS": {"-": {"morpho": ["n-s---fn4-"], "uri": ["d2164"]}},
    "DONARIA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["98242"]}},
    "DONATICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d2169"]}},
    "DONATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d2170"]}},
    "DONATIVVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d2172"]}},
    "DONATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d2173"]}},
    "DONEC": {"-": {"morpho": ["----------"], "uri": []}},
    "DONO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d2179"]}},
    "DONVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d2180"]}},
    "DORCAS": {"-": {"morpho": ["n-s---cn3-"], "uri": ["d2181"]}},
    "DORICE": {"-": {"morpho": ["rp--------"], "uri": []}},
    "DORMIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["d2184"]}},
    "DORMITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d2188"]}},
    "DORON": {"-": {"morpho": ["n-s---nn2g"], "uri": ["d2191"]}},
    "DOROTHEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DORSVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d7193"]}},
    "DORYPHORA": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "DORYPHOROS": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "DOS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d2197"]}},
    "DOTALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["d2198"]}},
    "DOTATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["48855"]}},
    "DOTO": {
        "-": {"morpho": ["v1spia--1-"], "uri": ["d2200"]},
        "N": {"morpho": ["n-s----n1-"], "uri": []},
    },
    "DRACHMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d2203"]}},
    "DRACO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d2206"]}},
    "DREPANITANVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "DROMAS": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d2231"]}},
    "DROPAX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d2239"]}},
    "DRVIDAE": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "DRVIDES": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "DRVSILLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "DRVSVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DVBIE": {"-": {"morpho": ["rp--------"], "uri": ["48311"]}},
    "DVBITABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["d2257"]}},
    "DVBITANTER": {"-": {"morpho": ["rp--------"], "uri": ["36581"]}},
    "DVBITATIM": {"-": {"morpho": ["rp--------"], "uri": ["d2258"]}},
    "DVBITATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d2259"]}},
    "DVBITATIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d2260"]}},
    "DVBITATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d2261"]}},
    "DVBITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d2262"]}},
    "DVBIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["22727"]}},
    "DVBIVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["d2263"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "DVCENI": {"-": {"morpho": ["rp--------"], "uri": []}},
    "DVCENTESIMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d2274"]}},
    "DVCENTESIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d2274"]}},
    "DVCENTI": {"-": {"morpho": ["----------"], "uri": []}},
    "DVCENTIES": {"-": {"morpho": ["rp--------"], "uri": ["98249"]}},
    "DVCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d2279"]}},
    "DVCTARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d2281"]}},
    "DVCTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d2284"]}},
    "DVCTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d2286"]}},
    "DVCTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d2287"]}},
    "DVCTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["d9564"]}},
    "DVDVM": {"-": {"morpho": ["rp--------"], "uri": ["d2289"]}},
    "DVELLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["22735"]}},
    "DVILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DVLCE": {
        "1": {"morpho": ["n-s----n3-"], "uri": []},
        "2": {"morpho": ["rp--------"], "uri": ["54867"]},
    },
    "DVLCEDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d2297"]}},
    "DVLCESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d2298"]}},
    "DVLCIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["d2309"]}},
    "DVM": {
        "1": {"morpho": ["rp--------"], "uri": ["d2317"]},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "DVMETVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d2320"]}},
    "DVMMODO": {"-": {"morpho": ["----------"], "uri": []}},
    "DVMOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d2323"]}},
    "DVMTAXAT": {"-": {"morpho": ["rp--------"], "uri": ["d2324"]}},
    "DVMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["d2325"]}},
    "DVO": {"-": {"morpho": ["----------"], "uri": []}},
    "DVODECIM": {"-": {"morpho": ["----------"], "uri": []}},
    "DVODECIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d2341"]}},
    "DVODENI": {"-": {"morpho": ["rp--------"], "uri": []}},
    "DVODEQVADRAGINTA": {"-": {"morpho": ["----------"], "uri": []}},
    "DVODETRICESIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d2357"]}},
    "DVODETRICIES": {"-": {"morpho": ["rp--------"], "uri": ["108331"]}},
    "DVODETRIGINTA": {"-": {"morpho": ["----------"], "uri": []}},
    "DVODEVICESIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["98257"]}},
    "DVODEVIGINTI": {"-": {"morpho": ["----------"], "uri": []}},
    "DVOETVICESIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["33715"]}},
    "DVPLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d2379"]}},
    "DVPLEX": {
        "-": {"morpho": ["aps---an3-", "aps---an3i"], "uri": ["d2371"]}
    },
    "DVPLICATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d2373"]}},
    "DVPLICITER": {"-": {"morpho": ["rp--------"], "uri": ["22747"]}},
    "DVPLICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d2376"]}},
    "DVPLO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["d2378"]},
        "2": {"morpho": ["rp--------"], "uri": ["22749"]},
    },
    "DVPLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d2379"]}},
    "DVPLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d2379"]}},
    "DVRABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["d2382"]}},
    "DVRACINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d2384"]}},
    "DVRAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["d2385"]}},
    "DVRAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d2386"]}},
    "DVRATEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d2387"]}},
    "DVRE": {"-": {"morpho": ["rp--------"], "uri": ["41749"]}},
    "DVREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["d2389"]}},
    "DVRESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d2389"]}},
    "DVRITER": {"-": {"morpho": ["rp--------"], "uri": ["48092"]}},
    "DVRITIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d2395"]}},
    "DVRITIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["d2395"]}},
    "DVRITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d2397"]}},
    "DVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d2400"]}},
    "DVRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["56055"]}},
    "DVRVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["d2401"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "DVVMVIR": {"-": {"morpho": ["n-s---mn2r"], "uri": ["34219"]}},
    "DVX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d2409"]}},
    "DYNASTES": {"-": {"morpho": ["n-s---mn1g"], "uri": ["d2413"]}},
    "DYRRACHIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "DYSEROS": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "DYSPEPSIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d2419"]}},
    "EA": {"-": {"morpho": ["rp--------"], "uri": ["38160"]}},
    "EADEM": {"-": {"morpho": ["rp--------"], "uri": ["22756"]}},
    "EATENVS": {"-": {"morpho": ["rp--------"], "uri": ["e0003"]}},
    "EBENVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["22758"]}},
    "EBIBO": {"2": {"morpho": ["v1spia--3-"], "uri": ["e0006"]}},
    "EBLANDIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["e0009"]}},
    "EBRIETAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0015"]}},
    "EBRIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e0019"]}},
    "EBRIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e0021"]}},
    "EBVLLIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["e0023"]}},
    "EBVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e0025"]}},
    "EBVR": {"-": {"morpho": ["n-s---nn3-"], "uri": ["e0026"]}},
    "EBVRNEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e0028"]}},
    "ECCE": {"-": {"morpho": ["----------"], "uri": []}},
    "ECCLESIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["e0037"]}},
    "ECCLESIASTERION": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "ECETRANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ECHEA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["e0041"]}},
    "ECHINVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["e0048"]}},
    "ECHO": {
        "-": {"morpho": ["n-s---fn--"], "uri": ["e0052"]},
        "N": {"morpho": ["n-s----n--"], "uri": []},
    },
    "ECNVBO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "ECPHORA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["e0062"]}},
    "ECQVANDO": {"-": {"morpho": ["rp--------"], "uri": ["e1927"]}},
    "ECQVI": {"-": {"morpho": ["rp--------"], "uri": ["98280"]}},
    "ECQVID": {"-": {"morpho": ["rp--------"], "uri": ["54888"]}},
    "ECQVIS": {"-": {"morpho": ["----------"], "uri": []}},
    "ECQVISNAM": {"-": {"morpho": ["----------"], "uri": []}},
    "ECQVO": {"-": {"morpho": ["rp--------"], "uri": ["34886"]}},
    "ECTYPVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e0069"]}},
    "EDAX": {"-": {"morpho": ["aps---an3i"], "uri": ["e0073"]}},
    "EDICO": {"2": {"morpho": ["v1spia--3-"], "uri": ["e0081"]}},
    "EDICTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e9956"]}},
    "EDISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0085"]}},
    "EDISSERO": {"2": {"morpho": ["v1spia--3-"], "uri": ["e0086"]}},
    "EDITA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["98283"]}},
    "EDITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0092"]}},
    "EDITVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e0128"]}},
    "EDITVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["e0128"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["22779"]},
    },
    "EDO": {
        "1": {"morpho": ["v1spia--3-"], "uri": ["e0094", "e0095", "e0094"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["e0094", "e0095", "e0094"]},
    },
    "EDOCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["e0097"]}},
    "EDOMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0100"]}},
    "EDORMIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["e0101"]}},
    "EDVCATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0102"]}},
    "EDVCATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["e0103"]}},
    "EDVCO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["e0105"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["e0106"]},
    },
    "EDVCTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0107"]}},
    "EDVLIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["e0110"]}},
    "EDVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0112"]}},
    "EDVRVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e0113"]}},
    "EFFARCIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["31305"]}},
    "EFFATVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e9980"]}},
    "EFFECTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e9981"]}},
    "EFFECTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["e9981"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["98287"]},
    },
    "EFFEMINATE": {"-": {"morpho": ["rp--------"], "uri": ["49420"]}},
    "EFFEMINATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["37436"]}},
    "EFFEMINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0123"]}},
    "EFFERATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["52539"]}},
    "EFFERCIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["e0125"]}},
    "EFFERO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["e0127"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["e0127"]},
    },
    "EFFERVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["e0130"]}},
    "EFFERVESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0130"]}},
    "EFFERVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0130"]}},
    "EFFERVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e0129"]}},
    "EFFETVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e0133"]}},
    "EFFICACITER": {"-": {"morpho": ["rp--------"], "uri": ["49984"]}},
    "EFFICAX": {"-": {"morpho": ["aps---an3i"], "uri": ["e0137"]}},
    "EFFICIENS": {
        "1": {"morpho": ["n-s----n3-"], "uri": []},
        "2": {"morpho": ["aps---an3i"], "uri": ["98289"]},
    },
    "EFFICIENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["e0131"]}},
    "EFFICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["e0139"]}},
    "EFFIGIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["e0144"]}},
    "EFFIGIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["e0144"]}},
    "EFFINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0148"]}},
    "EFFLAGITATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0151"]}},
    "EFFLAGITATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["e9983"]}},
    "EFFLAGITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0152"]}},
    "EFFLIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0157"]}},
    "EFFLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0158"]}},
    "EFFLORESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0159"]}},
    "EFFLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0160"]}},
    "EFFLVVIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e0162"]}},
    "EFFODIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["e0165"]}},
    "EFFOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["e0167"]}},
    "EFFRACTARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["e0171"]}},
    "EFFRENATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["22804"]}},
    "EFFRENO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0175"]}},
    "EFFRENVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e0174"]}},
    "EFFRICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0176"]}},
    "EFFRINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0177"]}},
    "EFFVGIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["e0182"]}},
    "EFFVGIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e0183"]}},
    "EFFVLGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["e0186"]}},
    "EFFVNDO": {"2": {"morpho": ["v1spia--3-"], "uri": ["e0193"]}},
    "EFFVNSVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "EFFVSE": {"-": {"morpho": ["rp--------"], "uri": ["45750"]}},
    "EFFVSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0194"]}},
    "EFFVSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["22812"]}},
    "EFFVTIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["e0200"]}},
    "EFFVTVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0201"]}},
    "EGELIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e0203"]}},
    "EGENS": {
        "C": {"morpho": ["aps---an3i"], "uri": ["98292"]},
        "-": {"morpho": ["aps---an3i"], "uri": ["98292"]},
    },
    "EGENVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e0206"]}},
    "EGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["e0207"]}},
    "EGERIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "EGERIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "EGERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0210"]}},
    "EGESTAS": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["e0213"]},
        "N": {"morpho": ["n-s---fn3-"], "uri": ["e0213"]},
    },
    "EGIGNO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0216"]}},
    "EGNATIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "EGO": {"-": {"morpho": ["o-----an--"], "uri": []}},
    "EGREDIOR": {"-": {"morpho": ["v1spid--3i"], "uri": ["e0221"]}},
    "EGREGIE": {"-": {"morpho": ["rp--------"], "uri": ["22817"]}},
    "EGREGIVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "EGREGIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e0222"]}},
    "EGRESSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["e9966"]}},
    "EHEV": {"-": {"morpho": ["----------"], "uri": []}},
    "EI": {"-": {"morpho": ["----------"], "uri": []}},
    "EIA": {"-": {"morpho": ["----------"], "uri": []}},
    "EIACVLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["e0231"]}},
    "EICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["e0232"]}},
    "EIECTAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e0233"]}},
    "EIECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0235"]}},
    "EIECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e9237"]}},
    "EIECTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["e9967"]}},
    "EIVLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e9235"]}},
    "EIVLATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["e9968"]}},
    "EIVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0237"]}},
    "EIVRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0241"]}},
    "EIVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0242"]}},
    "EIVSMODI": {"-": {"morpho": ["rp--------"], "uri": ["e8100"]}},
    "ELABOR": {"2": {"morpho": ["v1spid--3-"], "uri": ["e0244"]}},
    "ELABORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0246"]}},
    "ELAEOTHESIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e0253"]}},
    "ELANGVESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0256"]}},
    "ELARGIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["e0260"]}},
    "ELATE": {
        "1": {"morpho": ["n-s---fn1g"], "uri": ["e0262"]},
        "2": {"morpho": ["rp--------"], "uri": ["48335"]},
    },
    "ELATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0265"]}},
    "ELATRO": {"2": {"morpho": ["v1spia--1-"], "uri": ["e0267"]}},
    "ELATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["40226"]}},
    "ELAVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0269"]}},
    "ELECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0274"]}},
    "ELECTO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["e1937", "e0276"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["e1937", "e0276"]},
    },
    "ELECTRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e0281"]}},
    "ELECTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["e9970"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["58144"]},
    },
    "ELEGANS": {"-": {"morpho": ["aps---an3i"], "uri": ["e1938"]}},
    "ELEGANTER": {"-": {"morpho": ["rp--------"], "uri": ["35134"]}},
    "ELEGANTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["e9993"]}},
    "ELEGEON": {"-": {"morpho": ["n-s---nn2g"], "uri": ["108661"]}},
    "ELEGI": {"-": {"morpho": ["n-s---mn2-"], "uri": ["55844"]}},
    "ELEGIDION": {"-": {"morpho": ["n-s---nn2g"], "uri": ["e0290"]}},
    "ELEGO": {"1": {"morpho": ["v1spia--1-"], "uri": ["e0292"]}},
    "ELEMENTA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["98310"]}},
    "ELEMENTARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e0294"]}},
    "ELEMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e0296"]}},
    "ELENCHVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["e0297"]}},
    "ELEPHANTVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["e0311"]}},
    "ELEPHAS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["98312"]}},
    "ELEVO": {"1": {"morpho": ["v1spia--1-"], "uri": ["e0316"]}},
    "ELICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["e0319"]}},
    "ELICIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ELIDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0320"]}},
    "ELIGO": {"3": {"morpho": ["v1spia--3-"], "uri": ["e0321"]}},
    "ELIMINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0324"]}},
    "ELINGVIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["e0329"]}},
    "ELIQVESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0333"]}},
    "ELIQVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0334"]}},
    "ELIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "ELISIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0336"]}},
    "ELIXVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e0340"]}},
    "ELLIPSIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["e0348"]}},
    "ELLYCHNIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e0349"]}},
    "ELOCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0350"]}},
    "ELOGIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e0356"]}},
    "ELOQVENS": {"-": {"morpho": ["aps---an3i"], "uri": ["22830"]}},
    "ELOQVENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["e9994"]}},
    "ELOQVIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e0359"]}},
    "ELOQVOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["e0360"]}},
    "ELORINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ELORINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "ELORVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ELVACRVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "ELVCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["e0363"]}},
    "ELVCTOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["e0368"]}},
    "ELVCVBRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0369"]}},
    "ELVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0372"]}},
    "ELVMBIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["e0374"]}},
    "ELVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0376"]}},
    "ELVTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["55757"]}},
    "ELVVIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["e0000"]}},
    "EMACERATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["98321"]}},
    "EMANCIPO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0398"]}},
    "EMANO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0401"]}},
    "EMARCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["45218"]}},
    "EMARCESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0404"]}},
    "EMAX": {"-": {"morpho": ["aps---an3i"], "uri": ["e0410"]}},
    "EMBATER": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "EMBLEMA": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "EMBOLIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e0423"]}},
    "EMBOLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["e0430"]}},
    "EMENDABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["e0436"]}},
    "EMENDATE": {"-": {"morpho": ["rp--------"], "uri": ["47981"]}},
    "EMENDATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0437"]}},
    "EMENDATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["e0438"]}},
    "EMENDATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["44987"]}},
    "EMENDICATICIVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["e0441"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["e0441"]},
    },
    "EMENDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0443"]}},
    "EMENTIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["e0445"]}},
    "EMERCOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["e0448"]}},
    "EMEREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["e0449"]}},
    "EMEREOR": {"-": {"morpho": ["v1spid--2-"], "uri": ["132455"]}},
    "EMERGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0450"]}},
    "EMERITVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e0426"]}},
    "EMERITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["57892"]}},
    "EMERSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["e0451"]}},
    "EMETIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["e0453"]}},
    "EMETO": {"2": {"morpho": ["v1spia--3-"], "uri": ["e0454"]}},
    "EMICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0457"]}},
    "EMIGRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0459"]}},
    "EMINENS": {"-": {"morpho": ["aps---an3i"], "uri": ["22853"]}},
    "EMINEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["e0461"]}},
    "EMINVS": {"-": {"morpho": ["rp--------"], "uri": ["e0468"]}},
    "EMIROR": {"-": {"morpho": ["v1spid--1-"], "uri": ["e0469"]}},
    "EMISSARIVS": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "EMISSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["e9971"]}},
    "EMITTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0477"]}},
    "EMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0479"]}},
    "EMODEROR": {"-": {"morpho": ["v1spid--1-"], "uri": ["e0480"]}},
    "EMODVLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["e0481"]}},
    "EMOLIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["e0483"]}},
    "EMOLLIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["e0485"]}},
    "EMOLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0486"]}},
    "EMOLVMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e0488"]}},
    "EMORIOR": {"-": {"morpho": ["v1spid--3i"], "uri": ["e0490"]}},
    "EMOVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["e0494"]}},
    "EMPLASTRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e0507"]}},
    "EMPLECTOS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "EMPORIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e0513"]}},
    "EMPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0520"]}},
    "EMPTITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0521"]}},
    "EMPTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["e0523"]}},
    "EMVLGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["e0531"]}},
    "EMVNDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0535"]}},
    "EMVNGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0536"]}},
    "EMVNIO": {"2": {"morpho": ["v1spia--4-"], "uri": ["e0537"]}},
    "EN": {"-": {"morpho": ["----------"], "uri": []}},
    "ENARRABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["e0549"]}},
    "ENARRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0550"]}},
    "ENARRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0553"]}},
    "ENASCOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["e0554"]}},
    "ENATO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0557"]}},
    "ENAVIGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0558"]}},
    "ENCARPI": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "ENCELADOS": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "ENCHYTVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "ENCLIMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["e0574"]}},
    "ENCYCLIOS": {"-": {"morpho": ["aps---mn1g"], "uri": ["e0579"]}},
    "ENDROMIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0584"]}},
    "ENECO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0587"]}},
    "ENERVATVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": ["30576"]},
        "-": {"morpho": ["aps---mn1-"], "uri": ["30576"]},
    },
    "ENERVIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["e0589"]}},
    "ENERVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e9585"]}},
    "ENGVINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ENGVINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "ENIM": {
        "1": {"morpho": ["rp--------"], "uri": ["e0594"]},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "ENIMVERO": {"-": {"morpho": ["rp--------"], "uri": ["e8102"]}},
    "ENITEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["e0595"]}},
    "ENITESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0595"]}},
    "ENITOR": {"2": {"morpho": ["v1spid--3-"], "uri": ["e0596"]}},
    "ENIXE": {"-": {"morpho": ["rp--------"], "uri": ["44115"]}},
    "ENIXVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["e0580"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["50725"]},
    },
    "ENNIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ENO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0609"]}},
    "ENODATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0611"]}},
    "ENODIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["e0612"]}},
    "ENODO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0613"]}},
    "ENORMIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["e0615"]}},
    "ENORMITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0616"]}},
    "ENOTESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0617"]}},
    "ENS": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "ENSIGER": {"-": {"morpho": ["aps---mn1r"], "uri": ["e0624"]}},
    "ENSIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["e0626"]}},
    "ENTASIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["108991"]}},
    "ENTELLINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ENTELLINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "ENTHEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e0634"]}},
    "ENVCLEATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["22888"]}},
    "ENVCLEO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0643"]}},
    "ENVMERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0645"]}},
    "ENVMERO": {"1": {"morpho": ["v1spia--1-"], "uri": ["e0646"]}},
    "ENVNTIATIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e0649"]}},
    "ENVNTIATVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e0659"]}},
    "ENVNTIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0651"]}},
    "ENVTRIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["e0653"]}},
    "EO": {
        "1": {"morpho": ["v1spia--4-"], "uri": ["e0655"]},
        "3": {"morpho": ["rp--------"], "uri": ["98345"]},
    },
    "EODEM": {"-": {"morpho": ["rp--------"], "uri": ["52062"]}},
    "EOVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": ["e1953"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["e1953"]},
    },
    "EPAEETIS": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "EPAGON": {"-": {"morpho": ["----------"], "uri": []}},
    "EPHEBEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e0686"]}},
    "EPHEBVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["e0689"]}},
    "EPHECTOS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "EPHEMERIDAE": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "EPHEMERIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0693"]}},
    "EPHESII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "EPHESIVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "EPHESVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "EPHIPPIATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e0696"]}},
    "EPHIPPIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e0697"]}},
    "EPIBATA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["e0700"]}},
    "EPIBATHRA": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "EPICRATES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "EPICVRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "EPIDIMOEROS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "EPIDROMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["e0727"]}},
    "EPIGRAMMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["e0728"]}},
    "EPIGRVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["e0734"]}},
    "EPIMENIVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "EPIPEMPTOS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "EPIRHEDIVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "EPIRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["98353"]}},
    "EPISCAENOS": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "EPISCOPVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["e0763"]}},
    "EPISTATES": {"-": {"morpho": ["n-s---mn1g"], "uri": ["e0766"]}},
    "EPISTOLA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["50212"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "EPISTOLIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e0771"]}},
    "EPISTOMIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["109136"]}},
    "EPISTYLIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e0774"]}},
    "EPITONIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e0791"]}},
    "EPITOXIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0792"]}},
    "EPITRITOS": {"-": {"morpho": ["aps---mn1g"], "uri": ["e0794"]}},
    "EPITYRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e0796"]}},
    "EPIZYGIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0799"]}},
    "EPOS": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "EPOTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0809"]}},
    "EPVLAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["e0482"]}},
    "EPVLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["e0817"]}},
    "EPVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e0503"]}},
    "EQVA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["e0530"]}},
    "EQVES": {"-": {"morpho": ["n-s---cn3-"], "uri": ["e0820"]}},
    "EQVESTER": {
        "-": {"morpho": ["aps---mn3-"], "uri": ["e0821"]},
        "A": {"morpho": ["n-s---mn3-"], "uri": ["46388"]},
    },
    "EQVESTRIA": {
        "Si\u00e8ges": {"morpho": ["n-p---nn3i"], "uri": ["98358"]},
        "-": {"morpho": ["n-p---nn3i"], "uri": ["98358"]},
    },
    "EQVIDEM": {"-": {"morpho": ["rp--------"], "uri": ["e0822"]}},
    "EQVILE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["e0824"]}},
    "EQVINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e0828"]}},
    "EQVITABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["e0834"]}},
    "EQVITATVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["e0863"]},
        "2": {"morpho": ["n-s---mn4-"], "uri": ["e0863"]},
    },
    "EQVITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0838"]}},
    "EQVOR": {"-": {"morpho": ["v1spid--3-"], "uri": []}},
    "EQVVLEVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["41202"]}},
    "EQVVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["e0555"]}},
    "ERADO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0847"]}},
    "ERCEO": {"-": {"morpho": ["v1spia--2-"], "uri": []}},
    "ERECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0853"]}},
    "ERECTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["22925"]}},
    "EREPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0859"]}},
    "EREPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0860"]}},
    "EREPTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["e0862"]}},
    "ERGA": {"-": {"morpho": ["p---------"], "uri": ["e0866"]}},
    "ERGASTVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e0872"]}},
    "ERGATA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["e0873"]}},
    "ERGO": {
        "1": {"morpho": ["p---------"], "uri": ["e0875"]},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "ERICIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["e0881"]}},
    "ERIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0884"]}},
    "ERIPHYLA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ERIPIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["e0892"]}},
    "ERISMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["e0893"]}},
    "ERNEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e0900"]}},
    "ERODO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0902"]}},
    "EROGATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0904"]}},
    "EROGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0908"]}},
    "EROMENION": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "EROS": {
        "2N": {"morpho": ["n-s----n-g"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "ERRABVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e0918"]}},
    "ERRATICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e0920"]}},
    "ERRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0922"]}},
    "ERRATVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e9958"]}},
    "ERRO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["e0926"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["e0925"]},
    },
    "ERROR": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["e9927"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "ERVBEO": {"-": {"morpho": ["v1spia--2-"], "uri": []}},
    "ERVBESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0929"]}},
    "ERVCA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["e0930"]}},
    "ERVCTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0932"]}},
    "ERVDERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0934"]}},
    "ERVDIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["e0936"]}},
    "ERVDITI": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "ERVDITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0937"]}},
    "ERVDITVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e0941"]}},
    "ERVDITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["22942"]}},
    "ERVMPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0946"]}},
    "ERVNCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0949"]}},
    "ERVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0950"]}},
    "ERVPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0951"]}},
    "ERVVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["22946"]}},
    "ERYCINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "ERYCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ERYMANTHIVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "ERYTHRAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "ESCA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["e0967"]}},
    "ESCARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e0969"]}},
    "ESCENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0971"]}},
    "ESCENSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0972"]}},
    "ESCENSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["e0955"]}},
    "ESCHARA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["e0973"]}},
    "ESITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0982"]}},
    "ESQVILIAE": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "ESQVILINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "ESSEDA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["e0987"]}},
    "ESSEDARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["e0261"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["54780"]},
    },
    "ESSEDVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e0987"]}},
    "ESSENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["e0988"]}},
    "ESVRIO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["e0994"]},
        "2": {"morpho": ["v1spia--4-"], "uri": ["e0993"]},
    },
    "ESVRITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0995"]}},
    "ET": {
        "1": {"morpho": ["rp--------"], "uri": ["32800"]},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "ETENIM": {"-": {"morpho": ["----------"], "uri": []}},
    "ETESIAE": {"-": {"morpho": ["n-p---mn1-"], "uri": ["e0307"]}},
    "ETESIVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["e0428"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "ETHOLOGIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["e1006"]}},
    "ETIAM": {"-": {"morpho": ["rp--------"], "uri": ["e1010"]}},
    "ETIAMNVNC": {"-": {"morpho": ["rp--------"], "uri": ["e8108"]}},
    "ETIAMSI": {"-": {"morpho": ["----------"], "uri": []}},
    "ETIAMTVM": {"-": {"morpho": ["rp--------"], "uri": ["e8110"]}},
    "ETIAMTVNC": {"-": {"morpho": ["rp--------"], "uri": ["e8110"]}},
    "ETRVRIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ETRVSCI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ETRVSCVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "ETSI": {
        "1": {"morpho": ["----------"], "uri": []},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "EV": {"-": {"morpho": ["----------"], "uri": []}},
    "EVADO": {"2": {"morpho": ["v1spia--3-"], "uri": ["e1020"]}},
    "EVAGATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1021"]}},
    "EVAGOR": {"2": {"morpho": ["v1spid--1-"], "uri": ["e1024"]}},
    "EVALESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1025"]}},
    "EVANDER": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "EVANESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1029"]}},
    "EVANIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e1030"]}},
    "EVANS": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "EVBVLIDA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "EVBVLIDAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "EVECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1052"]}},
    "EVECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1053"]}},
    "EVECTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["e9928"]}},
    "EVEHO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1054"]}},
    "EVELLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1056"]}},
    "EVENIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["e1057"]}},
    "EVENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e9929"]}},
    "EVENTVRA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["29639"]}},
    "EVENTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["e9929"]}},
    "EVERBERO": {"2": {"morpho": ["v1spia--1-"], "uri": ["e1060"]}},
    "EVERRICVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e1066"]}},
    "EVERRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1067"]}},
    "EVERSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1068"]}},
    "EVERSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["e1070"]}},
    "EVERTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1072"]}},
    "EVGANEI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "EVGE": {"-": {"morpho": ["----------"], "uri": []}},
    "EVGENEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e1077"]}},
    "EVHYIVS": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "EVIDENS": {"-": {"morpho": ["aps---an3i"], "uri": ["e1086"]}},
    "EVIDENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["e1087"]}},
    "EVIGILO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1091"]}},
    "EVINCIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["e1094"]}},
    "EVINCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1095"]}},
    "EVIRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1098"]}},
    "EVITABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["e1100"]}},
    "EVITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1102"]}},
    "EVMENIDAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "EVNVCHVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["e1112"]}},
    "EVOCATI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["98384"]}},
    "EVOCATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1113"]}},
    "EVOCATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["e1115"]}},
    "EVOCATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["98385"]}},
    "EVOCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1117"]}},
    "EVOE": {"-": {"morpho": ["----------"], "uri": []}},
    "EVOLO": {"2": {"morpho": ["v1spia--1-"], "uri": ["e1120"]}},
    "EVOLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1123"]}},
    "EVOMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1124"]}},
    "EVPHRANOR": {
        "-": {"morpho": ["n-s----n3-"], "uri": []},
        "N": {"morpho": ["n-s---mn3-"], "uri": []},
    },
    "EVPOLEMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "EVRIPVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["e1145"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["e1145"]},
    },
    "EVROPA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "EVROVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e1151"]}},
    "EVRVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["e1152"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["e1152"]},
    },
    "EVRYTHMIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["e1153"]}},
    "EVSTYLOS": {"-": {"morpho": ["aps---mn1g"], "uri": ["e1157"]}},
    "EVTHVS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "EVTHYGRAMMOS": {"-": {"morpho": ["aps---mn1g"], "uri": ["e1161"]}},
    "EVTHYMIA": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "EVVLGO": {"1": {"morpho": ["v1spia--1-"], "uri": ["e1163"]}},
    "EVVLSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["109498"]}},
    "EX": {"-": {"morpho": ["p---------"], "uri": ["e1167"]}},
    "EXACERBO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e8172"]}},
    "EXACTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1177"]}},
    "EXACTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["e1178"]}},
    "EXACTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["e9931"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["22995"]},
    },
    "EXACVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1181"]}},
    "EXADVERSVM": {
        "1": {"morpho": ["rp--------"], "uri": ["22997"]},
        "2": {"morpho": ["p---------"], "uri": ["e1184"]},
    },
    "EXAEDIFICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1186"]}},
    "EXAEQVATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1187"]}},
    "EXAEQVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1188"]}},
    "EXAESTVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1191"]}},
    "EXAGGERO": {"1": {"morpho": ["v1spia--1-"], "uri": ["e1194"]}},
    "EXAGITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1197"]}},
    "EXAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["e1212"]}},
    "EXAMINATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1213"]}},
    "EXAMINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1217"]}},
    "EXANCLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1223"]}},
    "EXANIMIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["e1228"]}},
    "EXANIMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1229"]}},
    "EXANIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e1228"]}},
    "EXARDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["e1237"]}},
    "EXARDESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0660"]}},
    "EXAREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["e1240"]}},
    "EXARESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1240"]}},
    "EXARMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1242"]}},
    "EXARO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1243"]}},
    "EXASPERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1250"]}},
    "EXAVCTORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1254"]}},
    "EXAVDIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["e1256"]}},
    "EXAVGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["e1259"]}},
    "EXAVGVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1261"]}},
    "EXCAECO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1269"]}},
    "EXCALCEO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1270"]}},
    "EXCANDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["e9808"]}},
    "EXCANDESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1279"]}},
    "EXCANTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1280"]}},
    "EXCARNIFICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1281"]}},
    "EXCAVO": {"1": {"morpho": ["v1spia--1-"], "uri": ["e1286"]}},
    "EXCEDO": {"1": {"morpho": ["v1spia--3-"], "uri": ["e1287"]}},
    "EXCELLENS": {"-": {"morpho": ["aps---an3i"], "uri": []}},
    "EXCELLENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["e9999"]}},
    "EXCELLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1288"]}},
    "EXCELSE": {"-": {"morpho": ["rp--------"], "uri": ["50834"]}},
    "EXCELSITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1289"]}},
    "EXCELSVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e0306"]}},
    "EXCELSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e1290"]}},
    "EXCEPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1293"]}},
    "EXCEPTIVNCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["e1294"]}},
    "EXCEPTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1295"]}},
    "EXCERNO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1300"]}},
    "EXCERPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1301"]}},
    "EXCESSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["e9932"]}},
    "EXCIDIVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["e8306"]},
        "2": {"morpho": ["n-s---nn2-"], "uri": ["e8306"]},
    },
    "EXCIDO": {
        "1": {"morpho": ["v1spia--3-"], "uri": ["e1308", "e1307", "e1307"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["e1308", "e1307", "e1307"]},
    },
    "EXCIEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["23032"]}},
    "EXCIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["e1309"]}},
    "EXCIPIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["e1311"]}},
    "EXCISIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1314"]}},
    "EXCITATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["50319"]}},
    "EXCITO": {"1": {"morpho": ["v1spia--1-"], "uri": ["e1322"]}},
    "EXCLAMATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1324"]}},
    "EXCLAMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1327"]}},
    "EXCLARO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1328"]}},
    "EXCLVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1329"]}},
    "EXCLVSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1330"]}},
    "EXCLVSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["109676"]}},
    "EXCOGITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1346"]}},
    "EXCOLO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["e1349"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["e1348"]},
    },
    "EXCOQVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1360"]}},
    "EXCORS": {"-": {"morpho": ["aps---an3-"], "uri": ["e1364"]}},
    "EXCREMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e1367"]}},
    "EXCREO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "EXCRESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1370"]}},
    "EXCRETVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["109703"]}},
    "EXCRVCIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1375"]}},
    "EXCVBIAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["e1378"]}},
    "EXCVBITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["e1381"]}},
    "EXCVBITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["e9935"]}},
    "EXCVBO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1384"]}},
    "EXCVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1385"]}},
    "EXCVLCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1387"]}},
    "EXCVRRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1392"]}},
    "EXCVRSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1395"]}},
    "EXCVRSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["e1397"]}},
    "EXCVRSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["e9936"]}},
    "EXCVSATE": {"-": {"morpho": ["rp--------"], "uri": ["38068"]}},
    "EXCVSATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1402"]}},
    "EXCVSATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["98397"]}},
    "EXCVSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1406"]}},
    "EXCVSSE": {"-": {"morpho": ["rp--------"], "uri": ["98398"]}},
    "EXCVSSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["57914"]}},
    "EXCVTIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["e1410"]}},
    "EXEDO": {"2": {"morpho": ["v1spia--3-"], "uri": ["e1416"]}},
    "EXEDRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["e0656"]}},
    "EXEMPLAR": {"-": {"morpho": ["n-s---nn3i"], "uri": ["e1423"]}},
    "EXEMPLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e1428"]}},
    "EXEMPTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["e9937"]}},
    "EXEO": {"1": {"morpho": ["v1spia--4-"], "uri": ["e1433"]}},
    "EXERCEO": {
        "2": {"morpho": ["v1spia--2-"], "uri": ["e1434"]},
        "-": {"morpho": ["v1spia--2-"], "uri": ["e1434"]},
    },
    "EXERCITATE": {"-": {"morpho": ["rp--------"], "uri": ["98404"]}},
    "EXERCITATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1437"]}},
    "EXERCITATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["23063"]}},
    "EXERCITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1441"]}},
    "EXERCITIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e1442"]}},
    "EXERCITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1443"]}},
    "EXERCITVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["e1414"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["98406"]},
    },
    "EXESOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["e1449"]}},
    "EXFLORESCO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "EXHALO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1457"]}},
    "EXHAVRIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["e1458"]}},
    "EXHEREDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1466"]}},
    "EXHIBEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["e1468"]}},
    "EXHILARO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1473"]}},
    "EXHORREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["e1476"]}},
    "EXHORRESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1476"]}},
    "EXHORTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1479"]}},
    "EXHORTOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["e1483"]}},
    "EXIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1487"]}},
    "EXIGVE": {"-": {"morpho": ["rp--------"], "uri": ["28632"]}},
    "EXIGVITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1488"]}},
    "EXIGVO": {"-": {"morpho": ["rp--------"], "uri": ["109828"]}},
    "EXIGVVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["e1489"]},
        "2": {"morpho": ["rp--------"], "uri": []},
    },
    "EXIGVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e1489"]}},
    "EXILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["e1492"]}},
    "EXILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1493"]}},
    "EXILITER": {"-": {"morpho": ["rp--------"], "uri": ["46402"]}},
    "EXIMIE": {"-": {"morpho": ["rp--------"], "uri": ["39161"]}},
    "EXIMIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e1496"]}},
    "EXIMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1497"]}},
    "EXINANIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["e1498"]}},
    "EXINDE": {"-": {"morpho": ["rp--------"], "uri": ["e1501"]}},
    "EXISTIMATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1504"]}},
    "EXISTIMATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["e1505"]}},
    "EXISTIMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1506"]}},
    "EXITIABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["e1507"]}},
    "EXITIALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["e1508"]}},
    "EXITIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e1510"]}},
    "EXITIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e1511"]}},
    "EXITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["e9938"]}},
    "EXODIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e1524"]}},
    "EXOLEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["e1526"]}},
    "EXOLESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0661"]}},
    "EXOLETI": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "EXOLETVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "EXONERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1531"]}},
    "EXOPTATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["37343"]}},
    "EXOPTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1535"]}},
    "EXORABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["e1536"]}},
    "EXORDIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["e1549"]}},
    "EXORDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e1550"]}},
    "EXORIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["e1551"]}},
    "EXORNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1553"]}},
    "EXORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1556"]}},
    "EXORSA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["e9939"]}},
    "EXORSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["e9939"]}},
    "EXORTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["e9940"]}},
    "EXOS": {"-": {"morpho": ["aps---an3-"], "uri": ["e1558"]}},
    "EXOSCVLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["e1560"]}},
    "EXOSSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1561"]}},
    "EXOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e1563"]}},
    "EXPALLEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["e1566"]}},
    "EXPALLESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1566"]}},
    "EXPANDO": {"2": {"morpho": ["v1spia--3-"], "uri": ["e1572"]}},
    "EXPATRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1579"]}},
    "EXPAVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["e1582"]}},
    "EXPAVESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1582"]}},
    "EXPEDIO": {"-": {"morpho": ["v1spia--4-"], "uri": []}},
    "EXPEDITE": {"-": {"morpho": ["rp--------"], "uri": ["39394"]}},
    "EXPEDITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1589"]}},
    "EXPEDITO": {"-": {"morpho": ["v1spia--4-"], "uri": []}},
    "EXPEDITVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "EXPEDITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["35510"]}},
    "EXPELLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1592"]}},
    "EXPENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1593"]}},
    "EXPENSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1594"]}},
    "EXPENSVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e9978"]}},
    "EXPERGEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["e1596"]}},
    "EXPERGISCOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["e1600"]}},
    "EXPERGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1601"]}},
    "EXPERIENS": {"-": {"morpho": ["aps---an3i"], "uri": ["23107"]}},
    "EXPERIENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["e9989"]}},
    "EXPERIMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e1603"]}},
    "EXPERIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["e1604"]}},
    "EXPERISCOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["e0663"]}},
    "EXPERS": {"-": {"morpho": ["aps---cn3i"], "uri": ["e1607"]}},
    "EXPERTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1608"]}},
    "EXPERTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["29781"]}},
    "EXPETIBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["e1610"]}},
    "EXPETO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1612"]}},
    "EXPILATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1618"]}},
    "EXPILO": {"2": {"morpho": ["v1spia--1-"], "uri": ["e1620", "e1943"]}},
    "EXPINSO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1622"]}},
    "EXPIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1623"]}},
    "EXPLANABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["e1625"]}},
    "EXPLANATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1626"]}},
    "EXPLANATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["28493"]}},
    "EXPLANO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1630"]}},
    "EXPLAVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["98431"]}},
    "EXPLEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["e1633"]}},
    "EXPLETVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["32435"]}},
    "EXPLICABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["e1637"]}},
    "EXPLICATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1638"]}},
    "EXPLICATVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["e9942"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["109952"]},
    },
    "EXPLICITVS": {
        "2": {"morpho": ["aps---mn1-"], "uri": ["53130"]},
        "-": {"morpho": ["aps---mn1-"], "uri": ["53130"]},
    },
    "EXPLICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1641"]}},
    "EXPLODO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1642"]}},
    "EXPLORATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1643"]}},
    "EXPLORATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["e1644"]}},
    "EXPLORATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["50276"]}},
    "EXPLORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1647"]}},
    "EXPOLIO": {"2": {"morpho": ["v1spia--4-"], "uri": ["e1649"]}},
    "EXPOLITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1650"]}},
    "EXPOLITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["52368"]}},
    "EXPONO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1653"]}},
    "EXPORRIGO": {"2": {"morpho": ["v1spia--3-"], "uri": ["e1654"]}},
    "EXPORTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1655"]}},
    "EXPORTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1656"]}},
    "EXPOSCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1657"]}},
    "EXPOSITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["31461"]}},
    "EXPOSTVLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1663"]}},
    "EXPOSTVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1664"]}},
    "EXPRESSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1666"]}},
    "EXPRESSVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["e9944"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["23138"]},
    },
    "EXPRIMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1668"]}},
    "EXPROBRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1671"]}},
    "EXPROBRATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["e1673"]}},
    "EXPROBRATRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1674"]}},
    "EXPROBRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1675"]}},
    "EXPROMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1677"]}},
    "EXPVGNABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["e1682"]}},
    "EXPVGNATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1683"]}},
    "EXPVGNATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["e1684"]}},
    "EXPVGNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1688"]}},
    "EXPVLTRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1693"]}},
    "EXPVNGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1698"]}},
    "EXPVRGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1700"]}},
    "EXQVIRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1703"]}},
    "EXQVISITE": {"-": {"morpho": ["rp--------"], "uri": ["44165"]}},
    "EXQVISITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["44884"]}},
    "EXSANGVINATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e1710"]}},
    "EXSANGVIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["e1711"]}},
    "EXSANIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1712"]}},
    "EXSATIO": {"2": {"morpho": ["v1spia--1-"], "uri": ["e1714"]}},
    "EXSATVRABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["e1715"]}},
    "EXSATVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1717"]}},
    "EXSCALPO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1718"]}},
    "EXSCINDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1719"]}},
    "EXSCREO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1724"]}},
    "EXSCRIBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1725"]}},
    "EXSCVLPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1726"]}},
    "EXSECO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1729"]}},
    "EXSECRABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["e1730"]}},
    "EXSECRANDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["45291"]}},
    "EXSECRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1733"]}},
    "EXSECRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["37505"]}},
    "EXSECROR": {"-": {"morpho": ["v1spid--1-"], "uri": ["e1736"]}},
    "EXSECVTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1739"]}},
    "EXSEQVIAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["e1742"]}},
    "EXSEQVOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["e1747"]}},
    "EXSERO": {"2": {"morpho": ["v1spia--3-"], "uri": ["e1748"]}},
    "EXSERTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1749"]}},
    "EXSERTVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "EXSIBILO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1750"]}},
    "EXSICCESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1746"]}},
    "EXSICCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1753"]}},
    "EXSILIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["e1755"]}},
    "EXSILIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["47900"]}},
    "EXSISTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1795"]}},
    "EXSOLESCO": {"-": {"morpho": ["v1spia--3-"], "uri": [None, "e1944"]}},
    "EXSOLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1759"]}},
    "EXSOLVTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1760"]}},
    "EXSOMNIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["e1761"]}},
    "EXSORBEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["e1763"]}},
    "EXSORS": {"-": {"morpho": ["aps---an3-"], "uri": ["e1765"]}},
    "EXSPARGO": {"2": {"morpho": ["v1spia--3-"], "uri": ["98447"]}},
    "EXSPATIOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["e1766"]}},
    "EXSPECTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1768"]}},
    "EXSPECTATVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "EXSPECTATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["34565"]}},
    "EXSPECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1771"]}},
    "EXSPES": {"-": {"morpho": ["aps---an--"], "uri": ["e8772"]}},
    "EXSPIRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1774"]}},
    "EXSPOLIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1778"]}},
    "EXSPVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1781"]}},
    "EXSTANS": {"-": {"morpho": ["aps---an3i"], "uri": ["110081"]}},
    "EXSTANTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["e9990"]}},
    "EXSTERNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1784"]}},
    "EXSTIMVLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["e1786"]}},
    "EXSTIMVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1787"]}},
    "EXSTINCTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["e1789"]}},
    "EXSTINGVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1791"]}},
    "EXSTIRPO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1794"]}},
    "EXSTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1795"]}},
    "EXSTRVCTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1796"]}},
    "EXSTRVCTVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "EXSTRVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1799"]}},
    "EXSVCATVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "EXSVCTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["43120"]}},
    "EXSVDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1804"]}},
    "EXSVGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1806"]}},
    "EXSVL": {"-": {"morpho": ["n-s---mn3-"], "uri": ["98451"]}},
    "EXSVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["58355"]}},
    "EXSVLTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1810"]}},
    "EXSVLTIM": {"-": {"morpho": ["rp--------"], "uri": ["e1811"]}},
    "EXSVLTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1812"]}},
    "EXSVPERABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["e1816"]}},
    "EXSVPERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1817"]}},
    "EXSVPERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1819"]}},
    "EXSVRDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1821"]}},
    "EXSVRGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1822"]}},
    "EXTA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["e1826"]}},
    "EXTEMPLO": {"-": {"morpho": ["rp--------"], "uri": ["e1831"]}},
    "EXTEMPORALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["e1832"]}},
    "EXTENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1834"]}},
    "EXTENTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1839"]}},
    "EXTENTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1840"]}},
    "EXTENTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["50941"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["42005"]},
    },
    "EXTENVATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1841"]}},
    "EXTENVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1843"]}},
    "EXTER": {"-": {"morpho": ["aps---mn1r"], "uri": ["e1844"]}},
    "EXTERGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["e1846"]}},
    "EXTERMINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1852"]}},
    "EXTERNI": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "EXTERNVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "EXTERNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e1854"]}},
    "EXTERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1855"]}},
    "EXTERREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["e1857"]}},
    "EXTERRITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["57178"]}},
    "EXTIMESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1861"]}},
    "EXTOLLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1864"]}},
    "EXTORQVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["e1867"]}},
    "EXTORRIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["e1869"]}},
    "EXTRA": {
        "1": {"morpho": ["rp--------"], "uri": ["23194"]},
        "2": {"morpho": ["p---------"], "uri": ["e1872"]},
    },
    "EXTRAHO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1875"]}},
    "EXTRANEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e1880"]}},
    "EXTRAORDINARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e1881"]}},
    "EXTRARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e1882"]}},
    "EXTREMA": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "EXTREMITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1883"]}},
    "EXTREMVM": {
        "A": {"morpho": ["n-s---nn2-"], "uri": ["98460"]},
        "-": {"morpho": ["n-s---nn2-"], "uri": ["98460"]},
    },
    "EXTRICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1885"]}},
    "EXTRINSECVS": {"-": {"morpho": ["rp--------"], "uri": ["e1887"]}},
    "EXTRVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1890"]}},
    "EXTVNDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1896"]}},
    "EXTVRBO": {"2": {"morpho": ["v1spia--1-"], "uri": ["e1897"]}},
    "EXVBERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1901"]}},
    "EXVLCERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1906"]}},
    "EXVLCERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1909"]}},
    "EXVLVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1911"]}},
    "EXVNDATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1912"]}},
    "EXVNDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1913"]}},
    "EXVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1917"]}},
    "EXVRGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["e1918"]}},
    "EXVRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1919"]}},
    "EXVVIAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["e1922"]}},
    "FABA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["f0017"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "FABAGINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0003"]}},
    "FABALIA": {"-": {"morpho": ["n-p---nn3i"], "uri": ["98465"]}},
    "FABARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0005"]}},
    "FABELLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0008"]}},
    "FABER": {
        "1": {"morpho": ["n-s---mn2r"], "uri": ["f0011"]},
        "2": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "FABIANVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "FABIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "FABRICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0013"]}},
    "FABRICATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0015"]}},
    "FABRICATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["f0016"]}},
    "FABRICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0021"]}},
    "FABRICOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132478"]}},
    "FABRILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["f0024"]}},
    "FABVLA": {
        "1": {"morpho": ["n-s---fn1-"], "uri": ["f0027", "f0035"]},
        "2": {"morpho": ["n-s---fn1-"], "uri": ["f0027", "f0035"]},
    },
    "FABVLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["f0030"]}},
    "FABVLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["f0032"]}},
    "FABVLOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0034"]}},
    "FABVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["f0035"]}},
    "FACESSO": {"-": {"morpho": ["v1spia--3-"], "uri": ["f0037"]}},
    "FACETE": {"-": {"morpho": ["rp--------"], "uri": ["36913"]}},
    "FACETIAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["31121"]}},
    "FACETVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0042"]}},
    "FACIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["f0043"]}},
    "FACILE": {"-": {"morpho": ["rp--------"], "uri": ["23217"]}},
    "FACILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["f0044"]}},
    "FACILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0045"]}},
    "FACINOROSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0046"]}},
    "FACINVS": {"-": {"morpho": ["n-s---nn3-"], "uri": ["f0047"]}},
    "FACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["f0048"]}},
    "FACTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0051"]}},
    "FACTIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0053"]}},
    "FACTITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0057"]}},
    "FACTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["f0059"]}},
    "FACTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f9201"]}},
    "FACTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["f9201"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["29797"]},
    },
    "FACVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0063"]}},
    "FACVLTAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0065"]}},
    "FACVNDE": {"-": {"morpho": ["rp--------"], "uri": ["33973"]}},
    "FACVNDIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0067"]}},
    "FACVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0070"]}},
    "FAECARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0072"]}},
    "FAECATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0073"]}},
    "FAECVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0076"]}},
    "FAENVS": {"-": {"morpho": ["n-s---nn3-"], "uri": ["f0100"]}},
    "FAEX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0101"]}},
    "FAGINEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0103"]}},
    "FAGINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0104"]}},
    "FAGVS": {"-": {"morpho": ["aps---mn1-"], "uri": []}},
    "FALA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0106"]}},
    "FALARICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0109"]}},
    "FALCARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["f0110"]}},
    "FALCATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0112"]}},
    "FALCIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["f0114"]}},
    "FALCVLA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["f0119"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["f0119"]},
    },
    "FALLACIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0122"]}},
    "FALLAX": {"-": {"morpho": ["aps---an3-"], "uri": ["f0126"]}},
    "FALLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["f0127"]}},
    "FALSARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["f0129"]}},
    "FALSIPARENS": {"-": {"morpho": ["aps---an3-"], "uri": ["f0142"]}},
    "FALSO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["f0144"]},
        "2": {"morpho": ["rp--------"], "uri": ["98485"]},
    },
    "FALSVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f0084"]}},
    "FALSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["23240"]}},
    "FALX": {"-": {"morpho": ["n-s---fn3i"], "uri": ["f0146"]}},
    "FAMA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["f0147"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "FAMELICVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["f0150"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["f0150"]},
    },
    "FAMES": {
        "-": {"morpho": ["n-s---fn3i"], "uri": ["f0153"]},
        "N": {"morpho": ["n-s---mn1g"], "uri": []},
    },
    "FAMILIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0163"]}},
    "FAMILIARICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0165"]}},
    "FAMILIARIS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["98488"]},
        "2": {"morpho": ["aps---cn3i"], "uri": ["f0166"]},
    },
    "FAMILIARITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0167"]}},
    "FAMILIARITER": {"-": {"morpho": ["rp--------"], "uri": ["51946"]}},
    "FAMOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0171"]}},
    "FAMVL": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "FAMVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0079"]}},
    "FAMVLARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["f0174"]}},
    "FAMVLATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["f0182"]}},
    "FAMVLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["f0180"]}},
    "FAMVLVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["f0172"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["f0083"]},
    },
    "FANATICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0183"]}},
    "FANDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["23254"]}},
    "FANNIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "FANOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132778"]}},
    "FANVM": {
        "-": {"morpho": ["n-s---nn2-"], "uri": ["f0187"]},
        "N": {"morpho": ["n-s---nn2-"], "uri": ["f0187"]},
    },
    "FAR": {"-": {"morpho": ["n-s---nn3-"], "uri": ["f0188"]}},
    "FARCIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["f0193"]}},
    "FARINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0196"]}},
    "FARINARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0198"]}},
    "FARNVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["f0204"]}},
    "FARRAGO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0208"]}},
    "FARRARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f0209"]}},
    "FARRARIVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "FARRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0211"]}},
    "FARREO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "FARTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["f9218"]}},
    "FARTVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0181"]}},
    "FAS": {
        "-": {"morpho": ["n-s---nn--"], "uri": ["f0221"]},
        "N": {"morpho": ["n-s----n--"], "uri": []},
    },
    "FASCIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0222"]}},
    "FASCICVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["f0224"]}},
    "FASCINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0227"]}},
    "FASCINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0231"]}},
    "FASCIOLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0253"]}},
    "FASCIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["f0237"]}},
    "FASTI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["f0248"]}},
    "FASTIDIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["f0240"]}},
    "FASTIDIOSE": {"-": {"morpho": ["rp--------"], "uri": ["35265"]}},
    "FASTIDIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0241"]}},
    "FASTIDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f0242"]}},
    "FASTIGATE": {"-": {"morpho": ["rp--------"], "uri": ["53894"]}},
    "FASTIGATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["45792"]}},
    "FASTIGIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f0245"]}},
    "FASTIGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0246"]}},
    "FASTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["f0248"]},
        "2": {"morpho": ["n-s---mn4-"], "uri": ["f0248"]},
        "3": {"morpho": ["aps---mn1-"], "uri": ["f0248"]},
    },
    "FATALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["f0249"]}},
    "FATEOR": {"-": {"morpho": ["v1spid--2-"], "uri": ["f0251"]}},
    "FATIDICVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["f0255"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["f0255"]},
    },
    "FATIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["f0256"]}},
    "FATIGATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0258"]}},
    "FATIGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0260"]}},
    "FATISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["f0264"]}},
    "FATISCOR": {"-": {"morpho": ["v1spid--2-"], "uri": ["132483"]}},
    "FATVA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f9271"]}},
    "FATVM": {
        "-": {"morpho": ["n-s---nn2-"], "uri": ["f0779"]},
        "N": {"morpho": ["n-s---nn2-"], "uri": ["f0779"]},
    },
    "FATVVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["f9271"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["f9271"]},
    },
    "FAVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["f0271"]}},
    "FAVILLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0275"]}},
    "FAVNVS": {
        "-": {"morpho": ["n-s----n2-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "FAVONIVS": {
        "-": {"morpho": ["n-s----n2-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["f0282"]},
    },
    "FAVOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["f0283"]}},
    "FAVORABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["f0284"]}},
    "FAVSTVLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["f0286"]}},
    "FAVSTVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["f0287"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "FAVTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["f0288"]}},
    "FAVVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["f0300"]}},
    "FAVX": {"-": {"morpho": ["n-s---mn3i"], "uri": ["f0272"]}},
    "FAX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0301"]}},
    "FEBRICITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0305"]}},
    "FEBRICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0307"]}},
    "FEBRICVLOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0308"]}},
    "FEBRIS": {
        "-": {"morpho": ["n-s---fn3i"], "uri": ["f0312"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "FEBRVARIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["f0815"]},
    },
    "FECTO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "FECVNDITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0317"]}},
    "FECVNDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0318"]}},
    "FECVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0319"]}},
    "FEL": {"-": {"morpho": ["n-s---nn3-"], "uri": ["f0320"]}},
    "FELES": {"-": {"morpho": ["n-s---fn3i"], "uri": ["f0321"]}},
    "FELICITAS": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["f0322"]},
        "N": {"morpho": ["n-s---fn3-"], "uri": ["f0322"]},
    },
    "FELICITER": {"-": {"morpho": ["rp--------"], "uri": ["30609"]}},
    "FELIX": {
        "-": {"morpho": ["aps---an3i"], "uri": ["f0327"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "FELLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["f0328"]}},
    "FELLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0339"]}},
    "FEMELLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0341"]}},
    "FEMINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0343"]}},
    "FEMINEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0345"]}},
    "FEMININVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0346"]}},
    "FEMINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0348"]}},
    "FEMVR": {"-": {"morpho": ["n-s---nn3-"], "uri": ["f0347"]}},
    "FENARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["98515"]}},
    "FENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["110529"]}},
    "FENEBRIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["110530"]}},
    "FENERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["98516"]}},
    "FENERATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["49166"]}},
    "FENERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["51204"]}},
    "FENEROR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132484"]}},
    "FENESTRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0352"]}},
    "FENESTRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0353"]}},
    "FENICVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["56961"]}},
    "FENILIA": {"-": {"morpho": ["n-p---nn3i"], "uri": ["110542"]}},
    "FENISECA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["110543"]}},
    "FENVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["28719"]}},
    "FENVS": {"-": {"morpho": ["n-s---nn3-"], "uri": ["23302"]}},
    "FERA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0996"]}},
    "FERALIA": {
        "-": {"morpho": ["n-s----n3-"], "uri": []},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "FERALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["f0357", "f0349"]}},
    "FERAX": {"-": {"morpho": ["aps---an3i"], "uri": ["f0359"]}},
    "FERCIO": {"-": {"morpho": ["v1spia--4-"], "uri": []}},
    "FERCVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f0360"]}},
    "FERE": {"-": {"morpho": ["rp--------"], "uri": ["f0361"]}},
    "FERENTARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["f0362"]}},
    "FERENTINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "FERETRIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "FERETRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f0364"]}},
    "FERIAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["f0365"]}},
    "FERIATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0367"]}},
    "FERINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0369"]}},
    "FERINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0369"]}},
    "FERIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["f0370"]}},
    "FERITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0371"]}},
    "FERME": {"-": {"morpho": ["rp--------"], "uri": ["23308"]}},
    "FERMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f0381"]}},
    "FERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["f0382"]}},
    "FEROCIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0383"]}},
    "FEROCITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0385"]}},
    "FEROCITER": {"-": {"morpho": ["rp--------"], "uri": ["23311"]}},
    "FEROCVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0386"]}},
    "FERONIA": {
        "1N": {"morpho": ["n-s----n1-"], "uri": []},
        "2N": {"morpho": ["n-s----n1-"], "uri": []},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "FEROX": {
        "-": {"morpho": ["aps---an3i"], "uri": ["f0387"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "FERRAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f0389"]}},
    "FERRARIA": {
        "1": {"morpho": ["n-s---fn1-"], "uri": ["f0390"]},
        "2": {"morpho": ["n-s---fn1-"], "uri": ["f0390"]},
    },
    "FERRARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["f0390"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["f0390"]},
    },
    "FERRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0393"]}},
    "FERREA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0394"]}},
    "FERREVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0394"]}},
    "FERRVGINEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0400"]}},
    "FERRVGO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0402"]}},
    "FERRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f0403"]}},
    "FERTILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["f0407"]}},
    "FERTILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0408"]}},
    "FERTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f0411"]}},
    "FERVEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["f0413"]}},
    "FERVENS": {"-": {"morpho": ["aps---an3i"], "uri": ["98527"]}},
    "FERVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["f0414"]}},
    "FERVESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["f9911"]}},
    "FERVIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0415"]}},
    "FERVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0416"]}},
    "FERVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["f0414"]}},
    "FERVOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["f0421"]}},
    "FERVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["f0997"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["f0949"]},
    },
    "FESCENNINI": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "FESCENNINVS": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "FESSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0426"]}},
    "FESTINANTER": {"-": {"morpho": ["rp--------"], "uri": ["23324"]}},
    "FESTINATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0431"]}},
    "FESTINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0433"]}},
    "FESTIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0437"]}},
    "FESTVCA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0438"]}},
    "FESTVCVLA": {
        "2": {"morpho": ["n-s---fn1-"], "uri": ["f0443"]},
        "-": {"morpho": ["n-s---fn1-"], "uri": ["f0443"]},
    },
    "FESTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f0444"]}},
    "FESTVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["f0444"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "FETIALIS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["98531"]},
        "2": {"morpho": ["aps---cn3i"], "uri": ["f0446"]},
    },
    "FETVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0453"]}},
    "FETVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["f0456"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["f0456"]},
    },
    "FIBER": {"-": {"morpho": ["n-s---mn2r"], "uri": ["f0457"]}},
    "FIBRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0458"]}},
    "FIBVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0460"]}},
    "FIBVLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0461"]}},
    "FICANA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "FICARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0464"]}},
    "FICEDVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0466"]}},
    "FICIO": {"-": {"morpho": ["v1spia--3i"], "uri": []}},
    "FICTILE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["29144"]}},
    "FICTILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["f0473"]}},
    "FICTRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0478"]}},
    "FICTVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["58019"]},
        "2": {"morpho": ["rp--------"], "uri": []},
    },
    "FICTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["23338"]}},
    "FICVLEA": {"N": {"morpho": ["n-s---fn1-"], "uri": ["110645"]}},
    "FICVLNEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0480"]}},
    "FICVLNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0481"]}},
    "FICVS": {
        "1": {"morpho": ["n-s---cn2-"], "uri": ["f0482"]},
        "2": {"morpho": ["n-s---cn2-"], "uri": ["f0482"]},
    },
    "FIDELIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0490"]}},
    "FIDELIS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["110654"]},
        "2": {"morpho": ["aps---cn3i"], "uri": ["f0491"]},
    },
    "FIDELITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0492"]}},
    "FIDELITER": {"-": {"morpho": ["rp--------"], "uri": ["23343"]}},
    "FIDENAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "FIDENAS": {
        "N": {"morpho": ["aps---an3-"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "FIDENATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "FIDENS": {"-": {"morpho": ["aps---an3i"], "uri": ["98539"]}},
    "FIDENTER": {"-": {"morpho": ["rp--------"], "uri": ["58163"]}},
    "FIDES": {
        "1": {"morpho": ["n-s---fn3i"], "uri": ["f0496"]},
        "2": {"morpho": ["n-s---fn3i"], "uri": ["f0496"]},
        "1N": {"morpho": ["n-s---fn3i"], "uri": ["f0496"]},
        "2N": {"morpho": ["n-s---fn3i"], "uri": ["f0496"]},
        "3N": {"morpho": ["n-s---fn3i"], "uri": ["f0496"]},
        "N": {"morpho": ["n-s---mn1g"], "uri": []},
    },
    "FIDICEN": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["f0497"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "FIDICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0500"]}},
    "FIDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["f0502"]}},
    "FIDVCIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0503"]}},
    "FIDVCIARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0505"]}},
    "FIDVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["f0507"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "FIGLINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0998"]}},
    "FIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["f0512"]}},
    "FIGVLVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["f0517"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["f0517"]},
    },
    "FIGVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0518"]}},
    "FIGVRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0521"]}},
    "FIGVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0524"]}},
    "FILATIM": {"-": {"morpho": ["rp--------"], "uri": ["f0526"]}},
    "FILIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0528"]}},
    "FILICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0535"]}},
    "FILIOLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0537"]}},
    "FILIOLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["f0537"]}},
    "FILIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["f0538"]}},
    "FILIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0539"]}},
    "FILVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f0541"]}},
    "FIMBRIA": {
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
        "-": {"morpho": ["n-s----n1-"], "uri": []},
    },
    "FIMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["44913"]}},
    "FINDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["f0546"]}},
    "FINE": {"-": {"morpho": ["p---------"], "uri": []}},
    "FINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["f0548"]}},
    "FINIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["f0549"]}},
    "FINIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["f0550"]}},
    "FINITIMI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["f1128"]}},
    "FINITIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f1070"]}},
    "FINITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0553"]}},
    "FINITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["f0555"]}},
    "FINITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["37043"]}},
    "FIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["f0557"]}},
    "FIRMAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["f0558"]}},
    "FIRMAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f0559"]}},
    "FIRMATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["f0560"]}},
    "FIRME": {"-": {"morpho": ["rp--------"], "uri": ["47461"]}},
    "FIRMITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0561"]}},
    "FIRMITER": {"-": {"morpho": ["rp--------"], "uri": ["23367"]}},
    "FIRMITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0562"]}},
    "FIRMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0563"]}},
    "FIRMVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["f0564"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "FISCELLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0567"]}},
    "FISCINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0568"]}},
    "FISCVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["f0571"]}},
    "FISSILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["f0573"]}},
    "FISSVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f0581"]}},
    "FISSVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0552"]}},
    "FISTVCA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["110728"]}},
    "FISTVCATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["110729"]}},
    "FISTVCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["110730"]}},
    "FISTVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0577"]}},
    "FISTVLOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0583"]}},
    "FITEOR": {"-": {"morpho": ["v1spid--2-"], "uri": []}},
    "FITILLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0586"]}},
    "FIXVS": {
        "-": {
            "morpho": ["aps---mn1-", "n-s---mn4-"],
            "uri": ["28675", "f9204"],
        }
    },
    "FLABELLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f0592"]}},
    "FLABRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f0595"]}},
    "FLACCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["f0596"]}},
    "FLACCESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["f0596"]}},
    "FLACCIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0597"]}},
    "FLACCVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["f0598"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "FLAGELLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0601"]}},
    "FLAGELLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f0602"]}},
    "FLAGITATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0603"]}},
    "FLAGITATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["f0604"]}},
    "FLAGITIOSE": {"-": {"morpho": ["rp--------"], "uri": ["30775"]}},
    "FLAGITIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0606"]}},
    "FLAGITIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f0607"]}},
    "FLAGITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0608"]}},
    "FLAGRANS": {"-": {"morpho": ["aps---an3i"], "uri": ["98548"]}},
    "FLAGRANTER": {"-": {"morpho": ["rp--------"], "uri": ["53497"]}},
    "FLAGRANTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f9996"]}},
    "FLAGRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0613"]}},
    "FLAGRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f0614"]}},
    "FLAMEN": {
        "1": {"morpho": ["n-s---nn3-"], "uri": ["f1301"]},
        "2": {"morpho": ["n-s---nn3-"], "uri": ["f1301"]},
    },
    "FLAMINICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0619"]}},
    "FLAMININVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "FLAMINIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["57890"]}},
    "FLAMINIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": ["f0622"]},
    },
    "FLAMMA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["f0623"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "FLAMMEOLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f0628"]}},
    "FLAMMESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["f9644"]}},
    "FLAMMEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f0629"]}},
    "FLAMMEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0629"]}},
    "FLAMMIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["f0634"]}},
    "FLAMMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0644"]}},
    "FLATVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0666"]}},
    "FLATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["f9206"]}},
    "FLAVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["f0655"]}},
    "FLAVESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["f0655"]}},
    "FLAVIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "FLAVOLEIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "FLAVVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["f0658"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["98556"]},
    },
    "FLEBILE": {"-": {"morpho": ["rp--------"], "uri": ["37817"]}},
    "FLEBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["f0659"]}},
    "FLEBILITER": {"-": {"morpho": ["rp--------"], "uri": ["31603"]}},
    "FLECTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["f0660"]}},
    "FLEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["f0662"]}},
    "FLETVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["f9207"]}},
    "FLEXANIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0664"]}},
    "FLEXIBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["f0665"]}},
    "FLEXILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["f0668"]}},
    "FLEXO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0673"]}},
    "FLEXVOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0677"]}},
    "FLEXVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0696"]}},
    "FLEXVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["f9208"]}},
    "FLIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["f0678"]}},
    "FLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0679"]}},
    "FLOCCVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["f0683"]}},
    "FLORA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "FLORENS": {"-": {"morpho": ["aps---an3i"], "uri": ["98563"]}},
    "FLOREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["f0685"]}},
    "FLORESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["f9910"]}},
    "FLOREVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0687"]}},
    "FLORIDVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0689"]}},
    "FLORIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0690"]}},
    "FLORIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["f0691"]}},
    "FLOS": {"-": {"morpho": ["n-s---mn3-"], "uri": ["f0703"]}},
    "FLOSCVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["f9705"]}},
    "FLVCTIFRAGVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0708"]}},
    "FLVCTISONVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0712"]}},
    "FLVCTVATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f9715"]}},
    "FLVCTVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0716"]}},
    "FLVCTVOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132486"]}},
    "FLVCTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["f9210"]}},
    "FLVENS": {"-": {"morpho": ["aps---an3i"], "uri": ["98566"]}},
    "FLVENTER": {"-": {"morpho": ["rp--------"], "uri": ["31993"]}},
    "FLVENTISONVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0718"]}},
    "FLVENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f9719"]}},
    "FLVIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0722"]}},
    "FLVITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0723"]}},
    "FLVMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["f0725"]}},
    "FLVMINEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0729"]}},
    "FLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["f0730"]}},
    "FLVTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0723"]}},
    "FLVVIALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["f0737"]}},
    "FLVVIATICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0738"]}},
    "FLVVIATILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["f0739"]}},
    "FLVVIDVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "FLVVIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["f1195"]}},
    "FLVXVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["f9211"]},
        "2": {
            "morpho": ["aps---mn1-", "n-s---mn4-"],
            "uri": ["42805", "f9211"],
        },
    },
    "FOCALE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["f0750"]}},
    "FOCILLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["44254"]}},
    "FOCO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "FOCVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["f0754"]}},
    "FOCVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["f0755"]}},
    "FODICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0756"]}},
    "FODINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0757"]}},
    "FODIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["f0758"]}},
    "FOEDE": {"-": {"morpho": ["rp--------"], "uri": ["38166"]}},
    "FOEDERATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["53468"]}},
    "FOEDITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0761"]}},
    "FOEDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0762"]}},
    "FOEDVS": {
        "1": {"morpho": ["n-s---nn3-"], "uri": ["f0764"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["f0763"]},
    },
    "FOETIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0767"]}},
    "FOLIATVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f0774"]}},
    "FOLIATVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0775"]}},
    "FOLIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f0626"]}},
    "FOLLICVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["f0785"]}},
    "FOLLIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["f0788"]}},
    "FOMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f0794"]}},
    "FOMES": {"-": {"morpho": ["n-s---fn3-"], "uri": ["38132"]}},
    "FONS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["f0796"]}},
    "FONTANVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0800"]}},
    "FONTICVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["f0804"]}},
    "FOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["f0802"]}},
    "FORAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["f0807"]}},
    "FORAS": {"-": {"morpho": ["rp--------"], "uri": ["f0812"]}},
    "FORCEPS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0814"]}},
    "FORENSIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["f0818"]}},
    "FORFEX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0819"]}},
    "FORICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0824"]}},
    "FORIS": {
        "1": {"morpho": ["n-s---fn3i"], "uri": ["f0830"]},
        "2": {"morpho": ["rp--------"], "uri": ["f0831"]},
    },
    "FORMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0832"]}},
    "FORMAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f0836"]}},
    "FORMATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0838"]}},
    "FORMATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["f9838"]}},
    "FORMATVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f9997"]}},
    "FORMICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0840"]}},
    "FORMIDABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["f0847"]}},
    "FORMIDO": {
        "1": {"morpho": ["n-s---fn3-"], "uri": ["f0852"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["f0851"]},
        "N": {"morpho": ["n-s----n1-"], "uri": []},
    },
    "FORMIDOLOSE": {"-": {"morpho": ["rp--------"], "uri": ["44780"]}},
    "FORMIDOLOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0854"]}},
    "FORMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0857"]}},
    "FORMOSE": {"-": {"morpho": ["rp--------"], "uri": ["57759"]}},
    "FORMOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0861"]}},
    "FORMVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0863"]}},
    "FORNACVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0869"]}},
    "FORNAX": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["f0870"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "FORNICATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0873"]}},
    "FORNIX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["f0878"]}},
    "FORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0879"]}},
    "FORS": {
        "1": {"morpho": ["n-s---fn3-"], "uri": ["f0881"]},
        "2": {"morpho": ["rp--------"], "uri": ["f0882"]},
        "N": {"morpho": ["n-s----n--"], "uri": []},
    },
    "FORSAN": {"-": {"morpho": ["rp--------"], "uri": ["23449"]}},
    "FORSIT": {"-": {"morpho": ["rp--------"], "uri": ["23450"]}},
    "FORSITAN": {"-": {"morpho": ["rp--------"], "uri": ["23451"]}},
    "FORTASSE": {"-": {"morpho": ["rp--------"], "uri": ["23452"]}},
    "FORTAX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["f0883"]}},
    "FORTE": {"-": {"morpho": ["rp--------"], "uri": ["23454"]}},
    "FORTIS": {
        "-": {"morpho": ["aps---cn3i"], "uri": ["f0889"]},
        "A": {"morpho": ["aps---cn3i"], "uri": ["f0889"]},
    },
    "FORTITER": {"-": {"morpho": ["rp--------"], "uri": ["23457"]}},
    "FORTITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0891"]}},
    "FORTVITA": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "FORTVITO": {"-": {"morpho": ["rp--------"], "uri": ["23459"]}},
    "FORTVITV": {"-": {"morpho": ["rp--------"], "uri": ["29112"]}},
    "FORTVITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0893"]}},
    "FORTVNA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["f0895"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "FORTVNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0898"]}},
    "FORVLI": {
        "-": {"morpho": ["n-p---mn2-"], "uri": ["29786"]},
        "N": {"morpho": ["n-p---mn2-"], "uri": ["29786"]},
    },
    "FORVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["f0900"]},
        "2": {"morpho": ["n-s---nn2-"], "uri": ["f0900"]},
        "N": {"morpho": ["n-s---nn2-"], "uri": ["f0900"]},
    },
    "FORVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["f0901"]}},
    "FOSSA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0732"]}},
    "FOSSICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0903"]}},
    "FOSSILIS": {
        "1": {"morpho": ["n-s----n3-"], "uri": []},
        "2": {"morpho": ["aps---cn3i"], "uri": ["f0904"]},
    },
    "FOSSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0905"]}},
    "FOSSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["f0908"]}},
    "FOSSVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0910"]}},
    "FOSSVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f9998"]}},
    "FOVEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0912"]}},
    "FOVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["f0913"]}},
    "FRACES": {"-": {"morpho": ["n-s---mn3-"], "uri": ["f0916"]}},
    "FRACESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["f0915"]}},
    "FRACIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0918"]}},
    "FRACTVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0914"]}},
    "FRACTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["43415"]}},
    "FRAGA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["98596"]}},
    "FRAGILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["f0924"]}},
    "FRAGILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0925"]}},
    "FRAGMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["f0927"]}},
    "FRAGMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f0928"]}},
    "FRAGOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["f0929"]}},
    "FRAGOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0930"]}},
    "FRAGRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0931"]}},
    "FRAGVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f0922"]}},
    "FRAMEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0932"]}},
    "FRANGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["f0933"]}},
    "FRATER": {"-": {"morpho": ["n-s---mn3-"], "uri": ["f0934"]}},
    "FRATERCVLVS": {
        "-": {"morpho": ["aps---mn1-", "n-s---mn2-"], "uri": ["f9935"]},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "FRATERNITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0936"]}},
    "FRATERNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0937"]}},
    "FRAVDATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0947"]}},
    "FRAVDATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["f0948"]}},
    "FRAVDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0952"]}},
    "FRAVDVLENTER": {"-": {"morpho": ["rp--------"], "uri": ["57743"]}},
    "FRAVDVLENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0953"]}},
    "FRAVDVLENTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0954"]}},
    "FRAVS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0956"]}},
    "FRAXINEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0957"]}},
    "FRAXINVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["f0958"]}},
    "FREMITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["f9215"]}},
    "FREMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["f0962"]}},
    "FRENATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["f0964"]}},
    "FRENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["f0965"]}},
    "FRENO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0967"]}},
    "FRENVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f0968"]}},
    "FREQVENS": {"-": {"morpho": ["aps---an3i"], "uri": ["f0969"]}},
    "FREQVENTATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["40063"]}},
    "FREQVENTER": {"-": {"morpho": ["rp--------"], "uri": ["23489"]}},
    "FREQVENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0975"]}},
    "FREQVENTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0976"]}},
    "FRETVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f0980"]}},
    "FRETVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["f0981", "f0980"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["f0981"]},
    },
    "FRICATVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0990"]}},
    "FRICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0988"]}},
    "FRIGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["f0994"]}},
    "FRIGERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0995"]}},
    "FRIGESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["f0994"]}},
    "FRIGIDA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0294"]}},
    "FRIGIDARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0296"]}},
    "FRIGIDVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f1001"]}},
    "FRIGIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0293"]}},
    "FRIGO": {"1": {"morpho": ["v1spia--3-"], "uri": ["f1302"]}},
    "FRIGVS": {"-": {"morpho": ["n-s---nn3-"], "uri": ["f1008"]}},
    "FRINGO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "FRIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f1012"]}},
    "FRITILLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["f1300"]}},
    "FRIVOLA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["98602"]}},
    "FRIVOLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f1016"]}},
    "FRONDATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["f1023"]}},
    "FRONDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["f1024"]}},
    "FRONDESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["f1024"]}},
    "FRONDEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f1025"]}},
    "FRONDIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["f1027"]}},
    "FRONDOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f1029"]}},
    "FRONS": {
        "1": {"morpho": ["n-s---fn3i"], "uri": ["f1031"]},
        "2": {"morpho": ["n-s---fn3i"], "uri": ["f1031"]},
    },
    "FRONTATI": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "FRVCTIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["f8032"]}},
    "FRVCTVOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f1036"]}},
    "FRVCTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["f1058"]}},
    "FRVGALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["f1037"]}},
    "FRVGALITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f1038"]}},
    "FRVGALITER": {"-": {"morpho": ["rp--------"], "uri": ["23512"]}},
    "FRVGI": {
        "-": {"morpho": ["aps---an--"], "uri": ["23513"]},
        "C": {"morpho": ["n-s----n--"], "uri": []},
        "N": {"morpho": ["n-p---mn2-"], "uri": []},
    },
    "FRVGIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["f1040"]}},
    "FRVGIFERENS": {"-": {"morpho": ["aps---an3i"], "uri": ["f1043"]}},
    "FRVGIPARVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f1046"]}},
    "FRVMENTARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["f1052"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["f1052"]},
    },
    "FRVMENTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f1053"]}},
    "FRVMENTOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["f1056"]}},
    "FRVMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f1305"]}},
    "FRVOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["f1060"]}},
    "FRVSTRA": {"-": {"morpho": ["rp--------"], "uri": ["f1064"]}},
    "FRVSTRAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["f1066"]}},
    "FRVSTRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f1068"]}},
    "FRVSTROR": {"-": {"morpho": ["v1spid--1-"], "uri": ["f1071"]}},
    "FRVSTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f1074"]}},
    "FRVTEX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["f1077"]}},
    "FRVTICETVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f8076"]}},
    "FRVTICOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132493"]}},
    "FRVTICOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f1081"]}},
    "FRVX": {"-": {"morpho": ["n-s---fn3-"], "uri": []}},
    "FVCATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["54348"]}},
    "FVCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f1086"]}},
    "FVCVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["f1088"]},
        "2": {"morpho": ["n-s---mn2-"], "uri": ["f1088"]},
    },
    "FVFETIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "FVFICIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "FVGA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["f1090"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "FVGAX": {"-": {"morpho": ["aps---an3i"], "uri": ["f1091"]}},
    "FVGELA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f1095"]}},
    "FVGIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["f1096"]}},
    "FVGITIVVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["f1099"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["f0094"]},
    },
    "FVGITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f1100"]}},
    "FVGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f1103"]}},
    "FVLCIMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f1105"]}},
    "FVLCIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["f1106"]}},
    "FVLCRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f1107"]}},
    "FVLGENS": {"-": {"morpho": ["aps---an3i"], "uri": ["98615"]}},
    "FVLGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["f1109"]}},
    "FVLGIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f1112"]}},
    "FVLGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["f1109"]}},
    "FVLGOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["f1113"]}},
    "FVLGVR": {"-": {"morpho": ["n-s---nn3-"], "uri": ["f1114"]}},
    "FVLGVRATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["f1117"]}},
    "FVLGVRIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["f1118"]}},
    "FVLGVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f1125"]}},
    "FVLICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f1119"]}},
    "FVLIGO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f1123"]}},
    "FVLLO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["f1126"]}},
    "FVLLONICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0489"]}},
    "FVLLONICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0455"]}},
    "FVLLONIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f1129"]}},
    "FVLMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["f1130"]}},
    "FVLMENTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f8105"]}},
    "FVLMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f8105"]}},
    "FVLMINEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f1134"]}},
    "FVLMINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f1135"]}},
    "FVLTVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f9197"]}},
    "FVLVIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "FVLVVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["f1140"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "FVMEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f1147"]}},
    "FVMIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f1148"]}},
    "FVMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f1155"]}},
    "FVMOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f1156"]}},
    "FVMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["f1157"]}},
    "FVNALE": {"-": {"morpho": ["n-s---nn3-"], "uri": ["f1158"]}},
    "FVNALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["f1158"]}},
    "FVNCTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f1161"]}},
    "FVNCTVS": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "FVNDA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f1162"]}},
    "FVNDAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["f1163"]}},
    "FVNDAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f1165"]}},
    "FVNDATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f1167"]}},
    "FVNDATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["52669"]}},
    "FVNDITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["f1174"]}},
    "FVNDITVS": {"-": {"morpho": ["rp--------"], "uri": ["f1175"]}},
    "FVNDO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["f1177"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["f1176"]},
    },
    "FVNDVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["f0499"]}},
    "FVNDVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["f1179"]}},
    "FVNEBRIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["f1180"]}},
    "FVNEREVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f1186"]}},
    "FVNERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f1187"]}},
    "FVNESTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f1188"]}},
    "FVNESTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f8181"]}},
    "FVNGOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["f1184"]}},
    "FVNGVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["f8187"]}},
    "FVNICVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["f8188"]}},
    "FVNIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["f1189"]}},
    "FVNVS": {
        "-": {"morpho": ["n-s---nn3-"], "uri": ["f1191"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "FVR": {"-": {"morpho": ["n-s---cn3-"], "uri": ["f1192"]}},
    "FVRCA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f1200"]}},
    "FVRCIFER": {"-": {"morpho": ["n-s---mn2r"], "uri": ["f1201"]}},
    "FVRCILLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f1202"]}},
    "FVRENS": {"-": {"morpho": ["aps---an3i"], "uri": ["98631"]}},
    "FVRIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["f1215"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "FVRIALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["f1216"]}},
    "FVRIBVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f1219"]}},
    "FVRIO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["f1221"]},
        "2": {"morpho": ["v1spia--4-"], "uri": ["f1222"]},
    },
    "FVRIOSE": {"-": {"morpho": ["rp--------"], "uri": ["47938"]}},
    "FVRIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f1223"]}},
    "FVRIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "FVRNVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["f1227"]}},
    "FVRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["f1228"]}},
    "FVROR": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["f1231"]},
        "2": {"morpho": ["v1spid--1-", "n-s---mn3-"], "uri": [None, "f1231"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "FVRTIM": {"-": {"morpho": ["rp--------"], "uri": ["f1233"]}},
    "FVRTIVE": {"-": {"morpho": ["rp--------"], "uri": ["32852"]}},
    "FVRTIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f1234"]}},
    "FVRTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f1235"]}},
    "FVRVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f1239"]}},
    "FVSCINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f1242"]}},
    "FVSCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f1245"]}},
    "FVSCVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["f1246"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["33792"]},
    },
    "FVSILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["f1248"]}},
    "FVSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f1249"]}},
    "FVSIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "FVSTERNA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f1252"]}},
    "FVSTIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["f1256"]}},
    "FVSVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["f9221"]},
        "2": {"morpho": ["n-s---mn4-"], "uri": ["f9220"]},
        "3": {"morpho": ["aps---mn1-"], "uri": ["23570"]},
    },
    "FVTILE": {
        "1": {"morpho": ["n-s---nn3i"], "uri": ["111366"]},
        "2": {"morpho": ["rp--------"], "uri": ["111367"]},
    },
    "FVTILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["98637"]}},
    "FVTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f9706"]}},
    "FVTVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["f1265"]}},
    "FVTVRI": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "FVTVRVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "FVTVRVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["35629"]}},
    "FVTVTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f1266"]}},
    "GABII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "GABINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "GABINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "GADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "GAESVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["g0110"]}},
    "GAGATES": {"-": {"morpho": ["n-s---mn1g"], "uri": ["g0009"]}},
    "GAIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GALBA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "GALBANEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["g0017"]}},
    "GALBEVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["g0020"]}},
    "GALBINVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["g0026"]}},
    "GALEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["g0027"]}},
    "GALEATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["g0031"]}},
    "GALEO": {"-": {"morpho": ["v1spia--1-"], "uri": ["g0034"]}},
    "GALERVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["g0043"]}},
    "GALLA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["g0045"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "GALLI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "GALLIA": {"N": {"morpho": ["n-s---fn1-"], "uri": ["23575"]}},
    "GALLICVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": ["g0050"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["g0050"]},
    },
    "GALLINA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["g0057"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "GALLINACEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["g0058"]}},
    "GALLVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["g0229"]},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["g0229"]},
    },
    "GANEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["g0069"]}},
    "GANEO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["g0071"]}},
    "GANEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["g0069"]}},
    "GANGABA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["g0073"]}},
    "GANNEO": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "GANNIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["g0077"]}},
    "GANNITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["g2280"]}},
    "GANOSIS": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "GANYMEDES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "GARRIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["g0090"]}},
    "GARRVLITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["g0092"]}},
    "GARRVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["g0094"]}},
    "GARVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["g0030"]}},
    "GAVDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["g0102"]}},
    "GAVDIVM": {
        "-": {"morpho": ["n-s---nn2-"], "uri": ["g0105"]},
        "N": {"morpho": ["n-s---nn2-"], "uri": ["g0105"]},
    },
    "GAVIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GAVSAPATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["g0112"]}},
    "GAVSAPVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["g0113"]}},
    "GAZA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["g0115"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["g0115"]},
    },
    "GEGANIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GELENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "GELICIDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["g0125"]}},
    "GELIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["g0127"]}},
    "GELLIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "GELO": {"-": {"morpho": ["v1spia--1-"], "uri": ["g0128"]}},
    "GELV": {"-": {"morpho": ["n-s---nn4-"], "uri": ["g0132"]}},
    "GELVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["g0132"]}},
    "GELVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["55339"]}},
    "GEMELLVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["g0137"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["g0137"]},
    },
    "GEMINI": {
        "-": {"morpho": ["n-p---mn2-"], "uri": ["98650"]},
        "N": {"morpho": ["n-p---mn2-"], "uri": ["98650"]},
    },
    "GEMINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["g0140"]}},
    "GEMINVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["g0142"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["37240"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["g0142"]},
    },
    "GEMISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["g0144"]}},
    "GEMITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["g2282"]}},
    "GEMMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["g0147"]}},
    "GEMMATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["51847"]}},
    "GEMMEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["g0151"]}},
    "GEMMIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["g0152"]}},
    "GEMMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["g0153"]}},
    "GEMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["g0156"]}},
    "GENA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["g0160"]}},
    "GENEALOGVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["g0162"]}},
    "GENER": {"-": {"morpho": ["n-s---mn2r"], "uri": ["g0163"]}},
    "GENERALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["g0165"]}},
    "GENERALITER": {"-": {"morpho": ["rp--------"], "uri": ["35625"]}},
    "GENERASCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["g0168"]}},
    "GENERATIM": {"-": {"morpho": ["rp--------"], "uri": ["g0169"]}},
    "GENERATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["g0171"]}},
    "GENERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["g0174"]}},
    "GENEROSE": {"-": {"morpho": ["rp--------"], "uri": ["23605"]}},
    "GENEROSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["g0176"]}},
    "GENESIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["g0178"]}},
    "GENETHLIOLOGIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["111546"]}},
    "GENETRIX": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["g0182"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "GENIALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["g0183"]}},
    "GENIALITER": {"-": {"morpho": ["rp--------"], "uri": ["33529"]}},
    "GENICVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["g0192"]}},
    "GENISTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["g0194"]}},
    "GENITABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["g0196"]}},
    "GENITALIA": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "GENITALIS": {
        "-": {"morpho": ["aps---cn3i"], "uri": ["g0198"]},
        "A": {"morpho": ["aps---cn3i"], "uri": ["g0198"]},
    },
    "GENITALITER": {"-": {"morpho": ["rp--------"], "uri": ["98659"]}},
    "GENITOR": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["g0199"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "GENIVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["g0201"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["g0201"]},
    },
    "GENO": {"-": {"morpho": ["v1spia--3-"], "uri": ["g0202"]}},
    "GENS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["g0203"]}},
    "GENTICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["g0205"]}},
    "GENTILICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["g0206"]}},
    "GENTILIS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["98661"]},
        "2": {"morpho": ["aps---cn3i"], "uri": ["g0207"]},
    },
    "GENTILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["g0208"]}},
    "GENV": {"-": {"morpho": ["n-s---nn4-"], "uri": ["g0210"]}},
    "GENVALIA": {"-": {"morpho": ["n-p---nn3i"], "uri": ["111566"]}},
    "GENVCIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GENVINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["g0214"]}},
    "GENVS": {
        "1": {"morpho": ["n-s---nn3-"], "uri": ["g0213"]},
        "2": {"morpho": ["n-s---mn4-"], "uri": ["111568"]},
    },
    "GEOMETRES": {"-": {"morpho": ["n-s---mn1g"], "uri": ["g0220"]}},
    "GEOMETRIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["g0221"]}},
    "GEOMETRICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["g0224"]}},
    "GEOMETRICE": {
        "1": {"morpho": ["n-s---fn1g"], "uri": ["111577"]},
        "2": {"morpho": ["rp--------"], "uri": ["111578"]},
    },
    "GEOMETRICVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["111579"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["g0224"]},
    },
    "GERMANA": {"-": {"morpho": ["n-s---fn1-"], "uri": []}},
    "GERMANIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "GERMANICVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "GERMANVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["g0230"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["g0230"]},
        "N": {"morpho": ["aps---mn1-"], "uri": ["g0230"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["g0230"]},
    },
    "GERMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["g0233"]}},
    "GERMINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["g0237"]}},
    "GERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["g0238"]}},
    "GERVLVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["g0252"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["g0252"]},
    },
    "GERVSIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["g0254"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "GESTA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["31533"]}},
    "GESTAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["g0259"]}},
    "GESTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["g0260"]}},
    "GESTIO": {
        "1": {"morpho": ["n-s---fn3-"], "uri": ["g0270"]},
        "2": {"morpho": ["v1spia--4-"], "uri": ["g0271"]},
    },
    "GESTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["g0274"]}},
    "GESTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["g0279"]}},
    "GETA": {
        "C": {"morpho": ["n-s----n1-"], "uri": []},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "GETAE": {
        "-": {"morpho": ["n-s----n1-"], "uri": []},
        "N": {"morpho": ["n-p---fn1-"], "uri": []},
    },
    "GETICVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "GIBBER": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["g0280"]},
        "2": {"morpho": ["aps---mn1r"], "uri": ["g0280"]},
    },
    "GIBBVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["g0284"]}},
    "GIGAS": {
        "-": {"morpho": ["n-s----n-g"], "uri": []},
        "N": {"morpho": ["n-s---mn3-"], "uri": []},
    },
    "GIGNO": {"-": {"morpho": ["v1spia--3-"], "uri": ["g0288"]}},
    "GILVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["g0300"]}},
    "GINGIVA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["g0304"]}},
    "GLABER": {
        "1": {"morpho": ["n-s---mn2r"], "uri": ["98673"]},
        "2": {"morpho": ["aps---mn1r"], "uri": ["g0314"]},
    },
    "GLABRIO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "GLACIALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["g0319"]}},
    "GLACIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["g0320"]}},
    "GLACIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["g0322"]}},
    "GLADIATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["g0324"]}},
    "GLADIATORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["g0324"]}},
    "GLADIATVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["g0328"]}},
    "GLADIOLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["g0330"]}},
    "GLADIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["g0332"]}},
    "GLANDARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["g0336"]}},
    "GLANDIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["g0339"]}},
    "GLANS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["g0346"]}},
    "GLAREA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["g0348"]}},
    "GLAREOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["g0349"]}},
    "GLAVCIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "GLAVCVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["g0361"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["g0361"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["g0361"]},
    },
    "GLEBA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["g0364"]}},
    "GLEBVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["g0371"]}},
    "GLESVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["35642"]}},
    "GLISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["g0380"]}},
    "GLOBO": {"-": {"morpho": ["v1spia--1-"], "uri": ["g0385"]}},
    "GLOBOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["g0387"]}},
    "GLOBVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["g0389"]}},
    "GLOMERAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["g0393"]}},
    "GLOMERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["g0397"]}},
    "GLOMVS": {"-": {"morpho": ["n-s---nn3-"], "uri": ["g0399"]}},
    "GLORIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["g0400"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "GLORIOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["g0409"]}},
    "GLORIOSE": {"-": {"morpho": ["rp--------"], "uri": ["40337"]}},
    "GLORIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["g0410"]}},
    "GLOSSA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["g0414"]}},
    "GLVBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["g0422"]}},
    "GLVTEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["g0426"]}},
    "GLVTINATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["g0432"]}},
    "GLVTINVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["g0426"]}},
    "GLVTIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["29959"]}},
    "GLVTVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["111769"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["111768"]},
    },
    "GNAEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GNARVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["g0453"]}},
    "GNOMON": {"-": {"morpho": ["n-s---mn3-"], "uri": ["g0457"]}},
    "GNOMONICE": {"-": {"morpho": ["n-s---fn1g"], "uri": ["g0459"]}},
    "GNOMONICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["g0459"]}},
    "GOBIO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["g0195"]}},
    "GORGO": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "GRABATVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["g0475"]}},
    "GRACCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GRACILIS": {
        "-": {"morpho": ["aps---cn3i"], "uri": ["g0480"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "GRACILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["g0483"]}},
    "GRACVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["g0072"]}},
    "GRADARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["g0490"]}},
    "GRADATIM": {"-": {"morpho": ["rp--------"], "uri": ["g0491"]}},
    "GRADATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["g0492"]}},
    "GRADIOR": {"-": {"morpho": ["v1spid--3i"], "uri": ["g0498"]}},
    "GRADIVVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GRADVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["g0499"]}},
    "GRAECE": {"-": {"morpho": ["rp--------"], "uri": ["98686"]}},
    "GRAECI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "GRAECIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "GRAECOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["g0503"]}},
    "GRAECVLVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "GRAECVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "GRAII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "GRAIVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "GRAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["g0506"]}},
    "GRAMINEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["g0508"]}},
    "GRAMMATICA": {
        "1": {"morpho": ["n-p---nn2-"], "uri": ["111857"]},
        "2": {"morpho": ["n-p---nn2-"], "uri": ["111857"]},
    },
    "GRAMMATICVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["g0516"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["g0516"]},
    },
    "GRAMMICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["g0519"]}},
    "GRANARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["g0521"]}},
    "GRANATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["g0525"]}},
    "GRANDAEVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["g0528"]}},
    "GRANDE": {"-": {"morpho": ["rp--------"], "uri": ["54157"]}},
    "GRANDESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["g0529"]}},
    "GRANDIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["g0537"]}},
    "GRANDIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["g0538"]}},
    "GRANDISCAPIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["g0539"]}},
    "GRANDITER": {"-": {"morpho": ["rp--------"], "uri": ["42250"]}},
    "GRANDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["g0543"]}},
    "GRANEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["g0544"]}},
    "GRANEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["g0544"]}},
    "GRANIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["g0545"]}},
    "GRANIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GRANVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["g2278"]}},
    "GRAPHICVS": {
        "-": {
            "morpho": ["aps---mn1-", "aps---mn1-"],
            "uri": ["g0551", "g0551"],
        }
    },
    "GRAPHIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["g0554"]}},
    "GRAPHIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["g0555"]}},
    "GRASSATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["g0558"]}},
    "GRASSOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["g0560"]}},
    "GRATE": {"-": {"morpho": ["rp--------"], "uri": ["23692"]}},
    "GRATES": {"-": {"morpho": ["n-p---fn3-"], "uri": ["g0562"]}},
    "GRATIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["g0563"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "GRATIFICOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["g0566"]}},
    "GRATIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["g0570"]}},
    "GRATIS": {"-": {"morpho": ["rp--------"], "uri": ["53964"]}},
    "GRATOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["g0571"]}},
    "GRATVITO": {"-": {"morpho": ["rp--------"], "uri": ["23696"]}},
    "GRATVITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["g0572"]}},
    "GRATVLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["g0575"]}},
    "GRATVLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["g0578"]}},
    "GRATVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["g0579"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["111907"]},
    },
    "GRAVATIM": {"-": {"morpho": ["rp--------"], "uri": ["g0583"]}},
    "GRAVE": {"-": {"morpho": ["rp--------"], "uri": ["47787"]}},
    "GRAVEDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["g0586"]}},
    "GRAVEOLENS": {"-": {"morpho": ["aps---an3i"], "uri": ["g0587"]}},
    "GRAVESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["g0589"]}},
    "GRAVIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["g0594"]}},
    "GRAVIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["g0596"]}},
    "GRAVITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["g0598"]}},
    "GRAVITER": {"-": {"morpho": ["rp--------"], "uri": ["23706"]}},
    "GRAVITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["g0599"]}},
    "GRAVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["g0601"]}},
    "GRAVOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132502"]}},
    "GREDIOR": {"-": {"morpho": ["v1spid--3i"], "uri": []}},
    "GREGALES": {"-": {"morpho": ["n-p---mn3i"], "uri": ["98701"]}},
    "GREGALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["g0602"]}},
    "GREGARIVS": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": ["g0604"]},
    },
    "GREGATIM": {"-": {"morpho": ["rp--------"], "uri": ["g0605"]}},
    "GREGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["g0607"]}},
    "GREMIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["g0608"]}},
    "GRESSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["g0481"]}},
    "GREX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["g0612"]}},
    "GROSPHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GROSSVS": {"-": {"morpho": ["n-s---cn2-"], "uri": ["g0701"]}},
    "GRVMVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["g0627"]}},
    "GRVMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["g0628"]}},
    "GRVNNIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["37470"]}},
    "GRVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["g0631"]}},
    "GRVS": {"-": {"morpho": ["n-s---fn3-"], "uri": []}},
    "GRYPS": {"-": {"morpho": ["n-s---mn3-"], "uri": ["g2295"]}},
    "GVBERNACVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["g0637"]}},
    "GVBERNATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["g0638"]}},
    "GVBERNATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["g0639"]}},
    "GVBERNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["g0643"]}},
    "GVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["g0646"]}},
    "GVLOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["g0650"]}},
    "GVMMI": {"-": {"morpho": ["n-s---nn--"], "uri": ["40865"]}},
    "GVRGES": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["g0656"]},
        "N": {"morpho": ["n-p---mn3-"], "uri": []},
    },
    "GVRGVLIO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["g0657"]}},
    "GVSTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["g0666"]}},
    "GVSTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["g0668"]}},
    "GVTTA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["g0669"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["g0669"]},
    },
    "GVTTVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["g0672"]}},
    "GVTTVR": {"-": {"morpho": ["n-s---nn3-"], "uri": ["g0673"]}},
    "GVTTVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["44529"]}},
    "GYMNASIARCHVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["g0678"]}},
    "GYMNASIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["g0679"]}},
    "GYMNICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["g0682"]}},
    "GYNAECONITIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["g0687"]}},
    "GYPSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["g0690"]}},
    "GYPSVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["g0692"]}},
    "GYRVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["g0696"]}},
    "HABENA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["h0001"]}},
    "HABEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["h0004"]}},
    "HABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["h0005"]}},
    "HABILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["h0006"]}},
    "HABITABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["h0007"]}},
    "HABITANS": {"-": {"morpho": ["n-s---cn3-"], "uri": ["39940"]}},
    "HABITATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["h0009"]}},
    "HABITATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["h0011"]}},
    "HABITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["h0015"]}},
    "HABITVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["h0050"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["98716"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "HABONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HAC": {"-": {"morpho": ["rp--------"], "uri": ["39629"]}},
    "HACTENVS": {"-": {"morpho": ["rp--------"], "uri": ["h0956"]}},
    "HADRIANVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "HADRIATICVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "HADRVMENTVM": {
        "-": {"morpho": ["n-s----n2-"], "uri": []},
        "N": {"morpho": ["n-s---nn2-"], "uri": []},
    },
    "HAEDILIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["h0027"]}},
    "HAEDINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0042"]}},
    "HAEDVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["h0030"]}},
    "HAEDVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["h0031"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["h0031"]},
    },
    "HAEREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["h0044"]}},
    "HAERESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["h0029"]}},
    "HAERESIS": {"-": {"morpho": ["n-s---fn3g"], "uri": ["h0047"]}},
    "HAESITATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["h0051"]}},
    "HAESITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["h0053"]}},
    "HALAESA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "HALAESINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "HALAESINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "HALICARNASSVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HALICYENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "HALICYENSIS": {
        "N": {"morpho": ["aps---fn3i"], "uri": []},
        "A": {"morpho": ["aps---fn3i"], "uri": []},
    },
    "HALITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["h0067"]}},
    "HALO": {"-": {"morpho": ["v1spia--1-"], "uri": ["h0074"]}},
    "HAMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["h0081"]}},
    "HAMATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0087"]}},
    "HAMAXOPODES": {"-": {"morpho": ["n-s---mn3-"], "uri": ["112095"]}},
    "HAMILCAR": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "HAMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["h0101"]}},
    "HANNIBAL": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "HAPHE": {"-": {"morpho": ["n-s---fn1g"], "uri": ["h0104"]}},
    "HARA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["h0106"]}},
    "HARIOLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["h0041"]}},
    "HARMONIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["h0122"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["h0122"]},
    },
    "HARMONICE": {"-": {"morpho": ["n-s---fn1g"], "uri": ["h0123"]}},
    "HARPAGO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["h0128"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["h0127"]},
    },
    "HARVSPEX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["h0132"]}},
    "HARVSPICINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["h0135"]}},
    "HARVSPICIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["h0136"]}},
    "HASDRVBAL": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "HASTA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["h0137"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "HASTARIVS": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": ["h0138"]},
    },
    "HASTATVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["h0139"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["h0139"]},
    },
    "HASTILE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["h0142"]}},
    "HAVD": {"-": {"morpho": ["rp--------"], "uri": ["h0146"]}},
    "HAVDDVM": {"-": {"morpho": ["rp--------"], "uri": ["h0950"]}},
    "HAVDQVAQVAM": {"-": {"morpho": ["rp--------"], "uri": ["h0951"]}},
    "HAVRIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["h0147"]}},
    "HAVSTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["h0159"]}},
    "HEBEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["h0158"]}},
    "HEBES": {"-": {"morpho": ["aps---an3-"], "uri": ["h0160"]}},
    "HEBESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["h0158"]}},
    "HEBETATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["h0161"]}},
    "HEBETO": {"-": {"morpho": ["v1spia--1-"], "uri": ["h0163"]}},
    "HEBRVS": {
        "1N": {"morpho": ["n-s----n2-"], "uri": []},
        "2N": {"morpho": ["n-s----n2-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "HECATOMBE": {"-": {"morpho": ["n-s---fn1g"], "uri": ["h0165"]}},
    "HEDERA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["h0172"]}},
    "HEDERACEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0173"]}},
    "HEDERIGER": {"-": {"morpho": ["aps---mn1r"], "uri": ["h0175"]}},
    "HEI": {"-": {"morpho": ["----------"], "uri": []}},
    "HEIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HELENA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "HELEPOLIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["h0190"]}},
    "HELICON": {
        "-": {"morpho": ["n-s----n-g"], "uri": []},
        "N": {"morpho": ["n-s---nn3-"], "uri": []},
    },
    "HELIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["h0200"]}},
    "HELLEBORVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["33472"]}},
    "HELLEBORVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["112188"]}},
    "HELLESPONTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HELO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "HELVOLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0209"]}},
    "HELVOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132956"]}},
    "HEM": {"-": {"morpho": ["----------"], "uri": []}},
    "HEMICYCLIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["h0116"]}},
    "HEMINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["h0226"]}},
    "HEMIOLIOS": {"-": {"morpho": ["aps---mn1g"], "uri": ["h0228"]}},
    "HEMISPHAERIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["h0230"]}},
    "HEMITONIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["51868"]}},
    "HEMITRIGLYPHVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["h0235"]}},
    "HENDECASYLLABVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["55022"]}},
    "HENNA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "HENNENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "HENNENSIS": {
        "N": {"morpho": ["aps---cn3i"], "uri": []},
        "A": {"morpho": ["aps---cn3i"], "uri": []},
    },
    "HERA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["29841"]}},
    "HERACLEA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "HERACLEO": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "HERACLEVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "HERACLIENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "HERACLIENSIS": {
        "N": {"morpho": ["aps---cn3i"], "uri": []},
        "A": {"morpho": ["aps---cn3i"], "uri": []},
    },
    "HERACLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HERBA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["h0268"]}},
    "HERBIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0277"]}},
    "HERBITA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "HERBITENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "HERBITENSIS": {
        "N": {"morpho": ["aps---cn3i"], "uri": []},
        "A": {"morpho": ["aps---cn3i"], "uri": []},
    },
    "HERBOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0284"]}},
    "HERBVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["h0285"]}},
    "HERCISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["98744"]}},
    "HERCVLES": {
        "-": {"morpho": ["----------"], "uri": []},
        "N": {"morpho": ["n-s---mn3i"], "uri": []},
    },
    "HEREDITARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0293"]}},
    "HEREDITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["h0294"]}},
    "HEREDITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["h0295"]}},
    "HERENNIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HERES": {"-": {"morpho": ["n-s---cn3-"], "uri": ["h0297"]}},
    "HERI": {"-": {"morpho": ["rp--------"], "uri": ["h0298"]}},
    "HERIFVGA": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "HERILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["112290"]}},
    "HERMINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HERNICI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "HERNICVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "HERODOTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HEROICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0312"]}},
    "HEROINE": {"-": {"morpho": ["n-s---fn1g"], "uri": ["h0314"]}},
    "HEROIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["h0316"]}},
    "HEROS": {"-": {"morpho": ["n-s---mn3-"], "uri": ["h0317"]}},
    "HEROVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["h0318"]}},
    "HEROVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0318"]}},
    "HERSILIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "HERVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["e0843"]}},
    "HESIODVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HESPERIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "HESPERIVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": ["40938"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["40938"]},
    },
    "HESTERNO": {"-": {"morpho": ["rp--------"], "uri": ["112317"]}},
    "HESTERNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0322"]}},
    "HEV": {"-": {"morpho": ["----------"], "uri": []}},
    "HEVS": {"-": {"morpho": ["----------"], "uri": []}},
    "HEXACHORDOS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "HEXAPHOROS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "HEXASTYLOS": {"-": {"morpho": ["aps---mn1g"], "uri": ["h0345"]}},
    "HIASCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["h0351"]}},
    "HIATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["h0406"]}},
    "HIBEO": {"-": {"morpho": ["v1spia--2-"], "uri": []}},
    "HIBERNA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["55734"]}},
    "HIBERNACVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["h0354"]}},
    "HIBERNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["h0358"]}},
    "HIBERNVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["h0355"]}},
    "HIBERNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0355"]}},
    "HIBISCVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["h0359"]}},
    "HIC": {
        "1": {"morpho": ["aps---mn1-"], "uri": []},
        "2": {"morpho": ["rp--------"], "uri": ["23798"]},
    },
    "HICCINE": {"-": {"morpho": ["----------"], "uri": []}},
    "HIEMALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["h0363"]}},
    "HIEMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["h0365"]}},
    "HIEMS": {
        "-": {"morpho": ["n-s---fn3-"], "uri": []},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "HIERA": {
        "1": {"morpho": ["n-s---fn1-"], "uri": ["h0379"]},
        "2": {"morpho": ["n-s---fn1-"], "uri": ["h0379"]},
    },
    "HIERO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "HIERONICVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "HILARE": {"-": {"morpho": ["rp--------"], "uri": ["49750"]}},
    "HILARICVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0382"]}},
    "HILARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["h0388"]}},
    "HILARITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["h0383"]}},
    "HILARO": {"-": {"morpho": ["v1spia--1-"], "uri": ["h0385"]}},
    "HILARVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["h0388"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "HILLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["h0299"]}},
    "HILVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["h0390"]}},
    "HIMERA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "HINC": {"-": {"morpho": ["rp--------"], "uri": ["h0360"]}},
    "HINNIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["h0397"]}},
    "HINNITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["h0413"]}},
    "HINNVLEVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["h0399"]}},
    "HIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["h0402"]}},
    "HIPPOCAMELVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["h9411"]}},
    "HIPPOMANES": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "HIPPOPERAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["h0421"]}},
    "HIPPOPOTAMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["h0428"]}},
    "HIPPOTOTAE": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "HIPPOTOXOTAE": {"-": {"morpho": ["n-p---mn1-"], "uri": ["98758"]}},
    "HIRCINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0305"]}},
    "HIRCOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0440"]}},
    "HIRCVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["h0443"]}},
    "HIRNEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["h0444"]}},
    "HIRSVTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0453"]}},
    "HIRTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0455"]}},
    "HIRVNDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["h0459"]}},
    "HISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["h0460"]}},
    "HISPANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "HISPANIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "HISPANVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "HISPIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0461"]}},
    "HISTORIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["h0463"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "HISTORICVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["h0465"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["h0465"]},
    },
    "HISTORVMENOS": {"-": {"morpho": ["----------"], "uri": []}},
    "HISTRIO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["h0470"]}},
    "HISTRIONALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["h0471"]}},
    "HIVLCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["h0475"]}},
    "HIVLCVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0476"]}},
    "HODIE": {"-": {"morpho": ["rp--------"], "uri": ["h0478"]}},
    "HODIERNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0479"]}},
    "HOMERVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HOMICIDA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["h0511"]}},
    "HOMICIDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["h0515"]}},
    "HOMO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["h0518"]}},
    "HOMOEOMERIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["h0519"]}},
    "HOMOTONVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0528"]}},
    "HOMVLLVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["h0530"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["h0530"]},
    },
    "HOMVNCIO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["h0531"]}},
    "HOMVNCVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["h0532"]}},
    "HONESTAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["h0533"]}},
    "HONESTAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["h0534"]}},
    "HONESTE": {"-": {"morpho": ["rp--------"], "uri": ["23833"]}},
    "HONESTI": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "HONESTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["h0537"]}},
    "HONESTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": []}},
    "HONESTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0538"]}},
    "HONOR": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["23835"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "HONORARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0541"]}},
    "HONORATVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["53785"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "HONORIFICE": {"-": {"morpho": ["rp--------"], "uri": ["49260"]}},
    "HONORIFICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0545"]}},
    "HONORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["h0548"]}},
    "HONORVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0550"]}},
    "HORA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["h0553"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "HORATIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "HORATIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "HORDEACEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0559"]}},
    "HORDEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["h0563"]}},
    "HORIZON": {"-": {"morpho": ["n-s---mn3i"], "uri": ["h0571"]}},
    "HORNOTINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0574"]}},
    "HORNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0575"]}},
    "HOROLOGIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["h0577"]}},
    "HOROSCOPVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["h0581"]}},
    "HORRENDVM": {"-": {"morpho": ["rp--------"], "uri": ["53608"]}},
    "HORRENDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["48154"]}},
    "HORRENS": {"-": {"morpho": ["aps---an3i"], "uri": ["98771"]}},
    "HORREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["h0586"]}},
    "HORRESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["h0586"]}},
    "HORREVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["h0589"]}},
    "HORRIBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["h0590"]}},
    "HORRIDVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0592"]}},
    "HORRIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0593"]}},
    "HORRIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["h0594"]}},
    "HORRIFICE": {"-": {"morpho": ["rp--------"], "uri": ["98772"]}},
    "HORRIFICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["h0596"]}},
    "HORRIFICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0597"]}},
    "HORRISONVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0600"]}},
    "HORROR": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["h0601"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "HORTAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["h0603"]}},
    "HORTAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["h0604"]}},
    "HORTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["h0606"]}},
    "HORTATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["h0608"]}},
    "HORTATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["h0584"]}},
    "HORTENSIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HORTENTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HORTOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["h0614"]}},
    "HORTVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["h0618"]}},
    "HORTVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["h0619"]}},
    "HOSPES": {"-": {"morpho": ["n-s---mn3-"], "uri": ["23855"]}},
    "HOSPITA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["h0620"]}},
    "HOSPITALIA": {"-": {"morpho": ["n-p---nn3i"], "uri": ["112613"]}},
    "HOSPITALIS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["112614"]},
        "2": {"morpho": ["aps---cn3i"], "uri": ["h0621"]},
    },
    "HOSPITALITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["h0622"]}},
    "HOSPITALITER": {"-": {"morpho": ["rp--------"], "uri": ["29524"]}},
    "HOSPITIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["h0625"]}},
    "HOSPITOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["h0626"]}},
    "HOSPITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["56971"]}},
    "HOSTES": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "HOSTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["h0628"]}},
    "HOSTICVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["h0632"]}},
    "HOSTICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0632"]}},
    "HOSTILE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["98774"]}},
    "HOSTILIS": {"-": {"morpho": ["aps---cn3i"], "uri": []}},
    "HOSTILITER": {"-": {"morpho": ["rp--------"], "uri": ["51740"]}},
    "HOSTILIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "HOSTIS": {"-": {"morpho": ["n-s---mn3i"], "uri": []}},
    "HOSTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HVC": {"-": {"morpho": ["rp--------"], "uri": ["h0360"]}},
    "HVCCINE": {"-": {"morpho": ["rp--------"], "uri": ["46033"]}},
    "HVIVSMODI": {"-": {"morpho": ["rp--------"], "uri": ["h0958"]}},
    "HVMANA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["98782"]}},
    "HVMANE": {"-": {"morpho": ["rp--------"], "uri": ["51891"]}},
    "HVMANI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["98783"]}},
    "HVMANITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["h0646"]}},
    "HVMANITER": {"-": {"morpho": ["rp--------"], "uri": ["23864"]}},
    "HVMANITVS": {"-": {"morpho": ["rp--------"], "uri": ["h0647"]}},
    "HVMANVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["42301"]}},
    "HVMANVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0649"]}},
    "HVMECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["41908"]}},
    "HVMECTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["44310"]}},
    "HVMEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["u0985"]}},
    "HVMERVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["u0988"]}},
    "HVMESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["98784"]}},
    "HVMIDITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["41802"]}},
    "HVMIDVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["112639"]}},
    "HVMIDVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["38196"]}},
    "HVMIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["53065"]}},
    "HVMILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["h0655"]}},
    "HVMILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["h0656"]}},
    "HVMILITER": {"-": {"morpho": ["rp--------"], "uri": ["53161"]}},
    "HVMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["h0660"]}},
    "HVMOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["u0998"]}},
    "HVMVS": {"-": {"morpho": ["n-s---cn2-"], "uri": ["h0661"]}},
    "HYACINTHINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0663"]}},
    "HYACINTHVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["h0665"]}},
    "HYALVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["h0672"]}},
    "HYBLENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "HYBRIDA": {"-": {"morpho": ["n-s---cn1-"], "uri": ["h0673"]}},
    "HYDOR": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "HYDRA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["h0676"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["h0676"]},
    },
    "HYDRALETES": {"-": {"morpho": ["n-s---mn1g"], "uri": ["h0680"]}},
    "HYDRAVLICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0686"]}},
    "HYDRIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["h0692"]}},
    "HYDRIAS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "HYDROPICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0707"]}},
    "HYDROPS": {"-": {"morpho": ["n-s---mn3-"], "uri": ["h0710"]}},
    "HYDRVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["h0711"]}},
    "HYMEN": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["h0716"]},
        "N": {"morpho": ["n-s---mn3-"], "uri": ["h0716"]},
    },
    "HYMENAEOS": {"-": {"morpho": ["n-s---mn2g"], "uri": ["112702"]}},
    "HYMENEIVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "HYPAETROS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "HYPATE": {"-": {"morpho": ["n-s---fn1g"], "uri": ["h0738"]}},
    "HYPATOS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "HYPERBOLAEOS": {
        "1": {"morpho": ["n-s----n-g"], "uri": []},
        "2": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "HYPERBOLE": {"-": {"morpho": ["n-s---fn1g"], "uri": ["h0747"]}},
    "HYPERTHYRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["h0760"]}},
    "HYPOCAVSIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["h0770"]}},
    "HYPOGEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["h0349"]}},
    "HYPOMOCHLION": {"-": {"morpho": ["n-s---nn2g"], "uri": ["h0800"]}},
    "HYPOTRACHELIVM": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "HYSGINVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["h0821"]}},
    "IACCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "IACEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["i0002"]}},
    "IACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["i0003"]}},
    "IACTANS": {"-": {"morpho": ["aps---an3i"], "uri": ["99157"]}},
    "IACTANTER": {"-": {"morpho": ["rp--------"], "uri": ["41891"]}},
    "IACTANTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i9833"]}},
    "IACTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0007"]}},
    "IACTATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["i9851"]}},
    "IACTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0013"]}},
    "IACTVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i9832"]}},
    "IACTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["i9852"]}},
    "IACVLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i0018"]}},
    "IACVLATRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0020"]}},
    "IACVLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["i0021"]}},
    "IACVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i0022"]}},
    "IACVLVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["i0022"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["i0022"]},
    },
    "IALYSVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["101277"]}},
    "IAM": {"-": {"morpho": ["rp--------"], "uri": ["i0023"]}},
    "IAMBVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["i0029"]}},
    "IAMDIV": {"-": {"morpho": ["rp--------"], "uri": ["i9022"]}},
    "IAMDVDVM": {"-": {"morpho": ["rp--------"], "uri": ["i9023"]}},
    "IAMIAM": {"-": {"morpho": ["rp--------"], "uri": ["i0023"]}},
    "IAMPRIDEM": {"-": {"morpho": ["rp--------"], "uri": ["i9024"]}},
    "IANICVLVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "IANITOR": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["i0031"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "IANVA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i0035"]}},
    "IANVARIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": ["53744"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "IANVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": ["i9035"]},
        "-": {"morpho": ["n-s----n2-"], "uri": []},
    },
    "IASPIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0039"]}},
    "IBERVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": ["101280"]},
    },
    "IBI": {"-": {"morpho": ["rp--------"], "uri": ["i0049"]}},
    "IBIDEM": {"-": {"morpho": ["rp--------"], "uri": ["i0051"]}},
    "IBIS": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["i0052"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "ICHNEVMON": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i0054"]}},
    "ICHNOGRAPHIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i0055"]}},
    "ICILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["i0057"]}},
    "ICO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0057"]}},
    "ICONISMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["i0061"]}},
    "ICTERICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0066"]}},
    "ICTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["i0094"]}},
    "IDAEVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": ["98794"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "IDCIRCO": {"-": {"morpho": ["rp--------"], "uri": ["i0073"]}},
    "IDEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i0074"]}},
    "IDEM": {"-": {"morpho": ["aps---mn1-"], "uri": ["48306"]}},
    "IDENTIDEM": {"-": {"morpho": ["rp--------"], "uri": ["i0076"]}},
    "IDEO": {"-": {"morpho": ["rp--------"], "uri": ["i0078"]}},
    "IDIOTA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["i0082"]}},
    "IDONEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0093"]}},
    "IDOS": {"-": {"morpho": ["n-s---nn--"], "uri": ["i0096"]}},
    "IDVS": {"-": {"morpho": ["n-s----n4-"], "uri": []}},
    "IECTO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "IECVR": {"-": {"morpho": ["n-s---nn3-"], "uri": ["50231"]}},
    "IEIVNITAS": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "IEIVNIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i0112"]}},
    "IEIVNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0114"]}},
    "IETINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "IGITVR": {"-": {"morpho": ["----------"], "uri": []}},
    "IGNARVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0120"]}},
    "IGNAVE": {"-": {"morpho": ["rp--------"], "uri": ["46319"]}},
    "IGNAVIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i0122"]}},
    "IGNAVITER": {"-": {"morpho": ["rp--------"], "uri": ["98798"]}},
    "IGNAVVS": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": ["i0124"]},
    },
    "IGNESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0125"]}},
    "IGNEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0127"]}},
    "IGNIARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i0129"]}},
    "IGNICVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["i0135"]}},
    "IGNIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["i0137"]}},
    "IGNIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["i0144"]}},
    "IGNOBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0151"]}},
    "IGNOBILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0152"]}},
    "IGNOMINIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i0154"]}},
    "IGNOMINIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0156"]}},
    "IGNORANS": {"-": {"morpho": ["aps---an3-"], "uri": ["31608"]}},
    "IGNORANTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i0196"]}},
    "IGNORATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0159"]}},
    "IGNORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0160"]}},
    "IGNOSCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0162"]}},
    "IGNOTITIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i1831"]}},
    "IGNOTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["35051"]}},
    "IGO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "ILEX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0165"]}},
    "ILIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["98803"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["98803"]},
    },
    "ILICET": {"-": {"morpho": ["rp--------"], "uri": ["i0167"]}},
    "ILIGNEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0172"]}},
    "ILIGNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0173"]}},
    "ILIVM": {
        "-": {"morpho": ["n-s---nn2-"], "uri": ["112950"]},
        "N": {"morpho": ["n-s---nn3-"], "uri": []},
    },
    "ILLA": {"-": {"morpho": ["rp--------"], "uri": ["48663"]}},
    "ILLABOR": {"2": {"morpho": ["v1spid--3-"], "uri": ["i0176"]}},
    "ILLABORATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["98804"]}},
    "ILLABORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0177"]}},
    "ILLAC": {"-": {"morpho": ["rp--------"], "uri": ["42305"]}},
    "ILLACESSITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0180"]}},
    "ILLACRIMABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0181"]}},
    "ILLACRIMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0182"]}},
    "ILLACRIMOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132781"]}},
    "ILLAESVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["52232"]}},
    "ILLAETABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0188"]}},
    "ILLAQVEO": {"2": {"morpho": ["v1spia--1-"], "uri": ["i0192"]}},
    "ILLARGIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["i0193"]}},
    "ILLAVDATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0204"]}},
    "ILLAVTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["112970"]}},
    "ILLE": {"-": {"morpho": ["aps---mn1-"], "uri": ["46623"]}},
    "ILLECEBRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i0206"]}},
    "ILLECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0211"]}},
    "ILLECTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["i0225"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["98812"]},
    },
    "ILLEPIDE": {"-": {"morpho": ["rp--------"], "uri": ["98813"]}},
    "ILLEPIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0216"]}},
    "ILLI": {"-": {"morpho": ["rp--------"], "uri": ["35225"]}},
    "ILLIBATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0221"]}},
    "ILLIBERALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0222"]}},
    "ILLIC": {
        "1": {"morpho": ["aps---mn1-"], "uri": []},
        "2": {"morpho": ["rp--------"], "uri": ["45900"]},
    },
    "ILLICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["i0229"]}},
    "ILLICITE": {
        "1": {"morpho": ["rp--------"], "uri": ["112989"]},
        "2": {"morpho": ["rp--------"], "uri": ["112989"]},
    },
    "ILLICITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0233"]}},
    "ILLICO": {"-": {"morpho": ["rp--------"], "uri": ["33052"]}},
    "ILLIDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0235"]}},
    "ILLIGO": {"2": {"morpho": ["v1spia--1-"], "uri": ["i0238"]}},
    "ILLIM": {"-": {"morpho": ["rp--------"], "uri": ["i0205"]}},
    "ILLINC": {"-": {"morpho": ["rp--------"], "uri": ["i0205"]}},
    "ILLINO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0245"]}},
    "ILLITTERATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0250"]}},
    "ILLIVSMODI": {"-": {"morpho": ["rp--------"], "uri": []}},
    "ILLO": {"-": {"morpho": ["rp--------"], "uri": ["42938"]}},
    "ILLOC": {"-": {"morpho": ["rp--------"], "uri": ["30606"]}},
    "ILLOTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0255"]}},
    "ILLVC": {"-": {"morpho": ["rp--------"], "uri": ["37414"]}},
    "ILLVCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["i0258"]}},
    "ILLVCESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0258"]}},
    "ILLVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0264"]}},
    "ILLVMINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0268"]}},
    "ILLVO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "ILLVSTRIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0280"]}},
    "ILLVSTRO": {"2": {"morpho": ["v1spia--1-"], "uri": ["i0281"]}},
    "ILLVVIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["i0284"]}},
    "ILLYRICVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "IMACHARENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "IMACHARENSIS": {
        "N": {"morpho": ["aps---fn3i"], "uri": []},
        "A": {"morpho": ["aps---fn3i"], "uri": []},
    },
    "IMAGINARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0289"]}},
    "IMAGINATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0290"]}},
    "IMAGINOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["i0293"]}},
    "IMAGINOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0294"]}},
    "IMAGO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0295"]}},
    "IMBECILLI": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "IMBECILLIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0301"]}},
    "IMBECILLITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0299"]}},
    "IMBECILLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0301"]}},
    "IMBELLIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0303"]}},
    "IMBER": {"-": {"morpho": ["n-s---mn3i"], "uri": ["i0304"]}},
    "IMBERBIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0305"]}},
    "IMBIBO": {"2": {"morpho": ["v1spia--3-"], "uri": ["i0306"]}},
    "IMBREX": {"-": {"morpho": ["n-s---cn3-"], "uri": ["i0311"]}},
    "IMBRICATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["113057"]}},
    "IMBRIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["i0319"]}},
    "IMBVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0326"]}},
    "IMITABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0328"]}},
    "IMITAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i0330"]}},
    "IMITATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0331"]}},
    "IMITATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i0333"]}},
    "IMITATRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0335"]}},
    "IMITOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["i0336"]}},
    "IMMANE": {"-": {"morpho": ["rp--------"], "uri": ["28729"]}},
    "IMMANIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0344"]}},
    "IMMANITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0345"]}},
    "IMMANSVETVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0346"]}},
    "IMMATVRVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0352"]}},
    "IMMEDIATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0353"]}},
    "IMMEIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["i0357"]}},
    "IMMEMOR": {"-": {"morpho": ["aps---an3-"], "uri": ["i0358"]}},
    "IMMEMORABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0359"]}},
    "IMMENSVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["i0367"]},
        "2": {"morpho": ["rp--------"], "uri": ["49409"]},
    },
    "IMMENSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0367"]}},
    "IMMERENS": {"-": {"morpho": ["aps---an3i"], "uri": ["i0369"]}},
    "IMMERGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0370"]}},
    "IMMERITO": {"-": {"morpho": ["rp--------"], "uri": ["40784"]}},
    "IMMERITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0371"]}},
    "IMMERSABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0372"]}},
    "IMMETATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0374"]}},
    "IMMIGRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0376"]}},
    "IMMINEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["i0377"]}},
    "IMMINVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0378"]}},
    "IMMISCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["i0380"]}},
    "IMMISERABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0381"]}},
    "IMMISSARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i0384"]}},
    "IMMITIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0387"]}},
    "IMMITTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0388"]}},
    "IMMO": {"-": {"morpho": ["rp--------"], "uri": ["i0389"]}},
    "IMMOBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0390"]}},
    "IMMODERATE": {"-": {"morpho": ["rp--------"], "uri": ["34421"]}},
    "IMMODERATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0394"]}},
    "IMMODESTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i0395"]}},
    "IMMODESTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0396"]}},
    "IMMODICE": {"-": {"morpho": ["rp--------"], "uri": ["31500"]}},
    "IMMODICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0397"]}},
    "IMMOLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0400"]}},
    "IMMOLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0403"]}},
    "IMMORDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["i0404"]}},
    "IMMORIOR": {"-": {"morpho": ["v1spid--3i"], "uri": ["i0405"]}},
    "IMMOROR": {
        "-": {"morpho": ["v1spid--1-"], "uri": ["i0406"]},
        "1": {"morpho": ["v1spid--1-"], "uri": ["i0406"]},
    },
    "IMMORTALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0407"]}},
    "IMMORTALITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0408"]}},
    "IMMOTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i3124"]}},
    "IMMVGIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["i0411"]}},
    "IMMVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0416"]}},
    "IMMVNIO": {"2": {"morpho": ["v1spia--4-"], "uri": ["i0417"]}},
    "IMMVNIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0418"]}},
    "IMMVNITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0419"]}},
    "IMMVNITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i3126"]}},
    "IMMVRMVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0421"]}},
    "IMMVTABILIS": {
        "1": {"morpho": ["aps---cn3i"], "uri": ["i1337"]},
        "2": {"morpho": ["aps---cn3i"], "uri": ["i1337"]},
    },
    "IMMVTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0426"]}},
    "IMMVTO": {"2": {"morpho": ["v1spia--1-"], "uri": ["i0430"]}},
    "IMO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "IMPACATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0431"]}},
    "IMPAGES": {"-": {"morpho": ["n-s---fn3i"], "uri": ["i0435"]}},
    "IMPALLESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0436"]}},
    "IMPAR": {"-": {"morpho": ["aps---an3i"], "uri": ["i0440"]}},
    "IMPARATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0442"]}},
    "IMPATIENS": {"-": {"morpho": ["aps---an3i"], "uri": ["i0452"]}},
    "IMPATIENTER": {"-": {"morpho": ["rp--------"], "uri": ["43139"]}},
    "IMPATIENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i0453"]}},
    "IMPAVIDE": {"-": {"morpho": ["rp--------"], "uri": ["28769"]}},
    "IMPAVIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0455"]}},
    "IMPEDIMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i0460"]}},
    "IMPEDIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["i0461"]}},
    "IMPEDITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i9461"]}},
    "IMPEDITVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "IMPEDITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["23944"]}},
    "IMPELLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0464"]}},
    "IMPENDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["i0465"]}},
    "IMPENDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i0468"]}},
    "IMPENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0469"]}},
    "IMPENETRABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0470"]}},
    "IMPENSA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i0491"]}},
    "IMPENSE": {"-": {"morpho": ["rp--------"], "uri": ["23951"]}},
    "IMPENSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["45422"]}},
    "IMPERATOR": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["i0480"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "IMPERATORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0481"]}},
    "IMPERATVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i0512"]}},
    "IMPERCVSSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0485"]}},
    "IMPERFECTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0488"]}},
    "IMPERIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0493"]}},
    "IMPERITIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i0494"]}},
    "IMPERITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0495"]}},
    "IMPERITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0497"]}},
    "IMPERIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i0496"]}},
    "IMPERMISSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0500"]}},
    "IMPERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0503"]}},
    "IMPERPETVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0504"]}},
    "IMPERTIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["i0514"]}},
    "IMPERTVRBATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0519"]}},
    "IMPERVIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0520"]}},
    "IMPES": {"-": {"morpho": ["n-s---mn3-"], "uri": ["50332"]}},
    "IMPETRABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0528"]}},
    "IMPETRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0533"]}},
    "IMPETVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["i0521"]}},
    "IMPEXVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0537"]}},
    "IMPIE": {"-": {"morpho": ["rp--------"], "uri": ["50837"]}},
    "IMPIETAS": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["i0540"]},
        "N": {"morpho": ["n-s---fn3-"], "uri": ["i0540"]},
    },
    "IMPIGER": {"-": {"morpho": ["aps---mn1r"], "uri": ["i0542"]}},
    "IMPIGRE": {"-": {"morpho": ["rp--------"], "uri": ["57880"]}},
    "IMPINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0547"]}},
    "IMPIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0552"]}},
    "IMPIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0553"]}},
    "IMPLACABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0554"]}},
    "IMPLACABILITER": {"-": {"morpho": ["rp--------"], "uri": ["55979"]}},
    "IMPLACATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0556"]}},
    "IMPLACIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0557"]}},
    "IMPLECTO": {"2": {"morpho": ["v1spia--3-"], "uri": ["i0564"]}},
    "IMPLEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["i0566"]}},
    "IMPLICATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["23972"]}},
    "IMPLICITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["39572"]}},
    "IMPLICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0574"]}},
    "IMPLORATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0576"]}},
    "IMPLORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0577"]}},
    "IMPLVMBO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0578"]}},
    "IMPLVMIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0579"]}},
    "IMPLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0580"]}},
    "IMPLVVIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i0582"]}},
    "IMPOLITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0588"]}},
    "IMPOLLVTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0589"]}},
    "IMPONO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0591"]}},
    "IMPORTATICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0597"]}},
    "IMPORTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0599"]}},
    "IMPORTVNITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0600"]}},
    "IMPORTVNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0601"]}},
    "IMPORTVOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0602"]}},
    "IMPOS": {"-": {"morpho": ["aps---an3-"], "uri": ["i0603"]}},
    "IMPOTENS": {"-": {"morpho": ["aps---an3i"], "uri": ["i0612"]}},
    "IMPOTENTER": {"-": {"morpho": ["rp--------"], "uri": ["33451"]}},
    "IMPOTENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i0613"]}},
    "IMPRAESENTIARVM": {"-": {"morpho": ["rp--------"], "uri": ["i9917"]}},
    "IMPRANSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0624"]}},
    "IMPRECATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0625"]}},
    "IMPRECOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["i0626"]}},
    "IMPRESSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0628"]}},
    "IMPRESSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0629"]}},
    "IMPRIMIS": {"-": {"morpho": ["rp--------"], "uri": ["23985"]}},
    "IMPRIMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0632"]}},
    "IMPROBABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0634"]}},
    "IMPROBATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0635"]}},
    "IMPROBE": {"-": {"morpho": ["rp--------"], "uri": ["33785"]}},
    "IMPROBITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0637"]}},
    "IMPROBO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0639"]}},
    "IMPROBVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0641"]}},
    "IMPROBVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0642"]}},
    "IMPROCERVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0643"]}},
    "IMPROMPTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0649"]}},
    "IMPROPERATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i3132"]}},
    "IMPROPERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0652"]}},
    "IMPROSPER": {"-": {"morpho": ["aps---mn1r"], "uri": ["i0658"]}},
    "IMPROSPERE": {"-": {"morpho": ["rp--------"], "uri": ["23990"]}},
    "IMPROVIDE": {"-": {"morpho": ["rp--------"], "uri": ["46339"]}},
    "IMPROVIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0663"]}},
    "IMPROVISO": {"-": {"morpho": ["rp--------"], "uri": ["42750"]}},
    "IMPROVISVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["98856"]}},
    "IMPROVISVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0664"]}},
    "IMPRVDENS": {"-": {"morpho": ["aps---an3i"], "uri": ["i0665"]}},
    "IMPRVDENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i0666"]}},
    "IMPVBES": {"-": {"morpho": ["aps---an3-"], "uri": ["i0667"]}},
    "IMPVDENS": {"-": {"morpho": ["aps---an3i"], "uri": ["i0668"]}},
    "IMPVDENTER": {"-": {"morpho": ["rp--------"], "uri": ["23995"]}},
    "IMPVDENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i0669"]}},
    "IMPVDICITIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i0671"]}},
    "IMPVDICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0672"]}},
    "IMPVGNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0677"]}},
    "IMPVLSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i0680"]}},
    "IMPVLSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["i9860"]}},
    "IMPVNE": {"-": {"morpho": ["rp--------"], "uri": ["58342"]}},
    "IMPVNIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0684"]}},
    "IMPVNITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0686"]}},
    "IMPVNITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0687"]}},
    "IMPVRVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0693"]}},
    "IMPVTATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i0696"]}},
    "IMPVTATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i3133"]}},
    "IMPVTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0697"]}},
    "IMVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0700"]}},
    "IMVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["98862"]}},
    "IN": {"-": {"morpho": ["p---------"], "uri": ["i0702"]}},
    "INACCEDO": {"1": {"morpho": ["v1spia--3-"], "uri": []}},
    "INACCESSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0714"]}},
    "INACESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0718"]}},
    "INADVSTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0725"]}},
    "INAEDIFICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0726"]}},
    "INAEQVABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0727"]}},
    "INAEQVALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0729"]}},
    "INAEQVALITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0730"]}},
    "INAEQVALITER": {"-": {"morpho": ["rp--------"], "uri": ["41668"]}},
    "INAEQVATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["47300"]}},
    "INAEQVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0731"]}},
    "INAESTIMABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0732"]}},
    "INAESTVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0734"]}},
    "INAFFECTATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0735"]}},
    "INAFFECTO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "INAGITATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0738"]}},
    "INAMABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0747"]}},
    "INAMARESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0748"]}},
    "INAMBVLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0752"]}},
    "INAMBVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0753"]}},
    "INAMOENVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0755"]}},
    "INANE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["i0767"]}},
    "INANIMALIS": {"-": {"morpho": ["aps---cn3i"], "uri": []}},
    "INANIMATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0763"]}},
    "INANIMIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i3135"]}},
    "INANIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0765"]}},
    "INANIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["i0766"]}},
    "INANIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0767"]}},
    "INANITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0768"]}},
    "INANITER": {"-": {"morpho": ["rp--------"], "uri": ["51968"]}},
    "INARATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i3137"]}},
    "INARDEO": {"-": {"morpho": ["v1spia--2-"], "uri": []}},
    "INARDESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0778"]}},
    "INARESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0779"]}},
    "INARO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0783"]}},
    "INASCENSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0786"]}},
    "INASSVETVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0791"]}},
    "INAVDAX": {"-": {"morpho": ["aps---an3i"], "uri": ["i0796"]}},
    "INAVDIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["i0798"]}},
    "INAVDITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i3138"]}},
    "INAVRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1003"]}},
    "INAVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0805"]}},
    "INAVSPICATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i9805"]}},
    "INAVSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0806"]}},
    "INCAEDVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0811"]}},
    "INCALESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0816"]}},
    "INCALLIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0819"]}},
    "INCANDEO": {"-": {"morpho": ["v1spia--2-"], "uri": []}},
    "INCANDESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0822"]}},
    "INCANESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0824"]}},
    "INCANTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0828"]}},
    "INCANVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0829"]}},
    "INCASSVM": {"-": {"morpho": ["rp--------"], "uri": ["i9919"]}},
    "INCASTIGO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "INCAVTE": {"-": {"morpho": ["rp--------"], "uri": ["51051"]}},
    "INCAVTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0847"]}},
    "INCEDO": {"1": {"morpho": ["v1spia--3-"], "uri": ["i0848"]}},
    "INCELEBRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0851"]}},
    "INCENATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0852"]}},
    "INCENDIARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["i0855"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["i0855"]},
    },
    "INCENDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i0857"]}},
    "INCENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0858"]}},
    "INCENSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0860"]}},
    "INCENSVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i3141"]}},
    "INCENSVS": {
        "1": {"morpho": ["aps---mn1-"], "uri": ["i0863"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["i0863"]},
    },
    "INCEPTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i9861"]}},
    "INCEPTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["i9861"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["113503"]},
    },
    "INCERNICVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i0873"]}},
    "INCERNO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0874"]}},
    "INCERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0875"]}},
    "INCERTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i0878"]}},
    "INCERTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0878"]}},
    "INCESSO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0882"]}},
    "INCESSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["i0887"]}},
    "INCESTE": {"-": {"morpho": ["rp--------"], "uri": ["54112"]}},
    "INCESTIFICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0884"]}},
    "INCESTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0885"]}},
    "INCESTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i0888"]}},
    "INCESTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["i0888"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["i0888"]},
    },
    "INCHOO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0960"]}},
    "INCIDENTIA": {
        "-": {
            "morpho": ["n-s---fn1-", "n-s---fn1-"],
            "uri": ["31317", "31317"],
        }
    },
    "INCIDO": {
        "1": {"morpho": ["v1spia--3-"], "uri": ["i0897", "i0898", "i0897"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["i0897", "i0898", "i0897"]},
    },
    "INCIEO": {"-": {"morpho": ["v1spia--2-"], "uri": []}},
    "INCILE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["i0900"]}},
    "INCILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0900"]}},
    "INCILO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0901"]}},
    "INCINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0904"]}},
    "INCINO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0906"]}},
    "INCIO": {"-": {"morpho": ["v1spia--4-"], "uri": []}},
    "INCIPIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["i0907"]}},
    "INCISVM": {
        "-": {"morpho": ["n-s---nn2-"], "uri": ["i9862"]},
        "L": {"morpho": ["n-s---nn2-"], "uri": ["i9862"]},
    },
    "INCITAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i0918"]}},
    "INCITATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0919"]}},
    "INCITATVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["113547"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["57905"]},
    },
    "INCITO": {"1": {"morpho": ["v1spia--1-"], "uri": ["i0923"]}},
    "INCITVS": {
        "1": {"morpho": ["aps---mn1-"], "uri": ["i9849"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["i9849"]},
    },
    "INCLAMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0930"]}},
    "INCLARESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0931"]}},
    "INCLEMENS": {"-": {"morpho": ["aps---an3i"], "uri": ["i0937"]}},
    "INCLEMENTER": {"-": {"morpho": ["rp--------"], "uri": ["24042"]}},
    "INCLEMENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i0938"]}},
    "INCLINABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0939"]}},
    "INCLINATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0941"]}},
    "INCLINATVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["i9863"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["44070"]},
    },
    "INCLINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0944"]}},
    "INCLVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0945"]}},
    "INCLVTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0948"]}},
    "INCOACTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1358"]}},
    "INCOGITATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i3146"]}},
    "INCOGITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0955"]}},
    "INCOGNITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i3147"]}},
    "INCOHIBEO": {"-": {"morpho": ["v1spia--2-"], "uri": []}},
    "INCOHO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0960"]}},
    "INCOLA": {"-": {"morpho": ["n-s---cn1-"], "uri": ["i0963"]}},
    "INCOLO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["113593"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["i0965"]},
    },
    "INCOLVMIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0968"]}},
    "INCOLVMITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0969"]}},
    "INCOMITATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0972"]}},
    "INCOMMODE": {"-": {"morpho": ["rp--------"], "uri": ["33137"]}},
    "INCOMMODVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i0985"]}},
    "INCOMMODVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0985"]}},
    "INCOMPERTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0998"]}},
    "INCOMPOSITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1001"]}},
    "INCOMPREHENSIBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1002"]}},
    "INCOMPTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1005"]}},
    "INCONCESSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1007"]}},
    "INCONCINNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1010"]}},
    "INCONCVSSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1012"]}},
    "INCONDEMNATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1013"]}},
    "INCONDITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1014"]}},
    "INCONFVSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1017"]}},
    "INCONSIDERATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1040"]}},
    "INCONSTANS": {"-": {"morpho": ["aps---an3i"], "uri": ["i1049"]}},
    "INCONSTANTER": {"-": {"morpho": ["rp--------"], "uri": ["24061"]}},
    "INCONSTANTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i1050"]}},
    "INCONSVETVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1051"]}},
    "INCONSVLTE": {"-": {"morpho": ["rp--------"], "uri": ["24062"]}},
    "INCONSVLTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["i1052"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["i1052"]},
    },
    "INCONTAMINATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1059"]}},
    "INCONTINENS": {"-": {"morpho": ["aps---an3i"], "uri": ["i1064"]}},
    "INCONVENIENS": {"-": {"morpho": ["aps---an3i"], "uri": ["i1071"]}},
    "INCOQVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1080"]}},
    "INCORPORALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1084"]}},
    "INCORRVPTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1097"]}},
    "INCREBRESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1101"]}},
    "INCREDIBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0685"]}},
    "INCREDIBILITER": {"-": {"morpho": ["rp--------"], "uri": ["45657"]}},
    "INCREDVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1107"]}},
    "INCREMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i1111"]}},
    "INCREPATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["113741"]}},
    "INCREPITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1116"]}},
    "INCREPITVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["i1108"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["113742"]},
    },
    "INCREPO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1117"]}},
    "INCRESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1118"]}},
    "INCRVENTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1124"]}},
    "INCRVSTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1126"]}},
    "INCVBO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["i1132"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["i1131"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "INCVLCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1137"]}},
    "INCVLTE": {"-": {"morpho": ["rp--------"], "uri": ["36320"]}},
    "INCVLTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["i9866"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["i9866"]},
    },
    "INCVMBA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i1142"]}},
    "INCVMBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1143"]}},
    "INCVNABVLA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["i1145"]}},
    "INCVRIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i1153"]}},
    "INCVRIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1155"]}},
    "INCVRO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "INCVRRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1156"]}},
    "INCVRSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i1160"]}},
    "INCVRSITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1161"]}},
    "INCVRSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1162"]}},
    "INCVRSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["i9867"]}},
    "INCVRVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1169"]}},
    "INCVRVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1170"]}},
    "INCVS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i1171"]}},
    "INCVSATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i1173"]}},
    "INCVSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1176"]}},
    "INCVSSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["i9868"]}},
    "INCVSTODITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1178"]}},
    "INCVSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i3151"]}},
    "INCVTIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["i1179"]}},
    "INDAGATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i1181"]}},
    "INDAGATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i1182"]}},
    "INDAGATRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i1183"]}},
    "INDAGO": {
        "1": {"morpho": ["n-s---fn3-"], "uri": ["i1186"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["i1185"]},
    },
    "INDE": {"-": {"morpho": ["rp--------"], "uri": ["i1188"]}},
    "INDEBITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1189"]}},
    "INDECENS": {"-": {"morpho": ["aps---an3i"], "uri": ["98918"]}},
    "INDECENTER": {"-": {"morpho": ["rp--------"], "uri": ["44989"]}},
    "INDECENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i9838"]}},
    "INDECLINABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1192"]}},
    "INDECORE": {"-": {"morpho": ["rp--------"], "uri": ["51722"]}},
    "INDECORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1197"]}},
    "INDECORVS": {"-": {"morpho": ["aps---mn1-"], "uri": []}},
    "INDEFATIGABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1199"]}},
    "INDEFATIGATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1200"]}},
    "INDEFENSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1202"]}},
    "INDEFESSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1204"]}},
    "INDEFLEXVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1209"]}},
    "INDELEBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i8211"]}},
    "INDEMNIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1218"]}},
    "INDEPRAVATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1224"]}},
    "INDEPREHENSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["113825"]}},
    "INDESERTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1231"]}},
    "INDEX": {"-": {"morpho": ["n-s---cn3-"], "uri": ["i1246"]}},
    "INDIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "INDICIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i1251"]}},
    "INDICO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["i1252"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["i1253"]},
    },
    "INDICTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i1255"]}},
    "INDICTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i3155"]}},
    "INDIDEM": {"-": {"morpho": ["rp--------"], "uri": ["i1259"]}},
    "INDIFFERENS": {"-": {"morpho": ["aps---an3i"], "uri": ["i1260"]}},
    "INDIGENA": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "INDIGENS": {
        "1": {"morpho": ["n-s----n3-"], "uri": []},
        "2": {"morpho": ["aps---an3i"], "uri": ["98927"]},
    },
    "INDIGENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i1279"]}},
    "INDIGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["i1264"]}},
    "INDIGES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "INDIGESTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1269"]}},
    "INDIGNANDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["42003"]}},
    "INDIGNATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i1275"]}},
    "INDIGNE": {"-": {"morpho": ["rp--------"], "uri": ["33530"]}},
    "INDIGNITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i1278"]}},
    "INDIGNOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["i1280"]}},
    "INDIGNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1374"]}},
    "INDIGVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1282"]}},
    "INDILIGENS": {"-": {"morpho": ["aps---an3i"], "uri": ["i1284"]}},
    "INDILIGENTER": {"-": {"morpho": ["rp--------"], "uri": ["32251"]}},
    "INDILIGENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i1285"]}},
    "INDIPISCOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["i1290"]}},
    "INDISCRETVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1299"]}},
    "INDISPOSITE": {"-": {"morpho": ["rp--------"], "uri": ["98932"]}},
    "INDISTINCTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1318"]}},
    "INDIVIDVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1320"]}},
    "INDIVISVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1323"]}},
    "INDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1325"]}},
    "INDOCILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1328"]}},
    "INDOCTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1330"]}},
    "INDOLENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i1334"]}},
    "INDOLES": {"-": {"morpho": ["n-s---fn3i"], "uri": ["i1335"]}},
    "INDOLESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1336"]}},
    "INDOMITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1340"]}},
    "INDORMIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["i1342"]}},
    "INDOTATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1344"]}},
    "INDVBITATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i3158"]}},
    "INDVBIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1349"]}},
    "INDVCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1352"]}},
    "INDVCTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i1355"]}},
    "INDVCTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["i9870"]}},
    "INDVLGENS": {"-": {"morpho": ["aps---an3i"], "uri": ["98939"]}},
    "INDVLGENTER": {"-": {"morpho": ["rp--------"], "uri": ["40425"]}},
    "INDVLGENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i9840"]}},
    "INDVLGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["i1362"]}},
    "INDVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1366"]}},
    "INDVRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["28674"]}},
    "INDVRESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1368"]}},
    "INDVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1369"]}},
    "INDVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "INDVSTRIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i1376"]}},
    "INDVSTRIE": {"-": {"morpho": ["rp--------"], "uri": ["43203"]}},
    "INDVSTRIOSE": {"-": {"morpho": ["rp--------"], "uri": ["55532"]}},
    "INDVSTRIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1376"]}},
    "INDVTIAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["i1378"]}},
    "INDVTILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1377"]}},
    "INDVTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["i9872"]}},
    "INEBRIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1383"]}},
    "INEDIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i1385"]}},
    "INEFFICAX": {"-": {"morpho": ["aps---an3i"], "uri": ["i1390"]}},
    "INELABORATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1394"]}},
    "INELEGANS": {"-": {"morpho": ["aps---an3i"], "uri": ["i1395"]}},
    "INELVCTABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1402"]}},
    "INEMENDABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1405"]}},
    "INEMORIOR": {"-": {"morpho": ["v1spid--3i"], "uri": ["i1408"]}},
    "INEMPTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1411"]}},
    "INENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "INEO": {"1": {"morpho": ["v1spia--4-"], "uri": ["i1421"]}},
    "INEPTE": {"-": {"morpho": ["rp--------"], "uri": ["33132"]}},
    "INEPTIAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["98951"]}},
    "INEPTIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["i1423"]}},
    "INEPTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1426"]}},
    "INERMIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1430"]}},
    "INERMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1430"]}},
    "INERRANS": {"-": {"morpho": ["aps---an3i"], "uri": ["i3159"]}},
    "INERRO": {"2": {"morpho": ["v1spia--1-"], "uri": ["i1433"]}},
    "INERS": {"-": {"morpho": ["aps---an3-"], "uri": ["i1434"]}},
    "INERTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i1435"]}},
    "INESCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1439"]}},
    "INEVITABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1441"]}},
    "INEXCITABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1446"]}},
    "INEXCVSABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1453"]}},
    "INEXERCITATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1456"]}},
    "INEXERCITO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "INEXHAVSTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1459"]}},
    "INEXORABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1461"]}},
    "INEXPERTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1470"]}},
    "INEXPIABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1471"]}},
    "INEXPLEBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1475"]}},
    "INEXPLICABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1477"]}},
    "INEXPLORATO": {"-": {"morpho": ["rp--------"], "uri": ["41960"]}},
    "INEXPLORATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1480"]}},
    "INEXPVGNABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1482"]}},
    "INEXSPECTATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1489"]}},
    "INEXSVPERABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1493"]}},
    "INEXTRICABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1496"]}},
    "INFABRE": {"-": {"morpho": ["rp--------"], "uri": ["i1499"]}},
    "INFABRICATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1500"]}},
    "INFACETIAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["114070"]}},
    "INFACETVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["30188"]}},
    "INFAMIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i1507"]}},
    "INFAMIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1508"]}},
    "INFAMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1509"]}},
    "INFANDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1510"]}},
    "INFANS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["24153"]},
        "2": {"morpho": ["aps---an3i"], "uri": ["i1511"]},
    },
    "INFANTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i1513"]}},
    "INFATIGABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1520"]}},
    "INFATVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1522"]}},
    "INFAVSTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1523"]}},
    "INFECTIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1526"]}},
    "INFECTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["i9873"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["i3164"]},
    },
    "INFECVNDITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i1531"]}},
    "INFECVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1532"]}},
    "INFELICITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i1533"]}},
    "INFELICITER": {"-": {"morpho": ["rp--------"], "uri": ["24162"]}},
    "INFELIX": {"-": {"morpho": ["aps---an3i"], "uri": ["i1535"]}},
    "INFENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["114111"]}},
    "INFENSE": {"-": {"morpho": ["rp--------"], "uri": ["38970"]}},
    "INFENSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1538"]}},
    "INFENSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1539"]}},
    "INFERI": {
        "-": {"morpho": ["n-p---mn2-"], "uri": ["i1543"]},
        "N": {"morpho": ["n-p---mn2-"], "uri": ["i1543"]},
    },
    "INFERIAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["i8542"]}},
    "INFERIORA": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "INFERIORES": {"-": {"morpho": ["n-p---mn3-"], "uri": ["98974"]}},
    "INFERIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i8542"]}},
    "INFERNA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["98977"]}},
    "INFERNAS": {"-": {"morpho": ["aps---an3-"], "uri": ["i1546"]}},
    "INFERNE": {"-": {"morpho": ["rp--------"], "uri": ["24164"]}},
    "INFERNI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["98978"]}},
    "INFERNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1547"]}},
    "INFERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1548"]}},
    "INFERVEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["i1552"]}},
    "INFERVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["i1553"]}},
    "INFERVESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1553"]}},
    "INFERVS": {"-": {"morpho": ["aps---mn1-"], "uri": []}},
    "INFESTE": {"-": {"morpho": ["rp--------"], "uri": ["24166"]}},
    "INFESTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1557"]}},
    "INFESTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1558"]}},
    "INFEXIO": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "INFICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["i1562"]}},
    "INFICISCO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "INFIDELIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1564"]}},
    "INFIDELITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i1565"]}},
    "INFIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1566"]}},
    "INFIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1567"]}},
    "INFIMVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["98983"]}},
    "INFINDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1574"]}},
    "INFINITVM": {"-": {"morpho": ["rp--------"], "uri": []}},
    "INFINITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1581"]}},
    "INFIO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1586"]}},
    "INFIRMIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1585"]}},
    "INFIRMITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i1583"]}},
    "INFIRMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1584"]}},
    "INFIRMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1585"]}},
    "INFITIAS": {"-": {"morpho": ["n-s----n--"], "uri": []}},
    "INFITIATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i1592"]}},
    "INFITIATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i1593"]}},
    "INFITIOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["i1595"]}},
    "INFLAMMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1602"]}},
    "INFLATE": {"-": {"morpho": ["rp--------"], "uri": ["53610"]}},
    "INFLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i1604"]}},
    "INFLATVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["i1591"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["24183"]},
    },
    "INFLECTIO": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "INFLECTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1607"]}},
    "INFLEXIBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1609"]}},
    "INFLEXILIS": {
        "-": {"morpho": ["aps---cn3i"], "uri": []},
        "A": {"morpho": ["aps---cn3i"], "uri": []},
    },
    "INFLEXIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i1610"]}},
    "INFLIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1613"]}},
    "INFLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1614"]}},
    "INFLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1617"]}},
    "INFODIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["i1620"]}},
    "INFORMIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1626"]}},
    "INFORMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1628"]}},
    "INFRA": {
        "1": {"morpho": ["rp--------"], "uri": ["24190"]},
        "2": {"morpho": ["p---------"], "uri": ["i1634"]},
    },
    "INFRACTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i3170"]}},
    "INFRAGILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1639"]}},
    "INFREMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1640"]}},
    "INFRENDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["98989"]}},
    "INFRENVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1643"]}},
    "INFREQVENS": {"-": {"morpho": ["aps---an3i"], "uri": ["i1645"]}},
    "INFREQVENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i1647"]}},
    "INFRINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1652"]}},
    "INFRIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1653"]}},
    "INFRVCTVOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1658"]}},
    "INFRVNITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1660"]}},
    "INFVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i1664"]}},
    "INFVLCIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["i1667"]}},
    "INFVNDIBVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i1663"]}},
    "INFVNDO": {"2": {"morpho": ["v1spia--3-"], "uri": ["i1671"]}},
    "INFVSCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1672"]}},
    "INGEMINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1678"]}},
    "INGEMISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["58218"]}},
    "INGEMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1680"]}},
    "INGENERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1681"]}},
    "INGENIOSE": {"-": {"morpho": ["rp--------"], "uri": ["24203"]}},
    "INGENIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1687"]}},
    "INGENITVS": {
        "1": {"morpho": ["aps---mn1-"], "uri": ["i3174"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["i3174"]},
    },
    "INGENIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i1688"]}},
    "INGENS": {"-": {"morpho": ["aps---an3i"], "uri": ["i1689"]}},
    "INGENVE": {"-": {"morpho": ["rp--------"], "uri": ["32794"]}},
    "INGENVITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i1690"]}},
    "INGENVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1691"]}},
    "INGERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1692"]}},
    "INGLORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1698"]}},
    "INGLVVIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["i1700"]}},
    "INGRATE": {"-": {"morpho": ["rp--------"], "uri": ["24208"]}},
    "INGRATIS": {"-": {"morpho": ["rp--------"], "uri": ["33309"]}},
    "INGRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1706"]}},
    "INGRAVESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1708"]}},
    "INGREDIOR": {"-": {"morpho": ["v1spid--3i"], "uri": ["i1712"]}},
    "INGRESSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["i9880"]}},
    "INGRVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1714"]}},
    "INGVEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["i1715"]}},
    "INGVSTATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i8720"]}},
    "INHABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1722"]}},
    "INHABITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1727"]}},
    "INHAEREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["i1728"]}},
    "INHIBEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["i1732"]}},
    "INHIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1735"]}},
    "INHONESTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1740"]}},
    "INHONORIFICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1744"]}},
    "INHORREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["i1747"]}},
    "INHORRESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1747"]}},
    "INHOSPITALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1749"]}},
    "INHOSPITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1751"]}},
    "INHVMANE": {"-": {"morpho": ["rp--------"], "uri": ["56309"]}},
    "INHVMANITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i1754"]}},
    "INHVMANITER": {"-": {"morpho": ["rp--------"], "uri": ["40278"]}},
    "INHVMANVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1755"]}},
    "INHVMATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i8755"]}},
    "INIBI": {"-": {"morpho": ["rp--------"], "uri": ["i1758"]}},
    "INICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["i1759"]}},
    "INIECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i1760"]}},
    "INIECTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["i9882"]}},
    "INIMICITIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i1767"]}},
    "INIMICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1766"]}},
    "INIMICVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["98997"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["i1768"]},
    },
    "INIMO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "INIQVE": {"-": {"morpho": ["rp--------"], "uri": ["33737"]}},
    "INIQVITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i1780"]}},
    "INIQVVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["98998"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["i1782"]},
    },
    "INITIAMENTA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["99000"]}},
    "INITIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1788"]}},
    "INITIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i1789"]}},
    "INITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["i9883"]}},
    "INIVCVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1792"]}},
    "INIVDICATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1793"]}},
    "INIVNGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1797"]}},
    "INIVRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1798"]}},
    "INIVRIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i1799"]}},
    "INIVRIOSE": {"-": {"morpho": ["rp--------"], "uri": ["56958"]}},
    "INIVRIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1801"]}},
    "INIVROSVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "INIVSSVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["99003"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["i1805"]},
    },
    "INIVSTE": {"-": {"morpho": ["rp--------"], "uri": ["47506"]}},
    "INIVSTITIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i1806"]}},
    "INIVSTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["99004"]}},
    "INIVSTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1807"]}},
    "INLEXILIS": {
        "-": {"morpho": ["aps---cn3i"], "uri": []},
        "A": {"morpho": ["aps---cn3i"], "uri": []},
    },
    "INNABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1809"]}},
    "INNASCOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["i1812"]}},
    "INNATO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1813"]}},
    "INNECTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1816"]}},
    "INNITOR": {"2": {"morpho": ["v1spid--3-"], "uri": ["i1820"]}},
    "INNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1821"]}},
    "INNOCENS": {"-": {"morpho": ["aps---an3i"], "uri": ["i1823"]}},
    "INNOCENTER": {"-": {"morpho": ["rp--------"], "uri": ["30011"]}},
    "INNOCENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i1824"]}},
    "INNOCVE": {"-": {"morpho": ["rp--------"], "uri": ["40633"]}},
    "INNOCVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1825"]}},
    "INNOTESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1830"]}},
    "INNOXIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1837"]}},
    "INNVBILVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1840"]}},
    "INNVBIS": {
        "-": {"morpho": ["aps---cn3i"], "uri": ["i1841"]},
        "A": {"morpho": ["aps---cn3i"], "uri": ["i1841"]},
    },
    "INNVBVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1838"]}},
    "INNVMERABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1845"]}},
    "INNVMERABILITER": {"-": {"morpho": ["rp--------"], "uri": ["43718"]}},
    "INNVMERALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1847"]}},
    "INNVMERVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1850"]}},
    "INNVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1851"]}},
    "INNVPTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i1852"]}},
    "INNVPTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1852"]}},
    "INNVTRIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["i1854"]}},
    "INOBSEQVENS": {"-": {"morpho": ["aps---an3i"], "uri": ["i1864"]}},
    "INOBSERVABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1866"]}},
    "INOBSERVATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1869"]}},
    "INODORVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1878"]}},
    "INOFFENSE": {"-": {"morpho": ["rp--------"], "uri": ["99015"]}},
    "INOFFENSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1879"]}},
    "INOFFICIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1880"]}},
    "INOLENS": {"-": {"morpho": ["aps---an3i"], "uri": ["i1881"]}},
    "INOLESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1882"]}},
    "INOMINATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1884"]}},
    "INOPERATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i3182"]}},
    "INOPIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i1888"]}},
    "INOPINANS": {"-": {"morpho": ["aps---an3i"], "uri": ["i1891"]}},
    "INOPINATVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["99019"]}},
    "INOPINATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1892"]}},
    "INOPINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1893"]}},
    "INOPS": {"-": {"morpho": ["aps---an3-"], "uri": ["i1897"]}},
    "INORDINATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1905"]}},
    "INORNATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i8907"]}},
    "INQVIES": {
        "1": {"morpho": ["n-s----n3-"], "uri": []},
        "2": {"morpho": ["aps---an3-"], "uri": ["i1921"]},
    },
    "INQVIETO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1919"]}},
    "INQVIETVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1921"]}},
    "INQVILINVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["i1922"]}},
    "INQVINAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i1924"]}},
    "INQVINATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["45232"]}},
    "INQVINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1926"]}},
    "INQVIO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "INQVIRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1928"]}},
    "INQVISITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i1930"]}},
    "INQVISITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i1932"]}},
    "INREGILLVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "INSAEPTVS": {
        "1": {"morpho": ["aps---mn1-"], "uri": ["i3185"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["i3185"]},
    },
    "INSALVBRIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1936"]}},
    "INSANABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1938"]}},
    "INSANE": {"-": {"morpho": ["rp--------"], "uri": ["24263"]}},
    "INSANIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i1940"]}},
    "INSANIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["i1941"]}},
    "INSANVM": {"-": {"morpho": ["rp--------"], "uri": ["28373"]}},
    "INSANVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1943"]}},
    "INSAPIENS": {
        "-": {"morpho": ["aps---an3-"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "INSATIABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1946"]}},
    "INSATIABILITER": {"-": {"morpho": ["rp--------"], "uri": ["43167"]}},
    "INSCENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1956"]}},
    "INSCIENS": {
        "1": {"morpho": ["aps---an3i"], "uri": ["i1964"]},
        "2": {"morpho": ["aps---an3i"], "uri": ["i1964"]},
    },
    "INSCIENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i1958"]}},
    "INSCITE": {"-": {"morpho": ["rp--------"], "uri": ["46193"]}},
    "INSCITIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i1960"]}},
    "INSCITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1962"]}},
    "INSCIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1963"]}},
    "INSCRIBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1965"]}},
    "INSCRIPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i1967"]}},
    "INSCVLPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i8955"]}},
    "INSECABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1971"]}},
    "INSECO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1973"]}},
    "INSECTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i1972"]}},
    "INSECTOR": {"2": {"morpho": ["v1spid--1-"], "uri": ["i1977"]}},
    "INSEDABILITER": {"-": {"morpho": ["rp--------"], "uri": ["i1982"]}},
    "INSEMINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1986"]}},
    "INSENESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1987"]}},
    "INSENSIBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1990"]}},
    "INSEPARABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1995"]}},
    "INSEPVLTVS": {
        "1": {"morpho": ["aps---mn1-"], "uri": ["i1997"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["i1997"]},
    },
    "INSEQVOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["i1998"]}},
    "INSERO": {
        "2": {"morpho": ["v1spia--3-"], "uri": ["i2002"]},
        "3": {"morpho": ["v1spia--3-"], "uri": ["i2002"]},
    },
    "INSERTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2007"]}},
    "INSERVIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["i2008"]}},
    "INSIDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["i2016"]}},
    "INSIDIAE": {
        "-": {"morpho": ["n-p---fn1-"], "uri": ["43269"]},
        "N": {"morpho": ["n-p---fn1-"], "uri": ["43269"]},
    },
    "INSIDIATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i2018"]}},
    "INSIDIOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["i2020"]}},
    "INSIDIOSE": {"-": {"morpho": ["rp--------"], "uri": ["29765"]}},
    "INSIDIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2021"]}},
    "INSIDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2022"]}},
    "INSIGNE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["i2023"]}},
    "INSIGNIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["i2026"]}},
    "INSIGNIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i2023"]}},
    "INSIGNITA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["114576"]}},
    "INSIGNITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["29732"]}},
    "INSILIA": {"-": {"morpho": ["n-p---nn3i"], "uri": ["99039"]}},
    "INSILIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["i2030"]}},
    "INSIMVLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2033"]}},
    "INSIMVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2031"]}},
    "INSINCERVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2035"]}},
    "INSINVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2039"]}},
    "INSIPIENS": {"-": {"morpho": ["aps---an3i"], "uri": ["i2041"]}},
    "INSIPIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["i7042"]}},
    "INSISTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2043"]}},
    "INSITICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2044"]}},
    "INSITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2045"]}},
    "INSITIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2046"]}},
    "INSITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i2047"]}},
    "INSITVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["i9885"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["57089"]},
    },
    "INSOCIABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i2048"]}},
    "INSOLABILITER": {"-": {"morpho": ["rp--------"], "uri": ["i2050"]}},
    "INSOLENS": {"-": {"morpho": ["aps---an3i"], "uri": ["i2052"]}},
    "INSOLENTER": {"-": {"morpho": ["rp--------"], "uri": ["24296"]}},
    "INSOLENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i2053"]}},
    "INSOLESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2054"]}},
    "INSOLITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2056"]}},
    "INSOLVBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i2059"]}},
    "INSOMNIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i2062"]}},
    "INSOMNIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i2067"]}},
    "INSONO": {"1": {"morpho": ["v1spia--1-"], "uri": ["i2069"]}},
    "INSONS": {"-": {"morpho": ["aps---an3i"], "uri": ["i2070"]}},
    "INSOPOR": {"-": {"morpho": ["aps---an3-"], "uri": ["i2074"]}},
    "INSPECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2082"]}},
    "INSPECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2084"]}},
    "INSPECTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i2085"]}},
    "INSPECTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["i9886"]}},
    "INSPERANS": {"-": {"morpho": ["aps---an3i"], "uri": ["i2088"]}},
    "INSPERATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2089"]}},
    "INSPERGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2090"]}},
    "INSPICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["i2093"]}},
    "INSPICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2095"]}},
    "INSPIRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2100"]}},
    "INSPVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2104"]}},
    "INSPVRCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2105"]}},
    "INSTABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i2108"]}},
    "INSTANS": {"-": {"morpho": ["aps---an3i"], "uri": ["99056"]}},
    "INSTANTER": {"-": {"morpho": ["rp--------"], "uri": ["53974"]}},
    "INSTAR": {"-": {"morpho": ["n-s---nn--"], "uri": ["i2110"]}},
    "INSTAVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2115"]}},
    "INSTERNO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2117"]}},
    "INSTIGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2122"]}},
    "INSTILLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2124"]}},
    "INSTINCTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["i9889"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["114660"]},
    },
    "INSTINGVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2128"]}},
    "INSTIPO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2129"]}},
    "INSTITA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i2131"]}},
    "INSTITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i2134"]}},
    "INSTITVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2136"]}},
    "INSTITVTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2137"]}},
    "INSTITVTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i3191"]}},
    "INSTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2140"]}},
    "INSTRAGVLVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "INSTRATVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i9890"]}},
    "INSTREPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2144"]}},
    "INSTRINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2146"]}},
    "INSTRVCTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2148"]}},
    "INSTRVCTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["i9891"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["24319"]},
    },
    "INSTRVMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i2150"]}},
    "INSTRVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2151"]}},
    "INSVAVIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i2157"]}},
    "INSVDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2167"]}},
    "INSVEFACTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["38437"]}},
    "INSVESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2169"]}},
    "INSVETVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i3192"]}},
    "INSVLA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["i2175"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "INSVLANVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["i2176"]}},
    "INSVLSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2184"]}},
    "INSVLTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2190"]}},
    "INSVM": {"1": {"morpho": ["v1spia--1-"], "uri": ["i2191"]}},
    "INSVMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2193"]}},
    "INSVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2195"]}},
    "INSVPER": {
        "1": {"morpho": ["rp--------"], "uri": ["57801"]},
        "2": {"morpho": ["p---------"], "uri": ["i2196"]},
    },
    "INSVPERABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i2197"]}},
    "INSVRGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2200"]}},
    "INSVSVRRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2206"]}},
    "INTABESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2207"]}},
    "INTACTILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i2208"]}},
    "INTACTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["i2209"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["i2209"]},
    },
    "INTAMINATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i7210"]}},
    "INTECTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i3194"]}},
    "INTEGELLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2213"]}},
    "INTEGER": {"-": {"morpho": ["aps---mn1r"], "uri": ["i2214"]}},
    "INTEGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2215"]}},
    "INTEGRE": {"-": {"morpho": ["rp--------"], "uri": ["24335"]}},
    "INTEGRITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2218"]}},
    "INTEGRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2220"]}},
    "INTELLECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2225"]}},
    "INTELLECTVALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i2228"]}},
    "INTELLECTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["i9893"]}},
    "INTELLIGENS": {"-": {"morpho": ["aps---an3i"], "uri": ["114758"]}},
    "INTELLIGENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["52770"]}},
    "INTELLIGIBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["132151"]}},
    "INTELLIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["39311"]}},
    "INTEMERATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i7234"]}},
    "INTEMPERANS": {"-": {"morpho": ["aps---an3i"], "uri": ["i2235"]}},
    "INTEMPERANTER": {"-": {"morpho": ["rp--------"], "uri": ["58299"]}},
    "INTEMPERANTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i2236"]}},
    "INTEMPERATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2237"]}},
    "INTEMPERIAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["i2238"]}},
    "INTEMPERIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["i2238"]}},
    "INTEMPESTIVE": {"-": {"morpho": ["rp--------"], "uri": ["37692"]}},
    "INTEMPESTIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2240"]}},
    "INTEMPESTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2241"]}},
    "INTENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2245"]}},
    "INTENTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2249"]}},
    "INTENTATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["114783"]}},
    "INTENTE": {"-": {"morpho": ["rp--------"], "uri": ["24348"]}},
    "INTENTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2248"]}},
    "INTENTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2252"]}},
    "INTENTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["i2276"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["24351"]},
    },
    "INTEPEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["i2253"]}},
    "INTEPESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2253"]}},
    "INTER": {"-": {"morpho": ["p---------"], "uri": ["i2254"]}},
    "INTERARESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2264"]}},
    "INTERCALARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2269"]}},
    "INTERCALO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2273"]}},
    "INTERCAPEDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2275"]}},
    "INTERCARDINATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2277"]}},
    "INTERCEDO": {"1": {"morpho": ["v1spia--3-"], "uri": ["i2278"]}},
    "INTERCESSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2281"]}},
    "INTERCESSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i2282"]}},
    "INTERCESSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["i9895"]}},
    "INTERCIDO": {
        "1": {"morpho": ["v1spia--3-"], "uri": ["i2283", "i2284", "i2283"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["i2283", "i2284", "i2283"]},
    },
    "INTERCIPIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["i2287"]}},
    "INTERCLVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2292"]}},
    "INTERCOLVMNIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i2293"]}},
    "INTERCVRRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2294"]}},
    "INTERCVRSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2295"]}},
    "INTERCVRSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["i9896"]}},
    "INTERCVS": {"-": {"morpho": ["aps---an3-"], "uri": ["i2296"]}},
    "INTERDATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["99075"]}},
    "INTERDICO": {"2": {"morpho": ["v1spia--3-"], "uri": ["i2300"]}},
    "INTERDICTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i9928"]}},
    "INTERDIV": {"-": {"morpho": ["rp--------"], "uri": ["i2306"]}},
    "INTERDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2307"]}},
    "INTERDVM": {"-": {"morpho": ["rp--------"], "uri": ["i2310"]}},
    "INTEREA": {"-": {"morpho": ["rp--------"], "uri": ["i2311"]}},
    "INTEREMPTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i2314"]}},
    "INTEREO": {"1": {"morpho": ["v1spia--4-"], "uri": ["i2317"]}},
    "INTEREQVITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2318"]}},
    "INTERFECTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i2324"]}},
    "INTERFECTRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2326"]}},
    "INTERFICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["i2329"]}},
    "INTERFIO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2331"]}},
    "INTERFLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2333"]}},
    "INTERFODIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["i2335"]}},
    "INTERFOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["i2336"]}},
    "INTERFRINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2338"]}},
    "INTERFVGIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["i2339"]}},
    "INTERFVNDO": {"2": {"morpho": ["v1spia--3-"], "uri": ["i2341"]}},
    "INTERFVSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["33861"]}},
    "INTERIACEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["i2349"]}},
    "INTERIACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["46230"]}},
    "INTERICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["i2353"]}},
    "INTERIECTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["i2350"]}},
    "INTERIM": {"-": {"morpho": ["rp--------"], "uri": ["i2356"]}},
    "INTERIMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2357"]}},
    "INTERIOR": {
        "-": {"morpho": ["acs---cn3-"], "uri": ["i2358"]},
        "2": {"morpho": ["acs---cn3-"], "uri": ["i2358"]},
    },
    "INTERIORA": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "INTERITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2359"]}},
    "INTERITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["i9897"]}},
    "INTERIVNGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2360"]}},
    "INTERLABOR": {"2": {"morpho": ["v1spid--3-"], "uri": ["i2362"]}},
    "INTERLEGO": {"2": {"morpho": ["v1spia--3-"], "uri": ["i2366"]}},
    "INTERLIGO": {"2": {"morpho": ["v1spia--1-"], "uri": ["i2368"]}},
    "INTERLINO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2370"]}},
    "INTERLOQVOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["i2373"]}},
    "INTERLVCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["i2375"]}},
    "INTERLVNIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i2378"]}},
    "INTERLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2380"]}},
    "INTERMESTRIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i2387"]}},
    "INTERMINO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "INTERMISCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["i2393"]}},
    "INTERMISSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2394"]}},
    "INTERMITTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2395"]}},
    "INTERMORIOR": {"-": {"morpho": ["v1spid--3i"], "uri": ["i2397"]}},
    "INTERNASCOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["i2402"]}},
    "INTERNECIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2407"]}},
    "INTERNECIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2409"]}},
    "INTERNITEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["i2414"]}},
    "INTERNODIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i2416"]}},
    "INTERNOSCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2417"]}},
    "INTERNVNTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i2420"]}},
    "INTERNVNTIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["i2420"]}},
    "INTERNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2421"]}},
    "INTERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2422"]}},
    "INTERPELLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2428"]}},
    "INTERPELLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i2429"]}},
    "INTERPELLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2431"]}},
    "INTERPENSIVA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["i2432"]}},
    "INTERPOLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2440"]}},
    "INTERPONO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2441"]}},
    "INTERPOSITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2442"]}},
    "INTERPRES": {"-": {"morpho": ["n-s---cn3-"], "uri": ["i2443"]}},
    "INTERPRETATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2446"]}},
    "INTERPRETOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["i2452"]}},
    "INTERPRIMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2453"]}},
    "INTERPVNGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2455"]}},
    "INTERPVTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2457"]}},
    "INTERQVIESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2458"]}},
    "INTERRITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2466"]}},
    "INTERROGATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2471"]}},
    "INTERROGATIVNCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i2472"]}},
    "INTERROGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2476"]}},
    "INTERRVMPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2477"]}},
    "INTERSAEPIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["i2480"]}},
    "INTERSCALMIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i2482"]}},
    "INTERSCINDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2485"]}},
    "INTERSECO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2487"]}},
    "INTERSECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2488"]}},
    "INTERSERO": {
        "2": {"morpho": ["v1spia--3-"], "uri": ["i2491"]},
        "3": {"morpho": ["v1spia--3-"], "uri": ["i2491"]},
    },
    "INTERSPIRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2498"]}},
    "INTERSTINCTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["32217"]}},
    "INTERSTINGVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2502"]}},
    "INTERSVM": {"1": {"morpho": ["v1spia--1-"], "uri": ["i2510"]}},
    "INTERTIGNIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i2513"]}},
    "INTERTRIGO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2515"]}},
    "INTERTRIMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i2516"]}},
    "INTERVALLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i7525"]}},
    "INTERVECTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["115036"]}},
    "INTERVELLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2527"]}},
    "INTERVENIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["i2529"]}},
    "INTERVENIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i2530"]}},
    "INTERVENTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["i2528"]}},
    "INTERVERSOR": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["i2535"]},
        "2": {"morpho": ["v1spid--1-"], "uri": ["i2534"]},
        "-": {"morpho": ["v1spid--1-"], "uri": ["i2534"]},
    },
    "INTERVERTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2536"]}},
    "INTERVISO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2541"]}},
    "INTERVOMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2546"]}},
    "INTERVTRASQVE": {"-": {"morpho": ["rp--------"], "uri": ["i2522"]}},
    "INTESTABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i2548"]}},
    "INTESTATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2549"]}},
    "INTESTINVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i2551"]}},
    "INTESTINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2551"]}},
    "INTEXO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2552"]}},
    "INTEXTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["i2542"]}},
    "INTIME": {"-": {"morpho": ["rp--------"], "uri": ["24411"]}},
    "INTIMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["57246"]}},
    "INTINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["47154"]}},
    "INTOLERABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i2563"]}},
    "INTOLERANDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i3200"]}},
    "INTOLERANS": {"-": {"morpho": ["aps---an3i"], "uri": ["i3201"]}},
    "INTOLERANTER": {"-": {"morpho": ["rp--------"], "uri": ["51331"]}},
    "INTOLERANTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i9846"]}},
    "INTONO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2568"]}},
    "INTONSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i3199"]}},
    "INTORQVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["i2569"]}},
    "INTRA": {
        "1": {"morpho": ["rp--------"], "uri": ["31808"]},
        "2": {"morpho": ["p---------"], "uri": ["i2571"]},
    },
    "INTRACTABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i2574"]}},
    "INTRACTATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2575"]}},
    "INTREMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2585"]}},
    "INTREPIDE": {"-": {"morpho": ["rp--------"], "uri": ["56211"]}},
    "INTREPIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2588"]}},
    "INTRINSECVS": {"-": {"morpho": ["rp--------"], "uri": ["i2593"]}},
    "INTRITVS": {
        "1": {"morpho": ["aps---mn1-"], "uri": ["i3203"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["i3203"]},
    },
    "INTRO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["i2596"]},
        "2": {"morpho": ["rp--------"], "uri": ["i2595"]},
    },
    "INTROCEDO": {"1": {"morpho": ["v1spia--3-"], "uri": ["i2597"]}},
    "INTRODVCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i9922"]}},
    "INTRODVCTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2601"]}},
    "INTROEO": {"1": {"morpho": ["v1spia--4-"], "uri": ["i9924"]}},
    "INTROFERO": {
        "-": {
            "morpho": ["v1spia--3-", "v1spia--3-"],
            "uri": ["i9923", "i9923"],
        }
    },
    "INTROGREDIOR": {"-": {"morpho": ["v1spid--3i"], "uri": ["i2606"]}},
    "INTROITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["i2599"]}},
    "INTROMITTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i9925"]}},
    "INTRORSVM": {"-": {"morpho": ["rp--------"], "uri": ["i2611"]}},
    "INTRORVMPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2612"]}},
    "INTROSPECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i9926"]}},
    "INTROSPICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["i2614"]}},
    "INTROTRVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2615"]}},
    "INTROVERSVS": {"-": {"morpho": ["rp--------"], "uri": []}},
    "INTVBVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i2619"]}},
    "INTVBVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["44697"]}},
    "INTVEOR": {"-": {"morpho": ["v1spid--2-"], "uri": []}},
    "INTVMESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2622"]}},
    "INTVMVLATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2623"]}},
    "INTVOR": {"-": {"morpho": ["v1spid--3-"], "uri": []}},
    "INTVRBIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2626"]}},
    "INTVRBO": {"2": {"morpho": ["v1spia--1-"], "uri": []}},
    "INTVS": {
        "1": {"morpho": ["rp--------"], "uri": ["i2628"]},
        "2": {"morpho": ["p---------"], "uri": []},
    },
    "INTVTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2629"]}},
    "INVADO": {"2": {"morpho": ["v1spia--3-"], "uri": ["i2630"]}},
    "INVALESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i9930"]}},
    "INVALIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2632"]}},
    "INVASIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2634"]}},
    "INVECTICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2639"]}},
    "INVECTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i2643"]}},
    "INVECTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["i9903"]}},
    "INVEHO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2644"]}},
    "INVENIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["i2649"]}},
    "INVENTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2651"]}},
    "INVENTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i2653"]}},
    "INVENTRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2654"]}},
    "INVENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i9904"]}},
    "INVENTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["i9904"]}},
    "INVENVSTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2655"]}},
    "INVERECVNDE": {"-": {"morpho": ["rp--------"], "uri": ["99117"]}},
    "INVERECVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2658"]}},
    "INVERGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2659"]}},
    "INVERSVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i2677"]}},
    "INVERSVS": {"1": {"morpho": ["aps---mn1-"], "uri": ["i3207"]}},
    "INVERTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2665"]}},
    "INVESPERASCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2666"]}},
    "INVESTIGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2671"]}},
    "INVESTIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["i2672"]}},
    "INVETERASCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2674"]}},
    "INVETERATVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "INVETERATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["38497"]}},
    "INVIA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["99122"]}},
    "INVICEM": {"-": {"morpho": ["rp--------"], "uri": ["i7526"]}},
    "INVICTVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["i2682"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["i2682"]},
    },
    "INVIDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["i2683"]}},
    "INVIDIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i2684"]}},
    "INVIDIOSE": {"-": {"morpho": ["rp--------"], "uri": ["43988"]}},
    "INVIDIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2685"]}},
    "INVIDVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["99123"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["i2686"]},
    },
    "INVIGILO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2687"]}},
    "INVINCTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2691"]}},
    "INVIOLABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i2694"]}},
    "INVIOLATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i7695"]}},
    "INVISITATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i7701"]}},
    "INVISO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2702"]}},
    "INVISVS": {
        "1": {"morpho": ["aps---mn1-"], "uri": ["i3208"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["i3208"]},
    },
    "INVITAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i2706"]}},
    "INVITATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2707"]}},
    "INVITE": {"-": {"morpho": ["rp--------"], "uri": ["53174"]}},
    "INVITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2713"]}},
    "INVITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2715"]}},
    "INVIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2716"]}},
    "INVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i2717"]}},
    "INVLTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["24437"]}},
    "INVMBRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2723"]}},
    "INVNDATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2727"]}},
    "INVNDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2728"]}},
    "INVNGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["115237"]}},
    "INVOCATVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["115243"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["i9915"]},
    },
    "INVOCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2733"]}},
    "INVOLITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2734"]}},
    "INVOLO": {"2": {"morpho": ["v1spia--1-"], "uri": ["i2735"]}},
    "INVOLVCRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i0274"]}},
    "INVOLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2741"]}},
    "INVOLVTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2742"]}},
    "INVOLVTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["37187"]}},
    "INVRBANE": {"-": {"morpho": ["rp--------"], "uri": ["29569"]}},
    "INVRGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["i2746"]}},
    "INVRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2748"]}},
    "INVSITATE": {"-": {"morpho": ["rp--------"], "uri": ["24442"]}},
    "INVSITATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2749"]}},
    "INVTILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i2752"]}},
    "INVTILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2753"]}},
    "INVTILITER": {"-": {"morpho": ["rp--------"], "uri": ["24445"]}},
    "INVVLNERABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i2755"]}},
    "INVVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "IO": {
        "-": {"morpho": ["----------"], "uri": []},
        "N": {"morpho": ["n-s---fn3-"], "uri": []},
    },
    "IOCATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2761"]}},
    "IOCOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["i2763"]}},
    "IOCOSE": {"-": {"morpho": ["rp--------"], "uri": ["34276"]}},
    "IOCOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2764"]}},
    "IOCVLARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i2765"]}},
    "IOCVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["i2772"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["i2772"]},
    },
    "IONIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "IONICE": {"-": {"morpho": ["rp--------"], "uri": ["101301"]}},
    "IPSE": {"-": {"morpho": ["aps---mn1-"], "uri": ["32864"]}},
    "IRA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["i2784"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "IRACVNDIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i2785"]}},
    "IRACVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2786"]}},
    "IRASCOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["i2788"]}},
    "IRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i3211"]}},
    "IRIS": {
        "-": {"morpho": ["n-s---fn3i"], "uri": []},
        "N": {"morpho": ["n-s---fn3-"], "uri": []},
    },
    "IRONIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i2798"]}},
    "IRPEX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i2800"]}},
    "IRRADO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2803"]}},
    "IRRATIONALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i2805"]}},
    "IRRELIGATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2826"]}},
    "IRRELIGIOSE": {"-": {"morpho": ["rp--------"], "uri": ["46638"]}},
    "IRREMEABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i2830"]}},
    "IRREMEDIABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i2831"]}},
    "IRREPARABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i2837"]}},
    "IRREPERTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2839"]}},
    "IRREPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2841"]}},
    "IRREQVIETVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2850"]}},
    "IRRESECTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2853"]}},
    "IRRETIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["i2858"]}},
    "IRRETORTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2859"]}},
    "IRREVERENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i2863"]}},
    "IRREVOCABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i2865"]}},
    "IRREVOCATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2867"]}},
    "IRRIDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["i2868"]}},
    "IRRIDICVLE": {"-": {"morpho": ["rp--------"], "uri": ["34054"]}},
    "IRRIGIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2872"]}},
    "IRRIGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2873"]}},
    "IRRIGVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2874"]}},
    "IRRISIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2876"]}},
    "IRRISOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i2878"]}},
    "IRRISVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["i9908"]}},
    "IRRITABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i2880"]}},
    "IRRITAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i2883"]}},
    "IRRITATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2884"]}},
    "IRRITATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i2885"]}},
    "IRRITATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["99148"]}},
    "IRRITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2887", "i2797"]}},
    "IRRITVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i2888"]}},
    "IRRITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2888"]}},
    "IRROGATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2890"]}},
    "IRROGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2891"]}},
    "IRRORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2893"]}},
    "IRRVMATIO": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "IRRVMATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i2904"]}},
    "IRRVMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2905"]}},
    "IRRVMPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2906"]}},
    "IRRVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2907"]}},
    "IRRVPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2908"]}},
    "IRRVPTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i3213"]}},
    "IS": {"-": {"morpho": ["aps---mn1-"], "uri": ["35656"]}},
    "ISATIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2916"]}},
    "ISCHIACVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2920"]}},
    "ISCHNOS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "ISIDORVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ISMENIAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ISODOMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2935"]}},
    "ISTAC": {"-": {"morpho": ["rp--------"], "uri": ["50662"]}},
    "ISTE": {"-": {"morpho": ["aps---mn1-"], "uri": []}},
    "ISTHMVS": {
        "-": {"morpho": ["n-s----n-g"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "ISTI": {"-": {"morpho": ["rp--------"], "uri": ["56546"]}},
    "ISTIC": {
        "1": {"morpho": ["aps---mn1-"], "uri": []},
        "2": {"morpho": ["rp--------"], "uri": ["48378"]},
    },
    "ISTINC": {"-": {"morpho": ["rp--------"], "uri": ["i2947"]}},
    "ISTIVSMODI": {"-": {"morpho": ["rp--------"], "uri": ["i2953"]}},
    "ISTO": {"-": {"morpho": ["rp--------"], "uri": ["35320"]}},
    "ISTOC": {"-": {"morpho": ["rp--------"], "uri": ["37189"]}},
    "ISTVC": {"-": {"morpho": ["rp--------"], "uri": ["45294"]}},
    "ITA": {"-": {"morpho": ["rp--------"], "uri": ["i2955"]}},
    "ITALIA": {"N": {"morpho": ["n-s---fn1-"], "uri": ["i3217"]}},
    "ITALICVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": ["47158"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["47158"]},
    },
    "ITALVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": ["45424"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["i2957"]},
    },
    "ITAQVE": {"-": {"morpho": ["rp--------"], "uri": ["i2958"]}},
    "ITEM": {"-": {"morpho": ["rp--------"], "uri": ["i2960"]}},
    "ITER": {"-": {"morpho": ["n-s---nn3-"], "uri": ["i2961", "i2961"]}},
    "ITERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2964"]}},
    "ITERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2966"]}},
    "ITERVM": {"-": {"morpho": ["rp--------"], "uri": ["i2967"]}},
    "ITIDEM": {"-": {"morpho": ["rp--------"], "uri": ["i2969"]}},
    "ITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["i9910"]}},
    "IVBA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["i2977"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["i2977"]},
    },
    "IVBAR": {"-": {"morpho": ["n-s---nn3-"], "uri": ["i2978"]}},
    "IVBATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2979"]}},
    "IVBEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["i2980"]}},
    "IVCVNDE": {"-": {"morpho": ["rp--------"], "uri": ["51244"]}},
    "IVCVNDITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2986"]}},
    "IVCVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2989"]}},
    "IVDEX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i2990"]}},
    "IVDICATVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i9912"]}},
    "IVDICATVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["i9912"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["115462"]},
    },
    "IVDICIALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i2995"]}},
    "IVDICIARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2996"]}},
    "IVDICIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i2998"]}},
    "IVDICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2999"]}},
    "IVGALES": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "IVGALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i3001"]}},
    "IVGERVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i3009"]}},
    "IVGIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i3011"]}},
    "IVGLANS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["i3014"]}},
    "IVGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i3015"]}},
    "IVGOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i3017"]}},
    "IVGVLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i3019"]}},
    "IVGVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i3023"]}},
    "IVGVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i3024"]}},
    "IVGVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["48396"]}},
    "IVGVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i3025"]}},
    "IVGVMENTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i3026"]}},
    "IVGVMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i3027"]}},
    "IVGVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i3028"]}},
    "IVLIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "IVLIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": ["i3221"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["43324"]},
    },
    "IVLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["i3030"]}},
    "IVMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i3033"]}},
    "IVNCTVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i3053"]}},
    "IVNCTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["54817"]}},
    "IVNCVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["i3043"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["i3043"]},
    },
    "IVNGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i3044"]}},
    "IVNIANVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": ["99173"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["99173"]},
    },
    "IVNIO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "IVNIOR": {
        "-": {"morpho": ["n-s----n3-"], "uri": []},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "IVNIPERVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["37714"]}},
    "IVNIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": ["i3222"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["32855"]},
    },
    "IVNIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i3103"]}},
    "IVNO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "IVPPITER": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "IVRAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i3048"]}},
    "IVRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["38876"]}},
    "IVRE": {"-": {"morpho": ["rp--------"], "uri": ["30841"]}},
    "IVRGIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i3059"]}},
    "IVRGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i3060"]}},
    "IVRIDICVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["i3064"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["i3064"]},
    },
    "IVRISCONSVLTVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["i6050"]}},
    "IVRISDICTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i6057"]}},
    "IVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i3067"]}},
    "IVROR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132783"]}},
    "IVS": {
        "1": {"morpho": ["n-s---nn3-"], "uri": ["i3100"]},
        "2": {"morpho": ["n-s---nn3-"], "uri": ["i3100"]},
    },
    "IVSCVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i3075"]}},
    "IVSIVRANDVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["99180"]}},
    "IVSSVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i9913"]}},
    "IVSSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["i9913"]}},
    "IVSTA": {
        "C": {"morpho": ["n-s---fn1-"], "uri": ["101313"]},
        "-": {"morpho": ["n-s---fn1-"], "uri": ["101313"]},
    },
    "IVSTE": {"-": {"morpho": ["rp--------"], "uri": ["56182"]}},
    "IVSTIFICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i3082"]}},
    "IVSTITIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["i3083"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "IVSTITIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i3085"]}},
    "IVSTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i3087"]}},
    "IVSTVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["i3087"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["99182"]},
    },
    "IVTO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "IVVAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["i3091"]}},
    "IVVENALIS": {
        "-": {"morpho": ["aps---cn3i"], "uri": ["i3093"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
        "A": {"morpho": ["aps---fn3i"], "uri": []},
    },
    "IVVENALITER": {"-": {"morpho": ["rp--------"], "uri": ["47541"]}},
    "IVVENCA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i3094"]}},
    "IVVENCVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["i3094"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["i3094"]},
    },
    "IVVENESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i3097"]}},
    "IVVENILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i3098"]}},
    "IVVENILITER": {"-": {"morpho": ["rp--------"], "uri": ["38921"]}},
    "IVVENIS": {
        "1": {"morpho": ["n-s---cn3-"], "uri": ["47290"]},
        "2": {"morpho": ["aps---cn3i"], "uri": ["i3102"]},
    },
    "IVVENTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i0271"]}},
    "IVVENTAS": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["i3106"]},
        "N": {"morpho": ["n-s---fn3-"], "uri": ["i3106"]},
    },
    "IVVENTVS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i3108"]}},
    "IVVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i3110"]}},
    "IVXTA": {
        "1": {"morpho": ["rp--------"], "uri": ["54775"]},
        "2": {"morpho": ["p---------"], "uri": ["i3111"]},
    },
    "IVXTIM": {"-": {"morpho": ["rp--------"], "uri": ["i3112"]}},
    "KALENDAE": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "KALENDARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["k0003"]}},
    "LABASCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["l0024"]}},
    "LABASCOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["132996"]}},
    "LABEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["l0007"]}},
    "LABEFACTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l0009"]}},
    "LABEFIO": {"-": {"morpho": ["v1spia--3-"], "uri": ["l0010"]}},
    "LABELLVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["l0015", "l0011"]},
        "2": {"morpho": ["n-s---nn2-"], "uri": ["l0015", "l0011"]},
    },
    "LABEOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0013"]}},
    "LABES": {"-": {"morpho": ["n-s---fn3i"], "uri": ["l0014"]}},
    "LABIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0141"]}},
    "LABICI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "LABIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0017"]}},
    "LABO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l0044"]}},
    "LABOR": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["24521"]},
        "2": {"morpho": ["v1spid--3-"], "uri": ["l0025"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "LABORIOSE": {"-": {"morpho": ["rp--------"], "uri": ["24523"]}},
    "LABORIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0029"]}},
    "LABORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l0030"]}},
    "LABRVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["l0020", "l0034"]},
        "2": {"morpho": ["n-s---nn2-"], "uri": ["l0020", "l0034"]},
    },
    "LABRVSCA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0035"]}},
    "LABYRINTHEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0037"]}},
    "LABYRINTHVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["24526"]}},
    "LAC": {"-": {"morpho": ["n-s---nn3-"], "uri": ["l0040"]}},
    "LACEDAEMON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "LACEDAEMONII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "LACEDAEMONIVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "LACER": {"-": {"morpho": ["aps---mn1r"], "uri": ["l0046"]}},
    "LACERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0048"]}},
    "LACERNA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0051"]}},
    "LACERNATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0052"]}},
    "LACERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l0054"]}},
    "LACERTA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["l0055"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "LACERTOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0056"]}},
    "LACERTVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["l0057"]},
        "2": {"morpho": ["n-s---mn2-"], "uri": ["l0057"]},
    },
    "LACESSO": {"-": {"morpho": ["v1spia--3-"], "uri": ["l0061"]}},
    "LACRIMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0069"]}},
    "LACRIMABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["l0070"]}},
    "LACRIMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l0074"]}},
    "LACRIMOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0075"]}},
    "LACRIMVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0076"]}},
    "LACTEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["l0080"]}},
    "LACTEOLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0081"]}},
    "LACTES": {"-": {"morpho": ["n-p---fn3i"], "uri": ["l0082"]}},
    "LACTEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0084"]}},
    "LACTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l0502"]}},
    "LACTVCA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0094"]}},
    "LACVNA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0101"]}},
    "LACVNAR": {"-": {"morpho": ["n-s---nn3i"], "uri": ["l0102"]}},
    "LACVNOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0104"]}},
    "LACVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["l0668"]}},
    "LAEDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["l0111"]}},
    "LAELIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LAENA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0114"]}},
    "LAESIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0113"]}},
    "LAETABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["l0116"]}},
    "LAETANS": {"-": {"morpho": ["aps---an3i"], "uri": ["99199"]}},
    "LAETE": {"-": {"morpho": ["rp--------"], "uri": ["24541"]}},
    "LAETIFICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l0122"]}},
    "LAETIFICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0123"]}},
    "LAETILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LAETITIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["l0126"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "LAETOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132542"]}},
    "LAETORIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LAETVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["l0130"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["l0130"]},
    },
    "LAEVA": {
        "1": {"morpho": ["n-s---fn1-"], "uri": ["l0133"]},
        "2": {"morpho": ["n-p---nn2-"], "uri": ["99201"]},
        "3": {"morpho": ["rp--------"], "uri": ["99202"]},
    },
    "LAEVE": {"-": {"morpho": ["rp--------"], "uri": ["50442"]}},
    "LAEVINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LAEVVM": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["rp--------"], "uri": ["115687"]},
        "-": {"morpho": ["rp--------"], "uri": ["115687"]},
    },
    "LAGANVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0135"]}},
    "LAGENA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["58291"]}},
    "LAGEVS": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "LAGOIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0140"]}},
    "LALLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l0151"]}},
    "LAMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0153"]}},
    "LAMBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["l0157"]}},
    "LAMELLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0158"]}},
    "LAMENTA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["99208"]}},
    "LAMENTABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["l0161"]}},
    "LAMENTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0163"]}},
    "LAMENTOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132543"]}},
    "LAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0160"]}},
    "LAMIA": {"N": {"morpho": ["n-s---fn1-"], "uri": ["l0168"]}},
    "LAMINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0169"]}},
    "LAMPAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0181"]}},
    "LAMPSACENI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "LAMPSACENVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "LAMPSACVS": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "LANA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0185"]}},
    "LANATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0188"]}},
    "LANCEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0189"]}},
    "LANCINATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0194"]}},
    "LANCINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l0196"]}},
    "LANCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0198"]}},
    "LANEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0201"]}},
    "LANGOR": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "LANGVENS": {"-": {"morpho": ["aps---an3i"], "uri": ["99210"]}},
    "LANGVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["l0206"]}},
    "LANGVESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["l9917"]}},
    "LANGVIDE": {"-": {"morpho": ["rp--------"], "uri": ["24559"]}},
    "LANGVIDVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0207"]}},
    "LANGVIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0208"]}},
    "LANGVOR": {"-": {"morpho": ["n-s---mn3-"], "uri": []}},
    "LANIATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["l9890"]}},
    "LANIFICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0222"]}},
    "LANIGER": {
        "1": {"morpho": ["n-s---mn2r"], "uri": ["l0226"]},
        "2": {"morpho": ["aps---mn1r"], "uri": ["l0226"]},
    },
    "LANIGERA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0226"]}},
    "LANIO": {
        "-": {"morpho": ["v1spia--1-"], "uri": ["l0228"]},
        "2": {"morpho": ["n-s---mn3-"], "uri": ["l0229"]},
        "1": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "LANISTA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["l0235"]}},
    "LANITIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["37832"]}},
    "LANVGO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0248"]}},
    "LANX": {"-": {"morpho": ["n-s---fn3i"], "uri": ["l0250"]}},
    "LAODICIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LAPATHVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0252"]}},
    "LAPATHVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["36480"]}},
    "LAPHIRO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "LAPICIDINAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["l0254"]}},
    "LAPIDATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0257"]}},
    "LAPIDEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0261"]}},
    "LAPIDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l0262"]}},
    "LAPIDOSA": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "LAPIDOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": []}},
    "LAPILLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l0268"]}},
    "LAPIS": {"-": {"morpho": ["n-s---fn3-"], "uri": []}},
    "LAPPA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["l0272"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "LAPSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l0280"]}},
    "LAPSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["l0003"]}},
    "LAQVEAR": {"-": {"morpho": ["n-s---nn3i"], "uri": ["99220"]}},
    "LAQVEO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["l0284"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["l0284"]},
    },
    "LAQVEVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l0285"]}},
    "LAR": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "LARCIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LARDVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["46839"]}},
    "LARENTIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "LARGE": {"-": {"morpho": ["rp--------"], "uri": ["52944"]}},
    "LARGIFICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0290"]}},
    "LARGIFLVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0291"]}},
    "LARGIOR": {"-": {"morpho": ["v1spid--4-"], "uri": []}},
    "LARGITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0295"]}},
    "LARGITER": {"-": {"morpho": ["rp--------"], "uri": ["24576"]}},
    "LARGITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0296"]}},
    "LARGITOR": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["l0298"]},
        "2": {"morpho": ["v1spid--1-"], "uri": []},
        "-": {"morpho": ["v1spid--1-"], "uri": []},
    },
    "LARGITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0299"]}},
    "LARGIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LARGO": {"-": {"morpho": ["rp--------"], "uri": []}},
    "LARGVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["l0302"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "LARIDVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0303"]}},
    "LARIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["24579"]}},
    "LARVA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["l0309"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "LARVALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["l0310"]}},
    "LASANVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0316"]}},
    "LASCIVE": {"-": {"morpho": ["rp--------"], "uri": ["99223"]}},
    "LASCIVIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0318"]}},
    "LASCIVIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["l0320"]}},
    "LASCIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0325"]}},
    "LASER": {"-": {"morpho": ["n-s---nn3-"], "uri": ["l0326"]}},
    "LASERPICIATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0329"]}},
    "LASERPICIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0330"]}},
    "LASERPITIFER": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "LASSITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0332"]}},
    "LASSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l0333"]}},
    "LASSVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0334"]}},
    "LASSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0335"]}},
    "LATE": {"-": {"morpho": ["rp--------"], "uri": ["53076"]}},
    "LATEBRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0339"]}},
    "LATEBROSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0341"]}},
    "LATENS": {"-": {"morpho": ["aps---an3i"], "uri": ["99227"]}},
    "LATEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["l1298"]}},
    "LATER": {"-": {"morpho": ["n-s---mn3-"], "uri": ["l0343"]}},
    "LATERAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["l0345"]}},
    "LATERARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0353"]}},
    "LATERCVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l0348"]}},
    "LATERICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0350"]}},
    "LATERNA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["42784"]}},
    "LATEX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["l0962"]}},
    "LATIBVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0357"]}},
    "LATIFVNDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0363"]}},
    "LATINE": {"-": {"morpho": ["rp--------"], "uri": []}},
    "LATINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "LATINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LATINVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "LATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0367"]}},
    "LATITO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["l0402", "l0373"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["l0402", "l0373"]},
    },
    "LATITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0374"]}},
    "LATIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "LATONA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["l0376"]}},
    "LATRATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["l9891"]}},
    "LATRO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["l0384"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["l0383"]},
    },
    "LATROCINIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0387"]}},
    "LATROCINOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["l0388"]}},
    "LATRVNCVLARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l9387"]}},
    "LATRVNCVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l0389"]}},
    "LATVS": {
        "1": {"morpho": ["n-s---nn3-"], "uri": ["l0393"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["l0392"]},
    },
    "LATVSCVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0394"]}},
    "LAVABRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["53124"]}},
    "LAVACRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0396"]}},
    "LAVATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0399"]}},
    "LAVDABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["l0403"]}},
    "LAVDABILITER": {"-": {"morpho": ["rp--------"], "uri": ["47794"]}},
    "LAVDATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0406"]}},
    "LAVDATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["l0408"]}},
    "LAVDATRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0410"]}},
    "LAVDATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["36485"]}},
    "LAVDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l0411"]}},
    "LAVINIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "LAVINIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "LAVO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["l0416"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["l0416"]},
    },
    "LAVREA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0223"]}},
    "LAVREATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0419"]}},
    "LAVRENS": {
        "N": {"morpho": ["aps---an3-"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "LAVRENTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "LAVRENTINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "LAVREVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0218"]}},
    "LAVRICOMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0424"]}},
    "LAVRIGER": {"-": {"morpho": ["aps---mn1r"], "uri": ["l0427"]}},
    "LAVRVS": {"-": {"morpho": ["n-s---fn4-"], "uri": ["l0431"]}},
    "LAVS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0432"]}},
    "LAVTE": {"-": {"morpho": ["rp--------"], "uri": ["24605"]}},
    "LAVTITIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0437"]}},
    "LAVTVMIAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["l0439"]}},
    "LAVTVS": {
        "C": {"morpho": ["aps---mn1-"], "uri": ["24607"]},
        "-": {"morpho": ["aps---mn1-"], "uri": ["24607"]},
    },
    "LAXAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0442"]}},
    "LAXATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0443"]}},
    "LAXATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["24612"]}},
    "LAXE": {"-": {"morpho": ["rp--------"], "uri": ["24613"]}},
    "LAXITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0445"]}},
    "LAXO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l0446"]}},
    "LAXVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0447"]}},
    "LEA": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "LEAENA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["l0449"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["l0449"]},
    },
    "LEBES": {"-": {"morpho": ["n-s---mn3-"], "uri": ["l0450"]}},
    "LECTICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0452"]}},
    "LECTICARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l0455"]}},
    "LECTICVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l0422"]}},
    "LECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0457"]}},
    "LECTISTERNIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0460"]}},
    "LECTITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l0461"]}},
    "LECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l0463"]}},
    "LECTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["l0464"]}},
    "LECTVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l0469"]}},
    "LECTVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["l0471"]},
        "2": {"morpho": ["n-s---mn4-"], "uri": ["l0488"]},
        "3": {"morpho": ["aps---mn1-"], "uri": ["24624"]},
    },
    "LEGALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["l0474"]}},
    "LEGATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0478"]}},
    "LEGATVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0497"]}},
    "LEGATVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l0497"]}},
    "LEGENS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["99240"]}},
    "LEGIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["l0485"]}},
    "LEGIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0486"]}},
    "LEGIONARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0487"]}},
    "LEGITIME": {"-": {"morpho": ["rp--------"], "uri": ["53830"]}},
    "LEGITIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0492"]}},
    "LEGO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["l0495"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["l0496"]},
    },
    "LEGVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l0499"]}},
    "LEGVMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["l0500"]}},
    "LEMBVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l0507"]}},
    "LEMVRES": {"-": {"morpho": ["n-s---mn3-"], "uri": ["l0511"]}},
    "LENA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0512"]}},
    "LENE": {
        "1": {"morpho": ["n-s----n3-"], "uri": []},
        "2": {"morpho": ["rp--------"], "uri": ["33618"]},
    },
    "LENIMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["l0514"]}},
    "LENIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["l0516"]}},
    "LENIS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["116022"]},
        "2": {"morpho": ["aps---cn3i"], "uri": ["l0517"]},
    },
    "LENITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0518"]}},
    "LENITER": {"-": {"morpho": ["rp--------"], "uri": ["24641"]}},
    "LENITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0519"]}},
    "LENO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["l0520"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["l0521"]},
    },
    "LENOCINIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0524"]}},
    "LENOCINOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["l0525"]}},
    "LENS": {
        "1": {"morpho": ["n-s---fn3i"], "uri": ["l0528", "l0527", "l0527"]},
        "2": {"morpho": ["n-s---fn3i"], "uri": ["l0528", "l0527", "l0527"]},
    },
    "LENTE": {"-": {"morpho": ["rp--------"], "uri": ["24646"]}},
    "LENTESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["l0529"]}},
    "LENTISCVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["l0538"]}},
    "LENTITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0540"]}},
    "LENTO": {
        "-": {"morpho": ["v1spia--1-"], "uri": ["l0541"]},
        "N": {"morpho": ["n-s----n1-"], "uri": []},
    },
    "LENTVLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LENTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0545"]}},
    "LENVNCVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l9919"]}},
    "LEO": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["l0549"]},
        "N": {"morpho": ["n-s----n-g"], "uri": []},
    },
    "LEONIDAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "LEONINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0551"]}},
    "LEONTINI": {
        "1N": {"morpho": ["n-s----n2-"], "uri": []},
        "2N": {"morpho": ["n-s----n2-"], "uri": []},
        "N": {"morpho": ["n-p---mn2-"], "uri": []},
    },
    "LEONTINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "LEPIDE": {"-": {"morpho": ["rp--------"], "uri": ["24656"]}},
    "LEPIDVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["l0565"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "LEPOR": {"-": {"morpho": ["n-s---nn3-"], "uri": ["99245"]}},
    "LEPTIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "LEPVS": {"-": {"morpho": ["n-s---cn3-"], "uri": ["l0581"]}},
    "LEPVSCVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l0582"]}},
    "LETALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["l0587"]}},
    "LETALITER": {"-": {"morpho": ["rp--------"], "uri": ["116083"]}},
    "LETHARGICVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l0589"]}},
    "LETHARGVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l0590"]}},
    "LETIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["l0591"]}},
    "LETIFICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0592"]}},
    "LETO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l0593"]}},
    "LETVM": {
        "-": {"morpho": ["n-s---nn2-"], "uri": ["l0594"]},
        "N": {"morpho": ["n-s---nn2-"], "uri": ["l0594"]},
    },
    "LEVAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["l0596"]}},
    "LEVAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0598", "l9905"]}},
    "LEVATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0599"]}},
    "LEVATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["116094"]}},
    "LEVCOPHAEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0620"]}},
    "LEVIGATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l9920"]}},
    "LEVIGO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["l0638", "l9893"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["l0638", "l9893"]},
    },
    "LEVIS": {
        "1": {"morpho": ["aps---cn3i"], "uri": ["l0641", "l0625"]},
        "2": {"morpho": ["aps---cn3i"], "uri": ["l0641", "l0625"]},
    },
    "LEVISOMNVS": {
        "2": {"morpho": ["aps---mn1-"], "uri": ["l0642"]},
        "-": {"morpho": ["aps---mn1-"], "uri": ["l0642"]},
    },
    "LEVITAS": {
        "1": {"morpho": ["n-s---fn3-"], "uri": ["l0643"]},
        "2": {"morpho": ["n-s---fn3-"], "uri": ["l0643"]},
    },
    "LEVITER": {"-": {"morpho": ["rp--------"], "uri": ["24673"]}},
    "LEVO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["l9894", "l0646"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["l9894", "l0646"]},
    },
    "LEVOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["l0647"]}},
    "LEX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0648"]}},
    "LIACVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0653"]}},
    "LIBAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["l0656"]}},
    "LIBAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0657"]}},
    "LIBARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l0662"]}},
    "LIBELLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0666"]}},
    "LIBELLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l0673"]}},
    "LIBENS": {"-": {"morpho": ["aps---an3i"], "uri": ["99252"]}},
    "LIBENTER": {"-": {"morpho": ["rp--------"], "uri": ["24683"]}},
    "LIBER": {
        "1": {"morpho": ["n-s---mn2r"], "uri": ["l0676"]},
        "2": {"morpho": ["aps---mn1r"], "uri": ["l0675"]},
        "N": {"morpho": ["n-s----n2-"], "uri": []},
    },
    "LIBERA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LIBERALIS": {
        "-": {"morpho": ["aps---cn3i"], "uri": ["l0677"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "LIBERALITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0678"]}},
    "LIBERALITER": {"-": {"morpho": ["rp--------"], "uri": ["44799"]}},
    "LIBERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0680"]}},
    "LIBERATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["l0681"]}},
    "LIBERE": {"-": {"morpho": ["rp--------"], "uri": ["37465"]}},
    "LIBERI": {"-": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LIBERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l0684"]}},
    "LIBERTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0238"]}},
    "LIBERTAS": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["l0686"]},
        "N": {"morpho": ["n-s---fn3-"], "uri": ["l0686"]},
    },
    "LIBERTINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0689"]}},
    "LIBERTINVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["l0689"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["l0689"]},
    },
    "LIBERTVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l0251"]}},
    "LIBET": {"-": {"morpho": ["v3spia--2-"], "uri": ["l0691", "57313"]}},
    "LIBIDINOSE": {"-": {"morpho": ["rp--------"], "uri": ["49328"]}},
    "LIBIDINOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0694"]}},
    "LIBIDO": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["l0695"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "LIBITA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["31927"]}},
    "LIBITINARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l0696"]}},
    "LIBO": {
        "-": {"morpho": ["v1spia--1-"], "uri": ["l0691", "l0697"]},
        "N": {"morpho": ["n-s---mn3-"], "uri": []},
    },
    "LIBRA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["l0699"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "LIBRAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0702"]}},
    "LIBRARIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l9881", "l0704"]}},
    "LIBRARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0704"]}},
    "LIBRARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["l0704"]},
        "2": {"morpho": ["n-s---mn2-"], "uri": ["l0704"]},
        "3": {"morpho": ["aps---mn1-"], "uri": ["l9881", "l0704"]},
        "4": {"morpho": ["aps---mn1-"], "uri": ["l9881", "l0704"]},
    },
    "LIBRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0705"]}},
    "LIBRATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["l0706"]}},
    "LIBRILE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["99259"]}},
    "LIBRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l0711"]}},
    "LIBVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0712"]}},
    "LIBVRNA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l9906"]}},
    "LIBVRNICA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["l9908"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "LICENS": {"-": {"morpho": ["aps---an3i"], "uri": ["99262"]}},
    "LICENTER": {"-": {"morpho": ["rp--------"], "uri": ["48176"]}},
    "LICENTIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["l9883"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "LICEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["l0715", "58090"]}},
    "LICEOR": {"-": {"morpho": ["v1spid--2-"], "uri": ["l0715", "132547"]}},
    "LICET": {
        "1": {"morpho": ["v1spia--2-"], "uri": ["58055"]},
        "2": {"morpho": ["rp--------"], "uri": ["99263"]},
        "3": {"morpho": ["----------"], "uri": []},
    },
    "LICHANOS": {"-": {"morpho": ["n-s---mn2g"], "uri": ["l0716"]}},
    "LICINIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "LICIO": {"-": {"morpho": ["v1spia--3i"], "uri": []}},
    "LICITATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0725"]}},
    "LICITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["35994"]}},
    "LICIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0728"]}},
    "LICTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["l0729"]}},
    "LIDO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "LIEN": {"-": {"morpho": ["n-s---mn3-"], "uri": ["l0732"]}},
    "LIENOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0734"]}},
    "LIGAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["l0737"]}},
    "LIGAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0738"]}},
    "LIGELLVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "LIGNATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0743"]}},
    "LIGNATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["l0744"]}},
    "LIGNEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0746"]}},
    "LIGNOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["l0748"]}},
    "LIGNVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0750"]}},
    "LIGO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["l0754"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["l0753"]},
        "3": {"morpho": ["v1spia--1-"], "uri": ["l0753"]},
    },
    "LIGVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0755"]}},
    "LIGVR": {
        "C": {"morpho": ["n-s----n3-"], "uri": []},
        "N": {"morpho": ["aps---an3-"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "LIGVRES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "LIGVRRIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["l0756"]}},
    "LIGVRRITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["l0758"]}},
    "LIGVSTRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0760"]}},
    "LILIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0764"]}},
    "LILYBAETANVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "LILYBAEVM": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "LIMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0765"]}},
    "LIMATVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l9885"]}},
    "LIMATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["34532"]}},
    "LIMBVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l0773"]}},
    "LIMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["l0774"]}},
    "LIMES": {"-": {"morpho": ["n-s---mn3-"], "uri": ["l0776"]}},
    "LIMNE": {"-": {"morpho": ["n-s---fn1g"], "uri": ["116259"]}},
    "LIMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l0855"]}},
    "LIMOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0797"]}},
    "LIMPIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0800"]}},
    "LIMVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["l0805", "l0806"]},
        "2": {"morpho": ["n-s---mn2-"], "uri": ["l0805", "l0806"]},
        "3": {"morpho": ["aps---mn1-"], "uri": ["l0805"]},
    },
    "LINEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0730"]}},
    "LINEAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0813"]}},
    "LINEATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0817"]}},
    "LINEO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l0818"]}},
    "LINEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0759"]}},
    "LINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["l0821"]}},
    "LINGVA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0822"]}},
    "LINGVARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0823"]}},
    "LINGVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["41266"]}},
    "LINGVLATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0828"]}},
    "LINIGER": {"-": {"morpho": ["aps---mn1r"], "uri": ["l0832"]}},
    "LINIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["l0835"]}},
    "LINO": {"-": {"morpho": ["v1spia--3-"], "uri": ["l0841"]}},
    "LINQVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["l0845"]}},
    "LINTEATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0851"]}},
    "LINTER": {"-": {"morpho": ["n-s---mn3i"], "uri": ["l0854"]}},
    "LINTEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0837"]}},
    "LINTEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0815"]}},
    "LINVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0858"]}},
    "LIPARENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "LIPARENSIS": {
        "N": {"morpho": ["aps---cn3i"], "uri": []},
        "A": {"morpho": ["aps---cn3i"], "uri": []},
    },
    "LIPPIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["l0867"]}},
    "LIPPITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0868"]}},
    "LIPPVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0870"]}},
    "LIQVEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["l0878"]}},
    "LIQVEFIO": {"-": {"morpho": ["v1spia--3-"], "uri": ["l0880"]}},
    "LIQVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["l0881"]}},
    "LIQVESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["l0881"]}},
    "LIQVIDO": {"-": {"morpho": ["rp--------"], "uri": ["51741"]}},
    "LIQVIDVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["l0885"]},
        "2": {"morpho": ["rp--------"], "uri": []},
    },
    "LIQVIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0885"]}},
    "LIQVOR": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["l0889"]},
        "2": {"morpho": ["v1spid--3-"], "uri": ["l0888"]},
    },
    "LIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["l0895"]}},
    "LITIGANS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["99271"]}},
    "LITIGATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["l0911"]}},
    "LITIGIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0914"]}},
    "LITIGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l0916"]}},
    "LITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l0917"]}},
    "LITORALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["l0918"]}},
    "LITOREVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0920"]}},
    "LITTERA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0922"]}},
    "LITTERARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0924"]}},
    "LITTERATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["l0926"]}},
    "LITTERATVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0929"]}},
    "LITTERATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0931"]}},
    "LITTERVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0934"]}},
    "LITVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l9887"]}},
    "LITVS": {
        "-": {"morpho": ["n-s---nn3-"], "uri": ["l0938"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "LITVVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l0939"]}},
    "LIVENS": {"-": {"morpho": ["aps---an3i"], "uri": ["99275"]}},
    "LIVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["l0940"]}},
    "LIVESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["l0940"]}},
    "LIVIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LIVIDVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0944"]}},
    "LIVIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0945"]}},
    "LIVIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "LIVOR": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["l0946"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "LIXA": {
        "1": {"morpho": ["n-s---fn1-"], "uri": ["47739"]},
        "-": {"morpho": ["n-s---fn1-"], "uri": ["47739"]},
    },
    "LIXIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0952"]}},
    "LOCALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["l0955"]}},
    "LOCATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0960"]}},
    "LOCATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["l0961"]}},
    "LOCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l0965"]}},
    "LOCRENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "LOCVLAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0966"]}},
    "LOCVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l0970"]}},
    "LOCVPLES": {"-": {"morpho": ["aps---an3-"], "uri": ["l0971"]}},
    "LOCVPLETO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l0974"]}},
    "LOCVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l0360"]}},
    "LOCVSTA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["l0977"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["l0977"]},
    },
    "LOCVTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0979"]}},
    "LODIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0983"]}},
    "LOGEION": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "LOGOS": {"-": {"morpho": ["n-s---mn2g"], "uri": ["l0994"]}},
    "LOLIGO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["116468"]}},
    "LOLIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0997"]}},
    "LOLLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LONGAEVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l1008"]}},
    "LONGE": {"-": {"morpho": ["rp--------"], "uri": ["41016"]}},
    "LONGINQVITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l1016"]}},
    "LONGINQVVM": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["rp--------"], "uri": ["116490"]},
    },
    "LONGINQVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l1018"]}},
    "LONGITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l1023"]}},
    "LONGVLA": {
        "N": {"morpho": ["n-s---fn1-"], "uri": []},
        "N12": {"morpho": ["n-s---fn1-"], "uri": []},
    },
    "LONGVM": {
        "2": {
            "morpho": ["rp--------", "rp--------"],
            "uri": ["58308", "58308"],
        },
        "1": {"morpho": ["n-s----n2-"], "uri": []},
    },
    "LONGVRIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l1032"]}},
    "LONGVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["l1033"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["l1033"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "LOQVACITER": {"-": {"morpho": ["rp--------"], "uri": ["33352"]}},
    "LOQVACVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l1037"]}},
    "LOQVAX": {"-": {"morpho": ["aps---an3i"], "uri": ["l1038"]}},
    "LOQVELA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l1039"]}},
    "LOQVOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["l1042"]}},
    "LORA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l1043"]}},
    "LOREVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l1047"]}},
    "LORICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l1048"]}},
    "LORICATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l1049"]}},
    "LORICATVS": {"-": {"morpho": ["a-s---mn1-"], "uri": ["57245"]}},
    "LORICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l1051"]}},
    "LORICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l1052"]}},
    "LORIPES": {"-": {"morpho": ["n-s---cn3-"], "uri": ["99285"]}},
    "LORVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l1055"]}},
    "LOTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l1058"]}},
    "LOTIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l1060"]}},
    "LOTOS": {"-": {"morpho": ["n-s---fn2g"], "uri": ["l1063"]}},
    "LVBRICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l1065"]}},
    "LVBRICVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l1067"]}},
    "LVBRICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l1067"]}},
    "LVCA": {
        "-": {"morpho": ["n-s----n1-"], "uri": []},
        "N": {"morpho": ["n-s---fn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "LVCAR": {"-": {"morpho": ["n-s---nn3-"], "uri": ["l1071"]}},
    "LVCCEIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LVCELLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l1074"]}},
    "LVCENS": {"-": {"morpho": ["aps---an3i"], "uri": ["99290"]}},
    "LVCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["l1075"]}},
    "LVCERES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "LVCERNA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l1076"]}},
    "LVCESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["l9918"]}},
    "LVCIDE": {"-": {"morpho": ["rp--------"], "uri": ["32929"]}},
    "LVCIDVM": {"-": {"morpho": ["rp--------"], "uri": ["99291"]}},
    "LVCIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l1086"]}},
    "LVCIFER": {
        "1": {"morpho": ["n-s---mn2r"], "uri": ["37206"]},
        "2": {"morpho": ["aps---mn1r"], "uri": ["l1087"]},
        "N": {"morpho": ["n-s----n2-"], "uri": []},
    },
    "LVCIFVGA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["l1093"]}},
    "LVCIFVGVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l1093"]}},
    "LVCILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LVCIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["l1100"]}},
    "LVCRETIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LVCRETIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LVCRIFICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l1106"]}},
    "LVCRINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "LVCROR": {"-": {"morpho": ["v1spid--1-"], "uri": ["l1111"]}},
    "LVCROSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l1112"]}},
    "LVCRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l1113"]}},
    "LVCTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l1116"]}},
    "LVCTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l1118"]}},
    "LVCTATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["l1119"]}},
    "LVCTIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["l1120"]}},
    "LVCTIFICABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["l1121"]}},
    "LVCTIFICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l1123"]}},
    "LVCTOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["l1126"]}},
    "LVCTVOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l1127"]}},
    "LVCTVS": {
        "-": {"morpho": ["n-s---mn4-"], "uri": ["l9901"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "LVCVBRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l1128"]}},
    "LVCVBRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l1131"]}},
    "LVCVLENTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l1136"]}},
    "LVCVLLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LVCVMO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "LVCVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["l1140"]},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["l1140"]},
    },
    "LVDIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l1156"]}},
    "LVDIBRIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l1144"]}},
    "LVDIBVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l1145"]}},
    "LVDICER": {"-": {"morpho": ["aps---mn1r"], "uri": ["l1146"]}},
    "LVDICRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l1146"]}},
    "LVDICRVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["52245"]}},
    "LVDIFICATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l1150"]}},
    "LVDIFICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["48449"]}},
    "LVDIFICOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["l1153"]}},
    "LVDIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l1156"]}},
    "LVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["l1157"]}},
    "LVDVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["l1158"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["l1158"]},
    },
    "LVELA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["99297"]}},
    "LVES": {
        "-": {"morpho": ["n-s---fn3i"], "uri": ["l1160"]},
        "N": {"morpho": ["n-s---mn1g"], "uri": []},
    },
    "LVGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["l1161"]}},
    "LVGVBRE": {"-": {"morpho": ["rp--------"], "uri": ["46915"]}},
    "LVGVBRIA": {"-": {"morpho": ["n-p---nn3i"], "uri": ["99298"]}},
    "LVGVBRIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["l1163"]}},
    "LVMBRICVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l1172"]}},
    "LVMBVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l1174"]}},
    "LVMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["l1176"]}},
    "LVMINARE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["l1177"]}},
    "LVMINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l1178"]}},
    "LVMINOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l1179"]}},
    "LVNA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["l1180"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["l1180"]},
    },
    "LVNARIS": {
        "-": {"morpho": ["aps---cn3i"], "uri": ["l1181"]},
        "A": {"morpho": ["aps---cn3i"], "uri": ["l1181"]},
    },
    "LVNATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["46620"]}},
    "LVNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l1183"]}},
    "LVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["l1185"]}},
    "LVPA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l1186"]}},
    "LVPANAR": {"-": {"morpho": ["n-s---nn3i"], "uri": ["l1188"]}},
    "LVPATA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["l1192"]}},
    "LVPATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l1192"]}},
    "LVPERCAL": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "LVPINARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l1195"]}},
    "LVPINVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l1197"]}},
    "LVPINVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["l1197"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["l1197"]},
    },
    "LVPVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["l1186"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["l1186"]},
    },
    "LVRIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l1207"]}},
    "LVROR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["l1208"]}},
    "LVSCINIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l1209"]}},
    "LVSCINIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l1209"]}},
    "LVSCITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l1214"]}},
    "LVSCITIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l1215"]}},
    "LVSCVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["l1216"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "LVSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["l1220"]}},
    "LVSORIAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["l1222"]}},
    "LVSORIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l1220"]}},
    "LVSORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l1220"]}},
    "LVSTRALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["l1224"]}},
    "LVSTRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l1227"]}},
    "LVSTRO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["l1233"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["l1232"]},
    },
    "LVSTRVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["l8222"]},
        "2": {"morpho": ["n-s---nn2-"], "uri": ["l8222"]},
    },
    "LVSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["l9903"]}},
    "LVTAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l1235"]}},
    "LVTATIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "LVTEOLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l1238"]}},
    "LVTEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l1241"]}},
    "LVTEVS": {
        "1": {"morpho": ["aps---mn1-"], "uri": ["l1241", "l9910"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["l1241", "l9910"]},
    },
    "LVTO": {"1": {"morpho": ["v1spia--1-"], "uri": ["l1244"]}},
    "LVTOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l1245"]}},
    "LVTRON": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "LVTVLENTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l1248"]}},
    "LVTVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["l1249", "l9911"]},
        "2": {"morpho": ["n-s---nn2-"], "uri": ["l1249", "l9911"]},
    },
    "LVX": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["l1250"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "LVXO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l1252"]}},
    "LVXVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l1257"]}},
    "LVXVRIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["l1253"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "LVXVRIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["l0327"]}},
    "LVXVRIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l1255"]}},
    "LVXVRIOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132558"]}},
    "LVXVRIOSE": {"-": {"morpho": ["rp--------"], "uri": ["49673"]}},
    "LVXVRIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l1256"]}},
    "LVXVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["l1258"]},
        "2": {"morpho": ["n-s---mn4-"], "uri": ["l1258"]},
        "3": {"morpho": ["aps---mn1-"], "uri": ["l1257"]},
    },
    "LYCAEVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "LYCHNOBIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l1266"]}},
    "LYCHNVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l1268"]}},
    "LYCIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "LYCII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "LYCISCVS": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "LYMPHA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["l1276"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "LYMPHATICVS": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": ["l1278"]},
    },
    "LYMPHATVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["l9904"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["24827"]},
    },
    "LYMPHO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l1280"]}},
    "LYNX": {"-": {"morpho": ["n-s---cn3-"], "uri": ["l1282"]}},
    "LYRA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["l1283"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "LYRICI": {
        "-": {"morpho": ["n-p---mn2-"], "uri": ["99320"]},
        "2": {"morpho": ["n-p---mn2-"], "uri": ["99320"]},
        "1": {"morpho": ["n-p---mn2-"], "uri": ["99320"]},
    },
    "LYRICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l1284"]}},
    "LYSIMACHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["116745"]}},
    "LYSIS": {
        "-": {"morpho": ["n-s---fn3i"], "uri": ["l1292"]},
        "N": {"morpho": ["n-s---fn3-"], "uri": []},
    },
    "LYSO": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "MACEDONES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "MACEDONIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "MACEDONICVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "MACELLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0008"]}},
    "MACER": {
        "-": {"morpho": ["aps---mn1r"], "uri": ["m0010"]},
        "N": {"morpho": ["n-s---mn3-"], "uri": []},
    },
    "MACERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m0011"]}},
    "MACERESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["m0012"]}},
    "MACERIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0013"]}},
    "MACERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m0017"]}},
    "MACHAERA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["m0019"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["m0019"]},
    },
    "MACHINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0022"]}},
    "MACHINAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0025"]}},
    "MACHINATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m0027"]}},
    "MACHINATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["m0028"]}},
    "MACHINATRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m0029"]}},
    "MACHINOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["m0030"]}},
    "MACIES": {
        "-": {"morpho": ["n-s---fn5-"], "uri": ["m0034"]},
        "N": {"morpho": ["n-s---mn1g"], "uri": []},
    },
    "MACRESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["m0042"]}},
    "MACRITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m0044"]}},
    "MACRO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "MACTABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["m0049"]}},
    "MACTATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["m0051"]}},
    "MACTATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["m9995"]}},
    "MACTE": {"-": {"morpho": ["----------"], "uri": []}},
    "MACTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m0053"]}},
    "MACTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0943"]}},
    "MACVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0055"]}},
    "MACVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m0058"]}},
    "MACVLOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0059"]}},
    "MADEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["m0060"]}},
    "MADEFIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m0061"]}},
    "MADEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["m0064"]}},
    "MADESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["m0064"]}},
    "MADIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0066"]}},
    "MAECENAS": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "MAECILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MAENA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0072"]}},
    "MAENAS": {
        "-": {"morpho": ["n-s----n-g"], "uri": []},
        "N": {"morpho": ["n-s---mn1g"], "uri": []},
    },
    "MAERENS": {"-": {"morpho": ["aps---an3i"], "uri": ["99329"]}},
    "MAEREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["m0077"]}},
    "MAEROR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["m0078"]}},
    "MAESTITIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0081"]}},
    "MAESTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0084"]}},
    "MAEVIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["33407"]}},
    "MAGA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0134"]}},
    "MAGALIA": {"-": {"morpho": ["n-p---nn3i"], "uri": ["m0086"]}},
    "MAGE": {"-": {"morpho": ["rp--------"], "uri": ["55358"]}},
    "MAGICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0091"]}},
    "MAGIS": {
        "1": {"morpho": ["n-s---fn3-"], "uri": ["m0096"]},
        "2": {"morpho": ["rp--------"], "uri": ["m0095"]},
    },
    "MAGISTER": {"-": {"morpho": ["n-s---mn2r"], "uri": ["m0097"]}},
    "MAGISTERIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0098"]}},
    "MAGISTRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0074"]}},
    "MAGISTRATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["m0103"]}},
    "MAGIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MAGNANIMITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m0113"]}},
    "MAGNANIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0112"]}},
    "MAGNAS": {"-": {"morpho": ["n-s---mn3-"], "uri": ["35073"]}},
    "MAGNES": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["m0118"]},
        "N": {"morpho": ["n-s---mn1g"], "uri": []},
    },
    "MAGNETICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["31429"]}},
    "MAGNIFICE": {"-": {"morpho": ["rp--------"], "uri": ["24862"]}},
    "MAGNIFICENS": {
        "-": {"morpho": ["aps---an3-"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "MAGNIFICENTER": {"-": {"morpho": ["rp--------"], "uri": ["28771"]}},
    "MAGNIFICENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0120"]}},
    "MAGNIFICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m0122"]}},
    "MAGNIFICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0123"]}},
    "MAGNILOQVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0127"]}},
    "MAGNITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m0130"]}},
    "MAGNOPERE": {"-": {"morpho": ["rp--------"], "uri": ["m0131"]}},
    "MAGNVM": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["rp--------"], "uri": ["99333"]},
    },
    "MAGNVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["m0132"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["m0132"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "MAGVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["m0134"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["m0134"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["m0134"]},
    },
    "MAIESTAS": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["m0137"]},
        "N": {"morpho": ["n-s---fn3-"], "uri": ["m0137"]},
    },
    "MAIORES": {"-": {"morpho": ["n-s---mn3-"], "uri": ["99335"]}},
    "MAIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": ["116865"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "MAIVSCVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0145"]}},
    "MALA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0146"]}},
    "MALACIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0149"]}},
    "MALAXO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m0158"]}},
    "MALE": {"-": {"morpho": ["rp--------"], "uri": ["24874"]}},
    "MALEDICENS": {"-": {"morpho": ["aps---an3i"], "uri": ["99339"]}},
    "MALEDICO": {"2": {"morpho": ["v1spia--3-"], "uri": ["m9606"]}},
    "MALEDICTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0162"]}},
    "MALEDICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["45544"]}},
    "MALEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["m9604"]}},
    "MALEFACTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m9499"]}},
    "MALEFICIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0173"]}},
    "MALEFICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0173"]}},
    "MALESVADVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0176"]}},
    "MALEVOLENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0181"]}},
    "MALEVOLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0182"]}},
    "MALI": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "MALIGNE": {"-": {"morpho": ["rp--------"], "uri": ["48638"]}},
    "MALIGNITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m0185"]}},
    "MALIGNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0186"]}},
    "MALITIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0194"]}},
    "MALITIOSA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MALITIOSE": {"-": {"morpho": ["rp--------"], "uri": ["36600"]}},
    "MALLEOLVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["m0099"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["m0099"]},
    },
    "MALLEVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["m0201"]}},
    "MALLVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "MALO": {"-": {"morpho": ["v1spia--3-"], "uri": ["m0206"]}},
    "MALOBATHRON": {"-": {"morpho": ["n-s---nn2g"], "uri": ["99345"]}},
    "MALVA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0214"]}},
    "MALVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["m0218", "m0216"]},
        "2": {"morpho": ["n-s---nn2-"], "uri": ["m0218", "m0216"]},
        "3": {"morpho": ["----------"], "uri": []},
    },
    "MALVS": {
        "1": {"morpho": ["n-s---fn2-"], "uri": ["m0219"]},
        "2": {"morpho": ["n-s---fn2-"], "uri": ["m0219"]},
        "3": {"morpho": ["aps---mn1-"], "uri": ["m0216"]},
    },
    "MAMERTINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MAMERTINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MAMILIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "MAMILLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0220"]}},
    "MAMMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0224"]}},
    "MAMMOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0233"]}},
    "MANABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["m0238"]}},
    "MANAEVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "MANCEPS": {"-": {"morpho": ["n-s---mn3-"], "uri": ["m0242"]}},
    "MANCIPIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0247"]}},
    "MANCIPO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m0252"]}},
    "MANCVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0253"]}},
    "MANDATVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m9996"]}},
    "MANDATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["m9996"]}},
    "MANDELA": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "MANDO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["m0262"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["m0260"]},
        "3": {"morpho": ["v1spia--3-"], "uri": ["m0261"]},
    },
    "MANDRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0263"]}},
    "MANDVCO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["m0270"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["m0269"]},
    },
    "MANE": {
        "1": {"morpho": ["n-s---nn--"], "uri": ["m0272"]},
        "2": {"morpho": ["rp--------"], "uri": ["42177"]},
    },
    "MANEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["m0273"]}},
    "MANES": {
        "-": {"morpho": ["n-p---mn3-"], "uri": ["m0274"]},
        "N": {"morpho": ["n-s---mn1g"], "uri": []},
    },
    "MANGO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["m0276"]}},
    "MANICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0281"]}},
    "MANICATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0284"]}},
    "MANICVLVS": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "MANIFESTE": {"-": {"morpho": ["rp--------"], "uri": ["43947"]}},
    "MANIFESTO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["m0291"]},
        "2": {"morpho": ["rp--------"], "uri": ["24906"]},
    },
    "MANIFESTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0292"]}},
    "MANILIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "MANIPVLARIS": {
        "-": {"morpho": ["aps---cn3i"], "uri": ["m0298"]},
        "2": {"morpho": ["aps---cn3i"], "uri": ["m0298"]},
        "1": {"morpho": ["aps---cn3i"], "uri": ["m0298"]},
    },
    "MANIPVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["m0301"]}},
    "MANIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MANLIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "MANNVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["m0304"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["m0304"]},
    },
    "MANO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m0305"]}},
    "MANSITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m0309"]}},
    "MANSVEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["m0313"]}},
    "MANSVEFIO": {"-": {"morpho": ["v1spia--3-"], "uri": ["m0314"]}},
    "MANSVES": {"-": {"morpho": ["aps---an3-"], "uri": ["m0316"]}},
    "MANSVESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["m0317"]}},
    "MANSVETE": {"-": {"morpho": ["rp--------"], "uri": ["44039"]}},
    "MANSVETVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m0320"]}},
    "MANSVETVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["32025"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "MANTELE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["24915"]}},
    "MANTICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0324"]}},
    "MANVALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["m0338"]}},
    "MANVBIAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["99358"]}},
    "MANVBRIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0347"]}},
    "MANVCLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["117045"]}},
    "MANVLEATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0352"]}},
    "MANVMISSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m0355"]}},
    "MANVMITTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["m9607"]}},
    "MANVPRETIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0360"]}},
    "MANVPRETIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0361"]}},
    "MANVS": {
        "1": {"morpho": ["n-s---fn4-"], "uri": ["m0362"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["m0363"]},
    },
    "MAPALIA": {"-": {"morpho": ["n-p---nn3i"], "uri": ["99362"]}},
    "MAPPA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m9367"]}},
    "MARATHVS": {
        "-": {"morpho": ["n-s----n2-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "MARCELLIA": {"N": {"morpho": ["n-s----n2-"], "uri": []}},
    "MARCELLINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MARCELLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["117061"]}},
    "MARCENS": {"-": {"morpho": ["aps---an3i"], "uri": ["24921"]}},
    "MARCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["m0371"]}},
    "MARCESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["m0371"]}},
    "MARCIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MARCIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0373"]}},
    "MARCIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "MARCOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["m0374"]}},
    "MARCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["m0944"]}},
    "MARE": {
        "-": {"morpho": ["n-s---nn3i"], "uri": ["m0378"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "MARGARITA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["m0381"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "MARGARITVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0381"]}},
    "MARGO": {"-": {"morpho": ["n-s---cn3-"], "uri": ["m0388"]}},
    "MARINVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["m0391"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "MARISCA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0392"]}},
    "MARITA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0117"]}},
    "MARITALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["m0393"]}},
    "MARITIMI": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "MARITIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0397"]}},
    "MARITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m0396"]}},
    "MARITVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["m0126"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["m0114"]},
    },
    "MARIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MARMOR": {"-": {"morpho": ["n-s---nn3-"], "uri": ["m0400"]}},
    "MARMORARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["m0401"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["m0401"]},
    },
    "MARMOREVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0403"]}},
    "MARRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0409"]}},
    "MARS": {"N": {"morpho": ["n-s---mn3i"], "uri": []}},
    "MARTIALIS": {
        "-": {"morpho": ["aps---cn3i"], "uri": []},
        "N": {"morpho": ["n-s---fn3-"], "uri": []},
        "A": {"morpho": ["aps---cn3i"], "uri": []},
    },
    "MARTIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "MAS": {"-": {"morpho": ["n-s---mn3-"], "uri": ["m0417"]}},
    "MASCVLINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0421"]}},
    "MASCVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0423"]}},
    "MASINISSA": {"N": {"morpho": ["n-s---mn1-"], "uri": []}},
    "MASSA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["m0425"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["m0425"]},
    },
    "MASSILIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "MATARA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["53937"]}},
    "MATELLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0445"]}},
    "MATELLIO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["m0446"]}},
    "MATEOLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0447"]}},
    "MATER": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["m0448"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "MATERCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0449"]}},
    "MATERIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0451"]}},
    "MATERIALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["m0452"]}},
    "MATERIATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m0454"]}},
    "MATERIATVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m9982"]}},
    "MATERIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["m0451"]}},
    "MATERINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0455"]}},
    "MATERIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m0456"]}},
    "MATERIOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132563"]}},
    "MATERNVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["m0458"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "MATERTERA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0459"]}},
    "MATHEMATICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0461"]}},
    "MATHEMATICVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["m0461"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["m0461"]},
    },
    "MATRIMONIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0472"]}},
    "MATRIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0473"]}},
    "MATRINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MATRONA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["m0475"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "MATRONALIA": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "MATRONALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["m0476"]}},
    "MATVRE": {"-": {"morpho": ["rp--------"], "uri": ["24955"]}},
    "MATVRESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["m0491"]}},
    "MATVRITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m0492"]}},
    "MATVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m0493"]}},
    "MATVRVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["m0494"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "MATVTINVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0497"]}},
    "MATVTINVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["m0497"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["m0497"]},
    },
    "MAVRETANIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MAXILLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0499"]}},
    "MAXIMITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m0501"]}},
    "MAXIMVM": {"-": {"morpho": ["rp--------"], "uri": []}},
    "MAXIMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MAZONOMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["m0505"]}},
    "MEATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["m9998"]}},
    "MECHANICOS": {"-": {"morpho": ["rp--------"], "uri": []}},
    "MEDEA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "MEDENS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["99381"]}},
    "MEDEOR": {"-": {"morpho": ["v1spid--2-"], "uri": ["m0525"]}},
    "MEDIANVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0528"]}},
    "MEDIASTINVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["34967"]}},
    "MEDICA": {
        "1": {"morpho": ["n-s---fn1-"], "uri": ["m0639", "m0549"]},
        "2": {"morpho": ["n-s---fn1-"], "uri": ["m0639", "m0549"]},
    },
    "MEDICABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["m0537"]}},
    "MEDICAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["m0539"]}},
    "MEDICAMENTOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0541"]}},
    "MEDICAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0542"]}},
    "MEDICATVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["m0536"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["32283"]},
    },
    "MEDICINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0545"]}},
    "MEDICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m0547"]}},
    "MEDICOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132565"]}},
    "MEDICVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["m0549"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["m0549"]},
        "N": {"morpho": ["aps---mn1-"], "uri": ["m0549"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["m0549"]},
    },
    "MEDIETAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m0552"]}},
    "MEDIMNVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0554"]}},
    "MEDIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m0555"]}},
    "MEDIOCRICVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0556"]}},
    "MEDIOCRIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["m0557"]}},
    "MEDIOCRITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m0558"]}},
    "MEDIOCRITER": {"-": {"morpho": ["rp--------"], "uri": ["44909"]}},
    "MEDIPONTVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["m0562"]}},
    "MEDITAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0564"]}},
    "MEDITATE": {"-": {"morpho": ["rp--------"], "uri": ["99386"]}},
    "MEDITATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m0565"]}},
    "MEDITERRANEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0570"]}},
    "MEDITERRANEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0570"]}},
    "MEDITOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["m0572"]}},
    "MEDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0574"]}},
    "MEDIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0574"]}},
    "MEDIVSFIDIVS": {"-": {"morpho": ["----------"], "uri": []}},
    "MEDVLLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0575"]}},
    "MEDVLLIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "MEDVSA": {
        "-": {"morpho": ["n-s----n1-"], "uri": []},
        "N": {"morpho": ["n-s---fn1-"], "uri": []},
    },
    "MEGALENSIS": {
        "-": {"morpho": ["aps---cn3i"], "uri": []},
        "A": {"morpho": ["aps---cn3i"], "uri": []},
    },
    "MEGALOGRAPHIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0584"]}},
    "MEGARIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "MEGISTANES": {"-": {"morpho": ["n-s---fn3-"], "uri": []}},
    "MEHERCVLE": {"-": {"morpho": ["----------"], "uri": []}},
    "MEI": {"-": {"morpho": ["aps---mn1-"], "uri": []}},
    "MEIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["m0588"]}},
    "MEL": {"-": {"morpho": ["n-s---nn3-"], "uri": ["m0589"]}},
    "MELA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MELANTHIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0603"]}},
    "MELICHROS": {"-": {"morpho": ["n-s---mn2g"], "uri": ["117281"]}},
    "MELICVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["m0620"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["m0620"]},
    },
    "MELIMELVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0622"]}},
    "MELIORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m0627"]}},
    "MELISPHYLLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["48440"]}},
    "MELITA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "MELITENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "MELITENSIA": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "MELITENSIS": {
        "N": {"morpho": ["aps---cn3i"], "uri": []},
        "A": {"morpho": ["aps---cn3i"], "uri": []},
    },
    "MELIVSCVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0640"]}},
    "MELLITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0660"]}},
    "MELOGRAPHIA": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "MELOS": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0673"]}},
    "MEMBRANA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0679"]}},
    "MEMBRATIM": {"-": {"morpho": ["rp--------"], "uri": ["m0685"]}},
    "MEMBRATVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0691"]}},
    "MEMBRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0690"]}},
    "MEMINI": {"-": {"morpho": ["v1sria--3-"], "uri": ["m0693"]}},
    "MEMOR": {"-": {"morpho": ["aps---an3-"], "uri": ["99398"]}},
    "MEMORABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["m0695"]}},
    "MEMORANDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["45807"]}},
    "MEMORATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["m0700"]}},
    "MEMORIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0702"]}},
    "MEMORITER": {"-": {"morpho": ["rp--------"], "uri": ["50888"]}},
    "MEMORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m0706"]}},
    "MENAENI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MENAENVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "MENDA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0709"]}},
    "MENDACIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0713"]}},
    "MENDAX": {"-": {"morpho": ["aps---an3i"], "uri": ["m0715"]}},
    "MENDICATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m0718"]}},
    "MENDICE": {"-": {"morpho": ["rp--------"], "uri": ["99401"]}},
    "MENDICITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m0720"]}},
    "MENDICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m0721"]}},
    "MENDICVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["m0723"]},
        "2": {"morpho": ["n-s---mn2-"], "uri": ["m0723"]},
        "1": {"morpho": ["n-s---mn2-"], "uri": ["m0723"]},
    },
    "MENDO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "MENDOSE": {"-": {"morpho": ["rp--------"], "uri": ["52741"]}},
    "MENDOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0725"]}},
    "MENDVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0709"]}},
    "MENENIANVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "MENENIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MENISCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MENS": {
        "-": {"morpho": ["n-s---fn3i"], "uri": ["m0733"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "MENSA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0734"]}},
    "MENSIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["m0739"]}},
    "MENSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["m0740"]}},
    "MENSTRVA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["99404"]}},
    "MENSTRVALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["m0742"]}},
    "MENSTRVVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0745"]}},
    "MENSTRVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0745"]}},
    "MENSVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m9983"]}},
    "MENTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0755"]}},
    "MENTIO": {
        "1": {"morpho": ["n-s---fn3-"], "uri": ["m0759"]},
        "2": {"morpho": ["v1spia--4-"], "uri": ["117420"]},
    },
    "MENTIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["m0760"]}},
    "MENTOR": {"N": {"morpho": ["n-s---mn3-"], "uri": ["117424"]}},
    "MENTVLA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["m0764"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "MENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0945"]}},
    "MEO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m0768"]}},
    "MEPHITIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m0582"]}},
    "MERACE": {"-": {"morpho": ["rp--------"], "uri": ["117427"]}},
    "MERACVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0771"]}},
    "MERCABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["m0774"]}},
    "MERCATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["m0777"]}},
    "MERCATVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0818"]}},
    "MERCATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["m9792"]}},
    "MERCEDVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0783"]}},
    "MERCENARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["43547"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["56204"]},
    },
    "MERCES": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m0785"]}},
    "MERCIMONIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0786"]}},
    "MERCOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["m0787"]}},
    "MERCVRIALIS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---cn3i"], "uri": []},
    },
    "MERCVRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["m0788"]}},
    "MERDA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0789"]}},
    "MERENS": {"-": {"morpho": ["aps---an3i"], "uri": ["99407"]}},
    "MEREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["m0792"]}},
    "MEREOR": {"-": {"morpho": ["v1spid--2-"], "uri": ["132570"]}},
    "MERETRICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0796"]}},
    "MERETRICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0798"]}},
    "MERETRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m0799"]}},
    "MEREVS": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "MERGA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0800"]}},
    "MERGES": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m0801"]}},
    "MERGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["m0803"]}},
    "MERGVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["m0805"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["m0805"]},
    },
    "MERIDIANO": {"-": {"morpho": ["rp--------"], "uri": ["117463"]}},
    "MERIDIANVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0808"]}},
    "MERIDIANVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0808"]}},
    "MERIDIES": {"-": {"morpho": ["n-s---mn5-"], "uri": []}},
    "MERIDIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m0812"]}},
    "MERITO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["m0817"]},
        "2": {"morpho": ["rp--------"], "uri": ["m0819"]},
    },
    "MERITORIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0820"]}},
    "MERITORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0820"]}},
    "MERITVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0778"]}},
    "MERITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["25029"]}},
    "MEROPS": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["m0823"]},
        "N": {"morpho": ["n-s---mn3-"], "uri": ["m0823"]},
    },
    "MEROS": {
        "-": {"morpho": ["n-s----n-g"], "uri": []},
        "N": {"morpho": ["n-s----n-g"], "uri": []},
    },
    "MERSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m0826"]}},
    "MERVLA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["m0828"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["m0828"]},
    },
    "MERVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0832"]}},
    "MERVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0832"]}},
    "MERX": {"-": {"morpho": ["n-s---fn3i"], "uri": ["m0833"]}},
    "MESAVLOE": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "MESE": {"-": {"morpho": ["n-s---fn1g"], "uri": ["m0837"]}},
    "MESIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MESOLABIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0843"]}},
    "MESOS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "MESSALA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MESSANA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "MESSIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["m0857"]}},
    "MESSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["m0858"]}},
    "META": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["m0860"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "METALLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0870"]}},
    "METAPONTVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "METAXA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["117535"]}},
    "METELLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["m9991"]}},
    "METHODVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["32500"]}},
    "METILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "METIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["m0893"]}},
    "METO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["117548"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["m0897"]},
    },
    "METOPA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0903"]}},
    "METOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["m0894"]}},
    "METRETA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0907"]}},
    "METRODORVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "METROPOLIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "METROPOLITAE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "METRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0914"]}},
    "METTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "METVENS": {"-": {"morpho": ["aps---an3i"], "uri": ["99415"]}},
    "METVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["m0917"]}},
    "METVS": {
        "-": {"morpho": ["n-s---mn4-"], "uri": ["m0918"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "MEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0920"]}},
    "MEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0919"]}},
    "MEZENTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0950"]}},
    "MICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0951"]}},
    "MICANS": {"-": {"morpho": ["aps---an3i"], "uri": ["99419"]}},
    "MICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m0957"]}},
    "MICTVRIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["m0966"]}},
    "MIGRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m0971"]}},
    "MIGRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m0973"]}},
    "MILES": {"-": {"morpho": ["n-s---mn3-"], "uri": ["m0974"]}},
    "MILESII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MILESIVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "MILETVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MILIA": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "MILIARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0859", "m9986"]}},
    "MILITARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["m0987"]}},
    "MILITIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0988"]}},
    "MILITO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["m0991"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["m0990"]},
    },
    "MILIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": []}},
    "MILLE": {"-": {"morpho": ["----------"], "uri": []}},
    "MILLESIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0889"]}},
    "MILLIARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["36611"]}},
    "MILLIARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["47941"]}},
    "MILLIES": {"-": {"morpho": ["rp--------"], "uri": ["99424"]}},
    "MILVVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["m1002"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["m1002"]},
    },
    "MILYAE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "MIMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m1005"]}},
    "MIMICE": {"-": {"morpho": ["rp--------"], "uri": ["43189", "99426"]}},
    "MIMICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1010"]}},
    "MIMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["m1005"]}},
    "MINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m1016"]}},
    "MINACITER": {"-": {"morpho": ["rp--------"], "uri": ["25058"]}},
    "MINAE": {
        "-": {"morpho": ["n-p---fn1-"], "uri": ["m1018"]},
        "N": {"morpho": ["n-p---fn1-"], "uri": ["m1018"]},
    },
    "MINANTER": {"-": {"morpho": ["rp--------"], "uri": ["29303"]}},
    "MINAX": {"-": {"morpho": ["aps---an3i"], "uri": ["m1022"]}},
    "MINEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["m1024"]}},
    "MINERVA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "MINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["m0588"]}},
    "MINIACEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1028"]}},
    "MINIMVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["99429"]},
        "2": {
            "morpho": ["rp--------", "n-s---nn2-"],
            "uri": ["29175", "99429"],
        },
    },
    "MINISCOR": {"-": {"morpho": ["v1spid--3-"], "uri": []}},
    "MINISTER": {
        "1": {"morpho": ["n-s---mn2r"], "uri": ["m1034"]},
        "2": {"morpho": ["aps---mn1r"], "uri": ["m1034"]},
    },
    "MINISTERIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m1037"]}},
    "MINISTRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m1034"]}},
    "MINISTRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m1038"]}},
    "MINISTRATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["m1039"]}},
    "MINISTRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m1042"]}},
    "MINITABVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1044"]}},
    "MINITOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["m1046"]}},
    "MINIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m1047"]}},
    "MINOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["m1048"]}},
    "MINVCIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "MINVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["m1057"]}},
    "MINVS": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "MINVSCVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1062"]}},
    "MINVTAL": {"-": {"morpho": ["n-s---nn3i"], "uri": ["m1063"]}},
    "MINVTATIM": {"-": {"morpho": ["rp--------"], "uri": ["m1064"]}},
    "MINVTE": {"-": {"morpho": ["rp--------"], "uri": ["54264"]}},
    "MINVTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m1065"]}},
    "MINVTIM": {"-": {"morpho": ["rp--------"], "uri": ["m1067"]}},
    "MINVTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["99431"]}},
    "MINVTVS": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": ["25071"]},
    },
    "MIRABILE": {"-": {"morpho": ["rp--------"], "uri": []}},
    "MIRABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["m1076"]}},
    "MIRABVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1078"]}},
    "MIRACVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0924"]}},
    "MIRANDE": {"-": {"morpho": ["rp--------"], "uri": []}},
    "MIRANDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["43797"]}},
    "MIRATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["m1084"]}},
    "MIRATRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m1085"]}},
    "MIRE": {"-": {"morpho": ["rp--------"], "uri": ["40769"]}},
    "MIRIFICE": {"-": {"morpho": ["rp--------"], "uri": ["54246"]}},
    "MIRIFICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1088"]}},
    "MIRMILLO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["117702"]}},
    "MIROR": {"-": {"morpho": ["v1spid--1-"], "uri": ["m1092"]}},
    "MIRVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "MIRVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1093"]}},
    "MISCELLANEA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["m0926"]}},
    "MISCELLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1096"]}},
    "MISCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["m1097"]}},
    "MISELLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1100"]}},
    "MISER": {"-": {"morpho": ["aps---mn1r"], "uri": ["m1101"]}},
    "MISERABILE": {"-": {"morpho": ["rp--------"], "uri": ["53968"]}},
    "MISERABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["m1102"]}},
    "MISERABILITER": {"-": {"morpho": ["rp--------"], "uri": ["54447"]}},
    "MISERANDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["33099"]}},
    "MISERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m1103"]}},
    "MISERE": {"-": {"morpho": ["rp--------"], "uri": ["25083"]}},
    "MISEREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["m1105"]}},
    "MISEREOR": {"-": {"morpho": ["v1spid--2-"], "uri": ["132577"]}},
    "MISERESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["m0784"]}},
    "MISERIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m1107"]}},
    "MISERICORDIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m1108"]}},
    "MISERICORS": {"-": {"morpho": ["aps---an3-"], "uri": ["m1109"]}},
    "MISERITER": {"-": {"morpho": ["rp--------"], "uri": ["48684"]}},
    "MISEROR": {"-": {"morpho": ["v1spid--1-"], "uri": ["m1113"]}},
    "MISSILE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["m1120"]}},
    "MISSILIA": {"-": {"morpho": ["n-p---nn3i"], "uri": ["99439"]}},
    "MISSILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["m1120"]}},
    "MISSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m1121"]}},
    "MISSITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m1122"]}},
    "MISSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["m9994"]}},
    "MITE": {"-": {"morpho": ["rp--------"], "uri": ["34979"]}},
    "MITESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["m1127"]}},
    "MITHRIDATES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "MITIGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m1135"]}},
    "MITIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["m1137"]}},
    "MITRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m1138"]}},
    "MITRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1140"]}},
    "MITTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["m1143"]}},
    "MITVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["m1144"]}},
    "MITYLENAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "MIXTIM": {"-": {"morpho": ["rp--------"], "uri": ["m1149"]}},
    "MIXTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m1150"]}},
    "MIXTVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m1069"]}},
    "MIXTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["32847"]}},
    "MNASISTRATVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MNEMOSYNON": {"-": {"morpho": ["n-s---nn2g"], "uri": ["99445"]}},
    "MOBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["m1156"]}},
    "MOBILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m1157"]}},
    "MOBILITER": {"-": {"morpho": ["rp--------"], "uri": ["25100"]}},
    "MOBILITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m1158"]}},
    "MODERABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["m1159"]}},
    "MODERAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["m1160"]}},
    "MODERANTER": {"-": {"morpho": ["rp--------"], "uri": ["57584"]}},
    "MODERATE": {"-": {"morpho": ["rp--------"], "uri": ["45492"]}},
    "MODERATIM": {"-": {"morpho": ["rp--------"], "uri": ["m1162"]}},
    "MODERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m1163"]}},
    "MODERATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["m1164"]}},
    "MODERATRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m1165"]}},
    "MODERATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["25104"]}},
    "MODEROR": {"-": {"morpho": ["v1spid--1-"], "uri": ["m1167"]}},
    "MODESTE": {"-": {"morpho": ["rp--------"], "uri": ["m1169"]}},
    "MODESTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m1168"]}},
    "MODESTVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["m1169"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "MODICE": {"-": {"morpho": ["rp--------"], "uri": ["43554"]}},
    "MODICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1174"]}},
    "MODIFICATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m1175"]}},
    "MODIOLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["m1180"]}},
    "MODIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m1181"]}},
    "MODIVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["m1181"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["m1181"]},
    },
    "MODO": {
        "1": {"morpho": ["rp--------"], "uri": ["m9610"]},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "MODVLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m1185"]}},
    "MODVLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["m1186"]}},
    "MODVLATVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["m9796"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["57027"]},
    },
    "MODVLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["m1188"]}},
    "MODVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["m1189"]}},
    "MODVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["m1190"]}},
    "MOECHA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m1199"]}},
    "MOECHOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["m1198"]}},
    "MOECHVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["m1199"]}},
    "MOENIA": {"-": {"morpho": ["n-p---nn3i"], "uri": ["m1725"]}},
    "MOLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m1203"]}},
    "MOLARIS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["99458"]},
        "2": {"morpho": ["aps---cn3i"], "uri": ["m1204"]},
    },
    "MOLARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1204"]}},
    "MOLES": {"-": {"morpho": ["n-s---fn3i"], "uri": ["m1213"]}},
    "MOLESTE": {"-": {"morpho": ["rp--------"], "uri": ["31934"]}},
    "MOLESTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m1214"]}},
    "MOLESTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m1215"]}},
    "MOLESTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1216"]}},
    "MOLETRINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m1217"]}},
    "MOLILE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["m1218"]}},
    "MOLIMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["m1219"]}},
    "MOLIMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m1220"]}},
    "MOLIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["m1223"]}},
    "MOLITIO": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["m1224", "m1247"]},
        "1": {"morpho": ["n-s---fn3-"], "uri": ["m1224", "m1247"]},
    },
    "MOLITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["m1225", "m1205"]}},
    "MOLLESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["m1227"]}},
    "MOLLICELLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1229"]}},
    "MOLLICVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1231"]}},
    "MOLLIMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m1235"]}},
    "MOLLIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["m1236"]}},
    "MOLLIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["m1238"]}},
    "MOLLITER": {"-": {"morpho": ["rp--------"], "uri": ["25130"]}},
    "MOLLITIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m1239"]}},
    "MOLLITIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["m1239"]}},
    "MOLLITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m1241"]}},
    "MOLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["m1245"]}},
    "MOMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["m1262"]}},
    "MOMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m1268"]}},
    "MONACHA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m1270"]}},
    "MONACHVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["m1270"]}},
    "MONAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m1279"]}},
    "MONASTERIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m1282"]}},
    "MONEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["m1292"]}},
    "MONETA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["m1293"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "MONETALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["m1294"]}},
    "MONILE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["m1297"]}},
    "MONIMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["35999"]}},
    "MONITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m1298"]}},
    "MONITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["m1300"]}},
    "MONITVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["47067"]}},
    "MONITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["m9797"]}},
    "MONOPTEROS": {"-": {"morpho": ["aps---mn1g"], "uri": ["m1331"]}},
    "MONOTRIGLYPHVS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "MONS": {
        "-": {"morpho": ["n-s---mn3i"], "uri": ["m1344"]},
        "N": {"morpho": ["n-s---mn3-"], "uri": []},
    },
    "MONSTRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m1346"]}},
    "MONSTRATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["m1348"]}},
    "MONSTRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m1354"]}},
    "MONSTRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m1357"]}},
    "MONSTRVOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1356"]}},
    "MONTANI": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "MONTANVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["m1360"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["m1360"]},
    },
    "MONTIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["m1366"]}},
    "MONTIVAGVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1368"]}},
    "MONTVOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0927"]}},
    "MONVMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0930"]}},
    "MORA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m1374"]}},
    "MORALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["m1376"]}},
    "MORATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m1380"]}},
    "MORATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["m1381"]}},
    "MORATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1382"]}},
    "MORBIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1387"]}},
    "MORBOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1391"]}},
    "MORBVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["m1392"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["m1392"]},
    },
    "MORDAX": {"-": {"morpho": ["aps---an3i"], "uri": ["m1394"]}},
    "MORDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["m1395"]}},
    "MORDICVS": {
        "-": {"morpho": ["rp--------"], "uri": ["25162"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["m1401"]},
    },
    "MORIBVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1405"]}},
    "MORIGEROR": {"-": {"morpho": ["v1spid--1-"], "uri": ["m1407"]}},
    "MORIGERVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1408"]}},
    "MORIOR": {"-": {"morpho": ["v1spid--3i"], "uri": ["m1411"]}},
    "MOROR": {
        "1": {"morpho": ["v1spid--1-"], "uri": ["m1416"]},
        "2": {"morpho": ["v1spid--1-"], "uri": ["m1416"]},
    },
    "MOROSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1418", "m1455"]}},
    "MORS": {
        "-": {"morpho": ["n-s---fn3i"], "uri": ["m1420"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "MORSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["m9799"]}},
    "MORTALIA": {"-": {"morpho": ["n-p---nn3i"], "uri": ["99476"]}},
    "MORTALIS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["99477"]},
        "2": {"morpho": ["aps---cn3i"], "uri": ["m1426"]},
    },
    "MORTALITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m1427"]}},
    "MORTARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m1429"]}},
    "MORTICINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1430"]}},
    "MORTIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["m1431"]}},
    "MORTVVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["30404"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["31870"]},
    },
    "MORVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m1441"]}},
    "MORVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["m1441"]}},
    "MOS": {"-": {"morpho": ["n-s---mn3-"], "uri": ["m1442"]}},
    "MOTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m1450"]}},
    "MOTIVNCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m1452"]}},
    "MOTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m1453"]}},
    "MOTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["m9800"]}},
    "MOVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["m1456"]}},
    "MOX": {"-": {"morpho": ["rp--------"], "uri": ["m1457"]}},
    "MVCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["m1459"]}},
    "MVCIA": {"N": {"morpho": ["n-s----n2-"], "uri": []}},
    "MVCIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1461"]}},
    "MVCIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "MVCRO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["m1466"]}},
    "MVCVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["m1471", "m1662"]}},
    "MVGILIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["99487"]}},
    "MVGILLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MVGIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["m1477"]}},
    "MVGITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["m9801"]}},
    "MVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m1479"]}},
    "MVLCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["m1485"]}},
    "MVLCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m1587"]}},
    "MVLCTRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m1588"]}},
    "MVLCTRARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m1589"]}},
    "MVLCTRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m1588"]}},
    "MVLGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["m1593"]}},
    "MVLIEBRIA": {"-": {"morpho": ["n-p---nn3i"], "uri": ["99488"]}},
    "MVLIEBRIS": {
        "-": {"morpho": ["aps---cn3i"], "uri": ["m1594"]},
        "N": {"morpho": ["aps---cn3i"], "uri": ["m1594"]},
        "A": {"morpho": ["aps---cn3i"], "uri": ["m1594"]},
    },
    "MVLIEBRITER": {"-": {"morpho": ["rp--------"], "uri": ["32197"]}},
    "MVLIER": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m1595"]}},
    "MVLIERARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["m1596"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["m1596"]},
    },
    "MVLIERCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m1597"]}},
    "MVLINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1603"]}},
    "MVLIO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["m1604"]}},
    "MVLLEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1607"]}},
    "MVLLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["m1609"]}},
    "MVLSVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m1948"]}},
    "MVLSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1948"]}},
    "MVLTA": {
        "1": {"morpho": ["n-s---fn1-"], "uri": ["m1613"]},
        "2": {"morpho": ["rp--------"], "uri": []},
    },
    "MVLTANGVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1615"]}},
    "MVLTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m1619"]}},
    "MVLTESIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1620"]}},
    "MVLTI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["30043"]}},
    "MVLTICIA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["m0204"]}},
    "MVLTIFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["m9605"]}},
    "MVLTIFARIAM": {"-": {"morpho": ["rp--------"], "uri": ["m1635"]}},
    "MVLTIFIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1638"]}},
    "MVLTIFORMIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["m1644"]}},
    "MVLTIFORVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1646"]}},
    "MVLTIGENVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1694"]}},
    "MVLTIMODIS": {"-": {"morpho": ["rp--------"], "uri": ["58182"]}},
    "MVLTIMODVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1657"]}},
    "MVLTIPLEX": {"-": {"morpho": ["aps---an3i"], "uri": ["m1665"]}},
    "MVLTIPLICATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m1667"]}},
    "MVLTIPLICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m1671"]}},
    "MVLTITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m1683"]}},
    "MVLTITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m1684"]}},
    "MVLTIVAGVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1685"]}},
    "MVLTIVOLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1690"]}},
    "MVLTO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["m1693"]},
        "2": {"morpho": ["rp--------"], "uri": ["25204"]},
    },
    "MVLTVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["m1696"]},
        "2": {"morpho": ["rp--------"], "uri": ["25205"]},
    },
    "MVLTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1695"]}},
    "MVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["m1479"]}},
    "MVMMIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MVNATIVS": {
        "-": {"morpho": ["n-s----n2-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "MVNDANVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1699"]}},
    "MVNDE": {"-": {"morpho": ["rp--------"], "uri": ["99504"]}},
    "MVNDITIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["m1711"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "MVNDITIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["m1711"]}},
    "MVNDVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["m1717"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["m1715"]},
    },
    "MVNERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m1723"]}},
    "MVNEROR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132586"]}},
    "MVNIA": {"-": {"morpho": ["n-p---nn3i"], "uri": ["m0932"]}},
    "MVNICEPS": {"-": {"morpho": ["n-s---nn3-"], "uri": []}},
    "MVNICIPALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["m1727"]}},
    "MVNICIPIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m1731"]}},
    "MVNIFICE": {"-": {"morpho": ["rp--------"], "uri": ["56158"]}},
    "MVNIFICENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m1737"]}},
    "MVNIFICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m1739"]}},
    "MVNIFICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1740"]}},
    "MVNIMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["m1741"]}},
    "MVNIMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m1742"]}},
    "MVNIO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["118187"]},
        "2": {"morpho": ["v1spia--4-"], "uri": ["m1743"]},
    },
    "MVNITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m1745"]}},
    "MVNITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["m1748"]}},
    "MVNITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["45250"]}},
    "MVNVS": {"-": {"morpho": ["n-s---nn3-"], "uri": ["m1751"]}},
    "MVNVSCVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m1752"]}},
    "MVRAENA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["45194"]}},
    "MVRALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["m1754"]}},
    "MVRCIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MVRENA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MVREX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["m1759"]}},
    "MVRGANTINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MVRGANTINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "MVRIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m1761"]}},
    "MVRIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["m1761"]}},
    "MVRMVR": {"-": {"morpho": ["n-s---nn3-"], "uri": ["m1774"]}},
    "MVRMVRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m1776"]}},
    "MVRMVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m1781"]}},
    "MVRVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["m1803"]}},
    "MVS": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["m1804"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "MVSA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["25228"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["25228"]},
    },
    "MVSAEVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["29416"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "MVSCA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m1807"]}},
    "MVSCARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1808"]}},
    "MVSCIPVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m1812"]}},
    "MVSCOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1814"]}},
    "MVSCVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["m1816"]}},
    "MVSCVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["m1818"]}},
    "MVSICA": {
        "1": {"morpho": ["n-p---nn2-", "n-s---fn1-"], "uri": ["m1824"]},
        "2": {"morpho": ["n-s---fn1-"], "uri": ["m1824"]},
    },
    "MVSICVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["m1824"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["m1824"]},
    },
    "MVSIMO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["m1826"]}},
    "MVSSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m1836"]}},
    "MVSTACEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m1837"]}},
    "MVSTACEVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["m9611"]}},
    "MVSTARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1838"]}},
    "MVSTELA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["m1840"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "MVSTEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1842"]}},
    "MVSTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MVSTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m1846"]}},
    "MVSTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1846"]}},
    "MVTABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["m1847"]}},
    "MVTABILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m1848"]}},
    "MVTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m1850"]}},
    "MVTILO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m1857"]}},
    "MVTILVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["m1858"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "MVTIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["m1866"]}},
    "MVTO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["25239"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["m1864"]},
    },
    "MVTVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["m1874"]}},
    "MVTVO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["51273"]},
        "2": {"morpho": ["rp--------"], "uri": ["30765"]},
    },
    "MVTVOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["m1876"]}},
    "MVTVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["m1877"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["99529"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["m1877"]},
    },
    "MVTVVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m1878"]}},
    "MVTVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1878"]}},
    "MVTYCENSIS": {
        "N": {"morpho": ["aps---cn3i"], "uri": []},
        "A": {"morpho": ["aps---cn3i"], "uri": []},
    },
    "MYNDVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MYOPARO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["99533"]}},
    "MYRICE": {"-": {"morpho": ["n-s---fn1g"], "uri": ["m1901"]}},
    "MYRO": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "MYRRHA": {
        "-": {"morpho": ["n-s----n1-"], "uri": []},
        "N": {"morpho": ["n-s---fn1-"], "uri": []},
    },
    "MYRRHEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["118364"]}},
    "MYRRHINA": {
        "1": {"morpho": ["n-s----n1-"], "uri": []},
        "2": {"morpho": ["n-s----n2-"], "uri": []},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "MYRTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m1950"]}},
    "MYRTETVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0940"]}},
    "MYRTEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1801"]}},
    "MYRTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m1951"]}},
    "MYRTVS": {"-": {"morpho": ["n-s---cn2-"], "uri": ["m1949"]}},
    "MYSTA": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "MYSTAGOGVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["m1930"]}},
    "MYSTERIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m1933"]}},
    "MYSTICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1935"]}},
    "MYTILENAE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "NABLIA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["n0001"]}},
    "NAEVIVS": {
        "A": {"morpho": ["n-s----n1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "NAEVVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["n9006"]}},
    "NAM": {"-": {"morpho": ["----------"], "uri": []}},
    "NAMQVE": {"-": {"morpho": ["----------"], "uri": []}},
    "NANCISCOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["n0013"]}},
    "NANVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["n0010"]}},
    "NAOS": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "NARCISSVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["n0024"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["n0024"]},
    },
    "NARDVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "NARDVS": {"-": {"morpho": ["n-s---fn2-"], "uri": []}},
    "NARIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["25254"]}},
    "NARRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["n9036"]}},
    "NARRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["n0039"]}},
    "NASCENTIA": {"-": {"morpho": ["n-p---nn3i"], "uri": ["n0033"]}},
    "NASCOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["n0044"]}},
    "NASITERNA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["118470"]}},
    "NASSA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["n0048"]}},
    "NASVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["n0053"]}},
    "NASVTE": {"-": {"morpho": ["rp--------"], "uri": ["99540"]}},
    "NASVTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0054"]}},
    "NATA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["n0088"]}},
    "NATALES": {"-": {"morpho": ["n-p---mn3i"], "uri": ["99541"]}},
    "NATALICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0058"]}},
    "NATALIS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["99543"]},
        "2": {"morpho": ["aps---cn3i"], "uri": ["n0059"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "NATATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["n0062"]}},
    "NATI": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "NATINOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["n0064"]}},
    "NATIO": {
        "1": {"morpho": ["n-s---fn3-"], "uri": ["n0065"]},
        "2": {"morpho": ["n-s---fn3-"], "uri": ["n0065"]},
    },
    "NATIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["n0066"]}},
    "NATIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0068"]}},
    "NATO": {"-": {"morpho": ["v1spia--1-"], "uri": ["n0069"]}},
    "NATRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["n0070"]}},
    "NATV": {"-": {"morpho": ["n-s----n4-"], "uri": []}},
    "NATVRA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["n0071"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "NATVRALIA": {"-": {"morpho": ["n-p---nn3i"], "uri": ["118501"]}},
    "NATVRALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["n0073"]}},
    "NATVRALITER": {"-": {"morpho": ["rp--------"], "uri": ["50840"]}},
    "NATVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["n0142"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["25271"]},
    },
    "NAVALE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["n0076"]}},
    "NAVALIA": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "NAVALIS": {
        "-": {"morpho": ["aps---cn3i"], "uri": ["n0076"]},
        "A": {"morpho": ["aps---cn3i"], "uri": ["n0076"]},
    },
    "NAVARCHVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["n0078"]}},
    "NAVE": {"-": {"morpho": ["rp--------"], "uri": ["40405"]}},
    "NAVFRAGIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["n0086"]}},
    "NAVFRAGVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["n0041"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["n0011"]},
    },
    "NAVICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["n0091"]}},
    "NAVICVLARIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["n0093"]}},
    "NAVICVLARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["n0093"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["n0093"]},
    },
    "NAVIFRAGVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["31286"]}},
    "NAVIGABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["n0095"]}},
    "NAVIGANS": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "NAVIGATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["n0096"]}},
    "NAVIGER": {"-": {"morpho": ["aps---mn1r"], "uri": ["n0098"]}},
    "NAVIGIOLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["n0100"]}},
    "NAVIGIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["n0101"]}},
    "NAVIGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["n0102"]}},
    "NAVIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["n0103"]}},
    "NAVITA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["25281"]}},
    "NAVITER": {"-": {"morpho": ["rp--------"], "uri": ["25282"]}},
    "NAVIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NAVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["n0105"]}},
    "NAVMACHIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["n0106"]}},
    "NAVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["n0108"]}},
    "NAVPACTVS": {"N": {"morpho": ["n-s---mn2-", "n-s---mn2-"], "uri": []}},
    "NAVSEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["n0113"]}},
    "NAVSEABVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0115"]}},
    "NAVSEATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["n0116"]}},
    "NAVSEO": {"-": {"morpho": ["v1spia--1-"], "uri": ["n0117"]}},
    "NAVTA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["n0122"]}},
    "NAVTICI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["n0126"]}},
    "NAVTICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0126"]}},
    "NAVTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NAVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0128"]}},
    "NE": {
        "1": {"morpho": ["rp--------"], "uri": ["n0131"]},
        "2": {"morpho": ["r---------"], "uri": []},
        "3": {"morpho": ["r---------"], "uri": []},
        "4": {"morpho": ["----------"], "uri": []},
    },
    "NEAPOLIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "NEBVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["n0135"]}},
    "NEBVLO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["n0136"]}},
    "NEBVLOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0138"]}},
    "NEC": {
        "1": {"morpho": ["rp--------"], "uri": ["25286"]},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "NECDVM": {
        "1": {"morpho": ["rp--------"], "uri": ["31057"]},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "NECESSARIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["n0141"]}},
    "NECESSARIO": {"-": {"morpho": ["rp--------"], "uri": ["56888"]}},
    "NECESSARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["n0141"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["n0141"]},
    },
    "NECESSE": {"-": {"morpho": ["aps---n---"], "uri": ["25291"]}},
    "NECESSIS": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "NECESSITAS": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["n0146"]},
        "N": {"morpho": ["n-s---fn3-"], "uri": ["n0146"]},
    },
    "NECESSITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["n0147"]}},
    "NECNE": {
        "-": {
            "morpho": ["rp--------", "rp--------"],
            "uri": ["57828", "57828"],
        }
    },
    "NECNON": {"-": {"morpho": ["----------"], "uri": []}},
    "NECO": {"-": {"morpho": ["v1spia--1-"], "uri": ["n0151"]}},
    "NECOPINANS": {"-": {"morpho": ["aps---an3i"], "uri": ["n9821"]}},
    "NECOPINATO": {"-": {"morpho": ["rp--------"], "uri": ["29757"]}},
    "NECOPINATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n9822"]}},
    "NECTAR": {"-": {"morpho": ["n-s---nn3i"], "uri": ["n0157"]}},
    "NECTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["n0160"]}},
    "NECVBI": {"-": {"morpho": ["----------"], "uri": []}},
    "NEDVM": {
        "1": {"morpho": ["rp--------"], "uri": ["n0755"]},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "NEFANDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0163"]}},
    "NEFARIE": {"-": {"morpho": ["rp--------"], "uri": ["45548"]}},
    "NEFARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["n0165"]}},
    "NEFARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0165"]}},
    "NEFAS": {"-": {"morpho": ["n-s----n--"], "uri": []}},
    "NEFASTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["n0167"]}},
    "NEFASTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0167"]}},
    "NEGITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["n0180"]}},
    "NEGLECTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["n9767"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["99565"]},
    },
    "NEGLIGENS": {
        "-": {"morpho": ["aps---an3-"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "NEGLIGENTER": {"-": {"morpho": ["rp--------"], "uri": ["52968"]}},
    "NEGLIGENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["30863"]}},
    "NEGLIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["51006"]}},
    "NEGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["n0185"]}},
    "NEGOTIALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["n0186"]}},
    "NEGOTIATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["n0187"]}},
    "NEGOTIATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["n0188"]}},
    "NEGOTIOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["n0193"]}},
    "NEGOTIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0195"]}},
    "NEGOTIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["n0196"]}},
    "NEMO": {"-": {"morpho": ["----------"], "uri": []}},
    "NEMORALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["n0201"]}},
    "NEMORIVAGVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0205"]}},
    "NEMOROSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0206"]}},
    "N": {"-": {"morpho": ["rp--------"], "uri": []}},
    "NEMVS": {"-": {"morpho": ["n-s---nn3-"], "uri": ["n0207"]}},
    "NENIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["n0209"]}},
    "NENVM": {"-": {"morpho": ["r---------"], "uri": []}},
    "NEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["n0211"]}},
    "NEPOS": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["n0224"]},
        "2": {"morpho": ["n-s---mn3-"], "uri": ["n0224"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "NEPOTOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["n0226"]}},
    "NEPTIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["n0229"]}},
    "NEPTVNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NEQVAM": {
        "1": {"morpho": ["n-s----n--"], "uri": []},
        "2": {"morpho": ["aps---an--"], "uri": ["n0758"]},
    },
    "NEQVAQVAM": {"-": {"morpho": ["rp--------"], "uri": ["n9756"]}},
    "NEQVE": {"-": {"morpho": ["----------"], "uri": []}},
    "NEQVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["n0233"]}},
    "NEQVIQVAM": {"-": {"morpho": ["rp--------"], "uri": ["n0761"]}},
    "NEQVIS": {"-": {"morpho": ["----------"], "uri": []}},
    "NEQVITER": {"-": {"morpho": ["rp--------"], "uri": ["n0234"]}},
    "NEQVITIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["n0236"]}},
    "NEQVITIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["n0236"]}},
    "NEQVO": {"-": {"morpho": ["rp--------"], "uri": ["118650"]}},
    "NERO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "NERVICVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["n0242"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["n0242"]},
    },
    "NERVOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0245"]}},
    "NERVVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["n0246"]}},
    "NERVVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["n0247"]}},
    "NESCIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["n0249"]}},
    "NESCIOQVA": {"-": {"morpho": ["rp--------"], "uri": []}},
    "NESCIOQVANTVS": {"-": {"morpho": ["rp--------"], "uri": []}},
    "NESCIOQVIS": {"-": {"morpho": ["----------"], "uri": []}},
    "NESCIOQVOT": {"-": {"morpho": ["----------"], "uri": []}},
    "NESCIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0250"]}},
    "NESTOR": {
        "-": {"morpho": ["n-s----n-g"], "uri": []},
        "N": {"morpho": ["n-s---mn3-"], "uri": []},
    },
    "NETE": {"-": {"morpho": ["n-s---fn1g"], "uri": ["n0253"]}},
    "NETINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "NETVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "NEVE": {
        "1": {"morpho": ["----------"], "uri": []},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "NEVTER": {"-": {"morpho": ["----------"], "uri": []}},
    "NEX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["n0264"]}},
    "NEXILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["n0266"]}},
    "NEXVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["n9770"]}},
    "NI": {
        "1": {"morpho": ["rp--------"], "uri": ["52497"]},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "NICASIO": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "NICE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "NICETERIVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "NICO": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "NICOMEDES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "NICTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["n0276"]}},
    "NIDEO": {"-": {"morpho": ["v1spia--2-"], "uri": []}},
    "NIDESCO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "NIDIFICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0282"]}},
    "NIDOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["n0283"]}},
    "NIDVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["n0288"]}},
    "NIGER": {
        "-": {"morpho": ["aps---mn1r"], "uri": ["n0291"]},
        "N": {"morpho": ["n-s---mn3-"], "uri": []},
    },
    "NIGRANS": {"-": {"morpho": ["aps---an3i"], "uri": ["99575"]}},
    "NIGRESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["n0294"]}},
    "NIGRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["n0301"]}},
    "NIGROR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["n0304"]}},
    "NIHIL": {"-": {"morpho": ["----------", "n-s---nn--"], "uri": ["25327"]}},
    "NIHILDVM": {"-": {"morpho": ["rp--------"], "uri": ["99576"]}},
    "NIHILOMINVS": {"-": {"morpho": ["rp--------"], "uri": ["n9910"]}},
    "NIHILOSETIVS": {"-": {"morpho": ["r---------"], "uri": []}},
    "NIHILVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["99577"]}},
    "NILVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["118734"]}},
    "NIMBOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0308"]}},
    "NIMBVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["n0309"]}},
    "NIMIO": {"-": {"morpho": ["rp--------"], "uri": ["54353"]}},
    "NIMIOPERE": {"-": {"morpho": ["rp--------"], "uri": ["n9310"]}},
    "NIMIRVM": {"-": {"morpho": ["rp--------"], "uri": ["n9811"]}},
    "NIMIS": {"-": {"morpho": ["rp--------"], "uri": ["n0311"]}},
    "NIMIVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["n0312"]},
        "2": {"morpho": ["rp--------"], "uri": ["25332"]},
    },
    "NIMIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0312"]}},
    "NINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["n0313"]}},
    "NINGVIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["n0315"]}},
    "NISI": {"-": {"morpho": ["----------"], "uri": []}},
    "NISVS": {
        "-": {"morpho": ["n-s---mn4-"], "uri": ["n0303"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["118743"]},
    },
    "NITENS": {"-": {"morpho": ["aps---an3i"], "uri": ["118750"]}},
    "NITEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["n0325"]}},
    "NITESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["n9823"]}},
    "NITIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0330"]}},
    "NITOR": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["n9323"]},
        "2": {"morpho": ["v1spid--3-"], "uri": ["n0322"]},
    },
    "NITROSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n9330"]}},
    "NITRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["n0331"]}},
    "NIVALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["n0332"]}},
    "NIVENDIALIS": {
        "-": {"morpho": ["aps---cn3i"], "uri": []},
        "A": {"morpho": ["aps---cn3i"], "uri": []},
    },
    "NIVEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0336", "43133"]}},
    "NIVOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0339"]}},
    "NIX": {"-": {"morpho": ["n-s---fn3i"], "uri": ["n0440"]}},
    "NIXOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["n0442"]}},
    "NO": {"-": {"morpho": ["v1spia--1-"], "uri": ["n0444"]}},
    "NOBILIS": {
        "1": {"morpho": ["n-s----n3-"], "uri": []},
        "2": {"morpho": ["aps---cn3i"], "uri": ["n0445"]},
    },
    "NOBILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["n0446"]}},
    "NOBILITER": {"-": {"morpho": ["rp--------"], "uri": ["118773"]}},
    "NOBILITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["n0447"]}},
    "NOCENS": {"-": {"morpho": ["aps---an3i"], "uri": ["99584"]}},
    "NOCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["n0448"]}},
    "NOCTE": {"-": {"morpho": ["rp--------"], "uri": ["57302"]}},
    "NOCTIVAGVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0462"]}},
    "NOCTV": {"-": {"morpho": ["rp--------"], "uri": ["n0464"]}},
    "NOCTVA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["n0465"]}},
    "NOCTVRNVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["n0469"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": ["n0469"]},
    },
    "NODATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["n0476"]}},
    "NODO": {"-": {"morpho": ["v1spia--1-"], "uri": ["n0478"]}},
    "NODOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0480"]}},
    "NODVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["n0482"]}},
    "NOLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["n0489"]}},
    "NOMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["n0492"]}},
    "NOMENCLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["n0493"]}},
    "NOMENCLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["n0494"]}},
    "NOMENCLATVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["n0495"]}},
    "NOMENTVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "NOMINATIM": {"-": {"morpho": ["rp--------"], "uri": ["n0499"]}},
    "NOMINATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["n0500"]}},
    "NOMINATVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["n9771"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["41718"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "NOMINITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["n0504"]}},
    "NOMINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["n0505"]}},
    "NOMISMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["n0506"]}},
    "NON": {"-": {"morpho": ["rp--------"], "uri": ["n9508"]}},
    "NONAE": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "NONAGENI": {"-": {"morpho": ["rp--------"], "uri": []}},
    "NONAGESIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0512"]}},
    "NONAGIES": {"-": {"morpho": ["rp--------"], "uri": ["118824"]}},
    "NONAGINTA": {"-": {"morpho": ["----------"], "uri": []}},
    "NONANVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0517"]}},
    "NONARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0518"]}},
    "NONDVM": {"-": {"morpho": ["rp--------"], "uri": ["n0766"]}},
    "NONGENTI": {"-": {"morpho": ["----------"], "uri": []}},
    "NONNE": {
        "-": {
            "morpho": ["rp--------", "rp--------"],
            "uri": ["n0767", "n0767"],
        }
    },
    "NONNVLLVS": {"-": {"morpho": ["----------"], "uri": []}},
    "NONNVMQVAM": {"-": {"morpho": ["rp--------"], "uri": ["n9815"]}},
    "NONVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0145"]}},
    "NORBA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "NORBANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NORMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["n0525"]}},
    "NOS": {"-": {"morpho": ["o-----an--"], "uri": []}},
    "NOSCITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["n0533"]}},
    "NOSCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["n0534"]}},
    "NOSTER": {"-": {"morpho": ["aps---mn1-"], "uri": []}},
    "NOSTRAS": {"-": {"morpho": ["aps---an3-"], "uri": ["n0771"]}},
    "NOSTRI": {"-": {"morpho": ["aps---mn1-", "n-s----n2-"], "uri": []}},
    "NOTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["n0537"]}},
    "NOTABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["n0538"]}},
    "NOTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["n0542"]}},
    "NOTESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["n0543"]}},
    "NOTHVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["n0545"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "NOTI": {
        "-": {"morpho": ["n-p---mn2-"], "uri": ["99599"]},
        "N": {"morpho": ["n-p---mn2-"], "uri": ["99599"]},
    },
    "NOTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["n0549"]}},
    "NOTITIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["n0550"]}},
    "NOTITIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["n0550"]}},
    "NOTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["n0552"]}},
    "NOTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["n0553"]}},
    "NOTVS": {
        "1": {"morpho": ["n-s----n-g"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": ["55561"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "NOVALE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["n0558"]}},
    "NOVALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["n0558"]}},
    "NOVATILLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "NOVE": {"-": {"morpho": ["rp--------"], "uri": ["32387"]}},
    "NOVELLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["35172"]}},
    "NOVELLVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["n0568"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "NOVEM": {
        "-": {"morpho": ["----------"], "uri": []},
        "N": {"morpho": ["n-s----n--"], "uri": []},
    },
    "NOVEMBER": {
        "A": {"morpho": ["aps---mn1r"], "uri": []},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "NOVENDIALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["n0340"]}},
    "NOVENVS": {"-": {"morpho": ["rp--------"], "uri": []}},
    "NOVERCA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["n0577"]}},
    "NOVERCALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["n0578"]}},
    "NOVICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0581"]}},
    "NOVIES": {"-": {"morpho": ["rp--------"], "uri": ["99606"]}},
    "NOVITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["n0585"]}},
    "NOVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["n0586"]}},
    "NOVVMCOMVM": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---nn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "NOVVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["n0587"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["n0587"]},
    },
    "NOX": {
        "-": {"morpho": ["n-s---fn3i"], "uri": ["n0588"]},
        "N": {"morpho": ["n-s----n--"], "uri": []},
    },
    "NOXA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["n0589"]}},
    "NOXIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["25378"]}},
    "NOXIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0591"]}},
    "NOXIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0594"]}},
    "NVBES": {"-": {"morpho": ["n-s---fn3i"], "uri": ["n0596"]}},
    "NVBIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["n0597"]}},
    "NVBILA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["99610"]}},
    "NVBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["n0601"]}},
    "NVBILO": {"-": {"morpho": ["v1spia--1-"], "uri": ["n0602"]}},
    "NVBILVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["n0604"]}},
    "NVBILVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0604"]}},
    "NVBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["n0606"]}},
    "NVCEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0611"]}},
    "NVCLEVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["n0618"]}},
    "NVDIVS": {"-": {"morpho": ["n-s----n--"], "uri": []}},
    "NVDIVSTERTIVS": {"-": {"morpho": ["rp--------"], "uri": ["51864"]}},
    "NVDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["n0627"]}},
    "NVDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0629"]}},
    "NVGAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["n0342"]}},
    "NVGATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["n0634"]}},
    "NVGATORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0634"]}},
    "NVGAX": {"-": {"morpho": ["aps---an3i"], "uri": ["n0637"]}},
    "NVGOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["n0641"]}},
    "NVLLATENVS": {"-": {"morpho": ["rp--------"], "uri": ["n0644"]}},
    "NVLLVS": {"-": {"morpho": ["----------"], "uri": []}},
    "NVM": {"-": {"morpho": ["rp--------"], "uri": ["n0486"]}},
    "NVMA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "NVMANTIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "NVMARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["118975"]}},
    "NVMATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["118976"]}},
    "NVMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["n0651"]}},
    "NVMENIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NVMERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["n0655"]}},
    "NVMERATVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["n0650"]}},
    "NVMERATVS": {
        "C": {"morpho": ["aps---mn1-"], "uri": ["54226"]},
        "-": {"morpho": ["aps---mn1-"], "uri": ["54226"]},
    },
    "NVMERO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["n0658"]},
        "2": {"morpho": ["rp--------"], "uri": ["n0659"]},
    },
    "NVMEROSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0661"]}},
    "NVMERVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["n0662"]}},
    "NVMICIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NVMICVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NVMIDICVS": {
        "C": {"morpho": ["n-s----n2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "NVMITORIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NVMNE": {"-": {"morpho": ["r---------"], "uri": []}},
    "NVMQVAM": {"-": {"morpho": ["rp--------"], "uri": ["n0780"]}},
    "NVMQVANDO": {"-": {"morpho": ["r---------"], "uri": []}},
    "NVMQVI": {"-": {"morpho": ["rp--------"], "uri": ["31601"]}},
    "NVMQVID": {"-": {"morpho": ["rp--------"], "uri": ["38043"]}},
    "NVMQVIDNAM": {"-": {"morpho": ["rp--------"], "uri": ["99620"]}},
    "NVMQVIS": {"-": {"morpho": ["----------"], "uri": []}},
    "NVMVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["118990"]}},
    "NVMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["118991"]}},
    "NVNC": {"-": {"morpho": ["rp--------"], "uri": ["n0670"]}},
    "NVNCVPATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["n0673"]}},
    "NVNCVPO": {"-": {"morpho": ["v1spia--1-"], "uri": ["n0676"]}},
    "NVNDINAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["n0094"]}},
    "NVNDINATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["n0681"]}},
    "NVNDINOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["n0684"]}},
    "NVNTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["n0345"]}},
    "NVNTIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["n0689"]}},
    "NVNTIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["n0347"]}},
    "NVNTIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["n0346"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["n0344"]},
    },
    "NVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["$0150"]}},
    "NVPER": {"-": {"morpho": ["rp--------"], "uri": ["n0692"]}},
    "NVPTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["n9764"]}},
    "NVPTIAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["n0695"]}},
    "NVPTIALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["n0696"]}},
    "NVRVS": {"-": {"morpho": ["n-s---fn4-"], "uri": ["n0703"]}},
    "NVSQVAM": {"-": {"morpho": ["rp--------"], "uri": ["n0784"]}},
    "NVTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["n0709"]}},
    "NVTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["n0710"]}},
    "NVTRICIA": {
        "1": {"morpho": ["n-s---fn1-"], "uri": ["n0349"]},
        "2": {"morpho": ["n-p---nn2-"], "uri": ["119030"]},
    },
    "NVTRICIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["n0351"]}},
    "NVTRICIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["n0350"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["n0348"]},
    },
    "NVTRICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["n0716"]}},
    "NVTRIMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["n0719"]}},
    "NVTRIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["n0720"]}},
    "NVTRIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["132597"]}},
    "NVTRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["n0724"]}},
    "NVTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["n0725"]}},
    "NVX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["n0726"]}},
    "NYMPHA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["n0735"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "NYMPHAEVM": {"N": {"morpho": ["n-s---nn2-"], "uri": ["n0353"]}},
    "NYMPHE": {"-": {"morpho": ["n-s---fn1-"], "uri": ["32644"]}},
    "NYMPHO": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "NYMPHODORVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NYSA": {
        "-": {"morpho": ["n-s----n-g"], "uri": []},
        "N": {"morpho": ["n-s---fn1-"], "uri": []},
    },
    "O": {"-": {"morpho": ["----------"], "uri": []}},
    "OB": {"-": {"morpho": ["p---------"], "uri": []}},
    "OBAERATVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["43569"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["o0007"]},
    },
    "OBAMBVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0010"]}},
    "OBARMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0013"]}},
    "OBBA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["o0021"]}},
    "OBBRVTESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0023"]}},
    "OBDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0025"]}},
    "OBDVCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0028"]}},
    "OBDVCTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0029"]}},
    "OBDVRESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0035"]}},
    "OBDVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o9036"]}},
    "OBEDIENS": {"-": {"morpho": ["aps---an3i"], "uri": ["119085"]}},
    "OBEDIENTER": {"-": {"morpho": ["rp--------"], "uri": ["45570"]}},
    "OBEDIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["39587"]}},
    "OBEO": {"1": {"morpho": ["v1spia--4-"], "uri": ["o0040"]}},
    "OBEQVITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0041"]}},
    "OBERRO": {"2": {"morpho": ["v1spia--1-"], "uri": ["o0042"]}},
    "OBESVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0036"]}},
    "OBEX": {"-": {"morpho": ["n-s---cn3-"], "uri": ["o0046"]}},
    "OBHAEREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["o0047"]}},
    "OBIACEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["o0051"]}},
    "OBICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["o0052"]}},
    "OBIECTA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["99635"]}},
    "OBIECTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0056"]}},
    "OBIECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0058"]}},
    "OBIECTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0060"]}},
    "OBIECTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["o0060"]}},
    "OBIRASCOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["o0061"]}},
    "OBITER": {"-": {"morpho": ["rp--------"], "uri": ["o0063"]}},
    "OBITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["o9990"]}},
    "OBIVRGATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0065"]}},
    "OBIVRGATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["o0066"]}},
    "OBIVRGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0068"]}},
    "OBLATRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0080"]}},
    "OBLECTAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0083"]}},
    "OBLECTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0084"]}},
    "OBLECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0087"]}},
    "OBLENIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["o0090"]}},
    "OBLIDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0091"]}},
    "OBLIGATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["99640"]}},
    "OBLIGO": {"2": {"morpho": ["v1spia--1-"], "uri": ["o0094"]}},
    "OBLIMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0096"]}},
    "OBLINO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0099"]}},
    "OBLIQVE": {"-": {"morpho": ["rp--------"], "uri": ["33639"]}},
    "OBLIQVO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["o0102"]},
        "2": {"morpho": ["rp--------"], "uri": ["99642"]},
    },
    "OBLIQVVM": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["rp--------"], "uri": []},
    },
    "OBLIQVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0103"]}},
    "OBLITESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0104"]}},
    "OBLITTERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0107"]}},
    "OBLIVIO": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["o0110"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "OBLIVIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0111"]}},
    "OBLIVISCOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["o0112"]}},
    "OBLIVIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0113"]}},
    "OBLONGVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0118"]}},
    "OBLOQVOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["o0120"]}},
    "OBLVCTOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["o0123"]}},
    "OBMOLIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["o0133"]}},
    "OBMOVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["o0135"]}},
    "OBMVRMVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0137"]}},
    "OBMVTESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0140"]}},
    "OBNITOR": {"2": {"morpho": ["v1spid--3-"], "uri": ["o0143"]}},
    "OBNIXVM": {"-": {"morpho": ["rp--------"], "uri": ["119184"]}},
    "OBNIXVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["25440"]}},
    "OBNOXIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0147"]}},
    "OBNVBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0152"]}},
    "OBOLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["o0161"]}},
    "OBORIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["o0162"]}},
    "OBREPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0167"]}},
    "OBRETIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["o0172"]}},
    "OBRIGESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0174"]}},
    "OBRVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0182"]}},
    "OBRVSSA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["o0183"]}},
    "OBSAEPIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["o0187"]}},
    "OBSCENA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["99651"]}},
    "OBSCENE": {"-": {"morpho": ["rp--------"], "uri": ["25450"]}},
    "OBSCENVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0623"]}},
    "OBSCENVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0609"]}},
    "OBSCVRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0196"]}},
    "OBSCVRE": {"-": {"morpho": ["rp--------"], "uri": ["52187"]}},
    "OBSCVRITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0201"]}},
    "OBSCVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0202"]}},
    "OBSCVRVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["o0203"]},
        "2": {"morpho": ["rp--------"], "uri": []},
    },
    "OBSCVRVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0203"]}},
    "OBSECRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0205"]}},
    "OBSECVNDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0209"]}},
    "OBSEDEO": {"-": {"morpho": ["v1spia--2-"], "uri": []}},
    "OBSEQVENS": {
        "O": {"morpho": ["aps---an3i"], "uri": ["99654"]},
        "A": {"morpho": ["aps---an3i"], "uri": ["99654"]},
        "-": {"morpho": ["aps---an3i"], "uri": ["99654"]},
    },
    "OBSEQVENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["o0224"]}},
    "OBSEQVIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0217"]}},
    "OBSEQVOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["o0218"]}},
    "OBSERO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["o0221"]},
        "3": {
            "morpho": ["v1spia--3-", "v1spia--1-"],
            "uri": ["o0222", "o0221"],
        },
    },
    "OBSERVABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["o0223"]}},
    "OBSERVATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0225"]}},
    "OBSERVATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["o0226"]}},
    "OBSERVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0229"]}},
    "OBSES": {"-": {"morpho": ["n-s---cn3-"], "uri": ["o0230"]}},
    "OBSESSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0231"]}},
    "OBSESSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["o0232"]}},
    "OBSIDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["o0236"]}},
    "OBSIDIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0239"]}},
    "OBSIDIVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["o9722"]},
        "2": {"morpho": ["n-s---nn2-"], "uri": ["o9722"]},
    },
    "OBSIDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0243"]}},
    "OBSIGNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0246"]}},
    "OBSISTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0248"]}},
    "OBSOLEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["o0249"]}},
    "OBSOLEFIO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o9249"]}},
    "OBSOLESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0250"]}},
    "OBSOLETE": {"-": {"morpho": ["rp--------"], "uri": ["43322"]}},
    "OBSOLETVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["35026"]}},
    "OBSOLIDATVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "OBSONATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["o0254"]}},
    "OBSONITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0255"]}},
    "OBSONIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0256"]}},
    "OBSONO": {"1": {"morpho": ["v1spia--1-"], "uri": ["o0257", "o0252"]}},
    "OBSTACVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0260"]}},
    "OBSTANTIA": {
        "1": {"morpho": ["n-p---nn3i"], "uri": ["o0261"]},
        "2": {"morpho": ["n-p---nn3i"], "uri": ["o0261"]},
    },
    "OBSTETRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0265"]}},
    "OBSTINATE": {"-": {"morpho": ["rp--------"], "uri": ["25476"]}},
    "OBSTINATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0266"]}},
    "OBSTINATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["25478"]}},
    "OBSTIPVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0270"]}},
    "OBSTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0271"]}},
    "OBSTREPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0277"]}},
    "OBSTRIGILLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0280"]}},
    "OBSTRINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0282"]}},
    "OBSTRVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0285"]}},
    "OBSTVPEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["o0287"]}},
    "OBSTVPEFIO": {"-": {"morpho": ["v1spia--2-"], "uri": ["o9287"]}},
    "OBSTVPESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["52985"]}},
    "OBSVM": {"1": {"morpho": ["v1spia--3-"], "uri": ["o0291"]}},
    "OBSVRDESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0293"]}},
    "OBTEGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0295"]}},
    "OBTEMPERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0297"]}},
    "OBTENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0298"]}},
    "OBTENTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["o9994"]}},
    "OBTERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0302"]}},
    "OBTESTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0303"]}},
    "OBTESTOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["o0304"]}},
    "OBTINEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["o0309"]}},
    "OBTINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0310"]}},
    "OBTINGVO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "OBTORPESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0312"]}},
    "OBTORQVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["o0313"]}},
    "OBTRECTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0316"]}},
    "OBTRECTATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["o0317"]}},
    "OBTRECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0318"]}},
    "OBTRVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0320"]}},
    "OBTRVNCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0322"]}},
    "OBTVNDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0324"]}},
    "OBTVRBO": {"2": {"morpho": ["v1spia--1-"], "uri": ["o0329"]}},
    "OBTVRGESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0330"]}},
    "OBTVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0331"]}},
    "OBTVSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["25498"]}},
    "OBTVTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["o9996"]}},
    "OBVCVLA": {
        "-": {"morpho": ["n-s----n1-"], "uri": []},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "OBVENIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["o0344"]}},
    "OBVERSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["99669"]}},
    "OBVERSOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["o0350"]}},
    "OBVERSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["35393"]}},
    "OBVERTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0352"]}},
    "OBVIAM": {"-": {"morpho": ["rp--------"], "uri": ["o7353"]}},
    "OBVIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0607"]}},
    "OBVMBRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o9350"]}},
    "OBVNCVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0354"]}},
    "OBVOLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0356"]}},
    "OCCAECO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0364"]}},
    "OCCALLESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0368"]}},
    "OCCANO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o9742"]}},
    "OCCASIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0370"]}},
    "OCCASVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["o9998"]}},
    "OCCEDO": {"1": {"morpho": ["v1spia--3-"], "uri": ["o0376"]}},
    "OCCIDENS": {
        "N": {"morpho": ["n-s----n3-"], "uri": []},
        "-": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "OCCIDIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0384"]}},
    "OCCIDO": {
        "1": {"morpho": ["v1spia--3-"], "uri": ["o0387", "o0386", "o0387"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["o0387", "o0386", "o0387"]},
    },
    "OCCIDVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0389"]}},
    "OCCIPIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["o0392"]}},
    "OCCIPITIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0393"]}},
    "OCCIPVT": {"-": {"morpho": ["n-s---nn3-"], "uri": ["o0394"]}},
    "OCCISIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0395"]}},
    "OCCISVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["119427"]}},
    "OCCLVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0401"]}},
    "OCCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0403"]}},
    "OCCVBO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0405"]}},
    "OCCVLCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0406"]}},
    "OCCVLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0407"]}},
    "OCCVLTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0408"]}},
    "OCCVLTE": {"-": {"morpho": ["rp--------"], "uri": ["32492"]}},
    "OCCVLTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0411"]}},
    "OCCVLTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["31173"]}},
    "OCCVLTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["40299"]}},
    "OCCVMBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0412"]}},
    "OCCVPATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0414"]}},
    "OCCVPATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["38555"]}},
    "OCCVPO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["o0418"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["o0417"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "OCCVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0419"]}},
    "OCCVRRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0420"]}},
    "OCCVRSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0427"]}},
    "OCCVRSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["o9710"]}},
    "OCEANVS": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["o9727"]},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "OCELLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["o0431"]}},
    "OCHRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["o0432"]}},
    "OCIMVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0434"]}},
    "OCINVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o9434"]}},
    "OCIOR": {"-": {"morpho": ["acs---cn3-"], "uri": ["o1200"]}},
    "OCITER": {"-": {"morpho": ["rp--------"], "uri": ["119458"]}},
    "OCLIFERIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0435"]}},
    "OCREA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["o0438"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["o0438"]},
    },
    "OCREATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0439"]}},
    "OCTACHORDOS": {"-": {"morpho": ["aps---mn1g"], "uri": ["31207"]}},
    "OCTANS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["o0448"]}},
    "OCTASTYLOS": {"-": {"morpho": ["aps---mn1g"], "uri": ["o0453"]}},
    "OCTAVAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["99679"]}},
    "OCTAVIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "OCTAVIANVS": {
        "C": {"morpho": ["n-s----n2-"], "uri": []},
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "OCTAVIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OCTAVVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["o0458"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "OCTINGENTESIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0464"]}},
    "OCTINGENTI": {"-": {"morpho": ["----------"], "uri": []}},
    "OCTO": {"-": {"morpho": ["----------"], "uri": []}},
    "OCTOBER": {
        "A": {"morpho": ["aps---mn1r"], "uri": []},
        "N": {"morpho": ["n-s---mn3-"], "uri": []},
    },
    "OCTOGENARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0472"]}},
    "OCTOGESIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0474"]}},
    "OCTOGINTA": {"-": {"morpho": ["----------"], "uri": []}},
    "OCTOGONOS": {"-": {"morpho": ["aps---mn1g"], "uri": ["119502"]}},
    "OCTONI": {"-": {"morpho": ["rp--------"], "uri": []}},
    "OCTOPHORON": {"-": {"morpho": ["n-s---nn2-"], "uri": ["31180"]}},
    "OCTVPLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0489"]}},
    "OCTVSSIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["o0490"]}},
    "OCVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["o0502"]}},
    "ODEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0507"]}},
    "ODI": {"-": {"morpho": ["v1sria--3-"], "uri": ["o0508"]}},
    "ODIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0512"]}},
    "ODIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0513"]}},
    "ODOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["o0516"]}},
    "ODORATVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["o9711"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["25528"]},
    },
    "ODORIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["o0523"]}},
    "ODORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0526"]}},
    "ODOROR": {"-": {"morpho": ["v1spid--1-"], "uri": ["o0488"]}},
    "ODORVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0527"]}},
    "ODYSSEA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "OEAX": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "OECONOMIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["o0528"]}},
    "OECVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["119551"]}},
    "OENOPHORVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0543"]}},
    "OESTRVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["o0548"]}},
    "OESYPVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0549"]}},
    "OETA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "OFELLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["o0551"]}},
    "OFFA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["o0552"]}},
    "OFFENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0562"]}},
    "OFFENSA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["o9987"]}},
    "OFFENSATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0565"]}},
    "OFFENSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0568"]}},
    "OFFENSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0570"]}},
    "OFFENSVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o9987"]}},
    "OFFENSVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["o9987"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["99696"]},
    },
    "OFFERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0573"]}},
    "OFFICINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["o0578"]}},
    "OFFICINATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["o0579"]}},
    "OFFICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["o0581"]}},
    "OFFICIOSE": {"-": {"morpho": ["rp--------"], "uri": ["46888"]}},
    "OFFICIOSVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["o0583"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["o0583"]},
    },
    "OFFICIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0585"]}},
    "OFFIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0586"]}},
    "OFFIRMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0588"]}},
    "OFFOCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0590"]}},
    "OFFVNDO": {"2": {"morpho": ["v1spia--3-"], "uri": ["o0597"]}},
    "OH": {"-": {"morpho": ["----------"], "uri": []}},
    "OHE": {"-": {"morpho": ["----------"], "uri": []}},
    "OLEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["o0617"]}},
    "OLEAGINEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0618"]}},
    "OLEAGINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0619"]}},
    "OLEARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["o0067"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["o0043"]},
    },
    "OLEASTER": {"-": {"morpho": ["n-s---mn2r"], "uri": ["42304"]}},
    "OLEITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0627"]}},
    "OLENS": {"-": {"morpho": ["aps---an3i"], "uri": ["99705"]}},
    "OLEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["o0637"]}},
    "OLESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0643"]}},
    "OLETVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["o0652"]},
        "2": {"morpho": ["n-s---nn2-"], "uri": ["o0652"]},
    },
    "OLEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0617"]}},
    "OLFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["o0648"]}},
    "OLFACTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["o0653"]}},
    "OLIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0654"]}},
    "OLIM": {"-": {"morpho": ["rp--------"], "uri": ["o0656"]}},
    "OLITIO": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "OLITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["47511"]}},
    "OLITORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["35259"]}},
    "OLIVA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["o0148"]}},
    "OLIVETVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0665"]}},
    "OLIVVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0184"]}},
    "OLOR": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["o0682"]},
        "2": {"morpho": ["n-s---mn3-"], "uri": ["o0682"]},
    },
    "OLORINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0684"]}},
    "OLVS": {"-": {"morpho": ["n-s---nn3-"], "uri": ["49356"]}},
    "OLVSCVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["119666"]}},
    "OLYMPICVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "OLYMPIVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["99712"]}},
    "OLYMPVS": {
        "1N": {"morpho": ["n-s----n2-"], "uri": []},
        "2N": {"morpho": ["n-s----n2-"], "uri": []},
        "3N": {"morpho": ["n-s----n2-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "OMASVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0688"]}},
    "OMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["o0690"]}},
    "OMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0692"]}},
    "OMINOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["o0696"]}},
    "OMITTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0699"]}},
    "OMNIGENVS": {
        "1": {"morpho": ["aps---mn1-"], "uri": ["o0709"]},
        "2": {"morpho": ["rp--------"], "uri": ["99715"]},
    },
    "OMNIMODIS": {"-": {"morpho": ["rp--------"], "uri": ["28413"]}},
    "OMNIMODVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0713"]}},
    "OMNINO": {"-": {"morpho": ["rp--------"], "uri": ["o0714"]}},
    "OMNIPARENS": {"-": {"morpho": ["aps---an3i"], "uri": ["o0716"]}},
    "OMNIPATENS": {
        "-": {"morpho": ["aps---an3-"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "OMNIPOTENS": {
        "-": {"morpho": ["aps---an3i"], "uri": ["o0721"]},
        "A": {"morpho": ["aps---an3i"], "uri": ["o0721"]},
    },
    "OMNIS": {"-": {"morpho": ["----------"], "uri": []}},
    "OMNITVENS": {"-": {"morpho": ["aps---an3i"], "uri": ["o0728"]}},
    "OMNIVOLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0731"]}},
    "ONAGER": {"-": {"morpho": ["n-s---mn2r"], "uri": ["o0742"]}},
    "ONASVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ONERARIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["o0746"]}},
    "ONERARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0746"]}},
    "ONERATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["99722"]}},
    "ONERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0747"]}},
    "ONEROSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0749"]}},
    "ONVS": {"-": {"morpho": ["n-s---nn3-"], "uri": ["o0768"]}},
    "ONVSTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0770"]}},
    "ONYX": {"-": {"morpho": ["n-s---cn3-"], "uri": ["o0775"]}},
    "OPACITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0776"]}},
    "OPACO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0777"]}},
    "OPACVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "OPACVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0778"]}},
    "OPE": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "OPELLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["o0781"]}},
    "OPERA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["o0783"]}},
    "OPERANS": {"-": {"morpho": ["aps---an3i"], "uri": ["119747"]}},
    "OPERARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["o0784"]},
        "2": {"morpho": ["n-s---mn2-"], "uri": ["o0784"]},
    },
    "OPERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0785"]}},
    "OPERATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["99723"]}},
    "OPERCVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0789"]}},
    "OPERIMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0791"]}},
    "OPERIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["o0792"]}},
    "OPEROR": {"-": {"morpho": ["v1spid--1-"], "uri": ["o0793"]}},
    "OPEROSE": {"-": {"morpho": ["rp--------"], "uri": ["30442"]}},
    "OPEROSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0795"]}},
    "OPERTORIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0799"]}},
    "OPERTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o9712"]}},
    "OPERTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["o9712"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["41510"]},
    },
    "OPICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o9737"]}},
    "OPIFEX": {"-": {"morpho": ["n-s---cn3-"], "uri": ["o0815"]}},
    "OPILIO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["o0819"]}},
    "OPIMIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OPIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0822"]}},
    "OPINATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["o9713"]}},
    "OPINIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0828"]}},
    "OPINOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["o0831"]}},
    "OPITER": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "OPITVLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["o0843"]}},
    "OPOBALSAMVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0848"]}},
    "OPORTET": {"-": {"morpho": ["v3spia--2-"], "uri": ["o0853"]}},
    "OPPEDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0858"]}},
    "OPPERIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["o0859"]}},
    "OPPETO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0861"]}},
    "OPPIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "OPPICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0863"]}},
    "OPPIDANI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["99734"]}},
    "OPPIDANVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["o0865"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["o0865"]},
    },
    "OPPIDO": {"-": {"morpho": ["rp--------"], "uri": ["o0867"]}},
    "OPPIDVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0868"]}},
    "OPPIDVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0869"]}},
    "OPPIGNERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0871"]}},
    "OPPILO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0874", "o0862"]}},
    "OPPLEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["o0875"]}},
    "OPPONO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0877"]}},
    "OPPORTVNE": {"-": {"morpho": ["rp--------"], "uri": ["25589"]}},
    "OPPORTVNITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0878"]}},
    "OPPORTVNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0879"]}},
    "OPPOSITVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["o9716"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["52307"]},
    },
    "OPPRESSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0881"]}},
    "OPPRESSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["o9717"]}},
    "OPPRIMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0884"]}},
    "OPPROBRIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0888"]}},
    "OPPVGNATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0890"]}},
    "OPPVGNATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["o0891"]}},
    "OPPVGNATORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0891"]}},
    "OPPVGNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0894"]}},
    "OPS": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["o0897"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "OPTABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["o0898"]}},
    "OPTATVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o9723"]}},
    "OPTATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["25601"]}},
    "OPTICE": {"-": {"morpho": ["n-s---fn1g"], "uri": ["o0902"]}},
    "OPTICOS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "OPTICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["36577"]}},
    "OPTIMAS": {"-": {"morpho": ["aps---an3-"], "uri": ["45999"]}},
    "OPTIMATES": {"-": {"morpho": ["n-p---mn3i"], "uri": ["31475"]}},
    "OPTIO": {
        "1": {"morpho": ["n-s---fn3-"], "uri": ["o0906"]},
        "2": {"morpho": ["n-s---fn3-"], "uri": ["o0906"]},
    },
    "OPTIONATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["o0907"]}},
    "OPTIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0908"]}},
    "OPTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0909"]}},
    "OPVLENS": {"-": {"morpho": ["aps---an3i"], "uri": ["o0914"]}},
    "OPVLENTER": {"-": {"morpho": ["rp--------"], "uri": ["36702"]}},
    "OPVLENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["o0911"]}},
    "OPVLENTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0913"]}},
    "OPVLENTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0914"]}},
    "OPVNTIVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "OPVS": {
        "1": {"morpho": ["n-s---fn3i"], "uri": ["o0918"]},
        "2": {"morpho": ["n-s---fn3i"], "uri": ["o0918"]},
    },
    "OPVSCVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0919"]}},
    "ORA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["o0923"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["o0923"]},
    },
    "ORACVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0925"]}},
    "ORATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0929"]}},
    "ORATIVNCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["o0931"]}},
    "ORATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["o0932"]}},
    "ORATORIE": {"-": {"morpho": ["rp--------"], "uri": ["45976"]}},
    "ORATORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0932"]}},
    "ORATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["o9718"]}},
    "ORBATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0936"]}},
    "ORBICVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["o0942"]}},
    "ORBIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["o0945"]}},
    "ORBITA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["o0946"]}},
    "ORBITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0947"]}},
    "ORBO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0951"]}},
    "ORBVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0952"]}},
    "ORCA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["o0953"]}},
    "ORCHESTRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["o0957"]}},
    "ORCHITIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["o0961"]}},
    "ORCHOMENOS": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "ORCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["o0966"]}},
    "ORCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["o9739"]}},
    "ORDINARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0971"]}},
    "ORDINATE": {"-": {"morpho": ["rp--------"], "uri": ["25619"]}},
    "ORDINATIM": {"-": {"morpho": ["rp--------"], "uri": ["o0972"]}},
    "ORDINATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0973"]}},
    "ORDINATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["o0975"]}},
    "ORDINATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["50618"]}},
    "ORDINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0977"]}},
    "ORDIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["o0978"]}},
    "ORDO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["o0979"]}},
    "OREAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["o0981"]}},
    "OREXIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["o0985"]}},
    "ORGANICOS": {"-": {"morpho": ["rp--------"], "uri": []}},
    "ORGANICVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["o0986"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["o0986"]},
    },
    "ORGANVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0988"]}},
    "ORGIA": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "ORIENS": {
        "N": {"morpho": ["n-s----n3-"], "uri": []},
        "-": {"morpho": ["aps---an3-"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "ORIFICIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0996"]}},
    "ORIGANITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0997"]}},
    "ORIGO": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["o1004"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "ORIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["o1006"]}},
    "ORIVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["48647"]}},
    "ORNAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o1012"]}},
    "ORNATE": {"-": {"morpho": ["rp--------"], "uri": ["57519"]}},
    "ORNATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o1013"]}},
    "ORNATRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o1016"]}},
    "ORNATVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["o9719"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["40146"]},
    },
    "ORNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o1025"]}},
    "ORNVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["o1026"]}},
    "ORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o1027"]}},
    "ORPHEVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": ["99755"]},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "ORSA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["o0962"]}},
    "ORTHOGONIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o1048"]}},
    "ORTHOGRAPHIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["o1050"]}},
    "ORTHOS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "ORTHOSTATA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["o1057"]}},
    "ORTONA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ORTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["o9720"]}},
    "ORYX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["o1062"]}},
    "ORYZA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["o1063"]}},
    "OS": {
        "1": {"morpho": ["n-s---nn3-"], "uri": ["o1065"]},
        "2": {"morpho": ["n-s---nn3-"], "uri": ["o1065"]},
    },
    "OSCEN": {"-": {"morpho": ["n-s---mn3-"], "uri": ["o1067"]}},
    "OSCILLVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["o1071"]},
        "2": {"morpho": ["n-s---nn2-"], "uri": ["o1071"]},
    },
    "OSCITANS": {"-": {"morpho": ["aps---an3i"], "uri": ["99759"]}},
    "OSCITATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o1074"]}},
    "OSCITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o1075"]}},
    "OSCITOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132789"]}},
    "OSCVLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o1077"]}},
    "OSCVLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["o1079"]}},
    "OSCVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o1080"]}},
    "OSSEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o1086"]}},
    "OSSICVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o1089"]}},
    "OSSIFRAGA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["o1090"]}},
    "OSSIFRAGVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["119996"]}},
    "OSTENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o1095"]}},
    "OSTENTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o1104"]}},
    "OSTENTATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["o1105"]}},
    "OSTENTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o1108"]}},
    "OSTENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o9721"]}},
    "OSTENTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["o9721"]}},
    "OSTIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "OSTIARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o1110"]}},
    "OSTIARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["o1110"]}},
    "OSTIATIM": {"-": {"morpho": ["rp--------"], "uri": ["o1111"]}},
    "OSTIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o1115"]}},
    "OSTREA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["o0576"]}},
    "OSTREOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o1126"]}},
    "OSTREVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0604"]}},
    "OSTRIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["o1129"]}},
    "OSTRINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o1130"]}},
    "OSTRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o1132"]}},
    "OTIOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["o1145"]}},
    "OTIOSE": {"-": {"morpho": ["rp--------"], "uri": ["36115"]}},
    "OTIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o1147"]}},
    "OTIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o1149"]}},
    "OVATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o1153"]}},
    "OVILE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["o1160"]}},
    "OVILLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0642"]}},
    "OVIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["o1166"]}},
    "OVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o1167"]}},
    "OVVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o1168"]}},
    "OXYMELI": {"-": {"morpho": ["n-s---nn3-"], "uri": ["o1182"]}},
    "PABVLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p0006"]}},
    "PABVLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p0007"]}},
    "PABVLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["p0009"]}},
    "PABVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p0011"]}},
    "PACATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p0014"]}},
    "PACATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["25660"]}},
    "PACHYNVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "PACIFICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p0019"]}},
    "PACILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PACISCOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["p0021"]}},
    "PACO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p0022"]}},
    "PACTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p0026"]}},
    "PACTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p9863"]}},
    "PACTVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["p9863"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["58353"]},
    },
    "PAEAN": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["p0060"]},
        "N": {"morpho": ["n-s---mn3-"], "uri": ["p0060"]},
    },
    "PAEDAGOGIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p0034"]}},
    "PAEDAGOGVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p0008"]}},
    "PAEDICO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["p0042"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["p0041"]},
    },
    "PAEDOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p0045"]}},
    "PAENE": {"-": {"morpho": ["rp--------"], "uri": ["p0049"]}},
    "PAENITENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p9827"]}},
    "PAENITEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p0052"]}},
    "PAENVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p0018"]}},
    "PAENVLATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0056"]}},
    "PAETVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["p0064"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "PAGANVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["p0067"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["p0067"]},
    },
    "PAGINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p0071"]}},
    "PAGVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p9075"]}},
    "PALA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p4062"]}},
    "PALAESTRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p0081"]}},
    "PALAESTRICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0082"]}},
    "PALAESTRITA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["p0083"]}},
    "PALAM": {
        "1": {"morpho": ["rp--------"], "uri": ["25678"]},
        "2": {"morpho": ["p---------"], "uri": ["p0086"]},
    },
    "PALATINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "PALATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p0092"]}},
    "PALATIVM": {
        "-": {"morpho": ["n-s---nn2-"], "uri": ["45678"]},
        "N": {"morpho": ["n-s---nn3-"], "uri": []},
    },
    "PALATVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p0093"]}},
    "PALEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p0095"]}},
    "PALEAR": {"-": {"morpho": ["n-s---nn3i"], "uri": ["p0097"]}},
    "PALICANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PALIMPSESTVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["25681"]}},
    "PALIVRVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p0110"]}},
    "PALLA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["p0111"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "PALLADIVM": {
        "-": {"morpho": ["n-s----n2-"], "uri": []},
        "N": {"morpho": ["n-s---nn2-"], "uri": []},
    },
    "PALLANTEVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "PALLANTIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "PALLAS": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["25684"]},
        "O": {"morpho": ["n-s---fn3-"], "uri": ["25684"]},
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "PALLENS": {"-": {"morpho": ["aps---an3i"], "uri": ["99779"]}},
    "PALLEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p0114"]}},
    "PALLESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p0114"]}},
    "PALLIATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p9118"]}},
    "PALLIDVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0116"]}},
    "PALLIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0117"]}},
    "PALLIOLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p0121"]}},
    "PALLIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p0122"]}},
    "PALLOR": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["p0124"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "PALMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p0229"]}},
    "PALMES": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p0132"]}},
    "PALMETVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p0134"]}},
    "PALMEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0135"]}},
    "PALMIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["p0137"]}},
    "PALMOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0142"]}},
    "PALMVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p0143"]}},
    "PALMVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["p0230"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["p0230"]},
    },
    "PALOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["p0148"]}},
    "PALPEBRAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["99782"]}},
    "PALPITATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p0159"]}},
    "PALPITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p0160"]}},
    "PALPO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["p0162"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["p0161"]},
    },
    "PALPOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132620"]}},
    "PALPVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["99784"]}},
    "PALVDAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p0164"]}},
    "PALVDATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0165"]}},
    "PALVMBES": {"-": {"morpho": ["n-s---mn3i"], "uri": ["p0174"]}},
    "PALVMBVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p0174"]}},
    "PALVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["p0177"]},
        "2": {"morpho": ["n-s---fn3-"], "uri": ["p0178"]},
    },
    "PALVSTER": {"-": {"morpho": ["aps---mn3-"], "uri": ["p0180"]}},
    "PAMPHILVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PAMPHYLIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PAMPINEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0187"]}},
    "PAMPINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p0188"]}},
    "PAMPINVS": {"-": {"morpho": ["n-s---cn2-"], "uri": ["p0190"]}},
    "PAN": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "PANACEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p0192"]}},
    "PANAX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["52940"]}},
    "PANCHRESTVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "PANCRATION": {"-": {"morpho": ["n-s---nn2g"], "uri": ["99789"]}},
    "PANDATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p0208"]}},
    "PANDO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["p0214"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["p0215"]},
    },
    "PANDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0218"]}},
    "PANGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p0223"]}},
    "PANHORMITANVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "PANHORMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PANICVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p0227"]}},
    "PANIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["p0232"]}},
    "PANNICVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p0237"]}},
    "PANNOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0239"]}},
    "PANNVCIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["46755"]}},
    "PANNVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p0243"]}},
    "PANTHERA": {"1": {"morpho": ["n-s---fn1-"], "uri": ["p7100"]}},
    "PANTOMIMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p0259"]}},
    "PANTOMIMICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0260"]}},
    "PANTOMIMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p0259"]}},
    "PANVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p0232"]}},
    "PAPAE": {"-": {"morpho": ["----------"], "uri": []}},
    "PAPAS": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p0267"]}},
    "PAPAVER": {"-": {"morpho": ["n-s---nn3-"], "uri": ["p0268"]}},
    "PAPHLAGONIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PAPILLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p0274"]}},
    "PAPINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PAPIRIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "PAPPO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p0276"]}},
    "PAPPVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p0277"]}},
    "PAPVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p0278"]}},
    "PAPYRVS": {"-": {"morpho": ["n-s---cn2-"], "uri": ["p0285"]}},
    "PAR": {
        "1": {"morpho": ["n-s---nn3i"], "uri": ["99797"]},
        "2": {"morpho": ["aps---an3i"], "uri": ["25714"]},
    },
    "PARABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p0287"]}},
    "PARABOLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p0290"]}},
    "PARADOXVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["52863"]}},
    "PARADROMIS": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "PARAENETICOS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "PARALVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PARALYSIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["p0325"]}},
    "PARAMESE": {"-": {"morpho": ["n-s---fn1g"], "uri": ["p0327"]}},
    "PARANETE": {"-": {"morpho": ["n-s---fn1g"], "uri": ["p0330"]}},
    "PARAPEGMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["p0334"]}},
    "PARARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p0343"]}},
    "PARASITA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p0352"]}},
    "PARASITVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p0352"]}},
    "PARASTAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p0353"]}},
    "PARASTATA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p0354"]}},
    "PARASTATICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p0355"]}},
    "PARATA": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "PARATE": {"-": {"morpho": ["rp--------"], "uri": ["54119"]}},
    "PARATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p0360"]}},
    "PARATVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["p9874"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["25725"]},
    },
    "PARCA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PARCE": {"-": {"morpho": ["rp--------"], "uri": ["45319"]}},
    "PARCENS": {"-": {"morpho": ["aps---an3i"], "uri": ["99805"]}},
    "PARCITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p0366"]}},
    "PARCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p0367"]}},
    "PARCVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0368"]}},
    "PARDALIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["p0371"]}},
    "PARDVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p0373"]}},
    "PARENS": {
        "1": {"morpho": ["n-s---cn3-"], "uri": []},
        "2": {
            "morpho": ["aps---an3i", "aps---an3i"],
            "uri": ["99808", "99808"],
        },
    },
    "PARENTES": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "PARENTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p0390"]}},
    "PAREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p0391"]}},
    "PARERGON": {"-": {"morpho": ["n-s---nn2g"], "uri": ["p0393"]}},
    "PARGO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "PARHYPATE": {"-": {"morpho": ["n-s---fn1g"], "uri": ["p0396"]}},
    "PARIES": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p0403"]}},
    "PARILIS": {
        "-": {"morpho": ["aps---cn3i"], "uri": ["p0407"]},
        "A": {"morpho": ["aps---cn3i"], "uri": ["p0407"]},
    },
    "PARIO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["p0410"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["p0410"]},
    },
    "PARITER": {"-": {"morpho": ["rp--------"], "uri": ["25734"]}},
    "PARMA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["p0414"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["p0414"]},
    },
    "PARMVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p0416"]}},
    "PARO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["120492"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["p4697"]},
    },
    "PAROCHVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p0421"]}},
    "PAROPSIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p0430"]}},
    "PAROTIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p0434"]}},
    "PARRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p0248"]}},
    "PARRICIDA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["p0437"]}},
    "PARRICIDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p0440"]}},
    "PARS": {"-": {"morpho": ["n-s---fn3i"], "uri": []}},
    "PARSIMONIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p0443"]}},
    "PARTHENICE": {"-": {"morpho": ["n-s---fn1g"], "uri": ["p0445"]}},
    "PARTHICVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "PARTHVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "PARTIARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0449"]}},
    "PARTICEPS": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["99820"]},
        "2": {"morpho": ["aps---an3-"], "uri": ["p0452"]},
    },
    "PARTICIPO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p0459"]}},
    "PARTICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p0460"]}},
    "PARTICVLARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p0461"]}},
    "PARTICVLATIM": {"-": {"morpho": ["rp--------"], "uri": ["p0462"]}},
    "PARTIM": {"-": {"morpho": ["rp--------"], "uri": ["p0466"]}},
    "PARTIO": {
        "1": {"morpho": ["n-s---fn3-"], "uri": ["p0467"]},
        "2": {"morpho": ["v1spia--4-"], "uri": ["p0468"]},
    },
    "PARTIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["132622"]}},
    "PARTITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p0471"]}},
    "PARTVRIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["p0477"]}},
    "PARTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": ["120553"]},
    },
    "PARVE": {"-": {"morpho": ["rp--------"], "uri": ["120554"]}},
    "PARVICAX": {
        "-": {"morpho": ["aps---an3-"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "PARVM": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["rp--------"], "uri": ["p0484"]},
    },
    "PARVMPER": {"-": {"morpho": ["rp--------"], "uri": ["p0486"]}},
    "PARVVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0489"]}},
    "PARVVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["99823"]}},
    "PARVVS": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": ["p0490"]},
    },
    "PASCALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p0492"]}},
    "PASCEOLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p0493"]}},
    "PASCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p0497"]}},
    "PASCOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["133049"]}},
    "PASCVVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p0500"]}},
    "PASCVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0500"]}},
    "PASSER": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p0502"]}},
    "PASSIM": {"-": {"morpho": ["rp--------"], "uri": ["p0510"]}},
    "PASSVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p9877"]}},
    "PASSVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["p9877"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["99826"]},
    },
    "PASTAS": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "PASTILLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p0525"]}},
    "PASTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p0531"]}},
    "PASTOR": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["p0534"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "PASTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["p0498"]}},
    "PATAVIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "PATEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["p0543"]}},
    "PATEFIO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p0544"]}},
    "PATELLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p0546"]}},
    "PATENS": {"-": {"morpho": ["aps---an3i"], "uri": ["99831"]}},
    "PATEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p0549"]}},
    "PATER": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["p0550"]},
        "N": {"morpho": ["n-s----n2-"], "uri": []},
    },
    "PATERA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p0551"]}},
    "PATERNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0554"]}},
    "PATESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p9947"]}},
    "PATHICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0557"]}},
    "PATIBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p0560"]}},
    "PATIBVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p0562"]}},
    "PATIENS": {"-": {"morpho": ["aps---an3i"], "uri": ["99832"]}},
    "PATIENTER": {"-": {"morpho": ["rp--------"], "uri": ["25773"]}},
    "PATIENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p9830"]}},
    "PATINA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["p0564"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["p0564"]},
    },
    "PATIOR": {"-": {"morpho": ["v1spid--3i"], "uri": ["p0566"]}},
    "PATRATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p0570"]}},
    "PATRIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["p0588"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "PATRICIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["p0578"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["p0578"]},
    },
    "PATRIMONIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p0582"]}},
    "PATRIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0588"]}},
    "PATRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p0589"]}},
    "PATROCINIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p0591"]}},
    "PATROCINOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["p0592"]}},
    "PATRONA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p0593"]}},
    "PATRONVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p0593"]}},
    "PATRVELIS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["99837"]},
        "2": {"morpho": ["aps---cn3i"], "uri": ["p0597"]},
    },
    "PATRVVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["p0598"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["p0598"]},
    },
    "PATVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0599"]}},
    "PAVA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p0626"]}},
    "PAVCA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["p0605"]}},
    "PAVCI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["p0605"]}},
    "PAVCITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p0603"]}},
    "PAVCVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0604"]}},
    "PAVCVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0605"]}},
    "PAVEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["p0606"]}},
    "PAVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p0607"]}},
    "PAVESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p0607"]}},
    "PAVICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p0609"]}},
    "PAVIDE": {"-": {"morpho": ["rp--------"], "uri": ["33397"]}},
    "PAVIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0611"]}},
    "PAVIMENTATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0613"]}},
    "PAVIMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p0615"]}},
    "PAVIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["p0616"]}},
    "PAVITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p0619"]}},
    "PAVLATIM": {
        "-": {
            "morpho": ["rp--------", "rp--------"],
            "uri": ["25783", "25783"],
        }
    },
    "PAVLISPER": {"-": {"morpho": ["rp--------"], "uri": ["57821"]}},
    "PAVLO": {
        "-": {
            "morpho": ["rp--------", "rp--------"],
            "uri": ["25784", "25784"],
        }
    },
    "PAVLVLVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["99841"]},
        "2": {"morpho": ["rp--------"], "uri": ["55567"]},
    },
    "PAVLVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["41363"]}},
    "PAVLVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["99842"]},
        "2": {"morpho": ["rp--------"], "uri": ["25785"]},
    },
    "PAVLVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["25786"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "PAVO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p0626"]}},
    "PAVOR": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["25795"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "PAVPER": {"-": {"morpho": ["aps---an3-"], "uri": ["p0630"]}},
    "PAVPERCVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0369"]}},
    "PAVPERIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["p0633"]}},
    "PAVPERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p0635"]}},
    "PAVPERTAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p0636"]}},
    "PAVSA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p0712"]}},
    "PAVSARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p0642"]}},
    "PAVSEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["43241"]}},
    "PAVVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p0626"]}},
    "PAVXILLVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["p0648"]},
        "2": {"morpho": ["n-s---nn2-"], "uri": ["p0648"]},
    },
    "PAVXILLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0649"]}},
    "PAX": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["p0650"]},
        "N": {"morpho": ["n-s---mn3-"], "uri": []},
    },
    "PECCANS": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "PECCATVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p9878"]}},
    "PECCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p0659"]}},
    "PECTEN": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p0664"]}},
    "PECTINATIM": {"-": {"morpho": ["rp--------"], "uri": ["p0666"]}},
    "PECTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p0670"]}},
    "PECTVS": {"-": {"morpho": ["n-s---nn3-"], "uri": ["p0674"]}},
    "PECV": {"-": {"morpho": ["n-s---nn4-"], "uri": ["p0677"]}},
    "PECVARIA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["p9955"]}},
    "PECVARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["p9954"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["p0679"]},
    },
    "PECVLATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["p9879"]}},
    "PECVLIARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p0686"]}},
    "PECVLIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p0691"]}},
    "PECVNIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["p0693"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "PECVNIARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["120724"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["p0695"]},
    },
    "PECVNIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0696"]}},
    "PECVS": {
        "1": {"morpho": ["n-s---nn3-"], "uri": ["p0698"]},
        "2": {"morpho": ["n-s---nn3-"], "uri": ["p0698"]},
    },
    "PEDALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p0701"]}},
    "PEDARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0705"]}},
    "PEDATVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "PEDATVS": {"1": {"morpho": ["n-s---mn4-"], "uri": []}},
    "PEDES": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p0708"]}},
    "PEDESTER": {"-": {"morpho": ["aps---mn3-"], "uri": ["p0709"]}},
    "PEDETENTIM": {"-": {"morpho": ["rp--------"], "uri": ["120737"]}},
    "PEDICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p0711"]}},
    "PEDICINVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["99859"]}},
    "PEDICVLVS": {"1": {"morpho": ["n-s---mn2-"], "uri": ["p4329"]}},
    "PEDIO": {"-": {"morpho": ["v1spia--4-"], "uri": []}},
    "PEDITATVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["p9724"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["p9724"]},
    },
    "PEDITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["120751"]}},
    "PEDITVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p0687"]}},
    "PEDO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["120752"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["p0724"]},
        "3": {"morpho": ["v1spia--3-"], "uri": ["p0725"]},
        "N": {"morpho": ["n-s---mn3-"], "uri": ["120752"]},
    },
    "PEDVCAEANVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "PEDVCAEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PEDVM": {
        "-": {"morpho": ["n-s---nn2-"], "uri": ["p0730"]},
        "N": {"morpho": ["n-s---nn2-"], "uri": ["p0730"]},
    },
    "PEGMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["p0732"]}},
    "PEIERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p0736"]}},
    "PELAGVS": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p0744"]}},
    "PELAMIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["44554"]}},
    "PELASTES": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "PELECINVS": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "PELLACIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p0749"]}},
    "PELLAX": {"-": {"morpho": ["aps---an3i"], "uri": ["p0751"]}},
    "PELLEX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["39430"]}},
    "PELLICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["p0757"]}},
    "PELLICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p0760"]}},
    "PELLIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["p0767"]}},
    "PELLITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0768"]}},
    "PELLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p0769"]}},
    "PELLVCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["99872"]}},
    "PELLVCIDVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["120794"]}},
    "PELLVCIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["120795"]}},
    "PELOPONNESVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PELORIS": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["p0772"]},
        "N": {"morpho": ["n-s---fn3-"], "uri": ["p0772"]},
    },
    "PELTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p0773"]}},
    "PELTATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0775"]}},
    "PELVIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["p0778"]}},
    "PELVSIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "PENARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0781"]}},
    "PENATES": {
        "-": {"morpho": ["n-s----n3-"], "uri": []},
        "N": {"morpho": ["n-s---mn1g"], "uri": []},
    },
    "PENATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p0784"]}},
    "PENDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p0785"]}},
    "PENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p0787"]}},
    "PENDVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0788"]}},
    "PENES": {"-": {"morpho": ["p---------"], "uri": ["p0790"]}},
    "PENETRABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p0792"]}},
    "PENETRALE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["p0793"]}},
    "PENETRALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p0793"]}},
    "PENETRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p0796"]}},
    "PENIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["p0800"]}},
    "PENITE": {"-": {"morpho": ["rp--------"], "uri": ["32839"]}},
    "PENITVS": {
        "1": {"morpho": ["aps---mn1-"], "uri": ["p9801"]},
        "2": {"morpho": ["rp--------"], "uri": ["p0801"]},
    },
    "PENNA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p0802"]}},
    "PENNATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0804"]}},
    "PENNIPES": {"-": {"morpho": ["aps---an3-"], "uri": ["48546"]}},
    "PENNIPOTENTES": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "PENSILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p0815"]}},
    "PENSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p0816"]}},
    "PENSITATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p0817"]}},
    "PENSITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p0819"]}},
    "PENSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p0821"]}},
    "PENSVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p0685"]}},
    "PENSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["99880"]}},
    "PENTADOROS": {
        "1": {"morpho": ["aps---mn1g"], "uri": ["p9826"]},
        "2": {"morpho": ["rp--------"], "uri": []},
    },
    "PENTASPASTON": {"-": {"morpho": ["n-s---nn2g"], "uri": ["p0839"]}},
    "PENTEMOEROS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "PENTEREMIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["p0850"]}},
    "PENTERES": {"-": {"morpho": ["n-s---fn3i"], "uri": ["99881"]}},
    "PENVRIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p0059"]}},
    "PENVS": {"-": {"morpho": ["n-s---cn4-"], "uri": ["p0858"]}},
    "PEPLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p0861"]}},
    "PER": {"-": {"morpho": ["p---------"], "uri": ["p0865"]}},
    "PERA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p0866"]}},
    "PERACVTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["38428"]}},
    "PERADOLESCENS": {
        "-": {"morpho": ["aps---an3-"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "PERAEQVE": {"-": {"morpho": ["rp--------"], "uri": ["p0844"]}},
    "PERAEQVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p0879"]}},
    "PERAGITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p0881"]}},
    "PERAGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p0882"]}},
    "PERAGRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p0885"]}},
    "PERAMBVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p0889"]}},
    "PERAMOENVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0892"]}},
    "PERAMPLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0893"]}},
    "PERANGVSTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0895"]}},
    "PERANTIQVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0896"]}},
    "PERARDVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0899"]}},
    "PERARIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0902"]}},
    "PERARMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p0904"]}},
    "PERARO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p0905"]}},
    "PERATTENTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0910"]}},
    "PERBIBO": {"2": {"morpho": ["v1spia--3-"], "uri": ["p0921"]}},
    "PERBONVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0924"]}},
    "PERBREVI": {"-": {"morpho": ["rp--------"], "uri": ["99890"]}},
    "PERBREVIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p0925"]}},
    "PERCALEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["p0928"]}},
    "PERCALESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p0930"]}},
    "PERCANDEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["p0933"]}},
    "PERCARVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0936"]}},
    "PERCELEBRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p0939"]}},
    "PERCELLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p0941"]}},
    "PERCENNIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PERCENSEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p0942"]}},
    "PERCEPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p0945"]}},
    "PERCIDO": {"2": {"morpho": ["v1spia--3-"], "uri": ["p0948"]}},
    "PERCIEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p0949"]}},
    "PERCIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["43567"]}},
    "PERCIPIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["p0951"]}},
    "PERCITO": {"1": {"morpho": ["v1spia--1-"], "uri": ["p0955"]}},
    "PERCOLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p0965"]}},
    "PERCOLO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["p0967"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["p0966"]},
    },
    "PERCOMMODE": {"-": {"morpho": ["rp--------"], "uri": ["42619"]}},
    "PERCOMPERIO": {"-": {"morpho": ["v1spia--4-"], "uri": []}},
    "PERCONTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p0973"]}},
    "PERCONTATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p0975"]}},
    "PERCOQVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p0980"]}},
    "PERCREBRESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p0983"]}},
    "PERCREPO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p0984"]}},
    "PERCRESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["120981"]}},
    "PERCRVDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0987"]}},
    "PERCVLSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["p0952"]}},
    "PERCVNCTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["99901"]}},
    "PERCVNCTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["120993"]}},
    "PERCVNCTOR": {"-": {"morpho": ["v1spid--1-"], "uri": []}},
    "PERCVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p0994"]}},
    "PERCVRRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p0995"]}},
    "PERCVRSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p0998"]}},
    "PERCVSSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p1000"]}},
    "PERCVSSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p1002"]}},
    "PERCVSSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["p1005"]}},
    "PERCVTIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["p1003"]}},
    "PERDELIRVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1009"]}},
    "PERDEPSO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1011"]}},
    "PERDIFFICILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p1017"]}},
    "PERDISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1020"]}},
    "PERDITE": {"-": {"morpho": ["rp--------"], "uri": ["25855"]}},
    "PERDITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p1024"]}},
    "PERDITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["25857"]}},
    "PERDIVES": {"-": {"morpho": ["aps---an3-"], "uri": ["p1027"]}},
    "PERDIVTVRNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1030"]}},
    "PERDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1032"]}},
    "PERDOCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p1033"]}},
    "PERDOLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1036"]}},
    "PERDOMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1039"]}},
    "PERDVCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1041"]}},
    "PERDVCTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p1042"]}},
    "PERDVCTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p1044"]}},
    "PERDVELLIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p1046"]}},
    "PERDVELLIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["p1047"]}},
    "PERDVLCIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p1049"]}},
    "PERDVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1050"]}},
    "PEREDO": {
        "1": {"morpho": ["v1spia--3-"], "uri": ["p1052", "p5000", "p1052"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["p1052", "p5000", "p1052"]},
    },
    "PEREGRE": {"-": {"morpho": ["rp--------"], "uri": ["32282"]}},
    "PEREGRINATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p1059"]}},
    "PEREGRINOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["p1062"]}},
    "PEREGRINVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["p1063"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["p1063"]},
    },
    "PEREMPTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p1069"]}},
    "PERENDINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1074"]}},
    "PERENNE": {"-": {"morpho": ["rp--------"], "uri": ["99912"]}},
    "PERENNIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p1075"]}},
    "PERENNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1079"]}},
    "PEREO": {"1": {"morpho": ["v1spia--4-"], "uri": ["p1081"]}},
    "PEREQVITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1082"]}},
    "PERERRO": {"2": {"morpho": ["v1spia--1-"], "uri": ["p1083"]}},
    "PEREXCELSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1085"]}},
    "PEREXIGVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1087"]}},
    "PERFACETE": {"-": {"morpho": ["rp--------"], "uri": ["41434"]}},
    "PERFACILE": {"-": {"morpho": ["rp--------"], "uri": ["33225"]}},
    "PERFACILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p1095"]}},
    "PERFAMILIARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p1099"]}},
    "PERFECTE": {"-": {"morpho": ["rp--------"], "uri": ["29378"]}},
    "PERFECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p1100"]}},
    "PERFECTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["p9881"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["25866"]},
    },
    "PERFERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1106"]}},
    "PERFICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["p1111"]}},
    "PERFICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1112"]}},
    "PERFIDE": {"-": {"morpho": ["rp--------"], "uri": ["99918"]}},
    "PERFIDIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["p1115"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "PERFIDIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1116"]}},
    "PERFIDVM": {"-": {"morpho": ["rp--------"], "uri": ["99919"]}},
    "PERFIDVS": {
        "1": {"morpho": ["aps---mn1-"], "uri": ["p1117"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["p1117"]},
    },
    "PERFIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1118"]}},
    "PERFLATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["p9882"]}},
    "PERFLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1127"]}},
    "PERFLVCTVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1129"]}},
    "PERFLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1130"]}},
    "PERFODIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["p1132"]}},
    "PERFORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1139"]}},
    "PERFRICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1149"]}},
    "PERFRIGESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1154"]}},
    "PERFRIGIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1155"]}},
    "PERFRINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1156"]}},
    "PERFRVOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["p1161"]}},
    "PERFVGA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["p1162"]}},
    "PERFVGIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["p1163"]}},
    "PERFVGIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p1164"]}},
    "PERFVNCTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p1167"]}},
    "PERFVNDO": {"2": {"morpho": ["v1spia--3-"], "uri": ["p1169"]}},
    "PERFVNGOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["p1170"]}},
    "PERFVRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1171"]}},
    "PERFVSORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1173"]}},
    "PERGA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PERGAEVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "PERGAMVS": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "PERGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1180"]}},
    "PERGRANDIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p1184"]}},
    "PERGRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1186"]}},
    "PERGVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p1188"]}},
    "PERHIBEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p1191"]}},
    "PERHILVM": {"-": {"morpho": ["rp--------"], "uri": ["p1193"]}},
    "PERHONORIFICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1195"]}},
    "PERHORRESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1196"]}},
    "PERIACTOS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "PERICLES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "PERICLITOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["p1210"]}},
    "PERICVLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["p1213"]}},
    "PERICVLOSE": {"-": {"morpho": ["rp--------"], "uri": ["53132"]}},
    "PERICVLOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1214"]}},
    "PERICVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p1215"]}},
    "PERIDONEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1216"]}},
    "PERIECHON": {"-": {"morpho": ["----------"], "uri": []}},
    "PERIMETROS": {"-": {"morpho": ["n-s---fn2g"], "uri": ["p1227"]}},
    "PERIMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1228"]}},
    "PERIMPEDITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1229"]}},
    "PERINDE": {"-": {"morpho": ["rp--------"], "uri": ["p1236"]}},
    "PERINGRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1245"]}},
    "PERINIQVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1246"]}},
    "PERINIVRIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1247"]}},
    "PERINTEMPESTIVVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["p1249"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["p1249"]},
    },
    "PERINVALIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1250"]}},
    "PERIPATETICI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PERIPATETICVS": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "PERIPETASMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["p1259"]}},
    "PERIPTEROS": {"-": {"morpho": ["aps---mn1g"], "uri": ["p1269"]}},
    "PERISCELIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p1271"]}},
    "PERISTYLIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p1283"]}},
    "PERISTYLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p1284"]}},
    "PERITE": {"-": {"morpho": ["rp--------"], "uri": ["29739"]}},
    "PERITHECIOS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "PERITIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p1285"]}},
    "PERITRETOS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "PERITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1257"]}},
    "PERIVRIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["25896"]}},
    "PERIVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["36742"]}},
    "PERIVRVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["49942"]}},
    "PERLABOR": {"2": {"morpho": ["v1spid--3-"], "uri": ["p1293"]}},
    "PERLATEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p1295"]}},
    "PERLEGO": {"2": {"morpho": ["v1spia--3-"], "uri": ["p1303"]}},
    "PERLIBRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p1307"]}},
    "PERLIBRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1308"]}},
    "PERLIMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1310"]}},
    "PERLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1325"]}},
    "PERLVSTRO": {"2": {"morpho": ["v1spia--1-"], "uri": ["p1327"]}},
    "PERMACERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1330"]}},
    "PERMADESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1332"]}},
    "PERMAGNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1335"]}},
    "PERMANANTER": {"-": {"morpho": ["rp--------"], "uri": ["99942"]}},
    "PERMANEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p1338"]}},
    "PERMANO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1339"]}},
    "PERMATVRESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1343"]}},
    "PERMEO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1355"]}},
    "PERMEREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p1356"]}},
    "PERMETIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["p1357"]}},
    "PERMINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1352"]}},
    "PERMISCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p1362"]}},
    "PERMISSVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["p9884"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["121296"]},
    },
    "PERMITIALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p1365"]}},
    "PERMITTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1367"]}},
    "PERMIXTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p1369"]}},
    "PERMIXTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["57761"]}},
    "PERMODESTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1370"]}},
    "PERMOLESTE": {"-": {"morpho": ["rp--------"], "uri": ["57190"]}},
    "PERMOLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1374"]}},
    "PERMOVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p1378"]}},
    "PERMVLCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p1379"]}},
    "PERMVLTO": {"-": {"morpho": ["rp--------"], "uri": ["37730"]}},
    "PERMVLTVM": {"-": {"morpho": ["rp--------"], "uri": ["46548"]}},
    "PERMVLTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1381"]}},
    "PERMVNIO": {"2": {"morpho": ["v1spia--4-"], "uri": ["p1384"]}},
    "PERMVTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p1387"]}},
    "PERMVTO": {"2": {"morpho": ["v1spia--1-"], "uri": ["p1388"]}},
    "PERNA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p1389"]}},
    "PERNEGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1395"]}},
    "PERNICIABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p1397"]}},
    "PERNICIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["p1399"]}},
    "PERNICIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1401"]}},
    "PERNICITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p1402"]}},
    "PERNICITER": {"-": {"morpho": ["rp--------"], "uri": ["29795"]}},
    "PERNIX": {"-": {"morpho": ["aps---an3i"], "uri": ["p1408"]}},
    "PERNOBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p1409"]}},
    "PERNOCTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1411"]}},
    "PERNOSCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1179"]}},
    "PERNOTESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1413"]}},
    "PERNOX": {"-": {"morpho": ["aps---an3-"], "uri": ["p1414"]}},
    "PERNVMERO": {"1": {"morpho": ["v1spia--1-"], "uri": ["p1417"]}},
    "PERO": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["p1419"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "PEROLEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p1427"]}},
    "PERONATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1428"]}},
    "PEROPPORTVNE": {"-": {"morpho": ["rp--------"], "uri": ["45041"]}},
    "PEROPPORTVNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1431"]}},
    "PERORNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1436"]}},
    "PERORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1437"]}},
    "PEROSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["55692"]}},
    "PERPARVVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1444"]}},
    "PERPARVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1445"]}},
    "PERPAVCI": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "PERPAVCVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1449"]}},
    "PERPELLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1455"]}},
    "PERPENDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["121364"]}},
    "PERPENDICVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p1459"]}},
    "PERPENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1460"]}},
    "PERPENSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1463"]}},
    "PERPERAM": {"-": {"morpho": ["rp--------"], "uri": ["p1464"]}},
    "PERPERE": {"-": {"morpho": ["rp--------"], "uri": ["121375"]}},
    "PERPERNA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PERPERVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1466"]}},
    "PERPES": {"-": {"morpho": ["aps---an3-"], "uri": ["p1467"]}},
    "PERPESSICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1468"]}},
    "PERPESSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p1469"]}},
    "PERPETIOR": {"-": {"morpho": ["v1spid--3i"], "uri": ["p1472"]}},
    "PERPETO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "PERPETRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1477"]}},
    "PERPETVE": {"-": {"morpho": ["rp--------"], "uri": ["42504"]}},
    "PERPETVITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p1481"]}},
    "PERPETVO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["p1483"]},
        "2": {"morpho": ["rp--------"], "uri": ["25920"]},
    },
    "PERPETVVM": {"-": {"morpho": ["rp--------"], "uri": ["99955"]}},
    "PERPETVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1484"]}},
    "PERPLEXVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1494"]}},
    "PERPLICATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1495"]}},
    "PERPLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1496"]}},
    "PERPOPVLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["p1502"]}},
    "PERPOTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1505"]}},
    "PERPRIMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1508"]}},
    "PERPVRGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1516"]}},
    "PERQVADRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1522"]}},
    "PERQVAM": {"-": {"morpho": ["rp--------"], "uri": ["p1523"]}},
    "PERQVIRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1525"]}},
    "PERRARO": {"-": {"morpho": ["rp--------"], "uri": ["30306"]}},
    "PERREPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1529"]}},
    "PERRIDICVLE": {"-": {"morpho": ["rp--------"], "uri": ["45674"]}},
    "PERROGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1536"]}},
    "PERRVMPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1537"]}},
    "PERSA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PERSAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "PERSAEPE": {"-": {"morpho": ["rp--------"], "uri": ["p1538"]}},
    "PERSALVTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1543"]}},
    "PERSANO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1545"]}},
    "PERSANVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1546"]}},
    "PERSCINDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1549"]}},
    "PERSCRIBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1552"]}},
    "PERSCRIPTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p1555"]}},
    "PERSCRVTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p1556"]}},
    "PERSCRVTOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["p1559"]}},
    "PERSECO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1562"]}},
    "PERSECTOR": {"2": {"morpho": ["v1spid--1-"], "uri": ["p1563"]}},
    "PERSEDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p1567"]}},
    "PERSENTIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["p1571"]}},
    "PERSENTISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["50541"]}},
    "PERSEQVOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["p1575"]}},
    "PERSERO": {
        "2": {"morpho": ["v1spia--3-"], "uri": ["p9933"]},
        "3": {"morpho": ["v1spia--3-"], "uri": ["p9933"]},
    },
    "PERSES": {
        "N": {"morpho": ["n-p---mn3-"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "PERSEVERANS": {"-": {"morpho": ["aps---an3i"], "uri": ["99962"]}},
    "PERSEVERANTER": {"-": {"morpho": ["rp--------"], "uri": ["58205"]}},
    "PERSEVERANTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p1572"]}},
    "PERSEVERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1581"]}},
    "PERSEVERVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1582"]}},
    "PERSIDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1586"]}},
    "PERSIMPLEX": {"-": {"morpho": ["aps---an3-"], "uri": ["p1590"]}},
    "PERSISTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1591"]}},
    "PERSOLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1596"]}},
    "PERSONA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p1597"]}},
    "PERSONATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0717"]}},
    "PERSONO": {"1": {"morpho": ["v1spia--1-"], "uri": ["p1605"]}},
    "PERSPECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1609"]}},
    "PERSPECTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["41679"]}},
    "PERSPERGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1613"]}},
    "PERSPICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["p1618"]}},
    "PERSPICVE": {"-": {"morpho": ["rp--------"], "uri": ["25936"]}},
    "PERSPICVITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p1619"]}},
    "PERSPICVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1620"]}},
    "PERSPIRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1621"]}},
    "PERSTERNO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1623"]}},
    "PERSTIMVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1624"]}},
    "PERSTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1627"]}},
    "PERSTREPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1628"]}},
    "PERSTRINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1631"]}},
    "PERSTRVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1632"]}},
    "PERSTVDIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1633"]}},
    "PERSVADEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p1634"]}},
    "PERSVASIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p1636"]}},
    "PERSVBTILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p1647"]}},
    "PERSVLTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1646"]}},
    "PERTAEDESCO": {"-": {"morpho": ["v1spia--3-"], "uri": [None, "p2137"]}},
    "PERTAEDET": {"-": {"morpho": ["v1spia--3-"], "uri": ["33351"]}},
    "PERTEGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1651"]}},
    "PERTENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1654"]}},
    "PERTENTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["25943"]}},
    "PERTENVIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p1655"]}},
    "PERTEREBRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1657"]}},
    "PERTERGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p1658"]}},
    "PERTERREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p1662"]}},
    "PERTERRICREPVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1663"]}},
    "PERTEXO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1665"]}},
    "PERTICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p1666"]}},
    "PERTIMEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p1671"]}},
    "PERTIMESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1671"]}},
    "PERTINACIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p1672"]}},
    "PERTINACITER": {"-": {"morpho": ["rp--------"], "uri": ["25947"]}},
    "PERTINAX": {"-": {"morpho": ["aps---an3i"], "uri": ["p1673"]}},
    "PERTINEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p1674"]}},
    "PERTINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1675"]}},
    "PERTIO": {"-": {"morpho": ["v1spia--4-"], "uri": []}},
    "PERTOLERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1676"]}},
    "PERTORQVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p1678"]}},
    "PERTRACTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1680"]}},
    "PERTRAHO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1681"]}},
    "PERTRITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["99979"]}},
    "PERTVNDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1691"]}},
    "PERTVRBATE": {"-": {"morpho": ["rp--------"], "uri": ["28707"]}},
    "PERTVRBATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p1692"]}},
    "PERTVRBATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["48276"]}},
    "PERTVRBO": {"2": {"morpho": ["v1spia--1-"], "uri": ["p1695"]}},
    "PERVADO": {"2": {"morpho": ["v1spia--3-"], "uri": ["p1697"]}},
    "PERVAGOR": {"2": {"morpho": ["v1spid--1-"], "uri": ["p1700"]}},
    "PERVAGVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1701"]}},
    "PERVALEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p1703"]}},
    "PERVASTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1709"]}},
    "PERVECTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p1711"]}},
    "PERVEHO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1712"]}},
    "PERVELLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1713"]}},
    "PERVENIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["p1714"]}},
    "PERVERSE": {"-": {"morpho": ["rp--------"], "uri": ["36413"]}},
    "PERVERSITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p1723"]}},
    "PERVERSVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["99984"]}},
    "PERVERSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["25959"]}},
    "PERVERTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1726"]}},
    "PERVESTIGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1730"]}},
    "PERVETVS": {"-": {"morpho": ["aps---an3-"], "uri": ["p1731"]}},
    "PERVICACIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p1734"]}},
    "PERVICACITER": {"-": {"morpho": ["rp--------"], "uri": ["47395"]}},
    "PERVICAX": {"-": {"morpho": ["aps---an3-"], "uri": ["p1735"]}},
    "PERVIDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p1737"]}},
    "PERVIGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p1738"]}},
    "PERVIGIL": {"-": {"morpho": ["aps---an3-"], "uri": ["p1742"]}},
    "PERVIGILIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p1742"]}},
    "PERVIGILO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1743"]}},
    "PERVINCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1745"]}},
    "PERVIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p1749"]}},
    "PERVIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1749"]}},
    "PERVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p1751"]}},
    "PERVNGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["45980"]}},
    "PERVOLITANTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p9836"]}},
    "PERVOLITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1754"]}},
    "PERVOLO": {
        "2": {"morpho": ["v1spia--1-"], "uri": ["p1755"]},
        "3": {"morpho": ["v1spia--1-"], "uri": ["p1755"]},
    },
    "PERVOLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1757"]}},
    "PERVRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1761"]}},
    "PERVSINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "PERVVLGO": {"1": {"morpho": ["v1spia--1-"], "uri": ["p1764"]}},
    "PES": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p1765"]}},
    "PESSVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["p1775"]},
        "2": {"morpho": ["rp--------"], "uri": ["p1774"]},
    },
    "PESSVMDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["25968"]}},
    "PESTIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["p1778"]}},
    "PESTILENS": {"-": {"morpho": ["aps---an3i"], "uri": ["p1779"]}},
    "PESTILENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p1780"]}},
    "PESTILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p1785"]}},
    "PESTIS": {
        "-": {"morpho": ["n-s---fn3i"], "uri": ["p1786"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "PETASVNCVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p4395"]}},
    "PETAVRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p1796"]}},
    "PETESSO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1799"]}},
    "PETILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PETIOR": {"-": {"morpho": ["v1spid--3i"], "uri": []}},
    "PETITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p1804"]}},
    "PETITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p1805"]}},
    "PETITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["p1776"]}},
    "PETO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1809"]}},
    "PETORITVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["44260"]}},
    "PETRA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["p1811"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["p1811"]},
    },
    "PETRINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PETRO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "PETVLANS": {"-": {"morpho": ["aps---an3i"], "uri": ["p1823"]}},
    "PETVLANTER": {"-": {"morpho": ["rp--------"], "uri": ["25980"]}},
    "PETVLANTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p1824"]}},
    "PETVLCVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1825"]}},
    "PEXVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["54675"]}},
    "PHAECASIATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1835"]}},
    "PHAECASIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p1834"]}},
    "PHALACRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PHALANGA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p1842"]}},
    "PHALANGARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p1843"]}},
    "PHALANX": {"-": {"morpho": ["n-s---fn3i"], "uri": ["p1847"]}},
    "PHALARIS": {"N": {"morpho": ["n-s---fn3-"], "uri": ["p1848"]}},
    "PHALERAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["p1849"]}},
    "PHARETRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p1862"]}},
    "PHARETRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1863"]}},
    "PHARMACOPOLA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["p1868"]}},
    "PHAROS": {
        "-": {"morpho": ["n-s----n-g"], "uri": []},
        "N": {"morpho": ["n-s----n-g"], "uri": []},
    },
    "PHARVS": {
        "-": {"morpho": ["n-s----n-g"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "PHASELIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "PHELLOS": {"-": {"morpho": ["n-s---mn2g"], "uri": ["p1885"]}},
    "PHIALA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p1892"]}},
    "PHILEMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["p1898"]}},
    "PHILINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PHILIPPVS": {
        "1N": {"morpho": ["n-s----n2-"], "uri": []},
        "2N": {"morpho": ["n-s----n2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "PHILO": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "PHILOCRATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PHILODAMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PHILOLOGIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p1906"]}},
    "PHILOLOGVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["p1907"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["p1907"]},
    },
    "PHILOMELIENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "PHILOMELIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "PHILOSOPHIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p1913"]}},
    "PHILOSOPHOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["p1915"]}},
    "PHILOSOPHVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["p2261"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["p2002"]},
    },
    "PHILOTECHNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["121755"]}},
    "PHILTRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p1921"]}},
    "PHILYRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p1922"]}},
    "PHIMES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PHIMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p1924"]}},
    "PHINTIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PHOCA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p1943"]}},
    "PHOCIS": {"N": {"morpho": ["n-s---fn3-"], "uri": ["p1944"]}},
    "PHOEBVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PHOENICE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "PHOENICES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PHOENICOPTERVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p1954"]}},
    "PHOENIX": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["p1956"]},
        "N": {"morpho": ["n-s---mn3-"], "uri": ["p1956"]},
    },
    "PHONASCVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p1957"]}},
    "PHRENESIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["p1963"]}},
    "PHRENETICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2309"]}},
    "PHRYGIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PHRYX": {
        "A": {"morpho": ["aps---mn1g"], "uri": []},
        "N": {"morpho": ["n-s---mn3-"], "uri": []},
    },
    "PHTHISICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2372"]}},
    "PHTHISIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["p1978"]}},
    "PHTHONGOS": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "PHYLARCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["p1994"]}},
    "PHYSETER": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p2001"]}},
    "PHYSICA": {
        "1": {"morpho": ["n-p---nn2-"], "uri": ["121831"]},
        "2": {"morpho": ["n-p---nn2-"], "uri": ["121831"]},
    },
    "PHYSICVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["p0964"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["p0959"]},
    },
    "PHYSIOLOGIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p2005"]}},
    "PIACVLARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p2012"]}},
    "PIACVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p2013"]}},
    "PIACVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p2014"]}},
    "PIAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p2016"]}},
    "PICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p2034"]}},
    "PICEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p2023"]}},
    "PICEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2023"]}},
    "PICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p2025"]}},
    "PICTOR": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["p2029"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "PICTVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p9837"]}},
    "PICTVRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2032"]}},
    "PICTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["54050"]}},
    "PICVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["p2034"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["p2034"]},
    },
    "PIE": {"-": {"morpho": ["rp--------"], "uri": ["54792"]}},
    "PIETAS": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["p2035"]},
        "N": {"morpho": ["n-s---mn3-"], "uri": []},
    },
    "PIGER": {"-": {"morpho": ["aps---mn1r"], "uri": ["p2037"]}},
    "PIGET": {"-": {"morpho": ["v3spia--2-"], "uri": ["p2038"]}},
    "PIGNERATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p2045"]}},
    "PIGNERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p2046"]}},
    "PIGNVS": {"-": {"morpho": ["n-s---nn3-"], "uri": ["p2048"]}},
    "PIGRE": {"-": {"morpho": ["rp--------"], "uri": ["29805"]}},
    "PIGRITIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p2051"]}},
    "PIGROR": {"-": {"morpho": ["v1spid--1-"], "uri": []}},
    "PILA": {
        "1": {"morpho": ["n-s---fn1-"], "uri": ["p2055", "p2030", "p9866"]},
        "2": {"morpho": ["n-s---fn1-"], "uri": ["p2055", "p2030", "p9866"]},
        "3": {"morpho": ["n-s---fn1-"], "uri": ["p2055", "p2030", "p9866"]},
    },
    "PILATIM": {"-": {"morpho": ["rp--------"], "uri": ["p4404"]}},
    "PILATVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["p4504"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "PILEATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["31008"]}},
    "PILENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p2063"]}},
    "PILEOLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["121891"]}},
    "PILEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["121892"]}},
    "PILEVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["32384"]}},
    "PILICREPVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p2064"]}},
    "PILO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["p2089", "p4534"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["p2089", "p4534"]},
    },
    "PILOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2069"]}},
    "PILVM": {
        "-": {"morpho": ["n-s---nn2-"], "uri": ["p2072"]},
        "N": {"morpho": ["n-s---nn2-"], "uri": ["p2072"]},
    },
    "PILVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["p2072", "p2074"]},
        "2": {"morpho": ["n-s---mn2-"], "uri": ["p2072", "p2074"]},
    },
    "PINACOTHECA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p2076"]}},
    "PINARIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "PINAX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p2080"]}},
    "PINETVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p2082"]}},
    "PINEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2083"]}},
    "PINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p2084"]}},
    "PINGVE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["p2093"]}},
    "PINGVESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p2094"]}},
    "PINGVIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p2093"]}},
    "PINGVITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2096"]}},
    "PINIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["p2099"]}},
    "PINNA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p4538"]}},
    "PINNIGER": {"-": {"morpho": ["aps---mn1r"], "uri": ["p2104"]}},
    "PINNIRAPVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p2100"]}},
    "PINSATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["121928"]}},
    "PINSO": {"-": {"morpho": ["n-s---fn3-"], "uri": []}},
    "PINVS": {"-": {"morpho": ["n-s---fn4-"], "uri": ["p2114"]}},
    "PIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p2115"]}},
    "PIPA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PIPER": {"-": {"morpho": ["n-s---nn3-"], "uri": ["p2119"]}},
    "PIPIO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["p2129"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["p2128"]},
        "3": {"morpho": ["v1spia--1-"], "uri": ["p2128"]},
    },
    "PIRAEEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PIRATA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["p2135"]}},
    "PIRATICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p2136"]}},
    "PIRATICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2136"]}},
    "PIRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p3591"]}},
    "PIRVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["p3389"]}},
    "PISATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2110"]}},
    "PISCARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2140"]}},
    "PISCATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p2142"]}},
    "PISCATORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2142"]}},
    "PISCATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["p2156"]}},
    "PISCICVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p2148"]}},
    "PISCINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p2149"]}},
    "PISCIS": {
        "-": {"morpho": ["n-s---mn3i"], "uri": ["p2154"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "PISCOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["p2157"]}},
    "PISCOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2158"]}},
    "PISCVLENTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3602"]}},
    "PISIDIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PISO": {"N": {"morpho": ["n-s---mn3-"], "uri": ["p2162"]}},
    "PISTOR": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["p2173"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "PISTRINVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p2183"]}},
    "PISTRIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["30801"]}},
    "PISTRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2184"]}},
    "PITVITA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p2197"]}},
    "PIVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["p2206"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "PIX": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["p4542"]},
        "2": {"morpho": ["n-s---fn3-"], "uri": ["p4542"]},
        "1": {"morpho": ["n-s---fn3-"], "uri": ["p4542"]},
    },
    "PLACABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p2208"]}},
    "PLACAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p2211"]}},
    "PLACATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2212"]}},
    "PLACATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["26048"]}},
    "PLACENTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p2215"]}},
    "PLACENTINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "PLACEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p2218"]}},
    "PLACIDE": {"-": {"morpho": ["rp--------"], "uri": ["26049"]}},
    "PLACIDVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["p2223"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "PLACITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p2226"]}},
    "PLACITVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p9932"]}},
    "PLACITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["54523"]}},
    "PLACO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p2227"]}},
    "PLAGA": {
        "1": {"morpho": ["n-s---fn1-"], "uri": ["p2229", "p2229", "p2230"]},
        "2": {"morpho": ["n-s---fn1-"], "uri": ["p2229", "p2229", "p2230"]},
        "3": {"morpho": ["n-s---fn1-"], "uri": ["p2229", "p2229", "p2230"]},
    },
    "PLAGIARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p2234"]}},
    "PLAGOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2242"]}},
    "PLANATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["122043"]}},
    "PLANCTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["p9886"]}},
    "PLANE": {"-": {"morpho": ["rp--------"], "uri": ["26056"]}},
    "PLANGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p2255"]}},
    "PLANGOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p2256"]}},
    "PLANIPES": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p2259"]}},
    "PLANITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2260"]}},
    "PLANITIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p2721"]}},
    "PLANITIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["p2727"]}},
    "PLANTA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["p4579"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "PLANTARIA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["p2267"]}},
    "PLANTARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p2266"]}},
    "PLANTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p2272"]}},
    "PLANVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p2273"]}},
    "PLANVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["p9867"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["p2273"]},
    },
    "PLASMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["p2275"]}},
    "PLASTES": {"-": {"morpho": ["n-s---mn1g"], "uri": ["p2280"]}},
    "PLASTICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2282"]}},
    "PLATANON": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p2289"]}},
    "PLATANVS": {"-": {"morpho": ["n-s---mn3-"], "uri": []}},
    "PLATEA": {"1": {"morpho": ["n-s---fn1-"], "uri": ["p2291"]}},
    "PLATO": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "PLAVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p2300"]}},
    "PLAVSIBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p2302"]}},
    "PLAVSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p2304"]}},
    "PLAVSTELLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["122097"]}},
    "PLAVSTRARIVS": {"2": {"morpho": ["aps---mn1-"], "uri": ["p3899"]}},
    "PLAVSTRVM": {
        "-": {"morpho": ["n-s---nn2-"], "uri": ["p1225"]},
        "N": {"morpho": ["n-s---nn2-"], "uri": ["p1225"]},
    },
    "PLAVSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["p9887"]}},
    "PLEBECVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p2311"]}},
    "PLEBEIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["p2312"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["p2312"]},
    },
    "PLEBICOLA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["p2313"]}},
    "PLEBISCITVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p2176"]}},
    "PLEBITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2315"]}},
    "PLEBS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2734"]}},
    "PLECTO": {
        "1": {"morpho": ["v1spia--3-"], "uri": ["p1977", "p2320"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["p1977", "p2320"]},
    },
    "PLECTOR": {"-": {"morpho": ["v1spid--3-"], "uri": []}},
    "PLECTRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p2324"]}},
    "PLENE": {"-": {"morpho": ["rp--------"], "uri": ["26073"]}},
    "PLENITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2329"]}},
    "PLENVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["100028"]}},
    "PLENVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2332"]}},
    "PLEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p2333"]}},
    "PLERAQVE": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "PLERIQVE": {"-": {"morpho": ["n-p---mn1-"], "uri": ["28978"]}},
    "PLERVMQVE": {"-": {"morpho": ["rp--------"], "uri": ["26076"]}},
    "PLERVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2338"]}},
    "PLERVSQVE": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2337"]}},
    "PLEVRITIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2343"]}},
    "PLICEO": {"-": {"morpho": ["v1spia--2-"], "uri": []}},
    "PLICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p2348"]}},
    "PLINTHIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2349"]}},
    "PLINTHIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p2350"]}},
    "PLINTHVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["p2351"]}},
    "PLODO": {"-": {"morpho": ["v1spia--3-"], "uri": ["34872"]}},
    "PLORABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p2356"]}},
    "PLORATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["p9888"]}},
    "PLORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p2361"]}},
    "PLOXEMVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["56733"]}},
    "PLVMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p2365"]}},
    "PLVMARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p2368"]}},
    "PLVMBARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1288"]}},
    "PLVMBEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p2373"]}},
    "PLVMBEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2373"]}},
    "PLVMBO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p2374"]}},
    "PLVMBVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p2376"]}},
    "PLVMIPES": {"-": {"morpho": ["aps---an3-"], "uri": ["p2380"]}},
    "PLVMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p2381"]}},
    "PLVMVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p2383"]}},
    "PLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p2384"]}},
    "PLVRIMVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["53655"]}},
    "PLVSCVLVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["p2398"]},
        "2": {"morpho": ["rp--------"], "uri": ["41589"]},
    },
    "PLVTEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["122187"]}},
    "PLVTEVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p2401"]}},
    "PLVVIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p2409"]}},
    "PLVVIALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p2405"]}},
    "PLVVIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2409"]}},
    "PNEVMATICOS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "PNIGEVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p2411"]}},
    "POCILLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p2416"]}},
    "POCVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p2419"]}},
    "PODAGRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p2420"]}},
    "PODAGRICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2421"]}},
    "PODEX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p2424"]}},
    "PODIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p2431"]}},
    "POEMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["p2432"]}},
    "POENA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["p2434"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "POENI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "POENVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "POETA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["p2438"]}},
    "POETICA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["p1601"]},
        "2": {"morpho": ["n-s---fn1-"], "uri": ["p1601"]},
    },
    "POETICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1599"]}},
    "POETRIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p2442"]}},
    "POETRIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2443"]}},
    "POL": {"-": {"morpho": ["----------"], "uri": []}},
    "POLEAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "POLEMARCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "POLENTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p2448"]}},
    "POLIO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["122234"]},
        "2": {"morpho": ["v1spia--4-"], "uri": ["p2455"]},
    },
    "POLITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2458"]}},
    "POLITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p2460"]}},
    "POLITORIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "POLITVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p9839"]}},
    "POLITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["43441"]}},
    "POLLEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["p2462"]}},
    "POLLENS": {"-": {"morpho": ["aps---an3i"], "uri": ["26098"]}},
    "POLLEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p2463"]}},
    "POLLEX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p2464"]}},
    "POLLICEOR": {"-": {"morpho": ["v1spid--2-"], "uri": ["p2466"]}},
    "POLLICITATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2467"]}},
    "POLLICITOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["p2469"]}},
    "POLLICITVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p2439"]}},
    "POLLINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p2473"]}},
    "POLLIO": {"N": {"morpho": ["n-s---mn3-"], "uri": ["p2475"]}},
    "POLLVCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p2476"]}},
    "POLLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p2479"]}},
    "POLLVTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["35779"]}},
    "POLLVX": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "POLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p2485"]}},
    "POLVSCA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "POLYBIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "POLYCLITVS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "POLYGONIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2501"]}},
    "POLYPVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p2517"]}},
    "POLYSPASTON": {"-": {"morpho": ["n-s---nn2g"], "uri": ["p2522"]}},
    "POMARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p2528"]}},
    "POMARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["p2528"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["p2528"]},
    },
    "POMERIDIANVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["58298"]}},
    "POMERIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p2530"]}},
    "POMETIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "POMETINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "POMIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["p2532"]}},
    "POMOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2533"]}},
    "POMPA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p2534"]}},
    "POMPEIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "POMPILIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "POMPONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "POMPTINVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "POMVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p2546"]}},
    "POMVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["26111"]}},
    "PONDERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2551"]}},
    "PONDERATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["122337"]}},
    "PONDERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p2555"]}},
    "PONDEROSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2556"]}},
    "PONDO": {"-": {"morpho": ["n-s----n--"], "uri": []}},
    "PONDVS": {"-": {"morpho": ["n-s---nn3-"], "uri": ["p2559"]}},
    "PONE": {
        "1": {"morpho": ["rp--------"], "uri": []},
        "2": {"morpho": ["p---------"], "uri": ["p2561"]},
    },
    "PONEVERSVS": {"-": {"morpho": ["p---------"], "uri": []}},
    "PONO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p2562"]}},
    "PONS": {
        "-": {"morpho": ["n-s---mn3i"], "uri": ["p2563"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "PONTICVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p2565"]}},
    "PONTICVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "PONTIFEX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p2566"]}},
    "PONTIFICALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p2567"]}},
    "PONTIFICATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["p2568"]}},
    "PONTIFICIVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["p2569"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "PONTO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p2573"]}},
    "PONTVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["p2575"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["p2575"]},
    },
    "POPA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["p2576"]}},
    "POPANVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p2577"]}},
    "POPELLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p2578"]}},
    "POPILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "POPINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p2579"]}},
    "POPINO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p2583"]}},
    "POPLES": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p2585"]}},
    "POPVLARIS": {
        "1": {"morpho": ["n-s----n3-"], "uri": []},
        "2": {"morpho": ["aps---cn3i"], "uri": ["p2592"]},
    },
    "POPVLARITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2593"]}},
    "POPVLARITER": {"-": {"morpho": ["rp--------"], "uri": ["30779"]}},
    "POPVLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p4607"]}},
    "POPVLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p2596"]}},
    "POPVLEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2598"]}},
    "POPVLIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["p2599"]}},
    "POPVLNEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2601"]}},
    "POPVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["100053"]}},
    "POPVLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["p2602"]}},
    "POPVLVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["p2606"]},
        "2": {"morpho": ["n-s---mn2-"], "uri": ["p2606"]},
    },
    "PORCA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p2627"]}},
    "PORCINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2621"]}},
    "PORCIVS": {
        "-": {"morpho": ["n-s----n2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "PORCVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p4681"]}},
    "PORCVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p2627"]}},
    "PORRECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2634"]}},
    "PORRECTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["36583"]}},
    "PORRIGO": {
        "1": {"morpho": ["n-s---fn3-"], "uri": ["p2640"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["p2638"]},
    },
    "PORRINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p2641"]}},
    "PORRO": {"-": {"morpho": ["rp--------"], "uri": ["p2643"]}},
    "PORRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p2644"]}},
    "PORRVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["56326"]}},
    "PORSENA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PORTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p2645"]}},
    "PORTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2648"]}},
    "PORTENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p2653"]}},
    "PORTENTOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2657"]}},
    "PORTENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p2650"]}},
    "PORTHMEVS": {
        "-": {"morpho": ["n-s----n-g"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "PORTICVS": {
        "-": {"morpho": ["n-s---fn4-"], "uri": ["p2658"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "PORTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2665"]}},
    "PORTISCVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p2667"]}},
    "PORTITOR": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["p2668"]},
        "2": {"morpho": ["n-s---mn3-"], "uri": ["p2668"]},
    },
    "PORTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PORTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p2671"]}},
    "PORTORIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p2672"]}},
    "PORTVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p2673"]}},
    "PORTVOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2675"]}},
    "PORTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["p2676"]}},
    "PORVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["p2677", "p2669"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["p2677", "p2669"]},
    },
    "POSCIMVS": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "POSCO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "POSIDES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "POSIDORVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "POSITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2684"]}},
    "POSITVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p9842"]}},
    "POSITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["p9890"]}},
    "POSOTES": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "POSSESSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2687"]}},
    "POSSESSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p2690"]}},
    "POSSIDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p2695"]}},
    "POSSIDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p2696"]}},
    "POSSVM": {"1": {"morpho": ["v1spia--3-"], "uri": ["p2697"]}},
    "POST": {
        "1": {"morpho": ["rp--------"], "uri": ["26149"]},
        "2": {"morpho": ["p---------"], "uri": ["p2700"]},
    },
    "POSTEA": {"-": {"morpho": ["rp--------"], "uri": ["p2702"]}},
    "POSTEAQVAM": {"-": {"morpho": ["----------"], "uri": []}},
    "POSTERGA": {"-": {"morpho": ["rp--------"], "uri": ["33287"]}},
    "POSTERGVM": {"-": {"morpho": ["rp--------"], "uri": ["50064"]}},
    "POSTERI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["p2712"]}},
    "POSTERITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2709"]}},
    "POSTERIVS": {"-": {"morpho": ["rp--------"], "uri": ["26151"]}},
    "POSTERVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p9446"]}},
    "POSTFERO": {"-": {"morpho": ["v1spia----"], "uri": ["p2714"]}},
    "POSTGENITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2716"]}},
    "POSTHABEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p2717"]}},
    "POSTHAC": {"-": {"morpho": ["rp--------"], "uri": ["p2718"]}},
    "POSTICVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p1920"]}},
    "POSTICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1739"]}},
    "POSTILLA": {"-": {"morpho": ["rp--------"], "uri": ["p5004"]}},
    "POSTIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["p5006"]}},
    "POSTMERIDIANVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p5009"]}},
    "POSTMODO": {"-": {"morpho": ["rp--------"], "uri": ["p2728"]}},
    "POSTMODVM": {"-": {"morpho": ["rp--------"], "uri": ["p2728"]}},
    "POSTPONO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p2731"]}},
    "POSTQVAM": {"-": {"morpho": ["----------"], "uri": []}},
    "POSTREMO": {"-": {"morpho": ["rp--------"], "uri": ["45353"]}},
    "POSTREMVM": {"-": {"morpho": ["rp--------"], "uri": ["52365"]}},
    "POSTRIDIE": {"-": {"morpho": ["rp--------"], "uri": ["p2737"]}},
    "POSTSCENIVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "POSTSCRIBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p2741"]}},
    "POSTVLATICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2746"]}},
    "POSTVLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2747"]}},
    "POSTVLATVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p9891"]}},
    "POSTVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p2751"]}},
    "POSTVMIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "POSTVMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["p2752"]}},
    "POTA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "POTAMO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "POTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2761"]}},
    "POTENS": {"-": {"morpho": ["aps---an3i"], "uri": ["100063"]}},
    "POTENTATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["p9892"]}},
    "POTENTER": {"-": {"morpho": ["rp--------"], "uri": ["35878"]}},
    "POTENTIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["p2698"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["p2698"]},
    },
    "POTESTAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2768"]}},
    "POTIO": {
        "1": {"morpho": ["n-s---fn3-"], "uri": ["p2773"]},
        "2": {"morpho": ["v1spia--4-"], "uri": ["p2774"]},
    },
    "POTIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["p2776"]}},
    "POTIS": {"-": {"morpho": ["aps---cn3-"], "uri": ["p2777"]}},
    "POTISSIME": {"-": {"morpho": ["rp--------"], "uri": ["36363"]}},
    "POTISSIMVM": {"-": {"morpho": ["rp--------"], "uri": ["26168"]}},
    "POTITII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "POTIVS": {"-": {"morpho": ["rp--------"], "uri": ["26170"]}},
    "POTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p2782"]}},
    "POTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p2784"]}},
    "POTORO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "POTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["p9930"]},
        "2": {"morpho": ["n-s---mn4-"], "uri": ["p9930"]},
    },
    "PRAE": {
        "1": {"morpho": ["rp--------"], "uri": ["26175"]},
        "2": {"morpho": ["p---------"], "uri": ["p2788"]},
    },
    "PRAEACVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p2790"]}},
    "PRAEALTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2791"]}},
    "PRAEBEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p2796"]}},
    "PRAEBITA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["122521"]}},
    "PRAECALIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2803"]}},
    "PRAECANVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2812"]}},
    "PRAECAVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p2815"]}},
    "PRAECEDO": {"1": {"morpho": ["v1spia--3-"], "uri": ["p2817"]}},
    "PRAECELLENS": {"-": {"morpho": ["aps---an3i"], "uri": ["28535"]}},
    "PRAECELLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p2820"]}},
    "PRAECELSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2821"]}},
    "PRAECEPS": {
        "1": {"morpho": ["n-s---nn3-"], "uri": ["100074"]},
        "2": {"morpho": ["aps---an3-"], "uri": ["p2826"]},
        "3": {"morpho": ["rp--------"], "uri": ["100073"]},
    },
    "PRAECEPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2827"]}},
    "PRAECEPTIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2828"]}},
    "PRAECEPTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p2830"]}},
    "PRAECEPTRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2831"]}},
    "PRAECEPTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p2785"]}},
    "PRAECERPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p2832"]}},
    "PRAECIDANEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2836"]}},
    "PRAECIDO": {"2": {"morpho": ["v1spia--3-"], "uri": ["p2838"]}},
    "PRAECINCTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2839"]}},
    "PRAECINCTVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p9844"]}},
    "PRAECINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p2841"]}},
    "PRAECINO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p2842"]}},
    "PRAECIPIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["p2843"]}},
    "PRAECIPITANTER": {"-": {"morpho": ["rp--------"], "uri": ["100079"]}},
    "PRAECIPITATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2845"]}},
    "PRAECIPITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p2847"]}},
    "PRAECIPVA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["122572"]}},
    "PRAECIPVE": {"-": {"morpho": ["rp--------"], "uri": ["26188"]}},
    "PRAECIPVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2849"]}},
    "PRAECISIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2853"]}},
    "PRAECISVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["52644"]}},
    "PRAECLARE": {"-": {"morpho": ["rp--------"], "uri": ["50583"]}},
    "PRAECLARVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2858"]}},
    "PRAECLVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p2860"]}},
    "PRAECLVSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2863"]}},
    "PRAECO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p2865"]}},
    "PRAECOGITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p2868"]}},
    "PRAECOLO": {"2": {"morpho": ["v1spia--3-"], "uri": ["p2872"]}},
    "PRAECONIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p2879"]}},
    "PRAECOQVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p2884"]}},
    "PRAECORDIA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["p2885"]}},
    "PRAECOX": {"-": {"morpho": ["aps---an3-"], "uri": ["p2887"]}},
    "PRAECVRRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p2893"]}},
    "PRAECVRSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p2896"]}},
    "PRAECVTIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["p2899"]}},
    "PRAEDA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p2900"]}},
    "PRAEDABVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2901"]}},
    "PRAEDATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2905"]}},
    "PRAEDATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p2906"]}},
    "PRAEDATORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2906"]}},
    "PRAEDICABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p2922"]}},
    "PRAEDICATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2924"]}},
    "PRAEDICO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["p2928"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["p2929"]},
    },
    "PRAEDICTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2930"]}},
    "PRAEDICTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p9944"]}},
    "PRAEDISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p2936"]}},
    "PRAEDITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2938"]}},
    "PRAEDIVES": {"-": {"morpho": ["aps---an3-"], "uri": ["p2939"]}},
    "PRAEDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p2944"]}},
    "PRAEDO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p2946"]}},
    "PRAEDOCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p2947"]}},
    "PRAEDOMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p2948"]}},
    "PRAEDONVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p2950"]}},
    "PRAEDOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["p2951"]}},
    "PRAEDVCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p2954"]}},
    "PRAEDVCTORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2955"]}},
    "PRAEDVRVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2958"]}},
    "PRAEEO": {"1": {"morpho": ["v1spia--4-"], "uri": ["p2962"]}},
    "PRAEFATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2968"]}},
    "PRAEFATVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["p9897"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["122685"]},
    },
    "PRAEFECTVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p9848"]}},
    "PRAEFECTVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p9943"]}},
    "PRAEFERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p2977"]}},
    "PRAEFEROX": {"-": {"morpho": ["aps---an3i"], "uri": ["p2978"]}},
    "PRAEFERRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2979"]}},
    "PRAEFERVIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2981"]}},
    "PRAEFESTINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p2983"]}},
    "PRAEFICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["p2985"]}},
    "PRAEFIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p2987"]}},
    "PRAEFINIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["p2990"]}},
    "PRAEFLORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p2995"]}},
    "PRAEFLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p2996"]}},
    "PRAEFOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["p3003"]}},
    "PRAEFRACTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["35306"]}},
    "PRAEFRINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3011"]}},
    "PRAEFVLGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p3014"]}},
    "PRAEFVRNIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p3018"]}},
    "PRAEGERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3024"]}},
    "PRAEGESTIO": {"2": {"morpho": ["v1spia--4-"], "uri": ["p3025"]}},
    "PRAEGNANS": {"-": {"morpho": ["aps---an3i"], "uri": ["100094"]}},
    "PRAEGRACILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p3030"]}},
    "PRAEGRANDIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p3032"]}},
    "PRAEGRAVIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p3035"]}},
    "PRAEGRAVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3036"]}},
    "PRAEGREDIOR": {"-": {"morpho": ["v1spid--3i"], "uri": ["p3037"]}},
    "PRAEGVSTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3041"]}},
    "PRAEIACEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p3045"]}},
    "PRAEIVDICIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p3051"]}},
    "PRAEIVDICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3052"]}},
    "PRAELABOR": {"2": {"morpho": ["v1spid--3-"], "uri": ["p3055"]}},
    "PRAELAMBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3056"]}},
    "PRAELARGVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3057"]}},
    "PRAELEGO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["p3068"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["p3069"]},
    },
    "PRAELIGANEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3076"]}},
    "PRAELIGO": {"2": {"morpho": ["v1spia--1-"], "uri": ["p3077"]}},
    "PRAELONGVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3083"]}},
    "PRAELVCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p3086"]}},
    "PRAEMATVRVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3098"]}},
    "PRAEMEDICATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3099"]}},
    "PRAEMEDITATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["100104"]}},
    "PRAEMEDITOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["p3102"]}},
    "PRAEMETVENTER": {"-": {"morpho": ["rp--------"], "uri": ["100105"]}},
    "PRAEMETVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3110"]}},
    "PRAEMINEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p2960"]}},
    "PRAEMIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3119"]}},
    "PRAEMITTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3124"]}},
    "PRAEMIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p3125"]}},
    "PRAEMONEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p3133"]}},
    "PRAEMONSTRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3140"]}},
    "PRAEMORDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p3141"]}},
    "PRAEMORIOR": {"-": {"morpho": ["v1spid--3i"], "uri": ["p3143"]}},
    "PRAEMVNIO": {"2": {"morpho": ["v1spia--4-"], "uri": ["p3147"]}},
    "PRAENATO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3151"]}},
    "PRAENAVIGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3153"]}},
    "PRAENESTE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PRAENITEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p3156"]}},
    "PRAENOMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["p3158"]}},
    "PRAENOSCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3161"]}},
    "PRAENOTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3163"]}},
    "PRAENVBILVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3165"]}},
    "PRAENVNTIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3172"]}},
    "PRAENVNTIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["57886"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["p3173"]},
    },
    "PRAEOBTVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3174"]}},
    "PRAEOCCVPO": {"2": {"morpho": ["v1spia--1-"], "uri": ["p3178"]}},
    "PRAEOPTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3182"]}},
    "PRAEPANDO": {"2": {"morpho": ["v1spia--3-"], "uri": ["p3186"]}},
    "PRAEPARATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3187"]}},
    "PRAEPARATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["54200"]}},
    "PRAEPARO": {"2": {"morpho": ["v1spia--1-"], "uri": ["p3191"]}},
    "PRAEPEDIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["p3195"]}},
    "PRAEPENDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p3196"]}},
    "PRAEPES": {
        "1": {"morpho": ["n-s----n3-"], "uri": []},
        "2": {"morpho": ["aps---cn3i"], "uri": ["p3197"]},
    },
    "PRAEPETO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3198"]}},
    "PRAEPILATVS": {"1": {"morpho": ["aps---mn1-"], "uri": ["p4682"]}},
    "PRAEPINGVIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p3201"]}},
    "PRAEPOLLENS": {"-": {"morpho": ["aps---an3i"], "uri": ["100115"]}},
    "PRAEPOLLEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p3202"]}},
    "PRAEPONDERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3204"]}},
    "PRAEPONO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3205"]}},
    "PRAEPORTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3206"]}},
    "PRAEPOSITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3207"]}},
    "PRAEPOSITVS": {"1": {"morpho": ["n-s---mn2-"], "uri": ["p9956"]}},
    "PRAEPOSTERO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["p3211"]},
        "2": {"morpho": ["rp--------"], "uri": []},
    },
    "PRAEPOSTERVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3212"]}},
    "PRAEPOTENS": {"-": {"morpho": ["aps---an3i"], "uri": ["p3213"]}},
    "PRAEPOTENTES": {"-": {"morpho": ["n-p---mn3i"], "uri": ["100119"]}},
    "PRAEPROPERANTER": {"-": {"morpho": ["rp--------"], "uri": ["p3216"]}},
    "PRAEPROPERE": {"-": {"morpho": ["rp--------"], "uri": ["40850"]}},
    "PRAEPVTIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p3223"]}},
    "PRAERADIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3226"]}},
    "PRAERAPIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3228"]}},
    "PRAERIGEO": {"-": {"morpho": ["v1spia--2-"], "uri": []}},
    "PRAERIPIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["p3234"]}},
    "PRAERODO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3236"]}},
    "PRAEROGATIVA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p3238"]}},
    "PRAEROGATIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3238"]}},
    "PRAERVMPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3242"]}},
    "PRAERVPTA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["100122"]}},
    "PRAERVPTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p9901"]}},
    "PRAERVPTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["38542"]}},
    "PRAES": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["p3245"]},
        "2": {"morpho": ["rp--------"], "uri": ["p3246"]},
    },
    "PRAESAEPE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["100123"]}},
    "PRAESAEPES": {"-": {"morpho": ["n-s---fn3i"], "uri": ["122954"]}},
    "PRAESAEPIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["p3247"]}},
    "PRAESAGIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["p3249"]}},
    "PRAESAGIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p3251"]}},
    "PRAESAGVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3254"]}},
    "PRAESCISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p2168"]}},
    "PRAESCIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3261"]}},
    "PRAESCRIBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3262"]}},
    "PRAESCRIPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3263"]}},
    "PRAESCRIPTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p9903"]}},
    "PRAESECO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3267"]}},
    "PRAESEMINATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3269"]}},
    "PRAESENS": {"-": {"morpho": ["aps---an3i"], "uri": ["100127"]}},
    "PRAESENSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3271"]}},
    "PRAESENTANEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3273"]}},
    "PRAESENTIA": {
        "1": {"morpho": ["n-p---nn3i"], "uri": ["p9852"]},
        "2": {"morpho": ["n-p---nn3i"], "uri": ["p9852"]},
    },
    "PRAESENTIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["p3276"]}},
    "PRAESENTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3277"]}},
    "PRAESERO": {"3": {"morpho": ["v1spia--3-"], "uri": ["$0147"]}},
    "PRAESERTIM": {"-": {"morpho": ["rp--------"], "uri": ["p3281"]}},
    "PRAESES": {"-": {"morpho": ["n-s---cn3-"], "uri": ["p3284"]}},
    "PRAESIDENS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["100131"]}},
    "PRAESIDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p3292"]}},
    "PRAESIDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p3297"]}},
    "PRAESIGNIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p3301"]}},
    "PRAESONO": {"1": {"morpho": ["v1spia--1-"], "uri": ["p3306"]}},
    "PRAESPARGO": {"2": {"morpho": ["v1spia--3-"], "uri": ["p3307"]}},
    "PRAESPICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["p3310"]}},
    "PRAESTABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p3311"]}},
    "PRAESTANS": {"-": {"morpho": ["aps---an3i"], "uri": ["100134"]}},
    "PRAESTANTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p9853"]}},
    "PRAESTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3312"]}},
    "PRAESTERNO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3314"]}},
    "PRAESTIGIAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["100136"]}},
    "PRAESTIGIATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p3317"]}},
    "PRAESTITVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3325"]}},
    "PRAESTO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["p3327"]},
        "2": {"morpho": ["rp--------"], "uri": ["p3326"]},
    },
    "PRAESTOLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["p3329"]}},
    "PRAESTRIGIAE": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "PRAESTRINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3332"]}},
    "PRAESTRVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3335"]}},
    "PRAESVM": {"1": {"morpho": ["v1spia--1-"], "uri": ["p3345"]}},
    "PRAESVMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3347"]}},
    "PRAESVMPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3348"]}},
    "PRAESVMPTIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3349"]}},
    "PRAESVMPTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p3351"]}},
    "PRAESVMPTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["100139"]}},
    "PRAETEGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3360"]}},
    "PRAETEMPTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3361"]}},
    "PRAETENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3362"]}},
    "PRAETENTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3361"]}},
    "PRAETEPESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3365"]}},
    "PRAETER": {
        "1": {"morpho": ["rp--------"], "uri": []},
        "2": {"morpho": ["p---------"], "uri": ["p3366"]},
    },
    "PRAETERAGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3367"]}},
    "PRAETEREA": {"-": {"morpho": ["rp--------"], "uri": ["p4721"]}},
    "PRAETEREO": {"1": {"morpho": ["v1spia--2-"], "uri": ["p3372"]}},
    "PRAETERFERO": {"-": {"morpho": ["v1spia----"], "uri": ["p3375"]}},
    "PRAETERFLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p6376"]}},
    "PRAETERGREDIOR": {"-": {"morpho": ["v1spid--3i"], "uri": ["p3376"]}},
    "PRAETERITA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["100142"]}},
    "PRAETERITVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["56139"]}},
    "PRAETERITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["26266"]}},
    "PRAETERLABOR": {"2": {"morpho": ["v1spid--3-"], "uri": ["p3380"]}},
    "PRAETERMEO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3383"]}},
    "PRAETERMITTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3385"]}},
    "PRAETERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3388"]}},
    "PRAETERQVAM": {"-": {"morpho": ["rp--------"], "uri": ["p5519"]}},
    "PRAETERVECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3391"]}},
    "PRAETERVEHOR": {"-": {"morpho": ["v1spid--3-"], "uri": []}},
    "PRAETERVERTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3393"]}},
    "PRAETERVOLO": {"2": {"morpho": ["v1spia--1-"], "uri": ["p3394"]}},
    "PRAETEXO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3396"]}},
    "PRAETEXTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p3424"]}},
    "PRAETEXTATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3397"]}},
    "PRAETEXTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p3424"]}},
    "PRAETEXTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["p3424"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["44379"]},
    },
    "PRAETIMEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p3398"]}},
    "PRAETOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p3403"]}},
    "PRAETORIANVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p4683"]}},
    "PRAETORIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p3407"]}},
    "PRAETORIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["100144"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["p3407"]},
    },
    "PRAETRACTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3410"]}},
    "PRAETREPIDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3413"]}},
    "PRAETREPIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3414"]}},
    "PRAETVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p3417"]}},
    "PRAEVALENS": {"-": {"morpho": ["aps---an3i"], "uri": ["123127"]}},
    "PRAEVALEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p3419"]}},
    "PRAEVALIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3420"]}},
    "PRAEVALLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3421"]}},
    "PRAEVARICATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3422"]}},
    "PRAEVARICOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["p3435"]}},
    "PRAEVEHOR": {"-": {"morpho": ["v1spid--3-"], "uri": []}},
    "PRAEVENIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["p3431"]}},
    "PRAEVERRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p6435"]}},
    "PRAEVERTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3436"]}},
    "PRAEVERTOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["132665"]}},
    "PRAEVIDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p3441"]}},
    "PRAEVINCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["123155"]}},
    "PRAEVIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3450"]}},
    "PRAEVMBRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3452"]}},
    "PRAEVRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3456"]}},
    "PRAGMATICVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["123173"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["p3460"]},
    },
    "PRANDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p3461"]}},
    "PRANDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p3463"]}},
    "PRANSITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3465"]}},
    "PRATENSIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p3478"]}},
    "PRATVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p3481"]}},
    "PRAVE": {"-": {"morpho": ["rp--------"], "uri": ["53172"]}},
    "PRAVITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3484"]}},
    "PRAVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3485"]}},
    "PRAXITELES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "PRECARIO": {"-": {"morpho": ["rp--------"], "uri": ["44032"]}},
    "PRECARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3488"]}},
    "PRECATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3489"]}},
    "PRECIAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["41938"]}},
    "PRECOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["p3496"]}},
    "PREHENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3497"]}},
    "PREHENSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3499"]}},
    "PRELVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p3501"]}},
    "PREMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3502"]}},
    "PRESBYTER": {"-": {"morpho": ["n-s---mn2r"], "uri": ["p3508"]}},
    "PRESBYTERIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p3510"]}},
    "PRESSE": {"-": {"morpho": ["rp--------"], "uri": ["36042"]}},
    "PRESSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3512"]}},
    "PRESSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3513"]}},
    "PRESSVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["p9907"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["49546"]},
    },
    "PRESTER": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p3517"]}},
    "PRETIOSE": {"-": {"morpho": ["rp--------"], "uri": ["49288"]}},
    "PRETIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3519"]}},
    "PRETIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p3520"]}},
    "PREX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3521"]}},
    "PRIAMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PRIAPVS": {
        "-": {"morpho": ["n-s----n2-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "PRIDEM": {"-": {"morpho": ["rp--------"], "uri": ["p3526"]}},
    "PRIDIE": {"-": {"morpho": ["rp--------"], "uri": ["p3528"]}},
    "PRIMA": {
        "1": {"morpho": ["n-p---nn2-"], "uri": ["100154"]},
        "2": {"morpho": ["rp--------"], "uri": []},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "PRIMAEVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3530"]}},
    "PRIMANI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["100155"]}},
    "PRIMARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3532"]}},
    "PRIMIGENVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3540"]}},
    "PRIMIPILARIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["100156"]}},
    "PRIMIPILVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p3545"]}},
    "PRIMITVS": {"-": {"morpho": ["rp--------"], "uri": ["p3551"]}},
    "PRIMO": {
        "-": {"morpho": ["v1spia--3-", "rp--------"], "uri": ["26297"]},
        "1": {"morpho": ["rp--------"], "uri": ["26297"]},
        "2": {"morpho": ["v1spia--3-"], "uri": []},
    },
    "PRIMORDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p3557"]}},
    "PRIMORES": {"-": {"morpho": ["n-s---mn3-"], "uri": ["100157"]}},
    "PRIMORIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p3558"]}},
    "PRIMVM": {
        "-": {
            "morpho": ["rp--------", "rp--------"],
            "uri": ["26299", "26299"],
        }
    },
    "PRIMVS": {
        "-": {
            "morpho": ["aps---mn1-", "n-s---nn3-"],
            "uri": ["p3562", "46637"],
        },
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "PRINCEPS": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["100159"]},
        "2": {"morpho": ["aps---an3-"], "uri": ["p3563"]},
    },
    "PRINCIPALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p3564"]}},
    "PRINCIPALITER": {"-": {"morpho": ["rp--------"], "uri": ["26301"]}},
    "PRINCIPATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["p3566"]}},
    "PRINCIPIALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p3567"]}},
    "PRINCIPIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p4684"]}},
    "PRIOR": {
        "-": {
            "morpho": ["acs---cn3-", "acs---cn3-"],
            "uri": ["p3572", "p3572"],
        }
    },
    "PRIORES": {"-": {"morpho": ["n-s---mn3-"], "uri": ["100161"]}},
    "PRISCI": {
        "-": {"morpho": ["n-p---mn2-"], "uri": ["100162"]},
        "N": {"morpho": ["n-p---mn2-"], "uri": ["100162"]},
    },
    "PRISCVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["p3576"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "PRISTINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p4685"]}},
    "PRIVATIM": {"-": {"morpho": ["rp--------"], "uri": ["p3582"]}},
    "PRIVATO": {"-": {"morpho": ["rp--------"], "uri": ["45526"]}},
    "PRIVATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["33402"]}},
    "PRIVIENVS": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "PRIVIGNA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p3587"]}},
    "PRIVIGNVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p3587"]}},
    "PRIVILEGIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p3589"]}},
    "PRIVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3590"]}},
    "PRIVS": {"-": {"morpho": ["rc--------", "rp--------"], "uri": ["26308"]}},
    "PRIVSQVAM": {"-": {"morpho": ["----------"], "uri": []}},
    "PRIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3592"]}},
    "PRO": {
        "1": {"morpho": ["p---------"], "uri": ["p2610"]},
        "2": {"morpho": ["----------"], "uri": []},
        "3": {"morpho": ["p---------"], "uri": ["p2610"]},
    },
    "PROAGORVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p3596"]}},
    "PROAVIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p3603"]}},
    "PROAVVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p3606"]}},
    "PROBABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p3607"]}},
    "PROBABILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3608"]}},
    "PROBATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3612"]}},
    "PROBATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p3615"]}},
    "PROBATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["38501"]}},
    "PROBE": {"-": {"morpho": ["rp--------"], "uri": ["37793"]}},
    "PROBITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3619"]}},
    "PROBO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3623"]}},
    "PROBOSCIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3624"]}},
    "PROBRO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "PROBROSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3628"]}},
    "PROBRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p3629"]}},
    "PROBVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["p3631"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["123303"]},
    },
    "PROCACITER": {"-": {"morpho": ["rp--------"], "uri": ["28828"]}},
    "PROCAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "PROCAX": {"-": {"morpho": ["aps---an3i"], "uri": ["p3638"]}},
    "PROCEDO": {"1": {"morpho": ["v1spia--3-"], "uri": ["p3639"]}},
    "PROCELLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p3640"]}},
    "PROCELLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3641"]}},
    "PROCELLOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3642"]}},
    "PROCER": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p3643"]}},
    "PROCERES": {"-": {"morpho": ["n-s---mn3-"], "uri": ["100167"]}},
    "PROCERITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3644"]}},
    "PROCERVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3647"]}},
    "PROCESSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3648"]}},
    "PROCESSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["p9908"]}},
    "PROCIDO": {"1": {"morpho": ["v1spia--3-"], "uri": ["p3652"]}},
    "PROCINCTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["p9909"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["100168"]},
    },
    "PROCLAMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3659"]}},
    "PROCLINATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3661"]}},
    "PROCLINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3662"]}},
    "PROCLIVE": {
        "1": {"morpho": ["n-s---nn3i"], "uri": ["p3663"]},
        "2": {"morpho": ["rp--------"], "uri": ["48766"]},
    },
    "PROCLIVIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p3663"]}},
    "PROCLIVITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3664"]}},
    "PROCLIVITER": {"-": {"morpho": ["rp--------"], "uri": ["123329"]}},
    "PROCLIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3663"]}},
    "PROCONSVL": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p3669"]}},
    "PROCONSVLARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p3670"]}},
    "PROCONSVLATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["p3671"]}},
    "PROCRASTINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3673"]}},
    "PROCREATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3675"]}},
    "PROCREO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3678"]}},
    "PROCRESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3679"]}},
    "PROCVBO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3681"]}},
    "PROCVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3683"]}},
    "PROCVL": {"-": {"morpho": ["rp--------"], "uri": ["p3684"]}},
    "PROCVLCATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3685"]}},
    "PROCVLCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3687"]}},
    "PROCVLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PROCVMBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3692"]}},
    "PROCVRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3694"]}},
    "PROCVRATIVNCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p3695"]}},
    "PROCVRATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p3696"]}},
    "PROCVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3699"]}},
    "PROCVRRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3700"]}},
    "PROCVRSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["p9910"]}},
    "PROCVRVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3706"]}},
    "PROCVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p4686"]}},
    "PRODEO": {"1": {"morpho": ["v1spia--4-"], "uri": ["p3709"]}},
    "PRODIGE": {"-": {"morpho": ["rp--------"], "uri": ["33676"]}},
    "PRODIGENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p9857"]}},
    "PRODIGIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3718"]}},
    "PRODIGIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p3720"]}},
    "PRODIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3721"]}},
    "PRODIGVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3722"]}},
    "PRODITIO": {
        "1": {"morpho": ["n-s---fn3-"], "uri": ["p4687"]},
        "2": {"morpho": ["n-s---fn3-"], "uri": ["p4687"]},
    },
    "PRODITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p3725"]}},
    "PRODO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3729"]}},
    "PRODOCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p3731"]}},
    "PRODVCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3734"]}},
    "PRODVCTA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["p9945"]}},
    "PRODVCTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3736"]}},
    "PRODVCTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["42030"]}},
    "PROELIATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p3746"]}},
    "PROELIOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["p3747"]}},
    "PROELIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p3748"]}},
    "PROFANO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["p4688"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["p4688"]},
    },
    "PROFANVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3753"]}},
    "PROFECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3755"]}},
    "PROFECTO": {"-": {"morpho": ["rp--------"], "uri": ["26351"]}},
    "PROFECTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["p9911"]}},
    "PROFERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3756"]}},
    "PROFESSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3757"]}},
    "PROFESSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p3760"]}},
    "PROFESSORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3761"]}},
    "PROFESSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["53499"]}},
    "PROFESTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3762"]}},
    "PROFICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["p3763"]}},
    "PROFICISCOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["p3764"]}},
    "PROFITEOR": {"-": {"morpho": ["v1spid--2-"], "uri": ["p3767"]}},
    "PROFLIGATOR": {"-": {"morpho": ["n-s---fn3-"], "uri": ["34147"]}},
    "PROFLIGATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["40688"]}},
    "PROFLIGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3771"]}},
    "PROFLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3773"]}},
    "PROFLVENS": {"-": {"morpho": ["aps---an3i"], "uri": ["100181"]}},
    "PROFLVENTER": {"-": {"morpho": ["rp--------"], "uri": ["57405"]}},
    "PROFLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3774"]}},
    "PROFLVVIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p1992"]}},
    "PROFLVVIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1976"]}},
    "PROFOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["p3777"]}},
    "PROFVGIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["p3780"]}},
    "PROFVGVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["p3781"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["p3781"]},
    },
    "PROFVNDO": {"2": {"morpho": ["v1spia--3-"], "uri": ["p3783"]}},
    "PROFVNDVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p3784"]}},
    "PROFVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3784"]}},
    "PROFVSE": {"-": {"morpho": ["rp--------"], "uri": ["38201"]}},
    "PROFVSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3785"]}},
    "PROFVSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["100182"]}},
    "PROGENER": {"-": {"morpho": ["n-s---mn2r"], "uri": ["p3788"]}},
    "PROGENERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3790"]}},
    "PROGENIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["p3791"]}},
    "PROGERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3795"]}},
    "PROGIGNO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3797"]}},
    "PROGNATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["100184"]}},
    "PROGNE": {
        "-": {"morpho": ["n-s----n-g"], "uri": []},
        "N": {"morpho": ["n-s---fn1-"], "uri": []},
    },
    "PROGREDIOR": {"-": {"morpho": ["v1spid--3i"], "uri": ["p3804"]}},
    "PROGRESSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3805"]}},
    "PROGRESSVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["p9913"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["100185"]},
    },
    "PROGYMNASTES": {"-": {"morpho": ["n-s---mn1g"], "uri": ["p3808"]}},
    "PROHIBEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p3810"]}},
    "PROHIBITA": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "PROHIBITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3811"]}},
    "PROICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["p3816"]}},
    "PROIECTVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p3839"]}},
    "PROIECTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["p9914"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["53242"]},
    },
    "PROINDE": {"-": {"morpho": ["rp--------"], "uri": ["p3821"]}},
    "PROLABOR": {"2": {"morpho": ["v1spid--3-"], "uri": ["p3822"]}},
    "PROLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3825"]}},
    "PROLATO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3827"]}},
    "PROLECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3828"]}},
    "PROLES": {"-": {"morpho": ["n-s---fn3i"], "uri": ["p3831"]}},
    "PROLETARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3833"]}},
    "PROLEVO": {"1": {"morpho": ["v1spia--1-"], "uri": ["p3834"]}},
    "PROLICEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p3836"]}},
    "PROLICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["p3837"]}},
    "PROLIXE": {"-": {"morpho": ["rp--------"], "uri": ["26381"]}},
    "PROLIXVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3844"]}},
    "PROLOGVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p3849"]}},
    "PROLONGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3850"]}},
    "PROLOQVOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["p3852"]}},
    "PROLVBIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p3854"]}},
    "PROLVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3856"]}},
    "PROLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3858"]}},
    "PROLVVIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["p3860"]}},
    "PROMEREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p3873"]}},
    "PROMEREOR": {"-": {"morpho": ["v1spid--2-"], "uri": ["132676"]}},
    "PROMERITVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["40757"]}},
    "PROMETHEVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "PROMINENS": {
        "1": {"morpho": ["n-s---nn3-"], "uri": ["123500"]},
        "2": {"morpho": ["aps---an3i"], "uri": ["100189"]},
    },
    "PROMINENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p3881"]}},
    "PROMINEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p3878"]}},
    "PROMISCVE": {"-": {"morpho": ["rp--------"], "uri": ["41778"]}},
    "PROMISCVVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "PROMISCVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3883"]}},
    "PROMISSVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p3896"]}},
    "PROMISSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["31796"]}},
    "PROMITTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3888"]}},
    "PROMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3890"]}},
    "PROMONTORIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["29458"]}},
    "PROMOVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p3895"]}},
    "PROMPTE": {"-": {"morpho": ["rp--------"], "uri": ["28790"]}},
    "PROMPTVARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2167"]}},
    "PROMPTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["p9918"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["26396"]},
    },
    "PROMVLGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3905"]}},
    "PROMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p3910"]}},
    "PROMVTVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3911"]}},
    "PRONAVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p3912"]}},
    "PRONE": {"-": {"morpho": ["rp--------"], "uri": ["50086"]}},
    "PRONEPOS": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p3916"]}},
    "PRONEPTIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["p3917"]}},
    "PRONVBA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p3925"]}},
    "PRONVBVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3925"]}},
    "PRONVNTIATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3928"]}},
    "PRONVNTIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3931"]}},
    "PRONVRVS": {"-": {"morpho": ["n-s---fn4-"], "uri": ["p3933"]}},
    "PRONVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3934"]}},
    "PROOEMIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p3938"]}},
    "PROPAGATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3939"]}},
    "PROPAGATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p3940"]}},
    "PROPAGO": {
        "1": {"morpho": ["n-s---fn3-"], "uri": ["p3946"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["p3945"]},
    },
    "PROPALAM": {"-": {"morpho": ["rp--------"], "uri": ["p3948"]}},
    "PROPATVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["52105"]}},
    "PROPATVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3956"]}},
    "PROPE": {
        "1": {"morpho": ["rp--------"], "uri": ["26407"]},
        "2": {"morpho": ["p---------"], "uri": ["p3957"]},
    },
    "PROPEDIEM": {"-": {"morpho": ["rp--------"], "uri": ["58175"]}},
    "PROPELLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3959"]}},
    "PROPEMODVM": {"-": {"morpho": ["rp--------"], "uri": ["p3960"]}},
    "PROPENDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p3962"]}},
    "PROPENSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3964"]}},
    "PROPENSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["40340"]}},
    "PROPERANS": {"-": {"morpho": ["aps---an3i"], "uri": ["100198"]}},
    "PROPERANTER": {"-": {"morpho": ["rp--------"], "uri": ["47017"]}},
    "PROPERANTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p9858"]}},
    "PROPERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3967"]}},
    "PROPERATO": {"-": {"morpho": ["rp--------"], "uri": ["46079"]}},
    "PROPERATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["53358"]}},
    "PROPERE": {"-": {"morpho": ["rp--------"], "uri": ["51997"]}},
    "PROPERIPES": {"-": {"morpho": ["aps---an3-"], "uri": ["p3968"]}},
    "PROPERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3969"]}},
    "PROPERVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3971"]}},
    "PROPINATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3987"]}},
    "PROPINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3989"]}},
    "PROPINQVI": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "PROPINQVITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3990"]}},
    "PROPINQVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3991"]}},
    "PROPINQVVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["100200"]}},
    "PROPINQVVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["p3992"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "PROPIOR": {"-": {"morpho": ["acs---cn3-"], "uri": ["p3994"]}},
    "PROPITIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p4000"]}},
    "PROPITIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p4001"]}},
    "PROPIVS": {"-": {"morpho": ["p---------"], "uri": []}},
    "PROPNIGEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p4003"]}},
    "PROPOLA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["p4689"]}},
    "PROPOLLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p4006"]}},
    "PROPONO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p4008"]}},
    "PROPORRO": {"-": {"morpho": ["rp--------"], "uri": ["p4009"]}},
    "PROPORTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p4010"]}},
    "PROPOSITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p4016"]}},
    "PROPOSITVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["26426"]}},
    "PROPRAETOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p4019"]}},
    "PROPRIATIM": {"-": {"morpho": ["rp--------"], "uri": ["p4020"]}},
    "PROPRIE": {"-": {"morpho": ["rp--------"], "uri": ["26427"]}},
    "PROPRIETAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p4022"]}},
    "PROPRITIM": {"-": {"morpho": ["rp--------"], "uri": ["p4024"]}},
    "PROPRIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["100205"]}},
    "PROPRIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p4025"]}},
    "PROPTER": {
        "1": {"morpho": ["rp--------"], "uri": []},
        "2": {"morpho": ["p---------"], "uri": ["p4027"]},
    },
    "PROPTEREA": {"-": {"morpho": ["rp--------"], "uri": ["p5011"]}},
    "PROPVGNACVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p4035"]}},
    "PROPVGNATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p4037"]}},
    "PROPVGNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p4039"]}},
    "PROPVLSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p4042"]}},
    "PROQVAESTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p4048"]}},
    "PROQVAM": {"-": {"morpho": ["rp--------"], "uri": ["45298"]}},
    "PRORA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p4051"]}},
    "PROREPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p4052"]}},
    "PRORIPIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["p4056"]}},
    "PRORIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["p4051"]}},
    "PRORITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p4057"]}},
    "PROROGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p4061"]}},
    "PRORSVM": {"-": {"morpho": ["rp--------"], "uri": ["26436"]}},
    "PRORSVS": {
        "1": {"morpho": ["aps---mn1-"], "uri": ["p4049"]},
        "2": {"morpho": ["rp--------"], "uri": ["26437"]},
    },
    "PRORVMPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p4066"]}},
    "PRORVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p4067"]}},
    "PROSAPIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p4074"]}},
    "PROSCAENIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p4076"]}},
    "PROSCENIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["34222"]}},
    "PROSCINDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p4078"]}},
    "PROSCRIBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p4080"]}},
    "PROSCRIPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p4081"]}},
    "PROSECO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p4084"]}},
    "PROSEMINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p4095"]}},
    "PROSEQVOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["p4098"]}},
    "PROSERO": {
        "2": {"morpho": ["v1spia--3-"], "uri": ["p4100"]},
        "3": {"morpho": ["v1spia--3-"], "uri": ["p4100"]},
    },
    "PROSERPINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PROSEVCHA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p4104"]}},
    "PROSILIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["p4107"]}},
    "PROSITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p4109"]}},
    "PROSLAMBANOMENOS": {"-": {"morpho": ["n-s---mn2g"], "uri": ["p4111"]}},
    "PROSOCER": {"-": {"morpho": ["n-s---mn2r"], "uri": ["p4112"]}},
    "PROSPECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p4121"]}},
    "PROSPECTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["p9922"]}},
    "PROSPECVLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["p4123"]}},
    "PROSPER": {"-": {"morpho": ["aps---mn1r"], "uri": ["p4124"]}},
    "PROSPERA": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "PROSPERE": {"-": {"morpho": ["rp--------"], "uri": ["28512"]}},
    "PROSPERITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p4125"]}},
    "PROSPERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p4126"]}},
    "PROSPICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["p4129"]}},
    "PROSTAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p4134"]}},
    "PROSTERNO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p4135"]}},
    "PROSTITVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p4139"]}},
    "PROSTITVTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p9870"]}},
    "PROSTITVTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["100209"]}},
    "PROSTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p4142"]}},
    "PROSTYLOS": {"-": {"morpho": ["aps---mn1g"], "uri": ["p4147"]}},
    "PROSVBIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p4149"]}},
    "PROSVM": {"1": {"morpho": ["v1spia--4-"], "uri": ["p4150"]}},
    "PROTECTVRA": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "PROTEGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p4160"]}},
    "PROTELVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p4163"]}},
    "PROTENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p4164"]}},
    "PROTENTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["123759"]}},
    "PROTERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p4168"]}},
    "PROTERREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p4169"]}},
    "PROTERVE": {"-": {"morpho": ["rp--------"], "uri": ["40366"]}},
    "PROTERVITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p4172"]}},
    "PROTERVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p4173"]}},
    "PROTHYRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p4179"]}},
    "PROTINVS": {"-": {"morpho": ["rp--------"], "uri": ["p4181"]}},
    "PROTOLLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p4183"]}},
    "PROTRAHO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p4196"]}},
    "PROTRITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["26469"]}},
    "PROTROPVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p4199"]}},
    "PROTRVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p4200"]}},
    "PROTVRBO": {"2": {"morpho": ["v1spia--1-"], "uri": ["p4203"]}},
    "PROVECTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["p9925"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["26473"]},
    },
    "PROVEHO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p4208"]}},
    "PROVENIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["p4210"]}},
    "PROVENTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["p9926"]}},
    "PROVERBIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p4212"]}},
    "PROVERTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["123810"]}},
    "PROVIDENS": {"-": {"morpho": ["aps---an3i"], "uri": ["100211"]}},
    "PROVIDENTER": {"-": {"morpho": ["rp--------"], "uri": ["51896"]}},
    "PROVIDENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p9860"]}},
    "PROVIDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p4214"]}},
    "PROVIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p4215"]}},
    "PROVINCIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p4216"]}},
    "PROVINCIALIS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["p4217"]},
        "2": {"morpho": ["aps---cn3i"], "uri": ["p4217"]},
    },
    "PROVINCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["30898"]}},
    "PROVISIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p4220"]}},
    "PROVISO": {
        "1": {"morpho": ["v1spia--3-"], "uri": ["p4221"]},
        "2": {"morpho": ["rp--------"], "uri": ["123814"]},
    },
    "PROVISOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p4222"]}},
    "PROVISVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["p9927"]}},
    "PROVIVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p4223"]}},
    "PROVOCATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p4227"]}},
    "PROVOCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p4232"]}},
    "PROVOLO": {"2": {"morpho": ["v1spia--1-"], "uri": ["p4233"]}},
    "PROVOLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p4234"]}},
    "PROVOMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p4235"]}},
    "PROVT": {"-": {"morpho": ["rp--------"], "uri": ["33375"]}},
    "PROXENETA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["p4240"]}},
    "PROXIME": {"-": {"morpho": ["p---------"], "uri": []}},
    "PROXIMITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p4243"]}},
    "PROXIMVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["100215"]}},
    "PRVDENS": {"-": {"morpho": ["aps---an3i"], "uri": ["p4246"]}},
    "PRVDENTER": {"-": {"morpho": ["rp--------"], "uri": ["26489"]}},
    "PRVDENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p4247"]}},
    "PRVINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p4251"]}},
    "PRVINOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p4252"]}},
    "PRVNA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p4253"]}},
    "PRVNVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p4256"]}},
    "PRVNVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["p4256"]}},
    "PRVRIGO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p4258"]}},
    "PRVRIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["p4259"]}},
    "PRYTANEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p4264"]}},
    "PRYTANIS": {
        "-": {"morpho": ["n-s---mn3i"], "uri": ["p4265"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "PSALLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p4267"]}},
    "PSALTRIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p4276"]}},
    "PSEVDOTHYRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p4318"]}},
    "PSITTACVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p4333"]}},
    "PSYLLVS": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "PSYTHIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["57655"]}},
    "PTEROMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["p4353"]}},
    "PTERYGOMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["p4356"]}},
    "PTISANARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["123942"]}},
    "PTOLEMAEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PVBENS": {"-": {"morpho": ["aps---an3i"], "uri": ["p4364"]}},
    "PVBERTAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p4366"]}},
    "PVBES": {
        "1": {"morpho": ["n-s---fn3i"], "uri": ["p4368"]},
        "2": {"morpho": ["aps---an3-"], "uri": ["54925"]},
    },
    "PVBESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p4369"]}},
    "PVBLICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p1917"]}},
    "PVBLICANVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["p4370"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["p4370"]},
    },
    "PVBLICATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p4372"]}},
    "PVBLICE": {"-": {"morpho": ["rp--------"], "uri": ["45222"]}},
    "PVBLICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p4376"]}},
    "PVBLICOLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PVBLICVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p1964"]}},
    "PVBLICVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["p1453"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["p4377"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["p1453"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["p4377"]},
    },
    "PVBLILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PVBLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PVDENDA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["100226"]}},
    "PVDENDVM": {"-": {"morpho": ["----------"], "uri": []}},
    "PVDENDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["51213"]}},
    "PVDENS": {
        "-": {"morpho": ["aps---an3i"], "uri": ["100227"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "PVDENTER": {"-": {"morpho": ["rp--------"], "uri": ["48784"]}},
    "PVDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p4380"]}},
    "PVDIBVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p4382"]}},
    "PVDICE": {"-": {"morpho": ["rp--------"], "uri": ["28603"]}},
    "PVDICITIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["p4383"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "PVDICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p4384"]}},
    "PVDOR": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["p4385"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "PVELLA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["p2919"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "PVELLARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p4389"]}},
    "PVELLVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p4394"]}},
    "PVELLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p3224"]}},
    "PVER": {"-": {"morpho": ["n-s---mn2r"], "uri": ["p2247"]}},
    "PVERILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p4400"]}},
    "PVERILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p4401"]}},
    "PVERILITER": {"-": {"morpho": ["rp--------"], "uri": ["55736"]}},
    "PVERITIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p3289"]}},
    "PVERPERA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p4408"]}},
    "PVERPERIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p4407"]}},
    "PVERPERVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p4408"]}},
    "PVERVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p4409"]}},
    "PVGA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p4410"]}},
    "PVGIL": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p4411"]}},
    "PVGILLAR": {"-": {"morpho": ["n-s---nn3i"], "uri": ["p4416"]}},
    "PVGILLARES": {"-": {"morpho": ["n-p---mn3i"], "uri": ["100234"]}},
    "PVGILLARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p4416"]}},
    "PVGILLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p4419"]}},
    "PVGIO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p4420"]}},
    "PVGNA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p4422"]}},
    "PVGNACITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p4424"]}},
    "PVGNACITER": {"-": {"morpho": ["rp--------"], "uri": ["33205"]}},
    "PVGNANS": {"-": {"morpho": ["aps---an3i"], "uri": ["123989"]}},
    "PVGNAX": {"-": {"morpho": ["aps---an3i"], "uri": ["p4429"]}},
    "PVGNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p4432"]}},
    "PVGNVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p4433"]}},
    "PVLCHER": {
        "-": {"morpho": ["aps---mn1r"], "uri": ["p4435"]},
        "C": {"morpho": ["n-s----n2-"], "uri": []},
        "N": {"morpho": ["n-s---mn2r"], "uri": []},
    },
    "PVLCHRE": {"-": {"morpho": ["rp--------"], "uri": ["26522"]}},
    "PVLCHRITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p4439"]}},
    "PVLEX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["26524"]}},
    "PVLLARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["38119"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["p4446"]},
    },
    "PVLLATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p4449"]}},
    "PVLLVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p4461"]}},
    "PVLLVM": {
        "2": {"morpho": ["n-s---nn2-"], "uri": ["p4459"]},
        "-": {"morpho": ["n-s---nn2-"], "uri": ["p4459"]},
    },
    "PVLLVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["p4464"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["p4691"]},
    },
    "PVLMENTARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p4467"]}},
    "PVLMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p4468"]}},
    "PVLMO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p4469"]}},
    "PVLPA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p4692"]}},
    "PVLPITVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p4478"]}},
    "PVLS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["p4481"]}},
    "PVLSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p4486"]}},
    "PVLSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["p0758"]}},
    "PVLTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p4494"]}},
    "PVLVEREVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p4498"]}},
    "PVLVERVLENTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p4501"]}},
    "PVLVILLVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": ["p4503"]},
        "-": {"morpho": ["n-s----n2-"], "uri": []},
    },
    "PVLVINAR": {"-": {"morpho": ["n-s---nn3-"], "uri": ["p2250"]}},
    "PVLVINATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p4506"]}},
    "PVLVINVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p4509"]}},
    "PVLVIS": {"-": {"morpho": ["n-s---mn3-"], "uri": ["26532"]}},
    "PVLVISCVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p4511"]}},
    "PVMEX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p4512"]}},
    "PVMICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p4514"]}},
    "PVMICOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p4515"]}},
    "PVMILIO": {"-": {"morpho": ["n-s---cn3-"], "uri": ["p4517"]}},
    "PVNCTIVNCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p4525"]}},
    "PVNCTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p9871"]}},
    "PVNCTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["p9871"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["100245"]},
    },
    "PVNGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p4528"]}},
    "PVNICEVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["p4530"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["p4530"]},
    },
    "PVNICIVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "PVNICVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "PVNICVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "PVNIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["p4531"]}},
    "PVNIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["132681"]}},
    "PVPA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p2331"]}},
    "PVPILLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p4535"]}},
    "PVPILLARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p4536"]}},
    "PVPILLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p4535"]}},
    "PVPPIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p4540"]}},
    "PVPVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p4541"]}},
    "PVPVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p4541"]}},
    "PVRE": {"-": {"morpho": ["rp--------"], "uri": ["49663"]}},
    "PVRGAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p4546"]}},
    "PVRGATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p4547"]}},
    "PVRGATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["38630"]}},
    "PVRGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p4553"]}},
    "PVRITER": {"-": {"morpho": ["rp--------"], "uri": ["36954"]}},
    "PVRPVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p4564"]}},
    "PVRPVRATVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["100251"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["p4567"]},
    },
    "PVRPVREVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p4568"]}},
    "PVRVLENTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p4575"]}},
    "PVRVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p4576"]}},
    "PVS": {"-": {"morpho": ["n-s---nn3-"], "uri": ["p4577"]}},
    "PVSILLVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["49949"]},
        "2": {"morpho": ["rp--------"], "uri": []},
    },
    "PVSILLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p4585"]}},
    "PVSIO": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["p4586"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "PVSTVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["26551"]}},
    "PVSVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p4588"]}},
    "PVTAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["p4595"]}},
    "PVTATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p4598"]}},
    "PVTEAL": {
        "-": {"morpho": ["n-s---nn3i"], "uri": ["p4600"]},
        "N": {"morpho": ["n-s---nn3i"], "uri": ["p4600"]},
    },
    "PVTEALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p4601"]}},
    "PVTEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p4605"]}},
    "PVTEOLI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PVTESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p9948"]}},
    "PVTEVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p4606"]}},
    "PVTIDE": {"-": {"morpho": ["rp--------"], "uri": ["34414"]}},
    "PVTIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p4610"]}},
    "PVTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p4613"]}},
    "PVTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p4614"]}},
    "PVTREFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["p4618"]}},
    "PVTREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p4621"]}},
    "PVTRESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p4621"]}},
    "PVTRIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p4624"]}},
    "PVTRIS": {
        "-": {"morpho": ["aps---cn3i"], "uri": []},
        "A": {"morpho": ["aps---cn3i"], "uri": []},
    },
    "PYCNOSTYLOS": {"-": {"morpho": ["aps---mn1g"], "uri": ["p4633"]}},
    "PYGARGVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p4637"]}},
    "PYLAEMENES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "PYRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p4642"]}},
    "PYRAGRO": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "PYRAMIS": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["p4645"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "PYRENAEVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "PYRETHRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p4649"]}},
    "PYROPVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p4657"]}},
    "PYTHAGORAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "PYTHAVLA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["124194"]}},
    "PYTHIA": {
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
        "O": {"morpho": ["n-s----n1-"], "uri": []},
    },
    "PYTHICVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "PYTISMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["p4672"]}},
    "PYXIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p4679"]}},
    "QVA": {
        "1": {"morpho": ["r---------"], "uri": []},
        "2": {"morpho": ["r---------"], "uri": []},
        "3": {"morpho": ["rp--------"], "uri": []},
    },
    "QVACVMQVE": {
        "1": {"morpho": ["rp--------"], "uri": ["55346"]},
        "2": {"morpho": ["rp--------"], "uri": ["55346"]},
    },
    "QVADRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["q0004"]}},
    "QVADRAGENARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["q0005"]}},
    "QVADRAGENI": {"-": {"morpho": ["rp--------"], "uri": []}},
    "QVADRAGESIMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["q0193"]}},
    "QVADRAGESIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["q0080"]}},
    "QVADRAGIES": {"-": {"morpho": ["rp--------"], "uri": []}},
    "QVADRAGINTA": {"-": {"morpho": ["----------"], "uri": []}},
    "QVADRANS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["100266"]}},
    "QVADRANTAL": {"-": {"morpho": ["n-s---nn3i"], "uri": ["q0013"]}},
    "QVADRANTARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["q0015"]}},
    "QVADRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["q0018"]}},
    "QVADRATVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["q9330"]}},
    "QVADRATVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["q9330"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["58105"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["q9330"]},
    },
    "QVADRIANGVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["q0245"]}},
    "QVADRIDENS": {"-": {"morpho": ["aps---an3i"], "uri": ["q0024"]}},
    "QVADRIDVVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["q0026"]}},
    "QVADRIENNIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["q0028"]}},
    "QVADRIFARIAM": {"-": {"morpho": ["rp--------"], "uri": ["q0030"]}},
    "QVADRIFIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["q0032"]}},
    "QVADRIFLVVIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["q0036"]}},
    "QVADRIGA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["q0040"]}},
    "QVADRIIVGI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["124251"]}},
    "QVADRIIVGIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["q0049"]}},
    "QVADRIIVGVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["q0049"]}},
    "QVADRIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["q0058"]}},
    "QVADRINGENI": {
        "-": {"morpho": ["aps---an3-", "app---mn1-"], "uri": ["q0060"]}
    },
    "QVADRINGENTESIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["q0063"]}},
    "QVADRINGENTI": {"-": {"morpho": ["----------"], "uri": []}},
    "QVADRINGENTIES": {"-": {"morpho": ["rp--------"], "uri": ["44979"]}},
    "QVADRINI": {"-": {"morpho": ["rp--------"], "uri": []}},
    "QVADRIPARTITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["34690"]}},
    "QVADRIREMIS": {
        "1": {"morpho": ["n-s---fn3i"], "uri": ["q0071"]},
        "2": {"morpho": ["aps---cn3i"], "uri": ["124271"]},
    },
    "QVADRIVIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["q0316"]}},
    "QVADRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["q0078"]}},
    "QVADRVPEDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["q0083"]}},
    "QVADRVPES": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["100276"]},
        "2": {"morpho": ["aps---an3-"], "uri": ["q0084"]},
    },
    "QVADRVPLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["q0085"]}},
    "QVADRVPLEX": {
        "-": {
            "morpho": ["aps---an3-", "n-s---nn3-"],
            "uri": ["q0086", "124285"],
        }
    },
    "QVADRVPLO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["q0090"]},
        "2": {"morpho": ["rp--------"], "uri": []},
    },
    "QVADRVPLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["33083"]}},
    "QVAERITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["q0095"]}},
    "QVAERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["q0096"]}},
    "QVAESITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["q0098"]}},
    "QVAESITVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["q9327"]}},
    "QVAESITVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["q9327"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["26578"]},
    },
    "QVAESO": {"-": {"morpho": ["v1spia--3-"], "uri": ["q0099"]}},
    "QVAESTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["q0101"]}},
    "QVAESTIVNCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["q0104"]}},
    "QVAESTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["q0105"]}},
    "QVAESTORIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["q0107"]}},
    "QVAESTORIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["q0107"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["q0107"]},
    },
    "QVAESTVARIA": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "QVAESTVOSE": {"-": {"morpho": ["rp--------"], "uri": ["100278"]}},
    "QVAESTVOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["q0109"]}},
    "QVAESTVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["q0110"]}},
    "QVAESTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["q0111"]}},
    "QVALIBET": {"-": {"morpho": ["rp--------"], "uri": ["46849"]}},
    "QVALIS": {
        "1": {"morpho": ["aps---mn1-"], "uri": []},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "QVALISCVMQVE": {
        "1": {"morpho": ["aps---mn1-"], "uri": []},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "QVALITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["q0114"]}},
    "QVALITER": {
        "1": {"morpho": ["rp--------"], "uri": ["44962"]},
        "2": {"morpho": ["rp--------"], "uri": ["44962"]},
    },
    "QVALVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["47449"]}},
    "QVALVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["q0117"]}},
    "QVAM": {
        "1": {"morpho": ["rp--------"], "uri": ["26584"]},
        "2": {"morpho": ["rp--------"], "uri": ["26584"]},
    },
    "QVAMDIV": {
        "1": {"morpho": ["rp--------"], "uri": ["q0118"]},
        "2": {"morpho": ["rp--------"], "uri": ["q0118"]},
    },
    "QVAMDVDVM": {"-": {"morpho": ["rp--------"], "uri": ["124306"]}},
    "QVAMLIBET": {
        "1": {"morpho": ["rp--------"], "uri": ["42273"]},
        "2": {"morpho": ["rp--------"], "uri": ["42273"]},
    },
    "QVAMOBREM": {
        "1": {"morpho": ["rp--------"], "uri": ["55699"]},
        "2": {"morpho": ["rp--------"], "uri": ["55699"]},
    },
    "QVAMPLVRES": {
        "-": {"morpho": ["aps---an3-"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "QVAMPRIMVM": {"-": {"morpho": ["rp--------"], "uri": ["48468"]}},
    "QVAMQVAM": {
        "1": {"morpho": ["rp--------"], "uri": ["53806"]},
        "2": {"morpho": ["----------"], "uri": []},
        "3": {"morpho": ["rp--------"], "uri": ["53806"]},
    },
    "QVAMVIS": {
        "1": {"morpho": ["rp--------"], "uri": ["q9411"]},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "QVANAM": {"-": {"morpho": ["rp--------"], "uri": ["41128"]}},
    "QVANDANTENVS": {"-": {"morpho": ["rp--------"], "uri": []}},
    "QVANDO": {
        "1": {"morpho": ["rp--------"], "uri": ["q0127"]},
        "2": {"morpho": ["rp--------"], "uri": ["q0127"]},
        "3": {"morpho": ["----------"], "uri": []},
        "4": {"morpho": ["rp--------"], "uri": ["q0127"]},
    },
    "QVANDOCVMQVE": {
        "1": {"morpho": ["rp--------"], "uri": ["q0128"]},
        "2": {"morpho": ["rp--------"], "uri": ["q0128"]},
    },
    "QVANDOQVE": {
        "1": {"morpho": ["rp--------"], "uri": ["q0139"]},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "QVANDOQVIDEM": {"-": {"morpho": ["----------"], "uri": []}},
    "QVANTITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["q0131"]}},
    "QVANTO": {
        "1": {"morpho": ["rp--------"], "uri": ["44013"]},
        "2": {"morpho": ["rp--------"], "uri": ["44013"]},
    },
    "QVANTOPERE": {
        "1": {"morpho": ["rp--------"], "uri": ["q0133"]},
        "2": {"morpho": ["rp--------"], "uri": ["q0133"]},
    },
    "QVANTVLVM": {
        "1": {"morpho": ["aps---mn1-"], "uri": []},
        "2": {"morpho": ["----------"], "uri": []},
        "3": {"morpho": ["rp--------"], "uri": ["50987"]},
        "4": {"morpho": ["rp--------"], "uri": ["50987"]},
    },
    "QVANTVLVS": {
        "1": {"morpho": ["aps---mn1-"], "uri": ["53581"]},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "QVANTVLVSCVMQVE": {
        "1": {"morpho": ["aps---mn1-"], "uri": []},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "QVANTVLVSLIBET": {
        "1": {"morpho": ["aps---mn1-"], "uri": []},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "QVANTVM": {
        "1": {"morpho": ["aps---mn1-"], "uri": []},
        "2": {"morpho": ["----------"], "uri": []},
        "3": {"morpho": ["rp--------"], "uri": ["33926"]},
        "4": {"morpho": ["rp--------"], "uri": ["33926"]},
    },
    "QVANTVMCVMQVE": {"-": {"morpho": ["rp--------"], "uri": ["51856"]}},
    "QVANTVMLIBET": {
        "1": {"morpho": ["----------"], "uri": []},
        "2": {"morpho": ["rp--------"], "uri": ["33517"]},
    },
    "QVANTVMVIS": {
        "1": {"morpho": ["rp--------"], "uri": ["46081"]},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "QVANTVS": {
        "1": {"morpho": ["aps---mn1-"], "uri": ["34986"]},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "QVANTVSCVMQVE": {
        "1": {"morpho": ["aps---mn1-"], "uri": ["55959"]},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "QVANTVSLIBET": {"-": {"morpho": ["----------"], "uri": []}},
    "QVANTVSVIS": {"-": {"morpho": ["----------"], "uri": []}},
    "QVAPROPTER": {
        "1": {
            "morpho": ["rp--------", "rp--------"],
            "uri": ["q0137", "q0137"],
        },
        "2": {"morpho": ["rp--------"], "uri": ["q0137"]},
    },
    "QVAQVAVERSVM": {"-": {"morpho": ["rp--------"], "uri": ["q0138"]}},
    "QVAQVE": {"-": {"morpho": ["rp--------"], "uri": ["100280"]}},
    "QVARE": {
        "1": {"morpho": ["rp--------"], "uri": ["q9399"]},
        "2": {"morpho": ["rp--------"], "uri": ["q9399"]},
    },
    "QVARTADECIMANI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["100281"]}},
    "QVARTANA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["q0142"]}},
    "QVARTANVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["q0142"]}},
    "QVARTARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["q0145"]}},
    "QVARTO": {"-": {"morpho": ["rp--------"], "uri": ["32066"]}},
    "QVARTVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["q0148"]},
        "2": {"morpho": ["rp--------"], "uri": ["58045"]},
    },
    "QVARTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["q0148"]}},
    "QVASI": {
        "1": {"morpho": ["rp--------"], "uri": ["35040"]},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "QVASILLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["33059"]}},
    "QVASSATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["q0154"]}},
    "QVASSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["q0155"]}},
    "QVASSVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["q9328"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["40492"]},
    },
    "QVATENVS": {
        "1": {"morpho": ["rp--------"], "uri": ["q0157"]},
        "2": {"morpho": ["rp--------"], "uri": ["q0157"]},
        "3": {"morpho": ["----------"], "uri": []},
    },
    "QVATER": {"-": {"morpho": ["rp--------"], "uri": ["q0158"]}},
    "QVATERNI": {"-": {"morpho": ["rp--------"], "uri": []}},
    "QVATIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["q0164"]}},
    "QVATRIDVVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["35021"]}},
    "QVATVOR": {"-": {"morpho": ["----------"], "uri": []}},
    "QVATVORDECIM": {"-": {"morpho": ["----------"], "uri": []}},
    "QVATVORVIR": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "QVAVIS": {"-": {"morpho": ["rp--------"], "uri": ["100288"]}},
    "QVE": {"-": {"morpho": ["----------"], "uri": []}},
    "QVEMADMODVM": {
        "1": {"morpho": ["rp--------"], "uri": ["q9406"]},
        "2": {"morpho": ["rp--------"], "uri": ["q9406"]},
    },
    "QVEO": {"-": {"morpho": ["v1spia----"], "uri": ["q0178"]}},
    "QVERCANEVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "QVERCEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["q0179"]}},
    "QVERCVS": {"-": {"morpho": ["n-s---fn4-"], "uri": ["q0181"]}},
    "QVERELA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["q0182"]}},
    "QVERIMONIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["q0186"]}},
    "QVERITOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["q0189"]}},
    "QVERNEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["q0190"]}},
    "QVERNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["29237"]}},
    "QVEROR": {"-": {"morpho": ["v1spid--3-"], "uri": ["q0191"]}},
    "QVERQVETVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["q0194"]}},
    "QVERVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["q0196"]}},
    "QVESTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["q9329"]}},
    "QVI": {
        "1": {"morpho": ["aps---mn1-"], "uri": []},
        "2": {"morpho": ["rp--------"], "uri": ["55142"]},
        "3": {"morpho": ["rp--------"], "uri": ["55142"]},
        "4": {"morpho": ["rp--------"], "uri": ["55142"]},
    },
    "QVIA": {"-": {"morpho": ["----------"], "uri": []}},
    "QVIANAM": {"-": {"morpho": ["rp--------"], "uri": ["q0211"]}},
    "QVICVMQVE": {
        "1": {"morpho": ["aps---mn1-"], "uri": []},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "QVIDAM": {"-": {"morpho": ["----------"], "uri": []}},
    "QVIDEM": {"-": {"morpho": ["rp--------"], "uri": ["q0201"]}},
    "QVIDNI": {
        "-": {
            "morpho": ["rp--------", "rp--------"],
            "uri": ["q9408", "q9408"],
        }
    },
    "QVIES": {"-": {"morpho": ["n-s---fn3-"], "uri": ["q0205"]}},
    "QVIESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["q0206"]}},
    "QVIETE": {"-": {"morpho": ["rp--------"], "uri": ["26597"]}},
    "QVIETO": {"-": {"morpho": ["v1spia--1-"], "uri": ["q0209"]}},
    "QVIETVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["100290"]}},
    "QVIETVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["q0204"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "QVILIBET": {"-": {"morpho": ["----------"], "uri": []}},
    "QVIN": {
        "1": {"morpho": ["rp--------"], "uri": ["q0212"]},
        "2": {"morpho": ["----------"], "uri": []},
        "3": {
            "morpho": ["rp--------", "rp--------"],
            "uri": ["q0212", "q0212"],
        },
    },
    "QVINARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["q0213"]}},
    "QVINCVNX": {"-": {"morpho": ["n-s---mn3i"], "uri": ["q0216"]}},
    "QVINDECIES": {"-": {"morpho": ["rp--------"], "uri": ["100292"]}},
    "QVINDECIM": {"-": {"morpho": ["----------"], "uri": []}},
    "QVINDECIMVIRALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["q0227"]}},
    "QVINDECIMVIRI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["100293"]}},
    "QVINGENARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["q0231"]}},
    "QVINGENI": {"-": {"morpho": ["rp--------"], "uri": []}},
    "QVINGENTESIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["q0236"]}},
    "QVINGENTI": {"-": {"morpho": ["----------"], "uri": []}},
    "QVINI": {"-": {"morpho": ["rp--------"], "uri": []}},
    "QVINIDENI": {"-": {"morpho": ["rp--------"], "uri": []}},
    "QVINQVAGENARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["q0242"]}},
    "QVINQVAGENI": {"-": {"morpho": ["rp--------"], "uri": []}},
    "QVINQVAGESIMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["q0136"]}},
    "QVINQVAGESIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["q0122"]}},
    "QVINQVAGINTA": {"-": {"morpho": ["----------"], "uri": []}},
    "QVINQVATRIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["51241"]}},
    "QVINQVE": {"-": {"morpho": ["----------"], "uri": []}},
    "QVINQVENNALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["q0261"]}},
    "QVINQVENNIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["q0264"]}},
    "QVINQVENNIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["q0265"]}},
    "QVINQVEREMIS": {
        "1": {"morpho": ["n-s---fn3i"], "uri": ["100304"]},
        "2": {"morpho": ["aps---cn3i"], "uri": ["q0269"]},
        "-": {"morpho": ["n-s---fn3i"], "uri": ["100304"]},
    },
    "QVINQVEVIR": {"-": {"morpho": ["n-s---mn2r"], "uri": ["q0274"]}},
    "QVINQVIES": {"-": {"morpho": ["rp--------"], "uri": ["100305"]}},
    "QVINQVIPLICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["q0281"]}},
    "QVINTADECIMANI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["56349"]}},
    "QVINTANI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["100306"]}},
    "QVINTARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["q0287"]}},
    "QVINTILIS": {
        "A": {"morpho": ["aps---cn3i"], "uri": []},
        "N": {"morpho": ["n-s---fn3-"], "uri": []},
    },
    "QVINTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "QVINTVM": {"-": {"morpho": ["rp--------"], "uri": ["33355"]}},
    "QVINTVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["q0144"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "QVIPPE": {
        "1": {"morpho": ["rp--------"], "uri": ["q0290"]},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "QVIQVI": {"-": {"morpho": ["----------"], "uri": []}},
    "QVIRINALIS": {
        "N": {"morpho": ["aps---cn3i"], "uri": []},
        "A": {"morpho": ["aps---cn3i"], "uri": []},
    },
    "QVIRINVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "QVIRITES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "QVIRITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["q0294", "q0293"]}},
    "QVIRO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "QVIS": {
        "1": {"morpho": ["----------"], "uri": []},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "QVISNAM": {"-": {"morpho": ["----------"], "uri": []}},
    "QVISPIAM": {"-": {"morpho": ["----------"], "uri": []}},
    "QVISQVAM": {"-": {"morpho": ["----------"], "uri": []}},
    "QVISQVE": {
        "1": {"morpho": ["aps---mn1-"], "uri": []},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "QVISQVIS": {
        "1": {"morpho": ["aps---mn1-"], "uri": []},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "QVIVIS": {"-": {"morpho": ["----------"], "uri": []}},
    "QVIVISCVMQVE": {"-": {"morpho": ["----------"], "uri": []}},
    "QVO": {
        "1": {"morpho": ["rp--------"], "uri": ["33941"]},
        "2": {"morpho": ["rp--------"], "uri": ["33941"]},
        "3": {"morpho": ["rp--------"], "uri": ["33941"]},
        "4": {"morpho": ["rp--------"], "uri": ["33941"]},
        "5": {"morpho": ["----------", "rp--------"], "uri": ["33941"]},
    },
    "QVOAD": {
        "1": {"morpho": ["rp--------"], "uri": ["q0297"]},
        "2": {"morpho": ["rp--------"], "uri": ["q0297"]},
    },
    "QVOCIRCA": {"-": {"morpho": ["rp--------"], "uri": ["q0299"]}},
    "QVOCVMQVE": {
        "1": {"morpho": ["rp--------"], "uri": ["47460"]},
        "2": {"morpho": ["rp--------"], "uri": ["47460"]},
    },
    "QVOD": {
        "1": {"morpho": ["----------"], "uri": []},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "QVODAMMODO": {"-": {"morpho": ["rp--------"], "uri": ["31098"]}},
    "QVOLIBET": {"-": {"morpho": ["rp--------"], "uri": ["40514"]}},
    "QVOMINVS": {"-": {"morpho": ["----------"], "uri": []}},
    "QVOMODO": {
        "1": {"morpho": ["rp--------"], "uri": ["q0308"]},
        "2": {"morpho": ["rp--------"], "uri": ["q0308"]},
    },
    "QVOMODOCVMQVE": {
        "1": {"morpho": ["rp--------"], "uri": ["q0305"]},
        "2": {"morpho": ["rp--------"], "uri": ["q0305"]},
    },
    "QVONAM": {"-": {"morpho": ["rp--------"], "uri": ["100311"]}},
    "QVONDAM": {"-": {"morpho": ["rp--------"], "uri": ["q0310"]}},
    "QVONIAM": {"-": {"morpho": ["----------"], "uri": []}},
    "QVOQVAM": {"-": {"morpho": ["rp--------"], "uri": ["29583"]}},
    "QVOQVE": {"-": {"morpho": ["rp--------"], "uri": ["26601"]}},
    "QVOQVO": {"-": {"morpho": ["rp--------"], "uri": ["55748"]}},
    "QVOQVOVERSVS": {"-": {"morpho": ["rp--------"], "uri": ["q0313"]}},
    "QVORSVM": {"-": {"morpho": ["rp--------"], "uri": ["42962"]}},
    "QVOT": {
        "1": {"morpho": ["aps---mn1-"], "uri": []},
        "2": {"morpho": ["----------"], "uri": []},
        "3": {"morpho": ["----------"], "uri": []},
    },
    "QVOTANNIS": {"-": {"morpho": ["rp--------"], "uri": ["36771"]}},
    "QVOTCVMQVE": {"-": {"morpho": ["aps---mn1-"], "uri": []}},
    "QVOTIENS": {
        "1": {"morpho": ["rp--------"], "uri": ["q0320"]},
        "2": {"morpho": ["rp--------"], "uri": ["q0320"]},
    },
    "QVOTIENSCVMQVE": {"-": {"morpho": ["rp--------"], "uri": ["q0332"]}},
    "QVOTQVOT": {
        "1": {"morpho": ["aps---mn1-"], "uri": []},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "QVOTVIS": {"-": {"morpho": ["----------"], "uri": []}},
    "QVOTVS": {"-": {"morpho": ["----------"], "uri": []}},
    "QVOTVSCVMQVE": {"-": {"morpho": ["aps---mn1-"], "uri": []}},
    "QVOTVSQVISQVE": {"-": {"morpho": ["----------"], "uri": []}},
    "QVOVIS": {"-": {"morpho": ["rp--------"], "uri": ["52926"]}},
    "QVOVSQVE": {
        "-": {"morpho": ["rp--------"], "uri": ["q9404"]},
        "2": {"morpho": ["rp--------"], "uri": ["q9404"]},
    },
    "R": {"-": {"morpho": ["n-s----n--"], "uri": []}},
    "RABIDE": {"-": {"morpho": ["rp--------"], "uri": ["40009"]}},
    "RABIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0001"]}},
    "RABIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["r0002"]}},
    "RABIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0004"]}},
    "RABO": {"2": {"morpho": ["v1spia--3-"], "uri": ["100314"]}},
    "RACEMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["r0016"]}},
    "RACILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "RADIANS": {"-": {"morpho": ["aps---an3i"], "uri": ["100315"]}},
    "RADIATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r9018"]}},
    "RADICESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r0964"]}},
    "RADICITVS": {"-": {"morpho": ["rp--------"], "uri": ["r0021"]}},
    "RADICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r0023"]}},
    "RADIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r0085"]}},
    "RADIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["r0026"]}},
    "RADIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0027"]}},
    "RADO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r0966"]}},
    "RAECIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "RAEDA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r0029"]}},
    "RAMALE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["r0034"]}},
    "RAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["r0035"]}},
    "RAMEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0037"]}},
    "RAMEX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["50237"]}},
    "RAMNENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "RAMNES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "RAMOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0040"]}},
    "RAMVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["r0042"]}},
    "RAMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["r0043"]}},
    "RANA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r0045"]}},
    "RANCENS": {"-": {"morpho": ["aps---an3i"], "uri": ["100321"]}},
    "RANCIDVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0046"]}},
    "RANCIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0047"]}},
    "RAPACITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0053"]}},
    "RAPAX": {
        "-": {"morpho": ["aps---an3i"], "uri": ["r0055"]},
        "A": {"morpho": ["n-s---cn3-"], "uri": ["124533"]},
    },
    "RAPHANVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["r0058"]}},
    "RAPICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0059"]}},
    "RAPIDE": {"-": {"morpho": ["rp--------"], "uri": ["31711"]}},
    "RAPIDITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0060"]}},
    "RAPIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0062"]}},
    "RAPINA": {
        "1": {"morpho": ["n-s---fn1-"], "uri": ["r0005", "r0063"]},
        "2": {"morpho": ["n-s---fn1-"], "uri": ["r0005", "r0063"]},
    },
    "RAPIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["r0970"]}},
    "RAPSO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "RAPTIM": {"-": {"morpho": ["rp--------"], "uri": ["r0071"]}},
    "RAPTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r0973"]}},
    "RAPTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["r0073"]}},
    "RAPTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["r0957"]}},
    "RAPTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["r0957"]}},
    "RAPVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["r0076"]}},
    "RAPVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["r0051"]}},
    "RARE": {"-": {"morpho": ["rp--------"], "uri": ["26624"]}},
    "RAREFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["r0974"]}},
    "RAREFIO": {"-": {"morpho": ["v1spia--3-"], "uri": [None, "r0975"]}},
    "RARENTER": {"-": {"morpho": ["rp--------"], "uri": ["r0078"]}},
    "RARESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r0976"]}},
    "RARITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0080"]}},
    "RARO": {"-": {"morpho": ["rp--------"], "uri": ["26626"]}},
    "RARVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0082"]}},
    "RASILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["r0084"]}},
    "RASTER": {"-": {"morpho": ["n-s---mn2r"], "uri": ["57779"]}},
    "RASTRI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["100327"]}},
    "RASTRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["r0091"]}},
    "RATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0095"]}},
    "RATIOCINATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0097"]}},
    "RATIOCINATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["r0099"]}},
    "RATIOCINO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "RATIOCINOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["r0978"]}},
    "RATIONALIS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["124578"]},
        "2": {"morpho": ["aps---cn3i"], "uri": ["r0103"]},
    },
    "RATIONALITER": {"-": {"morpho": ["rp--------"], "uri": ["100329"]}},
    "RATIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["r0283"]}},
    "RATIVNCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r0109"]}},
    "RATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["26640"]}},
    "RAVCISONVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0114"]}},
    "RAVCVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0110"]}},
    "RAVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0398"]}},
    "REA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r0643"]}},
    "REALIS": {"-": {"morpho": ["aps---cn3-"], "uri": ["53967"]}},
    "REAPSE": {"-": {"morpho": ["rp--------"], "uri": ["38576"]}},
    "REBELLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0130"]}},
    "REBELLIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0132"]}},
    "REBELLIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["r0133"]}},
    "REBELLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r0984"]}},
    "REBOO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r0986"]}},
    "RECALCITRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r0988"]}},
    "RECALESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r0991"]}},
    "RECALFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["r0990"]}},
    "RECANTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r0992"]}},
    "RECEDO": {"1": {"morpho": ["v1spia--3-"], "uri": ["r0997"]}},
    "RECELLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r0999"]}},
    "RECENS": {
        "1": {"morpho": ["aps---an3i"], "uri": ["r0144"]},
        "2": {"morpho": ["rp--------"], "uri": ["47634"]},
    },
    "RECENSEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["r1001"]}},
    "RECEPIO": {"-": {"morpho": ["v1spia--3i"], "uri": []}},
    "RECEPTACVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["r0149"]}},
    "RECEPTICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0154"]}},
    "RECEPTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1002"]}},
    "RECEPTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["r0156"]}},
    "RECEPTRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0158"]}},
    "RECEPTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["52591"]}},
    "RECEPTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["r7016"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["100338"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "RECESSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0160"]}},
    "RECESSVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["r7012"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["124660"]},
    },
    "RECHAMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["r0162"]}},
    "RECIDIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0164"]}},
    "RECIDO": {
        "1": {"morpho": ["v1spia--3-"], "uri": ["r1005", "r1004", "r1004"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["r1005", "r1004", "r1004"]},
    },
    "RECINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1008"]}},
    "RECINO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r0994"]}},
    "RECIPIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["r1009"]}},
    "RECIPROCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1010"]}},
    "RECIPROCVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0171"]}},
    "RECISVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["100341"]}},
    "RECITATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0175"]}},
    "RECITATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["r9175"]}},
    "RECITO": {"1": {"morpho": ["v1spia--1-"], "uri": ["r1011"]}},
    "RECLAMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1013"]}},
    "RECLINIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["r0180"]}},
    "RECLINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1015"]}},
    "RECLVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1016"]}},
    "RECOGITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1018"]}},
    "RECOGNITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0184"]}},
    "RECOGNOSCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1019"]}},
    "RECOLLIGO": {
        "-": {"morpho": ["v1spia--3-"], "uri": ["r1021"]},
        "3": {"morpho": ["v1spia--3-"], "uri": ["r1021"]},
    },
    "RECOLO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["r1024"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["r1025"]},
    },
    "RECOMPONO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1031"]}},
    "RECONCILIATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0186"]}},
    "RECONCILIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1032"]}},
    "RECONDITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["52965"]}},
    "RECONDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1035"]}},
    "RECONFLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1037"]}},
    "RECOQVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1043"]}},
    "RECORDATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0193"]}},
    "RECORDOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["r1047"]}},
    "RECORRIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1049"]}},
    "RECREO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1052"]}},
    "RECREPO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1053"]}},
    "RECRESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1054"]}},
    "RECRVDESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1056"]}},
    "RECTA": {"-": {"morpho": ["rp--------"], "uri": ["26671"]}},
    "RECTE": {"-": {"morpho": ["rp--------"], "uri": ["26672"]}},
    "RECTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["r0209"]}},
    "RECTRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0210"]}},
    "RECTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["38704"]}},
    "RECTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["26675"]}},
    "RECVBO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1057"]}},
    "RECVMBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1058"]}},
    "RECVPERATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["45505"]}},
    "RECVPERATORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["32656"]}},
    "RECVPERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["26678"]}},
    "RECVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1059"]}},
    "RECVRRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1060"]}},
    "RECVRSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1061"]}},
    "RECVRSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["r7019"]}},
    "RECVRVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1062"]}},
    "RECVRVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0216"]}},
    "RECVSATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0218"]}},
    "RECVSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1063"]}},
    "RECVTIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["r1064"]}},
    "RECVTITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0220"]}},
    "REDAMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1070"]}},
    "REDARDESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1073"]}},
    "REDARGVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1074"]}},
    "REDDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1075"]}},
    "REDDVCO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "REDEMPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0227"]}},
    "REDEMPTITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1076"]}},
    "REDEMPTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["r0228"]}},
    "REDEO": {"1": {"morpho": ["v1spia--4-"], "uri": ["r1080"]}},
    "REDHALO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1081"]}},
    "REDIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1085"]}},
    "REDIMICVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["r0234"]}},
    "REDIMIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["r1086"]}},
    "REDIMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1087"]}},
    "REDINTEGRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1088"]}},
    "REDITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0238"]}},
    "REDITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["r7023"]}},
    "REDIVIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0239"]}},
    "REDOLEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["r1092"]}},
    "REDONO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1093"]}},
    "REDVCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1098"]}},
    "REDVCTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0244"]}},
    "REDVCTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["56595"]}},
    "REDVNDANTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r9991"]}},
    "REDVNDATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0248"]}},
    "REDVNDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1101"]}},
    "REDVRESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1102"]}},
    "REDVX": {"-": {"morpho": ["aps---an3-"], "uri": ["r0252"]}},
    "REFECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0253"]}},
    "REFELLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1107"]}},
    "REFERCIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["r1109"]}},
    "REFERIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["r1110"]}},
    "REFERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1111"]}},
    "REFERT": {"-": {"morpho": ["v1spia--2-"], "uri": ["r1112"]}},
    "REFERTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["55475"]}},
    "REFICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["r1117"]}},
    "REFIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1118"]}},
    "REFINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1120"]}},
    "REFLAGITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1122"]}},
    "REFLECTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1124"]}},
    "REFLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1125"]}},
    "REFLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1128"]}},
    "REFLVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0261"]}},
    "REFORMATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0262"]}},
    "REFORMIDO": {"2": {"morpho": ["v1spia--1-"], "uri": ["r1131"]}},
    "REFORMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1132"]}},
    "REFOVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["r1133"]}},
    "REFRACTARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0266"]}},
    "REFRAGOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["r1134"]}},
    "REFRENATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0271"]}},
    "REFRENO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1135"]}},
    "REFRICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1136"]}},
    "REFRIGERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0273"]}},
    "REFRIGERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1137"]}},
    "REFRIGESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1138"]}},
    "REFRINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1139"]}},
    "REFVGIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["r1141"]}},
    "REFVGVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["57678"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["r0281"]},
    },
    "REFVLGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["r1144"]}},
    "REFVNDO": {"2": {"morpho": ["v1spia--3-"], "uri": ["r1145"]}},
    "REFVTATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["r7029"]}},
    "REFVTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1146"]}},
    "REGALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["r0292"]}},
    "REGELO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1147"]}},
    "REGENS": {"-": {"morpho": ["n-s---cn3-"], "uri": ["48969"]}},
    "REGERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1151"]}},
    "REGIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["r0310"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "REGIE": {"-": {"morpho": ["rp--------"], "uri": ["36812"]}},
    "REGIFICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0300"]}},
    "REGIGNO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1152"]}},
    "REGILLVS": {
        "1N": {"morpho": ["n-s---mn2-"], "uri": ["124904"]},
        "2N": {"morpho": ["n-s---mn2-"], "uri": ["124904"]},
        "1": {"morpho": ["n-s---mn2-"], "uri": ["124904"]},
        "2": {"morpho": ["n-s---mn2-"], "uri": ["124904"]},
    },
    "REGIMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["r0303"]}},
    "REGINA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["r0306"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "REGINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "REGIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0307"]}},
    "REGIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "REGIVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["r0310"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "REGLVTINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1154"]}},
    "REGNATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["r0311"]}},
    "REGNATRIX": {"-": {"morpho": ["aps---an3-"], "uri": ["56295"]}},
    "REGNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1155"]}},
    "REGNVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["r0314"]}},
    "REGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1156"]}},
    "REGREDIOR": {"-": {"morpho": ["v1spid--3i"], "uri": ["r1158"]}},
    "REGRESSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["r7030"]}},
    "REGVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r0320"]}},
    "REGVLVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": ["r0320"]},
        "-": {"morpho": ["n-s---mn2-"], "uri": ["r0320"]},
    },
    "REGVSTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1160"]}},
    "REICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["r1162"]}},
    "REICVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0322"]}},
    "REIECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0325"]}},
    "REIECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1163"]}},
    "RELABOR": {"2": {"morpho": ["v1spid--3-"], "uri": ["r1166"]}},
    "RELANGVESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1168"]}},
    "RELATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0327"]}},
    "RELATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["r7032"]}},
    "RELAXO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1170"]}},
    "RELEGATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0334"]}},
    "RELEGO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["r1171"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["r1172"]},
    },
    "RELENTESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1173"]}},
    "RELEVATVS": {
        "C": {"morpho": ["aps---mn1-"], "uri": []},
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "RELEVO": {"1": {"morpho": ["v1spia--1-"], "uri": ["r1175"]}},
    "RELICTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0338"]}},
    "RELICTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["r7033"]}},
    "RELICTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["r7033"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["58076"]},
    },
    "RELIGIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0343"]}},
    "RELIGIOSE": {"-": {"morpho": ["rp--------"], "uri": ["26740"]}},
    "RELIGIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0346"]}},
    "RELIGO": {"2": {"morpho": ["v1spia--1-"], "uri": ["r1178"]}},
    "RELINO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1180"]}},
    "RELINQVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1181"]}},
    "RELIQVA": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "RELIQVI": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "RELIQVIAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["43584"]}},
    "RELIQVVM": {"-": {"morpho": ["n-s---nn2-"], "uri": []}},
    "RELIQVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0123"]}},
    "RELVCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["r1185"]}},
    "RELVCTOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["r1186"]}},
    "REMANEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["r1197"]}},
    "REMANO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1198"]}},
    "REMEDIABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["r0362"]}},
    "REMEDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["r0366"]}},
    "REMEO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1203"]}},
    "REMETIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["r1205"]}},
    "REMEX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["r0369"]}},
    "REMIGIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["r0372"]}},
    "REMIGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1206"]}},
    "REMIGRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1207"]}},
    "REMINISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["125015"]}},
    "REMINISCOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["r1208"]}},
    "REMISCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["r9375"]}},
    "REMISSARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0376"]}},
    "REMISSE": {"-": {"morpho": ["rp--------"], "uri": ["53148"]}},
    "REMISSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0378"]}},
    "REMISSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["26749"]}},
    "REMITTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1210"]}},
    "REMOLIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["r1211"]}},
    "REMOLLESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r8212"]}},
    "REMOLLIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["r1212"]}},
    "REMORDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["r1215"]}},
    "REMOROR": {"1": {"morpho": ["v1spid--1-"], "uri": ["r1216"]}},
    "REMOTA": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "REMOTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["51744"]}},
    "REMOVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["r1217"]}},
    "REMVGIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["r1218"]}},
    "REMVLCVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["r9389"]}},
    "REMVNERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0391"]}},
    "REMVNERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["47728"]}},
    "REMVNEROR": {"-": {"morpho": ["v1spid--1-"], "uri": ["r1232"]}},
    "REMVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["r0394"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["r0394"]},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "RENARRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1235"]}},
    "RENASCOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["r1236"]}},
    "RENAVIGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1238"]}},
    "RENES": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "RENIDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["r1241"]}},
    "RENIDESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1241"]}},
    "RENO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1244"]}},
    "RENODO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1245"]}},
    "RENOVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1249"]}},
    "RENVMERO": {"1": {"morpho": ["v1spia--1-"], "uri": ["r1252"]}},
    "RENVNTIATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0409"]}},
    "RENVNTIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1253"]}},
    "RENVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1254"]}},
    "RENVTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1255"]}},
    "REOR": {"-": {"morpho": ["v1spid--2-"], "uri": ["r1257"]}},
    "REPAGVLA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["100379"]}},
    "REPARABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["r0417"]}},
    "REPARCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1261"]}},
    "REPARO": {"2": {"morpho": ["v1spia--1-"], "uri": ["r1262"]}},
    "REPASTINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1264"]}},
    "REPECTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1268"]}},
    "REPELLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1270"]}},
    "REPENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1271"]}},
    "REPENS": {
        "1": {"morpho": ["aps---an3i"], "uri": ["r7054"]},
        "2": {"morpho": ["rp--------"], "uri": ["125094"]},
    },
    "REPENSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1272"]}},
    "REPENTE": {"-": {"morpho": ["rp--------"], "uri": ["49540"]}},
    "REPENTINO": {"-": {"morpho": ["rp--------"], "uri": ["52470"]}},
    "REPENTINVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["r0429"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "REPERCVSSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["r7036"]}},
    "REPERCVTIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["r1273"]}},
    "REPERIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["r1274"]}},
    "REPERTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["r0434"]}},
    "REPERTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["r7037"]}},
    "REPERTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["r7037"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["125111"]},
    },
    "REPETENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r9993"]}},
    "REPETITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0438"]}},
    "REPETITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["r0439"]}},
    "REPETO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1275"]}},
    "REPETVNDAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["100387"]}},
    "REPLEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["r1281"]}},
    "REPLETVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["36453"]}},
    "REPLICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1282"]}},
    "REPLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["r0447"]}},
    "REPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1286"]}},
    "REPONO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1289"]}},
    "REPORTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1291"]}},
    "REPOSCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1292"]}},
    "REPOSITORIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["r0453"]}},
    "REPOSITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["37403"]}},
    "REPOTIA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["r0455"]}},
    "REPRAESENTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0459"]}},
    "REPRAESENTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1293"]}},
    "REPREHENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1295"]}},
    "REPREHENSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0462"]}},
    "REPREHENSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["r0463"]}},
    "REPRIMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1297"]}},
    "REPTABVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0473"]}},
    "REPTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1301"]}},
    "REPVDIATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0477"]}},
    "REPVDIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1303"]}},
    "REPVDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["r0480"]}},
    "REPVGNANS": {"-": {"morpho": ["aps---an3i"], "uri": ["100392"]}},
    "REPVGNANTER": {"-": {"morpho": ["rp--------"], "uri": ["53255"]}},
    "REPVGNATORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0484"]}},
    "REPVGNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1305"]}},
    "REPVLLVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1307"]}},
    "REPVLSA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r7010"]}},
    "REPVLSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1308"]}},
    "REPVLSVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["r7010"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["100395"]},
    },
    "REPVRGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1310"]}},
    "REPVTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0492"]}},
    "REPVTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1312"]}},
    "REQVIES": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "REQVIESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1313"]}},
    "REQVIRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1315"]}},
    "RES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["r0497"]}},
    "RESALVTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1317"]}},
    "RESANESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1319"]}},
    "RESARCIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["r1321"]}},
    "RESCINDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1323"]}},
    "RESCIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["r1324"]}},
    "RESCISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r0352"]}},
    "RESCRIBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1325"]}},
    "RESCRIPTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["49370"]}},
    "RESECO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1327"]}},
    "RESEQVOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["r1331"]}},
    "RESERO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["r1333"]},
        "3": {"morpho": ["v1spia--3-"], "uri": ["r1332"]},
    },
    "RESERVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1334"]}},
    "RESES": {"-": {"morpho": ["aps---an3-"], "uri": ["r0508"]}},
    "RESIDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["r1337"]}},
    "RESIDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r0247"]}},
    "RESIDVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0510"]}},
    "RESIGNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1338"]}},
    "RESILIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["r1339"]}},
    "RESINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r0515"]}},
    "RESINATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0519"]}},
    "RESIPISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1341"]}},
    "RESISTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1342"]}},
    "RESOLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1343"]}},
    "RESOLVTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["125251"]}},
    "RESONABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["r0527"]}},
    "RESONANTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r9998"]}},
    "RESONO": {"1": {"morpho": ["v1spia--1-"], "uri": ["r1344"]}},
    "RESORBEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["r1346"]}},
    "RESPECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1349"]}},
    "RESPECTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["r7039"]}},
    "RESPERGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1348"]}},
    "RESPICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["r1350"]}},
    "RESPIRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0536"]}},
    "RESPIRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1351"]}},
    "RESPLENDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["r1352"]}},
    "RESPONDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["r1353"]}},
    "RESPONSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0540"]}},
    "RESPONSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1355"]}},
    "RESPONSVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["r7042"]}},
    "RESPONSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["r7042"]}},
    "RESPVBLICA": {"-": {"morpho": ["n-s----n5-"], "uri": []}},
    "RESPVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1356"]}},
    "RESTAGNO": {"1": {"morpho": ["v1spia--1-"], "uri": ["r1357"]}},
    "RESTAVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1358"]}},
    "RESTIBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["r0549"]}},
    "RESTICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r0550"]}},
    "RESTINGVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1362"]}},
    "RESTIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["r0555"]}},
    "RESTITVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1365"]}},
    "RESTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r0230"]}},
    "RESTRICTE": {"-": {"morpho": ["rp--------"], "uri": ["34552"]}},
    "RESTRICTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["45170"]}},
    "RESTRINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1366"]}},
    "RESVDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1368"]}},
    "RESVLTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1370"]}},
    "RESVMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1371"]}},
    "RESVPINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1373"]}},
    "RESVPINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0568"]}},
    "RESVRGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1374"]}},
    "RETARDATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0573"]}},
    "RETARDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1378"]}},
    "RETE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["r0574"]}},
    "RETEGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1380"]}},
    "RETENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1384"]}},
    "RETENTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0579"]}},
    "RETENTO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["r1506"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["r1506"]},
    },
    "RETERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1386"]}},
    "RETEXO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1387"]}},
    "RETICENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r7007"]}},
    "RETICEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["r1388"]}},
    "RETICESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r7053"]}},
    "RETICVLATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0590"]}},
    "RETICVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["r0591"]}},
    "RETINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["34496"]}},
    "RETINACVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["r0593"]}},
    "RETINENS": {"-": {"morpho": ["aps---an3i"], "uri": ["100405"]}},
    "RETINENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r7008"]}},
    "RETINEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["r1389"]}},
    "RETINGO": {
        "1": {"morpho": ["v1spia--3-"], "uri": []},
        "2": {"morpho": ["v1spia--3-"], "uri": []},
    },
    "RETIO": {"-": {"morpho": ["v1spia--4-"], "uri": []}},
    "RETONO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1394"]}},
    "RETORQVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["r1396"]}},
    "RETORRIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0596"]}},
    "RETRACTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0597"]}},
    "RETRACTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0599"]}},
    "RETRACTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1398"]}},
    "RETRAHO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1400"]}},
    "RETRIBVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1403"]}},
    "RETRO": {
        "1": {"morpho": ["rp--------"], "uri": ["r0607"]},
        "2": {"morpho": ["p---------"], "uri": []},
    },
    "RETROCEDO": {"1": {"morpho": ["v1spia--3-"], "uri": ["r8003"]}},
    "RETRODVCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r8005"]}},
    "RETROFLECTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r8006"]}},
    "RETROLEGO": {"2": {"morpho": ["v1spia--3-"], "uri": ["r8009"]}},
    "RETRORSVM": {"-": {"morpho": ["rp--------"], "uri": ["47642"]}},
    "RETROVERTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r8010"]}},
    "RETRVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1415"]}},
    "RETVNDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1416"]}},
    "RETVSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["34195"]}},
    "REVALESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1418"]}},
    "REVECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1419"]}},
    "REVEHO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1420"]}},
    "REVELLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1421"]}},
    "REVELO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1422"]}},
    "REVENIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["r1428"]}},
    "REVERA": {"-": {"morpho": ["rp--------"], "uri": ["38339"]}},
    "REVERBERO": {"2": {"morpho": ["v1spia--1-"], "uri": ["r1429"]}},
    "REVERENS": {"-": {"morpho": ["aps---an3i"], "uri": ["100410"]}},
    "REVERENTER": {"-": {"morpho": ["rp--------"], "uri": ["26852"]}},
    "REVERENTIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["r7009"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "REVEREOR": {"-": {"morpho": ["v1spid--2-"], "uri": ["r1430"]}},
    "REVERSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0625"]}},
    "REVERSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1433"]}},
    "REVERTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["55550"]}},
    "REVERTOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["r1434"]}},
    "REVILESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1439"]}},
    "REVINCIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["r1440"]}},
    "REVINCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1441"]}},
    "REVISO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1444"]}},
    "REVIVISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["54830"]}},
    "REVIVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r0388"]}},
    "REVOCABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["r0636"]}},
    "REVOCAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["r0637"]}},
    "REVOCATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0638"]}},
    "REVOCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1447"]}},
    "REVOLO": {"2": {"morpho": ["v1spia--1-"], "uri": ["r1448"]}},
    "REVOLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1449"]}},
    "REVOMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1450"]}},
    "REVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["r0643"]}},
    "REX": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["r0645"]},
        "N": {"morpho": ["n-s---mn3-"], "uri": ["r0645"]},
    },
    "RHADINE": {"-": {"morpho": ["n-s---fn1g"], "uri": ["100413"]}},
    "RHEA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "RHEDA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["125458"]}},
    "RHENO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["30870"]}},
    "RHETOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["r0658"]}},
    "RHETORICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0094"]}},
    "RHINOCEROS": {"-": {"morpho": ["n-s---mn3-"], "uri": ["r0672"]}},
    "RHODII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "RHODIVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "RHODOS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "RHOMBOS": {"-": {"morpho": ["n-s---mn2g"], "uri": ["100415"]}},
    "RHYTHMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["r0704"]}},
    "RICINVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["r0959"]}},
    "RICTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["r7048"]}},
    "RICTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["r7048"]}},
    "RIDENDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["125526"]}},
    "RIDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["r1455"]}},
    "RIDIBVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0712"]}},
    "RIDICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r0713"]}},
    "RIDICVLARIA": {"-": {"morpho": ["n-p---nn3i"], "uri": ["100418"]}},
    "RIDICVLE": {"-": {"morpho": ["rp--------"], "uri": ["34555"]}},
    "RIDICVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["45426"]}},
    "RIDICVLVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["46688"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["r0717"]},
    },
    "RIGENS": {"-": {"morpho": ["aps---an3i"], "uri": ["100419"]}},
    "RIGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["r1457"]}},
    "RIGESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1457"]}},
    "RIGIDA": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "RIGIDE": {"-": {"morpho": ["rp--------"], "uri": ["44860"]}},
    "RIGIDITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0720"]}},
    "RIGIDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1458"]}},
    "RIGIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0721"]}},
    "RIGO": {
        "-": {
            "morpho": ["v1spia--1-", "v1spia--1-"],
            "uri": ["r1459", "r1459"],
        }
    },
    "RIGOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["r0722"]}},
    "RIGVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0723"]}},
    "RIMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r0724"]}},
    "RIMOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["r1461"]}},
    "RIMOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0728"]}},
    "RIMVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r0729"]}},
    "RINGOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["r1462"]}},
    "RIPA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r0732"]}},
    "RIPIO": {"-": {"morpho": ["v1spia--3i"], "uri": []}},
    "RISVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["r7050"]}},
    "RITE": {"-": {"morpho": ["rp--------"], "uri": ["r0746"]}},
    "RITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["r0748"]}},
    "RIVALIS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["100421"]},
        "2": {"morpho": ["aps---cn3i"], "uri": ["r0750"]},
    },
    "RIVO": {
        "-": {"morpho": ["v1spia--1-"], "uri": ["r1465"]},
        "1": {"morpho": ["v1spia--1-"], "uri": ["r1465"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["r1465"]},
    },
    "RIVVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["r0756"]}},
    "RIXA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r0757"]}},
    "RIXOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["r1466"]}},
    "ROBIGALIA": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "ROBIGO": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["r0765"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "ROBORATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r9767"]}},
    "ROBORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1469"]}},
    "ROBVR": {"-": {"morpho": ["n-s---nn3-"], "uri": ["r0769"]}},
    "ROBVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r9763"]}},
    "ROBVSTEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0770"]}},
    "ROBVSTVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["r0771"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "RODEVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "RODO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1470"]}},
    "ROGALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["r0773"]}},
    "ROGATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0778"]}},
    "ROGATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["r1508"]}},
    "ROGITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1471"]}},
    "ROGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1472"]}},
    "ROGVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["r0782"]}},
    "ROMA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ROMANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ROMANVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": ["39569"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["39569"]},
    },
    "ROMVLARIS": {
        "N": {"morpho": ["aps---cn3i"], "uri": []},
        "A": {"morpho": ["aps---cn3i"], "uri": []},
    },
    "ROMVLEVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "ROMVLVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "RORIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0789"]}},
    "RORIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["r0790"]}},
    "RORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1474"]}},
    "RORVLENTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0793"]}},
    "ROS": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["r0794"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "ROSA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r0795"]}},
    "ROSARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["r0799"]}},
    "ROSCIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0802"]}},
    "ROSETVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["r0803"]}},
    "ROSEVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["r0804"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["r0804"]},
    },
    "ROSTRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0811"]}},
    "ROSTRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["r0812"]}},
    "ROTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r0816"]}},
    "ROTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0824"]}},
    "ROTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1476"]}},
    "ROTVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r0827"]}},
    "ROTVNDATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0829"]}},
    "ROTVNDITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0831"]}},
    "ROTVNDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1477"]}},
    "ROTVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0177"]}},
    "RVBELLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0838"]}},
    "RVBENS": {"-": {"morpho": ["aps---an3i"], "uri": ["100424"]}},
    "RVBEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["r1479"]}},
    "RVBER": {
        "-": {"morpho": ["aps---mn1r"], "uri": ["r0839"]},
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "RVBESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1479"]}},
    "RVBETA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r0840"]}},
    "RVBETVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["r0841"]}},
    "RVBEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["26896"]}},
    "RVBIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r0842"]}},
    "RVBICVNDVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0844"]}},
    "RVBICVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0845"]}},
    "RVBOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["r0848"]}},
    "RVBRICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r0851"]}},
    "RVBRICOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0852"]}},
    "RVBRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "RVBVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["r0854"]}},
    "RVCTABVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0855"]}},
    "RVCTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1482"]}},
    "RVCTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["r0861"]}},
    "RVDECTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0862"]}},
    "RVDENS": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "RVDERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0865"]}},
    "RVDICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r0869"]}},
    "RVDIMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["r0871"]}},
    "RVDIS": {
        "1": {"morpho": ["n-s---fn3i"], "uri": ["r0872"]},
        "2": {"morpho": ["aps---cn3i"], "uri": ["r9872"]},
    },
    "RVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1484"]}},
    "RVDVS": {"-": {"morpho": ["n-s---nn3-"], "uri": ["r1990"]}},
    "RVFVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["r0878"]},
        "N": {"morpho": ["aps---mn1-"], "uri": ["r0878"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["r0878"]},
    },
    "RVGA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r0879"]}},
    "RVGOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0883", "40600"]}},
    "RVINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r0886"]}},
    "RVINOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0887"]}},
    "RVMINALIS": {
        "N": {"morpho": ["aps---cn3i"], "uri": ["r0894"]},
        "A": {"morpho": ["aps---cn3i"], "uri": ["r0894"]},
    },
    "RVMINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1493"]}},
    "RVMOR": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["r0896"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "RVMPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1496"]}},
    "RVNCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1498"]}},
    "RVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1499"]}},
    "RVPES": {"-": {"morpho": ["n-s---fn3i"], "uri": ["r0908"]}},
    "RVPILIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "RVPTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["r0916"]}},
    "RVRICOLA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["r0921"]}},
    "RVRSVS": {"-": {"morpho": ["rp--------"], "uri": ["r0924"]}},
    "RVS": {"-": {"morpho": ["n-s---nn3-"], "uri": ["r0925"]}},
    "RVSCEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0927"]}},
    "RVSCVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["r0929"]}},
    "RVSCVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["34314"]}},
    "RVSSATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0930"]}},
    "RVSSEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0932"]}},
    "RVSSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0934"]}},
    "RVSTARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0935"]}},
    "RVSTICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["56569"]}},
    "RVSTICANVS": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": ["r1901"]},
    },
    "RVSTICE": {"-": {"morpho": ["rp--------"], "uri": ["39958"]}},
    "RVSTICITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0940"]}},
    "RVSTICVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["31185"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["r0943"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["31185"]},
    },
    "RVTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r0944"]}},
    "RVTABVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["r0945"]}},
    "RVTILANS": {"-": {"morpho": ["aps---an3i"], "uri": ["100441"]}},
    "RVTILIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "RVTILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "RVTILO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1505"]}},
    "RVTILVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["r0950"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "RVTRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["r0953"]}},
    "RVTVLI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SABBATVM": {
        "-": {"morpho": ["n-s---nn2-"], "uri": ["s0006"]},
        "N": {"morpho": ["n-s---nn2-"], "uri": ["s0006"]},
    },
    "SABINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SABINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "SABVLO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s9014"]}},
    "SABVLOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0013"]}},
    "SABVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0014"]}},
    "SABVRRA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["s0015"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "SACCATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["100443"]}},
    "SACCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s0031"]}},
    "SACCVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s0036"]}},
    "SACCVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s0037"]}},
    "SACELLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0024"]}},
    "SACER": {
        "-": {"morpho": ["aps---mn1r"], "uri": ["s0040"]},
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "SACERDOS": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["26922"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "SACERDOTIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0046"]}},
    "SACRAMENTALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["125828"]}},
    "SACRAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0052"]}},
    "SACRARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0053"]}},
    "SACRATIVIR": {"N": {"morpho": ["n-s----n2-"], "uri": []}},
    "SACRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["26926"]}},
    "SACRIFICALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s0059"]}},
    "SACRIFICIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0063"]}},
    "SACRIFICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s0064"]}},
    "SACRIFICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0066"]}},
    "SACRILEGA": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "SACRILEGIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0067"]}},
    "SACRILEGVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["100447"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["s0068"]},
    },
    "SACRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s0071"]}},
    "SACROSANCTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s9997"]}},
    "SACRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0040"]}},
    "SADALA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SAECVLARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s0073"]}},
    "SAECVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0075"]}},
    "SAEPE": {"-": {"morpho": ["rp--------"], "uri": ["s0076"]}},
    "SAEPENVMERO": {"-": {"morpho": ["rp--------"], "uri": ["35227"]}},
    "SAEPES": {"-": {"morpho": ["n-s---fn3i"], "uri": ["s0078"]}},
    "SAEPIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["125849"]}},
    "SAEPIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["s0083"]}},
    "SAEPTA": {
        "-": {"morpho": ["n-s----n2-"], "uri": []},
        "L": {"morpho": ["n-s----n2-"], "uri": []},
    },
    "SAEPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0088"]}},
    "SAEPTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": []}},
    "SAETA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0090"]}},
    "SAETIGER": {
        "1": {"morpho": ["n-s---mn2r"], "uri": ["100448"]},
        "2": {"morpho": ["aps---mn1r"], "uri": ["s0091"]},
    },
    "SAETOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0092"]}},
    "SAEVE": {"-": {"morpho": ["rp--------"], "uri": ["46055"]}},
    "SAEVIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["s0095"]}},
    "SAEVITIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0098"]}},
    "SAEVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0100"]}},
    "SAGA": {
        "1": {"morpho": ["n-s---fn1-"], "uri": ["s0134"]},
        "2": {"morpho": ["n-s---fn1-"], "uri": ["s0134"]},
    },
    "SAGACITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0103"]}},
    "SAGACITER": {"-": {"morpho": ["rp--------"], "uri": ["37310"]}},
    "SAGAX": {"-": {"morpho": ["aps---an3i"], "uri": ["s0109"]}},
    "SAGINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0113"]}},
    "SAGINATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["125879"]}},
    "SAGINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s0117"]}},
    "SAGIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["s0118"]}},
    "SAGITTA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["s0119"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "SAGITTARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["s0120"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["s0120"]},
    },
    "SAGITTIFER": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["aps---mn1r"], "uri": ["s0121"]},
    },
    "SAGVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0132"]}},
    "SAGVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0133"]}},
    "SAGVNTINI": {
        "-": {"morpho": ["n-s----n2-"], "uri": []},
        "N": {"morpho": ["n-p---mn2-"], "uri": []},
    },
    "SAGVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["125890"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["s0134"]},
    },
    "SAL": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s0136"]}},
    "SALAPVTIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0141"]}},
    "SALARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0144"]}},
    "SALAX": {"-": {"morpho": ["aps---an3i"], "uri": ["s0146"]}},
    "SALEBRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0147"]}},
    "SALEBROSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0150"]}},
    "SALGAMA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["s0151"]}},
    "SALIARIS": {
        "1": {"morpho": ["aps---cn3i"], "uri": ["s0154"]},
        "2": {"morpho": ["aps---cn3i"], "uri": ["s0154"]},
        "A": {"morpho": ["aps---cn3i"], "uri": ["s0154"]},
    },
    "SALICTARIVS": {"1": {"morpho": ["n-s---mn2-"], "uri": ["100453"]}},
    "SALICTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0160"]}},
    "SALIFODINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0161"]}},
    "SALIGNEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0162"]}},
    "SALIGNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0163"]}},
    "SALII": {
        "-": {"morpho": ["n-s----n2-"], "uri": []},
        "N": {"morpho": ["n-p---mn2-"], "uri": []},
    },
    "SALILLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0164"]}},
    "SALINAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["s0165"]}},
    "SALINATOR": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["s0166"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "SALINVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0165"]}},
    "SALIO": {
        "1": {"morpho": ["v1spia--4-"], "uri": ["s0182"]},
        "2": {"morpho": ["v1spia--4-"], "uri": ["s0182"]},
    },
    "SALIVA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0176"]}},
    "SALIVNCA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0174"]}},
    "SALIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0181"]}},
    "SALLVSTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SALSAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0191"]}},
    "SALSITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0198"]}},
    "SALSVGO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0201"]}},
    "SALSVM": {
        "-": {"morpho": ["n-s----n2-"], "uri": []},
        "N": {"morpho": ["n-s---nn2-"], "uri": []},
    },
    "SALSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["38775"]}},
    "SALTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0204"]}},
    "SALTATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s0206"]}},
    "SALTATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["s9947"]}},
    "SALTEM": {"-": {"morpho": ["rp--------"], "uri": ["s0210"]}},
    "SALTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s0213"]}},
    "SALTVOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0218"]}},
    "SALTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["s9948"]},
        "2": {"morpho": ["n-s---mn4-"], "uri": ["s9948"]},
    },
    "SALVBER": {"-": {"morpho": ["aps---mn3-"], "uri": ["s0223"]}},
    "SALVBRITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0224"]}},
    "SALVBRITER": {"-": {"morpho": ["rp--------"], "uri": ["50714"]}},
    "SALVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["s0225"]}},
    "SALVIDIENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SALVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0231"]}},
    "SALVS": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["s0233"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "SALVTARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s0235"]}},
    "SALVTARITER": {"-": {"morpho": ["rp--------"], "uri": ["29873"]}},
    "SALVTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0236"]}},
    "SALVTATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s0238"]}},
    "SALVTATRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0240"]}},
    "SALVTIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["s0241"]}},
    "SALVTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s0245"]}},
    "SALVVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["s0246"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "SAMBVCA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0249"]}},
    "SAMII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SAMIVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "SAMNIS": {
        "N": {"morpho": ["aps---an3-"], "uri": []},
        "A": {"morpho": ["aps---cn3i"], "uri": []},
    },
    "SAMOS": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "SANABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s0262"]}},
    "SANCIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["s0267"]}},
    "SANCTE": {"-": {"morpho": ["rp--------"], "uri": ["48764"]}},
    "SANCTIMONIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0278"]}},
    "SANCTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0281"]}},
    "SANCTITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0282"]}},
    "SANCTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s0284"]}},
    "SANCTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["100463"]}},
    "SANCTVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["26987"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["26987"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["126023"]},
    },
    "SANCVS": {
        "-": {"morpho": ["n-s----n--"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "SANDAPILA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0291"]}},
    "SANDARACA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0293"]}},
    "SANDYX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0301"]}},
    "SANE": {"-": {"morpho": ["rp--------"], "uri": ["26989"]}},
    "SANGVINARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0309"]}},
    "SANGVINEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0311"]}},
    "SANGVINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s0312"]}},
    "SANGVINOLENTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0314"]}},
    "SANGVIS": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s0316"]}},
    "SANIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["s0320"]}},
    "SANITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0323"]}},
    "SANNA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0324"]}},
    "SANO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s0328"]}},
    "SANVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0331"]}},
    "SAPA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0332"]}},
    "SAPERDA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["s0334"]}},
    "SAPIENS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["100468"]},
        "2": {"morpho": ["aps---an3i"], "uri": ["100467"]},
        "C": {"morpho": ["n-s---mn3i"], "uri": ["100468"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "SAPIENTER": {"-": {"morpho": ["rp--------"], "uri": ["58067"]}},
    "SAPIENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0302"]}},
    "SAPINEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["44930"]}},
    "SAPINVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["39228"]}},
    "SAPIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["s0339"]}},
    "SAPOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s0344"]}},
    "SAPPHO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "SARCINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0360"]}},
    "SARCINARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0362"]}},
    "SARCINVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0367"]}},
    "SARCIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["s0368"]}},
    "SARCOPHAGVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["34640"]}},
    "SARCVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0382"]}},
    "SARDA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s3319"]}},
    "SARDINIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SARDONYX": {"-": {"morpho": ["n-s---cn3-"], "uri": ["s0389"]}},
    "SARIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["s0391"]}},
    "SARISA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0393"]}},
    "SARISOPHORVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s0394"]}},
    "SARMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0401"]}},
    "SARRACVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["53037"]}},
    "SARTAGO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0405"]}},
    "SARTVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0355"]}},
    "SARTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["42922"]}},
    "SATA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["45937"]}},
    "SATAGEVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "SATAGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a9271"]}},
    "SATARIVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "SATELLES": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s0417"]}},
    "SATIAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0420"]}},
    "SATIETAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0422"]}},
    "SATIO": {
        "1": {"morpho": ["n-s---fn3-"], "uri": ["s0425"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["s0424"]},
    },
    "SATIS": {
        "1": {"morpho": ["aps---nn--"], "uri": ["27017"]},
        "2": {"morpho": ["rp--------"], "uri": ["100479"]},
    },
    "SATISACCIPIO": {"-": {"morpho": ["v1spia--3i"], "uri": []}},
    "SATISDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s0432"]}},
    "SATISFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["s0434", "s0433"]}},
    "SATISFACTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0435"]}},
    "SATISFIO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "SATOR": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["s0439"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "SATRAPEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["100480"]}},
    "SATRAPES": {"-": {"morpho": ["n-s---mn1g"], "uri": ["s0441"]}},
    "SATRICVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "SATRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SATVR": {"-": {"morpho": ["aps---mn1r"], "uri": ["s0445"]}},
    "SATVRA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["s0446"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "SATVRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["126175"]}},
    "SATVREIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["55627"]}},
    "SATVRITAS": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["s0453"]},
        "N": {"morpho": ["n-s---fn3-"], "uri": ["s0453"]},
    },
    "SATVRNALIA": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "SATVRNINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SATVRNVS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "SATVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s0454"]}},
    "SATVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["s9949"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["39002"]},
    },
    "SATYRICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0457"]}},
    "SATYRVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["s0460"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["s0460"]},
    },
    "SAVCIATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0462"]}},
    "SAVCIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s0464"]}},
    "SAVCIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0465"]}},
    "SAVILLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0468"]}},
    "SAVIOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["s0469"]}},
    "SAVIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0471"]}},
    "SAXEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0482"]}},
    "SAXIFICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0485"]}},
    "SAXOSA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["126208"]}},
    "SAXOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0491"]}},
    "SAXVM": {
        "-": {"morpho": ["n-s---nn2-"], "uri": ["s0493"]},
        "N": {"morpho": ["n-s---nn2-"], "uri": ["s0493"]},
    },
    "SCABELLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["32052"]}},
    "SCABER": {"-": {"morpho": ["aps---mn1r"], "uri": ["s0495"]}},
    "SCABIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["s0496"]}},
    "SCABIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0503"]}},
    "SCABO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s0505"]}},
    "SCAEVOLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SCAEVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s4025"]}},
    "SCALAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["40692"]}},
    "SCALMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s0528"]}},
    "SCALPELLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0530"]}},
    "SCALPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s0531"]}},
    "SCALPRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0533"]}},
    "SCALPTVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0567"]}},
    "SCAMILLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["126257"]}},
    "SCAMMONIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["126261"]}},
    "SCAMNVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0547"]}},
    "SCANDENS": {
        "-": {"morpho": ["aps---an3-"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "SCANDILIANVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "SCANDILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SCANDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s0553"]}},
    "SCANDVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0554"]}},
    "SCANSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0559"]}},
    "SCANSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["126273"]}},
    "SCANSORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0560"]}},
    "SCAPHE": {"-": {"morpho": ["n-s---fn1g"], "uri": ["s0563"]}},
    "SCAPHIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0565"]}},
    "SCAPVLA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["s7605"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["s7605"]},
    },
    "SCAPVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s0570"]}},
    "SCARABEVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["31716"]}},
    "SCARVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s0579"]}},
    "SCATEBRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0580"]}},
    "SCATEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["s0582"]}},
    "SCATO": {
        "-": {"morpho": ["v1spia--3-"], "uri": ["36021"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "SCATVRIGO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["31441"]}},
    "SCAVRVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["s0589"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "SCELERATE": {"-": {"morpho": ["rp--------"], "uri": ["48709"]}},
    "SCELERATI": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "SCELERATVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["27040"]},
        "N": {"morpho": ["aps---mn1-"], "uri": ["27040"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["27040"]},
    },
    "SCELERO": {
        "1": {"morpho": ["n-s----n3-"], "uri": []},
        "2": {"morpho": ["v1spia--1-"], "uri": ["s0594"]},
    },
    "SCELEROSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0595"]}},
    "SCELESTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0597"]}},
    "SCELVS": {
        "-": {"morpho": ["n-s---nn3-"], "uri": ["s0601"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "SCENA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["36764"]}},
    "SCENICA": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "SCENICVS": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "SCEPTRIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["s0607"]}},
    "SCEPTRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0609"]}},
    "SCEPTVCHVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s0610"]}},
    "SCHEDA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["56664"]}},
    "SCHEMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0614"]}},
    "SCHERINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SCHIDIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0616"]}},
    "SCHOENOBATES": {"-": {"morpho": ["n-s---mn1g"], "uri": ["s0622"]}},
    "SCHOENVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s0623"]}},
    "SCHOLA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["s0625"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "SCHOLASTICA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["s0628"]}},
    "SCHOLASTICVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["s0628"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["s0628"]},
    },
    "SCIATHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SCIENS": {"-": {"morpho": ["aps---an3i"], "uri": ["27050"]}},
    "SCIENTER": {"-": {"morpho": ["rp--------"], "uri": ["58223"]}},
    "SCIENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s9915"]}},
    "SCILICET": {"-": {"morpho": ["rp--------"], "uri": ["27052"]}},
    "SCILLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0641"]}},
    "SCINDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s0647"]}},
    "SCINTILLA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["s0649"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "SCINTILLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s0651"]}},
    "SCIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["s0654"]}},
    "SCIOTHERES": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "SCIPIO": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["s0658"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "SCIRPEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0145"]}},
    "SCIRPICVLVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["s1732"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["s1596"]},
    },
    "SCISCITOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["s0672"]}},
    "SCISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s0409"]}},
    "SCISSVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s9916"]}},
    "SCITE": {"-": {"morpho": ["rp--------"], "uri": ["40874"]}},
    "SCITOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["s0680"]}},
    "SCITVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s9950"]}},
    "SCITVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["s9950"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["27058"]},
    },
    "SCOMBER": {"-": {"morpho": ["n-s---mn2r"], "uri": ["s0700"]}},
    "SCOPAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["100497"]}},
    "SCOPIO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s0706"]}},
    "SCOPVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0710"]}},
    "SCOPVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s4026"]}},
    "SCORDALVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s0716"]}},
    "SCORPIO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s0727"]}},
    "SCORPIVS": {
        "-": {"morpho": ["n-s----n-g"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "SCORTATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s0735"]}},
    "SCORTILLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0739"]}},
    "SCORTVM": {
        "-": {"morpho": ["n-s---nn2-"], "uri": ["s0743"]},
        "N": {"morpho": ["n-s---nn2-"], "uri": ["s0743"]},
    },
    "SCOTEINOS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "SCOTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0744"]}},
    "SCREO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s0753"]}},
    "SCRIBA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["s0754"]}},
    "SCRIBLITA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["37553"]}},
    "SCRIBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s0758"]}},
    "SCRINIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0761"]}},
    "SCRIPTA": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "SCRIPTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s0768"]}},
    "SCRIPTVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0770"]}},
    "SCRIPTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": []}},
    "SCRIPTVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s9917"]}},
    "SCRIPVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0777"]}},
    "SCROBIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["s0780"]}},
    "SCROFA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0781"]}},
    "SCRVPEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0786"]}},
    "SCRVPOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0787"]}},
    "SCRVPVLOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0790"]}},
    "SCRVPVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s0791"]}},
    "SCRVTA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["s0793"]}},
    "SCRVTANS": {"-": {"morpho": ["aps---an3i"], "uri": ["126493"]}},
    "SCRVTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0796"]}},
    "SCRVTOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["s0801"]}},
    "SCVLPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s0805"]}},
    "SCVLPONEAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["s0806"]}},
    "SCVLPTVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s9918"]}},
    "SCVRRA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["s0813"]}},
    "SCVRRILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s0814"]}},
    "SCVRRILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0815"]}},
    "SCVTATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0822"]}},
    "SCVTICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0824"]}},
    "SCVTRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0826"]}},
    "SCVTRISCVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0828"]}},
    "SCVTVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s4027"]}},
    "SCVTVLATA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["100509"]}},
    "SCVTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0833"]}},
    "SCYLLA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SCYMNVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s0836"]}},
    "SCYPHVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s0837"]}},
    "SCYTHAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "SCYTHIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SEBVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0851"]}},
    "SECEDO": {"1": {"morpho": ["v1spia--3-"], "uri": ["s0856"]}},
    "SECENTI": {"-": {"morpho": ["----------"], "uri": []}},
    "SECERNO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s0858"]}},
    "SECESSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0860"]}},
    "SECESSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["s9952"]}},
    "SECLVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s0862"]}},
    "SECLVSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["53848"]}},
    "SECO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s0864"]}},
    "SECRETE": {"-": {"morpho": ["rp--------"], "uri": ["100511"]}},
    "SECRETO": {"-": {"morpho": ["rp--------"], "uri": ["27095"]}},
    "SECRETVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["27096"]}},
    "SECRETVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["37037"]}},
    "SECRO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "SECROR": {"-": {"morpho": ["v1spid--1-"], "uri": []}},
    "SECTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s9868"]}},
    "SECTATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s0870"]}},
    "SECTILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s0872"]}},
    "SECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0873"]}},
    "SECTIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0874"]}},
    "SECTOR": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["s0876"]},
        "2": {"morpho": ["v1spid--1-"], "uri": ["s0875"]},
    },
    "SECTVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0888"]}},
    "SECVBITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["s9881"]}},
    "SECVBO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s0881"]}},
    "SECVNDAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["100514"]}},
    "SECVNDANI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["35820"]}},
    "SECVNDANVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0883"]}},
    "SECVNDARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0884"]}},
    "SECVNDE": {"-": {"morpho": ["rp--------"], "uri": ["100515"]}},
    "SECVNDO": {
        "1": {"morpho": ["rp--------"], "uri": ["126575"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["s0887"]},
    },
    "SECVNDVM": {
        "1": {"morpho": ["rp--------"], "uri": ["58018"]},
        "2": {"morpho": ["p---------"], "uri": ["s9996"]},
    },
    "SECVNDVS": {
        "1": {"morpho": ["aps---mn1-"], "uri": ["s9990"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["s9990"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "SECVRE": {"-": {"morpho": ["rp--------"], "uri": ["100516"]}},
    "SECVRICLATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0889"]}},
    "SECVRICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0890"]}},
    "SECVRIGER": {"-": {"morpho": ["aps---mn1r"], "uri": ["s0893"]}},
    "SECVRIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["s0894"]}},
    "SECVRITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0895"]}},
    "SECVRVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0896"]}},
    "SECVS": {
        "1": {"morpho": ["n-s---nn--"], "uri": ["s9897"]},
        "2": {"morpho": ["rp--------"], "uri": ["27103"]},
        "3": {"morpho": ["p---------"], "uri": ["s0897"]},
    },
    "SECVTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s0899"]}},
    "SED": {"-": {"morpho": ["----------"], "uri": []}},
    "SEDAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["s0903"]}},
    "SEDATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0904"]}},
    "SEDATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["27104"]}},
    "SEDECIM": {"-": {"morpho": ["----------"], "uri": []}},
    "SEDENTARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0911"]}},
    "SEDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["s0912"]}},
    "SEDES": {"-": {"morpho": ["n-s---fn3i"], "uri": ["s0913"]}},
    "SEDILE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["s0916"]}},
    "SEDITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0919"]}},
    "SEDITIOSE": {"-": {"morpho": ["rp--------"], "uri": ["51907"]}},
    "SEDITIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0921"]}},
    "SEDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s0922"]}},
    "SEDVCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s0923"]}},
    "SEDVCTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["57845"]}},
    "SEDVLITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0931"]}},
    "SEDVLO": {"-": {"morpho": ["rp--------"], "uri": ["27113"]}},
    "SEDVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0932"]}},
    "SEGES": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0934"]}},
    "SEGESTA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SEGESTANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SEGESTANVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "SEGMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["s0938"]}},
    "SEGMENTATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0940"]}},
    "SEGMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0941"]}},
    "SEGNE": {"-": {"morpho": ["rp--------"], "uri": ["100520"]}},
    "SEGNIPES": {"-": {"morpho": ["aps---an3-"], "uri": ["s0943"]}},
    "SEGNIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s0944"]}},
    "SEGNITER": {"-": {"morpho": ["rp--------"], "uri": ["47832"]}},
    "SEGNITIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0946"]}},
    "SEGNITIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["s0946"]}},
    "SEGREGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s0952"]}},
    "SEGREX": {"-": {"morpho": ["aps---an3-"], "uri": ["s0953"]}},
    "SEIANIANVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "SEIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SEIVNCTIM": {"-": {"morpho": ["rp--------"], "uri": ["s0959"]}},
    "SEIVNCTVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "SEIVNGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s0961"]}},
    "SELECTVS": {
        "C": {"morpho": ["aps---mn1-"], "uri": []},
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "SELENE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SELIBRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0971"]}},
    "SELIGO": {"3": {"morpho": ["v1spia--3-"], "uri": ["s0972"]}},
    "SELLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0977"]}},
    "SELLARIVS": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "SELLISTERNIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0982"]}},
    "SEMEL": {"-": {"morpho": ["rp--------"], "uri": ["s0990"]}},
    "SEMEN": {
        "1": {"morpho": ["n-s---nn3-"], "uri": ["s0991"]},
        "2": {"morpho": ["n-s---nn3-"], "uri": ["s0991"]},
    },
    "SEMENTIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["s0996"]}},
    "SEMENTIVVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["s0997"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["s0997"]},
    },
    "SEMESTRIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["126656"]}},
    "SEMESVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0999"]}},
    "SEMIADAPERTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1002"]}},
    "SEMIANIMIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s1773"]}},
    "SEMIANIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1746"]}},
    "SEMIBOS": {
        "-": {"morpho": ["aps---an3-"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "SEMICANALICVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s1018"]}},
    "SEMIDEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1041"]}},
    "SEMIDEVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["s1041"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["s1041"]},
    },
    "SEMIERMIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s1813"]}},
    "SEMIERMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1775"]}},
    "SEMIFACTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1048"]}},
    "SEMIFASTIGIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1050"]}},
    "SEMIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["s1051"]}},
    "SEMIFVNIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1055"]}},
    "SEMIGERMANVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["34510"]}},
    "SEMIGRAVIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s1059"]}},
    "SEMIHIANS": {"-": {"morpho": ["aps---an3i"], "uri": ["s1061"]}},
    "SEMIHOMO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s1063"]}},
    "SEMIHORA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1064"]}},
    "SEMILATER": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s1070"]}},
    "SEMILAVTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1072"]}},
    "SEMILIBER": {"-": {"morpho": ["aps---mn1r"], "uri": ["s1073"]}},
    "SEMIMARINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1079"]}},
    "SEMIMAS": {"-": {"morpho": ["aps---an3-"], "uri": ["s1080"]}},
    "SEMIMETOPIA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["126730"]}},
    "SEMIMORTVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1087"]}},
    "SEMINARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1089"]}},
    "SEMINARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1089"]}},
    "SEMINATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s1091"]}},
    "SEMINEX": {"-": {"morpho": ["aps---an3-"], "uri": ["s1092"]}},
    "SEMINIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1094"]}},
    "SEMINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s1095"]}},
    "SEMINVDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1097"]}},
    "SEMIPAGANVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["100538"]}},
    "SEMIPES": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s1113"]}},
    "SEMIPLENVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1115"]}},
    "SEMIPVTATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1123"]}},
    "SEMIRASVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1125"]}},
    "SEMIREDVCTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1126"]}},
    "SEMIREFECTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1127"]}},
    "SEMIRVTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1132"]}},
    "SEMIS": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["s1133"]},
        "2": {"morpho": ["n-s---mn3-"], "uri": ["s1133"]},
        "1": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "SEMISEPVLTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1137"]}},
    "SEMISOMNIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s1140"]}},
    "SEMISOMNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1140"]}},
    "SEMISVPINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1152"]}},
    "SEMITA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1154"]}},
    "SEMITARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1158"]}},
    "SEMIVIR": {"-": {"morpho": ["n-s---mn2r"], "uri": ["s1171"]}},
    "SEMIVIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1172"]}},
    "SEMIVSTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1177"]}},
    "SEMODIALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s1182"]}},
    "SEMODIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s1183"]}},
    "SEMOTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["47862"]}},
    "SEMOVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["s1186"]}},
    "SEMPER": {"-": {"morpho": ["rp--------"], "uri": ["s1187"]}},
    "SEMPITERNO": {"-": {"morpho": ["rp--------"], "uri": ["27128"]}},
    "SEMPITERNVM": {"-": {"morpho": ["rp--------"], "uri": ["100542"]}},
    "SEMPITERNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1191"]}},
    "SEMPRONIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "SEMVNCIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1192"]}},
    "SEMVSTVLATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["56730"]}},
    "SENARIOLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s1197"]}},
    "SENATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s1199"]}},
    "SENATORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1200"]}},
    "SENATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["s1202"]}},
    "SENECA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SENECTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1205"]}},
    "SENECTVS": {
        "1": {"morpho": ["n-s---fn3-"], "uri": ["s1206"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["s1205"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "SENEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["s1207"]}},
    "SENESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s1207"]}},
    "SENEX": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["27134"]},
        "2": {"morpho": ["aps---an3-"], "uri": []},
    },
    "SENI": {"-": {"morpho": ["rp--------"], "uri": []}},
    "SENILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s1212"]}},
    "SENIO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s1213"]}},
    "SENIOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["48567"]}},
    "SENIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1214"]}},
    "SENIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1216"]}},
    "SENSA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["126841"]}},
    "SENSATIO": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "SENSIBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s1218"]}},
    "SENSIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["s1221"]}},
    "SENSILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s1225"]}},
    "SENSIM": {"-": {"morpho": ["rp--------"], "uri": ["s1227"]}},
    "SENSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["s1245"]}},
    "SENTENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1232"]}},
    "SENTINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1238"]}},
    "SENTINOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1242"]}},
    "SENTIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["s1243"]}},
    "SENTIOR": {"-": {"morpho": ["v1spid--4-"], "uri": []}},
    "SENTIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["s1244"]}},
    "SENTISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s1501"]}},
    "SENTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SENTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1248"]}},
    "SEORSVM": {
        "1": {"morpho": ["rp--------"], "uri": ["s1249"]},
        "2": {"morpho": ["p---------"], "uri": []},
    },
    "SEPARATE": {"-": {"morpho": ["rp--------"], "uri": ["30457"]}},
    "SEPARATIM": {"-": {"morpho": ["rp--------"], "uri": ["s1253"]}},
    "SEPARATVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "SEPARATVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": ["29373"]},
    },
    "SEPARO": {"2": {"morpho": ["v1spia--1-"], "uri": ["s1259"]}},
    "SEPELIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["s1261"]}},
    "SEPES": {"-": {"morpho": ["n-s---fn3i"], "uri": ["s1262"]}},
    "SEPONO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s1268"]}},
    "SEPOSITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["54474"]}},
    "SEPSE": {"-": {"morpho": ["o-----an--"], "uri": []}},
    "SEPTEM": {
        "-": {"morpho": ["----------"], "uri": []},
        "N": {"morpho": ["n-s----n--"], "uri": []},
    },
    "SEPTEMBER": {
        "A": {"morpho": ["aps---mn1r"], "uri": []},
        "N": {"morpho": ["n-s---mn2r"], "uri": []},
    },
    "SEPTEMGEMINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1277"]}},
    "SEPTEMPLEX": {
        "-": {
            "morpho": ["aps---an3-", "aps---an3-"],
            "uri": ["s1281", "s1281"],
        }
    },
    "SEPTEMREMIS": {
        "-": {"morpho": ["aps---an3-"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "SEPTEMVIR": {"-": {"morpho": ["n-s---mn2r"], "uri": ["32599"]}},
    "SEPTENDECIM": {"-": {"morpho": ["----------"], "uri": []}},
    "SEPTENDECIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": []}},
    "SEPTENTRIO": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["s1292"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "SEPTENTRIONALIA": {"-": {"morpho": ["n-p---nn3i"], "uri": ["100559"]}},
    "SEPTENTRIONALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s1289"]}},
    "SEPTENVS": {"-": {"morpho": ["rp--------"], "uri": []}},
    "SEPTICIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SEPTIMIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SEPTIMVM": {"-": {"morpho": ["rp--------"], "uri": ["54362"]}},
    "SEPTIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1305"]}},
    "SEPTINGENTESIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1310"]}},
    "SEPTINGENTI": {"-": {"morpho": ["----------"], "uri": []}},
    "SEPTVAGESIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1319"]}},
    "SEPTVAGINTA": {"-": {"morpho": ["----------"], "uri": []}},
    "SEPVLCRALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s1329"]}},
    "SEPVLCRETVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1330"]}},
    "SEPVLCRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1331"]}},
    "SEPVLTVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1290"]}},
    "SEQVAX": {"-": {"morpho": ["aps---an3i"], "uri": ["s1337"]}},
    "SEQVENS": {"-": {"morpho": ["aps---an3i"], "uri": ["100565"]}},
    "SEQVESTER": {
        "1": {"morpho": ["n-s---mn2r"], "uri": ["s1339"]},
        "2": {"morpho": ["aps---an--"], "uri": []},
        "A": {"morpho": ["aps---an--"], "uri": []},
    },
    "SEQVESTRO": {"2": {"morpho": ["rp--------"], "uri": []}},
    "SEQVOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["s1348"]}},
    "SERA": {
        "1": {"morpho": ["n-s---fn1-"], "uri": ["s1431", "s1349"]},
        "2": {"morpho": ["rp--------"], "uri": ["100566"]},
    },
    "SERAPIS": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "SERARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1351"]}},
    "SERENITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s1356"]}},
    "SERENO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s1357"]}},
    "SERENVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1358"]}},
    "SERENVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["s1358"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "SERESCO": {"1": {"morpho": ["v1spia--3-"], "uri": ["s4028"]}},
    "SERGIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "SERIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["s1375", "s1360", "s1366"]}
    },
    "SERICA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["100568"]}},
    "SERICVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["s3809"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["s3809"]},
    },
    "SERIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["s1366"]}},
    "SERIO": {"-": {"morpho": ["rp--------"], "uri": ["27163"]}},
    "SERIOLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1369"]}},
    "SERIPHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SERIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1375"]}},
    "SERIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1375"]}},
    "SERMO": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["s1376"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "SERMOCINOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["s1379"]}},
    "SERO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["s1386"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["s1384", "s1385"]},
        "3": {"morpho": ["v1spia--3-"], "uri": ["s1384", "s1385"]},
        "4": {"morpho": ["rp--------"], "uri": ["100571"]},
    },
    "SERPENS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["27169"]}},
    "SERPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s1395"]}},
    "SERPYLLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["51111"]}},
    "SERRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1400"]}},
    "SERRARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s1405"]}},
    "SERRATIM": {"-": {"morpho": ["rp--------"], "uri": ["s1406"]}},
    "SERRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["52473"]}},
    "SERTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s9938"]}},
    "SERTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SERTORIANVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "SERTORIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SERTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s9938"]}},
    "SERVA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1432"]}},
    "SERVANS": {"-": {"morpho": ["aps---an3i"], "uri": ["100573"]}},
    "SERVATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s1415"]}},
    "SERVIENS": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "SERVILIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SERVILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s1418"]}},
    "SERVILIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "SERVIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["s1419"]}},
    "SERVITIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1421"]}},
    "SERVITVS": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["s1426"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "SERVIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SERVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["s1431", "s1427"]},
        "2": {"morpho": ["n-s---nn2-"], "uri": ["s1431", "s1427"]},
    },
    "SERVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s1428"]}},
    "SERVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1431"]}},
    "SERVVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["38041"]}},
    "SERVVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["s1432"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["s1432"]},
    },
    "SESAMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1347"]}},
    "SESCENI": {"-": {"morpho": ["rp--------"], "uri": []}},
    "SESCENTESIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1440"]}},
    "SESCENTI": {"-": {"morpho": ["----------"], "uri": []}},
    "SESCENTIES": {"-": {"morpho": ["rp--------"], "uri": ["s1442"]}},
    "SESQVIALTER": {"-": {"morpho": ["----------"], "uri": []}},
    "SESQVIDIGITVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s1456"]}},
    "SESQVILIBRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1459"]}},
    "SESQVIMODIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s1461"]}},
    "SESQVIPEDALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s1468"]}},
    "SESQVIPES": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s1470"]}},
    "SESQVIPLAGA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1471"]}},
    "SESSILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s1488"]}},
    "SESSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s1491"]}},
    "SESSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s1494"]}},
    "SESTERTIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1499"]}},
    "SESTERTIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["100590"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["s1499"]},
    },
    "SEV": {
        "1": {"morpho": ["----------"], "uri": []},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "SEVEHO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "SEVERA": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "SEVERE": {"-": {"morpho": ["rp--------"], "uri": ["43925"]}},
    "SEVERI": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "SEVERITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s1505"]}},
    "SEVERO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "SEVERVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["s1507"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "SEVOCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s1511"]}},
    "SEX": {"-": {"morpho": ["----------"], "uri": []}},
    "SEXAGENI": {"-": {"morpho": ["rp--------"], "uri": []}},
    "SEXAGESIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["100593"]}},
    "SEXAGINTA": {"-": {"morpho": ["----------"], "uri": []}},
    "SEXCENARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["56248"]}},
    "SEXENNIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s1522"]}},
    "SEXIES": {"-": {"morpho": ["rp--------"], "uri": ["100596"]}},
    "SEXIS": {"-": {"morpho": ["n-s---nn--"], "uri": ["s1526"]}},
    "SEXTANS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["s1532"]}},
    "SEXTARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s1537"]}},
    "SEXTILIS": {
        "-": {"morpho": ["n-s----n3-"], "uri": []},
        "A": {"morpho": ["aps---cn3i"], "uri": ["s1539"]},
        "N": {"morpho": ["n-s---fn3-"], "uri": []},
    },
    "SEXTIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "SEXTVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1540"]}},
    "SEXTVM": {"-": {"morpho": ["rp--------"], "uri": ["53786"]}},
    "SEXTVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["s1541"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "SEXVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["s1543"]}},
    "SI": {
        "1": {"morpho": ["r---------"], "uri": []},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "SIBILO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s1551"]}},
    "SIBILVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "SIBILVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": ["s0664"]},
    },
    "SIBYLLA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SIBYLLINVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "SIC": {"-": {"morpho": ["rp--------"], "uri": ["s1555"]}},
    "SICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1556"]}},
    "SICARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s1557"]}},
    "SICCESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s1565"]}},
    "SICCITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s1568"]}},
    "SICCIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SICCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s1569"]}},
    "SICCVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["56685"]}},
    "SICCVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1571"]}},
    "SICILIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SICILICVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s1577"]}},
    "SICILIENSIS": {
        "N": {"morpho": ["aps---cn3i"], "uri": []},
        "A": {"morpho": ["aps---cn3i"], "uri": []},
    },
    "SICILIMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1578"]}},
    "SICINE": {"-": {"morpho": ["rp--------"], "uri": ["57194"]}},
    "SICINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SICVBI": {"-": {"morpho": ["----------"], "uri": []}},
    "SICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1585"]}},
    "SICVLI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SICVLVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "SICVT": {
        "1": {"morpho": ["rp--------"], "uri": ["27199"]},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "SICYONIVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "SIDEO": {"-": {"morpho": ["v1spia--2-"], "uri": []}},
    "SIDEREVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1593"]}},
    "SIDERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s1598"]}},
    "SIDO": {
        "-": {"morpho": ["v1spia--3-"], "uri": ["s9912"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "SIDVS": {"-": {"morpho": ["n-s---nn3-"], "uri": ["s1600"]}},
    "SIGILLARIA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["s8602"]}},
    "SIGILLATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["48470"]}},
    "SIGILLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1606"]}},
    "SIGNATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s1614"]}},
    "SIGNATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["100606"]}},
    "SIGNIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SIGNIFER": {
        "1": {"morpho": ["n-s---mn2r"], "uri": ["s1616"]},
        "2": {"morpho": ["aps---mn1r"], "uri": ["s1616"]},
    },
    "SIGNIFICANS": {"-": {"morpho": ["aps---an3i"], "uri": ["127208"]}},
    "SIGNIFICATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s1619"]}},
    "SIGNIFICATVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["127214"]}},
    "SIGNIFICATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["s9956"]}},
    "SIGNIFICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s1623"]}},
    "SIGNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s1627"]}},
    "SIGNVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1628"]}},
    "SIL": {"-": {"morpho": ["n-s---nn3-"], "uri": ["s4022"]}},
    "SILACEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1630"]}},
    "SILANIO": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "SILANVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": ["s1631"]},
        "-": {"morpho": ["n-s---mn2-"], "uri": ["s1631"]},
    },
    "SILENS": {"-": {"morpho": ["aps---an3i"], "uri": ["127228"]}},
    "SILENTIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1636"]}},
    "SILEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["s1638"]}},
    "SILER": {"-": {"morpho": ["n-s---nn3-"], "uri": ["s1639"]}},
    "SILESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s1638"]}},
    "SILEX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s1640"]}},
    "SILICEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1644"]}},
    "SILICVLVS": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "SILIGINEVS": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "SILIGO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s1651"]}},
    "SILIO": {"-": {"morpho": ["v1spia--4-"], "uri": []}},
    "SILIQVA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1652"]}},
    "SILPHIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1660"]}},
    "SILVA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1661"]}},
    "SILVATICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1662"]}},
    "SILVESTER": {"-": {"morpho": ["aps---mn3-"], "uri": ["27218"]}},
    "SILVESTRIA": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "SILVESTRIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["100611"]}},
    "SILVIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SILVICOLA": {"-": {"morpho": ["n-s---cn1-"], "uri": ["s1666"]}},
    "SILVICVLTRIX": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "SILVIFRAGVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1670"]}},
    "SILVIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "SILVOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1672"]}},
    "SILVRVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s1673"]}},
    "SIMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1726"]}},
    "SIMIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1678"]}},
    "SIMILAGO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s1682"]}},
    "SIMILE": {
        "1": {"morpho": ["n-s---nn3i"], "uri": []},
        "2": {"morpho": ["rp--------"], "uri": []},
    },
    "SIMILIA": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "SIMILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s1685"]}},
    "SIMILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s1686"]}},
    "SIMILITER": {"-": {"morpho": ["rp--------"], "uri": ["27220"]}},
    "SIMILITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s1688"]}},
    "SIMIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s1678"]}},
    "SIMO": {
        "-": {"morpho": ["v1spia--1-"], "uri": ["s1695"]},
        "N": {"morpho": ["n-s----n1-"], "uri": []},
    },
    "SIMPLEX": {
        "-": {
            "morpho": ["aps---an3-", "aps---an3-"],
            "uri": ["s1698", "s1698"],
        },
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "SIMPLICITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s1700"]}},
    "SIMPLICITER": {"-": {"morpho": ["rp--------"], "uri": ["27224"]}},
    "SIMPLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1707"]}},
    "SIMPVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1710"]}},
    "SIMPVVIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1712"]}},
    "SIMVL": {
        "1": {"morpho": ["rp--------"], "uri": ["s1713"]},
        "2": {"morpho": ["p---------"], "uri": []},
        "3": {"morpho": ["----------"], "uri": []},
    },
    "SIMVLAC": {
        "1": {"morpho": ["rp--------"], "uri": ["44410"]},
        "2": {"morpho": ["----------"], "uri": []},
        "-": {"morpho": ["rp--------"], "uri": ["44410"]},
    },
    "SIMVLACRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1714"]}},
    "SIMVLAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["s1715"]}},
    "SIMVLANS": {
        "-": {"morpho": ["aps---an3i"], "uri": ["100615"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "SIMVLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s1718"]}},
    "SIMVLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s1719"]}},
    "SIMVLATQVE": {
        "1": {"morpho": ["rp--------"], "uri": ["53092"]},
        "2": {"morpho": ["----------"], "uri": []},
        "-": {"morpho": ["rp--------"], "uri": ["53092"]},
    },
    "SIMVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s1722"]}},
    "SIMVLTAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s1723"]}},
    "SIMVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1725"]}},
    "SIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1726"]}},
    "SIN": {"-": {"morpho": ["----------"], "uri": []}},
    "SINCERE": {"-": {"morpho": ["rp--------"], "uri": ["39091"]}},
    "SINCERITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s1730"]}},
    "SINCERVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1480"]}},
    "SINCIPVT": {"-": {"morpho": ["n-s---nn3-"], "uri": ["s1735"]}},
    "SINE": {"-": {"morpho": ["p---------"], "uri": ["s1737"]}},
    "SINGILLARITER": {"-": {"morpho": ["rp--------"], "uri": ["127301"]}},
    "SINGILLATIM": {"-": {"morpho": ["rp--------"], "uri": ["s1740"]}},
    "SINGVLA": {
        "1": {"morpho": ["n-s---fn1-"], "uri": ["s0280"]},
        "2": {"morpho": ["n-s---fn1-"], "uri": ["s0280"]},
    },
    "SINGVLARES": {
        "-": {"morpho": ["n-p---mn3i"], "uri": ["127303"]},
        "N": {"morpho": ["n-s---mn1g"], "uri": []},
    },
    "SINGVLARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s1742"]}},
    "SINGVLARITER": {"-": {"morpho": ["rp--------"], "uri": ["30136"]}},
    "SINGVLATIM": {"-": {"morpho": ["rp--------"], "uri": ["30033"]}},
    "SINGVLTIM": {"-": {"morpho": ["rp--------"], "uri": ["s1747"]}},
    "SINGVLTIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["s1748"]}},
    "SINGVLTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["s1750"]}},
    "SINGVLVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "SINGVLVS": {"-": {"morpho": ["rp--------"], "uri": []}},
    "SINISTER": {"-": {"morpho": ["aps---mn1r"], "uri": ["s1751"]}},
    "SINISTRA": {
        "1": {"morpho": ["n-s---fn1-"], "uri": ["s1751"]},
        "2": {"morpho": ["n-s---fn1-"], "uri": ["s1751"]},
    },
    "SINISTRE": {"-": {"morpho": ["rp--------"], "uri": ["37431"]}},
    "SINISTRORSVM": {"-": {"morpho": ["rp--------"], "uri": ["s1754"]}},
    "SINISTRORSVS": {"-": {"morpho": ["rp--------"], "uri": ["29824"]}},
    "SINISTRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1751"]}},
    "SINO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s1756"]}},
    "SINOPE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SINVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s1762"]}},
    "SINVOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1763"]}},
    "SINVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["s1764"]}},
    "SIPARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1766"]}},
    "SIPHO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s1768"]}},
    "SIPIO": {"-": {"morpho": ["v1spia--3i"], "uri": []}},
    "SIQVIDEM": {"-": {"morpho": ["----------"], "uri": []}},
    "SIREMPS": {"-": {"morpho": ["----------"], "uri": []}},
    "SIRPICIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1786"]}},
    "SIRVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s1788"]}},
    "SISENNA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SISER": {"-": {"morpho": ["n-s---nn3-"], "uri": ["s1790"]}},
    "SISTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s1791"]}},
    "SISTRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1795"]}},
    "SISYPHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SITICVLOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1803"]}},
    "SITIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["s1804"]}},
    "SITIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["s1805"]}},
    "SITVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s0413"]}},
    "SITVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["s4029"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["100625"]},
    },
    "SIVE": {
        "1": {"morpho": ["----------", "----------"], "uri": []},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "SMARAGDVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s1820"]}},
    "SMYRNA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["s1827"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["s1827"]},
    },
    "SMYRNION": {"-": {"morpho": ["n-s---nn2g"], "uri": ["127380"]}},
    "SOBRIETAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s1833"]}},
    "SOBRINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1834"]}},
    "SOBRIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1836"]}},
    "SOCCATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1837"]}},
    "SOCCVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s1839"]}},
    "SOCCVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s1840"]}},
    "SOCER": {"-": {"morpho": ["n-s---mn2r"], "uri": ["s9995"]}},
    "SOCIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1844"]}},
    "SOCIABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s1845"]}},
    "SOCIALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s1846"]}},
    "SOCIALITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s1847"]}},
    "SOCIETAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s1851"]}},
    "SOCIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s1852"]}},
    "SOCIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["s1844"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["s1844"]},
    },
    "SOCORDIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1856"]}},
    "SOCORDITER": {"-": {"morpho": ["rp--------"], "uri": ["39005"]}},
    "SOCORS": {"-": {"morpho": ["aps---an3-"], "uri": ["s1857"]}},
    "SOCRATES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "SOCRVS": {"-": {"morpho": ["n-s---fn4-"], "uri": ["27243"]}},
    "SODALICIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1862"]}},
    "SODALIS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["27245"]},
        "2": {"morpho": ["aps---cn3i"], "uri": ["s1863"]},
    },
    "SODALITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s1864"]}},
    "SODALITIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["127406"]}},
    "SODES": {"-": {"morpho": ["rp--------"], "uri": ["42386"]}},
    "SOL": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["s1867"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "SOLACIOLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1868"]}},
    "SOLACIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1869"]}},
    "SOLAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["s1871"]}},
    "SOLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s1881"]}},
    "SOLDVRII": {"-": {"morpho": ["n-p---mn2-"], "uri": ["s1884"]}},
    "SOLEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1885"]}},
    "SOLEATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1888"]}},
    "SOLENNE": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "SOLENS": {"-": {"morpho": ["aps---an3i"], "uri": ["100634"]}},
    "SOLEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["s1890"]}},
    "SOLICISMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["127430"]}},
    "SOLIDATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s1895"]}},
    "SOLIDESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s1897"]}},
    "SOLIDITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s1899"]}},
    "SOLIDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s1900"]}},
    "SOLIDVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["45055"]},
        "2": {"morpho": ["rp--------"], "uri": ["127443"]},
    },
    "SOLIDVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["127444"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["s1901"]},
    },
    "SOLIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["s1902"]}},
    "SOLITARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1911"]}},
    "SOLITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s1915"]}},
    "SOLITVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["36358"]}},
    "SOLITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["27255"]}},
    "SOLIVAGVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1916"]}},
    "SOLIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1918"]}},
    "SOLLEMNE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["s0662"]}},
    "SOLLEMNIS": {"-": {"morpho": ["aps---cn3-"], "uri": []}},
    "SOLLERS": {
        "-": {"morpho": ["aps---cn3i"], "uri": ["s1923"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "SOLLERTER": {"-": {"morpho": ["rp--------"], "uri": ["27258"]}},
    "SOLLERTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1924"]}},
    "SOLLICITATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s1925"]}},
    "SOLLICITE": {"-": {"morpho": ["rp--------"], "uri": ["55763"]}},
    "SOLLICITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s1927"]}},
    "SOLLICITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s1928"]}},
    "SOLLICITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1929"]}},
    "SOLO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["s1932"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["s1932"]},
    },
    "SOLOECISMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s1933"]}},
    "SOLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["s1937"]}},
    "SOLSTITIALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s1941"]}},
    "SOLSTITIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1942"]}},
    "SOLVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["s1944"]},
        "2": {"morpho": ["rp--------"], "uri": ["27266"]},
    },
    "SOLVNTINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SOLVNTINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "SOLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s1946"]}},
    "SOLVS": {"-": {"morpho": ["----------"], "uri": []}},
    "SOLVTE": {"-": {"morpho": ["rp--------"], "uri": ["40886"]}},
    "SOLVTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s1950"]}},
    "SOLVTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["100640"]}},
    "SOLVTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["27269"]}},
    "SOMNIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s1961"]}},
    "SOMNIVM": {
        "-": {"morpho": ["n-s---nn2-"], "uri": ["s1963"]},
        "N": {"morpho": ["n-s---nn2-"], "uri": ["s1963"]},
    },
    "SOMNVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["s1967"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["s1967"]},
    },
    "SONANS": {"-": {"morpho": ["aps---an3i"], "uri": ["100642"]}},
    "SONAX": {"-": {"morpho": ["aps---an3i"], "uri": ["s1969"]}},
    "SONIPES": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["27274"]},
        "2": {"morpho": ["aps---an3-"], "uri": ["s1972"]},
    },
    "SONITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["s9957"]}},
    "SONIVIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1974"]}},
    "SONO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["s1975"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["s1975"]},
    },
    "SONOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s1976"]}},
    "SONORVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1978"]}},
    "SONS": {
        "1": {"morpho": ["n-s----n3-"], "uri": []},
        "2": {"morpho": ["aps---an3i"], "uri": ["s1979"]},
    },
    "SONTICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1980"]}},
    "SONVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1981"]}},
    "SOPATER": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "SOPHIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1982"]}},
    "SOPHISMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["s1983"]}},
    "SOPHISTES": {"-": {"morpho": ["n-s---mn1-"], "uri": []}},
    "SOPIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["s1991"]}},
    "SOPITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s1993"]}},
    "SOPOR": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["s1994"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "SOPORATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["44519"]}},
    "SOPORIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["s1995"]}},
    "SOPORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s1996"]}},
    "SOPORVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1997"]}},
    "SORBEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["s1999"]}},
    "SORBITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s2003"]}},
    "SORBVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s2006"]}},
    "SORBVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["s2007"]}},
    "SORDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["s2008"]}},
    "SORDES": {"-": {"morpho": ["n-p---fn3i"], "uri": ["s2009"]}},
    "SORDESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s2008"]}},
    "SORDIDATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["53481"]}},
    "SORDIDE": {"-": {"morpho": ["rp--------"], "uri": ["27288"]}},
    "SORDIDVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2012"]}},
    "SORDIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2013"]}},
    "SORITES": {"-": {"morpho": ["n-s---mn1g"], "uri": ["s2021"]}},
    "SOROR": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["s2023"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "SORS": {"-": {"morpho": ["n-s---fn3i"], "uri": []}},
    "SORTIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["s2036"]}},
    "SORTITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s2037"]}},
    "SORTITO": {"-": {"morpho": ["rp--------"], "uri": ["36669"]}},
    "SORTITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s2038"]}},
    "SORTITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["s2058"]}},
    "SOSIPPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SOSITHEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SOSPES": {"-": {"morpho": ["aps---an3-"], "uri": ["s2040"]}},
    "SOSPITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s2046"]}},
    "SOSTRATVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SPADIX": {"-": {"morpho": ["n-s---cn3-"], "uri": ["s2053"]}},
    "SPADO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s2054"]}},
    "SPARGO": {
        "1": {"morpho": ["n-s---fn3-"], "uri": ["s2063"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["s2064"]},
    },
    "SPARSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["100652"]}},
    "SPARTA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SPARTEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2070"]}},
    "SPARTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s2071"]}},
    "SPARVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s7075", "s2075"]}},
    "SPATHA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2083"]}},
    "SPATIATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s2085"]}},
    "SPATIOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["s2087"]}},
    "SPATIOSE": {"-": {"morpho": ["rp--------"], "uri": ["33901"]}},
    "SPATIOSVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "SPATIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2089"]}},
    "SPATIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s2090"]}},
    "SPECIALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s2093"]}},
    "SPECIATIM": {"-": {"morpho": ["rp--------"], "uri": ["s2096"]}},
    "SPECIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["s2098"]}},
    "SPECILLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s2101"]}},
    "SPECIMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["s2102"]}},
    "SPECIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["s2104"]}},
    "SPECIOSE": {"-": {"morpho": ["rp--------"], "uri": ["33147"]}},
    "SPECIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2106"]}},
    "SPECTABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s2107"]}},
    "SPECTACVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s2109"]}},
    "SPECTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s2112"]}},
    "SPECTATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s2114"]}},
    "SPECTATRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s2115"]}},
    "SPECTATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["40434"]}},
    "SPECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s2118"]}},
    "SPECVLA": {
        "1": {"morpho": ["n-s---fn1-"], "uri": ["s9941", "s2120"]},
        "2": {"morpho": ["n-s---fn1-"], "uri": ["s9941", "s2120"]},
    },
    "SPECVLARE": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "SPECVLARIA": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "SPECVLARIS": {
        "1": {"morpho": ["aps---cn3i"], "uri": ["s2124"]},
        "2": {"morpho": ["aps---cn3i"], "uri": ["s2124"]},
        "-": {"morpho": ["aps---cn3i"], "uri": ["s2124"]},
    },
    "SPECVLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s2128"]}},
    "SPECVLATORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2129"]}},
    "SPECVLATRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s2130"]}},
    "SPECVLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["s2132"]}},
    "SPECVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s2133"]}},
    "SPECVS": {"-": {"morpho": ["n-s---mn4-"], "uri": []}},
    "SPELAEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s2136"]}},
    "SPELVNCA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["s2138"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "SPERATA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["127612"]}},
    "SPERATVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "SPERATVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": ["100662"]},
    },
    "SPERGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["127615"]}},
    "SPERNENDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["51314"]}},
    "SPERNENS": {
        "-": {"morpho": ["aps---an3-"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "SPERNO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s2146"]}},
    "SPERNOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132724"]}},
    "SPERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s2147"]}},
    "SPES": {
        "-": {"morpho": ["n-s---fn5-"], "uri": ["s2149"]},
        "N": {"morpho": ["n-s----n5-"], "uri": []},
    },
    "SPHAERA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2154"]}},
    "SPHAERITA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["s2160"]}},
    "SPHAEROIDES": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "SPHAEROMACHIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2162"]}},
    "SPICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1284"]}},
    "SPICESCO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "SPICEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2178"]}},
    "SPICIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["s2179"]}},
    "SPICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["49675"]}},
    "SPICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s2181"]}},
    "SPICVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s2184"]}},
    "SPINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2201"]}},
    "SPINETVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s2188"]}},
    "SPINOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2194"]}},
    "SPINTRIA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["s2196"]}},
    "SPINVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["s2201", "s2201"]}},
    "SPIRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2204"]}},
    "SPIRABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s2205"]}},
    "SPIRACVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s2206"]}},
    "SPIRAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["s2208"]}},
    "SPIRAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s2209"]}},
    "SPIRITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["s2214"]}},
    "SPIRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s2215"]}},
    "SPISSAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s2217"]}},
    "SPISSE": {"-": {"morpho": ["rp--------"], "uri": ["54391"]}},
    "SPISSITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s2221"]}},
    "SPISSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s2223"]}},
    "SPISSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2224"]}},
    "SPLEN": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s2228"]}},
    "SPLENDENS": {"-": {"morpho": ["aps---an3i"], "uri": ["100668"]}},
    "SPLENDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["s2229"]}},
    "SPLENDESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s0486"]}},
    "SPLENDIDE": {"-": {"morpho": ["rp--------"], "uri": ["32230"]}},
    "SPLENDIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2232"]}},
    "SPLENDOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s2235"]}},
    "SPOLIARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s2248"]}},
    "SPOLIATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s2249"]}},
    "SPOLIATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s2250"]}},
    "SPOLIATVS": {
        "1": {"morpho": ["aps---mn1-"], "uri": ["38820"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["38820"]},
    },
    "SPOLIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s2252"]}},
    "SPOLIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s2253"]}},
    "SPONDA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2255"]}},
    "SPONDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["s2260"]}},
    "SPONDYLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["32743"]}},
    "SPONGIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1676"]}},
    "SPONS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s2271"]}},
    "SPONSA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s9942"]}},
    "SPONSALIA": {"-": {"morpho": ["n-p---nn3i"], "uri": ["s2273"]}},
    "SPONSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s2274"]}},
    "SPONSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s2277"]}},
    "SPONSVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s9942"]}},
    "SPONSVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["s9942", "s9942"]},
        "2": {"morpho": ["n-s---mn4-"], "uri": ["s9942", "s9942"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["s9942", "s9942"]},
    },
    "SPONTE": {"-": {"morpho": ["rp--------"], "uri": ["54456"]}},
    "SPORTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2282"]}},
    "SPORTVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2285"]}},
    "SPRETOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s2288"]}},
    "SPRETVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["s2148"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["127737"]},
    },
    "SPVMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2289"]}},
    "SPVMANS": {
        "-": {"morpho": ["aps---an3-"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "SPVMESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s2292"]}},
    "SPVMEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2293"]}},
    "SPVMIGER": {"-": {"morpho": ["aps---mn1r"], "uri": ["s2297"]}},
    "SPVMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s2298"]}},
    "SPVMOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2299"]}},
    "SPVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s2300"]}},
    "SPVRCE": {"-": {"morpho": ["rp--------"], "uri": ["28568"]}},
    "SPVRCITIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2305"]}},
    "SPVRCITIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["s2305"]}},
    "SPVRCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s2306"]}},
    "SPVRCVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2307"]}},
    "SPVRIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["s2308"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["s2308"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["s2308"]},
    },
    "SPVTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s9961"]}},
    "SPVTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["s9961"]}},
    "SQVALENS": {
        "-": {"morpho": ["aps---an3-"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "SQVALEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["s2314"]}},
    "SQVALIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2316"]}},
    "SQVALOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s2320"]}},
    "SQVAMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2322"]}},
    "SQVAMEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2326"]}},
    "SQVAMIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["s2328"]}},
    "SQVAMIGER": {"-": {"morpho": ["aps---mn1r"], "uri": ["s2329"]}},
    "SQVAMOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2330"]}},
    "SQVILLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2335"]}},
    "STABILIA": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "STABILIMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["s2337"]}},
    "STABILIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["s2339"]}},
    "STABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s2340"]}},
    "STABILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s2341"]}},
    "STABILITER": {"-": {"morpho": ["rp--------"], "uri": ["127779"]}},
    "STABILITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s2343"]}},
    "STABVLARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s2344"]}},
    "STABVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s2346"]}},
    "STABVLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["133128"]}},
    "STABVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s2347"]}},
    "STACTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2349"]}},
    "STADIATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2353"]}},
    "STADIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s2355"]}},
    "STAGNANS": {"-": {"morpho": ["aps---an3i"], "uri": ["100683"]}},
    "STAGNO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["s4031"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["s4031"]},
    },
    "STAGNVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["s2364"]},
        "2": {"morpho": ["n-s---nn2-"], "uri": ["s2364"]},
    },
    "STAIENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "STAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["s2370"]}},
    "STAMINEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2373"]}},
    "STATERA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2386"]}},
    "STATICVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s2389"]}},
    "STATIM": {"-": {"morpho": ["rp--------"], "uri": ["s2390"]}},
    "STATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s2391"]}},
    "STATIVA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["100686"]}},
    "STATIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "STATIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2394"]}},
    "STATOR": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["s4032"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "STATVA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2396"]}},
    "STATVARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["s2397"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["s2397"]},
    },
    "STATVMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["s2401"]}},
    "STATVMINATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s2402"]}},
    "STATVMINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s2403"]}},
    "STATVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s2405"]}},
    "STATVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1812"]}},
    "STATVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["s9962"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["100688"]},
    },
    "STATVTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s2407"]}},
    "STATVTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["127841"]}},
    "STATVTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["127842"]}},
    "STELLA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["s2415"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "STELLIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["s2419"]}},
    "STELLIGER": {"-": {"morpho": ["aps---mn1r"], "uri": ["s2420"]}},
    "STELLIO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s2422"]}},
    "STELLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s2424"]}},
    "STEMMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["s2427"]}},
    "STERCORARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2436"]}},
    "STERCORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s2439"]}},
    "STERCOROSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2440"]}},
    "STERCVLINVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s2441"]}},
    "STERCVS": {"-": {"morpho": ["n-s---nn3-"], "uri": ["s2442"]}},
    "STEREOBATES": {"-": {"morpho": ["n-s---mn1g"], "uri": ["s2443"]}},
    "STERILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s2449"]}},
    "STERILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s2450"]}},
    "STERILVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2449"]}},
    "STERNO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s2452"]}},
    "STERNVMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s2453"]}},
    "STERNVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s2454"]}},
    "STERNVTAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s2455"]}},
    "STERTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s2460"]}},
    "STESICHORVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "STHENIANVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "STHENIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "STIGMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["s2466"]}},
    "STILLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2470"]}},
    "STILLARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s2471"]}},
    "STILLICIDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s2476"]}},
    "STILLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s2478"]}},
    "STILVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s2480"]}},
    "STIMVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s2486"]}},
    "STIMVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s2487"]}},
    "STINGVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s2488"]}},
    "STIPATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s2493"]}},
    "STIPATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["100698"]}},
    "STIPENDIALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s2495"]}},
    "STIPENDIARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["48965"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["s2496"]},
    },
    "STIPENDIOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["s2497"]}},
    "STIPENDIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2498"]}},
    "STIPENDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s2499"]}},
    "STIPES": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s2500"]}},
    "STIPO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s2502"]}},
    "STIPS": {"-": {"morpho": ["n-s---mn3-"], "uri": []}},
    "STIPVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2839"]}},
    "STIPVLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s2505"]}},
    "STIPVLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["s2508"]}},
    "STIRIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2509"]}},
    "STIRPS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["s2515"]}},
    "STITVO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "STIVA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2516"]}},
    "STLATARIVS": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "STLOPPVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s2520"]}},
    "STO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s1430"]}},
    "STOECHION": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "STOICA": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["n-s----n-g"], "uri": []},
    },
    "STOICI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "STOICVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "STOLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2526"]}},
    "STOLATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2528"]}},
    "STOLIDE": {"-": {"morpho": ["rp--------"], "uri": ["27396"]}},
    "STOLIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2530"]}},
    "STOMACHICVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s2535"]}},
    "STOMACHOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["s2536"]}},
    "STOMACHOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2537"]}},
    "STOMACHVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s2538"]}},
    "STOREA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2546"]}},
    "STRABO": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["s2547"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "STRAGES": {"-": {"morpho": ["n-s---fn3i"], "uri": ["s2550"]}},
    "STRAGVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2551"]}},
    "STRAGVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s2551"]}},
    "STRAGVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2551"]}},
    "STRAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["s2553"]}},
    "STRAMENTARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2554"]}},
    "STRAMENTICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2555"]}},
    "STRAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s2557"]}},
    "STRAMINEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2558"]}},
    "STRANGVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s2564"]}},
    "STRANGVRIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2565"]}},
    "STRATVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s9943"]}},
    "STRENVE": {"-": {"morpho": ["rp--------"], "uri": ["48269"]}},
    "STRENVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2584"]}},
    "STREPITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s2585"]}},
    "STREPITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["s9965"]}},
    "STREPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s2586"]}},
    "STRIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2590"]}},
    "STRICTIM": {"-": {"morpho": ["rp--------"], "uri": ["s2592"]}},
    "STRICTIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2595"]}},
    "STRICTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s2596"]}},
    "STRICTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["35443"]}},
    "STRIDENS": {
        "-": {"morpho": ["aps---an3-"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "STRIDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["s2599"]}},
    "STRIDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s2599"]}},
    "STRIDOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s2600"]}},
    "STRIDVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2601"]}},
    "STRIGILIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["s2604"]}},
    "STRIGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s2607"]}},
    "STRINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s2612"]}},
    "STRINGOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s2613"]}},
    "STRIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s2614"]}},
    "STRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s4034"]}},
    "STROPHA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2624"]}},
    "STROPHIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s2629"]}},
    "STROPHVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s2632"]}},
    "STRVCTILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s2633"]}},
    "STRVCTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s2635"]}},
    "STRVCTVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2652"]}},
    "STRVES": {"-": {"morpho": ["n-s---fn3i"], "uri": ["s2637"]}},
    "STRVMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2640"]}},
    "STRVMOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2644"]}},
    "STRVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s2646"]}},
    "STRVTHEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s3105"]}},
    "STRVTHOCAMELVS": {"-": {"morpho": ["n-s---mn2-"], "uri": []}},
    "STVDENS": {
        "-": {"morpho": ["aps---an3-"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "STVDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["s2656"]}},
    "STVDIOSE": {"-": {"morpho": ["rp--------"], "uri": ["27423"]}},
    "STVDIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2658"]}},
    "STVDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s2659"]}},
    "STVLTE": {"-": {"morpho": ["rp--------"], "uri": ["27425"]}},
    "STVLTI": {
        "-": {"morpho": ["n-s----n2-"], "uri": []},
        "N": {"morpho": ["n-p---mn2-"], "uri": []},
    },
    "STVLTITIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2663"]}},
    "STVLTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2665"]}},
    "STVPEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["s2666"]}},
    "STVPEFIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["s2667"]}},
    "STVPEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["s2668"]}},
    "STVPESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s2156"]}},
    "STVPIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2670"]}},
    "STVPOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s2671"]}},
    "STVPPA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2673"]}},
    "STVPPEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2676"]}},
    "STVPRATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s2677"]}},
    "STVPRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s2679"]}},
    "STVPRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s2681"]}},
    "STYGIVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["44369"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["44369"]},
    },
    "STYLOBATA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["128067"]}},
    "STYX": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "SUPERINICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["s3602"]}},
    "SVADEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["s2695"]}},
    "SVASIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s2700"]}},
    "SVASOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s2701"]}},
    "SVASORIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2702"]}},
    "SVASVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["s2713"]}},
    "SVAVE": {"-": {"morpho": ["rp--------"], "uri": ["46784"]}},
    "SVAVIA": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "SVAVIDICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2705"]}},
    "SVAVILOQVENS": {"-": {"morpho": ["aps---an3i"], "uri": ["s2708"]}},
    "SVAVIOLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["53254"]}},
    "SVAVIOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132727"]}},
    "SVAVIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s2711"]}},
    "SVAVITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s2717"]}},
    "SVAVITER": {"-": {"morpho": ["rp--------"], "uri": ["27439"]}},
    "SVAVIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": []}},
    "SVB": {"-": {"morpho": ["p---------"], "uri": ["s2719"]}},
    "SVBACIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2723"]}},
    "SVBACTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s2725"]}},
    "SVBAERATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2732"]}},
    "SVBARESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s2751"]}},
    "SVBCAVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2782"]}},
    "SVBCERNO": {"-": {"morpho": ["v1spia--3-"], "uri": ["100723"]}},
    "SVBCRISPVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2795"]}},
    "SVBCRVDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2797"]}},
    "SVBDEFICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["s2808"]}},
    "SVBDIFFICILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s2814"]}},
    "SVBDITIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2823"]}},
    "SVBDIV": {"-": {"morpho": ["rp--------"], "uri": ["128212"]}},
    "SVBDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s2828"]}},
    "SVBDOLE": {"-": {"morpho": ["rp--------"], "uri": ["43881"]}},
    "SVBDOLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2833"]}},
    "SVBDVCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s2836"]}},
    "SVBDVCTARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2837"]}},
    "SVBDVCTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s2838"]}},
    "SVBEDO": {"2": {"morpho": ["v1spia--3-"], "uri": ["s2843"]}},
    "SVBEO": {
        "1": {"morpho": ["v1spia--4-"], "uri": ["s2844"]},
        "-": {"morpho": ["v1spia--4-"], "uri": ["s2844"]},
    },
    "SVBER": {"-": {"morpho": ["n-s---nn3-"], "uri": ["s2845"]}},
    "SVBFVSCVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "SVBIACEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["s2870"]}},
    "SVBICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["s2872"]}},
    "SVBIECTE": {"-": {"morpho": ["rp--------"], "uri": ["53203"]}},
    "SVBIECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s2875"]}},
    "SVBIECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s2877"]}},
    "SVBIECTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s2878"]}},
    "SVBIECTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s9969"]}},
    "SVBIECTVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["100736"]},
        "2": {"morpho": ["n-s---mn4-"], "uri": ["s9969"]},
        "3": {"morpho": ["aps---mn1-"], "uri": ["36202"]},
    },
    "SVBIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s2882"]}},
    "SVBINDE": {"-": {"morpho": ["rp--------"], "uri": ["s2886"]}},
    "SVBITARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2907"]}},
    "SVBITO": {"-": {"morpho": ["rp--------"], "uri": ["47747"]}},
    "SVBITVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["100733"]}},
    "SVBITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["27445"]}},
    "SVBIVGIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2913"]}},
    "SVBIVNGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s2920"]}},
    "SVBLABOR": {"2": {"morpho": ["v1spid--3-"], "uri": ["s2921"]}},
    "SVBLECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s2930"]}},
    "SVBLEGO": {"2": {"morpho": ["v1spia--3-"], "uri": ["s2931"]}},
    "SVBLEVO": {"1": {"morpho": ["v1spia--1-"], "uri": ["s2934"]}},
    "SVBLICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2935"]}},
    "SVBLICIVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["s2936"]},
        "N": {"morpho": ["aps---mn1-"], "uri": ["s2936"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["s2936"]},
    },
    "SVBLIDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s2937"]}},
    "SVBLIGAR": {"-": {"morpho": ["n-s---nn3i"], "uri": ["s2939"]}},
    "SVBLIGO": {
        "2": {"morpho": ["v1spia--1-"], "uri": ["s2941"]},
        "3": {"morpho": ["v1spia--1-"], "uri": ["s2941"]},
    },
    "SVBLIME": {
        "1": {"morpho": ["n-s---nn3i"], "uri": ["100739"]},
        "2": {"morpho": ["rp--------"], "uri": ["48889"]},
    },
    "SVBLIMIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s2944"]}},
    "SVBLIMITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s2945"]}},
    "SVBLIMITER": {"-": {"morpho": ["rp--------"], "uri": ["100740"]}},
    "SVBLIMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s2947"]}},
    "SVBLIMVS": {"1": {"morpho": ["aps---mn1-"], "uri": ["s2944"]}},
    "SVBLINO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s2949"]}},
    "SVBLVCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["s2952"]}},
    "SVBLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s2957"]}},
    "SVBLVSTRIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s2958"]}},
    "SVBMANO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "SVBMERGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s2968"]}},
    "SVBMINISTRATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s2975"]}},
    "SVBMINISTRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s2976"]}},
    "SVBMISSE": {"-": {"morpho": ["rp--------"], "uri": ["42635"]}},
    "SVBMISSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["38259"]}},
    "SVBMITTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s2979"]}},
    "SVBMOVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["s2984"]}},
    "SVBMVGIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["s2985"]}},
    "SVBNASCOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["s2993"]}},
    "SVBNECTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s2996"]}},
    "SVBNITOR": {
        "2": {"morpho": ["v1spid--3-"], "uri": []},
        "-": {"morpho": ["v1spid--3-"], "uri": []},
    },
    "SVBNIXVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s3003"]}},
    "SVBNOTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s3005"]}},
    "SVBNVBA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s3006"]}},
    "SVBNVBILVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s3007"]}},
    "SVBO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s3008"]}},
    "SVBOLEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["s3015"]}},
    "SVBOLES": {"-": {"morpho": ["n-s---fn3i"], "uri": ["s3016"]}},
    "SVBORIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["s3019"]}},
    "SVBORNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s3022"]}},
    "SVBORTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["s9974"]}},
    "SVBRADO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3042"]}},
    "SVBRECTITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s3046"]}},
    "SVBRECTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["128396"]}},
    "SVBREPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3055"]}},
    "SVBRIDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["s3059"]}},
    "SVBRIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["27465"]}},
    "SVBRIPIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["s3065"]}},
    "SVBROGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s3069"]}},
    "SVBROTATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s3071"]}},
    "SVBRVBEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["s3074"]}},
    "SVBRVBICVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s3076"]}},
    "SVBRVFVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s3077"]}},
    "SVBRVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3081"]}},
    "SVBSCRIBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3097"]}},
    "SVBSCRIPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s3098"]}},
    "SVBSCVS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s3101"]}},
    "SVBSECIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["31467"]}},
    "SVBSECO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s3102"]}},
    "SVBSELLIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s3107"]}},
    "SVBSEQVOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["s3110"]}},
    "SVBSERO": {
        "2": {"morpho": ["v1spia--3-"], "uri": ["s3114"]},
        "3": {"morpho": ["v1spia--3-"], "uri": ["s3114"]},
    },
    "SVBSESSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s3116"]}},
    "SVBSIDENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s9925"]}},
    "SVBSIDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["$0215"]}},
    "SVBSIDIARII": {"-": {"morpho": ["n-p---mn2-"], "uri": ["128465"]}},
    "SVBSIDIARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s3120"]}},
    "SVBSIDIOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["s3121"]}},
    "SVBSIDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s3122"]}},
    "SVBSIDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3123"]}},
    "SVBSILIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["s3128"]}},
    "SVBSILVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "SVBSISTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3133"]}},
    "SVBSORTIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["s3140"]}},
    "SVBSORTITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s3141"]}},
    "SVBSTANTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s7603"]}},
    "SVBSTERNO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3148"]}},
    "SVBSTILLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s3149"]}},
    "SVBSTITVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3150"]}},
    "SVBSTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s6133"]}},
    "SVBSTRINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3158"]}},
    "SVBSTRVCTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s3159"]}},
    "SVBSTRVCTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["128496"]}},
    "SVBSTRVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3160"]}},
    "SVBSVM": {"1": {"morpho": ["v1spia--2-"], "uri": ["s3163"]}},
    "SVBSVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3164"]}},
    "SVBTEGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3169"]}},
    "SVBTEMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["s3172"]}},
    "SVBTENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3173"]}},
    "SVBTER": {
        "1": {"morpho": ["rp--------"], "uri": ["27481"]},
        "2": {"morpho": ["p---------"], "uri": ["s3176"]},
    },
    "SVBTERFLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3183"]}},
    "SVBTERFVGIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["s3184"]}},
    "SVBTERLABOR": {"2": {"morpho": ["v1spid--3-"], "uri": ["s3190"]}},
    "SVBTERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3198"]}},
    "SVBTERRANEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["128541"]}},
    "SVBTERRANEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s3200"]}},
    "SVBTEXO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3208"]}},
    "SVBTILE": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "SVBTILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s3212"]}},
    "SVBTILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s3213"]}},
    "SVBTILITER": {"-": {"morpho": ["rp--------"], "uri": ["32440"]}},
    "SVBTRAHO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3219"]}},
    "SVBTVS": {
        "1": {"morpho": ["rp--------"], "uri": ["s3227"]},
        "2": {"morpho": ["p---------"], "uri": []},
    },
    "SVBTVSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["51252"]}},
    "SVBVCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s3231"]}},
    "SVBVECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s3232"]}},
    "SVBVECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s3233"]}},
    "SVBVECTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s3234"]}},
    "SVBVECTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["s3245"]}},
    "SVBVEHO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3235"]}},
    "SVBVENIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["s3238"]}},
    "SVBVERSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["128582"]}},
    "SVBVERSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s6236"]}},
    "SVBVERTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3237"]}},
    "SVBVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s3242"]}},
    "SVBVLCVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s3243"]}},
    "SVBVOLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3251"]}},
    "SVBVRBANI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["100757"]}},
    "SVBVRBANITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s3252"]}},
    "SVBVRBANVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["43729"]}},
    "SVBVRBANVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s3253"]}},
    "SVBVRGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["s3256"]}},
    "SVCCEDENS": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "SVCCEDO": {"1": {"morpho": ["v1spia--3-"], "uri": ["s3260"]}},
    "SVCCENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3261"]}},
    "SVCCENSEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["100758"]}},
    "SVCCESSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s3271"]}},
    "SVCCESSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s3272"]}},
    "SVCCESSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["s3269"]}},
    "SVCCIDIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s3275"]}},
    "SVCCIDO": {
        "1": {"morpho": ["v1spia--3-"], "uri": ["s3277", "s3277"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["s3277", "s3277"]},
    },
    "SVCCIDVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s3278"]}},
    "SVCCINCTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["100761"]}},
    "SVCCINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3283"]}},
    "SVCCINO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3285"]}},
    "SVCCIPIO": {"-": {"morpho": ["v1spia--3i"], "uri": []}},
    "SVCCRESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3294"]}},
    "SVCCVBO": {"2": {"morpho": ["v1spia--1-"], "uri": [None, "s3297"]}},
    "SVCCVMBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3300"]}},
    "SVCCVRRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3302"]}},
    "SVCCVTIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["s3308"]}},
    "SVCIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s3311"]}},
    "SVCINVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s3471"]}},
    "SVCINVS": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": ["s3448"]},
    },
    "SVCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s3316"]}},
    "SVCVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s3670"]}},
    "SVDARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s3323"]}},
    "SVDATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s3325"]}},
    "SVDATORIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s3327"]}},
    "SVDIS": {"-": {"morpho": ["n-s---fn3i"], "uri": []}},
    "SVDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s3330"]}},
    "SVDOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s3331"]}},
    "SVDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s3335"]}},
    "SVESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3338"]}},
    "SVESSA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SVETTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SVETVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["27507"]}},
    "SVFES": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s3339"]}},
    "SVFFERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3345"]}},
    "SVFFICIENS": {"-": {"morpho": ["aps---an3i"], "uri": ["27510"]}},
    "SVFFICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["s3351"]}},
    "SVFFIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3352"]}},
    "SVFFIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["s3356"]}},
    "SVFFLAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["s3361"]}},
    "SVFFLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s3366"]}},
    "SVFFOCATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s3368"]}},
    "SVFFOCO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["s3369"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["s3369"]},
    },
    "SVFFODIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["s3370"]}},
    "SVFFOSSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s3371"]}},
    "SVFFRAGATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s3373"]}},
    "SVFFRAGATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s3374"]}},
    "SVFFRAGIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s3379"]}},
    "SVFFRAGOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["s3381"]}},
    "SVFFVGIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["s3387"]}},
    "SVFFVGIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s3388"]}},
    "SVFFVLCIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["s3389"]}},
    "SVFFVNDO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["s3396"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["s3395"]},
    },
    "SVFFVSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s3399"]}},
    "SVGGERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3402"]}},
    "SVGGESTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s9979"]}},
    "SVGGESTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["s9979"]}},
    "SVGGREDIOR": {"-": {"morpho": ["v1spid--3i"], "uri": ["s3407"]}},
    "SVGGRVNDA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s3408"]}},
    "SVGGRVNDATIO": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "SVGGRVNDIVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "SVGILLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["38038"]}},
    "SVGILLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s3405"]}},
    "SVGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3414"]}},
    "SVI": {
        "1": {"morpho": ["o-----an--"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "SVILLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s3417"]}},
    "SVLCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s3425"]}},
    "SVLCVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s3426"]}},
    "SVLFVR": {"-": {"morpho": ["n-s---nn3-"], "uri": ["45365"]}},
    "SVLFVRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["40402"]}},
    "SVLFVREVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["56970"]}},
    "SVLFVROSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["128768"]}},
    "SVLLA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SVLPICIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "SVLTO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "SVM": {
        "1": {"morpho": ["v1spia--3-"], "uri": ["s3436"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["s3436"]},
    },
    "SVMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["s3438"]}},
    "SVMMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s3451"]}},
    "SVMMARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s3442"]}},
    "SVMMARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["s3442"]},
        "2": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "SVMMATIM": {"-": {"morpho": ["rp--------"], "uri": ["s3444"]}},
    "SVMMATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["s3445"]}},
    "SVMME": {"-": {"morpho": ["rp--------"], "uri": ["33379"]}},
    "SVMMOPERE": {"-": {"morpho": ["rp--------"], "uri": ["38702"]}},
    "SVMMVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s3450"]}},
    "SVMMVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["27536"]},
        "2": {"morpho": ["rp--------"], "uri": ["100777"]},
    },
    "SVMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3452"]}},
    "SVMPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s3453"]}},
    "SVMPTVARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s3455"]}},
    "SVMPTVOSE": {"-": {"morpho": ["rp--------"], "uri": ["45762"]}},
    "SVMPTVOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s3457"]}},
    "SVMPTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["s9980"]},
        "2": {"morpho": ["n-s---mn4-"], "uri": ["s9980"]},
    },
    "SVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3459"]}},
    "SVOVETAVRILIA": {"-": {"morpho": ["n-p---nn3i"], "uri": ["s3460"]}},
    "SVPELLEX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s3462"]}},
    "SVPER": {
        "1": {"morpho": ["rp--------"], "uri": ["27541"]},
        "2": {"morpho": ["p---------"], "uri": ["s3463"]},
    },
    "SVPERA": {
        "1": {"morpho": ["rp--------"], "uri": []},
        "2": {"morpho": ["p---------"], "uri": []},
    },
    "SVPERABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s3464"]}},
    "SVPERADDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["40222"]}},
    "SVPERANS": {"-": {"morpho": ["aps---an3i"], "uri": ["100784"]}},
    "SVPERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s3488"]}},
    "SVPERATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s3489"]}},
    "SVPERBE": {"-": {"morpho": ["rp--------"], "uri": ["38031"]}},
    "SVPERBIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s3493"]}},
    "SVPERBIFICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s3495"]}},
    "SVPERBIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["s3497"]}},
    "SVPERBVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "SVPERBVS": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": ["s3500"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "SVPERCADO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3501"]}},
    "SVPERCERNO": {"-": {"morpho": ["v1spia--3-"], "uri": ["128848"]}},
    "SVPERCILIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s3506"]}},
    "SVPERCILIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s3507"]}},
    "SVPERCRESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3512"]}},
    "SVPERCVRRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3514"]}},
    "SVPEREMINEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["s3529"]}},
    "SVPERFICIARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["128898"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["s3549"]},
    },
    "SVPERFIO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3553"]}},
    "SVPERFLVENS": {"-": {"morpho": ["aps---an3i"], "uri": ["100788"]}},
    "SVPERFLVO": {
        "1": {"morpho": ["v1spia--3-"], "uri": ["s3557"]},
        "2": {"morpho": ["rp--------"], "uri": ["128908"]},
    },
    "SVPERFVNDO": {"2": {"morpho": ["v1spia--3-"], "uri": ["s3565"]}},
    "SVPERGREDIOR": {"-": {"morpho": ["v1spid--3i"], "uri": ["s3571"]}},
    "SVPERI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["100789"]}},
    "SVPERIACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["s3576"]}},
    "SVPERIMPONO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3583"]}},
    "SVPERINCVMBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3590"]}},
    "SVPERIOR": {"-": {"morpho": ["n-s----n3-", "aps---cn3-"], "uri": []}},
    "SVPERIORA": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "SVPERIVS": {"-": {"morpho": ["rp--------"], "uri": ["42731"]}},
    "SVPERLABOR": {"2": {"morpho": ["v1spid--3-"], "uri": ["s3619"]}},
    "SVPERLIGO": {
        "2": {"morpho": ["v1spia--1-"], "uri": ["s3624"]},
        "-": {"morpho": ["v1spia--1-"], "uri": ["s3624"]},
    },
    "SVPERNAS": {"2": {"morpho": ["aps---an3-"], "uri": ["129019"]}},
    "SVPERNE": {"-": {"morpho": ["rp--------"], "uri": ["54858"]}},
    "SVPERNVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "SVPERNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s3649"]}},
    "SVPERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s3650"]}},
    "SVPERPENDENS": {"-": {"morpho": ["aps---an3i"], "uri": ["129042"]}},
    "SVPERPONO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3664"]}},
    "SVPERSCANDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3677"]}},
    "SVPERSEDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["s3680"]}},
    "SVPERSERO": {"3": {"morpho": ["v1spia--3-"], "uri": ["s3683"]}},
    "SVPERSIDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3685"]}},
    "SVPERSPICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["s3691"]}},
    "SVPERSTAGNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s3692"]}},
    "SVPERSTES": {"-": {"morpho": ["aps---an3i"], "uri": ["s3695"]}},
    "SVPERSTITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s3697"]}},
    "SVPERSTITIOSE": {"-": {"morpho": ["rp--------"], "uri": ["45621"]}},
    "SVPERSTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s3700"]}},
    "SVPERSTRVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3702"]}},
    "SVPERSVM": {"1": {"morpho": ["v1spia--3-"], "uri": ["s3704"]}},
    "SVPERTEGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3706"]}},
    "SVPERVACANEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s3715"]}},
    "SVPERVACVVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "SVPERVACVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s3718"]}},
    "SVPERVADO": {"2": {"morpho": ["v1spia--3-"], "uri": ["s3719"]}},
    "SVPERVECTO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "SVPERVEHO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3724"]}},
    "SVPERVEHOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["132736"]}},
    "SVPERVENIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["s3725"]}},
    "SVPERVINCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3728"]}},
    "SVPERVIVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3729"]}},
    "SVPERVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "SVPERVOLO": {"2": {"morpho": ["v1spia--1-"], "uri": ["s3734"]}},
    "SVPERVRGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["s3736"]}},
    "SVPERVS": {"-": {"morpho": ["aps---mn1-"], "uri": []}},
    "SVPINE": {"-": {"morpho": ["rp--------"], "uri": ["100794"]}},
    "SVPINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s3740"]}},
    "SVPINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s3741"]}},
    "SVPPARVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["47821"]}},
    "SVPPEDITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s3748"]}},
    "SVPPERNATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s3750"]}},
    "SVPPETIAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["s3753"]}},
    "SVPPETO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3757"]}},
    "SVPPILO": {"2": {"morpho": ["v1spia--1-"], "uri": ["s3758"]}},
    "SVPPLANTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s3764"]}},
    "SVPPLAVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["129155"]}},
    "SVPPLEMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s3765"]}},
    "SVPPLEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["s3766"]}},
    "SVPPLEX": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["38903"]},
        "2": {"morpho": ["aps---an3-"], "uri": ["s3768"]},
    },
    "SVPPLICATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s3770"]}},
    "SVPPLICITER": {"-": {"morpho": ["rp--------"], "uri": ["54374"]}},
    "SVPPLICIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s3772"]}},
    "SVPPLICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s3773"]}},
    "SVPPONO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3778"]}},
    "SVPPORTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s3780"]}},
    "SVPPRESSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s3787"]}},
    "SVPPRESSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["48181"]}},
    "SVPPRIMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3789"]}},
    "SVPPVRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s3792"]}},
    "SVPPVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s3794"]}},
    "SVPPVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s3795"]}},
    "SVPPVTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s3799"]}},
    "SVPRA": {
        "1": {"morpho": ["rp--------"], "uri": ["27566"]},
        "2": {"morpho": ["p---------"], "uri": ["s3800"]},
    },
    "SVPRADICO": {"2": {"morpho": ["v1spia--3-"], "uri": ["49572"]}},
    "SVPRASCANDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3808"]}},
    "SVPRAVIVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3811"]}},
    "SVPREMO": {
        "-": {"morpho": ["rp--------"], "uri": ["129207"]},
        "1": {"morpho": ["rp--------"], "uri": ["129207"]},
    },
    "SVPREMVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["29833"]},
        "2": {"morpho": ["rp--------"], "uri": ["49008"]},
    },
    "SVRA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["s3814"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["s3814"]},
    },
    "SVRCVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s3821"]}},
    "SVRDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s3828"]}},
    "SVRGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3061"]}},
    "SVRRIPIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["40416"]}},
    "SVRSVM": {"-": {"morpho": ["rp--------"], "uri": ["s3832"]}},
    "SVRSVMVORSVM": {"-": {"morpho": ["rp--------"], "uri": []}},
    "SVS": {"-": {"morpho": ["n-s---cn3-"], "uri": ["s3834"]}},
    "SVSCENSEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["s3836"]}},
    "SVSCEPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s3839"]}},
    "SVSCIPIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["s3844"]}},
    "SVSCITO": {"1": {"morpho": ["v1spia--1-"], "uri": ["s3848"]}},
    "SVSPECTO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["s3853"]},
        "2": {"morpho": ["rp--------"], "uri": ["129279"]},
    },
    "SVSPECTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["s9985"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["40940"]},
    },
    "SVSPENDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s3856"]}},
    "SVSPENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3857"]}},
    "SVSPENSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s3859"]}},
    "SVSPENSVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "SVSPENSVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s3830"]}},
    "SVSPENSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["55321"]}},
    "SVSPICAX": {"-": {"morpho": ["aps---an3i"], "uri": ["s3861"]}},
    "SVSPICIO": {
        "1": {"morpho": ["n-s---fn3-"], "uri": ["s3863"]},
        "2": {"morpho": ["v1spia--3i"], "uri": ["s3862"]},
    },
    "SVSPICIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s3864"]}},
    "SVSPICOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["s3866"]}},
    "SVSPIRIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s3870"]}},
    "SVSPIRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s3871"]}},
    "SVSTENTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s3877"]}},
    "SVSTINEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["s3878"]}},
    "SVSTOLLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3879"]}},
    "SVSVRRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s3884"]}},
    "SVSVRRVS": {"-": {"morpho": ["n-s---mn4-"], "uri": []}},
    "SVTILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s3888"]}},
    "SVTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s3889"]}},
    "SVTORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s3891"]}},
    "SVTRINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s3893"]}},
    "SVTRINVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s3893"]}},
    "SVTRINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s3893"]}},
    "SVVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["32704"]}},
    "SVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["49239"]}},
    "SYLLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SYLLABA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s3911"]}},
    "SYLLANVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "SYLLOGISMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s3918"]}},
    "SYMBOLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["27593"]}},
    "SYMMACHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SYMMETRIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s3924"]}},
    "SYMMETROS": {"-": {"morpho": ["aps---mn1g"], "uri": ["s3925"]}},
    "SYMPATHIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s3928"]}},
    "SYMPHONIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s3930"]}},
    "SYMPHONIACVS": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": ["s3931"]},
    },
    "SYNEMMENOS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "SYNGRAPHA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s3969"]}},
    "SYNGRAPHE": {"-": {"morpho": ["n-s---fn1g"], "uri": ["129395"]}},
    "SYNHECHON": {"-": {"morpho": ["----------"], "uri": []}},
    "SYRACVSAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "SYRACVSANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SYRACVSANVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "SYRI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SYRIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SYRINX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s4006"]}},
    "SYRMA": {"-": {"morpho": ["n-s---nn3-"], "uri": []}},
    "SYRTIS": {
        "N": {"morpho": ["n-s---fn3-"], "uri": ["44899"]},
        "-": {"morpho": ["n-s---fn3-"], "uri": ["44899"]},
    },
    "SYSTYLOS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "TABELLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t0004"]}},
    "TABELLARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["t0006"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["t0006"]},
    },
    "TABEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["t0007"]}},
    "TABERNA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t0462"]}},
    "TABERNACVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0010"]}},
    "TABERNARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["t0011"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["t0011"]},
    },
    "TABES": {
        "-": {"morpho": ["n-s---fn3i"], "uri": ["t0013"]},
        "N": {"morpho": ["n-s---mn1g"], "uri": []},
    },
    "TABESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t0007"]}},
    "TABIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0015"]}},
    "TABIFICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0018"]}},
    "TABVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t0025"]}},
    "TABVLARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0028"]}},
    "TABVLARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["t0028"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["t0028"]},
    },
    "TABVLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t0030"]}},
    "TABVLATVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0031"]}},
    "TABVLINVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0032"]}},
    "TABVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0034"]}},
    "TACEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["t0035"]}},
    "TACITE": {"-": {"morpho": ["rp--------"], "uri": ["33058"]}},
    "TACITVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["100818"]}},
    "TACITVRNITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t0037"]}},
    "TACITVRNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0038"]}},
    "TACITVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["58354"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": ["58354"]},
    },
    "TACTILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["t0039"]}},
    "TACTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t0041"]}},
    "TACTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["t9994"]}},
    "TADIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TAEDA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t0042"]}},
    "TAEDET": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "TAEDIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["t0044"]}},
    "TAEDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0047"]}},
    "TAENIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t0049"]}},
    "TALARIA": {"-": {"morpho": ["n-p---nn3i"], "uri": ["t0060"]}},
    "TALARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["t0060"]}},
    "TALEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t0063"]}},
    "TALENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0065"]}},
    "TALIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t0067"]}},
    "TALIS": {"-": {"morpho": ["aps---mn1-"], "uri": []}},
    "TALPA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t0073"]}},
    "TALVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["t0076"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["t0076"]},
    },
    "TAM": {"-": {"morpho": ["rp--------"], "uri": ["t1499"]}},
    "TAMDIV": {"-": {"morpho": ["rp--------"], "uri": ["37388"]}},
    "TAMEN": {"-": {"morpho": ["rp--------"], "uri": ["t0084"]}},
    "TAMENETSI": {"-": {"morpho": ["----------"], "uri": []}},
    "TAMETSI": {
        "1": {"morpho": ["----------"], "uri": []},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "TAMINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t0090"]}},
    "TAMQVAM": {
        "1": {"morpho": ["rp--------"], "uri": ["46698"]},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "TAMTVMDEM": {"1": {"morpho": ["aps---mn1-"], "uri": []}},
    "TANAQVIL": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "TANDEM": {"-": {"morpho": ["rp--------"], "uri": ["t0094"]}},
    "TANGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t0096"]}},
    "TANTILLVM": {
        "1": {"morpho": ["aps---mn1-"], "uri": []},
        "2": {"morpho": ["rp--------"], "uri": []},
    },
    "TANTILLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0101"]}},
    "TANTISPER": {"-": {"morpho": ["rp--------"], "uri": ["t0102"]}},
    "TANTO": {"-": {"morpho": ["rp--------"], "uri": ["100825"]}},
    "TANTOPERE": {"-": {"morpho": ["rp--------"], "uri": ["t0103"]}},
    "TANTVLVM": {
        "1": {"morpho": ["aps---mn1-"], "uri": []},
        "2": {"morpho": ["rp--------"], "uri": ["55227"]},
    },
    "TANTVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0104"]}},
    "TANTVM": {
        "1": {"morpho": ["aps---mn1-"], "uri": []},
        "2": {"morpho": ["rp--------"], "uri": ["27624"]},
    },
    "TANTVMDEM": {
        "1": {"morpho": ["aps---mn1-"], "uri": []},
        "2": {"morpho": ["rp--------"], "uri": ["100826"]},
    },
    "TANTVMMODO": {"-": {"morpho": ["rp--------"], "uri": ["27625"]}},
    "TANTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0106"]}},
    "TANTVSDEM": {"-": {"morpho": ["aps---mn1-"], "uri": []}},
    "TAPETVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0109"]}},
    "TARDE": {"-": {"morpho": ["rp--------"], "uri": ["52249"]}},
    "TARDESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t0117"]}},
    "TARDILOQVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["129556"]}},
    "TARDIPES": {"-": {"morpho": ["aps---an3-"], "uri": ["t0124"]}},
    "TARDITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t0125"]}},
    "TARDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t0129"]}},
    "TARDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0131"]}},
    "TARENTINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TARIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TARPEIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "TARQVINIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TARQVINIANVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "TARQVINIENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "TARQVINIENSIS": {
        "N": {"morpho": ["aps---cn3i"], "uri": []},
        "A": {"morpho": ["aps---cn3i"], "uri": []},
    },
    "TARQVINII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TARQVINIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "TARTARA": {"N": {"morpho": ["n-s----n2-"], "uri": []}},
    "TARTAREVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "TARTARVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TATIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "TAVREA": {
        "-": {"morpho": ["n-s----n1-"], "uri": []},
        "N": {"morpho": ["n-s---fn1-"], "uri": []},
    },
    "TAVREVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0147"]}},
    "TAVRIFORMIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["t0151"]}},
    "TAVRINVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["t0153"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["t0153"]},
    },
    "TAVRIS": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "TAVROBOLIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0157"]}},
    "TAVROMENITANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TAVROMENITANVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "TAVRVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["t0162"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["t0162"]},
    },
    "TAXATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t0165"]}},
    "TAXILLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["t0172"]}},
    "TAXIS": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "TAXO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t0174"]}},
    "TAXVS": {"-": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TECTE": {"-": {"morpho": ["rp--------"], "uri": ["44408"]}},
    "TECTINATVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "TECTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["t0186"]}},
    "TECTORIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0189"]}},
    "TECTORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0189"]}},
    "TECTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["27633"]}},
    "TECTVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["27634"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["27634"]},
    },
    "TEGES": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t0192"]}},
    "TEGIMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["t0198"]}},
    "TEGIMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0199"]}},
    "TEGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t0200"]}},
    "TEGVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t1500"]}},
    "TELA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t0205"]}},
    "TELEOS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "TELLENA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "TELLVS": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["t0212"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "TELVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0219"]}},
    "TEMENITES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "TEMERARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0221"]}},
    "TEMERE": {"-": {"morpho": ["rp--------"], "uri": ["t0225"]}},
    "TEMERITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t0224"]}},
    "TEMERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t0228"]}},
    "TEMESA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "TEMETVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0229"]}},
    "TEMNO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t0231"]}},
    "TEMO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["t0232"]}},
    "TEMPE": {
        "-": {"morpho": ["n-s----n-g"], "uri": []},
        "N": {"morpho": ["n-s---fn1-"], "uri": []},
    },
    "TEMPERAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0235"]}},
    "TEMPERANS": {"-": {"morpho": ["aps---an3i"], "uri": ["100836"]}},
    "TEMPERANTER": {"-": {"morpho": ["rp--------"], "uri": ["54327"]}},
    "TEMPERANTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t0218"]}},
    "TEMPERATE": {"-": {"morpho": ["rp--------"], "uri": ["37232"]}},
    "TEMPERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t0236"]}},
    "TEMPERATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["t0238"]}},
    "TEMPERATVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t9979"]}},
    "TEMPERATVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["t9995"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["27647"]},
    },
    "TEMPERIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["t0240"]}},
    "TEMPERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t0241"]}},
    "TEMPESTAS": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["t0242"]},
        "N": {"morpho": ["n-s---fn3-"], "uri": ["t0242"]},
    },
    "TEMPESTIVE": {"-": {"morpho": ["rp--------"], "uri": ["33934"]}},
    "TEMPESTIVO": {"-": {"morpho": ["rp--------"], "uri": ["100838"]}},
    "TEMPESTIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0245"]}},
    "TEMPLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0251"]}},
    "TEMPORARIVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "TEMPORARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0255"]}},
    "TEMPSANVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "TEMPTATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["t0261"]}},
    "TEMPVS": {
        "1": {"morpho": ["n-s---nn3-"], "uri": ["t0264"]},
        "2": {"morpho": ["n-s---nn3-"], "uri": ["t0264"]},
    },
    "TEMVLENTVS": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": ["t0267"]},
    },
    "TENACITER": {"-": {"morpho": ["rp--------"], "uri": ["56488"]}},
    "TENAX": {"-": {"morpho": ["aps---an3-"], "uri": ["t0271"]}},
    "TENDICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t0272"]}},
    "TENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t0273"]}},
    "TENEBRAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["38647"]}},
    "TENEBRICOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0281"]}},
    "TENEBROSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0286"]}},
    "TENEDII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TENEDVS": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "TENELLVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0287"]}},
    "TENEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["t0289"]}},
    "TENER": {"-": {"morpho": ["aps---mn1r"], "uri": ["t0290"]}},
    "TENERASCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["100839"]}},
    "TENERE": {"-": {"morpho": ["rp--------"], "uri": ["44982"]}},
    "TENERITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t0292"]}},
    "TENERVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "TENES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "TENOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["t0299"]}},
    "TENSA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t0300"]}},
    "TENSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["129695"]}},
    "TENTABVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["37727"]}},
    "TENTIGO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t0305"]}},
    "TENTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["129698"]}},
    "TENTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t0263"]}},
    "TENTORIOLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0308"]}},
    "TENTORIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0309"]}},
    "TENTVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "TENVE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["100844"]}},
    "TENVIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["t0317"]}},
    "TENVITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t9308"]}},
    "TENVITER": {"-": {"morpho": ["rp--------"], "uri": ["35711"]}},
    "TENVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t9309"]}},
    "TENVS": {
        "1": {"morpho": ["n-s---nn3-"], "uri": ["t9310"]},
        "2": {"morpho": ["p---------"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "TEPEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["t9312"]}},
    "TEPEFACTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t8313"]}},
    "TEPEFIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t9313"]}},
    "TEPEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["t9314"]}},
    "TEPESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t9314"]}},
    "TEPIDARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0318"]}},
    "TEPIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0320"]}},
    "TEPOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["t0321"]}},
    "TER": {"-": {"morpho": ["rp--------"], "uri": ["t0324"]}},
    "TERCENTVM": {"-": {"morpho": ["----------"], "uri": []}},
    "TERDECIES": {"-": {"morpho": ["rp--------"], "uri": ["129722"]}},
    "TEREBINTHVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["35275"]}},
    "TEREBRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t0333"]}},
    "TEREBRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t0335"]}},
    "TEREBRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t0336"]}},
    "TEREDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t0338"]}},
    "TERENTIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "TERES": {"-": {"morpho": ["aps---an3-"], "uri": ["t0340"]}},
    "TERGEMINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["54284"]}},
    "TERGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["t0343"]}},
    "TERGIVERSATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t0346"]}},
    "TERGIVERSOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["t0348"]}},
    "TERGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t0343"]}},
    "TERGVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0350"]}},
    "TERGVS": {"-": {"morpho": ["n-s---nn3-"], "uri": ["t0351"]}},
    "TERMES": {"-": {"morpho": ["n-s---mn3-"], "uri": ["t0355"]}},
    "TERMINALIA": {
        "-": {"morpho": ["n-s----n3-"], "uri": []},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "TERMINALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["t0357"]}},
    "TERMINATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t0358"]}},
    "TERMINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t0360"]}},
    "TERMINVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["t0361"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["t0361"]},
    },
    "TERNI": {"-": {"morpho": ["rp--------"], "uri": []}},
    "TERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t0371"]}},
    "TERRA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["t0372"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "TERRENA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["100853"]}},
    "TERRENI": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "TERRENVM": {"-": {"morpho": ["n-s---nn2-"], "uri": []}},
    "TERRENVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0377"]}},
    "TERREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["t0378"]}},
    "TERRESTRIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["t0379"]}},
    "TERREVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0380"]}},
    "TERRIBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["t0382"]}},
    "TERRIFICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t0388"]}},
    "TERRIFICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0389"]}},
    "TERRIGENA": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "TERRILOQVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0391"]}},
    "TERRITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t0395"]}},
    "TERRITORIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0398"]}},
    "TERROR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["t0399"]}},
    "TERROSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0400"]}},
    "TERSVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["t9999"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["33755"]},
    },
    "TERTIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "TERTIANVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["32956"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["t0405"]},
    },
    "TERTIARIVS": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": ["t0406"]},
    },
    "TERTIO": {"-": {"morpho": ["rp--------"], "uri": ["57806"]}},
    "TERTIVM": {"-": {"morpho": ["rp--------"], "uri": ["32303"]}},
    "TERTIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0413"]}},
    "TERTIVSDECIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["129796"]}},
    "TESCA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["129800"]}},
    "TESSELA": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "TESSERA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t0423"]}},
    "TESSERVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t0426"]}},
    "TESTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t0427"]}},
    "TESTACEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0429"]}},
    "TESTAMENTARIVS": {
        "2": {"morpho": ["aps---mn1-"], "uri": ["t0431"]},
        "1": {"morpho": ["n-s---mn2-"], "uri": ["42275"]},
    },
    "TESTAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0432"]}},
    "TESTATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["33127"]}},
    "TESTICVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["t0442"]}},
    "TESTIFICATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t0444"]}},
    "TESTIFICO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "TESTIFICOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["t0445"]}},
    "TESTIMONIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0447"]}},
    "TESTIS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["t1502"]},
        "2": {"morpho": ["n-s---mn3i"], "uri": ["t1502"]},
    },
    "TESTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["100865"]}},
    "TESTOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["t0450"]}},
    "TESTV": {"-": {"morpho": ["v1spid--1-"], "uri": []}},
    "TESTVDINEATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["52083"]}},
    "TESTVDINEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0455"]}},
    "TESTVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t0456"]}},
    "TESTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0451"]}},
    "TETER": {"-": {"morpho": ["aps---mn1r"], "uri": ["43877"]}},
    "TETRADOROS": {
        "1": {"morpho": ["aps---mn1g"], "uri": ["t0469"]},
        "2": {"morpho": ["rp--------"], "uri": []},
    },
    "TETRANS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["t0481"]}},
    "TETRARCHES": {
        "-": {"morpho": ["n-s---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn1g"], "uri": ["t0491"]},
    },
    "TETRASTYLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["129880"]}},
    "TETRE": {"-": {"morpho": ["rp--------"], "uri": ["129883"]}},
    "TETRICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0502"]}},
    "TETTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TEXO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t0513"]}},
    "TEXTILE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["100868"]}},
    "TEXTILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["t0514"]}},
    "TEXTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["t0515"]}},
    "TEXTORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0516"]}},
    "TEXTRINVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0518"]}},
    "TEXTRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t0519"]}},
    "TEXTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t9914"]}},
    "TEXTVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t9981"]}},
    "TEXTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["t9914"]}},
    "THALAMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["t0521"]}},
    "THALASSINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0526"]}},
    "THALASSIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "THAPSVS": {
        "-": {"morpho": ["n-s----n2-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "THARSVS": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "THEATRALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["t0539"]}},
    "THEATRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0542"]}},
    "THEBANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "THEBANVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "THECA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t0543"]}},
    "THEMATISMOS": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "THEMISTAGORAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "THEODORVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "THEOLOGVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["t0560"]}},
    "THEOMNASTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "THEOPANES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "THEORACTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "THERICLEVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "THERMAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["t0578"]}},
    "THERMITANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "THERMITANVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "THESAVRVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["t0593"]}},
    "THESPIADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "THESPIAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "THESPIENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "THESSALVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "THETA": {"-": {"morpho": ["n-s---nn--"], "uri": ["t0597"]}},
    "THIASVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["t0606"]}},
    "THOLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["t9612"]}},
    "THOMIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["129997"]}},
    "THORAX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["t0618"]}},
    "THRACES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "THRACIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "THRACIVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "THRASO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "THRAX": {
        "N": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "THVRINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": ["130030"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["130030"]},
    },
    "THYAS": {
        "-": {"morpho": ["n-s----n-g"], "uri": []},
        "N": {"morpho": ["n-s---mn1g"], "uri": []},
    },
    "THYIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0637"]}},
    "THYMBRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t0639"]}},
    "THYMELICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0643"]}},
    "THYMVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0652"]}},
    "THYNNVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["t0654"]}},
    "THYROMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["t0658"]}},
    "THYRORON": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "THYRSIGER": {"-": {"morpho": ["aps---mn1r"], "uri": ["t0662"]}},
    "THYRSVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["t0664"]}},
    "TIARA": {"-": {"morpho": ["n-s---fn1-"], "uri": []}},
    "TIARAS": {"-": {"morpho": ["n-s---fn1g"], "uri": ["100879"]}},
    "TIBERINVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "TIBERIS": {"N": {"morpho": ["n-s---cn3-"], "uri": []}},
    "TIBERIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TIBIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t0667"]}},
    "TIBICEN": {"-": {"morpho": ["n-s---mn3-"], "uri": ["t0670"]}},
    "TIBICINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t0671"]}},
    "TICESCO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "TIGILLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0681"]}},
    "TIGNARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0682"]}},
    "TIGNVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0685"]}},
    "TIGRIS": {
        "-": {"morpho": ["n-s---mn3-"], "uri": []},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "TILIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t0689"]}},
    "TIMARCHIDES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "TIMEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["t0695"]}},
    "TIMENS": {"-": {"morpho": ["aps---an3i"], "uri": ["100880"]}},
    "TIMEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["t0696"]}},
    "TIMESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t0696"]}},
    "TIMIDE": {"-": {"morpho": ["rp--------"], "uri": ["27722"]}},
    "TIMIDITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t0699"]}},
    "TIMIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0701"]}},
    "TIMOR": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["27724"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "TINEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t0709"]}},
    "TINEO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t0710"]}},
    "TINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t0712"]}},
    "TINNIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["t0716"]}},
    "TINNITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["t9916"]}},
    "TINNVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0719"]}},
    "TINTINNABVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0723"]}},
    "TINTINNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t0726"]}},
    "TIRO": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["t0731"]},
        "N": {"morpho": ["n-s---mn3-"], "uri": ["t0731"]},
    },
    "TIROCINIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0732"]}},
    "TIRVNCVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["t0736"]}},
    "TISSENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "TITIENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "TITILLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t0742"]}},
    "TITILLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t0743"]}},
    "TITINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TITIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "TITVBATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t0750"]}},
    "TITVBO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t0751"]}},
    "TITVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["t1300"]}},
    "TITVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "-": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "TLEPOLEMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TOFVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["t0767"]}},
    "TOGA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t0768"]}},
    "TOGATA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["38591"]}},
    "TOGATI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["100884"]}},
    "TOGATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0772"]}},
    "TOLERABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["t0775"]}},
    "TOLERABILITER": {"-": {"morpho": ["rp--------"], "uri": ["36887"]}},
    "TOLERANDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["56691"]}},
    "TOLERANS": {"-": {"morpho": ["aps---an3i"], "uri": ["100885"]}},
    "TOLERANTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t9983"]}},
    "TOLERATVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["t9918"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["130158"]},
    },
    "TOLERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t0777"]}},
    "TOLLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t0780"]}},
    "TOLVTARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t9781"]}},
    "TOMACVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0788"]}},
    "TOMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0790"]}},
    "TONANS": {
        "-": {"morpho": ["aps---an3i"], "uri": ["100888"]},
        "A": {"morpho": ["aps---an3i"], "uri": ["100888"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "TONDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["t0795"]}},
    "TONITRALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["100889"]}},
    "TONITRVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["t1486"]}},
    "TONITRVVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["39798"]}},
    "TONO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t0796"]}},
    "TONSA": {
        "1": {"morpho": ["n-s---fn1-"], "uri": ["t9992"]},
        "2": {"morpho": ["n-s---fn1-"], "uri": ["t9992"]},
    },
    "TONSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["t0812"]}},
    "TONSTRICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t0814"]}},
    "TONSVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t9984"]}},
    "TONVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["t0818"]}},
    "TOPEODVS": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "TOPIA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["t0825"]}},
    "TORAL": {"-": {"morpho": ["n-s---nn3i"], "uri": ["t0831"]}},
    "TORCVLAR": {"-": {"morpho": ["n-s---nn3i"], "uri": ["t0833"]}},
    "TORCVLARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0834"]}},
    "TORCVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0836"]}},
    "TOREVMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["t0838"]}},
    "TORMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0841"]}},
    "TORMINA": {"-": {"morpho": ["n-s---nn3-"], "uri": []}},
    "TORNATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["28409"]}},
    "TORNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t0849"]}},
    "TORNVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["t0850"]}},
    "TOROSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0852"]}},
    "TORPEDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t0853"]}},
    "TORPEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["t0855"]}},
    "TORPESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t0855"]}},
    "TORPIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0856"]}},
    "TORPOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["t0857"]}},
    "TORQVATVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["t0859"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "TORQVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["t0860"]}},
    "TORQVIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["100897"]}},
    "TORRENS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["27754"]},
        "2": {"morpho": ["aps---an3i"], "uri": ["100898"]},
    },
    "TORREO": {"-": {"morpho": ["v1spia--2-"], "uri": ["t0863"]}},
    "TORRESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t1498"]}},
    "TORRIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0866"]}},
    "TORRO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "TORSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t0868"]}},
    "TORTE": {"-": {"morpho": ["rp--------"], "uri": ["100900"]}},
    "TORTILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["t0870"]}},
    "TORTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t0871"]}},
    "TORTIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0873"]}},
    "TORTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t9873"]}},
    "TORTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["t0874"]}},
    "TORTVOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0877"]}},
    "TORTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["t9913"]}},
    "TORVA": {"-": {"morpho": ["rp--------"], "uri": ["100906"]}},
    "TORVITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t0879"]}},
    "TORVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["t0880"]}},
    "TORVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["t0881"]}},
    "TORVVM": {"-": {"morpho": ["rp--------"], "uri": []}},
    "TORVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0882"]}},
    "TOT": {"-": {"morpho": ["aps---mn1-"], "uri": []}},
    "TOTIDEM": {
        "1": {"morpho": ["aps---mn1-"], "uri": []},
        "2": {"morpho": ["rp--------"], "uri": ["100907"]},
    },
    "TOTIENS": {"-": {"morpho": ["rp--------"], "uri": ["t0888"]}},
    "TOTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["100910"]}},
    "TOTVS": {"-": {"morpho": ["----------"], "uri": []}},
    "TOXICVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0893"]}},
    "TRABALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["t0895"]}},
    "TRABEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t0897"]}},
    "TRABEATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0899"]}},
    "TRABECVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t0900"]}},
    "TRABS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["t0903"]}},
    "TRACTA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["130269"]}},
    "TRACTABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["t0910"]}},
    "TRACTABILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t0911"]}},
    "TRACTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t0913"]}},
    "TRACTATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["t0914"]}},
    "TRACTATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["t9920"]}},
    "TRACTIM": {"-": {"morpho": ["rp--------"], "uri": ["t0917"]}},
    "TRACTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t0918"]}},
    "TRACTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t0919"]}},
    "TRACTORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0922"]}},
    "TRACTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["t9993"]}},
    "TRADITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t0924"]}},
    "TRADO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t0926"]}},
    "TRADVCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t0927"]}},
    "TRADVCTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t0929"]}},
    "TRAGICE": {"-": {"morpho": ["rp--------"], "uri": ["42224"]}},
    "TRAGICVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["100916"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["t0938"]},
    },
    "TRAGOEDIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["t0940"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "TRAGOEDVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["t0942"]}},
    "TRAGVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t0948"]}},
    "TRAHA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["40154"]}},
    "TRAHO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t0956"]}},
    "TRAICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["t0957"]}},
    "TRAIECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t0959"]}},
    "TRAIECTVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t9986"]}},
    "TRAIECTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["t9923"]}},
    "TRAMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t0962"]}},
    "TRAMES": {"-": {"morpho": ["n-s---mn3-"], "uri": ["t0964"]}},
    "TRANGRESSVS": {"-": {"morpho": ["n-s----n4-"], "uri": []}},
    "TRANO": {"-": {"morpho": ["v1spia--1-"], "uri": ["45622"]}},
    "TRANQVILLE": {"-": {"morpho": ["rp--------"], "uri": ["39444"]}},
    "TRANQVILLITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t0966"]}},
    "TRANQVILLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t0967"]}},
    "TRANQVILLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0968"]}},
    "TRANQVILLVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["t0968"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "TRANS": {
        "1": {"morpho": ["rp--------"], "uri": []},
        "2": {"morpho": ["p---------"], "uri": ["t0969"]},
    },
    "TRANSACTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["t0972"]}},
    "TRANSALPINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": ["t0975"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["t0975"]},
    },
    "TRANSCENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t0980"]}},
    "TRANSCOLO": {"1": {"morpho": ["v1spia--1-"], "uri": ["t0983"]}},
    "TRANSCRIBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t0985"]}},
    "TRANSCVRRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t0988"]}},
    "TRANSCVRSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["t9925"]}},
    "TRANSDVCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["30156"]}},
    "TRANSEO": {"1": {"morpho": ["v1spia--4-"], "uri": ["t0993"]}},
    "TRANSERO": {"2": {"morpho": ["v1spia--3-"], "uri": ["t0994"]}},
    "TRANSFERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t0995"]}},
    "TRANSFIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t0997"]}},
    "TRANSFIGVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t1001"]}},
    "TRANSFODIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["t1009"]}},
    "TRANSFORMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t1013"]}},
    "TRANSFORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t1014"]}},
    "TRANSFVGA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["t1018"]}},
    "TRANSFVGIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["t1019"]}},
    "TRANSFVGIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t1020"]}},
    "TRANSFVNDO": {"2": {"morpho": ["v1spia--3-"], "uri": ["t1023"]}},
    "TRANSGERO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t1026"]}},
    "TRANSGREDIOR": {"-": {"morpho": ["v1spid--3i"], "uri": ["t1028"]}},
    "TRANSGRESSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t1029"]}},
    "TRANSGRESSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["t9926"]}},
    "TRANSIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t1032"]}},
    "TRANSILIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["t1033"]}},
    "TRANSITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t1035"]}},
    "TRANSITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["t9927"]}},
    "TRANSLATICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1043"]}},
    "TRANSLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t1044"]}},
    "TRANSLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["t1046"]}},
    "TRANSLATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["t9928"]}},
    "TRANSLEGO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["130406"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["t1047"]},
    },
    "TRANSLVCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["t1050"]}},
    "TRANSMARINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1052"]}},
    "TRANSMEO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t1055"]}},
    "TRANSMIGRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t1057"]}},
    "TRANSMISSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["t9929"]}},
    "TRANSMITTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t1061"]}},
    "TRANSMOVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["t1063"]}},
    "TRANSMVTO": {"2": {"morpho": ["v1spia--1-"], "uri": ["t1067"]}},
    "TRANSNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t1070"]}},
    "TRANSPARENS": {
        "-": {"morpho": ["aps---an3-"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "TRANSPECTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["t1076"]}},
    "TRANSPICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["t1080"]}},
    "TRANSPONO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t1082"]}},
    "TRANSPORTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t1083"]}},
    "TRANSPORTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t1084"]}},
    "TRANSSCRIBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["36511"]}},
    "TRANSTRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t1094"]}},
    "TRANSVECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t1097"]}},
    "TRANSVECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["130440"]}},
    "TRANSVEHO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t1099"]}},
    "TRANSVERBERO": {"2": {"morpho": ["v1spia--1-"], "uri": ["t1104"]}},
    "TRANSVERSE": {"-": {"morpho": ["rp--------"], "uri": ["36972"]}},
    "TRANSVERSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t1108"]}},
    "TRANSVERSVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["100934"]},
        "2": {"morpho": ["rp--------"], "uri": []},
    },
    "TRANSVERSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["47535"]}},
    "TRANSVERTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t1109"]}},
    "TRANSVIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t1110"]}},
    "TRANSVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t1116"]}},
    "TRANSVOLITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t1117"]}},
    "TRANSVOLO": {"2": {"morpho": ["v1spia--1-"], "uri": ["t1118"]}},
    "TRAPES": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "TRAPETVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["100937"]}},
    "TRAPETVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["t1122"]}},
    "TRAVLIZO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "TREBIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "TREBONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TRECENI": {"-": {"morpho": ["rp--------"], "uri": []}},
    "TRECENTENI": {"-": {"morpho": ["rp--------"], "uri": []}},
    "TRECENTESIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1134"]}},
    "TRECENTI": {
        "-": {"morpho": ["----------"], "uri": []},
        "N": {"morpho": ["n-p---mn2-"], "uri": []},
    },
    "TRECENTIES": {"-": {"morpho": ["rp--------"], "uri": ["30013"]}},
    "TRECHEDIPNVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t1137"]}},
    "TRECTO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "TREDECIM": {"-": {"morpho": ["----------"], "uri": []}},
    "TREMEBVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1142"]}},
    "TREMEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["t1143"]}},
    "TREMELLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TREMENDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["29510"]}},
    "TREMESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t0804"]}},
    "TREMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t1148"]}},
    "TREMOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["t1149"]}},
    "TREMVLVS": {
        "1": {"morpho": ["n-s---fn2-"], "uri": ["130490"]},
        "2": {"morpho": ["n-s---fn2-"], "uri": ["130490"]},
        "3": {"morpho": ["aps---mn1-"], "uri": ["t1150"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "TREPIDANTER": {"-": {"morpho": ["rp--------"], "uri": ["34308"]}},
    "TREPIDATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t1151"]}},
    "TREPIDE": {"-": {"morpho": ["rp--------"], "uri": ["37158"]}},
    "TREPIDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t1153"]}},
    "TREPIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1155"]}},
    "TRES": {
        "-": {"morpho": ["----------"], "uri": []},
        "N": {"morpho": ["n-s---mn1g"], "uri": []},
    },
    "TRESSIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["t1159"]}},
    "TRESVIRI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["100943"]}},
    "TRIARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["34278"]},
        "2": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["34278"]},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "TRIBON": {"-": {"morpho": ["n-s---mn3-"], "uri": ["t1169"]}},
    "TRIBVLIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["t1504"]}},
    "TRIBVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t1176"]}},
    "TRIBVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0337"]}},
    "TRIBVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["30124"]}},
    "TRIBVNAL": {"-": {"morpho": ["n-s---nn3i"], "uri": ["t1179"]}},
    "TRIBVNATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["t1181"]}},
    "TRIBVNICIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["35123"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["t1182"]},
    },
    "TRIBVNVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["t1183"]}},
    "TRIBVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t1184"]}},
    "TRIBVS": {"-": {"morpho": ["n-s---mn4-"], "uri": []}},
    "TRIBVTARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1186"]}},
    "TRIBVTIM": {"-": {"morpho": ["rp--------"], "uri": ["t1187"]}},
    "TRIBVTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t1173"]}},
    "TRIBVTVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["t1173"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["49498"]},
    },
    "TRICENARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1193"]}},
    "TRICENI": {"-": {"morpho": ["rp--------"], "uri": []}},
    "TRICEPS": {"-": {"morpho": ["aps---an3-"], "uri": ["t1198"]}},
    "TRICESIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["100950"]}},
    "TRICHALCVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t1202"]}},
    "TRICHILA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t1205"]}},
    "TRICHO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "TRICIENS": {"-": {"morpho": ["rp--------"], "uri": ["t1212"]}},
    "TRICIPITINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TRICLINIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t1217"]}},
    "TRICOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["t1223"]}},
    "TRICORPOR": {"-": {"morpho": ["aps---an3-"], "uri": ["t1227"]}},
    "TRIDENS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["100954"]},
        "2": {"morpho": ["aps---an3-"], "uri": ["t1231"]},
    },
    "TRIDVVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t1236"]}},
    "TRIENNIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t1239"]}},
    "TRIERARCHVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["t1244"]}},
    "TRIERIS": {"2": {"morpho": ["aps---cn3i"], "uri": ["t1245"]}},
    "TRIETERICA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["100957"]}},
    "TRIETERICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1246"]}},
    "TRIFARIAM": {"-": {"morpho": ["rp--------"], "uri": ["t1248"]}},
    "TRIFAVX": {"-": {"morpho": ["aps---an3-"], "uri": []}},
    "TRIFIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1253"]}},
    "TRIFISSILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["t1256"]}},
    "TRIFORMIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["t1258"]}},
    "TRIGINTA": {"-": {"morpho": ["----------"], "uri": []}},
    "TRIGLYPHVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["t1277"]}},
    "TRIGON": {"-": {"morpho": ["n-s---mn3-"], "uri": ["t1278"]}},
    "TRIGONIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t1282"]}},
    "TRIGONOS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "TRIGONVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["47707"]}},
    "TRILIBRIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["t1288"]}},
    "TRILINGVIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["t1290"]}},
    "TRILIX": {"-": {"morpho": ["aps---an3i"], "uri": ["100961"]}},
    "TRIMESTRIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["100963"]}},
    "TRIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1304"]}},
    "TRINODIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["t1315"]}},
    "TRINVS": {"-": {"morpho": ["rp--------"], "uri": []}},
    "TRIOBOLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TRIOCALINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "TRIONES": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "TRIPARTIO": {"-": {"morpho": ["v1spia--4-", "n-s----n3-"], "uri": []}},
    "TRIPARTITO": {
        "-": {
            "morpho": ["rp--------", "rp--------"],
            "uri": ["130633", "130633"],
        }
    },
    "TRIPECTORVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1333"]}},
    "TRIPEDANEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1335"]}},
    "TRIPERTITO": {"-": {"morpho": ["rp--------"], "uri": ["32543"]}},
    "TRIPES": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["130641"]},
        "2": {"morpho": ["aps---an3-"], "uri": ["t1338"]},
    },
    "TRIPLEX": {
        "-": {"morpho": ["aps---an3-", "n-s---nn3-"], "uri": ["130647"]}
    },
    "TRIPLINTHIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1350"]}},
    "TRIPODO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t1354"]}},
    "TRIPTOLEMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TRIPVDIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t1361"]}},
    "TRIPVDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t1362"]}},
    "TRIPVS": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["t1364"]},
        "2": {"morpho": ["aps---an3-"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "TRIQVETRVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["t1365"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["t1365"]},
    },
    "TRIREMIS": {
        "1": {"morpho": ["n-s---fn3i"], "uri": ["100967"]},
        "2": {"morpho": ["aps---cn3-"], "uri": []},
    },
    "TRISCVRRIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t1372"]}},
    "TRISPASTOS": {
        "-": {"morpho": ["aps---mn1g"], "uri": []},
        "A": {"morpho": ["aps---mn1g"], "uri": []},
    },
    "TRISTE": {
        "1": {"morpho": ["n-s---nn3i"], "uri": ["100969"]},
        "2": {"morpho": ["rp--------"], "uri": ["52314"]},
    },
    "TRISTIMONIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t1385"]}},
    "TRISTIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["t1386"]}},
    "TRISTITIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t1388"]}},
    "TRISTOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["t1390"]}},
    "TRISVLCVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1391"]}},
    "TRITE": {"-": {"morpho": ["n-s---fn1g"], "uri": ["t8395"]}},
    "TRITICEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1400"]}},
    "TRITICVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t1403"]}},
    "TRITVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t0408"]}},
    "TRITVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["t9930"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["27828"]},
    },
    "TRIVIALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["t1411"]}},
    "TRIVIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t1413"]}},
    "TRIVIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1414"]}},
    "TRIVMPHALIA": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "TRIVMPHALIS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["100973"]},
        "2": {"morpho": ["aps---cn3i"], "uri": ["t1415"]},
    },
    "TRIVMPHO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t1420"]}},
    "TRIVMPHVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["t1421"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["t1421"]},
    },
    "TRIVMVIR": {"-": {"morpho": ["n-s---mn2r"], "uri": ["t1422"]}},
    "TRIVMVIRALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["t1423"]}},
    "TRIVMVIRATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["t1424"]}},
    "TROCHILVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["t1430"]}},
    "TROCHLEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t1434"]}},
    "TROCHVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["t1436"]}},
    "TROIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "TROIANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TROIANVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "TROPAEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t1440"]}},
    "TROSSVLVS": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "TRVCIDATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t1450"]}},
    "TRVCIDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t1452"]}},
    "TRVCVLENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t1456"]}},
    "TRVCVLENTVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["t1457"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "TRVDIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["t1458"]}},
    "TRVDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t1459"]}},
    "TRVLLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t1460"]}},
    "TRVLLEVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["t1461"]}},
    "TRVLLISSATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t1462"]}},
    "TRVLLISSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t1463"]}},
    "TRVNCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t1464"]}},
    "TRVNCVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["t1466"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["t1466"]},
    },
    "TRVSATILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["t1469"]}},
    "TRVSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t1470"]}},
    "TRVTINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t1471"]}},
    "TRVTINOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132801"]}},
    "TRVX": {"-": {"morpho": ["aps---an3-"], "uri": ["t1474"]}},
    "TV": {"-": {"morpho": ["o-----an--"], "uri": []}},
    "TVBA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t1685"]}},
    "TVBER": {"-": {"morpho": ["n-s---nn3-"], "uri": ["t1687"]}},
    "TVBICEN": {"-": {"morpho": ["n-s---mn3-"], "uri": ["t1692"]}},
    "TVBVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t1699"]}},
    "TVBVLVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["t1702"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["t1702"]},
    },
    "TVBVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["t1706"]}},
    "TVCCETVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t1707"]}},
    "TVDITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t1711"]}},
    "TVEOR": {"-": {"morpho": ["v1spid--2-"], "uri": ["t1712"]}},
    "TVGVRIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t1715"]}},
    "TVI": {"-": {"morpho": ["aps---mn1-", "n-p---mn2-"], "uri": ["130806"]}},
    "TVLLIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TVLLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["t1719"]}},
    "TVLLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TVM": {"-": {"morpho": ["rp--------"], "uri": ["t1721"]}},
    "TVMEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["t1723"]}},
    "TVMEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["t1725"]}},
    "TVMESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t1725"]}},
    "TVMIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1730"]}},
    "TVMOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["t1731"]}},
    "TVMVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t1732"]}},
    "TVMVLOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1733"]}},
    "TVMVLTVARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1734"]}},
    "TVMVLTVOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["t1738"]}},
    "TVMVLTVOSE": {"-": {"morpho": ["rp--------"], "uri": ["46387"]}},
    "TVMVLTVOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1739"]}},
    "TVMVLTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["t1740"]}},
    "TVMVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["t1741"]}},
    "TVNC": {"-": {"morpho": ["rp--------"], "uri": ["t1742"]}},
    "TVNDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t1743"]}},
    "TVNICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t1745"]}},
    "TVNICATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1746"]}},
    "TVOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["132802"]}},
    "TVRBA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t1755"]}},
    "TVRBATE": {"-": {"morpho": ["rp--------"], "uri": ["54497"]}},
    "TVRBATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["t1758"]}},
    "TVRBATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["54581"]}},
    "TVRBEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["48085"]}},
    "TVRBIDE": {"-": {"morpho": ["rp--------"], "uri": ["56698"]}},
    "TVRBIDVM": {"-": {"morpho": ["rp--------"], "uri": ["100994"]}},
    "TVRBIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1764"]}},
    "TVRBO": {
        "1": {"morpho": ["n-s---fn3-"], "uri": []},
        "2": {"morpho": ["v1spia--1-"], "uri": ["t1770"]},
        "N": {"morpho": ["n-s----n1-"], "uri": []},
    },
    "TVRBVLENTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1776"]}},
    "TVRDA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t1778"]}},
    "TVRDVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["t1778"]}},
    "TVREVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1782"]}},
    "TVRGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["t1783"]}},
    "TVRGESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t0933"]}},
    "TVRGIDVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1784"]}},
    "TVRGIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1785"]}},
    "TVRIBVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t1787"]}},
    "TVRICREMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1788"]}},
    "TVRIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["t1789"]}},
    "TVRMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t1797"]}},
    "TVRMATIM": {"-": {"morpho": ["rp--------"], "uri": ["t1800"]}},
    "TVRNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TVRPE": {
        "1": {"morpho": ["n-s---nn3i"], "uri": ["100995"]},
        "2": {"morpho": ["rp--------"], "uri": ["30799"]},
    },
    "TVRPICVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1802"]}},
    "TVRPIO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "TVRPIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["t1808"]}},
    "TVRPITER": {"-": {"morpho": ["rp--------"], "uri": ["27871"]}},
    "TVRPITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t1809"]}},
    "TVRPO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t1810"]}},
    "TVRRICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t1811"]}},
    "TVRRIGER": {"-": {"morpho": ["aps---mn1r"], "uri": ["t1813"]}},
    "TVRRIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["t1814"]}},
    "TVRRITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1815"]}},
    "TVRTVR": {"-": {"morpho": ["n-s---fn3-"], "uri": []}},
    "TVRTVRILLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["100997"]}},
    "TVRVNDA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t0613"]}},
    "TVS": {"-": {"morpho": ["n-s---nn3-"], "uri": ["t1821", "51765"]}},
    "TVSCI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TVSCVLANVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "TVSCVLANVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "TVSCVLVM": {
        "N": {"morpho": ["n-s---nn2-"], "uri": ["t1822"]},
        "-": {"morpho": ["n-s---nn2-"], "uri": ["t1822"]},
    },
    "TVSCVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "TVSSIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["t1829"]}},
    "TVSSIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["t1830"]}},
    "TVTAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["t1832"]}},
    "TVTAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t1833"]}},
    "TVTE": {"-": {"morpho": ["rp--------"], "uri": ["36734"]}},
    "TVTELA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["t1837"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "TVTO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["27879"]},
        "2": {"morpho": ["rp--------"], "uri": ["100998"]},
        "-": {"morpho": ["rp--------"], "uri": ["100998"]},
    },
    "TVTOR": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["t1842"]},
        "2": {"morpho": ["v1spid--1-"], "uri": ["t1843"]},
        "N": {"morpho": ["n-s----n1-"], "uri": []},
    },
    "TVTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["100999"]}},
    "TVTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1848"]}},
    "TVVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "TVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["29740"]}},
    "TYCHA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "TYMPANVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t1862"]}},
    "TYNDARIS": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "TYNDARITANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TYNDARITANVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "TYRACINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TYRANNICE": {"-": {"morpho": ["rp--------"], "uri": ["28356"]}},
    "TYRANNICIDA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["t1868"]}},
    "TYRANNICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1870"]}},
    "TYRANNIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t1871"]}},
    "TYRANNVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["t0901"]}},
    "TYRIVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "TYROS": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "VACATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0002"]}},
    "VACCA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["u0003"]},
        "N": {"morpho": ["n-s---mn1-"], "uri": []},
    },
    "VACCINIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0004"]}},
    "VACEFIO": {"-": {"morpho": ["v1spia--3-"], "uri": ["u9015"]}},
    "VACILLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0010"]}},
    "VACO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0013"]}},
    "VACVEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["u0014"]}},
    "VACVITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0016"]}},
    "VACVO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["u0017"]},
        "2": {"morpho": ["rp--------"], "uri": []},
    },
    "VACVVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0072"]}},
    "VACVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0052"]}},
    "VADIMONIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0019"]}},
    "VADO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["u0024"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["u0021"]},
    },
    "VADOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["u0029"]}},
    "VADOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0022"]}},
    "VADVM": {
        "-": {"morpho": ["n-s---nn2-"], "uri": ["u0023"]},
        "N": {"morpho": ["n-s---nn2-"], "uri": ["u0023"]},
    },
    "VAE": {"-": {"morpho": ["----------"], "uri": []}},
    "VAFER": {"-": {"morpho": ["aps---an3-"], "uri": []}},
    "VAFRE": {"-": {"morpho": ["rp--------"], "uri": ["45829"]}},
    "VAFRITIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u9029"]}},
    "VAGABVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0030"]}},
    "VAGATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0031"]}},
    "VAGE": {"-": {"morpho": ["rp--------"], "uri": ["50264"]}},
    "VAGINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0044"]}},
    "VAGIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["u0035"]}},
    "VAGITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["u0020"]}},
    "VAGOR": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["u0039"]},
        "2": {"morpho": ["v1spid--1-"], "uri": ["u0038"]},
    },
    "VAGVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0043"]}},
    "VALDE": {"-": {"morpho": ["rp--------"], "uri": ["u0046"]}},
    "VALEDICO": {
        "2": {"morpho": ["v1spia--3-"], "uri": ["u0047"]},
        "-": {"morpho": ["v1spia--3-"], "uri": ["u0047"]},
    },
    "VALENS": {
        "-": {"morpho": ["aps---an3i"], "uri": ["27976"]},
        "N": {"morpho": ["n-s---mn3-"], "uri": []},
    },
    "VALENTINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "VALENTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VALEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["u0050"]}},
    "VALERIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "VALESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["u0725"]}},
    "VALETVDINARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0106"]}},
    "VALETVDINARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["u0066"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["u0000"]},
    },
    "VALETVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0053"]}},
    "VALIDE": {"-": {"morpho": ["rp--------"], "uri": ["57421"]}},
    "VALIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0057"]}},
    "VALLIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["101062"]}},
    "VALLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0064"]}},
    "VALLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0068"]}},
    "VALLVS": {
        "1": {"morpho": ["n-s---fn2-"], "uri": ["u0068"]},
        "2": {"morpho": ["n-s---fn2-"], "uri": ["u0068"]},
    },
    "VALOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["48931"]}},
    "VALVA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0069"]}},
    "VALVAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["101063"]}},
    "VALVATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0071"]}},
    "VANA": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "VANE": {"-": {"morpho": ["rp--------"], "uri": ["130972"]}},
    "VANESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["u0073"]}},
    "VANILOQVENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0077"]}},
    "VANILOQVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0079"]}},
    "VANITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0080"]}},
    "VANNVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["u0085"]}},
    "VANVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["58068"]}},
    "VANVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0087"]}},
    "VAPIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0088"]}},
    "VAPOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["27989"]}},
    "VAPORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0096"]}},
    "VAPPA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0099"]}},
    "VAPVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0102"]}},
    "VARA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0103"]}},
    "VARIANTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u9972"]}},
    "VARIATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["131002"]}},
    "VARICOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0113"]}},
    "VARICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0115"]}},
    "VARIE": {"-": {"morpho": ["rp--------"], "uri": ["55066"]}},
    "VARIETAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0118"]}},
    "VARIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0119"]}},
    "VARIVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "VARIVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["u0120", "u0120"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": ["u0120", "u0120"]},
    },
    "VARIX": {"-": {"morpho": ["n-s---mn3-"], "uri": []}},
    "VARRO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "VARVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["u0125"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["u0103"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["u0125"]},
    },
    "VAS": {
        "1": {"morpho": ["n-s---nn3-"], "uri": ["u0132"]},
        "2": {"morpho": ["n-s---nn3-"], "uri": ["u0132"]},
    },
    "VASARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0128"]}},
    "VASCVLARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["u0321"]}},
    "VASCVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0133"]}},
    "VASCVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0134"]}},
    "VASTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0140"]}},
    "VASTATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["u0141"]}},
    "VASTATRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0143"]}},
    "VASTE": {"-": {"morpho": ["rp--------"], "uri": ["32307"]}},
    "VASTITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0146"]}},
    "VASTITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0149"]}},
    "VASTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0150"]}},
    "VASTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0152"]}},
    "VATES": {"-": {"morpho": ["n-s---mn3-"], "uri": []}},
    "VATICINATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0156"]}},
    "VATICINOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["u0160"]}},
    "VATILLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0162"]}},
    "VBER": {
        "1": {"morpho": ["n-s---nn3-"], "uri": ["u9157"]},
        "2": {"morpho": ["aps---an3-"], "uri": ["u9157"]},
    },
    "VBERE": {"-": {"morpho": ["rp--------"], "uri": []}},
    "VBERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u9158"]}},
    "VBERTAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u9159"]}},
    "VBERTIM": {"-": {"morpho": ["rp--------"], "uri": ["u9160"]}},
    "VBI": {
        "1": {"morpho": ["rp--------"], "uri": ["u0165"]},
        "2": {"morpho": ["rp--------"], "uri": ["u0165"]},
        "3": {"morpho": ["----------"], "uri": []},
    },
    "VBICVMQVE": {
        "1": {"morpho": ["rp--------"], "uri": ["u0168"]},
        "2": {"morpho": ["rp--------"], "uri": ["u0168"]},
    },
    "VBILIBET": {"-": {"morpho": ["rp--------"], "uri": ["u0166"]}},
    "VBINAM": {
        "1": {"morpho": ["rp--------"], "uri": ["u0172"]},
        "2": {"morpho": ["rp--------"], "uri": ["u0172"]},
    },
    "VBIQVE": {"-": {"morpho": ["rp--------"], "uri": ["u0167"]}},
    "VBIVIS": {"-": {"morpho": ["rp--------"], "uri": ["u0170"]}},
    "VDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0178"]}},
    "VE": {"-": {"morpho": ["----------"], "uri": []}},
    "VECORDIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0179"]}},
    "VECORS": {"-": {"morpho": ["aps---an3-"], "uri": ["u0180"]}},
    "VECTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0184"]}},
    "VECTIARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["u0185"]}},
    "VECTICVLARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0186"]}},
    "VECTIGAL": {"-": {"morpho": ["n-s---nn3i"], "uri": ["u0187"]}},
    "VECTIGALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["u0332"]}},
    "VECTIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["u0191"]}},
    "VECTITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0192"]}},
    "VECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0193"]}},
    "VECTOR": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["u0194"]},
        "2": {"morpho": ["v1spid--1-", "n-s---mn3-"], "uri": ["u0194"]},
    },
    "VECTORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0195"]}},
    "VECTVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u9973"]}},
    "VEDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VEGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["u0198"]}},
    "VEGETO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0203"]}},
    "VEGETVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0204"]}},
    "VEGRANDIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["u0205"]}},
    "VEHEMENS": {"-": {"morpho": ["aps---an3i"], "uri": ["u0209"]}},
    "VEHEMENTER": {"-": {"morpho": ["rp--------"], "uri": ["28010"]}},
    "VEHEMENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0211"]}},
    "VEHES": {"-": {"morpho": ["n-s---fn3i"], "uri": ["101074"]}},
    "VEHICVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0216"]}},
    "VEHO": {"-": {"morpho": ["v1spia--3-"], "uri": ["u0217"]}},
    "VEHOR": {"-": {"morpho": ["v1spid--3-"], "uri": []}},
    "VEIENS": {
        "N": {"morpho": ["aps---an3-"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "VEIENTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "VEII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "VEL": {
        "1": {"morpho": ["rp--------"], "uri": ["50736"]},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "VELAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["u0222"]}},
    "VELAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0223"]}},
    "VELARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0225"]}},
    "VELES": {"-": {"morpho": ["n-s---mn3-"], "uri": ["u0229"]}},
    "VELIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "VELIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["u0230"]}},
    "VELIFICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0233"]}},
    "VELIFICOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132757"]}},
    "VELITARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["u0236"]}},
    "VELITERNVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "VELITRAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "VELIVOLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0240"]}},
    "VELLICATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0243"]}},
    "VELLICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0244"]}},
    "VELLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["u0246"]}},
    "VELLVS": {"-": {"morpho": ["n-s---nn3-"], "uri": ["u0247"]}},
    "VELO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0248"]}},
    "VELOCITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0249"]}},
    "VELOCITER": {"-": {"morpho": ["rp--------"], "uri": ["28021"]}},
    "VELOX": {"-": {"morpho": ["aps---an3i"], "uri": ["u0250"]}},
    "VELVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0251"]}},
    "VELVT": {
        "1": {"morpho": ["rp--------"], "uri": ["u0252"]},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "VELVTSI": {"-": {"morpho": ["----------"], "uri": []}},
    "VENA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0253"]}},
    "VENABVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0254"]}},
    "VENALES": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "VENALIA": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "VENALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["u0257"]}},
    "VENANS": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "VENATICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0260"]}},
    "VENATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0261"]}},
    "VENATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["u0262"]}},
    "VENATORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u9262"]}},
    "VENATRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0264"]}},
    "VENATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["u9980"]}},
    "VENDAX": {"-": {"morpho": ["aps---an3i"], "uri": ["u0265"]}},
    "VENDIBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["u0266"]}},
    "VENDITATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0268"]}},
    "VENDITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0270"]}},
    "VENDITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0271"]}},
    "VENDITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["u0272"]}},
    "VENDITVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["32160"]}},
    "VENDO": {"-": {"morpho": ["v1spia--3-"], "uri": ["u0274"]}},
    "VENEFICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0277"]}},
    "VENEFICIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0276"]}},
    "VENEFICVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["u0277"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["u0277"]},
    },
    "VENENATVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "VENENATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["28279"]}},
    "VENENO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0280"]}},
    "VENENVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0283"]}},
    "VENEO": {"-": {"morpho": ["v1spia--5-"], "uri": ["u0284"]}},
    "VENERABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["u0285"]}},
    "VENERABVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0287"]}},
    "VENERANDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["29632"]}},
    "VENERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0289"]}},
    "VENEREVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "-": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "VENERIA": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "VENEROR": {"-": {"morpho": ["v1spid--1-"], "uri": ["u0293"]}},
    "VENETI": {
        "1N": {"morpho": ["n-s----n2-"], "uri": []},
        "N": {"morpho": ["n-p---mn2-"], "uri": []},
    },
    "VENIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0296"]}},
    "VENIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["u0300"]}},
    "VENNONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VENOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["u0302"]}},
    "VENOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0303"]}},
    "VENTER": {"-": {"morpho": ["n-s---mn3-"], "uri": ["u0304"]}},
    "VENTILO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0309"]}},
    "VENTITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0312"]}},
    "VENTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0313"]}},
    "VENTOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0314"]}},
    "VENTRICVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["131160"]}},
    "VENTRICVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["u0503"]}},
    "VENTVRVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "VENTVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["u0328"]}},
    "VENVCVLVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "VENVLEIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VENVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["101085"]}},
    "VENVNDO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "VENVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["u0470"]},
        "2": {"morpho": ["n-s---mn4-"], "uri": ["u0470"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "VENVSTAS": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["u0333"]},
        "N": {"morpho": ["n-s---fn3-"], "uri": ["u0333"]},
    },
    "VENVSTE": {"-": {"morpho": ["rp--------"], "uri": ["34343"]}},
    "VENVSTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0336"]}},
    "VEPALLIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0337"]}},
    "VEPRES": {"-": {"morpho": ["n-s---mn3i"], "uri": ["101087"]}},
    "VER": {"-": {"morpho": ["n-s---nn3-"], "uri": ["u0342"]}},
    "VERATRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0343"]}},
    "VERAX": {
        "-": {"morpho": ["aps---an3i"], "uri": ["u0344"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "VERBENA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0348"]}},
    "VERBER": {"-": {"morpho": ["n-s---nn3-"], "uri": ["u0352"]}},
    "VERBERATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["u9982"]}},
    "VERBERO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["u0362"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["u0361"]},
    },
    "VERBOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u9369"]}},
    "VERBVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u9363"]}},
    "VERE": {"-": {"morpho": ["rp--------"], "uri": ["28055"]}},
    "VERECVNDE": {"-": {"morpho": ["rp--------"], "uri": ["45936"]}},
    "VERECVNDIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u9365"]}},
    "VERECVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u1500"]}},
    "VERENDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["33683"]}},
    "VEREOR": {"-": {"morpho": ["v1spid--2-"], "uri": ["u0369"]}},
    "VERGENS": {
        "-": {"morpho": ["aps---an3-"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "VERGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["u0372"]}},
    "VERGOBRETVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["u0373"]}},
    "VERIDICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0377"]}},
    "VERISIMILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["u9380"]}},
    "VERISIMILITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0380"]}},
    "VERITAS": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["u0381"]},
        "N": {"morpho": ["n-s---fn3-"], "uri": ["u0381"]},
    },
    "VERMICVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["u0387"]}},
    "VERMINA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["u0390"]}},
    "VERMINATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0391"]}},
    "VERMINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0392"]}},
    "VERMIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["u0394"]}},
    "VERNA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["u0395"]}},
    "VERNACVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0587"]}},
    "VERNACVLVS": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": ["u0514"]},
    },
    "VERNILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0403"]}},
    "VERNILITER": {"-": {"morpho": ["rp--------"], "uri": ["28065"]}},
    "VERNO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["u0406"]},
        "2": {"morpho": ["rp--------"], "uri": []},
    },
    "VERNVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": []}},
    "VERNVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0409"]}},
    "VERNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0409"]}},
    "VERO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["u0411"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["u0410"]},
        "3": {"morpho": ["rp--------"], "uri": ["28066"]},
        "4": {"morpho": ["----------"], "uri": []},
    },
    "VERPA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0413"]}},
    "VERPVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["u0413"]}},
    "VERRES": {
        "-": {"morpho": ["n-s---mn3i"], "uri": ["u0414"]},
        "N": {"morpho": ["n-s---mn1g"], "uri": []},
    },
    "VERRIA": {"N": {"morpho": ["n-s----n2-"], "uri": []}},
    "VERRINVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": ["u0416"]},
        "-": {"morpho": ["aps---mn1-"], "uri": ["u0416"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["u0416"]},
    },
    "VERRIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VERRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["u0417"]}},
    "VERRVCA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0418"]}},
    "VERRVCIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VERRVCOSVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["u0420"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "VERSABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["u0423"]}},
    "VERSABVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0424"]}},
    "VERSATILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["u0425"]}},
    "VERSATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0426"]}},
    "VERSATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["46788"]}},
    "VERSICOLOR": {"-": {"morpho": ["aps---an3-"], "uri": ["u0428"]}},
    "VERSICVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["u0429"]}},
    "VERSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0438"]}},
    "VERSOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132762"]}},
    "VERSVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0447"]}},
    "VERSVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["u0516"]},
        "2": {"morpho": ["aps---mn1-"], "uri": []},
        "3": {"morpho": ["rp--------"], "uri": ["u0440"]},
        "4": {"morpho": ["p---------"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "VERSVTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0443"]}},
    "VERTEBRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0444"]}},
    "VERTENS": {
        "-": {"morpho": ["aps---an3-"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "VERTEX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["u0446"]}},
    "VERTICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0453"]}},
    "VERTIGO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0456"]}},
    "VERTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["u0459"]}},
    "VERTVMNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["u0461"]}},
    "VERV": {"-": {"morpho": ["n-s---nn4-"], "uri": ["u0462"]}},
    "VERVACTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u9997"]}},
    "VERVEX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["u0467"]}},
    "VERVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["u8133", "u0462"]},
        "2": {"morpho": ["n-s---nn2-"], "uri": ["u8133", "u0462"]},
        "3": {"morpho": ["rp--------"], "uri": ["28081"]},
        "4": {"morpho": ["----------"], "uri": []},
        "-": {"morpho": ["n-s---nn2-"], "uri": ["u8133", "u0462"]},
    },
    "VERVMTAMEN": {"-": {"morpho": ["rp--------"], "uri": ["101108"]}},
    "VERVMVERO": {"-": {"morpho": ["rp--------"], "uri": []}},
    "VERVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["u0469"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "VERVTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0709"]}},
    "VERVTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0693"]}},
    "VESANIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0472"]}},
    "VESANIO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "VESANVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0474"]}},
    "VESCOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["u0476"]}},
    "VESCVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0478"]}},
    "VESICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0479"]}},
    "VESICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0483"]}},
    "VESPER": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["u0487"]},
        "N": {"morpho": ["n-s----n2-"], "uri": []},
    },
    "VESPERA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0487"]}},
    "VESPERASCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["u0489"]}},
    "VESPERTINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0492"]}},
    "VESTA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "VESTALIA": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "VESTALIS": {
        "1": {"morpho": ["n-s----n3-"], "uri": []},
        "2": {"morpho": ["aps---cn3i"], "uri": []},
        "A": {"morpho": ["aps---cn3i"], "uri": []},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "VESTER": {"-": {"morpho": ["aps---mn1-"], "uri": []}},
    "VESTIARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0497"]}},
    "VESTIARIVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["u0497"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["u0497"]},
        "1": {"morpho": ["n-s---mn2-"], "uri": ["u0497"]},
    },
    "VESTIBVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0498"]}},
    "VESTIGATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["u0508"]}},
    "VESTIGIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0509"]}},
    "VESTIGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0510"]}},
    "VESTIMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0511"]}},
    "VESTIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["u0512"]}},
    "VESTIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["u0515"]}},
    "VESTITVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["u9983"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["101117"]},
    },
    "VESTRI": {"-": {"morpho": ["aps---mn1-"], "uri": []}},
    "VETELIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "VETERA": {
        "-": {"morpho": ["n-s----n3-"], "uri": []},
        "N": {"morpho": ["n-s---nn3-"], "uri": []},
    },
    "VETERANVS": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": ["u0524"]},
    },
    "VETERARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0525"]}},
    "VETERATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["u0527"]}},
    "VETERATORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0528"]}},
    "VETERES": {"-": {"morpho": ["n-s---mn3-"], "uri": ["131341"]}},
    "VETERINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0533"]}},
    "VETERNOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0535"]}},
    "VETERNVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["u0536"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["u0536"]},
    },
    "VETITVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u9998"]}},
    "VETO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0539"]}},
    "VETTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VETVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0541"]}},
    "VETVLVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["u0541"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["u0541"]},
    },
    "VETVRIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "VETVRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VETVS": {
        "-": {"morpho": ["aps---an3-"], "uri": ["28102"]},
        "N": {"morpho": ["aps---an3-"], "uri": ["28102"]},
        "A": {"morpho": ["aps---an3-"], "uri": ["28102"]},
    },
    "VETVSTAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0544"]}},
    "VETVSTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0546"]}},
    "VEXAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["u0548"]}},
    "VEXATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0549"]}},
    "VEXATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["u0551"]}},
    "VEXILLARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["u0553"]}},
    "VEXILLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0556"]}},
    "VEXO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0557"]}},
    "VIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["u0558"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "VIATICVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0565"]}},
    "VIATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["u0566"]}},
    "VIATORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0567"]}},
    "VIBEX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0569"]}},
    "VIBIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VIBO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "VIBRANS": {
        "-": {"morpho": ["aps---an3-", "rp--------"], "uri": []},
        "V\u00e9h\u00e9ment": {"morpho": ["aps---an3-"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "VIBRATVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["u9229"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["131384"]},
    },
    "VIBRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u9574"]}},
    "VIBVRNVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u9575"]}},
    "VICA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "VICARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["u0580"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["u0580"]},
    },
    "VICATIM": {"-": {"morpho": ["rp--------"], "uri": ["u9580", "u9581"]}},
    "VICECOMES": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "VICENARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0583"]}},
    "VICENVS": {"-": {"morpho": ["rp--------"], "uri": []}},
    "VICESIMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0878"]}},
    "VICESIMANVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["33736"]}},
    "VICESIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0781"]}},
    "VICIA": {
        "1": {"morpho": ["n-s---fn1-"], "uri": ["u0632", "u0593"]},
        "2": {"morpho": ["n-s---fn1-"], "uri": ["u0632", "u0593"]},
    },
    "VICIES": {"-": {"morpho": ["rp--------"], "uri": ["u0596"]}},
    "VICINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0588"]}},
    "VICINALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["u0597"]}},
    "VICINIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0599"]}},
    "VICINITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0600"]}},
    "VICINVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0589"]}},
    "VICINVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["u0464"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["u0603"]},
    },
    "VICIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["u0604"]}},
    "VICISSIM": {"-": {"morpho": ["rp--------"], "uri": ["u0605"]}},
    "VICISSITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0607"]}},
    "VICTIMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0609"]}},
    "VICTOR": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["u0614"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "VICTORIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["u0615"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "VICTORIATVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["u0617"]}},
    "VICTRIX": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["58324"]},
        "2": {"morpho": ["aps---an3i"], "uri": ["u0620"]},
        "1": {"morpho": ["aps---an3i"], "uri": ["u0620"]},
    },
    "VICTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["u9986"]}},
    "VICVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["u0623"]}},
    "VICVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["u1342"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["u1342"]},
    },
    "VIDELICET": {"-": {"morpho": ["rp--------"], "uri": ["u0625"]}},
    "VIDENS": {
        "-": {"morpho": ["aps---an3-"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "VIDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["u0626"]}},
    "VIDVA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0633"]}},
    "VIDVERTAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0628"]}},
    "VIDVITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0629"]}},
    "VIDVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0631"]}},
    "VIDVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0633"]}},
    "VIEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["u0635"]}},
    "VIESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["u9635"]}},
    "VIETVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["57189"]}},
    "VIGENS": {"-": {"morpho": ["aps---an3-"], "uri": ["43955"]}},
    "VIGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["u0636"]}},
    "VIGESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["u0636"]}},
    "VIGIL": {
        "-": {"morpho": ["aps---an3-"], "uri": ["u0638"]},
        "2": {"morpho": ["n-s---mn3-"], "uri": ["101135"]},
        "1": {"morpho": ["n-s---mn3-"], "uri": ["101135"]},
    },
    "VIGILANS": {"-": {"morpho": ["aps---an3i"], "uri": ["101136"]}},
    "VIGILANTER": {"-": {"morpho": ["rp--------"], "uri": ["38650"]}},
    "VIGILANTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0657"]}},
    "VIGILES": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "VIGILIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0989"]}},
    "VIGILIARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0643"]}},
    "VIGILO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0644"]}},
    "VIGINTI": {"-": {"morpho": ["app---an--"], "uri": []}},
    "VIGINTIDVO": {"-": {"morpho": ["app---an--"], "uri": []}},
    "VIGINTIVIRATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["u0649"]}},
    "VIGOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["u0651"]}},
    "VILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["u0659"]}},
    "VILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0660"]}},
    "VILLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0662"]}},
    "VILLICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["32706"]}},
    "VILLICO": {"2": {"morpho": ["v1spia--1-"], "uri": ["131464"]}},
    "VILLICVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["32870"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["131467"]},
    },
    "VILLOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0668"]}},
    "VILLVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0669"]}},
    "VILLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["u0671"]}},
    "VIMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["u0672"]}},
    "VIMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0673"]}},
    "VIMINALIS": {
        "-": {"morpho": ["aps---cn3i"], "uri": ["u0674"]},
        "N": {"morpho": ["aps---cn3i"], "uri": ["u0674"]},
        "A": {"morpho": ["aps---cn3i"], "uri": ["u0674"]},
    },
    "VIMINEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0677"]}},
    "VINACEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u9735"]}},
    "VINACEVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["u0678"]}},
    "VINARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0682"]}},
    "VINARIVS": {
        "2": {"morpho": ["aps---mn1-"], "uri": ["u0682"]},
        "1": {"morpho": ["n-s---mn2-"], "uri": ["u0682"]},
    },
    "VINCIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["u0685"]}},
    "VINCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["u0688"]}},
    "VINCTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["u0666"]}},
    "VINCVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0692"]}},
    "VINDEMIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0694"]}},
    "VINDEMIATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["u0700"]}},
    "VINDEX": {
        "-": {"morpho": ["n-s---mn3-"], "uri": []},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "VINDICIAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["101147"]}},
    "VINDICIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VINDICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0707"]}},
    "VINDICTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0708"]}},
    "VINEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u1245"]}},
    "VINEATICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0713"]}},
    "VINETVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0715"]}},
    "VINITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["u0720"]}},
    "VINOLENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u1388"]}},
    "VINOLENTVS": {
        "1": {"morpho": ["n-s----n2-"], "uri": []},
        "2": {"morpho": ["aps---mn1-"], "uri": ["u0726"]},
    },
    "VINOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0728"]}},
    "VINVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0729"]}},
    "VIOLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0732"]}},
    "VIOLABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["u0733"]}},
    "VIOLACEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u1398"]}},
    "VIOLARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0736"]}},
    "VIOLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u1078"]}},
    "VIOLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["u0738"]}},
    "VIOLENS": {"-": {"morpho": ["aps---an3i"], "uri": ["u0740"]}},
    "VIOLENTER": {"-": {"morpho": ["rp--------"], "uri": ["57994"]}},
    "VIOLENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0741"]}},
    "VIOLENTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0740"]}},
    "VIOLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0744"]}},
    "VIPERA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0745"]}},
    "VIPEREVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0747"]}},
    "VIPERINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0748"]}},
    "VIR": {"-": {"morpho": ["n-s---mn2r"], "uri": ["u0750"]}},
    "VIRAGO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0753"]}},
    "VIRENS": {"-": {"morpho": ["aps---an3i"], "uri": ["101152"]}},
    "VIRENTIA": {"-": {"morpho": ["n-p---nn3i"], "uri": ["101153"]}},
    "VIREO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["u0758"]},
        "2": {"morpho": ["v1spia--2-"], "uri": ["u0757"]},
    },
    "VIRESCO": {
        "1": {"morpho": ["v1spia--3-"], "uri": ["u0757"]},
        "2": {"morpho": ["v1spia--3-"], "uri": ["u0757"]},
        "-": {"morpho": ["v1spia--3-"], "uri": ["u0757"]},
    },
    "VIRETVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["131531"]}},
    "VIRGA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0759"]}},
    "VIRGATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0762"]}},
    "VIRGEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0764"]}},
    "VIRGILIAE": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "VIRGILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VIRGINALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["u0767"]}},
    "VIRGINEVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["u0770"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["u0770"]},
    },
    "VIRGINITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0771"]}},
    "VIRGINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VIRGO": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["u0773"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "VIRGVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0775"]}},
    "VIRGVLTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0777"]}},
    "VIRGVNCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0778"]}},
    "VIRIDANS": {"-": {"morpho": ["aps---an3i"], "uri": ["101157"]}},
    "VIRIDE": {
        "1": {"morpho": ["n-s---nn3i"], "uri": ["u0591"]},
        "2": {"morpho": ["rp--------"], "uri": ["131549"]},
        "-": {"morpho": ["rp--------"], "uri": ["131549"]},
    },
    "VIRIDIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["u0785"]}},
    "VIRIDITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0786"]}},
    "VIRIDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0787"]}},
    "VIRILIS": {
        "-": {"morpho": ["aps---cn3i"], "uri": ["u0789"]},
        "A": {"morpho": ["aps---cn3i"], "uri": ["u0789"]},
    },
    "VIRILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0790"]}},
    "VIRILITER": {"-": {"morpho": ["rp--------"], "uri": ["37341"]}},
    "VIRITIM": {"-": {"morpho": ["rp--------"], "uri": ["u0796"]}},
    "VIROSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u9999", "u0799"]}},
    "VIRTVS": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["u0800"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "VIRVS": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0803"]}},
    "VIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0804"]}},
    "VISCERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0812"]}},
    "VISCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0807"]}},
    "VISCVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0817"]}},
    "VISCVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": []},
        "2": {"morpho": ["n-s---nn3-"], "uri": ["u0816"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "VISITATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0824"]}},
    "VISITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0826"]}},
    "VISO": {"-": {"morpho": ["v1spia--3-"], "uri": ["u0827"]}},
    "VISVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u9988"]}},
    "VISVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["u9988"]}},
    "VITA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0835"]}},
    "VITABVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0837"]}},
    "VITALIA": {"-": {"morpho": ["n-p---nn3i"], "uri": ["u0838"]}},
    "VITALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["u0838"]}},
    "VITALITER": {"-": {"morpho": ["rp--------"], "uri": ["29606"]}},
    "VITELLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VITELLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["u0844", "u0843"]}},
    "VITEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0845"]}},
    "VITEX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0846"]}},
    "VITIARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0848"]}},
    "VITICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0854"]}},
    "VITIGENVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0857"]}},
    "VITIGINEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0858"]}},
    "VITILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["u0860"]}},
    "VITILITIGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0862"]}},
    "VITIO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["u0865"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["u0865"]},
    },
    "VITIOSITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0866"]}},
    "VITIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0867"]}},
    "VITIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["u0869"]}},
    "VITIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0871"]}},
    "VITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0872"]}},
    "VITRARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["131620"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["131619"]},
    },
    "VITREVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u1345"]}},
    "VITREVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u1348"]}},
    "VITRICVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["u0880"]}},
    "VITRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u1132"]}},
    "VITTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0882"]}},
    "VITTATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0883"]}},
    "VITVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0889"]}},
    "VITVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["u0889"]}},
    "VITVPERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0891"]}},
    "VITVPERIVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "VITVPERO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["u0895"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["u0894"]},
    },
    "VIVARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0898"]}},
    "VIVATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0899"]}},
    "VIVAX": {"-": {"morpho": ["aps---an3i"], "uri": ["u0900"]}},
    "VIVENS": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "VIVESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["u1006"]}},
    "VIVIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0905"]}},
    "VIVIRADIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["40744"]}},
    "VIVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["u0913"]}},
    "VIVVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["101172"]}},
    "VIVVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["101173"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["u0914"]},
    },
    "VIX": {"-": {"morpho": ["rp--------"], "uri": ["u0915"]}},
    "VIXDVM": {"-": {"morpho": ["rp--------"], "uri": ["u0916"]}},
    "VLCERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0918"]}},
    "VLCERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0919"]}},
    "VLCEROSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0920"]}},
    "VLCISCOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["u0921"]}},
    "VLCVS": {"-": {"morpho": ["n-s---nn3-"], "uri": ["u0922"]}},
    "VLCVSCVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0923"]}},
    "VLIGO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0926"]}},
    "VLIXES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "VLIXEVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "VLLVS": {"-": {"morpho": ["----------"], "uri": []}},
    "VLMEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0933"]}},
    "VLMVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["u0935"]}},
    "VLNA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0936"]}},
    "VLPICVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0939"]}},
    "VLTERIOR": {
        "-": {"morpho": ["aps---cn3-"], "uri": []},
        "A": {"morpho": ["aps---cn3-"], "uri": []},
    },
    "VLTERIORA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["101005"]}},
    "VLTIMVM": {
        "1": {"morpho": ["n-s---nn2-"], "uri": ["u0018"]},
        "2": {
            "morpho": ["rp--------", "n-s---nn2-"],
            "uri": ["39171", "u0018"],
        },
    },
    "VLTIMVS": {
        "-": {
            "morpho": ["aps---mn1-", "aps---mn1-"],
            "uri": ["u0943", "u0943"],
        }
    },
    "VLTIO": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["u0944"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "VLTOR": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["u0945"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "VLTRA": {
        "1": {"morpho": ["rp--------"], "uri": ["27896"]},
        "2": {"morpho": ["p---------"], "uri": ["u0947"]},
    },
    "VLTRIX": {"-": {"morpho": ["n-s---fn3i"], "uri": ["u0950"]}},
    "VLTRO": {"-": {"morpho": ["rp--------"], "uri": ["u0951"]}},
    "VLVA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0955"]}},
    "VLVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0957"]}},
    "VLVLATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["u0976"]}},
    "VLVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0961"]}},
    "VMBELLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0963"]}},
    "VMBILICVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["u0966"]}},
    "VMBO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["u0967"]}},
    "VMBRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0968"]}},
    "VMBRACVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0970"]}},
    "VMBRATICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0972"]}},
    "VMBRIFER": {"-": {"morpho": ["aps---mn1r"], "uri": ["u0978"]}},
    "VMBRO": {
        "-": {"morpho": ["v1spia--1-"], "uri": ["u0979"]},
        "N": {"morpho": ["n-s----n1-"], "uri": []},
    },
    "VMBROSA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["u0980"]}},
    "VMBROSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0980"]}},
    "VMQVAM": {"-": {"morpho": ["rp--------"], "uri": ["u1000"]}},
    "VNA": {"-": {"morpho": ["rp--------"], "uri": ["27908"]}},
    "VNANIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u1399"]}},
    "VNCIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u1009"]}},
    "VNCIARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u1011"]}},
    "VNCINVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["u1014"]}},
    "VNCIOLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u1015"]}},
    "VNCTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u1018"]}},
    "VNCTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["u1021"]}},
    "VNCTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u9984"]}},
    "VNCTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["u9984"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["43318"]},
    },
    "VNCVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["u1026"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["u1026"]},
    },
    "VNDA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u1027"]}},
    "VNDE": {
        "1": {"morpho": ["rp--------"], "uri": ["u1031"]},
        "2": {"morpho": ["rp--------"], "uri": ["u1031"]},
    },
    "VNDECIM": {"-": {"morpho": ["----------"], "uri": []}},
    "VNDECIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u1401"]}},
    "VNDECVMQVE": {
        "1": {"morpho": ["rp--------"], "uri": ["u7500"]},
        "2": {"morpho": ["rp--------"], "uri": ["u7500"]},
    },
    "VNDENI": {"-": {"morpho": ["rp--------"], "uri": []}},
    "VNDEOCTOGINTA": {"-": {"morpho": ["----------"], "uri": []}},
    "VNDEQVINQVAGESIMVS": {"-": {"morpho": ["----------"], "uri": []}},
    "VNDEQVINQVAGINTA": {"-": {"morpho": ["----------"], "uri": []}},
    "VNDETRICESIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u1056"]}},
    "VNDETRIGINTA": {"-": {"morpho": ["----------"], "uri": []}},
    "VNDEVICESIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["29582"]}},
    "VNDEVIGINTI": {"-": {"morpho": ["----------"], "uri": []}},
    "VNDIQVE": {"-": {"morpho": ["rp--------"], "uri": ["u1064"]}},
    "VNDISONVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u1066"]}},
    "VNDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u1068"]}},
    "VNDOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u1069"]}},
    "VNETVICESIMANI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["37393"]}},
    "VNETVICESIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["48692"]}},
    "VNGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["u1091"]}},
    "VNGVEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["u1077"]}},
    "VNGVENTARIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u1404"]}},
    "VNGVENTARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u1406"]}},
    "VNGVENTARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["u1405"]},
        "2": {"morpho": ["aps---mn1-"], "uri": ["u1403"]},
    },
    "VNGVENTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u1079"]}},
    "VNGVENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u1080"]}},
    "VNGVICVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["u1081"]}},
    "VNGVIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["u1085"]}},
    "VNGVITO": {"-": {"morpho": ["v1spia--1-"], "uri": [None, "u1086"]}},
    "VNGVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u1246"]}},
    "VNICE": {"-": {"morpho": ["rp--------"], "uri": ["27926"]}},
    "VNICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u1098"]}},
    "VNIFORMIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["u1100"]}},
    "VNIGENA": {"-": {"morpho": ["aps---cn1-"], "uri": ["101026"]}},
    "VNIO": {
        "1": {"morpho": ["n-s---fn3-"], "uri": []},
        "2": {"morpho": ["v1spia--4-"], "uri": ["u1111"]},
    },
    "VNITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u1119"]}},
    "VNITER": {"-": {"morpho": ["rp--------"], "uri": ["49720"]}},
    "VNIVERSALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["u1121"]}},
    "VNIVERSE": {"-": {"morpho": ["rp--------"], "uri": ["35313"]}},
    "VNIVERSI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["101029"]}},
    "VNIVERSITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u1125"]}},
    "VNIVERSVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0654"]}},
    "VNIVERSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u1126"]}},
    "VNVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "VNVS": {"-": {"morpho": ["----------"], "uri": []}},
    "VNVSQVISQVE": {"-": {"morpho": ["----------"], "uri": []}},
    "VNVSQVISQVIS": {"-": {"morpho": ["----------"], "uri": []}},
    "VOCABVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u1138"]}},
    "VOCALES": {"-": {"morpho": ["n-p---mn3i"], "uri": ["131808"]}},
    "VOCALIS": {
        "1": {"morpho": ["n-s---fn3i"], "uri": ["131809"]},
        "2": {"morpho": ["aps---cn3i"], "uri": ["u1139"]},
    },
    "VOCAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["u1141"]}},
    "VOCATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u1142"]}},
    "VOCATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["u1144"]}},
    "VOCIFERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u1146"]}},
    "VOCIFEROR": {"-": {"morpho": ["v1spid--1-"], "uri": ["u1148"]}},
    "VOCITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u1151"]}},
    "VOCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u1152"]}},
    "VOCONIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "VOCVLA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["u1153"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "VOLANS": {"-": {"morpho": ["n-s---cn3-"], "uri": ["40174"]}},
    "VOLANTES": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "VOLATICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u1159"]}},
    "VOLATILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["u1160"]}},
    "VOLATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["u9991"]}},
    "VOLEMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["41682"]}},
    "VOLENS": {"-": {"morpho": ["aps---an3-"], "uri": ["33774"]}},
    "VOLERO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "VOLESVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VOLITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u1165"]}},
    "VOLO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["u1169"]},
        "2": {"morpho": ["v1spia--1-"], "uri": ["u1168"]},
        "3": {"morpho": ["v1spia--1-"], "uri": ["u1168"]},
    },
    "VOLSCI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "VOLSCVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "VOLTEIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VOLVBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["u1174"]}},
    "VOLVBILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u1175"]}},
    "VOLVCER": {"-": {"morpho": ["aps---mn3-"], "uri": ["u1176"]}},
    "VOLVCRIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["28198"]}},
    "VOLVMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["u1182"]}},
    "VOLVMNIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "VOLVNTARII": {"-": {"morpho": ["n-p---mn2-"], "uri": ["101182"]}},
    "VOLVNTARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u1184"]}},
    "VOLVNTAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u1185"]}},
    "VOLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["u1188"]}},
    "VOLVPTARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u1192"]}},
    "VOLVPTAS": {
        "-": {"morpho": ["n-s---fn3-"], "uri": ["u1193"]},
        "N": {"morpho": ["n-s---fn3-"], "uri": ["u1193"]},
    },
    "VOLVSIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VOLVTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u9978"]}},
    "VOLVTABRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u1199"]}},
    "VOLVTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u1201"]}},
    "VOLVTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u1202"]}},
    "VOMER": {"-": {"morpho": ["n-s---mn3-"], "uri": ["28203"]}},
    "VOMICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u7205"]}},
    "VOMIS": {"-": {"morpho": ["n-s---mn3-"], "uri": ["101184"]}},
    "VOMITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u1209"]}},
    "VOMITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u1211"]}},
    "VOMITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["u1212"]}},
    "VOMITORIVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "VOMITORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u1213"]}},
    "VOMITVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["u9995"]}},
    "VOMO": {"-": {"morpho": ["v1spia--3-"], "uri": ["u1214"]}},
    "VORAGINOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u1217"]}},
    "VORAGO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u1218"]}},
    "VORAX": {"-": {"morpho": ["aps---an3i"], "uri": ["u1221"]}},
    "VORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u1222"]}},
    "VOS": {"-": {"morpho": ["o-----an--"], "uri": []}},
    "VOTIVVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "VOTIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u1226"]}},
    "VOTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["55454"]}},
    "VOTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u9882"]}},
    "VOVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["u1227"]}},
    "VOX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u1228"]}},
    "VRBANI": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "VRBANITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u1236"]}},
    "VRBANVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u1237"]}},
    "VRBIVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "VRBS": {
        "-": {"morpho": ["n-s---fn3i"], "uri": ["u1243"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "VRCEOLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["u1412"]}},
    "VRCEVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["u7248"]}},
    "VREDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u1248"]}},
    "VRGENS": {"-": {"morpho": ["aps---an3i"], "uri": ["101039"]}},
    "VRGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["u1252"]}},
    "VRINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u1254"]}},
    "VRIVS": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "VRNA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u1260"]}},
    "VRNALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["u1262"]}},
    "VRO": {"-": {"morpho": ["v1spia--3-"], "uri": ["u1266"]}},
    "VRSA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["u1269"]},
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
    },
    "VRSVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["u1269"]}},
    "VRTICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u1274"]}},
    "VRVCA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["e0930"]}},
    "VRVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["u1277"]}},
    "VSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u1282"]}},
    "VSITATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["27944"]}},
    "VSPIAM": {"-": {"morpho": ["rp--------"], "uri": ["u1286"]}},
    "VSQVAM": {"-": {"morpho": ["rp--------"], "uri": ["u1287"]}},
    "VSQVE": {"-": {"morpho": ["rp--------"], "uri": ["27945"]}},
    "VSQVEQVAQVE": {"-": {"morpho": ["rp--------"], "uri": ["u1289"]}},
    "VSTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u9979"]}},
    "VSTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["u1295"]}},
    "VSTVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u1299"]}},
    "VSVALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["u1300"]}},
    "VSVCAPIO": {
        "1": {"morpho": ["n-s---fn3-"], "uri": ["u1303"]},
        "2": {"morpho": ["v1spia--3i"], "uri": ["u1302"]},
    },
    "VSVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u1334"]}},
    "VSVRPATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u1311"]}},
    "VSVRPO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u1317"]}},
    "VSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["u9985"]}},
    "VSVSFRVCTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["57540"]}},
    "VSVVENIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["131977"]}},
    "VT": {
        "1": {"morpho": ["rp--------"], "uri": ["37549"]},
        "2": {"morpho": ["rp--------"], "uri": ["37549"]},
        "3": {"morpho": ["rp--------"], "uri": ["37549"]},
        "4": {"morpho": ["----------"], "uri": []},
    },
    "VTCVMQVE": {
        "1": {"morpho": ["rp--------"], "uri": ["58357"]},
        "2": {"morpho": ["rp--------"], "uri": ["58357"]},
    },
    "VTENS": {"-": {"morpho": ["aps---an3i"], "uri": ["101044"]}},
    "VTENSILIA": {"-": {"morpho": ["n-p---nn3i"], "uri": ["u1320"]}},
    "VTENSILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["u1320"]}},
    "VTER": {
        "1": {"morpho": ["n-s---mn2r"], "uri": ["131984"]},
        "2": {"morpho": ["aps---mn1-"], "uri": []},
        "3": {"morpho": ["----------"], "uri": []},
        "4": {"morpho": ["----------"], "uri": []},
    },
    "VTERCVMQVE": {
        "1": {"morpho": ["aps---mn1-"], "uri": []},
        "2": {"morpho": ["----------"], "uri": []},
    },
    "VTERLIBET": {"-": {"morpho": ["----------"], "uri": []}},
    "VTERQVE": {"-": {"morpho": ["----------"], "uri": []}},
    "VTERVIS": {"-": {"morpho": ["----------"], "uri": []}},
    "VTERVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["u1324"]}},
    "VTICA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "VTICENSIS": {
        "N": {"morpho": ["aps---cn3i"], "uri": []},
        "A": {"morpho": ["aps---cn3i"], "uri": []},
    },
    "VTILE": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "VTILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["u1327"]}},
    "VTILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u1328"]}},
    "VTILITER": {"-": {"morpho": ["rp--------"], "uri": []}},
    "VTINAM": {"-": {"morpho": ["rp--------"], "uri": ["u1329"]}},
    "VTIQVE": {"-": {"morpho": ["rp--------"], "uri": ["u1332"]}},
    "VTOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["u1333"]}},
    "VTPOTE": {"-": {"morpho": ["rp--------"], "uri": ["u9891"]}},
    "VTQVI": {"-": {"morpho": ["----------"], "uri": []}},
    "VTRAQVE": {"-": {"morpho": ["rp--------"], "uri": ["101048"]}},
    "VTRICVS": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "VTRIMQVE": {"-": {"morpho": ["rp--------"], "uri": ["u1341"]}},
    "VTRIMQVESECVS": {"-": {"morpho": ["rp--------"], "uri": ["u9893"]}},
    "VTRINDE": {"-": {"morpho": ["rp--------"], "uri": ["u1343"]}},
    "VTRO": {
        "1": {"morpho": ["rp--------"], "uri": ["33152"]},
        "2": {"morpho": ["rp--------"], "uri": ["33152"]},
    },
    "VTROBI": {"-": {"morpho": ["rp--------"], "uri": ["131995"]}},
    "VTROBIQVE": {"-": {"morpho": ["rp--------"], "uri": ["49476"]}},
    "VTROQVE": {"-": {"morpho": ["rp--------"], "uri": ["39779"]}},
    "VTRVM": {"-": {"morpho": ["rp--------"], "uri": ["56265"]}},
    "VTRVMNE": {"-": {"morpho": ["r---------"], "uri": []}},
    "VVA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u1351"]}},
    "VVESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["u1353"]}},
    "VVIDVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u1355"]}},
    "VVIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u1356"]}},
    "VVLCANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VVLCATIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VVLGARIA": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "VVLGARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["u1415"]}},
    "VVLGATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["u1361"]}},
    "VVLGATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["28215"]}},
    "VVLGIVAGVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u1362"]}},
    "VVLGO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["u1363"]},
        "2": {"morpho": ["rp--------"], "uri": ["u1364"]},
    },
    "VVLGVS": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u1365"]}},
    "VVLNERABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["u1366"]}},
    "VVLNERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u1368"]}},
    "VVLNERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u1370"]}},
    "VVLNIFICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u1372"]}},
    "VVLNVS": {"-": {"morpho": ["n-s---nn3-"], "uri": ["u1373"]}},
    "VVLPECVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u1375"]}},
    "VVLPES": {"-": {"morpho": ["n-s---fn3i"], "uri": ["u1376"]}},
    "VVLTVR": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["u1383"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "VVLTVRIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["u1385"]}},
    "VVLTVRNVS": {
        "-": {"morpho": ["n-s----n2-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "VVLTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["u1418"]}},
    "VVLVA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u1187"]}},
    "VXOR": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u1391"]}},
    "VXORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u1394"]}},
    "XENIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["x0003"]}},
    "XENO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "XENOCRATES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "XERAMPELINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["x0008"]}},
    "XYSTVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["x0029"]}},
    "ZAMA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ZELOTYPVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["z0014"]}},
    "ZELVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["z0015"]}},
    "ZENO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ZEPHYRVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "-": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "ZMARAGDVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["44559"]}},
    "ZONA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["z0035"]}},
    "ZONVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["z0040"]}},
    "ZOPHORVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["z0042"]}},
    "ZOSIPPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ZYGIOS": {
        "-": {"morpho": ["aps---an3-"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "ZYGON": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "ALTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a1528"]}},
    "AMFRACTVS": {
        "1": {"morpho": ["n-s---mn4-"], "uri": ["44184"]},
        "2": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "BVTHROTOS": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "CHERSONENSVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "COMMINGO5L": {"COMMINISCOR": {"morpho": ["v1spia--3-"], "uri": []}},
    "CONSVLTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["21635"]}},
    "CORIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c4037"]}},
    "DISIVNCTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["38857"]}},
    "DIVNGO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "DISIVNGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1800"]}},
    "FVFIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "GERMANA11": {"GERMANIA": {"morpho": ["n-s----n1-"], "uri": []}},
    "HOSTVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["h0642"]}},
    "LAEVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0133"]}},
    "NEMPE": {"-": {"morpho": ["rp--------"], "uri": ["n0757"]}},
    "PAVLLO": {"-": {"morpho": ["rp--------"], "uri": ["42604"]}},
    "PETREIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PROMVNTVRIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p3908"]}},
    "RECERNO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "RVSTICATIO": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "SCOBIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["s0690"]}},
    "SVPERINICIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["s3602"]}},
    "_SVM": {
        "1": {"morpho": ["v1spia--3-"], "uri": []},
        "2": {"morpho": ["v1spia--3-"], "uri": []},
    },
    "VENERIVS": {
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "#": {"-": {"morpho": ["----------"], "uri": []}},
    "BABYLONICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MAMMIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MICIPSA": {"N": {"morpho": ["n-s---mn1-"], "uri": []}},
    "SVCRONENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "CANDAVIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CECROPIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LVCCIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SEGIMERVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BEDRIACVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "PYLIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PERILLEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "REMMIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "ICARIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "LATIARIS": {
        "A": {"morpho": ["aps---cn3i"], "uri": []},
        "N": {"morpho": ["n-s----n--"], "uri": []},
    },
    "ACHILLEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LIBANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["l0661"]}},
    "APAMEA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CRISPINILLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LOLLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CAIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MARCIANVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "SABELLVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ZOSIMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ILLYRIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "FVCINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARGYRIPPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PRIAMEIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ORTYGIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "VALA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "THYBRIS": {
        "A": {"morpho": ["aps---cn3i"], "uri": []},
        "N": {"morpho": ["n-s----n--"], "uri": []},
    },
    "AVERNA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SCAEA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PAPIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "CRVSTVMIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SVLMONENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "EPYTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "RAVOLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SALAMINIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "APOPHORETA": {
        "2": {"morpho": ["n-p---nn2-"], "uri": ["a2996"]},
        "1": {"morpho": ["n-p---nn2-"], "uri": ["a2996"]},
    },
    "VESTRICIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CEPHEIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ACVTIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CATVLLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "COLCHVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CVPIENNIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PHEMIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ASCRA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "DVCENIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CARVILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CHERVSCVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TESSERARIVS": {"1": {"morpho": ["n-s---mn2-"], "uri": ["t0424"]}},
    "LEVCTRA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SAENIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "OBARITVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VIBVLENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TARRACINENSIS": {
        "A": {"morpho": ["aps---cn3i"], "uri": []},
        "N": {"morpho": ["n-s----n--"], "uri": []},
    },
    "HIBERVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "ORCHIVIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TALAVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CLONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TIMOTHEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LATERANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "STATILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GABINIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "FLAVIANVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "HICETAONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CLARIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "ACILIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CINNAMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "COLMINIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CARATACVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VOPISCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["u1215"]}},
    "ACTAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["a0438"]}},
    "SIMBRVINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ARIOVISTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SOCRATICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "BIBVLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "OEENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "MEGAREA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SOPHOCLEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ALBINTIMILIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "HIPPARCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SILANA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "AMALTHEA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CASTALIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CAPHAREVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "ELYSIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "OTOGESA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "THAMYRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NERITIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ALCATHOVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TOGONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VOLTVRCIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SALENTINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TIRYNTHIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "FVSCINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PHENEVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "TOLVMNIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LATIVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["115912"]}},
    "SEMIPLACENTINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "AMORGVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PASITHEA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "VOLVMNIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ANCHARIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PHALEREVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SORICARIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "BAEBIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MODIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LYCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MASSICVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "CAVARILLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LABDACVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PALINVRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LARONIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "VTRICVLVS": {"1": {"morpho": ["n-s---mn2-"], "uri": ["u1340"]}},
    "TRAIANVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "GABOLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SVMMANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HYRIEVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "STYMPHALA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ILIONEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CEPHEVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "PEDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CAPVA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "VOLSINIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LAESTRYGONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "AETHIOPIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CAMPANVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": ["c9916"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "AMPSANCTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HORVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TRAVSIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VITELLIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PATVLCIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CVRTILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SCANTIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MEMNONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SATYRA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PERSICVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "TERPOLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "POMPONIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CAMVLOGENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AGNALIA": {"N": {"morpho": ["n-p---nn3i"], "uri": []}},
    "COLONEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SVADELA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TRYPHAENA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SARDANAPALVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "IASIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DECRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PENTHESILEA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CESTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TRINACRIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ADIATVANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MENA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "FIRMINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CLITVMNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TIBVRTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DAMASIPPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PEGASEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VOSEGVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CHLOREVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["c1465"]}},
    "NARNIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "RHOETVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LYCAONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CHAONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PEDANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TRIFOLINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TARENTVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "CYLLARVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CLARANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SALIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "LYSIPPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OFONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GYGAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "HYPERBOREVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["46332"]}},
    "ORESTILLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "POLYPHEMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ATRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LANVVIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "GALLOGRAECIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ARQVITIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DELIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "VBIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "NONACRIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "NERVLLINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "QVADRATILLA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "IVLIANVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": ["99172"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["49311"]},
    },
    "ANAGNIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ZALACVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MIMALLONEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ARGIVVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TARRACINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "VIBELLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TISAMENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TANTALEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "BIONEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CYDNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LARINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CAMARINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SEBOSIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "NERTHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CALVINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CLEOPATRA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PISANVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "RVLLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["r0888"]}},
    "SILARVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BVTA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "BAETICA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "FORTVNATVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["23461"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "HYPSIPYLAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ATHENAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["97705"]}},
    "FADIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CLAVDIALIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "MEGABOCCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CINCIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ATELLA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PERGAMVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "ANXVRVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MASTANESOSVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SVETONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CINARA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CATONIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ITHACA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ZACYNTHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TILLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "STICHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NEMETOCENNA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ALEXANDRINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "GRAECINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "HERODVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CAMVRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MARTIGENA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "HECTOREVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LVGDVNENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "CORELLIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TVRANNIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HERMIPPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LILYBAEA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ENTELLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PHOLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TROILVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VENETVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": ["u0294"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["u0294"]},
    },
    "PHTHIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "OPPIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DACICVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "BALEARICVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HELVIDIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "SILVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NISEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "THESSALIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "ACRAGANTINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "IANALIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "LVCTERIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LITERNINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PHERONACTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "IVNONICOLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MEDIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PARAETONICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SALONINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CARPATHIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CAENINA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "THYONIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LENNIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CAECILIANVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "BEDRIACVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PANDATERIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TARCONDARIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CANOPIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SALERNVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "BATAVVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CVRTILLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "QVARTILLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SVEBVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PACENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "IAPETVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NASIDIENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "FAVSTA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MOESIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "HARPAGVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "FAESIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HYBLA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "MESSALLA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "MESSIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ILISSVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["101287"]}},
    "AMBIVIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NAVPACTOVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LAECA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "TYPHOEVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "IPHITVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARGYNNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BAIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "HYPSAEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HARPALVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DECIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ANDROGEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LYCOMEDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PHOCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BETACEVS": {"1": {"morpho": ["n-s---mn2-"], "uri": ["b0542"]}},
    "SATRACHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TVRIENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "CROTONIENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "PHIDIPPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MAELIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OEBALVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PAVLLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LAVREOLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TAENARIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LICINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OENOTRIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ANICIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LVSTRICIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VEIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "VACCENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "DINIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "METHYMNAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PAGYDA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LAMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PAGASA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "VOLAGINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "REATINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MELISSA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "FLEVVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "DECIBALVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SINOPA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "GORDYAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "RABOCENTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LIBYA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CORDOBA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ANTEIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PARIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "SALVIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NEAPOLITANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TRITONIS": {
        "A": {"morpho": ["aps---cn3i"], "uri": []},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "CRINISVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MACCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["m0005"]}},
    "CLAVDIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SCYLACEVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "THEMISCYRA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "AGERMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ACYLLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "OECHALIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "VIRGINIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TITIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ILIONA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SOPHRONISCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NEAERA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TRANSPADANVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["t1075"]}},
    "HIMELLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PARRHASIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "APSVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CAPRICORNVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": ["c0543"]},
        "-": {"morpho": ["n-s----n2-"], "uri": []},
    },
    "PVLCHRA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PACHYNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "COSYRA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "LVGDVNVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "ACADEMICA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ZOILVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ORATA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PATROBIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MARSICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ONAGVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["40212"]}},
    "SIMONIDEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CADIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CLVVIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TRASIMENVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "SVLGIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "THEOPHRASTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PESSINVNTIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CHALYBEIVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["c1342"]}},
    "PSEVDOPHILIPPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HIRTIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "CORFIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARRIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "MECVLONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MARSAEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "FALANIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VICTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ORTINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LOLLIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PANNONICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "THRAESSA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CONTEREBROMNIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MYSIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "OLYNTHIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "DARDANIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MONOBAZVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "EVTHYNICVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARATVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["97658"]}},
    "CINNANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "EVANDRIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VORTVMNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GRAECANICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CYRENAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MELIBOEVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "FVNDANIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CATIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VOLVSVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TELEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PRISCA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "METABVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LIGVRINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VERVLAMIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "OGVLNIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "VINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SIGNINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "BACTRA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "GONOESSA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LEGERDA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "NEMEAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "NYCTEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MAGNESIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "RVSPINA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "LIBYCVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PHARNABAZVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AVRVNCVLEIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BERECYNTIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PHERAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "HELYMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AMBRACIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TIBERIOLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "METISCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PHILADELPHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PRIVERNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AESONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CLYTAEMESTRA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "DINDYMENA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ACCIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DANAVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "BOIOCALVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AQVILIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CASINVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "VICESIMARIVS": {"1": {"morpho": ["n-s---mn2-"], "uri": ["u0710"]}},
    "ALFIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARIADNAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VELLAVNODVNVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "SPOLETINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "POLYHYMNIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "POMEIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CAMERINVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "NVCERINVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "LAPITHA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PAVLINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TINGITANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "THABENA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "BLVCIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "HERMARCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AESCHYLEVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["a3720"]}},
    "PELETHRONIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "CORREVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["c4108"]}},
    "BADVHENNA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "COEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CALLISTRATVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CAENEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LEDAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "APICATA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CERCINA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "QVINTILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CARYSTIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MALTINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MAZIPPA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "DARDANIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "BITHYNICVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "GARGARA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ELIDENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "STRATONICA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "HELVETICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "IVNONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LAVRENTIVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["48767"]}},
    "CHALDAEVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "CROESVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AQVILIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SCIRONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CARTIMANDVA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ARCESILAVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VANNIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PLINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ERGENNA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "NIOBEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ARIMIENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "SVTRIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "ARVPINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MOCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CERBEREVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "IBYCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NABDALSA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SORANVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "AMPHOTERVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VENVSTINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PALAEPHARSALVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "QVADRIGARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["q0043"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["q0043"]},
    },
    "VINNIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MAGIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "COMMIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "THERAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PERIBOMIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HESVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CYTHEREA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LARGA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "VRBICVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["53478"]}},
    "CALABRIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "LAVSVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "IVVENALIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "HIPPOMENEIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "AMPHAEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DAVNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ASYLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "FAVENTINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DVRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GESSIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ILITHYIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PLAGVLEIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OCELVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "CEIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SABELLICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "HECATAEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GANGETIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "RVFVLVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["r0877"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "SEIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SINVESSA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "HORMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["h0900"]}},
    "VERCASSIVELLAVNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MAECIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "EPIDAVRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LVCILIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PHILIPPEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CALATINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TREVIRVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CHALINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NICOMEDIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "DACIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "DAMASCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ATACINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LAERTIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "CEBALINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ABVDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PLOTINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "AQVINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MVNIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ANDRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HEBRAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VENVSINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CLEOBVLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PANATHENAICVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARELLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CHIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "FONTEIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "QVIRINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TVTICANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BABYLONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "DIDIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "FABRICIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": ["f0020"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "MAEOTIS": {
        "A": {"morpho": ["aps---cn3i"], "uri": []},
        "N": {"morpho": ["n-s----n--"], "uri": []},
    },
    "IDYIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "OSTIENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "QVINQVATRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["q0250"]}},
    "LACAENA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "APONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PRAECONINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AVGAEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MAEONIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LILYBEIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "AMISVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CASMILLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ASVVIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SALARIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": ["s0144"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["s0144"]},
    },
    "ABYDENVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CAVDIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "HYMETTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PARAPANISVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PORTICANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DIANIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ICCIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MARICA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "LVSCIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BROGITARVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TAMPIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ADIABENVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CLVENTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PLAVTIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "TIFERNIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "BONONIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "MAEONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "NERIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CHAERIPPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ABDERA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "METTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ZALEVCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ELEGIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TYNDAREVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "COMETA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SEDVLLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TANTALVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SCYLLAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "BITHYNIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VELEDA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "HELIODORVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SCAEVA": {
        "2": {"morpho": ["n-s---mn1-"], "uri": ["s4025"]},
        "N": {"morpho": ["n-s----n2-"], "uri": []},
    },
    "ORICVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "PANAETIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HISPVLLA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "TMARVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "THERAPNAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CLVSIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "APPVLEIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": ["101235"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "GRATIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ERYTHEIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "TVDICIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "EPHYREIACVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CASTRICIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["c0795"]}},
    "ATALANTA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "BOSPORANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TELESPHORVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PLANESIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "AMISSVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CRITOGNATVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PRVSA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "LITERNVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "OSSA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "LETOVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TARENTINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "IVGVRTHA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CYNOSVRA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "NASIDIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "FENESTELLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CREMERENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "CASCA": {"N": {"morpho": ["n-s---fn1-"], "uri": ["c0734"]}},
    "CLYMENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["c2157"]}},
    "STAIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GIGANTEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "DIPHILVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PYTHIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "GORGONEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "BESTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MALCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HIERONYMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PYTAGOREVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LONDINIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "VINALIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ALFENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TRYPHERVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MEGARVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PARTHENOPAEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LECHAEVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "HERIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OCNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NEREIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SARNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BAZAIRA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "THASVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ROMILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "FORMIANVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "TRISANTONA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "EPIMACHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CIRTA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ANDROMACHA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CYRENENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "VERANIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LARINVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "GANNASCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ABNOBA": {"N": {"morpho": ["n-s---mn1-"], "uri": []}},
    "SERGESTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "EPPONINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "EVRYSTHEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CYTHERIACVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CALLIOPEA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MARVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MARRVCINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MYTILENENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "ARISTIPPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NOREIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TRANSRHENANVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["t1088"]}},
    "STATILIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "OCELLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SARDONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TITARESVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "FAENIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OENOTRVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "COSSINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NEMORENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": ["n0202"]}},
    "VIRBIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PANSA": {"N": {"morpho": ["n-s---mn1-"], "uri": ["p9864"]}},
    "FORTVNATA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SISYPHIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "HATERIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VMBRIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "GAVDA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PROTESILAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CYTHEREIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LAECANIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "RECENTORICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CONSVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARGANTHONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CAVILLONVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "LESBIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": ["30487"]},
    },
    "CLEONAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "DAVVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HASTENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "COMPSA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PYLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DARDANIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "THERMODONTIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "GALLINARIVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["g0059"]}},
    "CAESIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARISTOBVLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ADRASTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GALATIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CLOTA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "QVINTIANVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "VESTRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PAVSIACVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ARABVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": ["97656"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["101236"]},
    },
    "APRILIS": {
        "N": {"morpho": ["n-s---fn3-"], "uri": []},
        "A": {"morpho": ["n-s---fn3-"], "uri": []},
    },
    "TVSCVLVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "PATISIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CAMILLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["c0422"]}},
    "GRAVPIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ACERRONIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SOSPITA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LESBIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "EPIDAMNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "THESSANDRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TORONAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "BELVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TELEGONVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TEMNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BAGRADA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TELAMONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "STABERIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CYTORIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MINOTAVRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ANGITIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CRETANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MAEOTICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ANAPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ITYRAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TARRACINIENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "SAMAXVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LEMANNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "EVODVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DONVSA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "NILIACVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "DVRATIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "FLORALIS": {"A": {"morpho": ["aps---cn3i"], "uri": ["f0684"]}},
    "ZONARIVS": {"1": {"morpho": ["n-s---mn2-"], "uri": ["z0006"]}},
    "EPRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "FABVLLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "COTTIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "NIREVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARDEATINVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "SILA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "TYRRHENVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "MENELAVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MALIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VERATIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "GYTHEVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "NEAPOLITANVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "ZELA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ATLANTEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CARRVCA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LYGDAMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SAEVINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AREOPAGVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DEMODOCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NIGRINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MAROBODVVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VELABRA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ALTINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HIPPONOVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PHILOPOLEMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LAVINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VENILIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "VFENTINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LICYMNIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MARONEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SIDICINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PRIMIGENIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AMATHVSIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "BASSARICVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["b9694"]}},
    "LEVCADIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LYCTIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LYNCEVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "TORQVATA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ELATEA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "VENTIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "THRIASIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "NICEPHORIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "ERGASILVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SAPPHICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "GALLICANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PHOENISSVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "THIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ASTYPALAEA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "VENAFRVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "DAVCIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PASSIENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CABERA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TROVCILLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "IAPYDIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "DRVSIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "STYMPHALIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SERAPA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CREMONA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "TONGILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PACARIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LYCINNA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SATTIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TELESINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SACROVIRIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CERAVNVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SCAEVINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ADAMASTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "COCYTIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CARFINIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PIPLEIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "HYMETTIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "BIBACVLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SEQVANA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SVBVRANVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "EPPIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ERETVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "SVEVVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PAELIGNVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "SVFFICIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CORYCIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "HYRTACVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LATONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CAMERIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "GORTYNIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "SICYONIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LYCAONIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "BAREA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PAGASAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CHARIOVALDA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MOLORCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VELINVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "AVTOBVLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PANDARVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CATHEGVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TERMESTINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PYRRHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NERITVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ILIACVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["i9165"]}},
    "OPHIVCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ERYMANTHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARTEMONA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ALPHESIBOEA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PAMPHILA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "GALERIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARVLENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TERTVLLINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OLYMPISCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HESIONA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "DAEDALEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ARCHESTRATA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "GARVMNA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MATINVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "MNESTHEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MATVTA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PHTHIOTICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CATIENA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "MIMNERMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SVILLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "COMMAGENVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "AVENIENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "ANTONIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ARPINVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "PVNICANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SVRRENTINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TERENTIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ANCONITANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VLIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "THAPSACVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MAENIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "VACVNALIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "AENEIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "EVRYLOCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ASSYRIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "TMOLIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TEGEAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VELABRVM": {"N": {"morpho": ["n-s---nn2-"], "uri": ["u0221"]}},
    "RVFRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "COERANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CLEVSTRATA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PHILOCTETA": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "NABALIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CHRISTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PILVMNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["p2073"]}},
    "SANQVINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ANDANIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "HERBESVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CRISPINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARCTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["97666"]}},
    "TYRTAEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CARYSTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VERVLANA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MAVORTIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LESBOVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ICARIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "BOLA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "LVCANIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CANTIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "HIBERINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PERIDIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "EPHORVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["e0699"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["e0699"]},
    },
    "PROTHOVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "THEONINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CABVRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DICAEARCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PETVSIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NYCTELIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "ANTAEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TEGEA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "NYMPHIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ATTIANVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "IAERA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "POSIDONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BISTONIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "PALILIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "THYNVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PHOCEVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "DALMATICVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["106786"]}},
    "AVGVSTALIS": {"A": {"morpho": ["aps---cn3i"], "uri": ["97722"]}},
    "ARENACVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "ORCHOMENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LEDA": {"N": {"morpho": ["n-s---fn1-"], "uri": ["l0473"]}},
    "ARGVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "NAXIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CORNVTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["c4063"]}},
    "OLYMPIACVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "AMPHILOCHIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SAETABVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "GABINIANVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "ARPINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SEXTILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "REGINVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "EVRYALVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PESSINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SVEBIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "DALMATIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PISISTRATVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PICENVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "ATINA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "DINDYMVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "PICENTIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "THESEVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "NORENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "DYMNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BATAVODVRVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "FALERNVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "ICARIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": ["98791"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "PETELIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CATVLLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "RVBRENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SARMENTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PROXVMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VITIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CRESSIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ACADEMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NERONEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PATAREVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "NVMISIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DVELLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OPTIMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ZENOBIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PARILIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "OPPIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": ["99735"]},
    },
    "BVLLATIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SILLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OLLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CANVTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MAVSOLEVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["m0942"]}},
    "CLASSICIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LATINIENSIS": {
        "A": {"morpho": ["aps---cn3i"], "uri": []},
        "N": {"morpho": ["n-s----n--"], "uri": []},
    },
    "FRONTINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GAETVLVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ASPARAGIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "AMATHVSIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MVRCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["m1756"]}},
    "LYDIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CVTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OLEARVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NEARCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NASAMONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TYPHOIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "BISTONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PLOCAMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CEPHALVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["104587"]}},
    "PHORCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SCANTIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ANAXARCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LVCAGVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TIMONEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "POSTVMIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "BOVIANVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "ROVCILLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "FOLIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TVCCIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CAECVBVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "DANAEIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CRASSICIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "APONIANA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "HERCVLEIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BRENNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "FVRFANIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ADONEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BRVNDISINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CENTONARIVS": {"1": {"morpho": ["n-s---mn2-"], "uri": ["c4048"]}},
    "TEVTHRANTIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MOPSOPIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PLATENIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CALIGVLA": {"N": {"morpho": ["n-s---mn1-"], "uri": []}},
    "ENIPEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PERMESSVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "THESSALA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ACMONENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "STHENEBOEA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "TIRESIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SCYTHICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SCANTIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "BOCVLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DIOMEDEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PETRINVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "ATIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ILVA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "DOSSENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["108191"]}},
    "SQVILLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TARSVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CVRETIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "GORTYNIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ACCA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CECROPIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "BELLIENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TECMESSA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PHERECLEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SAMOTHRACIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "MAENALIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "PAPHIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LVCETIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DEXITHEA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MEMPHITIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "PHARIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "ISTHMIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SVBRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MEGALESIACVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SALIENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ALBESIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "XENOPHILVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DAEDALA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ALBIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ACVTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MANNICIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MAECILIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "BYZANTIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VAGELLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ALLOBROGICVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CORNICVLANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ARISBA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PROCILLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TYDEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BATTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MEDVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": ["m0581"]},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "LEONIDA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MARDVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "AEQVICVLVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PEDANIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AEGIMVRVS": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ACTIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "SERGIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CRETAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PADVSA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "GORGOBINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ENNIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TEMESAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CASSIVELLAVNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PVPINIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "NARBONENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "APPENNINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VIRDOMARVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CONGVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CAESELLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OSINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SCRIBONIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DIONYSVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "RVMINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "THEMISTA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "HYMENAEVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["h0718"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["h0718"]},
    },
    "NASIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CAICVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DVLICHIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ECBOLIMA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "AEGEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "THEODAMANTEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "OPITERGIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "VESPASIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AVIENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PISTORIENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": ["p2175"]}},
    "VALENTINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "THRAECIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "IONIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "FADVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VERVLANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VESVVIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LABERIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CLEOMBROTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TVLLINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ALINDIENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "DANVVIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CASTICVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AVRVNCVLEIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ADRANA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CALPVRNIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CERVARIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ZETA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "GALBIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ITHACVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": ["99154"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "ARPINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ACHAEMENIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PANGAEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AEGIENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "CLYTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "THREISSA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "NORICA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "COVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "FALISCVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["f0121"]}},
    "HYPSAEA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ALLIARIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ANICETVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MORITASGVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "FONTEIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "OLELVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GRATILLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "COROEBVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CANTHARA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "HADRIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PORRIMA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "EREBEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PARNETHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NICOMEDENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "LEONNATVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PSEVDOLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "COPTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OFELLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PARENTALIS": {"A": {"morpho": ["aps---cn3i"], "uri": ["p0383"]}},
    "HIPPONENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "BELLOVACVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "DREPANVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "ACRISIONEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "RVTVBA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LIBITINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LYSIDICVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ASCONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PHILOMELA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "VESCVLARIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SOLYMVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TARQVITVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AMPHIONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MARGIANA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "DVMNACVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PYGMAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SOZA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CAESENNIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MOSELLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PONTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CISSEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PHOEBEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SATICVLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "APENNINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VELLEIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ANEMVRIENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "NEPTVNIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "EGNATIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SCISSA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TEVTONICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ISMARA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "AMVLLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LVCERIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SELINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LIBERALIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MALLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TROXOBORVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CORFINIENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "TMARIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "HIRRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MINTVRNENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "FABATVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "FLACCILLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PANTHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BACCHEIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ANDRIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LAMYRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AVARICENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "AERIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "IDRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MILICHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DIVODVRVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "CAPHARIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "CORYBANTIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "FIMBRIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LEMNIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SELEVCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARGOLICVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "ALBINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "POLYPLVSIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "AMATHVS": {"N": {"morpho": ["n-s---mn3i"], "uri": []}},
    "SPEVSIPPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PIERIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CORVNCANIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "QVINTILIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CLVVIDIENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LEVCIPPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "POMPEIANVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "AGRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["101212"]}},
    "FVLVIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "DORIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "THRASYLLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DEIANIRA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SABRINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "BAIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ARTORIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TRALLIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MAVRICVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NERATIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BERYTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "POPVLIANA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TORANIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CAMPANICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LVGIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CHAERESTRATVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TANFANA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CATILVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DODONAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "DIVITIACVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MARSVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "BLOSSIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "THERMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ALBANA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "GALATEA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "TABRACA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LAODICEA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SYRACOSIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SYMAETHIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "AEACIDINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PROCILLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ANICIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "AQVITANVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "CIMBERIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MOLORCHAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CISPIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ADGANDESTRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PARDALISCA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CARPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PLANCINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "BERECYNTIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CETHEGVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "STROBILVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["s2618"]}},
    "DINDYMIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "ILLVRGAVONENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "SOHAEMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "STROPHIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "FVNDANVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": ["f1166"]},
    },
    "MEGARA": {
        "N": {"morpho": ["n-s---fn1-"], "uri": []},
        "O": {"morpho": ["n-s---fn1-"], "uri": []},
    },
    "AEOLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LIBVRNICVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["l9908"]}},
    "MEGALOBVLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ALLIENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "THARSIMACHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NONACRINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "AGYIEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "RIPHAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MEDOBRIGA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CAESARIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["c9835"]}},
    "MACRINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BELIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VOLVSENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SAFINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HIBERIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PAPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ILVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GNOSIS": {
        "A": {"morpho": ["aps---cn3i"], "uri": []},
        "N": {"morpho": ["n-s----n--"], "uri": []},
    },
    "AGYLLINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VERGILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "RAETIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ELAPTONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LASERPICIARIVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["l0328"]}},
    "SABAEA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "RVTVPINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PISONIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SAXA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "THALIARCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AGRIPPINA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "TVCCIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "APONIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PAXAEA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PATROCLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BROMIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CARTEINENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "CINGA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "BALBINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AEGISTHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ELISSA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PERRHAEBIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "OEOBAZVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SAGVNTINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ATHAMANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MELANIPPA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "STYMPHALIS": {
        "A": {"morpho": ["aps---cn3i"], "uri": []},
        "N": {"morpho": ["n-s----n--"], "uri": []},
    },
    "IVNIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "VENAFRANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SYCHAEVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "HOMERICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ALSIENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "ILLYRICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "APHOBETVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CAVCASEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ONESICRITVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ACADEMIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "VALERIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ARTABAZVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VNOMAMMIVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["u1133"]}},
    "POLYCLETVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HIBERNIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "FLAVINIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PENELOPEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CVRENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "TERETINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "COMANA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "HYACINTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AMISIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CALLIMACHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VISTILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CYNTHIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "PEVCOLAVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SOTERICVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "FVNISVLANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BITHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AMINNEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PHENEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "FIDENA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MYRTILVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SVRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["s3833"]}},
    "ASVDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ANAGNINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CADMEIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CYDONIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": ["47617"]}},
    "VALGIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LAIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SYRVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": ["s4014"]},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "LYCIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "NOVARIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MINCIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PACIDEIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "THESSALONICA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "AFRICANA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MASSICVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "PYTHAGOREVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ASIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": ["97698"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "CARMANIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MANTVA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "GARVTIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ITVRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BRIAREVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BRVCTERVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TITIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TREBELLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PACVVIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BVTRA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CHOBVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VINDONISSA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MENNIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AEGIALVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ACESTA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "DYRRACHINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LYCA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PAPIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "RHESVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ORFITVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CISALPINVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["c1950"]}},
    "GRANICVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "COSA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "RHOETEVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "VATICANVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "FIRMANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CADMEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CHARAXVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "FERENTINVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "MINICIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SISOCOSTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TREBELLENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GAVRANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PLANCIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "FRISIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "APERENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "EVMOLPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NAVPLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["n0111"]}},
    "FADIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ASCLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SITHONIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "LICINIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PIRITHOVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARRIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TVRSELIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CORMA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SOGDIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VIRGILIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VOLCACIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NAXVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MNESILOCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CINYPHIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ANTEIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PACONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ACHERVSIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VINICIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MARVLLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ENCOLPIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PICENVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "POENINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "BRIXIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "OSCVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MVLVIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "PALPELLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NOVATVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SYBILLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "REBILVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "FVRVNCVLVS": {"1": {"morpho": ["n-s---mn2-"], "uri": ["f1274"]}},
    "CARPOPHORVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARIADNEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TVRVLLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SESTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AEOLIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TVRRANIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OTACILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OGYGIVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["99699"]}},
    "ELIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SETINVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "VXELLODVNVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "RHADAMANTHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BYRSA": {"N": {"morpho": ["n-s---fn1-"], "uri": ["b0789"]}},
    "HAEMONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PANVRGVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ATELLANA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "EXAERAMBVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SAFINIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SVESSANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "APINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NVMITORIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PARADA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PERIMEDEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CHRYSALVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SIBYRTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SOSIBIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SEQVANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ITHACENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "TIMAVVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CIRRHAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "STALAGMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ENNOSIGAEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["e0608"]}},
    "PARTHIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "COSCONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SCIRONIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "APPIANVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "AGONALIS": {"A": {"morpho": ["aps---cn3i"], "uri": ["97565"]}},
    "COA": {"N": {"morpho": ["n-s---fn1-"], "uri": ["c2173"]}},
    "VICETIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ATHENA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PATALIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "AVENTICVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "CALVENTIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "TEANVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "BRIXIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PHALANTHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CYPRIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MYTILENAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "IVDA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CODRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TITYRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MOLOSSVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "BRISAEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DODONA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "FVRNIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ATTICINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AVTRONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ROSIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BALBVTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VRBINIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "AVERNALIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "ALBANVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "SYRIACVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CINYRA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PINARVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GAZABA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "INOVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PAETINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CHAONIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MARIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TANAGRAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PARTHENIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "ERYTHRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CONCANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "BESSICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "APPVLEIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MAMERCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BARBIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MAVRVSIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MILONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ACTVMERVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TIBERINIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "CAYCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MALEA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ISIACVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CENTAVREVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VORANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MVTVSCA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PRVSIADA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "APVLEIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OLENNIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ALPINVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "GORDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARCETIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LVSITANVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "ALSVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ANIENVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["97616"]}},
    "PHLEGRAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ALBINGAVNVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "GALVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "COTVATVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARVERNVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "AETHALVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["a1045"]}},
    "TANAGRA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "DINDYMA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "IVGVRTHINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VOTIENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SEMPRONIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PVLFENIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "STRYMONIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "STLOGA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PLANCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["42458"]}},
    "HYDAEPA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "HERACLITVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ACRATVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "REMVRIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "DEMOLEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARCHIPPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SARDOVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LEVCOPETRA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CAMPANEVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["c0331"]}},
    "CYBELVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NABATHAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SCETANIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CATIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PYTHAGORICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CHOERILVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PVLLIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CERINTHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["c1238"]}},
    "TYNDARVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PAEONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["p0061"]}},
    "RHAEBVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PELOPEIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "POLYTIMETVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LEBEDVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CETRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARVSEIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NIGIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CHAONIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "GERGOVIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MOSA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CENTAVROMACHIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "LVCANIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": ["l1114"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "AMANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LERNA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "EVRYTHMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MARMARICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MATTIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "HELLESPONTIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CORINTHIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SELEVCEA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "DECIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AEGALEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ANCONA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PITVANIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GALLONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SYBYLLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ANTEMNA": {"N": {"morpho": ["n-s---fn1-"], "uri": ["a2151"]}},
    "RAEDARIVS": {"1": {"morpho": ["n-s---mn2-"], "uri": ["30607"]}},
    "EVBOEA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ACIDALIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "HYRCANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "IDALIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ASPAVIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PHOCAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "HERCVLANEVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "FIDVLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TVSCILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ANA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CLVSIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PERGAMA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "AEQVIMAELIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "APENNINICOLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SORA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "IOLAVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "RADAMISTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "INSTEIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TANVSIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VEHILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MACAREVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "COMPITALIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["54067"]},
        "N": {"morpho": ["n-s----n2-"], "uri": []},
    },
    "CAESERNIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PANCHAIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CARSIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "EVNOMIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MEMMIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HECVBA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "FERENTVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "CLVENTIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PYRRHA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "COLCHIS": {
        "N": {"morpho": ["n-s---fn3-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "HYRCANIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "METLOSEDVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "PANOPEA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "IDALIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "HEGELOCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PELOPONNESIACVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ALBVNEA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "IASONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VISELLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CRODVNVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "QVINTILIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PERGAMENVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "DOMITIANVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "MARACANDA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "AECVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AVRVNCA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "EPPIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PARRHASIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "DEMIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARRETINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TIBVRNVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "AEMILIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "FORNACALIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CLINIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "NVMIDVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LAVINIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VRGVLANIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PHRIXVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VIBIDIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LVPIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ALBIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "BACTRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TOLOSA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "TAMPILVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AMPHRYSVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GLOBVLVS": {
        "-": {"morpho": ["n-s---mn2-"], "uri": ["g0388"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["g0388"]},
    },
    "FIDICVLANIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MESOPOTAMIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "QVINTA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ANIENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "BACCHIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["b9691"]}},
    "ELEGEIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SIDONICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LABIENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "THERODAMANTEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LAPPIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BODOTRIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ROMANIENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "CAECILIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ARIONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "THRONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DAVNIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "BALBILLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "STHENELVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MINOVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TREVIRICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PHINEIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SESCENTOPLAGVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "QVERQVETVLANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VECELLINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TARRACONENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "COCCEIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AEFVLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CARNVTIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "SALLVSTIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PODALIRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "INDVTIOMARVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SANGARIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MOENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LIPARAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "INTERAMNA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CAESENNIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ARTEMITA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "LVCANVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": ["l1141"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["116538"]},
    },
    "ALEXANDREA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CASTALIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ERYTHRAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["e9961"]}},
    "RVFRIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "THERMOPOLIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": ["t0584"]}},
    "SIRPICVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["100622"]}},
    "ERECHTHEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "APOLLONIDENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "TEIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ILLYRIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CYNTHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MVSONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PACTOLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AQVILEIENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "SVSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "TROICA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "HELLESPONTIACVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TIGELLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "RHEGINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CISPIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TERENTIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SICANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "IARBA": {"N": {"morpho": ["n-s---mn1-"], "uri": []}},
    "COMVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "PHARMACVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["p1869"]}},
    "ANDRONICVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NELEIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ABYDVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "RIGODVLVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "METHYMNA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PACIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GYARVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BOVILLANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CERAVNIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "IONICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "GARGONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NICAEA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CENCHREVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CYLLENA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "IVDAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["37896"]}},
    "HEGESIMACHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TYCHIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SENTICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CELADVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TARPA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "HELVINVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["112198"]}},
    "BABYLONIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ZEVGMA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "VESVLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VRSIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PLAETORIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "HAEMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "RHENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "XENIPPA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "VENALICIVS": {"1": {"morpho": ["n-s---mn2-"], "uri": ["u0256"]}},
    "AMASTRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HISTRIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MACRINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LICINIANVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "COENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ASINARIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "VALERIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "IGVIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "HELENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CVRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SICELIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "ALEIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PENTHEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AHALA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ERILLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PASSENNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CENABENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "DIOXENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "POSTVERTA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CISRHENANVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["c1956"]}},
    "SANGA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "NORTIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "THEOPHILVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ELECTRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["98304"]}},
    "BITHYNVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VENETIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PALLADIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "RHINOCOLVRA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PERSEVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "SCYTHA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PROTESILAVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BABA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "DIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "GRATIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OROANDICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PAPHLAGONA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PRISCINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BELGICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TREBVLANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "STERTINIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "DECETIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LYDVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "PALILIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "BIRRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PHRYGIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "AMEANA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LYGDVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "FANNIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ERIGONEIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "COPONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VIRIATVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "EVNAEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARGOVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SARRANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "FALCIDIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ABASCANTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ANDROGEONEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "HAMILLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LENAEVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "HERMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CAEPASIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DARDANVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "VARIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ARCADIVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["97661"]}},
    "BETILIENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HIRPINVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "CATVVOLCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DEIDAMIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SACRANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CLOACINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "RIPAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "RVFA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "VIBIENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MAZAEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CIRRHA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "TAVROMENVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "PERRHAEBVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MEGADORVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VIBIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "EPEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PHIDIACVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ECBATANA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PETRAEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MEMORATVS": {"1": {"morpho": ["n-s---mn4-"], "uri": ["m9791"]}},
    "SASERNA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CRVSTVMINVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "DORIS": {
        "N": {"morpho": ["n-s---fn3-"], "uri": []},
        "A": {"morpho": ["n-s---fn3-"], "uri": []},
    },
    "VERANIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ICELVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MAENALA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CLITARCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GERYONACEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ALCESIMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SILIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PARNASSIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PICENTINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "BONNA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MENOPHILA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MONOECVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PISTOCLERVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NAVTACA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MAMVRRA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CONCONNETODVMNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ANDROMACHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VOLTINIENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "ITIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CLEONICVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SITTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VARILLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CNEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "APOLLINEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "DEIPHOBVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HERMOBIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "FREGELLANVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "PHAEACIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ITYLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CVPENCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ORBIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VARGVNTEIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "COSANVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "CRISPINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CYBELEIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PYRAGRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TAXIMAGVLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ASSACANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SESTIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PVTEOLANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "GEMELLINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GLAVCIPPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ASIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "DELLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VOLOGAESVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TRICCA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "LOLLIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CEA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TIBVRTINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "HISPANIENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "EDESSA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "TREBVLA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PACTVMEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SYMAETHEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SYMPOSIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "AQVILEIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ACHERVNTIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MARICCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OXVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CALVINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MYRTOVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TELANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MOSCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HOMERISTA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LYRNESIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PLATAGIDORVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "IGILIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "SCYRIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "DORYLENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "TAVNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CRATINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SERVAEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HYLLEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PLISTHENIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LETHAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SVEBICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ORICIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CAECINA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "THRASEA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ERICHTHONIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": ["101261"]},
    },
    "PLACENTIA": {"N": {"morpho": ["n-s---fn1-"], "uri": ["p2217"]}},
    "SAMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VERONENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "VICANVS": {"1": {"morpho": ["n-s---mn2-"], "uri": ["55357"]}},
    "CAMVRTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NVMANTINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "IRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "METANIRA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "DEIOTARVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HAGNA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PEGASVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AQVITANIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "VIENNA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "HYDROCHOEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DESVLTORIVS": {"1": {"morpho": ["n-s---mn2-"], "uri": ["98148"]}},
    "OBVLTRONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NASVA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "GLAPHYRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TAYGETVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TRIPHALLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["t1339"]}},
    "ALCAEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PETRONIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ORCYNIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PETRIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "FORMIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "THREICIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CAELIANVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "TROIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "EVROPAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VATIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PELIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SEGOVIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "FALCIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "RVFILLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SYRTICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "NASICA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "LEVCTRICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ORITHYIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "BRATVSPANTIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "CHIONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TENTYRA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PYRRICHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GELOVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "THESPROTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OETAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CAEPARIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["c0541"]}},
    "AVARICVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "ORICVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "THYNIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "GALERIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MEGALOPOLITANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "AMPIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AVFIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "THYREATIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "ARIONVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PANDIONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LIBENTIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CORVINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CENTAVRICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MAMVRRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GNOSIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "GRYPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["g0007"]}},
    "MYCONVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ANTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MVLVCHA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PTOLOMAEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ACIDINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ALPINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SANTONICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ASINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "AXONA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TIGELLINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TETRARCHIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["t0492"]},
        "N": {"morpho": ["n-s----n2-"], "uri": []},
    },
    "NOTARIVS": {"1": {"morpho": ["n-s---mn2-"], "uri": ["n0541"]}},
    "PVSILLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "NVMERIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MAEOTIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PERILAVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HELVETIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MOPSOPIVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["56049"]}},
    "VENVTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HYLAEVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "CAMVLODVNVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "LAELIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SVLLANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "HALOTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SALONINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "INTIMILIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "CNIDVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ROSIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OLENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["101325"]}},
    "BOVDICCA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "DORYCLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GRVMENTVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "HYBLAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CYLLENIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "OEBALIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "HIPPOLYTA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ARSAMOSATA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MAECIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LONGARENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LONGINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CARA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "STHENELEIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MASSICA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "GRYLLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["g0634"]}},
    "DORVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "VISIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "IONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "RVSTIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HERCEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PROCHYTA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "AVSONIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ELYSIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "SEGIMVNDVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "EPOREDIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CORANVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "PANNONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "AMITERNVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LAODAMIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CAYSTRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LEPIDA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "NYSIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NOVAESIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "THEODOTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DIADEMATVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LENTIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "FISCELLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["f9567"]}},
    "IPHIANASSA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ARCTVRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["a2749"]}},
    "BRVTTIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "COTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LIMETANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PELVSIACVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "AXIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PTERELA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CHAERIBVLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "COBIAMACHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TAENARVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GENVSVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BLESAMIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LYBIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "AVIOLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CARYSTEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MISENVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "ACARNANIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PAVLINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CYRSILVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OCTAVIANIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "QVINCTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CLVSTVMINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ARPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DILLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "EPIGONVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PERGAMEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "STATIRA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ARARICVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HERMOLAVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AMBRACIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "MENTOREVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CRESSVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ISMARIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "HORTALVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ISMENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AVGVSTODVNVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "PALLANTIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "AEDVVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MVTINA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "EVRYTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SVADA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "IVLLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["101312"]}},
    "VIRIOLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "AVFIDIENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DORYPHORVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["d2196"]}},
    "CAESONINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LYAEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ACONTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CALENVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "SASSIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MESSALINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PETREVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CELEMNA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "AELIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "TOLBIACVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "VETILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CAVSINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PINDARVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CORINTHIENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "PHYLLEIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "CASTA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "IOLCIACVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LIGVSTICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LABIENIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ANTINOVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "RVSTICVLVS": {"1": {"morpho": ["n-s---mn2-"], "uri": ["r0942"]}},
    "PHARSALIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MENOECEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CINYRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OECONOMICVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HARMODIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SVNIVM": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "ITVRAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "IONIACVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["99129"]}},
    "AGANIPPAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "NEOPTOLEMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CEPHALIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "IVVENTIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": ["99189"]},
    },
    "DONATIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BOSPORVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LATONIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CHALCIDICVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["c1337"]}},
    "GANGETICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CANINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "POMPONIANVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "TANTALICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LABERIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HELIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BOLANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AVRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MYGDONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CAECVBVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "PRIVERNVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "CABALLINVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["c0006"]}},
    "ALTINVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "CARENVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "GAIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "THRASIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "EVRYMACHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CASINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SENIENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "BOIHAEMVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "PAEGNIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "ARGEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "XANTHIPPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VELANIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CAECVLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HERCVLEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "AMPHRYSIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CARFVLENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CATAMITVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "-": {"morpho": ["n-s----n2-"], "uri": []},
    },
    "AMPELVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CORYNAEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LACONICVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["l9068"]}},
    "ALBVCIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LAOMEDONTEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "AQVICVLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CALYDONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "POPVLONIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "MATRALIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "IOLLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PHILIPPENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "PHILODEMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "EVDOXVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GRYNEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "HESPERVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BERENICEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LACETANIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "OPHIONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ANCYRA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CREMONENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "ATABVLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["a3297"]}},
    "NICEPHORIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CYTHERA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ARTACANA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ANCHEMOLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SVLPICIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TYRRHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MENVLLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ISAVRICVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BARIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "METELLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PHAEDRIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ANTISTIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ARMILLATVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["a2881"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "ANDROMEDA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PISEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VARISIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LINGONVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARNIENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "AGRIPPINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MEGALESIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "EQVIRRIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LANARIVS": {
        "1": {"morpho": ["n-s---mn2-"], "uri": ["l0186"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": ["l0186"]},
    },
    "GRATTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TOXEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARECTEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MVTINENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "OEAGRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VSTICA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CALVISIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ORFIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CEDROSIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PTOLEMAEEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TITANIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "EPIDAVRIVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["101257"]}},
    "NIPHAEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "FABRICIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MEGYLLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "EXECESTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TICIDA": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "SATVREIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CANIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VOLANERIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LARIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CHARVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MESSAPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PANTILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PHLEGRA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "VACCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TALLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HERCVLIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PROPONTIACVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CORFINIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "PLANASIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CALVSIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VITELLIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ATRATINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MAEANDRIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "EPIDICVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CARSEOLANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MEDEIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "OCRICVLANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "RESTITVTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PADVA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "VIRDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BRVTTEDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ANIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AVNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PHRIXEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CLAVDIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ENNIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "QVINTIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CRATERVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ATIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "APVLVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ASELLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ILIPA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "BARBATIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GERELLANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ASTVRICVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "IMBRASVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DVLICHIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ATTIVS": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CADVRCVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["c0069"]}},
    "CIMINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TABVRNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VIBENNA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "HAEMONIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PHTHIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CRAGVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MVLLINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CRESIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VERBIGENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["u0364"]}},
    "PANDA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ONOMASTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "THEOTIMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OPIMIANVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "BOXVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AEACVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "EVRYPYLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DOMNILAVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ANCAEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "FAESVLANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "HYPERMESTRA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CARTEIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ELEAZARVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PHIGALESIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SIRIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": ["s8782"]},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "CEREALIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "RHIPEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LESBIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "PHARSALIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ASTERIA": {"N": {"morpho": ["n-s---fn1-"], "uri": ["a3223"]}},
    "VARENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PROPERTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SICCA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ATREVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ACHARNEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VOLCANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["u9904"]}},
    "EVCAERVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PAEONIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ROMVLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ACILIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SOLIMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PRAETEXTATA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PLAVTINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "NESSEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "GRAECINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DORICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "DIRCAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "FALERNVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "BOEOTIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "GAVILLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PERILLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BESSA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ACRISIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "XANTHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MAREOTICVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "MINOIS": {
        "N": {"morpho": ["n-s---fn3-"], "uri": []},
        "A": {"morpho": ["n-s---fn3-"], "uri": []},
    },
    "BVPALVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SIGEVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "NOVIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "VOLTINIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "RVTVLVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "POSTVMINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CARCHEDONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "GRATIDIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PAESTVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "HAEDVVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ARGOLIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "SILENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SICANIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SEMVRIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "PHAEACIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "POLYXENIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "THASIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TRINACRIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SCANTINIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TEDIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SIGEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TVRPILIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BANTINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "IVDAICVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["36185"]}},
    "BALVENTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MAENALIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "GARAMANTIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "IAVOLENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["39690"]}},
    "PHAEACVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ACHELOVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LITAVICCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SASSVRA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "THIRMIDA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "EPICHARNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DANDARICA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MAGNESSVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PHARSALICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SELEVCENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "FIRMIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "THYSDRA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ASCYLTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PISAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SPARTANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ASCLEPIODOTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TROICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MASSIVA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SERVIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VEIANIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ASCIBVRGIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "ERVCIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VOLANDVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "SAVREA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ACROCERAVNIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ATLANTICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "INACHIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MENELAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SVBVRA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ANNA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "POLYXENA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "BOEOTVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CVRICTA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CORSICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TIBVRTINVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "METIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ASCVRVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "CILLA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SORRENTVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "ERASINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ORBILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VCVBENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "FVLGINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARABIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PHYLLEIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "THYMBRAEVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": ["t0640"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "CLEOPHANTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARRETIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "CHAEREA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "NYSAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ARRHIDAEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VENETICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PHAEDROMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MESSALINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TIMAEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PERINTHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GEDVSANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SEXTIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "NEREVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "RAVIDVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AEOLIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PHILODORVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VIRIDOMARVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PHOSPHORVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CAMERTINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "EVHENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MANDVBRACIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NOCTILVCA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ERYCINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PELOPEIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "PANGAEA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CLODIANVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "IMBRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BASILICVS": {
        "2": {"morpho": ["aps---mn1-"], "uri": ["b0157"]},
        "1": {"morpho": ["n-s---mn2-"], "uri": ["b0157"]},
    },
    "CRVCIARIVS": {"1": {"morpho": ["n-s---mn2-"], "uri": ["c4425"]}},
    "NICOSTRATVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PLOTIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "NVRSIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "THERSIPPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CARISTIA": {"N": {"morpho": ["n-s---fn2-"], "uri": ["49765"]}},
    "SITHONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ROSCIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "ALESIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CELTIBERIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SPVRINNA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MATTIACVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VELIENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "FVLCINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARCHILOCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "EVGANEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SMYRNAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ASCVLVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "ARMENIVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "QVINTILLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ARISTONVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SATVRNALICIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PELIACVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TIBVRTIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "ARCHEMORVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GADITANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PROTEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TITHONVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "FOROIVLIENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "PANCHAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "POMPVLENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ALCMAEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CALPETVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MOGONTIACVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "CILICENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "ZETHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CASPERIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "BRVTTIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "APVLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SESTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VERTISCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "FABVLLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PLATONICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VADA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MVRRANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ASOPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VERGINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NESTOREVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "FVFIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "POLYBVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SALTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VERRITVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ATRIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TRAPEZVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "THERSILOCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PAVCVVIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TREBIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AMASTRIACVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PAPIRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MEDIOLANVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "VELIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BROCHVBELVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CAELICVLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ISTHMIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LARENTALIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PROCVLEIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NVCERIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SINCAIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MINICIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AVZEA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "THYESTVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ALBVRNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["a1310"]}},
    "METAVRVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "SATVRNIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "VENERARIVS": {
        "2": {"morpho": ["aps---mn1-"], "uri": ["u0288"]},
        "1": {"morpho": ["n-s----n2-"], "uri": []},
    },
    "PRODICVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CATIENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HORDEONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TRAVLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LAPIDARIVS": {
        "2": {"morpho": ["aps---mn1-"], "uri": ["l0255"]},
        "1": {"morpho": ["n-s---mn2-"], "uri": ["l0255"]},
    },
    "PARNASVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CALVINA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "LABICANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MVCIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ANTEGVA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SCRIBONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CANVSIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "DIODOTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VRANIA": {"N": {"morpho": ["n-s---fn1-"], "uri": ["101033"]}},
    "DELTA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "HADRIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CEBENNA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MEVIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DECIDIANA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "AMPHIARAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VETTIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CAESONIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ATHAMANIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "MAXIMILLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ISELASTICVS": {
        "2": {"morpho": ["aps---mn1-"], "uri": ["i2928"]},
        "1": {"morpho": ["n-s----n2-"], "uri": []},
    },
    "GELA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CERCIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "IDA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "POPPAEANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ERYTHEA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "MANLIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "AFRANIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DEXAMENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CASTABALVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CITA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CAMERINVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "VARINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CLOANTHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "APPVLEIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TEVTOBVRGIENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "BONONIENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "ALPHESIBOEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BENACVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "RAVENNA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "VOLATERRANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ITALICENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "FABRATERIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CLYTAEMNESTRA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "COCYTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HIEROSOLYMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ELEVSIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "VETTONIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "POPPAEVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "PIRAEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PHINEVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "CAVLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BRVTTIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "POLYSTRATVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "THOANTEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "GLITIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LVCINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ABGARVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SCYTHVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PYRAMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VORENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SARDANAPALLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AQVINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HELICONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "NONIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OCCIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "VINICIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CECROPIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ALLIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ILERDA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "RVDINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "BALEARIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "TERTVLLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LYRNESVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SERESTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SEXTILIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PHILETAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ASTRAEA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "NINNIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ATRACIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LATVM": {"N": {"morpho": ["n-s---nn2-"], "uri": ["l0392"]}},
    "AESOPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DICTYNNA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ICARVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AVSONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CALLISTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VERANIOLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SERVILIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "NAVPORTVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "ITONVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CVMAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "BATHYLLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CYRRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CLVTORIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CHRYSANTHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["c1529"]}},
    "MASSVGRADA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TARTESIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LARISAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "AMERIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "OLENIVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["99704"]}},
    "FIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ABELLA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PONTIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "NERVA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SINOPENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "AMAZONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ECHIONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ZANCLAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ASTRAEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AFRANIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "POTITIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "THRAECIDICA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "VESTINVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "MELITENA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "APPIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CASSANDREVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ATEIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VIPSTANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PHASMA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "DACVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TITANIS": {
        "A": {"morpho": ["aps---cn3i"], "uri": []},
        "N": {"morpho": ["n-s----n--"], "uri": []},
    },
    "ALPHEVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "CARMELVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PALAESTINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TVRONVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SEPTIMILLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MAIA": {"N": {"morpho": ["n-s---fn1-"], "uri": ["116859"]}},
    "OPPIANICVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TELEMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DERCENNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CVRTISIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BEDRIACENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "ARRECINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NATTA": {"N": {"morpho": ["n-s---mn1-"], "uri": []}},
    "NASTA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PINDARICVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["50214"]}},
    "AQVILONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["a2614"]}},
    "PRILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BOCCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PALLACINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LEMONIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MAVRA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "FVFIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MISENENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "XENOPHANTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PROETVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CVSVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PARSAGADA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CAPANEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SVEDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LEARCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GRACCHANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ARCHIDEMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TVNNIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TOLENVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "AVFIDVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DICTAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "BRIGANTICVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BELLICA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LAENIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MOPSVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CLATERNA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ALERNVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "MENSARIVS": {"1": {"morpho": ["n-s---mn2-"], "uri": ["m0736"]}},
    "CATVLINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ABDVCILLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VATINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["133196", "131037"]}},
    "PITTHEIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SYRIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "GARGILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BARDAICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VSSETA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "VIPSANIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "ANTANDRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ISAEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GAVRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AMYCLAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "AREMORICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "EPICVREVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PACONIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HEGESARETVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ANTICYRA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "BOSPHORVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DORYLAVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PADAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TVLLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CASPERIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AEETAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ABDALONYMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CANTABRICVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["c0438"]}},
    "AELIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "QVARINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SALEIVS": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SIPYLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SAVFEIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ARGYRIPA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "AVXIMVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "PHAEDRA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SAMOTHRACIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "STOICIDA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SILVANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["36617"]}},
    "CASSANDRA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SELEVCIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "THEAETETVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OSTORIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "THALLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["t0533"]}},
    "VOLSINIENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "PENTHEIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SARMATIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CALGACVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ATELLANICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MAVRVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SALMONEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MELANTHIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PHANAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TEIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "VISTILIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "VOLOGESVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "IXIONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PELOPIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MASSYLVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "BETVVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ALLIENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "CIMBRICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CIMMERIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "RHOETEVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "CORDALVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ERICTHONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LEMONVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "FERENTIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "ARCHIDAMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LAGVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PHOENISSA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TRIVICVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "APHIDNA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ISMARVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TRITONIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LIBVRNVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["l9906"]}},
    "IDOMENEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PHASIACVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CAVARINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "FLORALIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PALLANTEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "NAEVA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "DEMAENETVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SAVFEIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MANCIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "LEBINTHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "THVBVRSICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "GENAVA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "NVCVLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ARETHVSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "PERENNA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TALAIONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PARMENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "TITYVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CILICIVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["c1606"]}},
    "CORYTHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CELEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SERIPHIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ARACINTHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MARPESIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "HELORIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "EVDEMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SILIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "DINAEA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SPARTACVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OVIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OEBALIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ISAVRVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SAGANA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "VETTIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "IPHIGENIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PELOPEA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ZMINTHEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ENNA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MEGAERA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "MYTILENA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "VALETIACVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MASVRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NOVIODVNVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "ILIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PAPHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PANTOLABVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CYDONEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "BVRRIENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PLVTONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LATONIGENA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CORSICA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CRETENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "ADVATVCA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ARMINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARIVSIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "RHODOPEIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "THRACA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SVSIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "IBERICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "AMATA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "VIBIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SARIOLENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "THRAESSVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CHIMAERA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "BALACRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TVBILVSTRIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "EVBOICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "BOVMELVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AEOLIS": {
        "A": {"morpho": ["aps---cn3i"], "uri": []},
        "N": {"morpho": ["n-s---fn3-"], "uri": []},
    },
    "PADVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["p0029"]}},
    "DAVARA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "IVDAEA": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "TELEMACHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "EVTRAPELVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MAMMAEA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CAVCASVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "APOLLINARIS": {
        "N": {"morpho": ["n-s---fn3-"], "uri": []},
        "A": {"morpho": ["n-s---fn3-"], "uri": []},
    },
    "MYRINA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "STILONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TRIGEMINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "VOCETIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VVLCANIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "BASSVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CHAERONEA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SATELLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AEAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "GAETVLICVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "RVFILLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NORICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ANYTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LIGVRIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CANNVTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARGVETIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BASSAREVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TEVCRIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SEMELEIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SERRANA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MARDONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CAESAREVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["53300"]}},
    "ALBINOVANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CLODIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CAMERINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "FIBVLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "IDISTAVISVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "APIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CALEDONIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CAESAREA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CLEARETA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ADRAMYTTEVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "CERIALIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MITHRIDATICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SERRANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CORELLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GNOSIACVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TARVLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CAMERIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NARYCIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PARVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["49086"]}},
    "ISEVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "DECIMIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PELASGVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "LVSIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "EXTITIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ATIMETVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CLEANTHEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "BELICA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "DIGENTIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "NOVIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PACALIS": {"A": {"morpho": ["aps---cn3i"], "uri": ["p0012"]}},
    "POSTVMIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ORONTEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "AVFILENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AMMIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CYTHEREVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PHILOXENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ISINDIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VARIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["u0108"]}},
    "SVRENA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SINVESSANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "RAMNVSIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MONTANIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ROMANICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "RVBELLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MOLOSSICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CAVDINVS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "RHODVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SCRIBONIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "NEMAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PAGASVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TVDITANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "IDALIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "PERSIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OCTODVRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ALBRVNA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MENEDEMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "EVENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["101267"]}},
    "ARRIONILLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PALICVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GABBA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CLONVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ERSILAVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SVEBA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LACAENVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MATIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARIADNA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CAESILIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ACADIRA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "DORYLAEVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "PARNASIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "IPHICLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GELONVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LISCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARDA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ALCINOVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CAESARIENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "MELANTHEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SOSIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ORSILOCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "THESPIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PATRVITVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MENARCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TARPEIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SABINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "BONNENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "SVBLAQVEVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "VALENTIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "OLYNTHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SAMAROBRIVA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SCIRTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GALLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ADRAMYTTENVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "HADRVMETVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "TVRTA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CERCYONEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CATONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MARTINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "POEANTIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LAMPADIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": ["l0178"]}},
    "MEGARICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PHILVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BACTRIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TEREVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CICIRRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MAREOTIS": {
        "A": {"morpho": ["aps---cn3i"], "uri": []},
        "N": {"morpho": ["n-s----n--"], "uri": []},
    },
    "DIOXIPPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LIBYSTINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VIBILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "EPASNACTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "EVMELVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LEMNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AVIDIENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OENOTHEA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MISENA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PITTACVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PVBLICIVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "RVTILA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "AHENOBARBVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ALLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AMPHILOCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PERSEIS": {
        "A": {"morpho": ["aps---cn3i"], "uri": []},
        "N": {"morpho": ["n-s----n--"], "uri": []},
    },
    "DEIOPEA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "HERMVNDVRVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "BAETICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "AVIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ODRYSIVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["99692"]}},
    "PHLYVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PIRENIDVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CLVVIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "NAMMEIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CYNTHIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "FONTEIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ANQVILLARIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "THALA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TRACHALVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SVRRENTVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "VMBRENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ADMETVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DOMITIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "HIRTVLEIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PALFVRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ALEXA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "COGIDVMNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARTEMISIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DANVBIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MARCODVRVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "EVTYCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ENCELADVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CNIDIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LAVRENTINVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "PINTHIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MYSVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "NONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AGONIA": {
        "-": {"morpho": ["n-p---nn2-"], "uri": ["102077"]},
        "N": {"morpho": ["n-s----n2-"], "uri": []},
    },
    "CASILINVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "NESSVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MYCENA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CAPRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ANTONINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NVMANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "RVFINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LYCVRGVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PETROSIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ANTHEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CELTILLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PERSIS": {
        "A": {"morpho": ["aps---cn3i"], "uri": []},
        "N": {"morpho": ["n-s----n--"], "uri": []},
    },
    "SEPLASIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "BODVOGNATVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OLYMPIA": {
        "O": {"morpho": ["n-p---nn2-"], "uri": ["101327"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": []},
    },
    "PHAEDRVS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "TENCTERVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "AVRIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PORSENNA": {"N": {"morpho": ["n-s---mn1-"], "uri": []}},
    "TANAVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LIBYSTIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "LVPERCALIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "IONIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "LVTETIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "NOMENTANVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "MANILIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CORINNA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ARTORIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ERIGYIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PASIPPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MAECENATIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CAMILLIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "NERONIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PANDVSA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "OPSIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VERVCLOETIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ASVVIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TAVRIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CENABVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "CRETHEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CRABRA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MASSILIENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "PERELLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OCRESIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CINGONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SALVIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CRESSA": {
        "N": {"morpho": ["n-s---fn2-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "SAMNIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "THYESTEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "HELVIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HADRIACVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PERVSIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "DEMOPHILVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "INDICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SVFFENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "POLLITTA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ALTHAEA": {"N": {"morpho": ["n-s---fn1-"], "uri": ["a1500"]}},
    "AVGVRINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ELECTRA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "GALERIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "EVCHADIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "LINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ANDECVMBORIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CASCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["c0734"]}},
    "TROIVGENA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TASGETIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "THERMAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LARA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SARDVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "EVHIVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "GALESVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PLEMVRIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "TECTA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PHILARGYRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AMYCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BACTRIANA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ASCVLANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CVSINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SAMIARIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "IVNONALIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "ORMENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ERIDANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "FORNACALIS": {"A": {"morpho": ["aps---cn3i"], "uri": ["f0866"]}},
    "BEBRYCIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ARISTONICVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARMENIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "TORIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TAGVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GALLOGRAECVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VMMIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GLYCERA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "AMYTHAONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MEVANIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PHILAENIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "DIABOLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["d1217"]}},
    "GOMPHENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "CASCELLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AETIDEMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MAEVIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PHILIPPICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "GRAPTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VERGILIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "GELDVBA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "FIRMIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CIRCEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "NAVA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TYRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BINGIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "PANNONIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "LAOMEDONTIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CESTILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GORDIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "ALCMENA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ANTILOCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PACIDEIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HIPPOLYTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DVLICHIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "APIDANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PHOEBIGENA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CANVSINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "GVLVSA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CILNIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "EVXINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TIGRANOCERTA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "APELLEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CAVCILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "THYBRINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "XVTHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PHEGEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SARMATA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PHRYXIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "OEAGRIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CALLIPHANA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "APVLIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "GAIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TRIVIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CAELIMONTANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TIBERIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SAMBVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GEMINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ALCMAEONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PETILLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MAMVRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "IDVMAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PENEIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VZITTA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "AVLERCVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MEDIOLANENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "CLITVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TIMARCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SINGILIENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "ITALIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "MELLA": {"N": {"morpho": ["n-s---fn1-"], "uri": ["m0642"]}},
    "PEREDIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "NVMMVLARIVS": {"1": {"morpho": ["n-s---mn2-"], "uri": ["n0665"]}},
    "VIBENNIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "INTERAMNIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "ACASTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ALLEDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "INACHIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SICANIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TITHONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "BRITANNICVS": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "THERMODONTIACVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ABELLANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "NVMIDA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "TALASSIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PANCHARIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "AMANTIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CYCNEIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "HYLLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "STAPHYLA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ACHERONTICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "OENEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["99693"]}},
    "EMATHIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "HOSTILIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MENESTHEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CINCINNATVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["c1636"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "AGENOREVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "THYONEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "RHODIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CLEOMACHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CYTHNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "REMVLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["r0390"]}},
    "HADRIATICVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "SEPTIMINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PROMOLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BAVIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LATAGVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CALPVRNIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MINOIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CYCLOPIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CATVALDA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "BOIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VACVNA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "DOMNOTAVRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MVTILIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PETTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ERYLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LEARCHEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "THESSALIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "RAETICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "EPONA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LERNAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SYGAMBRA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "COSSVTIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TRIARIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LIVINEIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DOSSENNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LVSCINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["116668"]}},
    "NERVIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "HARPYIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "NERINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SABAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PHALAECEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LYRNESIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "TALTHYBIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TARSA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "BEIRA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "HELLANICVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CASINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "FLAMINIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "STIMVLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MELETEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "AELIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ARISTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SCAPTENSVLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PRASVTAGVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AEGAEVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "PERPENNA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PERVSINVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "SALONA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "FVLCINIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "BELGIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "MAEONIS": {
        "A": {"morpho": ["aps---cn3i"], "uri": []},
        "N": {"morpho": ["n-s----n--"], "uri": []},
    },
    "MEMPHITICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CASTVLONENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "POPPAEA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ARBELA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PARIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "IVRA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CAEDICVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LANVVINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "BATVLVM": {"N": {"morpho": ["n-s---nn2-"], "uri": ["b9695"]}},
    "HILAIRA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CORDVBENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "OILEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["99700"]}},
    "TARBELLVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "POLYDORVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MONA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PHILIPPA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PELOPEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "COLCHICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TMOLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "POLLENTIA": {
        "-": {"morpho": ["n-s---fn1-"], "uri": ["p9840"]},
        "N": {"morpho": ["n-s---fn1-"], "uri": ["p9840"]},
    },
    "DIONAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "HALAESVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VARILLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "VIPSANIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "AEGINA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "MOESICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LYCAMBEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TROGVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AENOBARBVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MANCINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TITVRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VENVLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PACIDEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MEVVLANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BROMIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "HARPALYCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CENTVMALVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CHRYSIPPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MYCENAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PACORVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MARRVVIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CAETRONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AVRELIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CALAVIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TETRICA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "AVSONIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "OCRICVLVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "BRIXELLVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "PIRAEA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ADIMANTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "APPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OVIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OEDIPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HALIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ELEVSINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PENEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PYLADVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NORICVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "LVSITANIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PENELEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AGRIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "AGRIPPINENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "MELINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TARQVITIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CALLAICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "HADRIACVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "HIEROCAESARIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "VIBVLLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LEVCADIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SANCIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PHRVGIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MONYCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SATRIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ATEGVA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ISSA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CENAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TRIMERVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NOMENTANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CARBONARIVS": {"1": {"morpho": ["n-s---mn2-"], "uri": ["c0600"]}},
    "PERBIBESIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "HALVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["112084"]}},
    "SIPONTINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CAETRA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PELLAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VELLOCATVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NVMANTINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MAGNETIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "AMYCLANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PHILYREIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "FLORVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CVRIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DVROCORTORVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "ARADVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "APELLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LAVERNA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "OTHONIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "HYRCANIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MEDVLLINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SAGVNTVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "ISMENIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "VLLEXODVNVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "ASCLEPIODORVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BAETVRIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ESQVILIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LATINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CARNA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LEVCOTHEA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "LYSANIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TERRASIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SANGVNNIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "POLYAENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "IVTVRNA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "GRAIVGENA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ATHENODORVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARTABANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ELEVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["98313"]}},
    "SACRATA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PRAENESTINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ANCHARIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ARNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["a2894"]}},
    "GYLIPPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ASCRAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LIGEA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ALBVCILLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LENTICVLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "BELLEROPHONTEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "NICOBVLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SIDONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "AVFILENA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "POTITVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VESAEVVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AEGIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "CARSVLANVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "IOLCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CVRIONIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PHOTINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ALLIFANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CYTORVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SCYRVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HIPPODAMIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CVPIDINEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "EPAPHRODITVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["98347"]}},
    "TVTILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PHILOSITVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "RHODANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AMENANVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["97595"]}},
    "PLAVTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AMOEBEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["97601"]}},
    "CALLIDROMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "INTERAMNANVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["i2260"]}},
    "TEGEEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SATVRNIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "STROLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ERECTHEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "HERCYNIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CAEDICIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "BARCANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CADRA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LISSVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SOSIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PARAETONIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": ["p0309"]}},
    "RECTINA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TITHONIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "BITTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AGESILAVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VALERVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARABIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CLANIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LACINIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "POENIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ACHELOIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CYMAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "AGAMEMNONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "QVIRINIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VILLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "IOCASTA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "THEODOSIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BEROEA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CAIETA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "VMMIDIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TVBERTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SETINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MARSYA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "LVCVLLIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ITALICA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ARVIRAGVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARCHIBVLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SATVRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AVILLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CIRCAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LATMIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VAGA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "EDONVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["98284"]}},
    "POSTVMINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AFINIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AVTOLYCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DINOMACHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VELLEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TIGVRINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ANTIOPA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SIGEIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "NVMIDIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SERGIOLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ORTYGIA": {"N": {"morpho": ["n-s---fn1-"], "uri": ["119978"]}},
    "BVCCILLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CALESTRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LENTINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LVPERCVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "BVCEPHALA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "INGVIOMERVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LYCISCA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ASSARACVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "EREBVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MELAMPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ALABANDA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CORYPHAEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["c4178"]}},
    "LATOVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LAELIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ETHYMANTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CYDONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PELORVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "RABIRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PENNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "EPHYRAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "APRONIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SVLPICIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ARTAXATA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "AGEDINCVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "PROCVLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "GALAESVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CASPIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MENANDREVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VASIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SVRRENTINVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "EBVSVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MISENVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "ANCHISEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "POMPEIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CEPARIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": ["34648"]}},
    "CAMILLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PACCIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CYRNEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "MEDVSAEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "AMASENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GAVIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CRVCISALVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TVRPILIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PEDIATIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TISIDIVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "DIDYMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TAVRICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PYRRIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "NEMEA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "BIVIVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["b0343"]},
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
    },
    "THESSALICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CRITOBVLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PINDVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TRITONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "IVLEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PACIAECVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NAEVOLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CAPENVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "COSMVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "APHIDNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MVNDENSIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "THALIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "BELGICA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "BANDVSIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "BLITIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NEAEREA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "SABINVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "PVPIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MAGETOBRIGA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "PISA": {"N": {"morpho": ["n-s---fn1-"], "uri": ["p2138"]}},
    "VESCINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VRBINVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "MNASYLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MAZACA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ANTIMACHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TREBATIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NOLANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "IPSITHILLA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ARRVNTIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CARANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "EGNATVLEIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "NICOMACHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TVLLIANVM": {"N": {"morpho": ["n-s---nn2-"], "uri": ["56777"]}},
    "BASILVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PORTVNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CORSVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "TVRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "THERAPONTIGONVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ASICIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "RHIDAGNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GYARA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "THEOPOMPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SVSA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CLVENTIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "VERONA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CYLLENEVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "SVSAGVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VEIENTANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "IBERIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ACERRONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "EVPHOLIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PELEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PETRONIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ATTA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SIPHNVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "LARISA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MALACA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "AQVINVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "SARMATICVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LESBIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ABDVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SILENA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TIBVLLVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "PYTHODORVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ANTIOCHINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PANTVLEIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MYLASA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "THRESSVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ARABICVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["a2627"]}},
    "SENIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "GNATIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ACONTEVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "INACHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VATINIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CAVCASIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PLEVRATVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CARIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ARDVENNA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "TEVTOMATVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SMINTHEVS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ASSYRIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "LIGARIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "STRATVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GNOSIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "NOLA": {"N": {"morpho": ["n-s---fn1-"], "uri": ["n9763", "n0488"]}},
    "CANOPVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TELEPHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "TICINVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "AMPHIARAVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "EPIDAVRVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "VACALVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CYMODOCEA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "BESSVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GARGANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "OCRICVLANVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "SOSIANVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CENSORINVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DECIDIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ATTIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "AVRIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "MACHAONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "AMERINVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ARGIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "VMBRICIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CLVVIENVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ARISTARCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MALLOVENDVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MVNDA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "ARCHIACVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "CRETIS": {"A": {"morpho": ["aps---cn3i"], "uri": []}},
    "LYDIA": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ARGANTHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "MAENALVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "AONIVS": {"A": {"morpho": ["aps---mn1-"], "uri": ["a3729"]}},
    "LEMVRIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "CELTAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "NVMIDAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "FORMIAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "SVIONAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "VOLCAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "ARSACIDAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "VERGILIAE": {"N": {"morpho": ["n-p---fn1-"], "uri": ["u9901"]}},
    "CYRENAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "SAVROMATAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "RVFRAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "PHARITAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "ARVAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "LAPITHAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "SALONAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "NERSAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "CARSVLAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "CLVVIAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "CYANEAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "AENEADAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "SYBARITAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "STABIAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "AEGEAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "ORNYTVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CARINAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "DAHAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "BORYSTHENIDAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "GORTVAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "NAPAEAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "BOVILLAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "ANTENORIDAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "IVGVLAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "MINYAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "MYCENAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "VLVBRAE": {"N": {"morpho": ["n-p---fn1-"], "uri": ["101012"]}},
    "BISALTAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "PHASELITAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "BASTARNAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "LYNCESTAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "CELAENAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "CIETAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "AVSONIDAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "PARAPAMISADAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "SARMATAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "BRANCHIDAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "ORESTAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "LIBVRNAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "FAESVLAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "BELITAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "THESIDAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "VOLATERRAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "TEGEATAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "MAZAGAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "EVMOLPIDAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "BELGAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "ARGINVSAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "OXIONAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "OCHVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CLVPEAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "TELEBOAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "THERMOPYLAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "CLEONAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "DANAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "ARCHELAVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SVDRACAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "PLATAEAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "PYLAE": {"N": {"morpho": ["n-p---fn1-"], "uri": ["p4640"]}},
    "SABARCAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "CANNAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "SAMARITAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "ODRYSAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "PISAVRVM": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "EVERGETAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "PIRVSTAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "AEGINETAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "TYRRHIDAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "MINTVRNAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "PHOENISSAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "SCAEAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "GORNEAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "DALMATAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "OEBALIDAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "PETREIANVS": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "ARGONAVTAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "THEBAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "HELVECONAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "HERACLEOTAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "CECROPIDAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "ANTEMNAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "TABAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "DENSELETAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "SABINAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "ALOIDAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "BAPTAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "GRAVISCAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "MYLAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "CERCETAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "CRANAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "ALAVDAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "MASSAGETAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "LVRIVS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "GEMONIAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "ACERRAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "HORITAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "ARABITAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "CENTVMCELLAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "CALYDNAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "GANGARIDAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "PISAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "APHIDNAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "COELALETAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "TYNDARIDAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "VERCELLAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "DRANGAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "ROMVLIDAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "BIAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "CANOPITAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "SACAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "GALATAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "CASPIAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "CAPREAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "CLAZOMENAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "ONCHAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "PASIPHAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "PISIDAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "MAEONIDAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "AMYCLAE": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "DANDARIDAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "DANAIDAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "PHOCAEI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TRIBALLI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SYCAMBRI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "VBII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TERMESTINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ALPINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "HIBERI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "FAVNI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ARGOLICI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TIANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CHAMAVI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ARECOMICI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "HERMVNDVRI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "BRITANNI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "POMPEII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "DII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "AVSCI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "COMMAGENI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "THVRII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MELDI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SEDOCHEZI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CALETI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "EPICVREI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "GRAI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "HAMMONII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SAPAEI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "AGATHYRSI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SABELLI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PARTHIENI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CINITHII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "VXII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "NOLANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "HELVII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "BRVNDISINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ADVATICI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MALLI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TAVRINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "OSISMI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MATTIACI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TVRONI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SEGVSIAVI": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "AVLERCI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SVLCITANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TVLINGI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "VELLAVII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SITHONII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "DARDANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PERGAMENI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PRAESTI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "DECANGI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MARSACI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CARALITANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TVNGRI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "BOTTIAEI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TEVTONI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MAGI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "RIPAEI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PHILADELPHENI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ACYLLITANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PARTHINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "AFRANIANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CADVRCI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SVEVI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "GRAECVLI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ARACHOSII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CHASVARII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "VSIPI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MARSI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MARRVCINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TELEBOI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "VATINIANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SANTONI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "THERAEI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "GELONI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MAGYNI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CASSI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "REATINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MENAPII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "FABRATERNI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CHALDEI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TERGESTINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MAVRI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MARSIGNI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CENOMANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PTIANII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TREVIRI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ARIMASPI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MOSCHI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "LEMOVII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "POMETII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SYRII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MEGARICI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PYRRHONEI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ELEVTETI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ANGRIVARII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PRAESTVTIANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MANII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "LARISAEI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ASSYRII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "VINDELICI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SARDIANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "VELIOCASSI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "LANGOBARDI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "HAEDVI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ALEXANDRINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "GEIDVMNI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ARVERNI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "AEGYPTII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CHALCEDONII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "GALLOGRAECI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ILII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MEDIOMATRICI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "HESPERII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "DACI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "AVSETANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PELASGI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "GORDYAEI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "COLOPHONII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PLATONICI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PETELINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "VERAGRI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "AMANTINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SIRACI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CAMPANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ITALI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "OPICI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MARDI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MOSSYNI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "VOLATERRANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SABAEI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "NVMANTINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "BOEOTI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "LIBVRNI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ITVRAEI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "FENNI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "FIRMANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "OLYMPENI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CONDRVSI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CIRCEII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CLAVDIOPOLITANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TIGVRINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ESSVBII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "COLCHI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ANNICERII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SALAPINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ALLOBROGICI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CEDROSII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SVEBI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "OSI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MORINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "BOII": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "PANNONII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "BAETICI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "BESSI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CVGERNI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "LVSITANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "AFRI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "GABALI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SEDVNI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CORSI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "LEVCOSYRI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "LANVVINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "INDI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MOSTENI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CAEROSI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PARENTANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SOLONI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "IVDAEI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SOLYMI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ANAGNINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "NABATAEI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CERAVNII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SVSIANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SICANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SOGDIANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TYRII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "VRII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ACHAICI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "LEVCI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "LVGII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "REMI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "IBERI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "LEPTITANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CHATTI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CIMBRI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "GARVMNI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "THESVNTI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "IADESTINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SOLI": {"N": {"morpho": ["n-p---fn3-"], "uri": []}},
    "PECOLI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "FRENTRANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "APAMENI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SARDI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "HYPAEPENI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SIMVI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TEMNII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SIBI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "AEDVI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "NERVII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CALAGVRRITANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SALENTINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CENIMAGNI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "IVLIOPOLITANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "GAETVLI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "HELISII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "QVADI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CAEMANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "AMBARRI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PVTEOLANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "BRVTTII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SEDVSII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PHILIPPI": {"N": {"morpho": ["n-p---mn2-"], "uri": ["121740"]}},
    "AMBIVARETI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "LIBVI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CHAVCI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SALAMINII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "AMBIBARII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MACROBII": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "SEQVANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "RVGII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "LATOBRIGI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MOESICI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "RHOXOLANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ATII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "DOMITIANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MYNDII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "OSCI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TENCTHERI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "VENETHI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "BELLOVACI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MVSICANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "GOMPHI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CYRENAICI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CERCINITANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PLEVMOXII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "STYMPHALIDAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "TAVRI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PYTHAGOREI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "VOLSINII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "FERENTANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ADIABENI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "VNELLI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "LEXOVII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TAPVRI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "LEPIDANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "GALILAEI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "IDAEI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SVNVCI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "DELIACI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CRVSTVMERI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "GABINIANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ANGLII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ARGIVI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "IVLIANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "LOCRI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CHERVSCI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CASSIANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "AETOLI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "AVSONII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "GOLGI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "GAMBRIVII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TRANSRHENANI": {"N": {"morpho": ["n-p---mn2-"], "uri": ["t1088"]}},
    "COSSAEI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "HYRCANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "AMERINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PANHORMITANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TYRRHENI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "POMEIANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ADELPHI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "COTINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "IGVINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MEDI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MANIMI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TVSCVLANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CAESARIANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "VANDILII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "VARINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SMYRNAEI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ICENI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "GEMELLI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PANNONICI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TENII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TAPSITANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CIMMERII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "HELVETII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "AFRICANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SOSII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CELTIBERI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CARMANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "AVERNI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "VOCONTII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CHORASMII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ANDECAVI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "HELLVSII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SIDICINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "EOI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TRANSPADANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "BRVCTERI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MASSYLI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "NARISTI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MEMACENI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "POMPEI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PEVCINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TARBELLI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "DANAI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MYTILENAEI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CAVNII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "GADITANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PAELIGNI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CVMANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "BREVNI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CADVSII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MARCOMANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PHILAENI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MAEDI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "LAODICENI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PATAVINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "HADRVMETINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PRASII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "AQVITANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TRALLIANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "HALICARNASSII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "DYRRACHINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MESSENII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "VIROMANDVI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SEBAGNINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "AESTII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "AMPSIVARII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "BABYLONII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "AORSI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "LVCANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SYNEPHEBI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ELYMAEI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "EDONI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "BAETASII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "HYPERBOREI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SEGNI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "BARCAEI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "HENIOCHI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ARGI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "EPIGONI": {"N": {"morpho": ["n-p---mn2-"], "uri": ["e1958"]}},
    "VMBRI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ASCVLANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "OMBI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "AMASTRIANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "LOTOPHAGI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ARAVISCI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PELOPONNESII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PHOCII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "VACCAEI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "RAETI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ALANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "THYSDRITANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MEGALOPOLITANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SEGONTIACI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ERETRICI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "HENETI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ARMENII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ABII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CARSEOLI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "FRISII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PYRGI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "GRVDII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "FVNDI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "DVLICHII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "RVBI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "NAHANARVALI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "RVTENI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TRIBOCI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ARRETINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MANDVBII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "AMISENI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MVSVLAMII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MYSI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "FOSI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "FALISCI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "BITHYNI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "BIBROCI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ACHAEI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "VICETINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "REVDIGNI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PLACENTINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SPARTANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PANICI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MOESI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TENCTERI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "POMPEIANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MINYI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TEVCRI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "BAVLI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ATTICI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CASPII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "LEVACI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "HARII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TVRDETANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CHII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "THESSALI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ELESIODVLI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "AMBILIATI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "METHYMNAEI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "RAVRACI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "OCTAVIANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "AVGVSTIANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ALBICI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "GERMANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TROGODYTAE": {"N": {"morpho": ["n-p---mn1-"], "uri": []}},
    "NEAPOLITANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "NORICI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "AGRIANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "BOSPORANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "TARTESSII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "FLORENTINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ILLYRII": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "ORDOVICES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "LAERTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ACESINES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "HEDYMELES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "MITHRES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "NICARCHIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "EPIPHANES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CINYRADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CORIOSOLITES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PRAEXASPES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "AMPHICTYONES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "ARSACIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "TERIOLTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "LEMNIADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ONOMASTORIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "MENETES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CALES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "NARNIENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "TANTALIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "LADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "COCOSATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "THAPSITANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MYRMIDONES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SOPHITES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "THESSALONICENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "CIRTENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "SVLMONENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "CAMERTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CARMONENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "GOTONES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ACMONENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "HERACLIDES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "DEMOCRATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PONTICI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "VRSAONENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "PHERETIADES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "METHYMNIADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "MEDATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PHTHIOTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ATRIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "AONIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PARRACES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "IXIONIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ANTICLES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "DRYADES": {"N": {"morpho": ["n-p---fn3-"], "uri": []}},
    "GARAMANTES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "DROPIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ATTALENSES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "ZAMENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "ORSINES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PYRRHIADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "LYCVRGIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ANCHISIADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "INTERAMNATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ARII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ERINDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "EVNONES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "MINTVRNENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "LEPTINES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "POLYBOETES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "DRANCES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ARSAMES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "EVANTHES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "LAOMEDONTIADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "BRONTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "NAMNETES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "OILIADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "NAVSIPHANES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "HYDASPES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "ERINNYES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "EBVRONES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SAMNITES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "GATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "IACETANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SOTIATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "AEOLIDES": {"N": {"morpho": ["n-s---mn1-"], "uri": []}},
    "HIPPOMENES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "SAVROMATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ARPI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ORODES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "PISCES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "MERIONES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "BORYSTHENES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "OAXES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "MALIENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "PHRADATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CATVRIGES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "AENIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "HELICONIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "AESCHINES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "LELEGES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "PLEIADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ECHECRATIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ASBYTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CORDVBENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "BVLIDENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "TAVRINITES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "TREVERI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ARTAVASDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SABACES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "HAMADRYADES": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "TYDIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "RHOSACES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PARMENIDES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "AMBRACIOTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "MEHERDATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "NAVTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ILIENSES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "IGVVINATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "GOMPHENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "BATAVI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ORESTES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "LACIADES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "DRYOPES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SPITAMENES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "HERMOGENES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "THYESTES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "SINNACES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "REDONES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CATANES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "MOYSES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "VELIOCASSES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "BARZAENTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "AGRIPPINENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "PANTHOIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "HARVDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ACHAEMENIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "TYRIOTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "IORDANES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "AEGIDES": {"N": {"morpho": ["n-s---mn1-"], "uri": []}},
    "PRVSENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "PENEIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "THEOPHANES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "ARCHIGENES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "APOLLONIDENSES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "VOCATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CHOASPES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "CEPHALLENES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "TROADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "STROPHADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ASTASPES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "AQVINATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "BRVTTIANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ARGOLIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PROETIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "OENIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ZORSINES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "IAZYGES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "HESPERIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CHARICLES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "TRINOBANTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "THESIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "NASAMONES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "APOECIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "IMBRASIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "TROIADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "LAERTIADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "BELIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SELEVCENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "CROTOPIADES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ABISARES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CROTONIATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "THABENENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "SITALCES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CVRETES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "ACARNANES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "BERONES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "INDIGETES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ARAXES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "STEROPES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PITTHIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PHAETHONTIADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "LEOPREPIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "TMOLITES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SOPHOCLES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "AMYNTIADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ACOETES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ELVSATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "MINDYRIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "NEALCES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "GYNDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ORNOSPADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "OXYARTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "BITVRIGES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "NAVCRATES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "ATARNITES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "EVMEDES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "HARMERIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SYRTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "EVRYMIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "TIMOCLES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ORIGINES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "MITHRENES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "BATTIADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ALYATTES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "ALETRINATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "TARVSATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PIERIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CEVTRONES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "MENOETIADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "BOOTES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "SALMACIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "AGYLLINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ASTACIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "NAVPLIADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "DATAPHERNES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "LIBETHRIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "DORYLENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "GERYONES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "MEDOBRIGENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "NORENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "ANTISTHENES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "HYSTASPES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "PETROCORII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CHREMES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "VONONES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SEVTHES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PHRAATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "STOECHADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "EPIMENES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "MASSILIENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "ADRYADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "RHEOMITHRES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CARES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CHALYBES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "CATAMANTALOES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "NIPHATES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "VSPENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "XERXES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "EVPHRATES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "PEVCESTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ELYSII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PHAEACES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "GENAVNI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CYCLOPES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "EVDOSES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "VTICENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "TEMNITES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SEMNONES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "IAPYDES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "PEGASIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PARMENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "IASIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "TROES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "MAZACES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "BRIGANTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CAMBYSES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "MASSAGETES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "LVPERCI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "THEODOROMEDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "IOANNES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PERIPHANES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "MONAESES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ARIMAZES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ONITES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "HIPPOCRATES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "PISTORENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "ACHILLES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "THALES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "INSVBRES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CEPHENES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "IPHICRATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "AMEDINES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "GRAIOCELI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "OTHRYADES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ACHATES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "PICTONES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PHILIPPENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "BVRI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "LIBONES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ATIZYES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "HISPELLATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "TERPSICHORES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PENTHELIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ACESTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SANTONES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SISYPHIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "DITAMENES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "VELIENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "VIENNENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "PHYLACIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PANES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "VELLIOCASSES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "NICOCLES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "PLISTHENES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "ANTIPHATES": {"N": {"morpho": ["n-s---mn1-"], "uri": []}},
    "ATINATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "AETHIOPES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ANDROMENES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "EBVROVICES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SVGAMBRI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "HERODES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "SARDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "AVXIMATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "NICOMEDENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "SALLVES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "DEMOSTHENES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "CLINIADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "OZINES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "AMYNTORIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "TIRIDATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "THVCYDIDES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "HYRTACIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ABORIGINES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "APOLLONIATES": {"N": {"morpho": ["n-s---mn1-"], "uri": []}},
    "HARPOCRATES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "POTNIADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "VSIPETES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SVARDONES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SORTIENTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SIDONII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ACHAEMENES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PHILYRIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "FOROIVLIENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "RHOEMETALCES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "HELICONIADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "EMERITENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "ARGYRASPIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "NITIOBROGES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "TVBANTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "HISPALENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "ETEOCLES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "ANARTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PETRAITES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "APHRODISIENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "CAPSENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "SERES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "TARRACINENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "AEGEATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ACMONIDES": {"N": {"morpho": ["n-s---mn1-"], "uri": []}},
    "SARRASTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "BASILIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PHINIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ERATOSTHENES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ANTORES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "LEVCIPPIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "DARDANIDES": {
        "N": {"morpho": ["n-s---mn1g"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "GRINNES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CRETES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "AREOPAGITES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "AGGRAMMES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CENABENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "ORCADES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "CHALDAEI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PARTHYAEI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "HERMES": {"N": {"morpho": ["n-p---fn1-"], "uri": []}},
    "ARETES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "TECTOSAGES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "BORESTI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "MITHRACENES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "APOLLONIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "MENOETES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ATREBATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "LYSIADES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "BAGISTANES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PHILOCTETES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "NVITONES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "GYGES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "TEVTONES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "MYSTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ARISTOPHONTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CORONIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ANTIGENES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "GANGARIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "NICETES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PALES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "STRATOCLES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "ELEI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ALETES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "THYMOETES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "BARGATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ABANTIADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "IDARNES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CARTEIENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "CALLICRATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SVIONES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SISENES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "LYNCESTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "DEMADES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "PHRYGES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "LEMOVICES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "AMBIANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "ASPRENATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CALETES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "DARES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PRIAMIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PALAMEDES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "THIMODES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "VASACES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CAPPADOCES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SICCENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "ORONTES": {"N": {"morpho": ["n-s---mn3i"], "uri": []}},
    "MEMMIADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ABVLITES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ERICHAETES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "HAVSTANES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "TRINOVANTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "BAGOPHANES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CREMONENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "CARNVTES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "HELIADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "MILTIADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SATROPATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "AVLESTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "LAESTRYGONES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SISIMITHRES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SEGESTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PYTHOCLES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "BVRSAVONENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "ARTAXERXES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "TAXILES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SINDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ARISTOMEDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "BYLLIDENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "IMMORTALES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ABDAGAESES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PHARNACES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "MYGDONIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "EPIDAPHNES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "BACTRIANI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "BERECYNTIADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "EPIRENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "VETTONES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "LYCAMBES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "STRATIPPOCLES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "OECLIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CYRENENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "AGRIANES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "NABATHES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SATIBARZANES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "VARDANES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PAPHLAGONES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CATAONES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "LEVCATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ARIARATHES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "NEREIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "RHODIENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "LARES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "POLEMOCRATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ISOCRATES": {"N": {"morpho": ["n-s---mn3i"], "uri": []}},
    "AMPHIAREIADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "IONES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SAMOTHRACES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PHILLYRIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CARALES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ORONTOBATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ZETES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "MIMALLONIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "LABEATES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "TIGRANES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ALCIBIADES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "NVCERINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "HYPERIDES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "MVNDENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "STRATONICENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "ZALACES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "AMPHITRYONIADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "TIBVRTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ABISTAMENES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "NICAEENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "ANCALITES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ILIADES": {
        "N": {"morpho": ["n-s---fn3-"], "uri": []},
        "O": {"morpho": ["n-s---fn3-"], "uri": []},
    },
    "MELICERTES": {"N": {"morpho": ["n-s---mn1-"], "uri": []}},
    "SINOPENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "HYADES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "LACONES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "NABARZANES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CYPRII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "OLYMPIONICES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "ERICES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "AEETES": {"N": {"morpho": ["n-s---mn1-"], "uri": []}},
    "PANOPES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "RAMESSES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "EVMENIDES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "PICENTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PELOPONNENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "DEMOCHARES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "CANNINEFATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SITHONES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "AVGVSTALES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SCIPIADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "EMPEDOCLES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "GOBARES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CHARES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "DIORES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "BVTES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "TIMAGENES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SYMPLEGADES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "APHELIOTES": {"N": {"morpho": ["n-s---mn1g"], "uri": ["a2338"]}},
    "GANGES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "PARTHI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PERIPLANES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "NEMETES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PAMMENES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "POLIORCETES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "LVGDVNENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "MAEONIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CALLICRATIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PROTOGENES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "COBARES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "DRAPPES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "OXYDATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "LINGONES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "LARINATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "TYRES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ELELEIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "HOMONADENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "AGAMEMNONIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ARISTOPHANES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "ACTORIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "EVMENES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "HYPSIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "DIABLINTES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "ANTIPHANES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ALTINATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "NANTVATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "THEMISTOCLES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "OEENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "CICONES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "THYESTIADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ISTAEVONES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PHRATAPHERNES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "OPHELTES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "SVESSIONES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "LYCONIDES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "CLEANTHES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "LYDI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "SOCRATICI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "IZATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "FICEDVLENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "ARSACES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "APAMENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "VERONENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "ITALICENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "ILLVRGAVONENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "INACHIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "DISCORIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ARISTOMENES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "TITIES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "AVIONES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ATLANTIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ANDES": {"N": {"morpho": ["n-p---mn3i"], "uri": []}},
    "ARABES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "DRYANTIADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "MARTIALES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "DOLOPES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "PAEONES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "VACCENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "SENIENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "ORSES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ARCHIDEMIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ATHAMANES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "BISTONES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "GAVDIMONIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["g2293"]}},
    "VNISVBSELLIVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "PIPVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["52911"]}},
    "LVMBIFRAGIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l1171"]}},
    "PVGNVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["100237"]}},
    "PROPYLAEVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "PRAEDIOLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p2934"]}},
    "CONSTRATVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c3619"]}},
    "PERMAGNVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["99941"]}},
    "PVLPAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p4476"]}},
    "HARPASTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "HAEMONIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CLOSTELLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["21023"]}},
    "PATARACINVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "AVRICHALCVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["34584"]}},
    "IDOLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["29125"]}},
    "GVRGVSTIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["g0660"]}},
    "TRIOBOLVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "EPYTIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "AVTOMATVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a3619"]}},
    "PERPAVXILLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p1877"]}},
    "BARBARVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["b0121"]}},
    "BICLINIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["b0320"]}},
    "STIBADIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s2461"]}},
    "LYCOMEDES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "CORYLETVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c4170"]}},
    "PARRHESIASTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SILVRES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "DVLGVBINI": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "CONTIGNATVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "PORCINVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "ALIORSVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "LINTEOLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0853"]}},
    "TORCVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0836"]}},
    "CARNEADES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "PATROCLES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SYNOECIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s3985"]}},
    "BAPTISTERIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["b0109"]}},
    "PSALTERIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p4274"]}},
    "POLITES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ACHILLIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PELIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CATINVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["34681"]}},
    "MVSCARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m1808"]}},
    "INGAEVONES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "INTERPRETAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i2445"]}},
    "CLAVARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c2056"]}},
    "OEBALIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "VIRIDARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["55362"]}},
    "SIBVLATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "NEXVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["n9770"]}},
    "PENICILLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p0797"]}},
    "SIMONIDES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "DEPRAESENTIARVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "COMPACTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["48823"]}},
    "DELIRAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d0578"]}},
    "DELECTAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d9513"]}},
    "PHERES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PERSICVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["43701"]}},
    "HAVSTRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["h0151"]}},
    "CAVAEDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c0927"]}},
    "REFVGIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["r0282"]}},
    "CORCILLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c4014"]}},
    "AONES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PRATVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p3480"]}},
    "TRISTIMONIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t1385"]}},
    "GALLINICIVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "FRIGIDVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f0295"]}},
    "PIGMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p2042"]}},
    "PANARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p0197"]}},
    "SPHAERISTERIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s2159"]}},
    "SVDVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["100767"]}},
    "PARTVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "CHIRAMAXIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c1446"]}},
    "OSTIOLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o1114"]}},
    "SPITACES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "DESIDIABVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d0971"]}},
    "SEXENNIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1523"]}},
    "BELLEROPHONTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PISVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p2138"]}},
    "STERQVILINVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["48731"]}},
    "METHODIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0890"]}},
    "TERMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0354"]}},
    "POSTLIMINIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p5008"]}},
    "ARCISELLIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a2744"]}},
    "VANGIONES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "VVLTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u1419"]}},
    "ALCIDES": {"N": {"morpho": ["n-s---mn1-"], "uri": []}},
    "CATILLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["97849"]}},
    "SESAMVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1429"]}},
    "QVASILLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["q0151"]}},
    "ALIQVOVORSVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "MAGNETES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SEMICINCTIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1023"]}},
    "EVRIPIDES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "LEPONTII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "PYLADES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "OSCENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "DIVERBIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["49209"]}},
    "TRALLES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "TVRBAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t1756"]}},
    "AEACIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CINGILLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c1650"]}},
    "IMMERITVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["113106"]}},
    "LARIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "HORAEVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "DIFFVGIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d1410"]}},
    "SVBLIGACVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s2938"]}},
    "NASVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["99539"]}},
    "SENONES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "BIFVRCVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["b0355"]}},
    "SEPLASIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1267"]}},
    "BLITEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["50582"]}},
    "OXATHRES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CORCVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c4016"]}},
    "VESTICONTVBERNIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0500"]}},
    "LENIMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0515"]}},
    "PITTACIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p2195"]}},
    "VIRECTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0756"]}},
    "SAXETVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0481"]}},
    "NAVCVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["n0082"]}},
    "BASILICVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["b0157"]}},
    "PROPVDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p4690"]}},
    "VTERVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["35863"]}},
    "CONTICINIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c3741"]}},
    "MANTELVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0321"]}},
    "CONSIDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c3533"]}},
    "EXVVIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e1922"]}},
    "VTROQVEVERSVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "RHAMNES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "VNCTORIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u1022"]}},
    "LITIGIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0915"]}},
    "PISTILLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["53609"]}},
    "CLIVVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["97908"]}},
    "FVSTVARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f1258"]}},
    "LATINIENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "CLIPEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["97907"]}},
    "ORICALCHVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "EVRYTIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "IENTACVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i0115"]}},
    "STABILIMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s2338"]}},
    "HIPPOTADES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "PRAEIVDICATVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["40840"]}},
    "GLANDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["g0346"]}},
    "OENOPOLIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0545"]}},
    "VERIVERBIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0382"]}},
    "BVBACES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CALLISTHENES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "TOLOSATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "LESBIADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CORFINIENSES": {"N": {"morpho": ["n-p---cn3i"], "uri": []}},
    "ARIOBARZANES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "CALAMISTRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c0162"]}},
    "POEMATIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p2433"]}},
    "CVPEDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["44702"]}},
    "DENTIFRANGIBVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["31849"]}},
    "CECROPIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "GVSTATORIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["g0664"]}},
    "EXPLEMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e1632"]}},
    "MYROPOLIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m1913"]}},
    "NASTVRTIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["118473"]}},
    "INTEGVMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i2222"]}},
    "THESTIADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "IRRIDICVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i2869"]}},
    "ODARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0504"]}},
    "HOSTIMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["h0637"]}},
    "CRYSTALLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c4479"]}},
    "NVCIFRANGIBVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["n0613"]}},
    "ASCLEPIADES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "TRIFOLIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t1257"]}},
    "MARSVPIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["47915"]}},
    "THESAVRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["100874"]}},
    "EMBOLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e0429"]}},
    "BACCIBALLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["b0024"]}},
    "PROCRVSTES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "PROMONTVRIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["54955"]}},
    "NOMADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "AMPHITEATRVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "CORYBANTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ILLORSVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "BISTONIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SVBVRBIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s3255"]}},
    "MENDICABVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0716"]}},
    "SPECVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "MALEICORIVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "EVRYBATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "TELINVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0210"]}},
    "GOTARZES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "VASCONES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "LABDACIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CLEOCHARES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "BRANNOVICES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "IDYLLIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i0098"]}},
    "GYNAECEVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["g0685"]}},
    "CONSERVITIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c3525"]}},
    "DILVCVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["d1591"]}},
    "OREADES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "MELCVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0609"]}},
    "MANVCIOLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m0348"]}},
    "STATVNCVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s2404"]}},
    "SISYMBRIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1797"]}},
    "BASIOLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["b0169"]}},
    "ZARIASPES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CHARITES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "THERSITES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "AESONIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "COLIPHIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["56403"]}},
    "OFFENDICVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0558"]}},
    "ARABARCHES": {"N": {"morpho": ["n-s---mn1g"], "uri": ["a2626"]}},
    "MELLICVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["117306"]}},
    "CRVCIAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c4424"]}},
    "ALTROVORSVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "VENALICIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u0256"]}},
    "MILIARIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m9986"]}},
    "BIGERRIONES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "APODYTERIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a2401"]}},
    "CONDITORIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c3171"]}},
    "LATERICIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0350"]}},
    "PROSTIBVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["p4138"]}},
    "TENTAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["33899"]}},
    "OXARTES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SVBNIMIVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "FASCINVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["f0233"]}},
    "OXYCOMINVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o1175"]}},
    "EXOTICVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["e1565"]}},
    "DIOMEDES": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "CALIDVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c2528"]}},
    "BLANDVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "LIBYES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CHRYSES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "LESBIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CORYMBIVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "HERMINONES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "DIARDINES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "MORETVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["m1404"]}},
    "GASTRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["g0097"]}},
    "TONSTRINVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0816"]}},
    "ARPINATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PARISII": {"N": {"morpho": ["n-p---mn2-"], "uri": []}},
    "INFORTVNIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i1632"]}},
    "TRINVNDINVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["130621"]}},
    "CEREBELLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["c1228"]}},
    "ELLVM": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "FEBRVVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["39456"]}},
    "LIBRVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["l0676"]}},
    "PHARASMANES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "COPHES": {"N": {"morpho": ["n-p---mn3-"], "uri": []}},
    "SENTICETVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1236"]}},
    "OFFVLGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["o0596"]}},
    "SACES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CAERACATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ARGESTES": {"N": {"morpho": ["n-s---mn3i"], "uri": []}},
    "DIRIBEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["d1676"]}},
    "SESSORIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s1495"]}},
    "MACEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["m0018"]}},
    "PRAEHIBEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["p3043"]}},
    "VERCVLVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["u9364"]}},
    "TITTIBILICIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["t0748"]}},
    "EXSVGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["98449"]}},
    "ALLEVAMENTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["a1410"]}},
    "ATTORQVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a3372"]}},
    "THVCIDIDES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SPICVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0988"]}},
    "ISELASTICVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i2928"]}},
    "OBOLEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["o0159"]}},
    "AFFLEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a1099"]}},
    "ORICHALCVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["o0993"]}},
    "FRENDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["f0965"]}},
    "REFERVEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["r1113"]}},
    "SPLENIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s2244"]}},
    "CONSVADEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c3633"]}},
    "ADDENSEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["a0525"]}},
    "LINTEO": {"-": {"morpho": ["v1spia--2-"], "uri": []}},
    "FOETEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["f0766"]}},
    "RECALEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["r0991"]}},
    "TYRIDATES": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CONFVLGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c3327"]}},
    "CIRCVMSIDEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c1861"]}},
    "SVBTIMEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["s3214"]}},
    "IMMVLGEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["i0412"]}},
    "SVPERIMMINEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["s3581"]}},
    "SCHEDIVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["s0612"]}},
    "INSCRIPTVM": {"-": {"morpho": ["n-s---nn2-"], "uri": ["i3186"]}},
    "COMPLACEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["c2847"]}},
    "REDHIBEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["r1082"]}},
    "REMVLCEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["r1219"]}},
    "OBSORBEO": {"-": {"morpho": ["v1spia--2-"], "uri": ["o0259"]}},
    "SEPARATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s1254"]}},
    "PERPOTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p1503"]}},
    "PERSALVTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p1542"]}},
    "PARASITATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p0349"]}},
    "COLLAVDATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2388"]}},
    "LIBATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0663"]}},
    "DISTVRBATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1984"]}},
    "CONSOCIATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3553"]}},
    "PELLICVLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p0761"]}},
    "CONCERTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2988"]}},
    "OCCVRSATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0422"]}},
    "OCCATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0373"]}},
    "EGESTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0214"]}},
    "RECEPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0155"]}},
    "SPECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s2116"]}},
    "OSCILLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o1069"]}},
    "PERSCRIPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p1553"]}},
    "IMMODERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0393"]}},
    "OBSECRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0204"]}},
    "ASSECTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a3114"]}},
    "PLACABILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2209"]}},
    "INSTAVRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2112"]}},
    "RESTITVTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0557"]}},
    "MVTITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["118306"]}},
    "NEGATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["n0174"]}},
    "BLATIO": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "EXANIMATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1227"]}},
    "SINISTERITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s1752"]}},
    "VERBIVELITATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0366"]}},
    "COEMPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2269"]}},
    "CONDEMNATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3139"]}},
    "AEQVABILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0913"]}},
    "INVECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2640"]}},
    "DERELICTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0896"]}},
    "EMANCIPATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0396"]}},
    "PERCVRSATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p0996"]}},
    "VILICATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u9655"]}},
    "IVGATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i3004"]}},
    "DIIVNCTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["50517"]}},
    "AFFABILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1062"]}},
    "EMBASICOETAS": {"-": {"morpho": ["n-s---mn1g"], "uri": ["e0416"]}},
    "AMANDATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1569"]}},
    "STIMVLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s2482"]}},
    "PVLSATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p4483"]}},
    "AMATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a8593"]}},
    "LOQVACITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l1036"]}},
    "OBSTRVCTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0283"]}},
    "EXPVRGATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1699"]}},
    "EMINATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0460"]}},
    "CONSORTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3579"]}},
    "INFLAMMATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i1600"]}},
    "CONVOCATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3928"]}},
    "INVESTIGATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2668"]}},
    "PRAECENTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2822"]}},
    "OBROGATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0178"]}},
    "COLLIGATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2424"]}},
    "ADMIRABILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0664"]}},
    "DOCILITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d2072"]}},
    "PENSATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p0812"]}},
    "FRIGVTIO": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "PROCACITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3633"]}},
    "SORTIO": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "DECLARATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0213"]}},
    "FESTIVITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0435"]}},
    "INTERFATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2320"]}},
    "GVSTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["g0663"]}},
    "DEFLAGRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0379"]}},
    "ANNOTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["45071"]}},
    "COMPELLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2808"]}},
    "MALEDICTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m0163"]}},
    "INCOMMODITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0983"]}},
    "CONGLVTINATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3378"]}},
    "CONSITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3546"]}},
    "EXPLICEO": {"-": {"morpho": ["v1spia--2-"], "uri": []}},
    "DISCRIPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d9938"]}},
    "ADMIXTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0677"]}},
    "COITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2351"]}},
    "GRANDITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["g0541"]}},
    "RELIGATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0342"]}},
    "NVNTIATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["n0686"]}},
    "PERACTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p0872"]}},
    "IACVLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0017"]}},
    "DIRITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1683"]}},
    "EFFRENATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0173"]}},
    "APPROPINQVATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a2541"]}},
    "CELLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c1074"]}},
    "DESTITVTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1033"]}},
    "OBSCENITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0195"]}},
    "RECVPERATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["26677"]}},
    "DICATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1333"]}},
    "INTERPVNCTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2454"]}},
    "INTERCEPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2279"]}},
    "VIVACITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0897"]}},
    "STIPATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s2492"]}},
    "QVAESITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["q0097"]}},
    "VITATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0841"]}},
    "EMACITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0390"]}},
    "INFIRMATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i1582"]}},
    "EXPIATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1614"]}},
    "SVBIGITATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s2879"]}},
    "APRICATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a2551"]}},
    "CONFICTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3236"]}},
    "IRRIGATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2870"]}},
    "STVPIDITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s1733"]}},
    "COMPOTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c7883"]}},
    "SCISCITATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0670"]}},
    "HELVATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["112192"]}},
    "GERMANITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["g0231"]}},
    "ABIECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0054"]}},
    "PVTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p4596"]}},
    "SEDVCTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0926"]}},
    "CONCENATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2974"]}},
    "INCOMITIO": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "PRAEROGATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3237"]}},
    "SERVATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s1414"]}},
    "DECERTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0180"]}},
    "EXSECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1737"]}},
    "PERAGRATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p0883"]}},
    "RESTIPVLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0554"]}},
    "EDICTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0083"]}},
    "AMPVTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1788"]}},
    "SCRIPTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0763"]}},
    "SVPPOSITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s3782"]}},
    "SATISDATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0430"]}},
    "PROPVLSATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p4040"]}},
    "OPIMITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0820"]}},
    "INSVLSITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2183"]}},
    "PROCRASTINATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3672"]}},
    "DISPERTITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1840"]}},
    "DIRIBITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1677"]}},
    "NEGLECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["n0182"]}},
    "PROPVGNATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p4036"]}},
    "ILLIBERALITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0223"]}},
    "TEMPESTIVITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t0244"]}},
    "OBNVNTIO": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "CONSVETIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3645"]}},
    "EPVLATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0813"]}},
    "IMPVRITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0689"]}},
    "RECLAMATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0176"]}},
    "MVTVATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m1871"]}},
    "AEQVATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0922"]}},
    "FORMOSITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0858"]}},
    "VINDICATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0702"]}},
    "GRATIFICATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["g0565"]}},
    "DIVITATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["98224"]}},
    "REPASTINATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0421"]}},
    "DEVEXITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1157"]}},
    "COLLOCVTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2443"]}},
    "MOROSITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m1417"]}},
    "DISPERCVTIO": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "DISSIPATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1912"]}},
    "INSATIETAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i1950"]}},
    "IMMATVRITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0351"]}},
    "DETRECTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d1113"]}},
    "COACTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2184"]}},
    "ASSIGNATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a3165"]}},
    "PROMVLGATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3903"]}},
    "IMPLICATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0571"]}},
    "ASSERTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a3142"]}},
    "INTERDICTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2301"]}},
    "DERIVATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0909"]}},
    "DEBITIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0107"]}},
    "OPTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0899"]}},
    "OBLIQVITAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["o0101"]}},
    "ELOCVTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0352"]}},
    "INSPECTATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2080"]}},
    "ROGITATIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0781"]}},
    "CONSVRRECTIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3665"]}},
    "CALVIA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "GALLITTA": {"N": {"morpho": ["n-s---fn2-"], "uri": []}},
    "IARBITAS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "AMYNTAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "DAMOETAS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "LYCOTAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "GORGATAS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "PHILETAS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "HORRATAS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "ARCHYTAS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "EVROTAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "FAVSTITAS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "PHILOTAS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "AERIFER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "DIFFIDENTER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "MELEAGER": {"N": {"morpho": ["n-s---mn2r"], "uri": []}},
    "FATALITER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "INSIPIENTER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "GRANDIFER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "ELOQVENTER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "TVRRIFER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "SCAMANDER": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "INCONTINENTER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "SOTER": {"N": {"morpho": ["n-s---mn3-"], "uri": ["s2047"]}},
    "IRREVERENTER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "THYMBER": {"N": {"morpho": ["n-s----n2-"], "uri": []}},
    "HISTER": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "AFER": {
        "N": {"morpho": ["n-s----n2-"], "uri": []},
        "A": {"morpho": ["aps---mn1r"], "uri": []},
    },
    "RACEMIFER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "CAPITALITER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "AMPHITHEATER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "NEGLEGENTER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "FVMIFER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "CONFIDENTER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "FVRIALITER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "FVRACITER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "INSIGNITER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "PVELLARITER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "LATENTER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "DENTER": {"N": {"morpho": ["n-s----n2-"], "uri": []}},
    "TEVCER": {
        "N": {"morpho": ["n-s---mn3-"], "uri": []},
        "A": {"morpho": ["aps---mn1r"], "uri": []},
    },
    "PERSAPIENTER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "CIMBER": {
        "N": {"morpho": ["n-s----n2-"], "uri": []},
        "A": {"morpho": ["aps---mn1r"], "uri": []},
    },
    "SOCIALITER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "ASSIMILITER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "INTELLIGENTER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "DVLCITER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "VILITER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "ELAVER": {"N": {"morpho": ["n-s----n2-"], "uri": []}},
    "LYSANDER": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "OLIVIFER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "TANAGER": {"N": {"morpho": ["n-s----n2-"], "uri": []}},
    "MONSTRIFER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "NOTABILITER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "PRODIGIALITER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "TITVBANTER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "TRISTITER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "EXCELLENTER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "ALCANDER": {"N": {"morpho": ["n-s----n2-"], "uri": []}},
    "OBSEQVENTER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "PINIGER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "PERBREVITER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "PERCELER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "ANTONIASTER": {"N": {"morpho": ["n-s----n2-"], "uri": []}},
    "SPHINTER": {"N": {"morpho": ["n-s----n2-"], "uri": []}},
    "AMPLITER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "BLANDITER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "CONGER": {"-": {"morpho": ["n-s---mn2r"], "uri": ["c3352"]}},
    "EXSVLTANTER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "ALABASTER": {"-": {"morpho": ["n-s---mn2r"], "uri": ["a1257"]}},
    "ASANDER": {"N": {"morpho": ["n-s----n2-"], "uri": []}},
    "SACRIFER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "SPINTHER": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "ENSIFER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "DOLENTER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "CREDIBILITER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "CASSANDER": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ANTIPATER": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SERVILITER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "SOMNIFER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "MILITARITER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "MVLCIBER": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "CLEANDER": {"N": {"morpho": ["n-s----n2-"], "uri": []}},
    "HERBIFER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "CORYMBIFER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "ARISTANDER": {"N": {"morpho": ["n-s----n2-"], "uri": []}},
    "SCVRRILITER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "IMPERABILITER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "PISANDER": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "VNGVLATER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "LEANDER": {"N": {"morpho": ["n-s---mn2r"], "uri": []}},
    "FALLACITER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "PACIFER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "PRAESTANTER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "SEMICAPER": {"-": {"morpho": ["n-s---mn2r"], "uri": ["s1020"]}},
    "TRVCVLENTER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "DIESPITER": {"N": {"morpho": ["n-s----n2-"], "uri": []}},
    "TVDER": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "REDVNDANTER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "IMPARITER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "PHAGER": {"-": {"morpho": ["n-s---mn2r"], "uri": ["p1838"]}},
    "CORPORALITER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "MNESTER": {"N": {"morpho": ["n-s---mn3-"], "uri": ["m1155"]}},
    "TRIFVRCIFER": {"-": {"morpho": ["n-s---mn2r"], "uri": ["t1261"]}},
    "MAEANDER": {
        "N": {"morpho": ["n-s---mn3-"], "uri": []},
        "-": {"morpho": ["n-s---mn3-"], "uri": []},
    },
    "APPETENTER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "BICORNIGER": {"N": {"morpho": ["n-s----n2-"], "uri": []}},
    "MALIFER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "ARVNDIFER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "MENANDER": {"N": {"morpho": ["n-s---mn2r"], "uri": []}},
    "ANTHEMVSIAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ARSANIAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ARCHIAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ADARRHIAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ATARRHIAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "HARPOCRAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "HABINNAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "TIRESIAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ATTINAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PNYTAGORAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SCIPIADAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "MIMAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "CASPIAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ONOMAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "MIDAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "DERDAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "LYCORMAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "DRYAS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "GNOSIAS": {
        "N": {"morpho": ["n-s---mn1g"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "LYSIAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "PHIDIAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ERYMAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "TROAS": {
        "N": {"morpho": ["n-s---mn1g"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "NAIAS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "SYMPLEGAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "IOLLAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "LADAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "CYDAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ANAXAGORAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "SIMIAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ABAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ARCAS": {
        "N": {"morpho": ["n-s---mn3-"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "PYTHIAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ATHENAGORAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "DYMAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "SCOPAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "HYLAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ACHILLAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "THESTIAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PERDICCAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "HIERAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ATYANAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CALLICRATIDAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PTERELAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "THOAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "PRVSIAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "GYAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CORYBAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ASILAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "EVHYAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PAVSANIAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "MENIDAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PERIPHAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "PHORBAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ATHARRIAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "AVGEAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "THAVMANTIAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PELIAS": {
        "N": {"morpho": ["n-s---mn3-"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "PANTAGIAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "GELAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "MAENIDAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "THAMYRAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "IARBAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PHINTIAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ORCAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ELIAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PALLANTIAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "LICHAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "PHOEBAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PLEIAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ACAMAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ATHANAGORAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "BELIAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PHTHIAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PROTAGORAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ALEVAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "AERIAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "HYAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "CINEAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "CINYRAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "PELORIAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "CARRINAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "TACFARINAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ARTAXIAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ARPINAS": {
        "N": {"morpho": ["n-s---mn1g"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "THOANTIAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "DIPSAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "THIODAMAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "LYCIDAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "APPIAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "EVRYDAMAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "BITIAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PASAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "IOPAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "LAENAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PHASIAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "BROTEAS": {"N": {"morpho": ["n-s---mn1-"], "uri": []}},
    "NOMAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "MENALCAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "MARSYAS": {"N": {"morpho": ["n-s---mn1-"], "uri": []}},
    "HELLAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CORAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "CLVSINAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "POEAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "PLIAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PSECAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "BIAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "CALCHAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "TEVTHRAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ATHAMAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "PHLEGYAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "IPHIAS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "BAGOAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "TAMIRAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "THEMILLAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "PERLIBERALITER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "PRAETERPROPTER": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "ARELAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "IDAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "FELGINAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ATLAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "THYIAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "LYCORIAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "GORGIAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "OLYMPIAS": {"N": {"morpho": ["n-s---mn3-"], "uri": ["101329"]}},
    "EPAMINONDAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "HIPPIAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ACRAGAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "XANTHIAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "BVCEPHALAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "ILIAS": {"N": {"morpho": ["n-s---mn1g"], "uri": []}},
    "HERMAGORAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ATHENION": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "LYCON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "STIMICHON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "EMATHION": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "OROMEDON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "SYNODON": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "MEDON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "CALCHEDON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "NICASION": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "PORPHYRION": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "AVEON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "MARATHON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "LYCAON": {"N": {"morpho": ["n-s---mn3-"], "uri": ["l1259"]}},
    "HAEMON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "LEVCON": {"N": {"morpho": ["n-s---nn2-"], "uri": []}},
    "AVTOMEDON": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "PELAGON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "BITON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "PYGMALION": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "CHIRON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "TIMON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "IXION": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "SASON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "HYPERION": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "TAVRON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "BABYLON": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "AGATHON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "CALYDON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "PLEVRON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "XENOPHON": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ALCIMEDON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "OLYMPION": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "PALAEMON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "AMAZON": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "PHITOLEON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "DEVCALION": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "CREON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "MAEON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "ENDYMION": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "IASON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "AGAMEMNON": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "POTAMON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "POLEMON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "PATRON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "ARISTION": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "AMYDON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "PROCOETON": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p3667"]}},
    "HECATON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "ARTAMON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "COLOPHON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "EVRYTION": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "ELEGIDARION": {"-": {"morpho": ["n-s---nn2g"], "uri": ["e0289"]}},
    "SILANION": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "HERACLEON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "APION": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "ZENON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "GERYON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "THEON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "LEON": {"N": {"morpho": ["n-s---mn3i"], "uri": []}},
    "XENON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "DEMOPHOON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "PARMENON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "TELAMON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "MACHAON": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "HIPPON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "LYCMON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "HEGION": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "BOLON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "CLEON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "LADON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "METRON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "DEMOPHON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "CRANNON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "HAMMON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "HISBON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "PHORMION": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "GITON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "PYRRHON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "ALISON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "VRION": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "STRYMON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "PHAON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "NICON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "CTESIPHON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "CITHAERON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "THERON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "PHAEDON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "AMMON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "POLYPEMON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "TELON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "DIDYMEON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "ACHERON": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "LACON": {
        "N": {"morpho": ["n-s----n-g"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "HIERON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "CONON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "PYTHON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "MATHON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "LYSON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "SIMON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "ANACREON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "MELON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "EVDAEMON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "LYCOPHRON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "AVLON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "DRACON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "LEOMEDON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "DAMASICTHON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "SCIRON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "PLVTON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "DAMON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "NYMPHON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "TARCHON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "BLEPHARON": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "VCALEGON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "DIDYMAON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "ACRON": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "THRASON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "ECHION": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "PHARON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "CROTON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "MENON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "CHAON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "SERAPION": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "DOLICHAON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "HERACON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "ANDROGEON": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "CAVLON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "GNATHON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "PROCYON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "HEPHAESTION": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "EVCTEMON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "PHAETHON": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "HIPPOCOON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "AMPHION": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "LAOMEDON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "ORION": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "OARION": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "SOCRATION": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "POLYPERCON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "ARISTON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "SOLON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "THAEMON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "IMAON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "STRABON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "NOEMON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "SOTION": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "PANDION": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "MILANION": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ARION": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "AEGON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "ATHLON": {"-": {"morpho": ["n-s---nn2g"], "uri": ["97707"]}},
    "DAPHNON": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d0049"]}},
    "ARTEMON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "CORYDON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "HYPERBATON": {"-": {"morpho": ["n-s---nn2g"], "uri": ["h0745"]}},
    "AESCHRION": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "MICON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "MYRON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "ACTAEON": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "MILON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "PARMENION": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "SARPEDON": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "CHARON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "STRATON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "IDMON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "DION": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "PHILON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "MYOPARON": {"-": {"morpho": ["n-s---nn2-"], "uri": ["29497"]}},
    "AMPHITRYON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "PYRACMON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "CYDON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "SINON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "THERMODON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "SICYON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "SVNION": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "HYPOCAVSTON": {"-": {"morpho": ["n-s---nn2g"], "uri": ["h0772"]}},
    "CIMON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "CALLIPHON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "SATYRION": {"-": {"morpho": ["n-s---nn2g"], "uri": ["s0458"]}},
    "AETHON": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ACMON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "THESPRION": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "MEMNON": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "PLATON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "THEMISON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "EETION": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "ALIACMON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "EVCLION": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "PELION": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "ALCON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "TETRASTYLON": {"-": {"morpho": ["n-s---nn2-"], "uri": ["100867"]}},
    "STILBON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "GLYCON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "DOLON": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["98234"]},
        "N": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "TRICHON": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "PRASINVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["56724"]}},
    "PRISCE": {"-": {"morpho": ["rp--------"], "uri": ["38499"]}},
    "TRICENVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["100949"]}},
    "AVSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["103422"]}},
    "SEPTIENS": {"-": {"morpho": ["rp--------"], "uri": ["s1297"]}},
    "SVLMO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "ADBITO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0517"]}},
    "MALITIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0196"]}},
    "PHAECASIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p1834"]}},
    "HECTOR": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "FVSTITVDINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f1257"]}},
    "PETAVRISTARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p1794"]}},
    "PROMISSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p3886"]}},
    "ENTHYMEMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["e0636"]}},
    "HAESITABVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0049"]}},
    "CANTAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["c0442"]}},
    "DISPOLIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["98203"]}},
    "CONGREGABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c3387"]}},
    "EVRYNOME": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "LYDE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "VLTIMO": {"2": {"morpho": ["v1spia--1-"], "uri": ["u0942"]}},
    "EXAVSPICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1263"]}},
    "HALOPHANTA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["h0075"]}},
    "DELIBERATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d0542"]}},
    "ARVNDINEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["103142"]}},
    "PHYLACA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p1991"]}},
    "LECTICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0456"]}},
    "SVBIGITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s2881"]}},
    "INTESTINA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["i2551"]}},
    "PERFIDIOSE": {"-": {"morpho": ["rp--------"], "uri": ["35851"]}},
    "CATEIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0868"]}},
    "STEGA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2410"]}},
    "CAVTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c1025"]}},
    "IMPASTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0451"]}},
    "SORORIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s2028"]}},
    "SPATALOCINAEDVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s2080"]}},
    "POSTILIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p5003"]}},
    "FESTINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0432"]}},
    "INDIGNATIVNCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i1276"]}},
    "PRAETERBITO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3368"]}},
    "INANIAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["i0757"]}},
    "AETATVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a1038"]}},
    "PRODIGIALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p3716"]}},
    "INFORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1629"]}},
    "CROTALISTRIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4415"]}},
    "STIGMOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2469"]}},
    "TONSORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0813"]}},
    "SVPERVOLITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s3733"]}},
    "LAMELLVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0159"]}},
    "RIDEOR": {"-": {"morpho": ["v1spid--2-"], "uri": ["132793"]}},
    "CORNEOLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4861"]}},
    "AEQVIPARABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["19595"]}},
    "CAVSIFICOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["53846"]}},
    "DELIQVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["54588"]}},
    "DEPVLSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d0881"]}},
    "PASTORICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0536"]}},
    "OPTEMPERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["36166"]}},
    "CIRCVMSPECTATRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c1873"]}},
    "CRITICA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["c4384"]}},
    "LAVDICENVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l9411"]}},
    "LATERNARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["115893"]}},
    "OBSCAEVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0192"]}},
    "ELEVSIN": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "SISAPO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "NEREIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "SCYLAX": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ROGATIVNCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r0779"]}},
    "BELLIPOTENS": {"-": {"morpho": ["aps---an3i"], "uri": ["b0237"]}},
    "MERCATORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0777"]}},
    "LATERENSIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "DVELLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["22733"]}},
    "REPVGNANTIA": {"-": {"morpho": ["n-p---nn3i"], "uri": ["r9995"]}},
    "COPIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["106329"]}},
    "HAEDILLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["h0028"]}},
    "MISCIX": {"-": {"morpho": ["aps---an3i"], "uri": ["99435"]}},
    "PASTORALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p0535"]}},
    "LVCRETILIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "CENATORIA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["97852"]}},
    "PROPINQVE": {"-": {"morpho": ["rp--------"], "uri": ["100199"]}},
    "SVPPINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3761"]}},
    "POLYDAMAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "NEVTIQVAM": {"-": {"morpho": ["rp--------"], "uri": ["n9809"]}},
    "VASVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["43983"]}},
    "INTERNECO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2410"]}},
    "AVTOPYROS": {"-": {"morpho": ["n-s---mn2g"], "uri": ["97732"]}},
    "QVATERNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["q0161"]}},
    "ARIETINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2833"]}},
    "LENONIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0520"]}},
    "MORIGERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["30265"]}},
    "EMINOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["e0465"]}},
    "TAPES": {"-": {"morpho": ["n-s---mn3-"], "uri": ["t0109"]}},
    "COMPITALICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c2843"]}},
    "SOLARIVM": {"2": {"morpho": ["n-s---nn2-"], "uri": ["s4023"]}},
    "BALISCVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["b0062"]}},
    "VIVE": {"-": {"morpho": ["rp--------"], "uri": ["101171"]}},
    "INDVSIATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1371"]}},
    "PALMVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p0143"]}},
    "IMPERATRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0482"]}},
    "SEPTIFLVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s8297"]}},
    "ABITO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0066"]}},
    "INSECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["51424"]}},
    "FIDICINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0497"]}},
    "PEREGRI": {"-": {"morpho": ["rp--------"], "uri": ["31066"]}},
    "LIMAX": {"-": {"morpho": ["n-s---cn3-"], "uri": ["l0770"]}},
    "CROCOTVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4410"]}},
    "LINGVLACA": {"-": {"morpho": ["n-s---cn1-"], "uri": ["l0827"]}},
    "CENTIENS": {"-": {"morpho": ["rp--------"], "uri": ["c1137"]}},
    "CISTVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c1976"]}},
    "PACTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p9863"]}},
    "LIBIDINOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["l0693"]}},
    "STRIGA": {"2": {"morpho": ["n-s---fn1-"], "uri": ["s4033"]}},
    "AMASIS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "DAPHNIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "LVSITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l1219"]}},
    "LIMONE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "DEVIRGINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1168"]}},
    "RVSTICOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["r1503"]}},
    "INSPVTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2106"]}},
    "CANTVRIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["c0470"]}},
    "COQVINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3986"]}},
    "APPINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a2501"]}},
    "SALAMIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "SENVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["100553"]}},
    "PVGNATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p4426"]}},
    "ARIETILLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2832"]}},
    "DVODECIENS": {"-": {"morpho": ["rp--------"], "uri": ["d2340"]}},
    "PARENTALIA": {"-": {"morpho": ["n-p---nn3i"], "uri": ["p0383"]}},
    "TRITON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "PROPONTIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "ATYS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "FALSE": {"-": {"morpho": ["rp--------"], "uri": ["98482"]}},
    "PERFVNCTORIE": {"-": {"morpho": ["rp--------"], "uri": ["34051"]}},
    "INSTRVCTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i2149"]}},
    "SIMITV": {"-": {"morpho": ["rp--------"], "uri": ["s1693"]}},
    "ECLOGA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["e0057"]}},
    "PARTHENOPE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "VESERIS": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "FLAMMARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["f0625"]}},
    "ARBITRARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2652"]}},
    "PANNVCIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["99791"]}},
    "INFIRME": {"-": {"morpho": ["rp--------"], "uri": ["24174"]}},
    "CONTVMVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3843"]}},
    "DIFFLAGITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1400"]}},
    "MYTILENE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PATINARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0565"]}},
    "CASTIGATORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0775"]}},
    "CELAENO": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "ZEPHYRITIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "OENONE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SCVRROR": {"-": {"morpho": ["v1spid--1-"], "uri": ["s0816"]}},
    "CONIVGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3428"]}},
    "STERNAX": {"-": {"morpho": ["aps---an3i"], "uri": ["s2451"]}},
    "CAPO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c0528"]}},
    "HIPPVRVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["h0432"]}},
    "HOSTIO": {"1": {"morpho": ["v1spia--4-"], "uri": ["h0638"]}},
    "INSIPIENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i7043"]}},
    "ORACVLARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0924"]}},
    "FVNDATOR": {"1": {"morpho": ["n-s---mn3-"], "uri": ["f1269"]}},
    "SCALPVRRIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["s0537"]}},
    "IRREVERENS": {"-": {"morpho": ["aps---an3i"], "uri": ["i2862"]}},
    "INDIPISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["98930"]}},
    "SITIENS": {"-": {"morpho": ["aps---cn3-"], "uri": ["27226"]}},
    "CONGRAECO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3383"]}},
    "DIARIA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["98165"]}},
    "ASSESSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a8906"]}},
    "MELILOTOS": {"-": {"morpho": ["n-s---fn2g"], "uri": ["m0621"]}},
    "COACESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c2177"]}},
    "AVELLANA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["97738"]}},
    "EXPETESSO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1609"]}},
    "OBSVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0292"]}},
    "COTYS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "MEROBIBVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0821"]}},
    "BEATITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["b0206"]}},
    "INTERSCRIBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2486"]}},
    "CONSVLTE": {"-": {"morpho": ["rp--------"], "uri": ["46160"]}},
    "GAVSAPE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["98644"]}},
    "HEREDIPETA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["h0292"]}},
    "PERNECESSARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1392"]}},
    "HECALE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "BVLEVTA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["b0722"]}},
    "DEAMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0085"]}},
    "DESOMNIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["d0990"]}},
    "HYDRAVLES": {"-": {"morpho": ["n-s---mn1-"], "uri": ["98786"]}},
    "VIATICVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["50528"]}},
    "ACLYS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0390"]}},
    "VITIOSE": {"-": {"morpho": ["rp--------"], "uri": ["54635"]}},
    "SEDECIES": {"-": {"morpho": ["rp--------"], "uri": ["s0907"]}},
    "EXSVSCITO": {"1": {"morpho": ["v1spia--1-"], "uri": ["e1825"]}},
    "CILO": {"N": {"morpho": ["n-s---mn3-"], "uri": ["c1615"]}},
    "MVLTIPOTENS": {"-": {"morpho": ["aps---an3i"], "uri": ["m1674"]}},
    "NOCTIVIGILVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["118788"]}},
    "ZOTHECA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["z0048"]}},
    "QVINDECIENS": {"-": {"morpho": ["rp--------"], "uri": ["q0223"]}},
    "PRAEDVLCIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p2956"]}},
    "TARRACO": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "COLLABEFACTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2366"]}},
    "COCIO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c2261"]}},
    "INSALVTATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1937"]}},
    "DEFAECO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0333"]}},
    "TITAN": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ALARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["a1268"]}},
    "DEONERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0745"]}},
    "LAGVNCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0144"]}},
    "PERSVASVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["p9885"]}},
    "BAIVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["b0046"]}},
    "ERADICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0846"]}},
    "APOCVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a2395"]}},
    "INTERCEPTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i2280"]}},
    "SERMONOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["s1381"]}},
    "AQVILEX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a2608"]}},
    "COMVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c2935"]}},
    "SVCOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s3315"]}},
    "PERFLAGITIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1122"]}},
    "ORESTIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "DIGNITOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d1441"]}},
    "SVPEREGREDIOR": {"-": {"morpho": ["v1spid--3i"], "uri": ["s3526"]}},
    "EXPVLSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1690"]}},
    "PERPLEXOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["p1493"]}},
    "SYBARIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "OPPESSVLATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["99732"]}},
    "DISSENSVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["d9556"]}},
    "REDIVIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["30973"]}},
    "BYLLIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "CENATICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c1097"]}},
    "CATAPIRATERIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0834"]}},
    "HELLE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "AGRICVLTVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a9238"]}},
    "INDELECTATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1212"]}},
    "CONSAEPIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["c3460"]}},
    "TRACHIN": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "NOVACVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["n0049"]}},
    "CONTRVNCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3833"]}},
    "INDEPLORATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1223"]}},
    "INCONTINENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i1065"]}},
    "CARINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0648"]}},
    "DAMMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d0025"]}},
    "CIRCVMVISO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c1931"]}},
    "OTHRYS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "SENTENTIOLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1234"]}},
    "OGGANNIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["o0602"]}},
    "INEXSTINCTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1490"]}},
    "MANNVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["m0303"]}},
    "HIPPODROMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": []}},
    "HERMIONE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "BVRDVBASTA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["b0749"]}},
    "PARALYTICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["25722"]}},
    "CONFLICTOR": {"2": {"morpho": ["v1spid--1-"], "uri": ["132346"]}},
    "DEOSCVLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["d0755"]}},
    "TOLVTIM": {"-": {"morpho": ["rp--------"], "uri": ["t0784"]}},
    "MALEDICE": {"-": {"morpho": ["rp--------"], "uri": ["47007"]}},
    "PLECTILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p2319"]}},
    "VIOLARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["u0736"]}},
    "REVIRESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1442"]}},
    "EVISCERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1099"]}},
    "EBVRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["54894"]}},
    "FALSIMONIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f0141"]}},
    "PERISTROMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["p1281"]}},
    "MODERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["25106"]}},
    "IMPERTERRITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0511"]}},
    "SEMIGRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s1060"]}},
    "CHORAVLA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["50845"]}},
    "VIDVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["u0630"]}},
    "LVSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l1217"]}},
    "EXTERSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["110156"]}},
    "FABARIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "POLYMITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2507"]}},
    "SERMVNCVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s1382"]}},
    "OPINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["99727"]}},
    "DELACERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0505"]}},
    "NATALICIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["n0058"]}},
    "CERASINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c1203"]}},
    "EXPRESSE": {"-": {"morpho": ["rp--------"], "uri": ["23136"]}},
    "ALMO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "AMBESTRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a1614"]}},
    "QVINQVEVIRATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["q0276"]}},
    "EPVLARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["e0811"]}},
    "DEVERSITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d1146"]}},
    "SEXCENTIENS": {"-": {"morpho": ["rp--------"], "uri": ["127112"]}},
    "SCROFIPASCVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0783"]}},
    "INCERTO": {"1": {"morpho": ["rp--------"], "uri": ["113507"]}},
    "ANXVR": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "SYCOPHANTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s3904"]}},
    "FVLIGINEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f1120"]}},
    "POETICE": {
        "2": {"morpho": ["n-s---fn1g"], "uri": ["100040"]},
        "-": {"morpho": ["n-s---fn1g"], "uri": ["100040"]},
        "1": {"morpho": ["n-s---fn1g"], "uri": ["100040"]},
    },
    "PROMPTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3898"]}},
    "PRISTIS": {
        "-": {"morpho": ["n-s---fn3i"], "uri": ["p3580"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "BINOMINIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["b0394"]}},
    "QVISQVILIA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["q0202"]}},
    "SVPRASCRIBO": {"-": {"morpho": ["v1spia--3-"], "uri": ["27567"]}},
    "COMMENDATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["31211"]}},
    "SPERABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s2140"]}},
    "BVCINVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["b0805"]}},
    "PRANSVS": {"2": {"morpho": ["n-s---mn4-"], "uri": ["p9905"]}},
    "TRANSFVSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t1025"]}},
    "PERELEGANS": {"-": {"morpho": ["aps---an3i"], "uri": ["p1064"]}},
    "VANITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["u0082"]}},
    "PRAESERVIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["p3282"]}},
    "HORDEIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["h0561"]}},
    "SVBAVSCVLTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s2765"]}},
    "THYONE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "EDICTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0084"]}},
    "GORGE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "DEMANDO": {"2": {"morpho": ["v1spia--1-"], "uri": ["d0609"]}},
    "ERATO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "INVIOLATE": {"-": {"morpho": ["rp--------"], "uri": ["38939"]}},
    "DIVEXO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d2024"]}},
    "VENERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["47458"]}},
    "EXSONO": {"1": {"morpho": ["v1spia--1-"], "uri": ["e1762"]}},
    "CORONIS": {"N": {"morpho": ["n-s---fn3-"], "uri": ["c4077"]}},
    "IMPVRE": {"-": {"morpho": ["rp--------"], "uri": ["24002"]}},
    "PVGILICE": {"-": {"morpho": ["rp--------"], "uri": ["p4415"]}},
    "MANTISCINOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["m0333"]}},
    "HELOPS": {"-": {"morpho": ["n-s---mn3-"], "uri": ["50527"]}},
    "MALEDICTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["m0164"]}},
    "SVCRO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "CVRSVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c9977"]}},
    "MANTO": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "EXHERES": {"-": {"morpho": ["n-s---mn3-"], "uri": ["e1467"]}},
    "PERTRACTATE": {"-": {"morpho": ["rp--------"], "uri": ["99975"]}},
    "SALTATRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0209"]}},
    "BARBVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["b0133"]}},
    "AFFIRMATE": {"-": {"morpho": ["rp--------"], "uri": ["31434"]}},
    "AMASIVNCVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a1592"]}},
    "RESTITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1364"]}},
    "IRRELIGIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2829"]}},
    "ARMAMENTA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["97676"]}},
    "PERCOPIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0979"]}},
    "EXHORTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["28860"]}},
    "MATVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0487"]}},
    "CALCITROSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0202"]}},
    "NERVIA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["n0239"]}},
    "RECENSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0145"]}},
    "EXOSTRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["e1562"]}},
    "AMPLEXO": {"-": {"morpho": ["v1spia--1-"], "uri": ["44933"]}},
    "PERLIBET": {"-": {"morpho": ["v3spia--2-"], "uri": ["p1306"]}},
    "TRAPEZITA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["130460"]}},
    "AVLOEDVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a3515"]}},
    "SAEPICVLE": {"-": {"morpho": ["rp--------"], "uri": ["s0080"]}},
    "CHRYSOPHRYS": {"-": {"morpho": ["n-s---fn3g"], "uri": ["c1556"]}},
    "INCLVSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0946"]}},
    "INSTIMVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2126"]}},
    "SPONDEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1595"]}},
    "ACTIVNCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a0446"]}},
    "OPERARIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["o0784"]}},
    "STENTOR": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ABSTINAX": {"-": {"morpho": ["aps---an3i"], "uri": ["a0189"]}},
    "OBSEQVELA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["99653"]}},
    "OSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["o1083"]}},
    "CARYOTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0625"]}},
    "POLLINCTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p2472"]}},
    "MEDVLLVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0580"]}},
    "DISTAEDET": {"-": {"morpho": ["v3spia--2-"], "uri": ["d1940"]}},
    "IAMIAMQVE": {"-": {"morpho": ["rp--------"], "uri": ["36578"]}},
    "PERSAPIENS": {"-": {"morpho": ["aps---an3i"], "uri": ["p1547"]}},
    "DEBLATERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0110"]}},
    "CENSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c1120"]}},
    "PYRENE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "VENTRIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0326"]}},
    "AMOENE": {"-": {"morpho": ["rp--------"], "uri": ["48543"]}},
    "LEVATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["l0600"]}},
    "HISPALIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "TARPEZITA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["t0135"]}},
    "HERBEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0275"]}},
    "PERSPICIENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p1643"]}},
    "CVBITORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4492"]}},
    "PERPLEXE": {"-": {"morpho": ["rp--------"], "uri": ["32899"]}},
    "REDAMBVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1068"]}},
    "NVTRICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["n0712"]}},
    "PANOPE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "REPOSTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["r0454"]}},
    "CRINALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c4356"]}},
    "MANIPVLATIM": {"-": {"morpho": ["rp--------"], "uri": ["m0300"]}},
    "ACCIPITRINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a0273"]}},
    "COLLARE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["c2379"]}},
    "ANAGNOSTES": {"-": {"morpho": ["n-s---mn1g"], "uri": ["a1854"]}},
    "BVXEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["b0783"]}},
    "OCTOGIENS": {"-": {"morpho": ["rp--------"], "uri": ["o0476"]}},
    "ARISTO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "BARBATORIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["b0123"]}},
    "GEMELLIPARVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["g0136"]}},
    "FAVSTE": {"-": {"morpho": ["rp--------"], "uri": ["50811"]}},
    "QVANTILLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["q0129"]}},
    "NVMVLARIOLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["118988"]}},
    "PATAGIATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0539"]}},
    "MORTIFERE": {"-": {"morpho": ["rp--------"], "uri": ["99479"]}},
    "HELIOPOLIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "CANTHARIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c1962"]}},
    "ATHLETICE": {"-": {"morpho": ["rp--------"], "uri": ["32584", "57497"]}},
    "INDVSIARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["i1370"]}},
    "CIRRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c1942"]}},
    "MARO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "SVPPLOSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s3776"]}},
    "ISTACTENVS": {"-": {"morpho": ["rp--------"], "uri": ["i2946"]}},
    "CORPVLENTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4090"]}},
    "PERVTILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p1763"]}},
    "NEMESIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "NAR": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "LANISTICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0237"]}},
    "INDVCTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i1357"]}},
    "IMPLVVIATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0581"]}},
    "ARABS": {
        "N": {"morpho": ["n-s----n3-"], "uri": []},
        "A": {"morpho": ["aps---mn3-"], "uri": []},
    },
    "PENELOPE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "DECEMPEDATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d0151"]}},
    "LVNVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l1184"]}},
    "CIRCVMSCRIPTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["38333"]}},
    "SEMICREMATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1036"]}},
    "ALCIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "TONSTRINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["130190"]}},
    "SOBRIE": {"-": {"morpho": ["rp--------"], "uri": ["34482"]}},
    "NIVATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0334"]}},
    "CASTERIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0768"]}},
    "ADONIS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "PHALERATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1850"]}},
    "RETROVERSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["39063"]}},
    "ALLACRIMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a1386"]}},
    "COENO": {"-": {"morpho": ["v1spia--1-"], "uri": ["36780"]}},
    "PERSVASTRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p1640"]}},
    "VIRGATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["u0761"]}},
    "BLENNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["b0446"]}},
    "DVCENVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["98251"]}},
    "PORCINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p2621"]}},
    "GOBIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["g0172"]}},
    "BABAECALVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["97745"]}},
    "PATAGIARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p0538"]}},
    "CALVOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["c0299"]}},
    "PROTOPRAXIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p4189"]}},
    "APPOTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2521"]}},
    "INGRATIFICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1704"]}},
    "PERINVISVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1251"]}},
    "INEXCITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1447"]}},
    "CINCTVTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c1641"]}},
    "RAPACIDA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["r0052"]}},
    "MELANVRVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["m0604"]}},
    "TRITICEIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t8400"]}},
    "SCOPVLOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0711"]}},
    "SIRPICVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["34247"]}},
    "SOPOLIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "AMYNTOR": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "PRAEGRESSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["122743"]}},
    "LAPIDATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["l0258"]}},
    "PERSONVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1606"]}},
    "INTERCINO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2286"]}},
    "INAVGVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0801"]}},
    "LATRATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["l0378"]}},
    "PEDALE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["p0701"]}},
    "TONSITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t0811"]}},
    "CAPSELLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0561"]}},
    "TVBVRCINABVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["130794"]}},
    "CRATHIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "SVBRANCIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s3043"]}},
    "VOLVPTVOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u1197"]}},
    "COMPLVRIENS": {"-": {"morpho": ["rp--------"], "uri": ["c2874"]}},
    "LATRINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0197"]}},
    "SEXIENS": {"-": {"morpho": ["rp--------"], "uri": ["s1525"]}},
    "FOETOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["f0768"]}},
    "DODRANTARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d2096"]}},
    "HYPASIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "ISIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "PETOSIRIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "APPOSCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a2517"]}},
    "XIPHIAS": {"-": {"morpho": ["n-s---mn1g"], "uri": ["x0015"]}},
    "CVRATVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c9976"]}},
    "SVBINVISVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2902"]}},
    "LABECVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0006"]}},
    "SIRMIO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "NERVOSE": {"-": {"morpho": ["rp--------"], "uri": ["30900"]}},
    "SOLIDE": {"-": {"morpho": ["rp--------"], "uri": ["42919"]}},
    "PACIFICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0020"]}},
    "MISSICVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m1119"]}},
    "MENSVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0747"]}},
    "ELPENOR": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "PARENTICIDA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["p0388"]}},
    "LATVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0390"]}},
    "POTITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p2779"]}},
    "PELOPS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "DESTIMVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1026"]}},
    "NARES": {"-": {"morpho": ["n-p---fn3i"], "uri": ["99538"]}},
    "RISOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["r0743"]}},
    "SITELLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1800"]}},
    "ASTITVO": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a3238"]}},
    "MVSSITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m1834"]}},
    "POLITE": {"-": {"morpho": ["rp--------"], "uri": ["56504"]}},
    "IATRALIPTA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["112840"]}},
    "COLLATIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c2384"]}},
    "SIREN": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "CLIO": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "OMPHALE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PATRICE": {"-": {"morpho": ["rp--------"], "uri": ["99834"]}},
    "NITEDVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["n0321"]}},
    "CONNVBIALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["105977"]}},
    "NESIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "QVERIBVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["q0185"]}},
    "CONCVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3112"]}},
    "SEVIRATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["s1510"]}},
    "SINGVLARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1744"]}},
    "ECHIDNA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["e0044"]}},
    "OBRODO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0177"]}},
    "AVSTERE": {"-": {"morpho": ["rp--------"], "uri": ["30261"]}},
    "APOPHORETVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2988"]}},
    "PERSEQVENS": {"-": {"morpho": ["aps---an3i"], "uri": ["99961"]}},
    "PIPERATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2123"]}},
    "LITTERATE": {"-": {"morpho": ["rp--------"], "uri": ["50890"]}},
    "SVTVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s9936"]}},
    "AHA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["35116"]}},
    "CARCERARIVS": {"2": {"morpho": ["n-s---mn2-"], "uri": ["c0610"]}},
    "AVLIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "CYMATILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["106736"]}},
    "HERO": {"N": {"morpho": ["n-s---mn3-"], "uri": ["112304"]}},
    "SVBREMIGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s3053"]}},
    "SALTVARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s0215"]}},
    "TERDECIENS": {"-": {"morpho": ["n-s---cn3-"], "uri": ["52565"]}},
    "COMMVRMVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2753"]}},
    "PERTVRPIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p1696"]}},
    "VBIVBI": {"-": {"morpho": ["rp--------"], "uri": ["45870"]}},
    "PERSOLLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p1593"]}},
    "GRACILIPES": {"-": {"morpho": ["aps---an3-"], "uri": ["g0479"]}},
    "SOMNICVLOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1956"]}},
    "CISTOPHORVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c1975"]}},
    "FERRICREPINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0395"]}},
    "ADIACENS": {"-": {"morpho": ["n-s---cn3-"], "uri": ["42572"]}},
    "CANDIDE": {"-": {"morpho": ["rp--------"], "uri": ["20608"]}},
    "QVINQVIENS": {"-": {"morpho": ["rp--------"], "uri": ["q0277"]}},
    "AGRICVLTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a9237"]}},
    "FIDELE": {"-": {"morpho": ["rp--------"], "uri": ["98537"]}},
    "DENVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["42429"]}},
    "SOBRINVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s1834"]}},
    "LOQVITOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["l1041"]}},
    "ECHENEIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["57700"]}},
    "LVDIFICABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["l1148"]}},
    "COCTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c7252"]}},
    "MESSORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0858"]}},
    "LATICLAVIVS": {"2": {"morpho": ["aps---mn1-"], "uri": ["l0359"]}},
    "HERMEROS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "CYME": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "AGGEMO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["39529"]}},
    "INVENVSTE": {"-": {"morpho": ["rp--------"], "uri": ["99116"]}},
    "PERREPTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p1530"]}},
    "NVMELLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["n0648"]}},
    "HISTORICE": {"-": {"morpho": ["n-s---fn1g"], "uri": ["h0465"]}},
    "MVGIL": {"-": {"morpho": ["n-s---mn3-"], "uri": ["m1474"]}},
    "PRAENVNTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["36802"]}},
    "THAIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "OBVIGILO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0353"]}},
    "TERNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["45691"]}},
    "COTISO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "NINIVE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CAPREIDA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0539"]}},
    "PERCITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["49680"]}},
    "BILYCHNIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["b0377"]}},
    "INCOMMODO": {"1": {"morpho": ["v1spia--1-"], "uri": ["i0984"]}},
    "PRAEFIDENS": {"-": {"morpho": ["aps---an3i"], "uri": ["p2986"]}},
    "MONSTRABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["m1345"]}},
    "PERLITTERATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1316"]}},
    "INSTINCTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i2127"]}},
    "SERVATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["100575"]}},
    "TONSTRINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t0815"]}},
    "CALCEOLARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c0195"]}},
    "CHIONE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SPONSIVNCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2275"]}},
    "QVOQVOMODO": {
        "1": {"morpho": ["rp--------"], "uri": ["36994"]},
        "2": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "QVADRAGENVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["100263"]}},
    "ALTARE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["a1480"]}},
    "OCCVLTATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["o0409"]}},
    "SEPTEMPEDALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s1280"]}},
    "QVADRIMVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["q0057"]}},
    "PERBACCHOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["p0914"]}},
    "BIMARITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["97772"]}},
    "COMISSATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c2583"]}},
    "BIFRONS": {"-": {"morpho": ["aps---an3i"], "uri": ["b0354"]}},
    "SPECVLABVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2122"]}},
    "MVRICIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1755"]}},
    "CAPERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["104228"]}},
    "COMPACISCOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["c2765"]}},
    "PERLEPIDE": {"-": {"morpho": ["rp--------"], "uri": ["p1302"]}},
    "EROGITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0907"]}},
    "ARCTOPHYLAX": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "DELECTOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132405"]}},
    "SVFFIMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["s3353"]}},
    "DORMITORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d2189"]}},
    "PRECATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p3491"]}},
    "XANTHIPPE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "STRICTVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s9924"]}},
    "CALLISTO": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "PERIVCVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1289"]}},
    "SCAZON": {"-": {"morpho": ["n-s---mn3i"], "uri": ["s0590"]}},
    "ANACHARSIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "VMBER": {"A": {"morpho": ["aps---mn1r"], "uri": ["101014"]}},
    "ACACIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a0218"]}},
    "FORDA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f1236"]}},
    "CONDVCTICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c3188"]}},
    "PORCELLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p4028"]}},
    "INEXPLETVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1476"]}},
    "GESTITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["g0272"]}},
    "INFECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1527"]}},
    "DESQVAMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1020"]}},
    "VRBANE": {"-": {"morpho": ["rp--------"], "uri": ["28429"]}},
    "PSAMATHE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PERSEPHONE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "INTERTEXO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2512"]}},
    "MASTRVCA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0438"]}},
    "EPILOGVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["e0741"]}},
    "SALIATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["s0155"]}},
    "BARDVS": {"2": {"morpho": ["aps---mn1-"], "uri": ["b0140"]}},
    "LVCTAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["l1117"]}},
    "MAIALIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["m0136"]}},
    "CATAPHRACTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["37015"]}},
    "COMMANIPVLARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["105478"]}},
    "AGAVE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "TRIPOLIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "CAPIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c0502"]}},
    "HAESITATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["h0052"]}},
    "FVCOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f1087"]}},
    "ECCERE": {"-": {"morpho": ["rp--------"], "uri": ["e1930"]}},
    "BVLE": {"-": {"morpho": ["n-s---fn1g"], "uri": ["b0721"]}},
    "GRAVATE": {"-": {"morpho": ["rp--------"], "uri": ["23701"]}},
    "REPROMITTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1299"]}},
    "NEPHELE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "FATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["110464"]}},
    "SOMNICVLOSE": {"-": {"morpho": ["rp--------"], "uri": ["28297"]}},
    "INCONCILIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1008"]}},
    "PERMIRVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1361"]}},
    "PAVONINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0628"]}},
    "SPOLIATRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s2251"]}},
    "VIRILIA": {"-": {"morpho": ["n-p---nn3i"], "uri": ["101160"]}},
    "IAPYX": {
        "A": {"morpho": ["aps---mn1-"], "uri": []},
        "N": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "OMINOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0697"]}},
    "GALBINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["g0025"]}},
    "INEQVITABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1427"]}},
    "SERISAPIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1372"]}},
    "AVGE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "OBSIGNATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["o0245"]}},
    "EXPISCOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["e1624"]}},
    "COEMPTIONALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["105262"]}},
    "FALSIDICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0133"]}},
    "REPRESSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["r0465"]}},
    "SAME": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CICARO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c1578"]}},
    "LANIENA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0221"]}},
    "MVLIONIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1605"]}},
    "FIMBRIAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["36073"]}},
    "GRAVASTELLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["98697"]}},
    "ANAS": {"1": {"morpho": ["n-s---fn3-"], "uri": ["a1883", "a9957"]}},
    "PRAESTIGIATRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3318"]}},
    "CONFOSSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["58253"]}},
    "RALLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0032"]}},
    "ATTHIS": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "EXPVLSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["e1692"]}},
    "SCALDIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "COMINCOMMODVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c2578"]}},
    "CHARONDAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "HALYS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "DELITIGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0584"]}},
    "REPROMISSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0471"]}},
    "NICEROS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "INIMICE": {"-": {"morpho": ["rp--------"], "uri": ["28446"]}},
    "CATVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0923"]}},
    "CHLORIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "VOLVPE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["131862"]}},
    "EBRIOLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e0017"]}},
    "TRANSFORMIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["t1012"]}},
    "AMPHITRITE": {"N": {"morpho": ["n-s---fn1g"], "uri": []}},
    "SORORICIDA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["s2025"]}},
    "VRCEATIM": {"-": {"morpho": ["rp--------"], "uri": ["u1244"]}},
    "EPVLO": {
        "1": {"morpho": ["v1spia--1-"], "uri": ["109192"]},
        "N": {"morpho": ["n-s----n1-"], "uri": []},
        "-": {"morpho": ["n-s----n2-"], "uri": []},
    },
    "IVNCEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i3035"]}},
    "INTERNECTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2411"]}},
    "DESVDASCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1050"]}},
    "CALIGARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0233"]}},
    "PEDISEQVA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p0715"]}},
    "PARAETACENE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PRIMITIAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["p3549"]}},
    "OBTVEOR": {"-": {"morpho": ["v1spid--2-"], "uri": ["o0323"]}},
    "FLEXIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0670"]}},
    "INCOMPOSITE": {"-": {"morpho": ["rp--------"], "uri": ["50370"]}},
    "DESPICOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["d8008"]}},
    "MIMIAMBI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["m1009"]}},
    "PACENSIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "NOMENCVLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["38959"]}},
    "SVBLATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["53020"]}},
    "MYCALE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ACOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a0402"]}},
    "ILLVSTRE": {"-": {"morpho": ["rp--------"], "uri": ["31076"]}},
    "PERNICIOSE": {"-": {"morpho": ["rp--------"], "uri": ["56632"]}},
    "SELLVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0983"]}},
    "MECASTOR": {"-": {"morpho": ["rp--------"], "uri": ["51291"]}},
    "FVRCIFERA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f1201"]}},
    "SVBBASILICANVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s2769"]}},
    "SCRIBILITA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0756"]}},
    "OCTIPES": {"-": {"morpho": ["aps---an3-"], "uri": ["o0467"]}},
    "CRANTOR": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "IMPETRIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["i0532"]}},
    "NISIBIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "QVADRVPEDANS": {
        "1": {"morpho": ["n-s---mn3i"], "uri": ["124280"]},
        "2": {"morpho": ["n-s---mn3i"], "uri": ["124280"]},
    },
    "EXLEX": {"-": {"morpho": ["aps---an3-"], "uri": ["e1513"]}},
    "DETRACTATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["46765"]}},
    "PERCONTOR": {"-": {"morpho": ["v1spid--1-"], "uri": []}},
    "DIAETA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["d1245"]}},
    "LABEO": {"N": {"morpho": ["n-s---mn3-"], "uri": ["l0012"]}},
    "RECESSIM": {"-": {"morpho": ["rp--------"], "uri": ["r0159"]}},
    "FICTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["f0477"]}},
    "COLLIQVEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["c2434"]}},
    "AGEMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["a1136"]}},
    "SANGVEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["100465"]}},
    "ROGATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["r0780"]}},
    "SEXTADECIMANI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["s1529"]}},
    "CVMBVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4583"]}},
    "PROMINVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3880"]}},
    "ASPELLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a3070"]}},
    "PARASITICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0350"]}},
    "INSPECIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2078"]}},
    "LIMVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0803"]}},
    "PHOLOE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PHLEGETHON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "BIBLIOPOLA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["b0306"]}},
    "CONCINNE": {"-": {"morpho": ["rp--------"], "uri": ["29644"]}},
    "PLVMOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2382"]}},
    "NARO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "EPIDIPNIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0725"]}},
    "ABSTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0192"]}},
    "RICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r0706"]}},
    "MARMORO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m0404"]}},
    "COMARCHVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c7553"]}},
    "DEDVCTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d0327"]}},
    "SAXATILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s0479"]}},
    "PRAEFODIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["p3001"]}},
    "RHETORICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r0211"]}},
    "THYMELE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "OCTONVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0482"]}},
    "FOLLITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0790"]}},
    "LOLIGVNCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["116469"]}},
    "CLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c7146"]}},
    "ANIMADVERSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a2031"]}},
    "HIPPOLYTE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "AVTONOE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "DIVINE": {"-": {"morpho": ["rp--------"], "uri": ["39463"]}},
    "APOLACTIZO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a2404"]}},
    "VERBEREVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0359"]}},
    "PRAECONIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2879"]}},
    "NOCTVINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0467"]}},
    "COMMENDATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c2642"]}},
    "INGIGNO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i1695"]}},
    "PVBLICITVS": {"-": {"morpho": ["rp--------"], "uri": ["p4375"]}},
    "INSTIMVLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i2125"]}},
    "DVCENARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["d2270"]}},
    "BABAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["47614"]}},
    "PHANTASIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p1857"]}},
    "EXENTERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1432"]}},
    "HORRIDE": {"-": {"morpho": ["rp--------"], "uri": ["33640"]}},
    "PYGISIACVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p4638"]}},
    "MESOCHORVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["m0841"]}},
    "ACCLINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0288"]}},
    "SMARIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s1821"]}},
    "BELLATVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["49454"]}},
    "EXEMPLARES": {"-": {"morpho": ["n-p---mn3i"], "uri": ["98401"]}},
    "RVBIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0846"]}},
    "PERDECORVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1007"]}},
    "TVBEROSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1691"]}},
    "VOLVPTABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["u1191"]}},
    "INDEMNATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1217"]}},
    "HELVO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["112200"]}},
    "CALYPSO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "CALLIOPE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "TISIPHONE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CONSVASOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c3634"]}},
    "OCVLISSIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0498"]}},
    "COLVMBVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c2530"]}},
    "SALVTIGERVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0244"]}},
    "LESBOS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "ALICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a1322"]}},
    "QVADRINGENTIENS": {"-": {"morpho": ["rp--------"], "uri": ["q0064"]}},
    "PERSIMILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p1589"]}},
    "ANCORARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["101224"]}},
    "PRAESTINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3323"]}},
    "HECATE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CYLLENE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "MENDICVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0722"]}},
    "PARTITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p0473"]}},
    "ODIOSE": {"-": {"morpho": ["rp--------"], "uri": ["39612"]}},
    "DECLAMITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0211"]}},
    "CALDICEREBRIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c9211"]}},
    "CIRCVMFORANEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c1757"]}},
    "SCABITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0504"]}},
    "ANTEGREDIOR": {"-": {"morpho": ["v1spid--3i"], "uri": ["a2134"]}},
    "DIPONDIARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["52224"]}},
    "MISCELLANEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0925"]}},
    "AGORANOMVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a1221"]}},
    "METICVLOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["117544"]}},
    "TRANSMISSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["t1059"]}},
    "IGNIPOTENS": {"-": {"morpho": ["aps---an3i"], "uri": ["i0142"]}},
    "OFFENSIVNCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["o0569"]}},
    "HEPATIA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["h0243"]}},
    "GENICVLATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["g0189"]}},
    "REMVRMVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1233"]}},
    "TRADITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["t0925"]}},
    "ERANVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["e0849"]}},
    "ILLICITATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i0231"]}},
    "PALAESTE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "BERENICE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "DEFENSITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0356"]}},
    "DEGRANDINAT": {"-": {"morpho": ["v3spia--1-"], "uri": ["d0448"]}},
    "PANCRATICE": {"-": {"morpho": ["rp--------"], "uri": ["p0206"]}},
    "MANTISSA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["99356"]}},
    "DVCTIM": {"-": {"morpho": ["rp--------"], "uri": ["d2283"]}},
    "AVGVRIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1366"]}},
    "SPHINX": {"N": {"morpho": ["n-s---fn3i"], "uri": []}},
    "DRYOPE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SCORTOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["s0741"]}},
    "INTERTVRBO": {"2": {"morpho": ["v1spia--1-"], "uri": ["i2518"]}},
    "BOTVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["b0569"]}},
    "TOPIARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["130199"]}},
    "QVIRITATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["q0307"]}},
    "LIMBOLARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["116245"]}},
    "POLENTARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2450"]}},
    "PROCRIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "SVSQVEDEQVE": {"-": {"morpho": ["rp--------"], "uri": ["129298"]}},
    "ADVENTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a0775"]}},
    "ATESTE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "QVIRIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "VIGILAX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["101137"]}},
    "APICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a2360"]}},
    "RVCTOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132707"]}},
    "SAEVITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s0099"]}},
    "HELICE": {"N": {"morpho": ["n-s---fn1g"], "uri": []}},
    "QVINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["q0239"]}},
    "NEPA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["n0218"]}},
    "PRAEPOSTERE": {"-": {"morpho": ["rp--------"], "uri": ["56197"]}},
    "PRAEBITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p2800"]}},
    "TRECENTIENS": {"-": {"morpho": ["rp--------"], "uri": ["t1136"]}},
    "SVFFLATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["s9978"]}},
    "MVLTIBIBVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1621"]}},
    "COPTOPLACENTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c3971"]}},
    "GELIDE": {"-": {"morpho": ["rp--------"], "uri": ["34674"]}},
    "PRAETERVEHO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p3392"]}},
    "HETAERIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["h0324"]}},
    "TRIENS": {"-": {"morpho": ["aps---an3i"], "uri": ["100955"]}},
    "VTIBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["u1326"]}},
    "DETONO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1102"]}},
    "DESTRICTE": {"-": {"morpho": ["rp--------"], "uri": ["22413"]}},
    "PROSPICIENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p9859"]}},
    "PIRENE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "COMMEMORABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c2632"]}},
    "PAVXILLATIM": {"-": {"morpho": ["rp--------"], "uri": ["p0645"]}},
    "DIGLADIOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["d1436"]}},
    "MANIFESTARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0287"]}},
    "DEFVNCTORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d0429"]}},
    "NEBRIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["n0132"]}},
    "PELAGIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0742"]}},
    "INHONORVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1745"]}},
    "CALVITIES": {"-": {"morpho": ["n-s---fn5-"], "uri": ["c0290"]}},
    "MOECHISSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m1196"]}},
    "INDOMABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1339"]}},
    "SCRIPTITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s0765"]}},
    "CONIECTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c3412"]}},
    "CONDORMISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3183"]}},
    "LEVCAS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ANTHRAX": {"N": {"morpho": ["n-s---mn3-"], "uri": ["a2201"]}},
    "TRICLINIARCHES": {"-": {"morpho": ["n-s---mn1-"], "uri": []}},
    "CYCLICVS": {"-": {"morpho": ["aps---mn1-"], "uri": []}},
    "PERBASIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p0915"]}},
    "EXDORSVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1411"]}},
    "EFFVLTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e0190"]}},
    "PERPLEXABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p1489"]}},
    "CROCOTARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4406"]}},
    "IMBERBVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0305"]}},
    "MEMPHIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "BESSALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["103613"]}},
    "BONATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["b0532"]}},
    "HISTRIONIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["h0473"]}},
    "VERBOSE": {"-": {"morpho": ["rp--------"], "uri": ["49690"]}},
    "OBSONATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["o9993"]}},
    "FVLCIPEDIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f1108"]}},
    "AMATRIX": {"-": {"morpho": ["aps---an3-"], "uri": ["41464"]}},
    "OBNOXIOSE": {"-": {"morpho": ["rp--------"], "uri": ["119187"]}},
    "NESAPIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0248"]}},
    "INTERDIVS": {"-": {"morpho": ["rp--------"], "uri": ["32001"]}},
    "RETRICES": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0605"]}},
    "DENTIFRANGIBVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d0724"]}},
    "PANTHERINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0256"]}},
    "INGRESSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i1713"]}},
    "PERISTASIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["p1277"]}},
    "PERINDVLGENS": {"-": {"morpho": ["aps---an3i"], "uri": ["p1239"]}},
    "QVADRILIBRIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["q0051"]}},
    "NIDVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["n0287"]}},
    "PERCALLESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p0932"]}},
    "PELLENE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CANTABVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0441"]}},
    "NAVTEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["n0124"]}},
    "DAVLIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "PRIMVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3560"]}},
    "VICTITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0613"]}},
    "INFECTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i1528"]}},
    "PRAEVARICATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p3433"]}},
    "GESTATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["g0262"]}},
    "ESCVLENTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e0980"]}},
    "BARCALA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["b0135"]}},
    "APPROMITTO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a2538"]}},
    "CALAMISTRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0163"]}},
    "MELINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["117285"]}},
    "ACTE": {"N": {"morpho": ["n-s---fn1g"], "uri": ["a0439"]}},
    "PILANVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p2056"]}},
    "DENASO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0679"]}},
    "COAEQVALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c2200"]}},
    "PHILOPATOR": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "INFRENO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1644"]}},
    "APELLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a2317"]}},
    "CALYMNE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PERTVSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["32158"]}},
    "CVRIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["21992"]}},
    "PVLVINARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p2248"]}},
    "DIPLOMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["d1651"]}},
    "IMPENDIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0467"]}},
    "DISPALESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1805"]}},
    "BILIX": {"-": {"morpho": ["aps---an3i"], "uri": ["b0374"]}},
    "CRVRALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c4456"]}},
    "FABRE": {"-": {"morpho": ["rp--------"], "uri": ["23206"]}},
    "AFFORMIDO": {"2": {"morpho": ["v1spia--1-"], "uri": ["a1113"]}},
    "TVSSICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t1824"]}},
    "SCRVPVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s0792"]}},
    "PERDOCTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["29756"]}},
    "CELOX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c1081"]}},
    "CONCVSTODIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["c3124"]}},
    "PAVXILLVLVM": {"2": {"morpho": ["n-s---nn2-"], "uri": ["p0435"]}},
    "LEPOS": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["l0571"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "ADMOLIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["a0683"]}},
    "SCISSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s0676"]}},
    "LODICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0982"]}},
    "DECIMO": {"-": {"morpho": ["rp--------"], "uri": ["106894"]}},
    "SACERDOTALIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["125810"]}},
    "VNOCVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["u1408"]}},
    "NIXVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["118771"]}},
    "VICIENS": {"-": {"morpho": ["rp--------"], "uri": ["38767"]}},
    "CLYMENE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PROFATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["123394"]}},
    "QVARTANI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["100283"]}},
    "FRATRICIDA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["f0941"]}},
    "DESCENSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0931"]}},
    "PHANTASMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["p1858"]}},
    "TAPANTA": {"-": {"morpho": ["n-s---nn--"], "uri": ["129542"]}},
    "HIPPODAME": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "INCORRVPTE": {"-": {"morpho": ["rp--------"], "uri": ["24068"]}},
    "TEPIDE": {"-": {"morpho": ["rp--------"], "uri": ["100848"]}},
    "AGNELLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a1186"]}},
    "ARCESSITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a2687"]}},
    "INTRIBVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2589"]}},
    "SPVRCATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["55175"]}},
    "CENATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["58103"]}},
    "TOGVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t0774"]}},
    "SIMVLATE": {"-": {"morpho": ["rp--------"], "uri": ["36520"]}},
    "CALAIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "ASTIPVLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a3236"]}},
    "HIPPOCRENE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PRASINATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3470"]}},
    "RVBICO": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "BALAENACEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["46907"]}},
    "IMMVTATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i3127"]}},
    "SCALPTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s0534"]}},
    "TVRBVLENTE": {"-": {"morpho": ["rp--------"], "uri": ["52207"]}},
    "SVBBLANDIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["s2771"]}},
    "DIRCE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "INTERIBI": {"-": {"morpho": ["rp--------"], "uri": ["i2351"]}},
    "CONCESSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2996"]}},
    "CONVADOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["c3854"]}},
    "TVNICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t1749"]}},
    "TVBERO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "PHOEBE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "AVRARIVS": {"2": {"morpho": ["aps---mn1-"], "uri": ["a3531"]}},
    "SVBDOMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s2834"]}},
    "PRAEQVAM": {"-": {"morpho": ["rp--------"], "uri": ["100120"]}},
    "HEPATARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["112233"]}},
    "CINCTICVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c1638"]}},
    "BOVILE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["b0572"]}},
    "CVIATIS": {"-": {"morpho": ["aps---cn3-"], "uri": ["42011"]}},
    "ANTIPOLIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "RENVTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["r7035"]}},
    "HAESITANTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["h0046"]}},
    "CVSTODITE": {"-": {"morpho": ["rp--------"], "uri": ["31925"]}},
    "INTESTATO": {"-": {"morpho": ["rp--------"], "uri": ["58136"]}},
    "MALEDICAX": {"-": {"morpho": ["aps---an3i"], "uri": ["m0161"]}},
    "PERFACETVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1094"]}},
    "RESECRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1328"]}},
    "OFFRENATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["56896"]}},
    "SOCIENNVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s1850"]}},
    "DERVNCINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0919"]}},
    "STROPHIARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s2626"]}},
    "RESTIPVLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["r1363"]}},
    "MILLESIMA": {"-": {"morpho": ["rp--------"], "uri": ["117609"]}},
    "BERBEX": {"-": {"morpho": ["n-s---nn3-"], "uri": ["50066"]}},
    "ASSVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a3209"]}},
    "GILLO": {"N": {"morpho": ["n-s---mn3-"], "uri": ["g0301"]}},
    "SCVTARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0821"]}},
    "REPVERASCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r1304"]}},
    "CAVILLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0959"]}},
    "RVSCINO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "BENEMORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["b0269"]}},
    "CONCILIATRICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c3017"]}},
    "LIBRARIOLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l0703"]}},
    "INSATVRABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1951"]}},
    "INCIPISSO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0908"]}},
    "PACIFICATORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0017"]}},
    "PRAETRVNCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3415"]}},
    "DEPVLSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0879"]}},
    "GVSTATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["111987"]}},
    "MELICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0620"]}},
    "VEPRECVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0339"]}},
    "PERSPICAX": {"-": {"morpho": ["aps---an3i"], "uri": ["p1616"]}},
    "MYROPOLA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["m1912"]}},
    "PERCVPIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["p0992"]}},
    "FIGVLARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["f0513"]}},
    "CARINARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c0644"]}},
    "ANNOTATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["37710"]}},
    "INCERTE": {"-": {"morpho": ["rp--------"], "uri": ["98880"]}},
    "LIQVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l0887"]}},
    "PROTINAM": {"-": {"morpho": ["rp--------"], "uri": ["p4180"]}},
    "COCHLEARE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["97913"]}},
    "PERCRVCIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p0986"]}},
    "IMPERSPICVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0509"]}},
    "PERPETVARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["121386"]}},
    "QVISQVILIAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["q0188"]}},
    "TECHNA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["29679"]}},
    "AMASTRIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "OBSERVANS": {"-": {"morpho": ["aps---an3i"], "uri": ["99657"]}},
    "ACERBO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0342"]}},
    "PROOEMIOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["p3937"]}},
    "PHILIPPOPOLIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "INCONSIDERATE": {"-": {"morpho": ["rp--------"], "uri": ["56805"]}},
    "VERENDA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["30144"]}},
    "DIOECETES": {"-": {"morpho": ["n-s---mn1g"], "uri": ["d1633"]}},
    "QVADRAGIENS": {"-": {"morpho": ["rp--------"], "uri": ["q0010"]}},
    "DEFLOCCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0387"]}},
    "CYMOTHOE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "POSTIDEA": {"-": {"morpho": ["rp--------"], "uri": ["p5001"]}},
    "CORYCVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c4168"]}},
    "NVGIGERVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["n0638"]}},
    "MILVA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m1002"]}},
    "IAMBEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0025"]}},
    "MALVGINENSIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "DEBELLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d0098"]}},
    "DISSORS": {"-": {"morpho": ["aps---an3i"], "uri": ["d1929"]}},
    "PLVSSCIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2400"]}},
    "FVLLONIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["f1129"]}},
    "MILO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "PELLICATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["120784"]}},
    "FLEMINA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["f0661"]}},
    "PRAECANTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p2810"]}},
    "PRAETERMISSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3384"]}},
    "SVBSICIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s3118"]}},
    "REVOLVBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["r0641"]}},
    "LAMENTARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0162"]}},
    "TROEZEN": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ACIDE": {"-": {"morpho": ["rp--------"], "uri": ["35191"]}},
    "PVRPVRISSATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p4569"]}},
    "OTHO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "DAPSILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["d0053"]}},
    "PERIODVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["25889"]}},
    "ADAEQVE": {"-": {"morpho": ["rp--------"], "uri": ["a0482"]}},
    "TIBVR": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "TRIMETER": {"1": {"morpho": ["n-s---mn2r"], "uri": ["t1219"]}},
    "INSTIGATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i2119"]}},
    "VXORCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u1392"]}},
    "NVMERABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["n0652"]}},
    "QVIPPINI": {"-": {"morpho": ["rp--------"], "uri": ["q0291"]}},
    "OBSEQVIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0216"]}},
    "MALACISSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m0150"]}},
    "VENVSTVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0335"]}},
    "OBTEXO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0305"]}},
    "SVBLESTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2932"]}},
    "ALIQVANTILLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["102177"]}},
    "IATRALIPTES": {"-": {"morpho": ["n-s---mn1g"], "uri": ["i0042"]}},
    "PSEVDOMENOS": {"-": {"morpho": ["n-s---mn2g"], "uri": ["p4304"]}},
    "EFFRENATE": {"-": {"morpho": ["rp--------"], "uri": ["46347"]}},
    "TEMPORI": {"-": {"morpho": ["rp--------"], "uri": ["46588"]}},
    "CIRCVMMINGO": {"-": {"morpho": ["rp--------"], "uri": ["54422"]}},
    "CVIAS": {"-": {"morpho": ["n-s---cn3-"], "uri": ["98059"]}},
    "CLIBANVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c2100"]}},
    "VITVLINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u1319"]}},
    "VINVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["101149"]}},
    "BOSTAR": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "LVRCHINABVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l1202"]}},
    "PATRIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0583"]}},
    "TRANSNATO": {"-": {"morpho": ["v1spia--1-"], "uri": ["t1068"]}},
    "VASA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["101065"]}},
    "TERVNCIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["48267"]}},
    "SARCINATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s3106"]}},
    "CROCOTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4405"]}},
    "MACRITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m0045"]}},
    "OBSTINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0268"]}},
    "CVNCTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["36548"]}},
    "RVMVSCVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["r0901"]}},
    "PROMVLSIS": {"-": {"morpho": ["n-s---nn3-"], "uri": ["56404"]}},
    "ALEC": {"-": {"morpho": ["n-s---nn3-"], "uri": ["102143"]}},
    "ABNVTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0113"]}},
    "ACIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "NOVIENS": {"-": {"morpho": ["rp--------"], "uri": ["n0582"]}},
    "INEPTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i1422"]}},
    "ILLOCABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0251"]}},
    "SPVTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s2313"]}},
    "SESQVIHORA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1457"]}},
    "SEPIOLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1264"]}},
    "QVAQVA": {"-": {"morpho": ["rp--------"], "uri": ["36527"]}},
    "PIAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["p2015"]}},
    "EXOBSECRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1517"]}},
    "VPVPA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u1229"]}},
    "DOMVSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d2166"]}},
    "DEPECVLATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["107235"]}},
    "PERPOLIO": {
        "-": {"morpho": ["v1spia--4-"], "uri": ["p1500"]},
        "2": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "INTERSTRINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2508"]}},
    "EVAN": {"N": {"morpho": ["n-s---mn--"], "uri": []}},
    "ZAMIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["z0003"]}},
    "AVDIENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a3439"]}},
    "ANNEXVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["53389"]}},
    "STERILICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2448"]}},
    "MORMYR": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m1413"]}},
    "FERICVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["98521"]}},
    "LARIFVGA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["l0305"]}},
    "PASTORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0534"]}},
    "ANATHYMIASIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["a1897"]}},
    "PRAELOQVOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["p3084"]}},
    "ELVVIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e0382"]}},
    "ANTIDHAC": {"-": {"morpho": ["rp--------"], "uri": ["43510"]}},
    "IMMEDICABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i0354"]}},
    "FENEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["110539"]}},
    "TALARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0061"]}},
    "CATILLO": {"2": {"morpho": ["n-s---mn3-"], "uri": ["c0899"]}},
    "DANISTA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["d0041"]}},
    "SPARVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s2074"]}},
    "EFFLICTIM": {"-": {"morpho": ["rp--------"], "uri": ["e0155"]}},
    "IEIVNIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0110"]}},
    "INCVRIOSE": {"-": {"morpho": ["rp--------"], "uri": ["24086"]}},
    "ABORTIO": {"1": {"morpho": ["v1spia--4-"], "uri": ["a0127"]}},
    "BALLIO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "HOMOLE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "COMMENDATRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2644"]}},
    "TETHYS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "SAGARIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "VINNVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0724"]}},
    "IMPLICISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0572"]}},
    "PRAEFRACTE": {"-": {"morpho": ["rp--------"], "uri": ["51844"]}},
    "LORVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l1055"]}},
    "PARRICIDALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p0438"]}},
    "DILATOR": {"-": {"morpho": ["v1spia--1-"], "uri": ["132437"]}},
    "NVPERVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0693"]}},
    "METATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["m0882"]}},
    "INDISPOSITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1308"]}},
    "SYNGRAPHVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s9015"]}},
    "FRENI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["98599"]}},
    "RHODOPE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "GIGERIA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["111633"]}},
    "LEMNISCATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0509"]}},
    "ADOLESCENTVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["36045"]}},
    "EXCETRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["e1304"]}},
    "CONSERVA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4629"]}},
    "IMPERDITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0486"]}},
    "BION": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "MELLILLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0656"]}},
    "INVVLNERATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2756"]}},
    "THESAVRARIVS": {"2": {"morpho": ["n-s---mn2-"], "uri": ["129973"]}},
    "MASCARPIO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["m0418"]}},
    "OCVLEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0496"]}},
    "COPS": {"-": {"morpho": ["aps---an3-"], "uri": ["c3969"]}},
    "HISPO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "MALEFACTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["m0171"]}},
    "PRAEDICATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p2926"]}},
    "COLLARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c2379"]}},
    "PRAXIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["p3486"]}},
    "RETRACTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["r7045"]}},
    "TRIFVR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["t1260"]}},
    "LAODICE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "DIIVNCTE": {"-": {"morpho": ["rp--------"], "uri": ["45599"]}},
    "CATAPHAGAS": {"-": {"morpho": ["n-s---mn1g"], "uri": ["c0832"]}},
    "DIERECTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d1377"]}},
    "SVTELA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s3887"]}},
    "INSVLARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["i2178"]}},
    "CONTEMPLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3701"]}},
    "FRONTO": {"N": {"morpho": ["n-s---mn3-"], "uri": ["f8034"]}},
    "ANTESCHOLANVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a2173"]}},
    "STERNVTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s2457"]}},
    "PLVMATILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p2369"]}},
    "SECVTVLEIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0901"]}},
    "CORBITA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4009"]}},
    "CLAVIGER": {"2": {"morpho": ["aps---mn1r"], "uri": ["c7933"]}},
    "OCCALESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["119405"]}},
    "VOCATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["u9989"]}},
    "DISPVTATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d1862"]}},
    "RECLAMITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1012"]}},
    "GAVSAPA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["98642"]}},
    "PROSERPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p4103"]}},
    "CISTELLVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c1969"]}},
    "APPENDICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a2483"]}},
    "POPPYSMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["p2587"]}},
    "NICANOR": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "CIVITATVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c2010"]}},
    "MINITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["25063"]}},
    "ALCITHOE": {"N": {"morpho": ["n-s---fn1g"], "uri": []}},
    "INFRENIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i1643"]}},
    "SACRICOLA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["s0057"]}},
    "DEPOPVLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d0822"]}},
    "ALCIMEDE": {"N": {"morpho": ["n-s---fn1g"], "uri": []}},
    "COLEVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c2358"]}},
    "COLLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c2385"]}},
    "LOCELLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l0963"]}},
    "ELEPHANTINE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SEVIR": {"-": {"morpho": ["n-s---mn2r"], "uri": ["s1508"]}},
    "LOQVENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l9888"]}},
    "RVBRICATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["45067"]}},
    "MINOS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "DEMARCHVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["d0614"]}},
    "ARENARIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["37872"]}},
    "PASSERCVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p0505"]}},
    "PHALANGAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["30824"]}},
    "CONCVSSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3122"]}},
    "CONTONAT": {"-": {"morpho": ["v3spia--1-"], "uri": ["c3772"]}},
    "EXPOSITICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e1658"]}},
    "LACTICVLOSVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["99197"]}},
    "LAMPETIE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SYROPHOENIX": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "TENELLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0288"]}},
    "PRAESEPIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["p3278"]}},
    "PRAEVOLO": {"2": {"morpho": ["v1spia--1-"], "uri": ["p3453"]}},
    "STELA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2412"]}},
    "CHRYSIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "DENVMERO": {"1": {"morpho": ["v1spia--1-"], "uri": ["d0736"]}},
    "PROVOCATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p4229"]}},
    "ADVERSARIA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["a1288"]}},
    "MVROBATHARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["m1783"]}},
    "PSEPHISMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["p4283"]}},
    "AREOLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a2779"]}},
    "HIRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["h0434"]}},
    "TERTIANA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["45277"]}},
    "INTERBIBO": {"2": {"morpho": ["v1spia--3-"], "uri": ["i2266"]}},
    "SVBLATE": {"-": {"morpho": ["rp--------"], "uri": ["39424"]}},
    "PRAESTES": {"A": {"morpho": ["n-s---cn3-"], "uri": ["p3315"]}},
    "QVINTVSDECIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["124460"]}},
    "TVATIM": {"-": {"morpho": ["rp--------"], "uri": ["t1484"]}},
    "PERLONGINQVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1317"]}},
    "CONTRITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["21699"]}},
    "PALMARIS": {"2": {"morpho": ["aps---cn3i"], "uri": ["p0128"]}},
    "OBSONOR": {"-": {"morpho": ["v1spid--1-"], "uri": []}},
    "EXPVTRESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["110005"]}},
    "INGRAVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1711"]}},
    "CALAMITOSE": {"-": {"morpho": ["rp--------"], "uri": ["20546"]}},
    "MELANIPPE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "RVGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1487"]}},
    "COALO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c0283"]}},
    "REDORMIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["r1097"]}},
    "AMPVLLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a1782"]}},
    "AGINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a1169"]}},
    "MAVORS": {"N": {"morpho": ["n-s---mn3i"], "uri": ["m0498"]}},
    "CONDICTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c3157"]}},
    "OLLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["o0672"]}},
    "SINGVLTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s1749"]}},
    "FABALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["f0004"]}},
    "AFFLICTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a1102"]}},
    "MEROE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "IMBITO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0307"]}},
    "BATVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["50261"]}},
    "LATERICVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["55213"]}},
    "ASCRIPTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["a3038"]}},
    "SACCELLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s0025"]}},
    "ANAPAESTVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a2681"]}},
    "SYENE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CANACE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SVBREPTICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s7608"]}},
    "TRIBVARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1172"]}},
    "CERNVVS": {"2": {"morpho": ["aps---mn1-"], "uri": ["c1249"]}},
    "OFFLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["99698"]}},
    "GERVLIFIGVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["g0250"]}},
    "PERACESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p0871"]}},
    "CAVPONVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0988"]}},
    "INTERTRAHO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i2514"]}},
    "INVESTIGATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i2669"]}},
    "FORENSIA": {"-": {"morpho": ["n-p---nn3i"], "uri": ["98580"]}},
    "AVGVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a3490"]}},
    "EXPALLIATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["98419"]}},
    "BVCCO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["b0684"]}},
    "PVMICEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p4513"]}},
    "GINGILIPHVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["g0303"]}},
    "RECVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r0213"]}},
    "IVLIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i3029"]}},
    "INFLETVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1608"]}},
    "AVDACVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a3435"]}},
    "PVRGAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["p4545"]}},
    "CATAPVLTARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0842"]}},
    "DETVMESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d1126"]}},
    "MVNITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m1747"]}},
    "ELINGVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0330"]}},
    "LVCVLENTE": {"-": {"morpho": ["rp--------"], "uri": ["43832"]}},
    "MOLOCHINARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["117840"]}},
    "PHYLLIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "HONORABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["h0540"]}},
    "SPECTANS": {"-": {"morpho": ["aps---an3i"], "uri": ["127594"]}},
    "PECVLATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p0684"]}},
    "TONSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["54607"]}},
    "MALACVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0153"]}},
    "NOCVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0473"]}},
    "COPVLATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["c9829"]}},
    "VRSINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u1273"]}},
    "IVRATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i3051"]}},
    "SYCOPHANTA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["s3903"]}},
    "PETASVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p1793"]}},
    "DEFRAVDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0408"]}},
    "PVGIVNCVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p4421"]}},
    "PERSEPOLIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "AGANIPPE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ZEVXIS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "PARONYCHIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["120499"]}},
    "COCCINEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c2244"]}},
    "LACHESIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "PSYCHROLVTA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["123930"]}},
    "EXCATARISSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1284"]}},
    "LVPATRIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l1191"]}},
    "INARIME": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "LONGIVSCVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l1027"]}},
    "VACIVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0012"]}},
    "PLEIONE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "VITISATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["u0870"]}},
    "EDITICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e0091"]}},
    "OBTICESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0307"]}},
    "MELIE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "FVRFVR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["f1207"]}},
    "PINNVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p2106"]}},
    "COMATORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c2555"]}},
    "COMPOTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c7888"]}},
    "DELIQVIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0575"]}},
    "ARCVLARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a2760"]}},
    "ILLECEBROSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0208"]}},
    "SCAPHA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0561"]}},
    "INCOMMODESTICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0982"]}},
    "PVGNEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p4430"]}},
    "FORTVNATE": {"-": {"morpho": ["rp--------"], "uri": ["31460"]}},
    "SARGVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s0390"]}},
    "LEMMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["l0508"]}},
    "FVGIENS": {"2": {"morpho": ["aps---an3i"], "uri": ["98613"]}},
    "COLONICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c2496"]}},
    "ARIS": {"N": {"morpho": ["n-s---fn3-"], "uri": ["a2839"]}},
    "PERNECESSE": {"-": {"morpho": ["rp--------"], "uri": ["99945"]}},
    "COGNOMINIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c2319"]}},
    "CRYPTOPORTICVS": {"-": {"morpho": ["n-s---fn4-"], "uri": ["c4474"]}},
    "EXTEXO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1859"]}},
    "PERBENE": {"-": {"morpho": ["rp--------"], "uri": ["p0918"]}},
    "PROPEXVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3974"]}},
    "PROSECTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["123670"]}},
    "FASTOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0247"]}},
    "IMPERCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["i0484"]}},
    "VINEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u1196"]}},
    "FVRAX": {"-": {"morpho": ["aps---an3i"], "uri": ["f1199"]}},
    "FREMOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["f0963"]}},
    "DEFERVESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0365"]}},
    "FRATERNE": {"-": {"morpho": ["rp--------"], "uri": ["47228"]}},
    "OCCLAMITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0398"]}},
    "HECATOMPYLOS": {"N": {"morpho": ["n-s---fn2g"], "uri": []}},
    "VERSIPELLIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["101105"]}},
    "DESVEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["d1051"]}},
    "EBOREVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e0012"]}},
    "OCCENTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0379"]}},
    "SORICINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2020"]}},
    "EVILESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e1093"]}},
    "ABSVMEDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0201"]}},
    "ELECEBRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["e0271"]}},
    "NVNDINALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["n0678"]}},
    "ADMVTILO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0698"]}},
    "ANTICATO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ESVRIALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["e0991"]}},
    "AMASIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a1593"]}},
    "THEMIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "VENDITATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["u0269"]}},
    "TEREBINTHINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t0330"]}},
    "GYPSATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["47713"]}},
    "ARTVTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2984"]}},
    "DVOETVICESIMANI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["43020"]}},
    "DELETRIX": {"-": {"morpho": ["aps---an3-"], "uri": ["40310"]}},
    "IMITAMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["i0329"]}},
    "NVBECVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["n0595"]}},
    "MATTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["117158"]}},
    "MOERIS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "DECEMVIRATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["33460"]}},
    "PACISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["99773"]}},
    "QVINQVAGIENS": {"-": {"morpho": ["rp--------"], "uri": ["q0247"]}},
    "CATOMIDIO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c0910"]}},
    "PAVPERVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["99845"]}},
    "TERVENEFICVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["t0414"]}},
    "AMOENA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["a1710"]}},
    "SVPERSTITIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s3698"]}},
    "TVBVLATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1701"]}},
    "SVFFLAMINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s3362"]}},
    "INGRATIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i1703"]}},
    "VALENTVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u0049"]}},
    "VTVT": {"-": {"morpho": ["rp--------"], "uri": ["u9895"]}},
    "BOGVD": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "SVAVITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s2718"]}},
    "TAVRVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["t0161"]}},
    "ORARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0927"]}},
    "DEPREHENSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["d0858"]}},
    "MASTIGIA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["m0435"]}},
    "PISTORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2173"]}},
    "TVGIO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "MACILENTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0036"]}},
    "SCORDALIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0715"]}},
    "MONEDVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m1288"]}},
    "PVLCHRALIA": {"-": {"morpho": ["n-p---nn3i"], "uri": ["p4436"]}},
    "PERVAGATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["56835"]}},
    "AGAGA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a1124"]}},
    "AMATORIE": {"-": {"morpho": ["rp--------"], "uri": ["54368"]}},
    "INDECOR": {"-": {"morpho": ["aps---an3-"], "uri": []}},
    "OPTATO": {"-": {"morpho": ["rp--------"], "uri": ["35243"]}},
    "DEBATVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["106820"]}},
    "CERCYRVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["29504"]}},
    "TYRO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "DILVTE": {"-": {"morpho": ["rp--------"], "uri": ["36383"]}},
    "OVILLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["o0647"]}},
    "OPIPARE": {"-": {"morpho": ["rp--------"], "uri": ["25578"]}},
    "EDENTVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e0077"]}},
    "SVFFRINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3385"]}},
    "SEMIDOCTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1046"]}},
    "VSVRARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["u1308"]}},
    "CRYSTALLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["54068"]}},
    "IVVENOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["i3104"]}},
    "REFORMATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["r0263"]}},
    "ROBOREVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["r0768"]}},
    "RESTITVTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["r0558"]}},
    "ALOGVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a1462"]}},
    "CLIMACTERICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c2110"]}},
    "ADLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["19388"]}},
    "TROPICA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["100977"]}},
    "SVBMISSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s2978"]}},
    "CYMBALISTRIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4800"]}},
    "SEVERITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["s1506"]}},
    "LYRICA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["l1284"]}},
    "DISCOLORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d1727"]}},
    "ADMISSARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0670"]}},
    "SOMNIOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132794"]}},
    "CONTINVE": {"-": {"morpho": ["rp--------"], "uri": ["44662"]}},
    "GERRAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["98665"]}},
    "PVNITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p4533"]}},
    "MEROPE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "INCRETO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1119"]}},
    "EPASTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["34427"]}},
    "LITICEN": {"-": {"morpho": ["n-s---mn3-"], "uri": ["l0909"]}},
    "MORIO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["m1409"]}},
    "INCALLIDE": {"-": {"morpho": ["rp--------"], "uri": ["31088"]}},
    "STEROPE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "BAETIS": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "DIRIBITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["d1678"]}},
    "CENATIVNCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c1099"]}},
    "FESTIVE": {"-": {"morpho": ["rp--------"], "uri": ["57382"]}},
    "FEMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["32193"]}},
    "MIMVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m1013"]}},
    "EMISSICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e0472"]}},
    "SVBSCRIPTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s3099"]}},
    "CAESICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0133"]}},
    "REBITO": {"-": {"morpho": ["v1spia--3-"], "uri": ["r0985"]}},
    "EFFERTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["52425"]}},
    "OBSERVANTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["o0213"]}},
    "TRANSENNA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t0992"]}},
    "MATTEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0483"]}},
    "SEPTEMVIRI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["100557"]}},
    "REDEMPTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1079"]}},
    "ELECTA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["98302"]}},
    "SEXAGIENS": {"-": {"morpho": ["rp--------"], "uri": ["s1517"]}},
    "POSITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p2686"]}},
    "MAGMENTARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0107"]}},
    "CECROPS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "OBVALLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0338"]}},
    "RHOMBVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["r0689"]}},
    "PIETATICVLTRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2036"]}},
    "SECTARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0868"]}},
    "SEPTEMVIRATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["s1286"]}},
    "CYRENE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PERDOLESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p2043"]}},
    "ITYS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "INIVRIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1799"]}},
    "ANATICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a1898"]}},
    "OLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0637"]}},
    "ARMENTALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["a2865"]}},
    "CELERE": {"-": {"morpho": ["rp--------"], "uri": ["20790"]}},
    "CVCVMVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4512"]}},
    "SVBALBVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2740"]}},
    "ENYO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ALCYONE": {"N": {"morpho": ["n-s---fn1g"], "uri": ["a1320"]}},
    "DVCTITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d2285"]}},
    "TVRPIFICATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1803"]}},
    "DISCORDABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["d1736"]}},
    "BYBLOS": {"N": {"morpho": ["n-s---mn2g"], "uri": []}},
    "PHASIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "EXAMBVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1211"]}},
    "DRVIDA": {"-": {"morpho": ["n-s---mn1-"], "uri": []}},
    "DEIERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0484"]}},
    "CYDIPPE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CONSECTATRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3494"]}},
    "CLOACALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c2134"]}},
    "CONCACO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2943"]}},
    "VAVATO": {"-": {"morpho": ["n-s---mn3-"], "uri": ["u9156"]}},
    "SVBTERDVCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s3182"]}},
    "FRVSTILLATIM": {"-": {"morpho": ["rp--------"], "uri": ["f1062"]}},
    "ADAMBVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a0498"]}},
    "EDISSERTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0090"]}},
    "RHOEA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r0685"]}},
    "CHANE": {"-": {"morpho": ["n-s---fn1g"], "uri": ["104705"]}},
    "DELINGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["d0570"]}},
    "CONGREDIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["105941"]}},
    "PLVMEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2378"]}},
    "OBVSTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0360"]}},
    "ASSVLATIM": {"-": {"morpho": ["rp--------"], "uri": ["a3198"]}},
    "COMMISEROR": {"-": {"morpho": ["v1spid--1-"], "uri": ["c2689"]}},
    "ENOTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0619"]}},
    "FERRVMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["f0404"]}},
    "CORBVLO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "HARIOLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["h0118"]}},
    "COMPRANSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c2892"]}},
    "OCVLATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0495"]}},
    "OBEDIENTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["40144"]}},
    "SALAMANDRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0140"]}},
    "VERVINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u0468"]}},
    "CONTOLLO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c3770"]}},
    "CAPPARIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["c0532"]}},
    "DVVMVIRATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["d2408"]}},
    "DESIPIENS": {"-": {"morpho": ["aps---an3i"], "uri": ["22381"]}},
    "REDVCTOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["r0245"]}},
    "COLLACRIMO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2373"]}},
    "IMPVRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["48493"]}},
    "MVFRIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["m1472"]}},
    "OPIME": {"-": {"morpho": ["rp--------"], "uri": ["99725"]}},
    "PVTILLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p4611"]}},
    "AEGLE": {"N": {"morpho": ["n-s---fn1g"], "uri": []}},
    "CIRCITOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c1683"]}},
    "LVSCINIOLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l1210"]}},
    "SITVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0354"]}},
    "MODICO": {"-": {"morpho": ["rp--------"], "uri": ["99451"]}},
    "BALITANS": {"-": {"morpho": ["aps---an3i"], "uri": ["97748"]}},
    "INCLITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["56181"]}},
    "AXICIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["55149"]}},
    "ERVILIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["e0945"]}},
    "INVOLVCRE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["i0334"]}},
    "CACCABVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c0021"]}},
    "NOCTIFER": {"A": {"morpho": ["n-s---mn2r"], "uri": ["n0457"]}},
    "SVSPICIOSE": {"-": {"morpho": ["rp--------"], "uri": ["53628"]}},
    "CIRCVMLVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c1804"]}},
    "TVRBELLAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["100993"]}},
    "AMYMONE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "STIGMATIAS": {"-": {"morpho": ["n-s---mn1-"], "uri": []}},
    "DRAPETA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["d2220"]}},
    "NARRATIVNCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["n0037"]}},
    "SALTATORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0207"]}},
    "SVBOLFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["s3017"]}},
    "CARIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c0651"]}},
    "PERTRECTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["99976"]}},
    "APICATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2357"]}},
    "DISPVDET": {"-": {"morpho": ["v3spia--2-"], "uri": ["d1854"]}},
    "LAOCOON": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ALOGIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a1460"]}},
    "EVRYDICE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SVBSVLTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s3162"]}},
    "LARVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["l0313"]}},
    "LYMPHOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["116734"]}},
    "INCONGRVENS": {"-": {"morpho": ["aps---an3i"], "uri": ["i1020"]}},
    "ZOTHECVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["z0049"]}},
    "CONCERPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["c2987"]}},
    "DEVOTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d1189"]}},
    "INERVDITVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1438"]}},
    "PERTRIBVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1687"]}},
    "BARCE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "BAVCIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "PROLVSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3859"]}},
    "APODIXIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["a2400"]}},
    "SVBSIGNO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s3127"]}},
    "COMPETITRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c2835"]}},
    "AVSPICATO": {"-": {"morpho": ["rp--------"], "uri": ["28664"]}},
    "SEMIVSTVLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s1176"]}},
    "SVBFODIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["51114"]}},
    "CHROMIS": {
        "-": {"morpho": ["n-s---mn3i"], "uri": ["c1519"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "MACEDO": {
        "N": {"morpho": ["n-s---mn3-"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "PRAESEGMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["p3268"]}},
    "DIONE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "OCIVS": {"-": {"morpho": ["rp--------"], "uri": ["49188"]}},
    "POTIVNCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p2781"]}},
    "CVCVMA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4507"]}},
    "INSPOLIATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2102"]}},
    "PERMAESTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1333"]}},
    "PVELLARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p4390"]}},
    "SEXCENTI": {"-": {"morpho": ["app---mn1-"], "uri": ["100594"]}},
    "IMPERITE": {"-": {"morpho": ["rp--------"], "uri": ["34611"]}},
    "INFACVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1502"]}},
    "PRAECVRSORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2896"]}},
    "AGIS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "PRAEGVSTATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p3040"]}},
    "EBVRNVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e0029"]}},
    "REDAVSPICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1078"]}},
    "ERYTHINVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["e9960"]}},
    "LETHE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "MAESTITVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["m0082"]}},
    "INFORTVNATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1630"]}},
    "SORTILEGVS": {"2": {"morpho": ["n-s---mn2-"], "uri": ["127537"]}},
    "ANSERINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a2100"]}},
    "COMPES": {"-": {"morpho": ["aps---an3-"], "uri": ["105601"]}},
    "CALAMIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "SVSPICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["100807"]}},
    "INSTIGATRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i2121"]}},
    "IEIVNE": {"-": {"morpho": ["rp--------"], "uri": ["29788", "99161"]}},
    "CONGIALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c3367"]}},
    "CONDECET": {"-": {"morpho": ["v3spia--2-"], "uri": ["c3128"]}},
    "CERINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c1239"]}},
    "LOTIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "CATAPLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c0839"]}},
    "EXCVRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["38433"]}},
    "SOPHOS": {"-": {"morpho": ["n-s---mn2g"], "uri": ["s1988"]}},
    "PERGRAVIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p1187"]}},
    "SYMBOLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s3921"]}},
    "VISCATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["46065"]}},
    "SAGATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2504"]}},
    "FRVSTRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["53073"]}},
    "CORDATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c4021"]}},
    "COGITATE": {"-": {"morpho": ["rp--------"], "uri": ["42902"]}},
    "SITICEN": {"-": {"morpho": ["n-s---mn3-"], "uri": ["s1801"]}},
    "PROLAPSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3823"]}},
    "DECLINIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["d0219"]}},
    "SERRVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s1410"]}},
    "PESSVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p1773"]}},
    "EXASCIATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["109586"]}},
    "OPPANGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0856"]}},
    "INTERMINOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["i2391"]}},
    "BELLVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["b0243"]}},
    "MORDEX": {"-": {"morpho": ["aps---an3-"], "uri": ["99475"]}},
    "TRAGVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["t0951"]}},
    "SARDIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "LIRIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "AVTHEPSA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a3611"]}},
    "CAPREAGINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c0538"]}},
    "SARTOR": {"2": {"morpho": ["n-s---mn3-"], "uri": ["s0407"]}},
    "SVBSIGNANVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s3125"]}},
    "NAPE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PERCVRIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0993"]}},
    "PERACERBVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0870"]}},
    "EOS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "SIPHVNCVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s1770"]}},
    "CALTHVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["55570"]}},
    "TRAGICOMOEDIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["t0937"]}},
    "SEPTEMVIRALIS": {
        "2": {"morpho": ["aps---cn3i"], "uri": ["s1285"]},
        "1": {"morpho": ["n-s----n2-"], "uri": []},
    },
    "ADRADO": {"-": {"morpho": ["v1spia--3-"], "uri": ["a0757"]}},
    "THETIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "VERBERITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["u0360"]}},
    "PECOROSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0663"]}},
    "FVNDITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f1173"]}},
    "PERLEVIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p1304"]}},
    "SINAPE": {"-": {"morpho": ["n-s---nn3-"], "uri": ["55164"]}},
    "FRVSTVLENTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f1072"]}},
    "ECDICVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["e0040"]}},
    "SOLITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s1914"]}},
    "PERCOLOPO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p0968"]}},
    "ASSIDVO": {"2": {"morpho": ["v1spia--1-"], "uri": ["a3162"]}},
    "PALVDOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0172"]}},
    "ANTHIAS": {"-": {"morpho": ["n-s---mn1g"], "uri": ["a2192"]}},
    "RVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1500"]}},
    "COMMETO": {"1": {"morpho": ["v1spia--1-"], "uri": ["c3528"]}},
    "SEMIRAMIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "SESTERTIARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s1497"]}},
    "LACINIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["l0064"]}},
    "ADVESPERASCIT": {"-": {"morpho": ["v3spia--3-"], "uri": ["a0797"]}},
    "HEDEROSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0176"]}},
    "IMPRODICTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0645"]}},
    "ZAPLVTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["42061"]}},
    "COMMENTARIOLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["42642"]}},
    "OFFVCIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["o0593"]}},
    "PERATVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["99888"]}},
    "INTROVOCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i9927"]}},
    "METVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["m0916"]}},
    "CLANDESTINO": {"-": {"morpho": ["rp--------"], "uri": ["30248"]}},
    "MACESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["m0018"]}},
    "MERACVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0770"]}},
    "CYPEROS": {"-": {"morpho": ["n-s---mn2g"], "uri": ["c4843"]}},
    "FATVOR": {"1": {"morpho": ["v1spid--1-"], "uri": ["f0270"]}},
    "INFELICO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1534"]}},
    "VINDEMITOR": {
        "-": {"morpho": ["n-s---mn3-"], "uri": ["32540"]},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "NAVFRAGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["n0087"]}},
    "LVSTROR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132557"]}},
    "DEPLORABVNDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d0804"]}},
    "PEDICELLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["120738"]}},
    "VENTVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["$0222"]}},
    "ARQVATVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["97683"]}},
    "ANIMVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["a2045"]}},
    "ATTACTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["a8909"]}},
    "DAPINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0051"]}},
    "AMASIVNCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a1592"]}},
    "BEROE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "BIFARIAM": {"-": {"morpho": ["rp--------"], "uri": ["20431"]}},
    "FOEDIFRAGVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f0760"]}},
    "BACALVSIAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["b0009"]}},
    "ALCESTIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "OCCILLO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0390"]}},
    "ANTIQVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["a2252"]}},
    "PEDISEQVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0714"]}},
    "VERNILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["u0402"]}},
    "OBROGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["o0179"]}},
    "SALVE": {"-": {"morpho": ["n-s---nn3-"], "uri": ["42820"]}},
    "VNAETVICESIMANI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["131699"]}},
    "REMELIGO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0367"]}},
    "MILVINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m1004"]}},
    "PRAEFISCINE": {"-": {"morpho": ["rp--------"], "uri": ["30720"]}},
    "MVNDO": {"-": {"morpho": ["v1spia--1-"], "uri": ["m1713"]}},
    "ASTYANAX": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ACCVSATRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a0327"],}},
    "PRAEDIATORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p2918"]}},
    "AGENOR": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "INCONDITE": {"-": {"morpho": ["rp--------"], "uri": ["29218"]}},
    "FVRTIFICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["f1232"]}},
    "CORRVPTE": {"-": {"morpho": ["rp--------"], "uri": ["37879"]}},
    "HIRVDO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["h0456"]}},
    "PERTRISTIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p1689"]}},
    "LANIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["l0239"]}},
    "FAVTRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["f0289"]}},
    "PVGILATVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["p9928"]}},
    "STRATEGVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s2571"]}},
    "MANVLEARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["m0351"]}},
    "INDVCVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i1360"]}},
    "SOLEARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["42709"]}},
    "FLICTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["f9209"]}},
    "POMPILVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p2543"]}},
    "TORRIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["t0867"]}},
    "PHASELVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p1874"]}},
    "APPARATE": {"-": {"morpho": ["rp--------"], "uri": ["19983"]}},
    "EGVRGITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e0225"]}},
    "RETICVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["54959"]}},
    "ARVINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a2989"]}},
    "IRACVNDE": {"-": {"morpho": ["rp--------"], "uri": ["34398"]}},
    "PROMISSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3884"]}},
    "CANDELABRVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c0385"]}},
    "OFFIRMATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["33825"]}},
    "DVODETRICIENS": {"-": {"morpho": ["rp--------"], "uri": ["d2358"]}},
    "PERATTENTE": {"-": {"morpho": ["rp--------"], "uri": ["30512"]}},
    "PRAESAGIOR": {"-": {"morpho": ["v1spid--4-"], "uri": ["132791"]}},
    "CONSILIATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c6533"]}},
    "PACATE": {"-": {"morpho": ["rp--------"], "uri": ["99771"]}},
    "FVLCIMEN": {"-": {"morpho": ["n-s---nn3-"], "uri": ["f1104"]}},
    "RAPA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r0051"]}},
    "PIGNEROR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132642"]}},
    "POSTILENA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p5002"]}},
    "CALCITRO": {
        "1": {"morpho": ["n-s---mn3-"], "uri": ["c0201"]},
        "2": {"morpho": ["n-s---mn3-"], "uri": ["c0201"]},
    },
    "SILVESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["s1663"]}},
    "EXANIMALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["e1225"]}},
    "GALERITVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["g0042"]}},
    "PERLECEBRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["99932"]}},
    "PRAEPROPERVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p3217"]}},
    "MALEVOLENS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["116898"]}},
    "OMINATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["o0695"]}},
    "DECIENS": {"-": {"morpho": ["rp--------"], "uri": ["d0192"]}},
    "PERSCRVTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["99960"]}},
    "CHALCIOPE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SVPELLECTICARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["100779"]}},
    "CAMELLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0316"]}},
    "LEPAS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["l0562"]}},
    "MODIALIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["m1170"]}},
    "SORBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s2000"]}},
    "INCLYTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["45160"]}},
    "ANATINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a2842"]}},
    "DEARTVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["d0089"]}},
    "REPORRIGO": {"2": {"morpho": ["v1spia--3-"], "uri": ["r1290"]}},
    "LINGVOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["l0829"]}},
    "MORTVALIA": {"-": {"morpho": ["n-p---nn3i"], "uri": ["m1436"]}},
    "TANAIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "PERIVCVNDE": {"-": {"morpho": ["rp--------"], "uri": ["36381"]}},
    "SEMELE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ERIGONE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "FRVNISCOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["f1059"]}},
    "NEVTRO": {"-": {"morpho": ["rp--------"], "uri": ["37085"]}},
    "ADVENTICIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a0773"]}},
    "NAVCLERICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["n0079"]}},
    "FERRVMINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["f0406"]}},
    "RAMENTA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["r0035"]}},
    "CALPE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "PROELIARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p3744"]}},
    "PIABILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["p2011"]}},
    "ARGVTOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132257"]}},
    "SCORTEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0219"]}},
    "PROMVLSIDARE": {"-": {"morpho": ["n-s---nn3i"], "uri": ["100193"]}},
    "ANTIGONE": {"N": {"morpho": ["n-s---fn1g"], "uri": []}},
    "GERENS": {"-": {"morpho": ["aps---an3-"], "uri": ["35019"]}},
    "SVPPARASITOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132800"]}},
    "TRYGONVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["t1492"]}},
    "EXAGGERATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["98386"]}},
    "OCLOPETA": {"-": {"morpho": ["n-s---mn1-"], "uri": ["99677"]}},
    "CONDVCIBILIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["c3185"]}},
    "PRAECVRSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2895"]}},
    "INDIREPTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1291"]}},
    "SPORTELLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s2283"]}},
    "PEROPTATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1432"]}},
    "EMPTICIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e0518"]}},
    "ILLEX": {"1": {"morpho": ["n-s---mn3-"], "uri": ["i0218"]}},
    "TRICAE": {"-": {"morpho": ["n-p---fn1-"], "uri": ["t1191"]}},
    "BASILICE": {"-": {"morpho": ["n-s---fn1g"], "uri": ["b0157"]}},
    "DVRIVSCVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d2399"]}},
    "LYRISTES": {"-": {"morpho": ["n-s---mn1g"], "uri": ["l1285"]}},
    "PVLTARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p4488"]}},
    "INTERAESTVO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i2256"]}},
    "FALSILOQVVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["110335"]}},
    "PEDIS": {"-": {"morpho": ["n-s---cn3i"], "uri": ["p9709"]}},
    "QVADRIPARTITO": {"-": {"morpho": ["rp--------"], "uri": ["124267"]}},
    "NONAGIENS": {"-": {"morpho": ["rp--------"], "uri": ["n0515"]}},
    "SVBACCVSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s2721"]}},
    "SCELESTE": {"-": {"morpho": ["rp--------"], "uri": ["36786"]}},
    "MICARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0952"]}},
    "SEXTVSDECIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["34258"]}},
    "OPHTHALMIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["54498"]}},
    "INDVBITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1351"]}},
    "INSTANTIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["i9843"]}},
    "STRABONVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["100704"]}},
    "NVPTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["119020"]}},
    "INTERCALARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["i2269"]}},
    "PRAEFOCO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p3000"]}},
    "OBDORMIO": {"-": {"morpho": ["v1spia--4-"], "uri": ["o0026"]}},
    "SEPTIMANI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["51650"]}},
    "REMANSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["r0357"]}},
    "PERENDIE": {"-": {"morpho": ["rp--------"], "uri": ["p1072"]}},
    "GRADATVS": {"2": {"morpho": ["n-s---mn4-"], "uri": ["g2287"]}},
    "CONSESSOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c3529"]}},
    "DVELLICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["22734"]}},
    "CALAVTICA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0175"]}},
    "PERCA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["p0926"]}},
    "SVBRVMPO": {"-": {"morpho": ["v1spia--3-"], "uri": ["128423"]}},
    "LACO": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "LICITATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["l0726"]}},
    "DIABATHRARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["d1212"]}},
    "IMPERIVRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i0498"]}},
    "STAMINATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2372"]}},
    "MASTRVCATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["m0439"]}},
    "PALLENE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "MVLTIGENERIS": {"-": {"morpho": ["aps---cn3i"], "uri": []}},
    "CYANE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "TABVLARIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["t0027"]}},
    "INCONSVMPTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1055"]}},
    "PAPAVEREVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0271"]}},
    "TRICLINIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["100953"]}},
    "AMICVLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["a1685"]}},
    "SVFFARCINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["s3341"]}},
    "IANITRIX": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0032"]}},
    "TROS": {
        "N": {"morpho": ["n-s---mn2-"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "INHALO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1729"]}},
    "EXAMVSSIM": {"-": {"morpho": ["rp--------"], "uri": ["e1220"]}},
    "OPIPARVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["o0833"]}},
    "ALCEDONIA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["a1315"]}},
    "DENTILEGVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["38896"]}},
    "MENDICOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["132567"]}},
    "COMBARDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c2557"]}},
    "INVRBANVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i2745"]}},
    "ERVDITE": {"-": {"morpho": ["rp--------"], "uri": ["54529"]}},
    "PYTHO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "CATASTROPHA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c0856"]}},
    "ERYX": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "BOLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["b0510"]}},
    "PRAEDIATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["p2918"]}},
    "NEVTRVBI": {"-": {"morpho": ["rp--------"], "uri": ["n0763"]}},
    "TRANSADIGO": {"-": {"morpho": ["v1spia--3-"], "uri": ["t0973"]}},
    "PARIS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "INSIGNITE": {"-": {"morpho": ["rp--------"], "uri": ["37401"]}},
    "CROCIO": {"-": {"morpho": ["v1spia--3-"], "uri": ["21908"]}},
    "PRAEMERCOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["p3105"]}},
    "PERBLANDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p0923"]}},
    "CASSIOPE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "DATATIM": {"-": {"morpho": ["rp--------"], "uri": ["d0062"]}},
    "APOSPHRAGISMA": {"-": {"morpho": ["n-s---nn3-"], "uri": ["a2432"]}},
    "SIDON": {"N": {"morpho": ["n-s---nn3-"], "uri": []}},
    "INGVRGITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i1719"]}},
    "INORATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["i1901"]}},
    "INCEPTO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0871"]}},
    "NITIDE": {"-": {"morpho": ["rp--------"], "uri": ["99580"]}},
    "VOLSELLA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["u1171"]}},
    "COMMARITVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["c2616"]}},
    "HYPSIPYLE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "OCCATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["o0374"]}},
    "SCIRPEVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s0101"]}},
    "LEVCOTHOE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "CARDVELIS": {"-": {"morpho": ["n-s---mn3i"], "uri": ["c0634"]}},
    "SVBHORRIDVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2867"]}},
    "PINSITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["p2111"]}},
    "BOIA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["b0499"]}},
    "DOMEFACTVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d2134"]}},
    "TRADVX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["t0932"]}},
    "CONSOBRINA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["c4653"]}},
    "AENEOLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["a0903"]}},
    "MATRICIDA": {"-": {"morpho": ["n-s---cn1-"], "uri": ["m0468"]}},
    "CONDIGNE": {"-": {"morpho": ["rp--------"], "uri": ["32195"]}},
    "MASO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "ALEATORIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["38678"]}},
    "HOSTIFICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["h0634"]}},
    "COMMINISCOR": {"-": {"morpho": ["v1spid--3-"], "uri": ["c2677"]}},
    "HERBESCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["h0274"]}},
    "IMMISSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["i0385"]}},
    "PRAELVSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p3091"]}},
    "CAERVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["35541"]}},
    "CYPASSIS": {"N": {"morpho": ["n-s---fn3-"], "uri": []}},
    "COMMEMINI": {"-": {"morpho": ["v1sria--3-"], "uri": ["c2631"]}},
    "SVCCEDANEVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["128608"]}},
    "PORCINARIVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["p2620"]}},
    "TRANSVERSARIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1105"]}},
    "INCLAMITO": {"-": {"morpho": ["v1spia--1-"], "uri": ["i0929"]}},
    "IOLE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "ANTECESSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["a2120"]}},
    "CONCASTIGO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c2961"]}},
    "RAVIS": {"-": {"morpho": ["n-s---fn3i"], "uri": ["r0122"]}},
    "NOTATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["47598"]}},
    "NARBO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "PERGRAECOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["p1182"]}},
    "FENERATO": {"-": {"morpho": ["rp--------"], "uri": ["98517"]}},
    "PERPROPINQVVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1510"]}},
    "HELIOCAMINVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["h0195"]}},
    "NYCTIMENE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "SALPA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s0186"]}},
    "GESTICVLOR": {"-": {"morpho": ["v1spid--1-"], "uri": ["g0268"]}},
    "SALSVRA": {"-": {"morpho": ["n-s---fn1-"], "uri": ["s9914"]}},
    "FRVMENTATOR": {"-": {"morpho": ["v1spia--1-"], "uri": ["132489"]}},
    "EVADNE": {"N": {"morpho": ["n-s---fn1-"], "uri": []}},
    "DELOS": {"N": {"morpho": ["n-s---mn2-"], "uri": []}},
    "EXCIDIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["e1306"]}},
    "INTERREX": {"-": {"morpho": ["n-s---mn3-"], "uri": ["i2463"]}},
    "PERLONGVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["p1318"]}},
    "SVBAVRATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["100721"]}},
    "QVARTVSDECIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["56259"]}},
    "OBDORMISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["o0194"]}},
    "CONGEMINO": {"-": {"morpho": ["v1spia--1-"], "uri": ["c3345"]}},
    "BELLATORIVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["b0221"]}},
    "FEBRVA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["f0313"]}},
    "GLIS": {"-": {"morpho": ["n-s---mn3-"], "uri": ["g0379"]}},
    "PRODICO": {"2": {"morpho": ["v1spia--3-"], "uri": ["p3712"]}},
    "CALIDE": {"-": {"morpho": ["rp--------"], "uri": ["50645"]}},
    "CONTENTIOSVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["c3718"]}},
    "SCVTIGERVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["s0825"]}},
    "RIVVLVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["r0755"]}},
    "CONGRESSIO": {"-": {"morpho": ["n-s---fn3-"], "uri": ["c3396"]}},
    "POESIS": {"-": {"morpho": ["n-s---fn3-"], "uri": ["p2437"]}},
    "DICACVLVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["d1332"]}},
    "TERTIADECIMANI": {"-": {"morpho": ["n-p---mn2-"], "uri": ["t0404"]}},
    "SEPTVENNIS": {"-": {"morpho": ["aps---cn3i"], "uri": ["s1322"]}},
    "ORPHVS": {"-": {"morpho": ["n-s---mn2-"], "uri": ["o1038"]}},
    "SPVRCIDICVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["s2302"]}},
    "CROTALIA": {"-": {"morpho": ["n-p---nn2-"], "uri": ["c4413"]}},
    "DVCENTIENS": {"-": {"morpho": ["rp--------"], "uri": ["d2276"]}},
    "PERVIVO": {"-": {"morpho": ["v1spia--3-"], "uri": ["p1750"]}},
    "VEIENTO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "FICTE": {"-": {"morpho": ["rp--------"], "uri": ["23337"]}},
    "VSVFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": ["u1305"]}},
    "SVPERVENTVS": {"-": {"morpho": ["n-s---mn4-"], "uri": ["s9983"]}},
    "BADIZO": {"-": {"morpho": ["v1spia--1-"], "uri": ["b0037"]}},
    "CONGRVENTIA": {"-": {"morpho": ["n-p---nn3i"], "uri": ["c9956"]}},
    "CHALCIS": {"N": {"morpho": ["n-s---fn3-"], "uri": ["c1338"]}},
    "EDORMISCO": {"-": {"morpho": ["v1spia--3-"], "uri": ["e0101"]}},
    "MYS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "FRVSINO": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "EXPVDORATVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["e1681"]}},
    "CONFIRMATOR": {"-": {"morpho": ["n-s---mn3-"], "uri": ["c3253"]}},
    "BETO": {"-": {"morpho": ["v1spia--3-"], "uri": ["43679"]}},
    "DICTAMNVS": {"-": {"morpho": ["n-s---fn2-"], "uri": ["d1353"]}},
    "TVRILEGVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["t1793"]}},
    "VFENS": {"N": {"morpho": ["n-s---mn3-"], "uri": []}},
    "DROMO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "HALOSIS": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "VRANIE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "TYPHON": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "RAPO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "INACHIS": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "SENECIO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "LVCRO": {"N": {"morpho": ["n-s----n1-"], "uri": []}},
    "TRIO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "CEYX": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "LVRCO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "PHALERIS": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "SACRATOR": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "BETIS": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "LEVCASPIS": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "ACANTHIS": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "AEGOCEROS": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "PELLIO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "MINIO": {"N": {"morpho": ["n-s----n1-"], "uri": []}},
    "SVAVISAVIATIO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "SILO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "POLLIS": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "INO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "PVLLO": {"N": {"morpho": ["n-s----n1-"], "uri": []}},
    "MEFITIS": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "PROSELENOS": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "ARGITIS": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "VESPERVGO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "VESPILLO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "DAPHNE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "METAMORPHOSIS": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "MVTTO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "BVRDO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "LATOIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "DAVNIAS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "ASOPIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "DODONIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "OMPHIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "LARINAS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "ILIBERRIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "BACCHIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "AMARYLLIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "PITTHEIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "ZIOBETIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "TOMYRIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "THYRSIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "GARAMAS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "STRATONICIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "LYCORIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "IGVVINAS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "BARDYLIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "SABIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "POTNIAS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "CADMEIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "ABARIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "SAGARITIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "HIAROTIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "ATREBAS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "ALETRINAS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "EPICHARIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "VENAFER": {"A": {"morpho": ["aps---mn1r"], "uri": []}},
    "AQVINAS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "PHEGIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "TRINACRIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "AMAZONIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "ATLANTIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "CEPHEIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "LIMNATIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "CERIALIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "BASSARIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "BOEBEIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "ACHAIAS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "INSVBER": {"A": {"morpho": ["aps---mn1r"], "uri": []}},
    "VADIMONIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "PALLANTIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "ERECHTHIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "AGANIPPIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "CISSEIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "RHESCVPORIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "TAENARIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "THERMVTIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "BYBLIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "SIDONIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "LOCRIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "NISIAS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "THYNIAS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "SESOSIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "CARTHASIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "POMPEIOPOLIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "DANAIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "BELIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "STELLAS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "OSIRIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "ATRACIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "PANNYCHIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "SCYRIAS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "SALMONIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "THEBAIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "AETOLIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "AZANIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "ATHESIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "OENEIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "ACTIAS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "SEBETHIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "ARARIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "THYMBRIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "TRITONIDIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "CANTABER": {"A": {"morpho": ["aps---mn1r"], "uri": []}},
    "CYTAEIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "PASITIGRIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "THALESTRIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "VRBINAS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "TAMESIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "ICARIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "SOTERIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "MANAPIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "VEIOVIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "DENTHALIAS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "SISIGAMBIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "CALABER": {"A": {"morpho": ["aps---mn1r"], "uri": []}},
    "LAIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "LYRNESSIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "RAVENNAS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "CANINEFAS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "ICARIOTIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "EVRYTIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "APOLLONIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "DICARCHIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "NYSIAS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "NAIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "CRAVGIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "INACHIOTIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "SINIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "SCHOENEIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "SICORIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "ARCHEPOLIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "MYCENIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "AORNIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "ANTIAS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "BOCCHORIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "HAEMONIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "ARDEAS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "PRIVERNAS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "CHRYSEIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "ALEXIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "CYROPOLIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "IBER": {"A": {"morpho": ["aps---mn1r"], "uri": []}},
    "MENECRATIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "DAMALIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "CELENDRIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "VCVBIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "DINIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "CLEOPHIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "SIMOIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "SVGAMBER": {"A": {"morpho": ["aps---mn1r"], "uri": []}},
    "DITIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "TRVCCVLENSIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "BITHYNIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "PHARIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "PTOLEMAIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "TANAITIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "AGONIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "MENNIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "SAMIRAMIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "OPIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "PELOPEIAS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "PIMPLEIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "ATINAS": {
        "A": {"morpho": ["aps---an3-"], "uri": []},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "PHORONIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "THESPIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "NYCTEIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "LEVCIPPIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "EDONIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "CELTIBER": {"A": {"morpho": ["aps---mn1r"], "uri": []}},
    "ANVBIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "PELASGIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "AMANICAE": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "BACENIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "LATRIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "PEDASIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "IASIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "SITHON": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "TELESTIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "CAPPADOX": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "HYPANIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "TANTALIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "CASINAS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "DYSPARIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "ALBIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "TRALLIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "CAR": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "ACARNAN": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "CLANIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "TVNGER": {"A": {"morpho": ["aps---mn1r"], "uri": []}},
    "CILIX": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "LYCAONIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "CANNINEFAS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "CRES": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "SCYDROTHEMIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "FVLGINAS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "THESTYLIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "ASIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "OEBALIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "CILISSA": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "DAVLIAS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "BRISEIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "RHACOTIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "NYSIGENA": {"A": {"morpho": ["aps---mn1-"], "uri": []}},
    "PELASGIAS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "THESEIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "CALIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "VISVRGIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "CONVICTOLITAVIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "SCYTHES": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "SVDANS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "ATHAMANTIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "DVBIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "LIBYS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "PRIAMEIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "RHASCYPOLIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "CAERES": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "PICENS": {
        "A": {"morpho": ["aps---an3-"], "uri": []},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "TIBVRS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "EVPOLIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "SPALIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "OCEANITIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "TVRESIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "THRAEX": {
        "A": {"morpho": ["aps---an3-"], "uri": []},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "AETHIOPS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "NISEIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "PHORCIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "PIERIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "PHAEAX": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "MEMPHITES": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "ELEATES": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "ACROPOLISTIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "IAPYS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "CAMERS": {
        "A": {"morpho": ["aps---an3-"], "uri": []},
        "N": {"morpho": ["n-s----n3-"], "uri": []},
    },
    "THESPIAS": {"A": {"morpho": ["aps---an3-"], "uri": []}},
    "ALPIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "NICOPOLIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "STRYMONIS": {"N": {"morpho": ["n-s----n--"], "uri": []}},
    "PHYLLODOCE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "BVBACENE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "BIBRACTE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "HELENE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "EVANTHE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "BARINE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "PHIDYLE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "TRICRENE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "ARTACIE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "PROCNE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "CHLOE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "PSYCHE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "ALCIPPE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "ACME": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "ANDROMEDE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "NEPTVNINE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "LEVCOPHRYNE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "CEYCE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "ALSIENSE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "CAMERE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "CHLIDE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "SOPHENE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "EVTERPE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "COELE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "THEBE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "VSPE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "HELLANICE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "CARENE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "ALCMENE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "RHODE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "PHIALE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "EPHYRE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "SITTACENE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "SORACTE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "CYMODOCE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "AMYCLE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "MYRTALE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "LIPARE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "DIDYME": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "ORTYGIE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "HARPALYCE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "LIBYE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "CYBELE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "HEBE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "DEIPHOBE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "NIOBE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "ILIONE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "EVROPE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "ATALANTE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "NEOBVLE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "LEVCONOE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "ROXANE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "ISCHOMACHE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "BARSINE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "CALYBE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "MOTHONE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "NISAEE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "IDE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "CALLIRHOE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "MELPOMENE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "PARTHIENE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "PARTHENIE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "THVLE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "PEGE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "SCARPHE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "HYPERMESTRE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "DINOMACHE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "HEDYLE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "ARELATE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "ANTIOPE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "NESAEE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "PALLACINE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "PHRYNE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "HERCVLANENSE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "PETALE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "CYBEBE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "HEROPHILE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "AEETINE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "CYTAEINE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "EVPHROSYNE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "ERIPHYLE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "PANTAGIE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "TAYGETE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "ANDROMACHE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "LALAGE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "ENISPE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "ANACTORIE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "LEDE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "THRACE": {"N": {"morpho": ["n-p---fn1g"], "uri": []}},
    "CATABATHMOS": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "ORSILOS": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "ATHOS": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "ILIOS": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "CROCOS": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "PHILEROS": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "ISSOS": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "SANBVLOS": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "PEPARETHOS": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "ARGOS": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "SPERCHIOS": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "PALAETYROS": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "EVXENIPPOS": {"N": {"morpho": ["n-s----n-g"], "uri": []}},
    "OROPS": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "PYRGO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "CONGRIO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "ALCANOR": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "ARISTOGITON": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "MACHAERIO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "ARGO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "CLOTHO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "EPOREDORIX": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "ASPRENAS": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "METRONAX": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "HYLAX": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "NVMITOR": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "CVPAVO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "BRIMO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "THESAVROCHRYSONICOCH.": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "RVFIO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "SER": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "ASPAR": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "CINGETORIX": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "VOLVX": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "TREVIR": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "BOCCAR": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "ATHO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "LVGOTORIX": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "ARCENS": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "AGGAR": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "VANGIO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "AMBIORIX": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "VOLCENS": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "DIVICO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "DRYOPS": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "BVRSAVO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "MASTANABAL": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "ADHERBAL": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "VRSAO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "GOBANNITIO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "HIEROCAESARIEN.": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "CRVPTORIX": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "DVMNORIX": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "TIMO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "SACROVIR": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "OLLOVICO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "SENO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "SEGOVAX": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "STASANOR": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "TAVROENS": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "PVLEIO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "SPIO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "ZOPYRIO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "ANIEN": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "TAPPO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "SVTHVL": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "BRITTO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "VOCCIO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "POLYMESTOR": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "CRESCENS": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "COTYTO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "XANTHO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "BIANOR": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "CARIO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "COTYLO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "DEMONAX": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "RVSO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "MALORIX": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "MATISCO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "PLATOR": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "VIRIDOVIX": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "IERVSALEM": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "TIPHYS": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "AGATHO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "NOBILIOR": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "DRYMO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "CALAS": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "MVTHVL": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "TVISTO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "VERGILIO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "ARAR": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "ORGETORIX": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "ERINNYS": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "VVLCHALO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "DOLOPS": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "LATO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "BRINNO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "PITHON": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "MACELO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "CAFO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "TVLLIO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "THEANO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "SYPHAX": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "VENTIPO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "BAMBALIO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "GRAECVLIO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "PITHO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "VESONTIO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "TYPHO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "MERCELLO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "QVERCENS": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "POETOVIO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "LIBAS": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "LIGER": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "ISION": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "LEPISO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "HIEMPSAL": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "CYDRO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "VERCINGETORIX": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "NASO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "HELENOR": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "ACCO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "CARCASO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "BOMILCAR": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "BIBRAX": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "ATAX": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "DABAR": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "TRIMALCHIO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "CORDALIO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "GALLIO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "SEMO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "VERTICO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "ASTYR": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "FELICIO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "VIRRO": {"N": {"morpho": ["n-s----n3-"], "uri": []}},
    "COCOCOCO": {"-": {"morpho": ["----------"], "uri": []}},
    "EXPRETVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "DISMARITVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "NONAGENVS": {"-": {"morpho": ["aps---mn1-"], "uri": []}},
    "MILITATVRA": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "OPTESTOR": {"-": {"morpho": ["v1spid--1-"], "uri": []}},
    "EM": {"-": {"morpho": ["----------"], "uri": []}},
    "PTHISICVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "STOICE": {"-": {"morpho": ["rp--------"], "uri": []}},
    "ODIOSSICVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["o0511"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["o0511"]},
    },
    "NVBIGENA": {"-": {"morpho": ["aps---cn1-"], "uri": ["99609"]}},
    "THYIADES": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "QVADRIGENTI": {"-": {"morpho": ["----------"], "uri": []}},
    "PERMEDITOR": {"-": {"morpho": ["v1spid--1-"], "uri": []}},
    "LARS": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "FIDVCARIVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "INEXSTINGVO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "ANTECIPIO": {"-": {"morpho": ["v1spia--3i"], "uri": []}},
    "CIRCVMVEHO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "BAT": {"-": {"morpho": ["----------"], "uri": []}},
    "EDEPOL": {"-": {"morpho": ["----------"], "uri": []}},
    "ABSENTIVVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "VNDENVS": {"-": {"morpho": ["aps---mn1-"], "uri": []}},
    "EVGEPAE": {"-": {"morpho": ["----------"], "uri": []}},
    "DVODENVS": {"-": {"morpho": ["aps---mn1-"], "uri": []}},
    "APAGE": {"-": {"morpho": ["----------"], "uri": []}},
    "EHEM": {"-": {"morpho": ["----------"], "uri": []}},
    "VAH": {"-": {"morpho": ["----------"], "uri": []}},
    "ARRADO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "ECASTOR": {"-": {"morpho": ["----------"], "uri": []}},
    "QVINQVAGENVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["q0243"]}},
    "EHO": {"-": {"morpho": ["----------"], "uri": []}},
    "SESCENTIENS": {"-": {"morpho": ["rp--------"], "uri": []}},
    "CONCILIOR": {"-": {"morpho": ["v1spid--1-"], "uri": []}},
    "IMMODVLOR": {"-": {"morpho": ["v1spid--1-"], "uri": []}},
    "TRICENTIENS": {"-": {"morpho": ["rp--------"], "uri": []}},
    "SEXCENVS": {"-": {"morpho": ["----------"], "uri": []}},
    "BALINEVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "DRACONIGENVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": ["d2209"]},
        "A": {"morpho": ["aps---mn1-"], "uri": ["d2209"]},
    },
    "INVERENCVNDVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "_EO": {"2": {"morpho": ["v1spia--3-"], "uri": []}},
    "QVASSILARIA": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "SVDES": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "BENEDICE": {"-": {"morpho": ["rp--------"], "uri": []}},
    "QVADRINVS": {"-": {"morpho": ["aps---mn1-"], "uri": ["q0065"]}},
    "QVINGENVS": {"-": {"morpho": ["----------"], "uri": []}},
    "PHYRGIO": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "AVERTOR": {"-": {"morpho": ["v1spid--3-"], "uri": []}},
    "ST": {"-": {"morpho": ["----------"], "uri": []}},
    "MOSCILLVS": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "HEDIA": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "VVLCANALIS": {
        "-": {"morpho": ["aps---cn3i"], "uri": []},
        "A": {"morpho": ["aps---cn3i"], "uri": []},
    },
    "AVCTORATI": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "SESCENVS": {"-": {"morpho": ["aps---mn1-"], "uri": []}},
    "VERBERABILIS": {
        "-": {"morpho": ["aps---cn3i"], "uri": ["u0353"]},
        "A": {"morpho": ["aps---cn3i"], "uri": ["u0353"]},
    },
    "MV": {"-": {"morpho": ["----------"], "uri": []}},
    "MA": {"-": {"morpho": ["----------"], "uri": []}},
    "VNVSETVICESIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": []}},
    "MILLIENS": {"-": {"morpho": ["rp--------"], "uri": []}},
    "FVI": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "ATTATAE": {"-": {"morpho": ["----------"], "uri": []}},
    "INCALEFACIO": {"-": {"morpho": ["v1spia--3i"], "uri": []}},
    "OCCLVSVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "TRECENVS": {"-": {"morpho": ["----------"], "uri": []}},
    "PRAEPOLLVO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "REGILLA": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "SOTADICVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "TARRACONENSES": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "TRIVENEFICVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "TRECENTENVS": {"-": {"morpho": ["----------"], "uri": []}},
    "SEXAGENVS": {"-": {"morpho": ["aps---mn1-"], "uri": []}},
    "SEPTVAGIENS": {"-": {"morpho": ["rp--------"], "uri": ["s1318"]}},
    "SVMMANIS": {
        "-": {"morpho": ["aps---cn3i"], "uri": []},
        "A": {"morpho": ["aps---cn3i"], "uri": []},
    },
    "APOCALO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "PRAECOMPONO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "EXIVRO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1512"]}},
    "THYLACISTA": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "INDEMNO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "RVMIFERO": {"-": {"morpho": ["v1spia--1-"], "uri": ["r1489"]}},
    "HAMADRYAS": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "VENERIAE": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "ACIA": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "THIMIAMAE": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "OFFERRO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "COMPEDES": {
        "-": {"morpho": ["aps---an3-"], "uri": []},
        "A": {"morpho": ["aps---an3-"], "uri": []},
    },
    "INNOXIE": {"-": {"morpho": ["rp--------"], "uri": []}},
    "PVNCTATORIOLA": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "ILLVNIS": {"-": {"morpho": ["aps---an3-"], "uri": []}},
    "PLAGIPATIDA": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "ANDROGYNOS": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "PRASINIANVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "OCCENDO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "PERNVLA": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "EXCELLENTIS": {
        "-": {"morpho": ["aps---cn3i"], "uri": []},
        "A": {"morpho": ["aps---cn3i"], "uri": []},
    },
    "STRIGOR": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "SESQVEOPVS": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "CINIPES": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "INGVRGITOR": {"-": {"morpho": ["v1spid--1-"], "uri": []}},
    "CAVNEA": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "INEXSPECTO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "LOQVO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "IMMEMORO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "CALVAE": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "NESCIOQVID": {"-": {"morpho": ["----------"], "uri": []}},
    "NOMVS": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "DVETVICESIMVS": {"-": {"morpho": ["aps---mn1-"], "uri": []}},
    "LOGVS": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "GRAMA": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "CLATHRA": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "VETVO": {"-": {"morpho": ["v1spia--1-"], "uri": []}},
    "SANGVNCVLVS": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "BISSACIVS": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "HIPPODROMOS": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "INTERVIAS": {
        "-": {"morpho": ["aps---an--"], "uri": []},
        "A": {"morpho": ["aps---an--"], "uri": []},
    },
    "LAMYROS": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "CLASSIA": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "DIVIDIA": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "CHIRAGRICI": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "FIVO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "CHRYSOS": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "PRAEVT": {"-": {"morpho": ["rp--------"], "uri": ["52057"]}},
    "THEBAICAE": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "PENTHIACVS": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "APSIS": {"-": {"morpho": ["n-s----n-g"], "uri": []}},
    "VESTES": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "COMMISSVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "D": {"-": {"morpho": ["----------"], "uri": []}},
    "IMPOLLITVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "VELANDA": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "QVAQVAM": {"-": {"morpho": ["r---------"], "uri": []}},
    "AV": {"-": {"morpho": ["----------"], "uri": []}},
    "STERTEIVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "CAELVS": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "VLS": {"-": {"morpho": ["p---------"], "uri": ["u0940"]}},
    "POTICIVS": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "OPISTOGRAPHVS": {
        "-": {"morpho": ["aps---mn1-"], "uri": []},
        "A": {"morpho": ["aps---mn1-"], "uri": []},
    },
    "MARTIOLVS": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "STOLATA": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "ASSVDASCO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "INDECET": {"-": {"morpho": ["v1spia--2-"], "uri": []}},
    "CEDRA": {"-": {"morpho": ["n-s----n1-"], "uri": []}},
    "NONNIHIL": {"-": {"morpho": ["aps---mn--"], "uri": []}},
    "EVAX": {"-": {"morpho": ["----------"], "uri": []}},
    "TAMIAS": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "NVMNAM": {"-": {"morpho": ["r---------"], "uri": []}},
    "LASANVS": {"-": {"morpho": ["n-s----n2-"], "uri": []}},
    "NVMQVO": {"-": {"morpho": ["r---------"], "uri": []}},
    "NVMQVISNAM": {"-": {"morpho": ["----------"], "uri": []}},
    "ADVSQVE": {"2": {"morpho": ["p---------"], "uri": ["a3717"]}},
    "CANALE": {"-": {"morpho": ["n-s----n3-"], "uri": []}},
    "NESCIOQVOMODO": {"-": {"morpho": ["r---------"], "uri": []}},
    "PORGO": {"-": {"morpho": ["v1spia--3-"], "uri": []}},
    "SERE": {"-": {"morpho": ["rp--------"], "uri": []}},
    "EXOPINISSO": {"-": {"morpho": ["v1spia--1-"], "uri": ["e1533"]}},
}

AUTHOR_TAB = {
    "A": {
        "author": "C. Iulius Caesar",
        "code": ["phi0448"],
        "phi0448": {
            "01": {
                "title": "De Bello Gallico",
                "meta": "book-chapter-section",
                "code": "phi001",
            },
            "02": {
                "title": "De Bello Gallico Liber VIII",
                "code": "phi001",
                "meta": "book-chapter-section",
                "spurious": {"author": "A. Hirtius", "code": "phi0530"},
            },
            "03": {
                "title": "Bellum Civile",
                "meta": "book-chapter-section",
                "code": "phi002",
            },
            "05": {
                "title": "Bellum Africum",
                "code": "phi001",
                "meta": "section",
                "spurious": {"author": "Anonymous", "code": ["phi0426"]},
            },
            "04": {
                "title": "Bellum Alexandrinum",
                "code": "phi001",
                "meta": "section",
                "spurious": {"author": "Anonymous", "code": "phi0428"},
            },
            "06": {
                "title": "Bellum Hispaniense",
                "code": "phi001",
                "meta": "section",
                "spurious": {"author": "Anonymous", "code": ["phi0430"]},
            },
        },
    },
    "B": {
        "author": "M. Porcius Cato",
        "code": ["phi0022"],
        "phi0022": {
            "01": {
                "title": "De Agri Cultura",
                "meta": "chapter-section",
                "code": "phi001",
            },
            "03": {
                "title": "orationes",
                "meta": "fragment-line",
                "code": "phi010",
            },
            "02": {
                "title": "Origines",
                "meta": "fragment-line",
                "code": "phi011",
            },
            "04": {
                "title": "incertorum librorum fragmenta",
                "meta": "page-fragment-line",  # work-
                "code": "phi006",
            },
        },
    },
    "C": {
        "author": "C. Valerius Catullus",
        "code": ["phi0472"],
        "phi0472": {
            "00": {"title": "carmina", "meta": "poem-verse", "code": "phi001"}
        },
    },
    "D": {
        "author": "M. Tullius Cicero",
        "code": ["phi0474"],
        "phi0474": {
            "34": {"title": "Pro Caelio", "code": "phi024", "meta": "section"},
            "35": {
                "title": "De Provinciis Consularibus",
                "code": "phi025",
                "meta": "section",
            },
            "13": {
                "title": "Pro Caecina",
                "code": "phi008",
                "meta": "section",
            },
            "15": {
                "title": "Pro Cluentio",
                "code": "phi010",
                "meta": "section",
            },
            "26": {"title": "Pro Archia", "code": "phi016", "meta": "section"},
            "36": {"title": "Pro Balbo", "code": "phi026", "meta": "section"},
            "37": {"title": "In Pisonem", "code": "phi027", "meta": "section"},
            "04": {"title": "Pro Tullio", "code": "phi006", "meta": "section"},
            "24": {"title": "Pro Murena", "code": "phi014", "meta": "section"},
            "38": {
                "title": "Pro Plancio",
                "code": "phi028",
                "meta": "section",
            },
            "01": {
                "title": "Pro Quinctio",
                "code": "phi001",
                "meta": "section",
            },
            "19": {
                "title": "Pro C. Rabirio perduellionis reo",
                "code": "phi012",
                "meta": "section",
            },
            "39": {
                "title": "Pro Rabirio Postumo",
                "code": "phi030",
                "meta": "section",
            },
            "27": {"title": "Pro Flacco", "code": "phi017", "meta": "section"},
            "12": {
                "title": "Pro Fonteio",
                "code": "phi007",
                "meta": "section",
            },
            "40": {"title": "Pro Milone", "code": "phi031", "meta": "section"},
            "41": {
                "title": "Pro Marcello",
                "code": "phi032",
                "meta": "section",
            },
            "29": {
                "title": "Post Reditum ad Populum",
                "code": "phi018",
                "meta": "section",
            },
            "28": {
                "title": "Post Reditum in Senatu",
                "code": "phi019",
                "meta": "section",
            },
            "42": {
                "title": "Pro Ligario",
                "code": "phi033",
                "meta": "section",
            },
            "43": {
                "title": "Pro Rege Deiotaro",
                "code": "phi034",
                "meta": "section",
            },
            "06": {
                "title": "In Verrem",
                "code": "phi005",
                "meta": "actio-book-chapter-section",
            },
            "07": {
                "title": "In Verrem",
                "code": "phi005",
                "meta": "actio-book-chapter-section",
            },
            "08": {
                "title": "In Verrem",
                "code": "phi005",
                "meta": "actio-book-chapter-section",
            },
            "09": {
                "title": "In Verrem",
                "code": "phi005",
                "meta": "actio-book-chapter-section",
            },
            "10": {
                "title": "In Verrem",
                "code": "phi005",
                "meta": "actio-book-chapter-section",
            },
            "11": {
                "title": "In Verrem",
                "code": "phi005",
                "meta": "actio-book-chapter-section",
            },
            "44": {
                "title": "Philippicae",
                "code": "phi035",
                "meta": "oration-section",
            },
            "45": {
                "title": "Philippicae",
                "code": "phi035",
                "meta": "oration-section",
            },
            "46": {
                "title": "Philippicae",
                "code": "phi035",
                "meta": "oration-section",
            },
            "48": {
                "title": "Philippicae",
                "code": "phi035",
                "meta": "oration-section",
            },
            "47": {
                "title": "Philippicae",
                "code": "phi035",
                "meta": "oration-section",
            },
            "49": {
                "title": "Philippicae",
                "code": "phi035",
                "meta": "oration-section",
            },
            "50": {
                "title": "Philippicae",
                "code": "phi035",
                "meta": "oration-section",
            },
            "51": {
                "title": "Philippicae",
                "code": "phi035",
                "meta": "oration-section",
            },
            "52": {
                "title": "Philippicae",
                "code": "phi035",
                "meta": "oration-section",
            },
            "53": {
                "title": "Philippicae",
                "code": "phi035",
                "meta": "oration-section",
            },
            "54": {
                "title": "Philippicae",
                "code": "phi035",
                "meta": "oration-section",
            },
            "55": {
                "title": "Philippicae",
                "code": "phi035",
                "meta": "oration-section",
            },
            "56": {
                "title": "Philippicae",
                "code": "phi035",
                "meta": "oration-section",
            },
            "57": {
                "title": "Philippicae",
                "code": "phi035",
                "meta": "oration-section",
            },
            "69": {
                "title": "Contra Contionem Q. Metelli",
                "code": "phi68x12",
                "meta": "fragment-line",
            },
            "70": {
                "title": "In P. Clodium et C. Curionem",
                "code": "phi68x13",
                "meta": "fragment-line",
            },
            "63": {
                "title": "Pro C. Fundanio",
                "code": "phi68x05",
                "meta": "fragment-line",
            },
            "62": {
                "title": "Pro C. Manilio",
                "code": "phi68x04",
                "meta": "fragment-line",
            },
            "65": {
                "title": "Pro Cornelio",
                "code": "phi68x07",
                "meta": "fragment-line",
            },
            "74": {
                "title": "Pro Negotiatoribus Achaeis",
                "code": "phi68x17",
                "meta": "fragment-line",
            },
            "61": {
                "title": "Pro Oppio",
                "code": "phi68x03",
                "meta": "fragment-line",
            },
            "73": {
                "title": "Pro P. Vatinio",
                "code": "phi68x16",
                "meta": "fragment-line",
            },
            "64": {
                "title": "Pro Q. Gallio",
                "code": "phi68x06",
                "meta": "fragment-line",
            },
            "68": {
                "title": "Cum A Ludis Contionem Avocavit",
                "code": "phi68x10",
                "meta": "fragment-line",
            },
            "59": {
                "title": "Cum Quaestor Lilybaeo Decederet",
                "code": "phi68x01",
                "meta": "fragment-line",
            },
            "72": {
                "title": "De Rege Alexandrino",
                "code": "phi68x15",
                "meta": "fragment-line",
            },
            "76": {
                "title": "Laelius de Amicitia",
                "code": "phi052",
                "meta": "section",
            },
            "30": {
                "title": "De Domo Sua",
                "code": "phi020",
                "meta": "section",
            },
            "31": {
                "title": "De Haruspicum Responso",
                "code": "phi021",
                "meta": "section",
            },
            "14": {
                "title": "Pro Lege Manilia",
                "code": "phi009",
                "meta": "section",
            },
            "16": {
                "title": "De Lege Agraria",
                "code": "phi011",
                "meta": "oration-section",
            },
            "78": {
                "title": "Cato Maior de Senectute",
                "code": "phi051",
                "meta": "section",
            },
            "77": {
                "title": "De Officiis",
                "code": "phi055",
                "meta": "book-chapter-section",
            },
            "05": {
                "title": "Divinatio in Q. Caecilium",
                "code": "phi004",
                "meta": "section",
            },
            "21": {
                "title": "In Catilinam",
                "code": "phi013",
                "meta": "oration-section",
            },
            "20": {
                "title": "In Catilinam",
                "code": "phi013",
                "meta": "oration-section",
            },
            "22": {
                "title": "In Catilinam",
                "code": "phi013",
                "meta": "oration-section",
            },
            "23": {
                "title": "In Catilinam",
                "code": "phi013",
                "meta": "oration-section",
            },
            "58": {"title": "Pro Scauro", "code": "phi029", "meta": "section"},
            "67": {
                "title": "In Toga Candida",
                "code": "phi005",
                "meta": "page-line",
            },
            "60": {
                "title": "Pro L. Vareno",
                "code": "phi68x02",
                "meta": "fragment-line",
            },
            "33": {
                "title": "In Vatinium",
                "code": "phi023",
                "meta": "section",
            },
            "71": {
                "title": "Interrogatio de aere alieno Milonis",
                "code": "phi68x14",
                "meta": "fragment-line",
            },
            "75": {
                "title": "orationum incertarum frr.",
                "code": "phi069",
                "meta": "fragment-line",
            },
            "03": {
                "title": "Roscio Comoedo",
                "code": "phi003",
                "meta": "section",
            },
            "32": {"title": "Pro Sestio", "code": "phi022", "meta": "section"},
            "02": {
                "title": "Roscio Amerino",
                "code": "phi002",
                "meta": "section",
            },
            "25": {"title": "Pro Sulla", "code": "phi015", "meta": "section"},
        },
    },
    "E": {
        "author": "Q. Curtius Rufus",
        "code": ["phi0860"],
        "phi0860": {
            "00": {
                "title": "Historiae Alexandri Magni",
                "code": "phi001",
                "meta": "book-chapter-section",
            }
        },
    },
    "F": {
        "author": "Q. Horatius Flaccus",
        "code": ["phi0893"],
        "phi0893": {
            "01": {
                "title": "Carmina",
                "code": "phi001",
                "meta": "book-poem-verse",
            },
            "02": {
                "title": "Carmen Saeculare",
                "code": "phi002",
                "meta": "verse",
            },
            "03": {"title": "Epodi", "code": "phi003", "meta": "poem-verse"},
            "04": {
                "title": "Sermones",
                "code": "phi004",
                "meta": "book-poem-verse",
            },
            "05": {
                "title": "Epistulae",
                "code": "phi005",
                "meta": "book-poem-verse",
            },
            "06": {"title": "Ars Poetica", "code": "phi006", "meta": "verse"},
        },
    },
    "G": {
        "author": "D. Iunius Iuvenalis",
        "code": ["phi1276"],
        "phi1276": {
            "00": {"title": "Saturae", "code": "phi001", "meta": "poem-verse"}
        },
    },
    "H": {
        "author": "T. Lucretius Carus",
        "code": ["phi0550"],
        "phi0550": {
            "00": {
                "title": "De Rerum Natura",
                "code": "phi001",
                "meta": "book-verse",
            }
        },
    },
    "I": {
        "author": "P. Ovidius Naso",
        "code": ["phi0959"],
        "phi0959": {
            "04": {
                "title": "Amores",
                "code": "phi001",
                "meta": "book-poem-verse",
            },
            "05": {
                "title": "Epistulae (vel Heroides)",
                "code": "phi002",
                "meta": "poem-verse",
            },
            "03": {
                "title": "Medicamina Faciei Femineae",
                "code": "phi003",
                "meta": "verse",
            },
            "01": {
                "title": "Ars Amatoria",
                "code": "phi004",
                "meta": "book-verse",
            },
            "02": {
                "title": "Remedia Amoris",
                "code": "phi005",
                "meta": "verse",
            },
            "06": {"title": "Fasti", "code": "phi007", "meta": "book-verse"},
            "11": {"title": "Ibis", "code": "phi010", "meta": "verse"},
            "10": {"title": "Halieutica", "code": "phi014", "meta": "verse"},
        },
    },
    "J": {
        "author": "A. Persius Flaccus",
        "code": ["phi0969"],
        "phi0969": {
            "00": {"title": "Saturae", "code": "phi001", "meta": "poem-verse"}
        },
    },
    "K": {
        "author": "Petronius",
        "code": ["phi0972"],
        "phi0972": {
            "00": {"title": "Satyrica", "code": "phi001", "meta": "section"}
        },
    },
    "L": {
        "author": "T. Maccius Plautus",
        "code": ["phi0119"],
        "phi0119": {
            "01": {"title": "Amphitruo", "code": "phi001", "meta": "verse"},
            "02": {"title": "Asinaria", "code": "phi002", "meta": "verse"},
            "03": {"title": "Aulularia", "code": "phi003", "meta": "verse"},
            "04": {"title": "Bacchides", "code": "phi004", "meta": "verse"},
            "05": {"title": "Captivi", "code": "phi005", "meta": "verse"},
            "06": {"title": "Casina", "code": "phi006", "meta": "verse"},
            "08": {"title": "Curculio", "code": "phi008", "meta": "verse"},
            "09": {"title": "Epidicus", "code": "phi009", "meta": "verse"},
        },
    },
    "M": {
        "author": "C. Plinius Caecilius Secundus",
        "code": ["phi1318"],
        "phi1318": {
            "01": {
                "title": "Epistulae",
                "code": "phi001",
                "meta": "book-letter-section",
            },
            "00": {
                "title": "Panegyricus",
                "code": "phi002",
                "meta": "section",
            },
        },
    },
    "N": {
        "author": "Sex. Propertius",
        "code": ["phi0620"],
        "phi0620": {
            "00": {
                "title": "Elegiae",
                "code": "phi001",
                "meta": "book-poem-verse",
            }
        },
    },
    "O": {
        "author": "C. Sallustius Crispus",
        "code": ["phi0631"],
        "phi0631": {
            "01": {
                "title": "Catilinae Coniuratio",
                "meta": "section",
                "code": "phi001",
            },
            "02": {
                "title": "Bellum Iugurthinum",
                "meta": "section",
                "code": "phi002",
            },
            "03": {
                "title": "Historiae",
                "meta": "book-fragment-line",
                "code": "phi003",
            },
        },
    },
    "P": {
        "author": "L. Annaeus Seneca iunior",
        "code": ["phi1017", "stoa0255"],
        "phi1017": {
            "15": {
                "title": "Hercules Furens",
                "code": "phi001",
                "meta": "verse",
            },
            "16": {"title": "Troades", "code": "phi002", "meta": "verse"},
            "17": {"title": "Phoenissae", "code": "phi003", "meta": "verse"},
            "18": {"title": "Medea", "code": "phi004", "meta": "verse"},
            "19": {"title": "Phaedra", "code": "phi005", "meta": "verse"},
            "20": {"title": "Oedipus", "code": "phi006", "meta": "verse"},
            "21": {"title": "Agamemnon", "code": "phi007", "meta": "verse"},
            "22": {"title": "Thyestes", "code": "phi008", "meta": "verse"},
            "23": {
                "title": "Hercules Oetaeus",
                "code": "phi009",
                "meta": "verse",
            },
            "01": {
                "title": "Apocolocyntosis",
                "code": "phi011",
                "meta": "section",
            },
            "": {
                "title": "Dialogi",
                "code": "phi012",
                "meta": "book-chapter-section",
            },
            "06": {
                "title": "De Beneficiis",
                "code": "phi013",
                "meta": "book-chapter-section",
            },
            "08": {
                "title": "De Clementia",
                "code": "phi014",
                "meta": "book-chapter-section",
            },
            "03": {
                "title": "Epistulae Morales ad Lucilium",
                "code": "phi015",
                "meta": "letter-section",
            },
        },
        "stoa0255": {
            "02": {
                "title": "de consolatione ad Helviam",
                "code": "stoa006",
                "meta": "section",
            },
            "04": {
                "title": "de consolatione ad Marciam",
                "code": "stoa007",
                "meta": "section",
            },
            "05": {
                "title": "de consolatione ad Polybium",
                "code": "stoa008",
                "meta": "section",
            },
            "07": {
                "title": "De Brevitate Vitae",
                "code": "stoa004",
                "meta": "section",
            },
            "09": {
                "title": "De Constantia",
                "code": "stoa009",
                "meta": "section",
            },
            "10": {
                "title": "De Ira",
                "code": "stoa010",
                "meta": "book-chapter-section",
            },
            "11": {
                "title": "De Otio Sapientis",
                "code": "stoa011",
                "meta": "section",
            },
            "12": {
                "title": "De Providentia",
                "code": "stoa012",
                "meta": "section",
            },
            "13": {
                "title": "De Tranquillitate Animi",
                "code": "stoa013",
                "meta": "section",
            },
            "14": {
                "title": "De Vita Beata",
                "code": "stoa014",
                "meta": "section",
            },
        },
    },
    "Q": {
        "author": "Cornelius Tacitus",
        "code": ["phi1351"],
        "phi1351": {
            "02": {
                "title": "De Vita Iulii Agricolae",
                "meta": "section",
                "code": "phi001",
            },
            "01": {
                "title": "De Origine et Situ Germanorum",
                "meta": "section",
                "code": "phi002",
            },
            "03": {
                "title": "Dialogus de Oratoribus",
                "meta": "section",
                "code": "phi003",
            },
            "05": {
                "title": "Historiae",
                "meta": "book-line",
                "code": "phi004",
            },
            "04": {
                "title": "Annales",
                "meta": "book-chapter-section",
                "code": "phi005",
            },
        },
    },
    "R": {
        "author": "Albius Tibullus",
        "code": ["phi0660"],
        "phi0660": {
            "00": {
                "title": "Elegiae",
                "meta": "book-poem-verse",
                "code": "phi001",
            }
        },
    },
    "S": {
        "author": "P. Vergilius Maro",
        "code": ["phi0690"],
        "phi0690": {
            "01": {"title": "Eclogae", "meta": "poem-verse", "code": "phi001"},
            "02": {
                "title": "Georgica",
                "meta": "book-verse",
                "code": "phi002",
            },
            "03": {"title": "Aeneis", "meta": "book-verse", "code": "phi003"},
        },
    },
}

FILE_TAB = {
    "Caesar": {"BellumCivile": ["A03"], "BellumGallicum": ["A01"]},
    "Cato": {
        "CatoAgricultura": ["B01"],
        "CatoFragmentaReliqua": ["B04"],
        "CatoOrationumFragmenta": ["B03"],
        "CatoOrigines": ["B02"],
    },
    "Catullus": {"Catullus": ["C00"]},
    "Cicero": {
        "ContraContionemQMetelliFragmenta": ["D69"],
        "CumALudisContionemAvocavitFragmenta": ["D68"],
        "CumQuaestorLilybaeoDecederetFragmenta": ["D59"],
        "DeAmicitia": ["D76"],
        "DeDomoSua": ["D30"],
        "DeHaruspicumResponso": ["D31"],
        "DeImperioCnPompei": ["D14"],
        "DeLegeAgraria": ["D16"],
        "DeOfficiis": ["D77"],
        "DeProvinciisConsularibus": ["D35"],
        "DeRegeAlexandrinoFragmenta": ["D72"],
        "DeSenectute": ["D78"],
        "DivinatioinQCaecilium": ["D05"],
        "InCatilinam": ["D20", "D21", "D22", "D23"],
        "Incertarum0rationumFragmenta": ["D75"],
        "InLCalpurniumPisonem": ["D37"],
        "InPClodiumEtCCurionemFragmenta": ["D70"],
        "InSenatuInTogaCandidaContraCAntoniumEtLCatilinamCompetitoresFragmenta": [
            "D67"
        ],
        "InterrogatioDeAereAlienoMilonisFragmenta": ["D71"],
        "InVatinium": ["D33"],
        "InVerremActioPrima": ["D06"],
        "InVerremActioSecunda": ["D07", "D08", "D09", "D10", "D11"],
        "PhilippicaOratio": [
            "D44",
            "D45",
            "D46",
            "D47",
            "D48",
            "D49",
            "D50",
            "D51",
            "D52",
            "D53",
            "D54",
            "D55",
            "D56",
            "D57",
        ],
        "PostReditumAdQuirites": ["D29"],
        "PostReditumInSenatu": ["D28"],
        "PrLigario": ["D42"],
        "ProACaecina": ["D13"],
        "ProACluentio": ["D15"],
        "ProArchia": ["D26"],
        "ProBalbo": ["D36"],
        "ProCaelio": ["D34"],
        "ProCFundanioFragmenta": ["D63"],
        "ProCManilioFragmenta": ["D62"],
        "ProCornelioFragmenta": ["D65"],
        "ProCRabirioPerduellionisReo": ["D19"],
        "ProFlacco": ["D27"],
        "ProLVarenoFragmenta": ["D60"],
        "ProMarcello": ["D41"],
        "ProMFonteio": ["D12"],
        "ProMilone": ["D40"],
        "ProMScauroFragmenta": ["D58"],
        "ProMTullio": ["D04"],
        "ProMurena": ["D24"],
        "ProNegotiatoribusAchaeisFragmenta": ["D74"],
        "ProOppioFragmenta": ["D61"],
        "ProPlancio": ["D38"],
        "ProPQuinctio": ["D01"],
        "ProPVatinioFragmenta": ["D73"],
        "ProQGallioFragmenta": ["D64"],
        "ProQRoscioComoedo": ["D03"],
        "ProRabirioPostumo": ["D39"],
        "ProRegeDeiotario": ["D43"],
        "ProSestio": ["D32"],
        "ProSRoscioAmerino": ["D02"],
        "ProSulla": ["D25"],
    },
    "Curtius": {"CurtiusHistoriaeAlexandriMagni": ["E00"]},
    "Hirtius": {"Bellum8Gallicum": ["A02"]},
    "Horatius": {
        "ArsPoetica": ["F06"],
        "CarmenSaeculare": ["F02"],
        "Carmina": ["F01"],
        "Epodi": ["F03"],
        "HorEpistulae": ["F05"],
        "HorSermones": ["F04"],
    },
    "Juvenalis": {"JuvSaturae": ["G00"]},
    "Lucretius": {"DeRerumNatura": ["H00"]},
    "Ovidius": {
        "Amores": ["I04"],
        "ArsAmatoria": ["I01"],
        "Fasti": ["I06"],
        "Halieutica": ["I10"],
        "HeroidumEpistulae": ["I05"],
        "InIbin": ["I11"],
        "MedicaminaFacieiFemineae": ["I03"],
        "RemediaAmoris": ["I02"],
    },
    "Persius": {"PersiusSatirae": ["J00"]},
    "Petronius": {"PetroniusSatiricon": ["K00"]},
    "Plautus": {
        "Amphitruo": ["L01"],
        "Asinaria": ["L02"],
        "Aulularia": ["L03"],
        "Bacchides": ["L04"],
        "Captiui": ["L05"],
        "Casina": ["L06"],
        "Curculio": ["L08"],
        "Epidicus": ["L09"],
    },
    "Plinius": {"PliniusEpistulae": ["M01"], "PliniusPanegyricus": ["M00"]},
    "Propertius": {"PropertiusElegiae": ["N00"]},
    "PseudoCaesar1": {"BellumAfricanum": ["A05"]},
    "PseudoCaesar2": {"BellumAlexandrinum": ["A04"]},
    "PseudoCaesar3": {"BellumHispanum": ["A06"]},
    "Sallustius": {
        "Catilina": ["O01"],
        "FragmentaHistoriarum": ["O03"],
        "Jugurtha": ["O02"],
    },
    "Seneca": {
        "AdHelviamMatremDeConsolatione": ["P02"],
        "AdLuciliumEpistulaeMorales": ["P03"],
        "AdMarciamDeConsolatione": ["P04"],
        "AdPolybiumDeConsolatione": ["P05"],
        "Agamemnon": ["P21"],
        "Apocolocyntosis": ["P01"],
        "DeBeneficiis": ["P06"],
        "DeBrevitateVitae": ["P07"],
        "DeClementia": ["P08"],
        "DeConstantia": ["P09"],
        "DeIra": ["P10"],
        "DeOtio": ["P11"],
        "DeProvidentia": ["P12"],
        "DeTranquillitateAnimi": ["P13"],
        "DeVitaBeata": ["P14"],
        "HerculesFurens": ["P15"],
        "HerculesOetaeus": ["P23"],
        "Medea": ["P18"],
        "Oedipus": ["P20"],
        "Phaedra": ["P19"],
        "Phoenissae": ["P17"],
        "Thyestes": ["P22"],
        "Troades": ["P16"],
    },
    "Tacitus": {
        "TacAgricola": ["Q02"],
        "TacAnnales": ["Q04"],
        "TacDialogusDeOratoribus": ["Q03"],
        "TacGermania": ["Q01"],
        "TacHistoriae": ["Q05"],
    },
    "Tibullus": {"TibElegiae": ["R00"]},
    "Vergilius": {"Aeneis": ["S03"], "Eclogae": ["S01"], "Georgica": ["S02"]},
}

digraphs = ({"ADMODVM": {"60": ["AD MODVM"]}},)
{"ALIQVAMDIV": {"60": ["ALIQVAM DIV"]}},
{"ALIQVAMMVLTVS": {"21": ["ALIQVAM MVLTVS"]}},
{"ALIVSMODI": {"60": ["ALIVS MODI"]}},
{"ALLICEFACIO": {"55": ["ALLICE FACIO"]}},
{"ALTERVTER": {"48": ["ALTER VTER"]}},
{"ANTEACTVS": {"21": ["ANTE ACTVS"]}},
{"ANTEAGO": {"5C": ["ANTE AGO"]}},
{"ANTEPONO": {"5C": ["ANTE PONO"]}},
{"ANTEQVAM": {"82": ["ANTE QVAM"]}},
{"AREFACIO": {"55": ["AREFACIO"]}},
{"ATTAMEN": {"60": ["AT TAMEN"]}},
{"BENEDICO": {"53": ["BENE DICO"]}},
{"BENEDICTVM": {"12": ["BENE DICTVM"]}},
{"BENEFACIO": {"55": ["BENE FACIO"]}},
{"BENEFACTVM": {"12": ["BENE FACTVM"]}},
{"BENEVOLENS": {"25": ["BENE VOLENS"]}},
{"CIRCVMAMPLECTOR": {"5L": ["CIRCVM AMPLECTOR"]}},
{"CIRCVMDO": {"51": ["CIRCVM DO"]}},
{"CIRCVMDVCO": {"53": ["CIRCVM DVCO"]}},
{"CIRCVMFVNDO": {"53": ["CIRCVM FVNDO"]}},
{"CIRCVMMVGIO": {"54": ["CIRCVM MVGIO"]}},
{"CIRCVMSAEPIO": {"54": ["CIRCVM SAEPIO"]}},
{"CIRCVMSPECTO": {"51": ["CIRCVM SPECTO"]}},
{"CIRCVMSTO": {"51": ["CIRCVM STO"]}},
{"CIRCVMVEHOR": {"5L": ["CIRCVM VEHOR"]}},
{"CONDOCEFACIO": {"55": ["CONDOCEFACIO"]}},
{"CONTRADICO": {"53": ["CONTRA DICO"]}},
{"CVIVSMODI": {"66": ["CVIVS MODI"]}},
{"CVIVSQVEMODI": {"60": ["CVIVSQVE MODI"]}},
{"CVMPRIMIS": {"60": ["CVM PRIMIS"]}},
{"DESVBITO": {"60": ["DE SVBITO"]}},
{"DVMMODO": {"82": ["DVM MODO"]}},
{"DVMTAXAT": {"60": ["DVM TAXAT"]}},
{"EIVSMODI": {"60": ["EIVS MODI"]}},
{"ENIMVERO": {"60": ["ENIM VERO"]}},
{"ETIAMNVNC": {"60": ["ETIAM NVNC"]}},
{"ETIAMSI": {"60": ["ETIAM SI"]}},
{"ETIAMTVM": {"82": ["ETIAM TVM"]}},
{"ETIAMTVNC": {"60": ["ETIAM TVNC"]}},
{"FERVEFACIO": {"55": ["FERVEFACIO"]}},
{"HACTENVS": {"60": ["HAC TENVS"]}},
{"HAVDDVM": {"68": ["HAVD DVM"]}},
{"HAVDQVAQVAM": {"68": ["HAVD QVAQVAM"]}},
{"HVIVSMODI": {"60": ["HVIVS MODI"]}},
{"IAMDIV": {"60": ["IAM DIV"]}},
{"IAMDVDVM": {"60": ["IAM DVDVM"]}},
{"IAMIAM": {"60": ["IAM IAM"]}},
{"IAMPRIDEM": {"60": ["IAM PRIDEM"]}},
{"ILLIVSMODI": {"60": ["ILLIVS MODI"]}},
{"IMPRAESENTIARVM": {"60": ["IN PRAESENTIARVM"]}},
{"IMPRIMIS": {"60": ["IN PRIMIS"]}},
{"INCASSVM": {"60": ["IN CASSVM"]}},
{"INTERLABOR": {"5L": ["INTER LABOR"]}},
{"INTERLEGO": {"5C": ["INTER LEGO"]}},
{"INTERLVNIVM": {"12": ["INTER LVNIVM"]}},
{"INTERNOSCO": {"53": ["INTER NOSCO"]}},
{"INTRODVCO": {"53": ["INTRO DVCO"]}},
{"INTROMITTO": {"53": ["INTRO MITTO"]}},
{"INTROTRVDO": {"53": ["INTRO TRVDO"]}},
{"INVICEM": {"60": ["IN VICEM"]}},
{"ISTIVSMODI": {"60": ["ISTIVS MODI"]}},
{"IVRISCONSVLTVS": {"12": ["IVRIS CONSVLTVS"]}},
{"IVRISDICTIO": {"13": ["IVRIS DICTIO"]}},
{"IVSIVRANDVM": {"13": ["IVS IVRANDVM"]}},
{"MAGNOPERE": {"60": ["MAGNOPERE"]}},
{"MALEDICENS": {"25": ["MALE DICENS"]}},
{"MALEDICO": {"53": ["MALE DICO"]}},
{"MALEDICTVM": {"12": ["MALE DICTVM"]}},
{"MALEFACIO": {"55": ["MALE FACIO"]}},
{"MALEFACTVM": {"12": ["MALE FACTVM"]}},
{"MANVMITTO": {"53": ["MANV MITTO"]}},
{"MEDIVSFIDIVS": {"90": ["MEDIVS FIDIVS"]}},
{"NECDVM": {"68": ["NEC DVM"]}},
{"NECNE": {"69": ["NEC NE"]}},
{"NECNON": {"81": ["NEC NON"]}},
{"NECOPINANS": {"25": ["NEC OPINANS"]}},
{"NECOPINATO": {"60": ["NEC OPINATO"]}},
{"NEQVEDVM": {"81": ["NEQVE DVM"]}},
{"NEQVIS": {"48": ["NE QVIS"]}},
{"NEQVO": {"60": ["NE QVO"]}},
{"NESCIOQVA": {"60": ["NESCIO QVA"]}},
{"NESCIOQVANTVS": {"48": ["NESCIO QVANTVS"]}},
{"NESCIOQVIS": {"48": ["NESCIO QVIS"]}},
{"NESCIOQVOT": {"48": ["NESCIO QVOT"]}},
{"NEVTIQVAM": {"68": ["NE VTIQVAM"]}},
{"NIHILOMINVS": {"68": ["NIHILO MINVS"]}},
{"NIHILOSETIVS": {"68": ["NIHILO SETIVS", "NIHILO SECIVS"]}},
{"NIMIOPERE": {"60": ["NIMIOPERE"]}},
{"NONNIHIL": {"48": ["NON NIHIL"]}},
{"NONNVLLVS": {"48": ["NON NVLLVS"]}},
{"NONNVMQVAM": {"60": ["NON NVMQVAM"]}},
{"NOVVMCOMVM": {"12": ["NOVVM COMVM"]}},
{"NVDIVSTERTIVS": {"60": ["NVDIVS TERTIVS"]}},
{"NVMNE": {"67": ["NVM NE"]}},
{"NVMQVANDO": {"67": ["NVM QVANDO"]}},
{"NVMQVIS": {"47": ["NVM QVIS"]}},
{"OBSTVPEFACIO": {"55": ["OB STVPEFACIO"]}},
{"OBVIAM": {"60": ["OB VIAM"]}},
{"PERBREVIS": {"24": ["PER BREVIS"]}},
{"PESSVMDO": {"51": ["PESSVM DO"]}},
{"PLEBISCITVM": {"12": ["PLEBI SCITVM"]}},
{"PONEVERSVS": {"70": ["PONE VERSVS"]}},
{"POSTEAQVAM": {"82": ["POSTEA QVAM"]}},
{"POSTQVAM": {"82": ["POST QVAM"]}},
{"PRAETERQVAM": {"60": ["PRAETER QVAM"]}},
{"PRAEVENIO": {"54": ["PRAE VENIO"]}},
{"PRIMIPILVS": {"12": ["PRIMI PILVS"]}},
{"PRIVSQVAM": {"82": ["PRIVS QVAM"]}},
{"PROCONSVL": {"13": ["PROCONSVL"]}},
{"PROPEDIEM": {"60": ["PROPE DIEM"]}},
{"PROPRAETOR": {"13": ["PRO PRAETOR"]}},
{"PROPEMODVM": {"60": ["PROPE MODVM"]}},
{"PROQVAESTOR": {"13": ["PRO QVAESTOR"]}},
{"QVALIBET": {"60": ["QVA LIBET"]}},
{"QVAMDIV": {"66": ["QVAMD IV"]}},
{"QVAMLIBET": {"60": ["QVAM LIBET"]}},
{"QVAMOBREM": {"66": ["QVAM OB REM"]}},
{"QVAMPRIMVM": {"60": ["QVAM PRIMVM"]}},
{"QVANDANTENVS": {"60": ["QVANDAM TENVS"]}},
{"QVANDOQVIDEM": {"82": ["QVANDO QVIDEM"]}},
{"QVANTOPERE": {"66": ["QVANTO OPERE"]}},
{"QVANTVLVSLIBET": {"48": ["QVANTVLVS LIBET"]}},
{"QVARE": {"66": ["QVA RE"]}},
{"QVEMADMODVM": {"66": ["QVEM AD MODVM"]}},
{"QVIDNI": {"69": ["QVID NI"]}},
{"QVILIBET": {"48": ["QVI LIBET"]}},
{"QVINIDENI": {"33": ["QVINI DENI"]}},
{"QVOCVMQVE": {"66": ["QVO CVMQVE"]}},
{"QVOLIBET": {"60": ["QVO LIBET"]}},
{"QVOMINVS": {"82": ["QVO MINVS"]}},
{"QVOMODO": {"66": ["QVO MODO"]}},
{"QVOQVOMODO": {"66": ["QVOQVO MODO"]}},
{"QVOQVOVERSVS": {"60": ["QVOQVO VERSVS"]}},
{"QVOTIENSCVMQVE": {"66": ["QVOTIENS CVMQVE"]}},
{"QVOTVSQVISQVE": {"47": ["QVOTVS QVISQVE"]}},
{"QVOVSQVE": {"67": ["QVO VSQVE"]}},
{"RESPVBLICA": {"15": ["RES PVBLICA"]}},
{"RETROFLECTO": {"53": ["RETRO FLECTO"]}},
{"REVERA": {"60": ["RE VERA"]}},
{"SAEPENVMERO": {"60": ["SAEPE NVMERO"]}},
{"SATISACCIPIO": {"55": ["SATIS ACCIPIO"]}},
{"SATISDO": {"51": ["SATIS DO"]}},
{"SATISFACIO": {"55": ["SATIS FACIO"]}},
{"SATISFACTIO": {"13": ["SATIS FACTIO"]}},
{"SATISFIO": {"56": ["SATIS FIO"]}},
{"SEPTENTRIO": {"13": ["SEPTENTRIO"]}},
{"SICVT": {"60": ["SIC VT"]}},
{"SIMVLAC": {"60": ["SIMVL AC"]}},
{"SIMVLATQVE": {"60": ["SIMVL ATQVE"]}},
{"SIQVIDEM": {"82": ["SI QVIDEM"]}},
{"SVPERFVNDO": {"53": ["SVPER FVNDO"]}},
{"SVPERSVM": {"56": ["SVPER SVM"]}},
{"SVPERTEGO": {"53": ["SVPER TEGO"]}},
{"SVRSVMVORSVM": {"60": ["SVRSVM VORSVM"]}},
{"TAMDIV": {"60": ["TAM DIV"]}},
{"TAMENETSI": {"82": ["TAMEN ETSI"]}},
{"TAMETSI": {"82": ["TAM ETSI"]}},
{"TANTOPERE": {"60": ["TANTO OPERE"]}},
{"TANTVMMODO": {"60": ["TANTVM MODO"]}},
{"TRIVMVIR": {"12": ["TRIVM VIR"]}},
{"VBIVBI": {"66": ["VBI VBI"]}},
{"VBICVMQVE": {"66": ["VBI CVMQVE"]}},
{"VELVTSI": {"82": ["VELVT SI"]}},
{"VERISIMILIS": {"24": ["VERI SIMILIS"]}},
{"VERISIMILITVDO": {"13": ["VERI SIMILITVDO"]}},
{"VERVMTAMEN": {"60": ["VERVM TAMEN"]}},
{"VNVSQVISQVE": {"48": ["VNVS QVISQVE"]}},
{"VNVSQVISQVIS": {"48": ["VNVS QVISQVIS"]}},
{"VSQVEQVAQVE": {"60": ["VSQVE QVAQVE"]}},
{"VSVCAPIO": {"13": ["VSV CAPIO"]}},
{"VSVFACIO": {"55": ["VSV FACIO"]}},
{"VSVSFRVCTVS": {"14": ["VSVS FRVCTVS"]}},
{"VSVVENIO": {"54": ["VSV VENIO"]}},
{"VTPOTE": {"60": ["VT POTE"]}},
{"VTRIMQVESECVS": {"60": ["VTRIMQVE SECVS"]}},
{"VTRVMNE": {"67": ["VTRVM NE"]}},

simplexes = {
    "AD MODVM": ("ADMODVM", "60"),
    "ALIQVAM DIV": ("ALIQVAMDIV", "60"),
    "ALIQVAM MVLTVS": ("ALIQVAMMVLTVS", "21"),
    "ALIVS MODI": ("ALIVSMODI", "60"),
    "ALLICE FACIO": ("ALLICEFACIO", "55"),
    "ALTER VTER": ("ALTERVTER", "48"),
    "ANTE ACTVS": ("ANTEACTVS", "21"),
    "ANTE AGO": ("ANTEAGO", "5C"),
    "ANTE PONO": ("ANTEPONO", "5C"),
    "ANTE QVAM": ("ANTEQVAM", "82"),
    "AREFACIO": ("AREFACIO", "55"),
    "AT TAMEN": ("ATTAMEN", "60"),
    "BENE DICO": ("BENEDICO", "53"),
    "BENE DICTVM": ("BENEDICTVM", "12"),
    "BENE FACIO": ("BENEFACIO", "55"),
    "BENE FACTVM": ("BENEFACTVM", "12"),
    "BENE VOLENS": ("BENEVOLENS", "25"),
    "CIRCVM AMPLECTOR": ("CIRCVMAMPLECTOR", "5L"),
    "CIRCVM DO": ("CIRCVMDO", "51"),
    "CIRCVM DVCO": ("CIRCVMDVCO", "53"),
    "CIRCVM FVNDO": ("CIRCVMFVNDO", "53"),
    "CIRCVM MVGIO": ("CIRCVMMVGIO", "54"),
    "CIRCVM SAEPIO": ("CIRCVMSAEPIO", "54"),
    "CIRCVM SPECTO": ("CIRCVMSPECTO", "51"),
    "CIRCVM STO": ("CIRCVMSTO", "51"),
    "CIRCVM VEHOR": ("CIRCVMVEHOR", "5L"),
    "CONDOCEFACIO": ("CONDOCEFACIO", "55"),
    "CONTRA DICO": ("CONTRADICO", "53"),
    "CVIVS MODI": ("CVIVSMODI", "66"),
    "CVIVSQVE MODI": ("CVIVSQVEMODI", "60"),
    "CVM PRIMIS": ("CVMPRIMIS", "60"),
    "DE SVBITO": ("DESVBITO", "60"),
    "DVM MODO": ("DVMMODO", "82"),
    "DVM TAXAT": ("DVMTAXAT", "60"),
    "EIVS MODI": ("EIVSMODI", "60"),
    "ENIM VERO": ("ENIMVERO", "60"),
    "ETIAM NVNC": ("ETIAMNVNC", "60"),
    "ETIAM SI": ("ETIAMSI", "60"),
    "ETIAM TVM": ("ETIAMTVM", "82"),
    "ETIAM TVNC": ("ETIAMTVNC", "60"),
    "FERVEFACIO": ("FERVEFACIO", "55"),
    "HAC TENVS": ("HACTENVS", "60"),
    "HAVD DVM": ("HAVDDVM", "68"),
    "HAVD QVAQVAM": ("HAVDQVAQVAM", "68"),
    "HVIVS MODI": ("HVIVSMODI", "60"),
    "IAM DIV": ("IAMDIV", "60"),
    "IAM DVDVM": ("IAMDVDVM", "60"),
    "IAM IAM": ("IAMIAM", "60"),
    "IAM PRIDEM": ("IAMPRIDEM", "60"),
    "ILLIVS MODI": ("ILLIVSMODI", "60"),
    "IN PRAESENTIARVM": ("IMPRAESENTIARVM", "60"),
    "IN PRIMIS": ("IMPRIMIS", "60"),
    "IN CASSVM": ("INCASSVM", "60"),
    "INTER LABOR": ("INTERLABOR", "5L"),
    "INTER LEGO": ("INTERLEGO", "5C"),
    "INTER LVNIVM": ("INTERLVNIVM", "12"),
    "INTER NOSCO": ("INTERNOSCO", "53"),
    "INTRO DVCO": ("INTRODVCO", "53"),
    "INTRO MITTO": ("INTROMITTO", "53"),
    "INTRO TRVDO": ("INTROTRVDO", "53"),
    "IN VICEM": ("INVICEM", "60"),
    "ISTIVS MODI": ("ISTIVSMODI", "60"),
    "IVRIS CONSVLTVS": ("IVRISCONSVLTVS", "12"),
    "IVRIS DICTIO": ("IVRISDICTIO", "13"),
    "IVS IVRANDVM": ("IVSIVRANDVM", "13"),
    "MAGNOPERE": ("MAGNOPERE", "60"),
    "MALE DICENS": ("MALEDICENS", "25"),
    "MALE DICO": ("MALEDICO", "53"),
    "MALE DICTVM": ("MALEDICTVM", "12"),
    "MALE FACIO": ("MALEFACIO", "55"),
    "MALE FACTVM": ("MALEFACTVM", "12"),
    "MANV MITTO": ("MANVMITTO", "53"),
    "MEDIVS FIDIVS": ("MEDIVSFIDIVS", "90"),
    "NEC DVM": ("NECDVM", "68"),
    "NEC NE": ("NECNE", "69"),
    "NEC NON": ("NECNON", "81"),
    "NEC OPINANS": ("NECOPINANS", "25"),
    "NEC OPINATO": ("NECOPINATO", "60"),
    "NEQVE DVM": ("NEQVEDVM", "81"),
    "NE QVIS": ("NEQVIS", "48"),
    "NE QVO": ("NEQVO", "60"),
    "NESCIO QVA": ("NESCIOQVA", "60"),
    "NESCIO QVANTVS": ("NESCIOQVANTVS", "48"),
    "NESCIO QVIS": ("NESCIOQVIS", "48"),
    "NESCIO QVOT": ("NESCIOQVOT", "48"),
    "NE VTIQVAM": ("NEVTIQVAM", "68"),
    "NIHILO MINVS": ("NIHILOMINVS", "68"),
    "NIHILO SETIVS": ("NIHILOSETIVS", "68"),
    "NIHILO SECIVS": ("NIHILOSETIVS", "68"),
    "NIMIOPERE": ("NIMIOPERE", "60"),
    "NON NIHIL": ("NONNIHIL", "48"),
    "NON NVLLVS": ("NONNVLLVS", "48"),
    "NON NVMQVAM": ("NONNVMQVAM", "60"),
    "NOVVM COMVM": ("NOVVMCOMVM", "12"),
    "NVDIVS TERTIVS": ("NVDIVSTERTIVS", "60"),
    "NVM NE": ("NVMNE", "67"),
    "NVM QVANDO": ("NVMQVANDO", "67"),
    "NVM QVIS": ("NVMQVIS", "47"),
    "OB STVPEFACIO": ("OBSTVPEFACIO", "55"),
    "OB VIAM": ("OBVIAM", "60"),
    "PER BREVIS": ("PERBREVIS", "24"),
    "PESSVM DO": ("PESSVMDO", "51"),
    "PLEBI SCITVM": ("PLEBISCITVM", "12"),
    "PONE VERSVS": ("PONEVERSVS", "70"),
    "POSTEA QVAM": ("POSTEAQVAM", "82"),
    "POST QVAM": ("POSTQVAM", "82"),
    "PRAETER QVAM": ("PRAETERQVAM", "60"),
    "PRAE VENIO": ("PRAEVENIO", "54"),
    "PRIMI PILVS": ("PRIMIPILVS", "12"),
    "PRIVS QVAM": ("PRIVSQVAM", "82"),
    "PROCONSVL": ("PROCONSVL", "13"),
    "PROPE DIEM": ("PROPEDIEM", "60"),
    "PRO PRAETOR": ("PROPRAETOR", "13"),
    "PROPE MODVM": ("PROPEMODVM", "60"),
    "PRO QVAESTOR": ("PROQVAESTOR", "13"),
    "QVA LIBET": ("QVALIBET", "60"),
    "QVAMD IV": ("QVAMDIV", "66"),
    "QVAM LIBET": ("QVAMLIBET", "60"),
    "QVAM OB REM": ("QVAMOBREM", "66"),
    "QVAM PRIMVM": ("QVAMPRIMVM", "60"),
    "QVANDAM TENVS": ("QVANDANTENVS", "60"),
    "QVANDO QVIDEM": ("QVANDOQVIDEM", "82"),
    "QVANTO OPERE": ("QVANTOPERE", "66"),
    "QVANTVLVS LIBET": ("QVANTVLVSLIBET", "48"),
    "QVA RE": ("QVARE", "66"),
    "QVEM AD MODVM": ("QVEMADMODVM", "66"),
    "QVID NI": ("QVIDNI", "69"),
    "QVI LIBET": ("QVILIBET", "48"),
    "QVINI DENI": ("QVINIDENI", "33"),
    "QVO CVMQVE": ("QVOCVMQVE", "66"),
    "QVO LIBET": ("QVOLIBET", "60"),
    "QVO MINVS": ("QVOMINVS", "82"),
    "QVO MODO": ("QVOMODO", "66"),
    "QVOQVO MODO": ("QVOQVOMODO", "66"),
    "QVOQVO VERSVS": ("QVOQVOVERSVS", "60"),
    "QVOTIENS CVMQVE": ("QVOTIENSCVMQVE", "66"),
    "QVOTVS QVISQVE": ("QVOTVSQVISQVE", "47"),
    "QVO VSQVE": ("QVOVSQVE", "67"),
    "RES PVBLICA": ("RESPVBLICA", "15"),
    "RETRO FLECTO": ("RETROFLECTO", "53"),
    "RE VERA": ("REVERA", "60"),
    "SAEPE NVMERO": ("SAEPENVMERO", "60"),
    "SATIS ACCIPIO": ("SATISACCIPIO", "55"),
    "SATIS DO": ("SATISDO", "51"),
    "SATIS FACIO": ("SATISFACIO", "55"),
    "SATIS FACTIO": ("SATISFACTIO", "13"),
    "SATIS FIO": ("SATISFIO", "56"),
    "SEPTENTRIO": ("SEPTENTRIO", "13"),
    "SIC VT": ("SICVT", "60"),
    "SIMVL AC": ("SIMVLAC", "60"),
    "SIMVL ATQVE": ("SIMVLATQVE", "60"),
    "SI QVIDEM": ("SIQVIDEM", "82"),
    "SVPER FVNDO": ("SVPERFVNDO", "53"),
    "SVPER SVM": ("SVPERSVM", "56"),
    "SVPER TEGO": ("SVPERTEGO", "53"),
    "SVRSVM VORSVM": ("SVRSVMVORSVM", "60"),
    "TAM DIV": ("TAMDIV", "60"),
    "TAMEN ETSI": ("TAMENETSI", "82"),
    "TAM ETSI": ("TAMETSI", "82"),
    "TANTO OPERE": ("TANTOPERE", "60"),
    "TANTVM MODO": ("TANTVMMODO", "60"),
    "TRIVM VIR": ("TRIVMVIR", "12"),
    "VBI VBI": ("VBIVBI", "66"),
    "VBI CVMQVE": ("VBICVMQVE", "66"),
    "VELVT SI": ("VELVTSI", "82"),
    "VERI SIMILIS": ("VERISIMILIS", "24"),
    "VERI SIMILITVDO": ("VERISIMILITVDO", "13"),
    "VERVM TAMEN": ("VERVMTAMEN", "60"),
    "VNVS QVISQVE": ("VNVSQVISQVE", "48"),
    "VNVS QVISQVIS": ("VNVSQVISQVIS", "48"),
    "VSQVE QVAQVE": ("VSQVEQVAQVE", "60"),
    "VSV CAPIO": ("VSVCAPIO", "13"),
    "VSV FACIO": ("VSVFACIO", "55"),
    "VSVS FRVCTVS": ("VSVSFRVCTVS", "14"),
    "VSV VENIO": ("VSVVENIO", "54"),
    "VT POTE": ("VTPOTE", "60"),
    "VTRIMQVE SECVS": ("VTRIMQVESECVS", "60"),
    "VTRVM NE": ("VTRVMNE", "67"),
}
